theory Cyc_Check
imports "../Impl/Data/Simple_Impl"
  "../../Modelchecker/CAVA_Automata/Digraph_Impl"
  "../Misc/Impl_Rev_Array_Stack"
begin


(* Simple cyclicity check *)

(* Define a state, based on the DFS-state. 
  In our case, we just add a break-flag.
*)
record 'v cycc_state = "'v state" +
  break :: bool

(* Some utility lemmas for the simplifier, to handle idiosyncrasies of
  the record package. *)
lemma break_more_cong: "state.more s = state.more s' \<Longrightarrow> break s = break s'"
  by (cases s, cases s', simp)

lemma [simp]: "s\<lparr> state.more := \<lparr> break = foo \<rparr> \<rparr> = s \<lparr> break := foo \<rparr>"
  by (cases s) simp

(* Define the parameterization. We start at a default parameterization, where
  all operations default to skip, and just add the operations we are 
  interested in: Initially, the break flag is false, it is set if we 
  encounter a back-edge, and once set, the algorithm shall terminate immediately. *)
definition cycc_params :: "('v,unit cycc_state_ext) parameterization"
where "cycc_params \<equiv> dflt_parametrization state.more 
  (RETURN \<lparr> break = False \<rparr>) \<lparr>
  on_back_edge := \<lambda>_ _ _. RETURN \<lparr> break = True \<rparr>,
  is_break := break \<rparr>"
lemmas cycc_params_simp[simp] = 
  gen_parameterization.simps[mk_record_simp, OF cycc_params_def[simplified]]

interpretation cycc: param_DFS_defs where param=cycc_params for G .

definition "cyc_checker G \<equiv> do {
  ASSERT (fb_graph G);
  s \<leftarrow> cycc.it_dfs TYPE('a) G;
  RETURN (break s)
}"

(* Next, we define a locale for the cyclicity checker's
  precondition and invariant, by specializing the *)

locale cycc = param_DFS G cycc_params for G :: "('v, 'more) graph_rec_scheme"
begin
  (* We can easily show that our parametrization does not fail, thus we also
    get the param_DFS-locale, which gives us the correctness theorem for
    the DFS-scheme *)
  sublocale DFS G cycc_params
    apply unfold_locales
    apply (simp_all add: cycc_params_def)
    done

  thm it_dfs_correct  
end

lemma cyccI: 
  assumes "fb_graph G" 
  shows "cycc G"
proof -
  interpret fb_graph G by fact
  show ?thesis by unfold_locales
qed

(* Next, we specialize the DFS_invar locale to our parameterization.
  This locale contains all proven invariants. When proving new invariants,
  this locale is available as assumption, thus allowing us to re-use already 
  proven invariants.
*)
locale cycc_invar = DFS_invar where param = cycc_params + cycc

(* The lemmas to establish invariants only provide the DFS_invar locale.
  This lemma is used to convert it into the cycc_invar locale.
*)
lemma cycc_invar_eq[simp]:
  shows "DFS_invar G cycc_params s \<longleftrightarrow> cycc_invar G s" 
proof
  assume "DFS_invar G cycc_params s"
  interpret DFS_invar G cycc_params s by fact
  show "cycc_invar G s" by unfold_locales
next
  assume "cycc_invar G s"
  then interpret cycc_invar G s .
  show "DFS_invar G cycc_params s" by unfold_locales
qed

context cycc_invar begin
  (* We fold the invariant in the introduction lemma *)
  (*lemmas establish_invarI = loc_establish_invarI[
    unfolded cycc_invar_eq,
    case_names init new_root finish cross_edge back_edge discover]*)

  lemma (in cycc) i_brk_eq_back: "is_invar (\<lambda>s. break s \<longleftrightarrow> back_edges s \<noteq> {})"
    apply (induct rule: establish_invarI)
    apply (simp_all add: cond_def cong: break_more_cong)
    apply (simp add: empty_state_def)
    done
  lemmas brk_eq_back = i_brk_eq_back[THEN make_invar_thm]

  lemma (in cycc) i_brk_eq_back_long_proof: "is_invar (\<lambda>s. break s \<longleftrightarrow> back_edges s \<noteq> {})"
  proof (induct rule: establish_invarI)
    case init thus ?case by (simp add: empty_state_def)
  next
    case (back_edge s s' u v)
    then interpret cycc_invar G s by simp
    from back_edge show ?case by simp
  qed (simp_all cong: break_more_cong)

  lemma cycc_correct_aux_short_proof:
    assumes NC: "\<not>cond s"
    shows "break s \<longleftrightarrow> \<not>acyclic (E \<inter> reachable \<times> UNIV)"
    using nc_edges_covered[OF NC] brk_eq_back cycle_iff_back_edges 
    by (auto dest: acyclic_subset[OF _ edges_ss_reachable_edges])

  lemma cycc_correct_aux: 
    assumes NC: "\<not>cond s"
    shows "break s \<longleftrightarrow> \<not>acyclic (E \<inter> reachable \<times> UNIV)"
  proof (cases "break s", simp_all)
    assume "break s"
    with brk_eq_back have "back_edges s \<noteq> {}" by simp
    with cycle_iff_back_edges have "\<not>acyclic (edges s)" by simp
    with acyclic_subset[OF _ edges_ss_reachable_edges] 
    show "\<not>acyclic (E \<inter> reachable \<times> UNIV)" by blast
  next
    assume A: "\<not>break s"
    from A brk_eq_back have "back_edges s = {}" by simp
    with cycle_iff_back_edges have "acyclic (edges s)" by simp
    also from A nc_edges_covered[OF NC] have "edges s = E \<inter> reachable \<times> UNIV" 
      by simp
    finally show "acyclic (E \<inter> reachable \<times> UNIV)" .
  qed
    
end

definition "cyc_checker_spec G \<equiv> do {
  ASSERT (fb_graph G);
  SPEC (\<lambda>r. r \<longleftrightarrow> \<not>acyclic (g_E G \<inter> ((g_E G)\<^sup>* `` g_V0 G) \<times> UNIV))}"

lemma cyc_checker_correct: "cyc_checker G \<le> cyc_checker_spec G"
  unfolding cyc_checker_def cyc_checker_spec_def
proof (refine_vcg le_ASSERTI order_trans[OF DFS.it_dfs_correct], clarsimp_all)
  assume "fb_graph G"
  then interpret fb_graph G .
  interpret cycc by unfold_locales
  show "DFS G cycc_params" by unfold_locales
next
  fix s
  assume "cycc_invar G s"
  then interpret cycc_invar G s .
  assume "\<not>cycc.cond TYPE('b) G s"
  thus "break s = (\<not> acyclic (g_E G \<inter> cycc.reachable TYPE('b) G \<times> UNIV))"
    by (rule cycc_correct_aux)
qed

subsection {* Implementation *}

(* Implementation with stack and sso_visited set *)
record 'v cycc_state_impl = "'v simple_state" +
  break :: bool

(* Definition of refinement relation: The break-flag is refined by identity.*)
definition "cycc_erel \<equiv> { 
  (\<lparr> cycc_state_impl.break = b \<rparr>, \<lparr> cycc_state.break = b\<rparr>) | b. True }"
abbreviation "cycc_rel \<equiv> \<langle>cycc_erel\<rangle>simple_state_rel"

(* Implementation of the parameters *)
definition cycc_params_impl 
  :: "('v,'v cycc_state_impl,unit cycc_state_impl_ext) gen_parameterization"
where "cycc_params_impl 
  \<equiv> dflt_parametrization simple_state.more (RETURN \<lparr> break = False \<rparr>) \<lparr>
  on_back_edge := \<lambda>u v s. RETURN \<lparr> break = True \<rparr>,
  is_break := break \<rparr>"
lemmas cycc_params_impl_simp[simp,DFS_code_unfold] = 
  gen_parameterization.simps[mk_record_simp, OF cycc_params_impl_def[simplified]]

(* Note: In this simple case, the reformulation of the extension state and 
  parameterization is just redundant, However, in general the refinement will 
  also affect the parameterization.  *)

lemma break_impl: "(si,s)\<in>cycc_rel 
  \<Longrightarrow> cycc_state_impl.break si = cycc_state.break s"
  by (cases si, cases s, simp add: simple_state_rel_def cycc_erel_def)

interpretation cycc_impl: simple_impl_defs G cycc_params_impl cycc_params 
  for G .

(* The above locale creates an iterative and a recursive implementation *)
term cycc_impl.tailrec_impl term cycc_impl.rec_impl

definition [DFS_code_unfold]: "cyc_checker_impl G \<equiv> do {
  ASSERT (fb_graph G);
  s \<leftarrow> cycc_impl.tailrec_impl TYPE('a) G;
  RETURN (break s)
}"

definition [DFS_code_unfold]: "cyc_checker_rec_impl G \<equiv> do {
  ASSERT (fb_graph G);
  s \<leftarrow> cycc_impl.rec_impl TYPE('a) G;
  RETURN (break s)
}"


context cycc begin
  (* Derive the implementation *)
  sublocale simple_impl G cycc_params cycc_params_impl cycc_erel 
    apply unfold_locales
    apply (intro fun_relI, clarsimp simp: simple_state_rel_def, parametricity) []
    apply (auto simp: cycc_erel_def break_impl simple_state_rel_def)
    done

  lemmas impl_refine = simple_tailrec_refine simple_rec_refine
  thm simple_refine
end

lemma cyc_checker_impl_refine: "cyc_checker_impl G \<le> \<Down>Id (cyc_checker G)"
  unfolding cyc_checker_impl_def cyc_checker_def
  apply (refine_vcg cycc.impl_refine)
  apply (simp_all add: break_impl cyccI)
  done
  
lemma cyc_checker_rec_impl_refine: 
  "cyc_checker_rec_impl G \<le> \<Down>Id (cyc_checker G)"
  unfolding cyc_checker_rec_impl_def cyc_checker_def
  apply (refine_vcg cycc.impl_refine)
  apply (simp_all add: break_impl cyccI)
  done

(* Autoref *)
record ('si,'nsi,'psi)cycc_state_impl' = "('si,'nsi)simple_state_impl" +
  break_impl :: bool

definition [to_relAPP]: "cycc_state_erel erel \<equiv> {
  (\<lparr>break_impl = bi, \<dots> =  mi\<rparr>,\<lparr>break = b, \<dots> = m\<rparr>) | bi mi b m.
    (bi,b)\<in>bool_rel \<and> (mi,m)\<in>erel}"

consts 
  i_cycc_state_ext :: "interface \<Rightarrow> interface"

lemmas [autoref_rel_intf] = REL_INTFI[of cycc_state_erel i_cycc_state_ext]


term cycc_state_impl_ext
lemma [autoref_rules]:
  fixes ns_rel vis_rel erel
  defines "R \<equiv> \<langle>ns_rel,vis_rel,\<langle>erel\<rangle>cycc_state_erel\<rangle>ss_impl_rel"
  shows 
    "(cycc_state_impl'_ext, cycc_state_impl_ext) \<in> bool_rel \<rightarrow> erel \<rightarrow> \<langle>erel\<rangle>cycc_state_erel"
    "(break_impl, cycc_state_impl.break) \<in> R \<rightarrow> bool_rel"
  unfolding cycc_state_erel_def ss_impl_rel_def R_def
  by auto

schematic_lemma cyc_checker_impl:
  defines "V \<equiv> Id :: ('v \<times> 'v::hashable) set"
  assumes [unfolded V_def,autoref_rules]:
    "(Gi, G) \<in> \<langle>Rm, V\<rangle>g_impl_rel_ext"
  notes [unfolded V_def,autoref_tyrel] = 
    TYRELI[where R="\<langle>V\<rangle>dflt_ahs_rel"]
    TYRELI[where R="\<langle>V \<times>\<^sub>r \<langle>V\<rangle>list_set_rel\<rangle>ras_rel"]
  shows "nres_of (?c::?'c dres) \<le>\<Down>?R (cyc_checker_impl G)"
  unfolding DFS_code_unfold
  using [[autoref_trace_failed_id, goals_limit=1]]
  apply (autoref_monadic (trace))
  done
concrete_definition cyc_checker_code uses cyc_checker_impl
export_code cyc_checker_code checking SML

lemma cyc_checker_code_correct:
  assumes 1: "fb_graph G"
  assumes 2: "(Gi, G) \<in> \<langle>Rm, Id\<rangle>g_impl_rel_ext"
  assumes 4: "cyc_checker_code Gi = dRETURN x"
  shows "x \<longleftrightarrow> (\<not>acyclic (g_E G \<inter> ((g_E G)\<^sup>* `` g_V0 G) \<times> UNIV))"
proof -
  note cyc_checker_code.refine[OF 2]
  also note cyc_checker_impl_refine
  also note cyc_checker_correct
  finally show ?thesis using 1 4
  unfolding cyc_checker_spec_def by auto
qed

schematic_lemma cyc_checker_rec_impl:
  defines "V \<equiv> Id :: ('v \<times> 'v::hashable) set"
  assumes [unfolded V_def,autoref_rules]:
    "(Gi, G) \<in> \<langle>Rm, V\<rangle>g_impl_rel_ext"
  notes [unfolded V_def,autoref_tyrel] = 
    TYRELI[where R="\<langle>V\<rangle>dflt_ahs_rel"]
    TYRELI[where R="\<langle>V \<times>\<^sub>r \<langle>V\<rangle>list_set_rel\<rangle>ras_rel"]
  shows "nres_of (?c::?'c dres) \<le>\<Down>?R (cyc_checker_rec_impl G)"
  unfolding DFS_code_unfold
  using [[autoref_trace_failed_id, goals_limit=1]]
  apply (autoref_monadic (trace))
  done
concrete_definition cyc_checker_rec_code uses cyc_checker_rec_impl
prepare_code_thms cyc_checker_rec_code_def
export_code cyc_checker_rec_code checking SML

lemma cyc_checker_rec_code_correct:
  assumes 1: "fb_graph G"
  assumes 2: "(Gi, G) \<in> \<langle>Rm, Id\<rangle>g_impl_rel_ext"
  assumes 4: "cyc_checker_rec_code Gi = dRETURN x"
  shows "x \<longleftrightarrow> (\<not>acyclic (g_E G \<inter> ((g_E G)\<^sup>* `` g_V0 G) \<times> UNIV))"
proof -
  note cyc_checker_rec_code.refine[OF 2]
  also note cyc_checker_rec_impl_refine
  also note cyc_checker_correct
  finally show ?thesis using 1 4 
  unfolding cyc_checker_spec_def by auto
qed
  
end

