
structure Uint : sig
  val set_bit : Word.word -> IntInf.int -> bool -> Word.word
  val shiftl : Word.word -> IntInf.int -> Word.word
  val shiftr : Word.word -> IntInf.int -> Word.word
  val shiftr_signed : Word.word -> IntInf.int -> Word.word
  val test_bit : Word.word -> IntInf.int -> bool
end = struct

fun set_bit x n b =
  let val mask = Word.<< (0wx1, Word.fromLargeInt (IntInf.toLarge n))
  in if b then Word.orb (x, mask)
     else Word.andb (x, Word.notb mask)
  end

fun shiftl x n =
  Word.<< (x, Word.fromLargeInt (IntInf.toLarge n))

fun shiftr x n =
  Word.>> (x, Word.fromLargeInt (IntInf.toLarge n))

fun shiftr_signed x n =
  Word.~>> (x, Word.fromLargeInt (IntInf.toLarge n))

fun test_bit x n =
  Word.andb (x, Word.<< (0wx1, Word.fromLargeInt (IntInf.toLarge n))) <> Word.fromInt 0

end; (* struct Uint *)


structure STArray = struct

datatype 'a Cell = Invalid | Value of 'a array;

exception AccessedOldVersion;

type 'a array = 'a Cell Unsynchronized.ref;

fun fromList l = Unsynchronized.ref (Value (Array.fromList l));
fun array (size, v) = Unsynchronized.ref (Value (Array.array (size,v)));
fun tabulate (size, f) = Unsynchronized.ref (Value (Array.tabulate(size, f)));
fun sub (Unsynchronized.ref Invalid, idx) = raise AccessedOldVersion |
    sub (Unsynchronized.ref (Value a), idx) = Array.sub (a,idx);
fun update (aref,idx,v) =
  case aref of
    (Unsynchronized.ref Invalid) => raise AccessedOldVersion |
    (Unsynchronized.ref (Value a)) => (
      aref := Invalid;
      Array.update (a,idx,v);
      Unsynchronized.ref (Value a)
    );

fun length (Unsynchronized.ref Invalid) = raise AccessedOldVersion |
    length (Unsynchronized.ref (Value a)) = Array.length a

fun grow (aref, i, x) = case aref of 
  (Unsynchronized.ref Invalid) => raise AccessedOldVersion |
  (Unsynchronized.ref (Value a)) => (
    let val len=Array.length a;
        val na = Array.array (len+i,x) 
    in
      aref := Invalid;
      Array.copy {src=a, dst=na, di=0};
      Unsynchronized.ref (Value na)
    end
    );

fun shrink (aref, sz) = case aref of
  (Unsynchronized.ref Invalid) => raise AccessedOldVersion |
  (Unsynchronized.ref (Value a)) => (
    if sz > Array.length a then 
      raise Size
    else (
      aref:=Invalid;
      Unsynchronized.ref (Value (Array.tabulate (sz,fn i => Array.sub (a,i))))
    )
  );

structure IsabelleMapping = struct
type 'a ArrayType = 'a array;

fun new_array (a:'a) (n:int) = array (n, a);

fun array_length (a:'a ArrayType) = length a;

fun array_get (a:'a ArrayType) (i:int) = sub (a, i);

fun array_set (a:'a ArrayType) (i:int) (e:'a) = update (a, i, e);

fun array_of_list (xs:'a list) = fromList xs;

fun array_grow (a:'a ArrayType) (i:int) (x:'a) = grow (a, i, x);

fun array_shrink (a:'a ArrayType) (sz:int) = shrink (a,sz);

end;

end;

structure FArray = struct
  datatype 'a Cell = Value of 'a Array.array | Upd of (int*'a*'a Cell Unsynchronized.ref);

  type 'a array = 'a Cell Unsynchronized.ref;

  fun array (size,v) = Unsynchronized.ref (Value (Array.array (size,v)));
  fun tabulate (size, f) = Unsynchronized.ref (Value (Array.tabulate(size, f)));
  fun fromList l = Unsynchronized.ref (Value (Array.fromList l));

  fun sub (Unsynchronized.ref (Value a), idx) = Array.sub (a,idx) |
      sub (Unsynchronized.ref (Upd (i,v,cr)),idx) =
        if i=idx then v
        else sub (cr,idx);

  fun length (Unsynchronized.ref (Value a)) = Array.length a |
      length (Unsynchronized.ref (Upd (i,v,cr))) = length cr;

  fun realize_aux (aref, v) =
    case aref of
      (Unsynchronized.ref (Value a)) => (
        let
          val len = Array.length a;
          val a' = Array.array (len,v);
        in
          Array.copy {src=a, dst=a', di=0};
          Unsynchronized.ref (Value a')
        end
      ) |
      (Unsynchronized.ref (Upd (i,v,cr))) => (
        let val res=realize_aux (cr,v) in
          case res of
            (Unsynchronized.ref (Value a)) => (Array.update (a,i,v); res)
        end
      );

  fun realize aref =
    case aref of
      (Unsynchronized.ref (Value _)) => aref |
      (Unsynchronized.ref (Upd (i,v,cr))) => realize_aux(aref,v);

  fun update (aref,idx,v) =
    case aref of
      (Unsynchronized.ref (Value a)) => (
        let val nref=Unsynchronized.ref (Value a) in
          aref := Upd (idx,Array.sub(a,idx),nref);
          Array.update (a,idx,v);
          nref
        end
      ) |
      (Unsynchronized.ref (Upd _)) =>
        let val ra = realize_aux(aref,v) in
          case ra of
            (Unsynchronized.ref (Value a)) => Array.update (a,idx,v);
          ra
        end
      ;

  fun grow (aref, inc, x) = case aref of 
    (Unsynchronized.ref (Value a)) => (
      let val len=Array.length a;
          val na = Array.array (len+inc,x) 
      in
        Array.copy {src=a, dst=na, di=0};
        Unsynchronized.ref (Value na)
      end
      )
  | (Unsynchronized.ref (Upd _)) => (
    grow (realize aref, inc, x)
  );

  fun shrink (aref, sz) = case aref of
    (Unsynchronized.ref (Value a)) => (
      if sz > Array.length a then 
        raise Size
      else (
        Unsynchronized.ref (Value (Array.tabulate (sz,fn i => Array.sub (a,i))))
      )
    ) |
    (Unsynchronized.ref (Upd _)) => (
      shrink (realize aref,sz)
    );

structure IsabelleMapping = struct
type 'a ArrayType = 'a array;

fun new_array (a:'a) (n:int) = array (n, a);

fun array_length (a:'a ArrayType) = length a;

fun array_get (a:'a ArrayType) (i:int) = sub (a, i);

fun array_set (a:'a ArrayType) (i:int) (e:'a) = update (a, i, e);

fun array_of_list (xs:'a list) = fromList xs;

fun array_grow (a:'a ArrayType) (i:int) (x:'a) = grow (a, i, x);

fun array_shrink (a:'a ArrayType) (sz:int) = shrink (a,sz);

fun array_get_oo (d:'a) (a:'a ArrayType) (i:int) = 
  sub (a,i) handle Subscript => d

fun array_set_oo (d:(unit->'a ArrayType)) (a:'a ArrayType) (i:int) (e:'a) = 
  update (a, i, e) handle Subscript => d ()

end;
end;




structure Bits_Integer : sig
  val set_bit : IntInf.int -> IntInf.int -> bool -> IntInf.int
  val shiftl : IntInf.int -> IntInf.int -> IntInf.int
  val shiftr : IntInf.int -> IntInf.int -> IntInf.int
  val test_bit : IntInf.int -> IntInf.int -> bool
end = struct

val maxWord = IntInf.pow (2, Word.wordSize);

fun set_bit x n b =
  if n < maxWord then
    if b then IntInf.orb (x, IntInf.<< (1, Word.fromInt n))
    else IntInf.andb (x, IntInf.notb (IntInf.<< (1, Word.fromInt n)))
  else raise (Fail ("Bit index too large: " ^ IntInf.toString n));

fun shiftl x n =
  if n < maxWord then IntInf.<< (x, Word.fromInt n)
  else raise (Fail ("Shift operand too large: " ^ IntInf.toString n));

fun shiftr x n =
  if n < maxWord then IntInf.~>> (x, Word.fromInt n)
  else raise (Fail ("Shift operand too large: " ^ IntInf.toString n));

fun test_bit x n =
  if n < maxWord then IntInf.andb (x, IntInf.<< (1, Word.fromInt n)) <> 0
  else raise (Fail ("Bit index too large: " ^ IntInf.toString n));

end; (*struct Bits_Integer*)

structure Stress_Test : sig
  datatype nat = Nat of IntInf.int
  datatype num = One | Bit0 of num | Bit1 of num
  val integer_of_nat : nat -> IntInf.int
  val plus_nat : nat -> nat -> nat
  val one_nat : nat
  val suc : nat -> nat
  datatype 'a itself = Type
  type 'a ord
  val less_eq : 'a ord -> 'a -> 'a -> bool
  val less : 'a ord -> 'a -> 'a -> bool
  val less_eq_nat : nat -> nat -> bool
  val less_nat : nat -> nat -> bool
  val ord_nat : nat ord
  datatype color = R | B
  datatype compare = Lt | Gt | Eq
  datatype ('a, 'b) rbt = Empty |
    Branch of color * ('a, 'b) rbt * 'a * 'b * ('a, 'b) rbt
  val foldli : 'a list -> ('b -> bool) -> ('a -> 'b -> 'b) -> 'b -> 'b
  val zero_nat : nat
  val equal_nat : nat -> nat -> bool
  val ord_integer : IntInf.int ord
  val max : 'a ord -> 'a -> 'a -> 'a
  val minus_nat : nat -> nat -> nat
  val times_nat : nat -> nat -> nat
  val fold : ('a -> 'b -> 'c -> 'c) -> ('a, 'b) rbt -> 'c -> 'c
  type dflt_size
  val gen_length : nat -> 'a list -> nat
  val size_list : 'a list -> nat
  val paint : color -> ('a, 'b) rbt -> ('a, 'b) rbt
  datatype 'a dres = DSUCCEEDi | DFAILi | DRETURN of 'a
  val sgn_integer : IntInf.int -> IntInf.int
  val abs_integer : IntInf.int -> IntInf.int
  val apsnd : ('a -> 'b) -> 'c * 'a -> 'c * 'b
  val divmod_integer : IntInf.int -> IntInf.int -> IntInf.int * IntInf.int
  val fst : 'a * 'b -> 'a
  val div_integer : IntInf.int -> IntInf.int -> IntInf.int
  val div_nat : nat -> nat -> nat
  val snd : 'a * 'b -> 'b
  val mod_integer : IntInf.int -> IntInf.int -> IntInf.int
  val mod_nat : nat -> nat -> nat
  val gen_add :
    ('a -> ('b -> bool) -> ('c * 'd -> 'e -> 'e) -> 'e -> 'e) ->
      ('c -> 'd -> 'e -> 'e) -> 'e -> 'a -> 'e
  val gen_ball :
    ('a -> ('b -> 'b) -> ('c -> 'd -> bool) -> bool -> 'e) ->
      'a -> ('c -> bool) -> 'e
  val balance : ('a, 'b) rbt -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val gen_entries :
    (('a * 'b) * ('a, 'b) rbt) list -> ('a, 'b) rbt -> ('a * 'b) list
  val entries : ('a, 'b) rbt -> ('a * 'b) list
  val rbt_ins :
    'a ord -> ('a -> 'b -> 'b -> 'b) -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val dbind : 'a dres -> ('a -> 'b dres) -> 'b dres
  val gen_equal : ('a -> 'b -> bool) -> ('b -> 'a -> bool) -> 'a -> 'b -> bool
  val rbt_insa :
    ('a -> 'a -> bool) ->
      ('a -> 'b -> 'b -> 'b) -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val skip_red : ('a, 'b) rbt -> ('a, 'b) rbt
  val set_iterator_image :
    ('a -> 'b) ->
      (('c -> bool) -> ('a -> 'c -> 'c) -> 'c -> 'c) ->
        ('c -> bool) -> ('b -> 'c -> 'c) -> 'c -> 'c
  val map_iterator_dom :
    (('a -> bool) -> ('b * 'c -> 'a -> 'a) -> 'a -> 'a) ->
      ('a -> bool) -> ('b -> 'a -> 'a) -> 'a -> 'a
  val it_to_list :
    ('a -> ('b -> bool) -> ('c -> 'c list -> 'c list) -> 'd list -> 'e) ->
      'a -> 'e
  val map2set_memb : ('a -> 'b -> 'c option) -> 'a -> 'b -> bool
  val rm_iterateoi :
    ('a, 'b) rbt -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
  val gen_subseteq :
    ('a -> ('b -> bool) -> bool) -> ('b -> 'c -> bool) -> 'a -> 'c -> bool
  val rbt_lookup : ('a -> 'a -> bool) -> ('a, 'b) rbt -> 'a -> 'b option
  val rs_eq : (nat, unit) rbt -> (nat, unit) rbt -> bool
  val the_res : 'a dres -> 'a
  val set_bit_integer : IntInf.int -> nat -> bool -> IntInf.int
  val bs_insert : nat -> IntInf.int -> IntInf.int
  val bs_empty : unit -> IntInf.int
  val nat_of_integer : IntInf.int -> nat
  val lcg_next : nat -> nat -> nat * nat
  val dWHILET : ('a -> bool) -> ('a -> 'a dres) -> 'a -> 'a dres
  val st_bs : nat -> nat -> nat * (nat * IntInf.int)
  val map2set_insert : ('a -> unit -> 'b -> 'c) -> 'a -> 'b -> 'c
  val rbt_insert_with_key :
    'a ord -> ('a -> 'b -> 'b -> 'b) -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val rbt_insert : 'a ord -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val st_rs : nat -> nat -> nat * (nat * (nat, unit) rbt)
  val len_of_dflt_size : dflt_size itself -> nat
  val set_bit_uint : Word.word -> nat -> bool -> Word.word
  val divmod_nat : nat -> nat -> nat * nat
  val bs_eq : IntInf.int -> IntInf.int -> bool
  val apfst : ('a -> 'b) -> 'a * 'c -> 'b * 'c
  val rbt_insert_with_keya :
    ('a -> 'a -> bool) ->
      ('a -> 'b -> 'b -> 'b) -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val skip_black : ('a, 'b) rbt -> ('a, 'b) rbt
  val compare_height :
    ('a, 'b) rbt -> ('a, 'b) rbt -> ('a, 'b) rbt -> ('a, 'b) rbt -> compare
  val sunion_with :
    ('a -> 'a -> bool) ->
      ('a -> 'b -> 'b -> 'b) ->
        ('a * 'b) list -> ('a * 'b) list -> ('a * 'b) list
  val rbtreeify_g : nat -> ('a * 'b) list -> ('a, 'b) rbt * ('a * 'b) list
  val rbtreeify_f : nat -> ('a * 'b) list -> ('a, 'b) rbt * ('a * 'b) list
  val rbtreeify : ('a * 'b) list -> ('a, 'b) rbt
  val rbt_union_with_key :
    ('a -> 'a -> bool) ->
      ('a -> 'b -> 'b -> 'b) -> ('a, 'b) rbt -> ('a, 'b) rbt -> ('a, 'b) rbt
  val rbt_union :
    ('a -> 'a -> bool) -> ('a, 'b) rbt -> ('a, 'b) rbt -> ('a, 'b) rbt
  val array_get : 'a FArray.IsabelleMapping.ArrayType -> nat -> 'a
  val iam_iteratei_aux :
    nat ->
      ('a option) FArray.IsabelleMapping.ArrayType ->
        ('b -> bool) -> (nat * 'a -> 'b -> 'b) -> 'b -> 'b
  val array_length : 'a FArray.IsabelleMapping.ArrayType -> nat
  val iam_iteratei :
    ('a option) FArray.IsabelleMapping.ArrayType ->
      ('b -> bool) -> (nat * 'a -> 'b -> 'b) -> 'b -> 'b
  val array_get_oo : 'a -> 'a FArray.IsabelleMapping.ArrayType -> nat -> 'a
  val iam_alpha :
    ('a option) FArray.IsabelleMapping.ArrayType -> nat -> 'a option
  val iam_lookup :
    nat -> ('a option) FArray.IsabelleMapping.ArrayType -> 'a option
  val ias_eq :
    (unit option) FArray.IsabelleMapping.ArrayType ->
      (unit option) FArray.IsabelleMapping.ArrayType -> bool
  val iam_increment : nat -> nat -> nat
  val array_set_oo :
    (unit -> 'a FArray.IsabelleMapping.ArrayType) ->
      'a FArray.IsabelleMapping.ArrayType ->
        nat -> 'a -> 'a FArray.IsabelleMapping.ArrayType
  val array_grow :
    'a FArray.IsabelleMapping.ArrayType ->
      nat -> 'a -> 'a FArray.IsabelleMapping.ArrayType
  val array_set :
    'a FArray.IsabelleMapping.ArrayType ->
      nat -> 'a -> 'a FArray.IsabelleMapping.ArrayType
  val iam_update :
    nat ->
      'a -> ('a option) FArray.IsabelleMapping.ArrayType ->
              ('a option) FArray.IsabelleMapping.ArrayType
  val iam_empty : unit -> ('a option) FArray.IsabelleMapping.ArrayType
  val st_ias :
    nat -> nat -> nat * (nat * (unit option) FArray.IsabelleMapping.ArrayType)
  val uv_empty : Word.word list
  val uv_zeroes : Word.word list -> bool
  val uv_equal : Word.word list -> Word.word list -> bool
  val uv_union : Word.word list -> Word.word list -> Word.word list
  val rs_empty : unit -> (nat, unit) rbt
  val rs_union : (nat, unit) rbt -> (nat, unit) rbt -> (nat, unit) rbt
  val bs_union : IntInf.int -> IntInf.int -> IntInf.int
  val ias_empty : unit -> (unit option) FArray.IsabelleMapping.ArrayType
  val ias_union :
    (unit option) FArray.IsabelleMapping.ArrayType ->
      (unit option) FArray.IsabelleMapping.ArrayType ->
        (unit option) FArray.IsabelleMapping.ArrayType
  val rs_insert : nat -> (nat, unit) rbt -> (nat, unit) rbt
  val uv_single_bit : nat -> Word.word list
  val uv_set_bit : nat -> Word.word list -> Word.word list
  val ias_insert :
    nat ->
      (unit option) FArray.IsabelleMapping.ArrayType ->
        (unit option) FArray.IsabelleMapping.ArrayType
end = struct

datatype nat = Nat of IntInf.int;

datatype num = One | Bit0 of num | Bit1 of num;

fun integer_of_nat (Nat x) = x;

fun plus_nat m n = Nat (IntInf.+ (integer_of_nat m, integer_of_nat n));

val one_nat : nat = Nat (1 : IntInf.int);

fun suc n = plus_nat n one_nat;

datatype 'a itself = Type;

type 'a ord = {less_eq : 'a -> 'a -> bool, less : 'a -> 'a -> bool};
val less_eq = #less_eq : 'a ord -> 'a -> 'a -> bool;
val less = #less : 'a ord -> 'a -> 'a -> bool;

fun less_eq_nat m n = IntInf.<= (integer_of_nat m, integer_of_nat n);

fun less_nat m n = IntInf.< (integer_of_nat m, integer_of_nat n);

val ord_nat = {less_eq = less_eq_nat, less = less_nat} : nat ord;

datatype color = R | B;

datatype compare = Lt | Gt | Eq;

datatype ('a, 'b) rbt = Empty |
  Branch of color * ('a, 'b) rbt * 'a * 'b * ('a, 'b) rbt;

fun foldli [] c f sigma = sigma
  | foldli (x :: xs) c f sigma =
    (if c sigma then foldli xs c f (f x sigma) else sigma);

val zero_nat : nat = Nat 0;

fun equal_nat m n = (((integer_of_nat m) : IntInf.int) = (integer_of_nat n));

val ord_integer =
  {less_eq = (fn a => fn b => IntInf.<= (a, b)),
    less = (fn a => fn b => IntInf.< (a, b))}
  : IntInf.int ord;

fun max A_ a b = (if less_eq A_ a b then b else a);

fun minus_nat m n =
  Nat (max ord_integer 0 (IntInf.- (integer_of_nat m, integer_of_nat n)));

fun times_nat m n = Nat (IntInf.* (integer_of_nat m, integer_of_nat n));

fun fold f (Branch (c, lt, k, v, rt)) x = fold f rt (f k v (fold f lt x))
  | fold f Empty x = x;

datatype dflt_size = EMPTY__

fun gen_length n (x :: xs) = gen_length (suc n) xs
  | gen_length n [] = n;

fun size_list x = gen_length zero_nat x;

fun paint c Empty = Empty
  | paint c (Branch (uu, l, k, v, r)) = Branch (c, l, k, v, r);

datatype 'a dres = DSUCCEEDi | DFAILi | DRETURN of 'a;

fun sgn_integer k =
  (if ((k : IntInf.int) = 0) then 0
    else (if IntInf.< (k, 0) then IntInf.~ (1 : IntInf.int)
           else (1 : IntInf.int)));

fun abs_integer k = (if IntInf.< (k, 0) then IntInf.~ k else k);

fun apsnd f (x, y) = (x, f y);

fun divmod_integer k l =
  (if ((k : IntInf.int) = 0) then (0, 0)
    else (if ((l : IntInf.int) = 0) then (0, k)
           else (apsnd o (fn a => fn b => IntInf.* (a, b)) o sgn_integer) l
                  (if (((sgn_integer k) : IntInf.int) = (sgn_integer l))
                    then IntInf.divMod (IntInf.abs k, IntInf.abs l)
                    else let
                           val (r, s) =
                             IntInf.divMod (IntInf.abs k, IntInf.abs l);
                         in
                           (if ((s : IntInf.int) = 0) then (IntInf.~ r, 0)
                             else (IntInf.- (IntInf.~ r, (1 : IntInf.int)),
                                    IntInf.- (abs_integer l, s)))
                         end)));

fun fst (a, b) = a;

fun div_integer k l = fst (divmod_integer k l);

fun div_nat m n = Nat (div_integer (integer_of_nat m) (integer_of_nat n));

fun snd (a, b) = b;

fun mod_integer k l = snd (divmod_integer k l);

fun mod_nat m n = Nat (mod_integer (integer_of_nat m) (integer_of_nat n));

fun gen_add it upd a b = it b (fn _ => true) (fn (aa, ba) => upd aa ba) a;

fun gen_ball it s p = it s (fn x => x) (fn x => fn _ => p x) true;

fun balance (Branch (R, a, w, x, b)) s t (Branch (R, c, y, z, d)) =
  Branch (R, Branch (B, a, w, x, b), s, t, Branch (B, c, y, z, d))
  | balance (Branch (R, Branch (R, a, w, x, b), s, t, c)) y z Empty =
    Branch (R, Branch (B, a, w, x, b), s, t, Branch (B, c, y, z, Empty))
  | balance (Branch (R, Branch (R, a, w, x, b), s, t, c)) y z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, Branch (B, a, w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance (Branch (R, Empty, w, x, Branch (R, b, s, t, c))) y z Empty =
    Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, Empty))
  | balance
    (Branch (R, Branch (B, va, vb, vc, vd), w, x, Branch (R, b, s, t, c))) y z
    Empty =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, Empty))
  | balance (Branch (R, Empty, w, x, Branch (R, b, s, t, c))) y z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, Branch (B, Empty, w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance
    (Branch (R, Branch (B, ve, vf, vg, vh), w, x, Branch (R, b, s, t, c))) y z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, Branch (B, Branch (B, ve, vf, vg, vh), w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance Empty w x (Branch (R, b, s, t, Branch (R, c, y, z, d))) =
    Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, d))
  | balance (Branch (B, va, vb, vc, vd)) w x
    (Branch (R, b, s, t, Branch (R, c, y, z, d))) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, d))
  | balance Empty w x (Branch (R, Branch (R, b, s, t, c), y, z, Empty)) =
    Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, Empty))
  | balance Empty w x
    (Branch (R, Branch (R, b, s, t, c), y, z, Branch (B, va, vb, vc, vd))) =
    Branch
      (R, Branch (B, Empty, w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance (Branch (B, va, vb, vc, vd)) w x
    (Branch (R, Branch (R, b, s, t, c), y, z, Empty)) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, Empty))
  | balance (Branch (B, va, vb, vc, vd)) w x
    (Branch (R, Branch (R, b, s, t, c), y, z, Branch (B, ve, vf, vg, vh))) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, ve, vf, vg, vh)))
  | balance Empty s t Empty = Branch (B, Empty, s, t, Empty)
  | balance Empty s t (Branch (B, va, vb, vc, vd)) =
    Branch (B, Empty, s, t, Branch (B, va, vb, vc, vd))
  | balance Empty s t (Branch (v, Empty, vb, vc, Empty)) =
    Branch (B, Empty, s, t, Branch (v, Empty, vb, vc, Empty))
  | balance Empty s t (Branch (v, Branch (B, ve, vf, vg, vh), vb, vc, Empty)) =
    Branch
      (B, Empty, s, t, Branch (v, Branch (B, ve, vf, vg, vh), vb, vc, Empty))
  | balance Empty s t (Branch (v, Empty, vb, vc, Branch (B, vf, vg, vh, vi))) =
    Branch
      (B, Empty, s, t, Branch (v, Empty, vb, vc, Branch (B, vf, vg, vh, vi)))
  | balance Empty s t
    (Branch (v, Branch (B, ve, vj, vk, vl), vb, vc, Branch (B, vf, vg, vh, vi)))
    = Branch
        (B, Empty, s, t,
          Branch
            (v, Branch (B, ve, vj, vk, vl), vb, vc, Branch (B, vf, vg, vh, vi)))
  | balance (Branch (B, va, vb, vc, vd)) s t Empty =
    Branch (B, Branch (B, va, vb, vc, vd), s, t, Empty)
  | balance (Branch (B, va, vb, vc, vd)) s t (Branch (B, ve, vf, vg, vh)) =
    Branch (B, Branch (B, va, vb, vc, vd), s, t, Branch (B, ve, vf, vg, vh))
  | balance (Branch (B, va, vb, vc, vd)) s t (Branch (v, Empty, vf, vg, Empty))
    = Branch
        (B, Branch (B, va, vb, vc, vd), s, t, Branch (v, Empty, vf, vg, Empty))
  | balance (Branch (B, va, vb, vc, vd)) s t
    (Branch (v, Branch (B, vi, vj, vk, vl), vf, vg, Empty)) =
    Branch
      (B, Branch (B, va, vb, vc, vd), s, t,
        Branch (v, Branch (B, vi, vj, vk, vl), vf, vg, Empty))
  | balance (Branch (B, va, vb, vc, vd)) s t
    (Branch (v, Empty, vf, vg, Branch (B, vj, vk, vl, vm))) =
    Branch
      (B, Branch (B, va, vb, vc, vd), s, t,
        Branch (v, Empty, vf, vg, Branch (B, vj, vk, vl, vm)))
  | balance (Branch (B, va, vb, vc, vd)) s t
    (Branch (v, Branch (B, vi, vn, vo, vp), vf, vg, Branch (B, vj, vk, vl, vm)))
    = Branch
        (B, Branch (B, va, vb, vc, vd), s, t,
          Branch
            (v, Branch (B, vi, vn, vo, vp), vf, vg, Branch (B, vj, vk, vl, vm)))
  | balance (Branch (v, Empty, vb, vc, Empty)) s t Empty =
    Branch (B, Branch (v, Empty, vb, vc, Empty), s, t, Empty)
  | balance (Branch (v, Empty, vb, vc, Branch (B, ve, vf, vg, vh))) s t Empty =
    Branch
      (B, Branch (v, Empty, vb, vc, Branch (B, ve, vf, vg, vh)), s, t, Empty)
  | balance (Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Empty)) s t Empty =
    Branch
      (B, Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Empty), s, t, Empty)
  | balance
    (Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Branch (B, ve, vj, vk, vl)))
    s t Empty =
    Branch
      (B, Branch
            (v, Branch (B, vf, vg, vh, vi), vb, vc, Branch (B, ve, vj, vk, vl)),
        s, t, Empty)
  | balance (Branch (v, Empty, vf, vg, Empty)) s t (Branch (B, va, vb, vc, vd))
    = Branch
        (B, Branch (v, Empty, vf, vg, Empty), s, t, Branch (B, va, vb, vc, vd))
  | balance (Branch (v, Empty, vf, vg, Branch (B, vi, vj, vk, vl))) s t
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (B, Branch (v, Empty, vf, vg, Branch (B, vi, vj, vk, vl)), s, t,
        Branch (B, va, vb, vc, vd))
  | balance (Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Empty)) s t
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (B, Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Empty), s, t,
        Branch (B, va, vb, vc, vd))
  | balance
    (Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Branch (B, vi, vn, vo, vp)))
    s t (Branch (B, va, vb, vc, vd)) =
    Branch
      (B, Branch
            (v, Branch (B, vj, vk, vl, vm), vf, vg, Branch (B, vi, vn, vo, vp)),
        s, t, Branch (B, va, vb, vc, vd));

fun gen_entries kvts (Branch (c, l, k, v, r)) =
  gen_entries (((k, v), r) :: kvts) l
  | gen_entries ((kv, t) :: kvts) Empty = kv :: gen_entries kvts t
  | gen_entries [] Empty = [];

fun entries x = gen_entries [] x;

fun rbt_ins A_ f k v Empty = Branch (R, Empty, k, v, Empty)
  | rbt_ins A_ f k v (Branch (B, l, x, y, r)) =
    (if less A_ k x then balance (rbt_ins A_ f k v l) x y r
      else (if less A_ x k then balance l x y (rbt_ins A_ f k v r)
             else Branch (B, l, x, f k y v, r)))
  | rbt_ins A_ f k v (Branch (R, l, x, y, r)) =
    (if less A_ k x then Branch (R, rbt_ins A_ f k v l, x, y, r)
      else (if less A_ x k then Branch (R, l, x, y, rbt_ins A_ f k v r)
             else Branch (R, l, x, f k y v, r)));

fun dbind DFAILi f = DFAILi
  | dbind DSUCCEEDi f = DSUCCEEDi
  | dbind (DRETURN x) f = f x;

fun gen_equal ss1 ss2 s1 s2 = ss1 s1 s2 andalso ss2 s2 s1;

fun rbt_insa less f k v (Branch (R, l, x, y, r)) =
  (if less k x then Branch (R, rbt_insa less f k v l, x, y, r)
    else (if less x k then Branch (R, l, x, y, rbt_insa less f k v r)
           else Branch (R, l, x, f k y v, r)))
  | rbt_insa less f k v (Branch (B, l, x, y, r)) =
    (if less k x then balance (rbt_insa less f k v l) x y r
      else (if less x k then balance l x y (rbt_insa less f k v r)
             else Branch (B, l, x, f k y v, r)))
  | rbt_insa less f k v Empty = Branch (R, Empty, k, v, Empty);

fun skip_red (Branch (R, l, k, v, r)) = l
  | skip_red Empty = Empty
  | skip_red (Branch (B, va, vb, vc, vd)) = Branch (B, va, vb, vc, vd);

fun set_iterator_image g it = (fn c => fn f => it c (fn x => f (g x)));

fun map_iterator_dom it = set_iterator_image fst it;

fun it_to_list it s = it s (fn _ => true) (fn x => fn l => l @ [x]) [];

fun map2set_memb l k s = (case l k s of NONE => false | SOME _ => true);

fun rm_iterateoi Empty c f sigma = sigma
  | rm_iterateoi (Branch (col, l, k, v, r)) c f sigma =
    (if c sigma
      then let
             val sigmaa = rm_iterateoi l c f sigma;
           in
             (if c sigmaa then rm_iterateoi r c f (f (k, v) sigmaa) else sigmaa)
           end
      else sigma);

fun gen_subseteq ball1 mem2 s1 s2 = ball1 s1 (fn x => mem2 x s2);

fun rbt_lookup less (Branch (uu, l, x, y, r)) k =
  (if less k x then rbt_lookup less l k
    else (if less x k then rbt_lookup less r k else SOME y))
  | rbt_lookup less Empty k = NONE;

fun rs_eq x =
  gen_equal
    (gen_subseteq
      (gen_ball
        (fn xa =>
          foldli
            (it_to_list (map_iterator_dom o (foldli o it_to_list rm_iterateoi))
              xa)))
      (map2set_memb (fn k => fn t => rbt_lookup less_nat t k)))
    (gen_subseteq
      (gen_ball
        (fn xa =>
          foldli
            (it_to_list (map_iterator_dom o (foldli o it_to_list rm_iterateoi))
              xa)))
      (map2set_memb (fn k => fn t => rbt_lookup less_nat t k)))
    x;

fun the_res (DRETURN x) = x;

fun set_bit_integer x i b = Bits_Integer.set_bit x (integer_of_nat i) b;

fun bs_insert i s = set_bit_integer s i true;

fun bs_empty x = (fn _ => 0) x;

fun nat_of_integer k = Nat (max ord_integer 0 k);

fun lcg_next s r =
  (mod_nat s r,
    mod_nat
      (plus_nat (times_nat (nat_of_integer (22695477 : IntInf.int)) s) one_nat)
      (nat_of_integer (268435456 : IntInf.int)));

fun dWHILET b f s = (if b s then dbind (f s) (dWHILET b f) else DRETURN s);

fun st_bs m n =
  the_res
    (dWHILET (fn (xa, _) => less_nat xa n)
      (fn (a, (aa, ba)) =>
        let
          val (ab, bb) = lcg_next aa m;
          val xb = bs_insert ab ba;
          val (_, bc) = lcg_next bb m;
          val (_, bd) = lcg_next bc m;
        in
          DRETURN (plus_nat a one_nat, (bd, xb))
        end)
      (zero_nat, (one_nat, bs_empty ())));

fun map2set_insert i k s = i k () s;

fun rbt_insert_with_key A_ f k v t = paint B (rbt_ins A_ f k v t);

fun rbt_insert A_ = rbt_insert_with_key A_ (fn _ => fn _ => fn nv => nv);

fun st_rs m n =
  the_res
    (dWHILET (fn (xa, _) => less_nat xa n)
      (fn (a, (aa, ba)) =>
        let
          val (ab, bb) = lcg_next aa m;
          val xb = map2set_insert (rbt_insert ord_nat) ab ba;
          val (_, bc) = lcg_next bb m;
          val (_, bd) = lcg_next bc m;
        in
          DRETURN (plus_nat a one_nat, (bd, xb))
        end)
      (zero_nat, (one_nat, Empty)));

fun len_of_dflt_size Type = nat_of_integer Word.wordSize;

fun set_bit_uint x n b =
  (if less_nat n (len_of_dflt_size Type)
    then Uint.set_bit x (integer_of_nat n) b else x);

fun divmod_nat m n = (div_nat m n, mod_nat m n);

fun bs_eq s1 s2 = ((s1 : IntInf.int) = s2);

fun apfst f (x, y) = (f x, y);

fun rbt_insert_with_keya less f k v t = paint B (rbt_insa less f k v t);

fun skip_black t =
  let
    val ta = skip_red t;
  in
    (case ta of Empty => ta | Branch (R, l, _, _, _) => ta
      | Branch (B, l, _, _, _) => l)
  end;

fun compare_height sx s t tx =
  (case (skip_red sx, (skip_red s, (skip_red t, skip_red tx)))
    of (Empty, (Empty, (_, Empty))) => Eq
    | (Empty, (Empty, (_, Branch (_, _, _, _, _)))) => Lt
    | (Empty, (Branch (_, sa, _, _, _), (Empty, b))) => Eq
    | (Empty, (Branch (_, sa, _, _, _), (Branch (_, ta, _, _, _), Empty))) => Eq
    | (Empty,
        (Branch (_, sa, _, _, _),
          (Branch (_, ta, _, _, _), Branch (_, txa, _, _, _))))
      => compare_height Empty sa ta (skip_black txa)
    | (Branch (_, sxa, _, _, _), (Empty, (Empty, Empty))) => Gt
    | (Branch (_, sxa, _, _, _), (Empty, (Empty, Branch (_, _, _, _, _)))) => Lt
    | (Branch (_, sxa, _, _, _), (Empty, (Branch (_, _, _, _, _), Empty))) => Eq
    | (Branch (_, sxa, _, _, _),
        (Empty, (Branch (_, _, _, _, _), Branch (_, _, _, _, _))))
      => Lt
    | (Branch (_, sxa, _, _, _), (Branch (_, sa, _, _, _), (Empty, xf))) => Gt
    | (Branch (_, sxa, _, _, _),
        (Branch (_, sa, _, _, _), (Branch (_, ta, _, _, _), Empty)))
      => compare_height (skip_black sxa) sa ta Empty
    | (Branch (_, sxa, _, _, _),
        (Branch (_, sa, _, _, _),
          (Branch (_, ta, _, _, _), Branch (_, txa, _, _, _))))
      => compare_height (skip_black sxa) sa ta (skip_black txa));

fun sunion_with less f asa [] = asa
  | sunion_with less f [] bs = bs
  | sunion_with less f ((ka, va) :: asa) ((k, v) :: bs) =
    (if less k ka then (k, v) :: sunion_with less f ((ka, va) :: asa) bs
      else (if less ka k then (ka, va) :: sunion_with less f asa ((k, v) :: bs)
             else (ka, f ka va v) :: sunion_with less f asa bs));

fun rbtreeify_g n kvs =
  (if equal_nat n zero_nat orelse equal_nat n one_nat then (Empty, kvs)
    else let
           val (na, r) = divmod_nat n (nat_of_integer (2 : IntInf.int));
         in
           (if equal_nat r zero_nat
             then let
                    val (t1, (k, v) :: kvsa) = rbtreeify_g na kvs;
                  in
                    apfst (fn a => Branch (B, t1, k, v, a))
                      (rbtreeify_g na kvsa)
                  end
             else let
                    val (t1, (k, v) :: kvsa) = rbtreeify_f na kvs;
                  in
                    apfst (fn a => Branch (B, t1, k, v, a))
                      (rbtreeify_g na kvsa)
                  end)
         end)
and rbtreeify_f n kvs =
  (if equal_nat n zero_nat then (Empty, kvs)
    else (if equal_nat n one_nat
           then let
                  val (k, v) :: kvsa = kvs;
                in
                  (Branch (R, Empty, k, v, Empty), kvsa)
                end
           else let
                  val (na, r) = divmod_nat n (nat_of_integer (2 : IntInf.int));
                in
                  (if equal_nat r zero_nat
                    then let
                           val (t1, (k, v) :: kvsa) = rbtreeify_f na kvs;
                         in
                           apfst (fn a => Branch (B, t1, k, v, a))
                             (rbtreeify_g na kvsa)
                         end
                    else let
                           val (t1, (k, v) :: kvsa) = rbtreeify_f na kvs;
                         in
                           apfst (fn a => Branch (B, t1, k, v, a))
                             (rbtreeify_f na kvsa)
                         end)
                end));

fun rbtreeify kvs = fst (rbtreeify_g (suc (size_list kvs)) kvs);

fun rbt_union_with_key less f t1 t2 =
  (case compare_height t1 t1 t2 t2
    of Lt =>
      fold (rbt_insert_with_keya less (fn k => fn v => fn w => f k w v)) t1 t2
    | Gt => fold (rbt_insert_with_keya less f) t2 t1
    | Eq => rbtreeify (sunion_with less f (entries t1) (entries t2)));

fun rbt_union less = rbt_union_with_key less (fn _ => fn _ => fn rv => rv);

fun array_get a = FArray.IsabelleMapping.array_get a o integer_of_nat;

fun iam_iteratei_aux i a c f sigma =
  (if equal_nat i zero_nat then sigma
    else (if c sigma
           then iam_iteratei_aux (minus_nat i one_nat) a c f
                  (case array_get a (minus_nat i one_nat) of NONE => sigma
                    | SOME x => f (minus_nat i one_nat, x) sigma)
           else sigma));

fun array_length x = (nat_of_integer o FArray.IsabelleMapping.array_length) x;

fun iam_iteratei a = iam_iteratei_aux (array_length a) a;

fun array_get_oo x a = FArray.IsabelleMapping.array_get_oo x a o integer_of_nat;

fun iam_alpha a i = array_get_oo NONE a i;

fun iam_lookup k a = iam_alpha a k;

fun ias_eq x =
  gen_equal
    (gen_subseteq
      (gen_ball
        (fn xa =>
          foldli
            (it_to_list (map_iterator_dom o (foldli o it_to_list iam_iteratei))
              xa)))
      (map2set_memb iam_lookup))
    (gen_subseteq
      (gen_ball
        (fn xa =>
          foldli
            (it_to_list (map_iterator_dom o (foldli o it_to_list iam_iteratei))
              xa)))
      (map2set_memb iam_lookup))
    x;

fun iam_increment l idx =
  max ord_nat (minus_nat (plus_nat idx one_nat) l)
    (plus_nat (times_nat (nat_of_integer (2 : IntInf.int)) l)
      (nat_of_integer (3 : IntInf.int)));

fun array_set_oo f a = FArray.IsabelleMapping.array_set_oo f a o integer_of_nat;

fun array_grow a = FArray.IsabelleMapping.array_grow a o integer_of_nat;

fun array_set a = FArray.IsabelleMapping.array_set a o integer_of_nat;

fun iam_update k v a =
  array_set_oo
    (fn _ =>
      let
        val l = array_length a;
      in
        array_set (array_grow a (iam_increment l k) NONE) k (SOME v)
      end)
    a k (SOME v);

fun iam_empty x = (fn _ => FArray.IsabelleMapping.array_of_list []) x;

fun st_ias m n =
  the_res
    (dWHILET (fn (xa, _) => less_nat xa n)
      (fn (a, (aa, ba)) =>
        let
          val (ab, bb) = lcg_next aa m;
          val xb = map2set_insert iam_update ab ba;
          val (_, bc) = lcg_next bb m;
          val (_, bd) = lcg_next bc m;
        in
          DRETURN (plus_nat a one_nat, (bd, xb))
        end)
      (zero_nat, (one_nat, iam_empty ())));

val uv_empty : Word.word list = [];

fun uv_zeroes (w :: ws) =
  ((w : Word.word) = (Word.fromInt 0)) andalso uv_zeroes ws
  | uv_zeroes [] = true;

fun uv_equal (v :: vs) (w :: ws) = ((v : Word.word) = w) andalso uv_equal vs ws
  | uv_equal (v :: va) [] = uv_zeroes (v :: va)
  | uv_equal [] (v :: va) = uv_zeroes (v :: va)
  | uv_equal [] [] = true;

fun uv_union (v :: vs) (w :: ws) = Word.orb (v, w) :: uv_union vs ws
  | uv_union (v :: va) [] = v :: va
  | uv_union [] ws = ws;

fun rs_empty x = (fn _ => Empty) x;

fun rs_union x = rbt_union less_nat x;

fun bs_union s1 s2 = IntInf.orb (s1, s2);

fun ias_empty x = iam_empty x;

fun ias_union x = gen_add (foldli o it_to_list iam_iteratei) iam_update x;

fun rs_insert x = map2set_insert (rbt_insert ord_nat) x;

fun uv_single_bit n =
  (if less_nat n (len_of_dflt_size Type)
    then [set_bit_uint (Word.fromInt 0) n true]
    else (Word.fromInt 0) ::
           uv_single_bit (minus_nat n (len_of_dflt_size Type)));

fun uv_set_bit i (w :: ws) =
  (if less_nat i (len_of_dflt_size Type) then set_bit_uint w i true :: ws
    else w :: uv_set_bit (minus_nat i (len_of_dflt_size Type)) ws)
  | uv_set_bit i [] = uv_single_bit i;

fun ias_insert x = map2set_insert iam_update x;

end; (*struct Stress_Test*)
