
structure STArray = struct

datatype 'a Cell = Invalid | Value of 'a array;

exception AccessedOldVersion;

type 'a array = 'a Cell Unsynchronized.ref;

fun fromList l = Unsynchronized.ref (Value (Array.fromList l));
fun array (size, v) = Unsynchronized.ref (Value (Array.array (size,v)));
fun tabulate (size, f) = Unsynchronized.ref (Value (Array.tabulate(size, f)));
fun sub (Unsynchronized.ref Invalid, idx) = raise AccessedOldVersion |
    sub (Unsynchronized.ref (Value a), idx) = Array.sub (a,idx);
fun update (aref,idx,v) =
  case aref of
    (Unsynchronized.ref Invalid) => raise AccessedOldVersion |
    (Unsynchronized.ref (Value a)) => (
      aref := Invalid;
      Array.update (a,idx,v);
      Unsynchronized.ref (Value a)
    );

fun length (Unsynchronized.ref Invalid) = raise AccessedOldVersion |
    length (Unsynchronized.ref (Value a)) = Array.length a

fun grow (aref, i, x) = case aref of 
  (Unsynchronized.ref Invalid) => raise AccessedOldVersion |
  (Unsynchronized.ref (Value a)) => (
    let val len=Array.length a;
        val na = Array.array (len+i,x) 
    in
      aref := Invalid;
      Array.copy {src=a, dst=na, di=0};
      Unsynchronized.ref (Value na)
    end
    );

fun shrink (aref, sz) = case aref of
  (Unsynchronized.ref Invalid) => raise AccessedOldVersion |
  (Unsynchronized.ref (Value a)) => (
    if sz > Array.length a then 
      raise Size
    else (
      aref:=Invalid;
      Unsynchronized.ref (Value (Array.tabulate (sz,fn i => Array.sub (a,i))))
    )
  );

structure IsabelleMapping = struct
type 'a ArrayType = 'a array;

fun new_array (a:'a) (n:int) = array (n, a);

fun array_length (a:'a ArrayType) = length a;

fun array_get (a:'a ArrayType) (i:int) = sub (a, i);

fun array_set (a:'a ArrayType) (i:int) (e:'a) = update (a, i, e);

fun array_of_list (xs:'a list) = fromList xs;

fun array_grow (a:'a ArrayType) (i:int) (x:'a) = grow (a, i, x);

fun array_shrink (a:'a ArrayType) (sz:int) = shrink (a,sz);

end;

end;

structure FArray = struct
  datatype 'a Cell = Value of 'a Array.array | Upd of (int*'a*'a Cell Unsynchronized.ref);

  type 'a array = 'a Cell Unsynchronized.ref;

  fun array (size,v) = Unsynchronized.ref (Value (Array.array (size,v)));
  fun tabulate (size, f) = Unsynchronized.ref (Value (Array.tabulate(size, f)));
  fun fromList l = Unsynchronized.ref (Value (Array.fromList l));

  fun sub (Unsynchronized.ref (Value a), idx) = Array.sub (a,idx) |
      sub (Unsynchronized.ref (Upd (i,v,cr)),idx) =
        if i=idx then v
        else sub (cr,idx);

  fun length (Unsynchronized.ref (Value a)) = Array.length a |
      length (Unsynchronized.ref (Upd (i,v,cr))) = length cr;

  fun realize_aux (aref, v) =
    case aref of
      (Unsynchronized.ref (Value a)) => (
        let
          val len = Array.length a;
          val a' = Array.array (len,v);
        in
          Array.copy {src=a, dst=a', di=0};
          Unsynchronized.ref (Value a')
        end
      ) |
      (Unsynchronized.ref (Upd (i,v,cr))) => (
        let val res=realize_aux (cr,v) in
          case res of
            (Unsynchronized.ref (Value a)) => (Array.update (a,i,v); res)
        end
      );

  fun realize aref =
    case aref of
      (Unsynchronized.ref (Value _)) => aref |
      (Unsynchronized.ref (Upd (i,v,cr))) => realize_aux(aref,v);

  fun update (aref,idx,v) =
    case aref of
      (Unsynchronized.ref (Value a)) => (
        let val nref=Unsynchronized.ref (Value a) in
          aref := Upd (idx,Array.sub(a,idx),nref);
          Array.update (a,idx,v);
          nref
        end
      ) |
      (Unsynchronized.ref (Upd _)) =>
        let val ra = realize_aux(aref,v) in
          case ra of
            (Unsynchronized.ref (Value a)) => Array.update (a,idx,v);
          ra
        end
      ;

  fun grow (aref, inc, x) = case aref of 
    (Unsynchronized.ref (Value a)) => (
      let val len=Array.length a;
          val na = Array.array (len+inc,x) 
      in
        Array.copy {src=a, dst=na, di=0};
        Unsynchronized.ref (Value na)
      end
      )
  | (Unsynchronized.ref (Upd _)) => (
    grow (realize aref, inc, x)
  );

  fun shrink (aref, sz) = case aref of
    (Unsynchronized.ref (Value a)) => (
      if sz > Array.length a then 
        raise Size
      else (
        Unsynchronized.ref (Value (Array.tabulate (sz,fn i => Array.sub (a,i))))
      )
    ) |
    (Unsynchronized.ref (Upd _)) => (
      shrink (realize aref,sz)
    );

structure IsabelleMapping = struct
type 'a ArrayType = 'a array;

fun new_array (a:'a) (n:int) = array (n, a);

fun array_length (a:'a ArrayType) = length a;

fun array_get (a:'a ArrayType) (i:int) = sub (a, i);

fun array_set (a:'a ArrayType) (i:int) (e:'a) = update (a, i, e);

fun array_of_list (xs:'a list) = fromList xs;

fun array_grow (a:'a ArrayType) (i:int) (x:'a) = grow (a, i, x);

fun array_shrink (a:'a ArrayType) (sz:int) = shrink (a,sz);

fun array_get_oo (d:'a) (a:'a ArrayType) (i:int) = 
  sub (a,i) handle Subscript => d

fun array_set_oo (d:(unit->'a ArrayType)) (a:'a ArrayType) (i:int) (e:'a) = 
  update (a, i, e) handle Subscript => d ()

end;
end;




structure SCC_Gabov : sig
  val id : 'a -> 'a
  type 'a equal
  val equal : 'a equal -> 'a -> 'a -> bool
  val eq : 'a equal -> 'a -> 'a -> bool
  datatype nat = Nat of IntInf.int
  datatype num = One | Bit0 of num | Bit1 of num
  val integer_of_nat : nat -> IntInf.int
  val plus_nat : nat -> nat -> nat
  val one_nata : nat
  val suc : nat -> nat
  val fold : ('a -> 'b -> 'b) -> 'a list -> 'b -> 'b
  val rev : 'a list -> 'a list
  val null : 'a list -> bool
  datatype 'a itself = Type
  type 'a one
  val one : 'a one -> 'a
  val one_nat : nat one
  type 'a ord
  val less_eq : 'a ord -> 'a -> 'a -> bool
  val less : 'a ord -> 'a -> 'a -> bool
  val less_eq_nat : nat -> nat -> bool
  val less_nat : nat -> nat -> bool
  val ord_nat : nat ord
  val the : 'a option -> 'a
  type 'a zero
  val zero : 'a zero -> 'a
  val filter : ('a -> bool) -> 'a list -> 'a list
  val zero_nata : nat
  val zero_nat : nat zero
  val foldli : 'a list -> ('b -> bool) -> ('a -> 'b -> 'b) -> 'b -> 'b
  val equal_nata : nat -> nat -> bool
  val equal_nat : nat equal
  val ord_integer : IntInf.int ord
  val max : 'a ord -> 'a -> 'a -> 'a
  val minus_nat : nat -> nat -> nat
  val times_nat : nat -> nat -> nat
  val is_none : 'a option -> bool
  val sgn_integer : IntInf.int -> IntInf.int
  val abs_integer : IntInf.int -> IntInf.int
  val apsnd : ('a -> 'b) -> 'c * 'a -> 'c * 'b
  val divmod_integer : IntInf.int -> IntInf.int -> IntInf.int * IntInf.int
  val snd : 'a * 'b -> 'b
  val mod_integer : IntInf.int -> IntInf.int -> IntInf.int
  val mod_nat : nat -> nat -> nat
  val fst : 'a * 'b -> 'a
  type 'a hashable
  val hashcode : 'a hashable -> 'a -> nat
  val bounded_hashcode : 'a hashable -> nat -> 'a -> nat
  val def_hashmap_size : 'a hashable -> 'a itself -> nat
  val list_map_update_aux :
    ('a -> 'a -> bool) ->
      'a -> 'b -> ('a * 'b) list -> ('a * 'b) list -> ('a * 'b) list
  val list_map_update :
    ('a -> 'a -> bool) -> 'a -> 'b -> ('a * 'b) list -> ('a * 'b) list
  val list_map_lookup : ('a -> 'a -> bool) -> 'a -> ('a * 'b) list -> 'b option
  datatype ('a, 'b) hashmap =
    HashMap of (('a * 'b) list) FArray.IsabelleMapping.ArrayType * nat
  val nat_of_integer : IntInf.int -> nat
  val array_length : 'a FArray.IsabelleMapping.ArrayType -> nat
  val array_set :
    'a FArray.IsabelleMapping.ArrayType ->
      nat -> 'a -> 'a FArray.IsabelleMapping.ArrayType
  val array_get : 'a FArray.IsabelleMapping.ArrayType -> nat -> 'a
  val ahm_update_aux :
    ('a -> 'a -> bool) ->
      (nat -> 'a -> nat) -> ('a, 'b) hashmap -> 'a -> 'b -> ('a, 'b) hashmap
  val idx_iteratei_aux :
    ('a -> nat -> 'b) ->
      nat -> nat -> 'a -> ('c -> bool) -> ('b -> 'c -> 'c) -> 'c -> 'c
  val idx_iteratei :
    ('a -> nat -> 'b) ->
      ('a -> nat) -> 'a -> ('c -> bool) -> ('b -> 'c -> 'c) -> 'c -> 'c
  val ahm_iteratei_aux :
    (('a * 'b) list) FArray.IsabelleMapping.ArrayType ->
      ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
  val ahm_rehash_auxa :
    (nat -> 'a -> nat) ->
      nat ->
        'a * 'b ->
          (('a * 'b) list) FArray.IsabelleMapping.ArrayType ->
            (('a * 'b) list) FArray.IsabelleMapping.ArrayType
  val new_array : 'a -> nat -> 'a FArray.IsabelleMapping.ArrayType
  val ahm_rehash_aux :
    (nat -> 'a -> nat) ->
      (('a * 'b) list) FArray.IsabelleMapping.ArrayType ->
        nat -> (('a * 'b) list) FArray.IsabelleMapping.ArrayType
  val ahm_rehash :
    (nat -> 'a -> nat) -> ('a, 'b) hashmap -> nat -> ('a, 'b) hashmap
  val load_factor : nat
  val ahm_filled : ('a, 'b) hashmap -> bool
  val hm_grow : ('a, 'b) hashmap -> nat
  val ahm_update :
    ('a -> 'a -> bool) ->
      (nat -> 'a -> nat) -> 'a -> 'b -> ('a, 'b) hashmap -> ('a, 'b) hashmap
  val list_map_delete_aux :
    ('a -> 'a -> bool) ->
      'a -> ('a * 'b) list -> ('a * 'b) list -> ('a * 'b) list
  val list_map_delete :
    ('a -> 'a -> bool) -> 'a -> ('a * 'b) list -> ('a * 'b) list
  val ahm_delete :
    ('a -> 'a -> bool) ->
      (nat -> 'a -> nat) -> 'a -> ('a, 'b) hashmap -> ('a, 'b) hashmap
  datatype ('a, 'b) impl_state_ext =
    Impl_state_ext of
      ('a FArray.IsabelleMapping.ArrayType * nat) *
        ((nat * bool) FArray.IsabelleMapping.ArrayType * nat) *
        ((nat * 'a list) FArray.IsabelleMapping.ArrayType * nat) *
        ('a, nat) hashmap * ('a, unit) hashmap * 'b
  val impl_on_stack_update :
    (('a, nat) hashmap -> ('a, nat) hashmap) ->
      ('a, 'b) impl_state_ext -> ('a, 'b) impl_state_ext
  val impl_s_stack_update :
    ((nat * bool) FArray.IsabelleMapping.ArrayType * nat ->
      (nat * bool) FArray.IsabelleMapping.ArrayType * nat) ->
      ('a, 'b) impl_state_ext -> ('a, 'b) impl_state_ext
  val impl_stack_update :
    ('a FArray.IsabelleMapping.ArrayType * nat ->
      'a FArray.IsabelleMapping.ArrayType * nat) ->
      ('a, 'b) impl_state_ext -> ('a, 'b) impl_state_ext
  val as_length : 'a FArray.IsabelleMapping.ArrayType * nat -> nat
  val map2set_insert : ('a -> unit -> 'b -> 'c) -> 'a -> 'b -> 'c
  val impl_done_update :
    (('a, unit) hashmap -> ('a, unit) hashmap) ->
      ('a, 'b) impl_state_ext -> ('a, 'b) impl_state_ext
  val array_shrink :
    'a FArray.IsabelleMapping.ArrayType ->
      nat -> 'a FArray.IsabelleMapping.ArrayType
  val as_shrink :
    'a FArray.IsabelleMapping.ArrayType * nat ->
      'a FArray.IsabelleMapping.ArrayType * nat
  val as_take :
    nat ->
      'a FArray.IsabelleMapping.ArrayType * nat ->
        'a FArray.IsabelleMapping.ArrayType * nat
  val as_top : 'a FArray.IsabelleMapping.ArrayType * nat -> 'a
  val as_pop :
    'a FArray.IsabelleMapping.ArrayType * nat ->
      'a FArray.IsabelleMapping.ArrayType * nat
  val as_get : 'a FArray.IsabelleMapping.ArrayType * nat -> nat -> 'a
  val impl_on_stack : ('a, 'b) impl_state_ext -> ('a, nat) hashmap
  val whilea : ('a -> bool) -> ('a -> 'a) -> 'a -> 'a
  val impl_s_stack :
    ('a, 'b) impl_state_ext ->
      (nat * bool) FArray.IsabelleMapping.ArrayType * nat
  val impl_stack :
    ('a, 'b) impl_state_ext -> 'a FArray.IsabelleMapping.ArrayType * nat
  val impl_done : ('a, 'b) impl_state_ext -> ('a, unit) hashmap
  val code_pop :
    'a equal * 'a hashable ->
      ('a, unit) impl_state_ext -> ('a, unit) impl_state_ext
  val is_None : 'a option -> bool
  val as_is_empty : 'a FArray.IsabelleMapping.ArrayType * nat -> bool
  val impl_i_stack_update :
    ((nat * 'a list) FArray.IsabelleMapping.ArrayType * nat ->
      (nat * 'a list) FArray.IsabelleMapping.ArrayType * nat) ->
      ('a, 'b) impl_state_ext -> ('a, 'b) impl_state_ext
  val impl_conc_seg_start :
    'a hashable -> ('a, unit) impl_state_ext -> nat -> nat
  val array_grow :
    'a FArray.IsabelleMapping.ArrayType ->
      nat -> 'a -> 'a FArray.IsabelleMapping.ArrayType
  val as_push :
    'a FArray.IsabelleMapping.ArrayType * nat ->
      'a -> 'a FArray.IsabelleMapping.ArrayType * nat
  val impl_num_seg : 'a hashable -> ('a, unit) impl_state_ext -> nat
  val impl_top_seg : 'a hashable -> ('a, unit) impl_state_ext -> nat
  val impl_i_stack :
    ('a, 'b) impl_state_ext ->
      (nat * 'a list) FArray.IsabelleMapping.ArrayType * nat
  val code_top_out_edge :
    'a hashable ->
      ('a, unit) impl_state_ext -> 'a option * ('a, unit) impl_state_ext
  val ahm_lookup_aux :
    ('a -> 'a -> bool) ->
      (nat -> 'a -> nat) ->
        'a -> (('a * 'b) list) FArray.IsabelleMapping.ArrayType -> 'b option
  val ahm_lookup :
    ('a -> 'a -> bool) ->
      (nat -> 'a -> nat) -> 'a -> ('a, 'b) hashmap -> 'b option
  val impl_is_on_stack :
    'a equal * 'a hashable -> ('a, unit) impl_state_ext -> 'a -> bool
  val as_singleton : 'b one -> 'a -> 'a FArray.IsabelleMapping.ArrayType * 'b
  val new_hashmap_with : nat -> ('a, 'b) hashmap
  val ahm_empty : nat -> ('a, 'b) hashmap
  val is_Nil : 'a list -> bool
  val as_empty : 'b zero -> unit -> 'a FArray.IsabelleMapping.ArrayType * 'b
  val code_init_state :
    'a equal * 'a hashable ->
      'a -> ('a -> bool) -> ('a -> 'a list) -> ('a, unit) impl_state_ext
  val slg_restrict_right :
    ('a -> 'b -> bool) -> ('c -> 'a list) -> 'b -> 'c -> 'a list
  val as_set :
    'a FArray.IsabelleMapping.ArrayType * nat ->
      nat -> 'a -> 'a FArray.IsabelleMapping.ArrayType * nat
  val code_collapse :
    'a equal * 'a hashable ->
      ('a -> 'a list) ->
        'a -> ('a, unit) impl_state_ext ->
                ('a * (('a -> 'a list) * ('a -> 'a list))) option *
                  ('a, unit) impl_state_ext
  val map2set_memb : ('a -> 'b -> 'c option) -> 'a -> 'b -> bool
  val impl_is_done :
    'a equal * 'a hashable -> ('a, unit) impl_state_ext -> 'a -> bool
  val code_push :
    'a equal * 'a hashable ->
      ('a -> bool) ->
        ('a -> 'a list) ->
          'a -> ('a, unit) impl_state_ext -> ('a, unit) impl_state_ext
  val code_scc_aux :
    'a equal * 'a hashable ->
      ('a -> 'a list) ->
        ('a -> bool) -> 'a -> ('a * (('a -> 'a list) * ('a -> 'a list))) option
  val code_scc :
    'a equal * 'a hashable -> ('a -> 'a list) -> ('a -> bool) -> 'a -> bool
  val def_hashmap_size_nat : nat itself -> nat
  val bounded_hashcode_nat : nat -> nat -> nat
  val hashcode_nat : nat -> nat
  val hashable_nat : nat hashable
  val list_map_pick_remove : 'a list -> 'a * 'a list
  val list_map_isEmpty : ('a * 'b) list -> bool
  val wset_find_path_code :
    'a equal * 'a hashable ->
      ('a -> 'a list) -> 'a list -> ('a -> bool) -> ('a list * 'a) option
  val find_path1_code :
    'a equal * 'a hashable ->
      ('a -> 'a list) -> 'a -> ('a -> bool) -> ('a list * 'a) option
  val code_scc_ce :
    'a equal * 'a hashable ->
      ('a -> 'a list) -> ('a -> bool) -> 'a -> ('a * ('a list * 'a list)) option
  val code_scc_nat : (nat -> nat list) -> (nat -> bool) -> nat -> bool
  val array_get_oo : 'a -> 'a FArray.IsabelleMapping.ArrayType -> nat -> 'a
  val array_set_oo :
    (unit -> 'a FArray.IsabelleMapping.ArrayType) ->
      'a FArray.IsabelleMapping.ArrayType ->
        nat -> 'a -> 'a FArray.IsabelleMapping.ArrayType
  val iam_alpha :
    ('a option) FArray.IsabelleMapping.ArrayType -> nat -> 'a option
  val iam_empty : unit -> ('a option) FArray.IsabelleMapping.ArrayType
  val code_scc_ce_nat :
    (nat -> nat list) ->
      (nat -> bool) -> nat -> (nat * (nat list * nat list)) option
  val iam_increment : nat -> nat -> nat
  val iam_update :
    nat ->
      'a -> ('a option) FArray.IsabelleMapping.ArrayType ->
              ('a option) FArray.IsabelleMapping.ArrayType
  val glist_member : ('a -> 'a -> bool) -> 'a -> 'a list -> bool
  val glist_insert : ('a -> 'a -> bool) -> 'a -> 'a list -> 'a list
  val acc_of_list_impl :
    nat list -> (unit option) FArray.IsabelleMapping.ArrayType
  val succ_of_list_impl : (nat * nat) list -> nat -> nat list
  val acc_of_list_impl_fun : nat list -> nat -> bool
end = struct

fun id x = (fn xa => xa) x;

type 'a equal = {equal : 'a -> 'a -> bool};
val equal = #equal : 'a equal -> 'a -> 'a -> bool;

fun eq A_ a b = equal A_ a b;

datatype nat = Nat of IntInf.int;

datatype num = One | Bit0 of num | Bit1 of num;

fun integer_of_nat (Nat x) = x;

fun plus_nat m n = Nat (IntInf.+ (integer_of_nat m, integer_of_nat n));

val one_nata : nat = Nat (1 : IntInf.int);

fun suc n = plus_nat n one_nata;

fun fold f (x :: xs) s = fold f xs (f x s)
  | fold f [] s = s;

fun rev xs = fold (fn a => fn b => a :: b) xs [];

fun null [] = true
  | null (x :: xs) = false;

datatype 'a itself = Type;

type 'a one = {one : 'a};
val one = #one : 'a one -> 'a;

val one_nat = {one = one_nata} : nat one;

type 'a ord = {less_eq : 'a -> 'a -> bool, less : 'a -> 'a -> bool};
val less_eq = #less_eq : 'a ord -> 'a -> 'a -> bool;
val less = #less : 'a ord -> 'a -> 'a -> bool;

fun less_eq_nat m n = IntInf.<= (integer_of_nat m, integer_of_nat n);

fun less_nat m n = IntInf.< (integer_of_nat m, integer_of_nat n);

val ord_nat = {less_eq = less_eq_nat, less = less_nat} : nat ord;

fun the (SOME x) = x;

type 'a zero = {zero : 'a};
val zero = #zero : 'a zero -> 'a;

fun filter p [] = []
  | filter p (x :: xs) = (if p x then x :: filter p xs else filter p xs);

val zero_nata : nat = Nat 0;

val zero_nat = {zero = zero_nata} : nat zero;

fun foldli [] c f sigma = sigma
  | foldli (x :: xs) c f sigma =
    (if c sigma then foldli xs c f (f x sigma) else sigma);

fun equal_nata m n = (((integer_of_nat m) : IntInf.int) = (integer_of_nat n));

val equal_nat = {equal = equal_nata} : nat equal;

val ord_integer =
  {less_eq = (fn a => fn b => IntInf.<= (a, b)),
    less = (fn a => fn b => IntInf.< (a, b))}
  : IntInf.int ord;

fun max A_ a b = (if less_eq A_ a b then b else a);

fun minus_nat m n =
  Nat (max ord_integer 0 (IntInf.- (integer_of_nat m, integer_of_nat n)));

fun times_nat m n = Nat (IntInf.* (integer_of_nat m, integer_of_nat n));

fun is_none (SOME x) = false
  | is_none NONE = true;

fun sgn_integer k =
  (if ((k : IntInf.int) = 0) then 0
    else (if IntInf.< (k, 0) then IntInf.~ (1 : IntInf.int)
           else (1 : IntInf.int)));

fun abs_integer k = (if IntInf.< (k, 0) then IntInf.~ k else k);

fun apsnd f (x, y) = (x, f y);

fun divmod_integer k l =
  (if ((k : IntInf.int) = 0) then (0, 0)
    else (if ((l : IntInf.int) = 0) then (0, k)
           else (apsnd o (fn a => fn b => IntInf.* (a, b)) o sgn_integer) l
                  (if (((sgn_integer k) : IntInf.int) = (sgn_integer l))
                    then IntInf.divMod (IntInf.abs k, IntInf.abs l)
                    else let
                           val (r, s) =
                             IntInf.divMod (IntInf.abs k, IntInf.abs l);
                         in
                           (if ((s : IntInf.int) = 0) then (IntInf.~ r, 0)
                             else (IntInf.- (IntInf.~ r, (1 : IntInf.int)),
                                    IntInf.- (abs_integer l, s)))
                         end)));

fun snd (a, b) = b;

fun mod_integer k l = snd (divmod_integer k l);

fun mod_nat m n = Nat (mod_integer (integer_of_nat m) (integer_of_nat n));

fun fst (a, b) = a;

type 'a hashable =
  {hashcode : 'a -> nat, bounded_hashcode : nat -> 'a -> nat,
    def_hashmap_size : 'a itself -> nat};
val hashcode = #hashcode : 'a hashable -> 'a -> nat;
val bounded_hashcode = #bounded_hashcode : 'a hashable -> nat -> 'a -> nat;
val def_hashmap_size = #def_hashmap_size : 'a hashable -> 'a itself -> nat;

fun list_map_update_aux eq k v [] accu = (k, v) :: accu
  | list_map_update_aux eq k v (x :: xs) accu =
    (if eq (fst x) k then (k, v) :: xs @ accu
      else list_map_update_aux eq k v xs (x :: accu));

fun list_map_update eq k v m = list_map_update_aux eq k v m [];

fun list_map_lookup eq uu [] = NONE
  | list_map_lookup eq k (y :: ys) =
    (if eq (fst y) k then SOME (snd y) else list_map_lookup eq k ys);

datatype ('a, 'b) hashmap =
  HashMap of (('a * 'b) list) FArray.IsabelleMapping.ArrayType * nat;

fun nat_of_integer k = Nat (max ord_integer 0 k);

fun array_length x = (nat_of_integer o FArray.IsabelleMapping.array_length) x;

fun array_set a = FArray.IsabelleMapping.array_set a o integer_of_nat;

fun array_get a = FArray.IsabelleMapping.array_get a o integer_of_nat;

fun ahm_update_aux eq bhc (HashMap (a, n)) k v =
  let
    val h = bhc (array_length a) k;
    val m = array_get a h;
    val insert = is_none (list_map_lookup eq k m);
  in
    HashMap
      (array_set a h (list_map_update eq k v m),
        (if insert then plus_nat n one_nata else n))
  end;

fun idx_iteratei_aux get sz i l c f sigma =
  (if equal_nata i zero_nata orelse not (c sigma) then sigma
    else idx_iteratei_aux get sz (minus_nat i one_nata) l c f
           (f (get l (minus_nat sz i)) sigma));

fun idx_iteratei get sz l c f sigma =
  idx_iteratei_aux get (sz l) (sz l) l c f sigma;

fun ahm_iteratei_aux a c f sigma =
  idx_iteratei array_get array_length a c (fn x => foldli x c f) sigma;

fun ahm_rehash_auxa bhc n kv a =
  let
    val h = bhc n (fst kv);
  in
    array_set a h (kv :: array_get a h)
  end;

fun new_array v = FArray.IsabelleMapping.new_array v o integer_of_nat;

fun ahm_rehash_aux bhc a sz =
  ahm_iteratei_aux a (fn _ => true) (ahm_rehash_auxa bhc sz) (new_array [] sz);

fun ahm_rehash bhc (HashMap (a, n)) sz = HashMap (ahm_rehash_aux bhc a sz, n);

val load_factor : nat = nat_of_integer (75 : IntInf.int);

fun ahm_filled (HashMap (a, n)) =
  less_eq_nat (times_nat (array_length a) load_factor)
    (times_nat n (nat_of_integer (100 : IntInf.int)));

fun hm_grow (HashMap (a, n)) =
  plus_nat (times_nat (nat_of_integer (2 : IntInf.int)) (array_length a))
    (nat_of_integer (3 : IntInf.int));

fun ahm_update eq bhc k v hm =
  let
    val hma = ahm_update_aux eq bhc hm k v;
  in
    (if ahm_filled hma then ahm_rehash bhc hma (hm_grow hma) else hma)
  end;

fun list_map_delete_aux eq k [] accu = accu
  | list_map_delete_aux eq k (x :: xs) accu =
    (if eq (fst x) k then xs @ accu
      else list_map_delete_aux eq k xs (x :: accu));

fun list_map_delete eq k m = list_map_delete_aux eq k m [];

fun ahm_delete eq bhc k (HashMap (a, n)) =
  let
    val h = bhc (array_length a) k;
    val m = array_get a h;
    val deleted = not (is_none (list_map_lookup eq k m));
  in
    HashMap
      (array_set a h (list_map_delete eq k m),
        (if deleted then minus_nat n one_nata else n))
  end;

datatype ('a, 'b) impl_state_ext =
  Impl_state_ext of
    ('a FArray.IsabelleMapping.ArrayType * nat) *
      ((nat * bool) FArray.IsabelleMapping.ArrayType * nat) *
      ((nat * 'a list) FArray.IsabelleMapping.ArrayType * nat) *
      ('a, nat) hashmap * ('a, unit) hashmap * 'b;

fun impl_on_stack_update impl_on_stacka
  (Impl_state_ext
    (impl_stack, impl_s_stack, impl_i_stack, impl_on_stack, impl_done, more))
  = Impl_state_ext
      (impl_stack, impl_s_stack, impl_i_stack, impl_on_stacka impl_on_stack,
        impl_done, more);

fun impl_s_stack_update impl_s_stacka
  (Impl_state_ext
    (impl_stack, impl_s_stack, impl_i_stack, impl_on_stack, impl_done, more))
  = Impl_state_ext
      (impl_stack, impl_s_stacka impl_s_stack, impl_i_stack, impl_on_stack,
        impl_done, more);

fun impl_stack_update impl_stacka
  (Impl_state_ext
    (impl_stack, impl_s_stack, impl_i_stack, impl_on_stack, impl_done, more))
  = Impl_state_ext
      (impl_stacka impl_stack, impl_s_stack, impl_i_stack, impl_on_stack,
        impl_done, more);

fun as_length x = snd x;

fun map2set_insert i k s = i k () s;

fun impl_done_update impl_donea
  (Impl_state_ext
    (impl_stack, impl_s_stack, impl_i_stack, impl_on_stack, impl_done, more))
  = Impl_state_ext
      (impl_stack, impl_s_stack, impl_i_stack, impl_on_stack,
        impl_donea impl_done, more);

fun array_shrink a = FArray.IsabelleMapping.array_shrink a o integer_of_nat;

fun as_shrink s =
  let
    val (a, n) = s;
    val aa =
      (if less_eq_nat (times_nat (nat_of_integer (128 : IntInf.int)) n)
            (array_length a) andalso
            less_nat (nat_of_integer (4 : IntInf.int)) n
        then array_shrink a n else a);
  in
    (aa, n)
  end;

fun as_take m s =
  let
    val (a, n) = s;
  in
    (if less_nat m n then as_shrink (a, m) else (a, n))
  end;

fun as_top s = let
                 val (a, n) = s;
               in
                 array_get a (minus_nat n one_nata)
               end;

fun as_pop s = let
                 val (a, n) = s;
               in
                 as_shrink (a, minus_nat n one_nata)
               end;

fun as_get s i = let
                   val (a, _) = s;
                 in
                   array_get a i
                 end;

fun impl_on_stack
  (Impl_state_ext
    (impl_stack, impl_s_stack, impl_i_stack, impl_on_stack, impl_done, more))
  = impl_on_stack;

fun whilea b c s = (if b s then whilea b c (c s) else s);

fun impl_s_stack
  (Impl_state_ext
    (impl_stack, impl_s_stack, impl_i_stack, impl_on_stack, impl_done, more))
  = impl_s_stack;

fun impl_stack
  (Impl_state_ext
    (impl_stack, impl_s_stack, impl_i_stack, impl_on_stack, impl_done, more))
  = impl_stack;

fun impl_done
  (Impl_state_ext
    (impl_stack, impl_s_stack, impl_i_stack, impl_on_stack, impl_done, more))
  = impl_done;

fun code_pop (A1_, A2_) s =
  let
    val x = fst (as_top (impl_s_stack s));
    val xa = as_pop (impl_s_stack s);
    val xb = impl_on_stack s;
    val xc = impl_done s;
    val xd = as_length (impl_stack s);
    val (_, (aa, ba)) =
      whilea (fn (xg, (_, _)) => less_nat xg xd)
        (fn (a, (aa, ba)) =>
          let
            val xf = as_get (impl_stack s) a;
          in
            (plus_nat a one_nata,
              (ahm_delete (eq A1_) (bounded_hashcode A2_) xf aa,
                map2set_insert (ahm_update (eq A1_) (bounded_hashcode A2_)) xf
                  ba))
          end)
        (x, (xb, xc));
    val xf = as_take x (impl_stack s);
  in
    impl_done_update (fn _ => ba)
      (impl_on_stack_update (fn _ => aa)
        (impl_stack_update (fn _ => xf) (impl_s_stack_update (fn _ => xa) s)))
  end;

fun is_None a = (case a of NONE => true | SOME _ => false);

fun as_is_empty s = equal_nata (snd s) zero_nata;

fun impl_i_stack_update impl_i_stacka
  (Impl_state_ext
    (impl_stack, impl_s_stack, impl_i_stack, impl_on_stack, impl_done, more))
  = Impl_state_ext
      (impl_stack, impl_s_stack, impl_i_stacka impl_i_stack, impl_on_stack,
        impl_done, more);

fun impl_conc_seg_start A_ =
  (fn x => fn xa =>
    (if less_nat xa (as_length (impl_s_stack x))
      then fst (as_get (impl_s_stack x) xa) else as_length (impl_stack x)));

fun array_grow a = FArray.IsabelleMapping.array_grow a o integer_of_nat;

fun as_push s x =
  let
    val (a, n) = s;
    val aa =
      (if equal_nata n (array_length a)
        then array_grow a
               (max ord_nat (nat_of_integer (4 : IntInf.int))
                 (times_nat (nat_of_integer (2 : IntInf.int)) n))
               x
        else a);
    val ab = array_set aa n x;
  in
    (ab, plus_nat n one_nata)
  end;

fun impl_num_seg A_ = (fn x => as_length (impl_s_stack x));

fun impl_top_seg A_ = (fn x => minus_nat (impl_num_seg A_ x) (suc zero_nata));

fun impl_i_stack
  (Impl_state_ext
    (impl_stack, impl_s_stack, impl_i_stack, impl_on_stack, impl_done, more))
  = impl_i_stack;

fun code_top_out_edge A_ s =
  (if as_is_empty (impl_i_stack s) then (NONE, s)
    else let
           val (a, b) = as_top (impl_i_stack s);
         in
           (if less_eq_nat (impl_conc_seg_start A_ s (impl_top_seg A_ s)) a
             then let
                    val (aa, ba) =
                      (case b of [] => (the NONE, []) | aa :: ba => (aa, ba));
                    val ab =
                      impl_i_stack_update
                        (fn _ =>
                          (case ba of [] => as_pop (impl_i_stack s)
                            | _ :: _ =>
                              as_push (as_pop (impl_i_stack s)) (a, ba)))
                        s;
                  in
                    (SOME aa, ab)
                  end
             else (NONE, s))
         end);

fun ahm_lookup_aux eq bhc k a =
  list_map_lookup eq k (array_get a (bhc (array_length a) k));

fun ahm_lookup eq bhc k (HashMap (a, uu)) = ahm_lookup_aux eq bhc k a;

fun impl_is_on_stack (A1_, A2_) =
  (fn x => fn xa =>
    not (is_None
          (ahm_lookup (eq A1_) (bounded_hashcode A2_) xa (impl_on_stack x))));

fun as_singleton B_ x = (FArray.IsabelleMapping.array_of_list [x], one B_);

fun new_hashmap_with size = HashMap (new_array [] size, zero_nata);

fun ahm_empty def_size = new_hashmap_with def_size;

fun is_Nil a = (case a of [] => true | _ :: _ => false);

fun as_empty B_ uu = (FArray.IsabelleMapping.array_of_list [], zero B_);

fun code_init_state (A1_, A2_) v0_impl a_impl e_impl =
  let
    val x = e_impl v0_impl;
  in
    Impl_state_ext
      (as_singleton one_nat v0_impl,
        as_singleton one_nat (zero_nata, a_impl v0_impl),
        (if is_Nil x then as_empty zero_nat ()
          else as_singleton one_nat (zero_nata, x)),
        ahm_update (eq A1_) (bounded_hashcode A2_) v0_impl zero_nata
          (ahm_empty (def_hashmap_size A2_ Type)),
        ahm_empty (def_hashmap_size A2_ Type), ())
  end;

fun slg_restrict_right mem succ v u = filter (fn va => mem va v) (succ u);

fun as_set s i x = let
                     val (a, b) = s;
                   in
                     (array_set a i x, b)
                   end;

fun code_collapse (A1_, A2_) e_impl v s =
  let
    val x =
      the (ahm_lookup (eq A1_) (bounded_hashcode A2_) v (impl_on_stack s));
  in
    (case (if not (snd (as_top (impl_s_stack s)))
            then whilea
                   (fn (xd, xe) =>
                     is_None xd andalso
                       less_nat x
                         (impl_conc_seg_start A2_ xe (impl_top_seg A2_ xe)))
                   (fn (_, b) =>
                     let
                       val xb = minus_nat (as_length (impl_s_stack b)) one_nata;
                       val (aa, ba) = as_get (impl_s_stack b) xb;
                       val (ab, bb) =
                         as_get (impl_s_stack b) (minus_nat xb one_nata);
                       val xe =
                         as_set (as_pop (impl_s_stack b))
                           (minus_nat xb one_nata) (ab, ba orelse bb);
                       val xf = impl_s_stack_update (fn _ => xe) b;
                     in
                       ((if ba then SOME (as_get (impl_stack xf) aa)
                          else (if bb then SOME (as_get (impl_stack xf) ab)
                                 else NONE)),
                         xf)
                     end)
                   (NONE, s)
            else let
                   val xa =
                     as_get (impl_stack s)
                       (impl_conc_seg_start A2_ s (impl_top_seg A2_ s));
                 in
                   (SOME xa, s)
                 end)
      of (NONE, b) => (NONE, b)
      | (SOME xb, b) =>
        let
          val xc =
            slg_restrict_right
              (fn xa => fn m =>
                not (is_None (ahm_lookup (eq A1_) (bounded_hashcode A2_) xa m)))
              e_impl (impl_on_stack b);
        in
          (SOME (xb, (xc, xc)), b)
        end)
  end;

fun map2set_memb l k s = (case l k s of NONE => false | SOME _ => true);

fun impl_is_done (A1_, A2_) =
  (fn x => fn xa =>
    map2set_memb (ahm_lookup (eq A1_) (bounded_hashcode A2_)) xa (impl_done x));

fun code_push (A1_, A2_) a_impl e_impl v s =
  let
    val x = e_impl v;
    val xa = as_length (impl_stack s);
    val xb =
      impl_on_stack_update
        (fn _ =>
          ahm_update (eq A1_) (bounded_hashcode A2_) v xa (impl_on_stack s))
        (impl_stack_update (fn _ => as_push (impl_stack s) v)
          (impl_i_stack_update
            (fn _ =>
              (if is_Nil x then impl_i_stack s
                else as_push (impl_i_stack s) (xa, x)))
            (impl_s_stack_update
              (fn _ => as_push (impl_s_stack s) (xa, a_impl v)) s)));
  in
    xb
  end;

fun code_scc_aux (A1_, A2_) e_impl a_impl v0_impl =
  let
    val x = code_init_state (A1_, A2_) v0_impl a_impl e_impl;
    val (a, _) =
      whilea
        (fn (xb, xc) => not (as_is_empty (impl_stack xc)) andalso is_None xb)
        (fn (_, b) =>
          (case code_top_out_edge A2_ b
            of (NONE, ba) =>
              let
                val a = code_pop (A1_, A2_) ba;
              in
                (NONE, a)
              end
            | (SOME xc, ba) =>
              (if impl_is_on_stack (A1_, A2_) ba xc
                then code_collapse (A1_, A2_) e_impl xc ba
                else (if not (impl_is_done (A1_, A2_) ba xc)
                       then let
                              val a = code_push (A1_, A2_) a_impl e_impl xc ba;
                            in
                              (NONE, a)
                            end
                       else (NONE, ba)))))
        (NONE, x);
  in
    a
  end;

fun code_scc (A1_, A2_) e_impl a_impl v0_impl =
  let
    val x = code_scc_aux (A1_, A2_) e_impl a_impl v0_impl;
  in
    not (is_None x)
  end;

fun def_hashmap_size_nat x = (fn _ => nat_of_integer (16 : IntInf.int)) x;

fun bounded_hashcode_nat na n = mod_nat n na;

fun hashcode_nat n = n;

val hashable_nat =
  {hashcode = hashcode_nat, bounded_hashcode = bounded_hashcode_nat,
    def_hashmap_size = def_hashmap_size_nat}
  : nat hashable;

fun list_map_pick_remove [] = (raise Fail "undefined")
  | list_map_pick_remove (kv :: l) = (kv, l);

fun list_map_isEmpty x = null x;

fun wset_find_path_code (A1_, A2_) e u0 p =
  let
    val x =
      let
        val (a, b) =
          foldli (id u0) (fn _ => true)
            (fn x => fn (a, b) =>
              (map2set_insert (ahm_update (eq A1_) (bounded_hashcode A2_)) x a,
                list_map_update (eq A1_) x [] b))
            (ahm_empty (def_hashmap_size A2_ Type), []);
      in
        (NONE, (a, b))
      end;
    val (a, (_, _)) =
      whilea (fn (xb, (_, xd)) => is_None xb andalso not (list_map_isEmpty xd))
        (fn (_, (aa, ba)) =>
          let
            val ((ac, bc), bb) = list_map_pick_remove ba;
          in
            (if p ac then (SOME (rev bc, ac), (aa, bb))
              else let
                     val (ad, bd) =
                       foldli (id (rev (e ac))) (fn _ => true)
                         (fn xc => fn (ad, bd) =>
                           (if map2set_memb
                                 (ahm_lookup (eq A1_) (bounded_hashcode A2_)) xc
                                 ad
                             then (ad, bd)
                             else (map2set_insert
                                     (ahm_update (eq A1_)
                                       (bounded_hashcode A2_))
                                     xc ad,
                                    (xc, ac :: bc) :: bd)))
                         (aa, bb);
                   in
                     (NONE, (ad, bd))
                   end)
          end)
        x;
  in
    a
  end;

fun find_path1_code (A1_, A2_) e u p =
  (case wset_find_path_code (A1_, A2_) e (e u) p of NONE => NONE
    | SOME (a, b) => SOME (u :: a, b));

fun code_scc_ce (A1_, A2_) e_impl a_impl v0_impl =
  (case code_scc_aux (A1_, A2_) e_impl a_impl v0_impl of NONE => NONE
    | SOME (a, (_, _)) =>
      let
        val xb = wset_find_path_code (A1_, A2_) e_impl [v0_impl] (eq A1_ a);
        val xc = fst (the xb);
        val xd = find_path1_code (A1_, A2_) e_impl a (eq A1_ a);
        val xe = fst (the xd);
      in
        SOME (a, (xc, xe))
      end);

fun code_scc_nat x = code_scc (equal_nat, hashable_nat) x;

fun array_get_oo x a = FArray.IsabelleMapping.array_get_oo x a o integer_of_nat;

fun array_set_oo f a = FArray.IsabelleMapping.array_set_oo f a o integer_of_nat;

fun iam_alpha a i = array_get_oo NONE a i;

fun iam_empty x = (fn _ => FArray.IsabelleMapping.array_of_list []) x;

fun code_scc_ce_nat x = code_scc_ce (equal_nat, hashable_nat) x;

fun iam_increment l idx =
  max ord_nat (minus_nat (plus_nat idx one_nata) l)
    (plus_nat (times_nat (nat_of_integer (2 : IntInf.int)) l)
      (nat_of_integer (3 : IntInf.int)));

fun iam_update k v a =
  array_set_oo
    (fn _ =>
      let
        val l = array_length a;
      in
        array_set (array_grow a (iam_increment l k) NONE) k (SOME v)
      end)
    a k (SOME v);

fun glist_member eq x [] = false
  | glist_member eq x (y :: ys) = eq x y orelse glist_member eq x ys;

fun glist_insert eq x xs = (if glist_member eq x xs then xs else x :: xs);

fun acc_of_list_impl x =
  (fn xa => fold (map2set_insert iam_update) xa (iam_empty ())) x;

fun succ_of_list_impl x =
  (fn xa =>
    let
      val xaa =
        fold (fn (xaa, xb) => fn xc =>
               (case iam_alpha xc xaa of NONE => iam_update xaa [xb] xc
                 | SOME xd =>
                   iam_update xaa (glist_insert equal_nata xb xd) xc))
          xa (iam_empty ());
    in
      (fn xb => (case iam_alpha xaa xb of NONE => [] | SOME xc => xc))
    end)
    x;

fun acc_of_list_impl_fun a =
  let
    val aa = acc_of_list_impl a;
  in
    (fn x => not (is_None (iam_alpha aa x)))
  end;

end; (*struct SCC_Gabov*)
