header {* Components of (symmetric digraphs *}

theory Digraph_Components
imports
  Digraphs
(*  Vertex_Walks *)
  Edge_Walks
begin

text {*
  This theory contains lemmas about sccs of symmetric digraphs. This
  is not really he same as undirected graphs, so use with care.
*}

section {* Basic Definitions *}

definition graph :: "('a,'b) pre_graph \<Rightarrow> bool" where
  "graph G \<equiv> digraph G \<and> symmetric G \<and> (\<forall>u. (u,u) \<notin> edges_ends G)"

definition compatible :: "('a,'b) pre_graph \<Rightarrow> ('a,'b) pre_graph \<Rightarrow> bool" where
  "compatible G H \<equiv>
      (\<forall>e \<in> edges G \<inter> edges H. edge_to_ends G e = edge_to_ends H e)"

(* XXX -- wohlgeformtheit erhalten? *)
definition subgraph :: "('a,'b) pre_graph \<Rightarrow> ('a,'b) pre_graph \<Rightarrow> bool" where
  "subgraph H G \<equiv> verts H \<subseteq> verts G \<and> edges H \<subseteq> edges G \<and> compatible H G \<and> wellformed_graph H"

definition induced_subgraph :: "('a,'b) pre_graph \<Rightarrow> ('a,'b) pre_graph \<Rightarrow> bool" where
  "induced_subgraph H G \<equiv> subgraph H G \<and> edges H = {e \<in> edges G. start G e \<in> verts H \<and> target G e \<in> verts H}"

definition spanning :: "('a,'b) pre_graph \<Rightarrow> ('a,'b) pre_graph \<Rightarrow> bool" where
  "spanning H G \<equiv> subgraph H G \<and> verts G = verts H"

definition strongly_connected :: "('a,'b) pre_graph \<Rightarrow> bool" where
  "strongly_connected G \<equiv> \<forall>u \<in> verts G. \<forall>v \<in> verts G. \<exists>p. pre_graph.ewalk G u p v"


text {* The underlying symmetric graph *}
definition mk_symmetric :: "('a,'b) pre_graph \<Rightarrow> ('a, 'a \<times> 'a) pre_graph" where
  "mk_symmetric G \<equiv> \<lparr> verts = verts G, edges = edges_ends G \<union> {(a,b). (b,a) \<in> edges_ends G},
    start = fst, target = snd\<rparr>"

definition connected :: "('a,'b) pre_graph \<Rightarrow> bool" where
  "connected G \<equiv> strongly_connected (mk_symmetric G)"

definition forest :: "('a,'b) pre_graph \<Rightarrow> bool" where
  "forest G \<equiv> \<not>(\<exists>p. cycle p G)"

definition tree :: "('a,'b) pre_graph \<Rightarrow> bool" where
  "tree G \<equiv> connected G \<and> forest G"

definition spanning_tree :: "('a,'b) pre_graph \<Rightarrow> ('a,'b) pre_graph \<Rightarrow> bool" where
  "spanning_tree H G \<equiv> tree H \<and> spanning H G"

definition scc :: "('a,'b) pre_graph \<Rightarrow> ('a,'b) pre_graph \<Rightarrow> bool" where
  "scc H G \<equiv> induced_subgraph H G \<and> strongly_connected H
    \<and> \<not>(\<exists>H'. induced_subgraph H' G
      \<and> strongly_connected H' \<and> verts H \<subset> verts H')"

definition union :: "('a,'b) pre_graph \<Rightarrow> ('a,'b) pre_graph \<Rightarrow> ('a,'b) pre_graph set" where
  "union G H \<equiv> {G'. verts G' = verts G \<union> verts H
    \<and> edges G' = edges G \<union> edges H
    \<and> compatible G' G \<and> compatible G' H}"

definition Union :: "('a,'b) pre_graph set \<Rightarrow> ('a,'b) pre_graph set" where
  "Union gs = {G. verts G = (\<Union>G' \<in> gs. verts G')
    \<and> edges G = (\<Union>G' \<in> gs. edges G')
    \<and> (\<forall>G' \<in> gs. compatible G G')}"

definition simple_graph :: "('a,'b) pre_graph \<Rightarrow> ('a,'b) pre_graph"
  where "simple_graph G \<equiv> \<lparr> verts = verts G, edges = edges G,
    start = restrict (start G) (edges G),
    target = restrict (target G) (edges G)\<rparr>"


section {* Basic lemmas *}

lemma graph_symmetric:
  assumes "graph G"
  shows "(u,v) \<in> edges_ends G \<Longrightarrow> (v,u) \<in> edges_ends G"
  using assms by (auto simp add: graph_def symmetric_def
    edges_ends_def edge_to_ends_def image_def)

lemma graphI[intro]:
  assumes "digraph G"
  assumes "symmetric G"
  assumes "\<And>u. (u,u) \<notin> edges_ends G"
  shows "graph G"
using assms by (auto simp add: graph_def)

lemma graphE[elim]:
  assumes "graph G"
  assumes "\<lbrakk>digraph G; symmetric G; \<And>u. (u,u) \<notin> edges_ends G\<rbrakk> \<Longrightarrow> P"
  shows "P"
using assms by (auto simp add: graph_def)

lemma strongly_connectedI[intro]:
  assumes "\<And>u v. u \<in> verts G \<Longrightarrow> v \<in> verts G \<Longrightarrow> \<exists>p. pre_graph.ewalk G u p v"
  shows "strongly_connected G"
using assms by (simp add: strongly_connected_def)

lemma strongly_connectedE[elim]:
  assumes "strongly_connected G"
  assumes "(\<And>u v. u \<in> verts G \<and> v \<in> verts G \<Longrightarrow> \<exists>p. pre_graph.ewalk G u p v) \<Longrightarrow> P"
  shows "P"
using assms by (auto simp add: strongly_connected_def)

lemma subgraph_imp_subverts[intro]:
  assumes "subgraph H G"
  shows "verts H \<subseteq> verts G"
using assms by (simp add: subgraph_def)

lemma induced_imp_subgraph:
  assumes "induced_subgraph H G"
  shows "subgraph H G"
using assms by (simp add: induced_subgraph_def)

lemma scc_imp_induced:
  assumes "scc c G"
  shows "induced_subgraph c G"
using assms by (simp add: scc_def)

lemma spanning_tree_imp_tree[dest]:
  assumes "spanning_tree H G"
  shows "tree H"
using assms by (simp add: spanning_tree_def)

lemma tree_imp_connected[dest]:
  assumes "tree G"
  shows "connected G"
using assms by (simp add: tree_def)

lemma spanning_treeI[intro]:
  assumes "spanning H G"
  assumes "tree H"
  shows "spanning_tree H G"
using assms by (simp add: spanning_tree_def)

lemma spanning_treeE[elim]:
  assumes "spanning_tree H G"
  assumes "tree H \<and> spanning H G \<Longrightarrow> P"
  shows "P"
using assms by (simp add: spanning_tree_def)

lemma spanningE[elim]:
  assumes "spanning H G"
  assumes "subgraph H G \<and> verts G = verts H \<Longrightarrow> P"
  shows "P"
using assms by (simp add: spanning_def)

lemma sccI[intro]:
  assumes "induced_subgraph c G"
  assumes "strongly_connected c"
  assumes "\<not>(\<exists>c'. induced_subgraph c' G \<and> strongly_connected c' \<and>
    verts c \<subset> verts c')"
  shows "scc c G"
using assms by (auto simp add: scc_def)

lemma sccE[elim]:
  assumes "scc c G"
  assumes "induced_subgraph c G \<Longrightarrow> strongly_connected c \<Longrightarrow> \<not> (\<exists>d.
    induced_subgraph d G \<and> strongly_connected d \<and> verts c \<subset> verts d) \<Longrightarrow> P"
  shows "P"
using assms by (auto simp add: scc_def)

lemma subgraphI[intro]:
  assumes "verts H \<subseteq> verts G"
  assumes "edges H \<subseteq> edges G"
  assumes "compatible H G"
  assumes "wellformed_graph H"
  shows "subgraph H G"
using assms by (auto simp add: subgraph_def)

lemma subgraphE[elim]:
  assumes "subgraph H G"
  assumes "\<lbrakk>verts H \<subseteq> verts G; edges H \<subseteq> edges G;
    compatible H G; wellformed_graph H\<rbrakk> \<Longrightarrow> P"
  shows "P"
using assms by (simp add: subgraph_def)

lemma induced_subgraphI[intro]:
  assumes "subgraph H G"
  assumes "edges H = {e \<in> edges G. start G e \<in> verts H \<and> target G e \<in> verts H}"
  shows "induced_subgraph H G"
using assms by (auto simp add: induced_subgraph_def)

lemma induced_subgraphE[elim]:
  assumes "induced_subgraph H G"
  assumes "\<lbrakk>subgraph H G; edges H = {e \<in> edges G. start G e \<in> verts H \<and> target G e \<in> verts H}\<rbrakk> \<Longrightarrow> P"
  shows "P"
using assms by (auto simp add: induced_subgraph_def)

lemma graph_union_case[cases pred: verts]:
  assumes "H \<in> (union G G')"
  assumes "u \<in> verts H"
  obtains (in_l) "u \<in> verts G" | (in_r) "u \<in> verts G'"
using assms by (auto simp: union_def)

lemma verts_mk_symmetric[simp]: "verts (mk_symmetric G) = verts G"
  and edges_mk_symmetric:
    "edges (mk_symmetric G) = edges_ends G \<union> {(b,a). (a,b) \<in> edges_ends G}"
  and start_mk_symmetric[simp]: "start (mk_symmetric G) = fst"
  and target_mk_symmetric[simp]: "target (mk_symmetric G) = snd"
  by (auto simp add: mk_symmetric_def)

section {* Other lemmas *}

lemma edges_ends_mono:
  assumes "subgraph H G"
  shows "edges_ends H \<subseteq> edges_ends G"
proof -
  have "edges_ends H = edge_to_ends G ` edges H"
    using assms by (simp add: compatible_def subgraph_def image_def edges_ends_def edge_to_ends_def Int_absorb2)
  then show ?thesis
    using assms by (auto simp add: subgraph_def edges_ends_def)
qed

lemma vwalk_subgraph_closed[intro]:
  assumes "subgraph H G"
  assumes "vwalk p H"
  shows "vwalk p G"
proof
  show "set p \<subseteq> verts G" and "p \<noteq> []"
    using assms by (auto simp add: subgraph_def vwalk_def)

  have "set (vwalk_edges p) \<subseteq> edges_ends H"
    using assms by (simp add: vwalk_def)
  also have "\<dots> \<subseteq> edges_ends G"
    using assms(1) by (rule edges_ends_mono)
  finally show "set (vwalk_edges p) \<subseteq> edges_ends G" .
qed

lemma subgraph_path_imp_path:
  assumes "path p G"
  assumes "subgraph G H"
  shows "path p H"
using assms by auto

lemma (in pseudo_digraph) edge_path:
  assumes "graph G"
  assumes "(a,b) \<in> edges_ends G"
  shows "path [a,b] G"
using assms wellformed
by (intro pathI vwalkI, auto)

lemma symp_compatible: "symp compatible"
  by (rule sympI) (auto simp: compatible_def)


subsection {* The underlying symmetric graph *}

lemma (in pseudo_digraph) pseudo_digraph_mk_symmetric[intro]: "pseudo_digraph (mk_symmetric G)"
proof -
  have "finite ((\<lambda>(a,b). (b,a))`edge_to_ends G ` edges G)" (is "finite ?X") by auto
  also have "?X = {(a, b). (b, a) \<in> edge_to_ends G ` edges G}" by auto
  finally have X: "finite ..." .
  show ?thesis
    by default
       (auto simp: mk_symmetric_def edges_ends_def edge_to_ends_def intro: wellformed X)
qed

lemma (in digraph) digraph_mk_symmetric[intro]: "digraph (mk_symmetric G)"
proof -
  have "finite ((\<lambda>(a,b). (b,a))`edge_to_ends G ` edges G)" (is "finite ?X") by auto
  also have "?X = {(a, b). (b, a) \<in> edge_to_ends G ` edges G}" by auto
  finally have X: "finite ..." .
  show ?thesis
    by default
       (auto simp: mk_symmetric_def edges_ends_def edge_to_ends_def no_loops intro: wellformed X)
qed

lemma (in pseudo_digraph) reachable_mk_symmetric:
  assumes "u \<rightarrow>*G v" shows "u \<rightarrow>*(mk_symmetric G) v"
proof -
  interpret S: pseudo_digraph "mk_symmetric G" ..
  from assms obtain p where walk: "ewalk u p v"
    by (auto simp: reachable_ewalk)
  with walk have "pre_graph.ewalk (mk_symmetric G) u (map (edge_to_ends G) p) v"
  by (induct arbitrary: u)
     (auto simp: S.ewalk_Cons_iff ewalk_Cons_iff ewalk_empty_iff S.ewalk_empty_iff
       edges_mk_symmetric edges_ends_def edge_to_ends_def intro: rev_image_eqI)
  then show ?thesis by (auto simp: S.reachable_ewalk)
qed

lemma (in digraph) mk_symmetric_ewalk_imp_ewalk:
  assumes sym: "symmetric G"
  assumes walk: "pre_graph.ewalk (mk_symmetric G) u p v"
  obtains q where "ewalk u q v"
proof
  let ?mG = "\<lparr>verts = verts G, edges = edges_ends G, start = fst, target = snd\<rparr>"

  from sym
  have "{(a, b). (b, a) \<in> edges_ends G} = edges_ends G"
    unfolding edges_ends_def symmetric_def
    by (auto simp: edge_to_ends_def image_def)
  then have ms_eq_mG: "mk_symmetric G \<equiv> ?mG"
    unfolding mk_symmetric_def by simp
  then interpret mG: digraph ?mG
    by (metis digraph_mk_symmetric)

  def to_G \<equiv> "\<lambda>ab. SOME e. e \<in> edges G \<and> edge_to_ends G e = ab"
  from walk[unfolded ms_eq_mG] show "ewalk u (map to_G p) v"
  proof (induct p arbitrary: u v)
    case Nil then show ?case by (auto simp: ewalk_empty_iff)
  next
    case (Cons e es)

    def w \<equiv> "target ?mG e"
    with Cons.prems have "e \<in> edges ?mG" "e = (u, w)"
      by (auto simp: mG.ewalk_Cons_iff)

    from `e \<in> edges ?mG`
    have "\<exists>ge. ge \<in> edges G \<and> edge_to_ends G ge = e"
      by (auto simp: edges_ends_def edge_to_ends_def)
    then have to_G_e1: "to_G e \<in> edges G \<and> edge_to_ends G (to_G e) = e"
      unfolding to_G_def by (rule exE_some)
    then have to_G_e2: "start G (to_G e) = u" "target G (to_G e) = target ?mG e"
      using `e = (u, w)` by (auto simp: edge_to_ends_def split: prod.split_asm)

    from Cons
    have "ewalk (target ?mG e) (map to_G es) v"
      by (auto simp: mG.ewalk_Cons_iff)
    then show "ewalk u (map to_G (e # es)) v"
      using to_G_e1 to_G_e2 by (auto simp: ewalk_Cons_iff)
  qed
qed

lemma symmetric_mk_symmetric:
  "symmetric (mk_symmetric G)"
proof -
  let ?msG = "mk_symmetric G"
  { fix e assume "e \<in> edges ?msG"
    then have "(snd e, fst e) \<in> edges ?msG"
      by (auto simp: mk_symmetric_def)
    then have "\<exists>e'\<in>edges ?msG. start ?msG e = target ?msG e' \<and> target ?msG e = start ?msG e'"
      by rule (auto simp: mk_symmetric_def) }
  then show ?thesis by (auto simp: symmetric_def)
qed

lemma (in wellformed_graph) wellformed_mk_symmetric[intro]:
  "wellformed_graph (mk_symmetric G)"
by unfold_locales (auto simp: mk_symmetric_def edges_ends_def edge_to_ends_def intro: wellformed)



section {* Compatible Graphs *}

lemma compatible_sym: "compatible G H = compatible H G"
  by (auto elim: sympE[OF symp_compatible])

lemma reflp_compatible: "reflp compatible"
  by (rule reflpI) (simp add: compatible_def)

lemma compatible_simple:
  assumes "start G = start H" "target G = target H"
  shows "compatible G H"
using assms by (auto simp: compatible_def edge_to_ends_def)

lemma compatE:
  assumes "compatible G H"
  assumes "e \<in> edges G" "e \<in> edges H"
  obtains "start G e = start H e" "target G e = target H e"
using assms by (auto simp: compatible_def edge_to_ends_def)

lemma compatE_rev:
  assumes "compatible G H"
  assumes "e \<in> edges G" "e \<in> edges H"
  obtains "start H e = start G e" "target H e = target G e"
using assms by (auto simp: compatible_def edge_to_ends_def)

lemma same_verts:
  assumes "set p \<subseteq> edges H" "set p \<subseteq> edges G" "compatible G H"
  shows "pre_graph.ewalk_verts G u p = pre_graph.ewalk_verts H u p"
using assms
by (cases p) (auto simp:  pre_graph.ewalk_verts_def elim!: compatE[OF assms(3)])

lemma compatible_subedges:
  assumes "compatible G H"
  assumes "edges G \<subseteq> edges H"
  assumes "e \<in> edges G"
  shows "edge_to_ends G e = edge_to_ends H e"
using assms unfolding compatible_def by auto

lemma compatible_inter_edge_to_ends:
  assumes "compatible G H"
  assumes "e \<in> edges G \<inter> edges H"
  shows "edge_to_ends G e \<in> edges_ends H"
        "edge_to_ends H e \<in> edges_ends G"
using assms
by (force simp: compatible_def edge_to_ends_def edges_ends_def
  image_def)+

locale compat_digraphs = G: digraph G + H: digraph H
    for G :: "('a,'b) pre_graph" and H :: "('a,'b) pre_graph" +
  assumes compat: "compatible G H"

lemma compat_digraphs_sym:
  "compat_digraphs G H = compat_digraphs H G"
by (auto simp: compat_digraphs_def compat_digraphs_axioms_def compatible_sym)

lemma (in compat_digraphs) same_verts:
  assumes "set p \<subseteq> edges H" "set p \<subseteq> edges G"
  shows "G.ewalk_verts u p = H.ewalk_verts u p"
using assms compat by (simp add: same_verts)

lemma (in digraph) compat_digraphsI:
  assumes "digraph H" "compatible G H"
  shows "compat_digraphs G H"
using assms
unfolding compat_digraphs_def compat_digraphs_axioms_def by simp default

lemma compatible_union_edge_to_ends:
  assumes comp: "compatible G H" "compatible G H'" "compatible H H'"
  assumes edge_sub: "edges G \<subseteq> edges H \<union> edges H'"
  shows "edges_ends G \<subseteq> edges_ends H \<union> edges_ends H'"
proof -
  { fix e assume "e \<in> edges G"
    then have "edge_to_ends G e \<in> edges_ends H \<union> edges_ends H'"
      using edge_sub comp by (blast intro: compatible_inter_edge_to_ends) }
  then show ?thesis by (fastforce simp only: edges_ends_def)
qed


section {* Subgraphs and Induced Subgraphs *}

text {*
  The @{term digraph} and @{term pseudo_digraph} properties are preserved under
  the (inverse) subgraph relation
*}
lemma (in pseudo_digraph) pseudo_digraph_subgraph:
  assumes "subgraph H G" shows "pseudo_digraph H"
proof (intro_locales)
  from assms show "wellformed_graph H" by auto

  have HG: "edges H \<subseteq> edges G" "verts H \<subseteq> verts G" "compatible H G"
    using assms by auto
  then have "finite (verts H)" "finite (edges H)"
    using finite_verts finite_edges by (blast intro: finite_subset)+
  then show "pseudo_digraph_axioms H"
    by unfold_locales
qed

(*XXX move *)
lemma (in digraph) digraph_subgraph:
  assumes "subgraph H G" shows "digraph H"
proof
  fix e assume e: "e \<in> edges H"
  with assms show "start H e \<in> verts H" "target H e \<in> verts H"
    apply (auto simp: subgraph_def intro: wellformed_graph.wellformed)
    done
  from e and assms have "e \<in> edges H \<inter> edges G" by auto
  with assms show "start H e \<noteq> target H e"
    using no_loops by (auto simp: subgraph_def compatible_def edge_to_ends_def)
next
  have "edges H \<subseteq> edges G" "verts H \<subseteq> verts G" using assms by auto
  then show "finite (edges H)" "finite (verts H)"
    using finite_verts finite_edges by (blast intro: finite_subset)+
next
  fix e1 e2 assume "e1 \<in> edges H" "e2 \<in> edges H"
    and eq: "edge_to_ends H e1 = edge_to_ends H e2"
  with assms have "e1 \<in> edges H \<inter> edges G" "e2 \<in> edges H \<inter> edges G"
    by auto
  with eq show "e1 = e2"
    using no_multi_edges assms
    by (auto simp: subgraph_def compatible_def edge_to_ends_def)
qed

text {*
  Edge walks and paths are preserved under the subgraph relation.
*}
lemma (in wellformed_graph) subgraph_ewalk_imp_ewalk:
  assumes walk: "pre_graph.ewalk H u p v"
  assumes sub: "subgraph H G"
  shows "ewalk u p v"
using assms
proof -
  interpret H: wellformed_graph H
    using sub by auto
  have *: "edges H \<subseteq> edges G" "verts H \<subseteq> verts G" "compatible H G"  using assms by auto
  then have A: "set p \<subseteq> edges H" "set p \<subseteq> edges G"
    using assms by (auto simp: pre_graph.ewalk_def)
  then have B: "pre_graph.ves H p = ves p"
  proof (induct p arbitrary: u v)
    case (Cons e es)
    then have "pre_graph.ves H es = ves es"
      by (auto simp: ewalk_Cons_iff)
    then show ?case
      using Cons by (cases es) (fastforce elim: compatE[OF *(3)])+
  qed simp

  from A have C: "ewalk_verts u p = H.ewalk_verts u p"
    using same_verts[OF _ _ *(3)] by auto
  then have D: "set (ewalk_verts u p) \<subseteq> verts G"
    using * assms(1) by auto

  from A B C D show ?thesis
    using assms(1) by (auto simp: ewalk_conv)
qed

lemma (in digraph) subgraph_epath_imp_epath:
  assumes path: "pre_graph.epath H u p v"
  assumes sub: "subgraph H G"
  shows "epath u p v"
proof -
  interpret H: digraph H
    using sub by (rule digraph_subgraph)
  from sub have "compatible G H" by (auto simp: compatible_sym)
  then show ?thesis
    using assms subgraph_ewalk_imp_ewalk
    by (fastforce simp: pre_graph.epath_def same_verts)
qed

lemma subgraph_ends:
  assumes "subgraph H G"
  assumes "e \<in> edges H"
  shows "edge_to_ends G e = edge_to_ends H e"
using assms by (elim subgraphE) (auto dest: compatible_subedges)

lemma subgraph_mk_symmetric:
  assumes "subgraph H G"
  shows "subgraph (mk_symmetric H) (mk_symmetric G)"
proof -
  from assms
  have V: "verts (mk_symmetric H) \<subseteq> verts (mk_symmetric G)"
    unfolding mk_symmetric_def by auto

  have C: "compatible (mk_symmetric H) (mk_symmetric G)"
    by (auto simp: compatible_simple mk_symmetric_def)

  have W: "wellformed_graph (mk_symmetric H)"
    using assms by (auto intro: wellformed_graph.wellformed_mk_symmetric)

  { fix e assume "e \<in> edges H"
    with assms have "edge_to_ends H e = edge_to_ends G e"
      by (auto simp: edge_to_ends_def subgraph_def compatible_def) }
  then have "edges_ends H = edge_to_ends G ` edges H"
    unfolding edges_ends_def by (auto simp: image_def)
  also have "\<dots> \<subseteq> edges_ends G"
    using assms by (auto simp: edges_ends_def)
  finally have "edges (mk_symmetric H) \<subseteq> edges (mk_symmetric G)"
    by (auto simp: mk_symmetric_def)
  then show ?thesis
    using V C W by (auto simp: subgraph_def)
qed

lemma subgraphs_are_compatible:
  assumes "subgraph g1 H"
  assumes "subgraph g2 H"
  shows "compatible g1 g2"
using assms by (metis IntE compatible_def subgraph_ends)


subsection {* induced subgraphs *}

lemma wellformed_induced:
  assumes "induced_subgraph H G"
  shows "wellformed_graph H"
proof -
  {
    fix e assume *: "e \<in> edges H"
    have "start G e \<in> verts H" "target G e \<in> verts H"
      using * and assms by auto
    moreover
    have "start G e = start H e" "target G e = target H e"
      using * and assms
      by (auto simp: induced_subgraph_def subgraph_def compatible_def edge_to_ends_def)
    ultimately
    have "start H e \<in> verts H" "target H e \<in> verts H"
      by auto
  }
  then show ?thesis
    by (auto simp: edges_ends_def edge_to_ends_def wellformed_graph_def)
qed

lemma (in digraph) digraph_induced:
  assumes "induced_subgraph H G"
  shows "digraph H"
proof
  from assms have "verts H \<subseteq> verts G" by blast
  then show "finite (verts H)" using finite_verts by (rule finite_subset)
next
  from assms have "edges H \<subseteq> edges G" by blast 
  then show "finite (edges H)" using finite_edges by (rule finite_subset)
next
  fix e assume "e \<in> edges H"
  moreover
  have "wellformed_graph H" using assms by (rule wellformed_induced)
  ultimately
  show "start H e \<in> verts H" "target H e \<in> verts H"
    using wellformed_graph.wellformed by fast+
  show "start H e \<noteq> target H e"
    using `e \<in> edges H` no_loops assms
    by (auto simp: induced_subgraph_def subgraph_def compatible_def edge_to_ends_def)
next
  fix e1 e2 assume "e1 \<in> edges H" "e2 \<in> edges H" "edge_to_ends H e1 = edge_to_ends H e2"
  then show "e1 = e2"
    using assms no_multi_edges
    by (auto simp: induced_subgraph_def subgraph_def compatible_def edge_to_ends_def)
qed

text {* Computes the subgraph of @{term G} induced by @{term vs} *}
definition induce_subgraph :: "('a,'b) pre_graph \<Rightarrow> 'a set \<Rightarrow> ('a,'b) pre_graph" (infix "\<restriction>" 67) where
  "G \<restriction> vs = \<lparr>
    verts = vs,
    edges = {e \<in> edges G. start G e \<in> vs \<and> target G e \<in> vs},
    start = start G, target = target G\<rparr>"

lemma induce_subgraph_verts[simp]:
 "verts (G \<restriction> vs) = vs"
by (auto simp add: induce_subgraph_def)

lemma induce_subgraph_edges[simp]:
 "edges (G \<restriction> vs) = {e \<in> edges G. start G e \<in> vs \<and> target G e \<in> vs}"
by (auto simp add: induce_subgraph_def)

lemma induce_subgraph_ends[simp]:
  "start (G \<restriction> vs) = start G"
  "target (G \<restriction> vs) = target G"
by (auto simp add: induce_subgraph_def)

lemma induced_induce[intro]:
  assumes "vs \<subseteq> verts G"
  shows "induced_subgraph (G \<restriction> vs) G"
using assms
by (intro subgraphI induced_subgraphI)
   (auto simp add: induce_subgraph_def compatible_def edge_to_ends_def)

lemma induced_graph_imp_symmetric:
  assumes "symmetric G"
  assumes "induced_subgraph H G"
  shows "symmetric H"
proof (unfold symmetric_def, safe)
  have "subgraph H G" using assms by auto
  note s_ends = edge_to_ends_decomp[OF subgraph_ends[OF this]]

  fix e1 assume "e1 \<in> edges H"
  then obtain e2 where "start G e1 = target G e2"
    "target G e1 = start G e2" "e2 \<in> edges G"
    using assms by (auto simp add: symmetric_def)
  moreover
  then have "e2 \<in> edges H"
    using assms and `e1 \<in> edges H` by auto
  ultimately
  show "\<exists>e2\<in>edges H.
            start H e1 = target H e2 \<and> target H e1 = start H e2"
    using assms `e1 \<in> edges H`
    by (auto simp add: s_ends)
qed

lemma (in digraph) induced_graph_imp_graph:
  assumes "graph G"
  assumes "induced_subgraph H G"
  shows "graph H"
proof
  from assms show "digraph H"
    by (blast intro: digraph_induced)
next
  fix u from assms show "(u,u) \<notin> edges_ends H"
    using edges_ends_mono[of H G] by auto
next
  show "symmetric H"
    using assms by (auto intro: induced_graph_imp_symmetric)
qed

lemma induce_reachable_preserves_paths:
  assumes "u \<rightarrow>*G v"
  shows "u \<rightarrow>*(G \<restriction> {w. u \<rightarrow>*G w}) v"
proof -
  from assms obtain p where "path p G" "hd p = u" "last p = v" by auto

  have "vwalk p (G \<restriction> {v. u \<rightarrow>*G v})"
  proof (rule vwalkI)
    from `path p G` show "p \<noteq> []" by auto

    have "set p \<subseteq> {v. u \<rightarrow>*G v}"
    proof (rule ccontr)
      assume l_assm: "\<not>?thesis"
      then obtain x where x_props: "x \<in> set p" "\<not> (u \<rightarrow>*G x)" by auto
      then obtain q xs ys
        where p_decomp: "p = xs @ x # ys" "q = xs @ [x]"
        using `path p G` by (auto simp add: in_set_conv_decomp)
      then have "vwalk (q @ ys) G" and "q \<noteq> []" using `path p G` by auto
      then have "vwalk q G"
        by (auto intro: vwalk_append1)
      moreover
      have "hd q = u" and "last q = x"
        using p_decomp and `hd p = u` by (auto simp: hd_append)
      ultimately
      show False using l_assm and x_props by auto
    qed
    then show "set p \<subseteq> verts (G \<restriction> {v. u \<rightarrow>*G v})" by simp

    show "set (vwalk_edges p) \<subseteq> edges_ends (G \<restriction> {v. u \<rightarrow>*G v})"
    proof -
      have "set (vwalk_edges p) \<subseteq> edges_ends G \<inter> (set p) \<times> (set p)"
        using `path p G` by (auto intro: vwalk_edges_verts)
      also have "\<dots> \<subseteq> edges_ends (G \<restriction> {v. u \<rightarrow>*G v})"
        using `set p \<subseteq> verts (G \<restriction> {v. u \<rightarrow>*G v})`
        by (auto simp: edges_ends_def edge_to_ends_def image_def)
      finally show ?thesis by assumption
    qed
  qed
  then show ?thesis using `hd p = u` and `last p = v` by auto
qed


section {* Unions of graphs *}

lemma union_imp_compatible:
  assumes "G \<in> union H H'"
  shows "compatible G H"
    and "compatible G H'"
    and "compatible H H'"
proof -
  show "compatible G H" "compatible G H'"
    using assms by (auto simp add: compatible_def union_def)

  {
    fix e assume e_prop: "e \<in> edges H" "e \<in> edges H'"
    have "edge_to_ends H e = edge_to_ends G e"
      using assms `e \<in> edges H` unfolding compatible_def union_def
      by auto
    also have "\<dots> = edge_to_ends H' e"
      using assms `e \<in> edges H'` unfolding compatible_def union_def
      by auto
    finally have "edge_to_ends H e = edge_to_ends H' e" .
  }
  then show "compatible H H'" by (auto simp: compatible_def)
qed

lemma wellformed_union:
  assumes wf: "wellformed_graph H" "wellformed_graph H'"
  assumes un: "G \<in> union H H'"
  shows "wellformed_graph G"
proof -
  have ends_props: "\<And>e. e \<in> edges H \<Longrightarrow> start G e = start H e \<and> target G e = target H e" 
    "\<And>e. e \<in> edges H' \<Longrightarrow> start G e = start H' e \<and> target G e = target H' e" 
    using un by (auto simp: union_def edge_to_ends_def image_def compatible_def)
  { fix e assume A: "e \<in> edges G" "e \<in> edges H \<or> e \<in> edges H'"
    have S: "start G e \<in> verts H \<union> verts H'" using A wf
      by (cases "e \<in> edges H") (auto simp: ends_props intro: wellformed_graph.wellformed)
    have T: "target G e \<in> verts H \<union> verts H'" using A wf
      by (cases "e \<in> edges H") (auto simp: ends_props intro: wellformed_graph.wellformed)
    note S T }
  then show ?thesis using un by (auto simp: union_def)
qed

lemma subgraph_union[intro]:
  assumes "subgraph h1 G"
  assumes "subgraph h2 G"
  assumes "G' \<in> union h1 h2"
  shows "subgraph G' G"
proof -
  from assms have "wellformed_graph G'"
    by (auto intro: wellformed_union)
  with assms show ?thesis
    by (auto simp add: subgraph_def union_def edge_to_ends_def compatible_def)
qed

(* XXX: make stronger? *)
lemma (in compat_digraphs) union_wellformed:
  assumes "x \<in> union G H" shows "wellformed_graph x"
proof -
  have "wellformed_graph G" "wellformed_graph H"
    by (intro_locales)
  then show ?thesis using assms by (rule wellformed_union)
qed

lemma (in compat_digraphs) union_pseudo_digraph:
  assumes "x \<in> union G H" shows "pseudo_digraph x"
proof intro_locales
  have "wellformed_graph G" "wellformed_graph H"
    by (intro_locales)
  then show "wellformed_graph x"
    using assms by (rule wellformed_union)

  show "pseudo_digraph_axioms x"
    using assms by unfold_locales (auto simp: union_def)
qed


section {* Connected and Strongly Connected Graphs*}

lemma (in digraph) strongly_connected_spanning_imp_strongly_connected:
  assumes "spanning H G"
  assumes "strongly_connected H"
  shows "strongly_connected G"
proof (unfold strongly_connected_def, (rule ballI)+)
  interpret H: digraph H
    using `spanning H G` by (auto intro: digraph_subgraph)
  fix u v
  assume "u \<in> verts G" and "v \<in> verts G"
  then have "u \<in> verts H" and "v \<in> verts H" "subgraph H G"
    using `spanning H G` by auto
  then obtain p where "H.ewalk u p v"
    using `strongly_connected H` by (auto simp add: strongly_connected_def)
  then have "ewalk u p v"
    using `subgraph H G` by (rule subgraph_ewalk_imp_ewalk)
  then show "\<exists>p. pre_graph.ewalk G u p v" ..
qed



lemma (in digraph) symmetric_connected_imp_strongly_connected:
  assumes "symmetric G" "connected G"
  shows "strongly_connected G"
proof -
  from `connected G`
  have "strongly_connected (mk_symmetric G)"
    unfolding connected_def by simp

  show "strongly_connected G"
  proof
    fix u v assume A: "u \<in> verts G" "v \<in> verts G"
    then obtain p where "pre_graph.ewalk (mk_symmetric G) u p v"
      using `strongly_connected (mk_symmetric G)`
      unfolding strongly_connected_def mk_symmetric_def by auto
    with `symmetric G` obtain q where "ewalk u q v"
      by (rule mk_symmetric_ewalk_imp_ewalk)
    then show "\<exists>p. ewalk u p v" ..
  qed
qed


lemma (in digraph) connected_spanning_imp_connected:
  assumes "spanning H G"
  assumes "connected H"
  shows "connected G"
proof (unfold connected_def strongly_connected_def, intro ballI)
  fix u v
  
  from pseudo_digraph_mk_symmetric have WF: "wellformed_graph (mk_symmetric G)" by blast

  assume "u \<in> verts (mk_symmetric G)" and "v \<in> verts (mk_symmetric G)"
  then have "u \<in> verts (mk_symmetric H)" and "v \<in> verts (mk_symmetric H)"
    using `spanning H G` by (auto simp: mk_symmetric_def)
  with `connected H`
  obtain p where "pre_graph.ewalk (mk_symmetric H) u p v"
    unfolding connected_def strongly_connected_def by auto
  then have "pre_graph.ewalk (mk_symmetric G) u p v"
  proof (rule wellformed_graph.subgraph_ewalk_imp_ewalk[OF WF])
    show  "subgraph (mk_symmetric H) (mk_symmetric G)"
      using `spanning H G`
      by (auto intro: subgraph_mk_symmetric)
  qed
  then show "\<exists>p. pre_graph.ewalk (mk_symmetric G) u p v " ..
qed

lemma (in digraph) spanning_tree_imp_connected:
  assumes "spanning_tree H G"
  shows "connected G"
using assms by (auto intro: connected_spanning_imp_connected)

(*XXX move*)
lemma (in wellformed_graph) symp_reachable:
  assumes "symmetric G"
  shows "symp (\<lambda>a b. reachable a G b)"
proof (rule sympI)
  fix u v assume "u \<rightarrow>*G v"
  then obtain p where "ewalk u p v"
    by (auto simp: reachable_ewalk)

  let "?rev_ep e e'" = "e' \<in> edges G \<and> start G e = target G e' \<and> target G e = start G e'"
  def rev_e \<equiv> "\<lambda>e. (SOME e'. ?rev_ep e e')"
  from assms have "\<And>e. e \<in> edges G \<Longrightarrow> \<exists>e'. ?rev_ep e e'"
    by (auto simp: symmetric_def)
  then have rev_ep: "\<And>e. e \<in> edges G \<Longrightarrow> ?rev_ep e (rev_e e)"
    unfolding rev_e_def by (rule exE_some)

  from `ewalk u p v` have "ewalk v (rev (map rev_e p)) u"
  proof (induct p arbitrary: u)
    case Nil then show ?case by (auto simp: ewalk_empty_iff)
  next
    case (Cons e es)
    then have e_props: "e \<in> edges G" "start G e = u"
      by (auto simp: ewalk_Cons_iff)

    from Cons have "ewalk v (rev (map rev_e es)) (target G e)"
      by (auto simp: ewalk_Cons_iff)
    moreover have "ewalk (target G e) [rev_e e] u"
      using e_props rev_ep[of e]
      by (auto simp add: ewalk_Cons_iff ewalk_empty_iff intro: wellformed)
    ultimately show "ewalk v (rev (map rev_e (e # es))) u"
      by (auto simp: ewalk_ewalk_joinI)
  qed
  then show "v \<rightarrow>*G u" by (rule reachable_ewalkI)
qed


lemma (in digraph) induce_reachable_is_scc:
  assumes "graph G"
  assumes "u \<in> verts G"
  shows "scc (G \<restriction> {v. u \<rightarrow>*G v}) G"
proof -
  let ?c = "(G \<restriction> {v. u \<rightarrow>*G v})"
  have isub_c: "induced_subgraph ?c G" by auto
  then interpret c: digraph ?c by (rule digraph_induced)

  have "symmetric G" using assms by auto
  then have sym_c: "symmetric (G \<restriction> {v. u \<rightarrow>*G v})"
    using isub_c by (rule induced_graph_imp_symmetric)

  note `induced_subgraph ?c G`
  moreover
  have "strongly_connected ?c"
  proof (rule strongly_connectedI)
    fix v w assume l_assms: "v \<in> verts ?c" "w \<in> verts ?c"
    have "u \<rightarrow>*(G \<restriction> {v. u \<rightarrow>*G v}) v"
      using l_assms by (auto intro: induce_reachable_preserves_paths)
    with c.symp_reachable[OF sym_c] 
    have "v \<rightarrow>*(G \<restriction> {v. u \<rightarrow>*G v}) u" by (rule sympE)
    also have "u \<rightarrow>*(G \<restriction> {v. u \<rightarrow>*G v}) w"
      using l_assms by (auto intro: induce_reachable_preserves_paths)
    finally show "\<exists>p. c.ewalk v p w"
      by (auto simp: c.reachable_ewalk)
  qed
  moreover
  have "\<not>(\<exists>d. induced_subgraph d G \<and> strongly_connected d \<and>
    verts ?c \<subset> verts d)"
  proof
    assume "\<exists>d. induced_subgraph d G \<and> strongly_connected d \<and>
      verts ?c \<subset> verts d"
    then obtain d where "induced_subgraph d G" "strongly_connected d"
      "verts ?c \<subset> verts d" by auto
    then obtain v where "v \<in> verts d" and "v \<notin> verts ?c"
      by auto

    have "u \<in> verts ?c" using `u \<in> verts G`
      unfolding reachable_def using path_self[of u G] by auto
    then have "u \<in> verts d" using `verts ?c \<subset> verts d` by auto 
    obtain q where "pre_graph.ewalk d u q v"
      using `strongly_connected d` `u \<in> verts d` `v \<in> verts d`
      by atomize_elim auto
    then have "ewalk u q v"
      using `induced_subgraph d G`
      by (auto intro: subgraph_ewalk_imp_ewalk)
    then have "v \<in> verts ?c" by (auto simp: reachable_ewalk)
    then show False using `v \<notin> verts ?c` by auto
  qed
  ultimately show ?thesis unfolding scc_def by auto   
qed


(*XXX better name?*)
lemma induce_reachable_contains:
  assumes "u \<in> verts G"
  shows "u \<in> verts (G \<restriction> {v. u \<rightarrow>*G v})"
using assms path_self[of u G] by (auto simp add: reachable_def)

lemma induced_eq_verts_imp_eq:
  assumes "induced_subgraph G H"
  assumes "induced_subgraph G' H"
  assumes "verts G = verts G'"
  shows "simple_graph G = simple_graph G'"
proof -
  from assms have props: "edges G = edges G'"
      "compatible G H" "compatible G' H"
    by (auto simp add: induced_subgraph_def)

  have edge_absorb: "edges G \<inter> edges H = edges G"
    "edges G' \<inter> edges H = edges G'"
    using assms by (blast elim: Int_absorb2)+
  have
    "restrict (start G) (edges G) = restrict (start G') (edges G')"
    "restrict (target G) (edges G) = restrict (target G') (edges G')"
    using props edge_absorb
    by (auto simp add: compatible_def restrict_def
      edge_to_ends_def)
  then show "simple_graph G = simple_graph G'"
    using `edges G = edges G'` `verts G = verts G'`
    by (auto simp: simple_graph_def)
qed

lemma scc_subset_imp_eq:
  assumes "scc c G"
  assumes "scc d G"
  assumes "verts c \<subseteq> verts d"
  shows "simple_graph c = simple_graph d"
using assms by (blast intro: induced_eq_verts_imp_eq)

lemma union_exI:
  fixes G H :: "('a,'b) pre_graph"
  assumes "compatible G H"
  shows "\<exists>x. x \<in> union G H"
proof -
  def s \<equiv> "\<lambda>e. if e \<in> edges G then start G e else start H e"
  def t \<equiv> "\<lambda>e. if e \<in> edges G then target G e else target H e"
  def x \<equiv> "\<lparr> verts = verts G \<union> verts H, edges = edges G \<union> edges H,
    start = s, target = t \<rparr>" 

  have "\<forall>e\<in>edges G. s e = start G e"
       "\<forall>e\<in>edges H. s e = start H e" 
       "\<forall>e\<in>edges G. t e = target G e"
       "\<forall>e\<in>edges H. t e = target H e"
    using assms unfolding compatible_def s_def t_def edge_to_ends_def
    by auto
  then have "compatible x G" "compatible x H"
    unfolding compatible_def edge_to_ends_def x_def by auto
  then have "x \<in> union G H"
    unfolding union_def x_def by auto
  then show ?thesis ..
qed

(*XXX move*)
lemma (in digraph) strongly_connected_imp_induce_subgraph_strongly_connected:
  assumes subg: "subgraph H G"
  assumes sc: "strongly_connected H"
  shows "strongly_connected (G \<restriction> (verts H))"
proof -
  let ?is_H = "G \<restriction> (verts H)"

  interpret H: digraph H
    using subg by (rule digraph_subgraph)
  interpret GrH: digraph "?is_H"
    by (metis induced_induce digraph_induced subg
      subgraph_imp_subverts)

  have "verts H \<subseteq> verts G" using assms by auto

  have "compatible G H"
    using assms by (auto simp: compatible_sym)
  have "verts H = verts ?is_H"
    by (metis induce_subgraph_verts)
  moreover
  have "edges H \<subseteq> edges ?is_H"
    using assms wellformed
      and subgraph_ends[OF `subgraph H G`, THEN edge_to_ends_decomp(1)]
      and subgraph_ends[OF `subgraph H G`, THEN edge_to_ends_decomp(2)]
    by (auto simp: edges_ends_def image_def edge_to_ends_def intro: H.wellformed)

  have "\<And>e. e \<in> edges H \<Longrightarrow> start G e = start H e"
    and "\<And>e. e \<in> edges H \<Longrightarrow> target G e = target H e"
    using subg by (auto elim!: compatE)
  then have "subgraph H (G \<restriction> verts H)"
    unfolding induce_subgraph_def using subg
    by (auto simp: subgraph_def compatible_def edge_to_ends_def intro: H.wellformed)
  then show ?thesis
    using induced_induce[OF `verts H \<subseteq> verts G`]
      and sc GrH.strongly_connected_spanning_imp_strongly_connected
    unfolding spanning_def by auto
qed

lemma subgraphs_of_union:
  assumes "wellformed_graph G" "wellformed_graph G'"
  assumes "H \<in> union G G'"
  shows "subgraph G H"
    and "subgraph G' H"
using assms by (auto simp: union_def compatible_sym)


section {* Components *}

theorem (in digraph) graph_is_union_sccs:
  assumes "graph G"
  shows "G \<in> Union {c. scc c G}"
proof -
  have "(\<Union>c \<in> {c. scc c G}. verts c) = verts G"
    using assms by (auto intro: induce_reachable_contains induce_reachable_is_scc)
  moreover
  have "(\<Union>c \<in> {c. scc c G}. edges c) = edges G"
  proof
    show "(\<Union>c \<in> {c. scc c G}. edges c) \<subseteq> edges G"
      by safe (metis sccE induced_imp_subgraph subgraphE subsetD)
    show "edges G \<subseteq> (\<Union>c \<in> {c. scc c G}. edges c)"
    proof (safe)
      fix e assume "e \<in> edges G"
      def a \<equiv> "start G e" and b \<equiv> "target G e"
      note a_def[simp] b_def[simp]

      have "e \<in> (\<Union>x \<in> {c. scc c G}. edges x)"
      proof cases
        assume "\<exists>x. scc x G \<and> {a,b } \<subseteq> verts x"
        then obtain c where "scc c G" and "{a,b} \<subseteq> verts c"
          by auto
        then have "e \<in> {e \<in> edges G. start G e \<in> verts c
          \<and> target G e \<in> verts c}" using `e \<in> edges G` by auto
        then have "e \<in> edges c" using `scc c G` by blast
        then show ?thesis using `scc c G` by auto
      next
        assume l_assm: "\<not>(\<exists>x. scc x G \<and> {a,b} \<subseteq> verts x)"

        have "(a,b) \<in> edges_ends G" and "a \<in> verts G"
          using `e \<in> edges G` wellformed
          by (fastforce simp: edges_ends_def edge_to_ends_def image_def)+
        have "\<exists>p. path p G \<and> hd p = a \<and> last p = b"
          apply (rule exI[where x="[a,b]"])
          using `(a,b) \<in> edges_ends G` and `graph G`
          by (auto intro: edge_path)
        then have "b \<in> verts (G \<restriction> {v. a \<rightarrow>*G v})" by auto
        moreover have "a \<in> verts (G \<restriction> {v. a \<rightarrow>*G v})"
          using `a \<in> verts G` by (intro induce_reachable_contains)
        ultimately
        have "{a,b} \<subseteq> verts (G \<restriction> {v. a \<rightarrow>*G v})" by auto
        moreover
        have "scc (G \<restriction> {v. a \<rightarrow>*G v}) G"
          using `a \<in> verts G` `graph G`
          by (auto intro: induce_reachable_is_scc)
        ultimately
        have False using l_assm by blast
        then show ?thesis by simp
      qed
      then show "e \<in> (\<Union>c \<in> {c. scc c G}. edges c)" by auto
    qed
  qed
  moreover have "\<And>H. scc H G \<Longrightarrow> compatible G H"
    by (auto simp: compatible_sym)
  ultimately show ?thesis
    by (auto simp add: Union_def intro: compatible_sym)
qed


(*FIXME: This lemma can be generalized to arbitrary graphs ... *)
lemma (in digraph) scc_for_vert_ex:
  assumes "graph G"
  assumes "u \<in> verts G"
  shows "\<exists>c. scc c G \<and> u \<in> verts c"
using assms
by (auto intro: induce_reachable_contains induce_reachable_is_scc)


lemma (in compat_digraphs) strongly_connected_non_disj:
  assumes "strongly_connected G"
  assumes "strongly_connected H"
  assumes "verts G \<inter> verts H \<noteq> {}"
  assumes "x \<in> union G H"
  shows "strongly_connected x"
proof (rule)
  fix u v w
  obtain w where w_in_both: "w \<in> verts G \<inter> verts H"
    using `verts G \<inter> verts H \<noteq> {}` by auto

  assume "u \<in> verts x" and "v \<in> verts x"

  interpret x: pseudo_digraph x
    using `x \<in> union G H` by (auto intro: union_pseudo_digraph)

  have subg: "subgraph G x" "subgraph H x"
    by (rule subgraphs_of_union[OF _ _ `x \<in> union G H`], intro_locales)+
  
  have reach_uw: "u \<rightarrow>*x w"
  using `x \<in> union G H` `u \<in> verts x` proof cases
    assume "u \<in> verts G"
    then obtain q where "G.ewalk u q w"
      using w_in_both and `strongly_connected G` by atomize_elim auto
    then show ?thesis using subg
      by (auto intro: x.subgraph_ewalk_imp_ewalk x.reachable_ewalkI)
  next
    assume "u \<in> verts H"
    then obtain q where "H.ewalk u q w"
      using w_in_both and `strongly_connected H` by atomize_elim auto
    then show ?thesis using subg
      by (auto intro: x.subgraph_ewalk_imp_ewalk x.reachable_ewalkI)
  qed
  also
  have reach_wv: "w \<rightarrow>*x v"
  using `x \<in> union G H` `v \<in> verts x` proof cases
    assume "v \<in> verts G"
    then obtain q where "G.ewalk w q v"
      using w_in_both and `strongly_connected G` by atomize_elim auto
    then show ?thesis using subg
      by (auto intro: x.subgraph_ewalk_imp_ewalk x.reachable_ewalkI)
  next
    assume "v \<in> verts H"
    then obtain q where "H.ewalk w q v"
      using w_in_both and `strongly_connected H` by atomize_elim auto
    then show ?thesis using subg
      by (auto intro: x.subgraph_ewalk_imp_ewalk x.reachable_ewalkI)
  qed
  finally show "\<exists>q. pre_graph.ewalk x u q v" by (simp add: x.reachable_ewalk)
qed

lemma (in digraph) scc_disj:
  assumes comp: "scc c G" "scc d G"
  assumes "simple_graph c \<noteq> simple_graph d"
  shows "verts c \<inter> verts d = {}"
proof (rule ccontr)
  assume contr: "\<not>?thesis"

  obtain x where "x \<in> union c d"
    using comp
    by atomize_elim (blast intro: subgraphs_are_compatible union_exI)

  have "digraph c" "digraph d"
    using comp by (metis digraph_induced scc_imp_induced)+
  moreover have "compatible c d"
    using comp by (metis `x \<in> union c d` union_imp_compatible(3))
  ultimately have compat_cd: "compat_digraphs c d"
    unfolding compat_digraphs_def by simp (default, simp)

  have "strongly_connected c" and "strongly_connected d"
    using comp by auto
  then have union_conn: "strongly_connected x"
    using contr and `x \<in> union c d`
    by (rule compat_digraphs.strongly_connected_non_disj[OF compat_cd])

  from comp have "induced_subgraph c G" "induced_subgraph d G"  by auto
  then have "wellformed_graph x" "subgraph x G"
    using `x \<in> union c d`
    by (blast intro: wellformed_induced wellformed_union)+
  with union_conn
  have induce_subgraph_conn: "strongly_connected (G \<restriction> verts x)"
      "induced_subgraph (G \<restriction> verts x) G"
    by (auto intro: strongly_connected_imp_induce_subgraph_strongly_connected)

  from assms have "\<not>verts c \<subseteq> verts d" and "\<not> verts d \<subseteq> verts c"
    by (metis scc_subset_imp_eq)+
  then have psub: "verts c \<subset> verts x"
    using `x \<in> union c d` by (auto simp: union_def)
  then show False using induce_subgraph_conn
    by (metis `scc c G` sccE induce_subgraph_verts)
qed

end
