theory Digraphs
imports
  Main
  Stuff
  "~~/src/HOL/Library/FuncSet"
begin

subsection {* Basic Definitions *}

type_synonym 'a ends = "'a \<times> 'a"

record ('a,'b) pre_graph =
  verts :: "'a set"
  edges :: "'b set"
  start :: "'b \<Rightarrow> 'a"
  target :: "'b \<Rightarrow> 'a"

definition edge_to_ends
    :: "('a,'b) pre_graph \<Rightarrow> 'b \<Rightarrow> 'a ends" where
  "edge_to_ends G e \<equiv> (start G e, target G e)"

definition edges_ends :: "('a,'b) pre_graph \<Rightarrow> 'a ends set" where
  "edges_ends G \<equiv> edge_to_ends G ` edges G"

locale pre_graph =
  fixes G :: "('a, 'b) pre_graph" (structure)

locale wellformed_graph = pre_graph +
  assumes start_in_verts: "e \<in> edges G \<Longrightarrow> start G e \<in> verts G"
  assumes target_in_verts: "e \<in> edges G \<Longrightarrow> target G e \<in> verts G"

lemmas (in wellformed_graph) wellformed = start_in_verts target_in_verts

text {* from Bang-Jensen, Gutin '00 *}
locale pseudo_digraph = wellformed_graph +
  assumes finite_verts[simp]: "finite (verts G)"
    and finite_edges[simp]: "finite (edges G)"

definition (in wellformed_graph) edge :: "'b \<Rightarrow> 'a \<times> 'a \<Rightarrow> bool" where
  "edge e uv \<equiv> e \<in> edges G \<and> start G e = fst uv \<and> target G e = snd uv"

text {* from Bang-Jensen, Gution '00 *}
locale digraph = pseudo_digraph +
  assumes no_loops: "e \<in> edges G \<Longrightarrow> start G e \<noteq> target G e"
  assumes no_multi_edges: "\<And>e1 e2. \<lbrakk>e1 \<in> edges G; e2 \<in> edges G;
     edge_to_ends G e1 = edge_to_ends G e2\<rbrakk> \<Longrightarrow> e1 = e2"

definition symmetric :: "('a,'b) pre_graph \<Rightarrow> bool" where
  "symmetric G \<equiv> \<forall>e1 \<in> edges G. \<exists>e2 \<in> edges G. start G e1 = target G e2 \<and> target G e1 = start G e2"


section {* Definition of Vertex Walks *}
(*XXX move? *)
type_synonym 'a vwalk = "'a list"

text {* Computes the list of edges belonging to a list of nodes *}
fun vwalk_edges :: "'a list \<Rightarrow> 'a ends list" where
    "vwalk_edges [] = []"
  | "vwalk_edges [x] = []"
  | "vwalk_edges (x#y#xs) = (x,y) # vwalk_edges (y#xs)"

definition vwalk_length :: "'a vwalk \<Rightarrow> nat" where
  "vwalk_length p \<equiv> length (vwalk_edges p)"

lemma vwalk_length_Nil[simp]:
  assumes "p = []"
  shows "vwalk_length p = 0"
using assms by (simp add: vwalk_length_def)

lemma vwalk_length_non_Nil[simp]:
  assumes "p \<noteq> []"
  shows "vwalk_length p = length p - 1"
using assms
by (induct p rule: vwalk_edges.induct) (auto simp: vwalk_length_def)

definition vwalk :: "'a vwalk \<Rightarrow> ('a,'b) pre_graph \<Rightarrow> bool" where
  "vwalk p G \<equiv> set p \<subseteq> verts G \<and> set (vwalk_edges p) \<subseteq> edges_ends G \<and> p \<noteq> []"

definition path :: "'a vwalk \<Rightarrow> ('a,'b) pre_graph \<Rightarrow> bool" where
  "path p G \<equiv> vwalk p G \<and> distinct p"

(*FIXME rename? remove component_of in lemma names ... *)
(*FIXME: change argument order? *)
definition reachable :: "'a \<Rightarrow> ('a,'b) pre_graph \<Rightarrow>
    'a \<Rightarrow> bool" ("_ \<rightarrow>*_ _" [100,100,100] 40) where
  "reachable u G v \<equiv> (\<exists>p. vwalk p G \<and> hd p = u \<and> last p = v)"

text {* For a given vwalk, compute a path with the same start and end *}
function vwalk_to_path :: "'a vwalk \<Rightarrow> 'a vwalk" where
    "vwalk_to_path [] = []"
  | "vwalk_to_path (x # xs) = (if (x \<in> set xs)
      then vwalk_to_path (dropWhile (\<lambda>y. y \<noteq> x) xs)
      else x # vwalk_to_path xs)"
by pat_completeness auto
termination by (lexicographic_order simp add: length_dropWhile_le)


(*XXX: Sane definition? Using the one from Volkmann at the moment *)
definition cycle :: "'a vwalk \<Rightarrow> ('a,'b) pre_graph \<Rightarrow> bool" where
  "cycle p G \<equiv> vwalk p G \<and> last p = hd p \<and> length p \<ge> 3 \<and> distinct (vwalk_edges p) \<and> distinct (tl p)"


section {* basic lemmas *}

lemma edge_to_ends_decomp:
  assumes "edge_to_ends G e1 = edge_to_ends H e2"
  shows "start G e1 = start H e2"
    and "target G e1 = target H e2"
using assms unfolding edge_to_ends_def by simp+

lemma edges_ends_symmetric:
  assumes "symmetric G"
  shows "(u,v) \<in> edges_ends G \<Longrightarrow> (v,u) \<in> edges_ends G"
  using assms by (auto simp add: symmetric_def
    edges_ends_def edge_to_ends_def image_def)


lemma wellformedI[intro]:
  assumes "\<And>e. e \<in> edges G \<Longrightarrow> start G e \<in> verts G"
  assumes "\<And>e. e \<in> edges G \<Longrightarrow> target G e \<in> verts G"
  shows "wellformed_graph G"
using assms by (auto simp: wellformed_graph_def)

(*XXX Use wellformed here?*)
lemma (in wellformed_graph) pseudo_digraphI[intro]:
  assumes "finite (verts G)"
  assumes "finite (edges G)"
  shows "pseudo_digraph G"
using assms wellformed by (auto simp: pseudo_digraph_def pseudo_digraph_axioms_def)

lemma edges_endsE[elim]:
  assumes "(u,v) \<in> edges_ends G"
  assumes "\<And>e. e \<in> edges G \<Longrightarrow> start G e = u \<Longrightarrow> target G e = v \<Longrightarrow> P"
  shows "P"
using assms by (auto simp add: edges_ends_def edge_to_ends_def)

lemma reachableI[intro]:
  assumes "vwalk p G"
  assumes "hd p = u"
  assumes "last p = v"
  shows "u \<rightarrow>*G v"
using assms by (auto intro: simp add: reachable_def)

end
