(*  Title:       Definitions about simulations and simulation preorders
                 on NFAs, without functions for computing them.
                 For code that computes the simulation preorder of an NFA,
                 see NFA_Simulations_INY.thy.
    Authors:     Manuel Eberl <eberlm@in.tum.de>
*)

theory NFA_Simulations
imports Main
  "../../Automata_Malik_Tuerk/NFA"
  "../Collections/Refine_Dflt_ICF"
begin

context NFA
begin


section {* Simulation relations *}


text {* One of the requirements to simulating states @{term "u"} and 
    @{term "v"}: for each successor @{term "u'"} of @{term "u"}, there is a 
    successor @{term "v'"} of @{term "v"} that simulates @{term "u'"}. *}
definition is_sim_pair where
"is_sim_pair \<S> u v \<equiv> 
    \<forall>u' c. (u,c,u')\<in>\<Delta> \<A> \<longrightarrow> (\<exists>v'. (u',v')\<in>\<S> \<and> (v,c,v')\<in>\<Delta> \<A>)"

text {* Definition of what a simulation relation of an NFA is. *}
definition is_sim where 
"is_sim \<S> \<equiv> \<S>\<subseteq>\<Q> \<A>\<times>\<Q> \<A> \<and> 
    (\<forall>u v. (u,v)\<in>\<S> \<longrightarrow> is_sim_pair \<S> u v) \<and>
    (\<forall>u v. (u,v)\<in>\<S> \<and> u\<in>\<F> \<A> \<longrightarrow> v\<in>\<F> \<A>)"

lemma is_simI[intro?]: assumes "\<S>\<subseteq>\<Q> \<A>\<times>\<Q> \<A>" and 
    "\<And>u u' v c. \<lbrakk>(u,v)\<in>\<S>; (u,c,u')\<in>\<Delta> \<A>\<rbrakk> \<Longrightarrow> 
         (\<exists>v'. (u',v')\<in>\<S> \<and> (v,c,v')\<in>\<Delta> \<A>)" and
    "\<And>u v. \<lbrakk>(u,v)\<in>\<S>; u\<in>\<F> \<A>\<rbrakk> \<Longrightarrow> v\<in>\<F> \<A>"
  shows "is_sim \<S>" 
  unfolding is_sim_def is_sim_pair_def by (auto simp add: assms)

lemma is_simI2: assumes "\<S>\<subseteq>\<Q> \<A>\<times>\<Q> \<A>" and 
    "\<And>u v. (u,v)\<in>\<S> \<Longrightarrow> is_sim_pair \<S> u v" and
    "\<And>u v. \<lbrakk>(u,v)\<in>\<S>; u\<in>\<F> \<A>\<rbrakk> \<Longrightarrow> v\<in>\<F> \<A>"
  shows "is_sim \<S>" unfolding is_sim_def by (auto simp add: assms)

lemma is_simD[dest]: assumes "is_sim \<S>"
  shows "\<S>\<subseteq>\<Q> \<A>\<times>\<Q> \<A>" "\<And>u u' v c. \<lbrakk>(u,v)\<in>\<S>; (u,c,u')\<in>\<Delta> \<A>\<rbrakk> \<Longrightarrow> 
        (\<exists>v'. (u',v')\<in>\<S> \<and> (v,c,v')\<in>\<Delta> \<A>)" and 
        "\<And>u v. \<lbrakk>(u,v)\<in>\<S>; u\<in>\<F> \<A>\<rbrakk> \<Longrightarrow> v\<in>\<F> \<A>"
 using assms unfolding is_sim_def is_sim_pair_def by blast+

lemma sim_subset_QQ[simp,intro]: "is_sim \<S> \<Longrightarrow> \<S> \<subseteq> \<Q> \<A> \<times> \<Q> \<A>"
    unfolding is_sim_def by simp

text {* simulation relations are always finite *}
lemma sim_finite: "is_sim \<S> \<Longrightarrow> finite \<S>" unfolding is_sim_def 
    using finite_\<Q> rev_finite_subset[of "\<Q> \<A>\<times>\<Q> \<A>"] by blast


subsection {* Closure properties *}

text {* The union of two simulations is a simulation. *}
lemma is_sim_union[intro]: "\<lbrakk>is_sim \<S>\<^sub>1; is_sim \<S>\<^sub>2\<rbrakk> \<Longrightarrow> is_sim (\<S>\<^sub>1\<union>\<S>\<^sub>2)"
  by (intro is_simI, blast+)


text {* The reflexie closure on @{term "\<Q> \<A>"} of a simulation is a 
    simulation. *}
lemma sim_refl_closure_is_sim[simp,intro]: assumes \<S>_sim: "is_sim \<S>"
    shows "is_sim (\<S>\<^sup>= \<inter> \<Q> \<A>\<times>\<Q> \<A>)"
proof (rule is_simI, simp)
  fix u u' v c assume "(u,v) \<in> \<S>\<^sup>= \<inter> \<Q> \<A>\<times>\<Q> \<A>" and "(u,c,u')\<in>\<Delta> \<A>"
  thus "\<exists>v'. (u',v')\<in>\<S>\<^sup>= \<inter> \<Q> \<A>\<times>\<Q> \<A> \<and> (v,c,v')\<in>\<Delta> \<A>" 
      using \<Delta>_consistent \<S>_sim by (cases "(u,v) \<in> \<S>", blast+)
next
  fix u v assume "(u,v) \<in> \<S>\<^sup>= \<inter> \<Q> \<A>\<times>\<Q> \<A>" and "u \<in> \<F> \<A>"
  thus "v \<in> \<F> \<A>" using \<S>_sim by (cases "(u,v) \<in> \<S>", blast+)
qed


lemma sim_trancl_sound1:
  assumes \<S>_sim: "is_sim \<S>" shows "(u,v) \<in> \<S>\<^sup>+ \<Longrightarrow> u\<in>\<Q> \<A> \<and> v\<in>\<Q> \<A>"
  by (induction rule: trancl_induct, insert \<S>_sim, blast+)

lemma sim_trancl_sound2:
  assumes \<S>_sim: "is_sim \<S>"
  shows "\<lbrakk>(u,v) \<in> \<S>\<^sup>+; (u,c,u')\<in>\<Delta> \<A>\<rbrakk> \<Longrightarrow> \<exists>v'. (u',v') \<in> \<S>\<^sup>+ \<and> (v,c,v')\<in>\<Delta> \<A>"
proof (induction rule: trancl_induct)
  case (step v w)
    obtain v' where "(u',v') \<in> \<S>\<^sup>+" "(v,c,v')\<in>\<Delta> \<A>" 
        using step.IH `(u,c,u')\<in>\<Delta> \<A>` by blast
    moreover then obtain w' where "(v',w') \<in> \<S>" and "(w,c,w')\<in>\<Delta> \<A>" 
        using \<S>_sim step `(v, w) \<in> \<S>` by blast
    ultimately have "(u',w') \<in> \<S>\<^sup>+" using trancl_into_trancl by simp
    thus ?case using `(w,c,w')\<in>\<Delta> \<A>` by blast
qed (insert \<S>_sim, blast)

lemma sim_trancl_sound3:
  assumes \<S>_sim: "is_sim \<S>"
  shows "(u,v) \<in> \<S>\<^sup>+ \<Longrightarrow> u \<in> \<F> \<A> \<longrightarrow> v \<in> \<F> \<A>"
  by (induction rule: trancl_induct, insert \<S>_sim, blast+)


text {* Transitive closures of simulations are again simulations. *}
lemma sim_trans_closure_is_sim[simp, intro]: "is_sim \<S> \<Longrightarrow> is_sim (\<S>\<^sup>+)"
  by (rule, auto simp add: sim_trancl_sound1 sim_trancl_sound2 
      sim_trancl_sound3)




text {* Abbreviations inclusion/equality of left/right language of states. *}
(* TODO: Consider moving this to NFA.thy *)
abbreviation \<L>_right_subset ("(1_) \<subseteq>\<^sub>R (1_)" [51,51] 50)
where "u \<subseteq>\<^sub>R v \<equiv> \<L>_right \<A> u \<subseteq> \<L>_right \<A> v"
abbreviation \<L>_right_eq ("(1_) =\<^sub>R (1_)" [51,51] 50)
where "u =\<^sub>R v \<equiv> \<L>_right \<A> u = \<L>_right \<A> v"
abbreviation \<L>_left_subset ("(1_) \<subseteq>\<^sub>L (1_)" [51,51] 50)
where "u \<subseteq>\<^sub>L v \<equiv> \<L>_left \<A> u \<subseteq> \<L>_left \<A> v"
abbreviation \<L>_left_eq ("(1_) =\<^sub>L (1_)" [51,51] 50)
where "u =\<^sub>L v \<equiv> \<L>_left \<A> u = \<L>_left \<A> v"

text {* simulation implies (right) language inclusion *}
lemma sim_imp_\<L>_right_subset:
  assumes sim: "is_sim \<S>" and uv_in_\<S>: "(u,v)\<in>\<S>" shows "u \<subseteq>\<^sub>R v"
proof
  fix x assume "x\<in>\<L>_right \<A> u" thus "x\<in>\<L>_right \<A> v" using assms
  proof (induction x arbitrary: u v)
    case Nil 
      hence "u\<in>\<F> \<A>" using in_\<L>_in_state_Nil by simp
      hence "v\<in>\<F> \<A>" using Nil.prems(2,3) unfolding is_sim_def by blast
      thus ?case using in_\<L>_in_state_Nil by simp
  next
    case (Cons c x)
      obtain u' where "c\<in>\<Sigma> \<A>" and
          u'_properties: "(u,c,u')\<in>\<Delta> \<A> \<and> x\<in>\<L>_right \<A> u'" 
          using in_\<L>_in_state_Cons Cons.prems(1) \<Delta>_consistent by fastforce
      moreover then obtain v' where v'_properties: "(u',v')\<in>\<S> \<and> (v,c,v')\<in>\<Delta> \<A>" 
          using Cons.prems(2,3) by blast
      ultimately show ?case using `c\<in>\<Sigma> \<A>` Cons.IH `is_sim \<S>`
          by (subst in_\<L>_in_state_Cons, blast)
    qed
qed



section {* Simulation preorders *}


text {* definition of what the largest simulation relation of an NFA is *}
definition is_largest_sim where
"is_largest_sim \<S> \<equiv> is_sim \<S> \<and> (\<forall>\<S>'. is_sim \<S>' \<longrightarrow> \<S>'\<subseteq>\<S>)"

lemma is_largest_simI[intro]: assumes "is_sim \<S>" and "\<And>\<S>'. is_sim \<S>' \<Longrightarrow> \<S>'\<subseteq>\<S>"
    shows "is_largest_sim \<S>" using assms unfolding is_largest_sim_def by blast

lemma is_largest_simD[dest]: 
    assumes "is_largest_sim \<S>"
    shows "is_sim \<S>" and "\<And>\<S>'. is_sim \<S>' \<Longrightarrow> \<S>'\<subseteq>\<S>" 
    using assms unfolding is_largest_sim_def by blast+

lemma is_largest_simD2[dest]: 
    assumes "is_largest_sim \<S>"
    shows "\<S>\<subseteq>\<Q> \<A>\<times>\<Q> \<A>" "\<And>u u' v c. \<lbrakk>(u,v)\<in>\<S>; (u,c,u')\<in>\<Delta> \<A>\<rbrakk> \<Longrightarrow> 
        (\<exists>v'. (u',v')\<in>\<S> \<and> (v,c,v')\<in>\<Delta> \<A>)" 
       "\<And>u v. \<lbrakk>(u,v)\<in>\<S>; u\<in>\<F> \<A>\<rbrakk> \<Longrightarrow> v\<in>\<F> \<A>" and "\<And>\<S>'. is_sim \<S>' \<Longrightarrow> \<S>'\<subseteq>\<S>" 
    using assms  by blast+

lemma is_largest_sim_unique:
  assumes "is_largest_sim \<S>" and "is_largest_sim \<S>'" 
  shows "\<S> = \<S>'" using assms unfolding is_largest_sim_def 
    by (intro equalityI, simp_all)



text {* We say that a state u "can be simulated" by a state @{term "v"} 
    iff there is a simulation that contains @{term "(u,v)"}. This is the 
    simulation preorder. *}
definition (in -) simulated_by ("(1_)/ \<le>\<^bsub>(_)\<^esub>/ (1_)" [51,1,51] 50) where 
    "u \<le>\<^bsub>\<A>\<^esub> v \<equiv> (\<exists>\<S>. NFA.is_sim \<A> \<S> \<and> (u,v)\<in>\<S>)"

text {* Two states are in the largest simulation relation iff there 
    is some simulation relation that contains them. *}
lemma in_a_largest_sim_iff_simulated: 
  "is_largest_sim \<S> \<Longrightarrow> (u,v)\<in>\<S> = (u \<le>\<^bsub>\<A>\<^esub> v)"
  unfolding simulated_by_def is_largest_sim_def by blast 

text {* Constructive definition of the largest simulation relation,
    i.e. the simulation preorder. *}
definition \<S>\<^sub>\<A> where "\<S>\<^sub>\<A> = \<Union>{\<S>. is_sim \<S>}"
lemmas largest_sim_def = \<S>\<^sub>\<A>_def

text {* @{term "\<S>\<^sub>\<A>"} is the largest simulation relation. *}
lemma \<S>\<^sub>\<A>_is_largest_sim[simp,intro]:
  "is_largest_sim \<S>\<^sub>\<A>" unfolding largest_sim_def
by (rule is_largest_simI, rule is_simI, blast+)

text {* @{term "\<S>\<^sub>\<A>"} is the simulation preorder. *}
lemma in_\<S>\<^sub>\<A>_iff_simulated: 
  "(u,v)\<in>\<S>\<^sub>\<A> = (u \<le>\<^bsub>\<A>\<^esub> v)" using in_a_largest_sim_iff_simulated by simp

lemma \<S>\<^sub>\<A>_is_sim[simp,intro]: "is_sim \<S>\<^sub>\<A>" using is_largest_simD(1) by simp

lemma \<S>\<^sub>\<A>_contains_sim[simp,intro]: "is_sim \<S> \<Longrightarrow> \<S> \<subseteq> \<S>\<^sub>\<A>"
    using is_largest_simD(2) by simp

lemma is_largest_sim_imp_eq_largest_sim[intro]:
  "is_largest_sim \<S> \<Longrightarrow> \<S> = \<S>\<^sub>\<A>" using is_largest_sim_unique by simp

lemma simulated_by_imp_\<L>_right_subset[simp]:
  "u \<le>\<^bsub>\<A>\<^esub> v \<Longrightarrow> \<L>_in_state \<A> u \<subseteq> \<L>_in_state \<A> v"
  by (rule sim_imp_\<L>_right_subset[of "\<S>\<^sub>\<A>"], 
      simp_all add: in_a_largest_sim_iff_simulated)


text {* Transitivity and reflexivity of the simulation preorder *}
lemma \<S>\<^sub>\<A>_sim_is_own_refl_closure[simp,intro]: 
    "is_largest_sim \<S> \<Longrightarrow> (\<S>\<^sup>= \<inter> \<Q> \<A>\<times>\<Q> \<A>) = \<S>"
    by (rule antisym, blast+)
lemma \<S>\<^sub>\<A>_is_own_trans_closure[simp,intro]: "is_largest_sim \<S> \<Longrightarrow> \<S>\<^sup>+ = \<S>"
  by (rule antisym, blast+)


text {* The simulation preorder is reflexive. *}
lemma is_largest_sim_refl[intro]: 
    "\<lbrakk>is_largest_sim \<S>; u\<in>\<Q> \<A>\<rbrakk> \<Longrightarrow> (u,u)\<in>\<S>" by blast

lemma \<S>\<^sub>\<A>_refl[intro]: "\<lbrakk>u\<in>\<Q> \<A>\<rbrakk> \<Longrightarrow> (u,u)\<in>\<S>\<^sub>\<A>" 
    using is_largest_sim_refl by simp

lemma simulated_by_refl[simp]: "u \<in> \<Q> \<A> \<Longrightarrow> u \<le>\<^bsub>\<A> \<^esub>u"
  using in_\<S>\<^sub>\<A>_iff_simulated by blast

text {* The simulation preorder is transitive. *}
lemma is_largest_sim_trans[trans]: 
  assumes "is_largest_sim \<S>" and "(u,v) \<in> \<S>" and "(v,w) \<in> \<S>"  
  shows "(u,w) \<in> \<S>"
proof-
  have "(u,w) \<in> \<S>\<^sup>+" using assms(2,3) by simp
  thus "(u,w) \<in> \<S>" using assms(1) by simp
qed

lemma \<S>\<^sub>\<A>_trans[trans]: "\<lbrakk>(u,v) \<in> \<S>\<^sub>\<A>; (v,w) \<in> \<S>\<^sub>\<A>\<rbrakk> \<Longrightarrow> (u,w) \<in> \<S>\<^sub>\<A>"
  by (rule is_largest_sim_trans, simp_all)

lemma simulated_by_trans[trans]: "\<lbrakk>u \<le>\<^bsub>\<A>\<^esub> v; v \<le>\<^bsub>\<A>\<^esub> w\<rbrakk> \<Longrightarrow> u \<le>\<^bsub>\<A>\<^esub> w"
  using \<S>\<^sub>\<A>_trans in_\<S>\<^sub>\<A>_iff_simulated  by blast


lemma sim_pair_in_\<S>\<^sub>\<A>:
  assumes "is_sim \<S>" and "is_sim_pair \<S> u v" and 
      "u \<in> \<Q> \<A>" and "v \<in> \<Q> \<A>" "u \<in> \<F> \<A> \<Longrightarrow> v \<in> \<F> \<A>" 
  shows "(u,v) \<in> \<S>\<^sub>\<A>"
proof-
  let ?\<S>' = "\<S> \<union> {(u,v)}"
  have "is_sim ?\<S>'" using assms 
    apply (intro is_simI2, blast) unfolding is_sim_pair_def apply blast+ done
  hence "?\<S>' \<subseteq> \<S>\<^sub>\<A>" using \<S>\<^sub>\<A>_contains_sim by blast
  thus "(u,v) \<in> \<S>\<^sub>\<A>" by simp
qed

lemma simrel_refl_on: "refl_on (\<Q> \<A>) \<S>\<^sub>\<A>" by (rule refl_onI, auto)
lemma simrel_trans: "trans \<S>\<^sub>\<A>" by (rule transI, rule \<S>\<^sub>\<A>_trans)
lemma simrel_preorder: "preorder_on (\<Q> \<A>) \<S>\<^sub>\<A>"
    unfolding preorder_on_def by (simp add: simrel_refl_on simrel_trans)



text {* Simulation on the reverse automaton *}
abbreviation (in -) sim_reverse ("(1_)/ \<le>\<^sup>r\<^bsub>(_)\<^esub>/ (1_)" [51,1,51] 50) where 
    "u \<le>\<^sup>r\<^bsub>\<A>\<^esub> v \<equiv> u \<le>\<^bsub>NFA_reverse \<A>\<^esub> v"


lemma sim_reverse_imp_\<L>_left_subset[simp]:
  assumes "u \<le>\<^sup>r\<^bsub>\<A>\<^esub> v" shows "u \<subseteq>\<^sub>L v"
proof
  fix w assume "w \<in> \<L>_left \<A> u"
  hence "rev w \<in> \<L>_right (NFA_reverse \<A>) u" 
      using NFA_reverse___\<L>_in_state[of \<A> u] by fast
  hence "rev w \<in> \<L>_right (NFA_reverse \<A>) v" 
      using NFA.simulated_by_imp_\<L>_right_subset[OF 
      NFA_reverse___is_well_formed[OF NFA_axioms] assms] by fast
  thus "w \<in> \<L>_left \<A> v" using NFA_reverse___\<L>_in_state[of \<A> v] by fast
qed




end

end
