(*  Title:       The algorithm by Ilie, Navarro and Yu for computation of 
                 simulation preorders in nondeterministic finite automata
    Authors:     Manuel Eberl <eberlm@in.tum.de>
*)
header {* \isaheader{The algorithm by Ilie, Navarro and Yu} *}
theory NFA_Simulations_INY
imports 
  Main 
  "Lib/Preorder_Equiv_Classes"
  NFA_Simulations 
  NFA_Refine
begin

text {*
  We verify the algorithm by Ilie, Navarro and Yu for computation of 
  simulation preorders in nondeterministic finite automata. We use the 
  Refinement Framework to produce efficiently executable code.
*}

context NFA
begin

subsection {* Preliminary definitions *}

text {* The complement relation of a relation @{term "\<S>"} over states of the 
    automaton @{term "\<A>"} *}
abbreviation compl where "compl \<S> \<equiv> \<Q> \<A> \<times> \<Q> \<A> - \<S>"

text {* The complement relation of a simulation contains all pairs @{term "(u,v)"}
    where @{text "u \<in> \<F> \<A>"} and @{text "v \<notin> \<F> \<A>"}, as final states can only be 
    simulated by other final states *}
lemma sim_compl_subset[intro]: 
  "is_sim \<S> \<Longrightarrow> \<F> \<A> \<times> (\<Q> \<A> - \<F> \<A>) \<subseteq> compl \<S>"
  using \<F>_consistent unfolding is_sim_def by blast

text {* the complement function is its own inverse function *}
lemma compl_compl[simp,intro]: "x\<subseteq>\<Q> \<A>\<times>\<Q> \<A> \<Longrightarrow> compl (compl x) = x"
    using \<F>_consistent by blast

text {* Some lemmata about how set inequalities between two relations reverse when
   taking the complements of both sides *}
lemma compl_subseteq_reverse: "\<lbrakk>x\<subseteq>\<Q> \<A>\<times>\<Q> \<A>; y\<subseteq>\<Q> \<A>\<times>\<Q> \<A>; x \<subseteq> compl y\<rbrakk>
    \<Longrightarrow> y \<subseteq> compl x" using \<F>_consistent by blast
lemma compl_subseteq: "\<lbrakk>x\<subseteq>\<Q> \<A>\<times>\<Q> \<A>; y\<subseteq>\<Q> \<A>\<times>\<Q> \<A>; x \<subseteq> y\<rbrakk>
    \<Longrightarrow> compl y \<subseteq> compl x" using \<F>_consistent by blast
lemma compl_subset_reverse: "\<lbrakk>x\<subseteq>\<Q> \<A>\<times>\<Q> \<A>; y\<subseteq>\<Q> \<A>\<times>\<Q> \<A>; x \<subset> compl y\<rbrakk>
    \<Longrightarrow> y \<subset> compl x" using \<F>_consistent by blast
lemma compl_subset: "\<lbrakk>x\<subseteq>\<Q> \<A>\<times>\<Q> \<A>; y\<subseteq>\<Q> \<A>\<times>\<Q> \<A>; x \<subset> y\<rbrakk>
    \<Longrightarrow> compl y \<subset> compl x" using \<F>_consistent by blast



subsection {* Abstract algorithm *}

text {* The invariant of the WHILE loop in the algorithm 
   It consists of five parts:
   1. @{term "\<omega>"} may only relate states of the automaton
   2. @{term "\<omega>"} must relate all final states with all nonfinal states
   3. @{term "\<C>"} must be a subset of @{term "\<omega>"}
   4. @{term "\<omega>"} must not relate two states that are also related by a simulation
      (i.e. @{term "\<omega>"} must be disjoint from all simulation relations)
      This means that @{term "\<omega>"} never gets "too large"
   5. if two states @{term "u"} and @{term "v"} are not related by @{term "\<omega>"}, each 
      successor @{term "u'"} of @{term "u"} must have a matching successor 
      @{term "v'"} of @{term "v"} that is either not in @{term "\<omega>"} or 
      still in @{term "\<C>"}. 
      This means that after each loop iteration, @{term "\<omega>"} is "large enough" 
      w.r.t. the state pairs processed so far.
*}
definition INY_abstr1_invar where
"INY_abstr1_invar \<equiv> \<lambda>(\<omega>, \<C>). 
    \<omega> \<subseteq> \<Q> \<A> \<times> \<Q> \<A>  \<and>  \<F> \<A> \<times> (\<Q> \<A> - \<F> \<A>) \<subseteq> \<omega>  \<and>  \<C> \<subseteq> \<omega>  \<and> \<S>\<^sub>\<A> \<inter> \<omega> = {}  \<and> 
    (\<forall>u v u' c. (u,v)\<in>compl \<omega> \<and> (u,c,u')\<in>\<Delta> \<A> \<longrightarrow> 
                    (\<exists>v'. (v,c,v')\<in>\<Delta> \<A> \<and> (u',v')\<in>compl (\<omega>-\<C>)))"

lemma INY_abstr1_invarI[intro]:
  assumes "\<omega> \<subseteq> \<Q> \<A> \<times> \<Q> \<A>" and 
    "\<F> \<A> \<times> (\<Q> \<A> - \<F> \<A>) \<subseteq> \<omega>" and "\<C> \<subseteq> \<omega>" and "\<S>\<^sub>\<A> \<inter> \<omega> = {}" and 
    "\<And>u v u' c. \<lbrakk>(u,v)\<in>compl \<omega>; (u,c,u')\<in>\<Delta> \<A>\<rbrakk> \<Longrightarrow> 
        \<exists>v'. (v,c,v')\<in>\<Delta> \<A> \<and> (u',v')\<in>compl (\<omega>-\<C>)"
  shows "INY_abstr1_invar (\<omega>, \<C>)" unfolding INY_abstr1_invar_def
apply (clarify, intro conjI) 
using assms apply (blast, blast, blast, blast) apply (blast intro: assms(5))
done

lemma INY_abstr1_invarD:
  assumes "INY_abstr1_invar (\<omega>, \<C>)"
  shows "\<omega> \<subseteq> \<Q> \<A> \<times> \<Q> \<A>" and "\<F> \<A> \<times> (\<Q> \<A> - \<F> \<A>) \<subseteq> \<omega>" and "\<C> \<subseteq> \<omega>" and "\<S>\<^sub>\<A> \<inter> \<omega> = {}"
      "\<And>u v u' c. \<lbrakk>(u,v)\<in>compl \<omega>; (u,c,u')\<in>\<Delta> \<A>\<rbrakk> \<Longrightarrow> 
                    \<exists>v'. (v,c,v')\<in>\<Delta> \<A> \<and> (u',v')\<in>compl (\<omega>-\<C>)" 
  using assms unfolding INY_abstr1_invar_def by blast+

lemma INY_abstr1_invar_emptyD:
  assumes "INY_abstr1_invar (\<omega>, {})"
  shows "\<omega> \<subseteq> \<Q> \<A> \<times> \<Q> \<A>" and "\<F> \<A> \<times> (\<Q> \<A> - \<F> \<A>) \<subseteq> \<omega>" and "\<S>\<^sub>\<A> \<inter> \<omega> = {}" and
        "\<And>u v u' c. \<lbrakk>(u,v)\<in>compl \<omega>; (u,c,u')\<in>\<Delta> \<A>\<rbrakk> \<Longrightarrow> 
                        \<exists>v'. (v,c,v')\<in>\<Delta> \<A> \<and> (u',v')\<in>compl \<omega>"
apply clarify
using INY_abstr1_invarD[OF assms] apply (blast, blast, blast)
using INY_abstr1_invarD(5)[OF assms] apply blast
done

text {* The initial @{term "\<omega>"}. These are all pairs of states @{term "(u,v)"} of 
   which we know that @{term "v"} does not simulate @{term "u"} from the start. 
   This can be because one of the following reasons:
   - @{term "u"} is a final state and @{term "v"} is not
   - @{term "u"} has a successor w.r.t. a character @{term "c"} whereas @{term "v"} 
     does not.
   Note that the second case is not taken into account in the paper by
   Ilie et al. This is a mistake, as it causes some state pairs in non-total 
   NFAs to be erroneously marked as simulating. *}
definition INY_initial where
"INY_initial = \<F> \<A> \<times> (\<Q> \<A> - \<F> \<A>) \<union> 
    {(u,v). u\<in>\<Q> \<A> \<and> v\<in>\<Q> \<A> \<and> (\<exists>c u'. (u,c,u')\<in>\<Delta> \<A> \<and> \<not>(\<exists>v'. (v,c,v')\<in>\<Delta> \<A>))}"


text {* The while loop invariant holds initially *}
lemma INY_abstr1_invar_initial:
    "INY_abstr1_invar (INY_initial, INY_initial)"
apply (rule INY_abstr1_invarI)
using \<F>_consistent INY_initial_def apply (fast, simp, simp)
unfolding INY_initial_def using \<S>\<^sub>\<A>_is_largest_sim apply blast
using \<Delta>_consistent apply blast
done


text {* The new entries that have to be added to @{term "\<omega>"} in one interation,
   i.e. the state pairs @{term "(u,v)"} that become unsimulatable if we now
   know that @{term "(u',v')"} is not simulatable. *}
definition INY_abstr1_set where
"INY_abstr1_set \<omega> \<C> u' v' = 
    {(u,v) |u v c. (u,v)\<in>compl \<omega> \<and> (u,c,u') \<in> \<Delta> \<A> \<and> (v,c,v') \<in> \<Delta> \<A> \<and> 
                   (\<forall>v''. (v,c,v'')\<in>\<Delta> \<A> \<longrightarrow> (u',v'')\<in>\<omega>-\<C>)}"

lemma INY_abstr1_set_subset_QQ[simp]:
    "INY_abstr1_set \<omega> \<C> u' v' \<subseteq> \<Q> \<A> \<times> \<Q> \<A>"
    unfolding INY_abstr1_set_def by blast

text{* if a pair @{term "(u,v)"} is in these entries, we know that it is not
   a simulating pair, but it is also not yet in @{term "\<omega>"}. *}
lemma INY_abstr1_set_memE:
  assumes "(u,v) \<in> INY_abstr1_set \<omega> \<C> u' v'" 
  obtains c where "(u,v)\<in>compl \<omega> \<and> (u,c,u')\<in>\<Delta> \<A> \<and> (v,c,v')\<in>\<Delta> \<A> \<and>
        (\<forall>v''. (v,c,v'')\<in>\<Delta> \<A> \<longrightarrow> (u',v'') \<in> \<omega>-\<C>)"
  using assms unfolding INY_abstr1_set_def by blast

text {* if a pair @{term "(u,v)"} is not in these entries, we know that it
   must fulfil the simulation criteria to the best of our current
   knowledge *}
lemma INY_abstr1_set_notmemE:
  assumes "(u, v) \<notin> INY_abstr1_set \<omega> \<C> u' v'" 
           "(u, v) \<in> compl \<omega>" "(u,c,u')\<in>\<Delta> \<A>" "(v,c,v')\<in>\<Delta> \<A>"
  obtains v'' where "(v,c,v'') \<in> \<Delta> \<A> \<and> (u',v'') \<in> compl (\<omega>-\<C>)"
  using assms \<Delta>_consistent unfolding INY_abstr1_set_def by blast

text {* None of the state pairs that are to be added in every step of the 
    while loop are in the simulation preorder.*}
lemma INY_abstr1_set_disjoint_\<S>\<^sub>\<A>:
  assumes I: "INY_abstr1_invar (\<omega>, \<C>)" and uv_in_\<C>: "(u',v')\<in>\<C>" 
  shows "\<S>\<^sub>\<A> \<inter> INY_abstr1_set \<omega> (\<C>-{(u',v')}) u' v' = {}" (is "\<S>\<^sub>\<A> \<inter> ?T = {}")
proof (intro equalityI subsetI, elim IntE, simp_all, clarify)
  note invar = INY_abstr1_invarD[OF I]
  note \<omega>_disjoint_\<S>\<^sub>\<A> = invar(4) 
  fix u v uv assume "(u,v) \<in> ?T" and "(u,v) \<in> \<S>\<^sub>\<A>"
  from INY_abstr1_set_memE[OF this(1)] guess c .
  moreover with `(u,v) \<in> \<S>\<^sub>\<A>` have "\<exists>v'. (u',v') \<in> \<S>\<^sub>\<A> \<and> (v,c,v')\<in>\<Delta> \<A>"
      using \<S>\<^sub>\<A>_is_largest_sim by blast
  ultimately show False using \<omega>_disjoint_\<S>\<^sub>\<A> invar(3) uv_in_\<C> by blast
qed


text {* The conditions that the new values for @{term "\<omega>"} and @{term "\<C>"} need
    to fulfil in order for the algorithm to work. *}
definition INY_abstr1_is_valid_\<omega>'\<C>' where
"INY_abstr1_is_valid_\<omega>'\<C>' \<omega> \<C> u' v' \<equiv> \<lambda>(\<omega>',\<C>'). \<omega>' \<subseteq> \<Q> \<A>\<times>\<Q> \<A> \<and>
    \<omega>'=\<omega>\<union>(\<omega>'-\<omega>) \<and> \<C>'=\<C>\<union>(\<omega>'-\<omega>) \<and> INY_abstr1_set \<omega> \<C> u' v' \<subseteq> \<omega>' \<and> 
    \<omega>' \<inter> \<S>\<^sub>\<A> = {}"

lemma INY_abstr1_is_valid_\<omega>'\<C>'I:
  assumes "\<omega>' \<subseteq> \<Q> \<A>\<times>\<Q> \<A>" and "\<omega> \<subseteq> \<omega>'" and "\<C>' = \<C> \<union> (\<omega>'-\<omega>)" and 
      "INY_abstr1_set \<omega> \<C> u' v' \<subseteq> \<omega>'" and " \<S>\<^sub>\<A> \<inter> \<omega>' = {}"
  shows "INY_abstr1_is_valid_\<omega>'\<C>' \<omega> \<C> u' v' (\<omega>',\<C>')"
  using assms unfolding INY_abstr1_is_valid_\<omega>'\<C>'_def by auto

lemma INY_abstr1_is_valid_\<omega>'\<C>'D:
  assumes "INY_abstr1_is_valid_\<omega>'\<C>' \<omega> \<C> u' v' (\<omega>',\<C>')"
  shows "\<omega>' \<subseteq> \<Q> \<A>\<times>\<Q> \<A>" and "\<omega> \<subseteq> \<omega>'" and "\<C>' = \<C> \<union> (\<omega>'-\<omega>)" and 
      "INY_abstr1_set \<omega> \<C> u' v' \<subseteq> \<omega>'" and "\<S>\<^sub>\<A> \<inter> \<omega>' = {}"
  using assms unfolding INY_abstr1_is_valid_\<omega>'\<C>'_def by auto

lemma INY_abstr1_is_valid_\<omega>'\<C>'D2:
  assumes "INY_abstr1_is_valid_\<omega>'\<C>' \<omega> (\<C>-{(u',v')}) u' v' (\<omega>',\<C>')" and
          I: "INY_abstr1_invar (\<omega>, \<C>)"
  shows "\<omega> \<subseteq> \<omega>'" "\<C>-{(u',v')} \<subseteq> \<C>'" "\<C>' = \<C> - {(u',v')} \<union> (\<omega>'-\<omega>)" 
        "\<omega>' \<subseteq> \<Q> \<A> \<times> \<Q> \<A>" "\<C>' \<subseteq> \<Q> \<A> \<times> \<Q> \<A>" "\<C>' \<subseteq> \<omega>'" 
        "INY_abstr1_set \<omega> (\<C>-{(u',v')}) u' v' \<subseteq> \<omega>'" "\<S>\<^sub>\<A> \<inter> \<omega>' = {}"
  using INY_abstr1_invarD[OF I] INY_abstr1_is_valid_\<omega>'\<C>'D[OF assms(1)] 
  by simp_all (blast+)
  
text {* The algorithm in its most abstract form: while there are unprocessed
   state pairs, we pick one pair at random and process it. The variables
   are as follows:
   - @{term "\<omega>"} is the relation that shall be the complement of @{text "\<le>\<^bsub>\<A>\<^esub>"} in the end
   - @{term "\<C>"} is the set of elements in @{term "\<omega>"} that have not yet been processed
   - @{term "u'"} and @{term "v'"} are the two states being processed in one 
     interation
*}
definition INY_abstr1 where
"INY_abstr1 \<equiv> WHILE\<^sub>T\<^bsup>INY_abstr1_invar\<^esup> (\<lambda>(\<omega>, \<C>). \<C>\<noteq>{}) (\<lambda>(\<omega>, \<C>). do {
    (u',v') \<leftarrow> SPEC (\<lambda>(u',v'). (u',v') \<in> \<C>);
    let \<C> = \<C> - {(u',v')};
    (\<omega>, \<C>) \<leftarrow> SPEC (INY_abstr1_is_valid_\<omega>'\<C>' \<omega> \<C> u' v');
    RETURN (\<omega>, \<C>)
}) (INY_initial, INY_initial)"


text {* the termination measure for the WHILE loop. The basic idea is that
   in each iteration, one pair @{term "(u',v')"} is processed, each pair is
   processed at most once and there is a finite number of pairs.
   @{term "\<omega>-\<C>"} is the set of pairs that are known to be non-simulating and
   of whose non-simulatability has been propagated as well.
*}
definition INY_abstr_measure::"((('q\<times>'q) set\<times>('q\<times>'q) set)) \<Rightarrow> nat"
where "INY_abstr_measure \<equiv> (\<lambda>(\<omega>,\<C>). card (compl (\<omega>-\<C>)))"

text {* The measure decreases in every iteration of the while loop. *}
lemma INY_measure_decreases:
  assumes "INY_abstr1_invar (\<omega>, \<C>)" and "(u',v')\<in>\<C>" 
      and "INY_abstr1_is_valid_\<omega>'\<C>' \<omega> (\<C>-{(u',v')}) u' v' (\<omega>',\<C>')"
  shows "INY_abstr_measure (\<omega>', \<C>') < INY_abstr_measure (\<omega>, \<C>)"
proof-
  let ?A = "compl (\<omega>-\<C>)" and ?A' = "compl (\<omega>'-\<C>')"
  note invar = INY_abstr1_invarD[OF assms(1)]
  note valid_\<omega>'\<C>' = INY_abstr1_is_valid_\<omega>'\<C>'D[OF assms(3)]  
  have "finite ?A" using finite_\<Q> by blast
  moreover have "?A' \<subset> ?A" using assms invar(1,3) valid_\<omega>'\<C>'(1-3) by blast
  ultimately show ?thesis unfolding INY_abstr_measure_def 
      by (simp add: psubset_card_mono)
qed


lemma INY_abstr1_invar5_preserved:
  assumes "INY_abstr1_invar (\<omega>, \<C>)" and
      "INY_abstr1_is_valid_\<omega>'\<C>' \<omega> (\<C>-{(u',v')}) u' v' (\<omega>',\<C>')" and 
      "(u',v')\<in>\<C>" and uv_notin_\<omega>': "(u,v) \<in> compl \<omega>'" and "(u,c,u'')\<in>\<Delta> \<A>"
  shows "\<exists>v''. (v,c,v'')\<in>\<Delta> \<A> \<and> (u'',v'') \<in> compl (\<omega>'-(\<C>'-{(u',v')}))"
proof-
  note invar = INY_abstr1_invarD[OF assms(1)]
  note valid_\<omega>'\<C>' = INY_abstr1_is_valid_\<omega>'\<C>'D2[OF assms(2,1)]
  let ?T = "\<omega>' - \<omega>" and ?T' = "INY_abstr1_set \<omega> (\<C>-{(u',v')}) u' v'"
  have uv_properties: "(u,v) \<notin> ?T'" "(u,v) \<in> compl \<omega>"
      using valid_\<omega>'\<C>'(1,7) uv_notin_\<omega>' by blast+
  then obtain v'' where 
      v''_properties: "(u'',v'') \<in> compl (\<omega> - \<C>) \<and> (v, c, v'')\<in>\<Delta> \<A>"
      using invar(5) `(u,c,u'')\<in>\<Delta> \<A>` `(u',v')\<in>\<C>` by blast
    show ?thesis

    -- {* case distinction: is the pair @{text "(u'',v'')"} we are looking at the *}
    -- {* same as the one we are processing at the moment *}
    proof (cases "u'=u'' \<and> v'=v''")
      case True
        -- {* yes, @{text "(u'',v'')=(u',v')"}, therefore, we cannot use *}
        -- {* @{text "(u'',v'')"} as it will no longer be in @{term "\<omega>-\<C>"} after *}
        -- {* this loop iteration. *}
        hence "(u,c,u')\<in>\<Delta> \<A>" and "(v,c,v')\<in>\<Delta> \<A>" 
            using `(u,c,u'')\<in>\<Delta> \<A>` v''_properties by blast+
        from INY_abstr1_set_notmemE[OF uv_properties this] guess v''' .
        -- {* this new v''' is definitely distinct from @{term "v'"}. *}
        moreover with invar(3) `(u',v')\<in>\<C>` have "v''' \<noteq> v'" by blast
        ultimately show ?thesis using True valid_\<omega>'\<C>'(3) by blast

    next
      case False
        -- {* no, @{text "(u'',v'') \<noteq> (u', v')"} therefore we can just use *}
        -- {* @{text "(u'',v'')"} itself. *}
        thus ?thesis using v''_properties valid_\<omega>'\<C>'(3) by blast
    qed
qed

text {* The while loop invariant is not violated by a loop iteration *}
lemma INY_abstr1_invar_preserved:
  assumes "INY_abstr1_invar (\<omega>, \<C>)" and "(u',v')\<in>\<C>" and
          "INY_abstr1_is_valid_\<omega>'\<C>' \<omega> (\<C>-{(u',v')}) u' v' (\<omega>',\<C>')"
  shows "INY_abstr1_invar (\<omega>', \<C>')"
apply (intro INY_abstr1_invarI)
using INY_abstr1_invarD(2)[OF assms(1)] INY_abstr1_is_valid_\<omega>'\<C>'D2[OF assms(3,1)]
apply (blast, fast, blast)
using INY_abstr1_is_valid_\<omega>'\<C>'D2(4,8)[OF assms(3,1)] apply blast
using INY_abstr1_invar5_preserved[OF assms(1,3,2)] apply blast
done


text {* If the invariant still holds and we have processed all elements 
   (i.e. @{term "\<C>"} is empty), @{term "\<omega>"} is now the complement of the
   simulation preorder. *}
lemma INY_abstr1_invar_imp_goal:
  assumes "INY_abstr1_invar (\<omega>, {})" shows "compl \<omega> = \<S>\<^sub>\<A>"
proof-
  note invar = INY_abstr1_invar_emptyD[OF assms(1)]
  have "is_largest_sim (compl \<omega>)"
    apply (intro is_largest_simI is_simI)
    using invar(2,4) apply (fast, fast, fast)
    apply (subgoal_tac "\<And>\<S>. is_sim \<S> \<Longrightarrow> \<S> \<inter> \<omega> = {}")
    using compl_subseteq_reverse[OF invar(1)] apply blast
    using invar(3) \<S>\<^sub>\<A>_is_largest_sim apply blast
    done
  thus "compl \<omega> = \<S>\<^sub>\<A>" using is_largest_sim_unique by simp
qed
   
 
text {* The abstract algorithm is correct. *}
theorem INY_abstr1_correct:
    shows "INY_abstr1 \<le> SPEC (\<lambda>(\<omega>,_). compl \<omega> = \<S>\<^sub>\<A>)"
    unfolding INY_abstr1_def
apply (intro refine_vcg)
apply (rule wf_measure[of INY_abstr_measure])
apply (fact INY_abstr1_invar_initial)
apply (simp_all add: INY_abstr1_invar_preserved INY_measure_decreases)[2]
apply (clarsimp simp add: INY_abstr1_invar_imp_goal)
done

text {* The set to be added to @{term "\<omega>"} in one iteration of the
    first for each loop. *}
definition INY_abstr2_set where
"INY_abstr2_set \<omega> \<C> u' v' c \<equiv> 
    {(u,v) |u v. (u,v)\<notin>\<omega> \<and> (u,c,u')\<in>\<Delta> \<A> \<and> (v,c,v') \<in> \<Delta> \<A> \<and> 
                 (\<forall>v''. (v,c,v'')\<in>\<Delta> \<A> \<longrightarrow> (u',v'') \<in>\<omega>-\<C>)}"

text {* The conditions the new values for @{term "\<omega>"} and @{term "\<C>"} need to
    fulfil in order for the first for each loop to work. *}
definition INY_abstr2_is_valid_\<omega>'\<C>' where
"INY_abstr2_is_valid_\<omega>'\<C>' \<omega> \<C> u' v' c \<equiv> \<lambda>(\<omega>',\<C>'). \<omega>'\<subseteq>\<Q> \<A>\<times>\<Q> \<A> \<and>
    \<omega>'=\<omega>\<union>(\<omega>'-\<omega>) \<and> \<C>'=\<C>\<union>(\<omega>'-\<omega>) \<and> INY_abstr2_set \<omega> \<C> u' v' c \<subseteq> \<omega>' \<and> \<S>\<^sub>\<A> \<inter> \<omega>' = {}"

lemma INY_abstr2_is_valid_\<omega>'\<C>'I:
  assumes "\<omega>'=\<omega>\<union>(\<omega>'-\<omega>)" "\<C>'=\<C>\<union>(\<omega>'-\<omega>)" "INY_abstr2_set \<omega> \<C> u' v' c \<subseteq> \<omega>'"
      "\<omega>'\<subseteq>\<Q> \<A>\<times>\<Q> \<A>" "\<S>\<^sub>\<A> \<inter> \<omega>' = {}"
  shows "INY_abstr2_is_valid_\<omega>'\<C>' \<omega> \<C> u' v' c (\<omega>',\<C>')"
  unfolding INY_abstr2_is_valid_\<omega>'\<C>'_def using assms by blast

lemma INY_abstr2_is_valid_\<omega>'\<C>'D:
  assumes "INY_abstr2_is_valid_\<omega>'\<C>' \<omega> \<C> u' v' c (\<omega>',\<C>')"
  shows "\<omega>'=\<omega>\<union>(\<omega>'-\<omega>)" "\<C>'=\<C>\<union>(\<omega>'-\<omega>)" "INY_abstr2_set \<omega> \<C> u' v' c \<subseteq> \<omega>'"
      "\<omega>' \<subseteq> \<Q> \<A>\<times>\<Q> \<A>" "\<S>\<^sub>\<A> \<inter> \<omega>' = {}"
  using assms unfolding INY_abstr2_is_valid_\<omega>'\<C>'_def by blast+

definition INY_abstr2_loopc_invar where
"INY_abstr2_loopc_invar \<omega> \<C> u' v' \<Sigma>' \<equiv> \<lambda>(\<omega>', \<C>').
  let T = \<omega>'-\<omega> in \<omega>' = \<omega> \<union> T \<and> \<C>' = \<C> \<union> T \<and> \<C> \<subseteq> \<omega> \<and> 
    \<omega>' \<subseteq> \<Q> \<A> \<times> \<Q> \<A> \<and> \<C>' \<subseteq> \<Q> \<A> \<times> \<Q> \<A> \<and>
    (\<Union>c\<in>(\<Sigma> \<A>-\<Sigma>'). INY_abstr2_set \<omega> \<C> u' v' c) \<subseteq> \<omega>' \<and>
    \<S>\<^sub>\<A> \<inter> \<omega>' = {} \<and> (u',v') \<in> \<omega> \<and> (u',v') \<notin> \<C>'"

lemma INY_abstr2_loopc_invarI[intro]: fixes \<omega> \<omega>' \<C> \<C>' \<Sigma>' u' v'
  assumes "\<omega>' = \<omega> \<union> (\<omega>'-\<omega>)" "\<C>' = \<C> \<union> (\<omega>'-\<omega>)" "\<C> \<subseteq> \<omega>"
      "\<omega>' \<subseteq> \<Q> \<A> \<times> \<Q> \<A>" "\<C>' \<subseteq> \<Q> \<A> \<times> \<Q> \<A>"
      "(\<Union>c\<in>\<Sigma> \<A>-\<Sigma>'. INY_abstr2_set \<omega> \<C> u' v' c) \<subseteq> \<omega>'" "\<S>\<^sub>\<A> \<inter> \<omega>' = {}" 
      "(u',v') \<in> \<omega>" "(u',v') \<notin> \<C>'"
  shows "INY_abstr2_loopc_invar \<omega> \<C> u' v' \<Sigma>' (\<omega>', \<C>')"
  unfolding INY_abstr2_loopc_invar_def using assms by (simp_all add: Let_def) 

lemma INY_abstr2_loopc_invarD[dest]: fixes \<omega> \<omega>' \<C> \<C>' \<Sigma>' u' v'
  assumes "INY_abstr2_loopc_invar \<omega> \<C> u' v' \<Sigma>' (\<omega>', \<C>')"
  shows "\<omega>' = \<omega> \<union> (\<omega>'-\<omega>)" "\<C>' = \<C> \<union> (\<omega>'-\<omega>)" "\<C> \<subseteq> \<omega>"
      "\<omega>' \<subseteq> \<Q> \<A> \<times> \<Q> \<A>" "\<C>' \<subseteq> \<Q> \<A> \<times> \<Q> \<A>"
      "(\<Union>c\<in>\<Sigma> \<A>-\<Sigma>'. INY_abstr2_set \<omega> \<C> u' v' c) \<subseteq> \<omega>'" 
      "\<S>\<^sub>\<A> \<inter> \<omega>' = {}" "(u',v') \<in> \<omega>" "(u',v') \<notin> \<C>'"
  using assms unfolding INY_abstr2_loopc_invar_def by (auto simp add: Let_def)


text {* The for each loop that iterates over all @{text "c \<in> \<Sigma>"}. *}
definition INY_abstr2_loopc where
"INY_abstr2_loopc \<omega> \<C> u' v' \<equiv> FOREACH\<^bsup>INY_abstr2_loopc_invar \<omega> \<C> u' v'\<^esup> 
(\<Sigma> \<A>) (\<lambda>c (\<omega>, \<C>). do {
    (\<omega>', \<C>') \<leftarrow> SPEC (INY_abstr2_is_valid_\<omega>'\<C>' \<omega> \<C> u' v' c);
    RETURN (\<omega>', \<C>')
}) (\<omega>, \<C>)"


text {* The first for each loop works correctly, i.e. it returns valid
    values for @{term "\<omega>"} and @{term "\<C>"}. *}
lemma INY_abstr2_loopc_correct: fixes \<omega> \<C> u' v'
  assumes I: "INY_abstr1_invar (\<omega>,\<C> \<union> {(u',v')})" and "(u',v')\<notin>\<C>"
  shows "INY_abstr2_loopc \<omega> \<C> u' v' \<le> SPEC (INY_abstr1_is_valid_\<omega>'\<C>' \<omega> \<C> u' v')"
unfolding INY_abstr2_loopc_def
proof (rule FOREACHi_rule)
  show "finite (\<Sigma> \<A>)" using finite_\<Sigma> .
next
  note invar_outer = INY_abstr1_invarD[OF I]
  show "INY_abstr2_loopc_invar \<omega> \<C> u' v' (\<Sigma> \<A>) (\<omega>, \<C>)"
      using invar_outer assms(2) by (intro INY_abstr2_loopc_invarI, blast+)
next
  case (goal3 c \<Sigma>')
  thus ?case
  proof (intro refine_vcg, clarify)
    case (goal1 \<omega>' \<C>' _  _ \<omega>'' \<C>'')
    note invar = INY_abstr2_loopc_invarD[OF goal1(3)]
    note valid = INY_abstr2_is_valid_\<omega>'\<C>'D[OF goal1(4)]
    let ?\<Sigma>'' = "\<Sigma>' - {c}"
    show "INY_abstr2_loopc_invar \<omega> \<C> u' v' ?\<Sigma>'' (\<omega>'', \<C>'')"
    proof (intro INY_abstr2_loopc_invarI)
      have "\<omega> \<union> INY_abstr2_set \<omega> \<C> u' v' c \<subseteq> \<omega>' \<union> INY_abstr2_set \<omega>' \<C>' u' v' c"
          unfolding INY_abstr2_set_def using invar(1-2) by blast
      moreover have "INY_abstr2_set \<omega> \<C> u' v' c \<inter> \<omega> = {}"
          unfolding INY_abstr2_set_def by blast
      ultimately show " (\<Union>c\<in>\<Sigma> \<A> - ?\<Sigma>''. INY_abstr2_set \<omega> \<C> u' v' c) \<subseteq>
          \<omega>''" using invar(6) valid(1,3) by blast
    next
      case goal5 thus ?case using invar(4,5) valid(1,2,4) by blast
    next
      from INY_abstr1_invarD(3)[OF I] show "(u',v') \<in> \<omega>" by simp
    next
      from invar valid(2) show "(u',v') \<notin> \<C>''" by blast
    qed (insert assms(2) invar(1-3) valid(1,2,4,5), auto)
  qed

next
  case (goal4 \<omega>'\<C>') thus ?case
    proof (cases \<omega>'\<C>', simp)
      fix \<omega>' \<C>' assume I: "INY_abstr2_loopc_invar \<omega> \<C> u' v' {} (\<omega>', \<C>')"
      note invar = INY_abstr2_loopc_invarD[OF I]
      have "(\<Union>c\<in>\<Sigma> \<A>. INY_abstr2_set \<omega> \<C> u' v' c) = INY_abstr1_set \<omega> \<C> u' v'" 
          unfolding INY_abstr2_set_def INY_abstr1_set_def
          using \<Delta>_consistent by blast
     thus "INY_abstr1_is_valid_\<omega>'\<C>' \<omega> \<C> u' v' (\<omega>', \<C>')"
         apply (intro INY_abstr1_is_valid_\<omega>'\<C>'I)
         using invar apply (blast, blast, blast, blast, simp)
         done
    qed
qed

lemma INY_abstr2_loopc_correct': 
  assumes "INY_abstr1_invar (\<omega>, \<C>)" and "(u',v') \<in> \<C>"
  shows   "INY_abstr2_loopc \<omega> (\<C> - {(u',v')}) u' v' \<le> 
               SPEC (INY_abstr1_is_valid_\<omega>'\<C>' \<omega> (\<C> - {(u',v')}) u' v')"
proof-
  from assms have "\<C> - {(u',v')} \<union> {(u',v')} = \<C>" by blast
  with INY_abstr2_loopc_correct assms show ?thesis by simp
qed


text {* The entire algorithm, now with a more concrete implementation
    of the computation of new @{term "\<omega>"} and @{term "\<C>"}. Since all steps
    from here on are analogous to this, they will not be commented any
    further. *}
definition INY_abstr2 where
"INY_abstr2 \<equiv> WHILE\<^sub>T\<^bsup>INY_abstr1_invar\<^esup> (\<lambda>(\<omega>, \<C>). \<C>\<noteq>{}) (\<lambda>(\<omega>, \<C>). do {
    (u',v') \<leftarrow> SPEC (\<lambda>(u', v'). (u',v') \<in> \<C>);
    let \<C> = \<C> - {(u',v')};
    (\<omega>, \<C>) \<leftarrow> (INY_abstr2_loopc \<omega> \<C> u' v');
    RETURN (\<omega>, \<C>)
}) (INY_initial, INY_initial)"

lemma INY_abstr2_correct: "INY_abstr2 \<le> \<Down>Id INY_abstr1"
  unfolding INY_abstr2_def INY_abstr1_def
  by (refine_rcg, simp_all add: INY_abstr2_loopc_correct')

definition INY_abstr3_set where
"INY_abstr3_set \<omega> \<C> u' v' c v \<equiv> 
    {(u,v) |u. (u,v)\<notin>\<omega> \<and> (u,c,u') \<in> \<Delta> \<A> \<and> (v,c,v') \<in> \<Delta> \<A> \<and> 
                 (\<forall>v''. (v,c,v'')\<in>\<Delta> \<A> \<longrightarrow> (u',v'')\<in>\<omega>-\<C>)}"

lemma INY_abstr3_set_simp:
    "(v,c,v')\<in>\<Delta> \<A> \<Longrightarrow> (if (\<forall>v''. (v,c,v'')\<in>\<Delta> \<A> \<longrightarrow> (u',v'')\<in>\<omega>-\<C>) then
    {(u,v) |u. (u,v)\<notin>\<omega> \<and> (u,c,u') \<in> \<Delta> \<A>} else {}) = 
    INY_abstr3_set \<omega> \<C> u' v' c v"
unfolding INY_abstr3_set_def by simp

definition INY_abstr3_is_valid_\<omega>'\<C>' where
"INY_abstr3_is_valid_\<omega>'\<C>' \<omega> \<C> u' v' c v \<equiv> \<lambda>(\<omega>',\<C>'). \<omega>' \<subseteq> \<Q> \<A>\<times>\<Q> \<A> \<and>
    \<omega>'=\<omega>\<union>(\<omega>'-\<omega>) \<and> \<C>'=\<C>\<union>(\<omega>'-\<omega>) \<and> INY_abstr3_set \<omega> \<C> u' v' c v \<subseteq> \<omega>' \<and> \<S>\<^sub>\<A> \<inter> \<omega>' = {}"

lemma INY_abstr3_is_valid_\<omega>'\<C>'I:
  assumes "\<omega>'=\<omega>\<union>(\<omega>'-\<omega>)" "\<C>'=\<C>\<union>(\<omega>'-\<omega>)" "INY_abstr3_set \<omega> \<C> u' v' c v \<subseteq> \<omega>'"
      "\<omega>' \<subseteq> \<Q> \<A>\<times>\<Q> \<A>" "\<S>\<^sub>\<A> \<inter> \<omega>' = {}"
  shows "INY_abstr3_is_valid_\<omega>'\<C>' \<omega> \<C> u' v' c v (\<omega>',\<C>')"
  unfolding INY_abstr3_is_valid_\<omega>'\<C>'_def using assms by blast

lemma INY_abstr3_is_valid_\<omega>'\<C>'D: 
  assumes "INY_abstr3_is_valid_\<omega>'\<C>' \<omega> \<C> u' v' c v (\<omega>',\<C>')"
  shows "\<omega>'=\<omega>\<union>(\<omega>'-\<omega>)" "\<C>'=\<C>\<union>(\<omega>'-\<omega>)" "INY_abstr3_set \<omega> \<C> u' v' c v \<subseteq> \<omega>'"
      "\<omega>' \<subseteq> \<Q> \<A>\<times>\<Q> \<A>" "\<S>\<^sub>\<A> \<inter> \<omega>' = {}"
  using assms unfolding INY_abstr3_is_valid_\<omega>'\<C>'_def by blast+

definition INY_abstr3_loopv_invar where
"INY_abstr3_loopv_invar \<omega> \<C> u' v' c V' \<equiv> \<lambda>(\<omega>', \<C>').
  let T = \<omega>'-\<omega> in \<omega>' = \<omega> \<union> T \<and> \<C>' = \<C> \<union> T \<and> \<C> \<subseteq> \<omega> \<and> 
    \<omega>' \<subseteq> \<Q> \<A> \<times> \<Q> \<A> \<and> \<C>' \<subseteq> \<Q> \<A>\<times>\<Q> \<A> \<and>
    (\<Union>v\<in>({v. (v,c,v')\<in>\<Delta> \<A>}-V'). INY_abstr3_set \<omega> \<C> u' v' c v) \<subseteq> \<omega>' \<and>
    \<S>\<^sub>\<A> \<inter> \<omega>' = {} \<and> (u',v') \<in> \<omega> \<and> (u',v') \<notin> \<C>'"

lemma INY_abstr3_loopv_invarI[intro]:
  assumes "\<omega>' = \<omega> \<union> (\<omega>'-\<omega>)" "\<C>' = \<C> \<union> (\<omega>'-\<omega>)" "\<C> \<subseteq> \<omega>" 
      "\<omega>'\<subseteq> \<Q> \<A> \<times> \<Q> \<A>" "\<C>' \<subseteq> \<Q> \<A>\<times>\<Q> \<A>"
      "(\<Union>v\<in>{v. (v,c,v')\<in>\<Delta> \<A>}-V'. INY_abstr3_set \<omega> \<C> u' v' c v) \<subseteq> \<omega>'"
      "\<S>\<^sub>\<A> \<inter> \<omega>' = {}" "(u',v') \<in> \<omega>" "(u',v') \<notin> \<C>'"
  shows "INY_abstr3_loopv_invar \<omega> \<C> u' v' c V' (\<omega>', \<C>')"
  using assms unfolding INY_abstr3_loopv_invar_def by (auto simp add: Let_def)

lemma INY_abstr3_loopv_invarD[dest]: fixes \<omega> \<omega>' \<C> \<C>' \<Sigma>' u' v' c
  assumes "INY_abstr3_loopv_invar \<omega> \<C> u' v' c V' (\<omega>', \<C>')"
  shows "\<omega>' = \<omega> \<union> (\<omega>'-\<omega>)" "\<C>' = \<C> \<union> (\<omega>'-\<omega>)" and "\<C> \<subseteq> \<omega>" 
      "\<omega>' \<subseteq> \<Q> \<A> \<times> \<Q> \<A>" and "\<C>' \<subseteq> \<Q> \<A>\<times>\<Q> \<A>" and    
      "(\<Union>v\<in>{v. (v,c,v')\<in>\<Delta> \<A>}-V'. INY_abstr3_set \<omega> \<C> u' v' c v) \<subseteq> \<omega>'" and
      "\<S>\<^sub>\<A> \<inter> \<omega>' = {}" "(u',v') \<in> \<omega>" "(u',v') \<notin> \<C>'"
  using assms unfolding INY_abstr3_loopv_invar_def by (auto simp add: Let_def)

definition INY_abstr3_loopv where
"INY_abstr3_loopv \<omega> \<C> u' v' c \<equiv> 
  FOREACH\<^bsup>INY_abstr3_loopv_invar \<omega> \<C> u' v' c\<^esup>
  {v. (v,c,v') \<in> \<Delta> \<A>} (\<lambda>v (\<omega>, \<C>).
    if (\<forall>v''. (v,c,v'')\<in>\<Delta> \<A> \<longrightarrow> (u',v'')\<in>\<omega>-\<C>) then do {
        (\<omega>', \<C>') \<leftarrow> SPEC (INY_abstr3_is_valid_\<omega>'\<C>' \<omega> \<C> u' v' c v);
        RETURN (\<omega>', \<C>')
    } else
        RETURN (\<omega>, \<C>)
  ) (\<omega>, \<C>)"

lemma INY_abstr3_loopv_correct:
  assumes I1: "INY_abstr2_loopc_invar \<omega> \<C> u' v' \<Sigma>' (\<omega>',\<C>')" 
  shows "INY_abstr3_loopv \<omega>' \<C>' u' v' c \<le> 
      SPEC (INY_abstr2_is_valid_\<omega>'\<C>' \<omega>' \<C>' u' v' c)"
unfolding INY_abstr3_loopv_def
proof (rule FOREACHi_rule)
  have "{v. (v,c,v')\<in>\<Delta> \<A>} \<subseteq> \<Q> \<A>" using \<Delta>_consistent by blast
  thus "finite {v. (v,c,v')\<in>\<Delta> \<A>}" using rev_finite_subset[OF finite_\<Q>] by simp
next
  note invar_loopc = INY_abstr2_loopc_invarD[OF I1]
  show "INY_abstr3_loopv_invar \<omega>' \<C>' u' v' c {v. (v, c, v') \<in> \<Delta> \<A>} (\<omega>', \<C>')"
    apply (rule INY_abstr3_loopv_invarI)
    using invar_loopc apply (blast, blast, blast, blast, blast, blast)
    using invar_loopc(7) apply blast
    using invar_loopc apply (blast, blast)
    done
next
  case (goal3 v V')
  hence v'_succ_v: "(v,c,v') \<in> \<Delta> \<A>" by blast
  show ?case using goal3(3)
  proof (intro refine_vcg, clarsimp)
    case (goal1 \<omega>'' \<C>'' \<omega>''' \<C>''')
    note invar = INY_abstr3_loopv_invarD[OF goal1(1)]
    note valid = INY_abstr3_is_valid_\<omega>'\<C>'D[OF goal1(4)]
    let ?V'' = "V' - {v}"
    show "INY_abstr3_loopv_invar \<omega>' \<C>' u' v' c ?V'' (\<omega>''', \<C>''')"
    proof (intro INY_abstr3_loopv_invarI)
      have "\<omega>'\<union>INY_abstr3_set \<omega>' \<C>' u' v' c v \<subseteq> 
          \<omega>''\<union>INY_abstr3_set \<omega>'' \<C>'' u' v' c v"
          unfolding INY_abstr3_set_def using invar(1,2) by blast
      moreover have "INY_abstr3_set \<omega>' \<C>' u' v' c v \<inter> \<omega>' = {}"
          unfolding INY_abstr3_set_def by blast
      ultimately show "(\<Union>v\<in>{v.(v,c,v')\<in>\<Delta> \<A>}-?V''. 
          INY_abstr3_set \<omega>' \<C>' u' v' c v) \<subseteq> \<omega>'''" 
         using invar(6) valid(1,3) by blast
    next
      case goal5 thus ?case using invar(4,5) valid(1,2,4) by blast
    next
      from invar(8) show "(u', v') \<in> \<omega>'" .
    next
      from invar valid(2) show "(u', v') \<notin> \<C>'''" by blast
    qed (insert invar(1-3) valid(1,2,4,5), fast, fast, force, force)
  next
    case (goal2 \<omega>'' \<C>'')
    have "INY_abstr3_loopv_invar \<omega>' \<C>' u' v' c V' (\<omega>'', \<C>'')" 
        using goal2(1,2) by simp
    note invar = INY_abstr3_loopv_invarD[OF this]
    let ?V'' = "V' - {v}"
    from goal2(3) have A: "\<not> (\<forall>v''. (v, c, v'') \<in> \<Delta> \<A> \<longrightarrow> (u', v'') \<in> \<omega>'-\<C>')"
      using invar(1,2) by blast
    have B: "INY_abstr3_set \<omega>' \<C>' u' v' c v = {}" using A
        by (subst INY_abstr3_set_simp[OF v'_succ_v, of u' \<omega>' \<C>', symmetric], 
            auto)
    thus "INY_abstr3_loopv_invar \<omega>' \<C>' u' v' c ?V'' (\<omega>'', \<C>'')"
      apply (intro INY_abstr3_loopv_invarI)
      using invar(1-5) apply (fast, fast, force, force, force)
      using invar(6) B apply blast
      using invar(7) apply force
      using invar(8,9) apply (simp, simp)
      done
  qed

next
  case (goal4 \<omega>''\<C>'') thus ?case
    proof (cases \<omega>''\<C>'', simp)
      fix \<omega>'' \<C>''  
      assume I: "INY_abstr3_loopv_invar \<omega>' \<C>' u' v' c {} (\<omega>'', \<C>'')"
      note invar = INY_abstr3_loopv_invarD[OF I]
      note invar_loopc = INY_abstr2_loopc_invarD[OF I1]
      have "(\<Union>v\<in>{v. (v,c,v')\<in>\<Delta> \<A>}. INY_abstr3_set \<omega>' \<C>' u' v' c v) = 
          INY_abstr2_set \<omega>' \<C>' u' v' c"
          unfolding INY_abstr3_set_def INY_abstr2_set_def by blast
      thus "INY_abstr2_is_valid_\<omega>'\<C>' \<omega>' \<C>' u' v' c (\<omega>'', \<C>'')"
         using invar by(intro INY_abstr2_is_valid_\<omega>'\<C>'I, simp_all)
    qed
qed


definition INY_abstr3_loopc where
"INY_abstr3_loopc \<omega> \<C> u' v' \<equiv> FOREACH\<^bsup>INY_abstr2_loopc_invar \<omega> \<C> u' v'\<^esup>
    (\<Sigma> \<A>) (\<lambda>c (\<omega>, \<C>). do {
        (\<omega>', \<C>') \<leftarrow> INY_abstr3_loopv \<omega> \<C> u' v' c;
        RETURN (\<omega>', \<C>')
    }) (\<omega>, \<C>)"

definition INY_abstr3 where
"INY_abstr3 \<equiv> WHILE\<^sub>T\<^bsup>INY_abstr1_invar\<^esup> (\<lambda>(\<omega>, \<C>). \<C>\<noteq>{}) (\<lambda>(\<omega>, \<C>). do {
    (u',v') \<leftarrow> SPEC (\<lambda>(u', v'). (u',v') \<in> \<C>);
    let \<C> = \<C> - {(u',v')};
    (\<omega>, \<C>) \<leftarrow> (INY_abstr3_loopc \<omega> \<C> u' v');
    RETURN (\<omega>, \<C>)
}) (INY_initial, INY_initial)"

lemma INY_abstr3_loopc_correct:
  assumes "(\<omega>\<^sub>1',\<omega>\<^sub>2') \<in> Id" and "(\<C>\<^sub>1',\<C>\<^sub>2') \<in> Id" 
      and "(u\<^sub>1',u\<^sub>2') \<in> Id" and "(v\<^sub>1',v\<^sub>2') \<in> Id"
  shows "INY_abstr3_loopc \<omega>\<^sub>1' \<C>\<^sub>1' u\<^sub>1' v\<^sub>1' \<le> \<Down>Id (INY_abstr2_loopc \<omega>\<^sub>2' \<C>\<^sub>2' u\<^sub>2' v\<^sub>2')"
  unfolding INY_abstr3_loopc_def INY_abstr2_loopc_def
  using assms by (refine_rcg inj_on_id, simp_all add: INY_abstr3_loopv_correct)

lemma INY_abstr3_correct: "INY_abstr3 \<le> \<Down>Id INY_abstr2"
  unfolding INY_abstr3_def INY_abstr2_def
  by (refine_rcg INY_abstr3_loopc_correct, simp_all)


definition INY_abstr4_set where
"INY_abstr4_set \<omega> \<C> u' v' c v u \<equiv> (if (u,v)\<notin>\<omega> then {(u,v)} else {})"
   
definition INY_abstr4_is_valid_\<omega>'\<C>' where
"INY_abstr4_is_valid_\<omega>'\<C>' \<omega> \<C> u' v' c v u \<equiv> \<lambda>(\<omega>',\<C>'). \<omega>' \<subseteq> \<Q> \<A>\<times>\<Q> \<A> \<and>
    \<omega>'=\<omega>\<union>(\<omega>'-\<omega>) \<and> \<C>'=\<C>\<union>(\<omega>'-\<omega>) \<and> INY_abstr4_set \<omega> \<C> u' v' c v u \<subseteq> \<omega>' \<and> 
    \<S>\<^sub>\<A> \<inter> \<omega>' = {}"

lemma INY_abstr4_is_valid_\<omega>'\<C>'I:
  assumes "\<omega>'=\<omega>\<union>(\<omega>'-\<omega>)" "\<C>'=\<C>\<union>(\<omega>'-\<omega>)" "INY_abstr4_set \<omega> \<C> u' v' c v u \<subseteq> \<omega>'"
      "\<omega>' \<subseteq> \<Q> \<A>\<times>\<Q> \<A>" "\<S>\<^sub>\<A> \<inter> \<omega>' = {}"
  shows "INY_abstr4_is_valid_\<omega>'\<C>' \<omega> \<C> u' v' c v u (\<omega>',\<C>')"
  unfolding INY_abstr4_is_valid_\<omega>'\<C>'_def using assms by blast

lemma INY_abstr4_is_valid_\<omega>'\<C>'E: 
  assumes "INY_abstr4_is_valid_\<omega>'\<C>' \<omega> \<C> u' v' c v u (\<omega>',\<C>')"
  shows "\<omega>'=\<omega>\<union>(\<omega>'-\<omega>)" "\<C>'=\<C>\<union>(\<omega>'-\<omega>)" "INY_abstr4_set \<omega> \<C> u' v' c v u \<subseteq> \<omega>'"
      "\<omega>' \<subseteq> \<Q> \<A>\<times>\<Q> \<A>" "\<S>\<^sub>\<A> \<inter> \<omega>' = {}"
  using assms unfolding INY_abstr4_is_valid_\<omega>'\<C>'_def by blast+

definition INY_abstr4_loopu_invar where
"INY_abstr4_loopu_invar \<omega> \<C> u' v' c v U' \<equiv> \<lambda>(\<omega>', \<C>').
  let T = \<omega>'-\<omega> in \<omega>' = \<omega> \<union> T \<and> \<C>' = \<C> \<union> T \<and> \<C> \<subseteq> \<omega> \<and> 
    \<omega>' \<subseteq> \<Q> \<A> \<times> \<Q> \<A> \<and> \<C>' \<subseteq> \<Q> \<A>\<times>\<Q> \<A> \<and>
    (\<Union>u\<in>({u. (u,c,u')\<in>\<Delta> \<A>}-U'). INY_abstr4_set \<omega> \<C> u' v' c v u) \<subseteq> \<omega>' \<and>
    \<S>\<^sub>\<A> \<inter> \<omega>' = {} \<and> (u',v') \<in> \<omega> \<and> (u',v') \<notin> \<C>'"

lemma INY_abstr4_loopu_invarI[intro]:
  assumes "\<omega>' = \<omega> \<union> \<omega>'" "\<C>' = \<C> \<union> (\<omega>'-\<omega>)" "\<C> \<subseteq> \<omega>" 
     "\<omega>' \<subseteq> \<Q> \<A> \<times> \<Q> \<A>" "\<C>' \<subseteq> \<Q> \<A>\<times>\<Q> \<A>"
     "(\<Union>u\<in>({u. (u,c,u')\<in>\<Delta> \<A>}-U'). INY_abstr4_set \<omega> \<C> u' v' c v u) \<subseteq> \<omega>'"
     "\<S>\<^sub>\<A> \<inter> \<omega>' = {}" "(u',v') \<in> \<omega>" "(u',v')\<notin>\<C>'"
  shows "INY_abstr4_loopu_invar \<omega> \<C> u' v' c v U' (\<omega>',\<C>')"
  using assms unfolding INY_abstr4_loopu_invar_def by (auto simp add: Let_def)

lemma INY_abstr4_loopu_invarD[dest]:
  assumes "INY_abstr4_loopu_invar \<omega> \<C> u' v' c v U' (\<omega>',\<C>')"
  shows "\<omega>' = \<omega> \<union> \<omega>'" "\<C>' = \<C> \<union> (\<omega>'-\<omega>)" "\<C> \<subseteq> \<omega>" 
     "\<omega>' \<subseteq> \<Q> \<A> \<times> \<Q> \<A>" "\<C>' \<subseteq> \<Q> \<A>\<times>\<Q> \<A>"
     "(\<Union>u\<in>({u. (u,c,u')\<in>\<Delta> \<A>}-U'). INY_abstr4_set \<omega> \<C> u' v' c v u) \<subseteq> \<omega>'"
     "\<S>\<^sub>\<A> \<inter> \<omega>' = {}" "(u',v') \<in> \<omega>" "(u',v')\<notin>\<C>"
  using assms unfolding INY_abstr4_loopu_invar_def by (auto simp add: Let_def)

definition INY_abstr4_loopu where
"INY_abstr4_loopu \<omega> \<C> u' v' c v \<equiv> 
  FOREACH\<^bsup>INY_abstr4_loopu_invar \<omega> \<C> u' v' c v\<^esup>
  {u. (u,c,u') \<in> \<Delta> \<A>} (\<lambda>u (\<omega>, \<C>).
    if (u,v)\<notin>\<omega> then do {
        ASSERT ((u,v)\<notin>\<omega>);
        ASSERT ((u,v)\<notin>\<C>);
        RETURN (insert (u,v) \<omega>, insert (u,v) \<C>)
    } else
        RETURN (\<omega>, \<C>)
  ) (\<omega>, \<C>)"

lemma INY_abstr4_loopu_correct:
  assumes I1: "INY_abstr3_loopv_invar \<omega> \<C> u' v' c V' (\<omega>',\<C>')" and
          v'_succ_v: "(v,c,v') \<in> \<Delta> \<A>" and 
          v'_properties: "\<forall>v''. (v,c,v'')\<in>\<Delta> \<A> \<longrightarrow> (u',v'')\<in>\<omega>'-\<C>'\<union>{(u',v')}"
  shows "INY_abstr4_loopu \<omega>' \<C>' u' v' c v \<le> 
             SPEC (INY_abstr3_is_valid_\<omega>'\<C>' \<omega>' \<C>' u' v' c v)"
unfolding INY_abstr4_loopu_def
proof (rule FOREACHi_rule)
  have "{u. (u,c,u')\<in>\<Delta> \<A>} \<subseteq> \<Q> \<A>" using \<Delta>_consistent by blast
  thus "finite {u. (u,c,u')\<in>\<Delta> \<A>}" using rev_finite_subset[OF finite_\<Q>] by simp
next
  note invar_loopv = INY_abstr3_loopv_invarD[OF I1]
  show "INY_abstr4_loopu_invar \<omega>' \<C>' u' v' c v {u. (u,c,u')\<in>\<Delta> \<A>} (\<omega>', \<C>')"
      using invar_loopv by auto
next
  case (goal3 u U')
  hence u'_succ_u: "(u,c,u')\<in>\<Delta> \<A>" by blast
  show ?case using goal3(3)
    apply (intro refine_vcg)
    apply (unfold INY_abstr4_loopu_invar_def, auto) [2]
    apply clarify
  proof -
    case (goal1 \<omega>'' \<C>'')
    note invar = INY_abstr4_loopu_invarD[OF goal1(1)]
    note uv_notin_\<omega>'' = goal1(2)
    let ?U'' = "U' - {u}" and ?\<omega>''' = "insert (u, v) \<omega>''" and 
        ?\<C>''' = "insert (u, v) \<C>''"
    show "INY_abstr4_loopu_invar \<omega>' \<C>' u' v' c v ?U'' (?\<omega>''', ?\<C>''')"
    proof (intro INY_abstr4_loopu_invarI)
      have "\<omega>' \<union> INY_abstr4_set \<omega>' \<C>' u' v' c v u \<subseteq> 
            \<omega>'' \<union> INY_abstr4_set \<omega>'' \<C>'' u' v' c v u"
          unfolding INY_abstr4_set_def using invar(1) by auto
      moreover have "INY_abstr4_set \<omega>' \<C>' u' v' c v u \<inter> \<omega>' = {}"
          unfolding INY_abstr4_set_def using uv_notin_\<omega>'' invar(1) by force
      moreover have "\<And>a b u.(a, b) \<in> INY_abstr4_set \<omega>' \<C>' u' v' c v u
          \<Longrightarrow> (a, b) = (u, v)" unfolding INY_abstr4_set_def 
         by (auto split: split_if_asm)
      ultimately show "(\<Union>u\<in>{u.(u,c,u')\<in>\<Delta> \<A>}-?U''. 
          INY_abstr4_set \<omega>' \<C>' u' v' c v u) \<subseteq> ?\<omega>'''" using invar(6) by auto
    next

      {
        assume "(u,v)\<in>\<S>\<^sub>\<A>"
         hence "\<exists>v'. (u',v') \<in> \<S>\<^sub>\<A> \<and> (v,c,v')\<in>\<Delta> \<A>"
            using u'_succ_u v'_succ_v `(u,v)\<in>\<S>\<^sub>\<A>` \<S>\<^sub>\<A>_is_largest_sim by blast
         moreover have " \<forall>v''. (v, c, v'') \<in> \<Delta> \<A> \<longrightarrow> (u', v'') \<in> \<omega>''"
             using v'_properties `(u',v')\<in>\<omega>'` invar(1) by blast
         ultimately have False using invar(1,7) v'_properties by blast
      }
      thus "\<S>\<^sub>\<A> \<inter> (insert (u,v) \<omega>'') = {}" using invar(7) by blast
    qed (insert invar uv_notin_\<omega>'' u'_succ_u v'_succ_v \<Delta>_consistent, auto)
  next

    case (goal2 \<omega>'' \<C>'')
    let ?U'' = "U' - {u}"
    have "INY_abstr4_loopu_invar \<omega>' \<C>' u' v' c v U' (\<omega>'', \<C>'')" 
        using goal2(1,2) by simp
    note invar = INY_abstr4_loopu_invarD[OF this]
    note invar_loopv = INY_abstr3_loopv_invarD[OF I1]
    hence A: "INY_abstr4_set \<omega> \<C> u' v' c v u \<subseteq> \<omega>''" using `\<not>(u,v) \<notin> \<omega>''`
       unfolding INY_abstr4_set_def by (auto split: split_if_asm)
    have B: "(\<Union>u\<in>{u. (u,c,u')\<in>\<Delta> \<A>} - ?U''.
        INY_abstr4_set \<omega>' \<C>' u' v' c v u) = (\<Union>u\<in>{u. (u,c,u')\<in>\<Delta> \<A>} - U'.
        INY_abstr4_set \<omega>' \<C>' u' v' c v u) \<union> INY_abstr4_set \<omega>' \<C>' u' v' c v u"
        using u'_succ_u by blast
    thus "INY_abstr4_loopu_invar \<omega>' \<C>' u' v' c v ?U'' (\<omega>'', \<C>'')"
      apply (intro INY_abstr4_loopu_invarI)
      using invar(1-5) apply (blast, blast, blast, blast, blast)
      apply (subst B) unfolding INY_abstr4_set_def
      using invar(6) `\<not>(u,v)\<notin>\<omega>''` apply auto[1]
      using invar apply simp_all[3]
      done
  qed

next
  case (goal4 \<omega>''\<C>'') thus ?case
    proof (cases \<omega>''\<C>'', simp)
      fix \<omega>'' \<C>'' 
      assume I: "INY_abstr4_loopu_invar \<omega>' \<C>' u' v' c v {} (\<omega>'', \<C>'')"
      note invar = INY_abstr4_loopu_invarD[OF I]
      note invar_loopv = INY_abstr3_loopv_invarD[OF I1]
      have "\<And>v''. (v,c,v'') \<in> \<Delta> \<A> \<Longrightarrow> (u',v'') \<in> \<omega>' - \<C>'" 
          using v'_properties invar_loopv(1) invar(8,9) by blast
      hence "(\<Union>u\<in>{u. (u,c,u')\<in>\<Delta> \<A>}. INY_abstr4_set \<omega>' \<C>' u' v' c v u) = 
          INY_abstr3_set \<omega>' \<C>' u' v' c v" using v'_succ_v
          unfolding INY_abstr4_set_def INY_abstr3_set_def
          by (auto split: split_if_asm)
     thus "INY_abstr3_is_valid_\<omega>'\<C>' \<omega>' \<C>' u' v' c v (\<omega>'', \<C>'')"
         apply (intro INY_abstr3_is_valid_\<omega>'\<C>'I)
         using invar(1-3) apply auto[2]
         using invar apply auto
         done
    qed
qed


definition INY_abstr4_loopv where
"INY_abstr4_loopv \<omega> \<C> u' v' c \<equiv> 
  FOREACH\<^bsup>INY_abstr3_loopv_invar \<omega> \<C> u' v' c\<^esup>
  {v. (v,c,v')\<in>\<Delta> \<A>} (\<lambda>v (\<omega>, \<C>).
    if (\<forall>v''. (v,c,v'')\<in>\<Delta> \<A> \<longrightarrow> (u',v'') \<in> \<omega>-\<C>) then do {
        (\<omega>', \<C>') \<leftarrow> INY_abstr4_loopu \<omega> \<C> u' v' c v;
        RETURN (\<omega>', \<C>')
    } else
        RETURN (\<omega>, \<C>)
  ) (\<omega>, \<C>)"

lemma INY_abstr4_loopv_correct:
  shows "INY_abstr4_loopv \<omega>' \<C>' u' v' c \<le> 
             \<Down>Id (INY_abstr3_loopv \<omega>' \<C>' u' v' c)"
unfolding INY_abstr4_loopv_def INY_abstr3_loopv_def
using assms apply (refine_rcg inj_on_id)
using assms apply simp_all[4]
apply simp
apply (rule INY_abstr4_loopu_correct, simp, blast)
apply simp_all
done


definition INY_abstr4_loopc where
"INY_abstr4_loopc \<omega> \<C> u' v' \<equiv> FOREACH\<^bsup>INY_abstr2_loopc_invar \<omega> \<C> u' v'\<^esup>
    (\<Sigma> \<A>) (\<lambda>c (\<omega>, \<C>). do {
        (\<omega>', \<C>') \<leftarrow> INY_abstr4_loopv \<omega> \<C> u' v' c;
        RETURN (\<omega>', \<C>')
    }) (\<omega>, \<C>)"

lemma INY_abstr4_loopc_correct:
  assumes "(\<omega>\<^sub>1',\<omega>\<^sub>2') \<in> Id" and "(\<C>\<^sub>1',\<C>\<^sub>2') \<in> Id" and 
          "(u\<^sub>1',u\<^sub>2') \<in> Id" and "(v\<^sub>1',v\<^sub>2') \<in> Id"
  shows "INY_abstr4_loopc \<omega>\<^sub>1' \<C>\<^sub>1' u\<^sub>1' v\<^sub>1' \<le> \<Down>Id (INY_abstr3_loopc \<omega>\<^sub>2' \<C>\<^sub>2' u\<^sub>2' v\<^sub>2')"
unfolding INY_abstr4_loopc_def INY_abstr3_loopc_def using assms
apply (refine_rcg inj_on_id INY_abstr4_loopv_correct)
apply simp_all[4]
apply (rule INY_abstr4_loopv_correct)
apply simp_all
done

definition INY_abstr4 where
"INY_abstr4 \<equiv> WHILE\<^sub>T\<^bsup>INY_abstr1_invar\<^esup> (\<lambda>(\<omega>, \<C>). \<C>\<noteq>{}) (\<lambda>(\<omega>, \<C>). do {
    (u',v') \<leftarrow> SPEC (\<lambda>(u', v'). (u',v') \<in> \<C>);
    let \<C> = \<C> - {(u',v')};
    (\<omega>, \<C>) \<leftarrow> (INY_abstr4_loopc \<omega> \<C> u' v');
    RETURN (\<omega>, \<C>)
}) (INY_initial, INY_initial)"

lemma INY_abstr4_correct:
  "INY_abstr4 \<le> \<Down>Id INY_abstr3"
  unfolding INY_abstr4_def INY_abstr3_def
  by (refine_rcg inj_on_id INY_abstr4_loopc_correct, simp_all)


subsection {* Optimisation with caching *}

text {*
  We now introduce the counter @{term "N"} and the constant data structures
  @{term "d"} and @{term "\<delta>r"}.
  @{text "N(c,u',v)"} stores the number of successors of @{term "v"} that are known 
                   to be unable to simulate v' and have been processed.
  @{text "d(v,c)"}    stores the number of successors of @{term "v"} w.r.t. @{term "c"},
                   i.e. @{text "|\<delta>(v,c)|"} (or @{text "|{v'. (v,c,v') \<in> \<Delta>}|"})
*}

text {* The counter's domain must be @{text "\<Sigma>\<times>\<Q>\<times>\<Q>"} 
    and initially, all values must be zero. *}
definition INY_is_valid_initial_counter::
    "(('a\<times>'q\<times>'q) \<rightharpoonup> nat) \<Rightarrow> bool" where
"INY_is_valid_initial_counter N \<equiv> (dom N = \<Sigma> \<A>\<times>\<Q> \<A>\<times>\<Q> \<A>) \<and>
    (\<forall>(c,u,v)\<in>dom N. N (c,u,v) = Some (card {v'. (v,c,v') \<in> \<Delta> \<A>}))"

text {* Increments the counter value for @{term "(c,u',v)"}.
  This is used when we know that another successor of @{term "v"} is unable
  to simulate @{term "u'"}. *}
definition INY_dec_counter where
"INY_dec_counter N c u' v = 
    (case N (c,u',v) of Some n \<Rightarrow> (let n = n - 1 in
          (N((c,u',v) \<mapsto> n), n = 0)) | None \<Rightarrow> (N, True))"

lemma INY_dec_counter_correct: assumes "N (c2,u'2,v2) = Some n"
  shows "(c,u',v)=(c2,u'2,v2) \<Longrightarrow> 
            fst (INY_dec_counter N c u' v) (c2,u'2,v2) = Some (n-(1::nat))"
        "(c,u',v)\<noteq>(c2,u'2,v2) \<Longrightarrow> 
            fst (INY_dec_counter N c u' v) (c2,u'2,v2) = Some n"
        "dom (fst (INY_dec_counter N c u' v)) = dom N"
        "N (c,u',v) = Some n \<Longrightarrow> snd (INY_dec_counter N c u' v) = (n - 1 = 0)"
using assms by (auto simp: INY_dec_counter_def 
    split: option.split split_if_asm)


text {* The @{text "dec_counter"} function does not change the counter's domain. *}
lemma INY_dec_counter_dom_unchanged[simp]:
  "dom (fst (INY_dec_counter N c u' v)) = dom N"
  unfolding INY_dec_counter_def dom_def by (auto split: option.split simp: Let_def)

text {* Only the incremented value changes. *}
lemma INY_dec_counter_unaffected: "(c,u',v)\<noteq>(c2,u'2,v2) \<Longrightarrow>
    fst (INY_dec_counter N c u' v) (c2,u'2,v2) = N (c2,u'2,v2)"
by (auto simp: assms INY_dec_counter_def split: option.split simp: Let_def)


text {* @{term "d"} is the data structure that stores @{text "|\<delta>(v,c)|"} 
    for any @{text "v\<in>\<Q>"}, @{text "c\<in>\<Sigma>"}. *}
definition INY_abstr5_d_correct where
"INY_abstr5_d_correct d \<equiv> (dom d = \<Q> \<A>\<times>\<Sigma> \<A>) \<and> 
    (\<forall>(v,c)\<in>dom d. d (v,c) = Some (card {v'. (v,c,v')\<in>\<Delta> \<A>}))"

lemma INY_abstr5_d_correctD[dest]:
  assumes "INY_abstr5_d_correct d" and "(v,c) \<in> \<Q> \<A>\<times>\<Sigma> \<A>"
  shows "d (v,c) = Some (card {v'. (v,c,v')\<in>\<Delta> \<A>})"
  using assms unfolding INY_abstr5_d_correct_def by blast

text {* @{term "\<delta>r"} is the data structure that stores the predecessors of 
    @{term "v"} w.r.t. @{term "c"} for any @{text "v\<in>\<Q>"}, @{text "c\<in>\<Sigma>"}.
*}
definition INY_abstr5_\<delta>r_correct where
"INY_abstr5_\<delta>r_correct d \<equiv> (dom d = \<Q> \<A>\<times>\<Sigma> \<A>) \<and> 
    (\<forall>(v,c)\<in>dom d. d (v,c) = Some {v'. (v',c,v)\<in>\<Delta> \<A>})"  

definition INY_abstr5_N\<delta>r_correct where
"INY_abstr5_N\<delta>r_correct \<omega> \<C> N \<delta>r \<equiv> (dom N = \<Sigma> \<A>\<times>\<Q> \<A>\<times>\<Q> \<A>) \<and>
    (\<forall>(c,u',v)\<in>dom N. N (c,u',v) = 
        Some (card {v''. (v,c,v'')\<in>\<Delta> \<A> \<and> (u',v'')\<notin>\<omega>-\<C>})) \<and>
     INY_abstr5_\<delta>r_correct \<delta>r"

lemma INY_abstr5_N\<delta>r_correctI: 
assumes "dom N = \<Sigma> \<A>\<times>\<Q> \<A>\<times>\<Q> \<A>"
    "\<And>c u' v. \<lbrakk>c\<in>\<Sigma> \<A>; u'\<in>\<Q> \<A>; v\<in>\<Q> \<A>\<rbrakk> \<Longrightarrow>
        N (c,u',v) = Some (card {v''. (v,c,v'')\<in>\<Delta> \<A> \<and> (u',v'')\<notin>\<omega>-\<C>})"
    "INY_abstr5_\<delta>r_correct \<delta>r"
shows "INY_abstr5_N\<delta>r_correct \<omega> \<C> N \<delta>r"
unfolding INY_abstr5_N\<delta>r_correct_def using assms by blast

lemma INY_abstr5_N\<delta>r_correctD[dest]: 
assumes "INY_abstr5_N\<delta>r_correct \<omega> \<C> N \<delta>r"
shows "dom N = \<Sigma> \<A>\<times>\<Q> \<A>\<times>\<Q> \<A>"
    "\<And>c u' v. \<lbrakk>c\<in>\<Sigma> \<A>; u'\<in>\<Q> \<A>; v\<in>\<Q> \<A>\<rbrakk> \<Longrightarrow>
        N (c,u',v) = Some (card {v''. (v,c,v'')\<in>\<Delta> \<A> \<and> (u',v'')\<notin>\<omega>-\<C>})"
   "INY_abstr5_\<delta>r_correct \<delta>r"
using assms unfolding INY_abstr5_N\<delta>r_correct_def by blast+

definition INY_abstr5_invar where
"INY_abstr5_invar \<delta>r \<equiv> \<lambda>(\<omega>, \<C>, N).
    INY_abstr1_invar (\<omega>, \<C>) \<and> INY_abstr5_N\<delta>r_correct \<omega> \<C> N \<delta>r"

lemma INY_abstr5_invarI:
  assumes "INY_abstr1_invar (\<omega>, \<C>)" and "INY_abstr5_N\<delta>r_correct \<omega> \<C> N \<delta>r"
  shows "INY_abstr5_invar \<delta>r (\<omega>, \<C>, N)"
  unfolding INY_abstr5_invar_def using assms by blast

lemma INY_abstr5_invarD[dest]:
  assumes "INY_abstr5_invar \<delta>r (\<omega>, \<C>, N)"
  shows "INY_abstr1_invar (\<omega>, \<C>)" "INY_abstr5_N\<delta>r_correct \<omega> \<C> N \<delta>r"
  using assms unfolding INY_abstr5_invar_def by blast+

definition INY_abstr5_loopc_N\<delta>r_correct where
"INY_abstr5_loopc_N\<delta>r_correct \<omega> \<C> N \<delta>r u' v' \<Sigma>' N' \<equiv> 
   (dom N' = \<Sigma> \<A>\<times>\<Q> \<A>\<times>\<Q> \<A>) \<and>
    (\<forall>(c,u'',v)\<in>dom N'. N' (c,u'',v) = (if c\<in>\<Sigma>'
          then N (c,u'',v)
          else Some (card {v''. (v,c,v'')\<in>\<Delta> \<A> \<and> (u'',v'')\<notin>(\<omega>-\<C>)})
    )) \<and> INY_abstr5_\<delta>r_correct \<delta>r"

lemma INY_abstr5_loopc_N\<delta>r_correctI: 
assumes "dom N' = \<Sigma> \<A>\<times>\<Q> \<A>\<times>\<Q> \<A>"
    "\<And>c u'' v. \<lbrakk>c\<in>\<Sigma> \<A>; c\<in>\<Sigma>'; u''\<in>\<Q> \<A>; v\<in>\<Q> \<A>\<rbrakk> \<Longrightarrow>
         N' (c,u'',v) = N (c,u'',v)" and
    "\<And>c u'' v. \<lbrakk>c\<in>\<Sigma> \<A>; c\<notin>\<Sigma>'; u''\<in>\<Q> \<A>; v\<in>\<Q> \<A>\<rbrakk> \<Longrightarrow>
       N' (c,u'',v) = Some (card {v''. (v,c,v'')\<in>\<Delta> \<A> \<and> 
       (u'',v'')\<notin>(\<omega>-\<C>)})"
     "INY_abstr5_\<delta>r_correct \<delta>r"
shows "INY_abstr5_loopc_N\<delta>r_correct \<omega> \<C> N \<delta>r u' v' \<Sigma>' N'"
unfolding INY_abstr5_loopc_N\<delta>r_correct_def using assms by auto

lemma INY_abstr5_loopc_N\<delta>r_correctD[dest]: 
assumes "INY_abstr5_loopc_N\<delta>r_correct \<omega> \<C> N \<delta>r u' v' \<Sigma>' N'"
shows "dom N' = \<Sigma> \<A>\<times>\<Q> \<A>\<times>\<Q> \<A>"
    "\<And>c u'' v. \<lbrakk>c\<in>\<Sigma> \<A>; c\<in>\<Sigma>'; u''\<in>\<Q> \<A>; v\<in>\<Q> \<A>\<rbrakk> \<Longrightarrow>
         N' (c,u'',v) = N (c,u'',v)" and
    "\<And>c u'' v. \<lbrakk>c\<in>\<Sigma> \<A>; c\<notin>\<Sigma>'; u''\<in>\<Q> \<A>; v\<in>\<Q> \<A>\<rbrakk> \<Longrightarrow>
       N' (c,u'',v) = Some (card {v''. (v,c,v'')\<in>\<Delta> \<A> \<and> 
       (u'',v'')\<notin>(\<omega>-\<C>)})"
    "INY_abstr5_\<delta>r_correct \<delta>r"
using assms unfolding INY_abstr5_loopc_N\<delta>r_correct_def by auto

definition INY_abstr5_loopc_invar where
"INY_abstr5_loopc_invar \<omega> \<C> N \<delta>r u' v' \<Sigma>' \<equiv> \<lambda>(\<omega>', \<C>', N').
    INY_abstr2_loopc_invar \<omega> \<C> u' v' \<Sigma>' (\<omega>', \<C>') \<and> 
    INY_abstr5_loopc_N\<delta>r_correct \<omega> \<C> N \<delta>r u' v' \<Sigma>' N'"

lemma INY_abstr5_loopc_invarI:
  assumes "INY_abstr2_loopc_invar \<omega> \<C> u' v' \<Sigma>' (\<omega>', \<C>')" and 
      "INY_abstr5_loopc_N\<delta>r_correct \<omega> \<C> N \<delta>r u' v' \<Sigma>' N'"
  shows "INY_abstr5_loopc_invar \<omega> \<C> N \<delta>r u' v' \<Sigma>' (\<omega>', \<C>', N')"
  unfolding INY_abstr5_loopc_invar_def using assms by blast

lemma INY_abstr5_loopc_invarD:
  assumes "INY_abstr5_loopc_invar \<omega> \<C> N \<delta>r u' v' \<Sigma>' (\<omega>', \<C>', N')"
  shows "INY_abstr2_loopc_invar \<omega> \<C> u' v' \<Sigma>' (\<omega>', \<C>')" and 
      "INY_abstr5_loopc_N\<delta>r_correct \<omega> \<C> N \<delta>r u' v' \<Sigma>' N'"
  using assms unfolding INY_abstr5_loopc_invar_def by blast+




definition INY_abstr5_loopv_N\<delta>r_correct where
"INY_abstr5_loopv_N\<delta>r_correct \<omega> \<C> N \<delta>r u' v' c' V' N' \<equiv> 
  (dom N' = \<Sigma> \<A>\<times>\<Q> \<A>\<times>\<Q> \<A>) \<and>
    (\<forall>(c,u'',v)\<in>dom N'. N' (c,u'',v) = (if c\<noteq>c' \<or> v\<in>V'
        then N (c,u'',v)
        else Some (card {v''. (v,c,v'')\<in>\<Delta> \<A> \<and> (u'',v'')\<notin>(\<omega>-\<C>)})
  )) \<and> INY_abstr5_\<delta>r_correct \<delta>r"

lemma INY_abstr5_loopv_N\<delta>r_correctI: 
assumes "dom N' = \<Sigma> \<A>\<times>\<Q> \<A>\<times>\<Q> \<A>"
    "\<And>c u'' v. \<lbrakk>c\<in>\<Sigma> \<A>; c\<noteq>c'; u''\<in>\<Q> \<A>; v\<in>\<Q> \<A>\<rbrakk> \<Longrightarrow>
         N' (c,u'',v) = N (c,u'',v)" and
    "\<And>c u'' v. \<lbrakk>c\<in>\<Sigma> \<A>; v\<in>V'; u''\<in>\<Q> \<A>; v\<in>\<Q> \<A>\<rbrakk> \<Longrightarrow>
         N' (c,u'',v) = N (c,u'',v)"
    "\<And>c u'' v. \<lbrakk>c\<in>\<Sigma> \<A>; c=c'; v\<notin>V'; u''\<in>\<Q> \<A>; v\<in>\<Q> \<A>\<rbrakk> \<Longrightarrow>
       N' (c,u'',v) = Some (card {v''. (v,c,v'')\<in>\<Delta> \<A> \<and> 
            (u'',v'')\<notin>(\<omega>-\<C>)})"
    "INY_abstr5_\<delta>r_correct \<delta>r"
shows "INY_abstr5_loopv_N\<delta>r_correct \<omega> \<C> N \<delta>r u' v' c' V' N'"
unfolding INY_abstr5_loopv_N\<delta>r_correct_def using assms by auto

lemma INY_abstr5_loopv_N\<delta>r_correctD:
assumes "INY_abstr5_loopv_N\<delta>r_correct \<omega> \<C> N \<delta>r u' v' c' V' N'"
shows "dom N' = \<Sigma> \<A>\<times>\<Q> \<A>\<times>\<Q> \<A>"
    "\<And>c u'' v. \<lbrakk>c\<in>\<Sigma> \<A>; c\<noteq>c'; u''\<in>\<Q> \<A>; v\<in>\<Q> \<A>\<rbrakk> \<Longrightarrow>
         N' (c,u'',v) = N (c,u'',v)" and
    "\<And>c u'' v. \<lbrakk>c\<in>\<Sigma> \<A>; v\<in>V'; u''\<in>\<Q> \<A>; v\<in>\<Q> \<A>\<rbrakk> \<Longrightarrow>
         N' (c,u'',v) = N (c,u'',v)"
    "\<And>c u'' v. \<lbrakk>c\<in>\<Sigma> \<A>; c=c'; v\<notin>V'; u''\<in>\<Q> \<A>; v\<in>\<Q> \<A>\<rbrakk> \<Longrightarrow>
         N' (c,u'',v) = Some (card {v''. (v,c,v'')\<in>\<Delta> \<A> \<and> 
              (u'',v'')\<notin>(\<omega>-\<C>)})"
    "INY_abstr5_\<delta>r_correct \<delta>r"
using assms unfolding INY_abstr5_loopv_N\<delta>r_correct_def by auto

definition INY_abstr5_loopv_invar where
"INY_abstr5_loopv_invar \<omega> \<C> N \<delta>r u' v' c V' \<equiv> \<lambda>(\<omega>', \<C>', N').
    INY_abstr3_loopv_invar \<omega> \<C> u' v' c V' (\<omega>', \<C>') \<and> 
    INY_abstr5_loopv_N\<delta>r_correct \<omega> \<C> N \<delta>r u' v' c V' N'"

lemma INY_abstr5_loopv_invarI:
  assumes "INY_abstr3_loopv_invar \<omega> \<C> u' v' c V' (\<omega>', \<C>')" and 
      "INY_abstr5_loopv_N\<delta>r_correct \<omega> \<C> N \<delta>r u' v' c V' N'"
  shows "INY_abstr5_loopv_invar \<omega> \<C> N \<delta>r u' v' c V' (\<omega>', \<C>', N')"
  unfolding INY_abstr5_loopv_invar_def using assms by blast

lemma INY_abstr5_loopv_invarD[intro]:
  assumes "INY_abstr5_loopv_invar \<omega> \<C> N \<delta>r u' v' c V' (\<omega>', \<C>', N')"
  shows "INY_abstr3_loopv_invar \<omega> \<C> u' v' c V' (\<omega>', \<C>')" and 
      "INY_abstr5_loopv_N\<delta>r_correct \<omega> \<C> N \<delta>r u' v' c V' N'"
  using assms unfolding INY_abstr5_loopv_invar_def by blast+


text {* The new, optimised loops of the algorithm using the cache variables
    we have just introduces. *}
definition INY_abstr5_loopu where
"INY_abstr5_loopu \<omega> \<C> \<delta>r u' v' c v \<equiv> 
  FOREACH\<^bsup>INY_abstr4_loopu_invar \<omega> \<C> u' v' c v\<^esup>
  (case \<delta>r(u',c) of None \<Rightarrow> {} | Some s \<Rightarrow> s) (\<lambda>u (\<omega>, \<C>).
    if (u,v)\<notin>\<omega> then do {
        ASSERT ((u,v)\<notin>\<omega>);
        ASSERT ((u,v)\<notin>\<C>);
        RETURN (insert (u,v) \<omega>, insert (u,v) \<C>)
    } else
        RETURN (\<omega>, \<C>)
  ) (\<omega>, \<C>)"

lemma INY_abstr5_\<delta>r_correct:
  assumes "INY_abstr2_loopc_invar \<omega> \<C> u' v' \<Sigma>' (\<omega>',\<C>')"
          "INY_abstr5_N\<delta>r_correct \<omega> \<C>\<^sub>2 N \<delta>r"
          "c\<in>\<Sigma>'" "\<Sigma>'\<subseteq>\<Sigma> \<A>"
    shows "(\<delta>r(u',c)) = Some {u. (u,c,u')\<in>\<Delta> \<A>}" (is ?A) and
           "(\<delta>r(v',c)) = Some {v. (v,c,v')\<in>\<Delta> \<A>}" (is ?B)
proof-
  from INY_abstr2_loopc_invarD(1,4,8)[OF assms(1)] 
      have "u'\<in>\<Q> \<A>" "v'\<in>\<Q> \<A>" by blast+
  moreover from `c\<in>\<Sigma>'` and `\<Sigma>'\<subseteq>\<Sigma> \<A>` have "c\<in>\<Sigma> \<A>" by blast
  moreover note INY_abstr5_N\<delta>r_correctD(3)[OF assms(2)]
  ultimately show ?A and ?B unfolding INY_abstr5_\<delta>r_correct_def by auto
qed


lemma INY_abstr5_loopu_correct:
  assumes "INY_abstr2_loopc_invar \<omega> \<C> u' v' \<Sigma>' (\<omega>',\<C>')"
          "INY_abstr5_N\<delta>r_correct \<omega> (\<C>\<union>{(u',v')}) N \<delta>r"
          "c\<in>\<Sigma>'" "\<Sigma>'\<subseteq>\<Sigma> \<A>"
  shows "INY_abstr5_loopu \<omega>'' \<C>'' \<delta>r u' v' c v \<le> \<Down>Id 
        (INY_abstr4_loopu \<omega>'' \<C>'' u' v' c v)"
unfolding INY_abstr5_loopu_def INY_abstr4_loopu_def
thm INY_abstr5_\<delta>r_correct[OF assms(1)]
by (refine_rcg inj_on_id, simp_all add: INY_abstr5_\<delta>r_correct[OF assms])


definition INY_abstr5_loopv where
"INY_abstr5_loopv \<omega> \<C> N \<delta>r u' v' c \<equiv> 
  FOREACH\<^bsup>INY_abstr5_loopv_invar \<omega> \<C> N \<delta>r u' v' c\<^esup>
  (case \<delta>r(v',c) of None \<Rightarrow> {} | Some s \<Rightarrow> s) (\<lambda>v (\<omega>, \<C>, N). do {
    let (N, iszero) = INY_dec_counter N c u' v;
    if iszero then do {
        (\<omega>', \<C>') \<leftarrow> INY_abstr5_loopu \<omega> \<C> \<delta>r u' v' c v;
        RETURN (\<omega>', \<C>', N)
    } else
        RETURN (\<omega>, \<C>, N)
  }) (\<omega>, \<C>, N)"

text {* If @{text "v \<notin> \<delta>\<inverse>(v',)"}, we don't have to do any updates on @{term "v"}, 
    it is not affected by the new information about @{term "(u',v')"}. *}
lemma INY_abstr5_loopv_N_unaffected:
  assumes "(v,c,v')\<notin>\<Delta> \<A>"
  shows "card {v''.(v,c,v'')\<in>\<Delta> \<A> \<and> (u'',v'')\<notin>\<omega>-\<C>} =
         card {v''. (v,c,v'')\<in>\<Delta> \<A> \<and> (u'',v'')\<notin>\<omega>-(\<C>\<union>{(u',v')})}" 
         (is "card ?U = card ?V")
by (subgoal_tac "?U=?V", simp, insert assms, blast)


text {* If the outer invariant holds, the counters have the correct values
  initially. *}
lemma INY_abstr5_loopv_N\<delta>r_correct_initial:
  assumes "INY_abstr5_loopc_N\<delta>r_correct \<omega> \<C> N \<delta>r u' v' \<Sigma>' N'"
          "INY_abstr2_loopc_invar \<omega> \<C> u' v' \<Sigma>' (\<omega>',\<C>')" "c\<in>\<Sigma>'"
          "INY_abstr5_N\<delta>r_correct \<omega> (\<C> \<union> {(u',v')}) N \<delta>r"
  shows "INY_abstr5_loopv_N\<delta>r_correct \<omega> \<C> N' \<delta>r u' v' c {v. (v,c,v') \<in> \<Delta> \<A>} N'"
apply (rule INY_abstr5_loopv_N\<delta>r_correctI)
defer 4
apply (fact INY_abstr5_loopc_N\<delta>r_correctD(1)[OF assms(1)])
apply simp
apply blast
apply (fact INY_abstr5_N\<delta>r_correctD(3)[OF assms(4)])
proof-
  case (goal1 c' u'' v) 
    note invar_loopc = INY_abstr2_loopc_invarD[OF assms(2)]
    have "(v,c',v')\<notin>\<Delta> \<A>" using goal1(2,3) by blast
    have "c'\<in>\<Sigma>'" using goal1(2) assms(3) by simp
    note INY_abstr5_loopc_N\<delta>r_correctD(2)[OF assms(1) goal1(1) this goal1(4,5)]
    also note INY_abstr5_N\<delta>r_correctD(2)[OF assms(4) goal1(1,4,5)]
    also note INY_abstr5_loopv_N_unaffected[OF `(v,c',v')\<notin>\<Delta> \<A>`, symmetric]
    finally show ?case .
qed

text {* The original if condition (@{text "v"} has no successor w.r.t. @{term "c"} 
  that can simulate @{term "u'"}) and the new one (the counter for @{term "(c,u',v)"} 
  is at its maximum) are equivalent. *}
lemma INY_abstr5_loopv_N_eq_0_iff:
  assumes I1: "INY_abstr3_loopv_invar \<omega>' \<C>' u' v' c V' (\<omega>'', \<C>'')" and
          C1: "INY_abstr5_loopv_N\<delta>r_correct \<omega> \<C> N' \<delta>r u' v' c V' N''" and
          I2: "INY_abstr2_loopc_invar \<omega> \<C> u' v' \<Sigma>' (\<omega>',\<C>')" and
          C2: "INY_abstr5_loopc_N\<delta>r_correct \<omega> \<C> N \<delta>r u' v' \<Sigma>' N'" and
          C3: "INY_abstr5_N\<delta>r_correct \<omega> (\<C>\<union>{(u',v')}) N \<delta>r" and "v\<in>V'" and
              "V'\<subseteq>{v. (v,c,v')\<in>\<Delta> \<A>}" and "c\<in>\<Sigma>'" and "\<Sigma>'\<subseteq>\<Sigma> \<A>"
  shows "snd (INY_dec_counter N'' c u' v) =
          (\<forall>v''. (v,c,v'')\<in>\<Delta> \<A> \<longrightarrow> (u',v'') \<in> \<omega>'' - \<C>'')"
proof-
  note correct = INY_abstr5_loopv_N\<delta>r_correctD[OF C1]
  note correct_loopc = INY_abstr5_loopc_N\<delta>r_correctD[OF C2]
  note correct_while = INY_abstr5_N\<delta>r_correctD[OF C3]
  note invar =  INY_abstr3_loopv_invarD[OF I1]
  note invar_loopc = INY_abstr2_loopc_invarD[OF I2]

  let ?X1 = "{v''.(v,c,v'')\<in>\<Delta> \<A> \<and> (u',v'')\<notin>\<omega> - (\<C> \<union> {(u',v')})}"
  let ?X2 = "{v''.(v,c,v'')\<in>\<Delta> \<A> \<and> (u',v'')\<notin>\<omega> - \<C>}"
  let ?X3 = "{v''.(v,c,v'')\<in>\<Delta> \<A> \<and> (u',v'')\<notin>\<omega>'- \<C>'}"

  from `v\<in>V'` and `V'\<subseteq>{v. (v,c,v')\<in>\<Delta> \<A>}` have "(v,c,v')\<in>\<Delta> \<A>" "v\<in>\<Q> \<A>"
      using \<Delta>_consistent by blast+
  from `c\<in>\<Sigma>'` `\<Sigma>'\<subseteq>\<Sigma> \<A>` have "c\<in>\<Sigma> \<A>" by blast
  have "u'\<in>\<Q> \<A>" using invar_loopc(1,4,8) by blast

  have "?X1 \<subseteq> \<Q> \<A>" "?X2 \<subseteq> \<Q> \<A>" using \<Delta>_consistent by blast+
  hence fin: "finite ?X1" "finite ?X2" using finite_\<Q> finite_\<Sigma>
    by (blast intro: finite_subset)+

  have "?X2 = ?X1 - {v'}" and "v'\<in>?X1" using `(v,c,v')\<in>\<Delta> \<A>` 
      invar_loopc(2,8,9) by auto
  hence new_card: "card ?X2 = card ?X1 - 1" using fin invar_loopc(8) by simp

  have "N'' (c,u',v) = N' (c,u',v)" 
      using correct(3)[OF `c\<in>\<Sigma> \<A>` _ `u'\<in>\<Q> \<A>` `v\<in>\<Q> \<A>`] `v\<in>V'` by simp
  also have "... =  N (c,u',v)" 
      using correct_loopc(2)[OF `c\<in>\<Sigma> \<A>` `c\<in>\<Sigma>'` `u'\<in>\<Q> \<A>` `v\<in>\<Q> \<A>`] .
  also have "... = Some (card ?X1)"
      using correct_while(2)[OF `c\<in>\<Sigma> \<A>` `u'\<in>\<Q> \<A>` `v\<in>\<Q> \<A>`] .
  finally have "snd (INY_dec_counter N'' c u' v) \<longleftrightarrow> (card ?X2 = 0)"
      using INY_dec_counter_correct(4)[of N'' c u' v "card ?X1"] new_card 
      by simp
  hence "snd (INY_dec_counter N'' c u' v) = (?X2={})" 
      using fin by simp
  thus ?thesis using invar(1,2) correct(2) invar_loopc(1,2) by blast
qed

text {* The counter is updated correctly, i.e. after an iteration,
  it will correctly reflect the fact that @{term "v'"} is a successor of @{term "v"} 
  that cannot simulate @{term "u'"}.
*}
lemma INY_abstr5_loopv_N_correctness_preserved:
  assumes I1: "INY_abstr3_loopv_invar \<omega>' \<C>' u' v' c V' (\<omega>'', \<C>'')" and
          C1: "INY_abstr5_loopv_N\<delta>r_correct \<omega> \<C> N' \<delta>r u' v' c V' N''" and
          I2: "INY_abstr2_loopc_invar \<omega> \<C> u' v' \<Sigma>' (\<omega>',\<C>')" and
          C2: "INY_abstr5_loopc_N\<delta>r_correct \<omega> \<C> N \<delta>r u' v' \<Sigma>' N'" and
          I3: "INY_abstr1_invar (\<omega>,\<C>\<C> \<union> {(u',v')})" and
          C3: "INY_abstr5_N\<delta>r_correct \<omega> (\<C>\<union>{(u',v')}) N \<delta>r" and "v\<in>V'" and
              "V'\<subseteq>{v. (v,c,v')\<in>\<Delta> \<A>}" and "c\<in>\<Sigma>'"
  shows "INY_abstr5_loopv_N\<delta>r_correct \<omega> \<C> N' \<delta>r u' v' c (V' - {v})
         (fst (INY_dec_counter N'' c u' v))"
proof-
  from `v\<in>V'` and `V'\<subseteq>{v. (v,c,v')\<in>\<Delta> \<A>}` have "(v,c,v')\<in>\<Delta> \<A>" by blast
  note correct = INY_abstr5_loopv_N\<delta>r_correctD[OF C1]
  note correct_loopc = INY_abstr5_loopc_N\<delta>r_correctD[OF C2]
  note correct_while = INY_abstr5_N\<delta>r_correctD[OF C3]
  note invar =  INY_abstr3_loopv_invarD[OF I1]
  note invar_loopc = INY_abstr2_loopc_invarD[OF I2]
  note invar_while = INY_abstr1_invarD[OF I3]
  from invar_loopc(1,4,8) have "u'\<in>\<Q> \<A>" by blast
  show ?thesis
    apply (rule INY_abstr5_loopv_N\<delta>r_correctI)
    using correct(1) apply simp
    using INY_dec_counter_unaffected[of c u' v _ _ _ N''] 
        correct(2) apply auto []
    using INY_dec_counter_unaffected[of c u' v _ _ _ N''] 
        correct(3) apply auto []
    defer
    apply (fact correct(5))
     
    proof-
    case (goal1 c' u'' v'')
      hence "c'\<in>\<Sigma>'" using `c\<in>\<Sigma>'` by simp
      show ?case proof(cases "v''=v")
        case False 
          hence "(c,u',v) \<noteq> (c',u'',v'')" by blast
          note INY_dec_counter_unaffected[OF this, of N'']
          moreover have "v''\<notin>V'" using goal1(3) `v''\<noteq>v` by simp
          note correct(4)[OF goal1(1,2) this goal1(4,5)]
          moreover have "\<omega>'-(\<C>'-{(u',v')})=\<omega>-(\<C>-{(u',v')})"
             using invar_loopc(1-3,8) invar(1,2) by blast
          ultimately show ?thesis by simp
      next
 
        case True
          hence "v\<in>\<Q> \<A>" and "v''\<in>V'" using  `v\<in>V'` and `(v,c,v')\<in>\<Delta> \<A>`
              \<Delta>_consistent by blast+
          thus ?thesis proof (cases "u''=u'")
            case False
              let ?X1 = "{v'''.(v'',c',v''')\<in>\<Delta> \<A> \<and> (u'',v''')\<notin>\<omega> - 
                  (\<C> \<union> {(u',v')})}"
              let ?X2 = "{v'''.(v'',c',v''')\<in>\<Delta> \<A> \<and> (u'',v''')\<notin>\<omega>-\<C>}"
              from False have set_unchanged: "?X1 = ?X2" by blast

              from False have "(c,u',v) \<noteq> (c',u'',v'')" by blast
              note INY_dec_counter_unaffected[OF this, of N'']
              also note correct(3)[OF goal1(1) `v''\<in>V'`goal1(4) `v''\<in>\<Q> \<A>`]
              also note correct_loopc(2)[OF goal1(1) `c'\<in>\<Sigma>'`goal1(4,5)]
              also note correct_while(2)[OF goal1(1,4,5)]
              finally have "fst (INY_dec_counter N'' c u' v) (c',u'',v'') = 
                  Some (card ?X1)" using `v''=v` by simp
              hence "fst (INY_dec_counter N'' c u' v) (c',u'',v'') = 
                  Some (card ?X2)" using set_unchanged by simp
              moreover have "\<omega>-(\<C>-{(u',v')})=\<omega>'-(\<C>'-{(u',v')})"
                 using invar_loopc(1-3,8) invar(1-3) by blast
              ultimately show ?thesis by simp
            next

            case True
              let ?X1 = "{v'''.(v'',c',v''')\<in>\<Delta> \<A> \<and> (u'',v''')\<notin>\<omega>-
                  (\<C> \<union> {(u',v')})}"
              let ?X2 = "{v'''.(v'',c',v''')\<in>\<Delta> \<A> \<and> (u'',v''')\<notin>\<omega>-\<C>}"
              have "?X1 \<subseteq> \<Q> \<A>" "?X2 \<subseteq> \<Q> \<A>" using \<Delta>_consistent by blast+
              hence fin: "finite ?X1" "finite ?X2" using finite_\<Q> 
                  using rev_finite_subset by blast+
              have "?X2 = ?X1 - {v'}" and "v' \<in> ?X1" using invar(2,8) 
                   `u'' = u'` `v'' = v``(v,c,v')\<in>\<Delta> \<A>` `c'=c`
                   invar_loopc(2,8,9) by auto

              hence new_card: "card ?X2 = card ?X1 - 1" 
                  using fin invar_loopc(8) by simp

              from True have param_eq: "(c,u',v) = (c',u'',v'')"
                  using `v''=v` `c'=c` `c\<in>\<Sigma>'` by simp
              note correct(3)[OF goal1(1) `v\<in>V'` `u'\<in>\<Q> \<A>` `v\<in>\<Q> \<A>`]
              also note correct_loopc(2)[OF goal1(1) `c'\<in>\<Sigma>'` `u'\<in>\<Q> \<A>` `v\<in>\<Q> \<A>`]
              also note correct_while(2)[OF goal1(1) `u'\<in>\<Q> \<A>` `v\<in>\<Q> \<A>`]
              finally have "N'' (c',u'',v'') = Some (card ?X1)" 
                  using `u''=u'` `v''=v` by simp
              moreover note INY_dec_counter_correct(1)[OF _ 
                  param_eq, of N'' "card ?X1"]
              ultimately have "fst (INY_dec_counter N'' c u' v) (c',u'',v'') =  
                  Some (card ?X2)" using new_card by simp
              moreover have "\<omega>-(\<C>-{(u',v')})=\<omega>'-(\<C>'-{(u',v')})"
                 using invar_loopc(1-3,8) invar(1-3) by blast
              ultimately show ?thesis using `u''=u'``v''=v` by simp
          qed
      qed 
    qed
qed


lemma INY_abstr5_loopv_invarI2:
  assumes "INY_abstr2_loopc_invar \<omega> \<C> u' v' \<Sigma>' (\<omega>',\<C>')" and
     "INY_abstr3_loopv_invar \<omega>' \<C>' u' v' c V' (\<omega>'', \<C>'')" and
     "INY_abstr5_loopv_N\<delta>r_correct \<omega> \<C> N' \<delta>r u' v' c V' N''"
  shows "INY_abstr5_loopv_invar \<omega>' \<C>' N' \<delta>r u' v' c V' (\<omega>'',\<C>'',N'')"
apply (intro INY_abstr5_loopv_invarI)
using assms(2) apply simp
apply (subgoal_tac "\<omega>'-\<C>' = \<omega>-\<C>")
using assms(3) unfolding INY_abstr5_loopv_N\<delta>r_correct_def apply clarsimp
using INY_abstr2_loopc_invarD(1-3,8)[OF assms(1)] apply blast
done

abbreviation INY_abstr5_refrel_loopv_it where
  "INY_abstr5_refrel_loopv_it \<omega> \<C> N \<delta>r u' v' c \<equiv> 
    br (\<lambda>(it,(\<omega>',\<C>',N')). (it,(\<omega>',\<C>')))
        (\<lambda>(V',(\<omega>',\<C>',N')). INY_abstr5_loopv_N\<delta>r_correct \<omega> \<C> N \<delta>r u' v' c V' N')"

abbreviation INY_abstr5_refrel_loopv where
  "INY_abstr5_refrel_loopv \<omega> \<C> N \<delta>r u' v' \<Sigma>' c \<equiv> 
    br (\<lambda>(\<omega>',\<C>',N'). (\<omega>',\<C>')) (\<lambda>(\<omega>',\<C>',N'). 
    INY_abstr5_loopc_N\<delta>r_correct \<omega> \<C> N \<delta>r u' v' (\<Sigma>'-{c}) N')"

lemma INY_abstr5_loopv_N\<delta>r_correct_transfer:
  assumes "INY_abstr2_loopc_invar \<omega> \<C> u' v' \<Sigma>' (\<omega>',\<C>')"
          "INY_abstr5_loopv_N\<delta>r_correct \<omega> \<C> N' \<delta>r u' v' c it b"
    shows "INY_abstr5_loopv_N\<delta>r_correct \<omega>' \<C>' N' \<delta>r u' v' c it b"
apply (rule INY_abstr5_loopv_N\<delta>r_correctI)
using INY_abstr5_loopv_N\<delta>r_correctD[OF assms(2)] apply (simp, simp, simp)
apply (subgoal_tac "\<omega>'-\<C>' = \<omega>-\<C>")
using INY_abstr5_loopv_N\<delta>r_correctD[OF assms(2)] apply simp
using INY_abstr2_loopc_invarD(1-3,8)[OF assms(1)] apply blast
using INY_abstr5_loopv_N\<delta>r_correctD[OF assms(2)] apply simp
done

lemma INY_abstr5_loopv_N\<delta>r_correct_lift:
  assumes "INY_abstr5_loopc_N\<delta>r_correct \<omega> \<C> N \<delta>r u' v' \<Sigma>' N'"
          "INY_abstr5_loopv_N\<delta>r_correct \<omega> \<C> N' \<delta>r u' v' c {} b"
    shows "INY_abstr5_loopc_N\<delta>r_correct \<omega> \<C> N \<delta>r u' v' (\<Sigma>'-{c}) b"
apply (rule INY_abstr5_loopc_N\<delta>r_correctI)
using INY_abstr5_loopv_N\<delta>r_correctD(1)[OF assms(2)] apply simp
apply (rename_tac c' u'' v, case_tac "c=c'")
using INY_abstr5_loopv_N\<delta>r_correctD(2)[OF assms(2)]
    INY_abstr5_loopc_N\<delta>r_correctD(2)[OF assms(1)] apply simp_all[2]
apply (rename_tac c' u'' v, case_tac "c=c'")
using INY_abstr5_loopv_N\<delta>r_correctD(2,4,5)[OF assms(2)] 
    INY_abstr5_loopc_N\<delta>r_correctD[OF assms(1)] apply simp_all[3]
done


lemma INY_abstr5_loopv_correct:
  notes [simp] = br_def
  assumes "INY_abstr2_loopc_invar \<omega> \<C> u' v' \<Sigma>' (\<omega>',\<C>')"
          "INY_abstr5_loopc_N\<delta>r_correct \<omega> \<C> N \<delta>r u' v' \<Sigma>' N'"
          "INY_abstr1_invar (\<omega>,\<C> \<union> {(u',v')})"
          "INY_abstr5_N\<delta>r_correct \<omega> (\<C>\<union>{(u',v')}) N \<delta>r" "c\<in>\<Sigma>'" "\<Sigma>'\<subseteq>\<Sigma> \<A>"
  shows "INY_abstr5_loopv \<omega>' \<C>' N' \<delta>r u' v' c \<le> 
         \<Down>(INY_abstr5_refrel_loopv \<omega> \<C> N \<delta>r u' v' \<Sigma>' c)
         (INY_abstr4_loopv \<omega>' \<C>' u' v' c)"
unfolding INY_abstr5_loopv_def INY_abstr4_loopv_def
apply (refine_rcg 
  inj_on_id FOREACHi_refine_genR[where 
      R = "INY_abstr5_refrel_loopv_it \<omega> \<C> N' \<delta>r u' v' c"]
)
using INY_abstr5_\<delta>r_correct(2)[OF assms(1,4,5,6)] apply simp
apply (simp add: INY_abstr5_\<delta>r_correct(2)[OF assms(1,4,5,6)]
       INY_abstr5_loopv_N\<delta>r_correct_initial[OF assms(2,1,5,4)])
using INY_abstr5_loopv_N\<delta>r_correct_transfer[OF assms(1)] 
    apply (clarsimp, unfold INY_abstr5_loopv_invar_def, simp)
using assms apply (force dest!: INY_abstr5_loopv_N_eq_0_iff)
apply clarsimp
apply (rule INY_abstr5_loopu_correct[OF assms(1,4,5,6)])
apply (simp add: single_valued_def)
using assms apply (force dest!: INY_abstr5_loopv_N_correctness_preserved)
apply (simp add: single_valued_def)
using assms apply (force dest!: INY_abstr5_loopv_N_correctness_preserved)
using INY_abstr5_loopv_N\<delta>r_correct_lift[OF assms(2)] apply clarsimp
done


lemma INY_abstr5_loopc_N\<delta>r_correct_initial:
  assumes "INY_abstr5_N\<delta>r_correct \<omega> (\<C>\<union>{(u',v')}) N \<delta>r"
  shows "INY_abstr5_loopc_N\<delta>r_correct \<omega> \<C> N \<delta>r u' v' (\<Sigma> \<A>) N"
apply (rule INY_abstr5_loopc_N\<delta>r_correctI)
using INY_abstr5_N\<delta>r_correctD[OF assms] apply simp_all
done

definition INY_abstr5_loopc where
"INY_abstr5_loopc \<omega> \<C> N \<delta>r u' v' \<equiv> FOREACH\<^bsup>INY_abstr5_loopc_invar \<omega> \<C> N \<delta>r u' v'\<^esup> 
    (\<Sigma> \<A>) (\<lambda>c (\<omega>, \<C>, N). do {
        (\<omega>',\<C>',N') \<leftarrow> INY_abstr5_loopv \<omega> \<C> N \<delta>r u' v' c;
        RETURN (\<omega>',\<C>',N')
     }) (\<omega>, \<C>, N)"

abbreviation INY_abstr5_refrel_loopc_it where
  "INY_abstr5_refrel_loopc_it \<omega> \<C> N \<delta>r u' v' \<equiv> 
    br (\<lambda>(it,(\<omega>',\<C>',N')). (it,(\<omega>',\<C>')))
        (\<lambda>(\<Sigma>',(\<omega>',\<C>',N')). INY_abstr5_loopc_N\<delta>r_correct \<omega> \<C> N \<delta>r u' v' \<Sigma>' N')"

abbreviation INY_abstr5_refrel_loopc where
  "INY_abstr5_refrel_loopc \<omega> \<C> N \<delta>r u' v' \<equiv> 
    br (\<lambda>(\<omega>',\<C>',N'). (\<omega>',\<C>')) (\<lambda>(\<omega>',\<C>',N'). 
    INY_abstr5_N\<delta>r_correct \<omega>' \<C>' N' \<delta>r)"

lemma INY_abstr5_loopc_N\<delta>r_correct_lift[intro]:
  assumes "INY_abstr5_N\<delta>r_correct \<omega> (\<C>\<union>{(u',v')}) N \<delta>r"
          "INY_abstr5_loopc_N\<delta>r_correct \<omega> \<C> N \<delta>r u' v' {} N'"
    shows "INY_abstr5_N\<delta>r_correct \<omega> \<C> N' \<delta>r"
using INY_abstr5_loopc_N\<delta>r_correctD[OF assms(2)] 
      INY_abstr5_N\<delta>r_correctD(2)[OF assms(1)] 
   by (intro INY_abstr5_N\<delta>r_correctI, simp_all)

lemma INY_abstr5_N\<delta>r_correct_transfer:
  assumes "INY_abstr2_loopc_invar \<omega> \<C> u' v' \<Sigma>' (\<omega>',\<C>')"
          "INY_abstr5_N\<delta>r_correct \<omega> \<C> N' \<delta>r"
    shows "INY_abstr5_N\<delta>r_correct \<omega>' \<C>' N' \<delta>r"
apply (rule INY_abstr5_N\<delta>r_correctI)
using INY_abstr5_N\<delta>r_correctD[OF assms(2)] apply simp
apply (subgoal_tac "\<omega>-\<C> = \<omega>'-\<C>'")
using INY_abstr5_N\<delta>r_correctD[OF assms(2)] apply simp
using INY_abstr2_loopc_invarD(1-3,8)[OF assms(1)] apply blast
using INY_abstr5_N\<delta>r_correctD[OF assms(2)] apply simp
done


lemma INY_abstr5_loopc_correct:
  notes [simp] = br_def
  assumes "INY_abstr1_invar (\<omega>,\<C>\<union>{(u',v')})" 
      "INY_abstr5_N\<delta>r_correct \<omega> (\<C>\<union>{(u',v')}) N \<delta>r" 
  shows "INY_abstr5_loopc \<omega> \<C> N \<delta>r u' v' \<le> 
        \<Down>(INY_abstr5_refrel_loopc \<omega> \<C> N \<delta>r u' v')
        (INY_abstr4_loopc \<omega> \<C> u' v')"
  unfolding INY_abstr5_loopc_def INY_abstr4_loopc_def
apply (refine_rcg 
  inj_on_id FOREACHi_refine_genR[where R = 
      "INY_abstr5_refrel_loopc_it \<omega> \<C> N \<delta>r u' v'"]
)
apply simp
using assms INY_abstr5_loopc_N\<delta>r_correct_initial apply simp
apply (clarsimp simp add: INY_abstr5_loopc_invarI)
apply clarsimp
apply (rule INY_abstr5_loopv_correct[OF _ _ assms(1,2)], assumption+)
apply (simp add: single_valued_def)
using INY_abstr5_loopc_N\<delta>r_correct_lift[OF assms(2)] apply clarsimp
using INY_abstr5_loopc_N\<delta>r_correct_lift[OF assms(2)]
      INY_abstr5_N\<delta>r_correct_transfer apply clarsimp
done

lemma INY_abstr5_loopc_correct':
  assumes "INY_abstr1_invar (\<omega>,\<C>)" 
      "INY_abstr5_N\<delta>r_correct \<omega> \<C> N \<delta>r" "(u',v') \<in> \<C>"
  shows "INY_abstr5_loopc \<omega> (\<C>-{(u',v')}) N \<delta>r u' v' \<le> 
        \<Down>(INY_abstr5_refrel_loopc \<omega> (\<C>-{(u',v')}) N \<delta>r u' v')
        (INY_abstr4_loopc \<omega> (\<C>-{(u',v')}) u' v')"
proof-
  have "\<C> - {(u',v')} \<union> {(u',v')} = \<C>" using assms(3) by blast
  thus ?thesis using INY_abstr5_loopc_correct assms by simp
qed


definition INY_abstr5' where
"INY_abstr5' \<omega> \<C> N \<delta>r \<equiv> WHILE\<^sub>T\<^bsup>INY_abstr5_invar \<delta>r\<^esup> (\<lambda>(\<omega>, \<C>, N). \<C>\<noteq>{}) 
  (\<lambda>(\<omega>, \<C>, N). do {
      ASSERT (\<C>\<noteq>{});
      (u',v') \<leftarrow> SPEC (\<lambda>(u',v'). (u',v')\<in>\<C>);
      let \<C> = \<C> - {(u',v')};
      (\<omega>,\<C>,N) \<leftarrow> INY_abstr5_loopc \<omega> \<C> N \<delta>r u' v';
      RETURN (\<omega>, \<C>, N)
  }) (\<omega>, \<C>, N)"

abbreviation INY_abstr5_refrel where
"INY_abstr5_refrel \<delta>r \<equiv> br (\<lambda>(\<omega>,\<C>,_). (\<omega>,\<C>)) (\<lambda>(\<omega>,\<C>,N). 
    INY_abstr5_N\<delta>r_correct \<omega> \<C> N \<delta>r)"


lemma INY_abstr5'_correct:
  notes [simp] = br_def
  assumes "INY_is_valid_initial_counter N"
          "INY_abstr5_\<delta>r_correct \<delta>r" "\<omega> = INY_initial"
  shows "INY_abstr5' \<omega> \<omega> N \<delta>r \<le> \<Down>(INY_abstr5_refrel \<delta>r) INY_abstr4"
unfolding INY_abstr5'_def INY_abstr4_def
apply (refine_rcg)
using assms unfolding INY_is_valid_initial_counter_def 
    apply (simp, intro INY_abstr5_N\<delta>r_correctI, simp, simp, fast, simp)
apply simp
using INY_abstr5_invarI apply clarsimp
apply simp
apply simp
apply simp
apply clarsimp
apply (rule INY_abstr5_loopc_correct', assumption+)
apply simp
done


abbreviation INY_is_empty_d where
"INY_is_empty_d d \<equiv> dom d = \<Q> \<A>\<times>\<Sigma> \<A> \<and> (\<forall>(q,a)\<in>dom d. d(q,a) = Some (0::nat))"
abbreviation INY_is_empty_\<delta>r where
"INY_is_empty_\<delta>r \<delta>r \<equiv> dom \<delta>r = \<Q> \<A>\<times>\<Sigma> \<A> \<and> 
    (\<forall>(q,a)\<in>dom \<delta>r. \<delta>r(q,a) = Some ({}::'q set))"

definition INY_abstr5 where
"INY_abstr5 \<equiv> do {
  (d,\<delta>r) \<leftarrow> SPEC (\<lambda>(d,\<delta>r). INY_is_empty_d d \<and> INY_is_empty_\<delta>r \<delta>r);
  (d,\<delta>r) \<leftarrow> SPEC (\<lambda>(d,\<delta>r). INY_abstr5_d_correct d \<and> INY_abstr5_\<delta>r_correct \<delta>r);
  N \<leftarrow> SPEC (\<lambda>N. INY_is_valid_initial_counter N);
  (\<omega>,\<C>) \<leftarrow> SPEC (\<lambda>(\<omega>,\<C>). \<omega> = INY_initial \<and> \<C> = INY_initial);
  (\<omega>,\<C>,N) \<leftarrow> INY_abstr5' \<omega> \<C> N \<delta>r;
  RETURN (\<omega>,\<C>)
}"

lemma INY_abstr5_correct: "INY_abstr5 \<le> \<Down>(Id) INY_abstr4"
  unfolding INY_abstr5_def
  apply refine_rcg
  using INY_abstr5'_correct
  apply (simp add: pw_le_iff refine_pw_simps br_def)
  apply force
  done


subsection {* Implementation of the initialisation *}


text {* @{term "INY_abstr6_empty_Nd\<delta>r"} returns @{term "N"} and @{term "d"} filled 
  with 0 and @{term "\<delta>r"} filled with the empty set for each value in their 
  respective domains.
*}
abbreviation INY_abstr6_empty_d\<delta>r_invar_loopc where
"INY_abstr6_empty_d\<delta>r_invar_loopc \<Sigma>' \<equiv> \<lambda>(d,\<delta>r).
    dom d = \<Q> \<A>\<times>(\<Sigma> \<A>-\<Sigma>') \<and> (\<forall>x\<in>dom d. d x = Some (0::nat)) \<and>
    dom \<delta>r = \<Q> \<A>\<times>(\<Sigma> \<A>-\<Sigma>') \<and> (\<forall>x\<in>dom \<delta>r. \<delta>r x = Some ({}::'q set))"

abbreviation INY_abstr6_empty_d\<delta>r_invar_loopu where
"INY_abstr6_empty_d\<delta>r_invar_loopu d \<delta>r c U' \<equiv> \<lambda>(d',\<delta>r').
    dom d' = dom d \<union> (\<Q> \<A>-U')\<times>{c} \<and> (\<forall>x\<in>dom d'. d' x = Some (0::nat)) \<and>
    dom \<delta>r' = dom \<delta>r \<union> (\<Q> \<A>-U')\<times>{c} \<and> (\<forall>x\<in>dom \<delta>r'. \<delta>r' x = Some ({}::'q set))"

definition INY_abstr6_empty_d\<delta>r where
"INY_abstr6_empty_d\<delta>r \<equiv> FOREACH\<^bsup>INY_abstr6_empty_d\<delta>r_invar_loopc\<^esup> (\<Sigma> \<A>) (\<lambda>c (d,\<delta>r).
    FOREACH\<^bsup>INY_abstr6_empty_d\<delta>r_invar_loopu d \<delta>r c\<^esup> (\<Q> \<A>) (\<lambda>u (d,\<delta>r). 
      RETURN (d((u,c) \<mapsto> 0::nat), \<delta>r((u,c) \<mapsto> {}::'q set))
    ) (d,\<delta>r)
  ) ( Map.empty, Map.empty)"


lemma INY_abstr6_empty_d\<delta>r_correct:
  "INY_abstr6_empty_d\<delta>r \<le>  SPEC (\<lambda>(d,\<delta>r). INY_is_empty_d d \<and> 
     INY_is_empty_\<delta>r \<delta>r)"
unfolding INY_abstr6_empty_d\<delta>r_def
apply (intro refine_vcg)
apply (simp_all add: finite_\<Q> finite_\<Sigma>)[5]
apply (clarsimp, blast)
apply (clarsimp, blast)
apply (simp add: INY_is_valid_initial_counter_def)
done


text {* @{term "INY_abstr6_init_d\<delta>r"} fills @{text "d(u,c)"} with @{text "|\<delta>(u,c)|"} 
  for all @{text "u\<in>\<Q>"}, @{text "c\<in>\<Sigma>"} and @{text "\<delta>r(u,c)"} with all @{text "\<delta>\<inverse>(u,c)"}
  for all @{text "u\<in>\<Q>"}, @{text "c\<in>\<Sigma>"}, i.e. the set @{text "{u'. (u',c,u)\<in>\<Delta>}"}.
*}

definition INY_abstr6_init_d\<delta>r_invar where
"INY_abstr6_init_d\<delta>r_invar \<Delta>' \<equiv> \<lambda>(d,\<delta>r). (dom d = \<Q> \<A>\<times>\<Sigma> \<A>) \<and>
    (\<forall>(v,c)\<in>dom d. d (v,c) = Some (card {v'. (v,c,v')\<in>\<Delta> \<A>-\<Delta>'})) \<and>
    (dom \<delta>r = \<Q> \<A>\<times>\<Sigma> \<A>) \<and>
    (\<forall>(v,c)\<in>dom \<delta>r. \<delta>r (v,c) = Some {v'. (v',c,v)\<in>\<Delta> \<A>-\<Delta>'})"

lemma INY_abstr6_init_d\<delta>r_invarI[intro]:
  assumes "dom d = \<Q> \<A>\<times>\<Sigma> \<A>"
      "\<And>v c. (v,c) \<in> \<Q> \<A>\<times>\<Sigma> \<A> \<Longrightarrow> d (v,c) = 
             Some (card {v'. (v,c,v')\<in>\<Delta> \<A>-\<Delta>'})"
      "dom \<delta>r = \<Q> \<A>\<times>\<Sigma> \<A>"
      "\<And>v c. (v,c) \<in> \<Q> \<A>\<times>\<Sigma> \<A> \<Longrightarrow> \<delta>r (v,c) = Some {v'. (v',c,v)\<in>\<Delta> \<A>-\<Delta>'}"
  shows "INY_abstr6_init_d\<delta>r_invar \<Delta>' (d,\<delta>r)"
  using assms unfolding INY_abstr6_init_d\<delta>r_invar_def by simp

lemma INY_abstr6_init_d\<delta>r_invarD:
  assumes "INY_abstr6_init_d\<delta>r_invar \<Delta>' (d,\<delta>r)"
  shows "dom d = \<Q> \<A>\<times>\<Sigma> \<A>"
      "\<And>v c. (v,c) \<in> \<Q> \<A>\<times>\<Sigma> \<A> \<Longrightarrow> d (v,c) = Some (card {v'. (v,c,v')\<in>\<Delta> \<A>-\<Delta>'})"
      "dom \<delta>r = \<Q> \<A>\<times>\<Sigma> \<A>"
      "\<And>v c. (v,c) \<in> \<Q> \<A>\<times>\<Sigma> \<A> \<Longrightarrow> \<delta>r (v,c) = Some {v'. (v',c,v)\<in>\<Delta> \<A>-\<Delta>'}"
  using assms unfolding INY_abstr6_init_d\<delta>r_invar_def by blast+

abbreviation INY_abstr6_inc_d where
"INY_abstr6_inc_d d v c \<equiv>
    (case d (v,c) of Some n \<Rightarrow> d((v,c) \<mapsto> n+(1::nat)) | None \<Rightarrow> d)"

abbreviation INY_abstr6_update_\<delta>r where
"INY_abstr6_update_\<delta>r \<delta>r v' c v \<equiv>
    (case \<delta>r (v,c) of Some A \<Rightarrow> \<delta>r((v,c) \<mapsto> insert v' A) | None \<Rightarrow> \<delta>r)"

text {* The initialisation of @{term "d"} and @{term "\<delta>r"} *}
definition INY_abstr6_init_d\<delta>r where
"INY_abstr6_init_d\<delta>r d \<delta>r = FOREACH\<^bsup>INY_abstr6_init_d\<delta>r_invar\<^esup> (\<Delta> \<A>) 
    (\<lambda>(v,c,v') (d,\<delta>r). RETURN (INY_abstr6_inc_d d v c, 
                               INY_abstr6_update_\<delta>r \<delta>r v c v'))
    (d, \<delta>r)"

lemma INY_abstr6_init_d\<delta>r_correct:
  assumes "dom d = \<Q> \<A>\<times>\<Sigma> \<A> \<and> (\<forall>x\<in>dom d. d x = Some 0)"
          "dom \<delta>r = \<Q> \<A>\<times>\<Sigma> \<A> \<and> (\<forall>x\<in>dom \<delta>r. \<delta>r x = Some {})"
  shows  "INY_abstr6_init_d\<delta>r d \<delta>r \<le> 
      SPEC (\<lambda>(d,\<delta>r). INY_abstr5_d_correct d \<and> INY_abstr5_\<delta>r_correct \<delta>r)"
unfolding INY_abstr6_init_d\<delta>r_def
proof (intro refine_vcg)
  show "finite (\<Delta> \<A>)" using finite_\<Q> finite_\<Sigma> \<Delta>_consistent
      rev_finite_subset[of "\<Q> \<A>\<times>\<Sigma> \<A>\<times>\<Q> \<A>" "\<Delta> \<A>"] by force
next
  show "INY_abstr6_init_d\<delta>r_invar (\<Delta> \<A>) (d,\<delta>r)"
      using assms by (intro INY_abstr6_init_d\<delta>r_invarI, simp_all)
next
  case (goal3 vcv' \<Delta>' d\<delta>r)
    thus ?case proof (cases vcv', cases d\<delta>r, clarsimp)
    case (goal1 v c v' d \<delta>r)
    note invar = INY_abstr6_init_d\<delta>r_invarD[OF goal1(2)]
    let ?d' = "INY_abstr6_inc_d d v c"
    let ?\<delta>r' = "INY_abstr6_update_\<delta>r \<delta>r v c v'"

    let ?X1 = "{v''. (v,c,v'')\<in>\<Delta> \<A>-\<Delta>'}"
    let ?X2 = "{v''. (v,c,v'')\<in>\<Delta> \<A>-(\<Delta>'-{(v,c,v')})}"
    have "?X2 = ?X1 \<union> {v'}" "v'\<notin> ?X1" using goal1 by blast+
    moreover have "?X2 \<subseteq> \<Q> \<A>" using \<Delta>_consistent by blast+
    hence "finite ?X2" using rev_finite_subset[OF finite_\<Q>] by blast+
    ultimately have new_card: "card ?X2 = card ?X1 + 1" by simp

    let ?Y1 = "{v''. (v'',c,v')\<in>\<Delta> \<A>-\<Delta>'}"
    let ?Y2 = "{v''. (v'',c,v')\<in>\<Delta> \<A>-(\<Delta>'-{(v,c,v')})}"
    have new_set: "?Y2 = ?Y1 \<union> {v}" using goal1 by blast
    have "d (v,c) = Some (card {v'. (v,c,v')\<in>\<Delta> \<A>-\<Delta>'})" 
        using goal1(1,5) invar(2) \<Delta>_consistent by blast
    moreover have "\<delta>r (v',c) = Some {v''. (v'',c,v')\<in>\<Delta> \<A>-\<Delta>'}"
        using goal1(1,5) invar(4) \<Delta>_consistent by blast
    ultimately show "INY_abstr6_init_d\<delta>r_invar (\<Delta>'-{(v,c,v')}) (?d',?\<delta>r')" 
      using goal1(1,2) apply (intro INY_abstr6_init_d\<delta>r_invarI)
      using invar(1) \<Delta>_consistent apply (simp, blast)
      using invar(2) new_card apply fastforce
      using invar(3) \<Delta>_consistent apply (simp, blast)
      using invar(4) new_set apply fastforce
    done
  qed

next
  case goal4 thus ?case
  proof (clarsimp)
    fix d \<delta>r assume "INY_abstr6_init_d\<delta>r_invar {} (d,\<delta>r)"
    thus "INY_abstr5_d_correct d \<and> INY_abstr5_\<delta>r_correct \<delta>r" 
        unfolding INY_abstr5_d_correct_def INY_abstr5_\<delta>r_correct_def
        INY_abstr6_init_d\<delta>r_invar_def by simp
  qed
qed


text {* The initialisation of @{term "N"}. *}
abbreviation INY_abstr6_init_N_invar_loopc where
"INY_abstr6_init_N_invar_loopc \<Sigma>' N \<equiv> (dom N = (\<Sigma> \<A>-\<Sigma>')\<times>\<Q> \<A>\<times>\<Q> \<A> \<and> 
    (\<forall>(c,u,v)\<in>dom N. N(c,u,v) = Some (card {v'. (v,c,v')\<in>\<Delta> \<A>})))"

abbreviation INY_abstr6_init_N_invar_loopu where
"INY_abstr6_init_N_invar_loopu N c U' N' \<equiv> 
    (dom N' = dom N \<union> {c}\<times>(\<Q> \<A>-U')\<times>\<Q> \<A> \<and> 
    (\<forall>(c,u,v)\<in>dom N'. N'(c,u,v) = Some (card {v'. (v,c,v')\<in>\<Delta> \<A>})))"

abbreviation INY_abstr6_init_N_invar_loopv where
"INY_abstr6_init_N_invar_loopv N c u V' N' \<equiv> (dom N' = 
    dom N \<union> {c}\<times>{u}\<times>(\<Q> \<A>-V') \<and>
    (\<forall>(c,u,v)\<in>dom N'. N'(c,u,v) = Some (card {v'. (v,c,v')\<in>\<Delta> \<A>})))"

definition INY_abstr6_init_N where
"INY_abstr6_init_N d \<equiv> FOREACH\<^bsup>INY_abstr6_init_N_invar_loopc\<^esup> (\<Sigma> \<A>) (\<lambda>c N.
    FOREACH\<^bsup>INY_abstr6_init_N_invar_loopu N c\<^esup> (\<Q> \<A>) (\<lambda>u N. 
      FOREACH\<^bsup>INY_abstr6_init_N_invar_loopv N c u\<^esup> (\<Q> \<A>) (\<lambda>v N. do {
          ASSERT (d(v,c) \<noteq> None);
          RETURN (N((c,u,v) \<mapsto> (the (d(v,c)))::nat))
        }
      ) N
    ) N
  ) Map.empty"

thm INY_abstr5_d_correctD
lemma INY_abstr5_d_correctD_ne_None_aux:
  assumes C: "INY_abstr5_d_correct d"
  assumes M: "v\<in>\<Q> \<A>"  "c\<in>\<Sigma> \<A>"
  shows "d (v,c) \<noteq> None"
  using INY_abstr5_d_correctD[OF C] assms
  by blast

lemma INY_abstr6_init_N_correct:
  assumes "INY_abstr5_d_correct d"
  shows "INY_abstr6_init_N d \<le> SPEC INY_is_valid_initial_counter"
  unfolding INY_abstr6_init_N_def
apply (intro refine_vcg)
apply (simp_all add: finite_\<Sigma> finite_\<Q>)[6]
apply (rule INY_abstr5_d_correctD_ne_None_aux[OF assms], (erule (1) set_mp)+) []
defer
using INY_abstr5_d_correctD[OF assms] apply auto[2]
apply (simp add: INY_is_valid_initial_counter_def)
proof-
  thm INY_abstr5_d_correctD[OF assms, simplified, rule_format]
  case (goal8 c \<Sigma>' N u U' N' v V' N'')
    def N''' == "N''((c, u, v) \<mapsto> the (d (v, c)))"
    note this[THEN meta_eq_to_obj_eq]
    also have "(v,c) \<in> \<Q> \<A> \<times> \<Sigma> \<A>" using goal8(1,2,7,8) by blast
    hence "d (v, c) = Some (card {v'. (v,c,v')\<in>\<Delta> \<A>})"
      using INY_abstr5_d_correctD[OF assms] by simp
    finally have A: "N''' = N''((c,u,v)\<mapsto>(card {v'. (v,c,v')\<in>\<Delta> \<A>}))" by force
    have "dom N''' = dom N' \<union> {c} \<times> {u} \<times> (\<Q> \<A> - (V' - {v}))"
        using goal8(1-9) by (subst A, auto)
    moreover have "\<And>n c u v. N'' (c,u,v) = Some n \<Longrightarrow> 
        N'' (c,u,v) = Some (card {v'. (v,c,v')\<in>\<Delta> \<A>})" using goal8(9) by fast
    hence C: "\<And>n c u v. N'' (c,u,v) = Some n \<Longrightarrow> 
        n = (card {v'. (v,c,v')\<in>\<Delta> \<A>})" by simp
    have  "\<forall>(c, u, v)\<in>dom N'''. N''' (c, u, v) =
        Some (card {v'. (v, c, v') \<in> \<Delta> \<A>})" by (clarsimp simp add: A C)
    ultimately show ?case by (simp add: N'''_def)
qed


text {* Now we compute the nontrivial part of the initial @{term "\<omega>"}, i.e.
    @{text "{(u,v)| \<exists>c. \<delta>(u,c) \<noteq> {} \<and> \<delta>(v,c) = {}}"}
    For this, we iterate over all @{term "c"}, @{term "u"}, @{term "v"} with for each 
    loops, checking the @{text "\<delta>(u,c)"} / @{text "\<delta>(v,c)"} conditions as soon as 
    possible. *}
definition INY_abstr6_init_\<omega>\<C>_loopc_invar where
"INY_abstr6_init_\<omega>\<C>_loopc_invar (d::('q\<times>'a)\<rightharpoonup>nat) \<Sigma>' \<equiv> \<lambda>(\<omega>,\<C>).
    (\<omega> = {(u,v) |u v c. u\<in>\<Q> \<A> \<and> v\<in>\<Q> \<A> \<and> c\<in>(\<Sigma> \<A>-\<Sigma>') \<and> 
        d (u,c)\<noteq>Some 0 \<and> d (v,c)=Some 0} \<and> \<C> = \<omega>)"

definition INY_abstr6_init_\<omega>\<C>_loopu_invar where
"INY_abstr6_init_\<omega>\<C>_loopu_invar (d::('q\<times>'a)\<rightharpoonup>nat) c \<omega> U' \<equiv> \<lambda>(\<omega>',\<C>').
    (\<omega>' = \<omega> \<union> {(u,v) |u v. u\<in>\<Q> \<A>-U' \<and> v\<in>\<Q> \<A> \<and>
    d (u,c)\<noteq>Some 0 \<and> d (v,c)=Some 0} \<and> \<C>' = \<omega>')"

definition INY_abstr6_init_\<omega>\<C>_loopv_invar where
"INY_abstr6_init_\<omega>\<C>_loopv_invar (d::('q\<times>'a)\<rightharpoonup>nat) c u \<omega> V' \<equiv> \<lambda>(\<omega>',\<C>').
    (\<omega>' = \<omega> \<union> {(u,v) |v. v\<in>\<Q> \<A>-V' \<and> d (v,c)=Some 0} \<and> \<C>' = \<omega>')"

definition INY_abstr6_init_\<omega>\<C>_loopv where
"INY_abstr6_init_\<omega>\<C>_loopv d c u \<omega> \<C> \<equiv> 
    FOREACH\<^bsup>INY_abstr6_init_\<omega>\<C>_loopv_invar d c u \<omega>\<^esup> (\<Q> \<A>)
      (\<lambda>v (\<omega>',\<C>'). if d (v, c) = Some 0 then 
        RETURN (insert (u,v) \<omega>', insert (u,v) \<C>') 
      else RETURN (\<omega>',\<C>')) (\<omega>,\<C>)"

definition INY_abstr6_init_\<omega>\<C>_loopu where
"INY_abstr6_init_\<omega>\<C>_loopu d c \<omega> \<C> \<equiv>
    FOREACH\<^bsup>INY_abstr6_init_\<omega>\<C>_loopu_invar d c \<omega>\<^esup> (\<Q> \<A>)
      (\<lambda>u (\<omega>',\<C>'). if d (u, c) \<noteq> Some 0 
        then INY_abstr6_init_\<omega>\<C>_loopv d c u \<omega>' \<C>'
        else RETURN (\<omega>',\<C>'))
      (\<omega>,\<C>)"

definition INY_abstr6_init_\<omega>\<C>_loopc where
"INY_abstr6_init_\<omega>\<C>_loopc d \<equiv>
    FOREACH\<^bsup>INY_abstr6_init_\<omega>\<C>_loopc_invar d\<^esup> (\<Sigma> \<A>)
      (\<lambda>c (\<omega>,\<C>). INY_abstr6_init_\<omega>\<C>_loopu d c \<omega> \<C>)
      ({}, {})"

text {* This computes the initial @{term "\<omega>"} and @{term "\<C>"} 
    using the precomputed @{text "|\<delta>(u,c)|"} values. *}
definition INY_abstr6_init_\<omega>\<C> where
"INY_abstr6_init_\<omega>\<C> d \<equiv> do {
  (\<omega>,\<C>) \<leftarrow> INY_abstr6_init_\<omega>\<C>_loopc d;
  let FN = \<F> \<A> \<times> (\<Q> \<A> - \<F> \<A>);
  ASSERT (\<omega> \<union> FN = INY_initial);
  ASSERT (\<C> \<union> FN = INY_initial);
  RETURN (\<omega> \<union> FN, \<C> \<union> FN)
}"

lemma INY_abstr6_init_\<omega>\<C>_loopv_correct:
  assumes "u\<in>U'" "U' \<subseteq> \<Q> \<A>" "INY_abstr6_init_\<omega>\<C>_loopu_invar d c \<omega> U' (\<omega>',\<C>')"
          "d (u,c) \<noteq> Some 0"
  shows "INY_abstr6_init_\<omega>\<C>_loopv d c u \<omega>' \<C>' \<le> SPEC
             (INY_abstr6_init_\<omega>\<C>_loopu_invar d c \<omega> (U'-{u}))"
proof-
  let ?T = "{(u,v) |v. v\<in>\<Q> \<A> \<and> d(v,c)=Some 0}"
  from assms(3) have [simp]: "\<C>' = \<omega>'" 
      unfolding INY_abstr6_init_\<omega>\<C>_loopu_invar_def by simp
  hence "INY_abstr6_init_\<omega>\<C>_loopv d c u \<omega>' \<C>' \<le> 
      SPEC (\<lambda>(\<omega>'',\<C>''). \<omega>'' = \<omega>' \<union> ?T \<and> \<C>'' = \<omega>'')"
      unfolding INY_abstr6_init_\<omega>\<C>_loopv_def
      by (intro refine_vcg, auto simp add: 
          INY_abstr6_init_\<omega>\<C>_loopv_invar_def finite_\<Q>)
  also have "INY_abstr6_init_\<omega>\<C>_loopu_invar d c \<omega> (U' - {u}) (\<omega>' \<union> ?T, \<C>' \<union> ?T)"
      using assms unfolding INY_abstr6_init_\<omega>\<C>_loopu_invar_def by blast
  hence "SPEC (\<lambda>(\<omega>'',\<C>''). \<omega>'' = \<omega>' \<union> ?T \<and> \<C>'' = \<omega>'') \<le>  SPEC (
      INY_abstr6_init_\<omega>\<C>_loopu_invar d c \<omega> (U'-{u}))" using SPEC_rule by force
  finally show ?thesis .
qed
     
lemma INY_abstr6_init_\<omega>\<C>_loopu_correct:
  assumes "c\<in>\<Sigma>'" "\<Sigma>' \<subseteq> \<Sigma> \<A>" "INY_abstr6_init_\<omega>\<C>_loopc_invar d \<Sigma>' (\<omega>',\<C>')"
  shows "INY_abstr6_init_\<omega>\<C>_loopu d c \<omega>' \<C>' \<le> SPEC
             (INY_abstr6_init_\<omega>\<C>_loopc_invar d (\<Sigma>'-{c}))"
proof-
  let ?T = "{(u,v) |u v. u\<in>\<Q> \<A> \<and> v\<in>\<Q> \<A> \<and> d(u,c)\<noteq>Some 0 \<and> d(v,c)=Some 0}"
  from assms(3) have [simp]: "\<C>' = \<omega>'" 
      unfolding INY_abstr6_init_\<omega>\<C>_loopc_invar_def by simp
  have "INY_abstr6_init_\<omega>\<C>_loopu d c \<omega>' \<C>' \<le>
      SPEC (\<lambda>(\<omega>'',\<C>''). \<omega>'' = \<omega>' \<union> ?T \<and> \<C>'' = \<omega>'')"
      unfolding INY_abstr6_init_\<omega>\<C>_loopu_def
        apply (intro refine_vcg finite_\<Q>)
        apply (simp add: INY_abstr6_init_\<omega>\<C>_loopu_invar_def)
        apply (intro INY_abstr6_init_\<omega>\<C>_loopv_correct)
        apply (auto simp add: INY_abstr6_init_\<omega>\<C>_loopu_invar_def)
        done
  also have "INY_abstr6_init_\<omega>\<C>_loopc_invar d (\<Sigma>' - {c}) (\<omega>' \<union> ?T, \<C>' \<union> ?T)"
      using assms unfolding INY_abstr6_init_\<omega>\<C>_loopc_invar_def by blast
  hence "SPEC (\<lambda>(\<omega>'',\<C>''). \<omega>'' = \<omega>' \<union> ?T \<and> \<C>'' = \<omega>'') \<le>
      SPEC (INY_abstr6_init_\<omega>\<C>_loopc_invar d (\<Sigma>'-{c}))" using SPEC_rule by auto
  finally show ?thesis .
qed

lemma INY_abstr6_init_\<omega>_loopc_correct:
  "INY_abstr6_init_\<omega>\<C>_loopc d \<le> SPEC (\<lambda>(\<omega>,\<C>). \<omega>={(u,v) |c u v. 
      u\<in>\<Q> \<A> \<and> v\<in>\<Q> \<A> \<and> c\<in>\<Sigma> \<A> \<and> d(u,c)\<noteq>Some 0 \<and> d(v,c)=Some 0} \<and> \<C> = \<omega>)"
  unfolding INY_abstr6_init_\<omega>\<C>_loopc_def INY_initial_def
apply (intro refine_vcg finite_\<Sigma>)
apply (simp add: INY_abstr6_init_\<omega>\<C>_loopc_invar_def)
apply (intro INY_abstr6_init_\<omega>\<C>_loopu_correct)
apply (auto simp: INY_abstr6_init_\<omega>\<C>_loopc_invar_def)
done

lemma INY_abstr6_init_\<omega>\<C>_loopc_correct_aux:
  assumes "INY_abstr5_d_correct d"
  shows "\<F> \<A> \<times> (\<Q> \<A> - \<F> \<A>) \<union> {(u,v) |c u v. u\<in>\<Q> \<A> \<and> v\<in>\<Q> \<A> \<and> 
      c\<in>\<Sigma> \<A> \<and> d(u,c)\<noteq>Some 0 \<and> d(v,c)=Some 0}  =  INY_initial"
      (is "?A \<union> ?B = INY_initial")
proof-
  let ?C = "{(u,v). u \<in> \<Q> \<A> \<and> v \<in> \<Q> \<A> \<and> 
      (\<exists>c u'. (u,c,u')\<in>\<Delta> \<A> \<and> \<not>(\<exists>v'.(v,c,v')\<in>\<Delta> \<A>))}"
  {
    fix q c assume "q\<in>\<Q> \<A>" "c\<in>\<Sigma> \<A>"
    hence "d(q,c) = Some (card {q'. (q,c,q')\<in>\<Delta> \<A>})" using assms
       unfolding INY_abstr5_d_correct_def by blast
    also have "{q'. (q,c,q')\<in>\<Delta> \<A>} \<subseteq> \<Q> \<A>" using \<Delta>_consistent by blast
    hence "finite {q'. (q,c,q')\<in>\<Delta> \<A>}" 
        using rev_finite_subset finite_\<Q> by blast
    hence "Some (card {q'. (q,c,q')\<in>\<Delta> \<A>}) = Some 0 \<longleftrightarrow> 
        \<not>(\<exists>q'. (q,c,q')\<in>\<Delta> \<A>)" by simp
    finally have "(d (q,c) = Some 0) = (\<not>(\<exists>q'. (q,c,q')\<in>\<Delta> \<A>))" .
  } 

  with \<Delta>_consistent have "?B = ?C" by blast
  thus ?thesis unfolding INY_initial_def by simp
qed

lemma INY_abstr6_init_\<omega>\<C>_loopc_correct':
  assumes "INY_abstr5_d_correct d"
  shows "INY_abstr6_init_\<omega>\<C>_loopc d \<le> 
      SPEC (\<lambda>(\<omega>,\<C>). \<F> \<A> \<times> (\<Q> \<A> - \<F> \<A>) \<union> \<omega> = INY_initial \<and> \<C> = \<omega>)"
proof-
  let ?\<omega>\<^sub>1' = "\<F> \<A> \<times> (\<Q> \<A> - \<F> \<A>)" and
      ?\<omega>\<^sub>2' = "{(u,v) |c u v. u\<in>\<Q> \<A> \<and> v\<in>\<Q> \<A> \<and> c\<in>\<Sigma> \<A> \<and> 
                             d(u, c)\<noteq>Some 0 \<and> d(v,c) = Some 0}"
  note INY_abstr6_init_\<omega>_loopc_correct[of d]
  also have "SPEC (\<lambda>(\<omega>,\<C>). \<omega> = ?\<omega>\<^sub>2' \<and> \<C> = \<omega>) \<le> 
      SPEC (\<lambda>(\<omega>,\<C>). ?\<omega>\<^sub>1'\<union>\<omega> = ?\<omega>\<^sub>1'\<union>?\<omega>\<^sub>2' \<and> \<C> = \<omega>)"
      by (rule SPEC_rule, force)
  also note INY_abstr6_init_\<omega>\<C>_loopc_correct_aux[OF assms]
  finally show ?thesis .
qed


lemma INY_abstr6_init_\<omega>\<C>_correct:
  assumes "INY_abstr5_d_correct d"
  shows "INY_abstr6_init_\<omega>\<C> d \<le> 
      SPEC (\<lambda>(\<omega>,\<C>). \<omega> = INY_initial \<and> \<C> = INY_initial)"
unfolding INY_abstr6_init_\<omega>\<C>_def
apply (intro refine_vcg)
apply (rule order_trans[OF INY_abstr6_init_\<omega>\<C>_loopc_correct'[OF assms]])
apply (intro refine_vcg)
apply (auto simp: Let_def)
done


text {* The final version of the abstract algorithm in which all operations 
  have been implemented. The only SPEC remaining is the one that obtains a 
  @{term "(u',v')"} from @{term "\<C>"}. *}
definition INY_abstr6 where
"INY_abstr6 \<equiv> do {
  (d,\<delta>r) \<leftarrow> INY_abstr6_empty_d\<delta>r;
  (d,\<delta>r) \<leftarrow> INY_abstr6_init_d\<delta>r d \<delta>r;
  N \<leftarrow> INY_abstr6_init_N d;
  (\<omega>,\<C>) \<leftarrow> INY_abstr6_init_\<omega>\<C> d;
  (\<omega>,\<C>,N) \<leftarrow> INY_abstr5' \<omega> \<C> N \<delta>r;
  RETURN (\<omega>,\<C>)
}"

lemma INY_abstr6_correct: "INY_abstr6 \<le> \<Down>Id INY_abstr5"
unfolding INY_abstr6_def INY_abstr5_def
apply (refine_rcg)
apply (rule INY_abstr6_empty_d\<delta>r_correct)
apply (rule INY_abstr6_init_d\<delta>r_correct, simp, simp)
apply (rule INY_abstr6_init_N_correct, simp)
apply (rule INY_abstr6_init_\<omega>\<C>_correct, simp)
apply (simp, rule Id_refine)
apply simp
done


text {* Refinement of @{term "\<omega>"} from @{typ "('q\<times>'q) set"} to
  @{typ "'q \<rightharpoonup> 'q set"} *}

(* TODO: Use rel_\<alpha> directly! *)
abbreviation "\<omega>_\<alpha> \<equiv> rel_\<alpha>"
lemmas \<omega>_\<alpha>_def = rel_\<alpha>_def

definition \<omega>_insert :: "('q\<times>'q) \<Rightarrow> ('q\<rightharpoonup>'q set) \<Rightarrow> 'q\<rightharpoonup>'q set"where 
  "\<omega>_insert \<equiv> (\<lambda>(x,y) \<omega>. case \<omega> x of 
      None \<Rightarrow> \<omega>(x \<mapsto> {y}) | Some \<omega>x \<Rightarrow> \<omega>(x \<mapsto> insert y \<omega>x))"

definition \<omega>_member :: "('q\<times>'q) \<Rightarrow> ('q\<rightharpoonup>'q set) \<Rightarrow> bool" where 
  "\<omega>_member = (\<lambda>(x,y) \<omega>. case \<omega> x of 
      None \<Rightarrow> False | Some \<omega>x \<Rightarrow> y \<in> \<omega>x)"

abbreviation "\<omega>_union_invar \<omega> S \<equiv> (\<lambda>it \<omega>'. \<omega>_\<alpha> \<omega>' \<union> it = \<omega>_\<alpha> \<omega> \<union> S)"

definition \<omega>_union :: "('q\<rightharpoonup>'q set) \<Rightarrow> ('q\<times>'q) set \<Rightarrow> ('q\<rightharpoonup>'q set) nres" where
  "\<omega>_union \<omega> S \<equiv> FOREACH\<^bsup>\<omega>_union_invar \<omega> S\<^esup> S 
      (\<lambda>xy \<omega>. RETURN (\<omega>_insert xy \<omega>)) \<omega>"

lemma \<omega>_insert_correct[simp]: "\<omega>_\<alpha> (\<omega>_insert xy \<omega>) = insert xy (\<omega>_\<alpha> \<omega>)"
  unfolding \<omega>_\<alpha>_def \<omega>_insert_def
  by (cases xy, force split: option.split_asm  option.split  split_if_asm)

lemma \<omega>_member_correct[simp]: "\<omega>_member xy \<omega> = (xy \<in> \<omega>_\<alpha> \<omega>)"
  unfolding \<omega>_\<alpha>_def \<omega>_member_def
  by (cases xy, force split: option.split)

lemma \<omega>_union_correct: 
  assumes "finite S"
  shows  "\<omega>_union \<omega> S \<le> \<Down>(br \<omega>_\<alpha> (\<lambda>_. True)) (RETURN (\<omega>_\<alpha> \<omega> \<union> S))"
  unfolding \<omega>_union_def
  by (refine_rcg, intro refine_vcg assms, auto simp: br_def)
  

definition INY_abstr7_init_\<omega>\<C>_loopv where
"INY_abstr7_init_\<omega>\<C>_loopv d c u \<omega> \<C> \<equiv> 
    FOREACH (\<Q> \<A>)
      (\<lambda>v (\<omega>',\<C>'). if d (v, c) = Some 0 then 
        RETURN (\<omega>_insert (u,v) \<omega>', insert (u,v) \<C>') 
      else RETURN (\<omega>',\<C>')) (\<omega>,\<C>)"

definition INY_abstr7_init_\<omega>\<C>_loopu where
"INY_abstr7_init_\<omega>\<C>_loopu d c \<omega> \<C> \<equiv>
    FOREACH (\<Q> \<A>)
      (\<lambda>u (\<omega>',\<C>'). if d (u, c) \<noteq> Some 0 
        then INY_abstr7_init_\<omega>\<C>_loopv d c u \<omega>' \<C>'
        else RETURN (\<omega>',\<C>'))
      (\<omega>,\<C>)"

definition INY_abstr7_init_\<omega>\<C>_loopc where
"INY_abstr7_init_\<omega>\<C>_loopc d \<equiv>
    FOREACH (\<Sigma> \<A>)
      (\<lambda>c (\<omega>,\<C>). INY_abstr7_init_\<omega>\<C>_loopu d c \<omega> \<C>)
      (Map.empty, {})"

definition INY_abstr7_init_\<omega>\<C> where
"INY_abstr7_init_\<omega>\<C> d \<equiv> do {
  (\<omega>,\<C>) \<leftarrow> INY_abstr7_init_\<omega>\<C>_loopc d;
  let FN = \<F> \<A> \<times> (\<Q> \<A> - \<F> \<A>);
  \<omega>' \<leftarrow> \<omega>_union \<omega> FN;
  RETURN (\<omega>', \<C> \<union> FN)
}"

abbreviation "\<omega>\<C>_rel \<equiv> br (\<lambda>(\<omega>,\<C>). (\<omega>_\<alpha> \<omega>, \<C>)) (\<lambda>_. True)"
lemma \<omega>\<C>_rel_sv: "single_valued \<omega>\<C>_rel" by (fact br_sv)

lemma finite_FN: "finite (\<F> \<A> \<times> (\<Q> \<A> - \<F> \<A>))"
    by (intro finite_Diff finite_SigmaI finite_\<F> finite_\<Q>)

lemma nofail_\<omega>_union:  "finite S \<Longrightarrow> nofail (\<omega>_union \<omega> S)"
    by (drule \<omega>_union_correct, force simp add: pw_le_iff pw_conc_nofail)
 
lemma inres_\<omega>_union:
  assumes "finite S" "inres (\<omega>_union \<omega> S) \<omega>'"
  shows "\<omega>_\<alpha> \<omega>' = \<omega>_\<alpha> \<omega> \<union> S"
using \<omega>_union_correct[OF assms(1)] assms(2)
  by (simp add: pw_ref_iff nofail_\<omega>_union[OF assms(1)])
     (simp add: br_def)

lemma INY_abstr7_init_\<omega>\<C>_refine:
  notes [[goals_limit = 3]]
  shows "INY_abstr7_init_\<omega>\<C> d \<le> \<Down>\<omega>\<C>_rel (INY_abstr6_init_\<omega>\<C> d)"
unfolding INY_abstr7_init_\<omega>\<C>_def INY_abstr6_init_\<omega>\<C>_def
INY_abstr7_init_\<omega>\<C>_loopc_def INY_abstr6_init_\<omega>\<C>_loopc_def
INY_abstr7_init_\<omega>\<C>_loopu_def INY_abstr6_init_\<omega>\<C>_loopu_def
INY_abstr7_init_\<omega>\<C>_loopv_def INY_abstr6_init_\<omega>\<C>_loopv_def
  apply (refine_rcg inj_on_id)
  apply simp_all
  apply (rule introR[where R=\<omega>\<C>_rel])
  apply (simp add: \<omega>_\<alpha>_def br_def)
  apply (simp add: br_def)
  apply (simp add: pw_le_iff refine_pw_simps br_def
    nofail_\<omega>_union[OF finite_FN]  inres_\<omega>_union[OF finite_FN])
done

definition INY_abstr7_loopu where
"INY_abstr7_loopu \<omega> \<C> \<delta>r u' v' c v \<equiv> 
  FOREACH (case \<delta>r(u',c) of None \<Rightarrow> {} | Some s \<Rightarrow> s) (\<lambda>u (\<omega>, \<C>).
    if \<not>\<omega>_member (u,v) \<omega> then
        RETURN (\<omega>_insert (u,v) \<omega>, insert (u,v) \<C>)
    else
        RETURN (\<omega>, \<C>)
  ) (\<omega>, \<C>)"

abbreviation "\<omega>\<C>_\<alpha> \<equiv> (\<lambda>(\<omega>,\<C>). (\<omega>_\<alpha> \<omega>,\<C>))"
abbreviation "\<omega>\<C>N_\<alpha> \<equiv> (\<lambda>(\<omega>,\<C>,N). (\<omega>_\<alpha> \<omega>,\<C>,N))"
abbreviation "\<omega>\<C>N_rel \<equiv> br \<omega>\<C>N_\<alpha> (\<lambda>_. True)"
lemma \<omega>\<C>N_rel_sv: "single_valued \<omega>\<C>N_rel" by (fact br_sv)

lemma INY_abstr7_loopu_refine:
  notes [[goals_limit = 1]]
  shows  "INY_abstr7_loopu \<omega> \<C> \<delta>r u' v' c v \<le>\<Down>\<omega>\<C>_rel 
             (INY_abstr5_loopu (\<omega>_\<alpha> \<omega>) \<C> \<delta>r u' v' c v)"
unfolding INY_abstr7_loopu_def INY_abstr5_loopu_def
    by (refine_rcg inj_on_id \<omega>\<C>N_rel_sv, simp_all add: br_def)

definition INY_abstr7_loopv where
"INY_abstr7_loopv \<omega> \<C> N \<delta>r u' v' c \<equiv> 
  FOREACH (case \<delta>r(v',c) of None \<Rightarrow> {} | Some s \<Rightarrow> s) (\<lambda>v (\<omega>, \<C>, N). do {
    let (N, iszero) = INY_dec_counter N c u' v;
    if iszero then do {
        (\<omega>', \<C>') \<leftarrow> INY_abstr7_loopu \<omega> \<C> \<delta>r u' v' c v;
        RETURN (\<omega>', \<C>', N)
    } else
        RETURN (\<omega>, \<C>, N)
  }) (\<omega>, \<C>, N)"

lemma INY_abstr7_loopv_refine:
  notes [[goals_limit = 1]]
  shows  "INY_abstr7_loopv \<omega> \<C> N \<delta>r u' v' c \<le>\<Down>\<omega>\<C>N_rel 
             (INY_abstr5_loopv (\<omega>_\<alpha> \<omega>) \<C> N \<delta>r u' v' c)"
unfolding INY_abstr7_loopv_def INY_abstr5_loopv_def
apply (refine_rcg inj_on_id \<omega>\<C>N_rel_sv)
apply (simp_all add: br_def) [3]    
apply (force simp: br_def intro!: INY_abstr7_loopu_refine)
apply (simp_all add: br_def)
done


definition INY_abstr7_loopc where
"INY_abstr7_loopc \<omega> \<C> N \<delta>r u' v' \<equiv> FOREACH
    (\<Sigma> \<A>) (\<lambda>c (\<omega>, \<C>, N). do {
        (\<omega>',\<C>',N') \<leftarrow> INY_abstr7_loopv \<omega> \<C> N \<delta>r u' v' c;
        RETURN (\<omega>',\<C>',N')
     }) (\<omega>, \<C>, N)"

lemma INY_abstr7_loopc_refine:
  notes [[goals_limit = 1]]
  shows  "INY_abstr7_loopc \<omega> \<C> N \<delta>r u' v' \<le>\<Down>\<omega>\<C>N_rel 
             (INY_abstr5_loopc (\<omega>_\<alpha> \<omega>) \<C> N \<delta>r u' v')"
unfolding INY_abstr7_loopc_def INY_abstr5_loopc_def
apply (refine_rcg inj_on_id \<omega>\<C>N_rel_sv)
apply (simp_all add: br_def) [2]    
apply (force simp: br_def intro!: INY_abstr7_loopv_refine)
apply (simp_all add: br_def)
done


definition INY_abstr7' where
"INY_abstr7' \<omega> \<C> N \<delta>r \<equiv> WHILE\<^sub>T (\<lambda>(\<omega>, \<C>, N). \<C>\<noteq>{}) 
  (\<lambda>(\<omega>, \<C>, N). do {
      ASSERT (\<C>\<noteq>{});
      (u',v') \<leftarrow> SPEC (\<lambda>(u',v'). (u',v')\<in>\<C>);
      let \<C> = \<C> - {(u',v')};
      (\<omega>,\<C>,N) \<leftarrow> INY_abstr7_loopc \<omega> \<C> N \<delta>r u' v';
      RETURN (\<omega>, \<C>, N)
  }) (\<omega>, \<C>, N)"

lemma INY_abstr7'_refine:
  notes [[goals_limit = 1]]
  shows  "INY_abstr7' \<omega> \<C> N \<delta>r \<le>\<Down>\<omega>\<C>N_rel 
             (INY_abstr5' (\<omega>_\<alpha> \<omega>) \<C> N \<delta>r)"
unfolding INY_abstr7'_def INY_abstr5'_def
apply (refine_rcg inj_on_id \<omega>\<C>N_rel_sv)
apply (simp_all add: br_def) [4]
apply (force simp: br_def intro!: INY_abstr7_loopc_refine)
apply (simp_all add: br_def)
done


definition INY_abstr7 where
"INY_abstr7 \<equiv> do {
  (d,\<delta>r) \<leftarrow> INY_abstr6_empty_d\<delta>r;
  (d,\<delta>r) \<leftarrow> INY_abstr6_init_d\<delta>r d \<delta>r;
  N \<leftarrow> INY_abstr6_init_N d;
  (\<omega>,\<C>) \<leftarrow> INY_abstr7_init_\<omega>\<C> d;
  (\<omega>,\<C>,N) \<leftarrow> INY_abstr7' \<omega> \<C> N \<delta>r;
  RETURN (\<omega>,\<C>)
}"
lemma INY_abstr7_correct:
  notes [[goals_limit = 1]]
  shows  "INY_abstr7 \<le>\<Down>\<omega>\<C>_rel INY_abstr6"
unfolding INY_abstr7_def INY_abstr6_def
by (refine_rcg inj_on_id \<omega>\<C>_rel_sv)
   (auto simp: br_def intro!: Id_refine 
        INY_abstr7_init_\<omega>\<C>_refine INY_abstr7'_refine)


text {* Summarize the definitions of all the internal constants that
  occur in @{term INY_abstr6}
*}

abbreviation "\<omega>_compl_invar \<omega> it \<omega>' \<equiv> 
    (\<forall>x. \<omega>' x = (if x\<in>it \<or> x\<notin>\<Q> \<A> then None else 
        Some {y. (x,y) \<in> compl (\<omega>_\<alpha> \<omega>)}))"

definition \<omega>_compl :: "('q\<rightharpoonup>'q set) \<Rightarrow> ('q\<rightharpoonup>'q set) nres" where
  "\<omega>_compl \<omega> = FOREACH\<^bsup>\<omega>_compl_invar \<omega>\<^esup> (\<Q> \<A>)
      (\<lambda>q \<omega>'. case \<omega> q of
          None \<Rightarrow> RETURN (\<omega>'(q \<mapsto> \<Q> \<A>)) |
          Some \<omega>q \<Rightarrow> RETURN (\<omega>'(q \<mapsto> \<Q> \<A> - \<omega>q))
      ) Map.empty"

lemma \<omega>_compl_correct:
    "\<omega>_compl \<omega> \<le>\<Down>(br \<omega>_\<alpha> (\<lambda>_. True)) (RETURN (compl (\<omega>_\<alpha> \<omega>)))"
  unfolding \<omega>_compl_def
  by (refine_rcg, intro refine_vcg finite_\<Q>, unfold \<omega>_\<alpha>_def br_def, 
      auto split: split_if_asm)

lemma nofail_\<omega>_compl:  "nofail (\<omega>_compl \<omega>)"
    by (insert \<omega>_compl_correct, force simp add: pw_le_iff pw_conc_nofail)

lemma inres_\<omega>_compl:
  assumes "inres (\<omega>_compl \<omega>) \<omega>'"
  shows "\<omega>_\<alpha> \<omega>' = compl (\<omega>_\<alpha> \<omega>)"
using \<omega>_compl_correct assms
  by (simp add: pw_ref_iff nofail_\<omega>_compl)
     (simp add: br_def)


definition "compute_simrel \<equiv> do {
  (\<omega>,_) \<leftarrow> INY_abstr7;
  \<omega>_compl \<omega>
}"


lemma compute_simrel_correct: 
    "compute_simrel \<le>\<Down>(br \<omega>_\<alpha> (\<lambda>_.True)) (SPEC (\<lambda>s. s=\<S>\<^sub>\<A>))"
proof -
  note INY_abstr7_correct
  also note INY_abstr6_correct
  also note INY_abstr5_correct
  also note INY_abstr4_correct
  also note INY_abstr3_correct
  also note INY_abstr2_correct
  also note INY_abstr1_correct
  finally have A7: "INY_abstr7 \<le>\<Down>\<omega>\<C>_rel (SPEC (\<lambda>(\<omega>, _). compl \<omega> = \<S>\<^sub>\<A>))" .
  
  have nofail: "nofail INY_abstr7"
      by (insert A7, force simp add: pw_le_iff pw_conc_nofail)
  with A7 have inres: "\<And>\<omega> \<C>. inres INY_abstr7 (\<omega>,\<C>) \<Longrightarrow>
      compl (\<omega>_\<alpha> \<omega>) = \<S>\<^sub>\<A>"
      by (simp add: pw_ref_iff nofail_\<omega>_compl)
         (auto simp add: br_def)

  show ?thesis
    unfolding compute_simrel_def
        by (auto simp add: pw_le_iff refine_pw_simps br_def 
                 nofail inres nofail_\<omega>_compl inres_\<omega>_compl)
qed

lemmas INY_defs = 
  INY_abstr7_def
  INY_abstr7'_def
  INY_abstr7_init_\<omega>\<C>_def
  INY_abstr7_init_\<omega>\<C>_loopc_def
  INY_abstr7_init_\<omega>\<C>_loopu_def
  INY_abstr7_init_\<omega>\<C>_loopv_def
  INY_abstr7_loopc_def
  INY_abstr7_loopv_def
  INY_abstr7_loopu_def
  INY_abstr6_def
  INY_abstr6_empty_d\<delta>r_def
  INY_abstr6_init_d\<delta>r_def
  INY_abstr6_init_N_def
  INY_abstr6_init_\<omega>\<C>_def
  INY_abstr5'_def
  INY_abstr6_init_\<omega>\<C>_loopc_def
  INY_abstr5_loopc_def 
  INY_abstr6_init_\<omega>\<C>_loopu_def
  INY_abstr5_loopv_def
  INY_dec_counter_def
  INY_abstr5_loopu_def
  INY_abstr6_init_\<omega>\<C>_loopv_def
  INY_initial_def
  \<omega>_insert_def
  \<omega>_member_def
  \<omega>_union_def
  \<omega>_compl_def



subsubsection {* NFA reduction *}

(* TODO: Move *)
abbreviation "map_optional f x \<equiv> case f x of None \<Rightarrow> x | Some y \<Rightarrow> y"

definition NFA_reduce where
"NFA_reduce \<equiv> do {
  \<S>\<^sub>\<A> \<leftarrow> SPEC (\<lambda>\<S>. \<S> = \<S>\<^sub>\<A>);
  f \<leftarrow> SPEC (is_preord_eqclasses_map (\<Q> \<A>) \<S>\<^sub>\<A>);
  RETURN (NFA_rename_states \<A> (map_optional f :: 'q \<Rightarrow> 'q))
}"

term map_optional

lemma preord_eqclasses_map_is_rename_fun:
assumes "is_preord_eqclasses_map (\<Q> \<A>) \<S>\<^sub>\<A> f"
shows "NFA_is_equivalence_rename_fun \<A> (map_optional f)"
proof-
    let ?eq = "\<lambda>u v. (u, v) \<in> \<S>\<^sub>\<A> \<and> (v, u) \<in> \<S>\<^sub>\<A>"
    from assms have f_props: 
        "\<And>u. u\<in>\<Q> \<A> \<Longrightarrow> \<exists>v. f u = Some v"
        "\<And>u v. u\<in>\<Q> \<A> \<Longrightarrow> v \<in> \<Q> \<A> \<Longrightarrow> 
                   (f u = f v) \<longleftrightarrow> ?eq u v"
        unfolding is_preord_eqclasses_map_def by auto

  {
    fix u::'q and v::'q
    assume "u\<in>\<Q> \<A>" "v\<in>\<Q> \<A>"
    from f_props(1)[OF this(1)] f_props(1)[OF this(2)] 
        obtain u' v' where [simp]: "f u = Some u'" "f v = Some v'" by auto
    from f_props(2)[OF `u \<in> \<Q> \<A>` `v \<in> \<Q> \<A>`]
        have "(map_optional f u = map_optional f v) \<longleftrightarrow> ?eq u v" by simp
    also have "... \<longrightarrow> u =\<^sub>R v" using sim_imp_\<L>_right_subset by blast
    finally have "map_optional f u = map_optional f v \<longrightarrow> u =\<^sub>R v" by simp
  }
  thus ?thesis
      unfolding NFA_is_equivalence_rename_fun_def by blast
qed

lemma NFA_reduce_correct:
  "NFA_reduce \<le> SPEC (\<lambda>\<A>'. \<L> \<A>' = \<L> \<A>)"
unfolding NFA_reduce_def
by (intro refine_vcg, simp add: \<L>_rename_iff 
        preord_eqclasses_map_is_rename_fun)

definition NFA_reduce_impl where
"NFA_reduce_impl \<equiv> do {
  \<S>\<^sub>\<A> \<leftarrow> compute_simrel;
  f \<leftarrow> preord_eqclasses_map_impl (\<Q> \<A>) \<S>\<^sub>\<A>;
  RETURN (NFA_rename_states \<A> (map_optional f :: 'q \<Rightarrow> 'q))
}"

lemma NFA_reduce_impl_refine:
  "NFA_reduce_impl \<le>\<Down>Id NFA_reduce"
unfolding NFA_reduce_impl_def NFA_reduce_def
apply (refine_rcg br_sv[of "\<omega>_\<alpha>" "\<lambda>_. True"])
apply (rule compute_simrel_correct)
apply (rule preord_eqclasses_map_impl_correct[OF finite_\<Q>])
apply (simp add: simrel_preorder)
apply (simp add: \<omega>_\<alpha>_def[abs_def] rel_\<alpha>_def[abs_def])
apply simp
done

lemma NFA_reduce_impl_correct:
  "NFA_reduce_impl \<le> SPEC (\<lambda>\<A>'. \<L> \<A>' = \<L> \<A>)"
  by (rule order_trans, rule NFA_reduce_impl_refine, 
      simp add: NFA_reduce_correct)

definition rev_simrel ("\<S>\<^sub>\<A>\<inverse>")
  where "rev_simrel \<equiv> NFA.\<S>\<^sub>\<A> (NFA_reverse \<A>)"

definition NFA_reduce_rev where
"NFA_reduce_rev \<equiv> do {
  \<A>' \<leftarrow> SPEC (\<lambda>\<A>'. \<A>' = NFA_reverse \<A>);
  \<S>\<^sub>\<A> \<leftarrow> SPEC (\<lambda>\<S>. \<S> = \<S>\<^sub>\<A>\<inverse>);
  f \<leftarrow> SPEC (is_preord_eqclasses_map (\<Q> \<A>) \<S>\<^sub>\<A>\<inverse>);
  RETURN (NFA_rename_states \<A> (the o f :: 'q \<Rightarrow> 'q))
}"

definition NFA_reduce_rev_impl where
"NFA_reduce_rev_impl \<equiv> do {
  \<A>' \<leftarrow> SPEC (\<lambda>\<A>'. \<A>' = NFA_reverse \<A>);
  \<S> \<leftarrow> NFA.compute_simrel \<A>';
  f \<leftarrow> preord_eqclasses_map_impl (\<Q> \<A>) \<S>;
  RETURN (NFA_rename_states \<A> (the o f :: 'q \<Rightarrow> 'q))
}"



text {*  Renaming states and taking the reverse automaton can be performed
  in arbitrary order without changing the result. *}
lemma NFA_rename_reverse_commute: 
    "NFA_rename_states (NFA_reverse \<A>) f = NFA_reverse (NFA_rename_states \<A> f)"
 unfolding NFA_rename_states_def SemiAutomaton_rename_states_ext_def 
     NFA_reverse_def by auto

text {* States with the same left language can be merged without changing
  the acceptance behaviour of the automaton.
  Note: @{text "\<L>_right"} of individual states may change. The 
  renaming is therefore not an "equivalence rename function" as defined 
  in the NFA theory. *}
lemma NFA_left_equiv_rename:
  assumes "\<forall>q\<in>\<Q> \<A>. \<forall>q'\<in>\<Q> \<A>. (f q = f q') \<longrightarrow> \<L>_left \<A> q = \<L>_left \<A> q'" 
  shows "\<L> (NFA_rename_states \<A> f) = \<L> \<A>"
proof-
  have [simp]: "\<And>A. NFA_reverse (NFA_reverse A) = A"
      unfolding NFA_reverse_def by simp

  let ?\<A>' = "NFA_reverse \<A>" and ?\<A>'' = "NFA_reverse (NFA_rename_states \<A> f)"
  {
   fix q q' assume "q\<in>\<Q> \<A>" "q'\<in>\<Q> \<A>" "f q = f q'"
   with assms have "\<L>_left \<A> q = \<L>_left \<A> q'" by blast
   hence "\<L>_right ?\<A>' q = \<L>_right ?\<A>' q'"
    using NFA_reverse___\<L>_in_state[of \<A> q] 
        NFA_reverse___\<L>_in_state[of \<A> q'] by simp
  }
  moreover have "\<Q> ?\<A>' = \<Q> \<A>" by simp
  ultimately have "NFA_is_equivalence_rename_fun ?\<A>' f" 
      unfolding NFA_is_equivalence_rename_fun_def by blast
  with NFA.\<L>_rename_iff[OF NFA_reverse___is_well_formed[OF NFA_axioms], symmetric]
      have "\<L> ?\<A>' = \<L> (NFA_rename_states ?\<A>' f)" .
  also have "NFA_rename_states ?\<A>' f = ?\<A>''" 
      using NFA.NFA_rename_reverse_commute[OF NFA_axioms] .
  finally have "\<L> ?\<A>' = \<L> ?\<A>''" .
  thus ?thesis by (force simp: NFA_reverse___\<L>)
qed

lemma rev_simrel_eqclasses_map_is_rename_fun:
assumes "is_preord_eqclasses_map (\<Q> \<A>) \<S>\<^sub>\<A>\<inverse> f"
shows "\<L> (NFA_rename_states \<A> (the \<circ> f)) = \<L> \<A>"
proof (rule NFA_left_equiv_rename, intro ballI impI)
  note wf_rev = NFA.NFA_reverse___is_well_formed[OF NFA_axioms]
  fix u v assume A: "u \<in> \<Q> \<A>" "v \<in> \<Q> \<A>" "(the \<circ> f) u = (the \<circ> f) v"
  from assms have f_props: 
      "\<And>u. u\<in>\<Q> \<A> \<Longrightarrow> \<exists>v. f u = Some v"
      "\<And>u v. u\<in>\<Q> \<A> \<Longrightarrow> v \<in> \<Q> \<A> \<Longrightarrow> 
                 (f u = f v) \<longrightarrow> u =\<^sub>L v"
      unfolding is_preord_eqclasses_map_def 
      unfolding rev_simrel_def NFA.in_\<S>\<^sub>\<A>_iff_simulated[OF wf_rev]
      by (auto dest!: sim_reverse_imp_\<L>_left_subset)
  from f_props(1)[OF A(1)] f_props(1)[OF A(2)]
      obtain u' v' where [simp]: "f u = Some u'" "f v = Some v'" by auto
  from A(3) have "u' = v'" by simp
  with f_props(2)[OF A(1,2)] show "u =\<^sub>L v" by simp
qed

lemma NFA_reduce_rev_correct:
  "NFA_reduce_rev \<le> SPEC (\<lambda>\<A>'. \<L> \<A>' = \<L> \<A>)"
unfolding NFA_reduce_rev_def
by (intro refine_vcg, simp add: rev_simrel_eqclasses_map_is_rename_fun)

lemma NFA_reduce_rev_impl_refine:
  "NFA_reduce_rev_impl \<le>\<Down>Id NFA_reduce_rev"
unfolding NFA_reduce_rev_impl_def NFA_reduce_rev_def
apply (refine_rcg br_sv[of "\<omega>_\<alpha>" "\<lambda>_. True"])
using NFA.compute_simrel_correct[OF NFA.NFA_reverse___is_well_formed[OF NFA_axioms]]
apply (simp add: rev_simrel_def)
apply (rule preord_eqclasses_map_impl_correct[OF finite_\<Q>])
using NFA.simrel_preorder[OF NFA.NFA_reverse___is_well_formed[OF NFA_axioms]]
apply (simp add: rev_simrel_def)
apply (simp add: \<omega>_\<alpha>_def[abs_def] rel_\<alpha>_def[abs_def])
apply simp
done


lemma NFA_reduce_rev_impl_correct:
  "NFA_reduce_rev_impl \<le> SPEC (\<lambda>\<A>'. \<L> \<A>' = \<L> \<A>)"
  apply (rule order_trans)
  apply (rule NFA_reduce_rev_impl_refine)
  apply (simp add: NFA_reduce_rev_correct)
  done

lemmas NFA_reduce_impl_defs =
  preord_eqclasses_map_impl_def
  preord_eqclasses_map_impl2_loop_def

end


subsection "Refinement and Code Generation"

subsubsection {* Test *}
lemmas (in NFA) foo_uc = INY_abstr5_loopu_def[unfolded INY_defs]
concrete_definition foo uses NFA.foo_uc

schematic_lemma 
  notes [[goals_limit = 1]]
  assumes [autoref_rules]: "(\<A>impl,\<A>)\<in>\<langle>nat_rel,nat_rel\<rangle>dflt_NFA_rel"
  shows "(?f::?'c, foo \<A>)\<in>?R"
  unfolding foo_def[abs_def]
  apply (autoref (keep_goal,trace))
  done


subsubsection {* preord-eqclasses-map *}
lemmas (in NFA) pecm_uc = preord_eqclasses_map_impl_def[
  unfolded preord_eqclasses_map_impl2_loop_def
]

concrete_definition pecm_ex uses NFA.pecm_uc

schematic_lemma pecm_impl:
  notes [[goals_limit = 1]]
  assumes [autoref_rules]: "(\<Q>impl,Q)\<in>\<langle>nat_rel\<rangle>dflt_rs_rel"
  assumes [autoref_rules]: "(Simpl,S)\<in>\<langle>nat_rel,\<langle>nat_rel\<rangle>dflt_rs_rel\<rangle>dflt_rm_rel"
  shows "(?f::?'c, pecm_ex Q S)\<in>?R"
  unfolding pecm_ex_def[abs_def]
  apply (autoref_monadic (trace))
  done

concrete_definition pecm_impl uses pecm_impl
thm pecm_impl.refine
declare pecm_impl.refine[autoref_higher_order_rule, autoref_rules]

subsubsection "compute-simrel"

text {* We need to extract the definition from the NFA-locale, 
  and unfold the definitions of the internally used constants *}
lemmas (in NFA) compute_simrel_unfold_complete 
  = compute_simrel_def[unfolded INY_defs]

concrete_definition compute_simrel_ex uses NFA.compute_simrel_unfold_complete

schematic_lemma compute_simrel_impl:
  notes [[goals_limit = 1]]
  assumes [autoref_rules]: "(\<A>impl,\<A>)\<in>\<langle>nat_rel,nat_rel\<rangle>dflt_NFA_rel"
  shows "(?f,compute_simrel_ex \<A>)\<in>(?R::(?'c\<times>_) set)"
  unfolding compute_simrel_ex_def[abs_def]
  apply (autoref_monadic (trace)) (* FIXME: This takes looooong, in 
    particular rel-inf phase *)
  done

concrete_definition compute_simrel_impl uses compute_simrel_impl
thm compute_simrel_impl.refine[no_vars]
lemma compute_simrel_impl_refine[autoref_rules]:
  "(\<lambda>\<A>impl. RETURN (compute_simrel_impl \<A>impl),compute_simrel_ex)\<in>
     \<langle>nat_rel, nat_rel\<rangle>dflt_NFA_rel 
  \<rightarrow> \<langle>\<langle>nat_rel,\<langle>nat_rel\<rangle>dflt_rs_rel\<rangle>dflt_rm_rel\<rangle>nres_rel"
  by (parametricity add: compute_simrel_impl.refine)


subsubsection {* Reduce *}

lemma (in NFA) is_NFA: "NFA \<A>" by unfold_locales

context NFA begin

lemmas pecm_unfold = pecm_ex.refine[OF is_NFA]
lemmas compute_simrel_unfold = compute_simrel_ex.refine[OF is_NFA]

lemmas reduce_unfold = NFA_reduce_impl_def[
  unfolded pecm_unfold, unfolded compute_simrel_unfold]

end

concrete_definition NFA_reduce_ex uses NFA.reduce_unfold
print_theorems


declare [[autoref_trace_intf_unif]]
declare [[autoref_trace_failed_id]]

schematic_lemma NFA_reduce_impl:
  notes [[goals_limit = 1]]
  assumes [autoref_rules]: "(\<A>impl,\<A>)\<in>\<langle>nat_rel,nat_rel\<rangle>dflt_NFA_rel"
  shows "(?f::?'c,NFA_reduce_ex \<A>)\<in>?R"
  unfolding NFA_reduce_ex_def
  apply (autoref_monadic (trace))
  done

concrete_definition NFA_reduce_impl uses NFA_reduce_impl

lemma NFA_reduce_impl_refine[autoref_rules]:
  "(\<lambda>\<A>impl. RETURN (NFA_reduce_impl \<A>impl),NFA_reduce_ex)\<in>
     \<langle>nat_rel, nat_rel\<rangle>dflt_NFA_rel 
  \<rightarrow> \<langle>\<langle>nat_rel, nat_rel\<rangle>dflt_NFA_rel\<rangle>nres_rel"
  by (parametricity add: NFA_reduce_impl.refine)


export_code 
  compute_simrel_impl 
  NFA_reduce_impl 
  in SML 


text {* Correctness lemmas for the constants we generated code from *}

definition simrel_rel_def_internal:
    "simrel_rel R \<equiv> \<langle>R,\<langle>R\<rangle>dflt_rs_rel\<rangle>dflt_rm_rel O br rel_\<alpha> (\<lambda>_. True)"
lemma simrel_rel_def:
    "\<langle>R\<rangle>simrel_rel \<equiv> \<langle>R,\<langle>R\<rangle>dflt_rs_rel\<rangle>dflt_rm_rel O br rel_\<alpha> (\<lambda>_. True)"
    unfolding relAPP_def simrel_rel_def_internal .

lemma compute_simrel_correct:
  shows "(compute_simrel_impl, NFA.\<S>\<^sub>\<A>) 
    \<in> \<langle>nat_rel,nat_rel\<rangle>dflt_NFA_rel \<rightarrow> \<langle>nat_rel\<rangle>simrel_rel"
proof (intro fun_relI)
  fix \<A>impl \<A>
  assume A: "(\<A>impl,\<A>)\<in>\<langle>nat_rel,nat_rel\<rangle>dflt_NFA_rel"
  from A interpret NFA \<A> by (auto simp add: NFA_rel_def)

  note compute_simrel_impl.refine[OF A, THEN nres_relD]
  also note compute_simrel_ex.refine[OF is_NFA, symmetric, THEN meta_eq_to_obj_eq]
  also note compute_simrel_correct
  also note conc_fun_chain
  finally have "RETURN (compute_simrel_impl \<A>impl)
    \<le> \<Down> (\<langle>nat_rel, \<langle>nat_rel\<rangle>dflt_rs_rel\<rangle>dflt_rm_rel O br rel_\<alpha> (\<lambda>_. True))
         (SPEC (\<lambda>s. s = NFA.\<S>\<^sub>\<A> \<A>))"
    .
  thus "(compute_simrel_impl \<A>impl, NFA.\<S>\<^sub>\<A> \<A>) \<in> \<langle>nat_rel\<rangle>simrel_rel"
    unfolding simrel_rel_def
    apply (rule RETURN_ref_SPECD)
    by simp
qed

lemma NFA_reduce_impl_correct:
  assumes A: "(\<A>impl,\<A>)\<in>\<langle>nat_rel,nat_rel\<rangle>dflt_NFA_rel"
  shows "RETURN (NFA_reduce_impl \<A>impl)
  \<le> \<Down>(\<langle>nat_rel,nat_rel\<rangle>dflt_NFA_rel) (SPEC (\<lambda>\<A>'. \<L> \<A>' = \<L> \<A>))"
proof -
  from A interpret NFA \<A> by (auto simp add: NFA_rel_def)

  note NFA_reduce_impl.refine[OF A, THEN nres_relD]
  also note NFA_reduce_ex.refine[OF is_NFA, symmetric, THEN meta_eq_to_obj_eq]
  also note NFA_reduce_impl_correct
  finally show ?thesis .
qed

hide_const NFA.compl

end
