theory Abs_Heapmap
imports Abs_Heap
begin

  section \<open>Abstract Heap-Maps\<close>
  text \<open>
    Abstract heap-maps encode a map data structure, 
    where values are associated with priorities. 
    Apart from the usual map operations, there are operations to retrieve and
    delete a minimum value.
    \<close>

  subsection \<open>Basic Definitions\<close>

  type_synonym ('k,'v) heapmap = "'k list \<times> ('k \<rightharpoonup> 'v)"

  locale hmstruct = heapstruct prio for prio :: "'v \<Rightarrow> 'b::linorder" 
  context hmstruct begin
    definition hm_\<alpha> :: "('k,'v) heapmap \<Rightarrow> 'k \<rightharpoonup> 'v" where "hm_\<alpha> = snd"

    text \<open>
      In order to reason about heap-maps, we abstract them to ordinary min-heaps.
      \<close>
    definition hm_to_heap :: "('k,'v) heapmap \<Rightarrow> 'v heap" where
      "hm_to_heap \<equiv> \<lambda>(pq,m). map (the o m) pq"

    definition is_heapmap1 :: "('k,'v) heapmap \<Rightarrow> bool" where
      "is_heapmap1 \<equiv> \<lambda>(pq,m). distinct pq \<and> dom m = set pq"
    
    definition is_heapmap2 :: "('k,'v) heapmap \<Rightarrow> bool" where
      "is_heapmap2 \<equiv> \<lambda>hm. is_heap (hm_to_heap hm)"
      
    definition is_heapmap :: "('k,'v) heapmap \<Rightarrow> bool" where
      "is_heapmap hm \<equiv> is_heapmap1 hm \<and> is_heapmap2 hm"

    definition hm_valid :: "('k,'v) heapmap \<Rightarrow> nat \<Rightarrow> bool" where
      "hm_valid \<equiv> \<lambda>(pq,m) i. 0<i \<and> i\<le>length pq"

    definition "hm_prio_of hm i \<equiv> prio_of (hm_to_heap hm) i"

    abbreviation "hm_has_parent h i \<equiv> hm_valid h (parent i)"
    abbreviation "hm_has_left h i \<equiv> hm_valid h (left i)"
    abbreviation "hm_has_right h i \<equiv> hm_valid h (right i)"

    abbreviation "hm_pparent h i == hm_prio_of h (parent i)"
    abbreviation "hm_pleft h i == hm_prio_of h (left i)"
    abbreviation "hm_pright h i == hm_prio_of h (right i)"

    lemma hm_valid_empty[simp]: "\<not>hm_valid ([],m) i"  
      by (auto simp: hm_valid_def)

    lemma hm_valid_idx_less[simp]: "hm_valid (pq,m) i \<Longrightarrow> i - Suc 0 < length pq"  
      unfolding hm_valid_def by auto

    lemma hm_valid_parent[simp]: "\<lbrakk>hm_valid hm i; 1<i\<rbrakk> \<Longrightarrow> hm_valid hm (parent i)"  
      unfolding hm_valid_def parent_def by auto

    lemma hm_valid_to_heap_conv: "hm_valid hm i \<longleftrightarrow> valid (hm_to_heap hm) i"
      unfolding valid_def hm_valid_def hm_to_heap_def by auto

    lemma hm_valid_to_heap[simp]: "hm_valid hm i \<Longrightarrow> valid (hm_to_heap hm) i"  
      unfolding valid_def hm_valid_def hm_to_heap_def by auto

    lemma length_hm_to_heap[simp]: "length (hm_to_heap (pq,m)) = length pq"  
      by (auto simp: hm_to_heap_def)

    lemma hm_to_heap_empty[simp]: "hm_to_heap ([],m) = []"
      by (auto simp: hm_to_heap_def)
      
    subsection \<open>Auxiliary Operations\<close>  

    subsubsection \<open>Exchange\<close>
    definition hm_exch :: "('k,'v) heapmap \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> ('k,'v) heapmap"
      where "hm_exch \<equiv> \<lambda>(pq,m) i j. (swap pq (i - 1) (j - 1), m)"

    lemma hm_exch_\<alpha>[simp]: "hm_\<alpha> (hm_exch hm i j) = hm_\<alpha> hm"
      by (auto simp: hm_exch_def hm_\<alpha>_def split: prod.splits)

    lemma hm_exch_invar1[simp]: "\<lbrakk>is_heapmap1 hm; hm_valid hm i; hm_valid hm j\<rbrakk> 
      \<Longrightarrow> is_heapmap1 (hm_exch hm i j)"  
      by (auto simp: is_heapmap1_def hm_exch_def)

    lemma hm_exch_to_heap[simp]: "\<lbrakk>hm_valid hm i; hm_valid hm j\<rbrakk> 
      \<Longrightarrow> hm_to_heap (hm_exch hm i j) = exch (hm_to_heap hm) i j"  
      apply (cases hm)
      apply (auto simp: hm_to_heap_def swap_def hm_exch_def exch_def map_update)
      done

    lemma hm_valid_exch[simp]: "hm_valid (hm_exch hm i j) = hm_valid hm"  
      by (auto intro!: ext simp: hm_valid_def hm_exch_def)

    subsubsection \<open>Butlast\<close>  
    lemma hm_to_heap_butlast[simp]: 
      "hm_to_heap (butlast pq,m) = butlast (hm_to_heap (pq,m))"
      by (auto simp: hm_to_heap_def map_butlast)

    subsubsection \<open>Update\<close>  
    lemma hm_to_heap_update[simp]: 
      "hm_to_heap (pq[i := k], m) = hm_to_heap (pq,m) [i := the (m k)]"
      by (auto simp: hm_to_heap_def map_update)


    subsubsection \<open>Swim\<close>  
    function hm_swim :: "('k,'v) heapmap \<Rightarrow> nat \<Rightarrow> ('k,'v) heapmap" where
      "hm_swim h i = (
        if i>1 \<and> \<not> (hm_prio_of h (parent i) \<le> hm_prio_of h i) then 
          hm_swim (hm_exch h i (parent i)) (parent i)
        else
          h
      )"
      by pat_completeness auto  

    termination 
      apply (relation "inv_image less_than snd")
      apply (auto simp: valid_def parent_def)
      done

    declare hm_swim.simps[simp del]
        
    lemma hm_to_heap_swim[simp]: 
      "hm_valid hm i \<Longrightarrow> hm_to_heap (hm_swim hm i) = swim (hm_to_heap hm) i"
      apply (induction hm i rule: hm_swim.induct)
      apply (subst hm_swim.simps)
      apply (subst swim.simps)
      apply (split split_if, intro impI conjI)
      apply (auto simp: hm_prio_of_def)
      done

    lemma is_heapmap1_swim[simp]: 
      "\<lbrakk>is_heapmap1 hm; hm_valid hm i\<rbrakk> \<Longrightarrow> is_heapmap1 (hm_swim hm i)"
      apply (induction hm i rule: hm_swim.induct)
      apply (subst hm_swim.simps)
      apply (auto)
      done

    lemma hm_\<alpha>_swim[simp]: "hm_\<alpha> (hm_swim hm i) = hm_\<alpha> hm"
      apply (induction hm i rule: hm_swim.induct)
      apply (subst hm_swim.simps)
      apply (auto)
      done

    subsubsection \<open>Sink\<close>  
    function hm_sink :: "('k,'v) heapmap \<Rightarrow> nat \<Rightarrow> ('k,'v) heapmap" where
    "hm_sink hm i = (
      if hm_valid hm i \<and> hm_has_left hm i then
        if hm_has_right hm i then
          if (hm_pleft hm i < hm_prio_of hm i \<and> hm_pright hm i \<ge> hm_pleft hm i) then
            hm_sink (hm_exch hm i (left i)) (left i)
          else if (hm_pright hm i < hm_pleft hm i \<and> hm_pright hm i < hm_prio_of hm i) then
            hm_sink (hm_exch hm i (right i)) (right i)
          else hm
        else if (hm_pleft hm i < hm_prio_of hm i) then
          hm_sink (hm_exch hm i (left i)) (left i)
        else hm  
      else hm
    )"
      by pat_completeness auto
    
    termination
      by (relation "measure (\<lambda>((pq,m),i). length pq - i)")  
         (auto simp add: hm_valid_def left_def right_def Let_def hm_exch_def)
    declare hm_sink.simps[simp del]

    lemma hm_to_heap_sink[simp]: 
      "hm_valid hm i \<Longrightarrow> hm_to_heap (hm_sink hm i) = sink (hm_to_heap hm) i"
      apply (induction hm i rule: hm_sink.induct)
      apply (subst hm_sink.simps)
      apply (subst sink.simps)
      unfolding hm_prio_of_def hm_valid_to_heap_conv[symmetric]
      apply simp_all
      done

    lemma is_heapmap1_sink[simp]: 
      "\<lbrakk>is_heapmap1 hm; hm_valid hm i\<rbrakk> \<Longrightarrow> is_heapmap1 (hm_sink hm i)"
      apply (induction hm i rule: hm_sink.induct)
      apply (subst hm_sink.simps)
      apply (auto)
      done

    lemma hm_\<alpha>_sink[simp]: "hm_\<alpha> (hm_sink hm i) = hm_\<alpha> hm"
      apply (induction hm i rule: hm_sink.induct)
      apply (subst hm_sink.simps)
      apply (auto)
      done

    lemma hm_sink_empty[simp]: "hm_sink ([],m) i = ([],m)"
      apply (subst hm_sink.simps)
      apply (auto simp: hm_valid_def)
      done

    lemma hm_sink_valid[simp]: 
      "hm_valid hm i \<Longrightarrow> hm_valid (hm_sink hm i) j = hm_valid hm j"
      apply (subst hm_valid_to_heap_conv)+
      apply (simp)
      done


    subsubsection \<open>Repair\<close>  
      
    definition "hm_repair hm i \<equiv> 
      if hm_valid hm i then (* Adding this check as swim and 
        sink depend on valid indexes, but repair may be called with
        index of last element that has been removed. In implementation, we
        may change this if we have sink and swim that are robust against 
        invalid index, e.g., when implemented on fixed-size array.
        *)
        hm_swim (hm_sink hm i) i
      else hm"

    lemma hm_repair_invar1[simp]: 
      assumes "is_heapmap1 hm"
      shows "is_heapmap1 (hm_repair hm i)"
      unfolding hm_repair_def
      apply (simp add: assms)
      done

    lemma hm_repair_\<alpha>[simp]: "hm_\<alpha> (hm_repair hm i) = hm_\<alpha> hm"
      by (simp add: hm_repair_def)


    lemma hm_repair_empty[simp]: "hm_repair ([],m) j = ([],m)"
      unfolding hm_repair_def
      apply auto
      done

    lemma hm_repair_to_heap[simp]: "hm_to_heap (hm_repair hm i) = repair (hm_to_heap hm) i"
      by (simp add: repair_def hm_repair_def hm_valid_to_heap_conv)

    subsection \<open>Operations\<close>  

    subsubsection \<open>Empty\<close>
    definition hm_empty :: "('k,'v) heapmap" where
      "hm_empty \<equiv> ([],Map.empty)"

    lemma hm_empty_invar[simp]: "is_heapmap hm_empty"
      unfolding hm_empty_def is_heapmap_def is_heapmap1_def is_heapmap2_def 
        hm_to_heap_def
      by auto

    lemma hm_empty_\<alpha>[simp]: "hm_\<alpha> hm_empty = Map.empty"  
      unfolding hm_empty_def hm_\<alpha>_def by simp

    subsubsection \<open>Insert\<close>  
    definition hm_insert :: "('k,'v) heapmap \<Rightarrow> 'k \<Rightarrow> 'v \<Rightarrow> ('k,'v) heapmap" 
      where
      "hm_insert \<equiv> \<lambda>(pq,m) k v. hm_swim (pq@[k], m(k \<mapsto> v)) (length pq + 1)"

    lemma hm_insert_invar[simp]:
      assumes INV: "is_heapmap hm"
      assumes NID: "k \<notin> dom (hm_\<alpha> hm)"
      shows "is_heapmap (hm_insert hm k v)"
    proof -  
      obtain pq m where [simp]: "hm = (pq,m)" by (cases hm)

      have [simp]: "hm_valid (pq@[k], m(k \<mapsto> v)) (Suc (length pq))"
        by (auto simp: hm_valid_def)

      from INV NID have [simp]: "k\<notin>set pq"  
        by (auto simp: hm_\<alpha>_def is_heapmap_def is_heapmap1_def)

      have "is_heapmap1 (pq@[k], m(k \<mapsto> v))"
        using INV NID 
        unfolding is_heapmap_def is_heapmap1_def hm_\<alpha>_def
        by auto
      hence G1: "is_heapmap1 (hm_insert hm k v)"
        unfolding hm_insert_def
        by auto

      have [simp]: "hm_to_heap (pq@[k], m(k \<mapsto> v)) = hm_to_heap hm @ [v]"  
        by (auto simp: hm_to_heap_def)

      have G2: "is_heapmap2 (hm_insert hm k v)"  
        unfolding is_heapmap2_def hm_insert_def
        apply simp
        apply (rule swim_restores)
        using swim_invar_insert[of "hm_to_heap (pq, m)" v] INV
        apply (simp add: is_heapmap_def is_heapmap2_def)
        done
        
      from G1 G2 show ?thesis by (simp add: is_heapmap_def)  
    qed  

    lemma hm_insert_\<alpha>[simp]:
      shows "hm_\<alpha> (hm_insert hm k v) = hm_\<alpha> hm (k \<mapsto> v)"
      apply (cases hm)
      apply (auto simp: hm_insert_def)
      apply (auto simp: hm_\<alpha>_def)
      done

    subsubsection \<open>Decrease Key\<close>

    definition hm_decrease_key :: "('k,'v) heapmap \<Rightarrow> 'k \<Rightarrow> 'v \<Rightarrow> ('k,'v) heapmap"
      where "hm_decrease_key \<equiv> \<lambda>(pq,m) k v. let
        i = index pq k + 1;
        m = m(k \<mapsto> v)
      in hm_swim (pq,m) i"

    lemma hm_decrease_key_\<alpha>[simp]: 
      "hm_\<alpha> (hm_decrease_key hm k v) = (hm_\<alpha> hm)(k\<mapsto>v)"
      apply (cases hm)
      apply (auto simp: hm_decrease_key_def)
      apply (auto simp: hm_\<alpha>_def)
      done
      
    lemma hm_decrease_key_invar[simp]:
      assumes INV: "is_heapmap hm"
      assumes MEM: "hm_\<alpha> hm k = Some vh"
      assumes DECR: "prio v \<le> prio vh"
      shows "is_heapmap (hm_decrease_key hm k v)"
    proof -  
      obtain pq m where [simp]: "hm = (pq,m)" by (cases hm)

      have [simp]: "is_heapmap1 (pq, m(k \<mapsto> v))" "distinct pq"
        using INV MEM
        by (auto simp: is_heapmap_def is_heapmap1_def hm_\<alpha>_def)

      from MEM INV have "k \<in> set pq"  
        by (auto simp: hm_\<alpha>_def is_heapmap_def is_heapmap1_def)
      hence 1: "index pq k < length pq" "pq!index pq k = k"
        by simp_all
      hence [simp]: "Suc (index pq k) \<le> length pq" by auto
      note [simp] = 1

      have [simp]: "hm_valid (pq, m(k \<mapsto> v)) (Suc (index pq k))"
        by (auto simp: hm_valid_def)

      have G1: "is_heapmap1 (hm_decrease_key hm k v)"
        apply (auto simp: hm_decrease_key_def)
        done
        
      have [simp]: "hm_to_heap (pq, m(k \<mapsto> v)) = (hm_to_heap hm)[index pq k := v]"
        apply (clarsimp simp: hm_to_heap_def)
        apply (rule nth_equalityI)
        apply auto []
        apply (auto simp: nth_list_update index_nth_id)
        done

      have [simp]: "vh = hm_to_heap hm!(index pq k)"  
        using MEM
        by (auto simp: hm_\<alpha>_def hm_to_heap_def)

      have G2: "is_heapmap2 (hm_decrease_key hm k v)"
        using INV
        unfolding is_heapmap_def is_heapmap2_def
        apply (simp add: hm_decrease_key_def)
        apply (rule swim_restores)
        apply (rule swim_invar_decr')
        using DECR
        apply auto
        done
          
      from G1 G2 show ?thesis by (simp add: is_heapmap_def)    
    qed  

    subsubsection \<open>Increase Key\<close>
    definition hm_increase_key :: "('k,'v) heapmap \<Rightarrow> 'k \<Rightarrow> 'v \<Rightarrow> ('k,'v) heapmap"
      where "hm_increase_key \<equiv> \<lambda>(pq,m) k v. let
        i = index pq k + 1;
        m = m(k \<mapsto> v)
      in hm_sink (pq,m) i"

    lemma hm_increase_key_\<alpha>[simp]: 
      "hm_\<alpha> (hm_increase_key hm k v) = (hm_\<alpha> hm)(k\<mapsto>v)"
      apply (cases hm)
      apply (auto simp: hm_increase_key_def)
      apply (auto simp: hm_\<alpha>_def)
      done

    lemma hm_increase_key_invar[simp]:
      assumes INV: "is_heapmap hm"
      assumes MEM: "hm_\<alpha> hm k = Some vh"
      assumes INCR: "prio v \<ge> prio vh"
      shows "is_heapmap (hm_increase_key hm k v)"
    proof -  
      obtain pq m where [simp]: "hm = (pq,m)" by (cases hm)

      have [simp]: "is_heapmap1 (pq, m(k \<mapsto> v))" "distinct pq"
        using INV MEM
        by (auto simp: is_heapmap_def is_heapmap1_def hm_\<alpha>_def)

      from MEM INV have "k \<in> set pq"  
        by (auto simp: hm_\<alpha>_def is_heapmap_def is_heapmap1_def)
      hence 1: "index pq k < length pq" "pq!index pq k = k"
        by simp_all
      hence [simp]: "Suc (index pq k) \<le> length pq" by auto
      note [simp] = 1

      have [simp]: "hm_valid (pq, m(k \<mapsto> v)) (Suc (index pq k))"
        by (auto simp: hm_valid_def)

      have G1: "is_heapmap1 (hm_increase_key hm k v)"
        apply (auto simp: hm_increase_key_def)
        done
        
      have [simp]: "hm_to_heap (pq, m(k \<mapsto> v)) = (hm_to_heap hm)[index pq k := v]"
        apply (clarsimp simp: hm_to_heap_def)
        apply (rule nth_equalityI)
        apply auto []
        apply (auto simp: nth_list_update index_nth_id)
        done

      have [simp]: "vh = hm_to_heap hm!(index pq k)"  
        using MEM
        by (auto simp: hm_\<alpha>_def hm_to_heap_def)

      have G2: "is_heapmap2 (hm_increase_key hm k v)"
        using INV
        unfolding is_heapmap_def is_heapmap2_def
        apply (simp add: hm_increase_key_def)
        apply (rule sink_restores)
        apply (rule sink_invar_incr')
        using INCR
        apply auto
        done
          
      from G1 G2 show ?thesis by (simp add: is_heapmap_def)    
    qed  

    subsubsection \<open>Pop Minimum Element\<close>

    definition hm_pop_min :: "('k,'v) heapmap \<Rightarrow> ('k,'v) heapmap \<times> ('k \<times> 'v)"
    where "hm_pop_min \<equiv> \<lambda>(pq,m). let
      k = pq!0;
      v = the (m k);
      m = m(k := None);
      pq = butlast (pq[0 := last pq]);
      hm' = hm_sink (pq,m) 1
    in (hm',(k,v))"

    lemma hm_pop_min_correct_aux:
      assumes INV: "is_heapmap hm"
      assumes NE: "hm_\<alpha> hm \<noteq> Map.empty"
      assumes RES: "hm_pop_min hm = (hm',(k,v))"
      shows "is_heapmap hm'"
      and "hm_\<alpha> hm k = Some v"
      and "prio v = Min (prio`ran (hm_\<alpha> hm))"
      and "hm_\<alpha> hm' = (hm_\<alpha> hm)(k := None)"
    proof -
      obtain pq m where [simp]: "hm = (pq,m)" by (cases hm)

      from RES have [simp]:
        "hm' = hm_sink (butlast (pq[0 := last pq]), m(pq ! 0 := None)) (Suc 0)" 
        "k = pq ! 0"
        "v = the (m (pq ! 0))"
        unfolding hm_pop_min_def
        by (auto simp: Let_def)

      show "hm_\<alpha> hm k = Some v"
        using INV NE
        apply (cases pq)
        apply (force simp: hm_\<alpha>_def is_heapmap_def is_heapmap1_def)+
        done

      from NE INV have [simp]: "length pq \<noteq> 0" "pq \<noteq> []" 
        by (auto simp: hm_\<alpha>_def is_heapmap_def is_heapmap1_def)

      from INV have hm_heap: "is_heap (hm_to_heap hm)" 
        by (simp add: is_heapmap_def is_heapmap2_def)

      show "prio v = Min (prio`ran (hm_\<alpha> hm))"
      proof -
        have 1: "ran (hm_\<alpha> hm) = set (hm_to_heap hm)"
          using INV
          by (force simp: hm_\<alpha>_def hm_to_heap_def ran_def dom_def
            is_heapmap_def is_heapmap1_def)
        from heap_min_prop[OF hm_heap] show ?thesis 
          unfolding 1
          by (simp add: val_of_def hm_to_heap_def)
      qed 

      show "hm_\<alpha> hm' = (hm_\<alpha> hm)(k := None)"
        apply (auto simp: )
        apply (auto simp: hm_\<alpha>_def)
        done

      show "is_heapmap hm'"
      proof (cases "length pq = Suc 0")
        case True
        then obtain k where [simp]: "pq=[k]"
          by (cases pq, simp_all)

        have "hm' = ([],m(pq!0 := None))"
          by auto
        with INV have [simp]: "m(k := None) = Map.empty"  
          by (auto intro!: ext simp: is_heapmap_def is_heapmap1_def)
          
        show ?thesis
          by (simp add: is_heapmap_def is_heapmap1_def is_heapmap2_def)
      next    
        case False
        hence L1: "length pq > Suc 0"
          by (cases pq, simp_all)

        then obtain kf pqh kl where [simp]: "pq = kf#pqh@[kl]"  
          apply (cases pq, simp_all)
          apply (rename_tac list)
          apply (case_tac list rule: rev_cases, simp_all)
          done

        have G1: "is_heapmap1 hm'"
          apply simp
          apply (rule is_heapmap1_sink)
          prefer 2
          apply (auto simp: hm_valid_def) []
          using INV
          apply (auto simp: is_heapmap1_def is_heapmap_def)
          done
          
        have [simp]: "hm_valid (kl # pqh, m(kf := None)) (Suc 0)"
          by (auto simp: hm_valid_def)

        from INV have [simp]: "kl \<noteq> kf" "kf \<notin> set pqh"
          by (auto simp: is_heapmap_def is_heapmap1_def)

        have [simp]:
          "butlast (exch (hm_to_heap (kf # pqh @ [kl], m)) (Suc 0) (Suc (Suc (length pqh))))
          = hm_to_heap (kl # pqh, m(kf := None))"
          using INV
          by (auto simp: hm_to_heap_def exch_def nth_append swap_def list_update_append)

        from sink_invar_init[OF hm_heap] 
        have "sink_invar (hm_to_heap (kl # pqh, m(kf := None))) (Suc 0)"
          by simp
        hence G2: "is_heapmap2 hm'"
          apply (simp add: is_heapmap2_def)
          by (rule sink_restores)
        
        from G1 G2 show ?thesis by (simp add: is_heapmap_def)  
      qed
    qed  

    text \<open>For better reasoning, we split the result of @{const hm_pop_min}
      in its components\<close>
    definition "hm_pop_min_hm hm \<equiv> fst (hm_pop_min hm)"
    definition "hm_pop_min_k hm \<equiv> fst (snd (hm_pop_min hm))"
    definition "hm_pop_min_v hm \<equiv> snd (snd (hm_pop_min hm))"

    lemma hm_pop_min_parts_conv[simp]: "hm_pop_min hm 
      = (hm_pop_min_hm hm, (hm_pop_min_k hm, hm_pop_min_v hm))"
      unfolding hm_pop_min_hm_def hm_pop_min_k_def hm_pop_min_v_def
      by auto

    lemma 
      assumes "is_heapmap hm"  
      assumes "hm_\<alpha> hm \<noteq> Map.empty"
      shows hm_pop_min_invar[simp]: 
        "is_heapmap (hm_pop_min_hm hm)"
      and hm_pop_min_\<alpha>[simp]: 
        "hm_\<alpha> (hm_pop_min_hm hm) = (hm_\<alpha> hm)(hm_pop_min_k hm := None)"
      and hm_pop_min_kv[simp]: "hm_\<alpha> hm (hm_pop_min_k hm) = Some (hm_pop_min_v hm)"
      and hm_pop_min_v_prio[simp]: "prio (hm_pop_min_v hm) = Min (prio ` ran (hm_\<alpha> hm))"
      using hm_pop_min_correct_aux[OF assms hm_pop_min_parts_conv]
      by simp_all
      

    subsubsection \<open>Remove\<close>  

    definition hm_remove :: "('k,'v) heapmap \<Rightarrow> 'k \<Rightarrow> ('k,'v) heapmap" where
      "hm_remove \<equiv> \<lambda>(pq,m) k. let
        m = m(k := None);
        i = index pq k;
        pq = butlast (pq[i := last pq]);
        hm' = hm_repair (pq,m) (Suc i)
      in hm'"

    lemma hm_remove_invar[simp]:
      assumes INV: "is_heapmap hm"
      assumes KID: "k \<in> dom (hm_\<alpha> hm)"
      shows "is_heapmap (hm_remove hm k)"
    proof -  
      obtain pq m where [simp]: "hm = (pq,m)" by (cases hm)

      from INV KID have [simp]: "k \<in> set pq" "pq\<noteq>[]" and DIST: "distinct pq"
        by (auto simp: hm_\<alpha>_def is_heapmap_def is_heapmap1_def)

      from INV have [simp]: "is_heap (hm_to_heap (pq,m))"
        by (auto simp: is_heapmap_def is_heapmap2_def)

      from `k\<in>set pq`[unfolded in_set_conv_decomp] obtain pq1 pq2 where
        PQ: "pq = pq1@k#pq2" by blast

      {
        fix a b c and d :: 'k 
        have "butlast (a@b#c@[d]) = a@b#c" 
        by (auto simp: butlast_append)
      } note [simp] = this 


      have G1: "is_heapmap1 (hm_remove hm k)"
      proof (cases "length pq = 1")
        case True hence [simp]: "pq=[k]"
          by (auto simp: PQ)

        from INV have "dom m = {k}"  
          by (auto simp: is_heapmap_def is_heapmap1_def)

        thus ?thesis
          by (auto simp: is_heapmap1_def hm_remove_def dom_def)
      next
        case False

        show ?thesis
          apply (simp add: hm_remove_def Let_def)
          apply (rule hm_repair_invar1)
          apply (cases pq2 rule: rev_cases)
          using INV 
          apply (auto simp: is_heapmap_def is_heapmap1_def PQ index_append)
          done
      qed

      have [simp]: "butlast (hm_to_heap (pq, m(last pq := None))) 
        = butlast (hm_to_heap (pq, m))"
        using DIST
        apply (cases pq rule: rev_cases)
        apply (auto simp: hm_to_heap_def)
        done

      have [simp]: 
        "butlast (hm_to_heap (pq, m(k := None)))[index pq k := the (m (last pq))]
        = butlast (hm_to_heap (pq, m))[index pq k := the (m (last pq))]"  
        using DIST
        by (auto simp: hm_to_heap_def map_update PQ 
          fun_upd_comp index_append butlast_append list_update_append)
        
      have G2: "is_heapmap2 (hm_remove hm k)"
        apply (auto simp: is_heapmap2_def hm_remove_def Let_def butlast_list_update DIST)
        apply (auto simp: repair_def) []
        using DIST 
        apply (cases pq rule:rev_cases, 
          auto simp: index_append split: split_if_asm) []

        apply (rule is_heap_repair')
        apply auto []

        using DIST
        apply (auto simp: valid_def index_append PQ split: split_if_asm) []
        done

      from G1 G2 show "is_heapmap (hm_remove hm k)" 
        by (simp add: is_heapmap_def)

    qed

    lemma hm_remove_\<alpha>[simp]:
      shows "hm_\<alpha> (hm_remove hm k) = (hm_\<alpha> hm)(k := None)"
      unfolding hm_remove_def
      apply (cases hm)
      apply (auto simp: Let_def)
      apply (auto simp: hm_\<alpha>_def)
      done

    subsubsection \<open>Change Key\<close>  
    text \<open>Change the value underlying an existing key. 
      Unless decrease-key and increase-key, this one puts no constraints on
      the new priority.\<close>

    definition hm_change_key :: "('k,'v) heapmap \<Rightarrow> 'k \<Rightarrow> 'v \<Rightarrow> ('k,'v) heapmap"
      where "hm_change_key \<equiv> \<lambda>(pq,m) k v. let
        m = m(k \<mapsto> v);
        i = index pq k;
        hm' = hm_repair (pq,m) (Suc i)
      in hm'"  

    lemma hm_change_key_invar[simp]:
      assumes INV: "is_heapmap hm"
      assumes KID: "k \<in> dom (hm_\<alpha> hm)"
      shows "is_heapmap (hm_change_key hm k v)"
    proof -  
      obtain pq m where [simp]: "hm = (pq,m)" by (cases hm)

      from INV KID have [simp]: "k \<in> set pq" "pq\<noteq>[]" and DIST: "distinct pq"
        by (auto simp: hm_\<alpha>_def is_heapmap_def is_heapmap1_def)

      from INV have [simp]: "is_heap (hm_to_heap (pq,m))"
        by (auto simp: is_heapmap_def is_heapmap2_def)

      from `k\<in>set pq`[unfolded in_set_conv_decomp] obtain pq1 pq2 where
        PQ: "pq = pq1@k#pq2" by blast

      have G1: "is_heapmap1 (hm_change_key hm k v)"
        apply (simp add: hm_change_key_def Let_def)
        apply (rule hm_repair_invar1)
        apply (cases pq2 rule: rev_cases)
        using INV 
        apply (auto simp: is_heapmap_def is_heapmap1_def PQ index_append)
        done

      have [simp]: "hm_to_heap (pq,m(k\<mapsto>v)) = hm_to_heap (pq,m) [index pq k := v]" 
        using DIST
        by (auto simp: hm_to_heap_def PQ index_append list_update_append)

      have [simp]: "valid (hm_to_heap (pq, m)) (Suc (index pq k))"  
        using DIST
        by (auto simp: valid_def index_append PQ split: split_if_asm) []

      have G2: "is_heapmap2 (hm_change_key hm k v)"
        apply (auto simp: is_heapmap2_def hm_change_key_def Let_def)
        apply (rule is_heap_repair')
        by auto

      from G1 G2 show "is_heapmap (hm_change_key hm k v)" 
        by (simp add: is_heapmap_def)
    qed    

    lemma hm_change_key_\<alpha>[simp]:
      "hm_\<alpha> (hm_change_key hm k v) = (hm_\<alpha> hm)(k\<mapsto>v)"
      apply (cases hm)
      apply (auto simp: hm_change_key_def)
      apply (auto simp: hm_\<alpha>_def)
      done

    subsubsection \<open>Contains\<close>  
    definition hm_contains :: "('k,'v) heapmap \<Rightarrow> 'k \<Rightarrow> bool" where
      "hm_contains \<equiv> \<lambda>(_,m) k. k\<in>dom m"

    lemma hm_contains_correct[simp]: "hm_contains hm k \<longleftrightarrow> k\<in>dom (hm_\<alpha> hm)"  
      by (auto simp: hm_contains_def hm_\<alpha>_def)

    subsubsection \<open>Set\<close>  
    text \<open>Update the map. No constraints on priority, or whether the key
      is already in the queue or not.\<close>
    definition hm_set :: "('k,'v) heapmap \<Rightarrow> 'k \<Rightarrow> 'v \<Rightarrow> ('k,'v) heapmap" where
      "hm_set \<equiv> \<lambda>hm k v. 
        if (hm_contains hm k) then 
          hm_change_key hm k v 
        else hm_insert hm k v"

    lemma hm_set_invar[simp]: "is_heapmap hm \<Longrightarrow> is_heapmap (hm_set hm k v)"    
      by (auto simp: hm_set_def)

    lemma hm_set_\<alpha>[simp]:
      "hm_\<alpha> (hm_set hm k v) = (hm_\<alpha> hm)(k\<mapsto>v)"
      by (auto simp: hm_set_def)
      
    subsubsection \<open>Lookup\<close>  
    definition hm_lookup :: "('k,'v) heapmap \<Rightarrow> 'k \<rightharpoonup> 'v" where
      "hm_lookup \<equiv> \<lambda>(_,m) k. m k"

    lemma hm_lookup_correct[simp]: "hm_lookup hm = hm_\<alpha> hm"
      apply (cases hm)
      apply (auto simp: hm_\<alpha>_def hm_lookup_def)
      done

    subsubsection \<open>Minimum Key-Value pair\<close>  
    definition hm_min :: "('k,'v) heapmap \<Rightarrow> 'k\<times>'v" where
      "hm_min \<equiv> \<lambda>(pq,m). (pq!0,the (m (pq!0)))"

    definition "hm_min_k \<equiv> fst o hm_min"
    definition "hm_min_v \<equiv> snd o hm_min"

    lemma hm_min_decomp_conv[simp]: "hm_min hm = (hm_min_k hm, hm_min_v hm)"
      by (auto simp: hm_min_k_def hm_min_v_def)

    lemma hm_min_correct_aux:
      assumes INV: "is_heapmap hm"
      assumes NE: "hm_\<alpha> hm \<noteq> Map.empty"
      assumes RES: "hm_min hm = (k,v)"
      shows 
        "hm_\<alpha> hm k = Some v"
        "prio v = Min (prio`ran (hm_\<alpha> hm))"
    proof -
      obtain pq m where [simp]: "hm = (pq,m)" by (cases hm)

      from INV NE RES obtain pq' where 
        [simp]: "pq = k#pq'" and "k\<in>dom m" and DM: "dom m = set pq"
        apply (cases pq)
        apply (auto simp: hm_\<alpha>_def is_heapmap_def is_heapmap1_def hm_min_def)
        done
        
      from `k\<in>dom m` RES have MKV: "m k = Some v"
        by (auto simp: hm_min_def)

      thus "hm_\<alpha> hm k = Some v" by (auto simp: hm_\<alpha>_def)

      from INV have hm_heap: "is_heap (hm_to_heap (pq,m))" 
        by (simp add: is_heapmap_def is_heapmap2_def)

      have "insert (prio v) (prio ` (the \<circ> m) ` set pq') = prio`ran m" 
        using DM MKV
        apply (auto simp: ran_def dom_def intro!: imageI)
        apply force
        by (metis (mono_tags, hide_lams) DM `pq = k # pq'` 
          comp_apply domI image_eqI option.sel set_ConsD) 

      with heap_min_prop[OF hm_heap] show "prio v = Min (prio`ran (hm_\<alpha> hm))"
        by (auto simp: hm_to_heap_def val_of_def MKV hm_\<alpha>_def hm_min_def)
    qed

    lemma hm_min_correct[simp]:
      assumes INV: "is_heapmap hm"
      assumes NE: "hm_\<alpha> hm \<noteq> Map.empty"
      shows 
        "hm_\<alpha> hm (hm_min_k hm) = Some (hm_min_v hm)"
        "prio (hm_min_v hm) = Min (prio`ran (hm_\<alpha> hm))"
      using hm_min_correct_aux[OF assms] by auto

(*.xxx, ctd here: Implement with arrays.
  n :: nat - #entries in pq
  pq :: nat array - pq of keys
  qp :: nat array - key \<rightarrow> pq-idx
    pq-idx is based 1
    qp[k] = 0 means that k is not in pq
  vals :: 'v array - key \<rightarrow> value
*)

  end

end
