
structure STArray = struct

datatype 'a Cell = Invalid | Value of 'a array;

exception AccessedOldVersion;

type 'a array = 'a Cell Unsynchronized.ref;

fun fromList l = Unsynchronized.ref (Value (Array.fromList l));
fun array (size, v) = Unsynchronized.ref (Value (Array.array (size,v)));
fun tabulate (size, f) = Unsynchronized.ref (Value (Array.tabulate(size, f)));
fun sub (Unsynchronized.ref Invalid, idx) = raise AccessedOldVersion |
    sub (Unsynchronized.ref (Value a), idx) = Array.sub (a,idx);
fun update (aref,idx,v) =
  case aref of
    (Unsynchronized.ref Invalid) => raise AccessedOldVersion |
    (Unsynchronized.ref (Value a)) => (
      aref := Invalid;
      Array.update (a,idx,v);
      Unsynchronized.ref (Value a)
    );

fun length (Unsynchronized.ref Invalid) = raise AccessedOldVersion |
    length (Unsynchronized.ref (Value a)) = Array.length a

fun grow (aref, i, x) = case aref of 
  (Unsynchronized.ref Invalid) => raise AccessedOldVersion |
  (Unsynchronized.ref (Value a)) => (
    let val len=Array.length a;
        val na = Array.array (len+i,x) 
    in
      aref := Invalid;
      Array.copy {src=a, dst=na, di=0};
      Unsynchronized.ref (Value na)
    end
    );

fun shrink (aref, sz) = case aref of
  (Unsynchronized.ref Invalid) => raise AccessedOldVersion |
  (Unsynchronized.ref (Value a)) => (
    if sz > Array.length a then 
      raise Size
    else (
      aref:=Invalid;
      Unsynchronized.ref (Value (Array.tabulate (sz,fn i => Array.sub (a,i))))
    )
  );

structure IsabelleMapping = struct
type 'a ArrayType = 'a array;

fun new_array (a:'a) (n:int) = array (n, a);

fun array_length (a:'a ArrayType) = length a;

fun array_get (a:'a ArrayType) (i:int) = sub (a, i);

fun array_set (a:'a ArrayType) (i:int) (e:'a) = update (a, i, e);

fun array_of_list (xs:'a list) = fromList xs;

fun array_grow (a:'a ArrayType) (i:int) (x:'a) = grow (a, i, x);

fun array_shrink (a:'a ArrayType) (sz:int) = shrink (a,sz);

end;

end;

structure FArray = struct
  datatype 'a Cell = Value of 'a Array.array | Upd of (int*'a*'a Cell Unsynchronized.ref);

  type 'a array = 'a Cell Unsynchronized.ref;

  fun array (size,v) = Unsynchronized.ref (Value (Array.array (size,v)));
  fun tabulate (size, f) = Unsynchronized.ref (Value (Array.tabulate(size, f)));
  fun fromList l = Unsynchronized.ref (Value (Array.fromList l));

  fun sub (Unsynchronized.ref (Value a), idx) = Array.sub (a,idx) |
      sub (Unsynchronized.ref (Upd (i,v,cr)),idx) =
        if i=idx then v
        else sub (cr,idx);

  fun length (Unsynchronized.ref (Value a)) = Array.length a |
      length (Unsynchronized.ref (Upd (i,v,cr))) = length cr;

  fun realize_aux (aref, v) =
    case aref of
      (Unsynchronized.ref (Value a)) => (
        let
          val len = Array.length a;
          val a' = Array.array (len,v);
        in
          Array.copy {src=a, dst=a', di=0};
          Unsynchronized.ref (Value a')
        end
      ) |
      (Unsynchronized.ref (Upd (i,v,cr))) => (
        let val res=realize_aux (cr,v) in
          case res of
            (Unsynchronized.ref (Value a)) => (Array.update (a,i,v); res)
        end
      );

  fun realize aref =
    case aref of
      (Unsynchronized.ref (Value _)) => aref |
      (Unsynchronized.ref (Upd (i,v,cr))) => realize_aux(aref,v);

  fun update (aref,idx,v) =
    case aref of
      (Unsynchronized.ref (Value a)) => (
        let val nref=Unsynchronized.ref (Value a) in
          aref := Upd (idx,Array.sub(a,idx),nref);
          Array.update (a,idx,v);
          nref
        end
      ) |
      (Unsynchronized.ref (Upd _)) =>
        let val ra = realize_aux(aref,v) in
          case ra of
            (Unsynchronized.ref (Value a)) => Array.update (a,idx,v);
          ra
        end
      ;

  fun grow (aref, inc, x) = case aref of 
    (Unsynchronized.ref (Value a)) => (
      let val len=Array.length a;
          val na = Array.array (len+inc,x) 
      in
        Array.copy {src=a, dst=na, di=0};
        Unsynchronized.ref (Value na)
      end
      )
  | (Unsynchronized.ref (Upd _)) => (
    grow (realize aref, inc, x)
  );

  fun shrink (aref, sz) = case aref of
    (Unsynchronized.ref (Value a)) => (
      if sz > Array.length a then 
        raise Size
      else (
        Unsynchronized.ref (Value (Array.tabulate (sz,fn i => Array.sub (a,i))))
      )
    ) |
    (Unsynchronized.ref (Upd _)) => (
      shrink (realize aref,sz)
    );

structure IsabelleMapping = struct
type 'a ArrayType = 'a array;

fun new_array (a:'a) (n:int) = array (n, a);

fun array_length (a:'a ArrayType) = length a;

fun array_get (a:'a ArrayType) (i:int) = sub (a, i);

fun array_set (a:'a ArrayType) (i:int) (e:'a) = update (a, i, e);

fun array_of_list (xs:'a list) = fromList xs;

fun array_grow (a:'a ArrayType) (i:int) (x:'a) = grow (a, i, x);

fun array_shrink (a:'a ArrayType) (sz:int) = shrink (a,sz);

fun array_get_oo (d:'a) (a:'a ArrayType) (i:int) = 
  sub (a,i) handle Subscript => d

fun array_set_oo (d:(unit->'a ArrayType)) (a:'a ArrayType) (i:int) (e:'a) = 
  update (a, i, e) handle Subscript => d ()

end;
end;




structure Ta : sig
  val id : 'a -> 'a
  type 'a equal
  val equal : 'a equal -> 'a -> 'a -> bool
  val eq : 'a equal -> 'a -> 'a -> bool
  datatype nat = Nat of IntInf.int
  datatype num = One | Bit0 of num | Bit1 of num
  type 'a ord
  val less_eq : 'a ord -> 'a -> 'a -> bool
  val less : 'a ord -> 'a -> 'a -> bool
  type 'a preorder
  val ord_preorder : 'a preorder -> 'a ord
  type 'a order
  val preorder_order : 'a order -> 'a preorder
  type 'a linorder
  val order_linorder : 'a linorder -> 'a order
  datatype color = R | B
  datatype ('a, 'b) rbta = Empty |
    Branch of color * ('a, 'b) rbta * 'a * 'b * ('a, 'b) rbta
  datatype ('a, 'b) rbt = Rbt of ('a, 'b) rbta
  datatype ('a, 'b) ta_rule = Rule of 'a * 'b * 'a list
  val lhs : ('a, 'b) ta_rule -> 'a
  val hd : 'a list -> 'a
  val tl : 'a list -> 'a list
  val integer_of_nat : nat -> IntInf.int
  val plus_nat : nat -> nat -> nat
  val one_nat : nat
  val suc : nat -> nat
  val rhsl : ('a, 'b) ta_rule -> 'b
  val rhsq : ('a, 'b) ta_rule -> 'a list
  datatype ('a, 'b) ustate_wrapper = USW1 of 'a | USW2 of 'b
  val map : ('a -> 'b) -> 'a list -> 'b list
  val fold : ('a -> 'b -> 'b) -> 'a list -> 'b -> 'b
  val rev : 'a list -> 'a list
  val zip : 'a list -> 'b list -> ('a * 'b) list
  datatype 'a tree = Node of 'a * 'a tree list
  val null : 'a list -> bool
  datatype 'a itself = Type
  val empty : 'a linorder -> ('a, 'b) rbt
  val r_prod : ('a, 'b) ta_rule -> ('c, 'd) ta_rule -> (('a * 'c), 'b) ta_rule
  datatype 'a dlist = Dlist of 'a list
  val foldl : ('a -> 'b -> 'a) -> 'a -> 'b list -> 'a
  val map_of : 'a equal -> ('a * 'b) list -> 'a -> 'b option
  val less_eq_nat : nat -> nat -> bool
  val less_nat : nat -> nat -> bool
  val ord_nat : nat ord
  val the : 'a option -> 'a
  val balance : ('a, 'b) rbta -> 'a -> 'b -> ('a, 'b) rbta -> ('a, 'b) rbta
  val rbt_ins :
    'a ord ->
      ('a -> 'b -> 'b -> 'b) -> 'a -> 'b -> ('a, 'b) rbta -> ('a, 'b) rbta
  val paint : color -> ('a, 'b) rbta -> ('a, 'b) rbta
  val rbt_insert_with_key :
    'a ord ->
      ('a -> 'b -> 'b -> 'b) -> 'a -> 'b -> ('a, 'b) rbta -> ('a, 'b) rbta
  val rbt_insert : 'a ord -> 'a -> 'b -> ('a, 'b) rbta -> ('a, 'b) rbta
  val impl_of : 'a linorder -> ('a, 'b) rbt -> ('a, 'b) rbta
  val insert : 'a linorder -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val rbt_lookup : 'a ord -> ('a, 'b) rbta -> 'a -> 'b option
  val lookup : 'a linorder -> ('a, 'b) rbt -> 'a -> 'b option
  val emptya : 'a dlist
  val member : 'a equal -> 'a list -> 'a -> bool
  val inserta : 'a equal -> 'a -> 'a list -> 'a list
  val list_of_dlist : 'a dlist -> 'a list
  val insertb : 'a equal -> 'a -> 'a dlist -> 'a dlist
  val membera : 'a equal -> 'a dlist -> 'a -> bool
  val foldli : 'a list -> ('b -> bool) -> ('a -> 'b -> 'b) -> 'b -> 'b
  val remdups : 'a equal -> 'a list -> 'a list
  val equal_nata : nat -> nat -> bool
  val equal_nat : nat equal
  val preorder_nat : nat preorder
  val order_nat : nat order
  val zero_nat : nat
  val list_all : ('a -> bool) -> 'a list -> bool
  val ord_integer : IntInf.int ord
  val max : 'a ord -> 'a -> 'a -> 'a
  val minus_nat : nat -> nat -> nat
  val times_nat : nat -> nat -> nat
  val remap_rule : ('a -> 'b) -> ('a, 'c) ta_rule -> ('b, 'c) ta_rule
  val dlist_iteratei : 'a dlist -> ('b -> bool) -> ('a -> 'b -> 'b) -> 'b -> 'b
  val iteratei_set_op_list_it_ls_ops :
    'a dlist -> ('b -> bool) -> ('a -> 'b -> 'b) -> 'b -> 'b
  val g_image_filter_ls_ops_ls_ops :
    'b equal -> ('a -> 'b option) -> 'a dlist -> 'b dlist
  datatype ('a, 'b) assoc_list = Assoc_List of ('a * 'b) list
  type 'a hashable
  val hashcode : 'a hashable -> 'a -> nat
  val bounded_hashcode : 'a hashable -> nat -> 'a -> nat
  val def_hashmap_size : 'a hashable -> 'a itself -> nat
  datatype ('a, 'b) hashmap = RBT_HM of (nat, ('a, 'b) assoc_list) rbt
  val impl_of_RBT_HM :
    'a hashable -> ('a, 'b) hashmap -> (nat, ('a, 'b) assoc_list) rbt
  val impl_ofa : ('a, 'b) assoc_list -> ('a * 'b) list
  val lookupa : 'a equal -> ('a, 'b) assoc_list -> 'a -> 'b option
  val linorder_nat : nat linorder
  val lookupb :
    'a equal * 'a hashable -> 'a -> (nat, ('a, 'b) assoc_list) rbt -> 'b option
  val hm_lookup : 'a equal * 'a hashable -> 'a -> ('a, 'b) hashmap -> 'b option
  val lookup_hm_ops_ls_ops :
    'a equal * 'a hashable -> 'a -> ('a, 'b dlist) hashmap -> 'b dlist
  datatype ('a, 'b, 'c) hashedTa_ext =
    HashedTa_ext of
      ('a, unit) hashmap * ('a, 'b) ta_rule dlist *
        ('b, ('a, 'b) ta_rule dlist) hashmap option *
        ('a, ('a, 'b) ta_rule dlist) hashmap option *
        (('a * 'b), ('a, 'b) ta_rule dlist) hashmap option * 'c
  val hta_idx_f :
    ('a, 'b, 'c) hashedTa_ext -> ('b, ('a, 'b) ta_rule dlist) hashmap option
  val hta_lookup_f :
    'a equal * 'a hashable ->
      'a -> ('b, 'a, 'c) hashedTa_ext -> ('b, 'a) ta_rule dlist
  val list_all_zip : ('a -> 'b -> bool) -> 'a list -> 'b list -> bool
  val faccs :
    'a equal -> 'b equal * 'b hashable ->
      ('a, 'b, 'c) hashedTa_ext -> 'b tree -> 'a dlist
  val equal_lista : 'a equal -> 'a list -> 'a list -> bool
  val equal_list : 'a equal -> ('a list) equal
  val gen_length : nat -> 'a list -> nat
  val size_list : 'a list -> nat
  val is_none : 'a option -> bool
  val emptyc : 'a hashable -> unit -> (nat, ('a, 'b) assoc_list) rbt
  val hm_empty_const : 'a hashable -> ('a, 'b) hashmap
  val hm_empty : 'a hashable -> unit -> ('a, 'b) hashmap
  val empty_hm_basic_ops : 'a hashable -> unit -> ('a, unit) hashmap
  val snd : 'a * 'b -> 'b
  val fst : 'a * 'b -> 'a
  val update_with_aux :
    'b equal -> 'a -> 'b -> ('a -> 'a) -> ('b * 'a) list -> ('b * 'a) list
  val update_with :
    'a equal ->
      'b -> 'a -> ('b -> 'b) -> ('a, 'b) assoc_list -> ('a, 'b) assoc_list
  val update :
    'a equal -> 'a -> 'b -> ('a, 'b) assoc_list -> ('a, 'b) assoc_list
  val emptyb : ('a, 'b) assoc_list
  val updatea :
    'a equal * 'a hashable ->
      'a -> 'b -> (nat, ('a, 'b) assoc_list) rbt ->
                    (nat, ('a, 'b) assoc_list) rbt
  val hm_update :
    'a equal * 'a hashable -> 'a -> 'b -> ('a, 'b) hashmap -> ('a, 'b) hashmap
  val ins_hm_basic_ops :
    'a equal * 'a hashable -> 'a -> ('a, unit) hashmap -> ('a, unit) hashmap
  val g_image_filter_ls_ops_hs_ops :
    'b equal * 'b hashable ->
      ('a -> 'b option) -> 'a dlist -> ('b, unit) hashmap
  val brc_iq :
    'a equal * 'a hashable -> ('a, 'b) ta_rule dlist -> ('a, unit) hashmap
  val hta_Qi : ('a, 'b, 'c) hashedTa_ext -> ('a, unit) hashmap
  val sgn_integer : IntInf.int -> IntInf.int
  val abs_integer : IntInf.int -> IntInf.int
  val apsnd : ('a -> 'b) -> 'c * 'a -> 'c * 'b
  val divmod_integer : IntInf.int -> IntInf.int -> IntInf.int * IntInf.int
  val mod_integer : IntInf.int -> IntInf.int -> IntInf.int
  val mod_nat : nat -> nat -> nat
  val fifo_empty : unit -> 'a list * 'a list
  val equal_ta_rulea :
    'a equal -> 'b equal -> ('a, 'b) ta_rule -> ('a, 'b) ta_rule -> bool
  val equal_ta_rule : 'a equal -> 'b equal -> ('a, 'b) ta_rule equal
  val idx_build_stepfun_hm_ops_ls_ops :
    'a equal -> 'b equal * 'b hashable ->
      ('a -> 'b) -> 'a -> ('b, 'a dlist) hashmap -> ('b, 'a dlist) hashmap
  val idx_build_hm_ops_ls_ops_ls_ops :
    'a equal -> 'b equal * 'b hashable ->
      ('a -> 'b) -> 'a dlist -> ('b, 'a dlist) hashmap
  val build_rule_index :
    'a equal -> 'b equal -> 'c equal * 'c hashable ->
      (('a, 'b) ta_rule -> 'c) ->
        ('a, 'b) ta_rule dlist -> ('c, ('a, 'b) ta_rule dlist) hashmap
  val build_rule_index_f :
    'a equal -> 'b equal * 'b hashable ->
      ('a, 'b) ta_rule dlist -> ('b, ('a, 'b) ta_rule dlist) hashmap
  val hta_idx_f_update :
    (('a, ('b, 'a) ta_rule dlist) hashmap option ->
      ('a, ('b, 'a) ta_rule dlist) hashmap option) ->
      ('b, 'a, 'c) hashedTa_ext -> ('b, 'a, 'c) hashedTa_ext
  val hta_delta : ('a, 'b, 'c) hashedTa_ext -> ('a, 'b) ta_rule dlist
  val hta_ensure_idx_f :
    'a equal -> 'b equal * 'b hashable ->
      ('a, 'b, 'c) hashedTa_ext -> ('a, 'b, 'c) hashedTa_ext
  val iteratei_bset_op_list_it_ls_basic_ops :
    'a dlist -> ('b -> bool) -> ('a -> 'b -> 'b) -> 'b -> 'b
  val g_ball_ls_basic_ops : 'a dlist -> ('a -> bool) -> bool
  val memb_hm_basic_ops :
    'a equal * 'a hashable -> 'a -> ('a, unit) hashmap -> bool
  val g_disjoint_ls_ops_hs_ops :
    'a equal * 'a hashable -> 'a dlist -> ('a, unit) hashmap -> bool
  val hta_mema :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      'a tree -> ('b, 'a, 'c) hashedTa_ext -> bool
  val hta_mem :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      'a tree -> ('b, 'a, 'c) hashedTa_ext -> bool
  val pa_cond :
    'a hashable -> 'b hashable -> 'c hashable ->
      (('a * 'b), unit) hashmap *
        (('a * 'b) list * (('a * 'b), 'c) ta_rule dlist) ->
        bool
  val equal_proda : 'a equal -> 'b equal -> 'a * 'b -> 'a * 'b -> bool
  val equal_prod : 'a equal -> 'b equal -> ('a * 'b) equal
  val def_hashmap_size_prod :
    'a hashable -> 'b hashable -> ('a * 'b) itself -> nat
  val nat_of_integer : IntInf.int -> nat
  val bounded_hashcode_prod :
    'a hashable -> 'b hashable -> nat -> 'a * 'b -> nat
  val hashcode_prod : 'a hashable -> 'b hashable -> 'a * 'b -> nat
  val hashable_prod : 'a hashable -> 'b hashable -> ('a * 'b) hashable
  val hta_idx_sf :
    ('a, 'b, 'c) hashedTa_ext ->
      (('a * 'b), ('a, 'b) ta_rule dlist) hashmap option
  val hta_lookup_sf :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      'a -> 'b -> ('a, 'b, 'c) hashedTa_ext -> ('a, 'b) ta_rule dlist
  val hta_idx_s :
    ('a, 'b, 'c) hashedTa_ext -> ('a, ('a, 'b) ta_rule dlist) hashmap option
  val hta_lookup_s :
    'a equal * 'a hashable ->
      'a -> ('a, 'b, 'c) hashedTa_ext -> ('a, 'b) ta_rule dlist
  val pa_upd_rule :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      (('a * 'b), unit) hashmap ->
        ('a * 'b) list ->
          ('a * 'b) list -> (('a * 'b), unit) hashmap * ('a * 'b) list
  val pa_step :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      'c equal * 'c hashable ->
      ('a, 'b, unit) hashedTa_ext ->
        ('c, 'b, unit) hashedTa_ext ->
          (('a * 'c), unit) hashmap *
            (('a * 'c) list * (('a * 'c), 'b) ta_rule dlist) ->
            (('a * 'c), unit) hashmap *
              (('a * 'c) list * (('a * 'c), 'b) ta_rule dlist)
  val equal_dlista : 'a equal -> 'a dlist -> 'a dlist -> bool
  val equal_dlist : 'a equal -> 'a dlist equal
  val brc_cond :
    'a hashable -> 'b hashable ->
      ('a, unit) hashmap * ('a list * (('a, 'b) ta_rule, nat) hashmap) -> bool
  val def_hashmap_size_ta_rule :
    'a hashable -> 'b hashable -> ('a, 'b) ta_rule itself -> nat
  val hashcode_list : 'a hashable -> 'a list -> nat
  val hashcode_of_ta_rule :
    'a hashable -> 'b hashable -> ('a, 'b) ta_rule -> nat
  val bounded_hashcode_ta_rule :
    'a hashable -> 'b hashable -> nat -> ('a, 'b) ta_rule -> nat
  val hashcode_ta_rule : 'a hashable -> 'b hashable -> ('a, 'b) ta_rule -> nat
  val hashable_ta_rule : 'a hashable -> 'b hashable -> ('a, 'b) ta_rule hashable
  val brc_inner_step :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, 'b) ta_rule ->
        ('a, unit) hashmap * ('a list * (('a, 'b) ta_rule, nat) hashmap) ->
          ('a, unit) hashmap * ('a list * (('a, 'b) ta_rule, nat) hashmap)
  val rqrm_lookup :
    'a equal * 'a hashable -> ('a, 'b dlist) hashmap -> 'a -> 'b dlist
  val brc_step :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, ('a, 'b) ta_rule dlist) hashmap ->
        ('a, unit) hashmap * ('a list * (('a, 'b) ta_rule, nat) hashmap) ->
          ('a, unit) hashmap * ('a list * (('a, 'b) ta_rule, nat) hashmap)
  val brec_iqm :
    'a equal * 'a hashable -> 'b hashable ->
      ('a, 'b) ta_rule dlist -> ('a, 'b tree) hashmap
  val build_rule_index_sf :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, 'b) ta_rule dlist -> (('a * 'b), ('a, 'b) ta_rule dlist) hashmap
  val hta_idx_sf_update :
    ((('a * 'b), ('a, 'b) ta_rule dlist) hashmap option ->
      (('a * 'b), ('a, 'b) ta_rule dlist) hashmap option) ->
      ('a, 'b, 'c) hashedTa_ext -> ('a, 'b, 'c) hashedTa_ext
  val hta_ensure_idx_sf :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, 'b, 'c) hashedTa_ext -> ('a, 'b, 'c) hashedTa_ext
  val build_rule_index_s :
    'a equal * 'a hashable -> 'b equal ->
      ('a, 'b) ta_rule dlist -> ('a, ('a, 'b) ta_rule dlist) hashmap
  val hta_idx_s_update :
    (('a, ('a, 'b) ta_rule dlist) hashmap option ->
      ('a, ('a, 'b) ta_rule dlist) hashmap option) ->
      ('a, 'b, 'c) hashedTa_ext -> ('a, 'b, 'c) hashedTa_ext
  val hta_ensure_idx_s :
    'a equal * 'a hashable -> 'b equal ->
      ('a, 'b, 'c) hashedTa_ext -> ('a, 'b, 'c) hashedTa_ext
  val iteratei :
    ('a, 'b) assoc_list -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
  val iteratei_map_op_list_it_lm_ops :
    ('a, 'b) assoc_list -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
  val rm_iterateoi :
    ('a, 'b) rbta -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
  val iteratei_map_op_list_it_rm_ops :
    'a linorder ->
      ('a, 'b) rbt -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
  val iterateia :
    'a linorder ->
      ('a, ('b, 'c) assoc_list) rbt ->
        ('d -> bool) -> ('b * 'c -> 'd -> 'd) -> 'd -> 'd
  val hm_iteratei :
    'a hashable ->
      ('a, 'b) hashmap -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
  val iteratei_set_op_list_it_hs_ops :
    'a hashable ->
      ('a, unit) hashmap -> ('b -> bool) -> ('a -> 'b -> 'b) -> 'b -> 'b
  val ins_dj_hm_basic_ops :
    'a equal * 'a hashable -> 'a -> ('a, unit) hashmap -> ('a, unit) hashmap
  val cart_hs_ops_hs_ops_hs_ops :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, unit) hashmap -> ('b, unit) hashmap -> (('a * 'b), unit) hashmap
  val whilea : ('a -> bool) -> ('a -> 'a) -> 'a -> 'a
  val iteratei_bset_op_list_it_dflt_basic_ops_hm_basic_ops :
    'a hashable ->
      ('a, unit) hashmap -> ('b -> bool) -> ('a -> 'b -> 'b) -> 'b -> 'b
  val g_to_list_dflt_basic_ops_hm_basic_ops :
    'a hashable -> ('a, unit) hashmap -> 'a list
  val pa_initial :
    'a equal * 'a hashable -> 'b hashable -> 'c equal * 'c hashable ->
      ('a, 'b, unit) hashedTa_ext ->
        ('c, 'b, unit) hashedTa_ext ->
          (('a * 'c), unit) hashmap *
            (('a * 'c) list * (('a * 'c), 'b) ta_rule dlist)
  val init_hta :
    ('a, unit) hashmap -> ('a, 'b) ta_rule dlist -> ('a, 'b, unit) hashedTa_ext
  val hta_proda :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      'c equal * 'c hashable ->
      ('a, 'b, unit) hashedTa_ext ->
        ('c, 'b, unit) hashedTa_ext -> (('a * 'c), 'b, unit) hashedTa_ext
  val hta_prod :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      'c equal * 'c hashable ->
      ('a, 'b, unit) hashedTa_ext ->
        ('c, 'b, unit) hashedTa_ext -> (('a * 'c), 'b, unit) hashedTa_ext
  val def_hashmap_size_nat : nat itself -> nat
  val bounded_hashcode_nat : nat -> nat -> nat
  val hashcode_nat : nat -> nat
  val hashable_nat : nat hashable
  val htai_mem : nat tree -> (nat, nat, unit) hashedTa_ext -> bool
  val rqrm_add :
    'a equal * 'a hashable -> 'b equal ->
      'a -> 'b -> ('a, 'b dlist) hashmap -> ('a, 'b dlist) hashmap
  val fifo_appendr : 'a -> 'a list * 'a list -> 'a list * 'a list
  val fifo_isEmpty : 'a equal -> 'a list * 'a list -> bool
  val fifo_removel : 'a list * 'a list -> 'a * ('a list * 'a list)
  val brec_cond :
    'a equal ->
      ('a, 'b tree) hashmap *
        (('a list * 'a list) * ((('a, 'b) ta_rule, nat) hashmap * 'a option)) ->
        bool
  val brec_construct_witness :
    'a equal * 'a hashable -> 'b hashable ->
      ('a, 'b tree) hashmap -> ('a, 'b) ta_rule -> 'b tree
  val brec_inner_step :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, unit) hashmap ->
        ('a, 'b) ta_rule ->
          ('a, 'b tree) hashmap *
            (('a list * 'a list) *
              ((('a, 'b) ta_rule, nat) hashmap * 'a option)) ->
            ('a, 'b tree) hashmap *
              (('a list * 'a list) *
                ((('a, 'b) ta_rule, nat) hashmap * 'a option))
  val brec_step :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, ('a, 'b) ta_rule dlist) hashmap ->
        ('a, unit) hashmap ->
          ('a, 'b tree) hashmap *
            (('a list * 'a list) *
              ((('a, 'b) ta_rule, nat) hashmap * 'a option)) ->
            ('a, 'b tree) hashmap *
              (('a list * 'a list) *
                ((('a, 'b) ta_rule, nat) hashmap * 'a option))
  val hta_empty :
    'a hashable -> 'b hashable -> unit -> ('a, 'b, unit) hashedTa_ext
  val g_image_ls_ops_ls_ops : 'b equal -> ('a -> 'b) -> 'a dlist -> 'b dlist
  val g_image_hs_ops_hs_ops :
    'a hashable -> 'b equal * 'b hashable ->
      ('a -> 'b) -> ('a, unit) hashmap -> ('b, unit) hashmap
  val hta_remap :
    'a hashable -> 'b equal * 'b hashable -> 'c equal * 'c hashable ->
      ('a -> 'b) -> ('a, 'c, unit) hashedTa_ext -> ('b, 'c, unit) hashedTa_ext
  val g_union_dflt_basic_ops_hm_basic_ops :
    'a equal * 'a hashable ->
      ('a, unit) hashmap -> ('a, unit) hashmap -> ('a, unit) hashmap
  val g_union_dj_ls_basic_ops : 'a equal -> 'a dlist -> 'a dlist -> 'a dlist
  val def_hashmap_size_ustate_wrapper :
    'a hashable -> 'b hashable -> ('a, 'b) ustate_wrapper itself -> nat
  val bounded_hashcode_ustate_wrapper :
    'a hashable -> 'b hashable -> nat -> ('a, 'b) ustate_wrapper -> nat
  val hashcode_ustate_wrapper :
    'a hashable -> 'b hashable -> ('a, 'b) ustate_wrapper -> nat
  val hashable_ustate_wrapper :
    'a hashable -> 'b hashable -> ('a, 'b) ustate_wrapper hashable
  val equal_ustate_wrappera :
    'a equal -> 'b equal ->
      ('a, 'b) ustate_wrapper -> ('a, 'b) ustate_wrapper -> bool
  val equal_ustate_wrapper :
    'a equal -> 'b equal -> ('a, 'b) ustate_wrapper equal
  val hta_union :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      'c equal * 'c hashable ->
      ('a, 'b, unit) hashedTa_ext ->
        ('c, 'b, unit) hashedTa_ext ->
          (('a, 'c) ustate_wrapper, 'b, unit) hashedTa_ext
  val map_to_nat_hs_ops_hm_ops :
    'a equal * 'a hashable -> ('a, unit) hashmap -> ('a, nat) hashmap
  val g_union_ls_ops_hs_ops :
    'a equal * 'a hashable ->
      'a dlist -> ('a, unit) hashmap -> ('a, unit) hashmap
  val g_Union_image_ls_ops_ls_ops_hs_ops :
    'b equal * 'b hashable -> ('a -> 'b dlist) -> 'a dlist -> ('b, unit) hashmap
  val g_from_list_aux_ls_basic_ops : 'a equal -> 'a dlist -> 'a list -> 'a dlist
  val g_from_list_ls_basic_ops : 'a equal -> 'a list -> 'a dlist
  val rule_states_l : 'a equal -> ('a, 'b) ta_rule -> 'a dlist
  val hta_delta_states :
    'a equal * 'a hashable -> ('a, 'b, 'c) hashedTa_ext -> ('a, unit) hashmap
  val hta_states :
    'a equal * 'a hashable -> ('a, 'b, 'c) hashedTa_ext -> ('a, unit) hashmap
  val hta_reindex :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, 'b, unit) hashedTa_ext -> (nat, 'b, unit) hashedTa_ext
  val htai_prod :
    (nat, nat, unit) hashedTa_ext ->
      (nat, nat, unit) hashedTa_ext -> (nat, nat, unit) hashedTa_ext
  val build_rqrm :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, 'b) ta_rule dlist -> ('a, ('a, 'b) ta_rule dlist) hashmap
  val hta_add_qi :
    'a equal * 'a hashable -> 'b hashable ->
      'a -> ('a, 'b, unit) hashedTa_ext -> ('a, 'b, unit) hashedTa_ext
  val inj_image_filter_cartesian_product_ls_ops_ls_ops_ls_ops :
    'c equal -> ('a * 'b -> 'c option) -> 'a dlist -> 'b dlist -> 'c dlist
  val inj_image_filter_cp_ls_ops_ls_ops_ls_ops :
    'c equal ->
      ('a * 'b -> 'c) -> ('a * 'b -> bool) -> 'a dlist -> 'b dlist -> 'c dlist
  val delta_prod_h :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      'c equal * 'c hashable ->
      ('a, 'b) ta_rule dlist ->
        ('c, 'b) ta_rule dlist -> (('a * 'c), 'b) ta_rule dlist
  val hta_prodWR :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      'd equal * 'd hashable ->
      ('a, 'b, 'c) hashedTa_ext ->
        ('d, 'b, 'e) hashedTa_ext -> (('a * 'd), 'b, unit) hashedTa_ext
  val g_inter_dflt_basic_ops_hm_basic_ops :
    'a equal * 'a hashable ->
      ('a, unit) hashmap -> ('a, unit) hashmap -> ('a, unit) hashmap
  val hta_reduce :
    'a equal * 'a hashable -> 'b equal ->
      ('a, 'b, 'c) hashedTa_ext ->
        ('a, unit) hashmap -> ('a, 'b, unit) hashedTa_ext
  val htai_empty : unit -> (nat, nat, unit) hashedTa_ext
  val htai_union :
    (nat, nat, unit) hashedTa_ext ->
      (nat, nat, unit) hashedTa_ext -> (nat, nat, unit) hashedTa_ext
  val brc_rcm_init :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, 'b) ta_rule dlist -> (('a, 'b) ta_rule, nat) hashmap
  val brc_initial :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, 'b) ta_rule dlist ->
        ('a, unit) hashmap * ('a list * (('a, 'b) ta_rule, nat) hashmap)
  val htai_add_qi :
    nat -> (nat, nat, unit) hashedTa_ext -> (nat, nat, unit) hashedTa_ext
  val htai_prodWR :
    (nat, nat, unit) hashedTa_ext ->
      (nat, nat, unit) hashedTa_ext -> (nat, nat, unit) hashedTa_ext
  val g_sel_dflt_basic_ops_hm_basic_ops :
    'a hashable -> ('a, unit) hashmap -> ('a -> bool) -> 'a option
  val g_disjoint_witness_hs_ops_hs_ops :
    'a equal * 'a hashable ->
      ('a, unit) hashmap -> ('a, unit) hashmap -> 'a option
  val g_set_to_listr_hs_ops_fifo_ops :
    'a hashable -> ('a, unit) hashmap -> 'a list * 'a list
  val brec_initial :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, unit) hashmap ->
        ('a, 'b) ta_rule dlist ->
          ('a, 'b tree) hashmap *
            (('a list * 'a list) *
              ((('a, 'b) ta_rule, nat) hashmap * 'a option))
  val hta_add_rule :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, 'b) ta_rule ->
        ('a, 'b, unit) hashedTa_ext -> ('a, 'b, unit) hashedTa_ext
  val htai_add_rule :
    (nat, nat) ta_rule ->
      (nat, nat, unit) hashedTa_ext -> (nat, nat, unit) hashedTa_ext
  val hta_bwd_reduce :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, 'b, 'c) hashedTa_ext -> ('a, 'b, unit) hashedTa_ext
  val htai_bwd_reduce :
    (nat, nat, unit) hashedTa_ext -> (nat, nat, unit) hashedTa_ext
  val g_size_dflt_basic_ops_hm_basic_ops :
    'a hashable -> ('a, unit) hashmap -> nat
  val hs_size : 'a hashable -> ('a, unit) hashmap -> nat
  val g_size_ls_basic_ops : 'a dlist -> nat
  val ls_size : 'a dlist -> nat
  val iteratei_bset_op_list_it_dflt_basic_oops_rm_basic_ops :
    'a linorder ->
      ('a, unit) rbt -> ('b -> bool) -> ('a -> 'b -> 'b) -> 'b -> 'b
  val g_size_dflt_basic_oops_rm_basic_ops : 'a linorder -> ('a, unit) rbt -> nat
  val rs_size : 'a linorder -> ('a, unit) rbt -> nat
  val htai_ensure_idx_f :
    (nat, nat, unit) hashedTa_ext -> (nat, nat, unit) hashedTa_ext
  val htai_ensure_idx_s :
    (nat, nat, unit) hashedTa_ext -> (nat, nat, unit) hashedTa_ext
  val htai_ensure_idx_sf :
    (nat, nat, unit) hashedTa_ext -> (nat, nat, unit) hashedTa_ext
  val hta_is_empty_witness :
    'a equal * 'a hashable -> 'b equal * 'b hashable ->
      ('a, 'b, 'c) hashedTa_ext -> 'b tree option
  val htai_is_empty_witness : (nat, nat, unit) hashedTa_ext -> nat tree option
end = struct

fun id x = (fn xa => xa) x;

type 'a equal = {equal : 'a -> 'a -> bool};
val equal = #equal : 'a equal -> 'a -> 'a -> bool;

fun eq A_ a b = equal A_ a b;

datatype nat = Nat of IntInf.int;

datatype num = One | Bit0 of num | Bit1 of num;

type 'a ord = {less_eq : 'a -> 'a -> bool, less : 'a -> 'a -> bool};
val less_eq = #less_eq : 'a ord -> 'a -> 'a -> bool;
val less = #less : 'a ord -> 'a -> 'a -> bool;

type 'a preorder = {ord_preorder : 'a ord};
val ord_preorder = #ord_preorder : 'a preorder -> 'a ord;

type 'a order = {preorder_order : 'a preorder};
val preorder_order = #preorder_order : 'a order -> 'a preorder;

type 'a linorder = {order_linorder : 'a order};
val order_linorder = #order_linorder : 'a linorder -> 'a order;

datatype color = R | B;

datatype ('a, 'b) rbta = Empty |
  Branch of color * ('a, 'b) rbta * 'a * 'b * ('a, 'b) rbta;

datatype ('a, 'b) rbt = Rbt of ('a, 'b) rbta;

datatype ('a, 'b) ta_rule = Rule of 'a * 'b * 'a list;

fun lhs (Rule (q, l, qs)) = q;

fun hd (x :: xs) = x;

fun tl [] = []
  | tl (x :: xs) = xs;

fun integer_of_nat (Nat x) = x;

fun plus_nat m n = Nat (IntInf.+ (integer_of_nat m, integer_of_nat n));

val one_nat : nat = Nat (1 : IntInf.int);

fun suc n = plus_nat n one_nat;

fun rhsl (Rule (q, l, qs)) = l;

fun rhsq (Rule (q, l, qs)) = qs;

datatype ('a, 'b) ustate_wrapper = USW1 of 'a | USW2 of 'b;

fun map f [] = []
  | map f (x :: xs) = f x :: map f xs;

fun fold f (x :: xs) s = fold f xs (f x s)
  | fold f [] s = s;

fun rev xs = fold (fn a => fn b => a :: b) xs [];

fun zip (x :: xs) (y :: ys) = (x, y) :: zip xs ys
  | zip xs [] = []
  | zip [] ys = [];

datatype 'a tree = Node of 'a * 'a tree list;

fun null [] = true
  | null (x :: xs) = false;

datatype 'a itself = Type;

fun empty A_ = Rbt Empty;

fun r_prod (Rule (q1, l1, qs1)) (Rule (q2, l2, qs2)) =
  Rule ((q1, q2), l1, zip qs1 qs2);

datatype 'a dlist = Dlist of 'a list;

fun foldl f a [] = a
  | foldl f a (x :: xs) = foldl f (f a x) xs;

fun map_of A_ ((l, v) :: ps) k = (if eq A_ l k then SOME v else map_of A_ ps k)
  | map_of A_ [] k = NONE;

fun less_eq_nat m n = IntInf.<= (integer_of_nat m, integer_of_nat n);

fun less_nat m n = IntInf.< (integer_of_nat m, integer_of_nat n);

val ord_nat = {less_eq = less_eq_nat, less = less_nat} : nat ord;

fun the (SOME x) = x;

fun balance (Branch (R, a, w, x, b)) s t (Branch (R, c, y, z, d)) =
  Branch (R, Branch (B, a, w, x, b), s, t, Branch (B, c, y, z, d))
  | balance (Branch (R, Branch (R, a, w, x, b), s, t, c)) y z Empty =
    Branch (R, Branch (B, a, w, x, b), s, t, Branch (B, c, y, z, Empty))
  | balance (Branch (R, Branch (R, a, w, x, b), s, t, c)) y z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, Branch (B, a, w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance (Branch (R, Empty, w, x, Branch (R, b, s, t, c))) y z Empty =
    Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, Empty))
  | balance
    (Branch (R, Branch (B, va, vb, vc, vd), w, x, Branch (R, b, s, t, c))) y z
    Empty =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, Empty))
  | balance (Branch (R, Empty, w, x, Branch (R, b, s, t, c))) y z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, Branch (B, Empty, w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance
    (Branch (R, Branch (B, ve, vf, vg, vh), w, x, Branch (R, b, s, t, c))) y z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, Branch (B, Branch (B, ve, vf, vg, vh), w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance Empty w x (Branch (R, b, s, t, Branch (R, c, y, z, d))) =
    Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, d))
  | balance (Branch (B, va, vb, vc, vd)) w x
    (Branch (R, b, s, t, Branch (R, c, y, z, d))) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, d))
  | balance Empty w x (Branch (R, Branch (R, b, s, t, c), y, z, Empty)) =
    Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, Empty))
  | balance Empty w x
    (Branch (R, Branch (R, b, s, t, c), y, z, Branch (B, va, vb, vc, vd))) =
    Branch
      (R, Branch (B, Empty, w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance (Branch (B, va, vb, vc, vd)) w x
    (Branch (R, Branch (R, b, s, t, c), y, z, Empty)) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, Empty))
  | balance (Branch (B, va, vb, vc, vd)) w x
    (Branch (R, Branch (R, b, s, t, c), y, z, Branch (B, ve, vf, vg, vh))) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, ve, vf, vg, vh)))
  | balance Empty s t Empty = Branch (B, Empty, s, t, Empty)
  | balance Empty s t (Branch (B, va, vb, vc, vd)) =
    Branch (B, Empty, s, t, Branch (B, va, vb, vc, vd))
  | balance Empty s t (Branch (v, Empty, vb, vc, Empty)) =
    Branch (B, Empty, s, t, Branch (v, Empty, vb, vc, Empty))
  | balance Empty s t (Branch (v, Branch (B, ve, vf, vg, vh), vb, vc, Empty)) =
    Branch
      (B, Empty, s, t, Branch (v, Branch (B, ve, vf, vg, vh), vb, vc, Empty))
  | balance Empty s t (Branch (v, Empty, vb, vc, Branch (B, vf, vg, vh, vi))) =
    Branch
      (B, Empty, s, t, Branch (v, Empty, vb, vc, Branch (B, vf, vg, vh, vi)))
  | balance Empty s t
    (Branch (v, Branch (B, ve, vj, vk, vl), vb, vc, Branch (B, vf, vg, vh, vi)))
    = Branch
        (B, Empty, s, t,
          Branch
            (v, Branch (B, ve, vj, vk, vl), vb, vc, Branch (B, vf, vg, vh, vi)))
  | balance (Branch (B, va, vb, vc, vd)) s t Empty =
    Branch (B, Branch (B, va, vb, vc, vd), s, t, Empty)
  | balance (Branch (B, va, vb, vc, vd)) s t (Branch (B, ve, vf, vg, vh)) =
    Branch (B, Branch (B, va, vb, vc, vd), s, t, Branch (B, ve, vf, vg, vh))
  | balance (Branch (B, va, vb, vc, vd)) s t (Branch (v, Empty, vf, vg, Empty))
    = Branch
        (B, Branch (B, va, vb, vc, vd), s, t, Branch (v, Empty, vf, vg, Empty))
  | balance (Branch (B, va, vb, vc, vd)) s t
    (Branch (v, Branch (B, vi, vj, vk, vl), vf, vg, Empty)) =
    Branch
      (B, Branch (B, va, vb, vc, vd), s, t,
        Branch (v, Branch (B, vi, vj, vk, vl), vf, vg, Empty))
  | balance (Branch (B, va, vb, vc, vd)) s t
    (Branch (v, Empty, vf, vg, Branch (B, vj, vk, vl, vm))) =
    Branch
      (B, Branch (B, va, vb, vc, vd), s, t,
        Branch (v, Empty, vf, vg, Branch (B, vj, vk, vl, vm)))
  | balance (Branch (B, va, vb, vc, vd)) s t
    (Branch (v, Branch (B, vi, vn, vo, vp), vf, vg, Branch (B, vj, vk, vl, vm)))
    = Branch
        (B, Branch (B, va, vb, vc, vd), s, t,
          Branch
            (v, Branch (B, vi, vn, vo, vp), vf, vg, Branch (B, vj, vk, vl, vm)))
  | balance (Branch (v, Empty, vb, vc, Empty)) s t Empty =
    Branch (B, Branch (v, Empty, vb, vc, Empty), s, t, Empty)
  | balance (Branch (v, Empty, vb, vc, Branch (B, ve, vf, vg, vh))) s t Empty =
    Branch
      (B, Branch (v, Empty, vb, vc, Branch (B, ve, vf, vg, vh)), s, t, Empty)
  | balance (Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Empty)) s t Empty =
    Branch
      (B, Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Empty), s, t, Empty)
  | balance
    (Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Branch (B, ve, vj, vk, vl)))
    s t Empty =
    Branch
      (B, Branch
            (v, Branch (B, vf, vg, vh, vi), vb, vc, Branch (B, ve, vj, vk, vl)),
        s, t, Empty)
  | balance (Branch (v, Empty, vf, vg, Empty)) s t (Branch (B, va, vb, vc, vd))
    = Branch
        (B, Branch (v, Empty, vf, vg, Empty), s, t, Branch (B, va, vb, vc, vd))
  | balance (Branch (v, Empty, vf, vg, Branch (B, vi, vj, vk, vl))) s t
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (B, Branch (v, Empty, vf, vg, Branch (B, vi, vj, vk, vl)), s, t,
        Branch (B, va, vb, vc, vd))
  | balance (Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Empty)) s t
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (B, Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Empty), s, t,
        Branch (B, va, vb, vc, vd))
  | balance
    (Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Branch (B, vi, vn, vo, vp)))
    s t (Branch (B, va, vb, vc, vd)) =
    Branch
      (B, Branch
            (v, Branch (B, vj, vk, vl, vm), vf, vg, Branch (B, vi, vn, vo, vp)),
        s, t, Branch (B, va, vb, vc, vd));

fun rbt_ins A_ f k v Empty = Branch (R, Empty, k, v, Empty)
  | rbt_ins A_ f k v (Branch (B, l, x, y, r)) =
    (if less A_ k x then balance (rbt_ins A_ f k v l) x y r
      else (if less A_ x k then balance l x y (rbt_ins A_ f k v r)
             else Branch (B, l, x, f k y v, r)))
  | rbt_ins A_ f k v (Branch (R, l, x, y, r)) =
    (if less A_ k x then Branch (R, rbt_ins A_ f k v l, x, y, r)
      else (if less A_ x k then Branch (R, l, x, y, rbt_ins A_ f k v r)
             else Branch (R, l, x, f k y v, r)));

fun paint c Empty = Empty
  | paint c (Branch (uu, l, k, v, r)) = Branch (c, l, k, v, r);

fun rbt_insert_with_key A_ f k v t = paint B (rbt_ins A_ f k v t);

fun rbt_insert A_ = rbt_insert_with_key A_ (fn _ => fn _ => fn nv => nv);

fun impl_of A_ (Rbt x) = x;

fun insert A_ x xa xb =
  Rbt (rbt_insert ((ord_preorder o preorder_order o order_linorder) A_) x xa
        (impl_of A_ xb));

fun rbt_lookup A_ Empty k = NONE
  | rbt_lookup A_ (Branch (uu, l, x, y, r)) k =
    (if less A_ k x then rbt_lookup A_ l k
      else (if less A_ x k then rbt_lookup A_ r k else SOME y));

fun lookup A_ x =
  rbt_lookup ((ord_preorder o preorder_order o order_linorder) A_)
    (impl_of A_ x);

val emptya : 'a dlist = Dlist [];

fun member A_ [] y = false
  | member A_ (x :: xs) y = eq A_ x y orelse member A_ xs y;

fun inserta A_ x xs = (if member A_ xs x then xs else x :: xs);

fun list_of_dlist (Dlist x) = x;

fun insertb A_ x dxs = Dlist (inserta A_ x (list_of_dlist dxs));

fun membera A_ dxs = member A_ (list_of_dlist dxs);

fun foldli [] c f sigma = sigma
  | foldli (x :: xs) c f sigma =
    (if c sigma then foldli xs c f (f x sigma) else sigma);

fun remdups A_ [] = []
  | remdups A_ (x :: xs) =
    (if member A_ xs x then remdups A_ xs else x :: remdups A_ xs);

fun equal_nata m n = (((integer_of_nat m) : IntInf.int) = (integer_of_nat n));

val equal_nat = {equal = equal_nata} : nat equal;

val preorder_nat = {ord_preorder = ord_nat} : nat preorder;

val order_nat = {preorder_order = preorder_nat} : nat order;

val zero_nat : nat = Nat 0;

fun list_all p [] = true
  | list_all p (x :: xs) = p x andalso list_all p xs;

val ord_integer =
  {less_eq = (fn a => fn b => IntInf.<= (a, b)),
    less = (fn a => fn b => IntInf.< (a, b))}
  : IntInf.int ord;

fun max A_ a b = (if less_eq A_ a b then b else a);

fun minus_nat m n =
  Nat (max ord_integer 0 (IntInf.- (integer_of_nat m, integer_of_nat n)));

fun times_nat m n = Nat (IntInf.* (integer_of_nat m, integer_of_nat n));

fun remap_rule f (Rule (q, l, qs)) = Rule (f q, l, map f qs);

fun dlist_iteratei xs = foldli (list_of_dlist xs);

fun iteratei_set_op_list_it_ls_ops s = dlist_iteratei s;

fun g_image_filter_ls_ops_ls_ops B_ f s =
  iteratei_set_op_list_it_ls_ops s (fn _ => true)
    (fn x => fn res => (case f x of NONE => res | SOME v => insertb B_ v res))
    emptya;

datatype ('a, 'b) assoc_list = Assoc_List of ('a * 'b) list;

type 'a hashable =
  {hashcode : 'a -> nat, bounded_hashcode : nat -> 'a -> nat,
    def_hashmap_size : 'a itself -> nat};
val hashcode = #hashcode : 'a hashable -> 'a -> nat;
val bounded_hashcode = #bounded_hashcode : 'a hashable -> nat -> 'a -> nat;
val def_hashmap_size = #def_hashmap_size : 'a hashable -> 'a itself -> nat;

datatype ('a, 'b) hashmap = RBT_HM of (nat, ('a, 'b) assoc_list) rbt;

fun impl_of_RBT_HM A_ (RBT_HM x) = x;

fun impl_ofa (Assoc_List x) = x;

fun lookupa A_ al = map_of A_ (impl_ofa al);

val linorder_nat = {order_linorder = order_nat} : nat linorder;

fun lookupb (A1_, A2_) k m =
  (case lookup linorder_nat m (hashcode A2_ k) of NONE => NONE
    | SOME lm => lookupa A1_ lm k);

fun hm_lookup (A1_, A2_) k hm = lookupb (A1_, A2_) k (impl_of_RBT_HM A2_ hm);

fun lookup_hm_ops_ls_ops (A1_, A2_) i m =
  (case hm_lookup (A1_, A2_) i m of NONE => emptya | SOME s => s);

datatype ('a, 'b, 'c) hashedTa_ext =
  HashedTa_ext of
    ('a, unit) hashmap * ('a, 'b) ta_rule dlist *
      ('b, ('a, 'b) ta_rule dlist) hashmap option *
      ('a, ('a, 'b) ta_rule dlist) hashmap option *
      (('a * 'b), ('a, 'b) ta_rule dlist) hashmap option * 'c;

fun hta_idx_f
  (HashedTa_ext (hta_Qi, hta_delta, hta_idx_f, hta_idx_s, hta_idx_sf, more)) =
  hta_idx_f;

fun hta_lookup_f (A1_, A2_) f h =
  lookup_hm_ops_ls_ops (A1_, A2_) f (the (hta_idx_f h));

fun list_all_zip p [] [] = true
  | list_all_zip p (a :: asa) (b :: bs) = p a b andalso list_all_zip p asa bs
  | list_all_zip p (v :: va) [] = false
  | list_all_zip p [] (v :: va) = false;

fun faccs A_ (B1_, B2_) h (Node (f, ts)) =
  let
    val qs = map (faccs A_ (B1_, B2_) h) ts;
  in
    g_image_filter_ls_ops_ls_ops A_
      (fn r =>
        let
          val Rule (_, _, qsa) = r;
        in
          (if list_all_zip (membera A_) qs qsa then SOME (lhs r) else NONE)
        end)
      (hta_lookup_f (B1_, B2_) f h)
  end;

fun equal_lista A_ (a :: lista) [] = false
  | equal_lista A_ [] (a :: lista) = false
  | equal_lista A_ (aa :: listaa) (a :: lista) =
    eq A_ aa a andalso equal_lista A_ listaa lista
  | equal_lista A_ [] [] = true;

fun equal_list A_ = {equal = equal_lista A_} : ('a list) equal;

fun gen_length n (x :: xs) = gen_length (suc n) xs
  | gen_length n [] = n;

fun size_list x = gen_length zero_nat x;

fun is_none (SOME x) = false
  | is_none NONE = true;

fun emptyc A_ = (fn _ => empty linorder_nat);

fun hm_empty_const A_ = RBT_HM (emptyc A_ ());

fun hm_empty A_ = (fn _ => hm_empty_const A_);

fun empty_hm_basic_ops A_ = hm_empty A_;

fun snd (a, b) = b;

fun fst (a, b) = a;

fun update_with_aux B_ v k f [] = [(k, f v)]
  | update_with_aux B_ v k f (p :: ps) =
    (if eq B_ (fst p) k then (k, f (snd p)) :: ps
      else p :: update_with_aux B_ v k f ps);

fun update_with A_ v k f al =
  Assoc_List (update_with_aux A_ v k f (impl_ofa al));

fun update A_ k v = update_with A_ v k (fn _ => v);

val emptyb : ('a, 'b) assoc_list = Assoc_List [];

fun updatea (A1_, A2_) k v m =
  let
    val hc = hashcode A2_ k;
  in
    (case lookup linorder_nat m hc
      of NONE => insert linorder_nat hc (update A1_ k v emptyb) m
      | SOME bm => insert linorder_nat hc (update A1_ k v bm) m)
  end;

fun hm_update (A1_, A2_) k v hm =
  RBT_HM (updatea (A1_, A2_) k v (impl_of_RBT_HM A2_ hm));

fun ins_hm_basic_ops (A1_, A2_) x s = hm_update (A1_, A2_) x () s;

fun g_image_filter_ls_ops_hs_ops (B1_, B2_) f s =
  iteratei_set_op_list_it_ls_ops s (fn _ => true)
    (fn x => fn res =>
      (case f x of NONE => res | SOME v => ins_hm_basic_ops (B1_, B2_) v res))
    (empty_hm_basic_ops B2_ ());

fun brc_iq (A1_, A2_) delta =
  g_image_filter_ls_ops_hs_ops (A1_, A2_)
    (fn r => (if null (rhsq r) then SOME (lhs r) else NONE)) delta;

fun hta_Qi
  (HashedTa_ext (hta_Qi, hta_delta, hta_idx_f, hta_idx_s, hta_idx_sf, more)) =
  hta_Qi;

fun sgn_integer k =
  (if ((k : IntInf.int) = 0) then 0
    else (if IntInf.< (k, 0) then IntInf.~ (1 : IntInf.int)
           else (1 : IntInf.int)));

fun abs_integer k = (if IntInf.< (k, 0) then IntInf.~ k else k);

fun apsnd f (x, y) = (x, f y);

fun divmod_integer k l =
  (if ((k : IntInf.int) = 0) then (0, 0)
    else (if ((l : IntInf.int) = 0) then (0, k)
           else (apsnd o (fn a => fn b => IntInf.* (a, b)) o sgn_integer) l
                  (if (((sgn_integer k) : IntInf.int) = (sgn_integer l))
                    then IntInf.divMod (IntInf.abs k, IntInf.abs l)
                    else let
                           val (r, s) =
                             IntInf.divMod (IntInf.abs k, IntInf.abs l);
                         in
                           (if ((s : IntInf.int) = 0) then (IntInf.~ r, 0)
                             else (IntInf.- (IntInf.~ r, (1 : IntInf.int)),
                                    IntInf.- (abs_integer l, s)))
                         end)));

fun mod_integer k l = snd (divmod_integer k l);

fun mod_nat m n = Nat (mod_integer (integer_of_nat m) (integer_of_nat n));

fun fifo_empty x = (fn _ => ([], [])) x;

fun equal_ta_rulea A_ B_ (Rule (qa, la, listaa)) (Rule (q, l, lista)) =
  eq A_ qa q andalso (eq B_ la l andalso equal_lista A_ listaa lista);

fun equal_ta_rule A_ B_ = {equal = equal_ta_rulea A_ B_} :
  ('a, 'b) ta_rule equal;

fun idx_build_stepfun_hm_ops_ls_ops A_ (B1_, B2_) f x m =
  let
    val i = f x;
  in
    (case hm_lookup (B1_, B2_) i m
      of NONE => hm_update (B1_, B2_) i (insertb A_ x emptya) m
      | SOME s => hm_update (B1_, B2_) i (insertb A_ x s) m)
  end;

fun idx_build_hm_ops_ls_ops_ls_ops A_ (B1_, B2_) f t =
  iteratei_set_op_list_it_ls_ops t (fn _ => true)
    (idx_build_stepfun_hm_ops_ls_ops A_ (B1_, B2_) f) (hm_empty B2_ ());

fun build_rule_index A_ B_ (C1_, C2_) =
  idx_build_hm_ops_ls_ops_ls_ops (equal_ta_rule A_ B_) (C1_, C2_);

fun build_rule_index_f A_ (B1_, B2_) delta =
  build_rule_index A_ B1_ (B1_, B2_) rhsl delta;

fun hta_idx_f_update hta_idx_fa
  (HashedTa_ext (hta_Qi, hta_delta, hta_idx_f, hta_idx_s, hta_idx_sf, more)) =
  HashedTa_ext
    (hta_Qi, hta_delta, hta_idx_fa hta_idx_f, hta_idx_s, hta_idx_sf, more);

fun hta_delta
  (HashedTa_ext (hta_Qi, hta_delta, hta_idx_f, hta_idx_s, hta_idx_sf, more)) =
  hta_delta;

fun hta_ensure_idx_f A_ (B1_, B2_) h =
  (case hta_idx_f h
    of NONE =>
      hta_idx_f_update
        (fn _ => SOME (build_rule_index_f A_ (B1_, B2_) (hta_delta h))) h
    | SOME _ => h);

fun iteratei_bset_op_list_it_ls_basic_ops s = dlist_iteratei s;

fun g_ball_ls_basic_ops s p =
  iteratei_bset_op_list_it_ls_basic_ops s (fn c => c) (fn x => fn _ => p x)
    true;

fun memb_hm_basic_ops (A1_, A2_) x s = not (is_none (hm_lookup (A1_, A2_) x s));

fun g_disjoint_ls_ops_hs_ops (A1_, A2_) s1 s2 =
  g_ball_ls_basic_ops s1 (fn x => not (memb_hm_basic_ops (A1_, A2_) x s2));

fun hta_mema (A1_, A2_) (B1_, B2_) t h =
  not (g_disjoint_ls_ops_hs_ops (B1_, B2_) (faccs B1_ (A1_, A2_) h t)
        (hta_Qi h));

fun hta_mem (A1_, A2_) (B1_, B2_) t h =
  hta_mema (A1_, A2_) (B1_, B2_) t (hta_ensure_idx_f B1_ (A1_, A2_) h);

fun pa_cond A_ B_ C_ s = let
                           val (_, (w, _)) = s;
                         in
                           not (null w)
                         end;

fun equal_proda A_ B_ (aa, ba) (a, b) = eq A_ aa a andalso eq B_ ba b;

fun equal_prod A_ B_ = {equal = equal_proda A_ B_} : ('a * 'b) equal;

fun def_hashmap_size_prod A_ B_ =
  (fn _ => plus_nat (def_hashmap_size A_ Type) (def_hashmap_size B_ Type));

fun nat_of_integer k = Nat (max ord_integer 0 k);

fun bounded_hashcode_prod A_ B_ n x =
  mod_nat
    (plus_nat
      (times_nat (bounded_hashcode A_ n (fst x))
        (nat_of_integer (33 : IntInf.int)))
      (bounded_hashcode B_ n (snd x)))
    n;

fun hashcode_prod A_ B_ x =
  plus_nat (times_nat (hashcode A_ (fst x)) (nat_of_integer (33 : IntInf.int)))
    (hashcode B_ (snd x));

fun hashable_prod A_ B_ =
  {hashcode = hashcode_prod A_ B_,
    bounded_hashcode = bounded_hashcode_prod A_ B_,
    def_hashmap_size = def_hashmap_size_prod A_ B_}
  : ('a * 'b) hashable;

fun hta_idx_sf
  (HashedTa_ext (hta_Qi, hta_delta, hta_idx_f, hta_idx_s, hta_idx_sf, more)) =
  hta_idx_sf;

fun hta_lookup_sf (A1_, A2_) (B1_, B2_) q f h =
  lookup_hm_ops_ls_ops (equal_prod A1_ B1_, hashable_prod A2_ B2_) (q, f)
    (the (hta_idx_sf h));

fun hta_idx_s
  (HashedTa_ext (hta_Qi, hta_delta, hta_idx_f, hta_idx_s, hta_idx_sf, more)) =
  hta_idx_s;

fun hta_lookup_s (A1_, A2_) q h =
  lookup_hm_ops_ls_ops (A1_, A2_) q (the (hta_idx_s h));

fun pa_upd_rule (A1_, A2_) (B1_, B2_) q w [] = (q, w)
  | pa_upd_rule (A1_, A2_) (B1_, B2_) q w (qp :: qs) =
    (if not (memb_hm_basic_ops (equal_prod A1_ B1_, hashable_prod A2_ B2_) qp q)
      then pa_upd_rule (A1_, A2_) (B1_, B2_)
             (ins_hm_basic_ops (equal_prod A1_ B1_, hashable_prod A2_ B2_) qp q)
             (qp :: w) qs
      else pa_upd_rule (A1_, A2_) (B1_, B2_) q w qs);

fun pa_step (A1_, A2_) (B1_, B2_) (C1_, C2_) h1 h2 s =
  let
    val (q, (w, delta_d)) = s;
    val (q1, q2) = hd w;
  in
    iteratei_set_op_list_it_ls_ops (hta_lookup_s (A1_, A2_) q1 h1)
      (fn _ => true)
      (fn r1 =>
        iteratei_set_op_list_it_ls_ops
          (hta_lookup_sf (C1_, C2_) (B1_, B2_) q2 (rhsl r1) h2) (fn _ => true)
          (fn r2 => fn res =>
            (if equal_nata (size_list (rhsq r1)) (size_list (rhsq r2))
              then let
                     val rp = r_prod r1 r2;
                     val (qa, (wa, delta_da)) = res;
                     val (qb, wb) =
                       pa_upd_rule (A1_, A2_) (C1_, C2_) qa wa (rhsq rp);
                   in
                     (qb, (wb, insertb (equal_ta_rule (equal_prod A1_ C1_) B1_)
                                 rp delta_da))
                   end
              else res)))
      (q, (tl w, delta_d))
  end;

fun equal_dlista A_ dxs dys =
  equal_lista A_ (list_of_dlist dxs) (list_of_dlist dys);

fun equal_dlist A_ = {equal = equal_dlista A_} : 'a dlist equal;

fun brc_cond A_ B_ = (fn (_, (w, _)) => not (null w));

fun def_hashmap_size_ta_rule A_ B_ = (fn _ => nat_of_integer (32 : IntInf.int));

fun hashcode_list A_ =
  foldl (fn h => fn x =>
          plus_nat (times_nat h (nat_of_integer (33 : IntInf.int)))
            (hashcode A_ x))
    (nat_of_integer (5381 : IntInf.int));

fun hashcode_of_ta_rule A_ B_ (Rule (q, f, qs)) =
  plus_nat (plus_nat (hashcode A_ q) (hashcode B_ f)) (hashcode_list A_ qs);

fun bounded_hashcode_ta_rule A_ B_ n r =
  mod_nat (hashcode_of_ta_rule A_ B_ r) n;

fun hashcode_ta_rule A_ B_ = hashcode_of_ta_rule A_ B_;

fun hashable_ta_rule A_ B_ =
  {hashcode = hashcode_ta_rule A_ B_,
    bounded_hashcode = bounded_hashcode_ta_rule A_ B_,
    def_hashmap_size = def_hashmap_size_ta_rule A_ B_}
  : ('a, 'b) ta_rule hashable;

fun brc_inner_step (A1_, A2_) (B1_, B2_) r =
  (fn (q, (w, rcm)) =>
    let
      val c =
        the (hm_lookup (equal_ta_rule A1_ B1_, hashable_ta_rule A2_ B2_) r rcm);
      val rcma =
        hm_update (equal_ta_rule A1_ B1_, hashable_ta_rule A2_ B2_) r
          (minus_nat c one_nat) rcm;
      val qa =
        (if less_eq_nat c one_nat then ins_hm_basic_ops (A1_, A2_) (lhs r) q
          else q);
      val wa =
        (if less_eq_nat c one_nat andalso
              not (memb_hm_basic_ops (A1_, A2_) (lhs r) q)
          then lhs r :: w else w);
    in
      (qa, (wa, rcma))
    end);

fun rqrm_lookup (A1_, A2_) rqrm q =
  (case hm_lookup (A1_, A2_) q rqrm of NONE => emptya | SOME s => s);

fun brc_step (A1_, A2_) (B1_, B2_) rqrm =
  (fn (q, (w, rcm)) =>
    iteratei_set_op_list_it_ls_ops (rqrm_lookup (A1_, A2_) rqrm (hd w))
      (fn _ => true) (brc_inner_step (A1_, A2_) (B1_, B2_)) (q, (tl w, rcm)));

fun brec_iqm (A1_, A2_) B_ delta =
  iteratei_set_op_list_it_ls_ops delta (fn _ => true)
    (fn r => fn m =>
      (if null (rhsq r) then hm_update (A1_, A2_) (lhs r) (Node (rhsl r, [])) m
        else m))
    (hm_empty A2_ ());

fun build_rule_index_sf (A1_, A2_) (B1_, B2_) delta =
  build_rule_index A1_ B1_ (equal_prod A1_ B1_, hashable_prod A2_ B2_)
    (fn r => (lhs r, rhsl r)) delta;

fun hta_idx_sf_update hta_idx_sfa
  (HashedTa_ext (hta_Qi, hta_delta, hta_idx_f, hta_idx_s, hta_idx_sf, more)) =
  HashedTa_ext
    (hta_Qi, hta_delta, hta_idx_f, hta_idx_s, hta_idx_sfa hta_idx_sf, more);

fun hta_ensure_idx_sf (A1_, A2_) (B1_, B2_) h =
  (case hta_idx_sf h
    of NONE =>
      hta_idx_sf_update
        (fn _ => SOME (build_rule_index_sf (A1_, A2_) (B1_, B2_) (hta_delta h)))
        h
    | SOME _ => h);

fun build_rule_index_s (A1_, A2_) B_ delta =
  build_rule_index A1_ B_ (A1_, A2_) lhs delta;

fun hta_idx_s_update hta_idx_sa
  (HashedTa_ext (hta_Qi, hta_delta, hta_idx_f, hta_idx_s, hta_idx_sf, more)) =
  HashedTa_ext
    (hta_Qi, hta_delta, hta_idx_f, hta_idx_sa hta_idx_s, hta_idx_sf, more);

fun hta_ensure_idx_s (A1_, A2_) B_ h =
  (case hta_idx_s h
    of NONE =>
      hta_idx_s_update
        (fn _ => SOME (build_rule_index_s (A1_, A2_) B_ (hta_delta h))) h
    | SOME _ => h);

fun iteratei al c f = foldli (impl_ofa al) c f;

fun iteratei_map_op_list_it_lm_ops s = iteratei s;

fun rm_iterateoi Empty c f sigma = sigma
  | rm_iterateoi (Branch (col, l, k, v, r)) c f sigma =
    (if c sigma
      then let
             val sigmaa = rm_iterateoi l c f sigma;
           in
             (if c sigmaa then rm_iterateoi r c f (f (k, v) sigmaa) else sigmaa)
           end
      else sigma);

fun iteratei_map_op_list_it_rm_ops A_ s = rm_iterateoi (impl_of A_ s);

fun iterateia A_ m c f sigma_0 =
  iteratei_map_op_list_it_rm_ops A_ m c
    (fn (_, lm) => iteratei_map_op_list_it_lm_ops lm c f) sigma_0;

fun hm_iteratei A_ hm = iterateia linorder_nat (impl_of_RBT_HM A_ hm);

fun iteratei_set_op_list_it_hs_ops A_ s =
  (fn c => fn f => hm_iteratei A_ s c (f o fst));

fun ins_dj_hm_basic_ops (A1_, A2_) x s = hm_update (A1_, A2_) x () s;

fun cart_hs_ops_hs_ops_hs_ops (A1_, A2_) (B1_, B2_) s1 s2 =
  iteratei_set_op_list_it_hs_ops A2_ s1 (fn _ => true)
    (fn x =>
      iteratei_set_op_list_it_hs_ops B2_ s2 (fn _ => true)
        (fn y =>
          ins_dj_hm_basic_ops (equal_prod A1_ B1_, hashable_prod A2_ B2_)
            (x, y)))
    (empty_hm_basic_ops (hashable_prod A2_ B2_) ());

fun whilea b c s = (if b s then whilea b c (c s) else s);

fun iteratei_bset_op_list_it_dflt_basic_ops_hm_basic_ops A_ s =
  (fn c => fn f => hm_iteratei A_ s c (f o fst));

fun g_to_list_dflt_basic_ops_hm_basic_ops A_ s =
  iteratei_bset_op_list_it_dflt_basic_ops_hm_basic_ops A_ s (fn _ => true)
    (fn a => fn b => a :: b) [];

fun pa_initial (A1_, A2_) B_ (C1_, C2_) h1 h2 =
  let
    val qip =
      cart_hs_ops_hs_ops_hs_ops (A1_, A2_) (C1_, C2_) (hta_Qi h1) (hta_Qi h2);
  in
    (qip, (g_to_list_dflt_basic_ops_hm_basic_ops (hashable_prod A2_ C2_) qip,
            emptya))
  end;

fun init_hta qi delta = HashedTa_ext (qi, delta, NONE, NONE, NONE, ());

fun hta_proda (A1_, A2_) (B1_, B2_) (C1_, C2_) h1 h2 =
  let
    val (_, (_, a)) =
      whilea (pa_cond A2_ C2_ B2_)
        (pa_step (A1_, A2_) (B1_, B2_) (C1_, C2_) h1 h2)
        (pa_initial (A1_, A2_) B2_ (C1_, C2_) h1 h2);
  in
    init_hta
      (cart_hs_ops_hs_ops_hs_ops (A1_, A2_) (C1_, C2_) (hta_Qi h1) (hta_Qi h2))
      a
  end;

fun hta_prod (A1_, A2_) (B1_, B2_) (C1_, C2_) h1 h2 =
  hta_proda (A1_, A2_) (B1_, B2_) (C1_, C2_)
    (hta_ensure_idx_s (A1_, A2_) B1_ h1)
    (hta_ensure_idx_sf (C1_, C2_) (B1_, B2_) h2);

fun def_hashmap_size_nat x = (fn _ => nat_of_integer (16 : IntInf.int)) x;

fun bounded_hashcode_nat na n = mod_nat n na;

fun hashcode_nat n = n;

val hashable_nat =
  {hashcode = hashcode_nat, bounded_hashcode = bounded_hashcode_nat,
    def_hashmap_size = def_hashmap_size_nat}
  : nat hashable;

fun htai_mem x = hta_mem (equal_nat, hashable_nat) (equal_nat, hashable_nat) x;

fun rqrm_add (A1_, A2_) B_ q r res =
  (case hm_lookup (A1_, A2_) q res
    of NONE => hm_update (A1_, A2_) q (insertb B_ r emptya) res
    | SOME s => hm_update (A1_, A2_) q (insertb B_ r s) res);

fun fifo_appendr a f = (a :: fst f, snd f);

fun fifo_isEmpty A_ f = equal_proda (equal_list A_) (equal_list A_) f ([], []);

fun fifo_removel f =
  (case snd f of [] => let
                         val rp = rev (fst f);
                       in
                         (hd rp, ([], tl rp))
                       end
    | a :: l => (a, (fst f, l)));

fun brec_cond A_ =
  (fn (_, (w, (_, qwit))) => not (fifo_isEmpty A_ w) andalso is_none qwit);

fun brec_construct_witness (A1_, A2_) B_ qm r =
  Node (rhsl r, map (fn q => the (hm_lookup (A1_, A2_) q qm)) (rhsq r));

fun brec_inner_step (A1_, A2_) (B1_, B2_) qi r =
  (fn (q, (w, (rcm, qwit))) =>
    let
      val c =
        the (hm_lookup (equal_ta_rule A1_ B1_, hashable_ta_rule A2_ B2_) r rcm);
      val cond =
        less_eq_nat c one_nat andalso is_none (hm_lookup (A1_, A2_) (lhs r) q);
      val rcma =
        hm_update (equal_ta_rule A1_ B1_, hashable_ta_rule A2_ B2_) r
          (minus_nat c one_nat) rcm;
      val qa =
        (if cond
          then hm_update (A1_, A2_) (lhs r)
                 (brec_construct_witness (A1_, A2_) B2_ q r) q
          else q);
      val wa = (if cond then fifo_appendr (lhs r) w else w);
      val qwita =
        (if less_eq_nat c one_nat andalso
              memb_hm_basic_ops (A1_, A2_) (lhs r) qi
          then SOME (lhs r) else qwit);
    in
      (qa, (wa, (rcma, qwita)))
    end);

fun brec_step (A1_, A2_) (B1_, B2_) rqrm qi =
  (fn (q, (w, (rcm, qwit))) =>
    let
      val (qa, wa) = fifo_removel w;
    in
      iteratei_set_op_list_it_ls_ops (rqrm_lookup (A1_, A2_) rqrm qa)
        (fn _ => true) (brec_inner_step (A1_, A2_) (B1_, B2_) qi)
        (q, (wa, (rcm, qwit)))
    end);

fun hta_empty A_ B_ u = init_hta (empty_hm_basic_ops A_ ()) emptya;

fun g_image_ls_ops_ls_ops B_ f s =
  iteratei_set_op_list_it_ls_ops s (fn _ => true) (fn x => insertb B_ (f x))
    emptya;

fun g_image_hs_ops_hs_ops A_ (B1_, B2_) f s =
  iteratei_set_op_list_it_hs_ops A_ s (fn _ => true)
    (fn x => ins_hm_basic_ops (B1_, B2_) (f x)) (empty_hm_basic_ops B2_ ());

fun hta_remap A_ (B1_, B2_) (C1_, C2_) f h =
  init_hta (g_image_hs_ops_hs_ops A_ (B1_, B2_) f (hta_Qi h))
    (g_image_ls_ops_ls_ops (equal_ta_rule B1_ C1_) (remap_rule f)
      (hta_delta h));

fun g_union_dflt_basic_ops_hm_basic_ops (A1_, A2_) s1 s2 =
  iteratei_bset_op_list_it_dflt_basic_ops_hm_basic_ops A2_ s1 (fn _ => true)
    (ins_hm_basic_ops (A1_, A2_)) s2;

fun g_union_dj_ls_basic_ops A_ s1 s2 =
  iteratei_bset_op_list_it_ls_basic_ops s1 (fn _ => true) (insertb A_) s2;

fun def_hashmap_size_ustate_wrapper A_ B_ =
  (fn _ => plus_nat (def_hashmap_size A_ Type) (def_hashmap_size B_ Type));

fun bounded_hashcode_ustate_wrapper A_ B_ n x =
  (case x of USW1 a => bounded_hashcode A_ n a
    | USW2 b => minus_nat (minus_nat n one_nat) (bounded_hashcode B_ n b));

fun hashcode_ustate_wrapper A_ B_ x =
  (case x
    of USW1 a => times_nat (nat_of_integer (2 : IntInf.int)) (hashcode A_ a)
    | USW2 b =>
      plus_nat (times_nat (nat_of_integer (2 : IntInf.int)) (hashcode B_ b))
        one_nat);

fun hashable_ustate_wrapper A_ B_ =
  {hashcode = hashcode_ustate_wrapper A_ B_,
    bounded_hashcode = bounded_hashcode_ustate_wrapper A_ B_,
    def_hashmap_size = def_hashmap_size_ustate_wrapper A_ B_}
  : ('a, 'b) ustate_wrapper hashable;

fun equal_ustate_wrappera A_ B_ (USW2 q2) (USW1 q1) = false
  | equal_ustate_wrappera A_ B_ (USW1 q1) (USW2 q2) = false
  | equal_ustate_wrappera A_ B_ (USW2 q2a) (USW2 q2) = eq B_ q2a q2
  | equal_ustate_wrappera A_ B_ (USW1 q1a) (USW1 q1) = eq A_ q1a q1;

fun equal_ustate_wrapper A_ B_ = {equal = equal_ustate_wrappera A_ B_} :
  ('a, 'b) ustate_wrapper equal;

fun hta_union (A1_, A2_) (B1_, B2_) (C1_, C2_) h1 h2 =
  init_hta
    (g_union_dflt_basic_ops_hm_basic_ops
      (equal_ustate_wrapper A1_ C1_, hashable_ustate_wrapper A2_ C2_)
      (g_image_hs_ops_hs_ops A2_
        (equal_ustate_wrapper A1_ C1_, hashable_ustate_wrapper A2_ C2_) USW1
        (hta_Qi h1))
      (g_image_hs_ops_hs_ops C2_
        (equal_ustate_wrapper A1_ C1_, hashable_ustate_wrapper A2_ C2_) USW2
        (hta_Qi h2)))
    (g_union_dj_ls_basic_ops (equal_ta_rule (equal_ustate_wrapper A1_ C1_) B1_)
      (g_image_ls_ops_ls_ops (equal_ta_rule (equal_ustate_wrapper A1_ C1_) B1_)
        (remap_rule USW1) (hta_delta h1))
      (g_image_ls_ops_ls_ops (equal_ta_rule (equal_ustate_wrapper A1_ C1_) B1_)
        (remap_rule USW2) (hta_delta h2)));

fun map_to_nat_hs_ops_hm_ops (A1_, A2_) s =
  snd (iteratei_set_op_list_it_hs_ops A2_ s (fn _ => true)
        (fn x => fn (c, m) => (plus_nat c one_nat, hm_update (A1_, A2_) x c m))
        (zero_nat, hm_empty A2_ ()));

fun g_union_ls_ops_hs_ops (A1_, A2_) s1 s2 =
  iteratei_set_op_list_it_ls_ops s1 (fn _ => true) (ins_hm_basic_ops (A1_, A2_))
    s2;

fun g_Union_image_ls_ops_ls_ops_hs_ops (B1_, B2_) f s =
  iteratei_set_op_list_it_ls_ops s (fn _ => true)
    (fn x => g_union_ls_ops_hs_ops (B1_, B2_) (f x))
    (empty_hm_basic_ops B2_ ());

fun g_from_list_aux_ls_basic_ops A_ accs (x :: l) =
  g_from_list_aux_ls_basic_ops A_ (insertb A_ x accs) l
  | g_from_list_aux_ls_basic_ops A_ y [] = y;

fun g_from_list_ls_basic_ops A_ l = g_from_list_aux_ls_basic_ops A_ emptya l;

fun rule_states_l A_ (Rule (q, f, qs)) =
  insertb A_ q (g_from_list_ls_basic_ops A_ qs);

fun hta_delta_states (A1_, A2_) h =
  g_Union_image_ls_ops_ls_ops_hs_ops (A1_, A2_) id
    (g_image_filter_ls_ops_ls_ops (equal_dlist A1_)
      (fn r => SOME (rule_states_l A1_ r)) (hta_delta h));

fun hta_states (A1_, A2_) h =
  g_union_dflt_basic_ops_hm_basic_ops (A1_, A2_) (hta_Qi h)
    (hta_delta_states (A1_, A2_) h);

fun hta_reindex (A1_, A2_) (B1_, B2_) h =
  let
    val mp = map_to_nat_hs_ops_hm_ops (A1_, A2_) (hta_states (A1_, A2_) h);
  in
    hta_remap A2_ (equal_nat, hashable_nat) (B1_, B2_)
      (fn q => the (hm_lookup (A1_, A2_) q mp)) h
  end;

fun htai_prod h1 h2 =
  hta_reindex
    (equal_prod equal_nat equal_nat, hashable_prod hashable_nat hashable_nat)
    (equal_nat, hashable_nat)
    (hta_prod (equal_nat, hashable_nat) (equal_nat, hashable_nat)
      (equal_nat, hashable_nat) h1 h2);

fun build_rqrm (A1_, A2_) (B1_, B2_) delta =
  iteratei_set_op_list_it_ls_ops delta (fn _ => true)
    (fn r => fn res =>
      foldl (fn resa => fn q =>
              rqrm_add (A1_, A2_) (equal_ta_rule A1_ B1_) q r resa)
        res (rhsq r))
    (hm_empty A2_ ());

fun hta_add_qi (A1_, A2_) B_ qi h =
  init_hta (ins_hm_basic_ops (A1_, A2_) qi (hta_Qi h)) (hta_delta h);

fun inj_image_filter_cartesian_product_ls_ops_ls_ops_ls_ops C_ f s1 s2 =
  iteratei_set_op_list_it_ls_ops s1 (fn _ => true)
    (fn x =>
      iteratei_set_op_list_it_ls_ops s2 (fn _ => true)
        (fn y => fn res =>
          (case f (x, y) of NONE => res | SOME z => insertb C_ z res)))
    emptya;

fun inj_image_filter_cp_ls_ops_ls_ops_ls_ops C_ f p s1 s2 =
  inj_image_filter_cartesian_product_ls_ops_ls_ops_ls_ops C_
    (fn xy => (if p xy then SOME (f xy) else NONE)) s1 s2;

fun delta_prod_h (A1_, A2_) (B1_, B2_) (C1_, C2_) delta_1 delta_2 =
  inj_image_filter_cp_ls_ops_ls_ops_ls_ops
    (equal_ta_rule (equal_prod A1_ C1_) B1_) (fn (a, b) => r_prod a b)
    (fn (r1, r2) =>
      eq B1_ (rhsl r1) (rhsl r2) andalso
        equal_nata (size_list (rhsq r1)) (size_list (rhsq r2)))
    delta_1 delta_2;

fun hta_prodWR (A1_, A2_) (B1_, B2_) (D1_, D2_) h1 h2 =
  init_hta
    (cart_hs_ops_hs_ops_hs_ops (A1_, A2_) (D1_, D2_) (hta_Qi h1) (hta_Qi h2))
    (delta_prod_h (A1_, A2_) (B1_, B2_) (D1_, D2_) (hta_delta h1)
      (hta_delta h2));

fun g_inter_dflt_basic_ops_hm_basic_ops (A1_, A2_) s1 s2 =
  iteratei_bset_op_list_it_dflt_basic_ops_hm_basic_ops A2_ s1 (fn _ => true)
    (fn x => fn s =>
      (if memb_hm_basic_ops (A1_, A2_) x s2
        then ins_dj_hm_basic_ops (A1_, A2_) x s else s))
    (empty_hm_basic_ops A2_ ());

fun hta_reduce (A1_, A2_) B_ h q =
  init_hta (g_inter_dflt_basic_ops_hm_basic_ops (A1_, A2_) q (hta_Qi h))
    (g_image_filter_ls_ops_ls_ops (equal_ta_rule A1_ B_)
      (fn r =>
        (if memb_hm_basic_ops (A1_, A2_) (lhs r) q andalso
              list_all (fn qa => memb_hm_basic_ops (A1_, A2_) qa q) (rhsq r)
          then SOME r else NONE))
      (hta_delta h));

fun htai_empty x = hta_empty hashable_nat hashable_nat x;

fun htai_union h1 h2 =
  hta_reindex
    (equal_ustate_wrapper equal_nat equal_nat,
      hashable_ustate_wrapper hashable_nat hashable_nat)
    (equal_nat, hashable_nat)
    (hta_union (equal_nat, hashable_nat) (equal_nat, hashable_nat)
      (equal_nat, hashable_nat) h1 h2);

fun brc_rcm_init (A1_, A2_) (B1_, B2_) delta =
  iteratei_set_op_list_it_ls_ops delta (fn _ => true)
    (fn r =>
      hm_update (equal_ta_rule A1_ B1_, hashable_ta_rule A2_ B2_) r
        (size_list (remdups A1_ (rhsq r))))
    (hm_empty (hashable_ta_rule A2_ B2_) ());

fun brc_initial (A1_, A2_) (B1_, B2_) delta =
  let
    val iq = brc_iq (A1_, A2_) delta;
  in
    (iq, (g_to_list_dflt_basic_ops_hm_basic_ops A2_ iq,
           brc_rcm_init (A1_, A2_) (B1_, B2_) delta))
  end;

fun htai_add_qi x = hta_add_qi (equal_nat, hashable_nat) hashable_nat x;

fun htai_prodWR h1 h2 =
  hta_reindex
    (equal_prod equal_nat equal_nat, hashable_prod hashable_nat hashable_nat)
    (equal_nat, hashable_nat)
    (hta_prodWR (equal_nat, hashable_nat) (equal_nat, hashable_nat)
      (equal_nat, hashable_nat) h1 h2);

fun g_sel_dflt_basic_ops_hm_basic_ops A_ s p =
  iteratei_bset_op_list_it_dflt_basic_ops_hm_basic_ops A_ s is_none
    (fn x => fn _ => (if p x then SOME x else NONE)) NONE;

fun g_disjoint_witness_hs_ops_hs_ops (A1_, A2_) s1 s2 =
  g_sel_dflt_basic_ops_hm_basic_ops A2_ s1
    (fn x => memb_hm_basic_ops (A1_, A2_) x s2);

fun g_set_to_listr_hs_ops_fifo_ops A_ s =
  iteratei_set_op_list_it_hs_ops A_ s (fn _ => true) fifo_appendr
    (fifo_empty ());

fun brec_initial (A1_, A2_) (B1_, B2_) qi delta =
  let
    val iq = brc_iq (A1_, A2_) delta;
  in
    (brec_iqm (A1_, A2_) B2_ delta,
      (g_set_to_listr_hs_ops_fifo_ops A2_ iq,
        (brc_rcm_init (A1_, A2_) (B1_, B2_) delta,
          g_disjoint_witness_hs_ops_hs_ops (A1_, A2_) iq qi)))
  end;

fun hta_add_rule (A1_, A2_) (B1_, B2_) r h =
  init_hta (hta_Qi h) (insertb (equal_ta_rule A1_ B1_) r (hta_delta h));

fun htai_add_rule x =
  hta_add_rule (equal_nat, hashable_nat) (equal_nat, hashable_nat) x;

fun hta_bwd_reduce (A1_, A2_) (B1_, B2_) h =
  let
    val rqrm = build_rqrm (A1_, A2_) (B1_, B2_) (hta_delta h);
  in
    hta_reduce (A1_, A2_) B1_ h
      (fst (whilea (brc_cond A2_ B2_) (brc_step (A1_, A2_) (B1_, B2_) rqrm)
             (brc_initial (A1_, A2_) (B1_, B2_) (hta_delta h))))
  end;

fun htai_bwd_reduce x =
  hta_bwd_reduce (equal_nat, hashable_nat) (equal_nat, hashable_nat) x;

fun g_size_dflt_basic_ops_hm_basic_ops A_ s =
  iteratei_bset_op_list_it_dflt_basic_ops_hm_basic_ops A_ s (fn _ => true)
    (fn _ => suc) zero_nat;

fun hs_size A_ = g_size_dflt_basic_ops_hm_basic_ops A_;

fun g_size_ls_basic_ops s =
  iteratei_bset_op_list_it_ls_basic_ops s (fn _ => true) (fn _ => suc) zero_nat;

fun ls_size x = g_size_ls_basic_ops x;

fun iteratei_bset_op_list_it_dflt_basic_oops_rm_basic_ops A_ s =
  (fn c => fn f => rm_iterateoi (impl_of A_ s) c (f o fst));

fun g_size_dflt_basic_oops_rm_basic_ops A_ s =
  iteratei_bset_op_list_it_dflt_basic_oops_rm_basic_ops A_ s (fn _ => true)
    (fn _ => suc) zero_nat;

fun rs_size A_ = g_size_dflt_basic_oops_rm_basic_ops A_;

fun htai_ensure_idx_f x =
  hta_ensure_idx_f equal_nat (equal_nat, hashable_nat) x;

fun htai_ensure_idx_s x =
  hta_ensure_idx_s (equal_nat, hashable_nat) equal_nat x;

fun htai_ensure_idx_sf x =
  hta_ensure_idx_sf (equal_nat, hashable_nat) (equal_nat, hashable_nat) x;

fun hta_is_empty_witness (A1_, A2_) (B1_, B2_) h =
  let
    val rqrm = build_rqrm (A1_, A2_) (B1_, B2_) (hta_delta h);
  in
    (case whilea (brec_cond A1_)
            (brec_step (A1_, A2_) (B1_, B2_) rqrm (hta_Qi h))
            (brec_initial (A1_, A2_) (B1_, B2_) (hta_Qi h) (hta_delta h))
      of (q, (_, (_, NONE))) => NONE
      | (q, (_, (_, SOME qa))) => hm_lookup (A1_, A2_) qa q)
  end;

fun htai_is_empty_witness x =
  hta_is_empty_witness (equal_nat, hashable_nat) (equal_nat, hashable_nat) x;

end; (*struct Ta*)
