theory Digraph_New
imports Digraph_Basic
begin

section {* Directed Graphs *}

record 'v digraph = 
  dg_V :: "'v set"
  dg_E :: "'v digraph"

locale digraph = 
  fixes G :: "('v, 'more) digraph_scheme"
  assumes wf_E: "dg_E G \<subseteq> dg_V G \<times> dg_V G"
begin
  abbreviation "V \<equiv> dg_V G"
  abbreviation "E \<equiv> dg_E G"

  lemma path_in_V: "path E u p v \<Longrightarrow> set p \<subseteq> V"
    using wf_E by (fastforce dest: path_nodes_edges)

  lemma wf_pathD: "\<lbrakk>path E u p v; p\<noteq>[]\<rbrakk> \<Longrightarrow> {u,v} \<union> set p \<subseteq> V"
    using wf_E
    apply safe
    apply (auto elim: path.cases) []

    apply (cases p rule: rev_cases, simp)
    apply (auto simp: path_simps) []
    
    apply (auto dest: path_in_V) []
    done
  
  lemma wf_ipath: "ipath E r \<Longrightarrow> range r \<subseteq> V"
    using wf_E unfolding ipath_def
    by auto

end

definition "digraph_inv ext G \<equiv> digraph G \<and> ext G"

lemma digraph_inv_iff: "digraph_inv ext G \<longleftrightarrow> digraph G \<and> ext G"
  by (simp add: digraph_inv_def)

subsection "Finite Digraph"
locale finite_digraph = digraph G for G :: "('v,'more) digraph_scheme" +
  assumes finite_V[simp, intro!]: "finite V"
begin
  lemma super_dg: "digraph G" by unfold_locales

  lemma finite_E[simp, intro!]: "finite E"
    using wf_E by (auto intro: finite_subset)

  lemma finite_ipath: "ipath E r \<Longrightarrow> finite (range r)"
    using finite_subset[OF wf_ipath finite_V] .

end

definition "finite_digraph_inv_ext ext G \<equiv> finite_digraph G \<and> ext G"
abbreviation "finite_digraph_inv ext \<equiv> digraph_inv (finite_digraph_inv_ext ext)"

lemma finite_digraph_inv_iff: 
  "finite_digraph_inv ext G \<longleftrightarrow> finite_digraph G \<and> ext G"
  by (auto simp add: 
    digraph_inv_iff finite_digraph_inv_ext_def finite_digraph.super_dg)

subsection {*Digraph with @{text "UNIV"} Set of Nodes *}

locale nn_digraph = digraph G for G :: "('v,'more) digraph_scheme" +
  assumes [simp]: "V=UNIV"
begin
  lemma super_dg: "digraph G" by unfold_locales
  lemma iV_intro[intro!]: "x\<in>V" by simp
end

definition "nn_digraph_inv_ext ext G \<equiv> nn_digraph G \<and> ext G"
abbreviation "nn_digraph_inv ext \<equiv> digraph_inv (nn_digraph_inv_ext ext)"

lemma nn_digraph_inv_iff: 
  "nn_digraph_inv ext G \<longleftrightarrow> nn_digraph G \<and> ext G"
  by (auto simp add: 
    digraph_inv_iff nn_digraph_inv_ext_def nn_digraph.super_dg)

section {* Rooted Directed Graphs *}

record 'v rooted_digraph = "'v digraph" +
  rdg_V0 :: "'v set"

locale rooted_digraph = digraph G 
  for G :: "('v,'more) rooted_digraph_scheme" +
  assumes wf_V0: "rdg_V0 G \<subseteq> V"
begin
  abbreviation "V0 \<equiv> rdg_V0 G"

  lemma super_dg: "digraph G" by unfold_locales

  definition is_run
    -- "Infinite run, i.e., a rooted infinite path"
    where "is_run r \<equiv> r 0 \<in> V0 \<and> ipath E r"
  
  lemma run_reachable: "is_run r \<Longrightarrow> range r \<subseteq> E\<^sup>*``V0"
    unfolding is_run_def
    using ipath_to_rtrancl[of E r 0]
    by auto

  lemma wf_run: "is_run r \<Longrightarrow> range r \<subseteq> V"
    unfolding is_run_def
    using wf_ipath by auto
  
  lemma 
    assumes "is_run r"
    shows run_ipath: "ipath E r"
    and run_V0: "r 0 \<in> V0"
    using assms unfolding is_run_def by auto

  lemma wf_reachable: "E\<^sup>*``V0 \<subseteq> V"
    apply (rule rtrancl_reachable_induct)
    using wf_V0 wf_E
    by auto

end

definition "rooted_digraph_inv_ext ext G \<equiv> rooted_digraph G \<and> ext G"
abbreviation "rooted_digraph_inv ext \<equiv> digraph_inv (rooted_digraph_inv_ext ext)"

lemma rooted_digraph_inv_iff: 
  "rooted_digraph_inv ext G \<longleftrightarrow> rooted_digraph G \<and> ext G"
  by (auto simp add: 
    digraph_inv_iff rooted_digraph_inv_ext_def rooted_digraph.super_dg)

abbreviation "rdg_is_run \<equiv> rooted_digraph.is_run"

locale finite_reachable_digraph = rooted_digraph G 
  for G :: "('v,'more) rooted_digraph_scheme" +
  assumes finite_reachable[simp, intro!]: "finite (E\<^sup>*``V0)"
begin
  lemma super_rdg: "rooted_digraph G" by unfold_locales
end

(*sublocale finite_rooted_digraph < finite_reachable_digraph
  apply unfold_locales
  by (metis finite_E finite_V0 finite_rtrancl_Image)
*)

definition "finite_reachable_digraph_inv_ext ext G 
  \<equiv> finite_reachable_digraph G \<and> ext G"
abbreviation "finite_reachable_digraph_inv ext 
  \<equiv> rooted_digraph_inv (finite_reachable_digraph_inv_ext ext)"

lemma finite_reachable_digraph_inv_iff: 
  "finite_reachable_digraph_inv ext G \<longleftrightarrow> finite_reachable_digraph G \<and> ext G"
  by (auto simp add: 
    rooted_digraph_inv_iff 
    finite_reachable_digraph_inv_ext_def finite_reachable_digraph.super_rdg)

context finite_reachable_digraph
begin
  lemma is_run_finite: "is_run r \<Longrightarrow> finite (range r)"
    apply (rule finite_subset[OF _ finite_reachable])
    using run_reachable
    by auto

end


locale finite_rooted_digraph = finite_reachable_digraph G
  + finite_digraph G
  for G :: "('v,'more) rooted_digraph_scheme" +
  assumes finite_V0[simp, intro!]: "finite V0"
begin
  lemma super_frdg: "finite_reachable_digraph G" by unfold_locales
end

definition "finite_rooted_digraph_inv_ext ext G 
  \<equiv> finite_rooted_digraph G \<and> ext G"
abbreviation "finite_rooted_digraph_inv ext 
  \<equiv> finite_reachable_digraph_inv (finite_rooted_digraph_inv_ext ext)"

lemma finite_rooted_digraph_inv_iff: 
  "finite_rooted_digraph_inv ext G \<longleftrightarrow> finite_rooted_digraph G \<and> ext G"
  by (auto simp add: 
    finite_reachable_digraph_inv_iff 
    finite_rooted_digraph_inv_ext_def finite_rooted_digraph.super_frdg)

locale nn_rooted_digraph = nn_digraph G + rooted_digraph G
  for G :: "('v,'more) rooted_digraph_scheme"

locale nn_finite_rooted_digraph = nn_digraph G + finite_rooted_digraph G
  for G :: "('v,'more) rooted_digraph_scheme"


section "Renaming of Nodes"

definition "dg_rename ext f G \<equiv> \<lparr>
  dg_V = f`dg_V G,
  dg_E = {(f u,f v) | u v. (u,v)\<in>dg_E G},
  \<dots> = ext
\<rparr>"

lemmas dg_rename_defs = dg_rename_def

lemma dg_rename_fields:
  "dg_V (dg_rename ext f G) = f`dg_V G"
  "dg_E (dg_rename ext f G) = {(f u,f v) | u v. (u,v)\<in>dg_E G}"
  "digraph.more (dg_rename ext f G) = ext"
  unfolding dg_rename_defs by simp_all

locale dg_rename_precond = digraph G 
  for G :: "('u,'more) digraph_scheme" +
  fixes f :: "'u \<Rightarrow> 'v"
  fixes ext :: "'moret"
  assumes f_inj: "inj_on f V"
begin

  definition "fi \<equiv> the_inv_into V f"

  abbreviation "G' \<equiv> dg_rename ext f G"
  abbreviation "V' \<equiv> dg_V G'"
  abbreviation "E' \<equiv> dg_E G'"

  lemma E'_fmt: "E' = {(f u,f v) | u v. (u,v)\<in>E}"
    unfolding dg_rename_defs by auto
  lemma V'_fmt: "V' = f`V"
    unfolding dg_rename_defs by auto

  lemma G'_digraph: "digraph G'"
    apply unfold_locales
    unfolding dg_rename_defs
    using wf_E
    by auto
    
  sublocale G'!: digraph G' using G'_digraph .

  lemma 
    fixes x :: 'u and y :: 'v
    shows
    fi_f: "x\<in>V \<Longrightarrow> fi (f x) = x" and
    f_fi: "y\<in>f`V \<Longrightarrow> f (fi y) = y" and
    fi_f_eq: "\<lbrakk>f x = y; x\<in>V\<rbrakk> \<Longrightarrow> fi y = x"
    unfolding fi_def
    by (auto simp: the_inv_into_f_f f_the_inv_into_f the_inv_into_f_eq f_inj)

  lemma E'_to_E: "(u,v) \<in> E' \<Longrightarrow> (fi u, fi v)\<in>E"
    using wf_E
    by (auto simp: fi_f E'_fmt)
  lemma V'_to_V: "u\<in>V' \<Longrightarrow> fi u \<in> V"
    by (auto simp: fi_f V'_fmt)

  lemma E_fmt: "E = {(fi u,fi v) | u v. (u,v)\<in>E'}"
    apply rule
    apply (clarsimp simp: E'_fmt)
    apply (metis (lifting) fi_f_eq mem_Sigma_iff set_rev_mp wf_E)
    apply (auto intro: E'_to_E)
    done

  lemma V_fmt: "V = fi`V'"
    apply rule
    apply (clarsimp simp: V'_fmt)
    apply (metis fi_f_eq image_eqI)
    apply (auto intro: V'_to_V)
    done

  lemma rtrancl_E'_to_E: assumes "(u,v)\<in>E'\<^sup>*" shows "(fi u, fi v)\<in>E\<^sup>*"
    using assms apply induction
    by (fastforce intro: E'_to_E rtrancl_into_rtrancl)+

  lemma ipath_sim1: "ipath E r \<Longrightarrow> ipath E' (f o r)"
    unfolding ipath_def E'_fmt by auto

  lemma ipath_sim2: "ipath E' r \<Longrightarrow> ipath E (fi o r)"
    unfolding ipath_def E'_fmt
    apply clarsimp
    apply (drule_tac x=i in spec)
    using wf_E
    by (auto simp: fi_f)

  lemma path_sim1: assumes "path E u p v" shows "path E' (f u) (map f p) (f v)"
    using assms
    by induction (auto simp: path_simps E'_fmt)

  lemma path_sim2: 
    assumes "path E' u p v" shows "path E (fi u) (map fi p) (fi v)"
    using assms
    by induction (auto simp: path_simps E_fmt)
end

lemma dg_rename_inv: 
  assumes A: "digraph_inv ext G"
  assumes B: "inj_on f (dg_V G)"
  assumes C: "ext G \<Longrightarrow> ext' (dg_rename m f G)"
  shows "digraph_inv ext' (dg_rename m f G)"
proof -
  from A interpret digraph G by (simp add: digraph_inv_iff)
  from B interpret dg_rename_precond G f m by unfold_locales

  from A C show ?thesis using G'_digraph 
    by (simp add: digraph_inv_iff)
qed


definition "rdg_rename_ext ext f G \<equiv> \<lparr> rdg_V0 = f`rdg_V0 G, \<dots> = ext \<rparr>"
abbreviation "rdg_rename ext f G \<equiv> dg_rename (rdg_rename_ext ext f G) f G"

lemmas rdg_rename_defs = dg_rename_defs rdg_rename_ext_def

lemma rdg_rename_fields_ext:
  "rdg_V0 (rdg_rename ext f G) = f`rdg_V0 G"
  "rooted_digraph.more (rdg_rename ext f G) = ext"
  unfolding rdg_rename_defs
  apply simp_all
  done

lemmas rdg_rename_fields = dg_rename_fields rdg_rename_fields_ext

locale rdg_rename_precond = dg_rename_precond G f "rdg_rename_ext ext f G" 
  + rooted_digraph G
  for G :: "('u,'more) rooted_digraph_scheme" 
  and f :: "'u \<Rightarrow> 'v" 
  and ext :: "'moret"
begin
  abbreviation "V0' \<equiv> rdg_V0 G'"

  lemma V0'_to_V0: "v\<in>f`V0 \<Longrightarrow> fi v \<in> V0"
    using wf_V0
    by (auto simp: fi_f)

  lemma V0'_fmt: "V0' = f`V0"
    by (auto simp: rdg_rename_fields)

  lemma V0_fmt: "V0 = fi`V0'"
    unfolding V0'_fmt
    using wf_V0 fi_f[symmetric]
    by auto []

  lemma G'_rooted_digraph: "rooted_digraph G'"
    apply unfold_locales
    using wf_V0
    apply (auto simp: rdg_rename_fields)
    done

  sublocale G'!: rooted_digraph G' using G'_rooted_digraph .

  lemma run_sim1: "is_run r \<Longrightarrow> G'.is_run (f o r)"
    unfolding is_run_def G'.is_run_def
    by (auto simp: ipath_sim1 V0'_fmt)

  lemma run_sim2: "G'.is_run r \<Longrightarrow> is_run (fi o r)"
    unfolding G'.is_run_def is_run_def
    by (auto simp: ipath_sim2 V0_fmt)

end

lemma  
  assumes A: "rooted_digraph_inv ext G"
  assumes B: "inj_on f (dg_V G)"
  shows rdg_rename_is_run_sim1: 
    "\<And>r. rdg_is_run G r \<Longrightarrow> rdg_is_run (rdg_rename m f G) (f o r)" 
    (is "PROP ?G1")
  and rdg_rename_is_run_sim2: 
    "\<And>r. rdg_is_run (rdg_rename m f G) r 
      \<Longrightarrow> rdg_is_run G (the_inv_into (dg_V G) f o r)" (is "PROP ?G2")
  and rdg_rename_inv: "\<lbrakk>ext G \<Longrightarrow> ext' (rdg_rename m f G)\<rbrakk> 
    \<Longrightarrow> rooted_digraph_inv ext' (rdg_rename m f G)" (is "PROP ?G3")
proof -
  from A interpret rooted_digraph G by (simp add: rooted_digraph_inv_iff)
  from B interpret rdg_rename_precond G f m by unfold_locales

  show "PROP ?G1" using run_sim1 by simp
  show "PROP ?G2" using run_sim2 by (simp add: fi_def)
  show "PROP ?G3" 
    apply (rule dg_rename_inv[OF A B])
    unfolding rooted_digraph_inv_ext_def
    by (simp add: G'_rooted_digraph)
qed

locale frdg_rename_precond = rdg_rename_precond G f ext
  + finite_reachable_digraph G
  for G :: "('u,'more) rooted_digraph_scheme" 
  and f :: "'u \<Rightarrow> 'v" 
  and ext :: "'moret"
begin
  lemma G'_finite_reachable_digraph: "finite_reachable_digraph G'"
  proof
    have "E'\<^sup>*``V0' \<subseteq> f ` (E\<^sup>*``V0)"
    proof
      fix v'
      assume A: "v'\<in>E'\<^sup>*``V0'"
      then obtain u' where "(u',v')\<in>E'\<^sup>*" and "u'\<in>V0'" by auto
      then obtain u where [simp]: "u'=f u" and "u\<in>V0" unfolding V0'_fmt by auto
      hence [simp]: "fi (f u) = u" using wf_V0 by (auto simp: fi_f)
      have [simp]: "f (fi v') = v'"
        apply (subst f_fi)
        using set_mp[OF G'.wf_reachable A]
        by (auto simp: V'_fmt)

      from rtrancl_E'_to_E[OF `(u',v')\<in>E'\<^sup>*`] have "(u,fi v') \<in> E\<^sup>*" by simp
      hence "fi v' \<in> E\<^sup>*``V0" using `u\<in>V0` by auto
      hence "f (fi v') \<in> f ` E\<^sup>* `` V0" by blast
      thus "v' \<in> f ` E\<^sup>* `` V0" by simp
    qed
    thus "finite (E'\<^sup>*``V0')"
      by (rule finite_subset) auto
  qed

  sublocale G'!: finite_reachable_digraph G' 
    using G'_finite_reachable_digraph .


end  

lemma  
  assumes A: "finite_reachable_digraph_inv ext G"
  assumes B: "inj_on f (dg_V G)"
  shows frdg_rename_inv: "\<lbrakk>ext G \<Longrightarrow> ext' (rdg_rename m f G)\<rbrakk> 
    \<Longrightarrow> finite_reachable_digraph_inv ext' (rdg_rename m f G)" (is "PROP ?G1")
proof -
  from A interpret finite_reachable_digraph G 
    by (simp add: finite_reachable_digraph_inv_iff)
  from B interpret frdg_rename_precond G f m by unfold_locales

  show "PROP ?G1" 
    apply (rule rdg_rename_inv[OF A B])
    unfolding finite_reachable_digraph_inv_ext_def
    by (simp add: G'_finite_reachable_digraph)
qed

locale fdg_rename_precond = rdg_rename_precond G f ext
  + finite_rooted_digraph G
  for G :: "('u,'more) rooted_digraph_scheme" 
  and f :: "'u \<Rightarrow> 'v" 
  and ext :: "'moret"
begin

  sublocale frdg_rename_precond by unfold_locales

  lemma G'_finite_rooted_digraph: "finite_rooted_digraph G'"
    apply unfold_locales
    unfolding V0'_fmt V'_fmt
    by auto

  sublocale G'!: finite_rooted_digraph G' 
    using G'_finite_rooted_digraph .

end  

lemma  
  assumes A: "finite_rooted_digraph_inv ext G"
  assumes B: "inj_on f (dg_V G)"
  shows fdg_rename_inv: "\<lbrakk>ext G \<Longrightarrow> ext' (rdg_rename m f G)\<rbrakk> 
    \<Longrightarrow> finite_rooted_digraph_inv ext' (rdg_rename m f G)" (is "PROP ?G1")
proof -
  from A interpret finite_rooted_digraph G 
    by (simp add: finite_rooted_digraph_inv_iff)
  from B interpret fdg_rename_precond G f m by unfold_locales

  show "PROP ?G1" 
    apply (rule frdg_rename_inv[OF A B])
    unfolding finite_rooted_digraph_inv_ext_def
    by (simp add: G'_finite_rooted_digraph)
qed

locale nn_finite_reachable_digraph = 
  finite_reachable_digraph G 
  + nn_rooted_digraph G 
  for G :: "('v,'more) rooted_digraph_scheme"

sublocale nn_finite_rooted_digraph < nn_finite_reachable_digraph 
  by unfold_locales


section {* Buchi Graphs *}

record 'Q generalized_buchi_graph = "'Q rooted_digraph" +
  gbg_F :: "'Q set set"

locale generalized_inf_buchi_graph = rooted_digraph G 
  for G :: "('Q,'more) generalized_buchi_graph_scheme" +
  assumes wf_F: "gbg_F G \<subseteq> Pow V"
begin
  abbreviation "F \<equiv> gbg_F G"
  abbreviation "Q \<equiv> dg_V G" (* TODO: Can we hide V ? *)

  lemma super_rdg: "rooted_digraph G" by unfold_locales

  definition is_acc :: "'Q word \<Rightarrow> bool" 
    where "is_acc r \<equiv> (\<forall>A\<in>F. \<exists>\<^sub>\<infinity>i. r i \<in> A)"

  definition "is_acc_run r \<equiv> is_run r \<and> is_acc r"

  lemma acc_run_run: "is_acc_run r \<Longrightarrow> is_run r"
    unfolding is_acc_run_def by simp
  lemma acc_run_acc: "is_acc_run r \<Longrightarrow> is_acc r"
    unfolding is_acc_run_def by simp

  lemmas acc_run_reachable = run_reachable[OF acc_run_run]
  lemmas wf_acc_run = wf_run[OF acc_run_run]

  lemma acc_eq_limit_if_finite: 
    assumes FIN: "finite (range r)"  
    shows "is_acc r \<longleftrightarrow> (\<forall>A\<in>F. limit r \<inter> A \<noteq> {})"
  proof 
    assume "\<forall>A\<in>F. limit r \<inter> A \<noteq> {}"
    thus "is_acc r"
      unfolding is_acc_def
      by (metis limit_inter_INF)
  next
    from FIN have FIN': "\<And>A. finite (A \<inter> range r)"
      by simp

    assume "is_acc r"
    hence AUX: "\<forall>A\<in>F. \<exists>\<^sub>\<infinity>i. r i \<in> (A \<inter> range r)"
      unfolding is_acc_def by auto
    have "\<forall>A\<in>F. limit r \<inter> (A \<inter> range r) \<noteq> {}"
      apply (rule ballI)
      apply (drule bspec[OF AUX])
      apply (subst (asm) fin_ex_inf_eq_limit[OF FIN'])
      .
    thus "\<forall>A\<in>F. limit r \<inter> A \<noteq> {}" 
      by auto
  qed

  lemma is_acc_suffix[simp]: 
    "is_acc (suffix i r) \<longleftrightarrow> is_acc r"
    unfolding is_acc_def suffix_def
    apply (clarsimp simp: INFM_nat)
    apply (rule iffI)
    apply (metis trans_less_add2)
    by (metis add_lessD1 less_imp_add_positive nat_add_left_cancel_less)

end


definition "generalized_inf_buchi_graph_inv_ext ext G 
  \<equiv> generalized_inf_buchi_graph G \<and> ext G"
abbreviation "generalized_inf_buchi_graph_inv ext 
  \<equiv> rooted_digraph_inv (generalized_inf_buchi_graph_inv_ext ext)"

lemma generalized_inf_buchi_graph_inv_iff: 
  "generalized_inf_buchi_graph_inv ext G 
  \<longleftrightarrow> generalized_inf_buchi_graph G \<and> ext G"
  by (auto simp add: 
    rooted_digraph_inv_iff 
    generalized_inf_buchi_graph_inv_ext_def 
    generalized_inf_buchi_graph.super_rdg)


abbreviation "gbg_is_acc_run \<equiv> generalized_inf_buchi_graph.is_acc_run"

locale generalized_fr_buchi_graph = 
  generalized_inf_buchi_graph G +
  finite_reachable_digraph G
  for G :: "('Q,'more) generalized_buchi_graph_scheme"
begin

  lemma super_gbg_inf: "generalized_inf_buchi_graph G"
    and super_frdg: "finite_reachable_digraph G"
    by unfold_locales

  lemma is_acc_run_limit_alt: 
    "is_acc_run r \<longleftrightarrow> is_run r \<and> (\<forall>A\<in>F. limit r \<inter> A \<noteq> {})"
    using acc_eq_limit_if_finite[symmetric]
    unfolding is_acc_run_def by (auto dest: is_run_finite)

end

definition "generalized_fr_buchi_graph_inv_ext ext G 
  \<equiv> generalized_fr_buchi_graph G \<and> ext G"
abbreviation "generalized_fr_buchi_graph_inv ext 
  \<equiv> generalized_inf_buchi_graph_inv (generalized_fr_buchi_graph_inv_ext ext)"

lemma generalized_fr_buchi_graph_inv_iff: 
  "generalized_fr_buchi_graph_inv ext G 
  \<longleftrightarrow> generalized_fr_buchi_graph G \<and> ext G"
  by (auto simp add: 
    generalized_inf_buchi_graph_inv_iff 
    generalized_fr_buchi_graph_inv_ext_def 
    generalized_fr_buchi_graph.super_gbg_inf)

locale generalized_buchi_graph = 
  generalized_fr_buchi_graph G +
  finite_rooted_digraph G
  for G :: "('Q,'more) generalized_buchi_graph_scheme"
begin
  lemma super_gbg_fr: "generalized_fr_buchi_graph G" and
    super_frdg: "finite_rooted_digraph G"
    by unfold_locales
end

definition "generalized_buchi_graph_inv_ext ext G 
  \<equiv> generalized_buchi_graph G \<and> ext G"
abbreviation "generalized_buchi_graph_inv ext 
  \<equiv> generalized_fr_buchi_graph_inv (generalized_buchi_graph_inv_ext ext)"

lemma generalized_buchi_graph_inv_iff: 
  "generalized_buchi_graph_inv ext G 
  \<longleftrightarrow> generalized_buchi_graph G \<and> ext G"
  by (auto simp add: 
    generalized_fr_buchi_graph_inv_iff 
    generalized_buchi_graph_inv_ext_def 
    generalized_buchi_graph.super_gbg_fr)

locale inf_buchi_graph = generalized_inf_buchi_graph G 
  for G :: "('Q,'more) generalized_buchi_graph_scheme" +
  assumes F_sng: "\<exists>A. F = {A}"
begin
  definition "A \<equiv> the_elem F"

  lemma F_fmt[simp]: "F = {A}"
    unfolding A_def
    using F_sng
    by fastforce
end

locale fr_buchi_graph = inf_buchi_graph G + generalized_fr_buchi_graph G
  for G :: "('Q,'more) generalized_buchi_graph_scheme"

locale buchi_graph = inf_buchi_graph G + generalized_buchi_graph G
  for G :: "('Q,'more) generalized_buchi_graph_scheme"
begin
  sublocale fr_buchi_graph by unfold_locales
end

subsection "Renaming of Nodes"

definition "gbg_rename_ext ext f G \<equiv> \<lparr>
  gbg_F = { f`A | A. A\<in>gbg_F G }, \<dots> = ext
\<rparr>"
abbreviation "gbg_rename ext f G \<equiv> rdg_rename (gbg_rename_ext ext f G) f G"

lemmas gbg_rename_defs = rdg_rename_defs gbg_rename_ext_def

locale gbg_inf_rename_precond = rdg_rename_precond G f "gbg_rename_ext ext f G"
  + generalized_inf_buchi_graph G
  for G :: "('u,'more) generalized_buchi_graph_scheme" 
  and f :: "'u \<Rightarrow> 'v" 
  and ext :: "'moret"
begin
  abbreviation "F' \<equiv> gbg_F G'"

  lemma F'_fmt: "F' = { f`A | A. A\<in>F }"
    unfolding gbg_rename_defs
    by auto

  lemma G'_generalized_inf_buchi_graph: "generalized_inf_buchi_graph G'"
    apply unfold_locales
    unfolding F'_fmt V'_fmt
    using wf_F
    by auto

  sublocale G'!: generalized_inf_buchi_graph G' 
    using G'_generalized_inf_buchi_graph .

  lemma acc_sim1: "is_acc r \<Longrightarrow> G'.is_acc (f o r)"
    unfolding is_acc_def G'.is_acc_def F'_fmt
    by (fastforce intro: imageI simp: INFM_nat)

  lemma acc_sim2: 
    assumes "G'.is_acc r" shows "is_acc (fi o r)"
  proof -
    from assms have 1: "\<And>A m. A \<in> F \<Longrightarrow> \<exists>i>m. r i \<in> f`A"
      unfolding G'.is_acc_def F'_fmt
      by (auto simp: INFM_nat)

    { fix A m 
      assume 2: "A \<in> F"  
      from 1[OF this, of m] have "\<exists>i>m. fi (r i) \<in> A"
        using wf_F
        apply clarsimp
        by (metis Pow_iff 2 fi_f in_mono)
    } thus ?thesis
      unfolding is_acc_def
      by (auto simp: INFM_nat)
  qed

  lemma acc_run_sim1: "is_acc_run r \<Longrightarrow> G'.is_acc_run (f o r)"
    using acc_sim1 run_sim1 unfolding is_acc_run_def G'.is_acc_run_def
    by auto
  lemma acc_run_sim2: "G'.is_acc_run r \<Longrightarrow> is_acc_run (fi o r)"
    using acc_sim2 run_sim2 unfolding G'.is_acc_run_def is_acc_run_def
    by auto

end

lemma  
  assumes A: "generalized_inf_buchi_graph_inv ext G"
  assumes B: "inj_on f (dg_V G)"
  shows gbg_rename_is_acc_run_sim1: 
    "\<And>r. gbg_is_acc_run G r \<Longrightarrow> gbg_is_acc_run (gbg_rename m f G) (f o r)"
    (is "PROP ?G1")
  and gbg_rename_is_acc_run_sim2: 
    "\<And>r. gbg_is_acc_run (gbg_rename m f G) r 
      \<Longrightarrow> gbg_is_acc_run G (the_inv_into (dg_V G) f o r)"
    (is "PROP ?G2")
  and gbg_inf_rename_inv: "\<lbrakk>ext G \<Longrightarrow> ext' (gbg_rename m f G)\<rbrakk> 
    \<Longrightarrow> generalized_inf_buchi_graph_inv ext' (gbg_rename m f G)" (is "PROP ?G3")
proof -
  from A interpret generalized_inf_buchi_graph G 
    by (simp add: generalized_inf_buchi_graph_inv_iff)
  from B interpret gbg_inf_rename_precond G f m by unfold_locales

  show "PROP ?G1" by (simp add: acc_run_sim1)
  show "PROP ?G2" 
    using acc_run_sim2 by (simp add: fi_def)

  show "PROP ?G3" 
    apply (rule rdg_rename_inv[OF A B])
    unfolding generalized_inf_buchi_graph_inv_ext_def
    by (simp add: G'_generalized_inf_buchi_graph)
qed

locale gbg_fr_rename_precond = 
    gbg_inf_rename_precond G f ext
  + generalized_fr_buchi_graph G
  for G :: "('u,'more) generalized_buchi_graph_scheme" 
  and f :: "'u \<Rightarrow> 'v" 
  and ext :: "'moret"
begin

  sublocale frdg_rename_precond G f "gbg_rename_ext ext f G"
    by unfold_locales

  lemma G'_generalized_fr_buchi_graph: "generalized_fr_buchi_graph G'"
    by unfold_locales

  sublocale G'!: generalized_fr_buchi_graph G' 
    using G'_generalized_fr_buchi_graph .

end

lemma  
  assumes A: "generalized_fr_buchi_graph_inv ext G"
  assumes B: "inj_on f (dg_V G)"
  shows gbg_fr_rename_inv: "\<lbrakk>ext G \<Longrightarrow> ext' (gbg_rename m f G)\<rbrakk> 
    \<Longrightarrow> generalized_fr_buchi_graph_inv ext' (gbg_rename m f G)" (is "PROP ?G1")
proof -
  from A interpret generalized_fr_buchi_graph G 
    by (simp add: generalized_fr_buchi_graph_inv_iff)
  from B interpret gbg_fr_rename_precond G f m by unfold_locales

  show "PROP ?G1" 
    apply (rule gbg_inf_rename_inv[OF A B])
    unfolding generalized_fr_buchi_graph_inv_ext_def
    by (simp add: G'_generalized_fr_buchi_graph)
qed


locale gbg_rename_precond = 
    gbg_fr_rename_precond G f ext
  + generalized_buchi_graph G
  for G :: "('u,'more) generalized_buchi_graph_scheme" 
  and f :: "'u \<Rightarrow> 'v" 
  and ext :: "'moret"
begin

  sublocale fdg_rename_precond G f "gbg_rename_ext ext f G"
    by unfold_locales

  lemma G'_generalized_buchi_graph: "generalized_buchi_graph G'"
    by unfold_locales

  sublocale G'!: generalized_buchi_graph G' 
    using G'_generalized_buchi_graph .

end

lemma  
  assumes A: "generalized_buchi_graph_inv ext G"
  assumes B: "inj_on f (dg_V G)"
  shows gbg_rename_inv: "\<lbrakk>ext G \<Longrightarrow> ext' (gbg_rename m f G)\<rbrakk> 
    \<Longrightarrow> generalized_buchi_graph_inv ext' (gbg_rename m f G)" (is "PROP ?G1")
proof -
  from A interpret generalized_buchi_graph G 
    by (simp add: generalized_buchi_graph_inv_iff)
  from B interpret gbg_rename_precond G f m by unfold_locales

  show "PROP ?G1" 
    apply (rule gbg_fr_rename_inv[OF A B])
    unfolding generalized_buchi_graph_inv_ext_def
    by (simp add: G'_generalized_buchi_graph)
qed

section {* Buchi Automata *}

record ('Q,'L) generalized_buchi_automaton = "'Q generalized_buchi_graph" +
  gba_\<Sigma> :: "'L set"
  gba_L :: "'Q \<Rightarrow> 'L set"

locale generalized_inf_buchi_automaton = generalized_inf_buchi_graph G
  for G :: "('Q,'L,'more) generalized_buchi_automaton_scheme" +
  assumes wf_L: "\<And>q. q\<in>Q \<Longrightarrow> gba_L G q \<subseteq> gba_\<Sigma> G"
begin
  abbreviation "\<Sigma> \<equiv> gba_\<Sigma> G"
  abbreviation "L \<equiv> gba_L G"

  lemma super_gbg_inf: "generalized_inf_buchi_graph G" by unfold_locales

  definition "accept w \<equiv> \<exists>r. is_acc_run r \<and> (\<forall>i. w i \<in> L (r i))"
  lemma acceptI[intro?]: "\<lbrakk>is_acc_run r; \<And>i. w i \<in> L (r i)\<rbrakk> \<Longrightarrow> accept w"
    by (auto simp: accept_def)

  definition "lang \<equiv> Collect (accept)"
  lemma langI[intro?]: "accept w \<Longrightarrow> w\<in>lang" by (auto simp: lang_def)

end

definition "generalized_inf_buchi_automaton_inv_ext ext G 
  \<equiv> generalized_inf_buchi_automaton G \<and> ext G"
abbreviation "generalized_inf_buchi_automaton_inv ext 
  \<equiv> generalized_inf_buchi_graph_inv 
      (generalized_inf_buchi_automaton_inv_ext ext)"

lemma generalized_inf_buchi_automaton_inv_iff: 
  "generalized_inf_buchi_automaton_inv ext G 
  \<longleftrightarrow> generalized_inf_buchi_automaton G \<and> ext G"
  by (auto simp add: 
    generalized_inf_buchi_graph_inv_iff 
    generalized_inf_buchi_automaton_inv_ext_def 
    generalized_inf_buchi_automaton.super_gbg_inf)

abbreviation "gba_accept \<equiv> generalized_inf_buchi_automaton.accept"
abbreviation "gba_lang \<equiv> generalized_inf_buchi_automaton.lang"

locale generalized_fr_buchi_automaton = 
  generalized_inf_buchi_automaton G
  + generalized_fr_buchi_graph G
  for G :: "('Q,'L,'more) generalized_buchi_automaton_scheme"
begin
  lemma super_gba_inf: "generalized_inf_buchi_automaton G"
    and super_gbg_fr: "generalized_fr_buchi_graph G"
    by unfold_locales

end

definition "generalized_fr_buchi_automaton_inv_ext ext G 
  \<equiv> generalized_fr_buchi_automaton G \<and> ext G"
abbreviation "generalized_fr_buchi_automaton_inv ext 
  \<equiv> generalized_inf_buchi_automaton_inv 
      (generalized_fr_buchi_automaton_inv_ext ext)"

lemma generalized_fr_buchi_automaton_inv_iff: 
  "generalized_fr_buchi_automaton_inv ext G 
  \<longleftrightarrow> generalized_fr_buchi_automaton G \<and> ext G"
  by (auto simp add: 
    generalized_inf_buchi_automaton_inv_iff 
    generalized_fr_buchi_automaton_inv_ext_def 
    generalized_fr_buchi_automaton.super_gba_inf)


locale generalized_buchi_automaton = 
  generalized_fr_buchi_automaton G
  + generalized_buchi_graph G
  for G :: "('Q,'L,'more) generalized_buchi_automaton_scheme"
begin
  lemma super_gba_fr: "generalized_fr_buchi_automaton G"
    and super_gbg: "generalized_buchi_graph G"
    by unfold_locales
end

definition "generalized_buchi_automaton_inv_ext ext G 
  \<equiv> generalized_buchi_automaton G \<and> ext G"
abbreviation "generalized_buchi_automaton_inv ext 
  \<equiv> generalized_fr_buchi_automaton_inv 
      (generalized_buchi_automaton_inv_ext ext)"

lemma generalized_buchi_automaton_inv_iff: 
  "generalized_buchi_automaton_inv ext G 
  \<longleftrightarrow> generalized_buchi_automaton G \<and> ext G"
  by (auto simp add: 
    generalized_fr_buchi_automaton_inv_iff 
    generalized_buchi_automaton_inv_ext_def 
    generalized_buchi_automaton.super_gba_fr)

subsection {* Renaming of States *}

definition "gba_rename_ext ext f G \<equiv> \<lparr>
  gba_\<Sigma> = gba_\<Sigma> G,
  gba_L = \<lambda>q. if q \<in> f`dg_V G then gba_L G (the_inv_into (dg_V G) f q) else {}, 
  \<dots> = ext
\<rparr>"

abbreviation "gba_rename ext f G \<equiv> gbg_rename (gba_rename_ext ext f G) f G"

lemmas gba_rename_defs = gbg_rename_defs gba_rename_ext_def

locale gba_inf_rename_precond = 
    gbg_inf_rename_precond G f "gba_rename_ext ext f G"
  + generalized_inf_buchi_automaton G
  for G :: "('u,'L,'more) generalized_buchi_automaton_scheme" 
  and f :: "'u \<Rightarrow> 'v" 
  and ext :: "'moret"
begin

  abbreviation "L' \<equiv> gba_L G'"
  lemma L'_fmt: "L' = 
    (\<lambda>q. if q \<in> V' then L (fi q) else {})"
    unfolding gba_rename_defs fi_def
    by simp

  abbreviation "\<Sigma>' \<equiv> gba_\<Sigma> G'"
  lemma \<Sigma>'_fmt[simp]: "\<Sigma>' = \<Sigma>"
    unfolding gba_rename_defs
    by simp
    
  lemma G'_generalized_inf_buchi_automaton: "generalized_inf_buchi_automaton G'"
    apply unfold_locales
    unfolding L'_fmt V'_fmt \<Sigma>'_fmt
    using wf_L by (fastforce simp: fi_f)
  sublocale G'!: generalized_inf_buchi_automaton G' 
    using G'_generalized_inf_buchi_automaton .

  lemma L_sim1: "\<lbrakk>range r \<subseteq> V\<rbrakk> \<Longrightarrow> L (r i) \<subseteq> L' (f (r i))"
    by (auto
      simp: L'_fmt V'_fmt fi_f 
      dest: inj_onD[OF f_inj]
      dest!: set_rev_mp[OF rangeI[of _ i]])
    
  lemma L_sim2: "\<lbrakk> range r \<subseteq> f`V \<rbrakk> \<Longrightarrow> L' (r i) \<subseteq> L (fi (r i))"
    by (auto
      simp: L'_fmt V_fmt f_fi
      dest!: set_rev_mp[OF rangeI[of _ i]])
    
  lemma accept_eq[simp]: "G'.accept = accept"
    apply (rule ext)
    unfolding G'.accept_def accept_def
  proof safe
    fix w r
    assume R: "G'.is_acc_run r"
    assume L: "\<forall>i. w i \<in> L' (r i)"
    have RAN: "range r \<subseteq> f`V"
      using G'.wf_run[OF G'.acc_run_run[OF R]]
      unfolding V'_fmt
      by simp
      
    show "\<exists>r. is_acc_run r \<and> (\<forall>i. w i \<in> L (r i))"
      apply (rule exI[where x = "fi o r"])
      using L acc_run_sim2[OF R] L_sim2[OF RAN]
      by auto
  next
    fix w r
    assume R: "is_acc_run r" 
    assume L: "\<forall>i. w i \<in> L (r i)"
    
    from R have RAN: "range r \<subseteq> V"
      using wf_run[OF acc_run_run[OF R]] by simp
      
    show "\<exists>r. G'.is_acc_run r \<and> (\<forall>i. w i \<in> L' (r i))"
      apply (rule exI[where x="f o r"])
      using L acc_run_sim1[OF R] L_sim1[OF RAN]
      by auto
  qed

  lemma lang_eq[simp]: "G'.lang = lang"
    unfolding G'.lang_def lang_def by simp

end

lemma  
  assumes A: "generalized_inf_buchi_automaton_inv ext G"
  assumes B: "inj_on f (dg_V G)"
  shows gba_rename_accept: 
    "\<And>w. gba_accept (gba_rename m f G) w \<longleftrightarrow> gba_accept G w" (is "PROP ?G1")
  and gba_rename_lang: 
    "gba_lang (gba_rename m f G) = gba_lang G" (is "PROP ?G2")
  and gba_inf_rename_inv: "\<lbrakk>ext G \<Longrightarrow> ext' (gba_rename m f G)\<rbrakk> 
    \<Longrightarrow> generalized_inf_buchi_automaton_inv ext' (gba_rename m f G)"
    (is "PROP ?G3")
proof -
  from A interpret generalized_inf_buchi_automaton G 
    by (simp add: generalized_inf_buchi_automaton_inv_iff)
  from B interpret gba_inf_rename_precond G f m by unfold_locales

  show "PROP ?G1" by (simp)
  show "PROP ?G2" by simp

  show "PROP ?G3" 
    apply (rule gbg_inf_rename_inv[OF A B])
    unfolding generalized_inf_buchi_automaton_inv_ext_def
    by (simp add: G'_generalized_inf_buchi_automaton)
qed


locale gba_fr_rename_precond = 
    gba_inf_rename_precond G f ext
  + generalized_fr_buchi_automaton G
  for G :: "('u,'L,'more) generalized_buchi_automaton_scheme" 
  and f :: "'u \<Rightarrow> 'v" 
  and ext :: "'moret"
begin
  sublocale frdg_rename_precond G f "gbg_rename_ext (gba_rename_ext ext f G) f G"
    by unfold_locales

  lemma G'_generalized_fr_buchi_automaton: "generalized_fr_buchi_automaton G'"
    by unfold_locales

  sublocale G'!: generalized_fr_buchi_automaton G' 
    using G'_generalized_fr_buchi_automaton .

end  

lemma  
  assumes A: "generalized_fr_buchi_automaton_inv ext G"
  assumes B: "inj_on f (dg_V G)"
  shows gba_fr_rename_inv: "\<lbrakk>ext G \<Longrightarrow> ext' (gba_rename m f G)\<rbrakk> 
    \<Longrightarrow> generalized_fr_buchi_automaton_inv ext' (gba_rename m f G)"
    (is "PROP ?G1")
proof -
  from A interpret generalized_fr_buchi_automaton G 
    by (simp add: generalized_fr_buchi_automaton_inv_iff)
  from B interpret gba_fr_rename_precond G f m by unfold_locales

  show "PROP ?G1" 
    apply (rule gba_inf_rename_inv[OF A B])
    unfolding generalized_fr_buchi_automaton_inv_ext_def
    by (simp add: G'_generalized_fr_buchi_automaton)
qed


locale gba_rename_precond = 
    gba_fr_rename_precond G f ext
  + generalized_buchi_automaton G
  for G :: "('u,'L,'more) generalized_buchi_automaton_scheme" 
  and f :: "'u \<Rightarrow> 'v" 
  and ext :: "'moret"
begin
  sublocale fdg_rename_precond G f "gbg_rename_ext (gba_rename_ext ext f G) f G"
    by unfold_locales

  lemma G'_generalized_buchi_automaton: "generalized_buchi_automaton G'"
    by unfold_locales

  sublocale G'!: generalized_buchi_automaton G' 
    using G'_generalized_buchi_automaton .
end

lemma  
  assumes A: "generalized_buchi_automaton_inv ext G"
  assumes B: "inj_on f (dg_V G)"
  shows gba_rename_inv: "\<lbrakk>ext G \<Longrightarrow> ext' (gba_rename m f G)\<rbrakk> 
    \<Longrightarrow> generalized_buchi_automaton_inv ext' (gba_rename m f G)"
    (is "PROP ?G1")
proof -
  from A interpret generalized_buchi_automaton G 
    by (simp add: generalized_buchi_automaton_inv_iff)
  from B interpret gba_rename_precond G f m by unfold_locales

  show "PROP ?G1" 
    apply (rule gba_fr_rename_inv[OF A B])
    unfolding generalized_buchi_automaton_inv_ext_def
    by (simp add: G'_generalized_buchi_automaton)
qed


locale inf_buchi_automaton = 
  generalized_inf_buchi_automaton G 
  + inf_buchi_graph G
  for G :: "('Q,'L,'more) generalized_buchi_automaton_scheme"

locale fr_buchi_automaton = generalized_fr_buchi_automaton G + fr_buchi_graph G
  for G :: "('Q,'L,'more) generalized_buchi_automaton_scheme"

locale buchi_automaton = generalized_buchi_automaton G + buchi_graph G
  for G :: "('Q,'L,'more) generalized_buchi_automaton_scheme"
begin
  sublocale fr_buchi_automaton by unfold_locales
end

locale nn_generalized_inf_buchi_graph = 
  generalized_inf_buchi_graph G 
  + nn_rooted_digraph G
  for G :: "('Q,'more) generalized_buchi_graph_scheme"

locale nn_generalized_fr_buchi_graph = 
  generalized_fr_buchi_graph G 
  + nn_finite_reachable_digraph G
  for G :: "('Q,'more) generalized_buchi_graph_scheme"

locale nn_generalized_buchi_graph = 
  generalized_buchi_graph G 
  + nn_finite_rooted_digraph G
  for G :: "('Q,'more) generalized_buchi_graph_scheme"
begin
  sublocale nn_generalized_fr_buchi_graph by unfold_locales
end

locale nn_generalized_inf_buchi_automaton = 
  generalized_inf_buchi_automaton G 
  + nn_rooted_digraph G
  for G :: "('Q,'L,'more) generalized_buchi_automaton_scheme"

locale nn_generalized_fr_buchi_automaton = 
  generalized_fr_buchi_automaton G 
  + nn_finite_reachable_digraph G
  for G :: "('Q,'L,'more) generalized_buchi_automaton_scheme"

locale nn_generalized_buchi_automaton = 
  generalized_buchi_automaton G 
  + nn_finite_rooted_digraph G
  for G :: "('Q,'L,'more) generalized_buchi_automaton_scheme"
begin
  sublocale nn_generalized_fr_buchi_automaton by unfold_locales
end

locale nn_inf_buchi_graph = 
  inf_buchi_graph G 
  + nn_rooted_digraph G
  for G :: "('Q,'more) generalized_buchi_graph_scheme"

locale nn_fr_buchi_graph = 
  fr_buchi_graph G 
  + nn_finite_reachable_digraph G
  for G :: "('Q,'more) generalized_buchi_graph_scheme"

locale nn_buchi_graph = 
  buchi_graph G 
  + nn_finite_rooted_digraph G
  for G :: "('Q,'more) generalized_buchi_graph_scheme"
begin
  sublocale nn_fr_buchi_graph by unfold_locales
end

locale nn_inf_buchi_automaton = 
  inf_buchi_automaton G 
  + nn_rooted_digraph G
  for G :: "('Q,'L,'more) generalized_buchi_automaton_scheme"

locale nn_fr_buchi_automaton = 
  fr_buchi_automaton G 
  + nn_finite_reachable_digraph G
  for G :: "('Q,'L,'more) generalized_buchi_automaton_scheme"

locale nn_buchi_automaton = 
  buchi_automaton G 
  + nn_finite_rooted_digraph G
  for G :: "('Q,'L,'more) generalized_buchi_automaton_scheme"
begin
  sublocale nn_fr_buchi_automaton by unfold_locales
end


section {* Indexed Generalized Buchi Graphs *}
text {* A representation of generalized Buchi Graphs with ordered acceptance
  classes. This is an implementation of GBGs, on which some algorithms are 
  simpler to formulate.
*}

record 'Q idx_generalized_buchi_graph = "'Q rooted_digraph" +
  igbg_num_acc :: nat
  igbg_acc :: "'Q \<Rightarrow> nat set"

locale idx_generalized_inf_buchi_graph = rooted_digraph G
  for G :: "('Q,'more) idx_generalized_buchi_graph_scheme" +
  assumes acc_bound: "\<Union>range (igbg_acc G) \<subseteq> {0..<igbg_num_acc G}"
  assumes wf_acc: "igbg_acc G q \<noteq> {} \<Longrightarrow> q \<in> V"
begin
  abbreviation "Q \<equiv> dg_V G"
  abbreviation num_acc where "num_acc \<equiv> igbg_num_acc G"
  abbreviation acc where "acc \<equiv> igbg_acc G"

  definition "accn i \<equiv> {q . i\<in>acc q}"
  definition "F \<equiv> { accn i | i. i<num_acc }"

  lemma F_fin[simp,intro!]: "finite F"
    unfolding F_def
    by auto
  lemma wf_F: "F \<subseteq> Pow Q"
    unfolding F_def accn_def
    using wf_acc
    by auto

  definition 
    "to_gbg_ext m \<equiv> \<lparr> dg_V = V, dg_E = E, rdg_V0 = V0, gbg_F = F, \<dots>=m \<rparr>"

  lemma is_gbg_inf: "generalized_inf_buchi_graph (to_gbg_ext m)"
    apply unfold_locales
    unfolding to_gbg_ext_def
    apply simp_all
    using wf_E wf_V0 wf_F .

  sublocale gbg!: generalized_inf_buchi_graph "(to_gbg_ext m)"
    apply (rule is_gbg_inf)
    done

  lemma to_gbg_fields[simp]: 
    "gbg.V T m = V"
    "gbg.E T m = E"
    "gbg.V0 T m = V0"
    "gbg.F T m = F"
    "generalized_buchi_graph.more (to_gbg_ext m) = m"
    unfolding to_gbg_ext_def by auto

  lemma to_gbg_is_run[simp]: "gbg.is_run T m = is_run"
    unfolding gbg.is_run_def[abs_def] is_run_def[abs_def]
    by simp

  definition is_acc_internal_def: "is_acc \<equiv> gbg.is_acc TYPE(unit) ()"

  lemma is_acc_def: "is_acc r = (\<forall>n<num_acc. \<exists>\<^sub>\<infinity>i. n \<in> acc (r i))"
    unfolding is_acc_internal_def 
    unfolding gbg.is_acc_def[abs_def] 
    apply simp
    unfolding F_def accn_def
    apply (blast intro: INFM_mono)
    done

  lemma to_gbg_is_acc[simp]: "gbg.is_acc T m = is_acc"
    unfolding is_acc_internal_def 
    unfolding gbg.is_acc_def[abs_def] 
    by simp

  definition is_acc_run_internal_def: "is_acc_run \<equiv> gbg.is_acc_run TYPE(unit) ()"
  lemma is_acc_run_def: "is_acc_run r \<longleftrightarrow> is_run r \<and> is_acc r"
    unfolding is_acc_run_internal_def 
    unfolding gbg.is_acc_run_def[abs_def] 
    by simp

  lemma to_gbg_is_acc_run[simp]: "gbg.is_acc_run T m = is_acc_run"
    unfolding is_acc_run_internal_def 
    unfolding gbg.is_acc_run_def[abs_def] 
    by simp
end

abbreviation "igbg_is_acc_run \<equiv> idx_generalized_inf_buchi_graph.is_acc_run"

abbreviation "igbg_inf_\<alpha> e\<alpha> G 
  \<equiv> idx_generalized_inf_buchi_graph.to_gbg_ext G (e\<alpha> G)"
definition "igbg_inf_rel e\<alpha> einv \<equiv> 
  br (igbg_inf_\<alpha> e\<alpha>) (\<lambda>G. idx_generalized_inf_buchi_graph G \<and> einv G)"

lemma 
  igbg_is_run_refine: "(rdg_is_run, rdg_is_run) \<in> igbg_inf_rel e\<alpha> einv \<rightarrow> Id"
  and igbg_is_acc_run_refine: 
    "(igbg_is_acc_run, gbg_is_acc_run) \<in> igbg_inf_rel e\<alpha> einv \<rightarrow> Id"
  by (clarsimp_all
    simp: br_def igbg_inf_rel_def
    simp: idx_generalized_inf_buchi_graph.to_gbg_is_run
    simp: idx_generalized_inf_buchi_graph.to_gbg_is_acc_run)

locale idx_generalized_fr_buchi_graph = 
  idx_generalized_inf_buchi_graph G
  + finite_reachable_digraph G
  for G :: "('Q,'more) idx_generalized_buchi_graph_scheme"
begin
  lemma is_gbg_fr: "generalized_fr_buchi_graph (to_gbg_ext m)"
    apply unfold_locales
    apply simp
    done

  sublocale gbg!: generalized_fr_buchi_graph "(to_gbg_ext m)"
    by (rule is_gbg_fr)
end

definition "igbg_fr_einv einv G \<equiv> idx_generalized_fr_buchi_graph G \<and> einv G"
abbreviation "igbg_fr_rel e\<alpha> einv \<equiv> igbg_inf_rel e\<alpha> (igbg_fr_einv einv)"

locale idx_generalized_buchi_graph = 
  idx_generalized_fr_buchi_graph G
  + finite_rooted_digraph G
  for G :: "('Q,'more) idx_generalized_buchi_graph_scheme"
begin
  lemma is_gbg: "generalized_buchi_graph (to_gbg_ext m)"
    apply unfold_locales
    apply simp_all
    done

  sublocale gbg!: generalized_buchi_graph "(to_gbg_ext m)"
    by (rule is_gbg)
end

definition "igbg_einv einv G \<equiv> idx_generalized_buchi_graph G \<and> einv G"
abbreviation "igbg_rel e\<alpha> einv \<equiv> igbg_fr_rel e\<alpha> (igbg_einv einv)"

record ('Q,'L) idx_generalized_buchi_automaton 
  = "'Q idx_generalized_buchi_graph" +
  igba_\<Sigma> :: "'L set"
  igba_L :: "'Q \<Rightarrow> 'L set"

locale idx_generalized_inf_buchi_automaton = idx_generalized_inf_buchi_graph G 
  for G :: "('Q,'L,'more) idx_generalized_buchi_automaton_scheme" +
  assumes wf_L: "\<And>q. q\<in>Q \<Longrightarrow> igba_L G q \<subseteq> igba_\<Sigma> G"
begin
  abbreviation "\<Sigma> \<equiv> igba_\<Sigma> G"
  abbreviation "L \<equiv> igba_L G"

  abbreviation "to_gba_ext m \<equiv> to_gbg_ext \<lparr> gba_\<Sigma> = \<Sigma>, gba_L = L, \<dots> = m \<rparr>"

  lemma is_gba_inf: "generalized_inf_buchi_automaton (to_gba_ext m)"
    apply unfold_locales
    apply simp
    unfolding to_gbg_ext_def
    apply simp_all
    using wf_L .

  sublocale gbg!: generalized_inf_buchi_automaton "(to_gba_ext m)"
    apply (rule is_gba_inf)
    done
  
  lemma to_gba_fields[simp]: 
    "gbg.\<Sigma> T m = \<Sigma>"
    "gbg.L T m = L"
    "generalized_buchi_automaton.more (to_gba_ext m) = m"
    unfolding to_gbg_ext_def by auto

  definition accept_internal_def: "accept \<equiv> gbg.accept TYPE(unit) ()"
  lemma accept_def: "accept w = (\<exists>r. is_acc_run r \<and> (\<forall>i. w i \<in> L (r i)))"
    unfolding accept_internal_def gbg.accept_def
    by simp
  lemma to_gba_accept[simp]: "gbg.accept T m = accept"
    unfolding accept_def[abs_def] gbg.accept_def[abs_def]
    by simp

  definition lang_internal_def: "lang \<equiv> gbg.lang TYPE(unit) ()"
  lemma lang_def: "lang = Collect accept"
    unfolding lang_internal_def gbg.lang_def
    by simp
  lemma to_gba_lang[simp]: "gbg.lang T m = lang"
    unfolding lang_def gbg.lang_def
    by simp
end

abbreviation "igba_accept \<equiv> idx_generalized_inf_buchi_automaton.accept"
abbreviation "igba_lang \<equiv> idx_generalized_inf_buchi_automaton.lang"

definition "igba_inf_einv einv G 
  \<equiv> idx_generalized_inf_buchi_automaton G \<and> einv G"
definition "igba_e\<alpha> e\<alpha> G \<equiv> \<lparr> gba_\<Sigma> = igba_\<Sigma> G, gba_L = igba_L G, \<dots> = e\<alpha> G \<rparr>"

abbreviation "igba_inf_rel e\<alpha> einv 
  \<equiv> igbg_inf_rel (igba_e\<alpha> e\<alpha>) (igba_inf_einv einv)"

lemma 
  igba_accept_refine: 
    "(igba_accept, gba_accept) \<in> igba_inf_rel e\<alpha> einv \<rightarrow> Id \<rightarrow> bool_rel"
  and igba_lang_refine: 
    "(igba_lang, gba_lang) \<in> igba_inf_rel e\<alpha> einv \<rightarrow> Id"
  apply (clarsimp_all
    simp: br_def igba_inf_einv_def igbg_inf_rel_def igba_e\<alpha>_def
    simp: idx_generalized_inf_buchi_automaton.to_gba_accept
    simp: idx_generalized_inf_buchi_automaton.to_gba_lang)
  done


locale idx_generalized_fr_buchi_automaton = 
  idx_generalized_inf_buchi_automaton G 
  + idx_generalized_fr_buchi_graph G
  for G :: "('Q,'L,'more) idx_generalized_buchi_automaton_scheme"
begin
  lemma is_gba_fr: "generalized_fr_buchi_automaton (to_gba_ext m)"
    apply unfold_locales
    done
    
  sublocale gbg!: generalized_fr_buchi_automaton "(to_gba_ext m)" by (rule is_gba_fr)
end

definition "igba_fr_einv einv G 
  \<equiv> idx_generalized_fr_buchi_automaton G \<and> einv G"

abbreviation "igba_fr_rel e\<alpha> einv 
  \<equiv> igba_inf_rel e\<alpha> (igba_fr_einv einv)"


locale idx_generalized_buchi_automaton = 
  idx_generalized_fr_buchi_automaton G 
  + idx_generalized_buchi_graph G
  for G :: "('Q,'L,'more) idx_generalized_buchi_automaton_scheme"
begin
  lemma is_gba: "generalized_buchi_automaton (to_gba_ext m)"
    apply unfold_locales
    done
    
  sublocale gbg!: generalized_buchi_automaton "(to_gba_ext m)" by (rule is_gba)
end

definition "igba_einv einv G 
  \<equiv> idx_generalized_buchi_automaton G \<and> einv G"

abbreviation "igba_rel e\<alpha> einv 
  \<equiv> igba_fr_rel e\<alpha> (igba_einv einv)"

section "Equivalences on Records"
(* TODO: Move *)
definition "AND_REL l x y \<equiv> \<forall>e\<in>set l. e x y"
lemma asc_imp_AND_REL_I: "\<lbrakk> 
  (\<And>x y. f2 x y \<Longrightarrow> f1 x y); AND_REL (f2#e) x y \<rbrakk> \<Longrightarrow> AND_REL (f1#f2#e) x y"
  unfolding AND_REL_def by auto

definition "dg_eq_ext x y \<equiv> digraph.truncate x = digraph.truncate y"
abbreviation "dg_eq ext \<equiv> AND_REL (dg_eq_ext#ext)"

lemma dg_eqI[intro?]: 
  "\<lbrakk>digraph.truncate x = digraph.truncate y; AND_REL ext x y\<rbrakk> 
  \<Longrightarrow> dg_eq ext x y"
  unfolding dg_eq_ext_def[abs_def] AND_REL_def by auto

lemma dg_eq_fields[simp]: 
  assumes "dg_eq E x y"
  shows "dg_V x = dg_V y"
  and "dg_E x = dg_E y"
  using assms unfolding dg_eq_ext_def[abs_def] AND_REL_def
  apply -
  apply (simp_all add: digraph.defs)
  done

lemma dg_eq_loc[simp]:
  assumes "dg_eq E x y" 
  shows "digraph x \<longleftrightarrow> digraph y"
  and "finite_digraph x \<longleftrightarrow> finite_digraph y"
  using assms apply -
  by (simp_all add: digraph_def 
    finite_digraph_def finite_digraph_axioms_def)

definition "rdg_eq_ext x y 
  \<equiv> rooted_digraph.truncate x = rooted_digraph.truncate y"
abbreviation "rdg_eq ext \<equiv> dg_eq (rdg_eq_ext#ext)"

lemma rdg_eqI[intro?]: 
  "\<lbrakk>rooted_digraph.truncate x = rooted_digraph.truncate y; AND_REL ext x y\<rbrakk> 
  \<Longrightarrow> rdg_eq ext x y"
  apply (intro dg_eqI)
  unfolding rdg_eq_ext_def[abs_def] AND_REL_def 
  by (auto simp add: rooted_digraph.defs digraph.defs)

lemma rdg_eq_fields[simp]:
  assumes "rdg_eq E x y"
  shows "rdg_V0 x = rdg_V0 y"
  using assms 
  unfolding rdg_eq_ext_def[abs_def] AND_REL_def
  by (simp add: rooted_digraph.defs)


lemma rdg_eq_loc[simp]: 
  assumes "rdg_eq E x y" 
  shows "rooted_digraph x \<longleftrightarrow> rooted_digraph y"
  and "finite_reachable_digraph x \<longleftrightarrow> finite_reachable_digraph y"
  and "finite_rooted_digraph x \<longleftrightarrow> finite_rooted_digraph y"
  using assms apply -
  by (simp_all add: rooted_digraph_def 
    finite_rooted_digraph_def finite_reachable_digraph_def
    rooted_digraph_axioms_def
    finite_reachable_digraph_axioms_def
    finite_rooted_digraph_axioms_def)
  

lemma rdg_eq_is_run[simp]: "\<lbrakk> rdg_eq E x y; rooted_digraph y\<rbrakk>
  \<Longrightarrow> rooted_digraph.is_run x = rooted_digraph.is_run y"
  apply (rule ext)
  apply (simp add: rooted_digraph.is_run_def)
  done

definition "gbg_eq_ext x y 
  \<equiv> generalized_buchi_graph.truncate x = generalized_buchi_graph.truncate y"
abbreviation "gbg_eq ext \<equiv> rdg_eq (gbg_eq_ext#ext)"

lemma gbg_eqI[intro?]: 
  "\<lbrakk>generalized_buchi_graph.truncate x = generalized_buchi_graph.truncate y; 
    AND_REL ext x y\<rbrakk> 
  \<Longrightarrow> gbg_eq ext x y"
  apply (intro rdg_eqI)
  unfolding gbg_eq_ext_def[abs_def] AND_REL_def 
  by (auto simp add: generalized_buchi_graph.defs 
    rooted_digraph.defs digraph.defs)

lemma gbg_eq_fields[simp]:
  assumes "gbg_eq E x y"
  shows "gbg_F x = gbg_F y"
  using assms 
  unfolding gbg_eq_ext_def[abs_def] AND_REL_def
  by (simp add: generalized_buchi_graph.defs)

lemma gbg_eq_loc[simp]: 
  assumes "gbg_eq E x y" 
  shows "generalized_inf_buchi_graph x \<longleftrightarrow> generalized_inf_buchi_graph y"
  and "generalized_fr_buchi_graph x \<longleftrightarrow> generalized_fr_buchi_graph y"
  and "generalized_buchi_graph x \<longleftrightarrow> generalized_buchi_graph y"
  using assms apply -
  by (simp_all add: 
    generalized_inf_buchi_graph_def generalized_inf_buchi_graph_axioms_def
    generalized_fr_buchi_graph_def
    generalized_buchi_graph_def)

definition "gba_eq_ext x y 
  \<equiv> generalized_buchi_automaton.truncate x 
  = generalized_buchi_automaton.truncate y"
abbreviation "gba_eq ext \<equiv> gbg_eq (gba_eq_ext#ext)"

lemma gba_eqI[intro?]: 
  "\<lbrakk>generalized_buchi_automaton.truncate x 
    = generalized_buchi_automaton.truncate y; 
    AND_REL ext x y\<rbrakk> 
  \<Longrightarrow> gba_eq ext x y"
  apply (intro gbg_eqI)
  unfolding gba_eq_ext_def[abs_def] AND_REL_def 
  by (auto simp add: generalized_buchi_automaton.defs
    generalized_buchi_graph.defs)

lemma gba_eq_fields[simp]:
  assumes "gba_eq E x y"
  shows "gba_\<Sigma> x = gba_\<Sigma> y"
  and "gba_L x = gba_L y"
  using assms 
  unfolding gba_eq_ext_def[abs_def] AND_REL_def
  by (simp_all add: generalized_buchi_automaton.defs)

lemma gba_eq_loc[simp]: 
  assumes "gba_eq E x y" 
  shows "generalized_inf_buchi_automaton x \<longleftrightarrow> generalized_inf_buchi_automaton y"
  and "generalized_fr_buchi_automaton x \<longleftrightarrow> generalized_fr_buchi_automaton y"
  and "generalized_buchi_automaton x \<longleftrightarrow> generalized_buchi_automaton y"
  using assms apply -
  by (simp_all add: 
    generalized_inf_buchi_automaton_def 
    generalized_inf_buchi_automaton_axioms_def
    generalized_fr_buchi_automaton_def
    generalized_buchi_automaton_def)

section {* Conversion to Indexed Acceptance Set *}
definition F_to_idx :: "'Q set set \<Rightarrow> (nat \<times> ('Q \<Rightarrow> nat set)) nres" where
  "F_to_idx F \<equiv> do {
    Flist \<leftarrow> SPEC (\<lambda>Flist. distinct Flist \<and> set Flist = F);
    let num_acc = length Flist;
    let acc = (\<lambda>v. {i . i<num_acc \<and> v\<in>Flist!i});
    RETURN (num_acc,acc)
  }"

lemma F_to_idx_correct:
  shows "F_to_idx F \<le> SPEC (\<lambda>(num_acc,acc). F = { {q. i\<in>acc q} | i. i<num_acc } 
    \<and> \<Union>range acc \<subseteq> {0..<num_acc})"
  unfolding F_to_idx_def
  apply (refine_rcg refine_vcg)
  apply (clarsimp dest!: sym[where t=F])
  apply (intro equalityI subsetI)
  apply (auto simp: in_set_conv_nth) [2]

  apply auto []
  done

definition "mk_acc_impl Flist \<equiv> do {
  let acc = Map.empty;

  (_,acc) \<leftarrow> nfoldli Flist (\<lambda>_. True) (\<lambda>A (i,acc). do {
    acc \<leftarrow> FOREACHi (\<lambda>it acc'. 
      acc' = (\<lambda>v. 
        if v\<in>A-it then 
          Some (insert i (the_default {} (acc v))) 
        else 
          acc v
      )
    ) 
      A (\<lambda>v acc. RETURN (acc(v\<mapsto>insert i (the_default {} (acc v))))) acc;
    RETURN (Suc i,acc)
  }) (0,acc);
  RETURN (\<lambda>x. the_default {} (acc x))
}"

lemma mk_acc_impl_correct: 
  assumes F: "(Flist',Flist)\<in>Id"
  assumes FIN: "\<forall>A\<in>set Flist. finite A"
  shows 
    "mk_acc_impl Flist' \<le> \<Down>Id (RETURN (\<lambda>v. {i . i<length Flist \<and> v\<in>Flist!i}))"
  using F apply simp
  unfolding mk_acc_impl_def

  apply (refine_rcg 
    nfoldli_rule[where 
      I="\<lambda>l1 l2 (i,res). i=length l1 
        \<and> the_default {} o res = (\<lambda>v. {j . j<i \<and> v\<in>Flist!j})"
    ]
    refine_vcg 
  )
  using FIN apply (simp_all)
  apply (rule ext) apply auto []

  apply (rule ext) apply (auto split: split_if_asm simp: nth_append nth_Cons') []
  apply (rule ext) apply (auto split: split_if_asm simp: nth_append nth_Cons' 
    fun_comp_eq_conv) []

  apply (rule ext) apply (auto simp: fun_comp_eq_conv) []
  done

definition F_to_idx_impl :: "'Q set set \<Rightarrow> (nat \<times> ('Q \<Rightarrow> nat set)) nres" where
  "F_to_idx_impl F \<equiv> do {
    Flist \<leftarrow> SPEC (\<lambda>Flist. distinct Flist \<and> set Flist = F);
    let num_acc = length Flist;
    acc \<leftarrow> mk_acc_impl Flist;
    RETURN (num_acc,acc)
  }"

lemma F_to_idx_refine: 
  assumes FIN: "\<forall>A\<in>F. finite A"
  shows "F_to_idx_impl F \<le> \<Down>Id (F_to_idx F)"
  using assms
  unfolding F_to_idx_impl_def F_to_idx_def

  apply (refine_rcg bind_Let_refine2[OF mk_acc_impl_correct])

  apply auto
  done

lemma F_to_idx_impl_correct:
  assumes FIN: "\<forall>A\<in>F. finite A"
  shows "F_to_idx_impl F 
  \<le> SPEC (\<lambda>(num_acc,acc). F = { {q. i\<in>acc q} | i. i<num_acc } 
      \<and> \<Union>range acc \<subseteq> {0..<num_acc})"
proof -
  note F_to_idx_refine[OF FIN]
  also note F_to_idx_correct
  finally show ?thesis by simp
qed

context generalized_buchi_graph
begin

  (* TODO: Move *)
  lemma finite_F_elem[simp, intro]: "A\<in>F \<Longrightarrow> finite A"
    using wf_F finite_V 
    by (metis Pow_iff finite_V infinite_super set_rev_mp)

  definition to_idx_gbg_ext where
    "to_idx_gbg_ext m \<equiv> do {
      (num_acc,acc) \<leftarrow> F_to_idx_impl F; 
      RETURN \<lparr> 
        dg_V = V, 
        dg_E = E, 
        rdg_V0 = V0,
        igbg_num_acc = num_acc, 
        igbg_acc = acc,
        \<dots> = m \<rparr>
    }"


  lemma to_idx_gbg_fields: 
    "to_idx_gbg_ext m \<le> SPEC (\<lambda>Gi. 
      dg_V Gi = V 
    \<and> dg_E Gi = E
    \<and> rdg_V0 Gi = V0
    \<and> F = { {q. i\<in>igbg_acc Gi q} | i. i<igbg_num_acc Gi } 
    \<and> \<Union>range (igbg_acc Gi) \<subseteq> {0..<igbg_num_acc Gi}
    \<and> idx_generalized_buchi_graph.more Gi = m
    )"
    unfolding to_idx_gbg_ext_def
    apply (refine_rcg refine_vcg order_trans[OF F_to_idx_impl_correct])
    by auto

  lemma to_idx_gbg_loc:
    "to_idx_gbg_ext (m::'morei) \<le> SPEC idx_generalized_buchi_graph"
    apply (rule order_trans[OF to_idx_gbg_fields])
    apply (refine_rcg refine_vcg)
    apply unfold_locales
    apply (clarsimp_all simp: wf_E wf_V0)
  proof -
    fix Gi :: "('Q, 'morei) idx_generalized_buchi_graph_scheme" and q
    assume A: "igbg_acc Gi q \<noteq> {}" 
      and [simp]: "F = {{q. i \<in> igbg_acc Gi q} |i. i < igbg_num_acc Gi}"
      and BOUND: "(\<Union>q. igbg_acc Gi q) \<subseteq> {0..<igbg_num_acc Gi}"

    from A obtain i where "i \<in> igbg_acc Gi q" by auto
    moreover with BOUND have "i<igbg_num_acc Gi" by fastforce
    ultimately show "q\<in>Q" using wf_F by auto
  qed

  lemmas to_idx_gbg_correct = 
    SPEC_rule_conjI[OF to_idx_gbg_fields to_idx_gbg_loc, 
      simplified conj_assoc]


  lemma to_idx_gbg_eq:
    assumes A: "AND_REL ext (G\<lparr>generalized_buchi_graph.more:=m\<rparr>) G"
    shows "to_idx_gbg_ext mi \<le> SPEC (\<lambda>Gi. 
      gbg_eq ext (idx_generalized_inf_buchi_graph.to_gbg_ext Gi m) G
    )"
    apply (rule order_trans[OF to_idx_gbg_correct])
    apply clarsimp
    apply (rule gbg_eqI)
  proof -
    case (goal1 Gi)
    then interpret Gi!: idx_generalized_buchi_graph Gi by simp
    
    have [simp]: "Gi.to_gbg_ext m = G\<lparr>generalized_buchi_graph.more:=m\<rparr>"
      apply (simp add: goal1 Gi.F_def Gi.accn_def)
      done

    show "generalized_buchi_graph.truncate 
      (Gi.to_gbg_ext m) 
      = generalized_buchi_graph.truncate G"
      by (simp add: generalized_buchi_graph.defs)
    
    show "AND_REL ext (idx_generalized_inf_buchi_graph.to_gbg_ext Gi m) G"
      using A by simp
  qed
    
  lemma 
    assumes A: "to_idx_gbg_ext m \<le> SPEC einv"
    shows "(to_idx_gbg_ext m,RETURN G) \<in> \<langle>igbg_rel e\<alpha> einv\<rangle>nres_rel"
    apply (rule nres_relI)
    apply (rule order_trans[OF SPEC_rule_conjI[OF to_idx_gbg_correct A]])
    apply clarsimp
xxx, ctd here \<dots>



end

context generalized_buchi_automaton
begin
  abbreviation to_idx_gba_ext where
    "to_idx_gba_ext m \<equiv> to_idx_gbg_ext \<lparr> igba_\<Sigma> = \<Sigma>, igba_L = L, \<dots>=m \<rparr>"
  
  lemma to_idx_gba_fields_ext: 
    "to_idx_gba_ext m \<le> SPEC (\<lambda>Gi. 
      igba_\<Sigma> Gi = \<Sigma>
    \<and> igba_L Gi = L
    \<and> idx_generalized_buchi_automaton.more Gi = m)"
    apply (rule order_trans[OF to_idx_gbg_fields])
    apply (clarsimp)
    apply (rename_tac Gi)
    apply (case_tac Gi)
    apply simp
    done

  lemmas to_idx_gba_fields = SPEC_rule_conjI[OF 
    to_idx_gbg_fields to_idx_gba_fields_ext, simplified conj_assoc]

  lemma to_idx_gba_loc: 
    "to_idx_gba_ext m' \<le> SPEC idx_generalized_buchi_automaton"
    apply (rule order_trans[OF SPEC_rule_conjI[
      OF to_idx_gbg_correct to_idx_gba_fields_ext]])
    apply clarsimp
  proof -
    case (goal1 Gi) 
    then interpret Gi!: idx_generalized_buchi_graph Gi by simp
    show "idx_generalized_buchi_automaton Gi"
      apply unfold_locales
      using wf_L goal1 by simp
  qed

  lemmas to_idx_gba_correct = 
    SPEC_rule_conjI[OF to_idx_gba_fields to_idx_gba_loc, 
      simplified conj_assoc]

  lemma to_idx_gba_eq:
    assumes A: "AND_REL ext (G\<lparr>generalized_buchi_automaton.more:=m\<rparr>) G"
    shows "to_idx_gba_ext mi \<le> SPEC (\<lambda>Gi. 
      gba_eq ext (idx_generalized_inf_buchi_automaton.to_gba_ext Gi m) G
    )"
    apply (rule order_trans[OF SPEC_rule_conjI[OF to_idx_gba_correct
      to_idx_gbg_eq[where 
        ext="gba_eq_ext#ext" and m = "\<lparr> gba_\<Sigma> = \<Sigma>, gba_L = L, \<dots> = m \<rparr>" ]]])
    apply clarsimp_all
    using A
    apply (cases G)
    apply (simp add: AND_REL_def gba_eq_ext_def 
      generalized_buchi_automaton.defs) []
    done

end

section {* Degeneralization *}
context idx_generalized_inf_buchi_graph
begin

  definition 
    degeneralize_gbg_ext :: "_ \<Rightarrow> ('Q \<times> nat, _) generalized_buchi_graph_scheme"
    where
    "degeneralize_gbg_ext m \<equiv> 
      if num_acc = 0 then \<lparr>
        dg_V = V \<times> {0},
        dg_E = {((q,0),(q',0)) | q q'. (q,q')\<in>E}, 
        rdg_V0 = V0\<times>{0}, 
        gbg_F = {V \<times> {0}},
        \<dots> = m
      \<rparr>
      else \<lparr>
        dg_V = V \<times> {0..<num_acc},
        dg_E = { ((q,i),(q',i')) | i i' q q'. 
          i<num_acc 
        \<and> (q,q')\<in>E 
        \<and> i' = (if i\<in>acc q then (i+1) mod num_acc else i) },
        rdg_V0 = V0 \<times> {0},
        gbg_F = {{(q,0)| q. 0\<in>acc q}},
        \<dots> = m
      \<rparr>"

  lemma degen_is_inf_buchi_graph: "inf_buchi_graph (degeneralize_gbg_ext m)"
    apply unfold_locales
    using wf_E apply (auto simp: degeneralize_gbg_ext_def) []
    using wf_V0 apply (auto simp: degeneralize_gbg_ext_def) []
    using wf_acc apply (auto simp: degeneralize_gbg_ext_def) []
    apply (auto simp: degeneralize_gbg_ext_def) []
    done

  sublocale degen!: inf_buchi_graph "degeneralize_gbg_ext m"
    by (rule degen_is_inf_buchi_graph)

  lemma degen_fields_uc:
    "degen.V0 T m = V0 \<times> {0}"
    "generalized_buchi_graph.more (degeneralize_gbg_ext m) = m"
    unfolding degeneralize_gbg_ext_def
    by simp_all

  lemma degen_fields_0:
    assumes [simp]: "num_acc = 0"
    shows 
    "degen.V T m = V \<times> {0}"
    "degen.E T m = {((q,0),(q',0)) | q q'. (q,q')\<in>E}"
    "degen.F T m = {V \<times> {0}}"
    "degen.A T m = V \<times> {0}"
    unfolding degen.A_def
    unfolding degeneralize_gbg_ext_def
    by simp_all

  lemma degen_fields_n0:
    assumes [simp]: "num_acc \<noteq> 0"
    shows 
    "degen.V T m = V \<times> {0..<num_acc}"
    "degen.E T m = { ((q,i),(q',i')) | i i' q q'. 
          i<num_acc \<and> (q,q')\<in>E \<and> i' = (if i\<in>acc q then (i+1) mod num_acc else i) }"
    "degen.F T m = {{(q,0)| q. 0\<in>acc q}}"
    "degen.A T m = {(q,0)| q. 0\<in>acc q}"
    unfolding degen.A_def
    unfolding degeneralize_gbg_ext_def
    by simp_all

  lemmas degen_fields = degen_fields_uc degen_fields_0 degen_fields_n0

  lemma degen_ipath_sound: "ipath (degen.E T m) r \<Longrightarrow> ipath E (fst o r)"
    apply (cases "num_acc = 0")
    apply (simp_all add: degen_fields)

    unfolding ipath_def
    apply auto []
    apply (metis fst_conv)
    apply auto []
    apply (metis fst_conv)
    done
   
  lemma degen_is_run_sound: "degen.is_run T m r \<Longrightarrow> is_run (fst o r)"
    unfolding degen.is_run_def is_run_def
    apply (auto simp: degen_ipath_sound degen_fields)
    done

  lemma degen_path_sound: 
    assumes "path (degen.E T m) u p v" 
    shows "path E (fst u) (map fst p) (fst v)"
    using assms
    apply induction
    apply simp
    apply (cases "num_acc = 0")
    apply (auto simp: degen_fields path_simps)
    done

  lemma degen_visit_acc:
    assumes "path (degen.E T m) (q,n) p (q',n')"
    assumes "n\<noteq>n'"
    shows "\<exists>qa. (qa,n)\<in>set p \<and> n\<in>acc qa"
    using assms
  proof (induction _ "(q,n)" p "(q',n')" arbitrary: q rule: path.induct)
    case (path_prepend qnh p)
    then obtain qh nh where [simp]: "qnh=(qh,nh)" by (cases qnh)
    from `((q,n),qnh) \<in> degen.E T m` 
    have "nh=n \<or> (nh=(n+1) mod num_acc \<and> n\<in>acc q)"
      by (auto simp: degeneralize_gbg_ext_def split: split_if_asm)
    thus ?case proof
      assume [simp]: "nh=n"
      from path_prepend obtain qa where "(qa, n) \<in> set p" and "n \<in> acc qa" by auto
      thus ?case by auto
    next
      assume "(nh=(n+1) mod num_acc \<and> n\<in>acc q)" thus ?case by auto
    qed
  qed simp

  lemma degen_run_complete0:
    assumes [simp]: "num_acc = 0"
    assumes R: "is_run r"
    shows "degen.is_run T m (\<lambda>i. (r i,0))"
    using R
    unfolding degen.is_run_def is_run_def ipath_def
    by (auto simp: degen_fields)

  lemma degen_acc_run_complete0:
    assumes [simp]: "num_acc = 0"
    assumes R: "is_acc_run r"
    shows "degen.is_acc_run T m (\<lambda>i. (r i,0))"
    using R
    unfolding degen.is_acc_run_def is_acc_run_def degen.is_acc_def is_acc_def
    apply (simp add: degen_run_complete0 degen_fields)
    apply (drule wf_run)
    apply (auto simp: INFM_nat_le)
    done

  lemma degen_run_complete_n0:
    assumes [simp]: "num_acc \<noteq> 0"
    assumes R: "is_run r"
    shows "\<exists>r'. degen.is_run T m r' \<and> r = fst o r'"
    using R
    unfolding degen.is_run_def is_run_def ipath_def
    apply (elim conjE)
  proof -
    assume R0: "r 0 \<in> V0" and RS: "\<forall>i. (r i, r (Suc i)) \<in> E"

    def r'\<equiv>"nat_rec 
      (r 0,0) 
      (\<lambda>i (q,n). (r (Suc i), if n \<in> acc q then (n+1) mod num_acc else n))"

    have [simp]:
      "r' 0 = (r 0,0)"
      "\<And>i. r' (Suc i) = (
        let 
          (q,n)=r' i 
        in 
          (r (Suc i), if n \<in> acc q then (n+1) mod num_acc else n)
      )"
      unfolding r'_def
      by auto

    have R0': "r' 0 \<in> degen.V0 T m" using R0
      by (simp add: degen_fields)

    have MAP: "r = fst o r'" proof (rule ext)
      fix i
      show "r i = (fst o r') i"
        by (cases i) (auto simp: split: prod.split)
    qed

    have [simp]: "0<num_acc" by (cases num_acc) auto

    have SND_LESS: "\<And>i. snd (r' i) < num_acc"
    proof -
      fix i show "snd (r' i) < num_acc" by (induction i) (auto split: prod.split) 
    qed

    have RS': "\<forall>i. (r' i, r' (Suc i)) \<in> degen.E T m"
    proof
      fix i
      obtain n where [simp]: "r' i = (r i,n)"
        apply (cases i)
        apply (force)
        apply (force split: prod.split)
        done
      from SND_LESS[of i] have [simp]: "n<num_acc" by simp

      show "(r' i, r' (Suc i)) \<in> degen.E T m" using RS
        by (auto simp: degeneralize_gbg_ext_def)
    qed

    from R0' RS' MAP show 
      "\<exists>r'. (r' 0 \<in> degen.V0 T m 
      \<and> (\<forall>i. (r' i, r' (Suc i)) \<in> degen.E T m)) 
      \<and> r = fst \<circ> r'" by blast
  qed

  lemma degen_run_bound:
    assumes [simp]: "num_acc \<noteq> 0"
    assumes R: "degen.is_run T m r"
    shows "snd (r i) < num_acc"
    apply (induction i)
    using R unfolding degen.is_run_def ipath_def
    apply -
    apply (auto simp: degen_fields) []
    apply (clarsimp simp: degen_fields)
    by (metis snd_conv)
  
  lemma degen_acc_run_complete_aux1: 
    assumes NN0[simp]: "num_acc \<noteq> 0"
    assumes R: "degen.is_run T m r"
    assumes EXJ: "\<exists>j\<ge>i. n \<in> acc (fst (r j))"
    assumes RI: "r i = (q,n)"
    shows "\<exists>j\<ge>i. \<exists>q'. r j = (q',n) \<and> n \<in> acc q'"
  proof -
    def j \<equiv> "LEAST j. j\<ge>i \<and> n \<in> acc (fst (r j))"

    from RI have "n<num_acc" using degen_run_bound[OF NN0 R, of i] by auto
    (*with ACC have "n \<in> \<Union>(acc ` limit (fst o r))" by auto
    hence EXJ: "\<exists>j\<ge>i. n \<in> acc (fst (r j))" 
      unfolding limit_def INFM_nat_le by auto*)
    from EXJ have "j\<ge>i" "n \<in> acc (fst (r j))" "\<forall>k\<ge>i. n \<in> acc (fst (r k)) \<longrightarrow> j\<le>k"
      using LeastI_ex[OF EXJ]
      unfolding j_def
      apply (auto) [2]
      apply (metis (lifting) Least_le)
      done
    hence "\<forall>k\<ge>i. k<j \<longrightarrow> n \<notin> acc (fst (r k))" by auto

    have "\<forall>k. k\<ge>i \<and> k\<le>j \<longrightarrow> (snd (r k) = n)"
    proof (clarify)
      fix k
      assume "i\<le>k" "k\<le>j"
      thus "snd (r k) = n"
      proof (induction k rule: less_induct)
        case (less k)
        show ?case proof (cases "k=i")
          case True thus ?thesis using RI by simp
        next
          case False with less.prems have "k - 1 < k" "i \<le> k - 1" "k - 1\<le>j" by auto
          from less.IH[OF this] have "snd (r (k - 1)) = n" .
          moreover from R have 
            "(r (k - 1), r k) \<in> degen.E T m"
            unfolding degen.is_run_def ipath_def
            by clarsimp (metis One_nat_def Suc_diff_1 `k - 1 < k` 
              less_nat_zero_code neq0_conv)
          moreover have "n \<notin> acc (fst (r (k - 1)))"
            by (metis 
              `\<forall>k\<ge>i. k < j \<longrightarrow> n \<notin> local.acc (fst (r k))` 
              `i \<le> k - 1` `k - 1 < k` `k - 1 \<le> j`
              le_antisym less.prems(2) min.semilattice_strict_iff_order)
          ultimately show ?thesis
            by (auto simp: degen_fields)
        qed
      qed
    qed

    thus ?thesis 
      by (metis `i \<le> j` `n \<in> local.acc (fst (r j))` order_refl surjective_pairing)
  qed
      
  lemma degen_acc_run_complete_aux1': 
    assumes NN0[simp]: "num_acc \<noteq> 0"
    assumes R: "degen.is_run T m r"
    assumes ACC: "\<forall>n<num_acc. \<exists>\<^sub>\<infinity>i. n \<in> acc (fst (r i))"
    assumes RI: "r i = (q,n)"
    shows "\<exists>j\<ge>i. \<exists>q'. r j = (q',n) \<and> n \<in> acc q'"
  proof -
    from RI have "n<num_acc" using degen_run_bound[OF NN0 R, of i] by auto
    with ACC have EXJ: "\<exists>j\<ge>i. n \<in> acc (fst (r j))" 
      unfolding INFM_nat_le by blast

    from degen_acc_run_complete_aux1[OF NN0 R EXJ RI] show ?thesis .
  qed

  lemma degen_acc_run_complete_aux2:
    assumes NN0[simp]: "num_acc \<noteq> 0"
    assumes R: "degen.is_run T m r"
    assumes ACC: "\<forall>n<num_acc. \<exists>\<^sub>\<infinity>i. n \<in> acc (fst (r i))"
    assumes RI: "r i = (q,n)" and OFS: "ofs<num_acc"
    shows "\<exists>j\<ge>i. \<exists>q'. 
      r j = (q',(n + ofs) mod num_acc) \<and> (n + ofs) mod num_acc \<in> acc q'"
    using RI OFS
  proof (induction ofs arbitrary: q n i)
    case 0 
    from degen_run_bound[OF NN0 R, of i] `r i = (q, n)` 
    have NLE: "n<num_acc" 
      by simp

    with degen_acc_run_complete_aux1'[OF NN0 R ACC `r i = (q, n)`] show ?case
      by auto
  next
    case (Suc ofs)
    from Suc.IH[OF Suc.prems(1)] Suc.prems(2)
    obtain j q' where "j\<ge>i" and RJ: "r j = (q',(n+ofs) mod num_acc)" 
      and A: "(n+ofs) mod num_acc \<in> acc q'"
      by auto
    from R have "(r j, r (Suc j)) \<in> degen.E T m" 
      by (auto simp: degen_fields degen.is_run_def ipath_def)
    with RJ A obtain q2 where RSJ: "r (Suc j) = (q2,(n+Suc ofs) mod num_acc)" 
      by (auto simp: degeneralize_gbg_ext_def mod_Suc_eq_Suc_mod[symmetric])

    have aux: "\<And>j'. i\<le>j \<Longrightarrow> Suc j \<le> j' \<Longrightarrow> i\<le>j'" by auto
    from degen_acc_run_complete_aux1'[OF NN0 R ACC RSJ] `j\<ge>i` 
    show ?case 
      by (auto dest: aux)
  qed

  lemma degen_acc_run_complete:
    assumes AR: "is_acc_run r"
    obtains r' 
    where "degen.is_acc_run T m r'" and "r = fst o r'"
  proof (cases "num_acc = 0")
    case True 
    with AR degen_acc_run_complete0 
    show ?thesis 
      by (auto intro!: that[of "(\<lambda>i. (r i, 0))"])
  next
    case False
    assume NN0[simp]: "num_acc \<noteq> 0"

    from AR have R: "is_run r" and ACC: "\<forall>n<num_acc. \<exists>\<^sub>\<infinity>i. n \<in> acc (r i)"
      unfolding is_acc_run_def is_acc_def by auto

    from degen_run_complete_n0[OF NN0 R, of m] obtain r' where 
      R': "degen.is_run T m r'" 
      and [simp]: "r = fst \<circ> r'" 
      by blast

    from ACC have ACC': "\<forall>n<num_acc. \<exists>\<^sub>\<infinity>i. n \<in> acc (fst (r' i))" by simp

    have "\<forall>i. \<exists>j>i. r' j \<in> degen.A T m"
    proof
      fix i
      obtain q n where RI: "r' (Suc i) = (q,n)" by (cases "r' (Suc i)")
      have "(n + (num_acc - n mod num_acc)) mod num_acc = 0"
        by (metis NN0 R' `r' (Suc i) = (q, n)` add_diff_cancel_left' 
          degen_run_bound less_imp_add_positive mod_self nat_mod_eq' snd_conv)
      then obtain ofs where 
        OFS_LESS: "ofs<num_acc" 
        and [simp]: "(n + ofs) mod num_acc = 0"
        by (metis NN0 Nat.add_0_right diff_less neq0_conv)
      with degen_acc_run_complete_aux2[OF NN0 R' ACC' RI OFS_LESS]
      obtain j q' where 
        "j>i" "r' j = (q',0)" and "0\<in>acc q'" 
        by (auto simp: less_eq_Suc_le)
      thus "\<exists>j>i. r' j \<in> degen.A T m" 
        by (auto simp: degen_fields)
    qed
    hence "\<exists>\<^sub>\<infinity>i. r' i \<in> degen.A T m" 
      by (auto simp: INFM_nat)

    have "degen.is_acc_run T m r'"
      unfolding degen.is_acc_run_def degen.is_acc_def
      apply rule 
      apply fact
      apply simp
      by fact

    thus ?thesis by (auto intro: that)
  qed

  lemma degen_run_find_change:
    assumes NN0[simp]: "num_acc \<noteq> 0"
    assumes R: "degen.is_run T m r"
    assumes A: "i\<le>j" "r i = (q,n)" "r j = (q',n')" "n\<noteq>n'"
    obtains k qk where "i\<le>k" "k<j" "r k = (qk,n)" "n \<in> acc qk"
  proof -
    from degen_run_bound[OF NN0 R] A have "n<num_acc" "n'<num_acc"
      by (metis snd_conv)+

    def k \<equiv> "LEAST k. i<k \<and> snd (r k) \<noteq> n"

    have "i<k" "snd (r k) \<noteq> n"
      by (metis (lifting, mono_tags) LeastI_ex A k_def leD less_linear snd_conv)+

    from Least_le[where P="\<lambda>k. i<k \<and> snd (r k) \<noteq> n", folded k_def]
    have LEK_EQN: "\<forall>k'. i\<le>k' \<and> k'<k \<longrightarrow> snd (r k') = n"
      using `r i = (q,n)`
      by clarsimp (metis le_neq_implies_less not_le snd_conv)
    hence SND_RKMO: "snd (r (k - 1)) = n" using `i<k` by auto
    moreover from R have "(r (k - 1), r k) \<in> degen.E T m"
      unfolding degen.is_run_def ipath_def using `i<k`
      by clarsimp (metis Suc_pred gr_implies_not0 neq0_conv) 
    moreover note `snd (r k) \<noteq> n`
    ultimately have "n \<in> acc (fst (r (k - 1)))"
      by (auto simp: degeneralize_gbg_ext_def split: split_if_asm)
    moreover have "k - 1 < j" using A LEK_EQN 
      apply (rule_tac ccontr)
      apply clarsimp
      by (metis One_nat_def `snd (r (k - 1)) = n` less_Suc_eq 
        less_imp_diff_less not_less_eq snd_conv)
    ultimately show thesis
      apply -
      apply (rule that[of "k - 1" "fst (r (k - 1))"])
      using `i<k` SND_RKMO by auto
  qed


  lemma degen_run_find_acc_aux:
    assumes NN0[simp]: "num_acc \<noteq> 0"
    assumes AR: "degen.is_acc_run T m r"
    assumes A: "r i = (q,0)" "0 \<in> acc q" "n<num_acc"
    shows "\<exists>j qj. i\<le>j \<and> r j = (qj,n) \<and> n \<in> acc qj"
  proof -
    from AR have R: "degen.is_run T m r" 
      and ACC: "\<exists>\<^sub>\<infinity>i. r i \<in> degen.A T m"
      unfolding degen.is_acc_run_def degen.is_acc_def by auto
    from ACC have ACC': "\<forall>i. \<exists>j>i. r j \<in> degen.A T m"
      by (auto simp: INFM_nat)
    
    show ?thesis using `n<num_acc`
    proof (induction n)
      case 0 thus ?case using A by auto
    next
      case (Suc n)
      then obtain j qj where "i\<le>j" "r j = (qj,n)" "n\<in>acc qj" by auto
      moreover from R have "(r j, r (Suc j)) \<in> degen.E T m" 
        unfolding degen.is_run_def ipath_def
        by auto
      ultimately obtain qsj where RSJ: "r (Suc j) = (qsj,Suc n)"
        unfolding degeneralize_gbg_ext_def using `Suc n<num_acc` by auto
      
      from ACC' obtain k q0 where "Suc j \<le> k" "r k = (q0, 0)"
        apply (auto simp: degen_fields)
        by (metis less_imp_le_nat)
      from degen_run_find_change[OF NN0 R `Suc j \<le> k` RSJ `r k = (q0, 0)`] 
      obtain l ql where
        "Suc j \<le> l" "l < k" "r l = (ql, Suc n)" "Suc n \<in> acc ql" 
        by blast
      thus ?case using `i \<le> j`
        by (intro exI[where x=l] exI[where x=ql]) auto
    qed
  qed
      
  lemma degen_acc_run_sound:
    assumes A: "degen.is_acc_run T m r"
    shows "is_acc_run (fst o r)"
  proof -
    from A have R: "degen.is_run T m r" 
      and ACC: "\<exists>\<^sub>\<infinity>i. r i \<in> degen.A T m"
      unfolding degen.is_acc_run_def degen.is_acc_def by auto
    from degen_is_run_sound[OF R] have R': "is_run (fst o r)" .

    show ?thesis
    proof (cases "num_acc = 0")
      case False[simp] note NN0=this

      from ACC have ACC': "\<forall>i. \<exists>j>i. r j \<in> degen.A T m"
        by (auto simp: INFM_nat)

      have "\<forall>n<num_acc. \<forall>i. \<exists>j>i. n \<in> acc (fst (r j))" 
      proof (intro allI impI)
        fix n i

        obtain j qj where "j>i" and RJ: "r j = (qj,0)" and ACCJ: "0 \<in> acc (qj)"
          using ACC' by (fastforce simp: degen_fields)

        assume NLESS: "n<num_acc"
        show "\<exists>j>i. n \<in> acc (fst (r j))"
        proof (cases n)
          case 0 thus ?thesis using `j>i` RJ ACCJ by auto
        next
          case (Suc n')[simp]
          from degen_run_find_acc_aux[OF NN0 A RJ ACCJ NLESS] obtain k qk where
            "j\<le>k" "r k = (qk,n)" "n \<in> acc qk" by auto
          thus ?thesis
            by (metis `i < j` dual_order.strict_trans1 fst_conv)
        qed
      qed
      hence "\<forall>n<num_acc. \<exists>\<^sub>\<infinity>i. n \<in> acc (fst (r i))"
        by (auto simp: INFM_nat)
      with R' show ?thesis
        unfolding is_acc_run_def is_acc_def by auto
    next
      case True[simp]
      with R' show ?thesis
        unfolding is_acc_run_def is_acc_def
        by auto
    qed
  qed

end

context idx_generalized_fr_buchi_graph begin

  lemma degen_is_fr_buchi_graph: "fr_buchi_graph (degeneralize_gbg_ext (m::'a))"
  proof
    have "(degen.E TYPE('a) m)\<^sup>* `` degen.V0 TYPE('a) m \<subseteq> E\<^sup>*``V0 \<times> {0..num_acc}"
    proof -
      {
        fix q n q' n'
        assume "((q,n),(q',n'))\<in>(degen.E TYPE('a) m)\<^sup>*" 
          and 0: "(q,n)\<in>degen.V0 TYPE('a) m"
        hence G1: "(q,q')\<in>E\<^sup>* \<and> n'\<le>num_acc"
          apply (induction rule: rtrancl_induct2)
          apply (simp add: degen_fields)
          apply (cases "num_acc = 0")
          by (auto simp: degen_fields)
        
        from 0 have G2: "q\<in>V0 \<and> n\<le>num_acc"
          by (auto simp: degen_fields)
        note G1 G2
      } thus ?thesis by fastforce
    qed
    also have "finite \<dots>" by auto
    finally (finite_subset) show 
      "finite ((dg_E (degeneralize_gbg_ext m))\<^sup>* `` rdg_V0 (degeneralize_gbg_ext m))" 
      .
  qed

  sublocale degen!: fr_buchi_graph "degeneralize_gbg_ext m"
    by (rule degen_is_fr_buchi_graph)
end

context idx_generalized_buchi_graph begin

  lemma degen_is_buchi_graph: "buchi_graph (degeneralize_gbg_ext (m::'a))"
    apply unfold_locales
    apply (cases "num_acc = 0", simp_all add: degen_fields) []
    apply (simp add: degen_fields)
    done

  sublocale degen!: buchi_graph "degeneralize_gbg_ext m"
    by (rule degen_is_buchi_graph)
end


context idx_generalized_inf_buchi_automaton begin

term degeneralize_gbg_ext

  abbreviation degeneralize_gba_ext 
    :: "'a \<Rightarrow> ('Q \<times> nat, 'L, 'a) generalized_buchi_automaton_scheme"
    where
    "degeneralize_gba_ext m 
    \<equiv> degeneralize_gbg_ext \<lparr> gba_\<Sigma> = \<Sigma>, gba_L = L o fst, \<dots> = m \<rparr>"

  lemma degen_gba_fields:
    "gba_\<Sigma> (degeneralize_gba_ext m) = \<Sigma>"
    "gba_L (degeneralize_gba_ext m) = L o fst"
    unfolding degeneralize_gbg_ext_def
    by (auto split: split_if)

  lemma degen_is_inf_buchi_automaton: "inf_buchi_automaton (degeneralize_gba_ext m)"
    apply unfold_locales
    using wf_L
    apply (cases "num_acc = 0")
    apply (simp_all add: degen_gba_fields degen_fields)
    apply force
    apply force
    done

  sublocale degen!: inf_buchi_automaton "(degeneralize_gba_ext m)"
    by (rule degen_is_inf_buchi_automaton)

  lemma degen_accept[simp]: "degen.accept T m = accept"
    apply (rule ext)
    unfolding degen.accept_def accept_def
    apply (intro iffI)
    apply (auto
      simp: degen_gba_fields 
      dest: degen_acc_run_sound) []
    apply (fastforce simp: degen_gba_fields elim: degen_acc_run_complete)
    done

  lemma degen_lang[simp]: "degen.lang T m = lang"
    unfolding degen.lang_def lang_def
    by simp
end

context idx_generalized_fr_buchi_automaton begin
  lemma degen_is_fr_buchi_automaton: "fr_buchi_automaton (degeneralize_gba_ext m)"
    by unfold_locales

  sublocale degen!: fr_buchi_automaton "(degeneralize_gba_ext m)"
    by (rule degen_is_fr_buchi_automaton)
end  

context idx_generalized_buchi_automaton begin
  lemma degen_is_buchi_automaton: "buchi_automaton (degeneralize_gba_ext m)"
    by unfold_locales

  sublocale degen!: buchi_automaton "(degeneralize_gba_ext m)"
    by (rule degen_is_buchi_automaton)
end  



interpretation idx_generalized_buchi_automaton undefined sorry


;
xxx, ctd here: 

context igb_graph_loc
begin

  (*lemma degen_F_finite: 
    assumes NN0[simp]: "num_acc \<noteq> 0"
    shows "finite (bg_F (degeneralize_ext ecnv))"
    unfolding degeneralize_ext_def
    apply auto
    by (metis acc_fin finite_conv_nat_seg_image image_image)
  *)

\<rightarrow> with sublocale

  lemma degen_invar: "b_graph_loc (degeneralize_ext ecnv)"
  proof
    have "((frg_E (degeneralize_ext ecnv))\<^sup>* `` frg_V0 (degeneralize_ext ecnv))
      \<subseteq> E\<^sup>*``V0 \<times> {0..num_acc}"
    proof -
      {
        fix q n q' n'
        assume "((q,n),(q',n'))\<in>(frg_E (degeneralize_ext ecnv))\<^sup>*" 
          and 0: "(q,n)\<in>frg_V0 (degeneralize_ext ecnv)"
        hence G1: "(q,q')\<in>E\<^sup>* \<and> n'\<le>num_acc"
          apply (induction rule: rtrancl_induct2)
          by (auto simp: degeneralize_ext_def split: split_if_asm)
        
        from 0 have G2: "q\<in>V0 \<and> n\<le>num_acc"
          by (auto simp: degeneralize_ext_def split: split_if_asm)
        note G1 G2
      } thus ?thesis by fastforce
    qed
    also have "finite \<dots>" by auto
    finally (finite_subset) show "finite 
      ((frg_E (degeneralize_ext ecnv))\<^sup>* `` frg_V0 (degeneralize_ext ecnv))" .
  qed

\<rightarrow> degen.accept, degen.lang !


end













interpretation idx_generalized_buchi_automaton undefined sorry


record 'a cset = 
  cset_num :: nat
  cset_cls :: "'a \<Rightarrow> bitset"

definition cset_cc :: "'a cset \<Rightarrow> nat \<Rightarrow> 'a set" where
  "cset_cc s i \<equiv> { x . i \<in> bs_\<alpha> (cset_cls s x)}"
definition cset_\<alpha> :: "'a cset \<Rightarrow> 'a set set" 
  where "cset_\<alpha> s \<equiv> cset_cc s ` {0..<cset_num s}"

locale cset_invar =
  fixes cs :: "'a cset"
  assumes cls_in_range: "\<Union>range (bs_\<alpha> o cset_cls cs) \<subseteq> {0..<cset_num cs}"
  assumes cls_inj: "inj_on (cset_cc cs) {0..<cset_num cs}"
begin

  lemma cls_in_range': "bs_\<alpha> (cset_cls cs x) \<subseteq> {0..<cset_num cs}"
    using cls_in_range by auto
end

definition "cset_rel \<equiv> br cset_\<alpha> cset_invar"

type_synonym css = "bitset"

definition css_\<alpha> :: "'a cset \<Rightarrow> css \<Rightarrow> 'a set set" where
  "css_\<alpha> cs css \<equiv> cset_cc cs ` bs_\<alpha> css"
definition css_invar :: "'a cset \<Rightarrow> css \<Rightarrow> bool" where
  "css_invar cs css \<equiv> bs_\<alpha> css \<subseteq> {0..<cset_num cs}"

definition "css_rel cs \<equiv> br (css_\<alpha> cs) (css_invar cs)"

lemma css_empty_rel: "(bs_empty (),{}) \<in> css_rel cs"
  unfolding css_rel_def css_\<alpha>_def css_invar_def br_def
  by auto

lemma css_union_rel: "(bs_union, op \<union>) \<in> css_rel cs \<rightarrow> css_rel cs \<rightarrow> css_rel cs"
  unfolding css_rel_def css_\<alpha>_def css_invar_def br_def
  by fastforce 

lemma (in cset_invar) css_inter_rel: 
  "(bs_inter, op \<inter>) \<in> css_rel cs \<rightarrow> css_rel cs \<rightarrow> css_rel cs"
  unfolding css_rel_def css_\<alpha>_def css_invar_def br_def
  apply auto
  by (metis IntI cls_inj inj_on_image_Int rev_image_eqI)

lemma (in cset_invar) css_subset_rel: 
  "(bs_subset_eq, op \<subseteq>) \<in> css_rel cs \<rightarrow> css_rel cs \<rightarrow> bool_rel"
  unfolding css_rel_def css_\<alpha>_def css_invar_def br_def
  apply clarsimp
  apply safe
  apply force
  apply (smt cls_inj dual_order.trans inj_on_image_eq_iff set_rev_mp subset_image_iff)
  done

(* TODO: Move *)
lemma bs_\<alpha>_eq_simp[simp]: "bs_\<alpha> x = bs_\<alpha> y \<longleftrightarrow> x=y"
  by (metis bs_eq_correct bs_eq_def)

lemma (in cset_invar) css_equal_rel: 
  "(op =, op =) \<in> css_rel cs \<rightarrow> css_rel cs \<rightarrow> bool_rel"
  unfolding css_rel_def css_\<alpha>_def css_invar_def br_def
  using cls_inj apply (auto simp:  inj_on_image_eq_iff)
  done

definition [simp]: "op_css_construct cs x \<equiv> { s \<in> cs. x\<in>s }"

lemma css_construct: "\<lbrakk> (csi,cs)\<in>cset_rel \<rbrakk> \<Longrightarrow> 
  (cset_cls csi, op_css_construct cs) \<in> Id \<rightarrow> css_rel csi"
  unfolding cset_rel_def br_def
proof clarsimp
  fix x
  assume "cset_invar csi"
  then interpret cset_invar csi .
  
  show "(cset_cls csi x, {s \<in> cset_\<alpha> csi. x \<in> s}) \<in> css_rel csi"
    unfolding cset_\<alpha>_def css_rel_def css_\<alpha>_def css_invar_def br_def cset_cc_def[abs_def]
    using set_mp[OF cls_in_range']
    by auto
qed









term bs_set_rel


;

xxx, ctd here: See this as implementation of F. 
  Based on this, build igbg - implementation. Try it in two phases, first
  going to css, and then implementing css by bitvectors.
 * Perhaps, we should construct a stand-alone implementation of css with bitvectors.
   The autoref-setup for this could then live inside an gbg-impl-locale, that implements
   F by cset.

lemma css_eq_rel: "cset_invar cs \<Longrightarrow> 
  (op =, op =) \<in> css_rel cs \<rightarrow> css_rel cs \<rightarrow> bool_rel"
  unfolding css_rel_def css_\<alpha>_def css_invar_def br_def
  unfolding cset_invar_def
  apply auto

lemma css_inter_rel: "cset_invar cs 
  \<Longrightarrow> (op \<inter>, op \<inter>) \<in> css_rel cs \<rightarrow> css_rel cs \<rightarrow> css_rel cs"
  unfolding css_rel_def css_\<alpha>_def css_invar_def br_def cset_invar_def css_cc_def[abs_def]
  apply auto
  apply force




lemma "(op \<inter>, op \<inter>) \<in> css_rel cs \<rightarrow> css_rel cs \<rightarrow> css_rel cs"
  unfolding css_rel_def css_\<alpha>_def css_invar_def br_def
  by auto




;

Implementing a set of sets:


  'a set set is represented by:

    num :: nat
    idxs :: 'a \<Rightarrow> nat set

    invar: \<Union>range acc \<subseteq> {0..<num_acc}
    \<alpha> = { idxn i | i. i<num_acc}
      where idxn i = { x . i \<in> idxs x}

Operations:
  Subsets containing given element:
    ss_containing S x \<equiv> {s \<in> S. x \<in> s}

    ss_containing_impl (num,idxs) x = idxs x
    \<alpha>_res (num,idxs) r = { x . r \<inter> idxs x \<noteq> {} }



section {* Indexed Acceptance Classes *}
record 'Q indexed_buchi_graph = "'Q finite_rooted_digraph_rec" +
  ibg_num_acc :: nat
  ibg_acc :: "'Q \<Rightarrow> nat set"







;
xxx, ctd here: nn-combinations

  


;
xxx, ctd here:
  Implement constraints finite, no_nodes, finite_reachable, degen








end
definition "generalized_inf_buchi_graph_pred_e ext G
  \<equiv> generalized_inf_buchi_graph G \<and> ext G"
abbreviation "generalized_inf_buchi_graph_pred ext 
  \<equiv> rooted_digraph_pred (generalized_inf_buchi_graph_pred_e ext)"

lemma generalized_inf_buchi_graphI: 
  "generalized_inf_buchi_graph_pred ext G \<Longrightarrow> generalized_inf_buchi_graph G"
  by (simp add: digraph_pred_def rooted_digraph_pred_e_def 
    generalized_inf_buchi_graph_pred_e_def)

lemma (in generalized_inf_buchi_graph) generalized_inf_buchi_graph_pred: 
  "generalized_inf_buchi_graph_pred top G"
  unfolding digraph_pred_def rooted_digraph_pred_e_def 
    generalized_inf_buchi_graph_pred_e_def
  apply safe
  apply unfold_locales
  done

definition "generalized_fr_buchi_graph_pred_e ext G
  \<equiv> generalized_fr_buchi_graph G \<and> ext G"
abbreviation "generalized_fr_buchi_graph_pred ext 
  \<equiv> generalized_inf_buchi_graph_pred (generalized_fr_buchi_graph_pred_e ext)"

lemma generalized_fr_buchi_graphI: 
  "generalized_fr_buchi_graph_pred ext G \<Longrightarrow> generalized_fr_buchi_graph G"
  by (simp add: digraph_pred_def rooted_digraph_pred_e_def 
    generalized_inf_buchi_graph_pred_e_def generalized_fr_buchi_graph_pred_e_def)

lemma (in generalized_fr_buchi_graph) generalized_fr_buchi_graph_pred: 
  "generalized_fr_buchi_graph_pred top G"
  unfolding digraph_pred_def rooted_digraph_pred_e_def 
    generalized_inf_buchi_graph_pred_e_def
    generalized_fr_buchi_graph_pred_e_def
  apply safe
  apply unfold_locales
  done

definition "generalized_buchi_graph_pred_e ext G
  \<equiv> generalized_buchi_graph G \<and> ext G"
abbreviation "generalized_buchi_graph_pred ext 
  \<equiv> generalized_fr_buchi_graph_pred (generalized_buchi_graph_pred_e ext)"

lemma generalized_buchi_graphI: 
  "generalized_buchi_graph_pred ext G \<Longrightarrow> generalized_buchi_graph G"
  by (simp add: digraph_pred_def rooted_digraph_pred_e_def 
    generalized_inf_buchi_graph_pred_e_def 
    generalized_fr_buchi_graph_pred_e_def
    generalized_buchi_graph_pred_e_def)

lemma (in generalized_buchi_graph) generalized_buchi_graph_pred: 
  "generalized_buchi_graph_pred top G"
  unfolding digraph_pred_def rooted_digraph_pred_e_def 
    generalized_inf_buchi_graph_pred_e_def
    generalized_fr_buchi_graph_pred_e_def
    generalized_buchi_graph_pred_e_def
  apply safe
  apply unfold_locales
  done
