theory BoolProgs
imports 
  "../../../CAVA_Setup/CAVA_Refine"
  "../../../CAVA_Setup/CAVA_Code_Target"
begin


section {* Syntax *}

text {* Boolean Expressions *}
type_synonym vname = nat
datatype bexp = 
  TT 
  | FF 
  | V vname
  | Not bexp 
  | And bexp bexp 
  | Or bexp bexp 

text {* Boolean Programs *}
datatype
  com = 
    SKIP                            -- {* Skip *}
  | Assign "vname list" "bexp list" -- {* Parallel Assignment *}  
  | Seq com com                     -- {* Sequential Composition *}
  | GC "(bexp * com)list"           -- {* Guarded Nondeterministic Choice *}
  | IfTE bexp com com               -- {* If-Then-Else *}
  | While bexp com                  -- {* While loop*}

locale BoolProg_Syntax begin
  notation 
        Assign       ("_ ::= _" [999, 61] 61)
    and Seq          ("_;/ _"  [60, 61] 60)
    and GC           ("IF _ FI")
    and IfTE         ("(IF _/ THEN _/ ELSE _)"  [0, 61, 61] 61)
    and While        ("(WHILE _/ DO _)"  [0, 61] 61)
end

section {* Semantics *}

type_synonym state = "vname \<Rightarrow> bool"

primrec bval :: "bexp \<Rightarrow> state \<Rightarrow> bool" where
  "bval TT s = True" |
  "bval FF s = False" |
  "bval (V n) s = s n" |
  "bval (Not b) s = (\<not> bval b s)" |
  "bval (And b\<^sub>1 b\<^sub>2) s = (bval b\<^sub>1 s & bval b\<^sub>2 s)" |
  "bval (Or b\<^sub>1 b\<^sub>2) s = (bval b\<^sub>1 s | bval b\<^sub>2 s)"

context begin interpretation BoolProg_Syntax .

(*
 * Semantics Notice:
 * To be equivalent in semantics with SPIN, there is no such thing as a finite run:
 *   - Deadlocks (i.e. empty Choice) are self-loops
 *   - program termination is self-loop
 *)
inductive small_step :: "com \<times> state \<Rightarrow> com * state \<Rightarrow> bool" where
  skip: "small_step (SKIP,s) (SKIP,s)"
| assign: "\<lbrakk> 
    length xs = length es; 
    bvs = map (\<lambda>b. bval b s) es;
    s' = foldl (\<lambda>s (n,bv). s(n := bv)) s (zip xs bvs)
  \<rbrakk> \<Longrightarrow> small_step (xs ::= es,s) (SKIP, s')"
| seq1: "\<lbrakk> small_step (c1,s) (c1',s'); c1'\<noteq>SKIP \<rbrakk> 
  \<Longrightarrow> small_step (c1;c2,s) (c1';c2,s')"
| seq2: "\<lbrakk> small_step (c1,s) (SKIP,s') \<rbrakk> 
  \<Longrightarrow> small_step (c1;c2,s) (c2,s')"
| gc: "\<lbrakk> (b,c)\<in>set cs; bval b s \<rbrakk> \<Longrightarrow> small_step (IF cs FI, s) (c,s)"
| gc_deadlock: "\<lbrakk> \<forall>(b,c)\<in>set cs. \<not>bval b s \<rbrakk> 
  \<Longrightarrow> small_step (IF cs FI, s) (IF cs FI,s)"
| if_true: "\<lbrakk> bval b s \<rbrakk> \<Longrightarrow> small_step (IF b THEN c1 ELSE c2,s) (c1,s)"
| if_false: "\<lbrakk> \<not>bval b s \<rbrakk> \<Longrightarrow> small_step (IF b THEN c1 ELSE c2,s) (c2,s)"
| while_true: "\<lbrakk> bval b s \<rbrakk> 
  \<Longrightarrow> small_step (WHILE b DO c, s) (c; WHILE b DO c, s)"
| while_false: "\<lbrakk> \<not> bval b s \<rbrakk> 
  \<Longrightarrow> small_step (WHILE b DO c, s) (SKIP, s)"

lemmas small_step_induct[induct pred] = small_step.induct[split_format(complete)]
declare small_step.intros[simp,intro]

fun well_formed :: "com \<Rightarrow> bool" where
  "well_formed SKIP \<longleftrightarrow> True"
| "well_formed (xs ::= es) \<longleftrightarrow> length xs = length es"
| "well_formed (c1;c2) \<longleftrightarrow> well_formed c1 \<and> well_formed c2"
| "well_formed (IF cs FI) \<longleftrightarrow> (\<forall>(b,c)\<in>set cs. well_formed c)"
| "well_formed (IF b THEN c1 ELSE c2) \<longleftrightarrow>  well_formed c1 \<and> well_formed c2"
| "well_formed (WHILE b DO c) \<longleftrightarrow> well_formed c"

lemma wf_step: 
  assumes "well_formed c" 
  assumes "small_step (c,s) (c',s')"  
  shows "well_formed c'"
  using assms(2,1)
  by induction auto

print_statement com.induct

lemma progress: "well_formed c \<Longrightarrow> \<exists>cs'. small_step (c,s) cs'"
proof (induction c arbitrary: s rule: com.inducts(1)
    [where ?P2.0 = "\<lambda>_. True" and ?P3.0 = "\<lambda>_. True"])
  case Seq thus ?case by clarsimp (metis seq1 seq2)
next
  case GC thus ?case
    by clarsimp (metis (lifting) gc gc_deadlock split_beta surjective_pairing)
next
  case IfTE thus ?case
    by clarsimp (metis if_false if_true)
next 
  case While thus ?case     
    by clarsimp (metis while_false while_true)
qed auto

xxx, ctd here:
  define semantics assequence of states
  define atomic propositions

  define compiler, show correctness
  define optimizing compiler/vm, show correctness, i.e. stutter-equiv 
    wrt observable APs.
    


type_synonym state = "bool list"


datatype instr =
  AssI "nat list" "bexp list" |
  TestI bexp int |
  ChoiceI "(bexp * int) list" |
  GotoI int

type_synonym config = "nat * state"
type_synonym bprog = "instr array"

(*
 * Semantics Notice:
 * To be equivalent in semantics with SPIN, there is no such thing as a finite run:
 *   - Deadlocks (i.e. empty Choice) are self-loops
 *   - program termination is self-loop
 *)

fun exec :: "instr \<Rightarrow> config \<Rightarrow> config list" where
"exec instr (pc,s) = (case instr of
  AssI ns bs \<Rightarrow> let bvs = zip ns (map (\<lambda>b. bval b s) bs) in
               [(pc + 1, foldl (\<lambda>s (n,bv). s[n := bv]) s bvs)] |
  TestI b d \<Rightarrow> [if bval b s then (pc+1, s) else (nat(int(pc+1)+d), s)] |
  ChoiceI bis \<Rightarrow> let succs = [(nat(int(pc+1)+i), s) . (b,i) <- bis, bval b s] 
                in if succs = [] then [(pc,s)] else succs |
  GotoI d \<Rightarrow> [(nat(int(pc+1)+d),s)])"


function exec' :: "bprog \<Rightarrow> state \<Rightarrow> nat \<Rightarrow> nat list" where
"exec' ins s pc = (
  if pc < array_length ins then (
    case (array_get ins pc) of
      AssI ns bs \<Rightarrow> [pc] |
      TestI b d \<Rightarrow> (
        if bval b s then exec' ins s (pc+1) 
        else let pc'=(nat(int(pc+1)+d)) in if pc'>pc then exec' ins s pc'
        else [pc']
      ) |
      ChoiceI bis \<Rightarrow> let succs = [(nat(int(pc+1)+i)) . (b,i) <- bis, bval b s]
                    in if succs = [] then [pc] else concat (map (\<lambda>pc'. if pc'>pc then exec' ins s pc' else [pc']) succs) |
      GotoI d \<Rightarrow> let pc' = nat(int(pc+1)+d) in (if pc'>pc then exec' ins s pc' else [pc'])
  ) else [pc]
)"
by pat_completeness auto
termination 
  apply (relation "measure (%(ins,s,pc). array_length ins - pc)")
  apply auto
  done

fun nexts1 :: "bprog \<Rightarrow> config \<Rightarrow> config list" where
"nexts1 ins (pc,s) = (if pc < array_length ins then exec (array_get ins pc) (pc,s) else [(pc,s)])"

fun nexts :: "bprog \<Rightarrow> config \<Rightarrow> config list" where
"nexts ins (pc,s) = concat (map (\<lambda>(pc,s). map (\<lambda>pc. (pc,s)) (exec' ins s pc)) (nexts1 ins (pc,s)))"

declare nexts.simps [simp del]

datatype
  com = SKIP
      | Assign "nat list" "bexp list"    
      | Seq    com  com         
      | GC  "(bexp * com)list"  
      | IfTE  bexp com com      
      | While  bexp com         

locale BoolProg_Syntax begin
  notation 
        Assign       ("_ ::= _" [999, 61] 61)
    and Seq          ("_;/ _"  [60, 61] 60)
    and GC           ("IF _ FI")
    and IfTE         ("(IF _/ THEN _/ ELSE _)"  [0, 61, 61] 61)
    and While        ("(WHILE _/ DO _)"  [0, 61] 61)
end

context begin interpretation BoolProg_Syntax .
fun comp' :: "com \<Rightarrow> instr list" where
"comp' SKIP = []" |
"comp' (Assign n b) = [AssI n b]" |
"comp' (c1;c2) = comp' c1 @ comp' c2" |
"comp' (IF gcs FI) =
  (let cgcs = map (\<lambda>(b,c). (b,comp' c)) gcs in
   let addbc = (\<lambda>(b,cc) (bis,ins).
         let cc' = cc @ (if ins = [] then [] else [GotoI (int(length ins))]) in
         let bis' = map (\<lambda>(b,i). (b, i + int(length cc'))) bis
         in ((b,0)#bis', cc' @ ins)) in
   let (bis,ins) = foldr addbc cgcs ([],[])
   in ChoiceI bis # ins)" |
"comp' (IF b THEN c1 ELSE c2) =
  (let ins1 = comp' c1 in let ins2 = comp' c2 in
   let i1 = int(length ins1 + 1) in let i2 = int(length ins2)
   in TestI b i1 # ins1 @ GotoI i2 # ins2)" |
"comp' (WHILE b DO c) =
  (let ins = comp' c in
   let i = int(length ins + 1)
   in TestI b i # ins @ [GotoI (-(i+1))])"

(* a test: *)
value "comp' (IF [(V 0, [1,0] ::= [TT, FF]), (V 1, [0] ::= [TT])] FI)"

end

definition comp :: "com \<Rightarrow> bprog" where
  "comp = array_of_list \<circ> comp'"

(* 
  Optimization

- Gotos: Resolve Goto-Chains: If a Goto referres to another Goto -- add the second offset to the first one. 
                              If a Goto has a negative offset ignore it, to avoid problems with loops.

*)

fun opt' where
  "opt' (GotoI d) ys = (let next = \<lambda>i. (case i of GotoI d \<Rightarrow> d + 1 | _ \<Rightarrow> 0)
                        in if d < 0 \<or> nat d \<ge> length ys then (GotoI d)#ys
                           else let d' = d + next (ys ! nat d)
                           in (GotoI d' # ys))"
| "opt' x ys = x#ys"

definition opt :: "instr list \<Rightarrow> instr list" where
  "opt instr = foldr opt' instr []"

definition optcomp :: "com \<Rightarrow> bprog" where
  "optcomp \<equiv> array_of_list \<circ> opt \<circ> comp'"

subsection {* Finiteness of reachable configurations *}

inductive_set reachable_configs
  for bp :: bprog
  and c\<^sub>s :: config -- "start configuration"
where
"c\<^sub>s \<in> reachable_configs bp c\<^sub>s" |
"c \<in> reachable_configs bp c\<^sub>s \<Longrightarrow> x \<in> set (nexts bp c) \<Longrightarrow> x \<in> reachable_configs bp c\<^sub>s"

lemmas reachable_configs_induct = reachable_configs.induct[split_format(complete),case_names 0 1]

fun offsets :: "instr \<Rightarrow> int set" where
"offsets (AssI _ _) = {0}" |
"offsets (TestI _ i) = {0,i}" |
"offsets (ChoiceI bis) = set(map snd bis) \<union> {0}" |
"offsets (GotoI i) = {i}"

definition offsets_is :: "instr list \<Rightarrow> int set" where
"offsets_is ins = (UN instr : set ins. offsets instr)"

definition max_next_pcs :: "instr list \<Rightarrow> nat set" where
"max_next_pcs ins = {nat(int(length ins + 1) + i) |i. i : offsets_is ins}"

(* delete next line *)
lemma finite_max_next_pcs: "finite(max_next_pcs bp)"
proof-
  { fix instr have "finite (offsets instr)" by(cases instr) auto }
  moreover
  { fix ins have "max_next_pcs ins = (UN i : offsets_is ins. {nat(int(length ins + 1) + i)})"
    by(auto simp add: max_next_pcs_def) }
  ultimately show ?thesis by(auto simp add: offsets_is_def)
qed

(* mv *)
lemma (in linorder) le_Max_insertI1: "\<lbrakk> finite A; x \<le> b \<rbrakk> \<Longrightarrow> x \<le> Max (insert b A)"
by (metis Max_ge finite.insertI insert_iff order_trans)
lemma (in linorder) le_Max_insertI2: "\<lbrakk> finite A; A \<noteq> {}; x \<le> Max A \<rbrakk> \<Longrightarrow> x \<le> Max (insert b A)"
by(auto simp add: max_def not_le simp del: Max_less_iff)

lemma max_next_pcs_not_empty:
  "pc<length bp \<Longrightarrow> x : set (exec (bp!pc) (pc,s)) \<Longrightarrow> max_next_pcs bp \<noteq> {}"
apply(drule nth_mem)
apply(fastforce simp: max_next_pcs_def offsets_is_def split: instr.splits)
done

lemma Max_lem2: 
  assumes "pc < length bp" 
  and "(pc', s') \<in> set (exec (bp!pc) (pc, s))"
  shows "pc' \<le> Max (max_next_pcs bp)"
using assms
proof (cases "bp ! pc")
  case (ChoiceI l)
  show ?thesis
  proof (cases "pc' = pc")
    case True with assms ChoiceI show ?thesis
      by (auto simp: Max_ge_iff max_next_pcs_not_empty finite_max_next_pcs) 
         (force simp add: max_next_pcs_def offsets_is_def dest: nth_mem)
  next
    case False with ChoiceI assms obtain b i where 
      bi: "bval b s" "(b,i) \<in> set l" "pc' = nat(int(pc+1)+i)"
      by (auto split: split_if_asm)
    with ChoiceI assms have "i \<in> UNION (set bp) offsets" by (force dest: nth_mem)
    with bi assms have "\<exists>a. (a \<in> max_next_pcs bp \<and> pc' \<le> a)"
      unfolding max_next_pcs_def offsets_is_def by force
    thus ?thesis 
      by (auto simp: Max_ge_iff max_next_pcs_not_empty[OF assms] finite_max_next_pcs) 
    qed
qed (auto simp: Max_ge_iff max_next_pcs_not_empty finite_max_next_pcs,
     (force simp add: max_next_pcs_def offsets_is_def dest: nth_mem split: split_if_asm)+)

lemma Max_lem1: "\<lbrakk> pc < length bp; (pc', s') \<in> set (exec (bp ! pc) (pc, s))\<rbrakk>
    \<Longrightarrow> pc' \<le> Max (insert x (max_next_pcs bp))"
apply(rule le_Max_insertI2)
apply (simp add: finite_max_next_pcs)
apply(simp add: max_next_pcs_not_empty)
apply(auto intro!: Max_lem2 simp del:exec.simps)
done

definition "pc_bound bp \<equiv> max 
    (Max (max_next_pcs (list_of_array bp)) + 1)
    (array_length bp + 1)"

declare exec'.simps[simp del]

lemma [simp]: "length (list_of_array a) = array_length a" by (cases a) auto

lemma aux2: 
  assumes A: "pc < array_length ins" 
  assumes B: "ofs \<in> offsets_is (list_of_array ins)"
  shows "nat (1 + int pc + ofs) < pc_bound ins"
proof -
  have "nat (int (1 + array_length ins) + ofs) 
    \<in> max_next_pcs (list_of_array ins)"
    using B unfolding max_next_pcs_def
    by auto
  with A show ?thesis
    unfolding pc_bound_def
    apply -
    apply (rule min_max.less_supI1)
    apply auto
    apply (drule Max_ge[OF finite_max_next_pcs])
    apply simp
    done
qed

lemma array_idx_in_set: 
  "\<lbrakk> pc < array_length ins; array_get ins pc = x \<rbrakk> 
  \<Longrightarrow> x \<in> set (list_of_array ins)"
  by (induct ins) auto

lemma rcs_aux: 
  assumes "pc < pc_bound bp"    
  assumes "pc'\<in>set (exec' bp s pc)"
  shows "pc' < pc_bound bp"
  using assms
proof (induction bp s pc arbitrary: pc' rule: exec'.induct[case_names C])
  case (C ins s pc pc')
  from C.prems show ?case
    apply (subst (asm) exec'.simps)
    apply (split split_if_asm instr.split_asm)+
    apply (simp add: pc_bound_def)

    apply (simp split: split_if_asm add: Let_def)
    apply (frule (2) C.IH(1), auto) []
    apply (auto simp: pc_bound_def) []
    apply (frule (2) C.IH(2), auto) []
    apply (subgoal_tac "int \<in> offsets_is (list_of_array ins)")
    apply (blast intro: aux2)
    apply (auto simp: offsets_is_def) []
    apply (rule_tac x="TestI bexp int" in bexI, auto simp: array_idx_in_set) []

    apply (clarsimp split: split_if_asm simp add: Let_def)
    apply (elim disjE conjE, auto) []
    apply (frule (1) C.IH(3), auto) []
    apply (force)
    apply (force)
    apply (subgoal_tac "ba \<in> offsets_is (list_of_array ins)")
    apply (blast intro: aux2)
    apply (auto simp: offsets_is_def) []
    apply (rule_tac x="ChoiceI list" in bexI, auto simp: array_idx_in_set) []

    apply (simp split: split_if_asm add: Let_def)
    apply (frule (1) C.IH(4), auto) []
    apply (subgoal_tac "int \<in> offsets_is (list_of_array ins)")
    apply (blast intro: aux2)
    apply (auto simp: offsets_is_def) []
    apply (rule_tac x="GotoI int" in bexI, auto simp: array_idx_in_set) []
    
    apply simp
    done
qed

lemma reachable_configs_in_bound:
  "pc < pc_bound bp \<Longrightarrow> reachable_configs bp (pc,s) 
  \<subseteq> {0..< pc_bound bp} \<times> {bs. length bs = length s}"
proof auto
  fix pc' s' 
  assume "(pc', s') \<in> reachable_configs bp (pc, s)"
  and "pc < pc_bound bp"
  thus "pc' < pc_bound bp"
  proof(induction rule: reachable_configs_induct)
    case 0 thus ?case . 
  next
    case (1 pc' s' pc'' s'') 
    obtain ins where BP_eq[simp]: "bp = Array ins" by (cases bp)
    from 1 show ?case 
      apply (auto simp del: exec.simps exec'.simps simp add: nexts.simps 
        Max_lem1 le_imp_less_Suc split: split_if_asm)
      prefer 2
      apply (simp add: exec'.simps)
    proof -
      case (goal1 pch)
      from goal1(4,5) have "pch < pc_bound bp"
        unfolding pc_bound_def
        by (auto simp del: exec.simps dest: Max_lem2)
      with rcs_aux goal1(6) show ?case by (simp only: BP_eq)
    qed
  qed
next
  { fix s0 :: state and l 
    have "length (foldl (\<lambda>s (x, y). s[x := y]) s0 l) = length s0"
      by (induction l arbitrary: s0) (simp_all split: prod.splits)
  }
  note len = this

  fix pc' s' show 
    "(pc', s') \<in> reachable_configs bp (pc, s) \<Longrightarrow> length s' = length s"
  proof(induction rule: reachable_configs_induct)
    case (1 pc' s' pc'' s'')
    obtain ins where BP_eq[simp]: "bp = Array ins" by (cases bp)
    from 1 show ?case
      apply (auto simp: nexts.simps 
        simp del: exec.simps exec'.simps split: split_if_asm)
      apply (auto simp: len simp del: exec'.simps
        split: instr.splits split_if_asm)
      done
  qed simp
qed
      
lemma reachable_configs_out_of_bound: "(pc',s')\<in>reachable_configs bp (pc,s) 
  \<Longrightarrow> \<not> pc < pc_bound bp \<Longrightarrow> (pc',s') = (pc,s)"
proof (induct rule: reachable_configs_induct)
  case (1 pc' s' pc'' s'')
  hence [simp]: "pc'=pc" "s'=s" by auto
  from 1(4) have "\<not> pc < array_length bp" unfolding pc_bound_def by auto
  with 1(3) show ?case
    by (auto simp add: nexts.simps exec'.simps)
qed auto

lemma reachable_configs_subset:
  "reachable_configs bp (pc,s) \<subseteq>
   insert pc {0..< pc_bound bp} \<times> {bs. length bs = length s}"
  apply (cases "pc < pc_bound bp")
  using reachable_configs_in_bound
  apply auto []
  apply (auto dest: reachable_configs_out_of_bound)
  done
  
lemma finite_bool_array_set: "finite {bs :: bool list. length bs = n}"
proof-
  have "ALL xs. set xs <= {True,False}" by auto
  with finite_lists_length_eq[of "{True,False}"] show ?thesis by simp
qed

lemma reachable_configs_finite[simp, intro!]: 
  "finite (reachable_configs bp c)"
  apply (cases c, simp)
  apply (rule finite_subset[OF reachable_configs_subset])
  apply (blast intro: finite_bool_array_set)
  done


definition config_bound :: "bprog \<times> config \<Rightarrow> config set"
  where
  "config_bound bpc \<equiv> let (bp,(pc,s)) = bpc in
    insert pc {0..< pc_bound bp} \<times> {bs. length bs = length s}"

lemma reachable_configs_bound:
  "reachable_configs bp c \<subseteq> config_bound (bp,c)"
  using reachable_configs_subset
  unfolding config_bound_def
  by auto

lemma config_bound_finite[simp, intro!]: "finite (config_bound bpc)"
  by (auto 
    split: prod.splits
    intro: finite_bool_array_set 
    simp: config_bound_def)


definition "bpc_is_run bpc r \<equiv> let (bp,c)=bpc in r 0 = c \<and> (\<forall>i. r (Suc i) \<in> set (BoolProgs.nexts bp (r i)))"
definition "bpc_props c \<equiv> {i. i<length (snd c) \<and> snd c ! i}"
definition "bpc_lang bpc \<equiv> {bpc_props o r | r. bpc_is_run bpc r}"

(* Printing *)
definition print_list where
  "print_list f sep s = (let f' = (\<lambda>str s. if str = [] then f s
                                       else str @ sep @ f s)
                     in ''['' @ (foldl f' [] s) @ '']'')"

fun bool_list_to_prop_list where
  "bool_list_to_prop_list _ [] props = props"
| "bool_list_to_prop_list n (x#xs) props = (let props' = if x then n#props else props 
                                            in bool_list_to_prop_list (Suc n) xs props')"

fun print_config where
  "print_config f (p,s) = f p @ '' '' @ (print_list f '', '' (rev (bool_list_to_prop_list 0 s [])))"

hide_const (open) print_list bool_list_to_prop_list

end
