(* 
   Author:  René Neumann <rene.neumann@in.tum.de>
*)

theory Graph_Ext
imports 
  Main 
  "../../../../Libs/Graphs/Edge_Walks"
  "../../../../Libs/Graphs/Digraph_Components"
  "../../../../Libs/Automatic_Refinement/Lib/Misc"
begin

text {* 
This theory relates a simple graph representation by relations to the
Graph theories developed by Lars Noschinski.

Later on different automata structures are mapped onto this simple
to be able to use the graph-theoretic reasoning.
*}

context wellformed_graph
begin

text {* An alias allowing to use the reachable syntax in this locale without
 having to specify @{term G} explicitly.

Note that we use @{text "\<rightarrow>\<star>"} instead of @{text "\<rightarrow>*"}. *}
(* has to be defined in every "sublocale" too *)
abbreviation pd_reachable ("_ \<rightarrow>\<star> _" [100,100] 40) where
  "pd_reachable v w \<equiv> Digraphs.reachable v G w"

definition reachable1 ("_ \<rightarrow>+ _" [100,100] 40) where
  "reachable1 v w \<equiv> (\<exists>p. ewalk v p w \<and> p \<noteq> [])"

lemma reachable1I:
  "ewalk v p w \<Longrightarrow> p \<noteq> [] \<Longrightarrow> v \<rightarrow>+ w"
by (auto simp add: reachable1_def)

lemma reachable1_reachable[intro]:
  "v \<rightarrow>+ w \<Longrightarrow> v \<rightarrow>\<star> w"
unfolding reachable1_def reachable_ewalk
by metis

lemmas reachable1_reachableE[elim] = reachable1_reachable[elim_format]

lemma reachable_neq_reachable1[intro]:
  "v \<noteq> w \<Longrightarrow> v \<rightarrow>\<star> w \<Longrightarrow> v \<rightarrow>+ w"
proof -
  assume "v \<noteq> w" "v \<rightarrow>\<star> w"
  then obtain p where p: "ewalk v p w" unfolding reachable_ewalk by auto
  hence "p \<noteq> []" using `v \<noteq> w` by induct simp_all
  with p show "v \<rightarrow>+ w" by (rule reachable1I)
qed

lemmas reachable_neq_reachable1E[elim] = reachable_neq_reachable1[elim_format]

lemma reachable1_trans[trans]:
  "u \<rightarrow>+ v \<Longrightarrow> v \<rightarrow>+ w \<Longrightarrow> u \<rightarrow>+ w"
unfolding reachable1_def
by (force intro: ewalk_ewalk_joinI)

lemma reachable1_reachable_trans [trans]:
  "u \<rightarrow>+ v \<Longrightarrow> v \<rightarrow>\<star> w \<Longrightarrow> u \<rightarrow>+ w"
by (metis reachable1_trans reachable_neq_reachable1)

lemma reachable_reachable1_trans [trans]:
  "u \<rightarrow>\<star> v \<Longrightarrow> v \<rightarrow>+ w \<Longrightarrow> u \<rightarrow>+ w"
by (metis reachable1_trans reachable_neq_reachable1)

lemma ewalkI[intro]:
  "\<lbrakk> u \<in> verts G; set p \<subseteq> edges G; ewhd u p = u;  ewlast u p = v; ves p \<rbrakk> \<Longrightarrow> ewalk u p v"
unfolding ewalk_def
by metis

lemma edge_implies_ewalk:
  "e \<in> edges G \<Longrightarrow> ewalk (start G e) [e] (target G e)"
by (auto simp add: start_in_verts target_in_verts ewalk_verts_def)

lemma edge_implies_reach:
  "e \<in> edges G \<Longrightarrow> start G e \<rightarrow>\<star> target G e"
by (fact reachable_ewalkI[OF edge_implies_ewalk])

lemma edge_implies_reach1:
  "e \<in> edges G \<Longrightarrow> start G e \<rightarrow>+ target G e"
proof -
  assume "e \<in> edges G"
  hence "ewalk (start G e) [e] (target G e)" by (fact edge_implies_ewalk)
  thus "start G e \<rightarrow>+ target G e" unfolding reachable1_def by auto
qed

lemma reach_implies_vert[intro]:
  assumes "x \<rightarrow>\<star> y"
  shows "x \<in> verts G" and "y \<in> verts G"
proof -
  from assms obtain p where p: "ewalk x p y" using reachable_ewalk by blast
  
  from p show "x \<in> verts G" by (frule ewalkE_verts)
  from p show "y \<in> verts G" by (frule ewalkE_verts)
qed

lemmas reach1_implies_vert[intro] = reach_implies_vert[OF reachable1_reachable]

lemma self_reachable:
  "x \<in> verts G \<Longrightarrow> x \<rightarrow>\<star> x"
by (rule reachable_ewalkI[of _ "[]"]) (auto simp add: ewalk_empty_iff)

definition succs_edges :: "'a \<Rightarrow> 'b set"
where
  "succs_edges n = {e \<in> edges G. start G e = n }"

lemma succs_edges_reachable:
  "x \<in> succs_edges n \<Longrightarrow> (n \<rightarrow>+ (target G x))"
unfolding succs_edges_def
by (auto intro: edge_implies_reach1)

definition succs_nodes :: "'a \<Rightarrow> 'a set"
where
  "succs_nodes n = { v | v e. e \<in> edges G \<and> (n, v) = edge_to_ends G e }"

lemma succs_nodes_alt_def:
  "succs_nodes n = (target G) ` succs_edges n"
unfolding succs_nodes_def succs_edges_def edge_to_ends_def
by auto

lemma succs_nodes_reachable: 
  "v \<in> succs_nodes n \<Longrightarrow> n \<rightarrow>+ v"
proof -
  assume "v \<in> succs_nodes n"
  then obtain e where edef: "e \<in> succs_edges n \<and> target G e = v" using succs_nodes_alt_def by auto
  with succs_edges_reachable have "n \<rightarrow>+ (target G e)" by metis
  with edef show "n \<rightarrow>+ v" by simp
qed

lemma vwalk_implies_reachable:
  "vwalk p G \<Longrightarrow> n < length p \<Longrightarrow> m < n \<Longrightarrow> p ! m \<rightarrow>+ p ! n"
proof (induct arbitrary: n m rule: vwalk_induct)
  case Base thus ?case by simp
next
  case (Cons u v ps)
  then have "\<And>n. n < length (v # ps) \<Longrightarrow> n > 0 \<Longrightarrow> v \<rightarrow>+ (v#ps) ! n" by force
  hence v_reach: "\<And>n. n < length (u # v # ps) \<Longrightarrow> n > 1 \<Longrightarrow> v \<rightarrow>+ (u#v#ps) ! n" by simp smt
  
  from Cons(1) obtain e where e: "start G e = u" "target G e = v" "e \<in> edges G" by force 
  hence u_r_v: "u \<rightarrow>+ v" using edge_implies_reach1[of e] by simp
  show ?case
  proof (cases "m = 0")
    case True
    with Cons v_reach show ?thesis 
      using u_r_v reachable1_trans[OF u_r_v]
      by (cases "n > 1") simp_all
  next
    case False with Cons have "n \<noteq> 0" by simp
    with False obtain m' n' where m': "m = Suc m'" and n': "n = Suc n'" by (metis nat.exhaust)
    hence reduce: "(u # v # ps) ! m = (v # ps) ! m'" "(u # v # ps) ! n = (v # ps) ! n'" by simp_all

    from n' m' Cons have "(v # ps) ! m' \<rightarrow>+ (v # ps) ! n'" by simp
    with reduce show ?thesis by simp
  qed
qed

lemma vwalk_start_last_reachable:
  assumes walk: "vwalk p G"
  and len: "length p > 1"
  shows "hd p \<rightarrow>+ last p"
proof -
  from len have "p \<noteq> []" by auto
  with hd_conv_nth last_conv_nth have conv: "hd p = p ! 0" "last p = p ! (length p - 1)" by simp_all
  with len have "length p - 1 > 0" by simp
  with vwalk_implies_reachable[OF walk] conv show ?thesis by simp
qed

lemma vwalk_tl:
  "xs \<noteq> [] \<Longrightarrow> vwalk xs G  \<Longrightarrow> vwalk (tl xs) G \<or> tl xs = []"
by (metis list.exhaust list.sel(3) vwalk_consE)

lemma vwalk_butlast:
  "xs \<noteq> [] \<Longrightarrow> vwalk xs G  \<Longrightarrow> vwalk (butlast xs) G \<or> butlast xs = []"
by (metis append_butlast_last_id vwalk_append1)

lemma ewalk_induct_rev[case_names Base Snoc, consumes 1]:
  assumes "ewalk u p v"
  and  "\<And>v. v \<in> verts G \<Longrightarrow> P v [] v" 
  assumes "\<And>v1 v2 e es. e \<in> edges G \<Longrightarrow> edge_to_ends G e = (v1, v2) 
    \<Longrightarrow> P u es v1 \<Longrightarrow> P u (es@[e]) v2"
  shows "P u p v"
using assms
proof (induction p arbitrary: u v rule: rev_induct)
  case Nil then show ?case using ewalk_empty_iff by force
next
  case (snoc e es) then have *: "e \<in> edges G" "edge_to_ends G e = (start G e, v)" unfolding ewalk_conv edge_to_ends_def by auto
  with snoc wellformed have "ewalk (start G e) [e] v" unfolding ewalk_conv ewalk_verts_def by force
  with snoc have "ewalk u es (start G e)" using ewalk_join_ewalk_l' by metis
  with ewalk_imp_ves have "P u es (start G e)" using snoc.IH snoc.prems by auto
  thus ?case by (rule snoc.prems(3)[OF *])
qed

lemma ewalk_edge_in_ewalk_verts:
  assumes "ewalk x p y" 
  and "e \<in> set p" 
  shows "{start G e, target G e} \<subseteq> set (ewalk_verts x p)"
using assms assms(1)[THEN ewalk_imp_ves]
proof (induction arbitrary: e rule: ewalk_induct_raw)
  case (Cons w1 w2 x xs) thus ?case
    proof (cases "x = e")
      case False with Cons have "{start G e, target G e} \<subseteq> set (ewalk_verts w2 xs)" by (fastforce dest!: ves_ConsD)
      with Cons[unfolded edge_to_ends_def] show ?thesis using ewalk_verts_Cons by auto
    qed (simp add: ewalk_verts_def)
qed simp

lemma ewalk_verts_are_edges:
  assumes "ewalk u p v"
  and "ewalk_verts u p = xs @ a # b # ys"
  obtains e where "e \<in> edges G" "start G e = a" "target G e = b"
using assms
proof (induct p arbitrary: u xs)
  case (Cons e p) note i = Cons
  show ?case
  proof (cases xs)
    case Nil with Cons
    have "e \<in> edges G" and "start G e = a \<and> target G e = b"
      by (auto simp add: ewalk_verts_def)
    then show ?thesis using i by auto
  next
    case (Cons x xs')
    with i have "ewalk (target G e) p v"
      and "ewalk_verts (target G e) p = tl xs @ a # b # ys"
      using ewalk_join_decomp[of u "[e]" p v] by (auto simp: ewalk_verts_Cons)
    then show ?thesis using i by simp
  qed
qed simp

lemma split_ewalk:
  assumes "ewalk x p y" and "ewalk_verts x p = ys@zs" and "ys \<noteq> []"
  obtains q r where "p = q@r" "ewalk_verts x q = ys" and "ewalk x q (last ys)" and "ewalk (last ys) r y"
proof
  def q \<equiv> "take (length ys - 1) p"
  def r \<equiv> "drop (length ys - 1) p"
  show "p = q@r" unfolding q_def r_def by simp

  moreover
  from assms ewalk_def have "hd ys = x" by simp
  from assms have "ys = take (length ys) (ewalk_verts x p)" by simp
  with `ys \<noteq> []` have "ys = take (Suc (length ys - 1)) (ewalk_verts x p)" by auto
  thus "ewalk_verts x q = ys"
    using ewalk_verts_take_conv[of x "length ys - 1" p] `hd ys = x`
    unfolding q_def
    apply (cases "p = []")
      apply simp
      apply (cases "1 < length ys")
       apply (simp_all add: ewalk_verts_def)
    done
  ultimately show "ewalk x q (last ys)" using ewalk_join_ewalk_l[of x q r y "last ys"] assms(1) by simp

  from ewalk_join_ewalk_r'[OF _ this] assms(1) `p = q@r` show "ewalk (last ys) r y" by simp
qed
end 

subsection {* A finite digraph based on a successor function and a starting node. *}

text {* The starting node is needed to restrict to a finite graph. 
In effect, the set of vertices are all nodes that are reachable from the starting node. *}

locale succs_digraph =
  fixes succs :: "'n \<Rightarrow> 'n set"
  and v\<^sub>0 :: "'n"
begin

inductive_set V :: "'n set" where
  v\<^sub>0_in_V [simp,intro]: "v\<^sub>0 \<in> V"
| succs_in_V: "v \<in> V \<Longrightarrow> v' \<in> succs v \<Longrightarrow> v' \<in> V"

lemma V_alt_def:
  "V = {(v,v'). v' \<in> succs v}\<^sup>+ `` {v\<^sub>0} \<union> {v\<^sub>0}" (is "V = ?V")
proof -
  {
    fix x
    assume "x \<in> V" hence "x \<in> ?V" by (induct x) (auto intro: trancl_into_trancl)
  }
 
  moreover
  {
    fix x
    assume "x \<in> ?V" hence "x \<in> V" 
    proof (cases "x = v\<^sub>0")
      case True thus ?thesis by simp
    next
      case False with `x \<in> ?V` have "(v\<^sub>0, x) \<in> {(v, v'). v' \<in> succs v}\<^sup>+" by auto
      thus ?thesis by (induct rule: trancl_induct) (auto intro: succs_in_V)
    qed
  }
  ultimately show ?thesis by blast
qed

lemma V_alt2_def: "V = {(v,v'). v' \<in> succs v}\<^sup>* `` {v\<^sub>0}"
    by (metis trancl_image_by_rtrancl V_alt_def)

definition E :: "'n rel"where
  "E \<equiv> {(v,v'). v \<in> V \<and> v' \<in> succs v}"

lemma edge_in_nodes:
  assumes "(v,v') \<in> E"
  shows "v \<in> V" and "v' \<in> V"
using assms
by (auto simp add: E_def intro: succs_in_V)

lemma edges_sub_nodes:
  "E \<subseteq> V \<times> V"
by (auto intro: edge_in_nodes)

abbreviation sd_to_graph ("\<G>") where
  "sd_to_graph \<equiv> \<lparr> verts = V, edges = E, start = fst, target = snd \<rparr>"

abbreviation sd_reachable ("_ \<rightarrow>\<star> _" [100,100] 40) where
  "sd_reachable v w \<equiv> Digraphs.reachable v \<G> w"
end

sublocale succs_digraph \<subseteq> wellformed_graph "\<G>" 
  by unfold_locales (auto simp add: edge_in_nodes)

context succs_digraph
begin

abbreviation sd_reachable1  ("_ \<rightarrow>+ _" [100,100] 40) where
  "sd_reachable1 \<equiv> reachable1"

lemma edge_to_ends_reduce[simp]:
  "edge_to_ends \<G> e = e"
unfolding edge_to_ends_def
by simp

lemma edges_ends_reduce[simp]:
  "edges_ends \<G> = E" 
unfolding edges_ends_def image_def
by simp

lemma succs_eq_succs_nodes:
  "v \<in> V \<Longrightarrow> succs v = succs_nodes v"
unfolding succs_nodes_def edge_to_ends_def
by (simp add: E_def)

lemma succs_reachable:
  "v \<in> V \<Longrightarrow> w \<in> succs v \<Longrightarrow> v \<rightarrow>+ w"
using succs_nodes_reachable succs_eq_succs_nodes
by metis

lemma succs_implies_edge:
  "y \<in> V \<Longrightarrow> x \<in> succs y \<Longrightarrow> (y, x) \<in> E"
unfolding E_def by simp

lemma edge_is_succs:
  "(x,y) \<in> E \<Longrightarrow> y \<in> succs x"
by (simp add: E_def)

lemma edges_reduce[simp]:
  "edges \<G> = E"
by simp

lemma verts_reduce[simp]:
  "verts \<G> = V"
by simp

lemmas reach_vert[elim] = reach_implies_vert[simplified,elim_format] reach1_implies_vert[simplified,elim_format]

lemma V_restrict_sub:
  "x \<in> V \<Longrightarrow> {(v,v'). v' \<in> succs v}\<^sup>+ `` {x} \<union> {x} \<subseteq> V"
unfolding V_alt_def
by auto

lemma vert_rtrancl:
  assumes "x \<in> V"
  shows "(v\<^sub>0, x) \<in> E\<^sup>*"
using assms
by induction (auto simp add: E_def intro: rtrancl_into_rtrancl)

lemma subgraph_V_sub:
  assumes "{(v,v'). v' \<in> succs' v} \<subseteq> {(v,v'). v' \<in> succs v}"
  and "w \<in> V"
  shows "succs_digraph.V succs' w \<subseteq> V"
proof (cases "w = v\<^sub>0")
  case False
  have "\<And>x. (w, x) \<in> {(v,v'). v' \<in> succs' v}\<^sup>+ \<Longrightarrow> (v\<^sub>0, x) \<in> {(v,v'). v' \<in> succs v}\<^sup>+"
  proof -
    fix x
    assume "(w, x) \<in> {(v,v'). v' \<in> succs' v}\<^sup>+"
    with trancl_mono[OF _ assms(1)] have "(w, x) \<in> {(v,v'). v' \<in> succs v}\<^sup>+" .
    moreover from `w \<in> V` False have "(v\<^sub>0, w) \<in>  {(v,v'). v' \<in> succs v}\<^sup>+" by (auto simp add: V_alt_def)
    ultimately show "(v\<^sub>0, x) \<in> {(v,v'). v' \<in> succs v}\<^sup>+" by (auto intro: trancl_into_trancl)
  qed
  with assms show ?thesis by (auto simp add: V_alt_def succs_digraph.V_alt_def)
next
  case True
  from assms have "{(v, v'). v' \<in> succs' v}\<^sup>+ `` {v\<^sub>0} \<union> {v\<^sub>0} \<subseteq> V"
    unfolding V_alt_def
    by (blast intro: trancl_mono Image_mono)
  with True show ?thesis by (simp add: succs_digraph.V_alt_def)
qed

lemma subgraph_E_sub:
  assumes "{(v,v'). v' \<in> succs' v} \<subseteq> {(v,v'). v' \<in> succs v}"
  and wV: "w \<in> V"
  shows "succs_digraph.E succs' w \<subseteq> E"
proof -
  interpret sub!: succs_digraph succs w by unfold_locales
  interpret sub2!: succs_digraph succs' w by unfold_locales

  have "sub2.E \<subseteq> sub.E"
  proof -
    from sub.subgraph_V_sub assms(1) have "sub2.V \<subseteq> sub.V" by simp
    with assms(1) show ?thesis unfolding sub.E_def sub2.E_def by auto
  qed

  also
  have "sub.E \<subseteq> E"
  proof -
    from wV subgraph_V_sub[where succs'=succs] have "sub.V \<subseteq> V" by simp
    with wV show ?thesis by (auto simp add: E_def sub.E_def)
  qed

  finally show ?thesis .
qed

lemma lift_reachable1_trancl:
  assumes "x \<in> V"
  and r1_trancl: "x \<rightarrow>+ y \<longleftrightarrow> (x,y) \<in> R\<^sup>+"
  shows "x \<rightarrow>\<star> y \<longleftrightarrow> (x,y) \<in> R\<^sup>*"
proof (cases "x = y")
  case True thus ?thesis using self_reachable assms by simp
next
  case False with r1_trancl have "x \<rightarrow>\<star> y \<longleftrightarrow> (x,y) \<in> R\<^sup>+" by blast
  with False show ?thesis using rtrancl_eq_or_trancl by metis
qed

lemma reachable1_trancl:
  "x \<rightarrow>+ y \<longleftrightarrow> (x,y) \<in> E\<^sup>+"
proof
  assume "x \<rightarrow>+ y"
  then obtain p where "ewalk x p y" "p \<noteq> []" unfolding reachable1_def by auto
  with this(1) show "(x,y) \<in> E\<^sup>+"
  proof (induct)
    case (Cons v w e es) thus ?case using ewalk_Cons_iff by (cases es) auto
  qed simp
next
  assume "(x, y) \<in> E\<^sup>+"
  thus "x \<rightarrow>+ y"
  proof induct
    case base thus ?case using edge_implies_reach1 by force
  next
    case (step v w) with edge_implies_reach1 have "v \<rightarrow>+ w" by force
    with step reachable1_trans show ?case by blast
  qed
qed

lemma reachable_rtrancl:
  "x \<in> V \<Longrightarrow> x \<rightarrow>\<star> y \<longleftrightarrow> (x,y) \<in> E\<^sup>*"
by (metis lift_reachable1_trancl reachable1_trancl)

lemma v\<^sub>0_reach_eq_vert:
  "v\<^sub>0 \<rightarrow>\<star> x \<longleftrightarrow> x \<in> V"
apply (rule iffI)
  apply blast
  apply (metis vert_rtrancl reachable_rtrancl v\<^sub>0_in_V)
done

lemma V_is_accessible:
  "V = {x. v\<^sub>0 \<rightarrow>\<star> x}"
by (simp add: v\<^sub>0_reach_eq_vert)

lemma V_is_accessible1:
  "V = {x. v\<^sub>0 \<rightarrow>+ x} \<union> {v\<^sub>0}"
using V_is_accessible
by blast

lemma reachable_induct [case_names self trans, induct pred: reachable, consumes 2]:
  assumes reach: "x \<rightarrow>\<star> y"
  and self: "P x"
  and trans: "\<And> y z. \<lbrakk> x \<rightarrow>\<star> y; P y; z \<in> succs y \<rbrakk> \<Longrightarrow> P z"
  shows "P y"
proof -
  from reach have "x \<in> V" by auto
  with reach have "(x,y) \<in> E\<^sup>*" using reachable_rtrancl by blast
  thus ?thesis
  proof induction
    case base with self show ?case .
  next
    case (step y z) with trans reachable_rtrancl[OF `x \<in> V`] edge_is_succs show ?case by simp 
  qed
qed

lemma reachable1_induct [case_names base trans, induct pred: reachable1]:
  assumes reach: "x \<rightarrow>+ y"
  and base': "\<And>y. y \<in> succs x \<Longrightarrow> P y"
  and trans: "\<And> y z. \<lbrakk> x \<rightarrow>+ y; P y; z \<in> succs y \<rbrakk> \<Longrightarrow> P z"
  shows "P y"
proof -
  from reach have "(x,y) \<in> E\<^sup>+" using reachable1_trancl by blast
  thus ?thesis
  proof induction
    case base with base' show ?case by (simp add: edge_is_succs)
  next
    case (step y z) with trans reachable1_trancl edge_is_succs show ?case by simp
  qed
qed

lemma reachable1_split:
  assumes "x \<rightarrow>+ v"
  obtains w where "x \<rightarrow>\<star> w" "v \<in> succs w"
proof -
  from assms self_reachable have "x \<rightarrow>\<star> x" by auto
  with assms that show ?thesis by induction auto
qed

lemma reachable1_succs_trancl:
    assumes "v \<in> V"
    shows "v \<rightarrow>+ v' \<longleftrightarrow> (v,v') \<in> {(v,v'). v' \<in> succs v}\<^sup>+" (is "?L \<longleftrightarrow> ?R")
proof
  assume ?L 
  thus ?R
    by (induct rule: reachable1_induct) (force intro: trancl_into_trancl)+
next
  assume ?R
  thus ?L
  proof (induct rule: trancl_induct)
    case base thus ?case by (simp add: succs_reachable[OF assms])
  next
    case (step y z) with succs_reachable have "y \<rightarrow>+ z" by auto
    with `v \<rightarrow>+ y` show ?case by (metis reachable1_trans)
  qed
qed

lemma reachable_succs_rtrancl:
  "v \<in> V \<Longrightarrow> v \<rightarrow>\<star> v' \<longleftrightarrow> (v,v') \<in> {(v,v'). v' \<in> succs v}\<^sup>*"
by (metis reachable1_succs_trancl lift_reachable1_trancl)

lemma sub_reachable:
  assumes "{(v,v'). v' \<in> succs' v} \<subseteq> {(v,v'). v' \<in> succs v}"
  and "x \<in> V"
  and sub_reach: "succs_digraph.sd_reachable succs' x v w"
  shows "v \<rightarrow>\<star> w"
proof -
  interpret sub!: succs_digraph succs' x by unfold_locales

 from assms have "sub.E \<subseteq> E" and "sub.V \<subseteq> V" using subgraph_E_sub subgraph_V_sub by simp_all
 hence "subgraph sub.sd_to_graph sd_to_graph"
    by (intro subgraphI, simp_all add: compatible_simple) unfold_locales
  with sub_reach show ?thesis using vwalk_subgraph_closed by (auto simp add: reachable_def)
qed

text {* x \<rightarrow>\<setminus>R+ y iff there is a path from x to way without nodes in R *}
(* we could've defined reachable_restr (\<rightarrow>\<setminus>R\<star>) too, but that additional case is trivial and unimportant *)
definition restr_reachable1  ("_ \<rightarrow>\<setminus>_+ _" [100,100,100] 40) where
  "restr_reachable1 v R w \<equiv> (v,w) \<in> (rel_restrict E R)\<^sup>+"

lemma restr_reachable1_reachable1:
  "v \<rightarrow>\<setminus>R+ w \<Longrightarrow> v \<rightarrow>+ w"
proof -
  assume "v \<rightarrow>\<setminus>R+ w"
  hence "(v,w) \<in> (rel_restrict E R)\<^sup>+" by (simp add: restr_reachable1_def)
  hence "(v, w) \<in> E\<^sup>+" by (metis rel_restrict_trancl_mem rel_restrict_lift)
  with reachable1_trancl show ?thesis by blast
qed

lemmas restr_reachable1_reachable1E[elim] = restr_reachable1_reachable1[elim_format]

lemma restr_reachable1_reachable:
  "x \<rightarrow>\<setminus>R+ y \<Longrightarrow> x \<rightarrow>\<star> y"
by (metis reachable1_reachable restr_reachable1_reachable1E)

lemmas restr_reachable1_reachableE[elim] = restr_reachable1_reachable[elim_format]

lemma restr_reachable1_empty[intro,simp]:
  "R = {} \<Longrightarrow> v \<rightarrow>\<setminus>R+ w \<longleftrightarrow> v \<rightarrow>+ w"
proof -
  assume "R = {}"
  hence "v \<rightarrow>\<setminus>R+ w \<longleftrightarrow> (v,w) \<in> E\<^sup>+" by (simp add: restr_reachable1_def)
  with reachable1_trancl show ?thesis by blast
qed

lemma restr_reachable1_no_verts:
  "R \<inter> V = {} \<Longrightarrow> v \<rightarrow>\<setminus>R+ w \<longleftrightarrow> v \<rightarrow>+ w"
proof -
  assume "R \<inter> V = {}"
  hence "R \<inter> Field E = {}" using edges_sub_nodes by (auto simp add: Field_def)
  hence "v \<rightarrow>\<setminus>R+ w \<longleftrightarrow> (v, w) \<in> E\<^sup>+" by (simp add: restr_reachable1_def rel_restrict_Int_empty)
  with reachable1_trancl show ?thesis by blast
qed

lemma restr_reachable_union:
  "R \<subseteq> S \<Longrightarrow> {y. x \<rightarrow>\<setminus>S+ y} \<subseteq> {y. x \<rightarrow>\<setminus>R+ y}"
unfolding restr_reachable1_def
by (smt Collect_mono rel_restrict_mono2 trancl_mono)

lemma restr_reachable_restrict:
  assumes "R \<subseteq> S"
  and "\<And>y. x \<rightarrow>\<setminus>R+ y \<Longrightarrow> P y"
  shows "\<And>y. x \<rightarrow>\<setminus>S+ y \<Longrightarrow> P y"
using assms restr_reachable_union
by blast

lemma restricted_target:
  "w \<in> R \<Longrightarrow> \<not> v \<rightarrow>\<setminus>R+ w"
proof (rule notI)
  assume "w \<in> R" "v \<rightarrow>\<setminus>R+ w"
  hence "(v, w) \<in> (rel_restrict E R)\<^sup>+" by (simp add: restr_reachable1_def)
  with rel_restrict_trancl_mem have" (v,w) \<in> rel_restrict (E\<^sup>+) R" by metis
  with rel_restrict_notR `w \<in> R` show "False" by metis
qed

lemma restricted_start:
  "v \<in> R \<Longrightarrow> \<not> v \<rightarrow>\<setminus>R+ w"
proof (rule notI)
  assume "v \<in> R" "v \<rightarrow>\<setminus>R+ w"
  hence "(v, w) \<in> (rel_restrict E R)\<^sup>+" by (simp add: restr_reachable1_def)
  with rel_restrict_trancl_mem have" (v,w) \<in> rel_restrict (E\<^sup>+) R" by metis
  with rel_restrict_notR `v \<in> R` show "False" by metis
qed

lemma succs_restricted:
  assumes "x \<in> V"
  and not_R: "x \<notin> R" "y \<notin> R"
  and succ: "y \<in> succs x"
  shows "x \<rightarrow>\<setminus>R+ y"
proof -
  from assms rel_restrictI have "(x,y) \<in> rel_restrict E R" by (auto dest!: succs_implies_edge)
  with r_into_trancl' have "(x,y) \<in> (rel_restrict E R)\<^sup>+" by blast
  thus ?thesis by (simp add: restr_reachable1_def)
qed

lemma restricted_induct [case_names base trans, induct pred: restr_reachable1]:
  assumes reach: "x \<rightarrow>\<setminus>R+ y"
  and base': "\<And>y. \<lbrakk> y \<in> succs x; x \<notin> R; y \<notin> R \<rbrakk> \<Longrightarrow> P y"
  and trans: "\<And>y z. \<lbrakk> x \<rightarrow>\<setminus>R+ y; P y; z \<in> succs y; x \<notin> R; y \<notin> R; z \<notin> R \<rbrakk> \<Longrightarrow> P z"
  shows "P y"
proof -
  from reach have "(x,y) \<in> (rel_restrict E R)\<^sup>+" using restr_reachable1_def by simp
  thus ?thesis
  proof induction
    case base with base' show ?case unfolding rel_restrict_def by (auto simp: edge_is_succs)
  next
    case (step y z) hence "x \<rightarrow>\<setminus>R+ y" by (simp add: restr_reachable1_def)
    moreover hence "x \<notin> R" "y \<notin> R" by (metis restricted_start restricted_target)+
    moreover from step edge_is_succs have "z \<in> succs y" "z \<notin> R" by (metis rel_restrict_lift rel_restrict_notR)+
    ultimately show ?case using step trans by metis
  qed
qed

lemma restr_reachable1_trans[trans]:
  "x \<rightarrow>\<setminus>R+ y \<Longrightarrow> y \<rightarrow>\<setminus>R+ z \<Longrightarrow> x \<rightarrow>\<setminus>R+ z"
by (metis restr_reachable1_def trancl_trans)

lemma reach_impl_restr_reach:
  assumes TRANS: "\<And> x y. x \<in> R \<Longrightarrow> x \<rightarrow>+ y \<Longrightarrow> y \<in> R"
  and "v \<rightarrow>+ w" "v \<notin> R" "w \<notin> R"
  shows "v \<rightarrow>\<setminus>R+ w"
using assms(2-)
proof induction
  case (base y) from `v \<rightarrow>+ w` have "v \<in> V" by blast 
  with base show ?case using succs_restricted by simp
next
  case (trans y z) hence "y \<rightarrow>+ z" using succs_reachable by blast
  with TRANS trans have "y \<notin> R" by auto
  with trans have "v \<rightarrow>\<setminus>R+ y" by simp
  also from trans `y \<notin> R` have "y \<rightarrow>\<setminus>R+ z" using succs_restricted by blast
  finally show ?case .
qed

lemma sub_reachable1_iff_reachable_restrict:
  assumes succs'_def: "\<And>v. succs' v = succs v - R"
  and v: "v \<in> V" "v \<notin> R"
  shows "succs_digraph.sd_reachable1 succs' v v w \<longleftrightarrow> v \<rightarrow>\<setminus>R+ w"
proof -
  interpret sub: succs_digraph succs' v
    by (unfold_locales)
  
  {
    fix x y
    assume x: "x \<notin> R" "x \<in> V" and E: "(x,y) \<in> sub.E"
    hence "y \<in> succs' x" by (simp add: sub.edge_is_succs)
    with x succs_implies_edge have "(x,y) \<in> E" and "y \<notin> R" by (auto simp add: succs'_def)
    with x have "(x,y) \<in> rel_restrict E R" using rel_restrictI by metis
    with trancl_mono have "(x,y) \<in> (rel_restrict E R)\<^sup>+" by auto
    hence "x \<rightarrow>\<setminus>R+ y" by (simp add: restr_reachable1_def)
  } note subE_to_reach = this


  show ?thesis
  proof
    assume "sub.sd_reachable1 v w" hence "(v, w) \<in> sub.E\<^sup>+" by (metis sub.reachable1_trancl)
    thus "v \<rightarrow>\<setminus>R+ w"
    proof induction
      case (base y) with subE_to_reach v show ?case by blast 
    next
      case (step y z) note `v \<rightarrow>\<setminus>R+ y`
      also from step have "y \<notin> R" and "y \<in> V" using restricted_target by blast+
      with step subE_to_reach have "y \<rightarrow>\<setminus>R+ z" by blast
      finally show ?case .
    qed
  next
    assume A: "v \<rightarrow>\<setminus>R+ w"
    
    with A show "sub.sd_reachable1 v w"
    proof induction
      case (base y) with succs'_def have "y \<in> succs' v" by simp
      with sub.succs_reachable v show ?case by blast
    next
      case (trans y z) hence "sub.sd_reachable1 v y" by simp
      also from trans succs'_def have "z \<in> succs' y" by simp
      with trans sub.succs_reachable have "sub.sd_reachable1 y z" by blast
      finally show ?case .
    qed
  qed
qed

lemma restrict_succs_digraph_V_restrict:
  assumes succs'_def: "\<And>v. succs' v = succs v - R"
  and w: "w \<in> V" "w \<notin> R"
  shows "succs_digraph.V succs' w \<subseteq> V - R"
proof -
  interpret sub!: succs_digraph succs' w by unfold_locales

  from edges_sub_nodes have E_trancl_sub: "E\<^sup>+ \<subseteq> (V\<times>V)\<^sup>+" by (blast intro: trancl_mono)
  have A_trancl: "\<And>A x. (A \<times> A)\<^sup>+ `` {x} \<subseteq> A" by (metis Image_subset subset_refl trancl_subset_Sigma)

  note Image_mono[intro]

  from sub.V_is_accessible1 have "sub.V = {x. sub.sd_reachable1 w x} \<union> {w}" by blast
  also from sub_reachable1_iff_reachable_restrict[OF assms] have "... = {x. w \<rightarrow>\<setminus>R+ x} \<union> {w}" by blast
  also from restr_reachable1_def have "... = (rel_restrict E R)\<^sup>+ `` {w} \<union> {w}" by blast
  also have "... \<subseteq> rel_restrict (E\<^sup>+) R `` {w} \<union> {w}" using rel_restrict_trancl_sub by blast
  also have "... \<subseteq> rel_restrict ((V \<times> V)\<^sup>+) R `` {w} \<union> {w}" using rel_restrict_mono[OF E_trancl_sub] by blast
  also have "... \<subseteq> ((V - R) \<times> (V - R))\<^sup>+ `` {w} \<union> {w}" using rel_restrict_Sigma_sub by blast
  also have "... \<subseteq> (V - R) \<union> {w}" using A_trancl by force
  also have "... \<subseteq> V - R" using w by blast
  finally show ?thesis .
qed

lemma restr_reachable1_alt_def:
  "restr_reachable1 v R w \<longleftrightarrow> (\<exists>p. ewalk v p w \<and> p \<noteq> [] \<and> set (ewalk_verts v p) \<inter> R = {})"
proof -
  def s \<equiv> "\<lambda>x. succs x - R"
  hence s: "\<And>v. s v = succs v - R" by blast
  interpret sub!: succs_digraph s v by unfold_locales

  have "{(v, v'). v' \<in> s v} \<subseteq> {(v, v'). v' \<in> succs v}"
    by (auto simp add: s_def)
  note E = subgraph_E_sub[OF this] and V = subgraph_V_sub[OF this]
  
  show ?thesis
  proof
    case goal1 hence "v \<in> V - R " using restricted_start by auto
    with restrict_succs_digraph_V_restrict[OF s] have V_restr: "sub.V \<subseteq> V - R" by blast

    from goal1 E V have "sub.E \<subseteq> E" and "sub.V \<subseteq> V" by blast+
    hence subgraph: "subgraph sub.sd_to_graph sd_to_graph"
      by (intro subgraphI, simp_all add: compatible_simple) unfold_locales
    
    have ev_eq: "\<And>u p. sub.ewalk_verts u p = ewalk_verts u p"
    unfolding ewalk_verts_def sub.ewalk_verts_def
    by simp_all

    from `v \<in> V - R` sub_reachable1_iff_reachable_restrict[THEN iffD2, OF s] goal1 have "sub.reachable1 v w" by blast
    then obtain p where p: "sub.ewalk v p w" "p \<noteq> []" unfolding sub.reachable1_def by auto
    hence "set (sub.ewalk_verts v p) \<subseteq> sub.V" unfolding sub.ewalk_conv by auto
    with ev_eq have "set (ewalk_verts v p) \<subseteq> sub.V" by simp
    with V_restr have "set (ewalk_verts v p) \<inter> R = {}" by auto

    moreover from p subgraph have "ewalk v p w" by (simp add: subgraph_ewalk_imp_ewalk)
    ultimately show ?case using p by auto
  next
    case goal2
    then obtain p where p: "ewalk v p w" "p \<noteq> []" "set (ewalk_verts v p) \<inter> R = {}" by auto
    hence "v \<rightarrow>+ w" using reachable1_def by auto

    from p have "ewhd v p = v" by (simp add: ewalk_conv)
    with ewhd_in_verts have "v \<in> set (ewalk_verts v p)" by metis
    with p have "v \<notin> R" by auto

    from p p(1) have "sub.reachable1 v w"
    proof (induction rule: ewalk_induct_rev)
      case (Snoc w1 w2 e es) thus ?case
      proof (cases "es = []")
        case True with Snoc have "ewalk v [e] w2" by simp
        with Snoc have "w1 = v" "(v, w2) \<in> E" by (simp_all add: ewalk_def)
        with edge_is_succs have "w2 \<in> succs v" by simp
        
        moreover
        from Snoc have "w2 = ewlast v (es@[e])" by (simp add: ewalk_conv)
        with Snoc have "w2 \<notin> R" using ewlast_in_verts by blast

        ultimately have "w2 \<in> s v" by (auto simp add: s)
        with sub.succs_implies_edge have "(v, w2) \<in> sub.E" by simp
        with sub.edge_implies_reach1 show ?thesis by auto
      next
        case False 
        moreover
        from Snoc ewalk_verts_join_l have "ewalk_verts v (es@[e]) = ewalk_verts v es @ tl (ewalk_verts w [e])" by simp
        hence "set (ewalk_verts v es) \<subseteq> set (ewalk_verts v (es@[e]))" by simp
        with Snoc have "set (ewalk_verts v es) \<inter> R = {}" by auto
        
        moreover
        from Snoc edge_implies_ewalk have "ewalk w1 [e] w2" by auto
        with ewalk_join_ewalk_l' Snoc have "ewalk v es w1" by metis

        ultimately have "sub.reachable1 v w1" using Snoc.IH by blast
        
        also
        hence "w1 \<in> sub.V" by blast

        from Snoc ewalk_edge_in_ewalk_verts[where e=e] have "{w1, w2} \<subseteq> set (ewalk_verts v (es@[e]))" by simp
        with Snoc have "w1 \<notin> R" "w2 \<notin> R" by auto
        with edge_is_succs s Snoc have "w2 \<in> s w1" by auto
        with `w1 \<in> sub.V` sub.succs_implies_edge have "(w1, w2) \<in> sub.E" by simp
        with sub.edge_implies_reach1 have "sub.reachable1 w1 w2" by auto
        
        finally show ?thesis .
      qed
    qed simp
    with `v \<notin> R` `v \<rightarrow>+ w` sub_reachable1_iff_reachable_restrict[THEN iffD1, OF s] show ?case by blast
  qed
qed

lemma restr_reachable1I:
  "ewalk v p w \<Longrightarrow> p \<noteq> [] \<Longrightarrow> set (ewalk_verts v p) \<inter> R = {} \<Longrightarrow> v \<rightarrow>\<setminus>R+ w"
unfolding restr_reachable1_alt_def by auto

lemma ewalk_dissect:
  assumes p[intro!,simp]: "ewalk x p v"
  and ne[intro!, simp]: "p \<noteq> []"
  obtains bp a where "bp = butlast p" and 
                     "p = bp@[(a,v)]" and 
                     "(a,v) \<in> E" and
                     "ewalk x bp a" and
                     "ewalk_verts x p = ewalk_verts x bp @ [v]"
proof -
  def bp \<equiv> "butlast p"

  moreover
  then obtain a b where ab: "a = fst (last p)" "b = snd (last p)" by auto
  hence p_split: "p = bp@[(a,b)]" by (simp add: bp_def)

  from p have "ewlast x p = v" unfolding ewalk_conv by simp
  with p_split have "ewlast x (bp@[(a,b)]) = v" by simp
  with ewlast_join_edge have "b = v" by simp
  with p_split have "p = bp@[(a,v)]" by simp
  
  moreover
  with p ewalk_verts_join_l have "ewalk_verts x p = ewalk_verts x bp @ tl (ewalk_verts x [(a,v)])" by simp
  hence ev_p_split: "ewalk_verts x p = ewalk_verts x bp @ [v]" by (simp add: ewalk_verts_def)
  
  moreover
  from ab `b = v` have "last p = (a,v)" by simp
  with p have "(a,v) \<in> E" unfolding ewalk_conv using last_in_set[OF ne] by auto

  moreover
  hence "ewalk a [(a,v)] v" using edge_implies_ewalk by force

  moreover
  note e = this
  from p p_split `b = v` have "ewalk x (bp @[(a,v)]) v" by simp
  with e have "ewalk x bp a" using ewalk_join_ewalk_l' by blast

  ultimately show ?thesis using that by simp
qed

(* lift it *)
lemma ewalk_verts_in_edges:
  assumes "ewalk u p v"
  and "ewalk_verts u p = xs @ a # b # ys"
  shows "(a,b) \<in> E"
proof -
  from ewalk_verts_are_edges[OF assms] obtain e where "e \<in> E" "fst e = a" "snd e = b" by auto
  thus ?thesis by auto
qed

lemma epath_verts_sub_ewalk_verts:
  assumes "ewalk u p v"
  shows "set (ewalk_verts u (ewalk_to_epath p)) \<subseteq> set (ewalk_verts u p)"
using assms
proof (induct rule: ewalk_to_epath_induct)
  case (path p) hence "ewalk_to_epath p = p" by (auto simp add: ewalk_to_epath.simps)
  thus ?case by simp
next
  case (decomp p q r s) with step_ewalk_to_epath[symmetric] have "set (ewalk_verts u (ewalk_to_epath p)) \<subseteq> set (ewalk_verts u (q@s))" by simp
  also from decomp(1-3) have "... \<subseteq> set (ewalk_verts u p)"
  proof (elim cycle_decompE)
    assume p_def:"p = q @ r @ s" and "\<exists>w. ewalk u q w \<and> ewalk w r w \<and> ewalk w s v"
    then obtain w where "ewalk u q w" "ewalk w r w" "ewalk w s v" by blast
    hence "ewalk u (q@s) v" and "ewalk u (q@r) w" and "ewalk u (q@r@s) v" by (blast intro: ewalk_ewalk_joinI)+
    hence qs:"ewalk_verts u (q@s) = ewalk_verts u q @ tl (ewalk_verts u s)" and
          "ewalk_verts u (q@r) = ewalk_verts u q @ tl (ewalk_verts u r)" by (blast intro: ewalk_verts_join_l)+
    with `ewalk u (q@r@s) v` have "ewalk_verts u (q@r@s) = ewalk_verts u q @ tl (ewalk_verts u r) @ tl (ewalk_verts u s)" using ewalk_verts_join_l[where p="q@r"] by bestsimp
    with qs and p_def show ?thesis by auto
  qed
  finally show ?case .
qed

lemma restr_reachable1_ne_split:
  assumes "x \<noteq> v"
  and "x \<rightarrow>\<setminus>R+ v"
  and "v \<notin> succs x"
  obtains w where "w \<in> (succs x - R)" "w \<rightarrow>\<setminus>insert x R+ v"
proof -
  from assms(2) obtain p where p: "ewalk x p v" "p \<noteq> []" "set (ewalk_verts x p) \<inter> R = {}" unfolding restr_reachable1_alt_def by blast

  def pp \<equiv> "ewalk_to_epath p"
  with p have ep: "epath x pp v" using epath_ewalk_to_epath by simp
  with p have "pp \<noteq> []" using epath_empty_iff assms(1) by auto
  then obtain e es where es: "pp = e#es" by (metis list.exhaust)
  with ep have "e \<in> E" "fst e = x" "epath (snd e) es v" and not_X: "x \<notin> set (ewalk_verts (snd e) es)" using epath_Cons_iff by auto
  hence "ewalk (snd e) es v" by (simp add: epath_def)

  moreover
  from ep have "ves pp" by (auto simp add: epath_def ewalk_imp_ves)
  with es have "set (ewalk_verts (snd e) es) \<subseteq> set (ewalk_verts x pp)" using ewalk_verts_Cons by auto
  with p pp_def epath_verts_sub_ewalk_verts[of x p v] have "set (ewalk_verts (snd e) es) \<inter> R = {}" by auto
  with not_X have "set (ewalk_verts (snd e) es) \<inter> (insert x R) = {}" by auto
  
  moreover
  have "es \<noteq> []"
  proof (rule notI)
    assume "es = []"
    with ep es have "ewalk x [e] v" by (simp add: epath_def)
    hence "(x,v) \<in> E" using ewalk_Cons_iff[of x e "[]"] by auto
    with assms(3) show "False" by (simp add: edge_is_succs)
  qed

  ultimately have "snd e \<rightarrow>\<setminus>(insert x R)+ v" by (blast intro: restr_reachable1I)
  moreover with restricted_start have "snd e \<notin> R" by auto
  with `e \<in> E` `fst e = x` have "snd e \<in> succs x - R" by (auto simp add: edge_is_succs)
  ultimately show ?thesis using that by metis
qed

lemma restr_reachable1_split:
  assumes "x \<rightarrow>\<setminus>R+ v"
  and "v \<notin> succs x"
  obtains w where "w \<in> (succs x - R)" "w \<rightarrow>\<setminus>R+ v"
proof (cases "x = v")
  case False with restr_reachable1_ne_split assms obtain w where "w \<in> succs x - R" "w \<rightarrow>\<setminus>insert x R+v" by metis
  with restr_reachable_restrict[where R=R and S="insert x R"] that show ?thesis by auto
next
  case True note x_v = this
  moreover
  with assms(1) obtain p where p: "ewalk x p x" "p \<noteq> []" "set (ewalk_verts x p) \<inter> R = {}" unfolding restr_reachable1_alt_def by blast
  then obtain e es where es: "p = e#es" by (metis list.exhaust)
  with p have ep: "e \<in> E" "fst e = x" "ewalk (snd e) es x" using ewalk_Cons_iff by auto
  from p es have notR:"snd e \<notin> R" "x \<notin> R" unfolding ewalk_verts_def by auto
  
  moreover from ep have "snd e \<in> succs x" by (auto simp add: edge_is_succs)

  moreover 
  have "snd e \<rightarrow>\<setminus>R+ x"
  proof (cases "x \<in> succs (snd e)")
    case True from ep have "snd e \<in> V" using edge_in_nodes by (cases e) simp
    with True notR show ?thesis using succs_restricted by auto
  next
    case False
    
    have "es \<noteq> []"
    proof (rule notI)
      assume "es = []"
      with p es have "ewalk x [e] x" by simp
      hence "snd e = x" by (simp add: ewalk_def)
      with `snd e \<in> succs x` x_v assms(2) show "False" by metis
    qed

    from p have "ves p" by (auto simp add: ewalk_imp_ves)
    with es have "set (ewalk_verts (snd e) es) \<subseteq> set (ewalk_verts x p)" using ewalk_verts_Cons by auto
    with p have "set (ewalk_verts (snd e) es) \<inter> R = {}" by auto
    
    with `es \<noteq> []` ep show ?thesis by (blast intro: restr_reachable1I)
  qed
  ultimately show ?thesis using that by blast
qed

lemma ewalk_verts_vwalk_to_ewalk:
  "xs \<noteq> [] \<Longrightarrow> vwalk xs \<G> \<Longrightarrow> hd xs = u \<Longrightarrow> ewalk_verts u (vwalk_to_ewalk xs) = xs"
unfolding vwalk_to_ewalk_def ewalk_verts_def
proof (induct xs rule: list_nonempty_induct)
  case (single x) thus ?case by simp
next
  case (cons x xs) hence xu: "x = u" by simp
  note thms = cons.prems `xs \<noteq> []`

  moreover have *: "\<And>x y. (x,y) \<in> E \<Longrightarrow> (SOME e. e \<in> E \<and> e = (x, y)) = (x, y)" by (intro some_equality) simp_all
  moreover from thms have "set (vwalk_edges (x#xs)) \<subseteq> E" unfolding vwalk_def by simp
  moreover with thms have "set (vwalk_edges xs) \<subseteq> E" by simp
  with `xs \<noteq> []` have "map (snd \<circ> (\<lambda>ee. SOME e. e \<in> E \<and> e = ee)) (vwalk_edges xs) = tl xs" by (induct xs rule: list_nonempty_induct) (simp_all add: *)
  ultimately show ?case by simp
qed

lemma vwalk_restr_reachable1_intro:
  assumes "vwalk xs \<G>"
  and "y \<in> set xs"
  and "set xs \<inter> R = {}"
  and "y \<noteq> last xs"
  shows "y \<rightarrow>\<setminus>R+ last xs"
proof -
  from vwalk_join_split[OF assms(2)] obtain q r where xs_split: "xs = q \<oplus> r" "joinable q r" and r_def: "last r = last xs" "hd r = y" by (auto simp add: vwalk_join_last)
  with vwalk_join_vwalk2 assms(1,3) have "vwalk r \<G>" by simp
  with vwalk_has_ewalk r_def have "ewalk y (vwalk_to_ewalk r) (last xs)" by simp
  
  moreover from xs_split have "r \<noteq> []" unfolding joinable_def by simp
  with r_def assms(4) have "tl r \<noteq> []" by (metis hd_rev singleton_rev_conv tl_obtain_elem)
  hence "vwalk_edges r \<noteq> []" by (induct r) simp_all
  hence "vwalk_to_ewalk r \<noteq> []" unfolding vwalk_to_ewalk_def by simp

  moreover from xs_split have "xs = butlast q @ r" by (simp add: vwalk_join_def2)
  with assms(3) have "set r \<inter> R = {}" by auto

  moreover from `vwalk r \<G>` ewalk_verts_vwalk_to_ewalk r_def `r \<noteq> []` have "ewalk_verts y (vwalk_to_ewalk r) = r" by simp

  ultimately show ?thesis by (auto intro!: restr_reachable1I)
qed
end

(* Make it finite ... *)
locale finite_succs_digraph = succs_digraph succs v\<^sub>0
  for succs and v\<^sub>0 :: "'n" +
  assumes finite_V[simp]: "finite V"
begin

lemma finite_E[simp]:
  "finite E"
proof -
  from finite_V have "finite (V \<times> V)" by auto
  with edges_sub_nodes finite_subset show ?thesis by blast
qed

lemma finite_succs_subgraphI:
  assumes "{(v, v'). v' \<in> succs' v} \<subseteq> {(v, v'). v' \<in> succs v}"
  and "w \<in> V"
  shows "finite_succs_digraph succs' w"
proof -
  from assms have "succs_digraph.V succs' w \<subseteq> V" by (simp add: subgraph_V_sub)
  thus ?thesis by unfold_locales (metis finite_subset finite_V)
qed

lemma succs_finite [intro]: "x \<in> V \<Longrightarrow> finite (succs x)"
by (blast intro: finite_subset succs_in_V finite_V)
end

sublocale finite_succs_digraph \<subseteq> pseudo_digraph "\<G>"
by unfold_locales auto

(*
subsection {* Subgraphs *}
locale finite_subgraph = G\<^sub>1: finite_digraph V E + G\<^sub>2: finite_digraph V' E'
  for V V' :: "'n set" and  E E' :: "('n \<times> 'n) set" +
  assumes sub: "E' \<subseteq> E"
  and restricted: "V' = Field E'"
begin

notation G\<^sub>1.fd_reachable ("_ \<rightarrow>\<^sub>1\<star> _" [100,100] 40)
notation G\<^sub>1.fd_reachable1 ("_ \<rightarrow>\<^sub>1+ _" [100,100] 40)
notation G\<^sub>1.fd_to_graph ("\<G>\<^sub>1")
(* this needs to be an abbreviation for some reason *)
abbreviation g2_fd_to_graph ("\<G>\<^sub>2") where
  "g2_fd_to_graph \<equiv> G\<^sub>2.fd_to_graph"
abbreviation g2_fd_reachable ("_ \<rightarrow>\<^sub>2\<star> _" [100,100] 40) where
  "g2_fd_reachable \<equiv> G\<^sub>2.fd_reachable"
abbreviation g2_fd_reachable1 ("_ \<rightarrow>\<^sub>2+ _" [100,100] 40) where
  "g2_fd_reachable1 \<equiv> G\<^sub>2.fd_reachable1"

lemma vertices_subset:
  "V' \<subseteq> V"
proof -
  from restricted sub have "V' \<subseteq> Field E" by (metis mono_Field)
  with G\<^sub>1.valid_edges show ?thesis unfolding Field_def by auto
qed

lemma subgraph[intro]:
  "subgraph \<G>\<^sub>2 \<G>\<^sub>1"
using sub
by (intro subgraphI, auto simp add: compatible_simple vertices_subset) unfold_locales

lemma ewalk2_implies_ewalk:
  "G\<^sub>2.ewalk u p v \<Longrightarrow> G\<^sub>1.ewalk u p v"
by (rule G\<^sub>1.subgraph_ewalk_imp_ewalk[OF _ subgraph])

lemma reach2_implies_reach:
  "v \<rightarrow>\<^sub>2+ w \<Longrightarrow> v \<rightarrow>\<^sub>1+ w"
by (metis ewalk2_implies_ewalk G\<^sub>2.reachable1_def G\<^sub>1.reachable1_def)

lemma ewalk_verts_eq:
  "G\<^sub>1.ewalk_verts u p = G\<^sub>2.ewalk_verts u p"
unfolding G\<^sub>1.ewalk_verts_def G\<^sub>2.ewalk_verts_def
by simp

end*)
end
