(* Author: René Neumann <rene.neumann@in.tum.de> *)
header {* A nondeterministic DFS implementation *}

theory DFS_old
imports
  Main
  "../Graph_Ext"
  "../Refine_Additions"
  "../../General/Misc_Additions"
begin

text {* Based on the work on the nondeterministic while combinator
we develop a depth-first-search function that works on sets rather than
lists. This collapses different runs of the search into a set of results,
which share a common property. *}

subsection {* Data structures and algorithm specification *}

text {* The representation of a working state. *}
record ('n) dfs_ws =
  start :: "'n"  -- "the starting node"
  stack :: "'n list"   -- "the search stack of the current run"
  wl :: "'n set list" -- "the set of successors that still need to visited for each node"
  discover :: "('n, nat) map" -- "mapping a node to the timestamp of their discovery"
  finish :: "('n, nat) map" -- "mapping a node to the timestamp of the finishing, i.e. the moment the search backtracks"
  counter :: "nat" -- "the counter for discover and finish"

abbreviation "discovered s \<equiv> dom (discover s)"
abbreviation "finished s \<equiv> dom (finish s)"
abbreviation disc ("\<delta>") where "disc s x \<equiv> the (discover s x)"
abbreviation fin ("\<phi>") where "fin s x \<equiv> the (finish s x)"

text {* As the DFS algorithm is intended to be only a framework for
DFS-based algorithm, we need to carry around a state, that is used
by the concrete implementation. *}
record ('S, 'n) dfs_sws = "('n) dfs_ws" +
   state :: "'S" -- "the state of the implementation"

text {* The representation of the parametrized algorithm.*}
(* should roughly be read as:

  dfs_algorithm s e:
   if (dfs_cond s)
     s' := dfs_action s e
     for e' in succs e do
        if (e' already visited)
           s' := dfs_remove s' e'
        else
           s' := dfs_algorithm s' e'
        fi
     done
     s' := dfs_post s' e
     return s'
   else
     return s
   fi
*)
     
record ('S, 'n) dfs_algorithm =
  dfs_cond :: "'S \<Rightarrow> bool"  -- "the condition to be satisified by the state S to continue searching from here."
  dfs_action :: "'S \<Rightarrow> ('S,'n) dfs_sws \<Rightarrow> 'n \<Rightarrow> 'S"  -- "modifies the state for the current node BEFORE visiting the successors."
  dfs_post :: "'S \<Rightarrow> ('S, 'n) dfs_sws \<Rightarrow> 'n \<Rightarrow> 'S"  -- "modifies the state for the current node AFTER having visited the successors (i.e. during backtracking)."
  dfs_remove :: "'S \<Rightarrow> ('S, 'n) dfs_sws \<Rightarrow> 'n \<Rightarrow> 'S" -- "modifies the state if a node has already been visited and is removed from the stack"
  dfs_start :: "'n \<Rightarrow> 'S" -- "the starting state"
  dfs_restrict :: "'n set" -- "a set of states that should not be visited"

context finite_digraph
begin

text {* The @{term dfs_step} returns a set of working states, that may be reached from the current node. In general, it returns one state for each successor of the current node. *}

(* we define it first in an unfolded fashion, as primrec only allows one pattern at the lhs *)
fun dfs_step' :: "('S, 'n, 'X) dfs_algorithm_scheme \<Rightarrow> ('S,'n) dfs_sws \<Rightarrow> 'S \<Rightarrow> 'n \<Rightarrow> ('n, nat) map \<Rightarrow> ('n, nat) map \<Rightarrow> nat \<Rightarrow> 'n set list \<Rightarrow> 'n list \<Rightarrow> (('S, 'n) dfs_sws) set"
where "dfs_step' dfs S s n d f c w [] = {}"
     |"dfs_step' dfs S s n d f c w (x#xs) = (if hd w = {} then {dfs_sws.make n xs (tl w) d (f(x \<mapsto> c)) (Suc c) (dfs_post dfs s S x)}
                                           else {dfs_sws.make n st' w' d' f c' s' | e st' w' d' c' s'. e \<in> hd w \<and> 
                                                   (if e \<in> dfs_restrict dfs then st' = x#xs \<and> w' = (hd w - {e})#tl w \<and> d' = d \<and> c' = c \<and> s' = s 
                                                    else if e \<in> dom d then st' = x#xs \<and> w' = (hd w - {e})#tl w \<and> d' = d \<and> c' = c \<and> s' = dfs_remove dfs s S e
                                                    else st' = e#x#xs \<and> w' = succs e#(hd w - {e})#tl w \<and> d' = d(e \<mapsto> c) \<and> c' = Suc c \<and> s' = dfs_action dfs s S e)})"

(* now wrap it *)
definition dfs_step :: "('S, 'n, 'R) dfs_algorithm_scheme \<Rightarrow> ('S, 'n) dfs_sws \<Rightarrow> ('S ,'n) dfs_sws set"
where "dfs_step dfs s \<equiv> dfs_step' dfs s (state s) (start s) (discover s) (finish s) (counter s) (wl s) (stack s)"

lemma dfs_step_simps[simp]:
  "stack s = [] \<Longrightarrow> dfs_step dfs s = {}"
  "stack s = x#xs \<Longrightarrow> wl s \<noteq> [] \<Longrightarrow> hd (wl s) = {} 
    \<Longrightarrow> dfs_step dfs s = {s\<lparr>stack := xs, wl := tl (wl s), finish := (finish s)(x \<mapsto> counter s), counter := Suc (counter s), state := dfs_post dfs (state s) s x \<rparr>}"
  "stack s = x#xs \<Longrightarrow> wl s \<noteq> [] \<Longrightarrow> hd (wl s) \<noteq> {}
     \<Longrightarrow> dfs_step dfs s = {dfs_sws.make (start s) st' w' d' (finish s) c' s' | e st' w' d' c' s'. e \<in> hd (wl s) \<and> 
                           (if e \<in> dfs_restrict dfs then st' = x#xs \<and> w' = (hd (wl s) - {e})#tl (wl s) \<and> d' = discover s \<and> c' = counter s \<and> s' = (state s)
                            else if e \<in> discovered s then st' = x#xs \<and> w' = (hd (wl s) - {e})#tl (wl s) \<and> d' = discover s \<and> c' = counter s \<and> s' = dfs_remove dfs (state s) s e
                            else st' = e#x#xs \<and> w' = succs e#(hd (wl s) - {e})#tl (wl s) \<and> d' = (discover s)(e \<mapsto> counter s) \<and> c' = Suc (counter s) \<and> s' = dfs_action dfs (state s) s e)}"
unfolding dfs_step_def
by (simp_all add: dfs_sws.defs)

lemma dfs_step_intros:
  "stack s = x#xs \<Longrightarrow> wl s \<noteq> [] \<Longrightarrow> hd (wl s) = {} \<Longrightarrow> s' = s\<lparr>stack := xs, wl := tl (wl s), finish := (finish s)(x \<mapsto> counter s), counter := Suc (counter s), state := dfs_post dfs (state s) s x \<rparr> \<Longrightarrow> s' \<in> dfs_step dfs s"
  "stack s = x#xs \<Longrightarrow> wl s \<noteq> [] \<Longrightarrow> hd (wl s) \<noteq> {} \<Longrightarrow> e \<in> hd (wl s) \<Longrightarrow> e \<in> dfs_restrict dfs \<Longrightarrow> s' = s\<lparr>wl := (hd (wl s) - {e})#tl (wl s)\<rparr> \<Longrightarrow> s' \<in> dfs_step dfs s"
  "stack s = x#xs \<Longrightarrow> wl s \<noteq> [] \<Longrightarrow> hd (wl s) \<noteq> {} \<Longrightarrow> e \<in> hd (wl s) \<Longrightarrow> e \<notin> dfs_restrict dfs \<Longrightarrow> e \<in> discovered s \<Longrightarrow> s' = s\<lparr>wl := (hd (wl s) - {e})#tl (wl s), state := dfs_remove dfs (state s) s e\<rparr> \<Longrightarrow> s' \<in> dfs_step dfs s"
  "stack s = x#xs \<Longrightarrow> wl s \<noteq> [] \<Longrightarrow> hd (wl s) \<noteq> {} \<Longrightarrow> e \<in> hd (wl s) \<Longrightarrow> e \<notin> dfs_restrict dfs \<Longrightarrow> e \<notin> discovered s \<Longrightarrow> s' = s\<lparr> state := dfs_action dfs (state s) s e, stack := e#x#xs, wl := succs e#(hd (wl s) - {e})#(tl (wl s)), discover := (discover s)(e \<mapsto> counter s), counter := Suc (counter s)\<rparr> \<Longrightarrow> s' \<in> dfs_step dfs s"
unfolding dfs_step_def
apply (simp_all add: dfs_sws.defs)
  apply (rule exI[where x=e], simp)+
done

(* Several case-constructs on dfs_step. From simple to good :) *)
lemma dfs_step_cases:
  assumes inv: "stack s \<noteq> [] \<Longrightarrow> wl s \<noteq> []"
  and "\<lbrakk> stack s = [] \<rbrakk> \<Longrightarrow> P"
  and "\<And> x xs w ws . \<lbrakk> stack s = x#xs; wl s = w#ws; w = {} \<rbrakk> \<Longrightarrow> P"
  and "\<And> x xs w ws. \<lbrakk> stack s = x#xs; wl s = w#ws; w \<noteq> {} \<rbrakk> \<Longrightarrow> P"
  shows "P"
using assms
proof (cases "stack s")
  case (Cons x xs) with inv obtain w ws where "w#ws = wl s" by (cases "wl s") auto
  with Cons assms(2-) show ?thesis by (cases "w = {}") metis+
qed

lemma dfs_step_cases_elem [case_names empty restrict remove visit, cases pred: dfs_step, consumes 2]:
  assumes inv: "s' \<in> dfs_step dfs s" "wl s \<noteq> []"
  and empty: "\<And>x xs w ws. \<lbrakk> stack s = x#xs; wl s = w#ws; w = {} \<rbrakk> \<Longrightarrow> 
     s' = s\<lparr>stack := xs, wl := ws, finish := (finish s)(x \<mapsto> counter s), counter := Suc (counter s), state := dfs_post dfs (state s) s x \<rparr> \<Longrightarrow> P"
  and restrict: "\<And> e x xs w ws. \<lbrakk> stack s = x#xs; wl s = w#ws; w \<noteq> {}; e \<in> w; e \<in> dfs_restrict dfs \<rbrakk> \<Longrightarrow>
     s' = s\<lparr> wl := (w - {e})#ws \<rparr> \<Longrightarrow> P"
  and remove: "\<And> e x xs w ws. \<lbrakk> stack s = x#xs; wl s = w#ws; w \<noteq> {}; e \<in> w; e \<notin> dfs_restrict dfs; e \<in> discovered s \<rbrakk> \<Longrightarrow>
     s' = s\<lparr> wl := (w - {e})#ws, state := dfs_remove dfs (state s) s e\<rparr> \<Longrightarrow> P"
  and visit: "\<And> e x xs w ws. \<lbrakk> stack s = x#xs; wl s = w#ws; w \<noteq> {}; e \<in> w; e \<notin> dfs_restrict dfs; e \<notin> discovered s \<rbrakk> \<Longrightarrow>
     s' = s\<lparr> state := dfs_action dfs (state s) s e, stack := e#x#xs, wl := succs e#(w - {e})#ws, discover := (discover s)(e \<mapsto> counter s), counter := Suc (counter s)\<rparr> \<Longrightarrow> P"
  shows "P"
using inv
proof (cases s rule: dfs_step_cases)
  case "3" then show ?thesis using empty[OF "3"] inv
    using dfs_step_simps(2)
    by fastforce
next
  case ("4" x xs w ws) with inv dfs_step_simps(3)[of s x xs dfs]
  obtain e where e:
    "e \<in> w"
    "if e \<in> dfs_restrict dfs then stack s' = x#xs \<and> wl s' = (hd (wl s) - {e})#tl (wl s) \<and> discover s' = discover s \<and> counter s' = counter s \<and> state s' = state s \<and> start s' = start s \<and> finish s' = finish s
     else if e \<in> discovered s 
     then stack s' = x#xs \<and> wl s' = (hd (wl s) - {e})#tl (wl s) \<and> discover s' = discover s \<and> counter s' = counter s \<and> state s' = dfs_remove dfs (state s) s e \<and> start s' = start s \<and> finish s' = finish s
     else stack s' = e#x#xs \<and> wl s' = succs e#(hd (wl s) - {e})#tl (wl s) \<and> discover s' = (discover s)(e \<mapsto> counter s) \<and> counter s' = Suc (counter s) \<and> state s' = dfs_action dfs (state s) s e \<and> start s' = start s \<and> finish s' = finish s"
    by (simp add: dfs_sws.defs) force

  show ?thesis
  proof (cases "e \<in> dfs_restrict dfs")
    case True with "4" e restrict[OF "4" e(1)] show ?thesis by simp
  next
    case False with "4" e remove[OF "4" e(1)] visit[OF "4" e(1)] show ?thesis by (cases "e \<in> discovered s") simp_all
  qed
qed simp_all

(* Some nice induction lemmas to show some \<forall>n < length (stack s)... properties *)
lemma stack_wl_remove_induct [case_names "0" greater, consumes 5]:
  assumes "stack s = x#xs" and "wl s = w#ws"
  and "stack s' = x#xs" and "wl s' = (w - {e})#ws"
  and inv: "\<forall> n < length (stack s). P (stack s ! n) (wl s ! n) (discover s) (finish s) (start s)"
  and case_0: "P x w (discover s) (finish s) (start s) \<Longrightarrow> P x (w - {e}) (discover s') (finish s') (start s')"
  and step: "\<And> n. \<lbrakk> n < length (stack s); n > 0; P (stack s ! n) (wl s ! n) (discover s) (finish s) (start s) \<rbrakk> \<Longrightarrow> P (stack s ! n) (wl s ! n) (discover s') (finish s') (start s')"
  shows "\<forall>n < length (stack s'). P (stack s' ! n) (wl s' ! n) (discover s') (finish s') (start s')"
proof (rule allI, rule impI)
  fix n
  assume lt': "n < length (stack s')" with assms have lt: "n < length (stack s)" by simp
  show "P (stack s' ! n) (wl s' ! n)(discover s') (finish s') (start s') "
  proof (cases n)
    case "0" with assms lt' show ?thesis by auto
  next
    case (Suc m) with lt step[of n] inv have "P (stack s ! n) (wl s ! n) (discover s') (finish s') (start s')" by simp
    with Suc assms(1-4) show ?thesis by auto
  qed
qed

lemma stack_wl_visit_induct [case_names case_0 case_1 greater, consumes 5]:
  assumes "stack s = x#xs" and "wl s = w#ws"
  and "stack s' = e#x#xs" and "wl s' = succs e#(w - {e})#ws"
  and inv: "\<forall> n < length (stack s). P (stack s ! n) (wl s ! n) (discover s) (finish s) (start s)"
  and case_0: "P e (succs e) (discover s') (finish s') (start s')"
  and case_1: "P x w (discover s) (finish s) (start s) \<Longrightarrow> P x (w - {e}) (discover s') (finish s') (start s')"
  and step: "\<And> n. \<lbrakk> n < length (stack s); n > 0; P (stack s ! n) (wl s ! n) (discover s) (finish s) (start s) \<rbrakk> \<Longrightarrow> P (stack s ! n) (wl s ! n) (discover s') (finish s') (start s')"
  shows "\<forall>n < length (stack s'). P (stack s' ! n) (wl s' ! n) (discover s') (finish s') (start s')"
proof (rule allI, rule impI)
  fix n
  assume lt': "n < length (stack s')" hence lt: "n < Suc (length (stack s))" using assms(1,3) by simp
  
  show "P (stack s' ! n) (wl s' ! n) (discover s') (finish s') (start s')"
  proof (cases n)
    case "0" with assms lt' show ?thesis by simp
  next
    case (Suc m) note suc_m = this
    show ?thesis
    proof (cases m)
      case "0" with suc_m assms show ?thesis by auto
    next
      case (Suc m') with suc_m lt have "m < length (stack s)" and "m > 0" by simp_all
      with step[of m] inv have "P (stack s ! m) (wl s ! m) (discover s') (finish s') (start s')" by simp
      with suc_m assms(1-4) `m > 0` show ?thesis by auto
    qed
  qed
qed

text {* Basic lemmata over @{term dfs_step} wrt to properties of the DFS *}

lemma dfs_step_preserves_start:
  "s' \<in> dfs_step dfs s \<Longrightarrow> wl s \<noteq> [] \<Longrightarrow> start s' = start s"
by (cases rule: dfs_step_cases_elem) (auto simp add: dfs_sws.defs)

lemma dfs_step_discovered_subset:
  "s' \<in> dfs_step dfs s \<Longrightarrow> wl s \<noteq> [] \<Longrightarrow> discovered s' \<supseteq> discovered s"
by (cases rule: dfs_step_cases_elem) force+

lemma dfs_step_exists:
  assumes inv: "stack s \<noteq> []" "wl s \<noteq> []"
  shows "dfs_step dfs s \<noteq> {}"
using inv
proof (cases "stack s")
  case (Cons x xs) then show ?thesis
  proof (cases "hd(wl s) = {}")
    case True with dfs_step_simps(2)[OF Cons inv(2)] show ?thesis by blast
  next
    case False then obtain e where "e \<in> hd (wl s)" by blast
    thus ?thesis
      using dfs_step_simps(3)[OF Cons inv(2) `hd(wl s) \<noteq> {}`, of dfs]
      apply simp
      apply (rule ex_comm[THEN iffD1]) -- "commute the two ex.-quant. variables in the conclusion"
      apply (rule_tac x=e in exI) -- "instantiate the first one"
      apply (cases "e \<in> dfs_restrict dfs", simp)
      by (cases "e \<in> discovered s") simp_all
  qed
qed simp

lemma dfs_step_stack_not_discovered:
  assumes "s' \<in> dfs_step dfs s" "wl s \<noteq> []"
  and "length (stack s') > length (stack s)"
  shows "hd (stack s') \<notin> discovered s"
using assms
by cases auto

lemma dfs_step_stack_not_restricted:
  assumes "s' \<in> dfs_step dfs s" "wl s \<noteq> []"
  and "length (stack s') > length (stack s)"
  shows "hd (stack s') \<notin> dfs_restrict dfs"
using assms
by cases auto

text {* @{term dfs_cond_compl} specifies the condition that must hold to continue searching. It consists of the conditions needed by
the framework and the ones given by @{term "dfs_cond dfs"}, that are dependent on the concrete implementation. *}
definition dfs_cond_compl :: "('S, 'n, 'X) dfs_algorithm_scheme \<Rightarrow> ('S, 'n) dfs_sws \<Rightarrow> bool"
where
  "dfs_cond_compl dfs s \<equiv> stack s \<noteq> [] \<and> wl s \<noteq> [] \<and> dfs_cond dfs (state s)"

lemma dfs_cond_compl_cond [dest]:
  "dfs_cond_compl dfs s \<Longrightarrow> dfs_cond dfs (state s)"
unfolding dfs_cond_compl_def
by auto

lemma dfs_cond_compl_step_exists [dest]:
  "dfs_cond_compl dfs s \<Longrightarrow> dfs_step dfs s \<noteq> {}"
unfolding dfs_cond_compl_def
by (simp add: dfs_step_exists)

text {* @{term dfs_next} includes the above condition, s.t. it is only possible to make a step, if the condition is met. *}
definition dfs_next :: "('S, 'n, 'X) dfs_algorithm_scheme \<Rightarrow> ('S, 'n) dfs_sws \<Rightarrow> ('S, 'n) dfs_sws \<Rightarrow> bool"
where
  "dfs_next dfs s s' \<longleftrightarrow>
    dfs_cond_compl dfs s \<and>  s' \<in> dfs_step dfs s"

lemma dfs_nextE:
  "\<lbrakk> dfs_next dfs s s'; \<lbrakk>dfs_cond_compl dfs s; s' \<in> dfs_step dfs s\<rbrakk> \<Longrightarrow> P \<rbrakk> \<Longrightarrow> P"
unfolding dfs_next_def
by blast

lemma dfs_next_dfs_step[dest]:
  "dfs_next dfs s s' \<Longrightarrow> s' \<in> dfs_step dfs s"
unfolding dfs_next_def
by simp

lemma dfs_next_stack_notempty[simp, intro]:
  "dfs_next dfs s s' \<Longrightarrow> stack s \<noteq> []"
unfolding dfs_next_def dfs_cond_compl_def
by simp_all

lemma dfs_next_wl_notempty[simp, intro]:
  "dfs_next dfs s s' \<Longrightarrow> wl s \<noteq> []"
unfolding dfs_next_def dfs_cond_compl_def
by simp_all

(* lift the case-splitting from dfs_step to dfs_next *)
lemma lift_to_dfs_next:
  assumes "s' \<in> dfs_step dfs s \<Longrightarrow> wl s \<noteq> [] \<Longrightarrow> PROP S"
  shows "dfs_next dfs s s' \<Longrightarrow> PROP S"
using assms
by (simp add: dfs_next_dfs_step)

lemmas dfs_next_cases_elem [case_names empty restrict remove visit, cases pred: dfs_next] = dfs_step_cases_elem[COMP lift_to_dfs_next]

lemma dfs_nextI: 
  "dfs_cond_compl dfs s \<Longrightarrow> s' \<in> dfs_step dfs s \<Longrightarrow> dfs_next dfs s s'"
unfolding dfs_next_def
by auto

lemma dfs_next_cond_compl:
  "dfs_next dfs s s' \<Longrightarrow> dfs_cond_compl dfs s"
unfolding dfs_next_def
by auto

lemmas dfs_next_intros = dfs_step_intros[THEN dfs_nextI[rotated], rotated -1]

end

subsection {* Properties of the DFS *}
text {* Using some notion of a \emph{good state} via the predicate @{term dfs_constructable}, we enable the tool of induction over DFS. This is very useful for showing that certain (basic) properties hold throughout the search. *}
context finite_digraph
begin
text {* To be able to define good states, we have to define a basic state, i.e.\ the state to start with for every possible DFS. Then @{term dfs_constructable} is just the set of states, that can be constructed from this point with the given graph. *}

definition dfs_constr_start :: "('S, 'n, 'X) dfs_algorithm_scheme \<Rightarrow> 'n \<Rightarrow> ('S, 'n) dfs_sws"
where "dfs_constr_start dfs x \<equiv> \<lparr> start = x, stack = [x], wl = [succs x], discover = [x \<mapsto> 0], finish = Map.empty, counter = 1, state = dfs_start dfs x \<rparr>"

lemma dfs_constr_start_simps[simp]:
  "start (dfs_constr_start dfs x) = x"
  "stack (dfs_constr_start dfs x) = [x]"
  "wl (dfs_constr_start dfs x) = [succs x]"
  "discovered (dfs_constr_start dfs x) = {x}"
  "ran (discover (dfs_constr_start dfs x)) = {0}"
  "finished (dfs_constr_start dfs x) = {}"
  "ran (finish (dfs_constr_start dfs x)) = {}"
  "counter (dfs_constr_start dfs x) = 1"
  "state (dfs_constr_start dfs x) = dfs_start dfs x"
by (simp_all add: dfs_constr_start_def)

inductive dfs_constructable :: "('S, 'n, 'X) dfs_algorithm_scheme \<Rightarrow> ('S, 'n) dfs_sws \<Rightarrow> bool"
  for dfs :: "('S, 'n, 'X) dfs_algorithm_scheme" where
  start: "\<And>x. x \<in> V \<Longrightarrow> x \<notin> dfs_restrict dfs \<Longrightarrow> dfs_constructable dfs (dfs_constr_start dfs x)"
| step: "dfs_next dfs s s' \<Longrightarrow> dfs_constructable dfs s \<Longrightarrow> dfs_constructable dfs s'"

(* basic induction principle, combining the one from dfs_constructable with the case-splitting of dfs_step *)
lemma dfs_constructable_induct [case_names start empty restrict remove visit, induct pred: dfs_constructable]:
  assumes inv: "dfs_constructable dfs s"
  and start': "\<And>x. x \<in> V \<Longrightarrow> x \<notin> dfs_restrict dfs \<Longrightarrow> P (dfs_constr_start dfs x)"
  and empty': "\<And>s s' x xs w ws. \<lbrakk> dfs_next dfs s s'; dfs_constructable dfs s; dfs_constructable dfs s'; P s; stack s = x#xs; wl s = w#ws; w = {};
              s' = s\<lparr>stack := xs, wl := ws, finish := (finish s)(x \<mapsto> counter s), counter := Suc (counter s), state := dfs_post dfs (state s) s x\<rparr>\<rbrakk>
              \<Longrightarrow> P s'"
  and restrict': "\<And>s s' e x xs w ws. \<lbrakk> dfs_next dfs s s'; dfs_constructable dfs s; dfs_constructable dfs s'; P s; stack s = x#xs; wl s = w#ws; w \<noteq> {}; 
               e \<in> w; e \<in> dfs_restrict dfs; s' = s\<lparr> wl := (w - {e})#ws\<rparr> \<rbrakk>
              \<Longrightarrow> P s'"
  and remove': "\<And>s s' e x xs w ws. \<lbrakk> dfs_next dfs s s'; dfs_constructable dfs s; dfs_constructable dfs s'; P s; stack s = x#xs; wl s = w#ws; w \<noteq> {}; 
               e \<in> w; e \<notin> dfs_restrict dfs; e \<in> discovered s; s' = s\<lparr> wl := (w - {e})#ws, state := dfs_remove dfs (state s) s e\<rparr> \<rbrakk>
              \<Longrightarrow> P s'"
  and visit': "\<And>s s' e x xs w ws. \<lbrakk> dfs_next dfs s s'; dfs_constructable dfs s; dfs_constructable dfs s'; P s; stack s = x#xs; wl s = w#ws; w \<noteq> {};
              e \<in> w; e \<notin> dfs_restrict dfs; e \<notin> discovered s;
              s' = s\<lparr> state := dfs_action dfs (state s) s e, stack := e#x#xs, wl := succs e#(w - {e})#ws, discover := (discover s)(e \<mapsto> counter s), counter := Suc (counter s)\<rparr> \<rbrakk>
              \<Longrightarrow> P s'"
  shows "P s"
  using inv
proof (induct rule: dfs_constructable.induct)
  case start with start' show ?case by metis
next
  case (step s s') then show ?case by (cases rule: dfs_next_cases_elem) (metis dfs_constructable.step step empty' restrict' remove' visit')+
qed

text {* Formulate the constructable predicate with an explicit starting point. *}
(* FIXME: Should probably make dfs_constructable obsolete at some point *)

inductive_set dfs_constr_from :: "('S, 'n, 'X) dfs_algorithm_scheme \<Rightarrow> 'n \<Rightarrow> ('S,'n) dfs_sws set"
for dfs :: "('S, 'n, 'X) dfs_algorithm_scheme"
and x :: "'n"
where
  start[simp]: "x \<in> V \<Longrightarrow> x \<notin> dfs_restrict dfs \<Longrightarrow> dfs_constr_start dfs x \<in> dfs_constr_from dfs x"
| step: "s \<in> dfs_constr_from dfs x \<Longrightarrow> dfs_next dfs s s' \<Longrightarrow> s' \<in> dfs_constr_from dfs x"

lemma dfs_constr_from_rtranclp_dfs_next:
  assumes "x \<in> V" "x \<notin> dfs_restrict dfs"
  shows "s \<in> dfs_constr_from dfs x \<longleftrightarrow> (dfs_next dfs)\<^sup>*\<^sup>* (dfs_constr_start dfs x) s"
proof 
  assume "s \<in> dfs_constr_from dfs x" thus "(dfs_next dfs)\<^sup>*\<^sup>* (dfs_constr_start dfs x) s" by induction (simp_all add: assms)
next
  assume "(dfs_next dfs)\<^sup>*\<^sup>* (dfs_constr_start dfs x) s" thus "s \<in> dfs_constr_from dfs x" by induction (simp_all add: dfs_constr_from.step assms)
qed

lemma dfs_constr_from_constructable[dest]:
  "s \<in> dfs_constr_from dfs x \<Longrightarrow> dfs_constructable dfs s"
by (induction rule: dfs_constr_from.induct) (simp_all add: dfs_constructable.start dfs_constructable.step)

lemma dfs_constructable_constr_from_start:
  "dfs_constructable dfs s \<Longrightarrow> s \<in> dfs_constr_from dfs (start s)"
by (induction rule: dfs_constructable.induct) (simp, metis dfs_constr_from.step dfs_step_preserves_start[COMP lift_to_dfs_next])

lemma constr_from_implies_start:
  "s \<in> dfs_constr_from dfs x \<Longrightarrow> start s = x"
by (induction rule: dfs_constr_from.induct) (simp_all add: dfs_step_preserves_start[COMP lift_to_dfs_next])

(* basic properties #1 *)
lemma counter_larger_one:
  "dfs_constructable dfs s \<Longrightarrow> counter s \<ge> 1"
by (induct rule: dfs_constructable_induct) auto

lemma length_wl_eq_stack:
  "dfs_constructable dfs s \<Longrightarrow> length (wl s) = length (stack s)"
by (induction rule: dfs_constructable_induct) auto

lemma succs_discovered_subset_wl_all:
  "dfs_constructable dfs s \<Longrightarrow> \<forall>n < length (stack s). succs (stack s ! n) - discovered s - dfs_restrict dfs \<subseteq> wl s ! n"
proof (induction rule: dfs_constructable_induct)
  case (empty s s') hence "length (stack s) = length (wl s)" by (auto dest: length_wl_eq_stack)
  with list_take_induct_tl2[OF this(1)] empty.IH empty show ?case by simp
next
  case (remove s s' e x xs w ws)

  then have w': "wl s' = (w - {e}) # ws" and s': "stack s' = x#xs" and d': "discovered s' = discovered s" by auto
  from remove(4,5) s' w' remove.IH show ?case
   by (rule stack_wl_remove_induct) (simp_all add: d', blast intro!: remove(9))
next
  case (restrict s s' e x xs w ws)

  then have w': "wl s' = (w - {e}) # ws" and s': "stack s' = x#xs" and d': "discovered s' = discovered s" by auto
  from restrict(4,5) s' w' restrict.IH show ?case
   by (rule stack_wl_remove_induct) (simp_all add: d', blast intro!: restrict(8))
next
  case (visit s s' e x xs w ws) hence s': "stack s' = e#x#xs" and w': "wl s' = succs e#(w - {e})#ws" and d': "discovered s' = insert e (discovered s)" by simp_all
  from visit(4,5) s' w' visit.IH show ?case
    by (rule stack_wl_visit_induct) (auto simp add: d')
qed auto

lemma succs_discovered_subset_wl:
  "dfs_constructable dfs s \<Longrightarrow> n < length (stack s) \<Longrightarrow> succs (stack s ! n) - discovered s - dfs_restrict dfs \<subseteq> wl s ! n"
by (metis succs_discovered_subset_wl_all)

lemma empty_wl_succs_discovered:
  "dfs_constructable dfs s \<Longrightarrow> n < length (stack s) \<Longrightarrow> wl s ! n = {} \<Longrightarrow> succs (stack s ! n) - dfs_restrict dfs \<subseteq> discovered s"
using succs_discovered_subset_wl
by blast

lemma wl_subset_succs_all:
  "dfs_constructable dfs s \<Longrightarrow> \<forall>n < length (stack s). wl s ! n \<subseteq> succs (stack s ! n)"
proof (induction rule: dfs_constructable_induct)
  case (empty s s') hence "length (wl s) = length (stack s)" by (auto dest: length_wl_eq_stack)
  with list_take_induct_tl2[OF this(1)] empty.IH empty show ?case by simp
next
  case (remove s s' e x xs w ws) hence w': "wl s' = (w - {e}) # ws" and s': "stack s' = x#xs" by simp_all
  from remove(4,5) s' w' remove.IH show ?case by (rule stack_wl_remove_induct) auto
next
  case (restrict s s' e x xs w ws) hence w': "wl s' = (w - {e}) # ws" and s': "stack s' = x#xs" by simp_all
  from restrict(4,5) s' w' restrict.IH show ?case by (rule stack_wl_remove_induct) auto
next
  case (visit s s' e x xs w ws) hence e: "stack s' = e#x#xs" "wl s' = succs e#(w - {e})#ws" by simp_all
  from visit(4,5) e visit.IH show ?case by (rule stack_wl_visit_induct) auto
qed simp

lemma wl_subset_succs:
  "dfs_constructable dfs s \<Longrightarrow> n < length (stack s) \<Longrightarrow> wl s ! n \<subseteq> succs (stack s ! n)"
by (metis wl_subset_succs_all)

lemma wl_subset_verts:
  "dfs_constructable dfs s \<Longrightarrow> x \<in> set (wl s) \<Longrightarrow> x \<subseteq> V"
proof -
  assume c: "dfs_constructable dfs s" and "x \<in> set (wl s)"
  hence "\<exists> n. x = wl s ! n \<and> n < length (wl s)" using in_set_conv_nth[of x "wl s"] by blast
  then obtain n where "x = wl s ! n" "n < length (wl s)" by force
  with length_wl_eq_stack[OF c] wl_subset_succs[OF c] have "x \<subseteq> succs (stack s ! n)" by simp
  with succs_in_V show ?thesis by auto
qed

lemma wl_finite:
  "dfs_constructable dfs s \<Longrightarrow> x \<in> set (wl s) \<Longrightarrow> finite x"
by (metis wl_subset_verts finite_V finite_subset)

lemma discovered_subset_verts:
  "dfs_constructable dfs s \<Longrightarrow> discovered s \<subseteq> V"
proof (induct s rule: dfs_constructable_induct)
  case (visit s s' e _ _ w) hence "discovered s' = insert e (discovered s)" by simp
  moreover
  from visit wl_subset_verts[OF visit(2)] have "w \<subseteq> V" by simp
  with visit have "e \<in> V" by auto
  ultimately show ?case using `discovered s \<subseteq> V` by auto
qed simp+

lemma discovered_finite:
  "dfs_constructable dfs s \<Longrightarrow> finite (discovered s)"
by (metis finite_V finite_subset discovered_subset_verts)

lemma ran_discover_finite:
  "dfs_constructable dfs s \<Longrightarrow> finite (ran (discover s))"
by (metis discovered_finite map_dom_ran_finite)

lemma stack_subset_discovered:
  "dfs_constructable dfs s \<Longrightarrow> set (stack s) \<subseteq> discovered s"
by (induct rule: dfs_constructable_induct) auto

lemma stack_subset_verts:
  "dfs_constructable dfs s \<Longrightarrow> set (stack s) \<subseteq> V"
by (metis discovered_subset_verts stack_subset_discovered subset_trans)

lemma stack_distinct:
  "dfs_constructable dfs s \<Longrightarrow> distinct (stack s)"
proof (induct rule: dfs_constructable_induct)
  case (visit s s' _ x xs)
  hence tl: "stack s = tl (stack s')" by simp
  
  {
    assume "hd (stack s') \<notin> set (stack s)"
    with tl have "hd (stack s') \<notin> set (tl (stack s'))" by simp
    
    moreover
    from tl `distinct (stack s)` have "distinct (tl (stack s'))" by simp
    
    ultimately have "distinct (stack s')" by (cases "stack s'") simp_all
  }
  
  moreover
  from visit have "hd (stack s') \<notin> discovered s" using dfs_step_stack_not_discovered[OF dfs_next_dfs_step] by simp
  
  moreover 
  from visit stack_subset_discovered have "set (stack s) \<subseteq> discovered s" by blast
  ultimately show ?case by auto
qed simp+

lemma last_stack_is_start:
  "dfs_constructable dfs s \<Longrightarrow> stack s \<noteq> [] \<Longrightarrow> last (stack s) = start s"
by (induct rule: dfs_constructable_induct) (auto simp add: dfs_constr_start_def tl_last)

lemma start_discovered:
  "dfs_constructable dfs s \<Longrightarrow> start s \<in> discovered s"
proof (induct rule: dfs_constructable.induct)
  case step thus ?case by (metis set_rev_mp dfs_step_discovered_subset dfs_next_dfs_step dfs_step_preserves_start dfs_next_wl_notempty)
qed simp

lemma start_not_restr:
  "dfs_constructable dfs s \<Longrightarrow> start s \<notin> dfs_restrict dfs"
by (induction rule: dfs_constructable_induct) simp_all

lemma dfs_start_in_verts:
  "dfs_constructable dfs s \<Longrightarrow> start s \<in> V"
by (metis start_discovered discovered_subset_verts set_rev_mp) 

lemma finished_stack_eq_discovered:
  "dfs_constructable dfs s \<Longrightarrow> finished s \<union> set (stack s) = discovered s"
proof (induction rule: dfs_constructable_induct)
  case (empty s s' x xs)
  hence "set (stack s) = set (stack s') \<union> {hd (stack s)}" by (fastforce intro: list.exhaust)
  with empty have "finished s' \<union> set (stack s') = finished s \<union> set (stack s)" by auto
  also note empty.IH
  also from empty have "discovered s = discovered s'" by simp
  finally show ?case .
next
  case (visit s s' e x xs) then have fin_eq: "finished s' = finished s" by simp

  from visit have dis_eq: "discovered s' = insert e (discovered s)" and "stack s' = e # stack s" by simp_all
  then have "set (stack s') = insert e (set (stack s))" by auto
  hence "finished s' \<union> set (stack s') = insert e ((finished s') \<union> set (stack s))" by auto
  also with fin_eq have  "... = insert e (finished s \<union> set (stack s))" by simp
  also with visit.IH have "... = insert e (discovered s)" by simp
  also with dis_eq have "... = discovered s'" by simp
  finally show ?case .
qed simp+

lemma finished_subset_discovered:
  "dfs_constructable dfs s \<Longrightarrow> finished s \<subseteq> discovered s"
by (auto dest!: finished_stack_eq_discovered)

lemma finished_subset_verts:
  "dfs_constructable dfs s \<Longrightarrow> finished s \<subseteq> V"
by (metis finished_subset_discovered discovered_subset_verts subset_trans)

lemma finished_finite:
  "dfs_constructable dfs s \<Longrightarrow> finite (finished s)"
by (metis finished_subset_verts finite_V finite_subset)

lemma ran_finish_finite:
  "dfs_constructable dfs s \<Longrightarrow> finite (ran (finish s))"
by (metis map_dom_ran_finite finished_finite)

lemma finished_implies_discovered:
  "dfs_constructable dfs s \<Longrightarrow> v \<in> finished s \<Longrightarrow> v \<in> discovered s"
by (auto dest!: finished_subset_discovered)

lemma not_discovered_implies_not_finished:
  "dfs_constructable dfs s \<Longrightarrow> v \<notin> discovered s \<Longrightarrow> v \<notin> finished s"
by (metis finished_implies_discovered)

lemma discovered_non_stack_implies_finished:
  "dfs_constructable dfs s \<Longrightarrow> v \<in> discovered s \<Longrightarrow> v \<notin> set (stack s) \<Longrightarrow> v \<in> finished s"
by (auto dest!: finished_stack_eq_discovered)

lemma discovered_not_finished_implies_stack:
  "dfs_constructable dfs s \<Longrightarrow> v \<in> discovered s \<Longrightarrow> v \<notin> finished s \<Longrightarrow> v \<in> set (stack s)"
by (metis discovered_non_stack_implies_finished)

lemma discovered_not_restricted:
  "dfs_constructable dfs s \<Longrightarrow> v \<in> discovered s \<Longrightarrow> v \<notin> dfs_restrict dfs"
by (induction rule: dfs_constructable_induct) auto

lemma finished_not_restricted:
  "dfs_constructable dfs s \<Longrightarrow> v \<in> finished s \<Longrightarrow> v \<notin> dfs_restrict dfs"
using discovered_not_restricted finished_subset_discovered 
by blast

lemma stack_not_restricted:
  "dfs_constructable dfs s \<Longrightarrow> v \<in> set (stack s) \<Longrightarrow> v \<notin> dfs_restrict dfs"
using discovered_not_restricted stack_subset_discovered
by blast

lemma restricted_not_finished:
  "dfs_constructable dfs s \<Longrightarrow> v \<in> dfs_restrict dfs \<Longrightarrow> v \<notin> finished s"
using finished_not_restricted by blast

lemma restricted_not_discovered:
  "dfs_constructable dfs s \<Longrightarrow> v \<in> dfs_restrict dfs \<Longrightarrow> v \<notin> discovered s"
using discovered_not_restricted by blast

lemma restricted_not_stack:
  "dfs_constructable dfs s \<Longrightarrow> v \<in> dfs_restrict dfs \<Longrightarrow> v \<notin> set (stack s)"
using stack_not_restricted by blast

(* more advanced lemmas *)
lemma stack_implies_not_finished:
  "dfs_constructable dfs s \<Longrightarrow> v \<in> set (stack s) \<Longrightarrow> v \<notin> finished s"
proof (induction rule: dfs_constructable_induct)
  case (empty _ _ x) with stack_distinct[OF empty(2)] show ?case by (cases "v = x") auto
next
  case (visit s s' e) then have f_eq: "finish s = finish s'" and e_stack: "stack s' = e # stack s" by simp_all
  with visit(9) not_discovered_implies_not_finished[OF visit(2)] have "e \<notin> finished s" by simp
  with e_stack visit.IH visit(12) f_eq show ?case by auto
qed simp+

lemma finished_implies_not_stack:
  "dfs_constructable dfs s \<Longrightarrow> x \<in> finished s \<Longrightarrow> x \<notin> set (stack s)"
by (metis stack_implies_not_finished)

lemma stack_in_Suc_succs_restr_all:
  "dfs_constructable dfs s \<Longrightarrow> \<forall> n < (length (stack s)) - 1. (stack s ! n) \<in> succs (stack s ! Suc n) - dfs_restrict dfs"
proof (induction rule: dfs_constructable_induct)
  case (empty _ _ _ xs) thus ?case by (induct xs) auto
next
  case (visit s s' e _ _ w) hence "stack s' = e#stack s" by simp
  with wl_subset_succs[OF visit(2), of 0] visit have "stack s' ! 0 \<in> succs (stack s' ! Suc 0) - dfs_restrict dfs" using dfs_next_stack_notempty[OF visit(1), THEN hd_conv_nth] by auto

  moreover 
  from visit have  "\<And> n. n < (length (stack s')) - 1 \<Longrightarrow> n > 0 \<Longrightarrow> (stack s' ! n) \<in> succs (stack s' ! Suc n) - dfs_restrict dfs" by fastforce

  ultimately show ?case by blast
qed simp+

lemma stack_in_Suc_succs_restr:
  "dfs_constructable dfs s \<Longrightarrow> n < length (stack s) - 1 
   \<Longrightarrow> stack s ! n \<in> succs (stack s ! Suc n) - dfs_restrict dfs"
by (metis stack_in_Suc_succs_restr_all)

lemma Suc_stack_stack_in_restr_edges:
  "dfs_constructable dfs s \<Longrightarrow> n < length (stack s) - 1 \<Longrightarrow> (stack s ! Suc n, stack s ! n) \<in> (rel_restrict E (dfs_restrict dfs))"
using succs_implies_edge rel_restrict_notR stack_in_Suc_succs_restr restricted_not_stack
by (smt Diff_iff nth_mem)

lemma Suc_stack_stack_in_edges:
  "dfs_constructable dfs s \<Longrightarrow> n < length (stack s) - 1 \<Longrightarrow> (stack s ! Suc n, stack s ! n) \<in> E"
using Suc_stack_stack_in_restr_edges rel_restrict_lift
by metis

lemma Suc_reachable_stack:
  "dfs_constructable dfs s \<Longrightarrow> n < (length (stack s)) - 1 \<Longrightarrow> stack s ! Suc n \<rightarrow>\<setminus>dfs_restrict dfs+ stack s ! n" 
using restr_reachable1_trancl Suc_stack_stack_in_restr_edges
by blast

lemma prev_reachable_stack:
  "dfs_constructable dfs s \<Longrightarrow> n < length (stack s) \<Longrightarrow> m < n \<Longrightarrow> stack s ! n \<rightarrow>\<setminus>dfs_restrict dfs+ stack s ! m"
by (metis restr_reachable1_trancl nth_step_trancl Suc_stack_stack_in_restr_edges)

lemma tl_reachable_stack_hd:
  assumes constr: "dfs_constructable dfs s"
  and "x \<in> set (tl (stack s))"
  shows "x \<rightarrow>\<setminus>dfs_restrict dfs+ hd (stack s)"
proof -
  from assms have ne: "stack s \<noteq> []" by auto
  from assms obtain n where "n < length (tl (stack s))" "tl (stack s) ! n = x" unfolding in_set_conv_nth by auto
  with nth_tl[OF ne] have "Suc n < length (stack s)" "stack s ! (Suc n) = x"  by simp_all
  moreover with prev_reachable_stack[OF constr, of "Suc n" 0] have "stack s ! Suc n \<rightarrow>\<setminus>dfs_restrict dfs+ stack s ! 0" by simp
  ultimately show ?thesis using hd_conv_nth[OF ne] by simp
qed

lemma start_reachable_stack:
  assumes constr: "dfs_constructable dfs s"
  and "x \<in> set (stack s)"
  shows "start s \<rightarrow>\<star> x"
proof (cases "x = start s")
  case True with self_reachable dfs_start_in_verts[OF constr] show ?thesis by simp
next
  case False with assms have "stack s \<noteq> []" by auto
  with last_stack_is_start[OF constr] last_conv_nth have s_nth: "start s = stack s ! (length (stack s) - 1)" by metis
  
  from `x \<in> set (stack s)` obtain n where n: "n < length (stack s)" "stack s ! n = x" unfolding in_set_conv_nth by auto
  
  have "n < length (stack s) - 1"
  proof (rule ccontr)
    assume "\<not> n < length (stack s) - 1"
    with n have "n = length (stack s) - 1" by auto
    with s_nth n False show "False" by simp
  qed
  with prev_reachable_stack[OF constr] have "stack s ! (length (stack s) - 1) \<rightarrow>\<setminus>dfs_restrict dfs+ stack s ! n" by auto
  with s_nth n have "start s \<rightarrow>\<setminus>dfs_restrict dfs+ x" by auto
  thus ?thesis by blast
qed

lemma ran_discover_lt_counter:
  "dfs_constructable dfs s \<Longrightarrow> v \<in> ran (discover s) \<Longrightarrow> v < counter s"
proof (induction rule: dfs_constructable_induct)
  case (visit s s' e) hence
        d': "discover s' = (discover s)(e \<mapsto> counter s)"
    and c': "counter s' = Suc (counter s)"
  by simp_all
  with visit(9) have r': "ran (discover s') = ran (discover s) \<union> {counter s}" by auto
  
  from c' show ?case
  proof (cases "v = counter s")
    case False with visit.prems r' have "v \<in> ran (discover s)" by simp
    with visit.IH c' show ?thesis by simp
  qed simp
qed auto

lemma discover_lt_counter:
  assumes constr: "dfs_constructable dfs s"
  and     nn:     "v \<in> discovered s"
  shows "\<delta> s v < counter s"
proof -
  from nn have "\<delta> s v \<in> ran (discover s)" by (auto intro: ranI)
  with ran_discover_lt_counter[OF constr] show ?thesis .
qed

lemma ran_finish_lt_counter:
  "dfs_constructable dfs s \<Longrightarrow> v \<in> ran (finish s) \<Longrightarrow> v < counter s"
proof (induction rule: dfs_constructable_induct)
  case (empty s s' x xs)
  hence f': "finish s' = (finish s)(x \<mapsto> counter s)"
    and c': "counter s' = Suc (counter s)"
    by simp_all

  with stack_implies_not_finished[OF empty(2)] empty(4) have "x \<notin> finished s" by simp
  with f' have r': "ran (finish s') = ran (finish s) \<union> {counter s}" by auto

  from c' show ?case
  proof (cases "v = counter s")
    case False with empty.prems r' have "v \<in> ran (finish s)" by simp
    with f' c' empty.IH show ?thesis by simp
  qed simp
qed auto

lemma finish_lt_counter:
  assumes constr: "dfs_constructable dfs s"
  and     nn:     "v \<in> finished s"
  shows "\<phi> s v < counter s"
proof -
  from nn have "\<phi> s v \<in> ran (finish s)" by (auto intro: ranI)
  with ran_finish_lt_counter[OF constr] show ?thesis .
qed

lemma finish_gt_discover:
  "dfs_constructable dfs s \<Longrightarrow> v \<in> finished s \<Longrightarrow> \<phi> s v > \<delta> s v"
proof (induction rule: dfs_constructable_induct)
  case (empty s s' x xs) 
  hence f': "finish s' = (finish s)(x \<mapsto> counter s)"
    and d': "discover s = discover s'" 
    and c': "counter s' = Suc (counter s)"
    by simp_all
  
  show ?case
  proof (cases "v = x")
    case False with empty.prems f' have "v \<in> finished s" "\<phi> s v = \<phi> s' v" by auto
    with d' empty.IH show ?thesis by auto
  next
    case True with f' have "v \<in> finished s'" by simp
    with finished_implies_discovered[OF empty(3)] d' have "v \<in> discovered s" by simp
    with discover_lt_counter[OF empty(2)] d' have "\<delta> s' v < counter s" by simp
    with f' c' True show ?thesis by simp
  qed
next
  case (visit s s' e)
  hence d': "discover s' = (discover s)(e \<mapsto> counter s)"
    and f': "finish s = finish s'"
    by simp_all
  
  show ?case
  proof (cases "v = e")
    case False with d' f' visit show ?thesis by force
  next
    case True with f' have "v \<notin> finished s'" using not_discovered_implies_not_finished[OF visit(2) visit(9)] by simp
    with visit.prems show ?thesis by contradiction
  qed
qed simp+

lemma discover_finish_distinct:
  "dfs_constructable dfs s \<Longrightarrow> ran (finish s) \<inter> ran (discover s) = {}"
proof (induction rule: dfs_constructable_induct)
  case (empty s s' x xs) then have "x \<notin> finished s" using stack_implies_not_finished[OF empty(2)] hd_in_set by simp
  with empty have "ran (finish s') = insert (counter s) (ran (finish s))" "ran (discover s') = ran (discover s)"
    by auto
  moreover
  with ran_discover_lt_counter[OF empty(2)] have "counter s \<notin> ran (discover s')" by auto
  ultimately show ?case using empty.IH by auto
next
  case (visit s s' e) hence d':  "discover s' = discover s (e \<mapsto> counter s)" by simp
  with visit(9) have "ran (discover s') = insert (counter s) (ran (discover s))" by auto

  moreover
  from visit have "finish s = finish s'" by simp
  
  moreover
  with ran_finish_lt_counter[OF visit(2)] have "counter s \<notin> ran (finish s')" by auto
  
  ultimately show ?case using visit.IH by auto
qed simp+

lemma discover_neq_finish:
  assumes constr: "dfs_constructable dfs s"
  and verts: "v \<in> V" "w \<in> V"
  and discovered: "v \<in> discovered s"
  shows "discover s v \<noteq> finish s w"
using discovered
proof (cases "finish s w")
  case (Some x) then have "x \<notin> ran (discover s)" using discover_finish_distinct[OF constr] by (auto intro: ranI)
  with Some discovered show ?thesis by (metis ranI)
qed auto

lemma discover_card_ran_dom:
  "dfs_constructable dfs s \<Longrightarrow> card (ran (discover s)) = card (dom (discover s))"
proof (induction rule: dfs_constructable_induct)
  case (visit s s') with discovered_finite[OF visit(2)] visit(9) have "card (dom (discover s')) = card (dom (discover s)) + 1" by simp

  moreover
  from visit ran_discover_lt_counter[OF visit(2)] have "ran (discover s') = insert (counter s) (ran (discover s))" "counter s \<notin> ran (discover s)" by auto
  with ran_discover_finite[OF visit(2)] have "card (ran (discover s')) = card (ran (discover s)) + 1" by simp

  ultimately show ?case using visit.IH by simp
qed simp+
  
lemma discover_neq_discover:
  assumes constr: "dfs_constructable dfs s"
  and verts: "v \<in> V" "w \<in> V"
  and ne: "v \<noteq> w"
  and discovered: "v \<in> discovered s"
  shows "discover s v \<noteq> discover s w"
proof -
  from map_card_eq_iff[OF discovered_finite, OF constr] assms discover_card_ran_dom[OF constr] have "discover s v = discover s w \<longleftrightarrow> v = w" by auto
  with ne show ?thesis by simp
qed

lemma finish_card_ran_dom:
  "dfs_constructable dfs s \<Longrightarrow> card (ran (finish s)) = card (dom (finish s))"
proof (induction rule: dfs_constructable_induct)
  case (empty s s' x) then have f': "finish s' = finish s (x \<mapsto> counter s)" by simp
  with empty have *: "x \<notin> finished s" using stack_implies_not_finished by fastforce
  with finished_finite[OF empty(2)] f' have "card (dom (finish s')) = card (dom (finish s)) + 1" by simp

  moreover
  from f' ran_finish_lt_counter[OF empty(2)] * have "ran (finish s') = insert (counter s) (ran (finish s))" "counter s \<notin> ran (finish s)" by auto
  with ran_finish_finite[OF empty(2)] have "card (ran (finish s')) = card (ran (finish s)) + 1" by simp
  
  ultimately show ?case using empty.IH by simp
qed simp+

lemma finish_neq_finish:
  assumes constr: "dfs_constructable dfs s"
  and verts: "v \<in> V" "w \<in> V"
  and ne: "v \<noteq> w"
  and discovered: "v \<in> finished s"
  shows" finish s v \<noteq> finish s w"
proof -
  from map_card_eq_iff[OF finished_finite, OF constr] assms finish_card_ran_dom[OF constr] have "finish s v = finish s w \<longleftrightarrow> v = w" by auto
  with ne show ?thesis by simp
qed

lemma prev_stack_discover_all:
  "dfs_constructable dfs s \<Longrightarrow> \<forall> n < length (stack s). \<forall> v \<in> set (drop (Suc n) (stack s)). \<delta> s (stack s ! n) > \<delta> s v"
proof (induction rule: dfs_constructable_induct)
  case (visit s s' e) 
  hence d': "discover s' = (discover s)(e \<mapsto> counter s)"
    and s': "stack s' = e # stack s"
    by simp_all
  with stack_distinct[OF visit(3)] have t_ni: "e \<notin> set(stack s)" by simp

  from visit(2) have "\<And>v. v \<in> set (stack s) \<Longrightarrow> \<delta> s v < counter s" by (metis discover_lt_counter set_mp stack_subset_discovered)
  with t_ni d' have "\<And>v. v \<in> set (stack s) \<Longrightarrow> \<delta> s' v < \<delta> s' e" by auto
  with s' have "\<And>v. v \<in> set (drop (Suc 0) (stack s')) \<Longrightarrow> \<delta> s' v < \<delta> s' (stack s' ! 0)" by auto

  moreover
  from s' d' t_ni have "\<And> n. n < (length (stack s')) \<Longrightarrow> n > 0 \<Longrightarrow> \<delta> s' (stack s' ! n) = \<delta> s (stack s' ! n)" by auto
  with visit.IH s' d'  t_ni have  "\<And> n. n < (length (stack s')) - 1 \<Longrightarrow> n > 0 \<Longrightarrow> \<forall> v \<in> set (drop (Suc n) (stack s')). \<delta> s' (stack s' ! n) > \<delta> s' v" by (smt drop_Cons' fun_upd_other in_set_dropD length_drop nth_Cons_pos)

  ultimately show ?case by (smt length_drop length_pos_if_in_set)
qed auto

lemma prev_stack_discover:
  "dfs_constructable dfs s \<Longrightarrow> n < length (stack s) \<Longrightarrow> v \<in> set (drop (Suc n) (stack s)) \<Longrightarrow> \<delta> s (stack s ! n) > \<delta> s v"
by (metis prev_stack_discover_all)

lemma Suc_stack_discover:
  assumes constr: "dfs_constructable dfs s"
  and n: "n < (length (stack s)) - 1"
  shows "\<delta> s (stack s ! n) > \<delta> s (stack s ! Suc n)"
proof -
  from prev_stack_discover assms have "\<And> v. v \<in> set (drop (Suc n) (stack s)) \<Longrightarrow> \<delta> s (stack s ! n) > \<delta> s v" by fastforce
  moreover from n have "stack s ! Suc n \<in> set (drop (Suc n) (stack s))" using in_set_conv_nth by fastforce
  ultimately show ?thesis .
qed

lemma tl_lt_stack_hd_discover:
  assumes "dfs_constructable dfs s"
  and notempty: "stack s \<noteq> []"
  and "x \<in> set (tl (stack s))"
  shows "\<delta> s x < \<delta> s (hd (stack s))"
proof -
  from notempty obtain y ys where "stack s = y#ys" by (metis list.exhaust)
  with assms show ?thesis
    using  prev_stack_discover
    by (cases ys) force+
qed

lemma obtain_discovered_predecessor:
  assumes "dfs_constructable tdfs s"
  and "v \<in> discovered s" 
  and "v \<noteq> start s"
  obtains y where "v \<in> succs y" and "y \<in> discovered s" and "\<delta> s v > \<delta> s y"
using assms
proof induction
  case (visit s s' e x xs w) hence *: "discover s' = discover s (e \<mapsto> counter s)" "start s' = start s" "stack s' = e#x#xs" by simp_all

  show ?case
  proof (cases "v = e")
    case True with visit have "v \<in> w" by simp
    with visit wl_subset_succs[OF visit(2), of 0] have "v \<in> succs x" by auto
    moreover from tl_lt_stack_hd_discover[OF visit(3), of x] * True have "\<delta> s' x < \<delta> s' v" by simp
    moreover from stack_subset_discovered[OF visit(3)] * have "x \<in> discovered s'" by auto
    ultimately show ?thesis using visit.prems by simp
  next
    case False with visit * have **: "v \<in> discovered s" "v \<noteq> start s" by auto
    {
      fix y
      assume A: "v \<in> succs y" "y \<in> discovered s" "\<delta> s y < \<delta> s v"
      with visit(9) False * have "y \<in> discovered s'" "\<delta> s' y < \<delta> s' v" by auto
      with A visit.prems have thesis by simp
    }
    with visit.IH[OF _ **] show ?thesis .
  qed
qed simp_all

lemma stack_finish_discover:
  "dfs_constructable dfs s \<Longrightarrow> v \<in> set (stack s) \<Longrightarrow> w \<in> finished s \<Longrightarrow> \<phi> s w < \<delta> s v \<or> \<delta> s v < \<delta> s w"
proof (induction rule: dfs_constructable_induct)
  case (empty s s' x) then have
    d': "discover s = discover s'" and
    f': "finish s' = finish s (x \<mapsto> counter s)" and
    v: "v \<in> set (stack s)"
    by simp_all

  show ?case
  proof (cases "w = x")
    case True with empty have "\<delta> s w > \<delta> s v" using prev_stack_discover[OF empty(2)] by fastforce
    with d' show ?thesis by simp
  next
    case False with empty.prems empty.IH v f' d' show ?thesis by force
  qed
next
  case (visit s s' e) 
  hence d': "discover s' = discover s (e \<mapsto> counter s)" 
    and s': "stack s' = e # stack s" 
    and f': "finish s = finish s'"
    by simp_all

  show ?case
  proof (cases "v = e")
    case True with d' have "\<delta> s' v = counter s" by simp
    moreover from visit.prems f' have "\<phi> s' w < counter s" using finish_lt_counter[OF visit(2)] by simp
    ultimately show ?thesis by simp
  next
    case False with visit.prems s' d' have "v \<in> set (stack s)" and "discover s v = discover s' v" by simp_all
    with f' visit.prems visit.IH have "\<phi> s' w < \<delta> s' v \<or> \<delta> s' v < \<delta> s w" by force

    moreover from visit.prems s' have "w \<noteq> e" using finished_implies_not_stack[OF visit(3)] by simp
    with d' have "discover s w = discover s' w" by simp
    ultimately show ?thesis by simp
  qed
qed simp+

lemma interval_inclusion:
  assumes "dfs_constructable dfs s"
  and "v \<in> discovered s"
  and "w \<in> finished s"
  and "\<delta> s v < \<delta> s w"
  and "v \<notin> finished s \<or> \<delta> s w < \<phi> s v"
  shows "v \<notin> finished s \<or> \<phi> s w < \<phi> s v"
using assms
proof induction
  case (empty s s' x) hence ne: "v \<noteq> w" and f': "finish s' = finish s (x\<mapsto>counter s)" by auto
  show ?case
  proof (cases "v = x")
    case True with f' ne empty.prems have "w \<in> finished s" and "\<phi> s' v = counter s"  by auto
    moreover with finish_lt_counter[OF empty(2)] empty.prems have "counter s > \<phi> s w" by simp
    with f' ne True have "counter s > \<phi> s' w" by simp
    ultimately show ?thesis by simp
  next
    case False
    {
      assume "v \<in> finished s"
      moreover with empty False have "v \<in> finished s'" "v \<in> discovered s" by simp_all
      moreover with empty False have *: "\<delta> s w < \<phi> s v" "\<delta> s v < \<delta> s w" by auto
      moreover have "w \<noteq> x"
      proof (rule notI)
        assume "w = x"
        with empty.hyps have "w \<in> set (stack s)" by simp
        with stack_finish_discover[OF empty(2)] `v \<in> finished s` * show "False" by force
      qed
      with empty have "w \<in> finished s" by auto
      ultimately have "\<phi> s w < \<phi> s v" using empty.IH by simp
      with False f' `w\<noteq>x` have ?thesis by simp
    }
    with False f' show ?thesis by force
  qed
next
  case (visit s s' e) hence
    f': "finish s' = finish s" and
    d': "discover s' = discover s (e \<mapsto> counter s)" by simp_all

  from visit have "w \<noteq> e" using stack_implies_not_finished[OF visit(3)] by auto

  show ?case
  proof (cases "v = e")
    case True with visit(9) have "v \<notin> discovered s" by simp
    with not_discovered_implies_not_finished[OF visit(2)] have "v \<notin> finished s" by simp
    with f' show ?thesis by simp
  next
    case False with `w \<noteq> e` visit.prems f' d' have "v \<in> discovered s" "w \<in> finished s" "\<delta> s v < \<delta> s w" "v \<notin> finished s \<or> \<delta> s w < \<phi> s v" by auto
    with visit.IH have "v \<notin> finished s \<or> \<phi> s w < \<phi> s v" .
    with f' show ?thesis by simp
  qed
qed simp_all

lemma correct_order:
  assumes constr: "dfs_constructable dfs s"
  and verts: "v \<in> finished s" "w \<in> finished s"
  and disc: "\<delta> s v < \<delta> s w"
  shows "\<phi> s v < \<delta> s w \<or> \<phi> s v > \<phi> s w"
proof (cases "\<delta> s w < \<phi> s v")
  case True with interval_inclusion finished_implies_discovered assms show ?thesis by metis
next
  case False hence "\<phi> s v \<le> \<delta> s w" by simp
  moreover from assms finished_subset_verts have "v \<in> V" "w \<in> V" by auto
  with discover_neq_finish finished_implies_discovered assms have "discover s w \<noteq> finish s v" by metis
  with verts(1) finished_implies_discovered[OF constr verts(2)] have "\<delta> s w \<noteq> \<phi> s v" by auto
  ultimately have "\<phi> s v < \<delta> s w" by simp
  thus ?thesis ..
qed

lemma discover_finish_implies_reach:
  "dfs_constructable dfs s \<Longrightarrow> 
   v \<in> discovered s \<Longrightarrow> w \<in> discovered s \<Longrightarrow>
   \<delta> s v < \<delta> s w \<Longrightarrow> v \<notin> finished s \<or> ( w \<in> finished s \<and> \<phi> s v > \<phi> s w)
   \<Longrightarrow> v \<rightarrow>\<setminus>dfs_restrict dfs+ w"
proof (induction rule: dfs_constructable_induct)
  case (empty s s' x) hence
    s': "discover s = discover s'" and
    f': "finish s' = finish s (x \<mapsto> counter s)"
    by simp_all

  show ?case
  proof (cases "v = x")
    case True with empty f' have "v \<notin> finished s" and "finish s' w = finish s w" using stack_implies_not_finished[OF empty(2)] by auto
    with empty.prems empty.IH s' show ?thesis by force
  next
    case False  with f' have fv: "finish s v = finish s' v" by simp
    show ?thesis
    proof (cases "w \<in> finished s'")
      case False with empty.IH s' empty.prems fv show ?thesis by fastforce
    next
      case True note wf' = this
      thus ?thesis
      proof (cases "v \<in> finished s'")
        case True with empty(12) wf' have *: "\<phi> s' w < \<phi> s' v" by simp
        have "w \<noteq> x"
        proof (rule notI)
          assume "w = x"
          with f' have "\<phi> s' w = counter s" by simp
          
          moreover
          from True fv have "v \<in> finished s" by auto        
          hence "\<phi> s v < counter s" using finish_lt_counter[OF empty(2)] by simp
          with fv have "\<phi> s' v < counter s" by simp
          ultimately show "False" using * by simp
        qed 
        with * fv f' wf' have "w \<in> finished s \<and> \<phi> s w < \<phi> s v" by simp
        with empty.prems s' empty.IH show ?thesis by simp
      next
        case False with empty.prems empty.IH s' f' show ?thesis by simp
      qed
    qed
  qed
next
  case (visit s s' e) 
  hence d': "discover s' = discover s (e \<mapsto> counter s)" 
    and s': "stack s' = e # stack s"
    and f': "finish s' = finish s"
    and ne: "v \<noteq> w"
    by auto

  show ?case
  proof (cases "v = e")
    case True with d' ne have "\<delta> s' v = counter s" "discover s w = discover s' w" by auto

    moreover
    then have "w \<in> discovered s" using visit.prems by auto
    hence "\<delta> s w < counter s" using discover_lt_counter[OF visit(2)] by simp

    ultimately have "False" using visit.prems by simp
    thus ?thesis ..
  next
    case False with d' have dv: "discover s v = discover s' v" by simp
    show ?thesis
    proof (cases "w = e")
      case True with s' have "w \<notin> finished s'" using stack_implies_not_finished[OF visit(3)] by simp
      with visit.prems have "v \<in> set (stack s')" using discovered_not_finished_implies_stack[OF visit(3)] by simp
      with False True s' show ?thesis using tl_reachable_stack_hd[OF visit(3)] by simp
    next
      case False with d' have "discover s w = discover s' w" by simp
      with dv f' visit.prems visit.IH show ?thesis by (cases "v \<in> finished s'") force+
    qed
  qed
qed simp+

lemma no_reach_discover_finish:
  assumes constr: "dfs_constructable dfs s"
  and no_reach: "\<not>v \<rightarrow>\<star> w"
  shows "v \<notin> discovered s \<or> w \<notin> discovered s \<or> w \<notin> finished s \<or> \<delta> s v > \<delta> s w \<or> (v \<in> finished s \<and> \<phi> s v < \<phi> s w)"
proof (cases "v \<in> V \<and> w \<in> V")
  case False hence "w \<notin> discovered s \<or> v \<notin> discovered s" using discovered_subset_verts[OF constr] by auto
  thus ?thesis by simp
next
  case True hence verts: "v \<in> V" "w \<in> V" by simp_all
  with no_reach have ne: "v \<noteq> w" using reachable_ewalk[of v w] ewalk_empty_iff[of v w] by simp
  from no_reach have "\<not> v\<rightarrow>\<setminus>dfs_restrict dfs+ w" by (metis reachable1_reachable restr_reachable1_reachable1)
  thus ?thesis
    apply (rule contrapos_np)
    apply (rule discover_finish_implies_reach[OF constr, of v w])
    apply (insert discover_neq_discover[OF constr verts ne] finish_neq_finish[OF constr verts ne])
    by force+
qed

lemma both_no_reach_discover_finish:
  assumes constr: "dfs_constructable dfs s"
  and no_reach: "\<not> v \<rightarrow>\<star> w" "\<not> w \<rightarrow>\<star> v"
  and fin: "v \<in> finished s" "w \<in> finished s"
  shows "\<delta> s w > \<phi> s v \<or> \<phi> s w < \<delta> s v"
proof -
  from constr fin have disc: "v \<in> discovered s" "w \<in> discovered s" using finished_implies_discovered by metis+
  with constr no_reach no_reach_discover_finish fin have "\<delta> s v > \<delta> s w \<or> \<phi> s v < \<phi> s w" "\<delta> s w > \<delta> s v \<or> \<phi> s w < \<phi> s v" by metis+
  thus ?thesis using correct_order constr fin by smt
qed

lemma rev_stack_vwalk_or_empty:
  assumes "dfs_constructable dfs s"
  shows "vwalk (rev (stack s)) \<G> \<or> stack s = []"
using assms
proof (induction)
  case (empty s s') 
  hence "vwalk (butlast (rev (stack s))) \<G> \<or> butlast (rev (stack s)) = []"
    using vwalk_butlast[of "rev (stack s)"] by simp
  with empty show ?case by (simp add: butlast_rev_tl)
next
  case (visit s s' e x xs) with wl_subset_verts[OF visit(2)] have "e \<in> V" by auto

  moreover
  from visit have "x \<in> V" and "set xs \<subseteq> V" using stack_subset_verts[OF visit(2)] by auto

  moreover
  from visit wl_subset_succs[OF visit(2), of 0] have "e \<in> succs x" by auto
  with visit have "set (vwalk_edges (rev (stack s))) \<subseteq> E" and "(x,e) \<in> E" unfolding succs_def by force+
  with visit(4) have "set (vwalk_edges (rev (stack s) @ [e])) \<subseteq> E" using vwalk_edges_append[of "rev (stack s)"] by simp
    
  ultimately have "vwalk (rev (stack s) @ [e]) \<G>" 
    using visit by (intro vwalkI) simp_all
  
  with visit show ?case by simp
qed auto

lemma rev_stack_vwalk:
  "dfs_constructable dfs s \<Longrightarrow> stack s \<noteq> [] \<Longrightarrow> vwalk (rev (stack s)) \<G>"
by (metis rev_stack_vwalk_or_empty)

lemma discover_start_eq_0:
  "dfs_constructable dfs s \<Longrightarrow> \<delta> s (start s) = 0"
proof (induct rule: dfs_constructable_induct)
  case start thus ?case unfolding dfs_constr_start_def by simp
next
  case (visit s s' e) hence start': "\<delta> s (start s') = 0" using dfs_step_preserves_start[OF dfs_next_dfs_step] by auto
  
  from visit.hyps have "start s = last (stack s)" using last_stack_is_start[OF visit(2)] by simp
  with dfs_step_preserves_start[OF dfs_next_dfs_step] visit have "start s' = last (stack s)" by auto
  with visit stack_distinct[OF visit(3)]  have "e \<noteq> start s'" by auto
  with start' visit show ?case by simp
qed auto

lemma discovered_neq_0:
  assumes constr: "dfs_constructable dfs s"
  and discovered: "v \<in> discovered s"
  and ne: "v \<noteq> start s"
  shows "\<delta> s v > 0"
proof -
  from discovered constr have "v \<in> V" "start s \<in> V" "start s \<in> discovered s" using discovered_subset_verts dfs_start_in_verts start_discovered by blast+
  with ne discovered have "\<delta> s (start s) \<noteq> \<delta> s v" using discover_neq_discover[OF constr] by force
  hence "\<delta> s v \<noteq> 0" using discover_start_eq_0[OF constr] by simp
  thus ?thesis by simp
qed

lemma discover_gt_start:
  "dfs_constructable dfs s \<Longrightarrow> v \<in> discovered s \<Longrightarrow> v \<noteq> start s \<Longrightarrow> \<delta> s v > \<delta> s (start s)"
by (metis discovered_neq_0 discover_start_eq_0)

lemma start_restr_reach_discovered:
  assumes constr: "dfs_constructable dfs s"
  and stack: "stack s \<noteq> []"
  and discovered: "v \<in> discovered s"
  and ne: "v \<noteq> start s"
  shows "start s \<rightarrow>\<setminus>dfs_restrict dfs+ v"
proof -
  from assms have "\<delta> s v > \<delta> s (start s)" using discover_gt_start by metis

  moreover
  from stack have "start s \<in> set (stack s)" using last_in_set last_stack_is_start[OF constr] by force
  with stack_implies_not_finished[OF constr] have "start s \<notin> finished s" by simp

  ultimately show ?thesis using discover_finish_implies_reach[OF constr start_discovered[OF constr] discovered] by auto
qed

lemma start_reach_discovered:
  "dfs_constructable dfs s \<Longrightarrow> stack s \<noteq> [] \<Longrightarrow> v \<in> discovered s \<Longrightarrow> start s \<rightarrow>\<star> v"
by (metis dfs_start_in_verts restr_reachable1_reachable self_reachable start_restr_reach_discovered verts_reduce)

lemma constr_from_restr_reach_stack:
  assumes constr: "s \<in> dfs_constr_from dfs x"
  and stack: "v \<in> set (stack s)"
  and ne: "v \<noteq> x"
  shows "x \<rightarrow>\<setminus>dfs_restrict dfs+ v"
proof -
  from stack have sne: "stack s \<noteq> []" by auto

  from constr have c: "dfs_constructable dfs s" using dfs_constr_from_constructable by metis
  with stack have in_d: "v \<in> discovered s" using stack_subset_discovered by auto

  from constr have "start s = x" using constr_from_implies_start by blast
  thus ?thesis using start_restr_reach_discovered[OF c sne in_d] ne by simp
qed

lemma constr_from_reach_stack:
  assumes constr: "s \<in> dfs_constr_from dfs x"
  and stack: "v \<in> set (stack s)"
  shows "x \<rightarrow>\<star> v"
proof (cases "x = v")
  case True with stack stack_subset_verts constr have "v \<in> V" by auto
  with True self_reachable show ?thesis by simp
next
  case False with constr_from_restr_reach_stack[OF assms] show ?thesis by blast
qed

lemma finished_implies_succs_discovered:
  "dfs_constructable dfs s \<Longrightarrow> v \<in> finished s \<Longrightarrow> succs v - dfs_restrict dfs \<subseteq> discovered s"
proof (induction rule: dfs_constructable_induct)
  case (empty s s' x _ w) with empty_wl_succs_discovered have "succs x - dfs_restrict dfs \<subseteq> discovered s'" by fastforce
  moreover from empty have "finished s' = insert x (finished s)" by simp
  ultimately show ?case using empty by (cases "v = x") bestsimp+
qed auto

lemma finished_implies_succs_discoveredI:
  "\<lbrakk> dfs_constructable dfs s; v \<in> finished s; w \<in> succs v; w \<notin> dfs_restrict dfs \<rbrakk> \<Longrightarrow> w \<in> discovered s"
by (auto dest!: finished_implies_succs_discovered)

lemma finished_no_reach_implies_succs_finished:
  "dfs_constructable dfs s \<Longrightarrow> v \<in> finished s \<Longrightarrow> w \<in> succs v \<Longrightarrow> w \<notin> dfs_restrict dfs \<Longrightarrow> \<not> w \<rightarrow>\<star> v \<Longrightarrow> w \<in> finished s"
proof (induction rule: dfs_constructable_induct)
  case (empty s s' x) then have "finished s' = insert x (finished s)" and d': "discovered s = discovered s'" by simp_all
  with empty.prems empty.IH show ?case
  proof (cases "v = x")
    case True with empty.prems have wd: "w \<in> discovered s'" using finished_implies_succs_discoveredI[OF empty(3)] by simp
    
    {
      assume "w \<notin> finished s'"
      with wd have "w \<in> set (stack s')" using discovered_not_finished_implies_stack[OF empty(3)] by simp
      with empty True have "w \<rightarrow>\<star> v" using tl_reachable_stack_hd[OF empty(2)] by auto
      with empty.prems have "False" by simp
    }
    thus ?thesis by (rule ccontr)
  qed simp
qed simp+

lemma not_finished_succs_impl_not_finished:
  "dfs_constructable dfs s \<Longrightarrow> v \<in> discovered s \<Longrightarrow> w \<in> discovered s \<Longrightarrow> \<delta> s v < \<delta> s w \<Longrightarrow> w \<notin> finished s \<Longrightarrow> w \<in> succs v \<Longrightarrow> v \<notin> finished s"
proof (induction rule: dfs_constructable_induct)
  case (empty s s' x xs) then have
    d': "discover s = discover s'" and
    f': "finished s' = insert x (finished s)" and
    s:  "stack s = x # xs" and s': "stack s' = xs"
    by simp_all

  from empty.prems have "w \<in> set (stack s')" using discovered_not_finished_implies_stack[OF empty(3)] by simp
  with s s' have ws: "w \<in> set (stack s)" by simp

  have "v \<noteq> x"
  proof (rule notI)
    assume "v = x"
    with ws s empty.prems have "\<delta> s w < \<delta> s v" using tl_lt_stack_hd_discover[OF empty(2)] by auto
    with d' have "\<delta> s' w < \<delta> s' v" by simp
    with empty.prems show "False" by simp
  qed
  with empty.prems empty.IH f' d' show ?case by simp
next
  case (visit s s' e) hence
    d': "discover s' = discover s (e \<mapsto> counter s)" and
    s': "stack s' = e # stack s" and
    f': "finish s = finish s'"
    by simp_all
  
  from visit.prems have ws': "w \<in> set (stack s')" using discovered_not_finished_implies_stack[OF visit(3)] by simp
  show ?case
  proof (cases "w = e")
    case True with visit(8,9) have "w \<notin> discovered s" and "w \<notin> dfs_restrict dfs" by simp_all
    hence "v \<notin> finished s" using visit.prems finished_implies_succs_discoveredI[OF visit(2)] by metis
    with f' show ?thesis by simp
  next
    case False with d' have wd: "discover s w = discover s' w" by simp
    show ?thesis
    proof (cases "v = e")
      case True with s' stack_implies_not_finished[OF visit(3)] show ?thesis by simp
    next
      case False with d' have "discover s v = discover s' v" by simp
      with wd f' visit.prems visit.IH show ?thesis by fastforce
    qed
  qed
 qed simp+

lemma finished_succs_finish:
  "dfs_constructable dfs s \<Longrightarrow> v \<in> discovered s \<Longrightarrow> w \<in> finished s \<Longrightarrow> w \<in> succs v \<Longrightarrow> \<delta> s w > \<delta> s v \<Longrightarrow> v \<notin> finished s \<or> \<phi> s w < \<phi> s v"
proof (induction rule: dfs_constructable_induct)
  case (empty s s' x) hence
    d': "discover s = discover s'" and
    f': "finish s' = finish s (x \<mapsto> counter s)"
    by simp_all

  from empty.prems have ne: "w \<noteq> v" by auto
  show ?case
  proof (cases "w = x")
    case True with empty have "w \<notin> finished s" using stack_implies_not_finished[OF empty(2)] by simp
    with empty.prems d' have "v \<notin> finished s"
      using not_finished_succs_impl_not_finished[OF empty(2)] finished_implies_discovered[OF empty(3)]
      by metis
    with True ne f' show ?thesis by force
  next
    case False with f' empty.prems have fw: "finish s w = finish s' w" by auto
    show ?thesis
    proof (cases "v = x")
      case True with f' have "\<phi> s' v = counter s" by simp
      moreover from empty.prems fw finish_lt_counter[OF empty(2)] have "\<phi> s' w < counter s" by force
      ultimately show ?thesis by simp
    next
      case False with f' have "finish s v = finish s' v" by simp
      with fw empty.prems d' empty.IH show ?thesis by force
    qed
  qed
next
  case (visit s s' e) hence
    d': "discover s' = discover s (e \<mapsto> counter s)" and
    s': "stack s' = e # stack s" and
    f': "finish s = finish s'"
    by simp_all
  
  with visit.prems finished_implies_not_stack[OF visit(3)] s' have "w \<noteq> e" by simp
  thus ?case
  proof (cases "v = e")
    case True with s' show ?thesis using stack_implies_not_finished[OF visit(3)] by simp
  next
    case False with d' `w \<noteq> e` have "discover s v = discover s' v" and "discover s w = discover s' w" by auto
    with f' visit.prems visit.IH show ?thesis by force
  qed
qed simp+

lemma finished_no_reach_nondiscovered:
  assumes constr: "dfs_constructable dfs s"
  and noreach: "\<forall>x \<in> finished s. \<forall> y \<in> set(stack s). \<not> x \<rightarrow>\<setminus>dfs_restrict dfs+ y"
  and nondisc: "z \<notin> discovered s"
  shows "\<forall>x \<in> finished s.  \<not> x \<rightarrow>\<setminus>dfs_restrict dfs+ z"
proof (rule ccontr)
  assume " \<not> (\<forall>x\<in>finished s. \<not> x \<rightarrow>\<setminus>dfs_restrict dfs+ z)"
  hence "\<exists>x \<in> finished s. x \<rightarrow>\<setminus>dfs_restrict dfs+ z" by simp
  then obtain x where "x \<rightarrow>\<setminus>dfs_restrict dfs+ z" "x \<in> finished s" by blast
  thus "False" using nondisc
  proof induction
    case (base y) with finished_implies_succs_discoveredI constr show "False" by blast
  next
    case (trans y z) hence y_disc: "y \<in> discovered s" by blast
    have "y \<in> finished s"
    proof (rule ccontr)
      assume "y \<notin> finished s"
      with y_disc discovered_not_finished_implies_stack[OF constr] have "y \<in> set (stack s)" by simp
      with trans noreach show "False" by simp
    qed
    with finished_implies_succs_discoveredI[OF constr] trans have "z \<in> finished s" by blast
    with trans finished_subset_discovered[OF constr] show "False" by auto
  qed
qed

lemma epath_generate_sws [case_names step finished, consumes 2]:
  assumes "epath v p w" and not_R: "set (ewalk_verts v p) \<inter> dfs_restrict dfs = {}"
  obtains 
    (step) s where "s \<in> dfs_constr_from dfs v" and 
                   "stack s = rev (ewalk_verts v p)" and
                   "discovered s = set (ewalk_verts v p)" and
                   "wl s \<noteq> []" "hd (wl s) = succs w" and
                   "dfs_cond dfs (state s)"
  | (finished) s where "s \<in> dfs_constr_from dfs v" "\<not> dfs_cond dfs (state s)"
proof -
  from assms have "ewalk v p w" and "distinct (ewalk_verts v p)" unfolding epath_def by simp_all
  thus ?thesis using that `ewalk v p w` not_R
  proof (induct rule: ewalk_induct_rev)
    case (Base v) thus ?case 
    proof (cases "dfs_cond dfs (state (dfs_constr_start dfs v))")
      case True with Base(1,2,6)  Base(3)[of "dfs_constr_start dfs v"] show ?thesis by simp
    next
      case False with Base(1,6) Base(4)[of "dfs_constr_start dfs v"] show ?thesis by simp
    qed
  next
    case (Snoc x y e es)

    from Snoc(1,2,7) have everts: "ewalk_verts v (es@[e]) = (ewalk_verts v es)@[y]" using ewalk_verts_join_l[of v es "[e]"] by (simp add: ewalk_verts_def)
    with Snoc(4,8) have d_everts: "distinct (ewalk_verts v es)" and y_not_R: "y \<notin> dfs_restrict dfs" and es_not_R: "set (ewalk_verts v es) \<inter> dfs_restrict dfs = {}" by auto
    
    from ewalk_join_inner_ends[OF Snoc(7)] Snoc(2) have "last (ewalk_verts v es) = x" by (simp add: ewalk_verts_def)
    with ewalk_join_ewalk_l[OF Snoc(7)] have ewalk: "ewalk v es x" by simp
  
    {
      fix s
      assume *: "s \<in> dfs_constr_from dfs v" "stack s = rev (ewalk_verts v es)" "discovered s = set (ewalk_verts v es)" "wl s \<noteq> []" "hd (wl s) = succs x" "dfs_cond dfs (state s)"
      moreover with Snoc have "y \<in> succs x" "hd (wl s) \<noteq> {}" unfolding succs_def by auto
      moreover from * ewalk_verts_non_Nil have "stack s \<noteq> []" by simp
      then obtain a as where "stack s = a#as" by (metis list.exhaust)
      moreover 
      let ?s' = "s\<lparr> state := dfs_action dfs (state s) s y, stack := y#a#as, wl := succs y#(succs x - {y})#tl (wl s), discover := (discover s)(y \<mapsto> counter s), counter := Suc (counter s)\<rparr>"
      from Snoc(4) everts have "y \<notin> set (ewalk_verts v es)" by auto
      with * have "y \<notin> discovered s" by simp
      ultimately have "dfs_next dfs s ?s'"
        unfolding dfs_next_def dfs_cond_compl_def
        using dfs_step_simps(3)[of s a as dfs] y_not_R
          apply (simp add: dfs_sws.defs)
          apply (intro exI[where x=y])
          apply fastforce
        done

      with * have "?s' \<in> dfs_constr_from dfs v" using dfs_constr_from.step by simp
      with Snoc(6) have thesis
      proof (cases "dfs_cond dfs (state ?s')")
        case True with * everts `stack s = a#as` have "stack ?s' = rev (ewalk_verts v (es@[e]))" "discovered ?s' = set (ewalk_verts v (es@[e]))" "wl ?s' \<noteq> []" "hd (wl ?s') = succs y"
          by simp_all
        with True `?s' \<in> dfs_constr_from dfs v` show ?thesis using Snoc(5) by metis
      qed simp
    }
    with Snoc(3)[OF d_everts _ _ ewalk es_not_R] Snoc(6) show ?case by metis
  qed
qed

lemma reach_generate_sws [case_names step finished, consumes 1]:
  assumes "v \<rightarrow>\<setminus>dfs_restrict dfs+ w"
  obtains 
    s where "s \<in> dfs_constr_from dfs v" and 
            "stack s \<noteq> []" "hd (stack s) = w" and
            "wl s \<noteq> []" "hd (wl s) = succs w" and
            "dfs_cond dfs (state s)"
  | s where "s \<in> dfs_constr_from dfs v" "\<not> dfs_cond dfs (state s)"
proof -
  from assms obtain p where "ewalk v p w" "set (ewalk_verts v p) \<inter> dfs_restrict dfs = {}" using restr_reachable1_def by blast
  with epath_ewalk_to_epath epath_verts_sub_ewalk_verts have "epath v (ewalk_to_epath p) w" "set (ewalk_verts v (ewalk_to_epath p)) \<inter> dfs_restrict dfs = {}" by blast+
  thus ?thesis
  proof (cases rule: epath_generate_sws[where dfs=dfs])
    case (step s)
    from `epath v (ewalk_to_epath p) w` have "ewalk v (ewalk_to_epath p) w" unfolding epath_def by simp
    hence "ewlast v (ewalk_to_epath p) = w" using ewalk_conv by simp
    hence "hd (rev (ewalk_verts v (ewalk_to_epath p))) = w" using hd_rev[OF ewalk_verts_non_Nil] by simp
    with step have "hd (stack s) = w" by simp
    with that(1) step show ?thesis by simp
  qed (metis that(2))
qed

text {* Now introduce a notion of a finished search. This enables us to proof properties about the result of a search. *}

definition dfs_finished :: "('S, 'n, 'X) dfs_algorithm_scheme \<Rightarrow> ('S, 'n) dfs_sws \<Rightarrow> bool"  where
  "dfs_finished dfs s \<equiv> dfs_constructable dfs s \<and> \<not>(\<exists> s'. dfs_next dfs s s')"

lemma dfs_finishedE:
  "\<lbrakk>dfs_finished dfs s; \<lbrakk> dfs_constructable dfs s; \<not>(\<exists> s'. dfs_next dfs s s') \<rbrakk> \<Longrightarrow> P \<rbrakk> \<Longrightarrow> P"
unfolding dfs_finished_def
by metis

lemma dfs_finishedI:
  "dfs_constructable dfs s \<Longrightarrow> \<not>(\<exists> s'. dfs_next dfs s s') \<Longrightarrow> dfs_finished dfs s"
unfolding dfs_finished_def
by metis

lemma dfs_finished_stackI:
  "dfs_constructable dfs s \<Longrightarrow> stack s = [] \<Longrightarrow> dfs_finished dfs s"
unfolding dfs_finished_def dfs_next_def dfs_cond_compl_def
 by simp

lemma dfs_finished_wlI:
  "dfs_constructable dfs s \<Longrightarrow> wl s = [] \<Longrightarrow> dfs_finished dfs s"
using dfs_finished_stackI length_wl_eq_stack
by fastforce

lemma dfs_finished_constructable[intro]:
  "dfs_finished dfs s \<Longrightarrow> dfs_constructable dfs s"
unfolding dfs_finished_def
by simp

lemma not_dfs_cond_implies_finished:
  "dfs_constructable dfs s \<Longrightarrow> \<not> dfs_cond dfs (state s) \<Longrightarrow> dfs_finished dfs s"
unfolding dfs_finished_def dfs_next_def dfs_cond_compl_def
by simp

definition "dfs_completed dfs s \<equiv> dfs_finished dfs s \<and> dfs_cond dfs (state s)"

lemma dfs_completed_finished:
  "dfs_completed dfs s \<Longrightarrow> dfs_finished dfs s"
unfolding dfs_completed_def
by simp

lemma finished_is_completed_or_not_cond:
  "dfs_finished dfs s \<Longrightarrow> dfs_completed dfs s \<or> \<not> dfs_cond dfs (state s)"
unfolding dfs_completed_def
by simp

lemma dfs_completed_constructable[dest,simp]:
  "dfs_completed dfs s \<Longrightarrow> dfs_constructable dfs s"
by (metis dfs_completed_finished dfs_finished_constructable)

lemma dfs_completedE:
  "\<lbrakk> dfs_completed dfs s; \<lbrakk> dfs_finished dfs s; dfs_cond dfs (state s) \<rbrakk> \<Longrightarrow> P \<rbrakk> \<Longrightarrow> P"
unfolding dfs_completed_def by metis

lemma dfs_completed_empty:
  "dfs_completed dfs s \<Longrightarrow> stack s = [] \<or> wl s = []"
unfolding dfs_finished_def dfs_next_def dfs_cond_compl_def dfs_completed_def
using dfs_step_exists
by auto

lemma dfs_completed_empty_stack:
  "dfs_completed dfs s \<Longrightarrow> stack s = []"
using length_wl_eq_stack dfs_completed_empty
by force

lemma dfs_completed_empty_wl:
  "dfs_completed dfs s \<Longrightarrow> wl s = []"
using length_wl_eq_stack dfs_completed_empty
by force

lemma dfs_completed_stackI:
  "dfs_constructable dfs s \<Longrightarrow> stack s = [] \<Longrightarrow> dfs_cond dfs (state s) \<Longrightarrow> dfs_completed dfs s"
unfolding dfs_completed_def
using dfs_finished_stackI
by simp

lemma dfs_completed_wlI:
  "dfs_constructable dfs s \<Longrightarrow> wl s = [] \<Longrightarrow> dfs_cond dfs (state s) \<Longrightarrow> dfs_completed dfs s"
using dfs_completed_stackI length_wl_eq_stack
by fastforce

lemma dfs_completed_revE [case_names prev, induct pred: dfs_completed]:
  assumes finished: "dfs_completed dfs s"
  and     prev:     "\<And>s r. \<lbrakk> dfs_constructable dfs r; dfs_next dfs r s; dfs_completed dfs s; s = r\<lparr> stack := [], wl := [], finish := (finish r)(hd (stack r) \<mapsto> counter r), counter := Suc (counter r), state := dfs_post dfs (state r) r (hd (stack r))\<rparr> \<rbrakk> \<Longrightarrow> P s"
  shows "P s"
proof -
  from finished have sempty: "stack s = []" "wl s = []" using dfs_completed_empty_stack dfs_completed_empty_wl by auto
  
  from finished have "dfs_constructable dfs s" by (rule dfs_completed_constructable)
  thus ?thesis
  proof (cases rule: dfs_constructable.cases)
    case start thus ?thesis using sempty by (simp add: dfs_constr_start_def)
  next
    case (step r)
    then have "s = r\<lparr> stack := [], wl := [], finish := (finish r)(hd (stack r) \<mapsto> counter r), counter := Suc (counter r), state := dfs_post dfs (state r) r (hd (stack r)) \<rparr>"
      using sempty
      by (cases rule: dfs_next_cases_elem) auto
    with assms step show ?thesis by simp
  qed
qed

lemma dfs_completed_prev_stack:
  assumes finished: "dfs_completed dfs s"
  and     step:     "dfs_next dfs r s"
  and     constr:   "dfs_constructable dfs r"
  shows  "stack r = [start s]"
proof -
  note emp = dfs_completed_empty_stack[OF finished]
  show ?thesis
  using step emp assms
  proof (cases)
    case (empty x xs) with emp obtain e where "stack r = [e]" by simp
    with last_stack_is_start[OF constr] have "stack r = [start r]" by simp
    with step dfs_step_preserves_start[OF dfs_next_dfs_step, OF step] show ?thesis by auto
  qed auto
qed

lemma completed_start_finished:
  "dfs_completed dfs s \<Longrightarrow> start s \<in> finished s"
proof (induct rule: dfs_completed_revE)
  case (prev s r) then have "finish s = finish r (hd (stack r) \<mapsto> counter r)" by simp
  with prev have "finish s = finish r (start s \<mapsto> counter r)" using dfs_completed_prev_stack[OF prev(3,2,1)] by force
  thus ?case by auto
qed

lemma completed_start_finish_eq_counter:
  "dfs_completed dfs s \<Longrightarrow> \<phi> s (start s) = counter s - 1"
proof (induct rule: dfs_completed_revE)
  case (prev s r) then have "finish s = finish r (hd (stack r) \<mapsto> counter r)" by simp
  with prev have "finish s = finish r (start s \<mapsto> counter r)" using dfs_completed_prev_stack[OF prev(3,2,1)] by force
  with prev show ?case by simp
qed

lemma completed_finish_lt_start:
  "dfs_completed dfs s \<Longrightarrow> v \<in> finished s \<Longrightarrow> v \<noteq> start s \<Longrightarrow> \<phi> s v < \<phi> s (start s)"
proof (induct rule: dfs_completed_revE)
  case (prev s r) then have "finish s = finish r (hd (stack r) \<mapsto> counter r)" by simp
  with prev have *:"finish s = finish r (start s \<mapsto> counter r)" using dfs_completed_prev_stack[OF prev(3,2,1)] by force
  hence "\<phi> s (start s) = counter r" by simp

  moreover
  from prev.prems * have fv: "finish r v = finish s v" by simp
  with prev.prems finish_lt_counter[OF prev(1)] have "\<phi> r v < counter r" by force
  with fv have "\<phi> s v < counter r" by simp

  ultimately show ?case by simp
qed

lemma completed_finished_eq_discovered:
  "dfs_completed dfs s \<Longrightarrow> finished s = discovered s" 
using dfs_completed_empty_stack
using finished_stack_eq_discovered[OF dfs_completed_constructable]
by (metis List.set.simps(1) Un_empty_right)

lemma completed_start_finished_restr_reach:
  assumes compl: "dfs_completed dfs s"
  and finished: "v \<in> finished s"
  and ne: "v \<noteq> start s"
  shows "start s \<rightarrow>\<setminus>dfs_restrict dfs+ v"
using assms
proof (induct rule: dfs_completed_revE)
  case (prev s r) with finished_implies_discovered have "v \<in> discovered s" "stack r \<noteq> []" "discovered s = discovered r" by (blast, simp_all)
  with prev have "start r \<rightarrow>\<setminus>dfs_restrict dfs+ v" using start_restr_reach_discovered[OF prev(1)] by simp
  with prev show ?case by simp
qed

lemma completed_start_finished_reach:
  assumes compl: "dfs_completed dfs s"
  and finished: "v \<in> finished s"
  shows "start s \<rightarrow>\<star> v"
proof (cases "start s = v")
  case True with dfs_start_in_verts compl self_reachable show ?thesis by auto
next
  case False with assms completed_start_finished_restr_reach show ?thesis by blast
qed

lemma completed_reach_implies_finished:
  assumes compl: "dfs_completed dfs s"
  and fin: "v \<in> finished s"
  and reach: "v \<rightarrow>\<setminus>dfs_restrict dfs+ w"
  shows "w \<in> finished s"
proof -
  {
    fix v z
    assume "v \<in> finished s" "z \<in> succs v" "z \<notin> dfs_restrict dfs"
    with finished_implies_succs_discoveredI compl have "z \<in> discovered s" by blast
    with compl completed_finished_eq_discovered have "z \<in> finished s" by blast
  }
  with reach fin show ?thesis by induction blast+
qed

lemma completed_start_reach_finished:
  "dfs_completed dfs s \<Longrightarrow> start s \<rightarrow>\<setminus>dfs_restrict dfs+ v \<Longrightarrow> v \<in> finished s"
by (metis completed_start_finished completed_reach_implies_finished)

lemma completed_finished_eq_reachable:
  assumes compl: "dfs_completed dfs s"
  shows "finished s = {v. v = start s \<or> start s \<rightarrow>\<setminus>dfs_restrict dfs+ v}" (is "... = ?F")
proof (rule set_eqI[OF iffI])
  fix x
  assume "x \<in> finished s"
  hence "x = start s \<or> start s \<rightarrow>\<setminus>dfs_restrict dfs+ x"
    using  completed_start_finished_restr_reach[OF compl] completed_start_finished[OF compl]
    by (cases "x = start s") simp_all
  thus "x \<in> ?F" ..
next
  fix x
  assume "x \<in> ?F"
  hence "x = start s \<or> start s \<rightarrow>\<setminus>dfs_restrict dfs+ x" ..
  thus "x \<in> finished s" using completed_start_finished[OF compl] completed_start_reach_finished[OF compl] by blast
qed

lemma no_restrict_finished_eq_reachable:
  assumes compl: "dfs_completed dfs s"
  and no_restr: "dfs_restrict dfs \<inter> V = {}"
  shows "finished s = {v. start s \<rightarrow>\<star> v}"
proof -
  from compl have "finished s = {v. v = start s \<or> start s \<rightarrow>\<setminus>dfs_restrict dfs+ v}" by (fact completed_finished_eq_reachable)
  also from no_restr have "... = {v. v = start s \<or> start s \<rightarrow>+ v}" using restr_reachable1_no_verts by simp
  also from reachable1_reachable have "... = {v. v = start s \<or> start s \<rightarrow>\<star> v}" by blast
  also from self_reachable dfs_start_in_verts compl have "... = {v. start s \<rightarrow>\<star> v}" by force
  finally show ?thesis .
qed

lemma completed_finished_succs_finish:
  assumes compl: "dfs_completed dfs s"
  and finished: "v \<in> finished s"
  and d_lt: "\<delta> s v < \<delta> s w"
  and succs: "w \<in> succs v"
  and "w \<notin> dfs_restrict dfs"
  shows "\<phi> s w < \<phi> s v"
proof -
  from compl have constr: "dfs_constructable dfs s" by simp
  moreover from finished have "v \<in> discovered s" using finished_implies_discovered[OF constr] by simp
  moreover hence "v \<notin> dfs_restrict dfs" using discovered_not_restricted[OF constr] by simp
  with `w \<notin> dfs_restrict dfs` succs succs_restricted have "v \<rightarrow>\<setminus>dfs_restrict dfs+ w" by simp
  with completed_reach_implies_finished[OF compl finished] have "w \<in> finished s" .

  ultimately show ?thesis using finished_succs_finish succs d_lt finished by fastforce
qed

lemma dfs_finished_cases [case_names start step, consumes 1]:
  assumes fin: "dfs_finished dfs s"
  and start: "\<not> dfs_cond dfs (dfs_start dfs (start s)) \<Longrightarrow> P"
  and step: "\<And>r. \<lbrakk> dfs_constructable dfs r; dfs_next dfs r s \<rbrakk> \<Longrightarrow> P"
  shows "P"
proof (cases "s = dfs_constr_start dfs (start s)")
  case True hence "stack s \<noteq> []" and st: "state s = dfs_start dfs (start s)" using dfs_constr_start_simps[of dfs "start s"] by simp_all
  hence "\<not> dfs_completed dfs s" using dfs_completed_empty_stack by blast
  with fin finished_is_completed_or_not_cond have "\<not> dfs_cond dfs (state s)" by blast
  with st start show ?thesis by simp
next
  case False 
  from fin have "dfs_constructable dfs s" by auto
  hence "\<exists>r. dfs_constructable dfs r \<and> dfs_next dfs r s" using False
    by (cases rule: dfs_constructable.cases) auto
  with step show ?thesis by blast
qed
end

subsection {* Lifting into the while'd *}
text {* We add an invariant field to the dfs. *}

record ('S, 'n) dfs_algorithm_invar = "('S, 'n) dfs_algorithm" +
  dfs_invar :: "('S, 'n) dfs_sws \<Rightarrow> bool"  -- "an invariant that each state built during the search must satisfy. This is dropped in the implementation."

context finite_digraph
begin

text {* @{term dfs_invar_compl} specifies the invariants that must hold on each working state. 
If a state does not fulfill parts of the invariant, the behavior of the dfs-algorithm is undefined.
 As stated later on by @{term dfs_relation_terminates}, it is guaranteed, that when starting with an 
invariant-obeying state, all states reached by the algorithms also obey them. 

Similar to @{term dfs_cond_compl} it is split between invariants of the framework and invariants of the implementation.
Here the invariant of the framework is, that the state is constructable, as given by @{term dfs_constructable} defined earlier. *}

definition dfs_invar_compl :: "('S, 'n, 'X) dfs_algorithm_invar_scheme \<Rightarrow> 'n \<Rightarrow> ('S , 'n) dfs_sws \<Rightarrow> bool"
where
  "dfs_invar_compl dfs x s \<equiv> s \<in> dfs_constr_from dfs x \<and> dfs_invar dfs s"

lemma dfs_invar_compl_invar[dest]:
  "dfs_invar_compl dfs x s \<Longrightarrow> dfs_invar dfs s"
unfolding dfs_invar_compl_def
by simp

lemma dfs_invar_compl_constr_from[dest]:
  "dfs_invar_compl dfs x s \<Longrightarrow> s \<in> dfs_constr_from dfs x"
unfolding dfs_invar_compl_def
by auto

lemma dfs_invar_compl_constr[dest]:
  "dfs_invar_compl dfs x s \<Longrightarrow> dfs_constructable dfs s"
by auto

lemma dfs_invar_complE:
  "\<lbrakk> dfs_invar_compl dfs x s; \<lbrakk> s \<in> dfs_constr_from dfs x; dfs_invar dfs s \<rbrakk> \<Longrightarrow> P \<rbrakk> \<Longrightarrow> P"
unfolding dfs_invar_compl_def
by metis

lemma dfs_invar_complI:
  "\<lbrakk> s \<in> dfs_constr_from dfs x; dfs_invar dfs s \<rbrakk> \<Longrightarrow> dfs_invar_compl dfs x s"
unfolding dfs_invar_compl_def
by simp

text {* And now we initialize the datastructures of the while-framework and show, that we behave as the framework expects us to behave. *}

definition dfs_fun :: "('S,'n,'X) dfs_algorithm_invar_scheme \<Rightarrow> 'n \<Rightarrow> ('S,'n) dfs_sws nres" where
  "dfs_fun dfs x \<equiv> 
            WHILE\<^sub>T\<^bsup>(dfs_invar_compl dfs x)\<^esup> (dfs_cond_compl dfs) (\<lambda> s. SPEC (\<lambda>s'. dfs_next dfs s s')) (dfs_constr_start dfs x)"

text {* We now state certain predicates the implementations have to obey. These predicates are used as assumptions
for specifying properties of the DFS-algorithm. *}

text {* Start with the need of the implementation to preserve its invariants. *}
definition dfs_preserves_invar :: "('S, 'n, 'X) dfs_algorithm_invar_scheme \<Rightarrow> bool"
where
  "dfs_preserves_invar dfs \<equiv>
   (\<forall>x \<in> V. dfs_invar dfs (dfs_constr_start dfs x)) \<and>
   (\<forall>x s s'. dfs_invar_compl dfs x s \<and> dfs_next dfs s s' \<longrightarrow>
                     dfs_invar dfs s')"

lemma dfs_preserves_invarI:
  "\<lbrakk>\<And>x s s'. \<lbrakk> dfs_invar_compl dfs x s; dfs_next dfs s s' \<rbrakk> \<Longrightarrow> dfs_invar dfs s'; \<And>x. x \<in> V \<Longrightarrow> dfs_invar dfs (dfs_constr_start dfs x) \<rbrakk> 
  \<Longrightarrow> dfs_preserves_invar dfs"
unfolding dfs_preserves_invar_def
by blast

lemma dfs_preserves_invarD[intro]:
  "\<lbrakk> dfs_preserves_invar dfs; dfs_invar_compl dfs x s; dfs_next dfs s s' \<rbrakk> \<Longrightarrow> dfs_invar dfs s'"
unfolding dfs_preserves_invar_def
by blast

lemma dfs_preserves_invarE[elim]:
  assumes "dfs_preserves_invar dfs"
  and "\<lbrakk>\<And>x s s'. \<lbrakk> dfs_invar_compl dfs x s; dfs_next dfs s s' \<rbrakk> \<Longrightarrow> dfs_invar dfs s' \<rbrakk> \<Longrightarrow> P"
  shows "P"
  using assms
  unfolding dfs_preserves_invar_def
  by blast

(* Also show, that our algorithm also preserves its own invariants. Else the invariant-preservation of the
implementation would be worthless :) *)

lemma dfs_preserves_invar_compl:
  assumes "dfs_preserves_invar dfs"
  and     "dfs_invar_compl dfs x s"
  and     "dfs_next dfs s s'"
  shows "dfs_invar_compl dfs x s'"
using assms
by (metis dfs_constr_from.step dfs_invar_compl_def dfs_preserves_invarD)

lemma constr_from_invar_compl:
  "s \<in> dfs_constr_from dfs x \<Longrightarrow> dfs_preserves_invar dfs \<Longrightarrow> dfs_invar_compl dfs x s"
proof (induct rule: dfs_constr_from.induct)
  case (step s s') with dfs_preserves_invar_compl show ?case by simp
next
  case start thus ?case by (simp add: dfs_invar_compl_def dfs_preserves_invar_def)
qed

lemma constr_from_invarI[intro!]:
  "s \<in> dfs_constr_from dfs x \<Longrightarrow> dfs_preserves_invar dfs \<Longrightarrow> dfs_invar dfs s"
by (auto dest: constr_from_invar_compl)

lemma dfs_constructable_invarI[intro!]:
  "dfs_constructable dfs s \<Longrightarrow> dfs_preserves_invar dfs \<Longrightarrow> dfs_invar dfs s"
by (auto elim!: dfs_constructable_constr_from_start)

text {* Introduce a measure, that lessens with each search-step. 
Therefore we define a measure to be a four-tuple, s.t. with each step one element gets smaller and all the previous ones remain equal:
\begin{enumerate}
   \item undiscovered nodes (lessens in the @{text visit} phase)
   \item length of the search stack (lessens in the @{text empty} phase)
   \item unchecked successors (lessens in the phases @{text visit} and @{text remove}, but for @{text visit} the first item already applies)
\end{enumerate} *}

abbreviation 
  "dfs_step_measure \<equiv> less_than <*lex*> less_than <*lex*> less_than"

lemma wf_dfs_step_measure:
  "wf dfs_step_measure"
by auto

definition ws_to_measure :: "('n, 'X) dfs_ws_scheme \<Rightarrow> (nat \<times> nat \<times> nat)"
where
  "ws_to_measure ws \<equiv> ( card (V - discovered ws), length (stack ws), card (hd (wl ws)))"

definition ws_rel :: "(('n, 'X) dfs_ws_scheme \<times> ('n, 'X) dfs_ws_scheme) set" where
  "ws_rel = {(ws', ws). (ws_to_measure ws', ws_to_measure ws) \<in> dfs_step_measure}"

lemma ws_rel_alt_def:
  "ws_rel = inv_image dfs_step_measure ws_to_measure"
unfolding ws_rel_def inv_image_def
by simp

theorem wf_ws_rel:
  "wf ws_rel"
unfolding ws_rel_alt_def
by (rule wf_inv_image[OF wf_dfs_step_measure])

lemma ws_rel_intro:
  "(ws_to_measure ws', ws_to_measure ws) \<in> dfs_step_measure \<Longrightarrow> (ws', ws) \<in> ws_rel"
unfolding ws_rel_def
by simp

text {* Now show, that this measure indeed applies to the algorithm. *}
lemma dfs_next_in_ws_rel:
  assumes step: "dfs_next dfs s s'"
  and inv: "dfs_invar_compl dfs x s"
  shows "(s', s) \<in> ws_rel"
using assms
proof (cases rule: dfs_next_cases_elem)
  case (empty x xs)
  then have "card (V - discovered s) = card (V - discovered s')" by simp
  with empty show ?thesis by (simp add: ws_to_measure_def ws_rel_intro)
next
  case (remove e) 
  from inv have c: "dfs_constructable dfs s" by auto
  with wl_subset_succs[OF this, of 0] remove have *: "e \<in> succs (hd (stack s'))" by auto

  from remove wl_finite[OF c] have F: "finite (hd (wl s))" using hd_in_set by simp

  from remove have "card (V - discovered s) = card (V - discovered s')" by auto
  moreover from remove have "length (stack s) = length (stack s')" by auto
  moreover from remove * have "hd (wl s) \<supset> hd (wl s')" by auto
  with psubset_card_mono[OF F] have "card (hd (wl s)) > card (hd (wl s'))" .

  ultimately show ?thesis by (simp add: ws_to_measure_def ws_rel_intro)
next 
  case (restrict e)
  from inv have c: "dfs_constructable dfs s" by auto
  with wl_subset_succs[OF this, of 0] restrict have *: "e \<in> succs (hd (stack s'))" by auto

  from restrict wl_finite[OF c] have F: "finite (hd (wl s))" using hd_in_set by simp

  from restrict have "card (V - discovered s) = card (V - discovered s')" by auto
  moreover from restrict have "length (stack s) = length (stack s')" by auto
  moreover from restrict * have "hd (wl s) \<supset> hd (wl s')" by auto
  with psubset_card_mono[OF F] have "card (hd (wl s)) > card (hd (wl s'))" .

  ultimately show ?thesis by (simp add: ws_to_measure_def ws_rel_intro)
next
  case visit hence dsub: "discovered s \<subset> discovered s'" by auto
  
  from finite_V have "finite (V - discovered s)" ..
  
  moreover
  from step inv have "dfs_constructable dfs s'" by (metis dfs_constructable.step dfs_invar_compl_constr)
  with visit have "discovered s' \<subseteq> V" using discovered_subset_verts[of dfs s'] by (simp_all add: dfs_sws.defs)

  with dsub have "V - discovered s' \<subset> V - discovered s" by blast

  ultimately have "card (V - discovered s) > card (V - discovered s')" by (rule psubset_card_mono)
  then show ?thesis by (simp add: ws_to_measure_def ws_rel_intro)
qed

lemma exists_finished:
  assumes pre: "x \<in> V" "x \<notin> dfs_restrict dfs"
  and inv: "dfs_preserves_invar dfs"
  shows "\<exists>s \<in> dfs_constr_from dfs x. dfs_finished dfs s"
proof -
  from assms have "dfs_constr_start dfs x \<in> dfs_constr_from dfs x" by simp
  with wfE_min[OF wf_ws_rel] obtain z where z: "z \<in> dfs_constr_from dfs x" "\<And>y. (y,z) \<in> ws_rel \<Longrightarrow> y \<notin> dfs_constr_from dfs x" by blast
  moreover with inv have "dfs_invar_compl dfs x z" using constr_from_invar_compl by blast
  ultimately have "\<not> (\<exists>y. dfs_next dfs z y)" by (blast intro: dfs_constr_from.step dfs_next_in_ws_rel)
  with z show ?thesis by (blast intro: dfs_finishedI)
qed

theorem dfs_fun_correct:
  assumes "dfs_preserves_invar dfs"
  and "x \<in> V" "x \<notin> dfs_restrict dfs"
  shows "dfs_fun dfs x \<le> SPEC (\<lambda> s. s \<in> dfs_constr_from dfs x \<and> dfs_finished dfs s)"
unfolding dfs_fun_def
proof (intro refine_vcg)
  show "wf ws_rel" by (fact wf_ws_rel)
next
  from assms show "dfs_invar_compl dfs x (dfs_constr_start dfs x)" unfolding dfs_invar_compl_def dfs_preserves_invar_def by (auto intro!: dfs_constructable.start)
next
  from assms show "\<And>s s'. \<lbrakk>dfs_cond_compl dfs s; dfs_invar_compl dfs x s; dfs_next dfs s s'\<rbrakk> \<Longrightarrow> dfs_invar_compl dfs x s' \<and> (s', s) \<in> ws_rel"
    by (blast intro: dfs_preserves_invar_compl dfs_next_in_ws_rel)
next
  show "\<And>s. \<lbrakk>dfs_invar_compl dfs x s; \<not> dfs_cond_compl dfs s\<rbrakk> \<Longrightarrow> s \<in> dfs_constr_from dfs x \<and> dfs_finished dfs s" by (metis dfs_finishedI dfs_invar_complE dfs_invar_compl_constr dfs_nextE)
qed

lemma dfs_fun_nofail[refine_pw_simps]:
  assumes "dfs_preserves_invar dfs"
  and "x \<in> V" "x \<notin> dfs_restrict dfs"
  shows "nofail (dfs_fun dfs x)"
using dfs_fun_correct[OF assms]
by (rule pwD) simp

lemma dfs_fun_finished:
  assumes "dfs_preserves_invar dfs"
  and "x \<in> V" "x \<notin> dfs_restrict dfs"
  and "inres (dfs_fun dfs x) s"
  shows "dfs_finished dfs s"
proof -
  note dfs_fun_correct[OF assms(1-3)]
  with assms(4) have "inres (SPEC (\<lambda>s. s \<in> dfs_constr_from dfs x \<and> dfs_finished dfs s)) s" using pwD2 by best
  thus ?thesis by simp 
qed

lemma dfs_fun_constructable:
  assumes "dfs_preserves_invar dfs"
  and "x \<in> V" "x \<notin> dfs_restrict dfs"
  and "inres (dfs_fun dfs x) s"
  shows "s \<in> dfs_constr_from dfs x"
proof -
  note dfs_fun_correct[OF assms(1-3)]
  with assms(4) have "inres (SPEC (\<lambda>s. s \<in> dfs_constr_from dfs x \<and> dfs_finished dfs s)) s" using pwD2 by best
  thus ?thesis by simp
qed

lemma dfs_step_refine [refine_transfer]:
  assumes "dfs_cond_compl dfs s"
  shows "RETURN (SOME s'. s' \<in> dfs_step dfs s) \<le> SPEC (\<lambda> s'. dfs_cond_compl dfs s \<and> s' \<in> dfs_step dfs s)"
using dfs_cond_compl_step_exists[OF assms] assms
by (auto intro: someI2_ex)

schematic_lemma dfs_fun_code:
  "RETURN ?f \<le> dfs_fun dfs x"
unfolding dfs_fun_def dfs_next_def
by (refine_transfer while.transfer_WHILEIT_esc_cond)

lemma dfs_fun_nonempty:
  "\<exists>s. inres (dfs_fun dfs x) s"
proof (cases "dfs_fun dfs x")
  case (RES X) with dfs_fun_code[of dfs x] have "X \<noteq> {}" by auto
  with RES show ?thesis by auto
qed simp

(* A lemma that allows to show that an equality holds on all generated results.
This is f.ex. needed for these implementations that use the state to control abortion of the search, 
so that they can show, that all results are equal regarding the abortion.

Therefore the 'thrd_d' is tailored towards this use-case, where the first property P is directly connected
with the abortion condition. *)

lemma dfs_fun_pred:
  assumes x: "x \<in> V" "x \<notin> dfs_restrict dfs"
  and pres_inv: "dfs_preserves_invar dfs"
  and one_d: "\<And>s. \<lbrakk> s \<in> dfs_constr_from dfs x; \<not> dfs_cond_compl dfs s \<rbrakk> \<Longrightarrow> P s \<Longrightarrow> Q s"
  and snd_d: "\<And>s. \<lbrakk> s \<in> dfs_constr_from dfs x; dfs_completed dfs s \<rbrakk> \<Longrightarrow> Q s \<Longrightarrow> P s"
  and thrd_d: "\<And>s. \<lbrakk> s \<in> dfs_constr_from dfs x; \<not> dfs_cond dfs (state s) \<rbrakk> \<Longrightarrow> P s" 
  shows "dfs_fun dfs x \<le> SPEC (\<lambda>s. P s \<longleftrightarrow> Q s)"
unfolding dfs_fun_def
apply (intro refine_vcg)
  apply (fact wf_ws_rel)
  apply (simp add: dfs_invar_compl_def x pres_inv[unfolded dfs_preserves_invar_def])
  apply (simp add: pres_inv dfs_next_in_ws_rel dfs_preserves_invar_compl)
proof -
  fix s
  assume "dfs_invar_compl dfs x s" and ncond: "\<not> dfs_cond_compl dfs s"
  hence constr: "s \<in> dfs_constr_from dfs x" and
        cond: "stack s = [] \<or> wl s = [] \<or> \<not> dfs_cond dfs (state s)" 
    unfolding dfs_cond_compl_def dfs_invar_compl_def by simp_all
  show "P s \<longleftrightarrow> Q s"
  proof
    assume "P s"
    with one_d constr ncond show "Q s" by metis
  next
    assume "Q s" 
    with constr cond show "P s" 
    proof (cases "dfs_cond dfs (state s)")
      case True with cond have "stack s = [] \<or> wl s = []" by simp
      with length_wl_eq_stack constr have "stack s = []" by force
      with True have "dfs_completed dfs s" using dfs_completed_stackI constr by force
      with `Q s` show ?thesis using snd_d constr by metis
    qed (metis thrd_d)
  qed
qed

end

section {* Simple DFS *}
text {* Introduce a simple dfs algorithm, that ignores all post and pre functions and just returns the filled discover and finish sets. *}

definition simple_dfs :: "(unit, 'n) dfs_algorithm_invar" where
  "simple_dfs = \<lparr> dfs_cond = \<lambda>_. True, 
                   dfs_action = \<lambda>_ _ _ . (), 
                   dfs_post = \<lambda>_ _ _. (), 
                   dfs_remove = \<lambda>_ _ _. (), 
                   dfs_start = \<lambda>_. (), 
                   dfs_restrict = {}, 
                   dfs_invar = \<lambda>_. True \<rparr>"

lemma simple_dfs_simps [simp]:
  "dfs_cond simple_dfs S"
  "dfs_action simple_dfs S s x = ()"
  "dfs_post simple_dfs S s x = ()"
  "dfs_remove simple_dfs S s x = ()"
  "dfs_start simple_dfs x = ()"
  "dfs_invar simple_dfs s"
  "dfs_restrict simple_dfs = {}"
by (simp_all add: simple_dfs_def)

lemma (in finite_digraph) simple_dfs_preserves_invar:
  "dfs_preserves_invar simple_dfs"
by (rule dfs_preserves_invarI) simp_all
end
