theory Ample_Abstract
imports
  "System_Interpreted_Traces"
  "Extensions/Relation_Extensions"
begin

  (*
    formalization of the abstract partial order reduction theory
    definitions and proofs adapted from
      [1] "Combining Partial Order Reductions with On-the-Fly Model-Checking"
        Doron Peled, FMSD, 1996
      [2] "Formal Verification of a Partial-Order Reduction Technique for Model Checking"
        Ching-Tsun Chou and Doron Peled, TACAS, 1996
  *)

  locale ample_base =
    system_interpreted_traces en ex int ind +
    wellfounded_relation src
    for en :: "'state \<Rightarrow> 'action set"
    and ex :: "'action \<Rightarrow> 'state \<Rightarrow> 'state"
    and int :: "'state \<Rightarrow> 'interpretation"
    and ind :: "'action \<Rightarrow> 'action \<Rightarrow> bool"
    and src :: "'state \<Rightarrow> 'state \<Rightarrow> bool"
  begin

    definition ample_set :: "'state \<Rightarrow> 'action set \<Rightarrow> bool"
      where "ample_set q A \<equiv>
        A \<subseteq> en q \<and>
        (A \<subset> en q \<longrightarrow> A \<noteq> {}) \<and>
        (\<forall> a. A \<subset> en q \<longrightarrow> a \<in> A \<longrightarrow> src (ex a q) q) \<and>
        (A \<subset> en q \<longrightarrow> A \<subseteq> invisible) \<and>
        (\<forall> w. A \<subset> en q \<longrightarrow> w \<in> words_fin q \<longrightarrow> A \<inter> set w = {} \<longrightarrow> Ind A (set w))"

    (* implicit in [1, 2] *)
    lemma ample_subset:
      assumes "ample_set q A"
      shows "A \<subseteq> en q"
      using assms unfolding ample_set_def by auto
    (* implicit in [1], condition C0 in [2] *)
    lemma ample_nonempty:
      assumes "ample_set q A" "A \<subset> en q"
      shows "A \<noteq> {}"
      using assms unfolding ample_set_def by auto
    (* condition C2 in [1, 2] *)
    lemma ample_wellfounded:
      assumes "ample_set q A" "A \<subset> en q" "a \<in> A"
      shows "src (ex a q) q"
      using assms unfolding ample_set_def by auto
    (* condition C3' in [1], not needed in [2] *)
    lemma ample_invisible:
      assumes "ample_set q A" "A \<subset> en q"
      shows "A \<subseteq> invisible"
      using assms unfolding ample_set_def by auto
    (* condition C1 in [1, 2] *)
    lemma ample_independent:
      assumes "ample_set q A" "A \<subset> en q" "w \<in> words_fin q" "A \<inter> set w = {}"
      shows "Ind A (set w)"
      using assms unfolding ample_set_def by auto

    lemma ample_en[intro]: "ample_set q (en q)" unfolding ample_set_def by blast

  end

  locale ample_abstract =
    S: system_complete en ex init int +
    R!: system_complete ren ex init int +
    ample_base en ex int ind src
    for en :: "'state \<Rightarrow> 'action set"
    and ex :: "'action \<Rightarrow> 'state \<Rightarrow> 'state"
    and init :: "'state set"
    and int :: "'state \<Rightarrow> 'interpretation"
    and ind :: "'action \<Rightarrow> 'action \<Rightarrow> bool"
    and src :: "'state \<Rightarrow> 'state \<Rightarrow> bool"
    and ren :: "'state \<Rightarrow> 'action set"
    +
    assumes reduction_ample: "q \<in> reachable \<Longrightarrow> ample_set q (ren q)"
  begin

    lemma reduction_words_fin:
      assumes "q \<in> reachable"
      shows "R.words_fin q \<subseteq> S.words_fin q"
    proof
      fix w
      assume 1: "w \<in> R.words_fin q"
      show "w \<in> words_fin q" using 1 assms ample_subset reduction_ample by (induct, fast+)
    qed
    lemma reduction_words_inf: "q \<in> reachable \<Longrightarrow> R.words_inf q \<subseteq> S.words_inf q"
      using reduction_words_fin by (intro subsetI words_infI_construct, auto)

    (* lemma 3.7 in [1] *)
    lemma reduction_step:
      assumes "q \<in> reachable" "w \<in> words_inf q"
      obtains
        (deferred) a where "a \<in> ren q" "[a] \<preceq>\<^sub>F\<^sub>I w" |
        (omitted) "ren q \<subseteq> invisible" "Ind (ren q) (range w)"
    proof (cases "en q = ren q")
      case True
      have 1: "w 0 ## suffix 1 w \<in> words_inf q" using assms(2) by simp
      show ?thesis
      proof (rule deferred)
        show "w 0 \<in> ren q" using True 1 by blast
        show "[w 0] \<preceq>\<^sub>F\<^sub>I w" by simp
      qed
    next
      case False
      have 1: "ren q \<subset> en q" using ample_subset reduction_ample assms(1) False by auto
      show ?thesis
      proof (cases "ren q \<inter> range w = {}")
        case True
        show ?thesis
        proof (rule omitted)
          show "ren q \<subseteq> invisible" using ample_invisible reduction_ample assms(1) 1 by auto
          show "Ind (ren q) (range w)"
          proof safe
            fix a i
            assume 2: "a \<in> ren q"
            have 3: "Ind (ren q) (set (prefix (Suc i) w))"
            proof (rule ample_independent)
              show "ample_set q (ren q)" using reduction_ample assms(1) by this
              show "ren q \<subset> en q" using 1 by this
              show "prefix (Suc i) w \<in> words_fin q" using prefix_words_inf assms by blast
              show "ren q \<inter> set (prefix (Suc i) w) = {}" using True by auto
            qed
            show "ind a (w i)" using 2 3 by auto
          qed
        qed
      next
        case False
        obtain u a v where 2: "w = u \<frown> [a] \<frown> v" "ren q \<inter> set u = {}" "a \<in> ren q"
          using word_first_split_set False by this
        have 3: "u \<in> words_fin q" using assms(2) unfolding 2(1) by auto
        have 4: "Ind (ren q) (set u)"
          using ample_independent reduction_ample assms(1) 1 3 2(2) by this
        have 5: "Ind (set [a]) (set u)" using 4 2(3) by simp
        have 6: "[a] @ u =\<^sub>F u @ [a]" using 5 by blast
        show ?thesis
        proof (rule deferred)
          show "a \<in> ren q" using 2(3) by this
          have "[a] \<preceq>\<^sub>F\<^sub>I [a] \<frown> u \<frown> v" by blast
          also have "[a] \<frown> u \<frown> v = ([a] @ u) \<frown> v" by simp
          also have "([a] @ u) \<frown> v =\<^sub>I (u @ [a]) \<frown> v" using 6 by blast
          also have "(u @ [a]) \<frown> v = u \<frown> [a] \<frown> v" by simp
          also have "\<dots> = w" unfolding 2(1) by rule
          finally show "[a] \<preceq>\<^sub>F\<^sub>I w" by this
        qed
      qed
    qed

    (* theorem 3.9 in [1] *)
    lemma reduction_chunk:
      assumes "q \<in> reachable" "[a] \<frown> v \<in> words_inf q"
      obtains b b\<^sub>1 b\<^sub>2 u
      where
        "b @ [a] \<in> R.words_fin q"
        "Ind {a} (set b)" "set b \<subseteq> invisible"
        "b =\<^sub>F b\<^sub>1 @ b\<^sub>2" "b\<^sub>1 \<frown> u =\<^sub>I v" "Ind (set b\<^sub>2) (range u)"
    using wellfounded assms
    proof (induct q arbitrary: thesis v rule: wfP_induct_rule)
      case (less q)
      show ?case
      proof (cases "a \<in> ren q")
        case (True)
        show ?thesis
        proof (rule less(2))
          show "[] @ [a] \<in> R.words_fin q" using True by auto
          show "Ind {a} (set [])" by auto
          show "set [] \<subseteq> invisible" by auto
          show "[] =\<^sub>F [] @ []" by auto
          show "[] \<frown> v =\<^sub>I v" by auto
          show "Ind (set []) (range v)" by auto
        qed
      next
        case (False)
        have 0: "en q \<noteq> ren q" using False less(4) by auto
        show ?thesis
        using less(3, 4)
        proof (cases rule: reduction_step)
          case (deferred c)
          have 1: "c \<in> ren q" using deferred(1) by simp
          have 2: "ind a c"
          proof (rule le_fininf_ind'')
            show "[a] \<preceq>\<^sub>F\<^sub>I [a] \<frown> v" by blast
            show "[c] \<preceq>\<^sub>F\<^sub>I [a] \<frown> v" using deferred(2) by this
            show "a \<noteq> c" using False 1 by auto
          qed
          obtain v' where 3: "[a] \<frown> v =\<^sub>I [c] \<frown> [a] \<frown> v'"
          proof -
            have 10: "[c] \<preceq>\<^sub>F\<^sub>I v"
            proof (rule le_fininf_not_member')
              show "[c] \<preceq>\<^sub>F\<^sub>I [a] \<frown> v" using deferred(2) by this
              show "c \<notin> set [a]" using False 1 by auto
            qed
            obtain v' where 11: "v =\<^sub>I [c] \<frown> v'" using 10 by blast
            have 12: "Ind (set [a]) (set [c])" using 2 by auto
            have 13: "[a] @ [c] =\<^sub>F [c] @ [a]" using 12 by blast
            have "[a] \<frown> v =\<^sub>I [a] \<frown> [c] \<frown> v'" using 11 by blast
            also have "\<dots> = ([a] @ [c]) \<frown> v'" unfolding conc_conc by rule
            also have "\<dots> =\<^sub>I ([c] @ [a]) \<frown> v'" using 13 by blast
            also have "\<dots> = [c] \<frown> [a] \<frown> v'" unfolding conc_conc by rule
            finally show ?thesis using that by auto
          qed
          have 4: "[c] \<frown> [a] \<frown> v' \<in> words_inf q" using eq_inf_word 3 less(4) by this
          show ?thesis
          proof (rule less(1))
            show "src (ex c q) q"
              using ample_wellfounded ample_subset reduction_ample less(3) 0 1 by blast
            have 100: "c \<in> en q" using less(4) deferred(2) le_fininf_word by auto
            show "ex c q \<in> reachable" using less(3) 100 by auto
            show "[a] \<frown> v' \<in> words_inf (ex c q)" using 4 by auto
          next
            fix b b\<^sub>1 b\<^sub>2 u
            assume 5: "b @ [a] \<in> R.words_fin (ex c q)"
            assume 6: "Ind {a} (set b)"
            assume 7: "set b \<subseteq> invisible"
            assume 8: "b =\<^sub>F b\<^sub>1 @ b\<^sub>2"
            assume 9: "b\<^sub>1 \<frown> u =\<^sub>I v'"
            assume 10: "Ind (set b\<^sub>2) (range u)"
            show "thesis"
            proof (rule less(2))
              show "([c] @ b) @ [a] \<in> R.words_fin q" using 1 5 by auto
              show "Ind {a} (set ([c] @ b))" using 6 2 by auto
              have 11: "c \<in> invisible"
                using ample_invisible ample_subset reduction_ample less(3) 0 1 by fast
              show "set ([c] @ b) \<subseteq> invisible" using 7 11 by auto
              have "[c] @ b =\<^sub>F [c] @ b\<^sub>1 @ b\<^sub>2" using 8 by blast
              also have "[c] @ b\<^sub>1 @ b\<^sub>2 = ([c] @ b\<^sub>1) @ b\<^sub>2" by simp
              finally show "[c] @ b =\<^sub>F ([c] @ b\<^sub>1) @ b\<^sub>2" by this
              show "([c] @ b\<^sub>1) \<frown> u =\<^sub>I v"
              proof -
                have 10: "Ind (set [a]) (set [c])" using 2 by auto
                have 11: "[a] @ [c] =\<^sub>F [c] @ [a]" using 10 by blast
                have "[a] \<frown> v =\<^sub>I [c] \<frown> [a] \<frown> v'" using 3 by this
                also have "\<dots> = ([c] @ [a]) \<frown> v'" unfolding conc_conc by rule
                also have "\<dots> =\<^sub>I ([a] @ [c]) \<frown> v'" using 11 by blast
                also have "\<dots> = [a] \<frown> [c] \<frown> v'" unfolding conc_conc by rule
                finally have 12: "[a] \<frown> v =\<^sub>I [a] \<frown> [c] \<frown> v'" by this
                have 12: "v =\<^sub>I [c] \<frown> v'" using 12 by blast
                have "([c] @ b\<^sub>1) \<frown> u = [c] \<frown> b\<^sub>1 \<frown> u" unfolding conc_conc by rule
                also have "\<dots> =\<^sub>I [c] \<frown> v'" using 9 by blast
                also have "\<dots> =\<^sub>I v" using 12 by blast
                finally show ?thesis by this
              qed
              show "Ind (set b\<^sub>2) (range u)" using 10 by this
            qed
          qed
        next
          case (omitted)
          have 1: "ren q \<subseteq> invisible" using omitted(1) by simp
          have 2: "Ind (ren q) (range ([a] \<frown> v))" using omitted(2) by simp
          obtain c where 3: "c \<in> ren q"
          proof -
            have 1: "a \<in> en q" using less(4) by auto
            show ?thesis using reduction_ample ample_nonempty less(3) 1 that by blast
          qed
          have 4: "Ind (set [c]) (range ([a] \<frown> v))" using 2 3 by auto
          have 6: "[c] \<in> words_fin q" using reduction_ample ample_subset less(3) 3 by auto
          have 7: "[a] \<frown> v \<in> words_inf (fold ex [c] q)"
            using diamond_fin_word_inf_word 4 6 less(4) by this
          show ?thesis
          proof (rule less(1))
            show "src (ex c q) q"
              using reduction_ample ample_wellfounded ample_subset less(3) 0 3 by blast
            show "ex c q \<in> reachable" using less(3) 6 by auto
            show "[a] \<frown> v \<in> words_inf (ex c q)" using 7 by auto
          next
            fix b s b\<^sub>1 b\<^sub>2 u
            assume 5: "b @ [a] \<in> R.words_fin (ex c q)"
            assume 6: "Ind {a} (set b)"
            assume 7: "set b \<subseteq> invisible"
            assume 8: "b =\<^sub>F b\<^sub>1 @ b\<^sub>2"
            assume 9: "b\<^sub>1 \<frown> u =\<^sub>I v"
            assume 10: "Ind (set b\<^sub>2) (range u)"
            show "thesis"
            proof (rule less(2))
              show "([c] @ b) @ [a] \<in> R.words_fin q" using 3 5 by auto
              show "Ind {a} (set ([c] @ b))"
              proof -
                have 1: "ind c a" using 4 by simp
                have 2: "ind a c" using independence_symmetric 1 by this
                show ?thesis using 6 2 by auto
              qed
              have 11: "c \<in> invisible" using 1 3 by auto
              show "set ([c] @ b) \<subseteq> invisible" using 7 11 by auto
              have 12: "Ind (set [c]) (set b\<^sub>1)"
              proof -
                have 1: "set b\<^sub>1 \<subseteq> range v" using 9 by force
                have 2: "Ind (set [c]) (range v)" using 4 by simp
                show ?thesis using 1 2 by auto
              qed
              have "[c] @ b =\<^sub>F [c] @ b\<^sub>1 @ b\<^sub>2" using 8 by blast
              also have "\<dots> = ([c] @ b\<^sub>1) @ b\<^sub>2" by simp
              also have "\<dots> =\<^sub>F (b\<^sub>1 @ [c]) @ b\<^sub>2" using 12 by blast
              also have "\<dots> = b\<^sub>1 @ [c] @ b\<^sub>2" by simp
              finally show "[c] @ b =\<^sub>F b\<^sub>1 @ [c] @ b\<^sub>2" by this
              show "b\<^sub>1 \<frown> u =\<^sub>I v" using 9 by this
              have 13: "Ind (set [c]) (range u)"
              proof -
                have 1: "range u \<subseteq> range v" using 9 by force
                have 2: "Ind (set [c]) (range v)" using 4 by simp
                show ?thesis using 1 2 by blast
              qed
              show "Ind (set ([c] @ b\<^sub>2)) (range u)" using 10 13 by auto
            qed
          qed
        qed
      qed
    qed

    (*
      q: initial state
      v\<^sub>1: operations of the original run processed so far
      v\<^sub>2: operations of the original run remaining
      l: operations executed in the reduced run, but not yet in the original run
      w: reduced run processed so far
      w\<^sub>1: selected reduced run processed so far
      w\<^sub>2: unselected reduced run processed so far
      u: continuation of w
    *)
    inductive reduced_run :: "'state \<Rightarrow> 'action list \<Rightarrow> 'action word \<Rightarrow> 'action list \<Rightarrow>
      'action list \<Rightarrow> 'action list \<Rightarrow> 'action list \<Rightarrow> 'action word \<Rightarrow> bool"
      where
        init: "reduced_run q [] v [] [] [] [] v" |
        absorb: "reduced_run q v\<^sub>1 ([a] \<frown> v\<^sub>2) l w w\<^sub>1 w\<^sub>2 u \<Longrightarrow> a \<in> set l \<Longrightarrow>
          reduced_run q (v\<^sub>1 @ [a]) v\<^sub>2 (remove1 a l) w w\<^sub>1 w\<^sub>2 u" |
        extend: "reduced_run q v\<^sub>1 ([a] \<frown> v\<^sub>2) l w w\<^sub>1 w\<^sub>2 u \<Longrightarrow> a \<notin> set l \<Longrightarrow>
          b @ [a] \<in> R.words_fin (fold ex w q) \<Longrightarrow>
          Ind {a} (set b) \<Longrightarrow> set b \<subseteq> invisible \<Longrightarrow>
          b =\<^sub>F b\<^sub>1 @ b\<^sub>2 \<Longrightarrow> [a] \<frown> b\<^sub>1 \<frown> u' =\<^sub>I u \<Longrightarrow> Ind (set b\<^sub>2) (range u') \<Longrightarrow>
          reduced_run q (v\<^sub>1 @ [a]) v\<^sub>2 (l @ b\<^sub>1) (w @ b @ [a]) (w\<^sub>1 @ b\<^sub>1 @ [a]) (w\<^sub>2 @ b\<^sub>2) u'"

    lemma reduced_run_words_fin:
      assumes "reduced_run q v\<^sub>1 v\<^sub>2 l w w\<^sub>1 w\<^sub>2 u"
      shows "w \<in> R.words_fin q"
      using assms by (induct, auto)

    lemma reduced_run_invar_2:
      assumes "reduced_run q v\<^sub>1 v\<^sub>2 l w w\<^sub>1 w\<^sub>2 u"
      shows "v\<^sub>2 =\<^sub>I l \<frown> u"
    using assms
    proof induct
      case (init q v)
      show ?case by simp
    next
      case (absorb q v\<^sub>1 a v\<^sub>2 l w w\<^sub>1 w\<^sub>2 u)
      obtain l\<^sub>1 l\<^sub>2 where 10: "l = l\<^sub>1 @ [a] @ l\<^sub>2" "a \<notin> set l\<^sub>1"
        using split_list_first[OF absorb(3)] by auto
      have 11: "Ind {a} (set l\<^sub>1)"
      proof (rule le_fininf_ind')
        show "[a] \<preceq>\<^sub>F\<^sub>I l \<frown> u" using absorb(2) by auto
        show "l\<^sub>1 \<preceq>\<^sub>F\<^sub>I l \<frown> u" unfolding 10(1) by auto
        show "a \<notin> set l\<^sub>1" using 10(2) by this
      qed
      have 12: "Ind (set [a]) (set l\<^sub>1)" using 11 by auto
      have "[a] @ remove1 a l = [a] @ l\<^sub>1 @ l\<^sub>2" unfolding 10(1) remove1_append using 10(2) by auto
      also have "\<dots> =\<^sub>F ([a] @ l\<^sub>1) @ l\<^sub>2" by simp
      also have "\<dots> =\<^sub>F (l\<^sub>1 @ [a]) @ l\<^sub>2" using 12 by blast
      also have "\<dots> = l" unfolding 10(1) by simp
      finally have 13: "[a] @ remove1 a l =\<^sub>F l" by this
      have "[a] \<frown> remove1 a l \<frown> u = ([a] @ remove1 a l) \<frown> u" unfolding conc_conc by simp
      also have "\<dots> =\<^sub>I l \<frown> u" using 13 by blast
      also have "\<dots> =\<^sub>I [a] \<frown> v\<^sub>2" using absorb(2) by auto
      finally show ?case by blast
    next
      case (extend q v\<^sub>1 a v\<^sub>2 l w w\<^sub>1 w\<^sub>2 u b b\<^sub>1 b\<^sub>2 u')
      have 11: "Ind {a} (set l)"
      proof (rule le_fininf_ind')
        show "[a] \<preceq>\<^sub>F\<^sub>I l \<frown> u" using extend(2) by auto
        show "l \<preceq>\<^sub>F\<^sub>I l \<frown> u" by auto
        show "a \<notin> set l" using extend(3) by this
      qed
      have 11: "Ind (set [a]) (set l)" using 11 by auto
      have 12: "eq_fin ([a] @ l) (l @ [a])" using 11 by blast
      have 131: "set b\<^sub>1 \<subseteq> set b" using extend(7) by auto
      have 132: "Ind (set [a]) (set b)" using extend(5) by auto
      have 13: "Ind (set [a]) (set b\<^sub>1)" using 131 132 by auto
      have 14: "eq_fin ([a] @ b\<^sub>1) (b\<^sub>1 @ [a])" using 13 by blast
      have "[a] \<frown> ((l @ b\<^sub>1) \<frown> u') = ([a] @ l) \<frown> b\<^sub>1 \<frown> u'" by simp
      also have "eq_inf \<dots> ((l @ [a]) \<frown> b\<^sub>1 \<frown> u')" using 12 by blast
      also have "\<dots> = l \<frown> [a] \<frown> b\<^sub>1 \<frown> u'" by simp
      also have "eq_inf \<dots> (l \<frown> u)" using extend(8) by blast
      also have "eq_inf \<dots> ([a] \<frown> v\<^sub>2)" using extend(2) by auto
      finally show ?case by blast
    qed

    lemma reduced_run_invar_1:
      assumes "reduced_run q v\<^sub>1 v\<^sub>2 l w w\<^sub>1 w\<^sub>2 u"
      shows "v\<^sub>1 @ l =\<^sub>F w\<^sub>1"
    using assms
    proof induct
      case (init q v)
      show ?case by simp
    next
      case (absorb q v\<^sub>1 a v\<^sub>2 l w w\<^sub>1 w\<^sub>2 u)
      have 1: "[a] \<frown> v\<^sub>2 =\<^sub>I l \<frown> u" using reduced_run_invar_2 absorb(1) by this
      obtain l\<^sub>1 l\<^sub>2 where 10: "l = l\<^sub>1 @ [a] @ l\<^sub>2" "a \<notin> set l\<^sub>1"
        using split_list_first[OF absorb(3)] by auto
      have 11: "Ind {a} (set l\<^sub>1)"
      proof (rule le_fininf_ind')
        show "[a] \<preceq>\<^sub>F\<^sub>I l \<frown> u" using 1 by auto
        show "l\<^sub>1 \<preceq>\<^sub>F\<^sub>I l \<frown> u" unfolding 10(1) by auto
        show "a \<notin> set l\<^sub>1" using 10(2) by this
      qed
      have 12: "Ind (set [a]) (set l\<^sub>1)" using 11 by auto
      have "[a] @ remove1 a l = [a] @ l\<^sub>1 @ l\<^sub>2" unfolding 10(1) remove1_append using 10(2) by auto
      also have "\<dots> =\<^sub>F ([a] @ l\<^sub>1) @ l\<^sub>2" by simp
      also have "\<dots> =\<^sub>F (l\<^sub>1 @ [a]) @ l\<^sub>2" using 12 by blast
      also have "\<dots> = l" unfolding 10(1) by simp
      finally have 13: "[a] @ remove1 a l =\<^sub>F l" by this
      have "w\<^sub>1 =\<^sub>F v\<^sub>1 @ l" using absorb(2) by auto
      also have "\<dots> =\<^sub>F v\<^sub>1 @ ([a] @ remove1 a l)" using 13 by blast
      also have "\<dots> = (v\<^sub>1 @ [a]) @ remove1 a l" by simp
      finally show ?case by auto
    next
      case (extend q v\<^sub>1 a v\<^sub>2 l w w\<^sub>1 w\<^sub>2 u b b\<^sub>1 b\<^sub>2 u')
      have 1: "[a] \<frown> v\<^sub>2 =\<^sub>I l \<frown> u" using reduced_run_invar_2 extend(1) by this
      have 11: "Ind {a} (set l)"
      proof (rule le_fininf_ind')
        show "[a] \<preceq>\<^sub>F\<^sub>I l \<frown> u" using 1 by auto
        show "l \<preceq>\<^sub>F\<^sub>I l \<frown> u" by auto
        show "a \<notin> set l" using extend(3) by auto
      qed
      have 11: "Ind (set [a]) (set l)" using 11 by auto
      have 12: "eq_fin ([a] @ l) (l @ [a])" using 11 by blast
      have 131: "set b\<^sub>1 \<subseteq> set b" using extend(7) by auto
      have 132: "Ind (set [a]) (set b)" using extend(5) by auto
      have 13: "Ind (set [a]) (set b\<^sub>1)" using 131 132 by auto
      have 14: "eq_fin ([a] @ b\<^sub>1) (b\<^sub>1 @ [a])" using 13 by blast
      have "eq_fin (w\<^sub>1 @ b\<^sub>1 @ [a]) (w\<^sub>1 @ [a] @ b\<^sub>1)" using 14 by blast
      also have "eq_fin \<dots> ((v\<^sub>1 @ l) @ [a] @ b\<^sub>1)" using extend(2) by blast
      also have "eq_fin \<dots> (v\<^sub>1 @ (l @ [a]) @ b\<^sub>1)" by simp
      also have "eq_fin \<dots> (v\<^sub>1 @ ([a] @ l) @ b\<^sub>1)" using 12 by blast
      also have "\<dots> = (v\<^sub>1 @ [a]) @ l @ b\<^sub>1" by simp
      finally show ?case by auto
    qed

    lemma reduced_run_invisible:
      assumes "reduced_run q v\<^sub>1 v\<^sub>2 l w w\<^sub>1 w\<^sub>2 u"
      shows "set w\<^sub>2 \<subseteq> invisible"
    using assms
    proof induct
      case (init q v)
      show ?case by simp
    next
      case (absorb q v\<^sub>1 a v\<^sub>2 l w w\<^sub>1 w\<^sub>2 u)
      show ?case using absorb(2) by this
    next
      case (extend q v\<^sub>1 a v\<^sub>2 l w w\<^sub>1 w\<^sub>2 u b b\<^sub>1 b\<^sub>2 u')
      have 1: "set b\<^sub>2 \<subseteq> set b" using extend(7) by auto
      show ?case unfolding set_append using extend(2) extend(6) 1 by blast
    qed

    lemma reduced_run_ind:
      assumes "reduced_run q v\<^sub>1 v\<^sub>2 l w w\<^sub>1 w\<^sub>2 u"
      shows "Ind (set w\<^sub>2) (range u)"
    using assms
    proof induct
      case (init q v)
      show ?case by simp
    next
      case (absorb q v\<^sub>1 a v\<^sub>2 l w w\<^sub>1 w\<^sub>2 u)
      show ?case using absorb(2) by this
    next
      case (extend q v\<^sub>1 a v\<^sub>2 l w w\<^sub>1 w\<^sub>2 u b b\<^sub>1 b\<^sub>2 u')
      have 1: "range u' \<subseteq> range u" using extend(8) by force
      show ?case using extend(2) extend(9) 1 unfolding set_append by blast
    qed

    lemma reduced_run_decompose:
      assumes "reduced_run q v\<^sub>1 v\<^sub>2 l w w\<^sub>1 w\<^sub>2 u"
      shows "w =\<^sub>F w\<^sub>1 @ w\<^sub>2"
    using assms
    proof induct
      case (init q v)
      show ?case by simp
    next
      case (absorb q v\<^sub>1 a v\<^sub>2 l w w\<^sub>1 w\<^sub>2 u)
      show ?case using absorb(2) by this
    next
      case (extend q v\<^sub>1 a v\<^sub>2 l w w\<^sub>1 w\<^sub>2 u b b\<^sub>1 b\<^sub>2 u')
      have 1: "Ind (set [a]) (set b\<^sub>2)" using extend(5) extend(7) by auto
      have 2: "Ind (set w\<^sub>2) (set (b\<^sub>1 @ [a]))"
      proof -
        have 1: "Ind (set w\<^sub>2) (range u)" using reduced_run_ind extend(1) by this
        have 2: "u =\<^sub>I [a] \<frown> b\<^sub>1 \<frown> u'" using extend(8) by auto
        have 3: "range u = range ([a] \<frown> b\<^sub>1 \<frown> u')" using 2 by blast
        show ?thesis unfolding set_append using 1 3 by simp
      qed
      have "w @ b @ [a] =\<^sub>F (w\<^sub>1 @ w\<^sub>2) @ b @ [a]" using extend(2) by blast
      also have "\<dots> =\<^sub>F (w\<^sub>1 @ w\<^sub>2) @ (b\<^sub>1 @ b\<^sub>2) @ [a]" using extend(7) by blast
      also have "\<dots> = w\<^sub>1 @ w\<^sub>2 @ b\<^sub>1 @ (b\<^sub>2 @ [a])" by simp
      also have "\<dots> =\<^sub>F w\<^sub>1 @ w\<^sub>2 @ b\<^sub>1 @ ([a] @ b\<^sub>2)" using 1 by blast
      also have "\<dots> =\<^sub>F w\<^sub>1 @ (w\<^sub>2 @ (b\<^sub>1 @ [a])) @ b\<^sub>2" by simp
      also have "\<dots> =\<^sub>F w\<^sub>1 @ ((b\<^sub>1 @ [a]) @ w\<^sub>2) @ b\<^sub>2" using 2 by blast
      also have "\<dots> =\<^sub>F (w\<^sub>1 @ b\<^sub>1 @ [a]) @ w\<^sub>2 @ b\<^sub>2" by simp
      finally show ?case by this
    qed

    lemma reduced_run_project:
      assumes "reduced_run q v\<^sub>1 v\<^sub>2 l w w\<^sub>1 w\<^sub>2 u"
      shows "project visible w\<^sub>1 = project visible w"
    proof -
      have 1: "w\<^sub>1 @ w\<^sub>2 =\<^sub>F w" using reduced_run_decompose assms by auto
      have 2: "set w\<^sub>2 \<subseteq> invisible" using reduced_run_invisible assms by this
      have 3: "project visible w\<^sub>2 = []" unfolding filter_empty_conv using 2 by auto
      have "project visible w\<^sub>1 = project visible w\<^sub>1 @ project visible w\<^sub>2" using 3 by simp
      also have "\<dots> = project visible (w\<^sub>1 @ w\<^sub>2)" by simp
      also have "\<dots> = list_of (lproject visible (llist_of (w\<^sub>1 @ w\<^sub>2)))" by simp
      also have "\<dots> = list_of (lproject visible (llist_of w))"
        using eq_fin_lproject_visible 1 by metis
      also have "\<dots> = project visible w" by simp
      finally show ?thesis by this
    qed

    lemma reduced_run_length_1:
      assumes "reduced_run q v\<^sub>1 v\<^sub>2 l w w\<^sub>1 w\<^sub>2 u"
      shows "length v\<^sub>1 \<le> length w\<^sub>1"
      using reduced_run_invar_1 assms by force
    lemma reduced_run_length:
      assumes "reduced_run q v\<^sub>1 v\<^sub>2 l w w\<^sub>1 w\<^sub>2 u"
      shows "length v\<^sub>1 \<le> length w"
    proof -
      have "length v\<^sub>1 \<le> length w\<^sub>1" using reduced_run_length_1 assms by this
      also have "\<dots> \<le> length w" using reduced_run_decompose assms by force
      finally show ?thesis by this
    qed

    lemma reduced_run_step:
      assumes "q \<in> reachable" "v\<^sub>1 \<frown> [a] \<frown> v\<^sub>2 \<in> words_inf q"
      assumes "reduced_run q v\<^sub>1 ([a] \<frown> v\<^sub>2) l w w\<^sub>1 w\<^sub>2 u"
      obtains l' w' w\<^sub>1' w\<^sub>2' u'
      where "reduced_run q (v\<^sub>1 @ [a]) v\<^sub>2 l' (w @ w') (w\<^sub>1 @ w\<^sub>1') (w\<^sub>2 @ w\<^sub>2') u'"
    proof (cases "a \<in> set l")
      case True
      show ?thesis
      proof
        show "reduced_run q (v\<^sub>1 @ [a]) v\<^sub>2 (remove1 a l) (w @ []) (w\<^sub>1 @ []) (w\<^sub>2 @ []) u"
        proof
          show "reduced_run q v\<^sub>1 ([a] \<frown> v\<^sub>2) l (w @ []) (w\<^sub>1 @ []) (w\<^sub>2 @ []) u" using assms(3) by simp
          show "a \<in> set l" using True by this
        qed
      qed
    next
      case False
      have 1: "v\<^sub>1 @ l =\<^sub>F w\<^sub>1" using reduced_run_invar_1 assms(3) by this
      have 2: "[a] \<frown> v\<^sub>2 =\<^sub>I l \<frown> u" using reduced_run_invar_2 assms(3) by this
      have 3: "w =\<^sub>F w\<^sub>1 @ w\<^sub>2" using reduced_run_decompose assms(3) by this
      have "v\<^sub>1 @ l @ w\<^sub>2 = (v\<^sub>1 @ l) @ w\<^sub>2" by simp
      also have "\<dots> =\<^sub>F w\<^sub>1 @ w\<^sub>2" using 1 by blast
      also have "\<dots> =\<^sub>F w" using 3 by blast
      finally have 4: "v\<^sub>1 @ l @ w\<^sub>2 =\<^sub>F w" by this
      have 5: "(v\<^sub>1 @ l) \<frown> w\<^sub>2 \<frown> u \<in> words_inf q"
      proof (rule diamond_fin_word_inf_word')
        show "Ind (set w\<^sub>2) (range u)" using reduced_run_ind assms(3) by this
        have 6: "w \<in> R.words_fin q" using reduced_run_words_fin assms(3) by this
        have 7: "w \<in> words_fin q" using reduction_words_fin assms(1) 6 by auto
        show "(v\<^sub>1 @ l) @ w\<^sub>2 \<in> words_fin q" using eq_fin_word 4 7 by auto
        have 8: "v\<^sub>1 \<frown> [a] \<frown> v\<^sub>2 =\<^sub>I v\<^sub>1 \<frown> l \<frown> u" using 2 by blast
        show "(v\<^sub>1 @ l) \<frown> u \<in> words_inf q" using eq_inf_word assms(2) 8 by auto
      qed
      have 6: "w \<frown> u \<in> words_inf q" using eq_inf_word 4 5 by auto
      have 7: "[a] \<preceq>\<^sub>F\<^sub>I l \<frown> u" using 2 by blast
      have 8: "[a] \<preceq>\<^sub>F\<^sub>I u" using le_fininf_not_member' 7 False by this
      obtain u' where 9: "u =\<^sub>I [a] \<frown> u'" using 8 by rule
      have 101: "fold ex w q \<in> reachable" using assms(1) reduction_words_fin assms(1) 6 by auto
      have 10: "[a] \<frown> u' \<in> words_inf (fold ex w q)" using eq_inf_word 9 6 by blast
      obtain b b\<^sub>1 b\<^sub>2 u'' where 11:
        "b @ [a] \<in> R.words_fin (fold ex w q)"
        "Ind {a} (set b)" "set b \<subseteq> invisible"
        "b =\<^sub>F b\<^sub>1 @ b\<^sub>2" "b\<^sub>1 \<frown> u'' =\<^sub>I u'" "Ind (set b\<^sub>2) (range u'')"
        using reduction_chunk 101 10 by this
      show ?thesis
      proof
        show "reduced_run q (v\<^sub>1 @ [a]) v\<^sub>2 (l @ b\<^sub>1) (w @ b @ [a]) (w\<^sub>1 @ b\<^sub>1 @ [a]) (w\<^sub>2 @ b\<^sub>2) u''"
        proof
          show "reduced_run q v\<^sub>1 ([a] \<frown> v\<^sub>2) l w w\<^sub>1 w\<^sub>2 u" using assms(3) by this
          show "a \<notin> set l" using False by this
          show "b @ [a] \<in> R.words_fin (fold ex w q)" using 11(1) by this
          show "Ind {a} (set b)" using 11(2) by this
          show "set b \<subseteq> invisible" using 11(3) by this
          show "b =\<^sub>F b\<^sub>1 @ b\<^sub>2" using 11(4) by this
          show "[a] \<frown> b\<^sub>1 \<frown> u'' =\<^sub>I u" using 9 11(5) by blast
          show "Ind (set b\<^sub>2) (range u'')" using 11(6) by this
        qed
      qed
    qed

    lemma reduced_run_chain:
      assumes "q \<in> reachable" "v \<in> words_inf q"
      obtains w w\<^sub>1
      where "\<And> k. \<exists> l w\<^sub>2 u. reduced_run q (prefix k v) (suffix k v) l (w k) (w\<^sub>1 k) w\<^sub>2 u"
        "chain w" "chain w\<^sub>1"
    proof -
      def P \<equiv> "\<lambda> k x. \<exists> l w\<^sub>2 u. reduced_run q (prefix k v) (suffix k v) l (fst x) (snd x) w\<^sub>2 u"
      obtain Q where 1: "chain (fst \<circ> Q)" "chain (snd \<circ> Q)" "\<And> k. P k (Q k)"
      proof (rule chain_construct_2[of P])
        show "P 0 ([], [])" unfolding P_def using init by force
      next
        fix k x
        assume 1: "P k x"
        obtain l w w\<^sub>1 w\<^sub>2 u where 2: "reduced_run q (prefix k v) (suffix k v) l w w\<^sub>1 w\<^sub>2 u"
          "x = (w, w\<^sub>1)" using 1 unfolding P_def by auto
        obtain l' w' w\<^sub>1' w\<^sub>2' u' where 3:
          "reduced_run q (prefix (Suc k) v) (suffix (Suc k) v) l' (w @ w') (w\<^sub>1 @ w\<^sub>1') (w\<^sub>2 @ w\<^sub>2') u'"
        proof (rule reduced_run_step)
          show "q \<in> reachable" using assms(1) by this
          show "prefix k v \<frown> [v k] \<frown> suffix (Suc k) v \<in> words_inf q" using assms(2) by simp
          show "reduced_run q (prefix k v) ([v k] \<frown> suffix (Suc k) v) l w w\<^sub>1 w\<^sub>2 u"
            using 2(1) by simp
        qed auto
        show "\<exists> x'. P (Suc k) x' \<and> fst x \<le> fst x' \<and> snd x \<le> snd x'"
        proof (unfold P_def, intro exI conjI)
          show "reduced_run q (prefix (Suc k) v) (suffix (Suc k) v)
            l' (fst (w @ w', w\<^sub>1 @ w\<^sub>1')) (snd (w @ w', w\<^sub>1 @ w\<^sub>1')) (w\<^sub>2 @ w\<^sub>2') u'" using 3 by simp
          show "fst x \<le> fst (w @ w', w\<^sub>1 @ w\<^sub>1')" unfolding 2(2) by simp
          show "snd x \<le> snd (w @ w', w\<^sub>1 @ w\<^sub>1')" unfolding 2(2) by simp
        qed
        show "k \<le> length (fst x)" using reduced_run_length 2 by force
        show "k \<le> length (snd x)" using reduced_run_length_1 2 by force
      qed rule
      show ?thesis
      proof
        fix k
        show "\<exists> l w\<^sub>2 u. reduced_run q (prefix k v) (suffix k v) l ((fst \<circ> Q) k) ((snd \<circ> Q) k) w\<^sub>2 u"
          using 1(3) unfolding P_def by simp
      next
        show "chain (fst \<circ> Q)" using 1(1) by this
        show "chain (snd \<circ> Q)" using 1(2) by this
      qed
    qed

    (* theorem 3.11 in [1] *)
    lemma reduction_word:
      assumes "q \<in> reachable" "v \<in> words_inf q"
      obtains u w
      where
        "w \<in> R.words_inf q"
        "v =\<^sub>I u" "u \<preceq>\<^sub>I w"
        "lproject visible (inf_llist u) = lproject visible (inf_llist w)"
    proof -
      obtain w w\<^sub>1 where 1:
        "\<And> k. \<exists> l w\<^sub>2 u. reduced_run q (prefix k v) (suffix k v) l (w k) (w\<^sub>1 k) w\<^sub>2 u"
        "chain w" "chain w\<^sub>1"
        using reduced_run_chain[OF assms] by blast
      show ?thesis
      proof
        show "limit w \<in> R.words_inf q"
        proof
          show "chain w" using 1(2) by this
        next
          fix k
          obtain l w\<^sub>2 u where 2: "reduced_run q (prefix k v) (suffix k v) l (w k) (w\<^sub>1 k) w\<^sub>2 u"
            using 1(1) by blast
          show "w k \<in> R.words_fin q" using reduced_run_words_fin 2 by this
        qed
        show "v =\<^sub>I limit w\<^sub>1"
        proof
          show "v \<preceq>\<^sub>I limit w\<^sub>1"
          proof (rule le_infI_chain_right')
            show "chain w\<^sub>1" using 1(3) by this
          next
            fix k
            obtain l w\<^sub>2 u where 2: "reduced_run q (prefix k v) (suffix k v) l (w k) (w\<^sub>1 k) w\<^sub>2 u"
              using 1(1) by blast
            have 3: "prefix k v @ l =\<^sub>F w\<^sub>1 k" using reduced_run_invar_1 2 by this
            show "prefix k v \<preceq>\<^sub>F w\<^sub>1 k" using 3 by blast
          qed
          show "limit w\<^sub>1 \<preceq>\<^sub>I v"
          proof (rule le_infI_chain_left)
            show "chain w\<^sub>1" using 1(3) by this
          next
            fix k
            obtain l w\<^sub>2 u where 2: "reduced_run q (prefix k v) (suffix k v) l (w k) (w\<^sub>1 k) w\<^sub>2 u"
              using 1(1) by blast
            have "w\<^sub>1 k =\<^sub>F prefix k v @ l" using reduced_run_invar_1 2 by blast
            also have "\<dots> \<le>\<^sub>F\<^sub>I prefix k v \<frown> l \<frown> u" by auto
            also have "\<dots> =\<^sub>I prefix k v \<frown> suffix k v" using reduced_run_invar_2 2 by fastforce
            also have "\<dots> = v" by simp
            finally show "w\<^sub>1 k \<preceq>\<^sub>F\<^sub>I v" by this
          qed
        qed
        show "limit w\<^sub>1 \<preceq>\<^sub>I limit w"
        proof (rule le_infI_chain_left)
          show "chain w\<^sub>1" using 1(3) by this
        next
          fix k
          obtain l w\<^sub>2 u where 2: "reduced_run q (prefix k v) (suffix k v) l (w k) (w\<^sub>1 k) w\<^sub>2 u"
            using 1(1) by blast
          have 3: "w k =\<^sub>F w\<^sub>1 k @ w\<^sub>2" using reduced_run_decompose 2 by this
          have "w\<^sub>1 k \<preceq>\<^sub>F w k" using 3 by blast
          also have "\<dots> \<le>\<^sub>F\<^sub>I limit w" using chain_prefix_limit 1(2) by this
          finally show "w\<^sub>1 k \<preceq>\<^sub>F\<^sub>I limit w" by this
        qed
        show "lproject visible (inf_llist (limit w\<^sub>1)) = lproject visible (inf_llist (limit w))"
        proof (rule antisym)
          show "lproject visible (inf_llist (limit w\<^sub>1)) \<le> lproject visible (inf_llist (limit w))"
          proof (rule lproject_prefix_limit_chain)
            show "chain w\<^sub>1" using 1(3) by this
          next
            fix k
            obtain l w\<^sub>2 u where 2: "reduced_run q (prefix k v) (suffix k v) l (w k) (w\<^sub>1 k) w\<^sub>2 u"
              using 1(1) by blast
            have 3: "w k \<le>\<^sub>F\<^sub>I limit w" using chain_prefix_limit 1(2) by this
            have "lproject visible (llist_of (w\<^sub>1 k)) = llist_of (project visible (w\<^sub>1 k))" by simp
            also have "\<dots> = llist_of (project visible (w k))" using reduced_run_project 2 by simp
            also have "\<dots> = lproject visible (llist_of (w k))" by simp
            also have "\<dots> \<le> lproject visible (inf_llist (limit w))" using 3 by blast
            finally show "lproject visible (llist_of (w\<^sub>1 k)) \<le>
              lproject visible (inf_llist (limit w))" by this
          qed
          show "lproject visible (inf_llist (limit w)) \<le> lproject visible (inf_llist (limit w\<^sub>1))"
          proof (rule lproject_prefix_limit_chain)
            show "chain w" using 1(2) by this
          next
            fix k
            obtain l w\<^sub>2 u where 2: "reduced_run q (prefix k v) (suffix k v) l (w k) (w\<^sub>1 k) w\<^sub>2 u"
              using 1(1) by blast
            have 3: "w\<^sub>1 k \<le>\<^sub>F\<^sub>I limit w\<^sub>1" using chain_prefix_limit 1(3) by this
            have "lproject visible (llist_of (w k)) = llist_of (project visible (w k))" by simp
            also have "\<dots> = llist_of (project visible (w\<^sub>1 k))" using reduced_run_project 2 by simp
            also have "\<dots> = lproject visible (llist_of (w\<^sub>1 k))" by simp
            also have "\<dots> \<le> lproject visible (inf_llist (limit w\<^sub>1))" using 3 by blast
            finally show "lproject visible (llist_of (w k)) \<le>
              lproject visible (inf_llist (limit w\<^sub>1))" by this
          qed
        qed
      qed
    qed

    (* theorem 3.12 in [1] *)
    lemma reduction_equivalent:
      assumes "q \<in> reachable" "u \<in> S.words_inf q"
      obtains v
      where "v \<in> R.words_inf q" "int \<circ> scan ex u q \<approx> int \<circ> scan ex v q"
    proof -
      obtain v w where 1: "w \<in> R.words_inf q" "u =\<^sub>I v" "v \<preceq>\<^sub>I w"
        "lproject visible (inf_llist v) = lproject visible (inf_llist w)"
        using reduction_word assms by this
      show ?thesis
      proof
        show "w \<in> R.words_inf q" using 1(1) by this
        show "int \<circ> scan ex u q \<approx> int \<circ> scan ex w q"
        proof (rule execute_inf_visible)
          show "u \<in> words_inf q" using assms(2) by this
          show "w \<in> words_inf q" using reduction_words_inf assms(1) 1(1) by auto
          have "u =\<^sub>I v" using 1(2) by this
          also have "\<dots> \<preceq>\<^sub>I w" using 1(3) by this
          finally show "u \<preceq>\<^sub>I w" by this
          show "w \<preceq>\<^sub>I w" by simp
          have "lproject visible (inf_llist u) = lproject visible (inf_llist v)"
            using eq_inf_lproject_visible 1(2) by this
          also have "\<dots> = lproject visible (inf_llist w)" using 1(4) by this
          finally show "lproject visible (inf_llist u) = lproject visible (inf_llist w)" by this
        qed
      qed
    qed

    lemma reduction_language_subset: "R.language \<subseteq> S.language"
      unfolding S.language_def R.language_def using reduction_words_inf by auto

    lemma reduction_language_stuttering:
      assumes "u \<in> S.language"
      obtains v
      where "v \<in> R.language" "u \<approx> v"
    proof -
      obtain q v where 1: "u = int \<circ> scan ex v q" "q \<in> init" "v \<in> S.words_inf q" using assms by rule
      obtain v' where 2: "v' \<in> R.words_inf q" "int \<circ> scan ex v q \<approx> int \<circ> scan ex v' q"
        using reduction_equivalent 1(2, 3) by blast
      show ?thesis
      proof (intro that R.languageI)
        show "int \<circ> scan ex v' q = int \<circ> scan ex v' q" by rule
        show "q \<in> init" using 1(2) by this
        show "v' \<in> R.words_inf q" using 2(1) by this
        show "u \<approx> int \<circ> scan ex v' q" unfolding 1(1) using 2(2) by this
      qed
    qed

  end

end
