structure RegularExpression :
sig

	val parseAll : string -> char Regular_Exp.rexp list
	val parse : string -> char Regular_Exp.rexp option
	val show : char Regular_Exp.rexp -> string

	val generateRandom : int -> char Regular_Exp.rexp

	val liftToString : char Regular_Exp.rexp -> string Regular_Exp.rexp

end = struct

	fun filter f [] = []
	  | filter f (x :: xs) = if f x then x :: filter f xs else filter f xs
	fun concat xss = foldr (op @) [] xss
	fun bind xs f = concat (map f xs)
	fun return x = [x]

	fun parseSymbol [] = []
	  | parseSymbol (c :: cs) = if Char.isAlpha c then [(c, cs)] else []
	fun parseConstant k [] = []
	  | parseConstant k (c :: cs) = if c = k then [(c, cs)] else []

	(*
		grammar
			associates plus and times to the right
			precedence is STAR > TIMES > PLUS

		PLUS -> <TIMES> | <TIMES>+<PLUS>
		TIMES -> <STAR> | <STAR><TIMES>
		STAR = <ATOM> | <ATOM>*
		ATOM = 0 | 1 | <SYMBOL> | (<PLUS>)
		SYMBOL = [a-zA-Z]
	*)

	fun parseAtom input1 =
		bind (parseConstant #"0" input1) (fn (_, input2) =>
		return (Regular_Exp.Zero, input2))
		@
		bind (parseConstant #"1" input1) (fn (_, input2) =>
		return (Regular_Exp.One, input2))
		@
		bind (parseSymbol input1) (fn (symbol, input2) =>
		return (Regular_Exp.Atom symbol, input2))
		@
		bind (parseConstant #"(" input1) (fn (_, input2) =>
		bind (parsePlus input2) (fn (plusExp, input3) => 
		bind (parseConstant #")" input3) (fn (_, input4) =>
		return (plusExp, input4))))
	and parseStar input1 =
		bind (parseAtom input1) (fn (atomExp, input2) =>
		return (atomExp, input2))
		@
		bind (parseAtom input1) (fn (atomExp, input2) => 
		bind (parseConstant #"*" input2) (fn (_, input3) =>
		return (Regular_Exp.Star atomExp, input3)))
	and parseTimes input1 =
		bind (parseStar input1) (fn (starExp, input2) =>
		return (starExp, input2))
		@
		bind (parseStar input1) (fn (starExp, input2) => 
		bind (parseTimes input2) (fn (timesExp, input3) =>
		return (Regular_Exp.Times (starExp, timesExp), input3)))
	and parsePlus input1 =
		bind (parseTimes input1) (fn (timesExp, input2) =>
		return (timesExp, input2))
		@
		bind (parseTimes input1) (fn (timesExp, input2) => 
		bind (parseConstant #"+" input2) (fn (_, input3) =>
		bind (parsePlus input3) (fn (plusExp, input4) =>
		return (Regular_Exp.Plus (timesExp, plusExp), input4))))

	fun parseAll input = map (fn (exp, _) => exp) (filter (fn (exp, rest) => null rest) (parsePlus (explode input)))
	fun parse input =
		let val result = parseAll input
		in
			if length result = 1
			then SOME (hd result)
			else NONE
		end
	fun show Regular_Exp.Zero = "0"
	  | show Regular_Exp.One = "1"
	  | show (Regular_Exp.Atom a) = str a
	  | show (Regular_Exp.Plus (r1, r2)) = "(" ^ show r1 ^ "+" ^ show r2 ^ ")"
	  | show (Regular_Exp.Times (r1, r2)) = "(" ^ show r1 ^ show r2 ^ ")"
	  | show (Regular_Exp.Star r) = "(" ^ show r ^ ")" ^ "*"

	fun generateRandom 0 = if random_range 0 4 = 0 then Regular_Exp.One else Regular_Exp.Atom (Char.chr (random_range 97 106))
	  | generateRandom size =
			case random_range 0 2 of
				0 => Regular_Exp.Plus (generateRandom ((size - 1) div 2), generateRandom ((size - 1) div 2)) |
				1 => Regular_Exp.Times (generateRandom ((size - 1) div 2), generateRandom ((size - 1) div 2)) |
				2 => Regular_Exp.Star (generateRandom (size - 1))

	fun liftToString Regular_Exp.Zero = Regular_Exp.Zero
	  | liftToString Regular_Exp.One = Regular_Exp.One
	  | liftToString (Regular_Exp.Atom a) = Regular_Exp.Atom (str a)
	  | liftToString (Regular_Exp.Plus (r1, r2)) = Regular_Exp.Plus (liftToString r1, liftToString r2)
	  | liftToString (Regular_Exp.Times (r1, r2)) = Regular_Exp.Times (liftToString r1, liftToString r2)
	  | liftToString (Regular_Exp.Star r) = Regular_Exp.Star (liftToString r)

end

functor Thompson (S : Sigma_param) :
sig

	type NFA

	val tmTranslate : S.Sigma Regular_Exp.rexp -> NFA
	val tmTranslateNormalize : S.Sigma Regular_Exp.rexp -> NFA
	val taTranslate : S.Sigma Regular_Exp.rexp -> NFA
	val taTranslateNormalize : S.Sigma Regular_Exp.rexp -> NFA

end = struct

	structure NFA = NFA (S)
	type NFA = NFA.NFA

	val sigma_linord = #2 (S.Sigma_dsarg)

	fun tmTranslate r = TM_Implementation.rs_TM_translate_NFA_wrapper_impl sigma_linord r
	fun tmTranslateNormalize r = TM_Implementation.rs_TM_translate_NFA_wrapper_normalize_impl sigma_linord r
	fun taTranslate r = TA_Implementation.rs_TA_translate_NFA_wrapper_impl sigma_linord r
	fun taTranslateNormalize r = TA_Implementation.rs_TA_translate_NFA_wrapper_normalize_impl sigma_linord r

(*
			val stateCount = NFA.nfa_states_no result;
			val transitionCount = NFA.nfa_trans_no result;
			val _ = print ("states: " ^ Int.toString stateCount ^ ", transitions: " ^ Int.toString transitionCount ^ ", ratio: " ^ Real.toString (Real.fromInt transitionCount / Real.fromInt stateCount) ^ "\n");
*)

end

structure ThompsonString = Thompson (string_Sigma_param)
structure ThompsonInteger = Thompson (int_Sigma_param)
