
structure STArray = struct

datatype 'a Cell = Invalid | Value of 'a array;

exception AccessedOldVersion;

type 'a array = 'a Cell Unsynchronized.ref;

fun fromList l = Unsynchronized.ref (Value (Array.fromList l));
fun array (size, v) = Unsynchronized.ref (Value (Array.array (size,v)));
fun tabulate (size, f) = Unsynchronized.ref (Value (Array.tabulate(size, f)));
fun sub (Unsynchronized.ref Invalid, idx) = raise AccessedOldVersion |
    sub (Unsynchronized.ref (Value a), idx) = Array.sub (a,idx);
fun update (aref,idx,v) =
  case aref of
    (Unsynchronized.ref Invalid) => raise AccessedOldVersion |
    (Unsynchronized.ref (Value a)) => (
      aref := Invalid;
      Array.update (a,idx,v);
      Unsynchronized.ref (Value a)
    );

fun length (Unsynchronized.ref Invalid) = raise AccessedOldVersion |
    length (Unsynchronized.ref (Value a)) = Array.length a

fun grow (aref, i, x) = case aref of 
  (Unsynchronized.ref Invalid) => raise AccessedOldVersion |
  (Unsynchronized.ref (Value a)) => (
    let val len=Array.length a;
        val na = Array.array (len+i,x) 
    in
      aref := Invalid;
      Array.copy {src=a, dst=na, di=0};
      Unsynchronized.ref (Value na)
    end
    );

fun shrink (aref, sz) = case aref of
  (Unsynchronized.ref Invalid) => raise AccessedOldVersion |
  (Unsynchronized.ref (Value a)) => (
    if sz > Array.length a then 
      raise Size
    else (
      aref:=Invalid;
      Unsynchronized.ref (Value (Array.tabulate (sz,fn i => Array.sub (a,i))))
    )
  );

structure IsabelleMapping = struct
type 'a ArrayType = 'a array;

fun new_array (a:'a) (n:int) = array (n, a);

fun array_length (a:'a ArrayType) = length a;

fun array_get (a:'a ArrayType) (i:int) = sub (a, i);

fun array_set (a:'a ArrayType) (i:int) (e:'a) = update (a, i, e);

fun array_of_list (xs:'a list) = fromList xs;

fun array_grow (a:'a ArrayType) (i:int) (x:'a) = grow (a, i, x);

fun array_shrink (a:'a ArrayType) (sz:int) = shrink (a,sz);

end;

end;

structure FArray = struct
  datatype 'a Cell = Value of 'a Array.array | Upd of (int*'a*'a Cell Unsynchronized.ref);

  type 'a array = 'a Cell Unsynchronized.ref;

  fun array (size,v) = Unsynchronized.ref (Value (Array.array (size,v)));
  fun tabulate (size, f) = Unsynchronized.ref (Value (Array.tabulate(size, f)));
  fun fromList l = Unsynchronized.ref (Value (Array.fromList l));

  fun sub (Unsynchronized.ref (Value a), idx) = Array.sub (a,idx) |
      sub (Unsynchronized.ref (Upd (i,v,cr)),idx) =
        if i=idx then v
        else sub (cr,idx);

  fun length (Unsynchronized.ref (Value a)) = Array.length a |
      length (Unsynchronized.ref (Upd (i,v,cr))) = length cr;

  fun realize_aux (aref, v) =
    case aref of
      (Unsynchronized.ref (Value a)) => (
        let
          val len = Array.length a;
          val a' = Array.array (len,v);
        in
          Array.copy {src=a, dst=a', di=0};
          Unsynchronized.ref (Value a')
        end
      ) |
      (Unsynchronized.ref (Upd (i,v,cr))) => (
        let val res=realize_aux (cr,v) in
          case res of
            (Unsynchronized.ref (Value a)) => (Array.update (a,i,v); res)
        end
      );

  fun realize aref =
    case aref of
      (Unsynchronized.ref (Value _)) => aref |
      (Unsynchronized.ref (Upd (i,v,cr))) => realize_aux(aref,v);

  fun update (aref,idx,v) =
    case aref of
      (Unsynchronized.ref (Value a)) => (
        let val nref=Unsynchronized.ref (Value a) in
          aref := Upd (idx,Array.sub(a,idx),nref);
          Array.update (a,idx,v);
          nref
        end
      ) |
      (Unsynchronized.ref (Upd _)) =>
        let val ra = realize_aux(aref,v) in
          case ra of
            (Unsynchronized.ref (Value a)) => Array.update (a,idx,v);
          ra
        end
      ;

  fun grow (aref, inc, x) = case aref of 
    (Unsynchronized.ref (Value a)) => (
      let val len=Array.length a;
          val na = Array.array (len+inc,x) 
      in
        Array.copy {src=a, dst=na, di=0};
        Unsynchronized.ref (Value na)
      end
      )
  | (Unsynchronized.ref (Upd _)) => (
    grow (realize aref, inc, x)
  );

  fun shrink (aref, sz) = case aref of
    (Unsynchronized.ref (Value a)) => (
      if sz > Array.length a then 
        raise Size
      else (
        Unsynchronized.ref (Value (Array.tabulate (sz,fn i => Array.sub (a,i))))
      )
    ) |
    (Unsynchronized.ref (Upd _)) => (
      shrink (realize aref,sz)
    );

structure IsabelleMapping = struct
type 'a ArrayType = 'a array;

fun new_array (a:'a) (n:int) = array (n, a);

fun array_length (a:'a ArrayType) = length a;

fun array_get (a:'a ArrayType) (i:int) = sub (a, i);

fun array_set (a:'a ArrayType) (i:int) (e:'a) = update (a, i, e);

fun array_of_list (xs:'a list) = fromList xs;

fun array_grow (a:'a ArrayType) (i:int) (x:'a) = grow (a, i, x);

fun array_shrink (a:'a ArrayType) (sz:int) = shrink (a,sz);

end;
end;




structure Fun : sig
  val id : 'a -> 'a
end = struct

fun id x = (fn xa => xa) x;

end; (*struct Fun*)

structure HOL : sig
  type 'a equal
  val equal : 'a equal -> 'a -> 'a -> bool
  val eq : 'a equal -> 'a -> 'a -> bool
  datatype 'a itself = Type
end = struct

type 'a equal = {equal : 'a -> 'a -> bool};
val equal = #equal : 'a equal -> 'a -> 'a -> bool;

fun eq A_ a b = equal A_ a b;

datatype 'a itself = Type;

end; (*struct HOL*)

structure Map : sig
  val map_of : 'a HOL.equal -> ('a * 'b) list -> 'a -> 'b option
end = struct

fun map_of A_ ((l, v) :: ps) k =
  (if HOL.eq A_ l k then SOME v else map_of A_ ps k)
  | map_of A_ [] k = NONE;

end; (*struct Map*)

structure Orderings : sig
  type 'a ord
  val less_eq : 'a ord -> 'a -> 'a -> bool
  val less : 'a ord -> 'a -> 'a -> bool
  val max : 'a ord -> 'a -> 'a -> 'a
  type 'a preorder
  val ord_preorder : 'a preorder -> 'a ord
  type 'a order
  val preorder_order : 'a order -> 'a preorder
  type 'a linorder
  val order_linorder : 'a linorder -> 'a order
end = struct

type 'a ord = {less_eq : 'a -> 'a -> bool, less : 'a -> 'a -> bool};
val less_eq = #less_eq : 'a ord -> 'a -> 'a -> bool;
val less = #less : 'a ord -> 'a -> 'a -> bool;

fun max A_ a b = (if less_eq A_ a b then b else a);

type 'a preorder = {ord_preorder : 'a ord};
val ord_preorder = #ord_preorder : 'a preorder -> 'a ord;

type 'a order = {preorder_order : 'a preorder};
val preorder_order = #preorder_order : 'a order -> 'a preorder;

type 'a linorder = {order_linorder : 'a order};
val order_linorder = #order_linorder : 'a linorder -> 'a order;

end; (*struct Orderings*)

structure RBT_Impl : sig
  datatype color = R | B
  datatype ('a, 'b) rbt = Empty |
    Branch of color * ('a, 'b) rbt * 'a * 'b * ('a, 'b) rbt
  val paint : color -> ('a, 'b) rbt -> ('a, 'b) rbt
  val balance : ('a, 'b) rbt -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val balance_left : ('a, 'b) rbt -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val combine : ('a, 'b) rbt -> ('a, 'b) rbt -> ('a, 'b) rbt
  val balance_right : ('a, 'b) rbt -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val rbt_del : 'a Orderings.ord -> 'a -> ('a, 'b) rbt -> ('a, 'b) rbt
  val rbt_del_from_right :
    'a Orderings.ord ->
      'a -> ('a, 'b) rbt -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val rbt_del_from_left :
    'a Orderings.ord ->
      'a -> ('a, 'b) rbt -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val rbt_ins :
    'a Orderings.ord ->
      ('a -> 'b -> 'b -> 'b) -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val equal_color : color -> color -> bool
  val equal_rbt :
    'a HOL.equal -> 'b HOL.equal -> ('a, 'b) rbt -> ('a, 'b) rbt -> bool
  val rbt_delete : 'a Orderings.ord -> 'a -> ('a, 'b) rbt -> ('a, 'b) rbt
  val rbt_insert_with_key :
    'a Orderings.ord ->
      ('a -> 'b -> 'b -> 'b) -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val rbt_insert : 'a Orderings.ord -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val rbt_lookup : 'a Orderings.ord -> ('a, 'b) rbt -> 'a -> 'b option
end = struct

datatype color = R | B;

datatype ('a, 'b) rbt = Empty |
  Branch of color * ('a, 'b) rbt * 'a * 'b * ('a, 'b) rbt;

fun paint c Empty = Empty
  | paint c (Branch (uu, l, k, v, r)) = Branch (c, l, k, v, r);

fun balance (Branch (R, a, w, x, b)) s t (Branch (R, c, y, z, d)) =
  Branch (R, Branch (B, a, w, x, b), s, t, Branch (B, c, y, z, d))
  | balance (Branch (R, Branch (R, a, w, x, b), s, t, c)) y z Empty =
    Branch (R, Branch (B, a, w, x, b), s, t, Branch (B, c, y, z, Empty))
  | balance (Branch (R, Branch (R, a, w, x, b), s, t, c)) y z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, Branch (B, a, w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance (Branch (R, Empty, w, x, Branch (R, b, s, t, c))) y z Empty =
    Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, Empty))
  | balance
    (Branch (R, Branch (B, va, vb, vc, vd), w, x, Branch (R, b, s, t, c))) y z
    Empty =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, Empty))
  | balance (Branch (R, Empty, w, x, Branch (R, b, s, t, c))) y z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, Branch (B, Empty, w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance
    (Branch (R, Branch (B, ve, vf, vg, vh), w, x, Branch (R, b, s, t, c))) y z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, Branch (B, Branch (B, ve, vf, vg, vh), w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance Empty w x (Branch (R, b, s, t, Branch (R, c, y, z, d))) =
    Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, d))
  | balance (Branch (B, va, vb, vc, vd)) w x
    (Branch (R, b, s, t, Branch (R, c, y, z, d))) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, d))
  | balance Empty w x (Branch (R, Branch (R, b, s, t, c), y, z, Empty)) =
    Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, Empty))
  | balance Empty w x
    (Branch (R, Branch (R, b, s, t, c), y, z, Branch (B, va, vb, vc, vd))) =
    Branch
      (R, Branch (B, Empty, w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance (Branch (B, va, vb, vc, vd)) w x
    (Branch (R, Branch (R, b, s, t, c), y, z, Empty)) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, Empty))
  | balance (Branch (B, va, vb, vc, vd)) w x
    (Branch (R, Branch (R, b, s, t, c), y, z, Branch (B, ve, vf, vg, vh))) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, ve, vf, vg, vh)))
  | balance Empty s t Empty = Branch (B, Empty, s, t, Empty)
  | balance Empty s t (Branch (B, va, vb, vc, vd)) =
    Branch (B, Empty, s, t, Branch (B, va, vb, vc, vd))
  | balance Empty s t (Branch (v, Empty, vb, vc, Empty)) =
    Branch (B, Empty, s, t, Branch (v, Empty, vb, vc, Empty))
  | balance Empty s t (Branch (v, Branch (B, ve, vf, vg, vh), vb, vc, Empty)) =
    Branch
      (B, Empty, s, t, Branch (v, Branch (B, ve, vf, vg, vh), vb, vc, Empty))
  | balance Empty s t (Branch (v, Empty, vb, vc, Branch (B, vf, vg, vh, vi))) =
    Branch
      (B, Empty, s, t, Branch (v, Empty, vb, vc, Branch (B, vf, vg, vh, vi)))
  | balance Empty s t
    (Branch (v, Branch (B, ve, vj, vk, vl), vb, vc, Branch (B, vf, vg, vh, vi)))
    = Branch
        (B, Empty, s, t,
          Branch
            (v, Branch (B, ve, vj, vk, vl), vb, vc, Branch (B, vf, vg, vh, vi)))
  | balance (Branch (B, va, vb, vc, vd)) s t Empty =
    Branch (B, Branch (B, va, vb, vc, vd), s, t, Empty)
  | balance (Branch (B, va, vb, vc, vd)) s t (Branch (B, ve, vf, vg, vh)) =
    Branch (B, Branch (B, va, vb, vc, vd), s, t, Branch (B, ve, vf, vg, vh))
  | balance (Branch (B, va, vb, vc, vd)) s t (Branch (v, Empty, vf, vg, Empty))
    = Branch
        (B, Branch (B, va, vb, vc, vd), s, t, Branch (v, Empty, vf, vg, Empty))
  | balance (Branch (B, va, vb, vc, vd)) s t
    (Branch (v, Branch (B, vi, vj, vk, vl), vf, vg, Empty)) =
    Branch
      (B, Branch (B, va, vb, vc, vd), s, t,
        Branch (v, Branch (B, vi, vj, vk, vl), vf, vg, Empty))
  | balance (Branch (B, va, vb, vc, vd)) s t
    (Branch (v, Empty, vf, vg, Branch (B, vj, vk, vl, vm))) =
    Branch
      (B, Branch (B, va, vb, vc, vd), s, t,
        Branch (v, Empty, vf, vg, Branch (B, vj, vk, vl, vm)))
  | balance (Branch (B, va, vb, vc, vd)) s t
    (Branch (v, Branch (B, vi, vn, vo, vp), vf, vg, Branch (B, vj, vk, vl, vm)))
    = Branch
        (B, Branch (B, va, vb, vc, vd), s, t,
          Branch
            (v, Branch (B, vi, vn, vo, vp), vf, vg, Branch (B, vj, vk, vl, vm)))
  | balance (Branch (v, Empty, vb, vc, Empty)) s t Empty =
    Branch (B, Branch (v, Empty, vb, vc, Empty), s, t, Empty)
  | balance (Branch (v, Empty, vb, vc, Branch (B, ve, vf, vg, vh))) s t Empty =
    Branch
      (B, Branch (v, Empty, vb, vc, Branch (B, ve, vf, vg, vh)), s, t, Empty)
  | balance (Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Empty)) s t Empty =
    Branch
      (B, Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Empty), s, t, Empty)
  | balance
    (Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Branch (B, ve, vj, vk, vl)))
    s t Empty =
    Branch
      (B, Branch
            (v, Branch (B, vf, vg, vh, vi), vb, vc, Branch (B, ve, vj, vk, vl)),
        s, t, Empty)
  | balance (Branch (v, Empty, vf, vg, Empty)) s t (Branch (B, va, vb, vc, vd))
    = Branch
        (B, Branch (v, Empty, vf, vg, Empty), s, t, Branch (B, va, vb, vc, vd))
  | balance (Branch (v, Empty, vf, vg, Branch (B, vi, vj, vk, vl))) s t
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (B, Branch (v, Empty, vf, vg, Branch (B, vi, vj, vk, vl)), s, t,
        Branch (B, va, vb, vc, vd))
  | balance (Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Empty)) s t
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (B, Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Empty), s, t,
        Branch (B, va, vb, vc, vd))
  | balance
    (Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Branch (B, vi, vn, vo, vp)))
    s t (Branch (B, va, vb, vc, vd)) =
    Branch
      (B, Branch
            (v, Branch (B, vj, vk, vl, vm), vf, vg, Branch (B, vi, vn, vo, vp)),
        s, t, Branch (B, va, vb, vc, vd));

fun balance_left (Branch (R, a, k, x, b)) s y c =
  Branch (R, Branch (B, a, k, x, b), s, y, c)
  | balance_left Empty k x (Branch (B, a, s, y, b)) =
    balance Empty k x (Branch (R, a, s, y, b))
  | balance_left (Branch (B, va, vb, vc, vd)) k x (Branch (B, a, s, y, b)) =
    balance (Branch (B, va, vb, vc, vd)) k x (Branch (R, a, s, y, b))
  | balance_left Empty k x (Branch (R, Branch (B, a, s, y, b), t, z, c)) =
    Branch (R, Branch (B, Empty, k, x, a), s, y, balance b t z (paint R c))
  | balance_left (Branch (B, va, vb, vc, vd)) k x
    (Branch (R, Branch (B, a, s, y, b), t, z, c)) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), k, x, a), s, y,
        balance b t z (paint R c))
  | balance_left Empty k x Empty = Empty
  | balance_left Empty k x (Branch (R, Empty, vb, vc, vd)) = Empty
  | balance_left Empty k x (Branch (R, Branch (R, ve, vf, vg, vh), vb, vc, vd))
    = Empty
  | balance_left (Branch (B, va, vb, vc, vd)) k x Empty = Empty
  | balance_left (Branch (B, va, vb, vc, vd)) k x
    (Branch (R, Empty, vf, vg, vh)) = Empty
  | balance_left (Branch (B, va, vb, vc, vd)) k x
    (Branch (R, Branch (R, vi, vj, vk, vl), vf, vg, vh)) = Empty;

fun combine Empty x = x
  | combine (Branch (v, va, vb, vc, vd)) Empty = Branch (v, va, vb, vc, vd)
  | combine (Branch (R, a, k, x, b)) (Branch (R, c, s, y, d)) =
    (case combine b c
      of Empty => Branch (R, a, k, x, Branch (R, Empty, s, y, d))
      | Branch (R, b2, t, z, c2) =>
        Branch (R, Branch (R, a, k, x, b2), t, z, Branch (R, c2, s, y, d))
      | Branch (B, b2, t, z, c2) =>
        Branch (R, a, k, x, Branch (R, Branch (B, b2, t, z, c2), s, y, d)))
  | combine (Branch (B, a, k, x, b)) (Branch (B, c, s, y, d)) =
    (case combine b c
      of Empty => balance_left a k x (Branch (B, Empty, s, y, d))
      | Branch (R, b2, t, z, c2) =>
        Branch (R, Branch (B, a, k, x, b2), t, z, Branch (B, c2, s, y, d))
      | Branch (B, b2, t, z, c2) =>
        balance_left a k x (Branch (B, Branch (B, b2, t, z, c2), s, y, d)))
  | combine (Branch (B, va, vb, vc, vd)) (Branch (R, b, k, x, c)) =
    Branch (R, combine (Branch (B, va, vb, vc, vd)) b, k, x, c)
  | combine (Branch (R, a, k, x, b)) (Branch (B, va, vb, vc, vd)) =
    Branch (R, a, k, x, combine b (Branch (B, va, vb, vc, vd)));

fun balance_right a k x (Branch (R, b, s, y, c)) =
  Branch (R, a, k, x, Branch (B, b, s, y, c))
  | balance_right (Branch (B, a, k, x, b)) s y Empty =
    balance (Branch (R, a, k, x, b)) s y Empty
  | balance_right (Branch (B, a, k, x, b)) s y (Branch (B, va, vb, vc, vd)) =
    balance (Branch (R, a, k, x, b)) s y (Branch (B, va, vb, vc, vd))
  | balance_right (Branch (R, a, k, x, Branch (B, b, s, y, c))) t z Empty =
    Branch (R, balance (paint R a) k x b, s, y, Branch (B, c, t, z, Empty))
  | balance_right (Branch (R, a, k, x, Branch (B, b, s, y, c))) t z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, balance (paint R a) k x b, s, y,
        Branch (B, c, t, z, Branch (B, va, vb, vc, vd)))
  | balance_right Empty k x Empty = Empty
  | balance_right (Branch (R, va, vb, vc, Empty)) k x Empty = Empty
  | balance_right (Branch (R, va, vb, vc, Branch (R, ve, vf, vg, vh))) k x Empty
    = Empty
  | balance_right Empty k x (Branch (B, va, vb, vc, vd)) = Empty
  | balance_right (Branch (R, ve, vf, vg, Empty)) k x
    (Branch (B, va, vb, vc, vd)) = Empty
  | balance_right (Branch (R, ve, vf, vg, Branch (R, vi, vj, vk, vl))) k x
    (Branch (B, va, vb, vc, vd)) = Empty;

fun rbt_del A_ x Empty = Empty
  | rbt_del A_ x (Branch (c, a, y, s, b)) =
    (if Orderings.less A_ x y then rbt_del_from_left A_ x a y s b
      else (if Orderings.less A_ y x then rbt_del_from_right A_ x a y s b
             else combine a b))
and rbt_del_from_right A_ x a y s (Branch (B, lt, z, v, rt)) =
  balance_right a y s (rbt_del A_ x (Branch (B, lt, z, v, rt)))
  | rbt_del_from_right A_ x a y s Empty =
    Branch (R, a, y, s, rbt_del A_ x Empty)
  | rbt_del_from_right A_ x a y s (Branch (R, va, vb, vc, vd)) =
    Branch (R, a, y, s, rbt_del A_ x (Branch (R, va, vb, vc, vd)))
and rbt_del_from_left A_ x (Branch (B, lt, z, v, rt)) y s b =
  balance_left (rbt_del A_ x (Branch (B, lt, z, v, rt))) y s b
  | rbt_del_from_left A_ x Empty y s b = Branch (R, rbt_del A_ x Empty, y, s, b)
  | rbt_del_from_left A_ x (Branch (R, va, vb, vc, vd)) y s b =
    Branch (R, rbt_del A_ x (Branch (R, va, vb, vc, vd)), y, s, b);

fun rbt_ins A_ f k v Empty = Branch (R, Empty, k, v, Empty)
  | rbt_ins A_ f k v (Branch (B, l, x, y, r)) =
    (if Orderings.less A_ k x then balance (rbt_ins A_ f k v l) x y r
      else (if Orderings.less A_ x k then balance l x y (rbt_ins A_ f k v r)
             else Branch (B, l, x, f k y v, r)))
  | rbt_ins A_ f k v (Branch (R, l, x, y, r)) =
    (if Orderings.less A_ k x then Branch (R, rbt_ins A_ f k v l, x, y, r)
      else (if Orderings.less A_ x k
             then Branch (R, l, x, y, rbt_ins A_ f k v r)
             else Branch (R, l, x, f k y v, r)));

fun equal_color B R = false
  | equal_color R B = false
  | equal_color B B = true
  | equal_color R R = true;

fun equal_rbt A_ B_ (Branch (color, rbt1, a, b, rbt2)) Empty = false
  | equal_rbt A_ B_ Empty (Branch (color, rbt1, a, b, rbt2)) = false
  | equal_rbt A_ B_ (Branch (colora, rbt1a, aa, ba, rbt2a))
    (Branch (color, rbt1, a, b, rbt2)) =
    equal_color colora color andalso
      (equal_rbt A_ B_ rbt1a rbt1 andalso
        (HOL.eq A_ aa a andalso
          (HOL.eq B_ ba b andalso equal_rbt A_ B_ rbt2a rbt2)))
  | equal_rbt A_ B_ Empty Empty = true;

fun rbt_delete A_ k t = paint B (rbt_del A_ k t);

fun rbt_insert_with_key A_ f k v t = paint B (rbt_ins A_ f k v t);

fun rbt_insert A_ = rbt_insert_with_key A_ (fn _ => fn _ => fn nv => nv);

fun rbt_lookup A_ Empty k = NONE
  | rbt_lookup A_ (Branch (uu, l, x, y, r)) k =
    (if Orderings.less A_ k x then rbt_lookup A_ l k
      else (if Orderings.less A_ x k then rbt_lookup A_ r k else SOME y));

end; (*struct RBT_Impl*)

structure RBT : sig
  datatype ('a, 'b) rbt = Rbt of ('a, 'b) RBT_Impl.rbt
  val empty : 'a Orderings.linorder -> ('a, 'b) rbt
  val impl_of : 'a Orderings.linorder -> ('a, 'b) rbt -> ('a, 'b) RBT_Impl.rbt
  val delete : 'a Orderings.linorder -> 'a -> ('a, 'b) rbt -> ('a, 'b) rbt
  val insert : 'a Orderings.linorder -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val lookup : 'a Orderings.linorder -> ('a, 'b) rbt -> 'a -> 'b option
end = struct

datatype ('a, 'b) rbt = Rbt of ('a, 'b) RBT_Impl.rbt;

fun empty A_ = Rbt RBT_Impl.Empty;

fun impl_of A_ (Rbt x) = x;

fun delete A_ k t =
  Rbt (RBT_Impl.rbt_delete
        ((Orderings.ord_preorder o Orderings.preorder_order o
           Orderings.order_linorder)
          A_)
        k (impl_of A_ t));

fun insert A_ k v t =
  Rbt (RBT_Impl.rbt_insert
        ((Orderings.ord_preorder o Orderings.preorder_order o
           Orderings.order_linorder)
          A_)
        k v (impl_of A_ t));

fun lookup A_ t =
  RBT_Impl.rbt_lookup
    ((Orderings.ord_preorder o Orderings.preorder_order o
       Orderings.order_linorder)
      A_)
    (impl_of A_ t);

end; (*struct RBT*)

structure Option : sig
  val the : 'a option -> 'a
  val is_none : 'a option -> bool
end = struct

fun the (SOME x) = x;

fun is_none (SOME x) = false
  | is_none NONE = true;

end; (*struct Option*)

structure Product_Type : sig
  val fst : 'a * 'b -> 'a
  val snd : 'a * 'b -> 'b
  val equal_unita : unit -> unit -> bool
  val equal_unit : unit HOL.equal
  val equal_prod : 'a HOL.equal -> 'b HOL.equal -> 'a * 'b -> 'a * 'b -> bool
end = struct

fun fst (a, b) = a;

fun snd (a, b) = b;

fun equal_unita u v = true;

val equal_unit = {equal = equal_unita} : unit HOL.equal;

fun equal_prod A_ B_ (aa, ba) (a, b) = HOL.eq A_ aa a andalso HOL.eq B_ ba b;

end; (*struct Product_Type*)

structure List : sig
  val hd : 'a list -> 'a
  val tl : 'a list -> 'a list
  val map : ('a -> 'b) -> 'a list -> 'b list
  val null : 'a list -> bool
  val foldl : ('a -> 'b -> 'a) -> 'a -> 'b list -> 'a
  val filter : ('a -> bool) -> 'a list -> 'a list
  val partition : ('a -> bool) -> 'a list -> 'a list * 'a list
  val map_filter : ('a -> 'b option) -> 'a list -> 'b list
end = struct

fun hd (x :: xs) = x;

fun tl [] = []
  | tl (x :: xs) = xs;

fun map f [] = []
  | map f (x :: xs) = f x :: map f xs;

fun null [] = true
  | null (x :: xs) = false;

fun foldl f a [] = a
  | foldl f a (x :: xs) = foldl f (f a x) xs;

fun filter p [] = []
  | filter p (x :: xs) = (if p x then x :: filter p xs else filter p xs);

fun partition p [] = ([], [])
  | partition p (x :: xs) =
    let
      val (yes, no) = partition p xs;
    in
      (if p x then (x :: yes, no) else (yes, x :: no))
    end;

fun map_filter f [] = []
  | map_filter f (x :: xs) =
    (case f x of NONE => map_filter f xs | SOME y => y :: map_filter f xs);

end; (*struct List*)

structure Enum : sig
  val product : 'a list -> 'b list -> ('a * 'b) list
end = struct

fun product [] uu = []
  | product (x :: xs) ys = List.map (fn a => (x, a)) ys @ product xs ys;

end; (*struct Enum*)

structure Misc : sig
  val less_eq_prod_aux :
    'a HOL.equal * 'a Orderings.ord -> 'b Orderings.ord ->
      'a * 'b -> 'a * 'b -> bool
  val less_prod :
    'a HOL.equal * 'a Orderings.ord -> 'b HOL.equal * 'b Orderings.ord ->
      'a * 'b -> 'a * 'b -> bool
  val less_eq_prod :
    'a HOL.equal * 'a Orderings.ord -> 'b Orderings.ord ->
      'a * 'b -> 'a * 'b -> bool
  val ord_prod :
    'a HOL.equal * 'a Orderings.ord -> 'b HOL.equal * 'b Orderings.ord ->
      ('a * 'b) Orderings.ord
  val preorder_prod :
    'a HOL.equal * 'a Orderings.order -> 'b HOL.equal * 'b Orderings.order ->
      ('a * 'b) Orderings.preorder
  val order_prod :
    'a HOL.equal * 'a Orderings.order -> 'b HOL.equal * 'b Orderings.order ->
      ('a * 'b) Orderings.order
  val linorder_prod :
    'a HOL.equal * 'a Orderings.linorder ->
      'b HOL.equal * 'b Orderings.linorder -> ('a * 'b) Orderings.linorder
end = struct

fun less_eq_prod_aux (A1_, A2_) B_ (a1, a2) (b1, b2) =
  Orderings.less A2_ a1 b1 orelse
    HOL.eq A1_ a1 b1 andalso Orderings.less_eq B_ a2 b2;

fun less_prod (A1_, A2_) (B1_, B2_) a b =
  not (Product_Type.equal_prod A1_ B1_ a b) andalso
    less_eq_prod_aux (A1_, A2_) B2_ a b;

fun less_eq_prod (A1_, A2_) B_ a b = less_eq_prod_aux (A1_, A2_) B_ a b;

fun ord_prod (A1_, A2_) (B1_, B2_) =
  {less_eq = less_eq_prod (A1_, A2_) B2_,
    less = less_prod (A1_, A2_) (B1_, B2_)}
  : ('a * 'b) Orderings.ord;

fun preorder_prod (A1_, A2_) (B1_, B2_) =
  {ord_preorder =
     ord_prod (A1_, (Orderings.ord_preorder o Orderings.preorder_order) A2_)
       (B1_, (Orderings.ord_preorder o Orderings.preorder_order) B2_)}
  : ('a * 'b) Orderings.preorder;

fun order_prod (A1_, A2_) (B1_, B2_) =
  {preorder_order = preorder_prod (A1_, A2_) (B1_, B2_)} :
  ('a * 'b) Orderings.order;

fun linorder_prod (A1_, A2_) (B1_, B2_) =
  {order_linorder =
     order_prod (A1_, Orderings.order_linorder A2_)
       (B1_, Orderings.order_linorder B2_)}
  : ('a * 'b) Orderings.linorder;

end; (*struct Misc*)

structure AList : sig
  val delete : 'a HOL.equal -> 'a -> ('a * 'b) list -> ('a * 'b) list
  val update : 'a HOL.equal -> 'a -> 'b -> ('a * 'b) list -> ('a * 'b) list
end = struct

fun delete A_ k = List.filter (fn (ka, _) => not (HOL.eq A_ k ka));

fun update A_ k v [] = [(k, v)]
  | update A_ k v (p :: ps) =
    (if HOL.eq A_ (Product_Type.fst p) k then (k, v) :: ps
      else p :: update A_ k v ps);

end; (*struct AList*)

structure Arith : sig
  datatype num = One | Bit0 of num | Bit1 of num
  val suc : IntInf.int -> IntInf.int
  val ord_nat : IntInf.int Orderings.ord
  val equal_nat : IntInf.int HOL.equal
  val preorder_nat : IntInf.int Orderings.preorder
  val order_nat : IntInf.int Orderings.order
  val div_nat : IntInf.int -> IntInf.int -> IntInf.int
  val mod_nat : IntInf.int -> IntInf.int -> IntInf.int
  val linorder_nat : IntInf.int Orderings.linorder
end = struct

datatype num = One | Bit0 of num | Bit1 of num;

fun suc n = IntInf.+ (n, (1 : IntInf.int));

val ord_nat =
  {less_eq = (fn a => fn b => IntInf.<= (a, b)),
    less = (fn a => fn b => IntInf.< (a, b))}
  : IntInf.int Orderings.ord;

val equal_nat = {equal = (fn a => fn b => ((a : IntInf.int) = b))} :
  IntInf.int HOL.equal;

val preorder_nat = {ord_preorder = ord_nat} : IntInf.int Orderings.preorder;

val order_nat = {preorder_order = preorder_nat} : IntInf.int Orderings.order;

fun div_nat m n = Product_Type.fst (IntInf.divMod (m, n));

fun mod_nat m n = Product_Type.snd (IntInf.divMod (m, n));

val linorder_nat = {order_linorder = order_nat} : IntInf.int Orderings.linorder;

end; (*struct Arith*)

structure LTSGA : sig
  val ltsga_copy :
    ('a -> ('b -> bool) -> ('c * ('d * 'c) -> 'e -> 'e) -> 'e -> 'e) ->
      (unit -> 'e) -> ('c -> 'd -> 'c -> 'e -> 'e) -> 'a -> 'e
  val ltsga_reverse :
    (unit -> 'a) ->
      ('b -> 'c -> 'b -> 'a -> 'a) ->
        ('d -> ('e -> bool) -> ('b * ('c * 'b) -> 'a -> 'a) -> 'a -> 'a) ->
          'd -> 'a
  val ltsga_list_to_collect_list :
    'a HOL.equal -> 'c HOL.equal ->
      ('a * ('b * 'c)) list -> ('a * ('b list * 'c)) list
end = struct

fun ltsga_copy it emp add l =
  it l (fn _ => true) (fn (q, (a, b)) => add q a b) (emp ());

fun ltsga_reverse e a it l =
  it l (fn _ => true) (fn (v, (ea, va)) => a va ea v) (e ());

fun ltsga_list_to_collect_list A_ C_ [] = []
  | ltsga_list_to_collect_list A_ C_ ((va, (e, v)) :: l) =
    let
      val (yes, no) =
        List.partition
          (fn vev =>
            HOL.eq A_ (Product_Type.fst vev) va andalso
              HOL.eq C_ (Product_Type.snd (Product_Type.snd vev)) v)
          l;
    in
      (va, (e :: List.map (Product_Type.fst o Product_Type.snd) yes, v)) ::
        ltsga_list_to_collect_list A_ C_ no
    end;

end; (*struct LTSGA*)

structure MapGA : sig
  val eu_sng : (unit -> 'a) -> ('b -> 'c -> 'a -> 'a) -> 'b -> 'c -> 'a
end = struct

fun eu_sng empt update k v = update k v (empt ());

end; (*struct MapGA*)

structure SetGA : sig
  val ei_sng : (unit -> 'a) -> ('b -> 'a -> 'a) -> 'b -> 'a
  val it_copy :
    ('a -> ('b -> bool) -> ('c -> 'b -> 'b) -> 'b -> 'b) ->
      (unit -> 'b) -> ('c -> 'b -> 'b) -> 'a -> 'b
  val it_diff :
    ('a -> ('b -> bool) -> ('c -> 'b -> 'b) -> 'b -> 'b) ->
      ('c -> 'b -> 'b) -> 'b -> 'a -> 'b
  val it_size :
    ('a ->
      (IntInf.int -> bool) ->
        ('b -> IntInf.int -> IntInf.int) -> IntInf.int -> IntInf.int) ->
      'a -> IntInf.int
  val iti_ball :
    ('a -> (bool -> bool) -> ('b -> bool -> bool) -> bool -> bool) ->
      'a -> ('b -> bool) -> bool
  val iflt_image : (('a -> 'b option) -> 'c -> 'd) -> ('a -> 'b) -> 'c -> 'd
  val ball_disjoint :
    ('a -> ('b -> bool) -> 'c) -> ('b -> 'd -> bool) -> 'a -> 'd -> 'c
  val it_set_to_list :
    ('a ->
      ('b list -> bool) -> ('b -> 'b list -> 'b list) -> 'b list -> 'b list) ->
      'a -> 'b list
  val iti_sel_no_map :
    ('a ->
      ('b option -> bool) ->
        ('b -> 'b option -> 'b option) -> 'b option -> 'b option) ->
      'a -> ('b -> bool) -> 'b option
  val gen_list_to_set_aux : ('a -> 'b -> 'b) -> 'b -> 'a list -> 'b
  val gen_list_to_set : (unit -> 'a) -> ('b -> 'a -> 'a) -> 'b list -> 'a
  val it_image_filter :
    ('a -> ('b -> bool) -> ('c -> 'b -> 'b) -> 'b -> 'b) ->
      (unit -> 'b) -> ('d -> 'b -> 'b) -> ('c -> 'd option) -> 'a -> 'b
  val neg_ball_bexists :
    ('a -> ('b -> bool) -> bool) -> 'a -> ('b -> bool) -> bool
end = struct

fun ei_sng e i x = i x (e ());

fun it_copy iti emp ins s = iti s (fn _ => true) ins (emp ());

fun it_diff it del s1 s2 = it s2 (fn _ => true) del s1;

fun it_size iti s = iti s (fn _ => true) (fn _ => Arith.suc) 0;

fun iti_ball iti s p = iti s (fn c => c) (fn x => fn _ => p x) true;

fun iflt_image iflt f s = iflt (fn x => SOME (f x)) s;

fun ball_disjoint ball1 memb2 s1 s2 = ball1 s1 (fn x => not (memb2 x s2));

fun it_set_to_list iti s = iti s (fn _ => true) (fn a => fn b => a :: b) [];

fun iti_sel_no_map iti s f =
  iti s Option.is_none (fn x => fn _ => (if f x then SOME x else NONE)) NONE;

fun gen_list_to_set_aux ins accs [] = accs
  | gen_list_to_set_aux ins accs (x :: l) =
    gen_list_to_set_aux ins (ins x accs) l;

fun gen_list_to_set empt ins = gen_list_to_set_aux ins (empt ());

fun it_image_filter iteratei1 empty2 ins2 f s =
  iteratei1 s (fn _ => true)
    (fn x => fn res => (case f x of NONE => res | SOME v => ins2 v res))
    (empty2 ());

fun neg_ball_bexists ball s p = not (ball s (fn x => not (p x)));

end; (*struct SetGA*)

structure ListGA : sig
  val idx_iteratei_aux :
    ('a -> IntInf.int -> 'b) ->
      IntInf.int ->
        IntInf.int -> 'a -> ('c -> bool) -> ('b -> 'c -> 'c) -> 'c -> 'c
  val idx_iteratei :
    ('a -> IntInf.int -> 'b) ->
      ('a -> IntInf.int) -> 'a -> ('c -> bool) -> ('b -> 'c -> 'c) -> 'c -> 'c
end = struct

fun idx_iteratei_aux get sz i l c f sigma =
  (if ((i : IntInf.int) = 0) orelse not (c sigma) then sigma
    else idx_iteratei_aux get sz
           (IntInf.max (0, IntInf.- (i, (1 : IntInf.int)))) l c f
           (f (get l (IntInf.max (0, IntInf.- (sz, i)))) sigma));

fun idx_iteratei get sz l c f sigma =
  idx_iteratei_aux get (sz l) (sz l) l c f sigma;

end; (*struct ListGA*)

structure ListAdd : sig
  val merge :
    'a HOL.equal * 'a Orderings.linorder -> 'a list -> 'a list -> 'a list
  val merge_list :
    'a HOL.equal * 'a Orderings.linorder ->
      ('a list) list -> ('a list) list -> 'a list
  val insertion_sort :
    'a HOL.equal * 'a Orderings.ord -> 'a -> 'a list -> 'a list
end = struct

fun merge (A1_, A2_) [] l2 = l2
  | merge (A1_, A2_) (v :: va) [] = v :: va
  | merge (A1_, A2_) (x1 :: l1) (x2 :: l2) =
    (if Orderings.less
          ((Orderings.ord_preorder o Orderings.preorder_order o
             Orderings.order_linorder)
            A2_)
          x1 x2
      then x1 :: merge (A1_, A2_) l1 (x2 :: l2)
      else (if HOL.eq A1_ x1 x2 then x1 :: merge (A1_, A2_) l1 l2
             else x2 :: merge (A1_, A2_) (x1 :: l1) l2));

fun merge_list (A1_, A2_) [] [] = []
  | merge_list (A1_, A2_) [] [l] = l
  | merge_list (A1_, A2_) (la :: acc2) [] =
    merge_list (A1_, A2_) [] (la :: acc2)
  | merge_list (A1_, A2_) (la :: acc2) [l] =
    merge_list (A1_, A2_) [] (l :: la :: acc2)
  | merge_list (A1_, A2_) acc2 (l1 :: l2 :: ls) =
    merge_list (A1_, A2_) (merge (A1_, A2_) l1 l2 :: acc2) ls;

fun insertion_sort (A1_, A2_) x [] = [x]
  | insertion_sort (A1_, A2_) x (y :: xs) =
    (if Orderings.less A2_ y x then y :: insertion_sort (A1_, A2_) x xs
      else (if HOL.eq A1_ x y then y :: xs else x :: y :: xs));

end; (*struct ListAdd*)

structure RBT_add : sig
  val equal_rbta :
    'a HOL.equal * 'a Orderings.linorder -> 'b HOL.equal ->
      ('a, 'b) RBT.rbt -> ('a, 'b) RBT.rbt -> bool
  val equal_rbt :
    'a HOL.equal * 'a Orderings.linorder -> 'b HOL.equal ->
      ('a, 'b) RBT.rbt HOL.equal
  val rm_iterateoi :
    ('a, 'b) RBT_Impl.rbt -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
  val rm_reverse_iterateoi :
    ('a, 'b) RBT_Impl.rbt -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
end = struct

fun equal_rbta (A1_, A2_) B_ ra r =
  RBT_Impl.equal_rbt A1_ B_ (RBT.impl_of A2_ ra) (RBT.impl_of A2_ r);

fun equal_rbt (A1_, A2_) B_ = {equal = equal_rbta (A1_, A2_) B_} :
  ('a, 'b) RBT.rbt HOL.equal;

fun rm_iterateoi RBT_Impl.Empty c f sigma = sigma
  | rm_iterateoi (RBT_Impl.Branch (col, l, k, v, r)) c f sigma =
    (if c sigma
      then let
             val sigmaa = rm_iterateoi l c f sigma;
           in
             (if c sigmaa then rm_iterateoi r c f (f (k, v) sigmaa) else sigmaa)
           end
      else sigma);

fun rm_reverse_iterateoi RBT_Impl.Empty c f sigma = sigma
  | rm_reverse_iterateoi (RBT_Impl.Branch (col, l, k, v, r)) c f sigma =
    (if c sigma
      then let
             val sigmaa = rm_reverse_iterateoi r c f sigma;
           in
             (if c sigmaa then rm_reverse_iterateoi l c f (f (k, v) sigmaa)
               else sigmaa)
           end
      else sigma);

end; (*struct RBT_add*)

structure HashCode : sig
  type 'a hashable
  val hashcode : 'a hashable -> 'a -> IntInf.int
  val bounded_hashcode : 'a hashable -> IntInf.int -> 'a -> IntInf.int
  val def_hashmap_size : 'a hashable -> 'a HOL.itself -> IntInf.int
  val hashcode_nat : IntInf.int -> IntInf.int
  val bounded_hashcode_nat : IntInf.int -> IntInf.int -> IntInf.int
  val def_hashmap_size_nat : IntInf.int HOL.itself -> IntInf.int
  val hashable_nat : IntInf.int hashable
end = struct

type 'a hashable =
  {hashcode : 'a -> IntInf.int,
    bounded_hashcode : IntInf.int -> 'a -> IntInf.int,
    def_hashmap_size : 'a HOL.itself -> IntInf.int};
val hashcode = #hashcode : 'a hashable -> 'a -> IntInf.int;
val bounded_hashcode = #bounded_hashcode :
  'a hashable -> IntInf.int -> 'a -> IntInf.int;
val def_hashmap_size = #def_hashmap_size :
  'a hashable -> 'a HOL.itself -> IntInf.int;

fun hashcode_nat n = n;

fun bounded_hashcode_nat na n = Arith.mod_nat n na;

fun def_hashmap_size_nat x = (fn _ => (16 : IntInf.int)) x;

val hashable_nat =
  {hashcode = hashcode_nat, bounded_hashcode = bounded_hashcode_nat,
    def_hashmap_size = def_hashmap_size_nat}
  : IntInf.int hashable;

end; (*struct HashCode*)

structure RBTMapImpl : sig
  val rm_empty : 'a Orderings.linorder -> unit -> ('a, 'b) RBT.rbt
  val rm_delete :
    'a Orderings.linorder -> 'a -> ('a, 'b) RBT.rbt -> ('a, 'b) RBT.rbt
  val rm_lookup : 'a Orderings.linorder -> 'a -> ('a, 'b) RBT.rbt -> 'b option
  val rm_update :
    'a Orderings.linorder -> 'a -> 'b -> ('a, 'b) RBT.rbt -> ('a, 'b) RBT.rbt
  val rm_isEmpty :
    'a HOL.equal * 'a Orderings.linorder -> 'b HOL.equal ->
      ('a, 'b) RBT.rbt -> bool
  val rm_iterateoi :
    'a Orderings.linorder ->
      ('a, 'b) RBT.rbt -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
  val rm_iteratei :
    'a Orderings.linorder ->
      ('a, 'b) RBT.rbt -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
  val rm_update_dj :
    'a Orderings.linorder -> 'a -> 'b -> ('a, 'b) RBT.rbt -> ('a, 'b) RBT.rbt
  val rm_reverse_iterateoi :
    'a Orderings.linorder ->
      ('a, 'b) RBT.rbt -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
end = struct

fun rm_empty A_ = (fn _ => RBT.empty A_);

fun rm_delete A_ = RBT.delete A_;

fun rm_lookup A_ k m = RBT.lookup A_ m k;

fun rm_update A_ = RBT.insert A_;

fun rm_isEmpty (A1_, A2_) B_ m =
  HOL.eq (RBT_add.equal_rbt (A1_, A2_) B_) m (RBT.empty A2_);

fun rm_iterateoi A_ r = RBT_add.rm_iterateoi (RBT.impl_of A_ r);

fun rm_iteratei A_ = rm_iterateoi A_;

fun rm_update_dj A_ = rm_update A_;

fun rm_reverse_iterateoi A_ r = RBT_add.rm_reverse_iterateoi (RBT.impl_of A_ r);

end; (*struct RBTMapImpl*)

structure SetByMap : sig
  val s_ins : ('a -> unit -> 'b -> 'c) -> 'a -> 'b -> 'c
  val s_memb : ('a -> 'b -> 'c option) -> 'a -> 'b -> bool
  val s_ins_dj : ('a -> unit -> 'b -> 'c) -> 'a -> 'b -> 'c
  val s_iteratei :
    ('a -> ('b -> bool) -> ('c * unit -> 'b -> 'b) -> 'b -> 'b) ->
      'a -> ('b -> bool) -> ('c -> 'b -> 'b) -> 'b -> 'b
end = struct

fun s_ins update x s = update x () s;

fun s_memb lookup x s = not (Option.is_none (lookup x s));

fun s_ins_dj update_dj x s = update_dj x () s;

fun s_iteratei it s c f = it s c (fn x => f (Product_Type.fst x));

end; (*struct SetByMap*)

structure RBTSetImpl : sig
  val rs_ins :
    'a Orderings.linorder -> 'a -> ('a, unit) RBT.rbt -> ('a, unit) RBT.rbt
  val rs_memb : 'a Orderings.linorder -> 'a -> ('a, unit) RBT.rbt -> bool
  val rs_empty : 'a Orderings.linorder -> unit -> ('a, unit) RBT.rbt
  val rs_iteratei :
    'a Orderings.linorder ->
      ('a, unit) RBT.rbt -> ('b -> bool) -> ('a -> 'b -> 'b) -> 'b -> 'b
  val rs_image_filter :
    'a Orderings.linorder -> 'b Orderings.linorder ->
      ('a -> 'b option) -> ('a, unit) RBT.rbt -> ('b, unit) RBT.rbt
  val rs_image :
    'a Orderings.linorder -> 'b Orderings.linorder ->
      ('a -> 'b) -> ('a, unit) RBT.rbt -> ('b, unit) RBT.rbt
  val rs_delete :
    'a Orderings.linorder -> 'a -> ('a, unit) RBT.rbt -> ('a, unit) RBT.rbt
  val rs_ins_dj :
    'a Orderings.linorder -> 'a -> ('a, unit) RBT.rbt -> ('a, unit) RBT.rbt
  val list_to_rs : 'a Orderings.linorder -> 'a list -> ('a, unit) RBT.rbt
  val rs_isEmpty :
    'a HOL.equal * 'a Orderings.linorder -> ('a, unit) RBT.rbt -> bool
  val rs_reverse_iterateoi :
    'a Orderings.linorder ->
      ('a, unit) RBT.rbt -> ('b -> bool) -> ('a -> 'b -> 'b) -> 'b -> 'b
  val rs_to_list : 'a Orderings.linorder -> ('a, unit) RBT.rbt -> 'a list
end = struct

fun rs_ins A_ = SetByMap.s_ins (RBTMapImpl.rm_update A_);

fun rs_memb A_ = SetByMap.s_memb (RBTMapImpl.rm_lookup A_);

fun rs_empty A_ = RBTMapImpl.rm_empty A_;

fun rs_iteratei A_ = SetByMap.s_iteratei (RBTMapImpl.rm_iteratei A_);

fun rs_image_filter A_ B_ =
  SetGA.it_image_filter (rs_iteratei A_) (rs_empty B_) (rs_ins B_);

fun rs_image A_ B_ = SetGA.iflt_image (rs_image_filter A_ B_);

fun rs_delete A_ = RBTMapImpl.rm_delete A_;

fun rs_ins_dj A_ = SetByMap.s_ins_dj (RBTMapImpl.rm_update_dj A_);

fun list_to_rs A_ = SetGA.gen_list_to_set (rs_empty A_) (rs_ins A_);

fun rs_isEmpty (A1_, A2_) =
  RBTMapImpl.rm_isEmpty (A1_, A2_) Product_Type.equal_unit;

fun rs_reverse_iterateoi A_ =
  SetByMap.s_iteratei (RBTMapImpl.rm_reverse_iterateoi A_);

fun rs_to_list A_ = SetGA.it_set_to_list (rs_reverse_iterateoi A_);

end; (*struct RBTSetImpl*)

structure ListIterator : sig
  val foldli : 'a list -> ('b -> bool) -> ('a -> 'b -> 'b) -> 'b -> 'b
end = struct

fun foldli [] c f sigma = sigma
  | foldli (x :: xs) c f sigma =
    (if c sigma then foldli xs c f (f x sigma) else sigma);

end; (*struct ListIterator*)

structure ArrayHashMap_Impl : sig
  datatype ('a, 'b) hashmap =
    HashMap of (('a * 'b) list) FArray.IsabelleMapping.ArrayType * IntInf.int
  val hm_grow : 'a HashCode.hashable -> ('a, 'b) hashmap -> IntInf.int
  val ahm_alpha_aux :
    'a HOL.equal * 'a HashCode.hashable ->
      (('a * 'b) list) FArray.IsabelleMapping.ArrayType -> 'a -> 'b option
  val ahm_alpha :
    'a HOL.equal * 'a HashCode.hashable -> ('a, 'b) hashmap -> 'a -> 'b option
  val new_hashmap_with : 'a HashCode.hashable -> IntInf.int -> ('a, 'b) hashmap
  val ahm_empty : 'a HashCode.hashable -> unit -> ('a, 'b) hashmap
  val ahm_delete :
    'a HOL.equal * 'a HashCode.hashable ->
      'a -> ('a, 'b) hashmap -> ('a, 'b) hashmap
  val load_factor : IntInf.int
  val ahm_filled : 'a HashCode.hashable -> ('a, 'b) hashmap -> bool
  val ahm_lookup :
    'a HOL.equal * 'a HashCode.hashable -> 'a -> ('a, 'b) hashmap -> 'b option
  val ahm_rehash_auxa :
    'a HashCode.hashable ->
      IntInf.int ->
        'a * 'b ->
          (('a * 'b) list) FArray.IsabelleMapping.ArrayType ->
            (('a * 'b) list) FArray.IsabelleMapping.ArrayType
  val ahm_iteratei_aux :
    'a HashCode.hashable ->
      (('a * 'b) list) FArray.IsabelleMapping.ArrayType ->
        ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
  val ahm_rehash_aux :
    'a HashCode.hashable ->
      (('a * 'b) list) FArray.IsabelleMapping.ArrayType ->
        IntInf.int -> (('a * 'b) list) FArray.IsabelleMapping.ArrayType
  val ahm_rehash :
    'a HashCode.hashable -> ('a, 'b) hashmap -> IntInf.int -> ('a, 'b) hashmap
  val ahm_update_aux :
    'a HOL.equal * 'a HashCode.hashable ->
      ('a, 'b) hashmap -> 'a -> 'b -> ('a, 'b) hashmap
  val ahm_update :
    'a HOL.equal * 'a HashCode.hashable ->
      'a -> 'b -> ('a, 'b) hashmap -> ('a, 'b) hashmap
  val ahm_iteratei :
    'a HashCode.hashable ->
      ('a, 'b) hashmap -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
end = struct

datatype ('a, 'b) hashmap =
  HashMap of (('a * 'b) list) FArray.IsabelleMapping.ArrayType * IntInf.int;

fun hm_grow A_ (HashMap (a, n)) =
  IntInf.+ (IntInf.* ((2 : IntInf.int), FArray.IsabelleMapping.array_length a),
    (3 : IntInf.int));

fun ahm_alpha_aux (A1_, A2_) a k =
  Map.map_of A1_
    (FArray.IsabelleMapping.array_get a
      (HashCode.bounded_hashcode A2_ (FArray.IsabelleMapping.array_length a) k))
    k;

fun ahm_alpha (A1_, A2_) (HashMap (a, uu)) = ahm_alpha_aux (A1_, A2_) a;

fun new_hashmap_with A_ size =
  HashMap (FArray.IsabelleMapping.new_array [] size, 0);

fun ahm_empty A_ =
  (fn _ => new_hashmap_with A_ (HashCode.def_hashmap_size A_ HOL.Type));

fun ahm_delete (A1_, A2_) k (HashMap (a, n)) =
  let
    val h =
      HashCode.bounded_hashcode A2_ (FArray.IsabelleMapping.array_length a) k;
    val m = FArray.IsabelleMapping.array_get a h;
    val deleted = not (Option.is_none (Map.map_of A1_ m k));
  in
    HashMap
      (FArray.IsabelleMapping.array_set a h (AList.delete A1_ k m),
        (if deleted then IntInf.max (0, IntInf.- (n, (1 : IntInf.int))) else n))
  end;

val load_factor : IntInf.int = (75 : IntInf.int);

fun ahm_filled A_ (HashMap (a, n)) =
  IntInf.<= (IntInf.* (FArray.IsabelleMapping.array_length a, load_factor),
    IntInf.* (n, (100 : IntInf.int)));

fun ahm_lookup (A1_, A2_) k hm = ahm_alpha (A1_, A2_) hm k;

fun ahm_rehash_auxa A_ n kv a =
  let
    val h = HashCode.bounded_hashcode A_ n (Product_Type.fst kv);
  in
    FArray.IsabelleMapping.array_set a h
      (kv :: FArray.IsabelleMapping.array_get a h)
  end;

fun ahm_iteratei_aux A_ a c f sigma =
  ListGA.idx_iteratei FArray.IsabelleMapping.array_get
    FArray.IsabelleMapping.array_length a c (fn x => ListIterator.foldli x c f)
    sigma;

fun ahm_rehash_aux A_ a sz =
  ahm_iteratei_aux A_ a (fn _ => true) (ahm_rehash_auxa A_ sz)
    (FArray.IsabelleMapping.new_array [] sz);

fun ahm_rehash A_ (HashMap (a, n)) sz = HashMap (ahm_rehash_aux A_ a sz, n);

fun ahm_update_aux (A1_, A2_) (HashMap (a, n)) k v =
  let
    val h =
      HashCode.bounded_hashcode A2_ (FArray.IsabelleMapping.array_length a) k;
    val m = FArray.IsabelleMapping.array_get a h;
    val insert = Option.is_none (Map.map_of A1_ m k);
  in
    HashMap
      (FArray.IsabelleMapping.array_set a h (AList.update A1_ k v m),
        (if insert then IntInf.+ (n, (1 : IntInf.int)) else n))
  end;

fun ahm_update (A1_, A2_) k v hm =
  let
    val hma = ahm_update_aux (A1_, A2_) hm k v;
  in
    (if ahm_filled A2_ hma then ahm_rehash A2_ hma (hm_grow A2_ hma) else hma)
  end;

fun ahm_iteratei A_ (HashMap (a, n)) = ahm_iteratei_aux A_ a;

end; (*struct ArrayHashMap_Impl*)

structure ArrayHashMap : sig
  datatype ('a, 'b) hashmap = HashMap of ('a, 'b) ArrayHashMap_Impl.hashmap
  val impl_of :
    'a HashCode.hashable ->
      ('a, 'b) hashmap -> ('a, 'b) ArrayHashMap_Impl.hashmap
  val ahm_empty_const : 'a HashCode.hashable -> ('a, 'b) hashmap
  val ahm_empty : 'a HashCode.hashable -> unit -> ('a, 'b) hashmap
  val ahm_delete :
    'a HOL.equal * 'a HashCode.hashable ->
      'a -> ('a, 'b) hashmap -> ('a, 'b) hashmap
  val ahm_lookup :
    'a HOL.equal * 'a HashCode.hashable -> 'a -> ('a, 'b) hashmap -> 'b option
  val ahm_update :
    'a HOL.equal * 'a HashCode.hashable ->
      'a -> 'b -> ('a, 'b) hashmap -> ('a, 'b) hashmap
  val ahm_iteratei :
    'a HashCode.hashable ->
      ('a, 'b) hashmap -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
end = struct

datatype ('a, 'b) hashmap = HashMap of ('a, 'b) ArrayHashMap_Impl.hashmap;

fun impl_of A_ (HashMap x) = x;

fun ahm_empty_const A_ = HashMap (ArrayHashMap_Impl.ahm_empty A_ ());

fun ahm_empty A_ = (fn _ => ahm_empty_const A_);

fun ahm_delete (A1_, A2_) k hm =
  HashMap (ArrayHashMap_Impl.ahm_delete (A1_, A2_) k (impl_of A2_ hm));

fun ahm_lookup (A1_, A2_) k hm =
  ArrayHashMap_Impl.ahm_lookup (A1_, A2_) k (impl_of A2_ hm);

fun ahm_update (A1_, A2_) k v hm =
  HashMap (ArrayHashMap_Impl.ahm_update (A1_, A2_) k v (impl_of A2_ hm));

fun ahm_iteratei A_ hm = ArrayHashMap_Impl.ahm_iteratei A_ (impl_of A_ hm);

end; (*struct ArrayHashMap*)

structure ArrayHashSet : sig
  val ahs_ins :
    'a HOL.equal * 'a HashCode.hashable ->
      'a -> ('a, unit) ArrayHashMap.hashmap -> ('a, unit) ArrayHashMap.hashmap
  val ahs_empty :
    'a HashCode.hashable -> unit -> ('a, unit) ArrayHashMap.hashmap
  val ahs_delete :
    'a HOL.equal * 'a HashCode.hashable ->
      'a -> ('a, unit) ArrayHashMap.hashmap -> ('a, unit) ArrayHashMap.hashmap
  val ahs_iteratei :
    'a HashCode.hashable ->
      ('a, unit) ArrayHashMap.hashmap ->
        ('b -> bool) -> ('a -> 'b -> 'b) -> 'b -> 'b
end = struct

fun ahs_ins (A1_, A2_) = SetByMap.s_ins (ArrayHashMap.ahm_update (A1_, A2_));

fun ahs_empty A_ = ArrayHashMap.ahm_empty A_;

fun ahs_delete (A1_, A2_) = ArrayHashMap.ahm_delete (A1_, A2_);

fun ahs_iteratei A_ = SetByMap.s_iteratei (ArrayHashMap.ahm_iteratei A_);

end; (*struct ArrayHashSet*)

structure ArrayMapImpl : sig
  val iam_alpha :
    ('a option) FArray.IsabelleMapping.ArrayType -> IntInf.int -> 'a option
  val iam_empty : unit -> ('a option) FArray.IsabelleMapping.ArrayType
  val iam_lookup :
    IntInf.int -> ('a option) FArray.IsabelleMapping.ArrayType -> 'a option
  val iam_increment : IntInf.int -> IntInf.int -> IntInf.int
  val iam_update :
    IntInf.int ->
      'a -> ('a option) FArray.IsabelleMapping.ArrayType ->
              ('a option) FArray.IsabelleMapping.ArrayType
end = struct

fun iam_alpha a i =
  (if IntInf.< (i, FArray.IsabelleMapping.array_length a)
    then FArray.IsabelleMapping.array_get a i else NONE);

fun iam_empty x = (fn _ => FArray.IsabelleMapping.array_of_list []) x;

fun iam_lookup k a = iam_alpha a k;

fun iam_increment l idx =
  Orderings.max Arith.ord_nat
    (IntInf.max (0, IntInf.- (IntInf.+ (idx, (1 : IntInf.int)), l)))
    (IntInf.+ (IntInf.* ((2 : IntInf.int), l), (3 : IntInf.int)));

fun iam_update k v a =
  let
    val l = FArray.IsabelleMapping.array_length a;
    val aa =
      (if IntInf.< (k, l) then a
        else FArray.IsabelleMapping.array_grow a (iam_increment l k) NONE);
  in
    FArray.IsabelleMapping.array_set aa k (SOME v)
  end;

end; (*struct ArrayMapImpl*)

structure StdInst : sig
  val rm_sng : 'a Orderings.linorder -> 'a -> 'b -> ('a, 'b) RBT.rbt
  val rs_sng : 'a Orderings.linorder -> 'a -> ('a, unit) RBT.rbt
  val ahm_sng :
    'a HOL.equal * 'a HashCode.hashable ->
      'a -> 'b -> ('a, 'b) ArrayHashMap.hashmap
  val ahs_sng :
    'a HOL.equal * 'a HashCode.hashable -> 'a -> ('a, unit) ArrayHashMap.hashmap
  val iam_sng : IntInf.int -> 'a -> ('a option) FArray.IsabelleMapping.ArrayType
  val rr_diff :
    'a Orderings.linorder ->
      ('a, unit) RBT.rbt -> ('a, unit) RBT.rbt -> ('a, unit) RBT.rbt
  val rs_ball :
    'a Orderings.linorder -> ('a, unit) RBT.rbt -> ('a -> bool) -> bool
  val rs_size : 'a Orderings.linorder -> ('a, unit) RBT.rbt -> IntInf.int
  val rs_bexists :
    'a Orderings.linorder -> ('a, unit) RBT.rbt -> ('a -> bool) -> bool
  val rr_disjoint :
    'a Orderings.linorder -> ('a, unit) RBT.rbt -> ('a, unit) RBT.rbt -> bool
end = struct

fun rm_sng A_ = MapGA.eu_sng (RBTMapImpl.rm_empty A_) (RBTMapImpl.rm_update A_);

fun rs_sng A_ = SetGA.ei_sng (RBTSetImpl.rs_empty A_) (RBTSetImpl.rs_ins A_);

fun ahm_sng (A1_, A2_) =
  MapGA.eu_sng (ArrayHashMap.ahm_empty A2_)
    (ArrayHashMap.ahm_update (A1_, A2_));

fun ahs_sng (A1_, A2_) =
  SetGA.ei_sng (ArrayHashSet.ahs_empty A2_) (ArrayHashSet.ahs_ins (A1_, A2_));

fun iam_sng x = MapGA.eu_sng ArrayMapImpl.iam_empty ArrayMapImpl.iam_update x;

fun rr_diff A_ =
  SetGA.it_diff (RBTSetImpl.rs_iteratei A_) (RBTSetImpl.rs_delete A_);

fun rs_ball A_ = SetGA.iti_ball (RBTSetImpl.rs_iteratei A_);

fun rs_size A_ = SetGA.it_size (RBTSetImpl.rs_iteratei A_);

fun rs_bexists A_ = SetGA.neg_ball_bexists (rs_ball A_);

fun rr_disjoint A_ = SetGA.ball_disjoint (rs_ball A_) (RBTSetImpl.rs_memb A_);

end; (*struct StdInst*)

structure Workset : sig
  val worklist :
    ('a -> bool) -> ('a -> 'b -> 'a * 'b list) -> 'a * 'b list -> 'a * 'b list
end = struct

fun worklist b f (s, e :: wl) =
  (if b s then let
                 val (sa, n) = f s e;
               in
                 worklist b f (sa, n @ wl)
               end
    else (s, e :: wl))
  | worklist b f (s, []) = (s, []);

end; (*struct Workset*)

structure NFAByLTS : sig
  datatype 'a nfa_props_ext = Nfa_props_ext of bool * bool * 'a
  val fields : bool -> bool -> unit nfa_props_ext
  val hopcroft_class_map_alpha_impl :
    (IntInf.int -> 'a option) -> IntInf.int -> IntInf.int -> 'a list
  val nfa_prop_is_initially_connected : 'a nfa_props_ext -> bool
  val nfa_prop_is_complete_deterministic : 'a nfa_props_ext -> bool
end = struct

datatype 'a nfa_props_ext = Nfa_props_ext of bool * bool * 'a;

fun fields nfa_prop_is_complete_deterministic nfa_prop_is_initially_connected =
  Nfa_props_ext
    (nfa_prop_is_complete_deterministic, nfa_prop_is_initially_connected, ());

fun hopcroft_class_map_alpha_impl pim l u =
  (if IntInf.<= (l, u)
    then Option.the (pim l) :: hopcroft_class_map_alpha_impl pim (Arith.suc l) u
    else []);

fun nfa_prop_is_initially_connected
  (Nfa_props_ext
    (nfa_prop_is_complete_deterministic, nfa_prop_is_initially_connected, more))
  = nfa_prop_is_initially_connected;

fun nfa_prop_is_complete_deterministic
  (Nfa_props_ext
    (nfa_prop_is_complete_deterministic, nfa_prop_is_initially_connected, more))
  = nfa_prop_is_complete_deterministic;

end; (*struct NFAByLTS*)

structure Dlist_add : sig
  val dlist_remove1 : 'a HOL.equal -> 'a -> 'a list -> 'a list -> 'a list
end = struct

fun dlist_remove1 A_ x z [] = z
  | dlist_remove1 A_ x z (y :: ys) =
    (if HOL.eq A_ x y then z @ ys else dlist_remove1 A_ x (y :: z) ys);

end; (*struct Dlist_add*)

structure SetIteratorOperations : sig
  val set_iterator_emp : ('a -> bool) -> ('b -> 'a -> 'a) -> 'a -> 'a
  val set_iterator_sng : 'a -> ('b -> bool) -> ('a -> 'b -> 'b) -> 'b -> 'b
end = struct

fun set_iterator_emp c f sigma_0 = sigma_0;

fun set_iterator_sng x c f sigma_0 =
  (if c sigma_0 then f x sigma_0 else sigma_0);

end; (*struct SetIteratorOperations*)

structure RBT_DLTSImpl : sig
  val rs_dlts_it :
    'a Orderings.linorder -> 'b Orderings.linorder ->
      ('a, ('b, 'c) RBT.rbt) RBT.rbt ->
        ('d -> bool) -> ('a * ('b * 'c) -> 'd -> 'd) -> 'd -> 'd
  val rs_dlts_add :
    'a Orderings.linorder -> 'b Orderings.linorder ->
      'a -> 'b -> 'a -> ('a, ('b, 'a) RBT.rbt) RBT.rbt ->
                          ('a, ('b, 'a) RBT.rbt) RBT.rbt
  val rs_dlts_empty :
    'a Orderings.linorder -> 'b Orderings.linorder ->
      unit -> ('a, ('b, 'a) RBT.rbt) RBT.rbt
  val rs_dlts_succ_it :
    'a Orderings.linorder -> 'b Orderings.linorder ->
      ('a, ('b, 'a) RBT.rbt) RBT.rbt ->
        'a -> 'b -> ('c -> bool) -> ('a -> 'c -> 'c) -> 'c -> 'c
  val rs_dlts_filter_it :
    'a Orderings.linorder -> 'b Orderings.linorder ->
      ('a -> bool) ->
        ('b -> bool) ->
          ('c -> bool) ->
            ('a * ('b * 'c) -> bool) ->
              ('a, ('b, 'c) RBT.rbt) RBT.rbt ->
                ('d -> bool) -> ('a * ('b * 'c) -> 'd -> 'd) -> 'd -> 'd
  val rs_dlts_succ_label_it :
    'a Orderings.linorder -> 'b Orderings.linorder ->
      ('a, ('b, 'a) RBT.rbt) RBT.rbt ->
        'a -> ('c -> bool) -> ('b * 'a -> 'c -> 'c) -> 'c -> 'c
end = struct

fun rs_dlts_it A_ B_ =
  (fn m1 => fn c => fn f =>
    RBTMapImpl.rm_iteratei A_ m1 c
      (fn (v1, m2) =>
        RBTMapImpl.rm_iteratei B_ m2 c (fn (e, v2) => f (v1, (e, v2)))));

fun rs_dlts_add A_ B_ =
  (fn v => fn w => fn va => fn l =>
    (case RBTMapImpl.rm_lookup A_ v l
      of NONE => RBTMapImpl.rm_update A_ v (StdInst.rm_sng B_ w va) l
      | SOME m2 =>
        RBTMapImpl.rm_update A_ v (RBTMapImpl.rm_update B_ w va m2) l));

fun rs_dlts_empty A_ B_ = RBTMapImpl.rm_empty A_;

fun rs_dlts_succ_it A_ B_ =
  (fn m1 => fn v => fn e =>
    (case RBTMapImpl.rm_lookup A_ v m1
      of NONE => SetIteratorOperations.set_iterator_emp
      | SOME m2 =>
        (case RBTMapImpl.rm_lookup B_ e m2
          of NONE => SetIteratorOperations.set_iterator_emp
          | SOME a => SetIteratorOperations.set_iterator_sng a)));

fun rs_dlts_filter_it A_ B_ =
  (fn p_v1 => fn p_e => fn p_v2 => fn p => fn m1 => fn c => fn f =>
    RBTMapImpl.rm_iteratei A_ m1 c
      (fn (v1, m2) => fn sigma =>
        (if p_v1 v1
          then RBTMapImpl.rm_iteratei B_ m2 c
                 (fn (e, v2) => fn sigmaa =>
                   (if p_v2 v2 andalso (p_e e andalso p (v1, (e, v2)))
                     then f (v1, (e, v2)) sigmaa else sigmaa))
                 sigma
          else sigma)));

fun rs_dlts_succ_label_it A_ B_ =
  (fn m1 => fn v =>
    (case RBTMapImpl.rm_lookup A_ v m1
      of NONE => (fn _ => fn _ => fn sigma_0 => sigma_0)
      | SOME a => RBTMapImpl.rm_iteratei B_ a));

end; (*struct RBT_DLTSImpl*)

structure LTSByLTS_DLTS : sig
  datatype ('a, 'b) lTS_choice = Lts of 'a | Dlts of 'b
  val dltsbc_add :
    ('a -> 'b -> 'a -> 'c -> 'c) ->
      ('a -> 'b -> 'a -> 'd -> 'd) ->
        'a -> 'b -> 'a -> ('c, 'd) lTS_choice -> ('c, 'd) lTS_choice
  val ltsbc_emp_lts : (unit -> 'a) -> unit -> ('a, 'b) lTS_choice
  val ltsbc_emp_dlts : (unit -> 'a) -> unit -> ('b, 'a) lTS_choice
  val ltsbc_filter_it :
    (('a -> bool) ->
      ('b -> bool) ->
        ('a -> bool) ->
          ('a * ('b * 'a) -> bool) ->
            'c -> ('d -> bool) -> ('a * ('b * 'a) -> 'd -> 'd) -> 'd -> 'd) ->
      (('a -> bool) ->
        ('b -> bool) ->
          ('a -> bool) ->
            ('a * ('b * 'a) -> bool) ->
              'e -> ('d -> bool) -> ('a * ('b * 'a) -> 'd -> 'd) -> 'd -> 'd) ->
        ('a -> bool) ->
          ('b -> bool) ->
            ('a -> bool) ->
              ('a * ('b * 'a) -> bool) ->
                ('c, 'e) lTS_choice ->
                  ('d -> bool) -> ('a * ('b * 'a) -> 'd -> 'd) -> 'd -> 'd
  val ltsbc_add_simple :
    ('a -> 'b -> 'a -> 'c -> 'c) ->
      ('d -> 'c) -> 'a -> 'b -> 'a -> ('c, 'd) lTS_choice -> ('c, 'd) lTS_choice
end = struct

datatype ('a, 'b) lTS_choice = Lts of 'a | Dlts of 'b;

fun dltsbc_add l_add d_add qa a q (Lts l) = Lts (l_add qa a q l)
  | dltsbc_add l_add d_add qa a q (Dlts d) = Dlts (d_add qa a q d);

fun ltsbc_emp_lts l_emp = (fn _ => Lts (l_emp ()));

fun ltsbc_emp_dlts d_emp = (fn _ => Dlts (d_emp ()));

fun ltsbc_filter_it l_it d_it p_qa p_a p_q p (Lts l) = l_it p_qa p_a p_q p l
  | ltsbc_filter_it l_it d_it p_qa p_a p_q p (Dlts d) = d_it p_qa p_a p_q p d;

fun ltsbc_add_simple l_add copy qa a q (Lts l) = Lts (l_add qa a q l)
  | ltsbc_add_simple l_add copy qa a q (Dlts d) = Lts (l_add qa a q (copy d));

end; (*struct LTSByLTS_DLTS*)

structure SemiAutomaton : sig
  val states_enumerate_nat : IntInf.int -> IntInf.int
end = struct

fun states_enumerate_nat q = q;

end; (*struct SemiAutomaton*)

structure SetIteratorGA : sig
  val iterate_size :
    ((IntInf.int -> bool) ->
      ('a -> IntInf.int -> IntInf.int) -> IntInf.int -> IntInf.int) ->
      IntInf.int
end = struct

fun iterate_size it = it (fn _ => true) (fn _ => Arith.suc) 0;

end; (*struct SetIteratorGA*)

structure While_Combinator : sig
  val while_option : ('a -> bool) -> ('a -> 'a) -> 'a -> 'a option
  val whilea : ('a -> bool) -> ('a -> 'a) -> 'a -> 'a
end = struct

fun while_option b c s = (if b s then while_option b c (c s) else SOME s);

fun whilea b c s = Option.the (while_option b c s);

end; (*struct While_Combinator*)

structure RBT_TripleSetImpl : sig
  val rs_ts_it :
    'a Orderings.linorder -> 'b Orderings.linorder -> 'c Orderings.linorder ->
      ('a, ('b, ('c, unit) RBT.rbt) RBT.rbt) RBT.rbt ->
        ('d -> bool) -> ('a * ('b * 'c) -> 'd -> 'd) -> 'd -> 'd
  val rs_ts_add :
    'a Orderings.linorder -> 'b Orderings.linorder -> 'c Orderings.linorder ->
      'a -> 'b -> 'c -> ('a, ('b, ('c, unit) RBT.rbt) RBT.rbt) RBT.rbt ->
                          ('a, ('b, ('c, unit) RBT.rbt) RBT.rbt) RBT.rbt
  val rs_ts_C_it :
    'a Orderings.linorder -> 'b Orderings.linorder -> 'c Orderings.linorder ->
      ('a, ('b, ('c, unit) RBT.rbt) RBT.rbt) RBT.rbt ->
        'a -> 'b -> ('d -> bool) -> ('c -> 'd -> 'd) -> 'd -> 'd
  val rs_ts_BC_it :
    'a Orderings.linorder -> 'b Orderings.linorder -> 'c Orderings.linorder ->
      ('a, ('b, ('c, unit) RBT.rbt) RBT.rbt) RBT.rbt ->
        'a -> ('d -> bool) -> ('b * 'c -> 'd -> 'd) -> 'd -> 'd
  val rs_ts_empty :
    'a Orderings.linorder -> 'b Orderings.linorder -> 'c Orderings.linorder ->
      unit -> ('a, ('b, ('c, unit) RBT.rbt) RBT.rbt) RBT.rbt
  val rs_ts_filter_it :
    'a Orderings.linorder -> 'b Orderings.linorder -> 'c Orderings.linorder ->
      ('a -> bool) ->
        ('b -> bool) ->
          ('c -> bool) ->
            ('a * ('b * 'c) -> bool) ->
              ('a, ('b, ('c, unit) RBT.rbt) RBT.rbt) RBT.rbt ->
                ('d -> bool) -> ('a * ('b * 'c) -> 'd -> 'd) -> 'd -> 'd
end = struct

fun rs_ts_it A_ B_ C_ =
  (fn m1 => fn c => fn f =>
    RBTMapImpl.rm_iteratei A_ m1 c
      (fn (a, m2) =>
        RBTMapImpl.rm_iteratei B_ m2 c
          (fn (b, s3) =>
            RBTSetImpl.rs_iteratei C_ s3 c (fn ca => f (a, (b, ca))))));

fun rs_ts_add A_ B_ C_ =
  (fn v => fn w => fn va => fn l =>
    (case RBTMapImpl.rm_lookup A_ v l
      of NONE =>
        RBTMapImpl.rm_update A_ v (StdInst.rm_sng B_ w (StdInst.rs_sng C_ va)) l
      | SOME m2 =>
        (case RBTMapImpl.rm_lookup B_ w m2
          of NONE =>
            RBTMapImpl.rm_update A_ v
              (RBTMapImpl.rm_update B_ w (StdInst.rs_sng C_ va) m2) l
          | SOME s3 =>
            RBTMapImpl.rm_update A_ v
              (RBTMapImpl.rm_update B_ w (RBTSetImpl.rs_ins C_ va s3) m2) l)));

fun rs_ts_C_it A_ B_ C_ =
  (fn m1 => fn a => fn b =>
    (case RBTMapImpl.rm_lookup A_ a m1
      of NONE => (fn _ => fn _ => fn sigma_0 => sigma_0)
      | SOME m2 =>
        (case RBTMapImpl.rm_lookup B_ b m2
          of NONE => (fn _ => fn _ => fn sigma_0 => sigma_0)
          | SOME aa => RBTSetImpl.rs_iteratei C_ aa)));

fun rs_ts_BC_it A_ B_ C_ =
  (fn m1 => fn a =>
    (case RBTMapImpl.rm_lookup A_ a m1
      of NONE => (fn _ => fn _ => fn sigma_0 => sigma_0)
      | SOME m2 =>
        (fn c => fn f =>
          RBTMapImpl.rm_iteratei B_ m2 c
            (fn (b, s3) =>
              RBTSetImpl.rs_iteratei C_ s3 c (fn ca => f (b, ca))))));

fun rs_ts_empty A_ B_ C_ = RBTMapImpl.rm_empty A_;

fun rs_ts_filter_it A_ B_ C_ =
  (fn p_a => fn p_b => fn p_c => fn p => fn m1 => fn c => fn f =>
    RBTMapImpl.rm_iteratei A_ m1 c
      (fn (a, m2) => fn sigma =>
        (if p_a a
          then RBTMapImpl.rm_iteratei B_ m2 c
                 (fn (b, s3) => fn sigmaa =>
                   (if p_b b
                     then RBTSetImpl.rs_iteratei C_ s3 c
                            (fn ca => fn sigmab =>
                              (if p_c ca andalso p (a, (b, ca))
                                then f (a, (b, ca)) sigmab else sigmab))
                            sigmaa
                     else sigmaa))
                 sigma
          else sigma)));

end; (*struct RBT_TripleSetImpl*)

structure ListSetImpl_Sorted : sig
  val lss_ins : 'a HOL.equal * 'a Orderings.linorder -> 'a -> 'a list -> 'a list
  val lss_sng : 'a -> 'a list
  val lss_empty : unit -> 'a list
  val lss_iteratei : 'a list -> ('b -> bool) -> ('a -> 'b -> 'b) -> 'b -> 'b
  val lss_union_list :
    'a HOL.equal * 'a Orderings.linorder -> ('a list) list -> 'a list
end = struct

fun lss_ins (A1_, A2_) x l =
  ListAdd.insertion_sort
    (A1_, (Orderings.ord_preorder o Orderings.preorder_order o
            Orderings.order_linorder)
            A2_)
    x l;

fun lss_sng x = [x];

fun lss_empty x = (fn _ => []) x;

fun lss_iteratei l = ListIterator.foldli l;

fun lss_union_list (A1_, A2_) l = ListAdd.merge_list (A1_, A2_) [] l;

end; (*struct ListSetImpl_Sorted*)

structure RBT_LTS_DLTS_LTSImpl : sig
  val rs_dlts_lts_copy :
    'a Orderings.linorder -> 'b Orderings.linorder ->
      ('a, ('b, 'a) RBT.rbt) RBT.rbt ->
        ('a, ('b, ('a, unit) RBT.rbt) RBT.rbt) RBT.rbt
  val rs_lts_dlts_succ :
    'a Orderings.linorder -> 'b Orderings.linorder ->
      (('a, ('b, ('a, unit) RBT.rbt) RBT.rbt) RBT.rbt,
        ('a, ('b, 'a) RBT.rbt) RBT.rbt)
        LTSByLTS_DLTS.lTS_choice ->
        'a -> 'b -> 'a option
  val rs_lts_dlts_empty_lts :
    'a Orderings.linorder -> 'b Orderings.linorder -> 'c Orderings.linorder ->
      unit ->
        (('a, ('b, ('c, unit) RBT.rbt) RBT.rbt) RBT.rbt, 'd)
          LTSByLTS_DLTS.lTS_choice
  val rs_lts_dlts_filter_it :
    'a Orderings.linorder -> 'b Orderings.linorder ->
      ('a -> bool) ->
        ('b -> bool) ->
          ('a -> bool) ->
            ('a * ('b * 'a) -> bool) ->
              (('a, ('b, ('a, unit) RBT.rbt) RBT.rbt) RBT.rbt,
                ('a, ('b, 'a) RBT.rbt) RBT.rbt)
                LTSByLTS_DLTS.lTS_choice ->
                ('c -> bool) -> ('a * ('b * 'a) -> 'c -> 'c) -> 'c -> 'c
  val rs_lts_dlts_add_simple :
    'a Orderings.linorder -> 'b Orderings.linorder ->
      'a -> 'b -> 'a -> (('a, ('b, ('a, unit) RBT.rbt) RBT.rbt) RBT.rbt,
                          ('a, ('b, 'a) RBT.rbt) RBT.rbt)
                          LTSByLTS_DLTS.lTS_choice ->
                          (('a, ('b, ('a, unit) RBT.rbt) RBT.rbt) RBT.rbt,
                            ('a, ('b, 'a) RBT.rbt) RBT.rbt)
                            LTSByLTS_DLTS.lTS_choice
  val rs_lts_dlts_image_filter :
    'a Orderings.linorder -> 'b Orderings.linorder -> 'c Orderings.linorder ->
      'd Orderings.linorder ->
      ('a * ('b * 'a) -> 'c * ('d * 'c)) ->
        ('a -> bool) ->
          ('b -> bool) ->
            ('a -> bool) ->
              ('a * ('b * 'a) -> bool) ->
                (('a, ('b, ('a, unit) RBT.rbt) RBT.rbt) RBT.rbt,
                  ('a, ('b, 'a) RBT.rbt) RBT.rbt)
                  LTSByLTS_DLTS.lTS_choice ->
                  (('c, ('d, ('c, unit) RBT.rbt) RBT.rbt) RBT.rbt,
                    ('c, ('d, 'c) RBT.rbt) RBT.rbt)
                    LTSByLTS_DLTS.lTS_choice
  val rs_lts_dlts_image :
    'a Orderings.linorder -> 'b Orderings.linorder -> 'c Orderings.linorder ->
      'd Orderings.linorder ->
      ('a * ('b * 'a) -> 'c * ('d * 'c)) ->
        (('a, ('b, ('a, unit) RBT.rbt) RBT.rbt) RBT.rbt,
          ('a, ('b, 'a) RBT.rbt) RBT.rbt)
          LTSByLTS_DLTS.lTS_choice ->
          (('c, ('d, ('c, unit) RBT.rbt) RBT.rbt) RBT.rbt,
            ('c, ('d, 'c) RBT.rbt) RBT.rbt)
            LTSByLTS_DLTS.lTS_choice
  val rs_lts_dlts_reverse :
    'a Orderings.linorder -> 'b Orderings.linorder ->
      (('a, ('b, ('a, unit) RBT.rbt) RBT.rbt) RBT.rbt,
        ('a, ('b, 'a) RBT.rbt) RBT.rbt)
        LTSByLTS_DLTS.lTS_choice ->
        (('a, ('b, ('a, unit) RBT.rbt) RBT.rbt) RBT.rbt,
          ('a, ('b, 'a) RBT.rbt) RBT.rbt)
          LTSByLTS_DLTS.lTS_choice
  val rs_lts_dlts_to_list :
    'a Orderings.linorder -> 'b Orderings.linorder ->
      (('a, ('b, ('a, unit) RBT.rbt) RBT.rbt) RBT.rbt,
        ('a, ('b, 'a) RBT.rbt) RBT.rbt)
        LTSByLTS_DLTS.lTS_choice ->
        ('a * ('b * 'a)) list
  val rs_lts_dlts_add_dlts :
    'a Orderings.linorder -> 'b Orderings.linorder ->
      'a -> 'b -> 'a -> (('a, ('b, ('a, unit) RBT.rbt) RBT.rbt) RBT.rbt,
                          ('a, ('b, 'a) RBT.rbt) RBT.rbt)
                          LTSByLTS_DLTS.lTS_choice ->
                          (('a, ('b, ('a, unit) RBT.rbt) RBT.rbt) RBT.rbt,
                            ('a, ('b, 'a) RBT.rbt) RBT.rbt)
                            LTSByLTS_DLTS.lTS_choice
  val rs_lts_dlts_add_choice :
    'a Orderings.linorder -> 'b Orderings.linorder ->
      bool ->
        'a -> 'b -> 'a -> (('a, ('b, ('a, unit) RBT.rbt) RBT.rbt) RBT.rbt,
                            ('a, ('b, 'a) RBT.rbt) RBT.rbt)
                            LTSByLTS_DLTS.lTS_choice ->
                            (('a, ('b, ('a, unit) RBT.rbt) RBT.rbt) RBT.rbt,
                              ('a, ('b, 'a) RBT.rbt) RBT.rbt)
                              LTSByLTS_DLTS.lTS_choice
  val rs_lts_dlts_empty_dlts :
    'b Orderings.linorder -> 'c Orderings.linorder ->
      unit -> ('a, ('b, ('c, 'b) RBT.rbt) RBT.rbt) LTSByLTS_DLTS.lTS_choice
  val rs_lts_dlts_image_filter_dlts :
    'a Orderings.linorder -> 'b Orderings.linorder -> 'c Orderings.linorder ->
      'd Orderings.linorder ->
      ('a * ('b * 'a) -> 'c * ('d * 'c)) ->
        ('a -> bool) ->
          ('b -> bool) ->
            ('a -> bool) ->
              ('a * ('b * 'a) -> bool) ->
                (('a, ('b, ('a, unit) RBT.rbt) RBT.rbt) RBT.rbt,
                  ('a, ('b, 'a) RBT.rbt) RBT.rbt)
                  LTSByLTS_DLTS.lTS_choice ->
                  (('c, ('d, ('c, unit) RBT.rbt) RBT.rbt) RBT.rbt,
                    ('c, ('d, 'c) RBT.rbt) RBT.rbt)
                    LTSByLTS_DLTS.lTS_choice
  val rs_lts_dlts_image_dlts :
    'a Orderings.linorder -> 'b Orderings.linorder -> 'c Orderings.linorder ->
      'd Orderings.linorder ->
      ('a * ('b * 'a) -> 'c * ('d * 'c)) ->
        (('a, ('b, ('a, unit) RBT.rbt) RBT.rbt) RBT.rbt,
          ('a, ('b, 'a) RBT.rbt) RBT.rbt)
          LTSByLTS_DLTS.lTS_choice ->
          (('c, ('d, ('c, unit) RBT.rbt) RBT.rbt) RBT.rbt,
            ('c, ('d, 'c) RBT.rbt) RBT.rbt)
            LTSByLTS_DLTS.lTS_choice
  val rs_lts_dlts_to_collect_list :
    'a HOL.equal * 'a Orderings.linorder -> 'b Orderings.linorder ->
      (('a, ('b, ('a, unit) RBT.rbt) RBT.rbt) RBT.rbt,
        ('a, ('b, 'a) RBT.rbt) RBT.rbt)
        LTSByLTS_DLTS.lTS_choice ->
        ('a * ('b list * 'a)) list
end = struct

fun rs_dlts_lts_copy A_ B_ =
  LTSGA.ltsga_copy (RBT_DLTSImpl.rs_dlts_it A_ B_)
    (RBT_TripleSetImpl.rs_ts_empty A_ B_ A_)
    (RBT_TripleSetImpl.rs_ts_add A_ B_ A_);

fun rs_lts_dlts_succ A_ B_ =
  (fn l => fn q => fn a =>
    (case l of LTSByLTS_DLTS.Lts aa => RBT_TripleSetImpl.rs_ts_C_it A_ B_ A_ aa
      | LTSByLTS_DLTS.Dlts aa => RBT_DLTSImpl.rs_dlts_succ_it A_ B_ aa)
      q
      a
      Option.is_none
      (fn x => fn _ => SOME x)
      NONE);

fun rs_lts_dlts_empty_lts A_ B_ C_ =
  LTSByLTS_DLTS.ltsbc_emp_lts (RBT_TripleSetImpl.rs_ts_empty A_ B_ C_);

fun rs_lts_dlts_filter_it A_ B_ =
  LTSByLTS_DLTS.ltsbc_filter_it (RBT_TripleSetImpl.rs_ts_filter_it A_ B_ A_)
    (RBT_DLTSImpl.rs_dlts_filter_it A_ B_);

fun rs_lts_dlts_add_simple A_ B_ =
  LTSByLTS_DLTS.ltsbc_add_simple (RBT_TripleSetImpl.rs_ts_add A_ B_ A_)
    (rs_dlts_lts_copy A_ B_);

fun rs_lts_dlts_image_filter A_ B_ C_ D_ =
  (fn f => fn p_v1 => fn p_e => fn p_v2 => fn p => fn l =>
    rs_lts_dlts_filter_it A_ B_ p_v1 p_e p_v2 p l (fn _ => true)
      (fn vev => fn la =>
        let
          val (v, (e, va)) = f vev;
        in
          rs_lts_dlts_add_simple C_ D_ v e va la
        end)
      (rs_lts_dlts_empty_lts C_ D_ C_ ()));

fun rs_lts_dlts_image A_ B_ C_ D_ =
  (fn f =>
    rs_lts_dlts_image_filter A_ B_ C_ D_ f (fn _ => true) (fn _ => true)
      (fn _ => true) (fn _ => true));

fun rs_lts_dlts_reverse A_ B_ =
  (fn l =>
    (case l of LTSByLTS_DLTS.Lts a => RBT_TripleSetImpl.rs_ts_it A_ B_ A_ a
      | LTSByLTS_DLTS.Dlts a => RBT_DLTSImpl.rs_dlts_it A_ B_ a)
      (fn _ => true)
      (fn (v, (e, va)) => rs_lts_dlts_add_simple A_ B_ va e v)
      (rs_lts_dlts_empty_lts A_ B_ A_ ()));

fun rs_lts_dlts_to_list A_ B_ =
  (fn x =>
    (case x of LTSByLTS_DLTS.Lts a => RBT_TripleSetImpl.rs_ts_it A_ B_ A_ a
      | LTSByLTS_DLTS.Dlts a => RBT_DLTSImpl.rs_dlts_it A_ B_ a)
      (fn _ => true)
      (fn a => fn b => a :: b)
      []);

fun rs_lts_dlts_add_dlts A_ B_ =
  LTSByLTS_DLTS.dltsbc_add (RBT_TripleSetImpl.rs_ts_add A_ B_ A_)
    (RBT_DLTSImpl.rs_dlts_add A_ B_);

fun rs_lts_dlts_add_choice A_ B_ b =
  (if b then rs_lts_dlts_add_dlts A_ B_ else rs_lts_dlts_add_simple A_ B_);

fun rs_lts_dlts_empty_dlts B_ C_ =
  LTSByLTS_DLTS.ltsbc_emp_dlts (RBT_DLTSImpl.rs_dlts_empty B_ C_);

fun rs_lts_dlts_image_filter_dlts A_ B_ C_ D_ =
  (fn f => fn p_v1 => fn p_e => fn p_v2 => fn p => fn l =>
    rs_lts_dlts_filter_it A_ B_ p_v1 p_e p_v2 p l (fn _ => true)
      (fn vev => fn la =>
        let
          val (v, (e, va)) = f vev;
        in
          rs_lts_dlts_add_dlts C_ D_ v e va la
        end)
      (rs_lts_dlts_empty_dlts C_ D_ ()));

fun rs_lts_dlts_image_dlts A_ B_ C_ D_ =
  (fn f =>
    rs_lts_dlts_image_filter_dlts A_ B_ C_ D_ f (fn _ => true) (fn _ => true)
      (fn _ => true) (fn _ => true));

fun rs_lts_dlts_to_collect_list (A1_, A2_) B_ =
  (fn l =>
    LTSGA.ltsga_list_to_collect_list A1_ A1_ (rs_lts_dlts_to_list A2_ B_ l));

end; (*struct RBT_LTS_DLTS_LTSImpl*)

structure Hopcroft_Minimisation : sig
  val l_insert_impl :
    IntInf.int -> ('a * IntInf.int) list -> 'a list -> ('a * IntInf.int) list
end = struct

fun l_insert_impl i l [] = l
  | l_insert_impl i l (a :: asa) = l_insert_impl i ((a, i) :: l) asa;

end; (*struct Hopcroft_Minimisation*)

structure RBT_NFAImpl : sig
  val rs_nfa_props :
    'a Orderings.linorder ->
      (IntInf.int, unit) RBT.rbt *
        (('a, unit) RBT.rbt *
          (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
             (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
             LTSByLTS_DLTS.lTS_choice *
            ((IntInf.int, unit) RBT.rbt *
              ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
        unit NFAByLTS.nfa_props_ext
  val rs_nfa_accept_dfa :
    'a Orderings.linorder ->
      (IntInf.int, unit) RBT.rbt *
        (('a, unit) RBT.rbt *
          (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
             (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
             LTSByLTS_DLTS.lTS_choice *
            ((IntInf.int, unit) RBT.rbt *
              ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
        'a list -> bool
  val rs_nfa_accept_nfa :
    'a Orderings.linorder ->
      (IntInf.int, unit) RBT.rbt *
        (('a, unit) RBT.rbt *
          (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
             (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
             LTSByLTS_DLTS.lTS_choice *
            ((IntInf.int, unit) RBT.rbt *
              ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
        'a list -> bool
  val rs_nfa_accept :
    'a Orderings.linorder ->
      (IntInf.int, unit) RBT.rbt *
        (('a, unit) RBT.rbt *
          (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
             (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
             LTSByLTS_DLTS.lTS_choice *
            ((IntInf.int, unit) RBT.rbt *
              ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
        'a list -> bool
  val rs_hop_lts_add :
    'a Orderings.linorder ->
      IntInf.int ->
        'a -> IntInf.int ->
                ('a, (((IntInf.int list) option)
                       FArray.IsabelleMapping.ArrayType))
                  RBT.rbt ->
                  ('a, (((IntInf.int list) option)
                         FArray.IsabelleMapping.ArrayType))
                    RBT.rbt
  val rs_nfa_reverse :
    'a Orderings.linorder ->
      (IntInf.int, unit) RBT.rbt *
        (('a, unit) RBT.rbt *
          (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
             (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
             LTSByLTS_DLTS.lTS_choice *
            ((IntInf.int, unit) RBT.rbt *
              ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
        (IntInf.int, unit) RBT.rbt *
          (('a, unit) RBT.rbt *
            (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
               (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
               LTSByLTS_DLTS.lTS_choice *
              ((IntInf.int, unit) RBT.rbt *
                ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext))))
  val rs_hop_lts_copy : 'a -> 'a
  val rs_nfa_sm_update :
    IntInf.int ->
      (IntInf.int option) FArray.IsabelleMapping.ArrayType ->
        (IntInf.int option) FArray.IsabelleMapping.ArrayType ->
          IntInf.int ->
            IntInf.int -> (IntInf.int option) FArray.IsabelleMapping.ArrayType
  val rs_hop_lts_get_succs :
    'a Orderings.linorder ->
      ('a, (((IntInf.int list) option) FArray.IsabelleMapping.ArrayType))
        RBT.rbt ->
        IntInf.int list -> 'a -> IntInf.int list
  val rs_nfa_rename_states_dfa :
    'a Orderings.linorder ->
      (IntInf.int, unit) RBT.rbt *
        (('a, unit) RBT.rbt *
          (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
             (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
             LTSByLTS_DLTS.lTS_choice *
            ((IntInf.int, unit) RBT.rbt *
              ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
        (IntInf.int -> IntInf.int) ->
          (IntInf.int, unit) RBT.rbt *
            (('a, unit) RBT.rbt *
              (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
                 (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
                 LTSByLTS_DLTS.lTS_choice *
                ((IntInf.int, unit) RBT.rbt *
                  ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext))))
  val rs_nfa_Hopcroft :
    'a Orderings.linorder ->
      (IntInf.int, unit) RBT.rbt *
        (('a, unit) RBT.rbt *
          (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
             (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
             LTSByLTS_DLTS.lTS_choice *
            ((IntInf.int, unit) RBT.rbt *
              ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
        (IntInf.int, unit) RBT.rbt *
          (('a, unit) RBT.rbt *
            (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
               (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
               LTSByLTS_DLTS.lTS_choice *
              ((IntInf.int, unit) RBT.rbt *
                ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext))))
  val rs_nfa_destruct :
    'a Orderings.linorder ->
      (IntInf.int, unit) RBT.rbt *
        (('a, unit) RBT.rbt *
          (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
             (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
             LTSByLTS_DLTS.lTS_choice *
            ((IntInf.int, unit) RBT.rbt *
              ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
        IntInf.int list *
          ('a list *
            ((IntInf.int * ('a list * IntInf.int)) list *
              (IntInf.int list * IntInf.int list)))
  val rs_nfa_final_no :
    'a Orderings.linorder ->
      (IntInf.int, unit) RBT.rbt *
        (('a, unit) RBT.rbt *
          (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
             (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
             LTSByLTS_DLTS.lTS_choice *
            ((IntInf.int, unit) RBT.rbt *
              ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
        IntInf.int
  val rs_nfa_trans_no :
    'a Orderings.linorder ->
      (IntInf.int, unit) RBT.rbt *
        (('a, unit) RBT.rbt *
          (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
             (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
             LTSByLTS_DLTS.lTS_choice *
            ((IntInf.int, unit) RBT.rbt *
              ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
        IntInf.int
  val rs_nfa_construct_gen :
    'a Orderings.linorder ->
      unit NFAByLTS.nfa_props_ext ->
        IntInf.int list *
          ('a list *
            ((IntInf.int * ('a list * IntInf.int)) list *
              (IntInf.int list * IntInf.int list))) ->
          (IntInf.int, unit) RBT.rbt *
            (('a, unit) RBT.rbt *
              (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
                 (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
                 LTSByLTS_DLTS.lTS_choice *
                ((IntInf.int, unit) RBT.rbt *
                  ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext))))
  val rs_dfa_construct :
    'a Orderings.linorder ->
      IntInf.int list *
        ('a list *
          ((IntInf.int * ('a list * IntInf.int)) list *
            (IntInf.int list * IntInf.int list))) ->
        (IntInf.int, unit) RBT.rbt *
          (('a, unit) RBT.rbt *
            (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
               (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
               LTSByLTS_DLTS.lTS_choice *
              ((IntInf.int, unit) RBT.rbt *
                ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext))))
  val rs_nfa_dfa_construct_reachable :
    'b Orderings.linorder -> 'c Orderings.linorder ->
      bool ->
        ('a -> 'b) ->
          'a list ->
            ('c, unit) RBT.rbt ->
              ('a -> bool) ->
                ('a ->
                  ((('b, IntInf.int) RBT.rbt * IntInf.int) *
                     (((IntInf.int, ('c, (IntInf.int, unit) RBT.rbt) RBT.rbt)
                         RBT.rbt,
                        (IntInf.int, ('c, IntInf.int) RBT.rbt) RBT.rbt)
                        LTSByLTS_DLTS.lTS_choice *
                       'a list) ->
                    bool) ->
                    ('c * 'a ->
                      (('b, IntInf.int) RBT.rbt * IntInf.int) *
                        (((IntInf.int, ('c, (IntInf.int, unit) RBT.rbt) RBT.rbt)
                            RBT.rbt,
                           (IntInf.int, ('c, IntInf.int) RBT.rbt) RBT.rbt)
                           LTSByLTS_DLTS.lTS_choice *
                          'a list) ->
                        (('b, IntInf.int) RBT.rbt * IntInf.int) *
                          (((IntInf.int,
                              ('c, (IntInf.int, unit) RBT.rbt) RBT.rbt)
                              RBT.rbt,
                             (IntInf.int, ('c, IntInf.int) RBT.rbt) RBT.rbt)
                             LTSByLTS_DLTS.lTS_choice *
                            'a list)) ->
                      (('b, IntInf.int) RBT.rbt * IntInf.int) *
                        (((IntInf.int, ('c, (IntInf.int, unit) RBT.rbt) RBT.rbt)
                            RBT.rbt,
                           (IntInf.int, ('c, IntInf.int) RBT.rbt) RBT.rbt)
                           LTSByLTS_DLTS.lTS_choice *
                          'd list) ->
                        (('b, IntInf.int) RBT.rbt * IntInf.int) *
                          (((IntInf.int,
                              ('c, (IntInf.int, unit) RBT.rbt) RBT.rbt)
                              RBT.rbt,
                             (IntInf.int, ('c, IntInf.int) RBT.rbt) RBT.rbt)
                             LTSByLTS_DLTS.lTS_choice *
                            'a list)) ->
                  (IntInf.int, unit) RBT.rbt *
                    (('c, unit) RBT.rbt *
                      (((IntInf.int, ('c, (IntInf.int, unit) RBT.rbt) RBT.rbt)
                          RBT.rbt,
                         (IntInf.int, ('c, IntInf.int) RBT.rbt) RBT.rbt)
                         LTSByLTS_DLTS.lTS_choice *
                        ((IntInf.int, unit) RBT.rbt *
                          ((IntInf.int, unit) RBT.rbt *
                            unit NFAByLTS.nfa_props_ext))))
  val rs_nfa_bool_comb_gen :
    'a Orderings.linorder ->
      ('a, unit) RBT.rbt ->
        (bool -> bool -> bool) ->
          (IntInf.int, unit) RBT.rbt *
            (('a, unit) RBT.rbt *
              (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
                 (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
                 LTSByLTS_DLTS.lTS_choice *
                ((IntInf.int, unit) RBT.rbt *
                  ((IntInf.int, unit) RBT.rbt *
                    unit NFAByLTS.nfa_props_ext)))) ->
            (IntInf.int, unit) RBT.rbt *
              (('a, unit) RBT.rbt *
                (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt)
                    RBT.rbt,
                   (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
                   LTSByLTS_DLTS.lTS_choice *
                  ((IntInf.int, unit) RBT.rbt *
                    ((IntInf.int, unit) RBT.rbt *
                      unit NFAByLTS.nfa_props_ext)))) ->
              (IntInf.int, unit) RBT.rbt *
                (('a, unit) RBT.rbt *
                  (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt)
                      RBT.rbt,
                     (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
                     LTSByLTS_DLTS.lTS_choice *
                    ((IntInf.int, unit) RBT.rbt *
                      ((IntInf.int, unit) RBT.rbt *
                        unit NFAByLTS.nfa_props_ext))))
  val rs_nfa_bool_comb :
    'a Orderings.linorder ->
      (bool -> bool -> bool) ->
        (IntInf.int, unit) RBT.rbt *
          (('a, unit) RBT.rbt *
            (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
               (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
               LTSByLTS_DLTS.lTS_choice *
              ((IntInf.int, unit) RBT.rbt *
                ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
          (IntInf.int, unit) RBT.rbt *
            (('a, unit) RBT.rbt *
              (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
                 (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
                 LTSByLTS_DLTS.lTS_choice *
                ((IntInf.int, unit) RBT.rbt *
                  ((IntInf.int, unit) RBT.rbt *
                    unit NFAByLTS.nfa_props_ext)))) ->
            (IntInf.int, unit) RBT.rbt *
              (('a, unit) RBT.rbt *
                (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt)
                    RBT.rbt,
                   (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
                   LTSByLTS_DLTS.lTS_choice *
                  ((IntInf.int, unit) RBT.rbt *
                    ((IntInf.int, unit) RBT.rbt *
                      unit NFAByLTS.nfa_props_ext))))
  val rs_nfa_construct :
    'a Orderings.linorder ->
      IntInf.int list *
        ('a list *
          ((IntInf.int * ('a list * IntInf.int)) list *
            (IntInf.int list * IntInf.int list))) ->
        (IntInf.int, unit) RBT.rbt *
          (('a, unit) RBT.rbt *
            (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
               (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
               LTSByLTS_DLTS.lTS_choice *
              ((IntInf.int, unit) RBT.rbt *
                ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext))))
  val rs_nfa_labels_no :
    'a Orderings.linorder ->
      (IntInf.int, unit) RBT.rbt *
        (('a, unit) RBT.rbt *
          (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
             (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
             LTSByLTS_DLTS.lTS_choice *
            ((IntInf.int, unit) RBT.rbt *
              ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
        IntInf.int
  val rs_nfa_normalise :
    'a Orderings.linorder ->
      (IntInf.int, unit) RBT.rbt *
        (('a, unit) RBT.rbt *
          (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
             (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
             LTSByLTS_DLTS.lTS_choice *
            ((IntInf.int, unit) RBT.rbt *
              ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
        (IntInf.int, unit) RBT.rbt *
          (('a, unit) RBT.rbt *
            (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
               (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
               LTSByLTS_DLTS.lTS_choice *
              ((IntInf.int, unit) RBT.rbt *
                ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext))))
  val rs_nfa_states_no :
    'a Orderings.linorder ->
      (IntInf.int, unit) RBT.rbt *
        (('a, unit) RBT.rbt *
          (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
             (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
             LTSByLTS_DLTS.lTS_choice *
            ((IntInf.int, unit) RBT.rbt *
              ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
        IntInf.int
  val rs_dfa_construct_reachable :
    'b Orderings.linorder -> 'c Orderings.linorder ->
      ('a -> 'b) ->
        'a list ->
          ('c, unit) RBT.rbt ->
            ('a -> bool) ->
              ('a ->
                ((('b, IntInf.int) RBT.rbt * IntInf.int) *
                   (((IntInf.int, ('c, (IntInf.int, unit) RBT.rbt) RBT.rbt)
                       RBT.rbt,
                      (IntInf.int, ('c, IntInf.int) RBT.rbt) RBT.rbt)
                      LTSByLTS_DLTS.lTS_choice *
                     'a list) ->
                  bool) ->
                  ('c * 'a ->
                    (('b, IntInf.int) RBT.rbt * IntInf.int) *
                      (((IntInf.int, ('c, (IntInf.int, unit) RBT.rbt) RBT.rbt)
                          RBT.rbt,
                         (IntInf.int, ('c, IntInf.int) RBT.rbt) RBT.rbt)
                         LTSByLTS_DLTS.lTS_choice *
                        'a list) ->
                      (('b, IntInf.int) RBT.rbt * IntInf.int) *
                        (((IntInf.int, ('c, (IntInf.int, unit) RBT.rbt) RBT.rbt)
                            RBT.rbt,
                           (IntInf.int, ('c, IntInf.int) RBT.rbt) RBT.rbt)
                           LTSByLTS_DLTS.lTS_choice *
                          'a list)) ->
                    (('b, IntInf.int) RBT.rbt * IntInf.int) *
                      (((IntInf.int, ('c, (IntInf.int, unit) RBT.rbt) RBT.rbt)
                          RBT.rbt,
                         (IntInf.int, ('c, IntInf.int) RBT.rbt) RBT.rbt)
                         LTSByLTS_DLTS.lTS_choice *
                        'd list) ->
                      (('b, IntInf.int) RBT.rbt * IntInf.int) *
                        (((IntInf.int, ('c, (IntInf.int, unit) RBT.rbt) RBT.rbt)
                            RBT.rbt,
                           (IntInf.int, ('c, IntInf.int) RBT.rbt) RBT.rbt)
                           LTSByLTS_DLTS.lTS_choice *
                          'a list)) ->
                (IntInf.int, unit) RBT.rbt *
                  (('c, unit) RBT.rbt *
                    (((IntInf.int, ('c, (IntInf.int, unit) RBT.rbt) RBT.rbt)
                        RBT.rbt,
                       (IntInf.int, ('c, IntInf.int) RBT.rbt) RBT.rbt)
                       LTSByLTS_DLTS.lTS_choice *
                      ((IntInf.int, unit) RBT.rbt *
                        ((IntInf.int, unit) RBT.rbt *
                          unit NFAByLTS.nfa_props_ext))))
  val rs_nfa_determinise :
    'a Orderings.linorder ->
      (IntInf.int, unit) RBT.rbt *
        (('a, unit) RBT.rbt *
          (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
             (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
             LTSByLTS_DLTS.lTS_choice *
            ((IntInf.int, unit) RBT.rbt *
              ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
        (IntInf.int, unit) RBT.rbt *
          (('a, unit) RBT.rbt *
            (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
               (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
               LTSByLTS_DLTS.lTS_choice *
              ((IntInf.int, unit) RBT.rbt *
                ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext))))
  val rs_nfa_Brzozowski :
    'a Orderings.linorder ->
      (IntInf.int, unit) RBT.rbt *
        (('a, unit) RBT.rbt *
          (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
             (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
             LTSByLTS_DLTS.lTS_choice *
            ((IntInf.int, unit) RBT.rbt *
              ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
        (IntInf.int, unit) RBT.rbt *
          (('a, unit) RBT.rbt *
            (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
               (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
               LTSByLTS_DLTS.lTS_choice *
              ((IntInf.int, unit) RBT.rbt *
                ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext))))
  val rs_nfa_complement :
    'a Orderings.linorder ->
      (IntInf.int, unit) RBT.rbt *
        (('a, unit) RBT.rbt *
          (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
             (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
             LTSByLTS_DLTS.lTS_choice *
            ((IntInf.int, unit) RBT.rbt *
              ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
        (IntInf.int, unit) RBT.rbt *
          (('a, unit) RBT.rbt *
            (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
               (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
               LTSByLTS_DLTS.lTS_choice *
              ((IntInf.int, unit) RBT.rbt *
                ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext))))
  val rs_nfa_initial_no :
    'a Orderings.linorder ->
      (IntInf.int, unit) RBT.rbt *
        (('a, unit) RBT.rbt *
          (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
             (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
             LTSByLTS_DLTS.lTS_choice *
            ((IntInf.int, unit) RBT.rbt *
              ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
        IntInf.int
  val rs_nfa_Hopcroft_NFA :
    'a Orderings.linorder ->
      (IntInf.int, unit) RBT.rbt *
        (('a, unit) RBT.rbt *
          (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
             (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
             LTSByLTS_DLTS.lTS_choice *
            ((IntInf.int, unit) RBT.rbt *
              ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
        (IntInf.int, unit) RBT.rbt *
          (('a, unit) RBT.rbt *
            (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
               (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
               LTSByLTS_DLTS.lTS_choice *
              ((IntInf.int, unit) RBT.rbt *
                ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext))))
  val rs_nfa_rename_labels_gen :
    'a Orderings.linorder ->
      bool ->
        (IntInf.int, unit) RBT.rbt *
          (('a, unit) RBT.rbt *
            (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
               (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
               LTSByLTS_DLTS.lTS_choice *
              ((IntInf.int, unit) RBT.rbt *
                ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
          'b -> ('a -> 'a) ->
                  (IntInf.int, unit) RBT.rbt *
                    ('b * (((IntInf.int,
                              ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt)
                              RBT.rbt,
                             (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
                             LTSByLTS_DLTS.lTS_choice *
                            ((IntInf.int, unit) RBT.rbt *
                              ((IntInf.int, unit) RBT.rbt *
                                unit NFAByLTS.nfa_props_ext))))
  val rs_nfa_rename_labels :
    'a Orderings.linorder ->
      bool ->
        (IntInf.int, unit) RBT.rbt *
          (('a, unit) RBT.rbt *
            (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
               (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
               LTSByLTS_DLTS.lTS_choice *
              ((IntInf.int, unit) RBT.rbt *
                ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
          ('a -> 'a) ->
            (IntInf.int, unit) RBT.rbt *
              (('a, unit) RBT.rbt *
                (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt)
                    RBT.rbt,
                   (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
                   LTSByLTS_DLTS.lTS_choice *
                  ((IntInf.int, unit) RBT.rbt *
                    ((IntInf.int, unit) RBT.rbt *
                      unit NFAByLTS.nfa_props_ext))))
  val rs_nfa_language_is_empty :
    'a Orderings.linorder ->
      (IntInf.int, unit) RBT.rbt *
        (('a, unit) RBT.rbt *
          (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
             (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
             LTSByLTS_DLTS.lTS_choice *
            ((IntInf.int, unit) RBT.rbt *
              ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
        bool
  val rs_nfa_language_is_subset :
    'a Orderings.linorder ->
      (IntInf.int, unit) RBT.rbt *
        (('a, unit) RBT.rbt *
          (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
             (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
             LTSByLTS_DLTS.lTS_choice *
            ((IntInf.int, unit) RBT.rbt *
              ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
        (IntInf.int, unit) RBT.rbt *
          (('a, unit) RBT.rbt *
            (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
               (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
               LTSByLTS_DLTS.lTS_choice *
              ((IntInf.int, unit) RBT.rbt *
                ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
          bool
  val rs_nfa_language_is_eq :
    'a Orderings.linorder ->
      (IntInf.int, unit) RBT.rbt *
        (('a, unit) RBT.rbt *
          (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
             (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
             LTSByLTS_DLTS.lTS_choice *
            ((IntInf.int, unit) RBT.rbt *
              ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
        (IntInf.int, unit) RBT.rbt *
          (('a, unit) RBT.rbt *
            (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
               (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
               LTSByLTS_DLTS.lTS_choice *
              ((IntInf.int, unit) RBT.rbt *
                ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
          bool
  val rs_nfa_destruct_simple :
    'a Orderings.linorder ->
      (IntInf.int, unit) RBT.rbt *
        (('a, unit) RBT.rbt *
          (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
             (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
             LTSByLTS_DLTS.lTS_choice *
            ((IntInf.int, unit) RBT.rbt *
              ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
        IntInf.int list *
          ('a list *
            ((IntInf.int * ('a * IntInf.int)) list *
              (IntInf.int list * IntInf.int list)))
  val rs_nfa_language_is_univ :
    'a Orderings.linorder ->
      (IntInf.int, unit) RBT.rbt *
        (('a, unit) RBT.rbt *
          (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
             (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
             LTSByLTS_DLTS.lTS_choice *
            ((IntInf.int, unit) RBT.rbt *
              ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
        bool
  val rs_nfa_right_quotient_lists :
    'a Orderings.linorder ->
      ('a -> bool) ->
        (IntInf.int, unit) RBT.rbt *
          (('a, unit) RBT.rbt *
            (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
               (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
               LTSByLTS_DLTS.lTS_choice *
              ((IntInf.int, unit) RBT.rbt *
                ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
          (IntInf.int, unit) RBT.rbt *
            (('a, unit) RBT.rbt *
              (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
                 (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
                 LTSByLTS_DLTS.lTS_choice *
                ((IntInf.int, unit) RBT.rbt *
                  ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext))))
end = struct

fun rs_nfa_props A_ (q, (a, (d, (i, (f, flags))))) = flags;

fun rs_nfa_accept_dfa A_ =
  (fn (_, (_, (d, (i, (f, _))))) => fn w =>
    StdInst.rs_bexists Arith.linorder_nat i
      (fn q =>
        (case ListIterator.foldli w (fn q_opt => not (Option.is_none q_opt))
                (fn sigma => fn q_opt =>
                  RBT_LTS_DLTS_LTSImpl.rs_lts_dlts_succ Arith.linorder_nat A_ d
                    (Option.the q_opt) sigma)
                (SOME q)
          of NONE => false
          | SOME qa => RBTSetImpl.rs_memb Arith.linorder_nat qa f)));

fun rs_nfa_accept_nfa A_ =
  (fn (_, (_, (d, (i, (f, _))))) => fn w =>
    not (StdInst.rr_disjoint Arith.linorder_nat
          (List.foldl
            (fn q => fn a =>
              RBTSetImpl.rs_iteratei Arith.linorder_nat q (fn _ => true)
                (fn qa =>
                  (case d
                    of LTSByLTS_DLTS.Lts aa =>
                      RBT_TripleSetImpl.rs_ts_C_it Arith.linorder_nat A_
                        Arith.linorder_nat aa
                    | LTSByLTS_DLTS.Dlts aa =>
                      RBT_DLTSImpl.rs_dlts_succ_it Arith.linorder_nat A_ aa)
                    qa
                    a
                    (fn _ => true)
                    (RBTSetImpl.rs_ins Arith.linorder_nat))
                (RBTSetImpl.rs_empty Arith.linorder_nat ()))
            i w)
          f));

fun rs_nfa_accept A_ a w =
  (if NFAByLTS.nfa_prop_is_complete_deterministic (rs_nfa_props A_ a)
    then rs_nfa_accept_dfa A_ a w else rs_nfa_accept_nfa A_ a w);

fun rs_hop_lts_add A_ =
  (fn q => fn a => fn v => fn l =>
    (case RBTMapImpl.rm_lookup A_ a l
      of NONE =>
        RBTMapImpl.rm_update A_ a
          (StdInst.iam_sng q (ListSetImpl_Sorted.lss_sng v)) l
      | SOME m2 =>
        (case ArrayMapImpl.iam_lookup q m2
          of NONE =>
            RBTMapImpl.rm_update A_ a
              (ArrayMapImpl.iam_update q (ListSetImpl_Sorted.lss_sng v) m2) l
          | SOME s3 =>
            RBTMapImpl.rm_update A_ a
              (ArrayMapImpl.iam_update q
                (ListSetImpl_Sorted.lss_ins
                  (Arith.equal_nat, Arith.linorder_nat) v s3)
                m2)
              l)));

fun rs_nfa_reverse A_ =
  (fn (q, (a, (d, (i, (f, _))))) =>
    (q, (a, (RBT_LTS_DLTS_LTSImpl.rs_lts_dlts_reverse Arith.linorder_nat A_ d,
              (f, (i, NFAByLTS.fields false false))))));

fun rs_hop_lts_copy x = Fun.id x;

fun rs_nfa_sm_update n sm pim p v =
  (if ((v : IntInf.int) = 0) then sm
    else let
           val q = Option.the (ArrayMapImpl.iam_lookup p pim);
           val sma = ArrayMapImpl.iam_update q n sm;
         in
           rs_nfa_sm_update n sma pim (Arith.suc p)
             (IntInf.max (0, IntInf.-
               (Arith.suc (IntInf.max (0, IntInf.- (v, (1 : IntInf.int)))),
               (1 : IntInf.int))))
         end);

fun rs_hop_lts_get_succs A_ =
  (fn l => fn vs => fn a =>
    (case RBTMapImpl.rm_lookup A_ a l of NONE => ListSetImpl_Sorted.lss_empty ()
      | SOME im =>
        ListSetImpl_Sorted.lss_union_list (Arith.equal_nat, Arith.linorder_nat)
          (List.map_filter (fn q => ArrayMapImpl.iam_lookup q im) vs)));

fun rs_nfa_rename_states_dfa A_ (q, (a, (d, (i, (fa, p))))) f =
  (RBTSetImpl.rs_image Arith.linorder_nat Arith.linorder_nat f q,
    (a, (RBT_LTS_DLTS_LTSImpl.rs_lts_dlts_image_dlts Arith.linorder_nat A_
           Arith.linorder_nat A_ (fn (qb, (aa, qa)) => (f qb, (aa, f qa))) d,
          (RBTSetImpl.rs_image Arith.linorder_nat Arith.linorder_nat f i,
            (RBTSetImpl.rs_image Arith.linorder_nat Arith.linorder_nat f fa,
              NFAByLTS.fields true
                (NFAByLTS.nfa_prop_is_initially_connected p))))));

fun rs_nfa_Hopcroft A_ (q, (a, (d, (i, (f, p))))) =
  let
    val al = RBTSetImpl.rs_to_list A_ a;
    val rv_D =
      rs_hop_lts_copy
        (LTSGA.ltsga_reverse (RBTMapImpl.rm_empty A_) (rs_hop_lts_add A_)
          (fn aa =>
            (case aa
              of LTSByLTS_DLTS.Lts ab =>
                RBT_TripleSetImpl.rs_ts_it Arith.linorder_nat A_
                  Arith.linorder_nat ab
              | LTSByLTS_DLTS.Dlts ab =>
                RBT_DLTSImpl.rs_dlts_it Arith.linorder_nat A_ ab))
          d);
    val pre_fun =
      (fn aa => fn pim => fn (l, u) =>
        rs_hop_lts_get_succs A_ rv_D
          (NFAByLTS.hopcroft_class_map_alpha_impl
            (fn ia => ArrayMapImpl.iam_lookup ia pim) l u)
          aa);
    val (b, c) =
      let
        val x = StdInst.rr_diff Arith.linorder_nat q f;
        val (xa, (xb, xc)) =
          let
            val cm0 = (ArrayMapImpl.iam_empty (), ArrayMapImpl.iam_empty ());
            val (cm1, s) =
              RBTSetImpl.rs_iteratei Arith.linorder_nat x (fn _ => true)
                (fn qa => fn (aa, b) =>
                  let
                    val (pm, pim) = aa;
                  in
                    (fn ia =>
                      ((ArrayMapImpl.iam_update qa ia pm,
                         ArrayMapImpl.iam_update ia qa pim),
                        Arith.suc ia))
                  end
                    b)
                (cm0, 0);
            val (cm2, m) =
              RBTSetImpl.rs_iteratei Arith.linorder_nat f (fn _ => true)
                (fn qa => fn (aa, b) =>
                  let
                    val (pm, pim) = aa;
                  in
                    (fn ia =>
                      ((ArrayMapImpl.iam_update qa ia pm,
                         ArrayMapImpl.iam_update ia qa pim),
                        Arith.suc ia))
                  end
                    b)
                (cm1, s);
          in
            (cm2, (s, m))
          end;
      in
        (if ((xc : IntInf.int) = 0)
          then ((ArrayMapImpl.iam_empty (), (ArrayMapImpl.iam_empty (), 0)), xa)
          else (if ((xc : IntInf.int) = xb)
                 then ((StdInst.iam_sng 0
                          (0, IntInf.max (0, IntInf.- (xc, (1 : IntInf.int)))),
                         (RBTSetImpl.rs_iteratei Arith.linorder_nat q
                            (fn _ => true)
                            (fn qa => ArrayMapImpl.iam_update qa 0)
                            (ArrayMapImpl.iam_empty ()),
                           (1 : IntInf.int))),
                        xa)
                 else let
                        val ((xd, _), xaa) =
                          While_Combinator.whilea
                            (fn pl =>
                              not (List.null
                                    (Product_Type.snd (Product_Type.fst pl))))
                            (fn ((xd, xba), (xaa, xca)) =>
                              let
                                val (xda, xe) = List.hd xba;
                                val xdb =
                                  pre_fun xda xca
                                    (Option.the
                                      (ArrayMapImpl.iam_lookup xe
(Product_Type.fst xd)));
                                val (aa, b) =
                                  let
                                    val (xab, xcb) =
                                      let
val (im, (sm, _)) = xd;
                                      in
ListSetImpl_Sorted.lss_iteratei xdb (fn _ => true)
  (fn qa => fn (iM, (pm, pim)) =>
    let
      val ia = Option.the (ArrayMapImpl.iam_lookup qa sm);
      val s =
        (case RBTMapImpl.rm_lookup Arith.linorder_nat ia iM
          of NONE =>
            let
              val (l, _) = Option.the (ArrayMapImpl.iam_lookup ia im);
            in
              l
            end
          | SOME s => s);
      val iq = Option.the (ArrayMapImpl.iam_lookup qa pm);
      val iMa = RBTMapImpl.rm_update Arith.linorder_nat ia (Arith.suc s) iM;
    in
      (if ((iq : IntInf.int) = s) then (iMa, (pm, pim))
        else let
               val qs = Option.the (ArrayMapImpl.iam_lookup s pim);
               val pima =
                 ArrayMapImpl.iam_update qs iq
                   (ArrayMapImpl.iam_update qa s pm);
               val pma =
                 ArrayMapImpl.iam_update iq qs
                   (ArrayMapImpl.iam_update s qa pim);
             in
               (iMa, (pima, pma))
             end)
    end)
  (RBTMapImpl.rm_empty Arith.linorder_nat (), (xaa, xca))
                                      end;
                                    val (xf, xac) =
                                      RBTMapImpl.rm_iteratei Arith.linorder_nat
xab (fn _ => true)
(fn xf => fn xac =>
  let
    val (xg, xbb) = xf;
    val ((xad, (xea, xfa)), xdc) = xac;
    val (xbc, (xga, (xh, xi))) =
      let
        val (l, u) = Option.the (ArrayMapImpl.iam_lookup xg xad);
      in
        ((l, IntInf.max (0, IntInf.- (xbb, (1 : IntInf.int)))),
          (IntInf.max (0, IntInf.- (xbb, l)),
            ((xbb, u), IntInf.max (0, IntInf.- (Arith.suc u, xbb)))))
      end;
  in
    (if ((xi : IntInf.int) = 0) then ((xad, (xea, xfa)), xdc)
      else let
             val (xbd, xgb) =
               (if IntInf.< (xi, xga) then (xh, xbc) else (xbc, xh));
             val xdd = Hopcroft_Minimisation.l_insert_impl xfa xdc al;
             val xj =
               let
                 val (_, pim) = xcb;
                 val (im, (sm, n)) = (xad, (xea, xfa));
                 val (pmin_l, pmin_u) = xbd;
               in
                 (ArrayMapImpl.iam_update n (pmin_l, pmin_u)
                    (ArrayMapImpl.iam_update xg xgb im),
                   (rs_nfa_sm_update n sm pim pmin_l
                      (IntInf.max (0, IntInf.- (Arith.suc pmin_u, pmin_l))),
                     Arith.suc n))
               end;
           in
             (xj, xdd)
           end)
  end)
(xd, List.tl xba);
                                  in
                                    ((xf, xac), xcb)
                                  end;
                              in
                                (aa, b)
                              end)
                            (((if ((xb : IntInf.int) = 0)
                                then (StdInst.iam_sng 0
(0, IntInf.max (0, IntInf.- (xc, (1 : IntInf.int)))),
                                       (RBTSetImpl.rs_iteratei
  Arith.linorder_nat q (fn _ => true) (fn qa => ArrayMapImpl.iam_update qa 0)
  (ArrayMapImpl.iam_empty ()),
 (1 : IntInf.int)))
                                else (ArrayMapImpl.iam_update (Arith.suc 0)
(0, IntInf.max (0, IntInf.- (xb, (1 : IntInf.int))))
(StdInst.iam_sng 0 (xb, IntInf.max (0, IntInf.- (xc, (1 : IntInf.int))))),
                                       (RBTSetImpl.rs_iteratei
  Arith.linorder_nat f (fn _ => true) (fn qa => ArrayMapImpl.iam_update qa 0)
  (RBTSetImpl.rs_iteratei Arith.linorder_nat x (fn _ => true)
    (fn qa => ArrayMapImpl.iam_update qa (1 : IntInf.int))
    (ArrayMapImpl.iam_empty ())),
 (2 : IntInf.int)))),
                               Hopcroft_Minimisation.l_insert_impl 0 [] al),
                              xa);
                      in
                        (xd, xaa)
                      end))
      end;
  in
    let
      val (_, (sm, _)) = b;
    in
      (fn _ =>
        rs_nfa_rename_states_dfa A_ (q, (a, (d, (i, (f, p)))))
          (fn qa =>
            SemiAutomaton.states_enumerate_nat
              (Option.the (ArrayMapImpl.iam_lookup qa sm))))
    end
      c
  end;

fun rs_nfa_destruct A_ (q, (a, (d, (i, (f, p))))) =
  (RBTSetImpl.rs_to_list Arith.linorder_nat q,
    (RBTSetImpl.rs_to_list A_ a,
      (RBT_LTS_DLTS_LTSImpl.rs_lts_dlts_to_collect_list
         (Arith.equal_nat, Arith.linorder_nat) A_ d,
        (RBTSetImpl.rs_to_list Arith.linorder_nat i,
          RBTSetImpl.rs_to_list Arith.linorder_nat f))));

fun rs_nfa_final_no A_ (q, (a, (d, (i, (f, p))))) =
  StdInst.rs_size Arith.linorder_nat f;

fun rs_nfa_trans_no A_ (q, (a, (d, (i, (f, p))))) =
  SetIteratorGA.iterate_size
    (case d
      of LTSByLTS_DLTS.Lts aa =>
        RBT_TripleSetImpl.rs_ts_it Arith.linorder_nat A_ Arith.linorder_nat aa
      | LTSByLTS_DLTS.Dlts aa =>
        RBT_DLTSImpl.rs_dlts_it Arith.linorder_nat A_ aa);

fun rs_nfa_construct_gen A_ p (ql, (al, (dl, (il, fl)))) =
  List.foldl
    (fn (q, (a, (d, (i, (f, props))))) => fn (q1, (l, q2)) =>
      (RBTSetImpl.rs_ins Arith.linorder_nat q1
         (RBTSetImpl.rs_ins Arith.linorder_nat q2 q),
        (List.foldl (fn s => fn x => RBTSetImpl.rs_ins A_ x s) a l,
          (List.foldl
             (fn da => fn aa =>
               RBT_LTS_DLTS_LTSImpl.rs_lts_dlts_add_simple Arith.linorder_nat A_
                 q1 aa q2 da)
             d l,
            (i, (f, props))))))
    (RBTSetImpl.list_to_rs Arith.linorder_nat (ql @ il @ fl),
      (RBTSetImpl.list_to_rs A_ al,
        (RBT_LTS_DLTS_LTSImpl.rs_lts_dlts_empty_dlts Arith.linorder_nat A_ (),
          (RBTSetImpl.list_to_rs Arith.linorder_nat il,
            (RBTSetImpl.list_to_rs Arith.linorder_nat fl, p)))))
    dl;

fun rs_dfa_construct A_ = rs_nfa_construct_gen A_ (NFAByLTS.fields true false);

fun rs_nfa_dfa_construct_reachable B_ C_ det ff i a fp d_it =
  let
    val (b, c) =
      List.foldl
        (fn (aa, b) =>
          let
            val (qm, n) = aa;
          in
            (fn is => fn q =>
              ((RBTMapImpl.rm_update_dj B_ (ff q)
                  (SemiAutomaton.states_enumerate_nat n) qm,
                 Arith.suc n),
                RBTSetImpl.rs_ins_dj Arith.linorder_nat
                  (SemiAutomaton.states_enumerate_nat n) is))
          end
            b)
        ((RBTMapImpl.rm_empty B_ (), 0),
          RBTSetImpl.rs_empty Arith.linorder_nat ())
        i;
  in
    let
      val (qm, n) = b;
    in
      (fn is =>
        let
          val (aa, ba) =
            Workset.worklist (fn _ => true)
              (fn (aa, ba) =>
                let
                  val (qma, na) = aa;
                in
                  (fn (qs, (asa, (dd, (isa, (fs, p))))) => fn q =>
                    let
                      val r = Option.the (RBTMapImpl.rm_lookup B_ (ff q) qma);
                    in
                      (if RBTSetImpl.rs_memb Arith.linorder_nat r qs
                        then (((qma, na), (qs, (asa, (dd, (isa, (fs, p)))))),
                               [])
                        else let
                               val (ab, bb) =
                                 d_it q (fn _ => true)
                                   (fn (ab, qa) => fn (bb, ca) =>
                                     let
                                       val (qmb, nb) = bb;
                                     in
                                       (fn (dda, naa) =>
 let
   val r_opt = RBTMapImpl.rm_lookup B_ (ff qa) qmb;
   val (bc, cb) =
     (if Option.is_none r_opt
       then let
              val ra = SemiAutomaton.states_enumerate_nat nb;
            in
              ((RBTMapImpl.rm_update_dj B_ (ff qa) ra qmb, Arith.suc nb), ra)
            end
       else ((qmb, nb), Option.the r_opt));
 in
   let
     val (qmc, nc) = bc;
   in
     (fn ra =>
       ((qmc, nc),
         (RBT_LTS_DLTS_LTSImpl.rs_lts_dlts_add_choice Arith.linorder_nat C_ det
            r ab ra dda,
           qa :: naa)))
   end
     cb
 end)
                                     end
                                       ca)
                                   ((qma, na), (dd, []));
                             in
                               let
                                 val (qmb, nb) = ab;
                               in
                                 (fn (dda, ac) =>
                                   (((qmb, nb),
                                      (RBTSetImpl.rs_ins_dj Arith.linorder_nat r
 qs,
(asa, (dda, (isa, ((if fp q then RBTSetImpl.rs_ins_dj Arith.linorder_nat r fs
                     else fs),
                    p)))))),
                                     ac))
                               end
                                 bb
                             end)
                    end)
                end
                  ba)
              (((qm, n),
                 (RBTSetImpl.rs_empty Arith.linorder_nat (),
                   (a, (RBT_LTS_DLTS_LTSImpl.rs_lts_dlts_empty_dlts
                          Arith.linorder_nat C_ (),
                         (is, (RBTSetImpl.rs_empty Arith.linorder_nat (),
                                NFAByLTS.fields det true)))))),
                i);
        in
          let
            val (_, aaa) = aa;
          in
            (fn _ => aaa)
          end
            ba
        end)
    end
      c
  end;

fun rs_nfa_bool_comb_gen A_ a bc (q1, (a1, (d1, (i1, (f1, p1)))))
  (q2, (a2, (d2, (i2, (f2, p2))))) =
  rs_nfa_dfa_construct_reachable
    (Misc.linorder_prod (Arith.equal_nat, Arith.linorder_nat)
      (Arith.equal_nat, Arith.linorder_nat))
    A_ (NFAByLTS.nfa_prop_is_complete_deterministic p1 andalso
         NFAByLTS.nfa_prop_is_complete_deterministic p2)
    rs_hop_lts_copy
    (Enum.product (RBTSetImpl.rs_to_list Arith.linorder_nat i1)
      (RBTSetImpl.rs_to_list Arith.linorder_nat i2))
    a (fn (q1a, q2a) =>
        bc (RBTSetImpl.rs_memb Arith.linorder_nat q1a f1)
          (RBTSetImpl.rs_memb Arith.linorder_nat q2a f2))
    (fn (q1a, q2a) => fn c => fn f =>
      (case d1
        of LTSByLTS_DLTS.Lts aa =>
          RBT_TripleSetImpl.rs_ts_BC_it Arith.linorder_nat A_ Arith.linorder_nat
            aa
        | LTSByLTS_DLTS.Dlts aa =>
          RBT_DLTSImpl.rs_dlts_succ_label_it Arith.linorder_nat A_ aa)
        q1a
        c
        (fn (aa, q1b) =>
          (case d2
            of LTSByLTS_DLTS.Lts ab =>
              RBT_TripleSetImpl.rs_ts_C_it Arith.linorder_nat A_
                Arith.linorder_nat ab
            | LTSByLTS_DLTS.Dlts ab =>
              RBT_DLTSImpl.rs_dlts_succ_it Arith.linorder_nat A_ ab)
            q2a
            aa
            c
            (fn q2b => f (aa, (q1b, q2b)))));

fun rs_nfa_bool_comb A_ bc (q1, (a1, (d1, (i1, (f1, p1)))))
  (q2, (a2, (d2, (i2, (f2, p2))))) =
  rs_nfa_bool_comb_gen A_ a1 bc (q1, (a1, (d1, (i1, (f1, p1)))))
    (q2, (a2, (d2, (i2, (f2, p2)))));

fun rs_nfa_construct A_ = rs_nfa_construct_gen A_ (NFAByLTS.fields false false);

fun rs_nfa_labels_no A_ (q, (a, (d, (i, (f, p))))) = StdInst.rs_size A_ a;

fun rs_nfa_normalise A_ =
  (fn (q, (a, (d, (i, (f, p))))) =>
    (if NFAByLTS.nfa_prop_is_initially_connected p
      then (q, (a, (d, (i, (f, p)))))
      else rs_nfa_dfa_construct_reachable Arith.linorder_nat A_
             (NFAByLTS.nfa_prop_is_complete_deterministic p) rs_hop_lts_copy
             (RBTSetImpl.rs_to_list Arith.linorder_nat i) a
             (fn qa => RBTSetImpl.rs_memb Arith.linorder_nat qa f)
             (case d
               of LTSByLTS_DLTS.Lts aa =>
                 RBT_TripleSetImpl.rs_ts_BC_it Arith.linorder_nat A_
                   Arith.linorder_nat aa
               | LTSByLTS_DLTS.Dlts aa =>
                 RBT_DLTSImpl.rs_dlts_succ_label_it Arith.linorder_nat A_ aa)));

fun rs_nfa_states_no A_ (q, (a, (d, (i, (f, p))))) =
  StdInst.rs_size Arith.linorder_nat q;

fun rs_dfa_construct_reachable B_ C_ =
  rs_nfa_dfa_construct_reachable B_ C_ true;

fun rs_nfa_determinise A_ (q1, (a1, (d1, (i1, (f1, p1))))) =
  (if NFAByLTS.nfa_prop_is_initially_connected p1 andalso
        NFAByLTS.nfa_prop_is_complete_deterministic p1
    then (q1, (a1, (d1, (i1, (f1, p1)))))
    else let
           val re_map =
             Product_Type.fst
               (RBTSetImpl.rs_iteratei Arith.linorder_nat q1 (fn _ => true)
                 (fn q => fn (m, n) =>
                   (RBTMapImpl.rm_update_dj Arith.linorder_nat q n m,
                     IntInf.* ((2 : IntInf.int), n)))
                 (RBTMapImpl.rm_empty Arith.linorder_nat (), (1 : IntInf.int)));
         in
           rs_dfa_construct_reachable Arith.linorder_nat A_
             (fn q =>
               RBTSetImpl.rs_iteratei Arith.linorder_nat q (fn _ => true)
                 (fn qa => fn n =>
                   IntInf.+ (n,
                     Option.the
                       (RBTMapImpl.rm_lookup Arith.linorder_nat qa re_map)))
                 0)
             [i1] a1 (fn q => not (StdInst.rr_disjoint Arith.linorder_nat q f1))
             (fn s => fn c => fn f =>
               RBTSetImpl.rs_iteratei A_ a1 c
                 (fn x =>
                   f (x, RBTSetImpl.rs_iteratei Arith.linorder_nat s
                           (fn _ => true)
                           (fn a =>
                             (case d1
                               of LTSByLTS_DLTS.Lts aa =>
                                 RBT_TripleSetImpl.rs_ts_C_it Arith.linorder_nat
                                   A_ Arith.linorder_nat aa
                               | LTSByLTS_DLTS.Dlts aa =>
                                 RBT_DLTSImpl.rs_dlts_succ_it Arith.linorder_nat
                                   A_ aa)
                               a
                               x
                               (fn _ => true)
                               (RBTSetImpl.rs_ins Arith.linorder_nat))
                           (RBTSetImpl.rs_empty Arith.linorder_nat ()))))
         end);

fun rs_nfa_Brzozowski A_ a =
  rs_nfa_determinise A_
    (rs_nfa_reverse A_ (rs_nfa_determinise A_ (rs_nfa_reverse A_ a)));

fun rs_nfa_complement A_ =
  (fn (q, (a, (d, (i, (f, p))))) =>
    (q, (a, (d, (i, (StdInst.rr_diff Arith.linorder_nat q f, p))))));

fun rs_nfa_initial_no A_ (q, (a, (d, (i, (f, p))))) =
  StdInst.rs_size Arith.linorder_nat i;

fun rs_nfa_Hopcroft_NFA A_ =
  (fn a => rs_nfa_Hopcroft A_ (rs_nfa_determinise A_ a));

fun rs_nfa_rename_labels_gen A_ det (q, (aa, (d, (i, (fa, p))))) a f =
  (q, (a, (RBT_LTS_DLTS_LTSImpl.rs_lts_dlts_image Arith.linorder_nat A_
             Arith.linorder_nat A_ (fn (qb, (ab, qa)) => (qb, (f ab, qa))) d,
            (i, (fa, NFAByLTS.fields det
                       (NFAByLTS.nfa_prop_is_initially_connected p))))));

fun rs_nfa_rename_labels A_ det =
  (fn (q, (a, (d, (i, f)))) => fn fa =>
    rs_nfa_rename_labels_gen A_ det (q, (a, (d, (i, f))))
      (RBTSetImpl.rs_image A_ A_ fa a) fa);

fun rs_nfa_language_is_empty A_ n =
  let
    val (_, (_, (_, (_, (f, _))))) = rs_nfa_normalise A_ n;
  in
    RBTSetImpl.rs_isEmpty (Arith.equal_nat, Arith.linorder_nat) f
  end;

fun rs_nfa_language_is_subset A_ a1 a2 =
  rs_nfa_language_is_empty A_
    (rs_nfa_bool_comb A_ (fn a => fn b => a andalso b) a1
      (rs_nfa_complement A_ (rs_nfa_determinise A_ a2)));

fun rs_nfa_language_is_eq A_ a1 a2 =
  rs_nfa_language_is_subset A_ a1 a2 andalso rs_nfa_language_is_subset A_ a2 a1;

fun rs_nfa_destruct_simple A_ (q, (a, (d, (i, (f, p))))) =
  (RBTSetImpl.rs_to_list Arith.linorder_nat q,
    (RBTSetImpl.rs_to_list A_ a,
      (RBT_LTS_DLTS_LTSImpl.rs_lts_dlts_to_list Arith.linorder_nat A_ d,
        (RBTSetImpl.rs_to_list Arith.linorder_nat i,
          RBTSetImpl.rs_to_list Arith.linorder_nat f))));

fun rs_nfa_language_is_univ A_ a =
  rs_nfa_language_is_empty A_ (rs_nfa_complement A_ (rs_nfa_determinise A_ a));

fun rs_nfa_right_quotient_lists A_ ap (q, (a, (d, (i, (f, p))))) =
  let
    val m =
      RBT_LTS_DLTS_LTSImpl.rs_lts_dlts_filter_it Arith.linorder_nat A_
        (fn _ => true) ap (fn _ => true) (fn _ => true) d (fn _ => true)
        (fn (qb, (_, qa)) => fn m =>
          let
            val s =
              (case RBTMapImpl.rm_lookup Arith.linorder_nat qa m
                of NONE => RBTSetImpl.rs_empty Arith.linorder_nat ()
                | SOME s => s);
          in
            RBTMapImpl.rm_update Arith.linorder_nat qa
              (RBTSetImpl.rs_ins Arith.linorder_nat qb s) m
          end)
        (RBTMapImpl.rm_empty Arith.linorder_nat ());
    val fa =
      Product_Type.fst
        (Workset.worklist (fn _ => true)
          (fn s => fn e =>
            (if RBTSetImpl.rs_memb Arith.linorder_nat e s then (s, [])
              else (RBTSetImpl.rs_ins Arith.linorder_nat e s,
                     (case RBTMapImpl.rm_lookup Arith.linorder_nat e m
                       of NONE => []
                       | SOME aa =>
                         RBTSetImpl.rs_to_list Arith.linorder_nat aa))))
          (RBTSetImpl.rs_empty Arith.linorder_nat (),
            RBTSetImpl.rs_to_list Arith.linorder_nat f));
  in
    (q, (a, (d, (i, (fa, p)))))
  end;

end; (*struct RBT_NFAImpl*)

structure Regular_Exp : sig
  datatype 'a rexp = Zero | One | Atom of 'a | Plus of 'a rexp * 'a rexp |
    Times of 'a rexp * 'a rexp | Star of 'a rexp
  val equal_rexp : 'a HOL.equal -> 'a rexp -> 'a rexp -> bool
end = struct

datatype 'a rexp = Zero | One | Atom of 'a | Plus of 'a rexp * 'a rexp |
  Times of 'a rexp * 'a rexp | Star of 'a rexp;

fun equal_rexp A_ (Star rexp) (Times (rexp1, rexp2)) = false
  | equal_rexp A_ (Times (rexp1, rexp2)) (Star rexp) = false
  | equal_rexp A_ (Star rexp) (Plus (rexp1, rexp2)) = false
  | equal_rexp A_ (Plus (rexp1, rexp2)) (Star rexp) = false
  | equal_rexp A_ (Times (rexp1a, rexp2a)) (Plus (rexp1, rexp2)) = false
  | equal_rexp A_ (Plus (rexp1a, rexp2a)) (Times (rexp1, rexp2)) = false
  | equal_rexp A_ (Star rexp) (Atom a) = false
  | equal_rexp A_ (Atom a) (Star rexp) = false
  | equal_rexp A_ (Times (rexp1, rexp2)) (Atom a) = false
  | equal_rexp A_ (Atom a) (Times (rexp1, rexp2)) = false
  | equal_rexp A_ (Plus (rexp1, rexp2)) (Atom a) = false
  | equal_rexp A_ (Atom a) (Plus (rexp1, rexp2)) = false
  | equal_rexp A_ (Star rexp) One = false
  | equal_rexp A_ One (Star rexp) = false
  | equal_rexp A_ (Times (rexp1, rexp2)) One = false
  | equal_rexp A_ One (Times (rexp1, rexp2)) = false
  | equal_rexp A_ (Plus (rexp1, rexp2)) One = false
  | equal_rexp A_ One (Plus (rexp1, rexp2)) = false
  | equal_rexp A_ (Atom a) One = false
  | equal_rexp A_ One (Atom a) = false
  | equal_rexp A_ (Star rexp) Zero = false
  | equal_rexp A_ Zero (Star rexp) = false
  | equal_rexp A_ (Times (rexp1, rexp2)) Zero = false
  | equal_rexp A_ Zero (Times (rexp1, rexp2)) = false
  | equal_rexp A_ (Plus (rexp1, rexp2)) Zero = false
  | equal_rexp A_ Zero (Plus (rexp1, rexp2)) = false
  | equal_rexp A_ (Atom a) Zero = false
  | equal_rexp A_ Zero (Atom a) = false
  | equal_rexp A_ One Zero = false
  | equal_rexp A_ Zero One = false
  | equal_rexp A_ (Star rexpa) (Star rexp) = equal_rexp A_ rexpa rexp
  | equal_rexp A_ (Times (rexp1a, rexp2a)) (Times (rexp1, rexp2)) =
    equal_rexp A_ rexp1a rexp1 andalso equal_rexp A_ rexp2a rexp2
  | equal_rexp A_ (Plus (rexp1a, rexp2a)) (Plus (rexp1, rexp2)) =
    equal_rexp A_ rexp1a rexp1 andalso equal_rexp A_ rexp2a rexp2
  | equal_rexp A_ (Atom aa) (Atom a) = HOL.eq A_ aa a
  | equal_rexp A_ One One = true
  | equal_rexp A_ Zero Zero = true;

end; (*struct Regular_Exp*)

structure nfa_to_rexp : sig
  datatype 'a gnfastate = Start | End | State of 'a
  val equal_gnfastatea : 'a HOL.equal -> 'a gnfastate -> 'a gnfastate -> bool
  val equal_gnfastate : 'a HOL.equal -> 'a gnfastate HOL.equal
  val hashcode_gnfastate : 'a HashCode.hashable -> 'a gnfastate -> IntInf.int
  val bounded_hashcode_gnfastate :
    'a HashCode.hashable -> IntInf.int -> 'a gnfastate -> IntInf.int
  val def_hashmap_size_gnfastate :
    'a HashCode.hashable -> 'a gnfastate HOL.itself -> IntInf.int
  val hashable_gnfastate :
    'a HashCode.hashable -> 'a gnfastate HashCode.hashable
  val rexp_simped_concat :
    'a HOL.equal ->
      'a Regular_Exp.rexp -> 'a Regular_Exp.rexp -> 'a Regular_Exp.rexp
  val rexp_simped_contract :
    'a HOL.equal ->
      'a Regular_Exp.rexp ->
        'a Regular_Exp.rexp -> 'a Regular_Exp.rexp -> 'a Regular_Exp.rexp
end = struct

datatype 'a gnfastate = Start | End | State of 'a;

fun equal_gnfastatea A_ (State q) End = false
  | equal_gnfastatea A_ End (State q) = false
  | equal_gnfastatea A_ (State q) Start = false
  | equal_gnfastatea A_ Start (State q) = false
  | equal_gnfastatea A_ End Start = false
  | equal_gnfastatea A_ Start End = false
  | equal_gnfastatea A_ (State qa) (State q) = HOL.eq A_ qa q
  | equal_gnfastatea A_ End End = true
  | equal_gnfastatea A_ Start Start = true;

fun equal_gnfastate A_ = {equal = equal_gnfastatea A_} : 'a gnfastate HOL.equal;

fun hashcode_gnfastate A_ q =
  (case q of Start => 0 | End => (1 : IntInf.int)
    | State a => HashCode.hashcode A_ a);

fun bounded_hashcode_gnfastate A_ n q =
  (case q of Start => 0 | End => (1 : IntInf.int)
    | State qa =>
      Arith.mod_nat (IntInf.+ ((2 : IntInf.int), HashCode.hashcode A_ qa)) n);

fun def_hashmap_size_gnfastate A_ = (fn _ => (16 : IntInf.int));

fun hashable_gnfastate A_ =
  {hashcode = hashcode_gnfastate A_,
    bounded_hashcode = bounded_hashcode_gnfastate A_,
    def_hashmap_size = def_hashmap_size_gnfastate A_}
  : 'a gnfastate HashCode.hashable;

fun rexp_simped_concat A_ r s =
  (if Regular_Exp.equal_rexp A_ r Regular_Exp.One then s
    else (if Regular_Exp.equal_rexp A_ s Regular_Exp.One then r
           else Regular_Exp.Times (r, s)));

fun rexp_simped_contract A_ r s t =
  (if Regular_Exp.equal_rexp A_ s Regular_Exp.Zero
    then rexp_simped_concat A_ r t
    else rexp_simped_concat A_ r
           (rexp_simped_concat A_ (Regular_Exp.Star s) t));

end; (*struct nfa_to_rexp*)

structure Nat_Bijection : sig
  val triangle : IntInf.int -> IntInf.int
  val prod_encode : IntInf.int * IntInf.int -> IntInf.int
end = struct

fun triangle n = Arith.div_nat (IntInf.* (n, Arith.suc n)) (2 : IntInf.int);

fun prod_encode x = (fn (m, n) => IntInf.+ (triangle (IntInf.+ (m, n)), m)) x;

end; (*struct Nat_Bijection*)

structure ListSetImpl_Invar : sig
  val lsi_iteratei : 'a list -> ('b -> bool) -> ('a -> 'b -> 'b) -> 'b -> 'b
  val lsi_sel : 'a list -> ('a -> bool) -> 'a option
  val lsi_empty : unit -> 'a list
  val lsi_delete : 'a HOL.equal -> 'a -> 'a list -> 'a list
  val lsi_ins_dj : 'a -> 'a list -> 'a list
  val lsi_isEmpty : 'a list -> bool
end = struct

fun lsi_iteratei x = ListIterator.foldli x;

fun lsi_sel x = SetGA.iti_sel_no_map lsi_iteratei x;

fun lsi_empty x = (fn _ => []) x;

fun lsi_delete A_ x l = Dlist_add.dlist_remove1 A_ x [] l;

fun lsi_ins_dj x l = x :: l;

fun lsi_isEmpty s = List.null s;

end; (*struct ListSetImpl_Invar*)

structure nfa_to_rexp_Export : sig
  val rs_nfa_to_rexp :
    'a HOL.equal * 'a HashCode.hashable * 'a Orderings.linorder ->
      (IntInf.int, unit) RBT.rbt *
        (('a, unit) RBT.rbt *
          (((IntInf.int, ('a, (IntInf.int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
             (IntInf.int, ('a, IntInf.int) RBT.rbt) RBT.rbt)
             LTSByLTS_DLTS.lTS_choice *
            ((IntInf.int, unit) RBT.rbt *
              ((IntInf.int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext)))) ->
        'a Regular_Exp.rexp
end = struct

fun rs_nfa_to_rexp (A1_, A2_, A3_) a_impl =
  let
    val (a, (aa, (ab, bb))) =
      let
        val x =
          let
            val x =
              SetGA.it_copy (RBTSetImpl.rs_iteratei Arith.linorder_nat)
                ListSetImpl_Invar.lsi_empty ListSetImpl_Invar.lsi_ins_dj
                (Product_Type.fst a_impl);
            val xa =
              ListSetImpl_Invar.lsi_iteratei x (fn _ => true)
                (fn xa => fn a =>
                  ArrayHashMap.ahm_update
                    (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                      nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
                    (nfa_to_rexp.State xa)
                    (ArrayHashSet.ahs_empty
                      (nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat) ())
                    a)
                (ArrayHashMap.ahm_empty
                  (nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat) ());
            val xb =
              ArrayHashMap.ahm_update
                (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                  nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
                nfa_to_rexp.Start
                (ArrayHashSet.ahs_empty
                  (nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat) ())
                (ArrayHashMap.ahm_update
                  (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                    nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
                  nfa_to_rexp.End
                  (ArrayHashSet.ahs_empty
                    (nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat) ())
                  xa);
            val xc =
              ArrayHashMap.ahm_empty
                (nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat) ();
          in
            (x, (xc, (xb, xb)))
          end;
        val xa =
          RBTSetImpl.rs_iteratei Arith.linorder_nat
            (Product_Type.fst
              (Product_Type.snd (Product_Type.snd (Product_Type.snd a_impl))))
            (fn _ => true)
            (fn xa => fn (a, (aa, (ab, bb))) =>
              (a, ((case ArrayHashMap.ahm_lookup
                           (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                             nfa_to_rexp.hashable_gnfastate
                               HashCode.hashable_nat)
                           nfa_to_rexp.Start aa
                     of NONE =>
                       ArrayHashMap.ahm_update
                         (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                           nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
                         nfa_to_rexp.Start
                         (StdInst.ahm_sng
                           (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                             nfa_to_rexp.hashable_gnfastate
                               HashCode.hashable_nat)
                           (nfa_to_rexp.State xa) Regular_Exp.One)
                         aa
                     | SOME du =>
                       ArrayHashMap.ahm_update
                         (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                           nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
                         nfa_to_rexp.Start
                         (ArrayHashMap.ahm_update
                           (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                             nfa_to_rexp.hashable_gnfastate
                               HashCode.hashable_nat)
                           (nfa_to_rexp.State xa) Regular_Exp.One du)
                         aa),
                    (ArrayHashMap.ahm_update
                       (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                         nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
                       (nfa_to_rexp.State xa)
                       (StdInst.ahs_sng
                         (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                           nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
                         nfa_to_rexp.Start)
                       ab,
                      (case ArrayHashMap.ahm_lookup
                              (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                                nfa_to_rexp.hashable_gnfastate
                                  HashCode.hashable_nat)
                              nfa_to_rexp.Start bb
                        of NONE => bb
                        | SOME mu =>
                          ArrayHashMap.ahm_update
                            (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                              nfa_to_rexp.hashable_gnfastate
                                HashCode.hashable_nat)
                            nfa_to_rexp.Start
                            (ArrayHashSet.ahs_ins
                              (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                                nfa_to_rexp.hashable_gnfastate
                                  HashCode.hashable_nat)
                              (nfa_to_rexp.State xa) mu)
                            bb)))))
            x;
        val xb =
          RBTSetImpl.rs_iteratei Arith.linorder_nat
            (Product_Type.fst
              (Product_Type.snd
                (Product_Type.snd
                  (Product_Type.snd (Product_Type.snd a_impl)))))
            (fn _ => true)
            (fn xb => fn (a, (aa, (ab, bb))) =>
              (a, ((case ArrayHashMap.ahm_lookup
                           (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                             nfa_to_rexp.hashable_gnfastate
                               HashCode.hashable_nat)
                           (nfa_to_rexp.State xb) aa
                     of NONE =>
                       ArrayHashMap.ahm_update
                         (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                           nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
                         (nfa_to_rexp.State xb)
                         (StdInst.ahm_sng
                           (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                             nfa_to_rexp.hashable_gnfastate
                               HashCode.hashable_nat)
                           nfa_to_rexp.End Regular_Exp.One)
                         aa
                     | SOME du =>
                       ArrayHashMap.ahm_update
                         (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                           nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
                         (nfa_to_rexp.State xb)
                         (ArrayHashMap.ahm_update
                           (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                             nfa_to_rexp.hashable_gnfastate
                               HashCode.hashable_nat)
                           nfa_to_rexp.End Regular_Exp.One du)
                         aa),
                    ((case ArrayHashMap.ahm_lookup
                             (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                               nfa_to_rexp.hashable_gnfastate
                                 HashCode.hashable_nat)
                             nfa_to_rexp.End ab
                       of NONE => ab
                       | SOME mu =>
                         ArrayHashMap.ahm_update
                           (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                             nfa_to_rexp.hashable_gnfastate
                               HashCode.hashable_nat)
                           nfa_to_rexp.End
                           (ArrayHashSet.ahs_ins
                             (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                               nfa_to_rexp.hashable_gnfastate
                                 HashCode.hashable_nat)
                             (nfa_to_rexp.State xb) mu)
                           ab),
                      ArrayHashMap.ahm_update
                        (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                          nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
                        (nfa_to_rexp.State xb)
                        (StdInst.ahs_sng
                          (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                            nfa_to_rexp.hashable_gnfastate
                              HashCode.hashable_nat)
                          nfa_to_rexp.End)
                        bb))))
            xa;
        val xc =
          (case Product_Type.fst (Product_Type.snd (Product_Type.snd a_impl))
            of LTSByLTS_DLTS.Lts a =>
              RBT_TripleSetImpl.rs_ts_it Arith.linorder_nat A3_
                Arith.linorder_nat a
            | LTSByLTS_DLTS.Dlts a =>
              RBT_DLTSImpl.rs_dlts_it Arith.linorder_nat A3_ a)
            (fn _ => true)
            (fn xc => fn s =>
              let
                val (a, (aa, ba)) = xc;
                val (ab, (ac, (ad, bd))) = s;
              in
                (ab, ((case ArrayHashMap.ahm_lookup
                              (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                                nfa_to_rexp.hashable_gnfastate
                                  HashCode.hashable_nat)
                              (nfa_to_rexp.State a) ac
                        of NONE =>
                          ArrayHashMap.ahm_update
                            (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                              nfa_to_rexp.hashable_gnfastate
                                HashCode.hashable_nat)
                            (nfa_to_rexp.State a)
                            (StdInst.ahm_sng
                              (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                                nfa_to_rexp.hashable_gnfastate
                                  HashCode.hashable_nat)
                              (nfa_to_rexp.State ba) (Regular_Exp.Atom aa))
                            ac
                        | SOME du =>
                          let
                            val duv =
                              (case ArrayHashMap.ahm_lookup
                                      (nfa_to_rexp.equal_gnfastate
 Arith.equal_nat,
nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
                                      (nfa_to_rexp.State ba) du
                                of NONE => Regular_Exp.Atom aa
                                | SOME duv =>
                                  Regular_Exp.Plus (duv, Regular_Exp.Atom aa));
                          in
                            ArrayHashMap.ahm_update
                              (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                                nfa_to_rexp.hashable_gnfastate
                                  HashCode.hashable_nat)
                              (nfa_to_rexp.State a)
                              (ArrayHashMap.ahm_update
                                (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                                  nfa_to_rexp.hashable_gnfastate
                                    HashCode.hashable_nat)
                                (nfa_to_rexp.State ba) duv du)
                              ac
                          end),
                       ((case ArrayHashMap.ahm_lookup
                                (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                                  nfa_to_rexp.hashable_gnfastate
                                    HashCode.hashable_nat)
                                (nfa_to_rexp.State ba) ad
                          of NONE => ad
                          | SOME mu =>
                            ArrayHashMap.ahm_update
                              (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                                nfa_to_rexp.hashable_gnfastate
                                  HashCode.hashable_nat)
                              (nfa_to_rexp.State ba)
                              (ArrayHashSet.ahs_ins
                                (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                                  nfa_to_rexp.hashable_gnfastate
                                    HashCode.hashable_nat)
                                (nfa_to_rexp.State a) mu)
                              ad),
                         (case ArrayHashMap.ahm_lookup
                                 (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                                   nfa_to_rexp.hashable_gnfastate
                                     HashCode.hashable_nat)
                                 (nfa_to_rexp.State a) bd
                           of NONE => bd
                           | SOME mu =>
                             ArrayHashMap.ahm_update
                               (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                                 nfa_to_rexp.hashable_gnfastate
                                   HashCode.hashable_nat)
                               (nfa_to_rexp.State a)
                               (ArrayHashSet.ahs_ins
                                 (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                                   nfa_to_rexp.hashable_gnfastate
                                     HashCode.hashable_nat)
                                 (nfa_to_rexp.State ba) mu)
                               bd))))
              end)
            xb;
      in
        xc
      end;
    val (_, (ad, (_, _))) =
      While_Combinator.whilea
        (fn (q, (_, (_, _))) => not (ListSetImpl_Invar.lsi_isEmpty q))
        (fn (ac, (ad, (ae, be))) =>
          let
            val xb = Option.the (ListSetImpl_Invar.lsi_sel ac (fn _ => true));
            val (af, (ag, (ah, bh))) =
              let
                val (af, (aaa, (aba, bba))) = (ac, (ad, (ae, be)));
                val x =
                  (case ArrayHashMap.ahm_lookup
                          (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                            nfa_to_rexp.hashable_gnfastate
                              HashCode.hashable_nat)
                          (nfa_to_rexp.State xb) aba
                    of NONE =>
                      ArrayHashSet.ahs_empty
                        (nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
                        ()
                    | SOME ag =>
                      ArrayHashSet.ahs_delete
                        (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                          nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
                        (nfa_to_rexp.State xb) ag);
                val xa =
                  (case ArrayHashMap.ahm_lookup
                          (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                            nfa_to_rexp.hashable_gnfastate
                              HashCode.hashable_nat)
                          (nfa_to_rexp.State xb) bba
                    of NONE =>
                      ArrayHashSet.ahs_empty
                        (nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
                        ()
                    | SOME ag =>
                      ArrayHashSet.ahs_delete
                        (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                          nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
                        (nfa_to_rexp.State xb) ag);
                val (aca, (ada, (aea, bea))) =
                  let
                    val (ag, (aab, (abb, bbb))) = (af, (aaa, (aba, bba)));
                  in
                    (case nfa_to_rexp.State xb
                      of nfa_to_rexp.Start => (ag, (aab, (abb, bbb)))
                      | nfa_to_rexp.End => (ag, (aab, (abb, bbb)))
                      | nfa_to_rexp.State q =>
                        let
                          val xc =
                            ArrayHashSet.ahs_iteratei
                              (nfa_to_rexp.hashable_gnfastate
                                HashCode.hashable_nat)
                              (case ArrayHashMap.ahm_lookup
                                      (nfa_to_rexp.equal_gnfastate
 Arith.equal_nat,
nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
                                      (nfa_to_rexp.State xb) bbb
                                of NONE =>
                                  ArrayHashSet.ahs_empty
                                    (nfa_to_rexp.hashable_gnfastate
                                      HashCode.hashable_nat)
                                    ()
                                | SOME mu => mu)
                              (fn _ => true)
                              (fn xc => fn s =>
                                (case ArrayHashMap.ahm_lookup
(nfa_to_rexp.equal_gnfastate Arith.equal_nat,
  nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
xc s
                                  of NONE => s
                                  | SOME mu =>
                                    ArrayHashMap.ahm_update
                                      (nfa_to_rexp.equal_gnfastate
 Arith.equal_nat,
nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
                                      xc (ArrayHashSet.ahs_delete
   (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
     nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
   (nfa_to_rexp.State xb) mu)
                                      s))
                              abb;
                          val xaa =
                            ArrayHashSet.ahs_iteratei
                              (nfa_to_rexp.hashable_gnfastate
                                HashCode.hashable_nat)
                              (case ArrayHashMap.ahm_lookup
                                      (nfa_to_rexp.equal_gnfastate
 Arith.equal_nat,
nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
                                      (nfa_to_rexp.State xb) abb
                                of NONE =>
                                  ArrayHashSet.ahs_empty
                                    (nfa_to_rexp.hashable_gnfastate
                                      HashCode.hashable_nat)
                                    ()
                                | SOME mu => mu)
                              (fn _ => true)
                              (fn xd => fn s =>
                                (case ArrayHashMap.ahm_lookup
(nfa_to_rexp.equal_gnfastate Arith.equal_nat,
  nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
xd s
                                  of NONE => s
                                  | SOME mu =>
                                    ArrayHashMap.ahm_update
                                      (nfa_to_rexp.equal_gnfastate
 Arith.equal_nat,
nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
                                      xd (ArrayHashSet.ahs_delete
   (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
     nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
   (nfa_to_rexp.State xb) mu)
                                      s))
                              bbb;
                        in
                          (ListSetImpl_Invar.lsi_delete Arith.equal_nat q ag,
                            (aab, (xc, xaa)))
                        end)
                  end;
                val (afa, (ag, (ah, bh))) =
                  ArrayHashSet.ahs_iteratei
                    (nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat) x
                    (fn _ => true)
                    (fn xc => fn (afa, (ag, (ah, bh))) =>
                      ArrayHashSet.ahs_iteratei
                        (nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
                        xa (fn _ => true)
                        (fn xaa => fn (ai, (aj, (ak, bk))) =>
                          (ai, (let
                                  val r1 =
                                    (case ArrayHashMap.ahm_lookup
    (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
      nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
    xc aj
                                      of NONE => Regular_Exp.Zero
                                      | SOME delta_u =>
(case ArrayHashMap.ahm_lookup
        (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
          nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
        (nfa_to_rexp.State xb) delta_u
  of NONE => Regular_Exp.Zero | SOME delta_uv => delta_uv));
                                  val r2 =
                                    (case ArrayHashMap.ahm_lookup
    (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
      nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
    (nfa_to_rexp.State xb) aj
                                      of NONE => Regular_Exp.Zero
                                      | SOME delta_u =>
(case ArrayHashMap.ahm_lookup
        (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
          nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
        (nfa_to_rexp.State xb) delta_u
  of NONE => Regular_Exp.Zero | SOME delta_uv => delta_uv));
                                  val r3 =
                                    (case ArrayHashMap.ahm_lookup
    (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
      nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
    (nfa_to_rexp.State xb) aj
                                      of NONE => Regular_Exp.Zero
                                      | SOME delta_u =>
(case ArrayHashMap.ahm_lookup
        (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
          nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
        xaa delta_u
  of NONE => Regular_Exp.Zero | SOME delta_uv => delta_uv));
                                in
                                  (case ArrayHashMap.ahm_lookup
  (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
    nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
  xc aj
                                    of NONE =>
                                      ArrayHashMap.ahm_update
(nfa_to_rexp.equal_gnfastate Arith.equal_nat,
  nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
xc (StdInst.ahm_sng
     (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
       nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
     xaa (nfa_to_rexp.rexp_simped_contract A1_ r1 r2 r3))
aj
                                    | SOME du =>
                                      let
val duv =
  (case ArrayHashMap.ahm_lookup
          (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
            nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
          xaa du
    of NONE => nfa_to_rexp.rexp_simped_contract A1_ r1 r2 r3
    | SOME duv =>
      Regular_Exp.Plus (duv, nfa_to_rexp.rexp_simped_contract A1_ r1 r2 r3));
                                      in
ArrayHashMap.ahm_update
  (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
    nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
  xc (ArrayHashMap.ahm_update
       (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
         nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
       xaa duv du)
  aj
                                      end)
                                end,
                                 ((case ArrayHashMap.ahm_lookup
  (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
    nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
  xaa ak
                                    of NONE => ak
                                    | SOME mu =>
                                      ArrayHashMap.ahm_update
(nfa_to_rexp.equal_gnfastate Arith.equal_nat,
  nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
xaa (ArrayHashSet.ahs_ins
      (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
        nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
      xc mu)
ak),
                                   (case ArrayHashMap.ahm_lookup
   (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
     nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
   xc bk
                                     of NONE => bk
                                     | SOME mu =>
                                       ArrayHashMap.ahm_update
 (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
   nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
 xc (ArrayHashSet.ahs_ins
      (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
        nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
      xaa mu)
 bk)))))
                        (afa, (ag, (ah, bh))))
                    (aca, (ada, (aea, bea)));
              in
                (afa, (ag, (ah, bh)))
              end;
          in
            (af, (ag, (ah, bh)))
          end)
        (a, (aa, (ab, bb)));
  in
    (case ArrayHashMap.ahm_lookup
            (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
              nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
            nfa_to_rexp.Start ad
      of NONE => Regular_Exp.Zero
      | SOME delta_u =>
        (case ArrayHashMap.ahm_lookup
                (nfa_to_rexp.equal_gnfastate Arith.equal_nat,
                  nfa_to_rexp.hashable_gnfastate HashCode.hashable_nat)
                nfa_to_rexp.End delta_u
          of NONE => Regular_Exp.Zero | SOME delta_uv => delta_uv))
  end;

end; (*struct nfa_to_rexp_Export*)
