use "Unsynchronized.sml";
use "Automata_Export_RBT.sml";

(* copied from  Isabelle2011-1/src/Pure/library.ML" *)
exception RANDOM;

local
  fun rmod x y = x - y * Real.realFloor (x / y);
  val a = 16807.0;
  val m = 2147483647.0;
  val random_seed = ref (Time.toReal (Time.now ()));
in

fun random () = 
  let val r = rmod (a * ! random_seed) m
  in (random_seed := r; r) end;

fun random_range l h =
  if h < l orelse l < 0 then raise RANDOM
  else l + Real.floor (rmod (random ()) (real (h - l + 1)));

end;

signature Sigma_param = sig
   type Sigma
   type Sigma_Word
   val Sigma_to_string : Sigma -> string
   val Sigma_Word_create : Sigma_Word -> Sigma list
   val Sigma_dsarg : Sigma HOL.equal * Sigma Orderings.linorder
   val Sigma_dsarg2 : Sigma HOL.equal * Sigma HashCode.hashable * Sigma Orderings.linorder
   val int_to_Sigma: int -> Sigma
end

structure string_Sigma_param : Sigma_param = struct
  type Sigma = string;
  type Sigma_Word = string

  val string_equal = {equal = fn (s1:string) => fn s2 => (s1 = s2)}

  val string_ord = {less_eq = fn s1 => fn s2 => (String.<= (s1, s2)), 
                   less = fn s1 => fn s2 => (String.< (s1, s2))};
  val string_preorder = {ord_preorder = string_ord};
  val string_order = {preorder_order = string_preorder};
  val string_linorder = {order_linorder=string_order};

  val Sigma_dsarg = (string_equal, string_linorder);


  fun hashcode_String s = 
     Char.ord (String.sub (s, 0))
     handle Subscript => 0
  fun bounded_hashcode_String na s = Arith.mod_nat (hashcode_String s) na;
  fun def_hashmap_size_String x = (fn _ => (16 : IntInf.int)) x;

  val string_hashable =
    {hashcode = hashcode_String, bounded_hashcode = bounded_hashcode_String,
      def_hashmap_size = def_hashmap_size_String}

  val Sigma_dsarg2 = (string_equal, string_hashable, string_linorder);

  fun Sigma_to_string s = s
  fun Sigma_Word_create s = map Char.toString (explode s)
  fun int_to_Sigma n = Char.toString (chr (n+97))
end


functor NFA (S : Sigma_param) : sig
  type NFA

  (* construct_NFA is used to construct NFAs from 
     a tuple "(Q, A, D, I, F)" where 
   
     Q : is a list of states
     A : is a list of labels 
     D : represents the transition relations. It consists of a list of triples.
         A triple "(q, al, q')" represents 
         all transitions from state q to state q' using a label in the list al.
     I : is the list of initial states
     F : is the list of final states

     Q and A get extended automatically to contain all the states and labels occuring in
     D, I and F. Therefore, they are usually empty lists. Only unreachable states and
     unused labels need to be added explicitly. 

     The result of construct_NFA is always a well-formed NFA,
  *)
  val nfa_construct : 
     (* Q *) int list * 
     (* A *) S.Sigma list * 
     (* D *) (int * S.Sigma list * int) list * 
     (* I *) int list *
     (* F *) int list 
     -> 
     NFA;

  (* Use internally a more efficient datastructure, if it can be guaranteed that the
     result is a DFA. *)
  val dfa_construct : 
     (* Q *) int list * 
     (* A *) S.Sigma list * 
     (* D *) (int * S.Sigma list * int) list * 
     (* I *) int list *
     (* F *) int list 
     -> 
     NFA;

  (* destruct_NFA is the inverse of construct_NFA. It destructs an NFA into a
     list representation. *)
  val nfa_destruct : NFA -> 
     int list * S.Sigma list * (int * S.Sigma list * int) list * int list * int list

  (* a simpler, faster version without grouping *)
  val nfa_destruct_simple : NFA -> 
     int list * S.Sigma list * (int * S.Sigma * int) list * int list * int list
  
  (* show_NFA uses the GraphViz package in order to display NFAs. *)
  val show_NFA : string option -> NFA -> unit

  (* export_NFA allows storing the GraphViz output in various file formats *)
  val export_NFA : string (*Format*) -> string (*File*) -> string option (*Label*) -> NFA -> unit

  (* random test generation *)
  val generate_random_NFAs : int -> int -> int -> real -> real -> NFA list
  val generate_random_ICDFAs : int -> int -> int -> real -> real -> NFA list


  (* Operations on NFAs *)  
  val nfa_accept_base : NFA -> S.Sigma list -> bool
  val nfa_accept : NFA -> S.Sigma_Word -> bool

  val nfa_rename_labels : NFA -> (S.Sigma -> S.Sigma) -> NFA
  val nfa_normalise : NFA -> NFA
  val nfa_complement : NFA -> NFA
  val nfa_bool_comb : (bool -> bool -> bool) -> NFA -> NFA -> NFA 
  val nfa_product : NFA -> NFA -> NFA 
  val nfa_determinise : NFA -> NFA
  val nfa_reverse : NFA -> NFA
  val nfa_Brzozowski : NFA -> NFA
  val nfa_Hopcroft : NFA -> NFA
  val nfa_Hopcroft_NFA : NFA -> NFA

  val nfa_to_rexp : NFA -> S.Sigma Regular_Exp.rexp

  (* Statistics on NFAs *)  
  val nfa_states_no : NFA -> int
  val nfa_labels_no : NFA -> int
  val nfa_trans_no : NFA -> int
  val nfa_final_no : NFA -> int
  val nfa_initial_no : NFA -> int

  (* Language checks on NFAs *)  
  val nfa_lang_is_empty : NFA -> bool
  val nfa_lang_is_univ : NFA -> bool
  val nfa_lang_is_subset : NFA -> NFA -> bool
  val nfa_lang_is_eq : NFA -> NFA -> bool

end = struct

type NFA = 
      (int, unit) RBT.rbt *
      ((S.Sigma, unit) RBT.rbt *
      (((int, (S.Sigma, (int, unit) RBT.rbt) RBT.rbt) RBT.rbt,
      (int, (S.Sigma, int) RBT.rbt) RBT.rbt) LTSByLTS_DLTS.lTS_choice *
      ((int, unit) RBT.rbt *
      ((int, unit) RBT.rbt * unit NFAByLTS.nfa_props_ext))))

(*
val nFA_states_nat = {states_enumerate = (fn i => i)}:(int NFA.nFA_states)
val equal_nat = {equal = fn (s1:int) => fn s2 => (s1 = s2)}
val state_args = (nFA_states_nat, Arith.linorder_nat)
val ext_state_args = (equal_nat, nFA_states_nat, Arith.linorder_nat)
*)

val sigma_linord = #2 (S.Sigma_dsarg)

fun nfa_construct (Q, A, D, I, F) = 
   (RBT_NFAImpl.rs_nfa_construct sigma_linord
    (Q, (A, (map (fn (a, b,c) => (a, (b,c))) D, (I, F))))):NFA;

fun dfa_construct (Q, A, D, I, F) = 
   RBT_NFAImpl.rs_dfa_construct sigma_linord
    (Q, (A, (map (fn (a, b,c) => (a, (b,c))) D, (I, F)))):NFA;

fun nfa_destruct (AA:NFA) =
   let val (Q, (A, (D, (I, F)))) = 
    RBT_NFAImpl.rs_nfa_destruct sigma_linord AA in
   (Q, A, map (fn (a, (b,c)) => (a, b,c)) D, I, F) end;

fun nfa_destruct_simple (AA:NFA) =
   let val (Q, (A, (D, (I, F)))) = 
    RBT_NFAImpl.rs_nfa_destruct_simple sigma_linord AA in
   (Q, A, map (fn (a, (b,c)) => (a, b,c)) D, I, F) end;

fun write_dot_graph os labelOpt AA =
let 
   val (Q, A, D, I, F) = nfa_destruct AA;
   val _ = TextIO.output (os, "digraph {\n")
   val _ = case labelOpt of NONE => () 
               | SOME l => TextIO.output (os, "label = \"" ^ l ^ "\"\n")
   val _ = TextIO.output (os, "rankdir=LR;\n")
   val _ = TextIO.output (os, "\"\" [shape=none];\n")

   fun separate_list x [] = []
     | separate_list x [s] = [s]
     | separate_list x (s1 :: s2 :: sl) =
       s1 :: x :: (separate_list x (s2 :: sl));

   fun separate x sl = String.concat (separate_list x sl) 

   fun mem x [] = false
     | mem x (x' :: xs) = if x = x' then true else mem x xs

   fun get_node_shape n =
      if (mem n F) then "doublecircle" else "circle"
   fun get_node_format n =
     String.concat ("shape = " :: (get_node_shape n) :: []);
   fun output_node n =
     TextIO.output (os, Int.toString n ^ " ["^get_node_format n^"];\n")
   fun output_initial_node n =
      (TextIO.output (os, String.concat [
          "\"\" -> ", Int.toString n, ";\n"]))
   val _ = map output_node Q;
   val _ = map output_initial_node I;

   fun output_edge (q0, al, q1) =
      (TextIO.output (os, String.concat [
          Int.toString q0, " -> ", Int.toString q1, " [ label = \"",
          separate ", " (rev (map S.Sigma_to_string al)), "\" ];\n"]))
   val _ = map output_edge D;
   val _ = TextIO.output (os, "}\n");   
in
   ()
end


fun export_NFA format file labelOpt AA =
let
   val proc = Unix.execute ("/usr/bin/dot", ["-T"^format, "-o"^file])
   val os = Unix.textOutstreamOf proc

   val _ = write_dot_graph os labelOpt AA
   val _ = TextIO.closeOut os;
in
  ()
end;

fun show_NFA labelOpt AA =
let
   val proc = Unix.execute ("/usr/bin/dot", ["-Tx11"])
   val os = Unix.textOutstreamOf proc

   val _ = write_dot_graph os labelOpt AA
   val _ = TextIO.closeOut os;
in
  ()
end;


open RBT_NFAImpl

(* Basic operations *)
val nfa_accept_base = rs_nfa_accept sigma_linord:(NFA -> S.Sigma list -> bool)
fun nfa_accept A w = nfa_accept_base A (S.Sigma_Word_create w)

val nfa_rename_labels = (rs_nfa_rename_labels sigma_linord false):(NFA -> (S.Sigma -> S.Sigma) -> NFA) 
val nfa_normalise = (rs_nfa_normalise sigma_linord):(NFA -> NFA)
val nfa_complement = (rs_nfa_complement sigma_linord):(NFA -> NFA)
val nfa_bool_comb = (rs_nfa_bool_comb sigma_linord):(
   (bool -> bool -> bool) -> NFA -> NFA -> NFA)
val nfa_product = nfa_bool_comb (fn b1 => fn b2 => b1 andalso b2)
val nfa_determinise = (rs_nfa_determinise sigma_linord):(NFA -> NFA)
val nfa_reverse = (rs_nfa_reverse sigma_linord):(NFA -> NFA)
val nfa_Brzozowski = (rs_nfa_Brzozowski sigma_linord):(NFA -> NFA)
val nfa_Hopcroft = (rs_nfa_Hopcroft sigma_linord):(NFA -> NFA)
val nfa_Hopcroft_NFA = (rs_nfa_Hopcroft_NFA sigma_linord):(NFA -> NFA)

val nfa_to_rexp = (nfa_to_rexp_Export.rs_nfa_to_rexp S.Sigma_dsarg2)

(* Statistic operations *)
val nfa_states_no = (rs_nfa_states_no sigma_linord):(NFA -> int)
val nfa_labels_no = (rs_nfa_labels_no sigma_linord):(NFA -> int)
val nfa_trans_no = (rs_nfa_trans_no sigma_linord) :(NFA -> int)
val nfa_final_no = (rs_nfa_final_no sigma_linord):(NFA -> int)
val nfa_initial_no = (rs_nfa_initial_no sigma_linord):(NFA -> int)

(* Language operations *)
val nfa_lang_is_empty = (rs_nfa_language_is_empty sigma_linord):(NFA -> bool)
val nfa_lang_is_univ = (rs_nfa_language_is_univ sigma_linord):(NFA -> bool)
val nfa_lang_is_subset = (rs_nfa_language_is_subset sigma_linord):(NFA -> NFA -> bool)
val nfa_lang_is_eq = (rs_nfa_language_is_eq sigma_linord):(NFA -> NFA -> bool)

(* Generating random automata *)

fun generate_random_NFAs number st_no label_no (den:real) (fden:real) =
let
  fun generate_labels 0 = []
    | generate_labels n = (S.int_to_Sigma (n-1)) :: generate_labels (n-1)

  fun generate_transition st_no l_no =
    (random_range 0 st_no, [S.int_to_Sigma (random_range 0 (label_no-1))], random_range 0 st_no)


  val labels = generate_labels label_no;
  val trans_no = Real.round(den * (Real.fromInt (st_no * label_no)));
  val accept_no = Real.round(fden * (Real.fromInt st_no)) + 1;

  fun generate_state_set 0 = []
    | generate_state_set n = (random_range 0 st_no) :: generate_state_set (n-1)

  fun replicate 0 x = []
    | replicate n x = x :: (replicate (n-1) x)

  fun generate_transitions () = map (fn f => f ()) (replicate trans_no 
      (fn () => generate_transition st_no label_no))

  fun generate_NFA () =
   nfa_construct ([], labels, generate_transitions(), [0], 
     generate_state_set accept_no)
in
  map (fn f => f ()) (replicate number generate_NFA)
end

fun generate_random_ICDFAs number st_no label_no (den:real) (fden:real) =
let
  val NFAs = generate_random_NFAs number st_no label_no (den:real) (fden:real)
in
  map nfa_determinise NFAs
end

end



structure int_Sigma_param : Sigma_param = struct
  type Sigma = int;
  type Sigma_Word = int list

  (*
  val nFA_states_nat = {states_prod_encode = Nat_Bijection.prod_encode, 
                      states_enumerate = (fn i => i)}:(int NFA.nFA_states)
  val ext_state_args = (Arith.equal_nat, HashCode.hashable_nat, nFA_states_nat)
  *)

  val equal_nat = {equal = fn (s1:int) => fn s2 => (s1 = s2)}
  val Sigma_dsarg = (equal_nat, Arith.linorder_nat);
  val Sigma_dsarg2 = (equal_nat, HashCode.hashable_nat, Arith.linorder_nat);


  fun Sigma_to_string s = Int.toString s
  fun Sigma_Word_create s = s
  fun int_to_Sigma n = n
end

structure NFA_string = NFA (string_Sigma_param)
structure NFA_int = NFA (int_Sigma_param)


fun rexp_to_string r = case r of 
    Regular_Exp.Zero => "{}" |
    Regular_Exp.One => "~" |
    Regular_Exp.Atom s => s |
    Regular_Exp.Plus (r1, r2) => rexp_to_string r1 ^ "|" ^ rexp_to_string r2 |
    Regular_Exp.Times (r1, r2) => 
      (case r1 of
        Regular_Exp.Plus (_, _) => "(" ^ rexp_to_string r1 ^ ")" |
        _ => rexp_to_string r1) ^ 
      (case r2 of
        Regular_Exp.Plus (_, _) => "(" ^ rexp_to_string r2 ^ ")" |
        _ => rexp_to_string r2) |
    Regular_Exp.Star r => 
      (case r of 
        Regular_Exp.Times (_, _) => "(" ^ rexp_to_string r ^ ")*" |
        Regular_Exp.Plus (_, _) => "(" ^ rexp_to_string r ^ ")*" |
        _ => rexp_to_string r ^ "*")






