(*
  Author: Manuel Eberl
*)
header {* \isaheader{Conversion of NFAs to Regular Expressions} *}
theory nfa_to_rexp
imports 
  Main
  "../../Automata_Malik_Tuerk/NFA"
  "../../Libs/Collections/Refine_Dflt_ICF"
  "../../Libs/Regular-Sets/Regular_Set"
  "../../Libs/Regular-Sets/Regular_Exp"
begin

text {*
  We verify an algorithm to convert NFAs to regular expressions.
  The Algorithm works by iteratively contracting the edges of the NFA 
  to regular expressions. The Refinement Framework is used to refine the
  algorithm to efficiently executable code.
*}

subsection "Basic Definitions"

datatype 'q gnfastate = Start | End | State 'q

instantiation gnfastate :: (hashable) hashable
begin
  definition [simp]: "hashcode q \<equiv> 
      (case q of Start \<Rightarrow> 0 | End \<Rightarrow> 1 | State q' \<Rightarrow> hashcode q')"
  (*definition [simp]: "bounded_hashcode n q \<equiv> 
      (case q of Start \<Rightarrow> 0 | End \<Rightarrow> 1 | State q' \<Rightarrow> (2 + hashcode q') mod n)"*)
  definition "def_hashmap_size = (\<lambda>_ :: (_ gnfastate) itself. 16)"
  instance by(intro_classes, simp_all split: gnfastate.split
      add: def_hashmap_size_gnfastate_def)
end

instantiation gnfastate :: (linorder) linorder
begin
  definition less_eq_def[simp]: "less_eq x y \<equiv> 
    case x of 
      Start \<Rightarrow> True |
      End \<Rightarrow> y \<noteq> Start |
      State x' \<Rightarrow> case y of
        State y' \<Rightarrow> x' \<le> y' |
        _ \<Rightarrow> False"

  definition less_def[simp]: "less x y \<equiv> 
    case x of 
      Start \<Rightarrow> y \<noteq> Start |
      End \<Rightarrow> y \<noteq> Start \<and> y \<noteq> End |
      State x' \<Rightarrow> case y of
        State y' \<Rightarrow> x' < y' |
        _ \<Rightarrow> False"

  instance apply (intro_classes)
    unfolding less_eq_def less_def
    apply (auto split: gnfastate.split gnfastate.split_asm)
    done
end

record ('q,'a) GNFA_rec =
  \<Q> :: "'q gnfastate set"          -- "The set of states"
  \<delta> :: "'q gnfastate \<Rightarrow> 'q gnfastate \<Rightarrow> 'a lang"  -- "The transition function"


locale GNFA =
  fixes \<A>::"('q,'a) GNFA_rec"
  assumes finite_\<Q>: "finite (\<Q> \<A>)" and
          start_correct: "Start \<in> \<Q> \<A>" "\<And>q. \<delta> \<A> q Start = {}" and
          end_correct: "End \<in> \<Q> \<A>" "\<And>q. \<delta> \<A> End q = {}" and
          no_epsilon: "\<And>u v. \<lbrakk>u \<noteq> Start; v \<noteq> End\<rbrakk> \<Longrightarrow> [] \<notin> \<delta> \<A> u v"
begin
lemmas GNFA_wf = finite_\<Q> start_correct end_correct no_epsilon
end


subsection {* Reachability and paths in GNFAs *}

inductive gnfa_is_reachable where
gnfa_eps[intro]: "u \<in> \<Q> \<A> \<Longrightarrow> gnfa_is_reachable \<A> u [] u" |
gnfa_step[intro]: "\<lbrakk>x = x1 @ x2; gnfa_is_reachable \<A>  u x1 v; w \<in> \<Q> \<A>; 
    x2 \<in> \<delta> \<A> v w\<rbrakk> \<Longrightarrow> gnfa_is_reachable \<A> u x w"

inductive_cases gnfa_is_reachableE[elim]: "gnfa_is_reachable \<A> u x v"


lemma gnfa_is_reachable_trans[trans, dest]:
  "\<lbrakk>gnfa_is_reachable \<A> u x1 v; gnfa_is_reachable \<A> v x2 w\<rbrakk>
      \<Longrightarrow> gnfa_is_reachable \<A> u (x1@x2) w"
proof (rotate_tac, induction rule: gnfa_is_reachable.induct)
  case (gnfa_step x2 x21 x22 v1 v2 w)
    have "x1 @ x2 = (x1 @ x21) @ x22" using `x2 = x21 @ x22` by simp
    thus ?case using gnfa_step by blast
qed simp

lemma gnfa_is_reachable_step[intro, dest]:
  "\<lbrakk>u \<in> \<Q> \<A>; v \<in> \<Q> \<A>; x \<in> \<delta> \<A> u v\<rbrakk> \<Longrightarrow> gnfa_is_reachable \<A> u x v" by blast

lemma gnfa_is_reachable_imp_in_\<Q>[dest, simp]:
  assumes "gnfa_is_reachable \<A> u x v"
  shows "u \<in> \<Q> \<A>" and "v \<in> \<Q> \<A>" using assms
  by (induction u x v rule: gnfa_is_reachable.induct, simp_all)

lemma gnfa_is_reachable_rev:
  assumes "x = x1 @ x2" "x1 \<in> \<delta> \<A> u v" "u \<in> \<Q> \<A>" "gnfa_is_reachable \<A> v x2 w"
  shows "gnfa_is_reachable \<A> u x w"
proof-
  from assms have "gnfa_is_reachable \<A> u x1 v" by blast
  also note `gnfa_is_reachable \<A> v x2 w`
  finally show ?thesis using `x = x1 @ x2` by simp
qed

lemma gnfa_is_reachable_revE:
  assumes "gnfa_is_reachable \<A> u x w"
  obtains "x = []" "w=u" "u \<in> \<Q> \<A>" |
          x1 x2 v where "x = x1 @ x2" "x1 \<in> \<delta> \<A> u v" "v \<in> \<Q> \<A>"
                        "gnfa_is_reachable \<A> v x2 w"
proof-
  case goal1
  from assms have "(x=[] \<and> w=u \<and> u \<in> \<Q> \<A>) \<or> (\<exists>x1 x2 v. x = x1 @ x2 \<and>
      gnfa_is_reachable \<A> v x2 w \<and> v \<in> \<Q> \<A> \<and> x1 \<in> \<delta> \<A> u v)" (is "?P \<or> ?Q")
  proof (induction \<A> u x w rule: gnfa_is_reachable.induct)
    case (gnfa_step x x1 x2 \<A> u v w)
      from gnfa_step.IH show ?case
      proof (rule disjE)
        assume "x1 = [] \<and> v = u \<and> u \<in> GNFA_rec.\<Q> \<A>"
        hence "x = x2 @ [] \<and> gnfa_is_reachable \<A> w [] w \<and> w \<in> \<Q> \<A> \<and> 
            x2 \<in> \<delta> \<A> u w" using gnfa_step by auto
        thus ?thesis by blast
      next
        assume "\<exists>x11 x12 v'. x1 = x11 @ x12 \<and> gnfa_is_reachable \<A> v' x12 v \<and> 
            v' \<in> \<Q> \<A> \<and> x11 \<in> \<delta> \<A> u v'" 
        then guess x11 x12 v' by (elim exE conjE)
        moreover from this have "gnfa_is_reachable \<A> v' (x12 @ x2) w" 
            using gnfa_step and gnfa_is_reachable_trans by blast
        ultimately have "x = x11 @ (x12 @ x2) \<and> 
            gnfa_is_reachable \<A> v' (x12 @ x2) w \<and> v' \<in> \<Q> \<A> \<and> x11 \<in> \<delta> \<A> u v'" 
            using gnfa_step by force
        thus ?thesis by blast
      qed  
  qed blast
  thus ?case using goal1 by auto
qed

lemma gnfa_is_reachable_revE2:
  assumes "gnfa_is_reachable \<A> u x w" and "u \<noteq> w"
  obtains x1 x2 v where "x = x1 @ x2" "x1 \<in> \<delta> \<A> u v" "v \<in> \<Q> \<A>"
                        "gnfa_is_reachable \<A> v x2 w"
  by (rule gnfa_is_reachable_revE[OF assms(1)], insert assms(2), simp_all)

inductive gnfa_is_reachable_in where
gnfa_in_eps[intro]: "u \<in> \<Q> \<A> \<Longrightarrow> gnfa_is_reachable_in \<A> u [] u 0" |
gnfa_in_step[intro]: "\<lbrakk>x = x1 @ x2; gnfa_is_reachable_in \<A> u x1 v n; 
    w \<in> \<Q> \<A>; x2 \<in> \<delta> \<A> v w\<rbrakk> \<Longrightarrow> gnfa_is_reachable_in \<A> u x w (Suc n)"

inductive_cases gnfa_is_reachable_inE[elim]:
    "gnfa_is_reachable_in \<A> u x v n"
inductive_cases gnfa_is_reachable_in_0E[elim!]: 
    "gnfa_is_reachable_in \<A> u x v 0"
inductive_cases gnfa_is_reachable_in_SucE[elim]: 
    "gnfa_is_reachable_in \<A> u x v (Suc n)"

lemma gnfa_is_reachable_in_imp_in_\<Q>[dest]:
  assumes "gnfa_is_reachable_in \<A> u x v n"
  shows "u \<in> \<Q> \<A>" and "v \<in> \<Q> \<A>" using assms
  by (induction rule: gnfa_is_reachable_in.induct, simp_all)

lemma gnfa_is_reachable_in_step:
  assumes "u \<in> \<Q> \<A>" and "v \<in> \<Q> \<A>" and "x \<in> \<delta> \<A> u v"
  shows "gnfa_is_reachable_in \<A> u x v (Suc 0)"
using assms by blast

lemma gnfa_is_reachable_iff_is_reachable_in:
    "gnfa_is_reachable \<A> u x v \<longleftrightarrow> (\<exists>n. gnfa_is_reachable_in \<A> u x v n)"
apply (rule iffI)
apply (induction u x v rule: gnfa_is_reachable.induct, blast, blast)
apply (elim exE, induct_tac rule: gnfa_is_reachable_in.induct, blast+)
done


subsection {* Operations on GNFAs *}

definition gnfa_add_transitions where
"gnfa_add_transitions \<A> \<delta>' \<equiv> \<lparr> \<Q> = \<Q> \<A>, \<delta> = (\<lambda>u v. \<delta> \<A> u v \<union> \<delta>' u v) \<rparr>"

lemma (in GNFA) gnfa_add_transitions_wf[simp,intro]:
  assumes "\<And>q. \<delta>' q Start = {}" and "\<And>q. \<delta>' End q = {}" and
          "\<And>u v. \<lbrakk>u \<noteq> Start; v \<noteq> End\<rbrakk> \<Longrightarrow> [] \<notin> \<delta>' u v"
  shows  "GNFA (gnfa_add_transitions \<A> \<delta>')"
  using assms and GNFA_wf unfolding gnfa_add_transitions_def GNFA_def 
  by simp_all

lemma gnfa_add_transitions_\<Q>[simp]: "\<Q> (gnfa_add_transitions \<A> \<delta>') = \<Q> \<A>"
    unfolding gnfa_add_transitions_def by simp

lemma gnfa_add_transitions_\<delta>[simp]: 
    "\<delta> (gnfa_add_transitions \<A> \<delta>') = (\<lambda>u v. \<delta> \<A> u v \<union> \<delta>' u v)"
    unfolding gnfa_add_transitions_def by simp

lemma gnfa_add_transitions_ge:
  shows "gnfa_is_reachable \<A> u x v \<Longrightarrow> 
      gnfa_is_reachable (gnfa_add_transitions \<A> \<delta>') u x v" 
proof (induction u x v rule: gnfa_is_reachable.induct)
  case (gnfa_step x x1 x2 \<A> u v w)
      let ?\<A>' = "gnfa_add_transitions \<A> \<delta>'"
      from gnfa_step.hyps have "w \<in> \<Q> ?\<A>'" and "x2 \<in> \<delta> ?\<A>' v w" by simp_all
      with `x = x1 @ x2` and gnfa_step.IH show ?case 
          using gnfa_is_reachable.intros(2) by blast
qed fastforce


text {* Computes, for all u, w, all the indirect transitions from u to w 
  that can be made by going through the intermediate state v.*}
definition gnfa_subsumed_transitions where
"gnfa_subsumed_transitions \<A> v \<equiv> \<lambda>u w. if u \<in> \<Q> \<A> - {v} \<and> w \<in> \<Q> \<A> - {v} then
    (let \<L>\<^sub>1 = \<delta> \<A> u v; \<L>\<^sub>2 = \<delta> \<A> v v; \<L>\<^sub>3 = \<delta> \<A> v w
     in \<L>\<^sub>1 @@ star \<L>\<^sub>2 @@ \<L>\<^sub>3) else {}"

lemma gnfa_is_reachable_loop:
  "\<lbrakk>x \<in> star (\<delta> \<A> q q); q \<in> \<Q> \<A>\<rbrakk> \<Longrightarrow> gnfa_is_reachable \<A> q x q"
proof (induction rule: star_induct)
  case Nil thus ?case by blast
next
  case (append u v)
    hence "gnfa_is_reachable \<A> q v q" using append by simp
    from gnfa_is_reachable_rev[OF _ append.hyps(1) append.prems this]
        show ?case by simp
qed

lemma gnfa_subsumed_transitions_is_reachable:
  assumes "x \<in> gnfa_subsumed_transitions \<A> v u w" and 
      "u \<in> \<Q> \<A> - {v}" and "v \<in> \<Q> \<A>" and "w \<in> \<Q> \<A> - {v}"
  shows "gnfa_is_reachable \<A> u x w"
proof-
  from assms have "u \<noteq> v" and "w \<noteq> v" and "u \<in> \<Q> \<A>" and "w \<in> \<Q> \<A>" by simp_all
  with assms(1) obtain x1 x2 x3 where x_split[simp]: 
      "x = x1 @ x2 @ x3" "x1 \<in> \<delta> \<A> u v"
      "x2 \<in> star (\<delta> \<A> v v)" "x3 \<in> \<delta> \<A> v w" 
      by (force simp: gnfa_subsumed_transitions_def elim!: concE)
  have "gnfa_is_reachable \<A> u x1 v" 
      using gnfa_is_reachable_step[OF `u \<in> \<Q> \<A>` `v \<in> \<Q> \<A>`] by simp
  also have "gnfa_is_reachable \<A> v x2 v"
      using gnfa_is_reachable_loop[OF _ `v \<in> \<Q> \<A>`] by simp
  also have "gnfa_is_reachable \<A> v x3 w"
      using gnfa_is_reachable_step[OF `v \<in> \<Q> \<A>` `w \<in> \<Q> \<A>`] by simp
  finally show ?thesis by simp
qed

abbreviation "gnfa_add_subsumed_transitions \<A> q \<equiv> 
    gnfa_add_transitions \<A> (gnfa_subsumed_transitions \<A> q)"

lemma (in GNFA) gnfa_add_subsumed_transitions_wf[simp,intro]:
  assumes "q \<noteq> Start" and "q \<noteq> End"
  shows  "GNFA (gnfa_add_subsumed_transitions \<A> q)"
  by (rule, unfold gnfa_subsumed_transitions_def, auto simp: GNFA_wf assms)

lemma gnfa_add_subsumed_transitions_le:
  assumes "q \<in> \<Q> \<A>" and
          "gnfa_is_reachable 
               (gnfa_add_subsumed_transitions \<A> q) u x w"
          (is "gnfa_is_reachable ?\<A>' _ _ _")
  shows "gnfa_is_reachable \<A> u x w"
using assms(2)
proof (induction ?\<A>' u x w rule: gnfa_is_reachable.induct, simp_all)
  fix u assume "u \<in> \<Q> \<A>" thus "gnfa_is_reachable \<A> u [] u" by blast
next
  fix x x1 x2 u v w
  assume "x = x1 @ x2" and "w \<in> \<Q> \<A>" and 
       u_to_v: "gnfa_is_reachable \<A> u x1 v" and 
       v_to_w_trans: "x2 \<in> \<delta> \<A> v w \<or> x2 \<in> gnfa_subsumed_transitions \<A> q v w"
  thus "gnfa_is_reachable \<A> u (x1 @ x2) w" 
  proof (cases rule: disjE[OF v_to_w_trans])

    (* case 1: old transition *)
    assume "x2 \<in> \<delta> \<A> v w" 
    from gnfa_is_reachable.intros(2)[OF `x = x1 @ x2` u_to_v `w \<in> \<Q> \<A>` this]
        show "gnfa_is_reachable \<A> u (x1 @ x2) w" using `x = x1 @ x2` by simp
  next

    (* case 2: new transition, show it can be built from old ones *)
    assume in_subsumed: "x2 \<in> gnfa_subsumed_transitions \<A> q v w"
    moreover from this have "v \<noteq> q" and "w \<noteq> q" 
        unfolding gnfa_subsumed_transitions_def by auto
    moreover from u_to_v have "v \<in> \<Q> \<A>" by simp_all
    ultimately have "gnfa_is_reachable \<A> v x2 w" using `q \<in> \<Q> \<A>` `w \<in> \<Q> \<A>`
        gnfa_subsumed_transitions_is_reachable  by fast
    thus "gnfa_is_reachable \<A> u (x1 @ x2) w" using u_to_v by blast
  qed
qed

lemma gnfa_add_subsumed_transitions_equiv[simp]:
  assumes "q \<in> \<Q> \<A>"
  shows "gnfa_is_reachable (gnfa_add_transitions \<A> 
       (gnfa_subsumed_transitions \<A> q)) u x w = gnfa_is_reachable \<A> u x w"
  by (rule iffI, fact gnfa_add_subsumed_transitions_le[OF assms],
      fact gnfa_add_transitions_ge)


definition gnfa_remove_state where
"gnfa_remove_state \<A> q \<equiv> \<lparr> \<Q> = \<Q> \<A> - {q}, \<delta> = \<delta> \<A> \<rparr>"

lemma gnfa_remove_state_\<Q>[simp]: "\<Q> (gnfa_remove_state \<A> q) = \<Q> \<A> - {q}"
    unfolding gnfa_remove_state_def by simp
lemma gnfa_remove_state_\<delta>[simp]: "\<delta> (gnfa_remove_state \<A> q) = \<delta> \<A>"
    unfolding gnfa_remove_state_def by simp

lemma (in GNFA) gnfa_remove_state_wf[simp,intro]:
  assumes "q \<noteq> Start" and "q \<noteq> End"
  shows "GNFA (gnfa_remove_state \<A> q)"
  by (unfold_locales, insert assms, simp_all add: GNFA_wf)


lemma gnfa_remove_state_le:
  assumes "gnfa_is_reachable (gnfa_remove_state \<A> q) u x v" 
          (is "gnfa_is_reachable ?\<A>' _ _ _")
  shows "gnfa_is_reachable \<A> u x v"
by (insert assms, induction ?\<A>' u x v rule: gnfa_is_reachable.induct, auto)

text {* For a sequence of steps @{text "u\<rightarrow>...\<rightarrow>w"} in the automaton with 
  @{text "u \<noteq> w"},
  this partitions the sequence into @{text "u\<rightarrow>...\<rightarrow>v\<rightarrow>w\<rightarrow>w\<rightarrow>...\<rightarrow>w\<rightarrow>w"} where 
  @{text "v \<noteq> w"}. *}
lemma gnfa_is_reachable_in_last_decompose:
  assumes "gnfa_is_reachable_in \<A> u x w n" and "u \<noteq> w"
  obtains v x1 x2 x3 n' where "n' < n" and "x = x1 @ x2 @ x3" and "v \<noteq> w" and 
      "gnfa_is_reachable_in \<A> u x1 v n'" and "x2 \<in> \<delta> \<A> v w" and 
      "x3 \<in> star (\<delta> \<A> w w)"
proof-
  case goal1
  have "\<exists>v x1 x2 x3 n'. n' < n \<and> x = x1 @ x2 @ x3 \<and> v \<noteq> w \<and> 
      gnfa_is_reachable_in \<A> u x1 v n' \<and> x2 \<in> \<delta> \<A> v w \<and> x3 \<in> star (\<delta> \<A> w w)"
  proof (insert assms, induction rule: gnfa_is_reachable_in.induct)
    case (gnfa_in_step x x1 x2 \<A> u v n w) thus ?case
    proof (cases "v = w")
      case False
        moreover have "x = x1 @ x2 @ []" using gnfa_in_step by simp
        ultimately show ?thesis using gnfa_in_step by blast

    next
      case True
        hence "u \<noteq> v" using gnfa_in_step by simp
        from gnfa_in_step.IH[OF this] and `v = w` 
            obtain v' x11 x12 x13 n' where
                    decomposition: "n' < n" "x1 = x11 @ x12 @ x13" "v' \<noteq> w"
                   "gnfa_is_reachable_in \<A> u x11 v' n'" "x12 \<in> \<delta> \<A> v' w"
                   "x13 \<in> star (\<delta> \<A> w w)" by blast
        with `x2 \<in> GNFA_rec.\<delta> \<A> v w` and `v = w` 
            have "x13 @ x2 \<in> star (\<delta> \<A> w w)" by simp
        moreover have "x = x11 @ x12 @ (x13 @ x2)" and "n' < Suc n"
            using gnfa_in_step and decomposition by simp_all
        ultimately show ?thesis using decomposition by blast
    qed
  qed simp
  with goal1 show ?thesis by blast
qed


lemma gnfa_remove_redundant_state_ge_helper:
  assumes "gnfa_is_reachable_in \<A> u x w n" and "u \<noteq> q" and "w \<noteq> q" and
          "\<And>u w. \<lbrakk>u \<in> \<Q> \<A> - {q}; w \<in> \<Q> \<A> - {q}\<rbrakk> \<Longrightarrow> 
               gnfa_subsumed_transitions \<A> q u w \<subseteq> 
                     \<delta> (gnfa_remove_state \<A> q) u w"
  shows "\<exists>n. gnfa_is_reachable_in (gnfa_remove_state \<A> q) u x w n"
        (is "\<exists>n. gnfa_is_reachable_in ?\<A>' _ _ _ _")
proof (insert assms, induction n arbitrary: x w rule: less_induct)
  case (less n)
    show ?case
    proof (cases rule: gnfa_is_reachable_inE[OF less.prems(1)])
      case 1
        with less.prems have "gnfa_is_reachable_in ?\<A>' u x w 0" by fastforce
        thus ?thesis ..
    next
      case (2 x1 x2 v n')
        show ?thesis
        proof (cases "v = q")
          case False
            from `v \<noteq> q` and less and 2 obtain n'' 
                where "gnfa_is_reachable_in ?\<A>' u x1 v n''" by blast
            moreover from 2 and less.prems
                have "x2 \<in> \<delta> ?\<A>' v w" and "w \<in> \<Q> ?\<A>'" by simp_all
            ultimately show ?thesis using 2 by blast
        next

          case True
            hence "u \<noteq> v" using less.prems by simp
            from gnfa_is_reachable_in_last_decompose[OF 
                `gnfa_is_reachable_in \<A> u x1 v n'` this]
            guess v' x11 x12 x13 n'' .
            note decomposition = this
            with `x2 \<in> \<delta> \<A> v w` and `v = q` and `w \<in> \<Q> \<A>` and `w \<noteq> q`
                have subsumed: "x12 @ x13 @ x2 \<in> 
                      gnfa_subsumed_transitions \<A> q v' w"
                unfolding gnfa_subsumed_transitions_def by (force simp: Let_def)
            have "v' \<in> \<Q> \<A> - {q}" "w \<in> \<Q> \<A> - {q}" "u \<noteq> q" "v' \<noteq> q"
                using decomposition(3,4) `v = q` less.prems by blast+
            from assms(4)[OF this(1,2)] and subsumed 
                have "x12 @ x13 @ x2 \<in> \<delta> ?\<A>' v' w" by blast

            moreover from `n'' < n'` and `n = Suc n'` have "n'' < n" by simp
            from less.IH[OF this decomposition(4) `u \<noteq> q` `v' \<noteq> q` assms(4)]
               obtain n''' where "gnfa_is_reachable_in ?\<A>' u x11 v' n'''" 
                   by blast
                
            ultimately have "gnfa_is_reachable_in ?\<A>' u x w (Suc n''')"
                using `w \<in> \<Q> \<A>` `w \<noteq> q` decomposition(2) 
                `x = x1 @ x2` by (force intro!: gnfa_in_step)
            thus ?thesis ..
        qed
    qed
qed

lemma gnfa_remove_redundant_state_ge:
  assumes "gnfa_is_reachable \<A> u x w" and "u \<noteq> q" and "w \<noteq> q" and
          "\<And>u w. \<lbrakk>u \<in> \<Q> \<A> - {q}; w \<in> \<Q> \<A> - {q}\<rbrakk> \<Longrightarrow> 
               gnfa_subsumed_transitions \<A> q u w \<subseteq> 
                     \<delta> (gnfa_remove_state \<A> q) u w"
  shows "gnfa_is_reachable (gnfa_remove_state \<A> q) u x w"
        (is "gnfa_is_reachable ?\<A>' _ _ _")
proof-
  from assms(1) and gnfa_is_reachable_iff_is_reachable_in[of \<A> u x w]
      obtain n where "gnfa_is_reachable_in \<A> u x w n" by blast
  from gnfa_remove_redundant_state_ge_helper[OF this assms(2-)]
      have "\<exists>n. gnfa_is_reachable_in ?\<A>' u x w n" by blast
  thus ?thesis using gnfa_is_reachable_iff_is_reachable_in[of ?\<A>' u x w] 
      by simp
qed

lemma gnfa_remove_redundant_state_equiv:
  assumes "u \<noteq> q" and "w \<noteq> q" and "\<And>u w. \<lbrakk>u \<in> \<Q> \<A> - {q}; w \<in> \<Q> \<A> - {q}\<rbrakk> \<Longrightarrow> 
               gnfa_subsumed_transitions \<A> q u w \<subseteq> 
                     \<delta> (gnfa_remove_state \<A> q) u w"
  shows "gnfa_is_reachable (gnfa_remove_state \<A> q) u x w = 
      gnfa_is_reachable \<A> u x w"
  by (rule iffI, fact gnfa_remove_state_le, 
      insert gnfa_remove_redundant_state_ge[OF _ assms], blast)


text {* Contraction of the automaton by "short-circuiting" ingoing 
    transitions, loops and outgoing transitions of the given state q 
    and then removing it. *}
abbreviation "gnfa_contract \<A> q \<equiv> gnfa_remove_state (
    gnfa_add_subsumed_transitions \<A> q) q"

lemma gnfa_contract_def: "gnfa_contract \<A> q =
    \<lparr>\<Q> = \<Q> \<A> - {q}, \<delta> = \<lambda>u v. \<delta> \<A> u v \<union> gnfa_subsumed_transitions \<A> q u v\<rparr>" 
    unfolding gnfa_remove_state_def gnfa_add_transitions_def by simp

lemma (in GNFA) gnfa_contract_wf[simp,intro]:
  assumes "q \<noteq> Start" and "q \<noteq> End"
  shows "GNFA (gnfa_contract \<A> q)"
  using assms by (intro GNFA.gnfa_remove_state_wf, blast)

lemma gnfa_subsumed_transitions_remove_state[simp]: 
    "gnfa_subsumed_transitions (gnfa_remove_state \<A> q) q = 
        gnfa_subsumed_transitions \<A> q"
  by (intro ext, simp add: gnfa_subsumed_transitions_def gnfa_remove_state_def)

lemma gnfa_contract_correct:
  assumes "q \<in> \<Q> \<A>" and "u \<in> \<Q> \<A> - {q}" and "v \<in> \<Q> \<A> - {q}"
  shows "gnfa_is_reachable (gnfa_contract \<A> q) u x v \<longleftrightarrow> 
      gnfa_is_reachable \<A> u x v"
proof-
  let ?P = "\<lambda>\<A>. gnfa_is_reachable \<A> u x v"
  let ?\<A>' = "gnfa_add_subsumed_transitions \<A> q"
  let ?\<A>'' = "gnfa_contract \<A> q"

  from gnfa_add_subsumed_transitions_equiv[OF assms(1)]
      have "?P \<A> \<longleftrightarrow> ?P ?\<A>'" ..
  also have "\<And>u v. \<lbrakk>u \<in> \<Q> ?\<A>' - {q}; v \<in> \<Q> ?\<A>' - {q}\<rbrakk>
      \<Longrightarrow> gnfa_subsumed_transitions ?\<A>' q u v \<subseteq> \<delta> ?\<A>'' u v"
      unfolding gnfa_add_transitions_def gnfa_subsumed_transitions_def by simp
  from assms and gnfa_remove_redundant_state_equiv[OF _ _ this]
      have "?P ?\<A>' \<longleftrightarrow> ?P ?\<A>''" by simp
  finally show ?thesis ..
qed


text {* The language that is accepted by the automaton. *}
definition gnfa_\<L> where "gnfa_\<L> \<A> = {x. gnfa_is_reachable \<A> Start x End}"



text {* Converts an NFA into an equivalent GNFA. *}
definition nfa_to_gnfa where
"nfa_to_gnfa A = \<lparr> \<Q> = {Start, End} \<union> State ` SemiAutomaton.\<Q> A, 
    \<delta> = \<lambda>u v. case u of 
                Start \<Rightarrow> (case v of
                  State v \<Rightarrow> if v \<in> \<I> A then {[]} else {} |
                  _ \<Rightarrow> {}) |
                End \<Rightarrow> {} |
                State u \<Rightarrow> (case v of 
                  Start \<Rightarrow> {} |
                  End \<Rightarrow> if u \<in> \<F> A then {[]} else {} |
                  State v \<Rightarrow> {[c] |c. (u,c,v) \<in> \<Delta> A}
                ) \<rparr>"

lemma nfa_to_gnfa_\<Q>[simp]:
  "\<Q> (nfa_to_gnfa \<A>) = {Start, End} \<union> State ` SemiAutomaton.\<Q> \<A>"
  unfolding nfa_to_gnfa_def by simp

lemma nfa_to_gnfa_\<delta>[simp]:
  "\<delta> (nfa_to_gnfa \<A>) u Start = {}"
  "\<delta> (nfa_to_gnfa \<A>) Start End = {}"
  "v' \<in> \<I> \<A> \<Longrightarrow> \<delta> (nfa_to_gnfa \<A>) Start (State v') = {[]}"
  "v' \<notin> \<I> \<A> \<Longrightarrow> \<delta> (nfa_to_gnfa \<A>) Start (State v') = {}"
  "\<delta> (nfa_to_gnfa \<A>) End v = {}"
  "u' \<in> \<F> \<A> \<Longrightarrow> \<delta> (nfa_to_gnfa \<A>) (State u') End = {[]}"
  "u' \<notin> \<F> \<A> \<Longrightarrow> \<delta> (nfa_to_gnfa \<A>) (State u') End = {}"
  "\<delta> (nfa_to_gnfa \<A>) (State u') (State v') = {[c] |c. (u',c,v') \<in> \<Delta> \<A>}"
unfolding nfa_to_gnfa_def by (cases u, simp_all)

lemma (in NFA) nfa_to_gnfa_wf[simp,intro]:
  "GNFA (nfa_to_gnfa \<A>)" (is "GNFA ?\<A>'")
proof (unfold_locales)
  fix u::"'q gnfastate" and v::"'q gnfastate" 
  assume "u \<noteq> Start" and "v \<noteq> End"
  thus "[] \<notin> \<delta> ?\<A>' u v" by (cases u, simp, simp, cases v, simp_all)
qed (simp_all add: finite_\<Q>)


text {* If the automaton consists only of Start and End, its entire language
  is in @{term "\<delta> \<A> Start end"}. *}
lemma (in GNFA) gnfa_\<L>_Start_End:
  assumes "\<Q> \<A> = {Start, End}"
  shows "gnfa_\<L> \<A> = \<delta> \<A> Start End"
unfolding gnfa_\<L>_def
proof (intro equalityI subsetI, simp_all)
  fix x assume "gnfa_is_reachable \<A> Start x End"
  thus "x \<in> \<delta> \<A> Start End"
  proof (rule gnfa_is_reachableE)
    fix x1 x2 v 
    assume "x = x1 @ x2" and "gnfa_is_reachable \<A> Start x1 v" and 
        "x2 \<in> \<delta> \<A> v End"
    moreover from this have "v = Start" using assms and end_correct by fast
    ultimately have "x = x2" using start_correct by blast
    thus "x \<in> \<delta> \<A> Start End" using `v = Start` and `x2 \<in> \<delta> \<A> v End` by simp
  qed simp
next
  fix x assume "x \<in> \<delta> \<A> Start End"
  thus "gnfa_is_reachable \<A> Start x End" 
      using start_correct and end_correct by blast
qed


lemma (in NFA) LTS_is_reachable_iff_gnfa_is_reachable:
  assumes "u \<in> SemiAutomaton.\<Q> \<A>" and "v \<in> SemiAutomaton.\<Q> \<A>"
  shows "LTS_is_reachable (\<Delta> \<A>) u x v \<longleftrightarrow> 
      gnfa_is_reachable (nfa_to_gnfa \<A>) (State u) x (State v)"
   (is "_ \<longleftrightarrow> gnfa_is_reachable ?\<A>' _ _ _")
proof (rule iffI)
  assume "LTS_is_reachable (\<Delta> \<A>) u x v"
  thus "gnfa_is_reachable ?\<A>' (State u) x (State v)" using assms(1,2)
  proof (induction x arbitrary: u)
    case (Cons c x)
      then obtain v' where v'_props: "(u,c,v') \<in> \<Delta> \<A>"
          "LTS_is_reachable (\<Delta> \<A>) v' x v" by force
      hence "v' \<in> SemiAutomaton.\<Q> \<A>"
          using \<Delta>_consistent by simp_all
      with Cons and v'_props 
          have "gnfa_is_reachable ?\<A>' (State v') x (State v)" by simp
      moreover from v'_props have "[c] \<in> \<delta> ?\<A>' (State u) (State v')" by simp
      moreover have "c # x = [c] @ x" and "State u \<in> \<Q> ?\<A>'" 
          using Cons by simp_all
      ultimately show ?case by (blast intro: gnfa_is_reachable_rev)
  qed force

next

  assume "gnfa_is_reachable ?\<A>' (State u) x (State v)"
  thus "LTS_is_reachable (\<Delta> \<A>) u x v"
  proof (induction ?\<A>' "(State u)" x "(State v)"  
      arbitrary: v rule: gnfa_is_reachable.induct)
    case (gnfa_step x x1 x2 v')
      from `x2 \<in> \<delta> ?\<A>' v' (State v)` have "v' \<noteq> End" by fastforce
      moreover from `gnfa_is_reachable ?\<A>' (State u) x1 v'`
          have "v' \<noteq> Start" by (rule gnfa_is_reachableE, simp, fastforce)
      ultimately obtain v'' where "v' = State v''" by (cases v', simp_all)
      with gnfa_step have "LTS_is_reachable (\<Delta> \<A>) u x1 v''" by simp
      moreover obtain c where "(v'',c,v) \<in> \<Delta> \<A>" and "x2 = [c]"
          using `x2 \<in> \<delta> ?\<A>' v' (State v)` and `v' = State v''` by force
      ultimately show "LTS_is_reachable (\<Delta> \<A>) u x v"
          using `x = x1 @ x2` by fastforce
  qed simp
qed

text {* If a word is in the language of a GNFA obtained from an NFA, 
  this obtains the corresponding initial state and finals state in 
  the original NFA. *}
lemma nfa_to_gnfa_Start_EndE:
  assumes "x \<in> gnfa_\<L> (nfa_to_gnfa \<A>)"
  obtains u v where "gnfa_is_reachable (nfa_to_gnfa \<A>) (State u) x (State v)"
     and "u \<in> \<I> \<A>" and "v \<in> \<F> \<A>"
proof-
  case goal1
  let ?\<A>' = "nfa_to_gnfa \<A>"
  from assms obtain v' x1 x2 where "gnfa_is_reachable ?\<A>' Start x1 v'" and
      "x2 \<in> \<delta> ?\<A>' v' End" and "x = x1 @ x2"
      unfolding gnfa_\<L>_def by blast
  moreover from `x2 \<in> \<delta> ?\<A>' v' End` obtain v where "v' = State v" and 
      "x2 = []" and "v \<in> \<F> \<A>" unfolding nfa_to_gnfa_def 
      by (cases v', auto split: split_if_asm)
  ultimately have "gnfa_is_reachable ?\<A>' Start x (State v)" and
      "Start \<noteq> State v" by simp_all
  from gnfa_is_reachable_revE2[OF this]
  obtain u' x1 x2 where "x = x1 @ x2" and "x1 \<in> \<delta> ?\<A>' Start u'" and 
      "gnfa_is_reachable ?\<A>' u' x2 (State v)" .
  moreover from `x1 \<in> \<delta> ?\<A>' Start u'` obtain u where "u' = State u" and
      "x1 = []" and "u \<in> \<I> \<A>" unfolding nfa_to_gnfa_def
      by (cases u', auto split: split_if_asm)
  ultimately have "gnfa_is_reachable ?\<A>' (State u) x (State v)" by simp
  with `u \<in> \<I> \<A>` and `v \<in> \<F> \<A>` and goal1 show ?thesis by simp
qed

lemma (in NFA) nfa_to_gnfa_correct[simp]:
  "gnfa_\<L> (nfa_to_gnfa \<A>) = \<L> \<A>" (is "gnfa_\<L> ?\<A>' = _")
proof (intro equalityI subsetI)
  fix x assume "x \<in> gnfa_\<L> ?\<A>'"
  then obtain u v where uv_props: "gnfa_is_reachable ?\<A>' (State u) x (State v)"
       "u \<in> \<I> \<A>" "v \<in> \<F> \<A>" by (blast elim: nfa_to_gnfa_Start_EndE)
  hence "u \<in> SemiAutomaton.\<Q> \<A>" and "v \<in> SemiAutomaton.\<Q> \<A>"
      using \<I>_consistent \<F>_consistent by blast+
  from LTS_is_reachable_iff_gnfa_is_reachable[OF this assms] and uv_props(1)
      have "LTS_is_reachable (\<Delta> \<A>) u x v" ..
  thus "x \<in> \<L> \<A>" using uv_props(2,3) 
      unfolding \<L>_def NFA_accept_def[abs_def] by blast
next

  interpret GNFA ?\<A>' using nfa_to_gnfa_wf[OF assms] .
  fix x assume "x \<in> \<L> \<A>"
  then obtain u v where uv_props: "u \<in> \<I> \<A>" "v \<in> \<F> \<A>"
      "LTS_is_reachable (\<Delta> \<A>) u x v" 
      unfolding \<L>_def NFA_accept_def[abs_def] by blast
  have "gnfa_is_reachable ?\<A>' Start [] (State u)"
      using start_correct uv_props \<I>_consistent by force
  also from uv_props have "u \<in> SemiAutomaton.\<Q> \<A>" and "v \<in> SemiAutomaton.\<Q> \<A>"
      using \<I>_consistent \<F>_consistent by blast+
  from LTS_is_reachable_iff_gnfa_is_reachable[OF this assms] and uv_props
      have "gnfa_is_reachable ?\<A>' (State u) x (State v)" by simp
  also have "gnfa_is_reachable ?\<A>' (State v) [] End"
      using end_correct uv_props \<F>_consistent by force
  finally show "x \<in> gnfa_\<L> ?\<A>'" unfolding gnfa_\<L>_def by simp
qed


  

subsection {* Conversion from NFA to RExp *}

text {* The invariant of the conversion algorithms main loop.
  The GNFA is well-formed, it is a subautomaton of teh original GNFA and
  all remaining states are connected with the same languages as initially. *}
definition nfa_to_rexp_invar where
"nfa_to_rexp_invar \<A> \<A>' \<equiv> GNFA \<A>' \<and> \<Q> \<A>' \<subseteq> \<Q> \<A> \<and> 
    (\<forall>u v x. u \<in> \<Q> \<A>' \<and> v \<in> \<Q> \<A>' \<longrightarrow> 
        (gnfa_is_reachable \<A>' u x v \<longleftrightarrow> gnfa_is_reachable \<A> u x v))"

lemma nfa_to_rexp_invarI[intro]:
  assumes "GNFA \<A>'" and "\<Q> \<A>' \<subseteq> \<Q> \<A>" and "\<And>u v x. \<lbrakk>u \<in> \<Q> \<A>'; v \<in> \<Q> \<A>'\<rbrakk> \<Longrightarrow> 
      gnfa_is_reachable \<A>' u x v \<longleftrightarrow> gnfa_is_reachable \<A> u x v"
  shows "nfa_to_rexp_invar \<A> \<A>'"
  unfolding nfa_to_rexp_invar_def by (intro conjI allI impI, 
      fact assms(1), fact assms(2), clarify, fact assms(3))

lemma nfa_to_rexp_invarD[dest]:
  assumes "nfa_to_rexp_invar \<A> \<A>'"
  shows "GNFA \<A>'" and "\<Q> \<A>' \<subseteq> \<Q> \<A>" and "\<And>u v x. \<lbrakk>u \<in> \<Q> \<A>'; v \<in> \<Q> \<A>'\<rbrakk> \<Longrightarrow> 
      gnfa_is_reachable \<A>' u x v \<longleftrightarrow> gnfa_is_reachable \<A> u x v"
  using assms unfolding nfa_to_rexp_invar_def by blast+

text {* Abstract algorithm that computes the language of a given NFA.
  This is, of course, not executable and will later be refined to use
  regular expressions as a concrete representation of these languages. *}
definition nfa_to_rexp_abstr where
"nfa_to_rexp_abstr \<A> \<equiv> do {
  \<A> \<leftarrow> SPEC (\<lambda>\<A>'. \<A>' = nfa_to_gnfa \<A>);
  \<A> \<leftarrow> 
    WHILE\<^sub>T\<^bsup>nfa_to_rexp_invar \<A>\<^esup> (\<lambda>\<A>. \<Q> \<A> \<noteq> {Start, End}) (\<lambda>\<A>. do {
      q \<leftarrow> SPEC (\<lambda>q. q \<in> \<Q> \<A> - {Start, End});
      \<A> \<leftarrow> SPEC (\<lambda>\<A>'. \<A>' = gnfa_contract \<A> q);
      RETURN \<A>
    }) \<A>;
  RETURN (\<delta> \<A> Start End)
}"

text {* The algorithm returns the language of the NFA *}
lemma (in NFA) nfa_to_rexp_abstr_correct:
  "nfa_to_rexp_abstr \<A> \<le> SPEC (\<lambda>L. L = \<L> \<A>)"
unfolding nfa_to_rexp_abstr_def
proof (intro refine_vcg)
  (* measure is number of states in the automaton *)
  show "wf (measure (card \<circ> \<Q>))" by simp
next
  (* invariant holds initially *)
  fix \<A>' assume "\<A>' = nfa_to_gnfa \<A>"
  with assms show "nfa_to_rexp_invar \<A>' \<A>'" by blast
next

  case (goal3 \<A>' \<A>'' q \<A>''' (*\<A>''''*))
    (*hence [simp]: "\<A>'''' = gnfa_contract \<A>'' q" by simp*)
    note inv = nfa_to_rexp_invarD[OF goal3(2)]
    then interpret GNFA \<A>'' by simp
    from goal3 have [simp]: "\<Q> \<A>''' = \<Q> \<A>'' - {q}" 
        unfolding gnfa_contract_def by simp

    (* the invariant is preserved *)
    have "nfa_to_rexp_invar \<A>' \<A>'''"
    proof (intro nfa_to_rexp_invarI) 
      from goal3 and wf show "GNFA \<A>'''" by blast
      from inv(2) show "\<Q> \<A>''' \<subseteq> \<Q> \<A>'" by (auto simp: gnfa_contract_def)
      fix u v x assume uv_assms: "u \<in> \<Q> \<A>'''" "v \<in> \<Q> \<A>'''"
      with `q \<in> \<Q> \<A>'' - {Start, End}` and gnfa_contract_correct
          have "gnfa_is_reachable (gnfa_contract \<A>'' q) u x v \<longleftrightarrow> 
              gnfa_is_reachable \<A>'' u x v"
          by (intro gnfa_contract_correct, simp_all)
      also have "gnfa_contract \<A>'' q = \<A>'''" using goal3 by simp
      also from uv_assms have "u \<in> \<Q> \<A>''" and "v \<in> \<Q> \<A>''"
          by (auto simp add: gnfa_contract_def)
      note inv(3)[OF this]
      finally show "gnfa_is_reachable \<A>''' u x v = 
          gnfa_is_reachable \<A>' u x v" .
    qed
    
    (* the measure decreases *)
    moreover have "card (\<Q> \<A>''') < card (\<Q> \<A>'')" using goal3 and 
        card_Diff1_less[OF finite_\<Q>] by simp
    
    ultimately show "nfa_to_rexp_invar \<A>' \<A>''' \<and> 
        (\<A>''',\<A>'') \<in> measure (card \<circ> \<Q>)" using goal3 by simp

next

  (* in the end, we have the automaton's language between Start and End. *)
  case (goal4 \<A>' \<A>'' (*L*))
    note inv = nfa_to_rexp_invarD[OF goal4(2)]
    from inv(1) interpret GNFA \<A>'' .
    from gnfa_\<L>_Start_End and goal4 have "GNFA_rec.\<delta> \<A>'' Start End = gnfa_\<L> \<A>''" by simp
    also from inv(3) have "gnfa_\<L> \<A>'' = gnfa_\<L> \<A>'" 
        unfolding gnfa_\<L>_def using start_correct(1) and end_correct(1) by blast
    also have "gnfa_\<L> \<A>' = \<L> \<A>" using assms and goal4 by simp
    finally show "GNFA_rec.\<delta> \<A>'' Start End = \<L> \<A>" .
qed




subsubsection {* Refinement step 1*}
text {* Implementation of @{term gnfa_contract} and @{term nfa_to_gnfa} *}

definition gnfa_contract_invar1 where
"gnfa_contract_invar1 \<A> q Q' \<A>' \<equiv> (\<Q> \<A>' = \<Q> \<A>) \<and> (\<forall>u v. 
    (u \<in> Q' \<longrightarrow> \<delta> \<A>' u v = \<delta> \<A> u v) \<and> 
    (u \<notin> Q' \<longrightarrow> \<delta> \<A>' u v = \<delta> \<A> u v \<union> gnfa_subsumed_transitions \<A> q u v))"

lemma gnfa_contract_invar1I[intro]:
  assumes "\<Q> \<A>' = \<Q> \<A>" and
          "\<And>u v. u \<in> Q' \<Longrightarrow> \<delta> \<A>' u v = \<delta> \<A> u v" and
          "\<And>u v. u \<notin> Q' \<Longrightarrow> \<delta> \<A>' u v = \<delta> \<A> u v \<union> 
                gnfa_subsumed_transitions \<A> q u v"
  shows "gnfa_contract_invar1 \<A> q Q' \<A>'"
  using assms unfolding gnfa_contract_invar1_def by simp

lemma gnfa_contract_invar1D[dest]:
  assumes "gnfa_contract_invar1 \<A> q Q' \<A>'"
  shows "\<Q> \<A>' = \<Q> \<A>" and
        "\<And>u v. u \<in> Q' \<Longrightarrow> \<delta> \<A>' u v = \<delta> \<A> u v" and
        "\<And>u v. u \<notin> Q' \<Longrightarrow> \<delta> \<A>' u v = \<delta> \<A> u v \<union> 
              gnfa_subsumed_transitions \<A> q u v"
  using assms unfolding gnfa_contract_invar1_def by simp_all


definition gnfa_contract_invar2 where
"gnfa_contract_invar2 \<equiv> \<lambda>\<A> \<A>' q u Q' \<A>''. \<Q> \<A>'' = \<Q> \<A> \<and>
  (\<forall>v. (\<forall>u'. u' \<noteq> u \<longrightarrow> \<delta> \<A>'' u' v = \<delta> \<A>' u' v) \<and>
    (v \<in> Q' \<longrightarrow> \<delta> \<A>'' u v = \<delta> \<A>' u v) \<and> 
    (v \<notin> Q' \<longrightarrow> \<delta> \<A>'' u v = \<delta> \<A>' u v \<union> gnfa_subsumed_transitions \<A> q u v))"

lemma gnfa_contract_invar2I[intro]:
  assumes "\<Q> \<A>'' = \<Q> \<A>" and
          "\<And>u' v. u' \<noteq> u \<Longrightarrow> \<delta> \<A>'' u' v = \<delta> \<A>' u' v" and
          "\<And>v. v \<in> Q' \<Longrightarrow> \<delta> \<A>'' u v = \<delta> \<A>' u v" and
          "\<And>v. v \<notin> Q' \<Longrightarrow> \<delta> \<A>'' u v = \<delta> \<A>' u v \<union> 
                gnfa_subsumed_transitions \<A> q u v"
  shows "gnfa_contract_invar2 \<A> \<A>' q u Q' \<A>''"
  using assms unfolding gnfa_contract_invar2_def by simp

lemma gnfa_contract_invar2D[dest]:
  assumes "gnfa_contract_invar2 \<A> \<A>' q u Q' \<A>''"
  shows "\<Q> \<A>'' = \<Q> \<A>" and
        "\<And>u' v. u' \<noteq> u \<Longrightarrow> \<delta> \<A>'' u' v = \<delta> \<A>' u' v" and
        "\<And>v. v \<in> Q' \<Longrightarrow> \<delta> \<A>'' u v = \<delta> \<A>' u v" and
        "\<And>v. v \<notin> Q' \<Longrightarrow> \<delta> \<A>'' u v = \<delta> \<A>' u v \<union> 
              gnfa_subsumed_transitions \<A> q u v"
  using assms unfolding gnfa_contract_invar2_def by simp_all


abbreviation "gnfa_contract_impl_update_\<delta> \<equiv> \<lambda>\<A> u q v u' v'.
     (if u'=u \<and> v'=v then \<delta> \<A> u v \<union> \<delta> \<A> u q @@ star (\<delta> \<A> q q) @@ \<delta> \<A> q v 
                       else \<delta> \<A> u' v')"

definition gnfa_contract_impl where
"gnfa_contract_impl \<A> q \<equiv> do {
  \<A> \<leftarrow> SPEC (\<lambda>\<A>'. \<A>' = gnfa_remove_state \<A> q);
  \<A> \<leftarrow> FOREACH\<^bsup>gnfa_contract_invar1 \<A> q\<^esup>  {u\<in>\<Q> \<A>. \<delta> \<A> u q \<noteq> {}} (\<lambda>u \<A>'.
    FOREACH\<^bsup>gnfa_contract_invar2 \<A> \<A>' q u\<^esup> {v\<in>\<Q> \<A>. \<delta> \<A> q v \<noteq> {}} (\<lambda>v \<A>'. 
        RETURN \<lparr> \<Q> = \<Q> \<A>', \<delta> = gnfa_contract_impl_update_\<delta> \<A>' u q v \<rparr>) \<A>'
  ) \<A>;
  ASSERT (GNFA \<A>);
  RETURN \<A>
}"



lemma (in GNFA) gnfa_contract_impl_correct:
  assumes "q \<noteq> Start" and "q \<noteq> End"
  shows "gnfa_contract_impl \<A> q \<le> SPEC (\<lambda>\<A>'. \<A>' = gnfa_contract \<A> q)"
unfolding gnfa_contract_impl_def
proof (intro refine_vcg)
  case goal1 thus ?case using finite_\<Q> by simp
next
  case goal2 show ?case by (rule gnfa_contract_invar1I, simp, simp,
      unfold gnfa_subsumed_transitions_def, clarsimp)
next
  case goal3 thus ?case using finite_\<Q> by simp
next
  case goal4
    note inv = gnfa_contract_invar1D(1)[OF goal4(4)]
    thus ?case by (intro gnfa_contract_invar2I, simp, simp, simp,
        unfold gnfa_subsumed_transitions_def, clarsimp)
next
  case (goal5 \<A> u it\<^sub>u \<A>' v it\<^sub>v \<A>'' (*\<A>'''*))
    (*from goal5 have [simp]: "\<A>''' = 
        \<lparr> \<Q> = \<Q> \<A>'', \<delta> = gnfa_contract_impl_update_\<delta> \<A>'' u q v \<rparr>" by simp*)
    from goal5(2,3,5,6) have uv_in_Q: "u \<in> \<Q> \<A>" "v \<in> \<Q> \<A>"  by auto
    note invu = gnfa_contract_invar1D[OF goal5(4)]
    note invv = gnfa_contract_invar2D[OF goal5(7)]
    from goal5 have "q \<notin> it\<^sub>u" "q \<noteq> u" "q \<notin> it\<^sub>v" "q \<noteq> v" by auto
    with invu invv `u \<in> it\<^sub>u` `v \<in> it\<^sub>v` have q_props: "\<delta> \<A>'' u q = \<delta> \<A> u q" 
        "\<delta> \<A>'' q q = \<delta> \<A> q q" "\<And>v'. \<delta> \<A>'' q v' = \<delta> \<A> q v'"
        by (simp_all add: gnfa_subsumed_transitions_def gnfa_remove_state_def)
    show ?case proof 
      fix v' assume v'_props: "v' \<notin> it\<^sub>v - {v}"
      show "\<delta> \<lparr> \<Q> = \<Q> \<A>'', \<delta> = gnfa_contract_impl_update_\<delta> \<A>'' u q v \<rparr> u v' 
          = \<delta> \<A>' u v' \<union> gnfa_subsumed_transitions \<A> q u v'"
        apply (cases "v = v'", insert uv_in_Q goal5(1,5) v'_props invv)
        apply (simp_all add: gnfa_subsumed_transitions_def q_props)
        done
    qed (simp_all add: invv)
next
  case (goal6 \<A> u it\<^sub>u \<A>' \<A>'')
    note invu = gnfa_contract_invar1D[OF goal6(4)]
    note invv = gnfa_contract_invar2D[OF goal6(5)]
    show ?case proof
      fix u' v' assume "u' \<in> it\<^sub>u - {u}" 
      with invu(2) invv(2) show "\<delta> \<A>'' u' v' = \<delta> \<A> u' v'" by simp
    next
      fix u' v' assume u'_props: "u' \<notin> it\<^sub>u - {u}"
      thus "\<delta> \<A>'' u' v' = \<delta> \<A> u' v' \<union> gnfa_subsumed_transitions \<A> q u' v'"
        apply (cases "u' = u")
        apply (insert invu(2) invv(4) `u \<in> it\<^sub>u`,
            auto simp: gnfa_subsumed_transitions_def)[1]      
        apply (simp add: u'_props invu(3) invv(2))
      done
    qed (simp add: invu invv)
next
  case (goal7 \<A>' \<A>'')
    note inv = gnfa_contract_invar1D(1,3)[OF goal7(2)]
    with goal7 have "\<delta> \<A>'' = (\<lambda>u v. \<delta> \<A> u v \<union> 
              gnfa_subsumed_transitions \<A> q u v)"
        by (intro ext, simp)
    with inv and goal7 have "\<A>'' = gnfa_contract \<A> q" by simp
    thus ?case using assms by simp
next    
  case (goal8 \<A>' \<A>'' (*\<A>'''*))
    note inv = gnfa_contract_invar1D(1,3)[OF goal8(2)]
    with goal8(1) have "\<delta> \<A>'' = (\<lambda>u v. \<delta> \<A> u v \<union> 
              gnfa_subsumed_transitions \<A> q u v)"
        by (intro ext, simp)
    with inv and goal8 show ?case by simp
qed

lemma gnfa_contract_impl_correct':
  fixes q\<^sub>1::"'q gnfastate" and \<A>\<^sub>1::"('q,'a,_) GNFA_rec_scheme"
  assumes "GNFA \<A>\<^sub>1" and "q\<^sub>1 \<noteq> Start" and "q\<^sub>1 \<noteq> End" and 
          "(q\<^sub>1,q\<^sub>2) \<in> Id" and "(\<A>\<^sub>1,\<A>\<^sub>2) \<in> Id"
  shows "gnfa_contract_impl \<A>\<^sub>1 q\<^sub>1 \<le> SPEC (\<lambda>\<A>'. \<A>' = gnfa_contract \<A>\<^sub>2 q\<^sub>2)"
  using assms and GNFA.gnfa_contract_impl_correct by blast

definition nfa_to_gnfa_invar1 where
"nfa_to_gnfa_invar1 \<A> Q \<equiv> \<lambda>I' \<A>'. (\<Q> \<A>' = Q \<and> \<delta> \<A>' = (\<lambda>u v. (case u of 
                Start \<Rightarrow> (case v of
                  State v \<Rightarrow> if v \<in> \<I> \<A> - I' then {[]} else {} |
                  _ \<Rightarrow> {}) |
                End \<Rightarrow> {} |
                State u \<Rightarrow> {})))"

definition nfa_to_gnfa_invar2 where
"nfa_to_gnfa_invar2 \<A> Q F' \<A>' \<equiv> (\<Q> \<A>' = Q \<and> \<delta> \<A>' = (\<lambda>u v. (case u of 
                Start \<Rightarrow> (case v of
                  State v \<Rightarrow> if v \<in> \<I> \<A> then {[]} else {} |
                  _ \<Rightarrow> {}) |
                End \<Rightarrow> {} |
                State u \<Rightarrow> (case v of 
                  Start \<Rightarrow> {} |
                  End \<Rightarrow> if u \<in> \<F> \<A> - F' then {[]} else {} |
                  State v \<Rightarrow> {}
                ))))"

definition nfa_to_gnfa_invar3 where
"nfa_to_gnfa_invar3 \<A> Q \<Delta>' \<A>' \<equiv> (\<Q> \<A>' = Q \<and> \<delta> \<A>' = (\<lambda>u v. (case u of 
                Start \<Rightarrow> (case v of
                  State v \<Rightarrow> if v \<in> \<I> \<A> then {[]} else {} |
                  _ \<Rightarrow> {}) |
                End \<Rightarrow> {} |
                State u \<Rightarrow> (case v of 
                  Start \<Rightarrow> {} |
                  End \<Rightarrow> if u \<in> \<F> \<A> then {[]} else {} |
                  State v \<Rightarrow> {[c] |c. (u,c,v) \<in> \<Delta> \<A> - \<Delta>'}
                ))))"

 
definition nfa_to_gnfa_impl where
"nfa_to_gnfa_impl \<A> = do {
  \<A>' \<leftarrow> SPEC (\<lambda>\<A>'. \<A>' = \<lparr> \<Q> = {Start, End} \<union> State ` SemiAutomaton.\<Q> \<A>,
      \<delta> = \<lambda>u v. {} \<rparr>);
  \<A>'' \<leftarrow> FOREACH\<^bsup>nfa_to_gnfa_invar1 \<A> (\<Q> \<A>')\<^esup> (\<I> \<A>) (\<lambda>v \<A>'. 
        RETURN \<lparr> \<Q> = \<Q> \<A>', \<delta> = \<lambda>u' v'.
            if u'=Start \<and> v'=State v then {[]} else \<delta> \<A>' u' v' \<rparr>) \<A>';
  \<A>'' \<leftarrow> FOREACH\<^bsup>nfa_to_gnfa_invar2 \<A> (\<Q> \<A>')\<^esup> (\<F> \<A>) (\<lambda>u \<A>'. 
        RETURN \<lparr> \<Q> = \<Q> \<A>', \<delta> = \<lambda>u' v'. 
            if u'=State u \<and> v'=End then {[]} else \<delta> \<A>' u' v' \<rparr>) \<A>'';
  \<A>'' \<leftarrow> FOREACH\<^bsup>nfa_to_gnfa_invar3 \<A> (\<Q> \<A>')\<^esup> (\<Delta> \<A>) (\<lambda>(u,c,v) \<A>'. 
        RETURN \<lparr> \<Q> = \<Q> \<A>', \<delta> = \<lambda>u' v'. if u'=State u \<and> v'=State v then 
            insert ([c]) (\<delta> \<A>' u' v') else \<delta> \<A>' u' v' \<rparr>) \<A>'';
  RETURN \<A>''
}"


lemma (in NFA) nfa_to_gnfa_impl_correct:
  "nfa_to_gnfa_impl \<A> \<le> SPEC (\<lambda>\<A>'. \<A>' = nfa_to_gnfa \<A>)"
unfolding nfa_to_gnfa_impl_def
proof (intro refine_vcg)
  case goal1 show ?case using finite_\<I> . next
  case goal2 thus ?case unfolding nfa_to_gnfa_invar1_def
      by (simp, intro ext, simp split: gnfastate.split) next
  case goal3 thus ?case unfolding nfa_to_gnfa_invar1_def
      by (simp, intro ext, auto split: gnfastate.split) next
  case goal4 show ?case using finite_\<F> . next
  case goal5 thus ?case unfolding nfa_to_gnfa_invar1_def nfa_to_gnfa_invar2_def
      by (simp, intro ext, simp split: gnfastate.split) next
  case goal6 thus ?case unfolding nfa_to_gnfa_invar2_def
      by (simp, intro ext, auto split: gnfastate.split) next
  case goal7 show ?case using finite_\<Delta> . next
  case goal8 thus ?case unfolding nfa_to_gnfa_invar2_def nfa_to_gnfa_invar3_def
      by (simp, intro ext, simp split: gnfastate.split) next
  case goal9 thus ?case unfolding nfa_to_gnfa_invar3_def[abs_def]
      by (simp split: prod.split, clarsimp, intro ext, 
          auto split: gnfastate.split)
next
  case (goal10 d _ _ \<A>' (*\<A>''*))
    hence "\<Q> \<A>' = \<Q> (nfa_to_gnfa \<A>)"
        unfolding nfa_to_gnfa_invar3_def by simp
    moreover have "\<delta> \<A>' = \<delta> (nfa_to_gnfa \<A>)" using goal10(4) 
        unfolding nfa_to_gnfa_def nfa_to_gnfa_invar3_def 
        by (intro ext, simp split: gnfastate.split)
    ultimately show ?case using goal10 by simp
qed


definition nfa_to_rexp_impl :: "('q,'a,_) NFA_rec_scheme \<Rightarrow> 'a lang nres" where
"nfa_to_rexp_impl \<A> \<equiv> do {
  \<A> \<leftarrow> nfa_to_gnfa_impl \<A>;
  \<A> \<leftarrow> 
    WHILE\<^sub>T (\<lambda>\<A>. \<exists>q\<in>\<Q> \<A>. q \<noteq> Start \<and> q \<noteq> End) (\<lambda>\<A>. do {
      q \<leftarrow> SPEC (\<lambda>q. q \<in> \<Q> \<A> - {Start, End});
      \<A> \<leftarrow> gnfa_contract_impl \<A> q;
      RETURN \<A>
    }) \<A>;
  RETURN (\<delta> \<A> Start End)
}"

lemma (in NFA) nfa_to_rexp_impl_refine:
  "nfa_to_rexp_impl \<A> \<le> \<Down>Id (nfa_to_rexp_abstr \<A>)"
unfolding nfa_to_rexp_impl_def nfa_to_rexp_abstr_def
apply (refine_rcg Id_refine single_valued_Id)
apply (fact nfa_to_gnfa_impl_correct)
apply assumption
apply (blast intro: GNFA.GNFA_wf)
apply simp
apply (drule nfa_to_rexp_invarD(1), blast intro: gnfa_contract_impl_correct')[]
apply simp_all
done




subsubsection {* Refinement step 2 *}
text {* concretisation of @{text "GNFA"} to @{text "(Q,\<delta>,P,S)"}, 
  where @{text P} and @{text S} are
  predecessor and successor maps *}


definition "gnfa_\<alpha> \<equiv> \<lambda>(Q,\<delta>,_,_). \<lparr> \<Q> = {Start, End} \<union> State`Q, \<delta> = \<delta> \<rparr>"
definition "gnfa_invar \<equiv> \<lambda>(Q,\<delta>,P,S). let \<A> = gnfa_\<alpha> (Q,\<delta>,P,S) in
    GNFA \<A> \<and> (\<forall>q\<in>\<Q> \<A>. P q = Some {u\<in>\<Q> \<A>. \<delta> u q \<noteq> {}} \<and>
                       S q = Some {v\<in>\<Q> \<A>. \<delta> q v \<noteq> {}})"

lemma gnfa_invarI[intro]:
  fixes \<delta>
  assumes "GNFA (gnfa_\<alpha> (Q,\<delta>,P,S))" and
          "\<And>q. q\<in>\<Q> (gnfa_\<alpha> (Q,\<delta>,P,S)) \<Longrightarrow> 
               P q = Some {u\<in>\<Q> (gnfa_\<alpha> (Q,\<delta>,P,S)). \<delta> u q \<noteq> {}}"
          "\<And>q. q\<in>\<Q> (gnfa_\<alpha> (Q,\<delta>,P,S)) \<Longrightarrow>
               S q = Some {v\<in>\<Q> (gnfa_\<alpha> (Q,\<delta>,P,S)). \<delta> q v \<noteq> {}}"
  shows "gnfa_invar (Q,\<delta>,P,S)"
  using assms unfolding gnfa_invar_def by simp

abbreviation "gnfa_refrel \<equiv> br gnfa_\<alpha> gnfa_invar"

lemma single_valued_gnfa_refrel: "single_valued gnfa_refrel" 
  by (fact br_sv)

lemma gnfa_refrel_imp_GNFA[simp,dest]:
  assumes "(\<A>\<^sub>1,\<A>\<^sub>2) \<in> gnfa_refrel"
  shows "GNFA \<A>\<^sub>2"
  using assms unfolding gnfa_invar_def br_def by (cases \<A>\<^sub>1, simp add: Let_def)


lemma GNFA_PS_correct:
  fixes \<A>::"('q,'a,_) GNFA_rec_scheme" and \<delta>
  assumes "((Q,\<delta>,P,S), \<A>) \<in> gnfa_refrel" and "q \<in> \<Q> \<A>"
  shows "P q = Some {u\<in>\<Q> \<A>. GNFA_rec.\<delta> \<A> u q \<noteq> {}}" 
        "S q = Some {v\<in>\<Q> \<A>. GNFA_rec.\<delta> \<A> q v \<noteq> {}}" 
  using assms unfolding gnfa_\<alpha>_def gnfa_invar_def by (auto simp add: br_def)


definition gnfa_remove_state_invar1 where
"gnfa_remove_state_invar1 \<equiv> \<lambda>Q \<delta> P q it P'. (\<forall>v\<in>{Start,End} \<union> State`Q.
    (v \<in> it \<longrightarrow> P' v = P v) \<and> 
    (v \<notin> it \<longrightarrow> P' v = Some {u\<in>{Start,End}\<union>State`Q-{q}. \<delta> u v \<noteq>{}}))"

lemma gnfa_remove_state_invar1I[intro]:
  fixes \<delta>
  assumes "\<And>v. v\<in>{Start,End} \<union> State`Q \<Longrightarrow> v\<in>it \<Longrightarrow> P' v = P v" and
          "\<And>v. v\<in>{Start,End} \<union> State`Q \<Longrightarrow> v\<notin>it \<Longrightarrow> P' v = 
               Some {u\<in>{Start,End}\<union>State`Q-{q}. \<delta> u v \<noteq>{}}"
  shows "gnfa_remove_state_invar1 Q \<delta> P q it P'"
  using assms unfolding gnfa_remove_state_invar1_def by simp

lemma gnfa_remove_state_invar1D[dest]:
  fixes \<delta>
  assumes "gnfa_remove_state_invar1 Q \<delta> P q it P'"
  shows "\<And>v. v\<in>{Start,End} \<union> State`Q \<Longrightarrow> v\<in>it \<Longrightarrow> P' v = P v" and
        "\<And>v. v\<in>{Start,End} \<union> State`Q \<Longrightarrow> v\<notin>it \<Longrightarrow> P' v = 
             Some {u\<in>{Start,End}\<union>State`Q-{q}. \<delta> u v \<noteq>{}}"
  using assms unfolding gnfa_remove_state_invar1_def by blast+

definition gnfa_remove_state_invar2 where
"gnfa_remove_state_invar2 \<equiv> \<lambda>Q \<delta> S q it S'. (\<forall>u\<in>{Start,End} \<union> State`Q. 
    (u \<in> it \<longrightarrow> S' u = S u) \<and> 
    (u \<notin> it \<longrightarrow> S' u = Some {v\<in>{Start,End}\<union>State`Q-{q}. \<delta> u v \<noteq>{}}))"

lemma gnfa_remove_state_invar2I[intro]:
  fixes \<delta>
  assumes "\<And>u. u\<in>{Start,End} \<union> State`Q \<Longrightarrow> u\<in>it \<Longrightarrow> S' u = S u" and
          "\<And>u. u\<in>{Start,End} \<union> State`Q \<Longrightarrow> u\<notin>it \<Longrightarrow> S' u = 
               Some {v\<in>{Start,End}\<union>State`Q-{q}. \<delta> u v \<noteq>{}}"
  shows "gnfa_remove_state_invar2 Q \<delta> S q it S'"
  using assms unfolding gnfa_remove_state_invar2_def by simp

lemma gnfa_remove_state_invar2D[dest]:
  fixes \<delta>
  assumes "gnfa_remove_state_invar2 Q \<delta> S q it S'"
  shows "\<And>u. u\<in>{Start,End} \<union> State`Q \<Longrightarrow> u\<in>it \<Longrightarrow> S' u = S u" and
        "\<And>u. u\<in>{Start,End} \<union> State`Q \<Longrightarrow> u\<notin>it \<Longrightarrow> S' u = 
             Some {v\<in>{Start,End}\<union>State`Q-{q}. \<delta> u v \<noteq>{}}"
  using assms unfolding gnfa_remove_state_invar2_def by blast+


definition "PS_add M u v \<equiv> case M u of None \<Rightarrow> M | 
    Some Mu \<Rightarrow> M(u \<mapsto> insert v Mu)"
definition "PS_remove M u v \<equiv> case M u of None \<Rightarrow> M | 
    Some Mu \<Rightarrow> M(u \<mapsto> Mu - {v})"
definition "PS_the M u \<equiv> case M u of None \<Rightarrow> {} | Some Mu \<Rightarrow> Mu"

definition gnfa_remove_state_impl2 where
"gnfa_remove_state_impl2 \<equiv> \<lambda>(Q,\<delta>,P,S) q.
  case q of Start \<Rightarrow> RETURN (Q,\<delta>,P,S) | End \<Rightarrow> RETURN (Q,\<delta>,P,S) | 
      State q' \<Rightarrow> do {
        P' \<leftarrow> FOREACH\<^bsup>gnfa_remove_state_invar1 Q \<delta> P q\<^esup> (PS_the S q) 
            (\<lambda>v P. RETURN (PS_remove P v q)) P;
        S' \<leftarrow> FOREACH\<^bsup>gnfa_remove_state_invar2 Q \<delta> S q\<^esup>  (PS_the P q) 
            (\<lambda>u S. RETURN (PS_remove S u q)) S;
        RETURN (Q - {q'}, \<delta>, P', S')
      }"


abbreviation "gnfa_state_refrel \<equiv> br State (\<lambda>_. True)"
lemma single_valued_gnfa_state_refrel:
    "single_valued gnfa_state_refrel" by (fact br_sv)
lemma gnfa_state_refrel_simp[simp]: 
    "(q',q) \<in> gnfa_state_refrel \<longleftrightarrow> q = State q'" by (simp add: br_def)
lemma gnfa_state_refrelD[dest]: 
    "(q',q) \<in> gnfa_state_refrel \<Longrightarrow> q = State q'" by simp



lemma gnfa_remove_state_impl2_correct:
  fixes \<A>\<^sub>1::"('q,'a,_) GNFA_rec_scheme" and 
      \<delta>::"'q gnfastate \<Rightarrow> 'q gnfastate \<Rightarrow> 'a lang"
  assumes "(q',q) \<in> gnfa_state_refrel" and "q' \<in> Q" and 
      "((Q,\<delta>,P,S),\<A>\<^sub>2) \<in> gnfa_refrel"
  shows "gnfa_remove_state_impl2 (Q,\<delta>,P,S) (State q') \<le> 
      \<Down>gnfa_refrel (SPEC (\<lambda>\<A>'. \<A>' = gnfa_remove_state \<A>\<^sub>2 q))"
unfolding gnfa_remove_state_impl2_def
using assms apply (simp add: br_def)
apply (refine_rcg, simp)
proof (intro refine_vcg)
  case goal1
    interpret GNFA \<A>\<^sub>2 using assms(3) by blast
    from goal1 GNFA_PS_correct(2)[OF assms(3)] finite_\<Q> assms
        show ?case by (simp add: gnfa_\<alpha>_def PS_the_def)
next
  case goal2
    from assms have "q \<in> \<Q> \<A>\<^sub>2" unfolding gnfa_\<alpha>_def by (simp add: br_def)
    with GNFA_PS_correct[OF assms(3)]
        have PS_props: "S q = Some {v\<in>\<Q> \<A>\<^sub>2. GNFA_rec.\<delta> \<A>\<^sub>2 q v \<noteq> {}}"
           "\<And>v. v \<in> \<Q> \<A>\<^sub>2 \<Longrightarrow> P v = Some {u\<in>\<Q> \<A>\<^sub>2. GNFA_rec.\<delta> \<A>\<^sub>2 u v \<noteq> {}}" 
           by simp_all
    {
      fix v assume v_props: "v \<in> \<Q> \<A>\<^sub>2" "v \<notin> the (S q)"
      hence "q \<notin> the (P v)" using PS_props assms unfolding gnfa_\<alpha>_def by simp
      hence "P v = Some {u\<in>\<Q> \<A>\<^sub>2 - {q}. GNFA_rec.\<delta> \<A>\<^sub>2 u v \<noteq>{}}" 
          using PS_props(2)[OF  v_props(1)] by auto
    }
    thus ?case using assms unfolding gnfa_remove_state_invar1_def 
        gnfa_\<alpha>_def PS_the_def gnfa_invar_def by (auto simp: br_def)
next
  case (goal3 v it P' (*P''*))
    let ?P''="(PS_remove P' v (State q'))"
    from GNFA_PS_correct[OF assms(3)] assms
        have S_props: "S q = Some {v\<in>\<Q> \<A>\<^sub>2. GNFA_rec.\<delta> \<A>\<^sub>2 q v \<noteq> {}}"
        unfolding gnfa_\<alpha>_def by (simp add: br_def)
    note inv = gnfa_remove_state_invar1D[OF goal3(6)]
    have v_props: "v \<in> {Start,End} \<union> State`Q" 
        using goal3(4,5) S_props assms(1,3)
        unfolding gnfa_\<alpha>_def PS_the_def by (auto simp: br_def)
    with GNFA_PS_correct[OF assms(3)] assms 
        have P_props: "P v = Some {u\<in>\<Q> \<A>\<^sub>2. GNFA_rec.\<delta> \<A>\<^sub>2 u v \<noteq> {}}"
        unfolding gnfa_\<alpha>_def by (simp add: br_def)
    hence P''_props: "?P'' v = Some ({u\<in>\<Q> \<A>\<^sub>2-{State q'}. 
          GNFA_rec.\<delta> \<A>\<^sub>2 u v \<noteq> {}})"
        using (*goal3(7)*) inv(1)[OF v_props `v \<in> it`] 
        by (auto simp add: PS_remove_def)
    show ?case 
      apply (intro gnfa_remove_state_invar1I)
      apply (insert (*goal3(7)*) inv, simp add: PS_remove_def 
          split: option.split) []
      apply (rename_tac v', case_tac "v' = v")
      apply (insert assms, simp add: P''_props gnfa_\<alpha>_def br_def) []
      apply (insert (*goal3(7)*) inv, simp add: PS_remove_def br_def 
          split: option.split) []
      done
next
  case goal4
    interpret GNFA \<A>\<^sub>2 using assms(3) by blast
    from goal4 GNFA_PS_correct(1)[OF assms(3)] finite_\<Q> assms
        show ?case by (simp add: gnfa_\<alpha>_def PS_the_def)
next
  case goal5
    from assms have "q \<in> \<Q> \<A>\<^sub>2" unfolding gnfa_\<alpha>_def by (simp add: br_def)
    with GNFA_PS_correct[OF assms(3)]
        have PS_props: "P q = Some {u\<in>\<Q> \<A>\<^sub>2. GNFA_rec.\<delta> \<A>\<^sub>2 u q \<noteq> {}}"
             "\<And>u. u \<in> \<Q> \<A>\<^sub>2 \<Longrightarrow> S u = Some {v\<in>\<Q> \<A>\<^sub>2. GNFA_rec.\<delta> \<A>\<^sub>2 u v \<noteq> {}}" 
             by simp_all
    {
      fix u assume u_props: "u \<in> \<Q> \<A>\<^sub>2" "u \<notin> the (P q)"
      hence "q \<notin> the (S u)" using PS_props assms unfolding gnfa_\<alpha>_def by simp
      hence "S u = Some {v\<in>\<Q> \<A>\<^sub>2 - {q}. GNFA_rec.\<delta> \<A>\<^sub>2 u v \<noteq>{}}" 
          using PS_props(2)[OF  u_props(1)] by auto
    }
    thus ?case using assms unfolding gnfa_remove_state_invar2_def 
        gnfa_\<alpha>_def PS_the_def gnfa_invar_def by (auto simp: br_def)
next
  case (goal6 _ u it S' (*S''*))
    let ?S'' = "(PS_remove S' u (State q'))"
    from GNFA_PS_correct[OF assms(3)] assms
        have P_props: "P q = Some {u\<in>\<Q> \<A>\<^sub>2. GNFA_rec.\<delta> \<A>\<^sub>2 u q \<noteq> {}}"
        unfolding gnfa_\<alpha>_def by (simp add: br_def)
    note inv = gnfa_remove_state_invar2D[OF goal6(7)]
    have u_props: "u \<in> {Start,End} \<union> State`Q" 
        using goal6(5,6) P_props assms(1,3)
        unfolding gnfa_\<alpha>_def PS_the_def by (auto simp: br_def)
    with GNFA_PS_correct[OF assms(3)] assms 
        have S_props: "S u = Some {v\<in>\<Q> \<A>\<^sub>2. GNFA_rec.\<delta> \<A>\<^sub>2 u v \<noteq> {}}"
        unfolding gnfa_\<alpha>_def by (simp add: br_def)
    hence S''_props: "?S'' u = Some ({v\<in>\<Q> \<A>\<^sub>2-{State q'}. 
          GNFA_rec.\<delta> \<A>\<^sub>2 u v \<noteq> {}})"
        using (*goal6(8)*) inv(1)[OF u_props `u \<in> it`] 
        by (auto simp: PS_remove_def)
    show ?case 
      apply (intro gnfa_remove_state_invar2I)
      apply (insert (*goal6(8)*) inv, simp add: PS_remove_def 
          split: option.split) []
      apply (rename_tac u', case_tac "u' = u")
      apply (insert assms, simp add: S''_props gnfa_\<alpha>_def br_def) []
      apply (insert (*goal6(8)*) inv, simp add: PS_remove_def 
          split: option.split) []
      done
next
  case (goal7 P' S' (*\<A>\<^sub>1'*))
    let ?\<A>\<^sub>1' = "(Q - {q'}, \<delta>, P', S')"
    (*hence \<A>\<^sub>1'_props: "\<A>\<^sub>1' = (Q - {q'}, \<delta>, P', S')" by simp*)
    interpret GNFA \<A>\<^sub>2 using assms(3) by blast
    note invP = gnfa_remove_state_invar1D(2)[OF goal7(4)]
    note invS = gnfa_remove_state_invar2D(2)[OF goal7(5)]
    from goal7 have A: "gnfa_remove_state \<A>\<^sub>2 (State q') = gnfa_\<alpha> ?\<A>\<^sub>1'"
        unfolding gnfa_remove_state_def gnfa_\<alpha>_def by fastforce
    moreover from assms have "GNFA (gnfa_\<alpha> (Q, \<delta>, P, S))" 
        unfolding gnfa_invar_def by (simp add: Let_def br_def)
    from GNFA.gnfa_remove_state_wf[OF this]
        have "GNFA (gnfa_remove_state \<A>\<^sub>2 (State q'))" by simp
    ultimately have B: "GNFA (gnfa_\<alpha> ?\<A>\<^sub>1')" by simp
    {fix q assume "q\<in>\<Q> (gnfa_\<alpha> ?\<A>\<^sub>1')"
     hence "q \<in> {Start,End} \<union> State`Q" using assms(3) (*\<A>\<^sub>1'_props*)
         unfolding gnfa_\<alpha>_def by auto
     from invP[OF this] invS[OF this] (*\<A>\<^sub>1'_props*)
         have "P' q = Some {u\<in>\<Q> (gnfa_\<alpha> ?\<A>\<^sub>1'). \<delta> u q \<noteq> {}}"
              "S' q = Some {v\<in>\<Q> (gnfa_\<alpha> ?\<A>\<^sub>1'). \<delta> q v \<noteq> {}}"
         unfolding gnfa_\<alpha>_def by auto
    } note P'S'_props = this
    
    show ?case (*using \<A>\<^sub>1'_props*) apply (intro conjI)
       apply (insert A goal7(3), simp) []
       apply ((*subst \<A>\<^sub>1'_props,*) rule gnfa_invarI)
       apply (insert B , simp) []
       apply (insert P'S'_props, simp_all)
       done
qed



definition gnfa_contract_update_\<delta>_impl2 where 
"gnfa_contract_update_\<delta>_impl2 \<equiv> \<lambda>\<delta> u q v u' v'.
     (if u'=u \<and> v'=v then \<delta> u v \<union> \<delta> u q @@ star (\<delta> q q) @@ \<delta> q v 
                       else \<delta> u' v')"

definition gnfa_contract_impl2 where
"gnfa_contract_impl2 \<equiv> \<lambda>(Q,\<delta>,P,S) q. do {
  let Pq = PS_the P q - {q}; let Sq = PS_the S q - {q};
  (Q,\<delta>,P,S) \<leftarrow> gnfa_remove_state_impl2 (Q,\<delta>,P,S) q;
  (Q,\<delta>,P,S) \<leftarrow> FOREACH Pq (\<lambda>u (Q',\<delta>',P',S').
    FOREACH Sq (\<lambda>v (Q',\<delta>',P',S').
        RETURN (Q', gnfa_contract_update_\<delta>_impl2 \<delta>' u q v,
            PS_add P' v u, PS_add S' u v))
      (Q',\<delta>',P',S')
  ) (Q,\<delta>,P,S);
  ASSERT (GNFA (gnfa_\<alpha> (Q,\<delta>,P,S)));
  RETURN (Q,\<delta>,P,S)
}"


lemma gnfa_contract_impl2_refine:
  fixes \<A>::"('q,'a,_) GNFA_rec_scheme" and \<delta>
  assumes "(q',q) \<in> gnfa_state_refrel" and "q' \<in> Q" and 
        "((Q,\<delta>,P,S), \<A>) \<in> gnfa_refrel"
  shows "gnfa_contract_impl2 (Q,\<delta>,P,S) (State q') \<le> 
      \<Down>gnfa_refrel (gnfa_contract_impl \<A> q)"
unfolding gnfa_contract_impl2_def gnfa_contract_impl_def Let_def
apply (refine_rcg single_valued_gnfa_refrel inj_on_id)
apply (clarify, rule gnfa_remove_state_impl2_correct[OF assms(1-3)])
apply (insert GNFA_PS_correct(1)[OF assms(3)] assms, unfold gnfa_\<alpha>_def, 
    auto simp: PS_the_def br_def split: option.split) []
apply simp
apply (insert GNFA_PS_correct(2)[OF assms(3)] assms, unfold gnfa_\<alpha>_def, 
    auto simp: br_def PS_the_def  split: option.split) []
apply simp
defer
apply (simp add: br_def)
apply (clarsimp simp: br_def)
proof-
  case (goal1 Q1 \<delta>1 P1 S1 u it\<^sub>u Q2 \<delta>2 P2 S2 v it\<^sub>v Q3 \<delta>3 P3 S3)
    note inv1 = gnfa_contract_invar1D[OF goal1(5)]
    note inv2 = gnfa_contract_invar2D[OF goal1(9)]

    let ?\<A>1 = "gnfa_\<alpha> (Q1, \<delta>1, P1, S1)"    
    let ?\<A>2 = "gnfa_\<alpha> (Q2, \<delta>2, P2, S2)"    
    let ?\<A>3 = "gnfa_\<alpha> (Q3, \<delta>3, P3, S3)"
    from goal1(10) have \<A>1_GNFA: "GNFA ?\<A>1" 
        unfolding gnfa_invar_def Let_def by simp
    from goal1(12) have \<A>3_GNFA: "GNFA ?\<A>3" 
        unfolding gnfa_invar_def Let_def by simp

    from inv2(1) goal1(1) have "{Start, End} \<union> 
          State`Q1 = {Start, End} \<union> State`Q3"
        unfolding gnfa_\<alpha>_def gnfa_remove_state_def by simp
    hence "\<Q> ?\<A>1 = \<Q> ?\<A>3" using inv2(1) unfolding gnfa_\<alpha>_def by simp
    moreover have uv_in_Q1: "u \<in> \<Q> ?\<A>1" "v \<in> \<Q> ?\<A>1" using goal1 
        unfolding gnfa_\<alpha>_def by auto
    ultimately have uv_in_Q3[simp]: "u \<in> \<Q> ?\<A>3" "v \<in> \<Q> ?\<A>3" by simp_all

    have "GNFA_rec.\<delta> \<A> = \<delta>1" using goal1(1) 
        unfolding gnfa_\<alpha>_def gnfa_remove_state_def by simp
    hence A: "\<lparr>GNFA_rec.\<Q> = \<Q> ?\<A>3,
       \<delta> = gnfa_contract_impl_update_\<delta> ?\<A>3 u q v\<rparr> =
       gnfa_\<alpha> (Q3, gnfa_contract_update_\<delta>_impl2 \<delta>3 u
       (State q') v, PS_add P3 v u, PS_add S3 u v)"
   unfolding gnfa_\<alpha>_def gnfa_remove_state_def
       gnfa_contract_update_\<delta>_impl2_def using assms(1) by fastforce
    
    have "gnfa_contract_update_\<delta>_impl2 \<delta>3 u q v = 
        gnfa_contract_impl_update_\<delta> ?\<A>3 u q v"
        unfolding gnfa_\<alpha>_def gnfa_contract_update_\<delta>_impl2_def by force
    hence B: "GNFA  \<lparr>GNFA_rec.\<Q> = insert Start (insert End (State ` Q3)),
        \<delta> = gnfa_contract_update_\<delta>_impl2 \<delta>3 u (State q') v\<rparr>"
       using GNFA.GNFA_wf[OF \<A>1_GNFA] GNFA.GNFA_wf[OF \<A>3_GNFA] assms(1)
           by (auto simp: GNFA_def gnfa_\<alpha>_def)

    have PS_props: 
         "PS_the P q - {q} = {u \<in> \<Q> ?\<A>1. \<delta>1 u q \<noteq> {}}"
         "PS_the S q - {q} = {v \<in> \<Q> ?\<A>1. \<delta>1 q v \<noteq> {}}"
         using assms goal1(1) unfolding gnfa_invar_def Let_def gnfa_\<alpha>_def
         gnfa_remove_state_def PS_the_def by (auto simp: br_def)

    have P3S3_props: 
         "\<And>q. q \<in>\<Q> ?\<A>3 \<Longrightarrow> P3 q = Some {u \<in> \<Q> ?\<A>3. \<delta>3 u q \<noteq> {}}"
         "\<And>q. q \<in>\<Q> ?\<A>3 \<Longrightarrow> S3 q = Some {v \<in> \<Q> ?\<A>3. \<delta>3 q v \<noteq> {}}"
         using goal1(12) unfolding gnfa_invar_def Let_def by simp_all

    from goal1 have "q \<notin> it\<^sub>u" "q \<noteq> u" "q \<notin> it\<^sub>v" by auto
    from inv2(2)[OF this(2)] inv1(3)[OF this(1)] inv2(4)[OF this(3)] 
         inv1(2)[OF `u \<in> it\<^sub>u`] goal1(1)
    have \<delta>3_q_props: "\<delta>3 u q = \<delta>1 u q" "\<delta>3 q v = \<delta>1 q v"  by (simp_all add: 
          gnfa_\<alpha>_def gnfa_subsumed_transitions_def gnfa_remove_state_def)

    have \<delta>_uq_qv_nonempty: "\<delta>1 u q \<noteq> {}" "\<delta>1 q v \<noteq> {}" 
        using assms(1) goal1(2,3,6,7) PS_props by auto
    with \<delta>3_q_props have \<delta>3_uq_qv_nonempty: "\<delta>3 u q \<noteq> {}" "\<delta>3 q v \<noteq> {}"
        by simp_all
    hence C: "\<delta>3 u q @@ star (\<delta>3 q q) @@ \<delta>3 q v \<noteq> {}" by blast

    show ?case
      apply (intro conjI)
      apply (rule A)
      apply rule
      apply (unfold gnfa_\<alpha>_def, insert B, simp) []
      unfolding gnfa_contract_update_\<delta>_impl2_def
      apply (rename_tac v', case_tac "v' = v",
          insert assms(1) P3S3_props(1) C uv_in_Q3 \<delta>_uq_qv_nonempty,
          unfold gnfa_\<alpha>_def PS_the_def PS_add_def, force, simp) []
      apply (rename_tac u', case_tac "u' = u",
          insert assms(1) P3S3_props(2) C uv_in_Q3 \<delta>_uq_qv_nonempty,
          unfold gnfa_\<alpha>_def PS_the_def PS_add_def, force, simp) []
      done
qed




definition gnfa_initial_invar where
"gnfa_initial_invar \<A> it M \<equiv> 
      (\<forall>q \<in> State`(SemiAutomaton.\<Q> \<A> - it). M q = Some {})"

definition gnfa_initial_impl2 where
"gnfa_initial_impl2 \<A> \<equiv> do {
    M \<leftarrow> FOREACH\<^bsup>gnfa_initial_invar \<A>\<^esup> (SemiAutomaton.\<Q> \<A>) 
        (\<lambda>q M. RETURN (M(State q \<mapsto> {}))) Map.empty;
    let M = M(Start \<mapsto> {}, End \<mapsto> {});
    d \<leftarrow> SPEC (\<lambda>d. \<forall>u v. d u v = {});
    RETURN (SemiAutomaton.\<Q> \<A>, d, M, M)
}"

lemma (in NFA) gnfa_initial_impl2_correct:
  "gnfa_initial_impl2 \<A> \<le> \<Down>gnfa_refrel (SPEC (\<lambda>\<A>'. \<A>' = 
      \<lparr> \<Q> = {Start,End} \<union> State ` SemiAutomaton.\<Q> \<A>, \<delta> = \<lambda>u v. {} \<rparr>))"
unfolding gnfa_initial_impl2_def
apply (simp add: br_def, refine_rcg single_valued_gnfa_refrel)
apply (simp add: single_valued_def)
apply (intro refine_vcg)
apply (fact finite_\<Q>)
apply (simp add: gnfa_initial_invar_def)
apply (force simp: gnfa_initial_invar_def)
unfolding gnfa_initial_invar_def gnfa_invar_def Let_def
apply (force simp add: GNFA_def gnfa_\<alpha>_def finite_\<Q>)
done


definition nfa_to_gnfa_impl2 where
"nfa_to_gnfa_impl2 \<A> = do {
  \<A>' \<leftarrow> gnfa_initial_impl2 \<A>;
  \<A>'' \<leftarrow> FOREACH (\<I> \<A>) (\<lambda>v (Q, \<delta>, P, S). 
        RETURN (Q, \<lambda>u' v'. if u'=Start \<and> v'=State v then {[]} else \<delta> u' v',
            P(State v  \<mapsto> {Start}), PS_add S Start (State v))) \<A>';
  \<A>'' \<leftarrow> FOREACH (\<F> \<A>) (\<lambda>u (Q, \<delta>, P, S). 
        RETURN (Q, \<lambda>u' v'. if u'=State u \<and> v'=End then {[]} else \<delta> u' v',
            PS_add P End (State u), S(State u \<mapsto> {End}))) \<A>'';
  \<A>'' \<leftarrow> FOREACH (\<Delta> \<A>) (\<lambda>(u,c,v) (Q, \<delta>, P, S). 
        RETURN (Q, \<lambda>u' v'. if u'=State u \<and> v'=State v then 
                insert [c] (\<delta> u' v') else \<delta> u' v',
            PS_add P (State v) (State u), PS_add S (State u) (State v))) \<A>'';
  RETURN \<A>''
}"

lemma (in NFA) nfa_to_gnfa_impl2_aux1:
  fixes \<delta>
  assumes "v \<in> it" "it \<subseteq> \<I> \<A>" "nfa_to_gnfa_invar1 \<A> 
        (insert Start (insert End (State`SemiAutomaton.\<Q> \<A>))) it 
        \<lparr>\<Q>=insert Start (insert End (State`Q)), GNFA_rec.\<delta> = \<delta> \<rparr>"
      "gnfa_invar (Q, \<delta>, P, S)"
  shows "gnfa_invar (Q, \<lambda>u' v'. if u'=Start\<and>v'=State v then 
      {[]} else \<delta> u' v', P(State v \<mapsto> {Start}), PS_add S Start (State v))" 
      (is "gnfa_invar (Q, ?\<delta>', ?P', ?S')")
proof
  let ?\<A> = "gnfa_\<alpha> (Q,\<delta>,P,S)" and  ?\<A>' = "gnfa_\<alpha> (Q,?\<delta>',?P',?S')"
  have [simp]: "\<Q> ?\<A>' = \<Q> ?\<A>" unfolding gnfa_\<alpha>_def by simp
  from assms have "GNFA ?\<A>" unfolding gnfa_invar_def Let_def by simp
  thus "GNFA ?\<A>'" by (simp_all add: GNFA_def gnfa_\<alpha>_def)

  hence [simp, intro]: "Start \<in> \<Q> ?\<A>" unfolding GNFA_def by simp

  fix q assume "q \<in> \<Q> ?\<A>'" hence "q \<in> \<Q> ?\<A>" by simp
  with assms have "P q = Some {u\<in>\<Q> ?\<A>. \<delta> u q \<noteq> {}}" 
      unfolding gnfa_invar_def by (simp add: Let_def)
  moreover from assms have "\<And>u v. u \<noteq> Start \<Longrightarrow> \<delta> u v = {}"
      unfolding nfa_to_gnfa_invar1_def gnfa_\<alpha>_def Let_def 
      by (simp split: gnfastate.split)
  ultimately show "?P' q = Some {u\<in>\<Q> ?\<A>'. ?\<delta>' u q \<noteq> {}}" by simp

  from assms have [simp]:"v \<in> Q" using \<I>_consistent 
      by (auto simp: nfa_to_gnfa_invar1_def)
  from `q \<in> \<Q> ?\<A>'` and assms have "S q = Some {v\<in>\<Q> ?\<A>. \<delta> q v \<noteq> {}}"
      unfolding gnfa_invar_def by (simp add: Let_def)
  thus "?S' q = Some {v\<in>\<Q> ?\<A>'. ?\<delta>' q v \<noteq> {}}" using assms
      by (force simp: gnfa_\<alpha>_def PS_add_def split: option.split)
qed


lemma (in NFA) nfa_to_gnfa_impl2_aux2:
  fixes \<delta>
  assumes "u \<in> it" "it \<subseteq> \<F> \<A>" "nfa_to_gnfa_invar2 \<A> 
        (insert Start (insert End (State`SemiAutomaton.\<Q> \<A>))) it 
        \<lparr>\<Q>=insert Start (insert End (State`Q)), GNFA_rec.\<delta> = \<delta> \<rparr>"
      "gnfa_invar (Q, \<delta>, P, S)"
  shows "gnfa_invar (Q, \<lambda>u' v'. if u'=State u\<and>v'=End then 
      {[]} else \<delta> u' v', PS_add P End (State u), S(State u \<mapsto> {End}))" 
      (is "gnfa_invar (Q, ?\<delta>', ?P', ?S')")
proof
  let ?\<A> = "gnfa_\<alpha> (Q,\<delta>,P,S)" and  ?\<A>' = "gnfa_\<alpha> (Q,?\<delta>',?P',?S')"
  have [simp]: "\<Q> ?\<A>' = \<Q> ?\<A>" unfolding gnfa_\<alpha>_def by simp
  from assms have "GNFA ?\<A>" unfolding gnfa_invar_def Let_def by simp
  thus "GNFA ?\<A>'" by (simp_all add: GNFA_def gnfa_\<alpha>_def)

  hence [simp, intro]: "End \<in> \<Q> ?\<A>" unfolding GNFA_def by simp

  from assms have [simp]:"u \<in> Q" using \<F>_consistent 
      by (auto simp: nfa_to_gnfa_invar2_def)

  fix q assume "q \<in> \<Q> ?\<A>'" hence "q \<in> \<Q> ?\<A>" by simp
  with assms have "S q = Some {v\<in>\<Q> ?\<A>. \<delta> q v \<noteq> {}}" 
      unfolding gnfa_invar_def by (simp add: Let_def)
  moreover from assms have "\<And>u v. v \<noteq> End \<Longrightarrow> \<delta> (State u) v = {}"
      unfolding nfa_to_gnfa_invar2_def gnfa_\<alpha>_def Let_def 
      by (simp split: gnfastate.split)
  ultimately show "?S' q = Some {v\<in>\<Q> ?\<A>'. ?\<delta>' q v \<noteq> {}}" by simp

  from `q \<in> \<Q> ?\<A>'` and assms have "P q = Some {u\<in>\<Q> ?\<A>. \<delta> u q \<noteq> {}}"
      unfolding gnfa_invar_def by (simp add: Let_def)
  thus "?P' q = Some {u\<in>\<Q> ?\<A>'. ?\<delta>' u q \<noteq> {}}" using assms
      by (force simp: gnfa_\<alpha>_def PS_add_def split: option.split)
qed


lemma (in NFA) nfa_to_gnfa_impl2_aux3:
  fixes \<delta>
  assumes "ucv \<in> it" "it \<subseteq> \<Delta> \<A>" "ucv = (u,c,v)" "nfa_to_gnfa_invar3 \<A> 
        (insert Start (insert End (State`SemiAutomaton.\<Q> \<A>))) it 
        \<lparr>\<Q>=insert Start (insert End (State`Q)), GNFA_rec.\<delta> = \<delta> \<rparr>"
      "gnfa_invar (Q, \<delta>, P, S)"
  shows "gnfa_invar (Q, \<lambda>u' v'. if u'=State u\<and>v'=State v then 
      insert [c] (\<delta> u' v') else \<delta> u' v', 
      PS_add P (State v) (State u), PS_add S (State u) (State v))" 
      (is "gnfa_invar (Q, ?\<delta>', ?P', ?S')")
proof
  let ?\<A> = "gnfa_\<alpha> (Q,\<delta>,P,S)" and  ?\<A>' = "gnfa_\<alpha> (Q,?\<delta>',?P',?S')"
  have [simp]: "\<Q> ?\<A>' = \<Q> ?\<A>" unfolding gnfa_\<alpha>_def by simp
  from assms have "GNFA ?\<A>" unfolding gnfa_invar_def Let_def by simp
  thus "GNFA ?\<A>'" by (simp_all add: GNFA_def gnfa_\<alpha>_def)

  from assms(1-3) and \<Delta>_consistent 
      have "u \<in> SemiAutomaton.\<Q> \<A>" "v \<in> SemiAutomaton.\<Q> \<A>" by auto
  hence [simp, intro]: "State u \<in> \<Q> ?\<A>" "State v \<in> \<Q> ?\<A>" "u \<in> Q" "v \<in> Q"
      using assms(4) by (auto simp: nfa_to_gnfa_invar3_def gnfa_\<alpha>_def)

  fix q assume "q \<in> \<Q> ?\<A>'" hence "q \<in> \<Q> ?\<A>" by simp
  with assms have A: "S q = Some {v\<in>\<Q> ?\<A>. \<delta> q v \<noteq> {}}" and
                  B: "P q = Some {u\<in>\<Q> ?\<A>. \<delta> u q \<noteq> {}}"
      unfolding gnfa_invar_def by (simp_all add: Let_def)
  thus "?S' q = Some {v\<in>\<Q> ?\<A>'. ?\<delta>' q v \<noteq> {}}"
       "?P' q = Some {u\<in>\<Q> ?\<A>'. ?\<delta>' u q \<noteq> {}}" 
       by (auto simp: gnfa_\<alpha>_def PS_add_def split: option.split)
qed



lemma (in NFA) nfa_to_gnfa_impl2_refine:
  "nfa_to_gnfa_impl2 \<A> \<le> \<Down>gnfa_refrel (nfa_to_gnfa_impl \<A>)"
unfolding nfa_to_gnfa_impl2_def nfa_to_gnfa_impl_def
apply (refine_rcg single_valued_gnfa_refrel inj_on_id)
apply (fact gnfa_initial_impl2_correct)
apply simp_all
apply (auto simp: gnfa_\<alpha>_def br_def intro!: ext nfa_to_gnfa_impl2_aux1) []
apply (auto simp: gnfa_\<alpha>_def br_def intro!: ext nfa_to_gnfa_impl2_aux2) []
apply (auto simp: gnfa_\<alpha>_def br_def intro!: ext nfa_to_gnfa_impl2_aux3) []
done


definition nfa_to_rexp_impl2 where
"nfa_to_rexp_impl2 \<A> \<equiv> do {
  (Q,\<delta>,P,S) \<leftarrow> nfa_to_gnfa_impl2 \<A>;
  (Q,\<delta>,P,S) \<leftarrow> 
    WHILE\<^sub>T (\<lambda>(Q,\<delta>,P,S). Q \<noteq> {}) (\<lambda>(Q,\<delta>,P,S). do {
      q \<leftarrow> SPEC (\<lambda>q. q \<in> Q);
      (Q,\<delta>,P,S) \<leftarrow> gnfa_contract_impl2 (Q,\<delta>,P,S) (State q);
      RETURN (Q,\<delta>,P,S)
    }) (Q,\<delta>,P,S);
  RETURN (\<delta> Start End)
}"

lemma (in NFA) nfa_to_rexp_impl2_refine:
  "nfa_to_rexp_impl2 \<A> \<le> \<Down>Id (nfa_to_rexp_impl \<A>)"
unfolding nfa_to_rexp_impl2_def nfa_to_rexp_impl_def
apply (refine_rcg )
apply (rule nfa_to_gnfa_impl2_refine)
apply simp
apply (force simp: gnfa_\<alpha>_def br_def)
apply (rule SPEC_refine) apply (rule SPEC_rule)
defer
apply (rule gnfa_contract_impl2_refine)
apply assumption
apply simp
apply blast
apply simp
apply (simp add: gnfa_\<alpha>_def br_def)
apply (simp add: gnfa_\<alpha>_def br_def)
done


abbreviation "gnfa_\<delta>_lookup d u v \<equiv>
  (case d u of None \<Rightarrow> Zero | Some du \<Rightarrow>
      (case du v of None \<Rightarrow> Zero | Some r \<Rightarrow> r))"

abbreviation "gnfa_\<delta>_\<alpha>2 d \<equiv> \<lambda>u v. 
    lang (gnfa_\<delta>_lookup d u v)"

abbreviation "gnfa_\<delta>_refrel2 \<equiv> br gnfa_\<delta>_\<alpha>2 (\<lambda>_. True)"
lemma single_valued_gnfa_\<delta>_refrel2:
   "single_valued gnfa_\<delta>_refrel2" by (fact br_sv)

abbreviation "rprod \<equiv> prod_rel"

abbreviation "gnfa_refrel2 \<equiv> \<langle>Id, \<langle>gnfa_\<delta>_refrel2, \<langle>Id, Id\<rangle>rprod\<rangle>rprod\<rangle>rprod"
lemma single_valued_gnfa_refrel2:
  "single_valued gnfa_refrel2" 
    by (intro prod_rel_sv single_valued_Id single_valued_gnfa_\<delta>_refrel2)


definition gnfa_initial_impl3 :: "('q, 'c, 'e) SemiAutomaton_rec_scheme
     \<Rightarrow> ('q set \<times>
        ('q gnfastate \<Rightarrow> ('q gnfastate \<Rightarrow> 'c rexp option) option) \<times>
        ('q gnfastate \<Rightarrow> ('q gnfastate) set option) \<times>
        ('q gnfastate \<Rightarrow> ('q gnfastate) set option)) nres" where
"gnfa_initial_impl3 \<A> \<equiv> do {
    M \<leftarrow> FOREACH (SemiAutomaton.\<Q> \<A>) 
        (\<lambda>q M. RETURN (M(State q \<mapsto> {}))) Map.empty;
    let M = M(Start \<mapsto> {}, End \<mapsto> {});
    d \<leftarrow> RETURN Map.empty;
    RETURN (SemiAutomaton.\<Q> \<A>, d, M, M)
}"

lemma gnfa_initial_impl3_\<delta>_correct:
  "RETURN Map.empty \<le> \<Down>gnfa_\<delta>_refrel2 
      (SPEC (\<lambda>d. \<forall>u v. d u v = {}))"
  by (rule SPEC_refine, simp add: br_def)

lemma gnfa_initial_impl3_refine:
  "gnfa_initial_impl3 \<A> \<le> \<Down>gnfa_refrel2 
      (gnfa_initial_impl2 \<A>)"
unfolding gnfa_initial_impl3_def gnfa_initial_impl2_def
apply (refine_rcg inj_on_id)
apply refine_dref_type
apply simp_all[3]
apply (rule gnfa_initial_impl3_\<delta>_correct)
apply (simp add: br_def)
done

definition "gnfa_\<delta>_update d u v r \<equiv>
  case d u of
    None \<Rightarrow> d(u \<mapsto> [v \<mapsto> r]) |
    Some du \<Rightarrow> d(u \<mapsto> du(v \<mapsto> r))"

definition "gnfa_\<delta>_insert d u v r \<equiv>
  case d u of
    None \<Rightarrow> d(u \<mapsto> [v \<mapsto> r]) |
    Some du \<Rightarrow> let duv' = (
        case du v of
          None \<Rightarrow> r |
          Some duv \<Rightarrow> Plus duv r)
      in d(u \<mapsto> du(v \<mapsto> duv'))"


lemma gnfa_\<delta>_update_correct[simp]:
  "gnfa_\<delta>_\<alpha>2 (gnfa_\<delta>_update d u v r) = (\<lambda>u' v'.
      (if u'=u \<and> v'=v then lang r else gnfa_\<delta>_\<alpha>2 d u' v'))"
  unfolding gnfa_\<delta>_update_def
  by (intro ext, auto split: option.split)

lemma gnfa_\<delta>_insert_correct[simp]:
     "gnfa_\<delta>_\<alpha>2 (gnfa_\<delta>_insert d u v r) = (\<lambda>u' v'.
      (if u'=u \<and> v'=v then gnfa_\<delta>_\<alpha>2 d u' v' \<union> lang r 
          else gnfa_\<delta>_\<alpha>2 d u' v'))"  
  unfolding gnfa_\<delta>_insert_def
  by (intro ext, auto split: option.split)


definition nfa_to_gnfa_impl3 where
"nfa_to_gnfa_impl3 \<A> = do {
  \<A>' \<leftarrow> gnfa_initial_impl3 \<A>;
  \<A>'' \<leftarrow> FOREACH (\<I> \<A>) (\<lambda>v (Q, \<delta>, P, S). 
        RETURN (Q, gnfa_\<delta>_update \<delta> Start (State v) rexp.One,
            P(State v  \<mapsto> {Start}), PS_add S Start (State v))) \<A>';
  \<A>'' \<leftarrow> FOREACH (\<F> \<A>) (\<lambda>u (Q, \<delta>, P, S). 
        RETURN (Q, gnfa_\<delta>_update \<delta> (State u) End rexp.One,
            PS_add P End (State u), S(State u \<mapsto> {End}))) \<A>'';
  \<A>'' \<leftarrow> FOREACH (\<Delta> \<A>) (\<lambda>(u,c,v) (Q, \<delta>, P, S). 
        RETURN (Q, gnfa_\<delta>_insert \<delta> (State u) (State v) (Atom c),
            PS_add P (State v) (State u), PS_add S (State u) (State v))) \<A>'';
  RETURN \<A>''
}"


lemma nfa_to_gnfa_impl3_refine:
  "nfa_to_gnfa_impl3 \<A> \<le> \<Down>gnfa_refrel2 (nfa_to_gnfa_impl2 \<A>)"
unfolding nfa_to_gnfa_impl3_def nfa_to_gnfa_impl2_def
apply (refine_rcg inj_on_id)
apply (rule gnfa_initial_impl3_refine)
apply simp
apply simp
apply (simp add: br_def, intro ext, simp)
apply simp
apply simp
apply (simp add: br_def, intro ext, simp)
apply simp
apply simp
apply (simp add: br_def, intro ext, auto) []
apply simp
done



definition gnfa_remove_state_impl3 where
"gnfa_remove_state_impl3 \<equiv> \<lambda>(Q,\<delta>,P,S) q.
  case q of Start \<Rightarrow> RETURN (Q,\<delta>,P,S) | End \<Rightarrow> RETURN (Q,\<delta>,P,S) | 
      State q' \<Rightarrow> do {
        P' \<leftarrow> FOREACH (PS_the S q) 
            (\<lambda>v P. RETURN (PS_remove P v q)) P;
        S' \<leftarrow> FOREACH (PS_the P q) 
            (\<lambda>u S. RETURN (PS_remove S u q)) S;
        RETURN (Q - {q'}, \<delta>, P', S')
      }"


lemma gnfa_remove_state_impl3_refine:
  fixes \<delta>
  assumes "((Q,\<delta>,P,S), (Q',\<delta>',P',S')) \<in> gnfa_refrel2" and "(q,q') \<in> Id"
  shows "gnfa_remove_state_impl3 (Q,\<delta>,P,S) q \<le> \<Down>gnfa_refrel2
      (gnfa_remove_state_impl2 (Q',\<delta>',P',S') q')"
unfolding gnfa_remove_state_impl3_def gnfa_remove_state_impl2_def
apply (simp add: br_def prod_rel_def split: gnfastate.split)
apply (intro conjI impI allI)
apply (insert assms)
apply (refine_rcg)
apply (simp_all add: br_def)[4]
apply (rule RETURN_refine)
apply (force simp: br_def)
apply (refine_rcg, (simp_all add: single_valued_def br_def)[4])
apply (refine_rcg inj_on_id)
apply refine_dref_type
apply (simp_all add: br_def prod_rel_def)
done


definition "rexp_simped_concat r s \<equiv>
  (if r = rexp.One then s
    else if s = rexp.One then r else Times r s)"

lemma rexp_simped_concat_correct[simp]:
  "lang (rexp_simped_concat r s) = lang r @@ lang s" 
   unfolding rexp_simped_concat_def by simp

definition "rexp_simped_contract r s t \<equiv>
  (if s = Zero then rexp_simped_concat r t 
      else rexp_simped_concat r 
          (rexp_simped_concat (Star s) t))"

lemma rexp_simped_contract_correct[simp]:
  "lang (rexp_simped_contract r1 r2 r3) = 
    lang r1 @@ star (lang r2) @@ lang r3"
  unfolding rexp_simped_contract_def by simp

definition gnfa_contract_impl3_update_\<delta> where 
"gnfa_contract_impl3_update_\<delta> \<equiv> \<lambda>\<delta> u q v.
    let r1 = gnfa_\<delta>_lookup \<delta> u q;
        r2 = gnfa_\<delta>_lookup \<delta> q q;
        r3 = gnfa_\<delta>_lookup \<delta> q v
    in gnfa_\<delta>_insert \<delta> u v (rexp_simped_contract r1 r2 r3)"


lemma gnfa_\<delta>_insert_correct':
  "gnfa_\<delta>_\<alpha>2 (gnfa_\<delta>_insert d u v r) u' v' = 
    (if u' = u \<and> v' = v then gnfa_\<delta>_\<alpha>2 d u' v' \<union> lang r
        else gnfa_\<delta>_\<alpha>2 d u' v')"
  by (simp add: gnfa_\<delta>_insert_def split: option.split)

lemma gnfa_contract_impl3_update_\<delta>_correct:
  assumes "(\<delta>\<^sub>1, \<delta>\<^sub>2) \<in> gnfa_\<delta>_refrel2" "(q,q') \<in> Id"
  shows "gnfa_contract_update_\<delta>_impl2 \<delta>\<^sub>2 u q' v =
      gnfa_\<delta>_\<alpha>2 (gnfa_contract_impl3_update_\<delta> \<delta>\<^sub>1 u q v)"
          (is "?g = gnfa_\<delta>_\<alpha>2 ?f")
proof-
  {fix u' v'
    from assms have "gnfa_\<delta>_\<alpha>2 ?f u' v' = ?g u' v'"
    unfolding gnfa_contract_impl3_update_\<delta>_def
        gnfa_contract_update_\<delta>_impl2_def
      by (auto simp: Let_def br_def gnfa_\<delta>_insert_correct')
  }
  hence "gnfa_\<delta>_\<alpha>2 ?f = ?g" by (intro ext)
  thus ?thesis by simp
qed


definition gnfa_contract_impl3 where
"gnfa_contract_impl3 \<equiv> \<lambda>(Q,\<delta>,P,S) q. do {
  let Pq = PS_the P q - {q}; let Sq = PS_the S q - {q};
  (Q,\<delta>,P,S) \<leftarrow> gnfa_remove_state_impl3 (Q,\<delta>,P,S) q;
  (Q,\<delta>,P,S) \<leftarrow> FOREACH Pq (\<lambda>u (Q',\<delta>',P',S').
    FOREACH Sq (\<lambda>v (Q',\<delta>',P',S').
        RETURN (Q', gnfa_contract_impl3_update_\<delta> \<delta>' u q v,
            PS_add P' v u, PS_add S' u v))
      (Q',\<delta>',P',S')
  ) (Q,\<delta>,P,S);
  RETURN (Q,\<delta>,P,S)
}"

lemma gnfa_contract_impl3_refine:
  fixes \<delta>
  assumes "((Q,\<delta>,P,S), (Q',\<delta>',P',S')) \<in> gnfa_refrel2" and 
      "(q,q') \<in> Id"
  shows "gnfa_contract_impl3 (Q,\<delta>,P,S) q \<le> \<Down>gnfa_refrel2
      (gnfa_contract_impl2 (Q',\<delta>',P',S') q')"
using [[goals_limit=1]]
unfolding gnfa_contract_impl3_def gnfa_contract_impl2_def
apply (refine_rcg inj_on_id)
thm gnfa_remove_state_impl3_refine
apply (insert assms, rule gnfa_remove_state_impl3_refine, simp_all) [2]
apply clarsimp
apply ((rule prod_relI IdI | assumption)+)[]
apply (insert assms, simp) []
apply simp
apply (clarsimp simp add: br_def
    gnfa_contract_impl3_update_\<delta>_correct[OF _ assms(2)])
apply simp
done

definition nfa_to_rexp_impl3 where
"nfa_to_rexp_impl3 \<A> \<equiv> do {
  (Q,\<delta>,P,S) \<leftarrow> nfa_to_gnfa_impl3 \<A>;
  (Q,\<delta>,P,S) \<leftarrow> 
    WHILE\<^sub>T (\<lambda>(Q,\<delta>,P,S). Q \<noteq> {}) (\<lambda>(Q,\<delta>,P,S). do {
      ASSERT (Q\<noteq>{});
      q \<leftarrow> SPEC (\<lambda>q. q \<in> Q);
      (Q,\<delta>,P,S) \<leftarrow> gnfa_contract_impl3 (Q,\<delta>,P,S) (State q);
      RETURN (Q,\<delta>,P,S)
    }) (Q,\<delta>,P,S);
  RETURN (gnfa_\<delta>_lookup \<delta> Start End)
}"


abbreviation "rexp_refrel \<equiv> br lang (\<lambda>_. True)"

lemma nfa_to_rexp_impl3_refine:
  "nfa_to_rexp_impl3 \<A> \<le> \<Down>rexp_refrel (nfa_to_rexp_impl2 \<A>)"
unfolding nfa_to_rexp_impl3_def nfa_to_rexp_impl2_def
using [[goals_limit=1]]
apply (refine_rcg inj_on_id)
apply (rule nfa_to_gnfa_impl3_refine)
apply clarsimp
apply ((rule prod_relI IdI | assumption)+) []
apply (simp_all add: br_def prod_rel_def)[3]
apply (rule gnfa_contract_impl3_refine)
apply (simp_all add: br_def prod_rel_def)
done

subsection "Implementation of NFAs"

abbreviation "rexp_rel \<equiv> Id :: (nat rexp\<times>nat rexp) set"
consts i_rexp :: "interface"

lemmas rexp_rel_def = TrueI

lemmas [autoref_rel_intf] = REL_INTFI[of "rexp_rel" i_rexp]

lemma rexp_rel_sv[relator_props]: "single_valued rexp_rel"
  unfolding rexp_rel_def by simp

lemma Zero_param[param,autoref_rules]:
    "(Zero,Zero) \<in> rexp_rel" unfolding rexp_rel_def by simp

lemma One_param[param,autoref_rules]:
    "(One,One) \<in> rexp_rel" unfolding rexp_rel_def by simp

lemma Atom_param[param,autoref_rules]:
    "(Atom,Atom) \<in> nat_rel \<rightarrow> rexp_rel" unfolding rexp_rel_def by simp

lemma Plus_param[param,autoref_rules]:
    "(Plus,Plus) \<in> rexp_rel \<rightarrow> rexp_rel \<rightarrow> rexp_rel"
    unfolding rexp_rel_def by simp

lemma Times_param[param,autoref_rules]:
    "(Times,Times) \<in> rexp_rel \<rightarrow> rexp_rel \<rightarrow> rexp_rel"
    unfolding rexp_rel_def by simp

lemma Star_param[param,autoref_rules]:
    "(Star,Star) \<in> rexp_rel \<rightarrow> rexp_rel"
    unfolding rexp_rel_def by simp


abbreviation "gnfastate_rel \<equiv> (Id::(nat gnfastate\<times>nat gnfastate) set)"

(*
definition gnfastate_rel_internal_def: 
    "gnfastate_rel \<equiv> (Id::(nat gnfastate\<times>nat gnfastate) set)"

lemma gnfastate_rel_def: "gnfastate_rel \<equiv> Id"
  by (fact gnfastate_rel_internal_def)

lemma gnfastate_rel_sv[relator_props]: "single_valued gnfastate_rel"
  unfolding gnfastate_rel_def by simp
*)

consts i_gnfastate :: "interface"

lemmas [autoref_rel_intf] = REL_INTFI[of "gnfastate_rel" i_gnfastate]

lemma param_State[param,autoref_rules]:
    "(State, State) \<in> nat_rel \<rightarrow> gnfastate_rel"
    (*unfolding gnfastate_rel_def*) by simp

lemma param_Start[param,autoref_rules]:
    "(Start, Start) \<in> gnfastate_rel" (*unfolding gnfastate_rel_def*) by simp

lemma param_End[param,autoref_rules]:
    "(End, End) \<in> gnfastate_rel" (*unfolding gnfastate_rel_def*) by simp

lemma param_case_gnfastate[param,autoref_rules]:
    "(case_gnfastate, case_gnfastate) \<in> 
      R \<rightarrow> R \<rightarrow> (nat_rel \<rightarrow> R) \<rightarrow> gnfastate_rel \<rightarrow> R"
  (*unfolding gnfastate_rel_def*)
  by (force split: gnfastate.split dest: fun_relD)

lemma param_gnfastate_eq[autoref_rules]:
  "(op =, op =) \<in> gnfastate_rel \<rightarrow> gnfastate_rel \<rightarrow> bool_rel"
  by simp

lemma rec_gnfastate_is_case_gnfastate[simp]:
    "rec_gnfastate = case_gnfastate"
    by (force split: gnfastate.split)

fun \<Q>_impl where "\<Q>_impl (Q,S,D,I,F) = Q"
fun \<Sigma>_impl where "\<Sigma>_impl (Q,S,D,I,F) = S"
fun \<Delta>_impl where "\<Delta>_impl (Q,S,D,I,F) = D"
fun \<I>_impl where "\<I>_impl (Q,S,D,I,F) = I"
fun \<F>_impl where "\<F>_impl (Q,S,D,I,F) = F"

definition NFA_rel_internal_def: "NFA_rel Rqs Rss Rds Ris Rfs R\<Q> R\<Sigma> \<equiv> 
  { ((Q,S,D,I,F),\<A>) . 
    NFA \<A> \<and> 
    (Q,SemiAutomaton.\<Q> \<A>)\<in>\<langle>R\<Q>\<rangle>Rqs \<and> 
    (S,\<Sigma> \<A>)\<in>\<langle>R\<Sigma>\<rangle>Rss \<and>       
    (D,\<Delta> \<A>)\<in>\<langle>\<langle>R\<Q>,\<langle>R\<Sigma>,R\<Q>\<rangle>prod_rel\<rangle>prod_rel\<rangle>Rds \<and>
    (I,\<I> \<A>)\<in>\<langle>R\<Q>\<rangle>Ris \<and>       
    (F,\<F> \<A>)\<in>\<langle>R\<Q>\<rangle>Rfs}"

lemma NFA_rel_def: "\<langle>R\<Q>,R\<Sigma>\<rangle>NFA_rel Rqs Rss Rds Ris Rfs \<equiv> { ((Q,S,D,I,F),\<A>) . 
    NFA \<A> \<and> 
    (Q,SemiAutomaton.\<Q> \<A>)\<in>\<langle>R\<Q>\<rangle>Rqs \<and> 
    (S,\<Sigma> \<A>)\<in>\<langle>R\<Sigma>\<rangle>Rss \<and>       
    (D,\<Delta> \<A>)\<in>\<langle>\<langle>R\<Q>,\<langle>R\<Sigma>,R\<Q>\<rangle>prod_rel\<rangle>prod_rel\<rangle>Rds \<and>
    (I,\<I> \<A>)\<in>\<langle>R\<Q>\<rangle>Ris \<and>       
    (F,\<F> \<A>)\<in>\<langle>R\<Q>\<rangle>Rfs}"
  unfolding NFA_rel_internal_def[abs_def] relAPP_def .
  
consts i_NFA :: "interface \<Rightarrow> interface \<Rightarrow> interface"

lemmas [autoref_rel_intf] = 
  REL_INTFI[of "NFA_rel Rqs Rss Rds Ris Rfs" i_NFA for Rqs Rss Rds Ris Rfs]

lemma \<Q>_autoref[autoref_rules]:
  "(\<Q>_impl,SemiAutomaton.\<Q>)\<in>\<langle>R\<Q>,R\<Sigma>\<rangle>NFA_rel Rqs Rss Rds Ris Rfs \<rightarrow> \<langle>R\<Q>\<rangle>Rqs"
  unfolding NFA_rel_def by auto
lemma \<Sigma>_autoref[autoref_rules]:
  "(\<Sigma>_impl,\<Sigma>)\<in>\<langle>R\<Q>,R\<Sigma>\<rangle>NFA_rel Rqs Rss Rds Ris Rfs \<rightarrow> \<langle>R\<Sigma>\<rangle>Rss"
  unfolding NFA_rel_def by auto
lemma \<Delta>_autoref[autoref_rules]:
  "(\<Delta>_impl,\<Delta>)\<in>\<langle>R\<Q>,R\<Sigma>\<rangle>NFA_rel Rqs Rss Rds Ris Rfs 
    \<rightarrow> \<langle>\<langle>R\<Q>,\<langle>R\<Sigma>,R\<Q>\<rangle>prod_rel\<rangle>prod_rel\<rangle>Rds"
  unfolding NFA_rel_def by auto
lemma \<I>_autoref[autoref_rules]:
  "(\<I>_impl,\<I>)\<in>\<langle>R\<Q>,R\<Sigma>\<rangle>NFA_rel Rqs Rss Rds Ris Rfs \<rightarrow> \<langle>R\<Q>\<rangle>Ris"
  unfolding NFA_rel_def by auto
lemma \<F>_autoref[autoref_rules]:
  "(\<F>_impl,\<F>)\<in>\<langle>R\<Q>,R\<Sigma>\<rangle>NFA_rel Rqs Rss Rds Ris Rfs \<rightarrow> \<langle>R\<Q>\<rangle>Rfs"
  unfolding NFA_rel_def by auto

(*abbreviation "dflt_NFA_rel 
  \<equiv> NFA_rel list_set_rel list_set_rel list_set_rel list_set_rel list_set_rel"
*)

abbreviation "dflt_NFA_rel 
  \<equiv> NFA_rel dflt_rs_rel dflt_rs_rel dflt_rs_rel dflt_rs_rel dflt_rs_rel"




lemmas nfa_to_rexp_unfold_complete =
     nfa_to_rexp_impl3_def[unfolded nfa_to_gnfa_impl3_def gnfa_initial_impl3_def
    gnfa_\<delta>_update_def gnfa_contract_impl3_def gnfa_contract_impl3_update_\<delta>_def
    rexp_simped_contract_def rexp_simped_concat_def PS_add_def 
    gnfa_remove_state_impl3_def PS_the_def gnfa_\<delta>_insert_def PS_remove_def]

concrete_definition nfa_to_rexp uses nfa_to_rexp_unfold_complete

(*
concrete_definition foo uses gnfa_initial_impl3_def
schematic_lemma 
  assumes [autoref_rules]: "(\<A>,\<A>')\<in>\<langle>nat_rel,nat_rel\<rangle>dflt_NFA_rel"
  notes [autoref_tyrel] = 
    ty_REL[where R="\<langle>gnfastate_rel\<rangle>dflt_rs_rel"]
    ty_REL[where R="\<langle>gnfastate_rel,rexp_rel\<rangle>dflt_rm_rel"]
    ty_REL[where R=
      "\<langle>gnfastate_rel,\<langle>gnfastate_rel,rexp_rel\<rangle>dflt_rm_rel\<rangle>dflt_rm_rel"]

  shows "(?f::?'c, foo \<A>')\<in>?R"
  unfolding foo_def
  apply (autoref (keep_goal,debug,trace) 
    phases: id_op rel_inf fix_rel trans
    
    )
    done
*)

lemma (in transfer) transfer_gnfastate[refine_transfer]:
  assumes "\<alpha> fs \<le> Fs"
  assumes "\<alpha> fe \<le> Fe"
  assumes "\<And>q. \<alpha> (fq q) \<le> Fq q"
  shows "\<alpha> (case_gnfastate fs fe fq x) \<le> case_gnfastate Fs Fe Fq x"
  using assms by (auto split: gnfastate.split)

lemma gnfastate_ne_bot[refine_transfer]:
  "\<And>fs fe fq x. 
    \<lbrakk> fs\<noteq>dSUCCEED; fe\<noteq>dSUCCEED; \<And>v. fq v \<noteq> dSUCCEED \<rbrakk> 
    \<Longrightarrow> case_gnfastate fs fe fq x \<noteq> dSUCCEED"
  by (auto split: gnfastate.split)

schematic_lemma nfa_to_rexp_impl:
  notes [[goals_limit = 1]]
  assumes [autoref_rules]: "(\<A>,\<A>')\<in>\<langle>nat_rel,nat_rel\<rangle>dflt_NFA_rel"
  shows "(?f::?'c, nfa_to_rexp \<A>')\<in>?R"
  using assms
  unfolding nfa_to_rexp_def
  apply (autoref_monadic (trace))  (* This takes loooong (several minutes)*)
  done

concrete_definition nfa_to_rexp_code uses nfa_to_rexp_impl

export_code nfa_to_rexp_code in SML 

theorem nfa_to_rexp_code_correct: 
  assumes A: "(\<A>impl,\<A>)\<in>\<langle>nat_rel,nat_rel\<rangle>dflt_NFA_rel"
  shows "lang (nfa_to_rexp_code \<A>impl) = \<L> \<A>" (is "lang ?r = _")
proof -
  interpret NFA \<A> using A[unfolded NFA_rel_def] by auto 

  note nfa_to_rexp_code.refine[OF A, THEN nres_relD]
  also note nfa_to_rexp.refine[symmetric, THEN meta_eq_to_obj_eq]
  also note nfa_to_rexp_impl3_refine
  also note nfa_to_rexp_impl2_refine
  also note nfa_to_rexp_impl_refine
  also note nfa_to_rexp_abstr_correct
  finally show ?thesis by (elim RETURN_ref_SPECD, simp add: br_def)
qed




(**** Old code generation ****)
(*
locale nfa_to_rexp_locale =
  NFA \<A> +
  NFA_Q_set : StdSet NFA_Q_set_ops +
  Q_set : StdSet Q_set_ops +
  \<delta>_map : StdMap \<delta>_map_ops +
  \<delta>u_map : StdMap \<delta>u_map_ops +
  PS_map : StdMap PS_map_ops +
  PSu_set : StdSet PSu_set_ops +
  NFA_Q_Q_it : set_iteratei NFA_Q_set.\<alpha> NFA_Q_set.invar NFA_Q_Q_it +
  Q_PS_it : set_iteratei Q_set.\<alpha> Q_set.invar Q_PS_it +
  Q_GNFA_it : set_iteratei Q_set.\<alpha> Q_set.invar Q_GNFA_it +
  PS_PS_it : set_iteratei PSu_set.\<alpha> PSu_set.invar PS_PS_it + 
  PS_GNFA_it : set_iteratei PSu_set.\<alpha> PSu_set.invar PS_GNFA_it

  for \<A> :: "('q,'c,_) NFA_rec_scheme"
  and NFA_Q_set_ops :: "('q, 'NFA_Q_set, _) set_ops_scheme"
  and Q_set_ops :: "('q, 'Q_set, _) set_ops_scheme"
  and \<delta>_map_ops :: "('q gnfastate, '\<delta>u_map, '\<delta>_map, _) map_ops_scheme"
  and \<delta>u_map_ops :: "('q gnfastate, 'c rexp, '\<delta>u_map, _) map_ops_scheme"
  and PS_map_ops :: "('q gnfastate, 'PSu_set, 'PS_map, _) map_ops_scheme"
  and PSu_set_ops :: "('q gnfastate, 'PSu_set, _) set_ops_scheme"
  and NFA_Q_Q_it :: "'NFA_Q_set \<Rightarrow> ('q, 'Q_set) set_iterator"
  and Q_PS_it :: "'Q_set \<Rightarrow> ('q, 'PS_map) set_iterator"
  and Q_GNFA_it :: "'Q_set \<Rightarrow> 
        ('q, 'Q_set\<times>'\<delta>_map\<times>'PS_map\<times>'PS_map) set_iterator"
  and PS_PS_it :: "'PSu_set \<Rightarrow> ('q gnfastate, 'PS_map) set_iterator"
  and PS_GNFA_it :: "'PSu_set \<Rightarrow> ('q gnfastate, 
      'Q_set\<times>'\<delta>_map\<times>'PS_map\<times>'PS_map) set_iterator" +

  fixes NFA_\<I>_GNFA_it :: "('q, 'Q_set\<times>'\<delta>_map\<times>'PS_map\<times>'PS_map) set_iterator"
  and NFA_\<F>_GNFA_it :: "('q, 'Q_set\<times>'\<delta>_map\<times>'PS_map\<times>'PS_map) set_iterator"
  and NFA_\<Delta>_GNFA_it :: 
        "('q\<times>'c\<times>'q, 'Q_set\<times>'\<delta>_map\<times>'PS_map\<times>'PS_map) set_iterator"
  and \<Q>_impl :: "'NFA_Q_set"
  assumes NFA_\<I>_GNFA_it_correct[refine_transfer]: 
      "set_iterator NFA_\<I>_GNFA_it (\<I> \<A>)"
  assumes NFA_\<F>_GNFA_it_correct[refine_transfer]: 
      "set_iterator NFA_\<F>_GNFA_it (\<F> \<A>)"
  assumes NFA_\<Delta>_GNFA_it_correct[refine_transfer]: 
      "set_iterator NFA_\<Delta>_GNFA_it (\<Delta> \<A>)"
  assumes \<Q>_impl_correct[refine_transfer]:
      "NFA_Q_set.invar \<Q>_impl" "NFA_Q_set.\<alpha> \<Q>_impl = SemiAutomaton.\<Q> \<A>"
begin

definition \<delta>_\<alpha> :: "'\<delta>_map \<Rightarrow> 'q gnfastate \<rightharpoonup> 'q gnfastate \<rightharpoonup> 'c rexp" 
  where "\<delta>_\<alpha> M u \<equiv> case \<delta>_map.\<alpha> M u of 
    None \<Rightarrow> None 
  | Some m \<Rightarrow> Some (\<delta>u_map.\<alpha> m)"
definition \<delta>_invar  where 
  "\<delta>_invar M \<equiv> \<delta>_map.invar M \<and> (\<forall>m\<in>ran (\<delta>_map.\<alpha> M). \<delta>u_map.invar m)"
abbreviation "\<delta>_refrel_impl \<equiv> br \<delta>_\<alpha> \<delta>_invar"
lemma single_valued_\<delta>_refrel_impl: "single_valued \<delta>_refrel_impl"
    by refine_post
lemma \<delta>_refrel_impl_def:
    "(d, d') \<in> \<delta>_refrel_impl \<longleftrightarrow> (d' = 
    (\<lambda>u. case map_op_\<alpha> \<delta>_map_ops d u of None \<Rightarrow> None 
    | Some m \<Rightarrow> Some (map_op_\<alpha> \<delta>u_map_ops m)) \<and>
     map_op_invar \<delta>_map_ops d \<and>
  (\<forall>x\<in>ran (map_op_\<alpha> \<delta>_map_ops d). map_op_invar \<delta>u_map_ops x))"
unfolding \<delta>_\<alpha>_def[abs_def] \<delta>_invar_def[abs_def] br_def  by simp




definition PS_\<alpha> :: "'PS_map \<Rightarrow> 'q gnfastate \<rightharpoonup> 'q gnfastate set" 
  where "PS_\<alpha> M u \<equiv> case PS_map.\<alpha> M u of 
    None \<Rightarrow> None 
  | Some s \<Rightarrow> Some (PSu_set.\<alpha> s)"
definition PS_invar  where 
  "PS_invar M \<equiv> PS_map.invar M \<and> (\<forall>s\<in>ran (PS_map.\<alpha> M). PSu_set.invar s)"
abbreviation "PS_refrel_impl \<equiv> br PS_\<alpha> PS_invar"
lemma single_valued_PS_refrel_impl: "single_valued PS_refrel_impl"
    by refine_post
lemma PS_refrel_impl_def: 
  "(PS,PS') \<in> PS_refrel_impl \<longleftrightarrow> (PS' =
     (\<lambda>u. case map_op_\<alpha> PS_map_ops PS u of None \<Rightarrow> None
         | Some s \<Rightarrow> Some (set_op_\<alpha> PSu_set_ops s)) \<and>
     map_op_invar PS_map_ops PS \<and>
     (\<forall>x\<in>ran (map_op_\<alpha> PS_map_ops PS). set_op_invar PSu_set_ops x))"
unfolding PS_\<alpha>_def[abs_def] PS_invar_def[abs_def] br_def by simp


definition gnfa_\<alpha> where
"gnfa_\<alpha> A \<equiv> case A of (Q, d, P, S) \<Rightarrow> (Q_set.\<alpha> Q, \<delta>_\<alpha> d, PS_\<alpha> P, PS_\<alpha> S)"
definition gnfa_invar where
"gnfa_invar A \<equiv> case A of (Q, d, P, S) \<Rightarrow> Q_set.invar Q \<and> \<delta>_invar d \<and>
    PS_invar P \<and> PS_invar S"
abbreviation "gnfa_refrel_impl \<equiv> br gnfa_\<alpha> gnfa_invar"
lemma single_valued_gnfa_refrel_impl: "single_valued gnfa_refrel_impl"
    by refine_post
lemma gnfa_refrel_impl_def:
  "((a,b,c,d),(a',b',c',d')) \<in> gnfa_refrel_impl \<longleftrightarrow> 
      (a' = set_op_\<alpha> Q_set_ops a \<and> b' = \<delta>_\<alpha> b \<and>
       c' = PS_\<alpha> c \<and> d' = PS_\<alpha> d \<and> set_op_invar Q_set_ops a \<and>
       \<delta>_invar b \<and> PS_invar c \<and> PS_invar d)"
  unfolding gnfa_\<alpha>_def[abs_def] gnfa_invar_def[abs_def] br_def by simp


abbreviation "NFA_Q_copy \<equiv> it_copy NFA_Q_Q_it Q_set.empty Q_set.ins_dj"

lemma NFA_Q_copy_correct[refine_transfer]:
  assumes "NFA_Q_set.invar Q"
  shows "Q_set.\<alpha> (NFA_Q_copy Q) = NFA_Q_set.\<alpha> Q" "Q_set.invar (NFA_Q_copy Q)"
proof-
  have "set_copy NFA_Q_set.\<alpha> NFA_Q_set.invar Q_set.\<alpha> Q_set.invar NFA_Q_copy"
      by (rule it_copy_correct, unfold_locales, simp_all add: Q_set.correct
          NFA_Q_Q_it.it_is_iterator[unfolded IT_tag_def])
  thus "Q_set.invar (NFA_Q_copy Q)" and 
      "Q_set.\<alpha> (NFA_Q_copy Q) = NFA_Q_set.\<alpha> Q"
      unfolding set_copy_def using assms by simp_all
qed


definition gnfa_initial_impl4 where
"gnfa_initial_impl4 \<equiv> do {
    let Q = NFA_Q_copy \<Q>_impl;
    ASSERT (Q_set.invar Q);
    M \<leftarrow> FOREACH (Q_set.\<alpha> Q)
        (\<lambda>q M. RETURN (PS_map.update (State q) (PSu_set.empty ()) M)) 
            (PS_map.empty ());
    let M = PS_map.update Start (PSu_set.empty ()) 
        (PS_map.update End (PSu_set.empty ()) M);
    d \<leftarrow> RETURN (\<delta>_map.empty ());
    RETURN (Q, d, M, M)
}"

lemma gnfa_initial_impl4_refine:
  "gnfa_initial_impl4 \<le> \<Down>gnfa_refrel_impl (gnfa_initial_impl3 \<A>)"
unfolding gnfa_initial_impl4_def gnfa_initial_impl3_def
unfolding Let_def
apply (refine_rcg inj_on_id single_valued_PS_refrel_impl
           single_valued_\<delta>_refrel_impl)
apply (simp_all add: NFA_Q_copy_correct \<Q>_impl_correct) [2]
apply (unfold PS_refrel_impl_def, simp add: PS_map.correct) []
apply (unfold PS_refrel_impl_def ran_def, 
       force simp: PS_map.correct PSu_set.correct) []
apply (unfold \<delta>_refrel_impl_def, simp add: \<delta>_map.correct) []
apply (unfold gnfa_refrel_impl_def PS_\<alpha>_def[abs_def] 
    PS_invar_def[abs_def] ran_def)
apply (auto simp: PS_map.correct PSu_set.correct NFA_Q_copy_correct 
    \<Q>_impl_correct)
done

schematic_lemma gnfa_initial_code_aux: "RETURN ?f \<le> gnfa_initial_impl4"
  unfolding gnfa_initial_impl4_def by refine_transfer
concrete_definition gnfa_initial_code uses gnfa_initial_code_aux



definition "PS_add_impl4 M u v \<equiv> case PS_map.lookup u M of
  None \<Rightarrow> M | Some Mu \<Rightarrow> PS_map.update u (PSu_set.ins v  Mu) M"


lemma PS_add_impl4_refine:
  assumes "(M,M') \<in> PS_refrel_impl"
  shows "(PS_add_impl4 M u v, PS_add M' u v) \<in> PS_refrel_impl"
  using assms unfolding PS_\<alpha>_def[abs_def] PS_invar_def[abs_def] 
       PS_add_def PS_add_impl4_def ran_def
    by (auto intro!: ext simp: PS_map.correct PSu_set.correct 
        split: option.split)

definition "gnfa_\<delta>_update_impl4 d u v r \<equiv>
  case \<delta>_map.lookup u d of
    None \<Rightarrow> \<delta>_map.update u (\<delta>u_map.sng v r) d |
    Some du \<Rightarrow> \<delta>_map.update u (\<delta>u_map.update v r du) d"

lemma gnfa_\<delta>_update_impl4_refine:
  assumes "(d, d') \<in> \<delta>_refrel_impl"
  shows "(gnfa_\<delta>_update_impl4 d u v r, gnfa_\<delta>_update d' u v r) \<in> \<delta>_refrel_impl"
  using assms unfolding gnfa_\<delta>_update_impl4_def 
    gnfa_\<delta>_update_def \<delta>_\<alpha>_def[abs_def] \<delta>_invar_def[abs_def] ran_def
    by (auto intro!: ext simp: \<delta>_map.correct \<delta>u_map.correct split: option.split)

definition "gnfa_\<delta>_insert_impl4 d u v r \<equiv>
  case \<delta>_map.lookup u d of
    None \<Rightarrow> \<delta>_map.update u (\<delta>u_map.sng v r) d |
    Some du \<Rightarrow> let duv' = (
        case \<delta>u_map.lookup v du of
          None \<Rightarrow> r |
          Some duv \<Rightarrow> Plus duv r)
      in \<delta>_map.update u (\<delta>u_map.update v duv' du) d"

lemma gnfa_\<delta>_insert_impl4_refine:
  assumes "(d, d') \<in> \<delta>_refrel_impl"
  shows "(gnfa_\<delta>_insert_impl4 d u v r, gnfa_\<delta>_insert d' u v r) \<in> \<delta>_refrel_impl"
  using assms unfolding gnfa_\<delta>_insert_impl4_def 
    gnfa_\<delta>_insert_def \<delta>_\<alpha>_def[abs_def] \<delta>_invar_def[abs_def] ran_def
    by (auto intro!: ext simp: \<delta>_map.correct \<delta>u_map.correct split: option.split)


definition nfa_to_gnfa_impl4 where
"nfa_to_gnfa_impl4 = do {
  \<A>' \<leftarrow> RETURN gnfa_initial_code;
  \<A>'' \<leftarrow> FOREACH (\<I> \<A>) (\<lambda>v (Q, \<delta>, P, S). 
        RETURN (Q, gnfa_\<delta>_update_impl4 \<delta> Start (State v) rexp.One,
            PS_map.update (State v) (PSu_set.sng Start) P, 
            PS_add_impl4 S Start (State v))) \<A>';
  \<A>'' \<leftarrow> FOREACH (\<F> \<A>) (\<lambda>u (Q, \<delta>, P, S). 
        RETURN (Q, gnfa_\<delta>_update_impl4 \<delta> (State u) End rexp.One,
            PS_add_impl4 P End (State u), 
            PS_map.update (State u) (PSu_set.sng End) S)) \<A>'';
  \<A>'' \<leftarrow> FOREACH (\<Delta> \<A>) (\<lambda>(u,c,v) (Q, \<delta>, P, S). 
        RETURN (Q, gnfa_\<delta>_insert_impl4 \<delta> (State u) (State v) (Atom c),
            PS_add_impl4 P (State v) (State u), 
            PS_add_impl4 S (State u) (State v))) \<A>'';
  RETURN \<A>''
}"

lemma nfa_to_gnfa_impl4_refine:
  "nfa_to_gnfa_impl4 \<le> \<Down>gnfa_refrel_impl (nfa_to_gnfa_impl3 \<A>)"
unfolding nfa_to_gnfa_impl4_def nfa_to_gnfa_impl3_def
apply (refine_rcg inj_on_id single_valued_gnfa_refrel_impl)
apply (rule order_trans[OF gnfa_initial_code.refine gnfa_initial_impl4_refine])

apply simp_all[2]
apply (unfold gnfa_refrel_impl_def gnfa_\<alpha>_def[abs_def] 
    gnfa_invar_def[abs_def], clarsimp, intro conjI) []
apply (insert gnfa_\<delta>_update_impl4_refine, fastforce) []
apply (auto simp add: PS_map.correct PSu_set.correct PS_\<alpha>_def PS_invar_def) []
apply (insert PS_add_impl4_refine, auto) []
apply (insert gnfa_\<delta>_update_impl4_refine, fastforce) []
apply (auto simp: PS_map.correct PSu_set.correct PS_invar_def ran_def) []
apply (insert PS_add_impl4_refine, fastforce) []

apply simp_all[2]
apply (unfold gnfa_refrel_impl_def gnfa_\<alpha>_def[abs_def] 
    gnfa_invar_def[abs_def], clarsimp, intro conjI) []
apply (insert gnfa_\<delta>_update_impl4_refine, fastforce) []
apply (insert PS_add_impl4_refine, fastforce) []
apply (auto simp add: PS_map.correct PSu_set.correct PS_\<alpha>_def PS_invar_def) []
apply (insert gnfa_\<delta>_update_impl4_refine, fastforce) []
apply (insert PS_add_impl4_refine, fastforce) []
apply (auto simp: PS_map.correct PSu_set.correct PS_invar_def ran_def) []

apply simp_all[2]
apply (insert gnfa_\<delta>_insert_impl4_refine PS_add_impl4_refine,
    simp add: gnfa_refrel_impl_def gnfa_\<alpha>_def[abs_def] gnfa_invar_def[abs_def])
apply assumption
done

schematic_lemma nfa_to_gnfa_code_aux: "RETURN ?f \<le> nfa_to_gnfa_impl4"
  unfolding nfa_to_gnfa_impl4_def PS_add_impl4_def 
      gnfa_\<delta>_update_impl4_def gnfa_\<delta>_insert_impl4_def by refine_transfer
concrete_definition nfa_to_gnfa_code uses nfa_to_gnfa_code_aux


definition "PS_the_impl4 M u \<equiv> case lookup u M of 
    None \<Rightarrow> PSu_set.empty () | Some Mu \<Rightarrow> Mu"

lemma PS_the_impl4_refine: 
    "(M,M') \<in> PS_refrel_impl \<Longrightarrow> 
      (PS_the_impl4 M u, PS_the M' u) \<in> br PSu_set.\<alpha> PSu_set.invar"
  unfolding PS_\<alpha>_def[abs_def] PS_invar_def[abs_def] PS_the_def PS_the_impl4_def
  by (auto simp: PS_map.correct PSu_set.correct ran_def split: option.split)

definition "PS_remove_impl4 M u v \<equiv> case lookup u M of
    None \<Rightarrow> M | Some Mu \<Rightarrow> PS_map.update u (PSu_set.delete v Mu) M"

lemma PS_remove_impl4_refine:
    assumes "(M,M') \<in> PS_refrel_impl" "(u,u') \<in> Id"
    shows "(PS_remove_impl4 M u v, PS_remove M' u v) \<in> PS_refrel_impl"
  using assms unfolding PS_\<alpha>_def[abs_def] PS_invar_def[abs_def] 
      PS_remove_def PS_remove_impl4_def ran_def
  by (auto intro!: ext simp: PS_map.correct PSu_set.correct split: option.split)

definition gnfa_remove_state_impl4 where
"gnfa_remove_state_impl4 \<equiv> \<lambda>(Q,\<delta>,P,S) q.
  case q of Start \<Rightarrow> RETURN (Q,\<delta>,P,S) | End \<Rightarrow> RETURN (Q,\<delta>,P,S) | 
      State q' \<Rightarrow> do {
        ASSERT (PSu_set.invar (PS_the_impl4 S q));
        P' \<leftarrow> FOREACH (PSu_set.\<alpha> (PS_the_impl4 S q)) 
            (\<lambda>v P. RETURN (PS_remove_impl4 P v q)) P;
        ASSERT (PSu_set.invar (PS_the_impl4 P q));
        S' \<leftarrow> FOREACH (PSu_set.\<alpha> (PS_the_impl4 P q)) 
            (\<lambda>u S. RETURN (PS_remove_impl4 S u q)) S;
        RETURN (Q_set.delete q' Q, \<delta>, P', S')
      }"

lemma gnfa_remove_state_impl4_refine:
  assumes "(A, A') \<in> gnfa_refrel_impl" "(q,q') \<in> Id"
  shows "gnfa_remove_state_impl4 A q \<le> \<Down>gnfa_refrel_impl
      (gnfa_remove_state_impl3 A' q')"
unfolding gnfa_remove_state_impl4_def gnfa_remove_state_impl3_def
apply (insert assms, refine_rcg, simp split: gnfastate.split)
apply (intro allI conjI impI)
apply (rule RETURN_refine_sv, simp add: single_valued_def, simp) []
apply (rule RETURN_refine_sv, simp add: single_valued_def, simp) []
apply (refine_rcg inj_on_id single_valued_PS_refrel_impl)
apply (insert PS_the_impl4_refine, simp_all add: gnfa_invar_def gnfa_\<alpha>_def) [3]
apply (insert PS_remove_impl4_refine PS_the_impl4_refine, 
    simp_all add: gnfa_invar_def gnfa_\<alpha>_def) [6]
apply (simp add: single_valued_def)
apply (simp add: gnfa_\<alpha>_def[abs_def] gnfa_invar_def[abs_def] Q_set.correct)
done

lemma (in transfer) transfer_gnfastate[refine_transfer]:
  assumes "\<alpha> fs \<le> Fs"
  assumes "\<alpha> fe \<le> Fe"
  assumes "\<And>q. \<alpha> (fq q) \<le> Fq q"
  shows "\<alpha> (gnfastate_case fs fe fq x) \<le> gnfastate_case Fs Fe Fq x"
  using assms by (auto split: gnfastate.split)

schematic_lemma gnfa_remove_state_code_aux: 
    "RETURN ?f \<le> gnfa_remove_state_impl4 A q"
  unfolding gnfa_remove_state_impl4_def by refine_transfer
concrete_definition gnfa_remove_state_code for A q 
    uses gnfa_remove_state_code_aux


definition "gnfa_\<delta>_lookup_impl4 \<equiv> \<lambda>\<delta> u v. 
    case \<delta>_map.lookup u \<delta> of 
      None \<Rightarrow> Zero |
      Some \<delta>u \<Rightarrow> case \<delta>u_map.lookup v \<delta>u of 
        None \<Rightarrow> Zero | 
        Some \<delta>uv \<Rightarrow> \<delta>uv"

lemma gnfa_\<delta>_lookup_impl4_refine:
  fixes \<delta>
  assumes "(\<delta>, \<delta>') \<in> \<delta>_refrel_impl"
  shows "(gnfa_\<delta>_lookup_impl4 \<delta> u v, gnfa_\<delta>_lookup \<delta>' u v) \<in> Id"
using assms unfolding gnfa_\<delta>_lookup_impl4_def 
    \<delta>_\<alpha>_def[abs_def] \<delta>_invar_def[abs_def] ran_def 
  by (auto simp: \<delta>_map.correct \<delta>u_map.correct split: option.split)

definition gnfa_contract_impl4_update_\<delta> where 
"gnfa_contract_impl4_update_\<delta> \<equiv> \<lambda>\<delta> u q v.
    let r1 = gnfa_\<delta>_lookup_impl4 \<delta> u q;
        r2 = gnfa_\<delta>_lookup_impl4 \<delta> q q;
        r3 = gnfa_\<delta>_lookup_impl4 \<delta> q v
    in gnfa_\<delta>_insert_impl4 \<delta> u v (rexp_simped_contract r1 r2 r3)"

lemma gnfa_\<delta>_contract_impl4_update_\<delta>_refine:
  fixes \<delta>
  assumes "(\<delta>\<^sub>1, \<delta>\<^sub>2) \<in> \<delta>_refrel_impl"
  shows "(gnfa_contract_impl4_update_\<delta> \<delta>\<^sub>1 u q v,
      gnfa_contract_impl3_update_\<delta> \<delta>\<^sub>2 u q v) \<in> \<delta>_refrel_impl"
unfolding gnfa_contract_impl4_update_\<delta>_def gnfa_contract_impl3_update_\<delta>_def 
  apply (clarsimp)
  apply (intro conjI ext)
  using assms gnfa_\<delta>_insert_impl4_refine
      gnfa_\<delta>_lookup_impl4_refine[OF assms(1)]
    apply (simp add: \<delta>_map.correct \<delta>u_map.correct) []
  using assms unfolding \<delta>_\<alpha>_def[abs_def] \<delta>_invar_def[abs_def] 
      gnfa_\<delta>_insert_impl4_def ran_def
    apply (auto simp add: \<delta>_map.correct \<delta>u_map.correct split: option.split)
done

definition gnfa_contract_impl4 where
"gnfa_contract_impl4 \<equiv> \<lambda>(Q,\<delta>,P,S) q. do {
  let Pq = (case PS_map.lookup q P of None \<Rightarrow> PSu_set.empty () | 
       Some Pq \<Rightarrow> PSu_set.delete q Pq); 
  let Sq = (case PS_map.lookup q S of None \<Rightarrow> PSu_set.empty () | 
       Some Sq \<Rightarrow> PSu_set.delete q Sq); 
  (Q,\<delta>,P,S) \<leftarrow> RETURN (gnfa_remove_state_code (Q,\<delta>,P,S) q);
  ASSERT (PSu_set.invar Pq);
  ASSERT (PSu_set.invar Sq);
  (Q,\<delta>,P,S) \<leftarrow> FOREACH (PSu_set.\<alpha> Pq) (\<lambda>u (Q',\<delta>',P',S').
    FOREACH (PSu_set.\<alpha> Sq) (\<lambda>v (Q',\<delta>',P',S').
        RETURN (Q', gnfa_contract_impl4_update_\<delta> \<delta>' u q v,
            PS_add_impl4 P' v u, PS_add_impl4 S' u v))
      (Q',\<delta>',P',S')
  ) (Q,\<delta>,P,S);
  RETURN (Q,\<delta>,P,S)
}"


lemma gnfa_contract_impl4_refine:
  assumes "(A, A') \<in> gnfa_refrel_impl" "(q,q') \<in> Id"
  shows "gnfa_contract_impl4 A q \<le> \<Down>gnfa_refrel_impl
      (gnfa_contract_impl3 A' q')"
unfolding gnfa_contract_impl4_def gnfa_contract_impl3_def
using assms apply (refine_rcg inj_on_id single_valued_gnfa_refrel_impl)
apply (blast intro!: order_trans[OF gnfa_remove_state_code.refine 
    gnfa_remove_state_impl4_refine])
apply (unfold PS_the_def gnfa_\<alpha>_def[abs_def] ran_def
         gnfa_invar_def[abs_def] PS_\<alpha>_def[abs_def] PS_invar_def[abs_def],
       auto simp add: PS_map.correct PSu_set.correct split: option.split) [3]
apply simp
apply (unfold PS_the_def gnfa_\<alpha>_def[abs_def] ran_def
         gnfa_invar_def[abs_def] PS_\<alpha>_def[abs_def] PS_invar_def[abs_def],
       auto simp: PS_map.correct PSu_set.correct split: option.split) []
apply simp
apply (insert gnfa_\<delta>_contract_impl4_update_\<delta>_refine PS_add_impl4_refine,
       simp add: gnfa_\<alpha>_def[abs_def] gnfa_invar_def[abs_def]) []
apply simp
done

schematic_lemma gnfa_contract_code_aux: 
    "RETURN ?f \<le> gnfa_contract_impl4 A q"
  unfolding gnfa_contract_impl4_def by refine_transfer
concrete_definition gnfa_contract_code for A q uses gnfa_contract_code_aux



definition "Q_get Q \<equiv> the (Q_set.sel Q (\<lambda>_. True))"

lemma Q_get_correct[simp]: 
  assumes "Q_set.invar Q" "Q_set.\<alpha> Q \<noteq> {}"
  shows "Q_get Q \<in> Q_set.\<alpha> Q" (is "?q \<in> ?Q")
proof-
  from assms obtain q where "q\<in>?Q" by blast
  from Q_set.sel'E[OF assms(1) this TrueI] guess q' .
  thus ?thesis unfolding Q_get_def by simp
qed
  


definition nfa_to_rexp_impl4 where
"nfa_to_rexp_impl4 \<equiv> do {
  (Q,\<delta>,P,S) \<leftarrow> RETURN nfa_to_gnfa_code;
  (Q,\<delta>,P,S) \<leftarrow> 
    WHILE\<^sub>T (\<lambda>(Q,\<delta>,P,S). \<not>Q_set.isEmpty Q) (\<lambda>(Q,\<delta>,P,S). do {
      q \<leftarrow> RETURN (Q_get Q);
      (Q,\<delta>,P,S) \<leftarrow> RETURN (gnfa_contract_code (Q,\<delta>,P,S) (State q));
      RETURN (Q,\<delta>,P,S)
    }) (Q,\<delta>,P,S);
  RETURN (gnfa_\<delta>_lookup_impl4 \<delta> Start End)
}"

lemma nfa_to_rexp_impl4_refine:
  assumes "(\<A>, \<A>') \<in> Id"
  shows "nfa_to_rexp_impl4 \<le> \<Down>Id (nfa_to_rexp_impl3 \<A>)"
unfolding nfa_to_rexp_impl4_def nfa_to_rexp_impl3_def
apply (refine_rcg single_valued_gnfa_refrel_impl)
apply (rule order_trans[OF nfa_to_gnfa_code.refine 
    nfa_to_gnfa_impl4_refine])
apply (force simp add: Q_set.correct split: prod.split)
apply (simp_all add: gnfa_\<alpha>_def[abs_def] 
    gnfa_invar_def[abs_def] Q_set.correct) [2]
apply (fast intro!: order_trans[OF gnfa_contract_code.refine 
    gnfa_contract_impl4_refine])
apply simp
apply (insert gnfa_\<delta>_lookup_impl4_refine, 
       simp add: gnfa_invar_def[abs_def] gnfa_\<alpha>_def[abs_def])
done

schematic_lemma nfa_to_rexp_code_aux: 
    "RETURN ?f \<le> nfa_to_rexp_impl4"
  unfolding nfa_to_rexp_impl4_def by refine_transfer
concrete_definition nfa_to_rexp_code uses nfa_to_rexp_code_aux


lemma nfa_to_rexp_code_correct: 
  "lang nfa_to_rexp_code = \<L> \<A>" (is "lang ?r = _")
proof-
  note nfa_to_rexp_code.refine
  also note nfa_to_rexp_impl4_refine[OF IdI]
  also note nfa_to_rexp_impl3_refine
  also note nfa_to_rexp_impl2_refine
  also note nfa_to_rexp_impl_refine
  also note nfa_to_rexp_abstr_correct
  finally show ?thesis by (elim RETURN_ref_SPECD, simp)
qed


end


record ('Q_set, '\<delta>_map, '\<delta>u_map, 'PS_map, 'PSu_set, 'NFA_Q_set, 'NFA_\<Sigma>_set, 
  'NFA_LTS, 'q::automaton_states, 'c, 'Q_set_more, '\<delta>_map_more, '\<delta>u_map_more, 
  'PS_map_more, 'PSu_set_more, 'NFA_Q_set_more, 'NFA_\<Sigma>_set_more, 'NFA_LTS_more
) nfa_to_rexp_by_lts_ops =
  NFA_Q_set_ops :: "('q, 'NFA_Q_set, 'NFA_Q_set_more) set_ops_scheme"
  NFA_\<Sigma>_set_ops :: "('c, 'NFA_\<Sigma>_set, 'NFA_\<Sigma>_set_more) set_ops_scheme"
  NFA_LTS_ops::"('q,'c,'NFA_LTS,'NFA_LTS_more) lts_ops_scheme"
  Q_set_ops :: "('q, 'Q_set, 'Q_set_more) set_ops_scheme"
  \<delta>_map_ops :: "('q gnfastate, '\<delta>u_map, '\<delta>_map, '\<delta>_map_more) map_ops_scheme"
  \<delta>u_map_ops :: "('q gnfastate, 'c rexp, '\<delta>u_map, '\<delta>u_map_more) map_ops_scheme"
  PS_map_ops :: "('q gnfastate, 'PSu_set, 'PS_map, 'PS_map_more) map_ops_scheme"
  PSu_set_ops :: "('q gnfastate, 'PSu_set, 'PSu_set_more) set_ops_scheme"
  NFA_Q_Q_it :: "'NFA_Q_set \<Rightarrow> ('q, 'Q_set) set_iterator"
  NFA_Q_GNFA_it::"'NFA_Q_set \<Rightarrow> ('q, 
      'Q_set\<times>'\<delta>_map\<times>'PS_map\<times>'PS_map) set_iterator"
  NFA_\<Delta>_GNFA_it::"'NFA_LTS \<Rightarrow> ('q\<times>'c\<times>'q, 
      'Q_set\<times>'\<delta>_map\<times>'PS_map\<times>'PS_map) set_iterator"
  Q_PS_it :: "'Q_set \<Rightarrow> ('q, 'PS_map) set_iterator"
  Q_GNFA_it :: "'Q_set \<Rightarrow> ('q, 'Q_set\<times>'\<delta>_map\<times>'PS_map\<times>'PS_map) set_iterator"
  PS_PS_it :: "'PSu_set \<Rightarrow> ('q gnfastate, 'PS_map) set_iterator"
  PS_GNFA_it :: "'PSu_set \<Rightarrow> ('q gnfastate, 
      'Q_set\<times>'\<delta>_map\<times>'PS_map\<times>'PS_map) set_iterator"

locale nfa_to_rexp_by_lts = 
  nfa!: nfa_by_lts_defs "NFA_Q_set_ops nfa_to_rexp_ops" 
    "NFA_\<Sigma>_set_ops nfa_to_rexp_ops" 
    "NFA_LTS_ops nfa_to_rexp_ops" +
  lts_iterator nfa.d.\<alpha> nfa.d.invar "NFA_\<Delta>_GNFA_it nfa_to_rexp_ops" +
  NFA "nfa.nfa_\<alpha> \<A>_impl" +
  nfa_to_rexp_locale "nfa.nfa_\<alpha> \<A>_impl" 
    "NFA_Q_set_ops nfa_to_rexp_ops" "Q_set_ops nfa_to_rexp_ops" 
    "\<delta>_map_ops nfa_to_rexp_ops" "\<delta>u_map_ops nfa_to_rexp_ops" 
    "PS_map_ops nfa_to_rexp_ops" "PSu_set_ops nfa_to_rexp_ops" 
    "NFA_Q_Q_it nfa_to_rexp_ops" "Q_PS_it nfa_to_rexp_ops"
    "Q_GNFA_it nfa_to_rexp_ops" "PS_PS_it nfa_to_rexp_ops"
    "PS_GNFA_it nfa_to_rexp_ops" 
    "NFA_Q_GNFA_it nfa_to_rexp_ops (nfa.nfa_initial \<A>_impl)"
    "NFA_Q_GNFA_it nfa_to_rexp_ops (nfa.nfa_accepting \<A>_impl)"
    "NFA_\<Delta>_GNFA_it nfa_to_rexp_ops (nfa.nfa_trans \<A>_impl)"
    "nfa.nfa_states \<A>_impl"
  for nfa_to_rexp_ops :: "('Q_set, '\<delta>_map, '\<delta>u_map, 'PS_map, 'PSu_set, 
      'NFA_Q_set, 'NFA_\<Sigma>_set, 'NFA_LTS, 'q::automaton_states, 'c, 
      _, _, _, _, _, _, _, _
  ) nfa_to_rexp_by_lts_ops"
  and \<A>_impl :: "('NFA_Q_set, 'NFA_\<Sigma>_set, 'NFA_LTS) NFA_impl" +
  assumes NFA_invar: "nfa.nfa_invar \<A>_impl"
begin
end


context nfa_by_lts_defs begin

  lemma intro_invars[simp]: 
    assumes "nfa_invar A"
    shows 
      "s.invar (nfa_states A)"
      "s.invar (nfa_accepting A)"
      "s.invar (nfa_initial A)"
      "l.invar (nfa_labels A)"
      "d.invar (nfa_trans A)"
    using assms
    unfolding nfa_invar_def nfa_invar_weak_def nfa_invar_no_props_def
    by simp_all

  lemma invar_imp_is_nfa[simp]: "nfa_invar A \<Longrightarrow> NFA (nfa_\<alpha> A)"
    unfolding nfa_invar_def nfa_invar_weak_def nfa_invar_no_props_def
    by simp_all
end

context nfa_to_rexp_by_lts begin
  definition "nfa_to_rexp \<equiv> nfa_to_rexp_code"
  lemmas nfa_to_rexp_correct = nfa_to_rexp_code_correct[
      folded nfa_to_rexp_def]

  schematic_lemma nfa_to_rexp_unfold: "nfa_to_rexp = ?f"
    unfolding nfa_to_rexp_def 
    unfolding nfa_to_rexp_code_def nfa_to_gnfa_code_def gnfa_initial_code_def
        gnfa_contract_code_def gnfa_remove_state_code_def
        gnfa_contract_impl4_update_\<delta>_def gnfa_\<delta>_lookup_impl4_def
        gnfa_\<delta>_update_impl4_def gnfa_\<delta>_insert_impl4_def PS_add_impl4_def
        PS_the_impl4_def PS_remove_impl4_def Q_get_def
    unfolding nfa.nfa_states_def nfa.nfa_initial_def 
        nfa.nfa_accepting_def nfa.nfa_trans_def 
    by (rule refl)
end
*)

end
