header {* Control Flow Graph *}
theory SM_Cfg
imports SM_Syntax "../Lib/LTS" "../Lib/SOS_Misc_Add" "~~/HOL/Library/Code_Char"
begin
  text {*
    We define the control flow graph of SM, in the style of
    an structural operational semantics.
  *}


  text {*
    The edges of the CFG are labeled with actions. These are
    assignments, tests, and skip.
    Assignment actions come in three versions: Assignment to local variables,
    assignment to global variables, and assignment to variables without
    specification of a scope. These will be mapped to local or global variables,
    depending on the scope.
  *}
  datatype_new action = 
      AAssign exp ident exp | AAssign_local exp ident exp | AAssign_global exp ident exp
    | ATest exp | ASkip

  text {*
    Moreover, the semantics uses internal actions to handle break and continue
    statements. For well-formed programs, this type of actions will
    never be exposed
  *}
  datatype_new brk_ctd = AIBreak | AIContinue

  text {*
    The control flow graph is a graph over commands, labeled with actions
  *}
  inductive cfg :: "cmd \<Rightarrow> (action + brk_ctd) \<Rightarrow> cmd \<Rightarrow> bool" where
    Ass: "cfg (Assign c x e) (Inl (AAssign c x e)) Skip"
  | Assl: "cfg (Assign_local c x e) (Inl (AAssign_local c x e)) Skip"
  | Assg: "cfg (Assign_global c x e) (Inl (AAssign_global c x e)) Skip"
  | Test: "cfg (Test e) (Inl (ATest e)) Skip"

  | Seq1: "cfg (Seq Skip c2) (Inl ASkip) c2"
  | Seq2': "\<lbrakk> cfg c1 e Skip \<rbrakk> \<Longrightarrow> cfg (Seq c1 c2) e c2"
  | Seq2: "\<lbrakk> cfg c1 e c1'; c1'\<noteq>Skip \<rbrakk> \<Longrightarrow> cfg (Seq c1 c2) e (Seq c1' c2)"

  | Or1: "cfg c1 e c1' \<Longrightarrow> cfg (Or c1 c2) e c1'"
  | Or2: "cfg c2 e c2' \<Longrightarrow> cfg (Or c1 c2) e c2'"

  (*| Loop: "cfg (Loop c) (Inl ASkip) (Iterate c c)"*)

  | Break: "cfg (Break) (Inr AIBreak) Invalid"
  | Continue: "cfg (Continue) (Inr AIContinue) Invalid"
  
  | Iterate1: "cfg (Iterate Skip c) (Inl ASkip) (Iterate c c)"
  | Iterate2': "\<lbrakk> cfg c (Inl e) Skip \<rbrakk> \<Longrightarrow> cfg (Iterate c cb) (Inl e) (Iterate cb cb)"
  | Iterate2: "\<lbrakk> cfg c (Inl e) c'; c'\<noteq>Skip \<rbrakk> \<Longrightarrow> cfg (Iterate c cb) (Inl e) (Iterate c' cb)"
  | IterateBrk: "\<lbrakk> cfg c (Inr AIBreak) _  \<rbrakk> \<Longrightarrow> cfg (Iterate c cb) (Inl ASkip) Skip"
  | IterateCtd: "\<lbrakk> cfg c (Inr AIContinue) _  \<rbrakk> \<Longrightarrow> cfg (Iterate c cb) (Inl ASkip) (Iterate cb cb)"

  inductive_cases Seq_cases: "cfg (Seq c1 c2) e c'"

(*
section {* Well-formedness checks *}
  (* Well-formedness check: Break/Continue inside loops *)
  primrec wf_cmd_aux :: "cmd \<Rightarrow> bool" where
    "wf_cmd_aux (Assign _ _) \<longleftrightarrow> True"
  | "wf_cmd_aux (Assign_local _ _) \<longleftrightarrow> True"
  | "wf_cmd_aux (Assign_global _ _) \<longleftrightarrow> True"
  | "wf_cmd_aux (Test _) \<longleftrightarrow> True"
  | "wf_cmd_aux (Skip) \<longleftrightarrow> True"
  | "wf_cmd_aux (Seq c1 c2) \<longleftrightarrow> wf_cmd_aux c1 \<and> wf_cmd_aux c2"
  | "wf_cmd_aux (Or c1 c2) \<longleftrightarrow> wf_cmd_aux c1 \<and> wf_cmd_aux c2"
  | "wf_cmd_aux (Loop c1) \<longleftrightarrow> wf_cmd_aux c1"
  | "wf_cmd_aux (Break) \<longleftrightarrow> True"
  | "wf_cmd_aux (Continue) \<longleftrightarrow> True"
  | "wf_cmd_aux (Iterate c1 c2) \<longleftrightarrow> wf_cmd_aux c1 \<and> wf_cmd_aux c2"
  | "wf_cmd_aux (Invalid) \<longleftrightarrow> False"

  primrec wf_cmd :: "cmd \<Rightarrow> bool" where
    "wf_cmd (Assign _ _) \<longleftrightarrow> True"
  | "wf_cmd (Assign_local _ _) \<longleftrightarrow> True"
  | "wf_cmd (Assign_global _ _) \<longleftrightarrow> True"
  | "wf_cmd (Test _) \<longleftrightarrow> True"
  | "wf_cmd (Skip) \<longleftrightarrow> True"
  | "wf_cmd (Seq c1 c2) \<longleftrightarrow> wf_cmd c1 \<and> wf_cmd c2"
  | "wf_cmd (Or c1 c2) \<longleftrightarrow> wf_cmd c1 \<and> wf_cmd c2"
  | "wf_cmd (Loop c1) \<longleftrightarrow> wf_cmd_aux c1"
  | "wf_cmd (Break) \<longleftrightarrow> False"
  | "wf_cmd (Continue) \<longleftrightarrow> False"
  | "wf_cmd (Iterate c1 c2) \<longleftrightarrow> wf_cmd_aux c1 \<and> wf_cmd_aux c2"
  | "wf_cmd (Invalid) \<longleftrightarrow> False"
  

  (* Well-formedness is preserved *)
  lemma cfg_wf_aux1:
    assumes "cfg c (Inl e) c'"
    assumes "wf_cmd_aux c"
    shows "wf_cmd_aux c'"
    using assms
    by (induction c "Inl e::action+brk_ctd" c') auto

  lemma cfg_wf_aux2:
    assumes "wf_cmd c" 
    assumes "cfg c e c'" 
    shows "wf_cmd c' \<and> isl e"
    using assms(2,1)
    by induction (auto simp: cfg_wf_aux1)

  theorem cfg_wf: "\<lbrakk> wf_cmd c; cfg c e c' \<rbrakk> \<Longrightarrow> wf_cmd c'"
    -- "Steps in CFG preserve well-formedness"
    using cfg_wf_aux2 by simp
   
  theorem cfg_no_internal: "\<lbrakk> wf_cmd c; cfg c e c' \<rbrakk> \<Longrightarrow> isl e"
    -- "No internal actions on CFG-edges"
    using cfg_wf_aux2 by simp

  (* Commands do not get stuck (except on skip) *)
  lemma cfg_no_stuck_aux: "\<lbrakk>wf_cmd_aux c; c\<noteq>Skip\<rbrakk> \<Longrightarrow> \<exists>e c'. cfg c e c'"
    apply (induction c)
    apply (auto intro!: cfg.intros)
    apply (case_tac "c1=Skip")
    apply (auto intro!: cfg.intros) []
    apply clarsimp
    apply (case_tac e)
    apply (auto intro!: cfg.intros) []
    apply (case_tac b)
    apply (auto intro!: cfg.intros) [2]
    done

  theorem cfg_no_stuck: "\<lbrakk>wf_cmd c; c\<noteq>Skip\<rbrakk> \<Longrightarrow> \<exists>e c'. cfg c e c'"
    -- "Well-formed commands have outgoing edges, except skip"
    using assms
    by (induction c) (auto intro: cfg.intros cfg_no_stuck_aux)
*)

section {* Reachable Nodes *}

interpretation cfg_lts: LTS cfg .

(* Finite set of reachable CFG nodes *)

primrec reachable_term_order_aux :: "cmd \<Rightarrow> nat" where
  "reachable_term_order_aux (Assign c x e) = 0"
| "reachable_term_order_aux (Assign_local c x e) = 0"
| "reachable_term_order_aux (Assign_global c x e) = 0"
| "reachable_term_order_aux (Test e) = 0"
| "reachable_term_order_aux (Skip) = 0"
| "reachable_term_order_aux (Seq c1 c2) = reachable_term_order_aux c1 + reachable_term_order_aux c2"
| "reachable_term_order_aux (Or c1 c2) = reachable_term_order_aux c1 + reachable_term_order_aux c2"
(*| "reachable_term_order_aux (Loop c) = 1 + 2*reachable_term_order_aux c"*)
| "reachable_term_order_aux (Break) = 0"
| "reachable_term_order_aux (Continue) = 0"
| "reachable_term_order_aux (Iterate c1 c2) = reachable_term_order_aux c1 + reachable_term_order_aux c2"
| "reachable_term_order_aux Invalid = 0"

function approx_reachable :: "cmd \<Rightarrow> cmd set" where
  "approx_reachable (Assign c x e) = {Assign c x e, Skip}"
| "approx_reachable (Assign_local c x e) = {Assign_local c x e, Skip}"
| "approx_reachable (Assign_global c x e) = {Assign_global c x e, Skip}"
| "approx_reachable (Test e) = {Test e, Skip}"
| "approx_reachable (Skip) = {Skip}"
| "approx_reachable (Seq c1 c2) = (approx_reachable c2) \<union> (do { c1\<leftarrow>approx_reachable c1; {Seq c1 c2} })"
| "approx_reachable (Or c1 c2) = insert (Or c1 c2) (approx_reachable c1 \<union> approx_reachable c2)"
(*| "approx_reachable (Loop c) = insert (Loop c) (approx_reachable (Iterate c c))"*)
| "approx_reachable (Break) = {Break,Invalid}"
| "approx_reachable (Continue) = {Continue,Invalid}"
| "approx_reachable (Iterate c1 c2) = insert Skip
    (do { c1\<leftarrow>approx_reachable c1; {Iterate c1 c2} })
  \<union> (do { c2'\<leftarrow>approx_reachable c2; {Iterate c2' c2} })"
| "approx_reachable Invalid = {Invalid}"
by pat_completeness auto

termination
  apply (relation "reachable_term_order_aux <*mlex*> measure size")
  apply (metis wf_measure wf_mlex)
  apply (auto simp: mlex_prod_def)
  done

lemma finite_approx_reachable[simp, intro!]: "finite (approx_reachable c)"
  by (induction c rule: approx_reachable.induct) auto

lemma approx_reachable_refl[simp]: "c\<in>approx_reachable c"
  by (induction c rule: approx_reachable.induct) auto

lemma approx_reachable_closed:
  assumes "c\<in>approx_reachable c0"
  assumes "cfg c a c'"
  shows "c'\<in>approx_reachable c0"
  using assms
  apply (induction c0 arbitrary: c c' rule: approx_reachable.induct)
  apply (auto elim: cfg.cases) [] 
  apply (auto elim: cfg.cases) [] 
  apply (auto elim: cfg.cases) [] 
  apply (auto elim: cfg.cases) [] 
  apply (auto elim: cfg.cases) [] 
  apply (auto elim: cfg.cases) [] (* Takes quite long. *)
  apply (auto, erule cfg.cases, simp_all, force+) [] (* Takes quite long. *)
  apply (auto elim: cfg.cases) [] 
  apply (auto elim: cfg.cases) [] 
  apply (auto elim: cfg.cases) [] (* Takes quite long. *)
  apply (auto elim: cfg.cases) [] 
  done

lemma approx_reachable_approx: "cfg_lts.reachable c c' \<Longrightarrow> c'\<in>approx_reachable c"
  unfolding cfg_lts.reachable_def
proof safe
  fix p
  assume "cfg_lts.path c p c'"
  thus "c'\<in>approx_reachable c"
    by (induction p arbitrary: c' rule: rev_induct)
       (auto intro: approx_reachable_closed)
qed

lemma finite_cfg_reachable: "finite (Collect (cfg_lts.reachable c))"
  apply (rule finite_subset[OF _ finite_approx_reachable])
  using approx_reachable_approx by auto

section {* Executable Successor Function *}
(*
(* CFG successors *)
primrec cfg_succ :: "cmd \<Rightarrow> ((action+brk_ctd) \<times> cmd) set" where
  "cfg_succ (Assign c x e) = {((Inl (AAssign c x e)), Skip)}"
| "cfg_succ (Assign_local c x e) = {((Inl (AAssign_local c x e)), Skip)}"
| "cfg_succ (Assign_global c x e) = {((Inl (AAssign_global c x e)), Skip)}"
| "cfg_succ (Test e) = {((Inl (ATest e)), Skip)}"
| "cfg_succ (Seq c1 c2) =
    (if c1=Skip then {((Inl ASkip), c2)} else {})
  \<union> { (e,Seq c1' c2) | c1' e.  (e,c1')\<in>cfg_succ c1}"
| "cfg_succ (Or c1 c2) = {(Inl ASkip, c1), (Inl ASkip, c2)}"
| "cfg_succ (Loop c) =  {((Inl ASkip), (Iterate c c))}"
| "cfg_succ (Break) = {(Inr AIBreak, Invalid)}"
| "cfg_succ (Continue) = {(Inr AIContinue, Invalid)}"
| cfg_succ_iterate: 
  "cfg_succ (Iterate c cb) =
  (if c=Skip then {(Inl ASkip, Iterate cb cb)} else {})
  \<union> (
    \<lambda> (Inl e,c') \<Rightarrow> (Inl e, Iterate c' cb)
    | (Inr AIBreak,_) \<Rightarrow> (Inl ASkip,Skip)
    | (Inr AIContinue,_) \<Rightarrow> (Inl ASkip, Iterate cb cb)
  )`cfg_succ c"
| "cfg_succ Skip = {}"
| "cfg_succ Invalid = {}"


declare cfg_succ_iterate[simp del]
lemma cfg_succ_iterate'[simp]: "cfg_succ (Iterate c cb) = 
  (if c=Skip then {(Inl ASkip, Iterate cb cb)} else {})
  \<union> { (Inl e, Iterate c' cb) | e c'. (Inl e, c') \<in> cfg_succ c }
  \<union> (if \<exists>c'. (Inr AIBreak,c')\<in> cfg_succ c then { (Inl ASkip, Skip)} else {})
  \<union> (if \<exists>c'. (Inr AIContinue,c')\<in> cfg_succ c then { (Inl ASkip, Iterate cb cb)} else {})"
  unfolding cfg_succ_iterate
  apply (rule sym)
  apply (cases "c=Skip")
  apply simp
  apply safe
  apply simp_all
  apply force
  apply (force split: split_if_asm)
  apply (force split: split_if_asm)
  apply (auto split: split_if_asm sum.splits brk_ctd.splits)
  done


export_code cfg_succ in SML

lemma cfg_succ_iff_cfg: "(a,c')\<in>cfg_succ c \<longleftrightarrow> cfg c a c'"
  apply rule
  apply (induction c arbitrary: a c')
  apply (auto intro: cfg.intros split: split_if_asm)
  
  apply (induction rule: cfg.induct)
  apply auto
  done

theorem cfg_succ_correct: "cfg_succ c = {(a,c'). cfg c a c'}"
  using cfg_succ_iff_cfg by auto

lemma cfg_succ_finite[simp, intro!]: "finite (cfg_succ c)"
  apply (induction c)
  using [[simproc add: finite_Collect]]
  by (simp_all del: cfg_succ_iterate' add: cfg_succ_iterate)
*)


end

