theory Explore_Tree
imports Tutorial_Basic
begin
  text \<open>
    Note that we represent directed graphs as relations of type @{typ "'a\<times>'a"}.
    The successors of a node can then be expressed using the relation-image operator
    @{term "op ``"}, i.e. we have @{lemma "E``{u} = {v . (u,v)\<in>E}" by auto}.
    \<close>

  subsection \<open>Recursion combinator\<close>

  text \<open> Here, we demonstrate a simple recursive function, 
    which scans a tree for a specific node, and returns true if this node
    is found. 

    We parameterize the function over the used recursion combinator,
    such that we can instantiate it for partial and total correctness.
  \<close>

  definition "find_node RECcomb T v0 vt \<equiv> RECcomb (\<lambda>rec v. 
    if v=vt then   (* Return true if node is found *)
      RETURN True
    else           (* Otherwise, iterate over successors *)
      FOREACHci (\<lambda>_ r. r \<longrightarrow> (v,vt)\<in>T\<^sup>*) (* Invariant annotated to loop *) 
        (T``{v})    (* Successors of node v *) 
        (\<lambda>x. \<not>x)    (* Iterate until result is true *)
        (\<lambda>v _. rec v) (* Invoke recursively *)
        False       (* Initial state for loop *) 
  ) v0"
      
  abbreviation "find_node_partial \<equiv> find_node REC"
  abbreviation "find_node_total \<equiv> find_node RECT"

  text \<open>The correctness lemma for the partial correct variant. 
    We only prove the simple direction of correctness, namely that
    a true result means that node @{term vt} was found.

    Note that the tree must be finitely branching, in order for 
    the foreach-loop to be defined. 
    
    However, we do not require @{term T} to be a tree or DAG. 
    It may be any graph, containing cycles, which lead to non-termination of the
    algorithm. \<close>
  lemma 
    assumes [simp]: "\<And>v. finite (T``{v})"
    shows "find_node_partial T v0 vt \<le> SPEC (\<lambda>r. r \<longrightarrow> (v0,vt)\<in>T\<^sup>*)"
    unfolding find_node_def
    apply (refine_vcg 
      REC_rule[where pre="\<lambda>v. (v0,v)\<in>T\<^sup>*"]
      ) -- \<open>Invoke VCG, hinting the precondition for the recursion\<close>
    apply auto [4] -- \<open>Solve some of the VCs automatically\<close>

    txt \<open>The next lines apply the induction/recursion hypothesis. 
      This is a standard pattern for proving recursive functions.
      The order-trans rule has the effect of a consequence rule.
      It is required if the current goal does not exactly match the
      inductive assumption. 
      
      Note that @{text rprems} applies a premise to solve the current subgoal,
      where the premise may be an implication with bound variables 
      (@{text "\<And>x\<^sub>1 \<dots> x\<^sub>n. \<lbrakk>P\<^sub>1;\<dots>;P\<^sub>m\<rbrakk> \<Longrightarrow> Q"}).\<close>
    apply (rule order_trans)
    apply rprems
    apply auto [2] 

    apply auto []
    done

  text \<open>For the total correct variant, we additionally require any path in T to
    be finite. 
    Note the opposite argument order required for the @{const wf} - predicate.
  \<close>
  lemma 
    assumes [simp]: "\<And>v. finite (T``{v})"
    assumes [simp]: "wf (T\<inverse>)"
    shows "find_node_total T v0 vt \<le> SPEC (\<lambda>r. r \<longrightarrow> (v0,vt)\<in>T\<^sup>*)"
    unfolding find_node_def
    apply (refine_vcg 
      RECT_rule[where pre="\<lambda>v. (v0,v)\<in>T\<^sup>*" and V="T\<inverse>"]
      )
    apply (auto) [5]
    apply (rule order_trans)
    apply rprems
    apply auto [2]
    apply auto [2]
    done

  text \<open>There are many ways to define the invariant of the foreach-loop.
    Note, however, that it depends on the variable $v$, which is not visible
    from outside. As an alternative to annotating the invariant, we can use
    Isabelle/HOL's apply style reasoning.\<close>  

  definition "find_node' T v0 vt \<equiv> REC (\<lambda>rec v. 
    if v=vt then   
      RETURN True
    else           
      FOREACHc (T``{v}) 
        (\<lambda>x. \<not>x)   
        (\<lambda>v _. rec v) 
        False
  ) v0"
      
  lemma 
    assumes [simp]: "\<And>v. finite (T``{v})"
    shows "find_node' T v0 vt \<le> SPEC (\<lambda>r. r \<longrightarrow> (v0,vt)\<in>T\<^sup>*)"
    unfolding find_node'_def
    apply (refine_vcg 
      REC_rule[where pre="\<lambda>v. (v0,v)\<in>T\<^sup>*"]
      )
    txt \<open>Note that the VCG has not resolved the foreach-loop\<close>
    apply auto [2]

    txt \<open>After discharging some VCs, we manually apply the rule for the foreach-loop,
      thereby sneaking in the invariant. \<close>
    apply (rename_tac rec v)
    apply (rule_tac I="(\<lambda>_ r. r \<longrightarrow> (v,vt)\<in>T\<^sup>*)" in FOREACHc_rule)
    apply auto [2]

    apply (rule order_trans)
    apply rprems
    apply auto [2]

    apply auto [2]
    done

  subsection \<open>While-Loop\<close>
  text \<open>An alternative to explore the nodes of a graph is a workset algorithm.
    It has the advantage that it allows multiple refinements of the workset, e.g.,
    to influence the search order.\<close>

  definition "abs_ws T v0 vt \<equiv> do {
    let s = ({v0},False);
    (* Iterate while workset is non-empty, and no path has been found.
      Note that we specify a partial correct loop here. *)
    s \<leftarrow> WHILE (\<lambda>(ws,r). \<not>r \<and> ws\<noteq>{}) (\<lambda>(ws,_). do {
      ASSERT (ws \<noteq> {});      (* Workaround quirk: Invariant not available for further refinement, so we assert it again *)
      v \<leftarrow> SPEC (\<lambda>x. x\<in>ws);  (* Pick element from workset *)
      let ws = ws - {v};     (* Remove it *)
      if v = vt then
        RETURN (ws,True)       (* If target node found, return true *)
      else do {
        let ws = ws \<union> T``{v};  (* Otherwise, add successors to workset *)
        RETURN (ws,False)
      }
    }) s;
    let (_,r) = s;    (* Extract the result *)
    RETURN r
  }"

  text \<open>The correctness proof is straightforward.
    Again, we only prove the easy direction.
    The only creativity goes into the loop invariant. 
    Here, we use that the workset only contains reachable nodes.
    \<close>

  lemma abs_ws_correct: "abs_ws T v0 vt \<le> SPEC (\<lambda>r. r \<longrightarrow> (v0,vt)\<in>T\<^sup>*)"
    unfolding abs_ws_def
    apply (refine_vcg 
      WHILE_rule[where I="\<lambda>(ws,r). ws \<subseteq> T\<^sup>*``{v0} \<and> (r \<longrightarrow> (v0,vt)\<in>T\<^sup>*)" ])
    txt \<open>The VCs go away automatically except one, for which sledgehammer
      quickly finds a solution:\<close>
    apply (auto)
    by (metis Image_singleton_iff rtrancl_image_advance subsetCE)


end

