
structure STArray = struct

datatype 'a Cell = Invalid | Value of 'a array;

exception AccessedOldVersion;

type 'a array = 'a Cell Unsynchronized.ref;

fun fromList l = Unsynchronized.ref (Value (Array.fromList l));
fun array (size, v) = Unsynchronized.ref (Value (Array.array (size,v)));
fun tabulate (size, f) = Unsynchronized.ref (Value (Array.tabulate(size, f)));
fun sub (Unsynchronized.ref Invalid, idx) = raise AccessedOldVersion |
    sub (Unsynchronized.ref (Value a), idx) = Array.sub (a,idx);
fun update (aref,idx,v) =
  case aref of
    (Unsynchronized.ref Invalid) => raise AccessedOldVersion |
    (Unsynchronized.ref (Value a)) => (
      aref := Invalid;
      Array.update (a,idx,v);
      Unsynchronized.ref (Value a)
    );

fun length (Unsynchronized.ref Invalid) = raise AccessedOldVersion |
    length (Unsynchronized.ref (Value a)) = Array.length a

fun grow (aref, i, x) = case aref of 
  (Unsynchronized.ref Invalid) => raise AccessedOldVersion |
  (Unsynchronized.ref (Value a)) => (
    let val len=Array.length a;
        val na = Array.array (len+i,x) 
    in
      aref := Invalid;
      Array.copy {src=a, dst=na, di=0};
      Unsynchronized.ref (Value na)
    end
    );

fun shrink (aref, sz) = case aref of
  (Unsynchronized.ref Invalid) => raise AccessedOldVersion |
  (Unsynchronized.ref (Value a)) => (
    if sz > Array.length a then 
      raise Size
    else (
      aref:=Invalid;
      Unsynchronized.ref (Value (Array.tabulate (sz,fn i => Array.sub (a,i))))
    )
  );

structure IsabelleMapping = struct
type 'a ArrayType = 'a array;

fun new_array (a:'a) (n:int) = array (n, a);

fun array_length (a:'a ArrayType) = length a;

fun array_get (a:'a ArrayType) (i:int) = sub (a, i);

fun array_set (a:'a ArrayType) (i:int) (e:'a) = update (a, i, e);

fun array_of_list (xs:'a list) = fromList xs;

fun array_grow (a:'a ArrayType) (i:int) (x:'a) = grow (a, i, x);

fun array_shrink (a:'a ArrayType) (sz:int) = shrink (a,sz);

end;

end;

structure FArray = struct
  datatype 'a Cell = Value of 'a Array.array | Upd of (int*'a*'a Cell Unsynchronized.ref);

  type 'a array = 'a Cell Unsynchronized.ref;

  fun array (size,v) = Unsynchronized.ref (Value (Array.array (size,v)));
  fun tabulate (size, f) = Unsynchronized.ref (Value (Array.tabulate(size, f)));
  fun fromList l = Unsynchronized.ref (Value (Array.fromList l));

  fun sub (Unsynchronized.ref (Value a), idx) = Array.sub (a,idx) |
      sub (Unsynchronized.ref (Upd (i,v,cr)),idx) =
        if i=idx then v
        else sub (cr,idx);

  fun length (Unsynchronized.ref (Value a)) = Array.length a |
      length (Unsynchronized.ref (Upd (i,v,cr))) = length cr;

  fun realize_aux (aref, v) =
    case aref of
      (Unsynchronized.ref (Value a)) => (
        let
          val len = Array.length a;
          val a' = Array.array (len,v);
        in
          Array.copy {src=a, dst=a', di=0};
          Unsynchronized.ref (Value a')
        end
      ) |
      (Unsynchronized.ref (Upd (i,v,cr))) => (
        let val res=realize_aux (cr,v) in
          case res of
            (Unsynchronized.ref (Value a)) => (Array.update (a,i,v); res)
        end
      );

  fun realize aref =
    case aref of
      (Unsynchronized.ref (Value _)) => aref |
      (Unsynchronized.ref (Upd (i,v,cr))) => realize_aux(aref,v);

  fun update (aref,idx,v) =
    case aref of
      (Unsynchronized.ref (Value a)) => (
        let val nref=Unsynchronized.ref (Value a) in
          aref := Upd (idx,Array.sub(a,idx),nref);
          Array.update (a,idx,v);
          nref
        end
      ) |
      (Unsynchronized.ref (Upd _)) =>
        let val ra = realize_aux(aref,v) in
          case ra of
            (Unsynchronized.ref (Value a)) => Array.update (a,idx,v);
          ra
        end
      ;

  fun grow (aref, inc, x) = case aref of 
    (Unsynchronized.ref (Value a)) => (
      let val len=Array.length a;
          val na = Array.array (len+inc,x) 
      in
        Array.copy {src=a, dst=na, di=0};
        Unsynchronized.ref (Value na)
      end
      )
  | (Unsynchronized.ref (Upd _)) => (
    grow (realize aref, inc, x)
  );

  fun shrink (aref, sz) = case aref of
    (Unsynchronized.ref (Value a)) => (
      if sz > Array.length a then 
        raise Size
      else (
        Unsynchronized.ref (Value (Array.tabulate (sz,fn i => Array.sub (a,i))))
      )
    ) |
    (Unsynchronized.ref (Upd _)) => (
      shrink (realize aref,sz)
    );

structure IsabelleMapping = struct
type 'a ArrayType = 'a array;

fun new_array (a:'a) (n:int) = array (n, a);

fun array_length (a:'a ArrayType) = length a;

fun array_get (a:'a ArrayType) (i:int) = sub (a, i);

fun array_set (a:'a ArrayType) (i:int) (e:'a) = update (a, i, e);

fun array_of_list (xs:'a list) = fromList xs;

fun array_grow (a:'a ArrayType) (i:int) (x:'a) = grow (a, i, x);

fun array_shrink (a:'a ArrayType) (sz:int) = shrink (a,sz);

fun array_get_oo (d:'a) (a:'a ArrayType) (i:int) = 
  sub (a,i) handle Subscript => d

fun array_set_oo (d:(unit->'a ArrayType)) (a:'a ArrayType) (i:int) (e:'a) = 
  update (a, i, e) handle Subscript => d ()

end;
end;




structure HOL : sig
  type 'a equal
  val equal : 'a equal -> 'a -> 'a -> bool
  val eq : 'a equal -> 'a -> 'a -> bool
  datatype 'a itself = Type
end = struct

type 'a equal = {equal : 'a -> 'a -> bool};
val equal = #equal : 'a equal -> 'a -> 'a -> bool;

fun eq A_ a b = equal A_ a b;

datatype 'a itself = Type;

end; (*struct HOL*)

structure Product_Type : sig
  val fst : 'a * 'b -> 'a
  val snd : 'a * 'b -> 'b
  val apsnd : ('a -> 'b) -> 'c * 'a -> 'c * 'b
  val scomp : ('a -> 'b * 'c) -> ('b -> 'c -> 'd) -> 'a -> 'd
  val map_pair : ('a -> 'b) -> ('c -> 'd) -> 'a * 'c -> 'b * 'd
  val equal_proda : 'a HOL.equal -> 'b HOL.equal -> 'a * 'b -> 'a * 'b -> bool
  val equal_prod : 'a HOL.equal -> 'b HOL.equal -> ('a * 'b) HOL.equal
  val equal_unita : unit -> unit -> bool
  val equal_unit : unit HOL.equal
  val equal_bool : bool -> bool -> bool
end = struct

fun fst (a, b) = a;

fun snd (a, b) = b;

fun apsnd f (x, y) = (x, f y);

fun scomp f g = (fn x => let
                           val (a, b) = f x;
                         in
                           g a b
                         end);

fun map_pair f g (a, b) = (f a, g b);

fun equal_proda A_ B_ (aa, ba) (a, b) = HOL.eq A_ aa a andalso HOL.eq B_ ba b;

fun equal_prod A_ B_ = {equal = equal_proda A_ B_} : ('a * 'b) HOL.equal;

fun equal_unita u v = true;

val equal_unit = {equal = equal_unita} : unit HOL.equal;

fun equal_bool p true = p
  | equal_bool p false = not p
  | equal_bool true p = p
  | equal_bool false p = not p;

end; (*struct Product_Type*)

structure Orderings : sig
  type 'a bot
  val bot : 'a bot -> 'a
  type 'a ord
  val less_eq : 'a ord -> 'a -> 'a -> bool
  val less : 'a ord -> 'a -> 'a -> bool
  val max : 'a ord -> 'a -> 'a -> 'a
  type 'a top
  val top : 'a top -> 'a
  type 'a preorder
  val ord_preorder : 'a preorder -> 'a ord
  type 'a order
  val preorder_order : 'a order -> 'a preorder
  type 'a linorder
  val order_linorder : 'a linorder -> 'a order
  type 'a order_bot
  val bot_order_bot : 'a order_bot -> 'a bot
  val order_order_bot : 'a order_bot -> 'a order
  type 'a order_top
  val order_order_top : 'a order_top -> 'a order
  val top_order_top : 'a order_top -> 'a top
end = struct

type 'a bot = {bot : 'a};
val bot = #bot : 'a bot -> 'a;

type 'a ord = {less_eq : 'a -> 'a -> bool, less : 'a -> 'a -> bool};
val less_eq = #less_eq : 'a ord -> 'a -> 'a -> bool;
val less = #less : 'a ord -> 'a -> 'a -> bool;

fun max A_ a b = (if less_eq A_ a b then b else a);

type 'a top = {top : 'a};
val top = #top : 'a top -> 'a;

type 'a preorder = {ord_preorder : 'a ord};
val ord_preorder = #ord_preorder : 'a preorder -> 'a ord;

type 'a order = {preorder_order : 'a preorder};
val preorder_order = #preorder_order : 'a order -> 'a preorder;

type 'a linorder = {order_linorder : 'a order};
val order_linorder = #order_linorder : 'a linorder -> 'a order;

type 'a order_bot = {bot_order_bot : 'a bot, order_order_bot : 'a order};
val bot_order_bot = #bot_order_bot : 'a order_bot -> 'a bot;
val order_order_bot = #order_order_bot : 'a order_bot -> 'a order;

type 'a order_top = {order_order_top : 'a order, top_order_top : 'a top};
val order_order_top = #order_order_top : 'a order_top -> 'a order;
val top_order_top = #top_order_top : 'a order_top -> 'a top;

end; (*struct Orderings*)

structure Stringa : sig
  datatype nibble = Nibble0 | Nibble1 | Nibble2 | Nibble3 | Nibble4 | Nibble5 |
    Nibble6 | Nibble7 | Nibble8 | Nibble9 | NibbleA | NibbleB | NibbleC |
    NibbleD | NibbleE | NibbleF
  val equal_char : char HOL.equal
  val equal_literal : string HOL.equal
end = struct

datatype nibble = Nibble0 | Nibble1 | Nibble2 | Nibble3 | Nibble4 | Nibble5 |
  Nibble6 | Nibble7 | Nibble8 | Nibble9 | NibbleA | NibbleB | NibbleC | NibbleD
  | NibbleE | NibbleF;

val equal_char = {equal = (fn a => fn b => ((a : char) = b))} : char HOL.equal;

val equal_literal = {equal = (fn a => fn b => ((a : string) = b))} :
  string HOL.equal;

end; (*struct Stringa*)

structure Fun : sig
  val id : 'a -> 'a
end = struct

fun id x = (fn xa => xa) x;

end; (*struct Fun*)

structure Arith : sig
  datatype int = Int_of_integer of IntInf.int
  datatype nat = Nat of IntInf.int
  datatype num = One | Bit0 of num | Bit1 of num
  val integer_of_int : int -> IntInf.int
  val ord_integer : IntInf.int Orderings.ord
  val nat : int -> nat
  val integer_of_nat : nat -> IntInf.int
  val plus_nat : nat -> nat -> nat
  val one_nat : nat
  val suc : nat -> nat
  type 'a one
  val one : 'a one -> 'a
  val one_inta : int
  val one_int : int one
  val minus_nat : nat -> nat -> nat
  val equal_nata : nat -> nat -> bool
  val zero_nat : nat
  val funpow : nat -> ('a -> 'a) -> 'a -> 'a
  val less_eq_nat : nat -> nat -> bool
  val less_nat : nat -> nat -> bool
  val ord_nat : nat Orderings.ord
  type 'a zero
  val zero : 'a zero -> 'a
  val uminus_int : int -> int
  val zero_inta : int
  val less_int : int -> int -> bool
  val abs_int : int -> int
  val zero_int : int zero
  type 'a times
  val times : 'a times -> 'a -> 'a -> 'a
  type 'a power
  val one_power : 'a power -> 'a one
  val times_power : 'a power -> 'a times
  val powera : 'a -> ('a -> 'a -> 'a) -> 'a -> nat -> 'a
  val power : 'a power -> 'a -> nat -> 'a
  val plus_int : int -> int -> int
  val times_inta : int -> int -> int
  val times_int : int times
  val power_int : int power
  val equal_nat : nat HOL.equal
  val preorder_nat : nat Orderings.preorder
  val order_nat : nat Orderings.order
  val equal_int : int -> int -> bool
  val minus_int : int -> int -> int
  val times_nat : nat -> nat -> nat
  val sgn_integer : IntInf.int -> IntInf.int
  val abs_integer : IntInf.int -> IntInf.int
  val divmod_integer : IntInf.int -> IntInf.int -> IntInf.int * IntInf.int
  val div_integer : IntInf.int -> IntInf.int -> IntInf.int
  val div_int : int -> int -> int
  val mod_integer : IntInf.int -> IntInf.int -> IntInf.int
  val mod_int : int -> int -> int
  val less_eq_int : int -> int -> bool
  val linorder_nat : nat Orderings.linorder
  datatype natural = Nata of IntInf.int
  type 'a zero_neq_one
  val one_zero_neq_one : 'a zero_neq_one -> 'a one
  val zero_zero_neq_one : 'a zero_neq_one -> 'a zero
  val zero_neq_one_int : int zero_neq_one
  val integer_of_natural : natural -> IntInf.int
  val div_natural : natural -> natural -> natural
  val mod_natural : natural -> natural -> natural
  val one_natural : natural
  val less_natural : natural -> natural -> bool
  val plus_natural : natural -> natural -> natural
  val zero_natural : natural
  val equal_natural : natural -> natural -> bool
  val minus_natural : natural -> natural -> natural
  val times_natural : natural -> natural -> natural
  val int_of_nat : nat -> int
  val nat_of_integer : IntInf.int -> nat
  val nat_of_natural : natural -> nat
  val natural_of_nat : nat -> natural
  val less_eq_natural : natural -> natural -> bool
  val natural_of_integer : IntInf.int -> natural
end = struct

datatype int = Int_of_integer of IntInf.int;

datatype nat = Nat of IntInf.int;

datatype num = One | Bit0 of num | Bit1 of num;

fun integer_of_int (Int_of_integer k) = k;

val ord_integer =
  {less_eq = (fn a => fn b => IntInf.<= (a, b)),
    less = (fn a => fn b => IntInf.< (a, b))}
  : IntInf.int Orderings.ord;

fun nat k = Nat (Orderings.max ord_integer 0 (integer_of_int k));

fun integer_of_nat (Nat x) = x;

fun plus_nat m n = Nat (IntInf.+ (integer_of_nat m, integer_of_nat n));

val one_nat : nat = Nat (1 : IntInf.int);

fun suc n = plus_nat n one_nat;

type 'a one = {one : 'a};
val one = #one : 'a one -> 'a;

val one_inta : int = Int_of_integer (1 : IntInf.int);

val one_int = {one = one_inta} : int one;

fun minus_nat m n =
  Nat (Orderings.max ord_integer 0
        (IntInf.- (integer_of_nat m, integer_of_nat n)));

fun equal_nata m n = (((integer_of_nat m) : IntInf.int) = (integer_of_nat n));

val zero_nat : nat = Nat 0;

fun funpow n f =
  (if equal_nata n zero_nat then Fun.id
    else f o funpow (minus_nat n one_nat) f);

fun less_eq_nat m n = IntInf.<= (integer_of_nat m, integer_of_nat n);

fun less_nat m n = IntInf.< (integer_of_nat m, integer_of_nat n);

val ord_nat = {less_eq = less_eq_nat, less = less_nat} : nat Orderings.ord;

type 'a zero = {zero : 'a};
val zero = #zero : 'a zero -> 'a;

fun uminus_int k = Int_of_integer (IntInf.~ (integer_of_int k));

val zero_inta : int = Int_of_integer 0;

fun less_int k l = IntInf.< (integer_of_int k, integer_of_int l);

fun abs_int i = (if less_int i zero_inta then uminus_int i else i);

val zero_int = {zero = zero_inta} : int zero;

type 'a times = {times : 'a -> 'a -> 'a};
val times = #times : 'a times -> 'a -> 'a -> 'a;

type 'a power = {one_power : 'a one, times_power : 'a times};
val one_power = #one_power : 'a power -> 'a one;
val times_power = #times_power : 'a power -> 'a times;

fun powera one times a n =
  (if equal_nata n zero_nat then one
    else times a (powera one times a (minus_nat n one_nat)));

fun power A_ = powera (one (one_power A_)) (times (times_power A_));

fun plus_int k l =
  Int_of_integer (IntInf.+ (integer_of_int k, integer_of_int l));

fun times_inta k l =
  Int_of_integer (IntInf.* (integer_of_int k, integer_of_int l));

val times_int = {times = times_inta} : int times;

val power_int = {one_power = one_int, times_power = times_int} : int power;

val equal_nat = {equal = equal_nata} : nat HOL.equal;

val preorder_nat = {ord_preorder = ord_nat} : nat Orderings.preorder;

val order_nat = {preorder_order = preorder_nat} : nat Orderings.order;

fun equal_int k l = (((integer_of_int k) : IntInf.int) = (integer_of_int l));

fun minus_int k l =
  Int_of_integer (IntInf.- (integer_of_int k, integer_of_int l));

fun times_nat m n = Nat (IntInf.* (integer_of_nat m, integer_of_nat n));

fun sgn_integer k =
  (if ((k : IntInf.int) = 0) then 0
    else (if IntInf.< (k, 0) then IntInf.~ (1 : IntInf.int)
           else (1 : IntInf.int)));

fun abs_integer k = (if IntInf.< (k, 0) then IntInf.~ k else k);

fun divmod_integer k l =
  (if ((k : IntInf.int) = 0) then (0, 0)
    else (if ((l : IntInf.int) = 0) then (0, k)
           else (Product_Type.apsnd o (fn a => fn b => IntInf.* (a, b)) o
                  sgn_integer)
                  l (if (((sgn_integer k) : IntInf.int) = (sgn_integer l))
                      then IntInf.divMod (IntInf.abs k, IntInf.abs l)
                      else let
                             val (r, s) =
                               IntInf.divMod (IntInf.abs k, IntInf.abs l);
                           in
                             (if ((s : IntInf.int) = 0) then (IntInf.~ r, 0)
                               else (IntInf.- (IntInf.~ r, (1 : IntInf.int)),
                                      IntInf.- (abs_integer l, s)))
                           end)));

fun div_integer k l = Product_Type.fst (divmod_integer k l);

fun div_int k l =
  Int_of_integer (div_integer (integer_of_int k) (integer_of_int l));

fun mod_integer k l = Product_Type.snd (divmod_integer k l);

fun mod_int k l =
  Int_of_integer (mod_integer (integer_of_int k) (integer_of_int l));

fun less_eq_int k l = IntInf.<= (integer_of_int k, integer_of_int l);

val linorder_nat = {order_linorder = order_nat} : nat Orderings.linorder;

datatype natural = Nata of IntInf.int;

type 'a zero_neq_one = {one_zero_neq_one : 'a one, zero_zero_neq_one : 'a zero};
val one_zero_neq_one = #one_zero_neq_one : 'a zero_neq_one -> 'a one;
val zero_zero_neq_one = #zero_zero_neq_one : 'a zero_neq_one -> 'a zero;

val zero_neq_one_int =
  {one_zero_neq_one = one_int, zero_zero_neq_one = zero_int} : int zero_neq_one;

fun integer_of_natural (Nata x) = x;

fun div_natural m n =
  Nata (div_integer (integer_of_natural m) (integer_of_natural n));

fun mod_natural m n =
  Nata (mod_integer (integer_of_natural m) (integer_of_natural n));

val one_natural : natural = Nata (1 : IntInf.int);

fun less_natural m n = IntInf.< (integer_of_natural m, integer_of_natural n);

fun plus_natural m n =
  Nata (IntInf.+ (integer_of_natural m, integer_of_natural n));

val zero_natural : natural = Nata 0;

fun equal_natural m n =
  (((integer_of_natural m) : IntInf.int) = (integer_of_natural n));

fun minus_natural m n =
  Nata (Orderings.max ord_integer 0
         (IntInf.- (integer_of_natural m, integer_of_natural n)));

fun times_natural m n =
  Nata (IntInf.* (integer_of_natural m, integer_of_natural n));

fun int_of_nat n = Int_of_integer (integer_of_nat n);

fun nat_of_integer k = Nat (Orderings.max ord_integer 0 k);

fun nat_of_natural x = (nat_of_integer o integer_of_natural) x;

fun natural_of_nat n = Nata (integer_of_nat n);

fun less_eq_natural m n =
  IntInf.<= (integer_of_natural m, integer_of_natural n);

fun natural_of_integer k = Nata (Orderings.max ord_integer 0 k);

end; (*struct Arith*)

structure List : sig
  val hd : 'a list -> 'a
  val tl : 'a list -> 'a list
  val map : ('a -> 'b) -> 'a list -> 'b list
  val nth : 'a list -> Arith.nat -> 'a
  val fold : ('a -> 'b -> 'b) -> 'a list -> 'b -> 'b
  val rev : 'a list -> 'a list
  val upt : Arith.nat -> Arith.nat -> Arith.nat list
  val zip : 'a list -> 'b list -> ('a * 'b) list
  val drop : Arith.nat -> 'a list -> 'a list
  val maps : ('a -> 'b list) -> 'a list -> 'b list
  val null : 'a list -> bool
  val take : Arith.nat -> 'a list -> 'a list
  val foldl : ('a -> 'b -> 'a) -> 'a -> 'b list -> 'a
  val foldr : ('a -> 'b -> 'b) -> 'a list -> 'b -> 'b
  val filter : ('a -> bool) -> 'a list -> 'a list
  val member : 'a HOL.equal -> 'a list -> 'a -> bool
  val insert : 'a HOL.equal -> 'a -> 'a list -> 'a list
  val remdups : 'a HOL.equal -> 'a list -> 'a list
  val remove1 : 'a HOL.equal -> 'a -> 'a list -> 'a list
  val distinct : 'a HOL.equal -> 'a list -> bool
  val list_all : ('a -> bool) -> 'a list -> bool
  val equal_lista : 'a HOL.equal -> 'a list -> 'a list -> bool
  val equal_list : 'a HOL.equal -> ('a list) HOL.equal
  val list_all2 : ('a -> 'b -> bool) -> 'a list -> 'b list -> bool
  val removeAll : 'a HOL.equal -> 'a -> 'a list -> 'a list
  val replicate : Arith.nat -> 'a -> 'a list
  val gen_length : Arith.nat -> 'a list -> Arith.nat
  val size_list : 'a list -> Arith.nat
  val insort_key :
    'b Orderings.linorder -> ('a -> 'b) -> 'a -> 'a list -> 'a list
  val list_update : 'a list -> Arith.nat -> 'a -> 'a list
  val all_interval_nat : (Arith.nat -> bool) -> Arith.nat -> Arith.nat -> bool
  val insort_insert_key :
    'b HOL.equal * 'b Orderings.linorder ->
      ('a -> 'b) -> 'a -> 'a list -> 'a list
end = struct

fun hd (x :: xs) = x;

fun tl [] = []
  | tl (x :: xs) = xs;

fun map f [] = []
  | map f (x :: xs) = f x :: map f xs;

fun nth (x :: xs) n =
  (if Arith.equal_nata n Arith.zero_nat then x
    else nth xs (Arith.minus_nat n Arith.one_nat));

fun fold f (x :: xs) s = fold f xs (f x s)
  | fold f [] s = s;

fun rev xs = fold (fn a => fn b => a :: b) xs [];

fun upt i j = (if Arith.less_nat i j then i :: upt (Arith.suc i) j else []);

fun zip (x :: xs) (y :: ys) = (x, y) :: zip xs ys
  | zip xs [] = []
  | zip [] ys = [];

fun drop n [] = []
  | drop n (x :: xs) =
    (if Arith.equal_nata n Arith.zero_nat then x :: xs
      else drop (Arith.minus_nat n Arith.one_nat) xs);

fun maps f [] = []
  | maps f (x :: xs) = f x @ maps f xs;

fun null [] = true
  | null (x :: xs) = false;

fun take n [] = []
  | take n (x :: xs) =
    (if Arith.equal_nata n Arith.zero_nat then []
      else x :: take (Arith.minus_nat n Arith.one_nat) xs);

fun foldl f a [] = a
  | foldl f a (x :: xs) = foldl f (f a x) xs;

fun foldr f [] = Fun.id
  | foldr f (x :: xs) = f x o foldr f xs;

fun filter p [] = []
  | filter p (x :: xs) = (if p x then x :: filter p xs else filter p xs);

fun member A_ [] y = false
  | member A_ (x :: xs) y = HOL.eq A_ x y orelse member A_ xs y;

fun insert A_ x xs = (if member A_ xs x then xs else x :: xs);

fun remdups A_ [] = []
  | remdups A_ (x :: xs) =
    (if member A_ xs x then remdups A_ xs else x :: remdups A_ xs);

fun remove1 A_ x [] = []
  | remove1 A_ x (y :: xs) =
    (if HOL.eq A_ x y then xs else y :: remove1 A_ x xs);

fun distinct A_ [] = true
  | distinct A_ (x :: xs) = not (member A_ xs x) andalso distinct A_ xs;

fun list_all p [] = true
  | list_all p (x :: xs) = p x andalso list_all p xs;

fun equal_lista A_ (a :: lista) [] = false
  | equal_lista A_ [] (a :: lista) = false
  | equal_lista A_ (aa :: listaa) (a :: lista) =
    HOL.eq A_ aa a andalso equal_lista A_ listaa lista
  | equal_lista A_ [] [] = true;

fun equal_list A_ = {equal = equal_lista A_} : ('a list) HOL.equal;

fun list_all2 p (x :: xs) (y :: ys) = p x y andalso list_all2 p xs ys
  | list_all2 p [] ys = null ys
  | list_all2 p xs [] = null xs;

fun removeAll A_ x [] = []
  | removeAll A_ x (y :: xs) =
    (if HOL.eq A_ x y then removeAll A_ x xs else y :: removeAll A_ x xs);

fun replicate n x =
  (if Arith.equal_nata n Arith.zero_nat then []
    else x :: replicate (Arith.minus_nat n Arith.one_nat) x);

fun gen_length n (x :: xs) = gen_length (Arith.suc n) xs
  | gen_length n [] = n;

fun size_list x = gen_length Arith.zero_nat x;

fun insort_key B_ f x [] = [x]
  | insort_key B_ f x (y :: ys) =
    (if Orderings.less_eq
          ((Orderings.ord_preorder o Orderings.preorder_order o
             Orderings.order_linorder)
            B_)
          (f x) (f y)
      then x :: y :: ys else y :: insort_key B_ f x ys);

fun list_update [] i y = []
  | list_update (x :: xs) i y =
    (if Arith.equal_nata i Arith.zero_nat then y :: xs
      else x :: list_update xs (Arith.minus_nat i Arith.one_nat) y);

fun all_interval_nat p i j =
  Arith.less_eq_nat j i orelse p i andalso all_interval_nat p (Arith.suc i) j;

fun insort_insert_key (B1_, B2_) f x xs =
  (if member B1_ (map f xs) (f x) then xs else insort_key B2_ f x xs);

end; (*struct List*)

structure Phantom_Type : sig
  datatype ('a, 'b) phantom = Phantom of 'b
  val of_phantom : ('a, 'b) phantom -> 'b
end = struct

datatype ('a, 'b) phantom = Phantom of 'b;

fun of_phantom (Phantom x) = x;

end; (*struct Phantom_Type*)

structure Lattices : sig
  type 'a inf
  val inf : 'a inf -> 'a -> 'a -> 'a
  type 'a sup
  val sup : 'a sup -> 'a -> 'a -> 'a
  type 'a semilattice_sup
  val sup_semilattice_sup : 'a semilattice_sup -> 'a sup
  val order_semilattice_sup : 'a semilattice_sup -> 'a Orderings.order
  type 'a semilattice_inf
  val inf_semilattice_inf : 'a semilattice_inf -> 'a inf
  val order_semilattice_inf : 'a semilattice_inf -> 'a Orderings.order
  type 'a lattice
  val semilattice_inf_lattice : 'a lattice -> 'a semilattice_inf
  val semilattice_sup_lattice : 'a lattice -> 'a semilattice_sup
  type 'a bounded_semilattice_inf_top
  val semilattice_inf_bounded_semilattice_inf_top :
    'a bounded_semilattice_inf_top -> 'a semilattice_inf
  val order_top_bounded_semilattice_inf_top :
    'a bounded_semilattice_inf_top -> 'a Orderings.order_top
  type 'a bounded_lattice_top
  val bounded_semilattice_inf_top_bounded_lattice_top :
    'a bounded_lattice_top -> 'a bounded_semilattice_inf_top
  val lattice_bounded_lattice_top : 'a bounded_lattice_top -> 'a lattice
  type 'a bounded_semilattice_sup_bot
  val semilattice_sup_bounded_semilattice_sup_bot :
    'a bounded_semilattice_sup_bot -> 'a semilattice_sup
  val order_bot_bounded_semilattice_sup_bot :
    'a bounded_semilattice_sup_bot -> 'a Orderings.order_bot
  type 'a bounded_lattice_bot
  val bounded_semilattice_sup_bot_bounded_lattice_bot :
    'a bounded_lattice_bot -> 'a bounded_semilattice_sup_bot
  val lattice_bounded_lattice_bot : 'a bounded_lattice_bot -> 'a lattice
  type 'a bounded_lattice
  val bounded_lattice_bot_bounded_lattice :
    'a bounded_lattice -> 'a bounded_lattice_bot
  val bounded_lattice_top_bounded_lattice :
    'a bounded_lattice -> 'a bounded_lattice_top
end = struct

type 'a inf = {inf : 'a -> 'a -> 'a};
val inf = #inf : 'a inf -> 'a -> 'a -> 'a;

type 'a sup = {sup : 'a -> 'a -> 'a};
val sup = #sup : 'a sup -> 'a -> 'a -> 'a;

type 'a semilattice_sup =
  {sup_semilattice_sup : 'a sup, order_semilattice_sup : 'a Orderings.order};
val sup_semilattice_sup = #sup_semilattice_sup : 'a semilattice_sup -> 'a sup;
val order_semilattice_sup = #order_semilattice_sup :
  'a semilattice_sup -> 'a Orderings.order;

type 'a semilattice_inf =
  {inf_semilattice_inf : 'a inf, order_semilattice_inf : 'a Orderings.order};
val inf_semilattice_inf = #inf_semilattice_inf : 'a semilattice_inf -> 'a inf;
val order_semilattice_inf = #order_semilattice_inf :
  'a semilattice_inf -> 'a Orderings.order;

type 'a lattice =
  {semilattice_inf_lattice : 'a semilattice_inf,
    semilattice_sup_lattice : 'a semilattice_sup};
val semilattice_inf_lattice = #semilattice_inf_lattice :
  'a lattice -> 'a semilattice_inf;
val semilattice_sup_lattice = #semilattice_sup_lattice :
  'a lattice -> 'a semilattice_sup;

type 'a bounded_semilattice_inf_top =
  {semilattice_inf_bounded_semilattice_inf_top : 'a semilattice_inf,
    order_top_bounded_semilattice_inf_top : 'a Orderings.order_top};
val semilattice_inf_bounded_semilattice_inf_top =
  #semilattice_inf_bounded_semilattice_inf_top :
  'a bounded_semilattice_inf_top -> 'a semilattice_inf;
val order_top_bounded_semilattice_inf_top =
  #order_top_bounded_semilattice_inf_top :
  'a bounded_semilattice_inf_top -> 'a Orderings.order_top;

type 'a bounded_lattice_top =
  {bounded_semilattice_inf_top_bounded_lattice_top :
     'a bounded_semilattice_inf_top,
    lattice_bounded_lattice_top : 'a lattice};
val bounded_semilattice_inf_top_bounded_lattice_top =
  #bounded_semilattice_inf_top_bounded_lattice_top :
  'a bounded_lattice_top -> 'a bounded_semilattice_inf_top;
val lattice_bounded_lattice_top = #lattice_bounded_lattice_top :
  'a bounded_lattice_top -> 'a lattice;

type 'a bounded_semilattice_sup_bot =
  {semilattice_sup_bounded_semilattice_sup_bot : 'a semilattice_sup,
    order_bot_bounded_semilattice_sup_bot : 'a Orderings.order_bot};
val semilattice_sup_bounded_semilattice_sup_bot =
  #semilattice_sup_bounded_semilattice_sup_bot :
  'a bounded_semilattice_sup_bot -> 'a semilattice_sup;
val order_bot_bounded_semilattice_sup_bot =
  #order_bot_bounded_semilattice_sup_bot :
  'a bounded_semilattice_sup_bot -> 'a Orderings.order_bot;

type 'a bounded_lattice_bot =
  {bounded_semilattice_sup_bot_bounded_lattice_bot :
     'a bounded_semilattice_sup_bot,
    lattice_bounded_lattice_bot : 'a lattice};
val bounded_semilattice_sup_bot_bounded_lattice_bot =
  #bounded_semilattice_sup_bot_bounded_lattice_bot :
  'a bounded_lattice_bot -> 'a bounded_semilattice_sup_bot;
val lattice_bounded_lattice_bot = #lattice_bounded_lattice_bot :
  'a bounded_lattice_bot -> 'a lattice;

type 'a bounded_lattice =
  {bounded_lattice_bot_bounded_lattice : 'a bounded_lattice_bot,
    bounded_lattice_top_bounded_lattice : 'a bounded_lattice_top};
val bounded_lattice_bot_bounded_lattice = #bounded_lattice_bot_bounded_lattice :
  'a bounded_lattice -> 'a bounded_lattice_bot;
val bounded_lattice_top_bounded_lattice = #bounded_lattice_top_bounded_lattice :
  'a bounded_lattice -> 'a bounded_lattice_top;

end; (*struct Lattices*)

structure Set : sig
  datatype 'a set = Set of 'a list | Coset of 'a list
  type 'a sup
  val sup : 'a sup -> 'a set -> 'a
  type 'a inf
  val inf : 'a inf -> 'a set -> 'a
  type 'a conditionally_complete_lattice
  val inf_conditionally_complete_lattice :
    'a conditionally_complete_lattice -> 'a inf
  val sup_conditionally_complete_lattice :
    'a conditionally_complete_lattice -> 'a sup
  val lattice_conditionally_complete_lattice :
    'a conditionally_complete_lattice -> 'a Lattices.lattice
  type 'a ccpo
  val sup_ccpo : 'a ccpo -> 'a sup
  val order_ccpo : 'a ccpo -> 'a Orderings.order
  type 'a complete_lattice
  val ccpo_complete_lattice : 'a complete_lattice -> 'a ccpo
  val conditionally_complete_lattice_complete_lattice :
    'a complete_lattice -> 'a conditionally_complete_lattice
  val bounded_lattice_complete_lattice :
    'a complete_lattice -> 'a Lattices.bounded_lattice
  val insert : 'a -> 'a set -> 'a set
  val member : 'a HOL.equal -> 'a -> 'a set -> bool
  val sup_seta : 'a HOL.equal -> 'a set -> 'a set -> 'a set
  val bot_seta : 'a set
  val sup_setc : 'a HOL.equal -> 'a set set -> 'a set
  val top_seta : 'a set
  val remove : 'a HOL.equal -> 'a -> 'a set -> 'a set
  val inf_seta : 'a HOL.equal -> 'a set -> 'a set -> 'a set
  val inf_setc : 'a HOL.equal -> 'a set set -> 'a set
  val sup_setb : 'a HOL.equal -> 'a set sup
  val inf_setb : 'a HOL.equal -> 'a set inf
  val less_eq_set : 'a HOL.equal -> 'a set -> 'a set -> bool
  val less_set : 'a HOL.equal -> 'a set -> 'a set -> bool
  val ord_set : 'a HOL.equal -> 'a set Orderings.ord
  val preorder_set : 'a HOL.equal -> 'a set Orderings.preorder
  val order_set : 'a HOL.equal -> 'a set Orderings.order
  val sup_set : 'a HOL.equal -> 'a set Lattices.sup
  val semilattice_sup_set : 'a HOL.equal -> 'a set Lattices.semilattice_sup
  val inf_set : 'a HOL.equal -> 'a set Lattices.inf
  val semilattice_inf_set : 'a HOL.equal -> 'a set Lattices.semilattice_inf
  val lattice_set : 'a HOL.equal -> 'a set Lattices.lattice
  val conditionally_complete_lattice_set :
    'a HOL.equal -> 'a set conditionally_complete_lattice
  val ccpo_set : 'a HOL.equal -> 'a set ccpo
  val top_set : 'a set Orderings.top
  val order_top_set : 'a HOL.equal -> 'a set Orderings.order_top
  val bounded_semilattice_inf_top_set :
    'a HOL.equal -> 'a set Lattices.bounded_semilattice_inf_top
  val bounded_lattice_top_set :
    'a HOL.equal -> 'a set Lattices.bounded_lattice_top
  val bot_set : 'a set Orderings.bot
  val order_bot_set : 'a HOL.equal -> 'a set Orderings.order_bot
  val bounded_semilattice_sup_bot_set :
    'a HOL.equal -> 'a set Lattices.bounded_semilattice_sup_bot
  val bounded_lattice_bot_set :
    'a HOL.equal -> 'a set Lattices.bounded_lattice_bot
  val bounded_lattice_set : 'a HOL.equal -> 'a set Lattices.bounded_lattice
  val complete_lattice_set : 'a HOL.equal -> 'a set complete_lattice
  val supr : 'b complete_lattice -> 'a set -> ('a -> 'b) -> 'b
  val bind : 'b HOL.equal -> 'a set -> ('a -> 'b set) -> 'b set
  val equal_set : 'a HOL.equal -> 'a set -> 'a set -> bool
  type 'a finite_UNIV
  val finite_UNIV : 'a finite_UNIV -> ('a, bool) Phantom_Type.phantom
  type 'a card_UNIV
  val finite_UNIV_card_UNIV : 'a card_UNIV -> 'a finite_UNIV
  val card_UNIV : 'a card_UNIV -> ('a, Arith.nat) Phantom_Type.phantom
  val card : 'a card_UNIV * 'a HOL.equal -> 'a set -> Arith.nat
  val card_UNIVa : 'a card_UNIV -> ('a, Arith.nat) Phantom_Type.phantom
  val union : 'a set -> 'a set -> 'a set
  val finite_UNIV_nata : (Arith.nat, bool) Phantom_Type.phantom
  val finite_UNIV_nat : Arith.nat finite_UNIV
  val card_UNIV_nata : (Arith.nat, Arith.nat) Phantom_Type.phantom
  val card_UNIV_nat : Arith.nat card_UNIV
  val is_list_UNIV : 'a card_UNIV * 'a HOL.equal -> 'a list -> bool
end = struct

datatype 'a set = Set of 'a list | Coset of 'a list;

type 'a sup = {sup : 'a set -> 'a};
val sup = #sup : 'a sup -> 'a set -> 'a;

type 'a inf = {inf : 'a set -> 'a};
val inf = #inf : 'a inf -> 'a set -> 'a;

type 'a conditionally_complete_lattice =
  {inf_conditionally_complete_lattice : 'a inf,
    sup_conditionally_complete_lattice : 'a sup,
    lattice_conditionally_complete_lattice : 'a Lattices.lattice};
val inf_conditionally_complete_lattice = #inf_conditionally_complete_lattice :
  'a conditionally_complete_lattice -> 'a inf;
val sup_conditionally_complete_lattice = #sup_conditionally_complete_lattice :
  'a conditionally_complete_lattice -> 'a sup;
val lattice_conditionally_complete_lattice =
  #lattice_conditionally_complete_lattice :
  'a conditionally_complete_lattice -> 'a Lattices.lattice;

type 'a ccpo = {sup_ccpo : 'a sup, order_ccpo : 'a Orderings.order};
val sup_ccpo = #sup_ccpo : 'a ccpo -> 'a sup;
val order_ccpo = #order_ccpo : 'a ccpo -> 'a Orderings.order;

type 'a complete_lattice =
  {ccpo_complete_lattice : 'a ccpo,
    conditionally_complete_lattice_complete_lattice :
      'a conditionally_complete_lattice,
    bounded_lattice_complete_lattice : 'a Lattices.bounded_lattice};
val ccpo_complete_lattice = #ccpo_complete_lattice :
  'a complete_lattice -> 'a ccpo;
val conditionally_complete_lattice_complete_lattice =
  #conditionally_complete_lattice_complete_lattice :
  'a complete_lattice -> 'a conditionally_complete_lattice;
val bounded_lattice_complete_lattice = #bounded_lattice_complete_lattice :
  'a complete_lattice -> 'a Lattices.bounded_lattice;

fun insert x (Set xs) = Set (x :: xs);

fun member A_ x (Coset xs) = not (List.member A_ xs x)
  | member A_ x (Set xs) = List.member A_ xs x;

fun sup_seta A_ (Coset xs) a =
  Coset (List.filter (fn x => not (member A_ x a)) xs)
  | sup_seta A_ (Set xs) a = List.fold insert xs a;

val bot_seta : 'a set = Set [];

fun sup_setc A_ (Set xs) = List.fold (sup_seta A_) xs bot_seta;

val top_seta : 'a set = Coset [];

fun remove A_ x (Coset xs) = Coset (List.insert A_ x xs)
  | remove A_ x (Set xs) = Set (List.removeAll A_ x xs);

fun inf_seta A_ a (Coset xs) = List.fold (remove A_) xs a
  | inf_seta A_ a (Set xs) = Set (List.filter (fn x => member A_ x a) xs);

fun inf_setc A_ (Set xs) = List.fold (inf_seta A_) xs top_seta;

fun sup_setb A_ = {sup = sup_setc A_} : 'a set sup;

fun inf_setb A_ = {inf = inf_setc A_} : 'a set inf;

fun less_eq_set A_ (Coset xs) (Set ys) =
  (if List.null xs andalso List.null ys then false
    else (raise Fail
           "subset_eq (List.coset _) (List.set _) requires type class instance card_UNIV")
           (fn _ => less_eq_set A_ (Coset xs) (Set ys)))
  | less_eq_set A_ a (Coset ys) = List.list_all (fn y => not (member A_ y a)) ys
  | less_eq_set A_ (Set xs) b = List.list_all (fn x => member A_ x b) xs;

fun less_set A_ a b = less_eq_set A_ a b andalso not (less_eq_set A_ b a);

fun ord_set A_ = {less_eq = less_eq_set A_, less = less_set A_} :
  'a set Orderings.ord;

fun preorder_set A_ = {ord_preorder = ord_set A_} : 'a set Orderings.preorder;

fun order_set A_ = {preorder_order = preorder_set A_} : 'a set Orderings.order;

fun sup_set A_ = {sup = sup_seta A_} : 'a set Lattices.sup;

fun semilattice_sup_set A_ =
  {sup_semilattice_sup = sup_set A_, order_semilattice_sup = order_set A_} :
  'a set Lattices.semilattice_sup;

fun inf_set A_ = {inf = inf_seta A_} : 'a set Lattices.inf;

fun semilattice_inf_set A_ =
  {inf_semilattice_inf = inf_set A_, order_semilattice_inf = order_set A_} :
  'a set Lattices.semilattice_inf;

fun lattice_set A_ =
  {semilattice_inf_lattice = semilattice_inf_set A_,
    semilattice_sup_lattice = semilattice_sup_set A_}
  : 'a set Lattices.lattice;

fun conditionally_complete_lattice_set A_ =
  {inf_conditionally_complete_lattice = inf_setb A_,
    sup_conditionally_complete_lattice = sup_setb A_,
    lattice_conditionally_complete_lattice = lattice_set A_}
  : 'a set conditionally_complete_lattice;

fun ccpo_set A_ = {sup_ccpo = sup_setb A_, order_ccpo = order_set A_} :
  'a set ccpo;

val top_set = {top = top_seta} : 'a set Orderings.top;

fun order_top_set A_ = {order_order_top = order_set A_, top_order_top = top_set}
  : 'a set Orderings.order_top;

fun bounded_semilattice_inf_top_set A_ =
  {semilattice_inf_bounded_semilattice_inf_top = semilattice_inf_set A_,
    order_top_bounded_semilattice_inf_top = order_top_set A_}
  : 'a set Lattices.bounded_semilattice_inf_top;

fun bounded_lattice_top_set A_ =
  {bounded_semilattice_inf_top_bounded_lattice_top =
     bounded_semilattice_inf_top_set A_,
    lattice_bounded_lattice_top = lattice_set A_}
  : 'a set Lattices.bounded_lattice_top;

val bot_set = {bot = bot_seta} : 'a set Orderings.bot;

fun order_bot_set A_ = {bot_order_bot = bot_set, order_order_bot = order_set A_}
  : 'a set Orderings.order_bot;

fun bounded_semilattice_sup_bot_set A_ =
  {semilattice_sup_bounded_semilattice_sup_bot = semilattice_sup_set A_,
    order_bot_bounded_semilattice_sup_bot = order_bot_set A_}
  : 'a set Lattices.bounded_semilattice_sup_bot;

fun bounded_lattice_bot_set A_ =
  {bounded_semilattice_sup_bot_bounded_lattice_bot =
     bounded_semilattice_sup_bot_set A_,
    lattice_bounded_lattice_bot = lattice_set A_}
  : 'a set Lattices.bounded_lattice_bot;

fun bounded_lattice_set A_ =
  {bounded_lattice_bot_bounded_lattice = bounded_lattice_bot_set A_,
    bounded_lattice_top_bounded_lattice = bounded_lattice_top_set A_}
  : 'a set Lattices.bounded_lattice;

fun complete_lattice_set A_ =
  {ccpo_complete_lattice = ccpo_set A_,
    conditionally_complete_lattice_complete_lattice =
      conditionally_complete_lattice_set A_,
    bounded_lattice_complete_lattice = bounded_lattice_set A_}
  : 'a set complete_lattice;

fun supr B_ (Set xs) f =
  List.fold
    (Lattices.sup
       ((Lattices.sup_semilattice_sup o Lattices.semilattice_sup_lattice o
          Lattices.lattice_bounded_lattice_bot o
          Lattices.bounded_lattice_bot_bounded_lattice o
          bounded_lattice_complete_lattice)
         B_) o
      f)
    xs (Orderings.bot
         ((Orderings.bot_order_bot o
            Lattices.order_bot_bounded_semilattice_sup_bot o
            Lattices.bounded_semilattice_sup_bot_bounded_lattice_bot o
            Lattices.bounded_lattice_bot_bounded_lattice o
            bounded_lattice_complete_lattice)
           B_));

fun bind B_ a f = supr (complete_lattice_set B_) a f;

fun equal_set A_ a b = less_eq_set A_ a b andalso less_eq_set A_ b a;

type 'a finite_UNIV = {finite_UNIV : ('a, bool) Phantom_Type.phantom};
val finite_UNIV = #finite_UNIV :
  'a finite_UNIV -> ('a, bool) Phantom_Type.phantom;

type 'a card_UNIV =
  {finite_UNIV_card_UNIV : 'a finite_UNIV,
    card_UNIV : ('a, Arith.nat) Phantom_Type.phantom};
val finite_UNIV_card_UNIV = #finite_UNIV_card_UNIV :
  'a card_UNIV -> 'a finite_UNIV;
val card_UNIV = #card_UNIV :
  'a card_UNIV -> ('a, Arith.nat) Phantom_Type.phantom;

fun card (A1_, A2_) (Coset xs) =
  Arith.minus_nat (Phantom_Type.of_phantom (card_UNIV A1_))
    (List.size_list (List.remdups A2_ xs))
  | card (A1_, A2_) (Set xs) = List.size_list (List.remdups A2_ xs);

fun card_UNIVa A_ = card_UNIV A_;

fun union (Set xs) (Set ys) = Set (xs @ ys);

val finite_UNIV_nata : (Arith.nat, bool) Phantom_Type.phantom =
  Phantom_Type.Phantom false;

val finite_UNIV_nat = {finite_UNIV = finite_UNIV_nata} : Arith.nat finite_UNIV;

val card_UNIV_nata : (Arith.nat, Arith.nat) Phantom_Type.phantom =
  Phantom_Type.Phantom Arith.zero_nat;

val card_UNIV_nat =
  {finite_UNIV_card_UNIV = finite_UNIV_nat, card_UNIV = card_UNIV_nata} :
  Arith.nat card_UNIV;

fun is_list_UNIV (A1_, A2_) xs =
  let
    val c = Phantom_Type.of_phantom (card_UNIVa A1_);
  in
    (if Arith.equal_nata c Arith.zero_nat then false
      else Arith.equal_nata (List.size_list (List.remdups A2_ xs)) c)
  end;

end; (*struct Set*)

structure FinFun : sig
  datatype ('a, 'b) finfun = Finfun_const of 'b |
    Finfun_update_code of ('a, 'b) finfun * 'a * 'b
  val finfun_All_except :
    'a Set.card_UNIV * 'a HOL.equal -> 'a list -> ('a, bool) finfun -> bool
  val finfun_All : 'a Set.card_UNIV * 'a HOL.equal -> ('a, bool) finfun -> bool
  val finfun_comp : ('a -> 'b) -> ('c, 'a) finfun -> ('c, 'b) finfun
  val finfun_update :
    'a HOL.equal -> 'b HOL.equal ->
      ('a, 'b) finfun -> 'a -> 'b -> ('a, 'b) finfun
  val finfun_apply : 'a HOL.equal -> ('a, 'b) finfun -> 'a -> 'b
  val finfun_Diag :
    'a HOL.equal -> 'b HOL.equal -> 'c HOL.equal ->
      ('a, 'b) finfun -> ('a, 'c) finfun -> ('a, ('b * 'c)) finfun
  val equal_finfuna :
    'a Set.card_UNIV * 'a HOL.equal -> 'b HOL.equal ->
      ('a, 'b) finfun -> ('a, 'b) finfun -> bool
  val equal_finfun :
    'a Set.card_UNIV * 'a HOL.equal -> 'b HOL.equal -> ('a, 'b) finfun HOL.equal
  val finfun_default : 'a Set.card_UNIV * 'a HOL.equal -> ('a, 'b) finfun -> 'b
  val finfun_to_list :
    'a Set.card_UNIV * 'a HOL.equal * 'a Orderings.linorder -> 'b HOL.equal ->
      ('a, 'b) finfun -> 'a list
end = struct

datatype ('a, 'b) finfun = Finfun_const of 'b |
  Finfun_update_code of ('a, 'b) finfun * 'a * 'b;

fun finfun_All_except (A1_, A2_) aa (Finfun_update_code (f, a, b)) =
  (List.member A2_ aa a orelse b) andalso
    finfun_All_except (A1_, A2_) (a :: aa) f
  | finfun_All_except (A1_, A2_) a (Finfun_const b) =
    b orelse Set.is_list_UNIV (A1_, A2_) a;

fun finfun_All (A1_, A2_) = finfun_All_except (A1_, A2_) [];

fun finfun_comp g (Finfun_update_code (f, a, b)) =
  Finfun_update_code (finfun_comp g f, a, g b)
  | finfun_comp g (Finfun_const c) = Finfun_const (g c);

fun finfun_update A_ B_ (Finfun_update_code (f, aa, ba)) a b =
  (if HOL.eq A_ aa a then finfun_update A_ B_ f aa b
    else Finfun_update_code (finfun_update A_ B_ f a b, aa, ba))
  | finfun_update A_ B_ (Finfun_const ba) a b =
    (if HOL.eq B_ ba b then Finfun_const ba
      else Finfun_update_code (Finfun_const ba, a, b));

fun finfun_apply A_ (Finfun_const b) a = b
  | finfun_apply A_ (Finfun_update_code (f, aa, b)) a =
    (if HOL.eq A_ aa a then b else finfun_apply A_ f a);

fun finfun_Diag A_ B_ C_ (Finfun_update_code (f, a, b)) g =
  finfun_update A_ (Product_Type.equal_prod B_ C_) (finfun_Diag A_ B_ C_ f g) a
    (b, finfun_apply A_ g a)
  | finfun_Diag A_ B_ C_ (Finfun_const b) (Finfun_update_code (g, a, c)) =
    Finfun_update_code (finfun_Diag A_ B_ C_ (Finfun_const b) g, a, (b, c))
  | finfun_Diag A_ B_ C_ (Finfun_const b) (Finfun_const c) =
    Finfun_const (b, c);

fun equal_finfuna (A1_, A2_) B_ f g =
  finfun_All (A1_, A2_)
    (finfun_comp (fn (a, b) => HOL.eq B_ a b) (finfun_Diag A2_ B_ B_ f g));

fun equal_finfun (A1_, A2_) B_ = {equal = equal_finfuna (A1_, A2_) B_} :
  ('a, 'b) finfun HOL.equal;

fun finfun_default (A1_, A2_) (Finfun_update_code (f, a, b)) =
  finfun_default (A1_, A2_) f
  | finfun_default (A1_, A2_) (Finfun_const c) =
    (if Arith.equal_nata (Set.card (A1_, A2_) Set.top_seta) Arith.zero_nat
      then c else (raise Fail "undefined"));

fun finfun_to_list (A1_, A2_, A3_) B_ (Finfun_update_code (f, a, b)) =
  (if HOL.eq B_ b (finfun_default (A1_, A2_) f)
    then List.remove1 A2_ a (finfun_to_list (A1_, A2_, A3_) B_ f)
    else List.insort_insert_key (A2_, A3_) (fn x => x) a
           (finfun_to_list (A1_, A2_, A3_) B_ f))
  | finfun_to_list (A1_, A2_, A3_) B_ (Finfun_const c) =
    (if Arith.equal_nata (Set.card (A1_, A2_) Set.top_seta) Arith.zero_nat
      then []
      else (raise Fail "finfun_to_list called on finite type")
             (fn _ => finfun_to_list (A1_, A2_, A3_) B_ (Finfun_const c)));

end; (*struct FinFun*)

structure Countable : sig
  type 'a countable
end = struct

type 'a countable = {};

end; (*struct Countable*)

structure Finite_Set : sig
  type 'a finite
  val countable_finite : 'a finite -> 'a Countable.countable
end = struct

type 'a finite = {countable_finite : 'a Countable.countable};
val countable_finite = #countable_finite : 'a finite -> 'a Countable.countable;

end; (*struct Finite_Set*)

structure Numeral_Type : sig
  datatype 'a bit0 = Abs_bit0 of Arith.int
  datatype num1 = One_num1
end = struct

datatype 'a bit0 = Abs_bit0 of Arith.int;

datatype num1 = One_num1;

end; (*struct Numeral_Type*)

structure Type_Length : sig
  type 'a len0
  val len_of : 'a len0 -> 'a HOL.itself -> Arith.nat
  type 'a len
  val len0_len : 'a len -> 'a len0
  val len_of_bit0 : 'a len0 -> 'a Numeral_Type.bit0 HOL.itself -> Arith.nat
  val len0_bit0 : 'a len0 -> 'a Numeral_Type.bit0 len0
  val len_bit0 : 'a len -> 'a Numeral_Type.bit0 len
  val len_of_num1 : Numeral_Type.num1 HOL.itself -> Arith.nat
  val len0_num1 : Numeral_Type.num1 len0
  val len_num1 : Numeral_Type.num1 len
end = struct

type 'a len0 = {len_of : 'a HOL.itself -> Arith.nat};
val len_of = #len_of : 'a len0 -> 'a HOL.itself -> Arith.nat;

type 'a len = {len0_len : 'a len0};
val len0_len = #len0_len : 'a len -> 'a len0;

fun len_of_bit0 A_ x =
  Arith.times_nat (Arith.nat_of_integer (2 : IntInf.int)) (len_of A_ HOL.Type);

fun len0_bit0 A_ = {len_of = len_of_bit0 A_} : 'a Numeral_Type.bit0 len0;

fun len_bit0 A_ = {len0_len = len0_bit0 (len0_len A_)} :
  'a Numeral_Type.bit0 len;

fun len_of_num1 x = Arith.one_nat;

val len0_num1 = {len_of = len_of_num1} : Numeral_Type.num1 len0;

val len_num1 = {len0_len = len0_num1} : Numeral_Type.num1 len;

end; (*struct Type_Length*)

structure Bit : sig
  datatype bit = Zero_bit | One_bit
end = struct

datatype bit = Zero_bit | One_bit;

end; (*struct Bit*)

structure Bit_Representation : sig
  val bitval : 'a Arith.zero_neq_one -> Bit.bit -> 'a
  val bit : Arith.int -> Bit.bit -> Arith.int
  val bin_last : Arith.int -> Bit.bit
  val bin_rest : Arith.int -> Arith.int
  val sbintrunc : Arith.nat -> Arith.int -> Arith.int
end = struct

fun bitval A_ =
  (fn a =>
    (case a of Bit.Zero_bit => Arith.zero (Arith.zero_zero_neq_one A_)
      | Bit.One_bit => Arith.one (Arith.one_zero_neq_one A_)));

fun bit k b =
  Arith.plus_int (Arith.plus_int (bitval Arith.zero_neq_one_int b) k) k;

fun bin_last w =
  (if Arith.equal_int (Arith.mod_int w (Arith.Int_of_integer (2 : IntInf.int)))
        Arith.zero_inta
    then Bit.Zero_bit else Bit.One_bit);

fun bin_rest w = Arith.div_int w (Arith.Int_of_integer (2 : IntInf.int));

fun sbintrunc n bin =
  (if Arith.equal_nata n Arith.zero_nat
    then (case bin_last bin of Bit.Zero_bit => Arith.zero_inta
           | Bit.One_bit => Arith.Int_of_integer (~1 : IntInf.int))
    else bit (sbintrunc (Arith.minus_nat n Arith.one_nat) (bin_rest bin))
           (bin_last bin));

end; (*struct Bit_Representation*)

structure Bit_Operations : sig
  val bitAND_bit : Bit.bit -> Bit.bit -> Bit.bit
end = struct

fun bitAND_bit Bit.Zero_bit y = Bit.Zero_bit
  | bitAND_bit Bit.One_bit y = y;

end; (*struct Bit_Operations*)

structure Bit_Int : sig
  val bitNOT_int : Arith.int -> Arith.int
  val bitAND_int : Arith.int -> Arith.int -> Arith.int
  val bitOR_int : Arith.int -> Arith.int -> Arith.int
  val bitXOR_int : Arith.int -> Arith.int -> Arith.int
end = struct

fun bitNOT_int x =
  (fn xa =>
    Arith.minus_int (Arith.uminus_int xa)
      (Arith.Int_of_integer (1 : IntInf.int)))
    x;

fun bitAND_int x y =
  (if Arith.equal_int x Arith.zero_inta then Arith.zero_inta
    else (if Arith.equal_int x (Arith.Int_of_integer (~1 : IntInf.int)) then y
           else Bit_Representation.bit
                  (bitAND_int (Bit_Representation.bin_rest x)
                    (Bit_Representation.bin_rest y))
                  (Bit_Operations.bitAND_bit (Bit_Representation.bin_last x)
                    (Bit_Representation.bin_last y))));

fun bitOR_int x =
  (fn xa => fn y => bitNOT_int (bitAND_int (bitNOT_int xa) (bitNOT_int y))) x;

fun bitXOR_int x =
  (fn xa => fn y =>
    bitOR_int (bitAND_int xa (bitNOT_int y)) (bitAND_int (bitNOT_int xa) y))
    x;

end; (*struct Bit_Int*)

structure Word : sig
  datatype 'a word = Word of Arith.int
  val uint : 'a Type_Length.len0 -> 'a word -> Arith.int
  val sint : 'a Type_Length.len -> 'a word -> Arith.int
  val unat : 'a Type_Length.len0 -> 'a word -> Arith.nat
  val word_of_int : 'a Type_Length.len0 -> Arith.int -> 'a word
  val shiftl1 : 'a Type_Length.len0 -> 'a word -> 'a word
  val shiftr1 : 'a Type_Length.len0 -> 'a word -> 'a word
  val sshiftr1 : 'a Type_Length.len -> 'a word -> 'a word
  val sshiftr : 'a Type_Length.len -> 'a word -> Arith.nat -> 'a word
  val word_sle : 'a Type_Length.len -> 'a word -> 'a word -> bool
  val plus_word : 'a Type_Length.len0 -> 'a word -> 'a word -> 'a word
  val zero_word : 'a Type_Length.len0 -> 'a word
  val bitOR_word : 'a Type_Length.len0 -> 'a word -> 'a word -> 'a word
  val equal_word : 'a Type_Length.len0 -> 'a word -> 'a word -> bool
  val minus_word : 'a Type_Length.len0 -> 'a word -> 'a word -> 'a word
  val times_word : 'a Type_Length.len0 -> 'a word -> 'a word -> 'a word
  val word_sless : 'a Type_Length.len -> 'a word -> 'a word -> bool
  val bitAND_word : 'a Type_Length.len0 -> 'a word -> 'a word -> 'a word
  val bitXOR_word : 'a Type_Length.len0 -> 'a word -> 'a word -> 'a word
  val shiftl_word : 'a Type_Length.len0 -> 'a word -> Arith.nat -> 'a word
  val shiftr_word : 'a Type_Length.len0 -> 'a word -> Arith.nat -> 'a word
end = struct

datatype 'a word = Word of Arith.int;

fun uint A_ (Word x) = x;

fun sint A_ w =
  Bit_Representation.sbintrunc
    (Arith.minus_nat (Type_Length.len_of (Type_Length.len0_len A_) HOL.Type)
      Arith.one_nat)
    (uint (Type_Length.len0_len A_) w);

fun unat A_ w = Arith.nat (uint A_ w);

fun word_of_int A_ k =
  Word (Arith.mod_int k
         (Arith.power Arith.power_int (Arith.Int_of_integer (2 : IntInf.int))
           (Type_Length.len_of A_ HOL.Type)));

fun shiftl1 A_ w =
  word_of_int A_ (Bit_Representation.bit (uint A_ w) Bit.Zero_bit);

fun shiftr1 A_ w = word_of_int A_ (Bit_Representation.bin_rest (uint A_ w));

fun sshiftr1 A_ w =
  word_of_int (Type_Length.len0_len A_)
    (Bit_Representation.bin_rest (sint A_ w));

fun sshiftr A_ w n = Arith.funpow n (sshiftr1 A_) w;

fun word_sle A_ a b = Arith.less_eq_int (sint A_ a) (sint A_ b);

fun plus_word A_ a b = word_of_int A_ (Arith.plus_int (uint A_ a) (uint A_ b));

fun zero_word A_ = word_of_int A_ Arith.zero_inta;

fun bitOR_word A_ a b =
  word_of_int A_ (Bit_Int.bitOR_int (uint A_ a) (uint A_ b));

fun equal_word A_ k l = Arith.equal_int (uint A_ k) (uint A_ l);

fun minus_word A_ a b =
  word_of_int A_ (Arith.minus_int (uint A_ a) (uint A_ b));

fun times_word A_ a b =
  word_of_int A_ (Arith.times_inta (uint A_ a) (uint A_ b));

fun word_sless A_ x y =
  word_sle A_ x y andalso not (equal_word (Type_Length.len0_len A_) x y);

fun bitAND_word A_ a b =
  word_of_int A_ (Bit_Int.bitAND_int (uint A_ a) (uint A_ b));

fun bitXOR_word A_ a b =
  word_of_int A_ (Bit_Int.bitXOR_int (uint A_ a) (uint A_ b));

fun shiftl_word A_ w n = Arith.funpow n (shiftl1 A_) w;

fun shiftr_word A_ w n = Arith.funpow n (shiftr1 A_) w;

end; (*struct Word*)

structure Predicate : sig
  datatype 'a seq = Empty | Insert of 'a * 'a pred | Join of 'a pred * 'a seq
  and 'a pred = Seq of (unit -> 'a seq)
  val is_empty : 'a pred -> bool
  val null : 'a seq -> bool
  val singleton : 'a HOL.equal -> (unit -> 'a) -> 'a pred -> 'a
  val the_only : 'a HOL.equal -> (unit -> 'a) -> 'a seq -> 'a
  val the : 'a HOL.equal -> 'a pred -> 'a
  val bind : 'a pred -> ('a -> 'b pred) -> 'b pred
  val apply : ('a -> 'b pred) -> 'a seq -> 'b seq
  val eval : 'a HOL.equal -> 'a pred -> 'a -> bool
  val member : 'a HOL.equal -> 'a seq -> 'a -> bool
  val holds : unit pred -> bool
  val bot_pred : 'a pred
  val single : 'a -> 'a pred
  val adjunct : 'a pred -> 'a seq -> 'a seq
  val sup_pred : 'a pred -> 'a pred -> 'a pred
  val if_pred : bool -> unit pred
  val not_pred : unit pred -> unit pred
  val set_of_seq : 'a seq -> 'a Set.set
  val set_of_pred : 'a pred -> 'a Set.set
  val pred_of_set : 'a Set.set -> 'a pred
end = struct

datatype 'a seq = Empty | Insert of 'a * 'a pred | Join of 'a pred * 'a seq
and 'a pred = Seq of (unit -> 'a seq);

fun is_empty (Seq f) = null (f ())
and null Empty = true
  | null (Insert (x, p)) = false
  | null (Join (p, xq)) = is_empty p andalso null xq;

fun singleton A_ dfault (Seq f) =
  (case f () of Empty => dfault ()
    | Insert (x, p) =>
      (if is_empty p then x
        else let
               val y = singleton A_ dfault p;
             in
               (if HOL.eq A_ x y then x else dfault ())
             end)
    | Join (p, xq) =>
      (if is_empty p then the_only A_ dfault xq
        else (if null xq then singleton A_ dfault p
               else let
                      val x = singleton A_ dfault p;
                      val y = the_only A_ dfault xq;
                    in
                      (if HOL.eq A_ x y then x else dfault ())
                    end)))
and the_only A_ dfault (Insert (x, p)) =
  (if is_empty p then x
    else let
           val y = singleton A_ dfault p;
         in
           (if HOL.eq A_ x y then x else dfault ())
         end)
  | the_only A_ dfault (Join (p, xq)) =
    (if is_empty p then the_only A_ dfault xq
      else (if null xq then singleton A_ dfault p
             else let
                    val x = singleton A_ dfault p;
                    val y = the_only A_ dfault xq;
                  in
                    (if HOL.eq A_ x y then x else dfault ())
                  end));

fun the A_ a =
  singleton A_ (fn _ => (raise Fail "not_unique") (fn _ => the A_ a)) a;

fun bind (Seq g) f = Seq (fn _ => apply f (g ()))
and apply f Empty = Empty
  | apply f (Insert (x, p)) = Join (f x, Join (bind p f, Empty))
  | apply f (Join (p, xq)) = Join (bind p f, apply f xq);

fun eval A_ (Seq f) = member A_ (f ())
and member A_ Empty x = false
  | member A_ (Insert (y, p)) x = HOL.eq A_ x y orelse eval A_ p x
  | member A_ (Join (p, xq)) x = eval A_ p x orelse member A_ xq x;

fun holds p = eval Product_Type.equal_unit p ();

val bot_pred : 'a pred = Seq (fn _ => Empty);

fun single x = Seq (fn _ => Insert (x, bot_pred));

fun adjunct p Empty = Join (p, Empty)
  | adjunct p (Insert (x, q)) = Insert (x, sup_pred q p)
  | adjunct p (Join (q, xq)) = Join (q, adjunct p xq)
and sup_pred (Seq f) (Seq g) =
  Seq (fn _ =>
        (case f () of Empty => g ()
          | Insert (x, p) => Insert (x, sup_pred p (Seq g))
          | Join (p, xq) => adjunct (Seq g) (Join (p, xq))));

fun if_pred b = (if b then single () else bot_pred);

fun not_pred p =
  (if eval Product_Type.equal_unit p () then bot_pred else single ());

fun set_of_seq (Join (p, xq)) = Set.union (set_of_pred p) (set_of_seq xq)
  | set_of_seq (Insert (x, p)) = Set.insert x (set_of_pred p)
  | set_of_seq Empty = Set.bot_seta
and set_of_pred (Seq f) =
  (case f () of Empty => Set.bot_seta
    | Insert (x, p) => Set.insert x (set_of_pred p)
    | Join (p, xq) => Set.union (set_of_pred p) (set_of_seq xq));

fun pred_of_set (Set.Set xs) = List.foldr (sup_pred o single) xs bot_pred;

end; (*struct Predicate*)

structure Type : sig
  datatype ty = Void | Boolean | Integer | Nt | Class of string | Array of ty
  datatype htype = Class_type of string | Array_type of ty * Arith.nat
  val object : string
  val thread : string
  val equal_tya : ty -> ty -> bool
  val equal_ty : ty HOL.equal
  val is_refT_i : ty -> unit Predicate.pred
  val is_refT : ty -> bool
  val the_Array : ty -> ty
  val the_Class : ty -> string
  val throwable : string
  val equal_htype : htype -> htype -> bool
  val ground_type : ty -> ty
  val ty_of_htype : htype -> ty
  val alen_of_htype : htype -> Arith.nat
  val class_type_of : ty -> string option
  val class_type_ofa : htype -> string
end = struct

datatype ty = Void | Boolean | Integer | Nt | Class of string | Array of ty;

datatype htype = Class_type of string | Array_type of ty * Arith.nat;

val object : string = "java/lang/Object";

val thread : string = "java/lang/Thread";

fun equal_tya (Array ty) (Class literal) = false
  | equal_tya (Class literal) (Array ty) = false
  | equal_tya (Array ty) Nt = false
  | equal_tya Nt (Array ty) = false
  | equal_tya (Class literal) Nt = false
  | equal_tya Nt (Class literal) = false
  | equal_tya (Array ty) Integer = false
  | equal_tya Integer (Array ty) = false
  | equal_tya (Class literal) Integer = false
  | equal_tya Integer (Class literal) = false
  | equal_tya Nt Integer = false
  | equal_tya Integer Nt = false
  | equal_tya (Array ty) Boolean = false
  | equal_tya Boolean (Array ty) = false
  | equal_tya (Class literal) Boolean = false
  | equal_tya Boolean (Class literal) = false
  | equal_tya Nt Boolean = false
  | equal_tya Boolean Nt = false
  | equal_tya Integer Boolean = false
  | equal_tya Boolean Integer = false
  | equal_tya (Array ty) Void = false
  | equal_tya Void (Array ty) = false
  | equal_tya (Class literal) Void = false
  | equal_tya Void (Class literal) = false
  | equal_tya Nt Void = false
  | equal_tya Void Nt = false
  | equal_tya Integer Void = false
  | equal_tya Void Integer = false
  | equal_tya Boolean Void = false
  | equal_tya Void Boolean = false
  | equal_tya (Array tya) (Array ty) = equal_tya tya ty
  | equal_tya (Class literala) (Class literal) = ((literala : string) = literal)
  | equal_tya Nt Nt = true
  | equal_tya Integer Integer = true
  | equal_tya Boolean Boolean = true
  | equal_tya Void Void = true;

val equal_ty = {equal = equal_tya} : ty HOL.equal;

fun is_refT_i x =
  (case x of Void => Predicate.bot_pred | Boolean => Predicate.bot_pred
    | Integer => Predicate.bot_pred | Nt => Predicate.single ()
    | Class _ => Predicate.single () | Array _ => Predicate.single ());

fun is_refT x1 = Predicate.holds (is_refT_i x1);

fun the_Array (Array t) = t;

fun the_Class (Class c) = c;

val throwable : string = "java/lang/Throwable";

fun equal_htype (Array_type (ty, nat)) (Class_type literal) = false
  | equal_htype (Class_type literal) (Array_type (ty, nat)) = false
  | equal_htype (Array_type (tya, nata)) (Array_type (ty, nat)) =
    equal_tya tya ty andalso Arith.equal_nata nata nat
  | equal_htype (Class_type literala) (Class_type literal) =
    ((literala : string) = literal);

fun ground_type (Array t) = ground_type t
  | ground_type Void = Void
  | ground_type Boolean = Boolean
  | ground_type Integer = Integer
  | ground_type Nt = Nt
  | ground_type (Class v) = Class v;

fun ty_of_htype (Class_type c) = Class c
  | ty_of_htype (Array_type (t, n)) = Array t;

fun alen_of_htype (Array_type (t, n)) = n;

fun class_type_of (Class c) = SOME c
  | class_type_of (Array t) = SOME object
  | class_type_of Void = NONE
  | class_type_of Boolean = NONE
  | class_type_of Integer = NONE
  | class_type_of Nt = NONE;

fun class_type_ofa (Class_type c) = c
  | class_type_ofa (Array_type (t, n)) = object;

end; (*struct Type*)

structure Value : sig
  datatype 'a vala = Unit | Null | Bool of bool |
    Intg of
      Numeral_Type.num1 Numeral_Type.bit0 Numeral_Type.bit0 Numeral_Type.bit0
        Numeral_Type.bit0
        Numeral_Type.bit0
        Word.word
    | Addr of 'a
  type 'a addr
  val hash_addr : 'a addr -> 'a -> Arith.int
  val monitor_finfun_to_list :
    'a addr -> ('a, Arith.nat) FinFun.finfun -> 'a list
  val equal_vala : 'a HOL.equal -> 'a vala -> 'a vala -> bool
  val equal_val : 'a HOL.equal -> 'a vala HOL.equal
  val the_Addr : 'a vala -> 'a
  val the_Intg :
    'a vala ->
      Numeral_Type.num1 Numeral_Type.bit0 Numeral_Type.bit0 Numeral_Type.bit0
        Numeral_Type.bit0
        Numeral_Type.bit0
        Word.word
  val default_val : Type.ty -> 'a vala
  val undefined_value : 'a vala
end = struct

datatype 'a vala = Unit | Null | Bool of bool |
  Intg of
    Numeral_Type.num1 Numeral_Type.bit0 Numeral_Type.bit0 Numeral_Type.bit0
      Numeral_Type.bit0
      Numeral_Type.bit0
      Word.word
  | Addr of 'a;

type 'a addr =
  {hash_addr : 'a -> Arith.int,
    monitor_finfun_to_list : ('a, Arith.nat) FinFun.finfun -> 'a list};
val hash_addr = #hash_addr : 'a addr -> 'a -> Arith.int;
val monitor_finfun_to_list = #monitor_finfun_to_list :
  'a addr -> ('a, Arith.nat) FinFun.finfun -> 'a list;

fun equal_vala A_ (Addr addr) (Intg word) = false
  | equal_vala A_ (Intg word) (Addr addr) = false
  | equal_vala A_ (Addr addr) (Bool boola) = false
  | equal_vala A_ (Bool boola) (Addr addr) = false
  | equal_vala A_ (Intg word) (Bool boola) = false
  | equal_vala A_ (Bool boola) (Intg word) = false
  | equal_vala A_ (Addr addr) Null = false
  | equal_vala A_ Null (Addr addr) = false
  | equal_vala A_ (Intg word) Null = false
  | equal_vala A_ Null (Intg word) = false
  | equal_vala A_ (Bool boola) Null = false
  | equal_vala A_ Null (Bool boola) = false
  | equal_vala A_ (Addr addr) Unit = false
  | equal_vala A_ Unit (Addr addr) = false
  | equal_vala A_ (Intg word) Unit = false
  | equal_vala A_ Unit (Intg word) = false
  | equal_vala A_ (Bool boola) Unit = false
  | equal_vala A_ Unit (Bool boola) = false
  | equal_vala A_ Null Unit = false
  | equal_vala A_ Unit Null = false
  | equal_vala A_ (Addr addra) (Addr addr) = HOL.eq A_ addra addr
  | equal_vala A_ (Intg worda) (Intg word) =
    Word.equal_word
      (Type_Length.len0_bit0
        (Type_Length.len0_bit0
          (Type_Length.len0_bit0
            (Type_Length.len0_bit0
              (Type_Length.len0_bit0 Type_Length.len0_num1)))))
      worda word
  | equal_vala A_ (Bool boolaa) (Bool boola) =
    Product_Type.equal_bool boolaa boola
  | equal_vala A_ Null Null = true
  | equal_vala A_ Unit Unit = true;

fun equal_val A_ = {equal = equal_vala A_} : 'a vala HOL.equal;

fun the_Addr (Addr a) = a;

fun the_Intg (Intg i) = i;

fun default_val Type.Void = Unit
  | default_val Type.Boolean = Bool false
  | default_val Type.Integer =
    Intg (Word.zero_word
           (Type_Length.len0_bit0
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0 Type_Length.len0_num1))))))
  | default_val Type.Nt = Null
  | default_val (Type.Class c) = Null
  | default_val (Type.Array a) = Null;

val undefined_value : 'a vala = Unit;

end; (*struct Value*)

structure MM : sig
  val monitor_finfun_to_list_nat :
    (Arith.nat, Arith.nat) FinFun.finfun -> Arith.nat list
  val hash_addr_nat : Arith.nat -> Arith.int
  val addr_nat : Arith.nat Value.addr
end = struct

fun monitor_finfun_to_list_nat x =
  FinFun.finfun_to_list (Set.card_UNIV_nat, Arith.equal_nat, Arith.linorder_nat)
    Arith.equal_nat x;

fun hash_addr_nat x = Arith.int_of_nat x;

val addr_nat =
  {hash_addr = hash_addr_nat,
    monitor_finfun_to_list = monitor_finfun_to_list_nat}
  : Arith.nat Value.addr;

end; (*struct MM*)

structure Semilat : sig
  val lesub : 'a -> ('a -> 'a -> bool) -> 'a -> bool
  val plussub : 'a -> ('a -> 'b -> 'c) -> 'b -> 'c
end = struct

fun lesub x r y = r x y;

fun plussub x f y = f x y;

end; (*struct Semilat*)

structure Err : sig
  datatype 'a err = Err | Ok of 'a
  val le : ('a -> 'a -> bool) -> 'a err -> 'a err -> bool
  val lift2 : ('a -> 'b -> 'c err) -> 'a err -> 'b err -> 'c err
  val sup : ('a -> 'b -> 'c) -> 'a err -> 'b err -> 'c err
  val ok_val : 'a err -> 'a
  val equal_erra : 'a HOL.equal -> 'a err -> 'a err -> bool
  val equal_err : 'a HOL.equal -> 'a err HOL.equal
end = struct

datatype 'a err = Err | Ok of 'a;

fun le r e_1 e_2 =
  (case e_2 of Err => true
    | Ok y => (case e_1 of Err => false | Ok x => Semilat.lesub x r y));

fun lift2 f e_1 e_2 =
  (case e_1 of Err => Err | Ok x => (case e_2 of Err => Err | Ok a => f x a));

fun sup f = lift2 (fn x => fn y => Ok (Semilat.plussub x f y));

fun ok_val (Ok x) = x;

fun equal_erra A_ (Ok a) Err = false
  | equal_erra A_ Err (Ok a) = false
  | equal_erra A_ (Ok aa) (Ok a) = HOL.eq A_ aa a
  | equal_erra A_ Err Err = true;

fun equal_err A_ = {equal = equal_erra A_} : 'a err HOL.equal;

end; (*struct Err*)

structure Option : sig
  val map : ('a -> 'b) -> 'a option -> 'b option
  val the : 'a option -> 'a
  val bind : 'a option -> ('a -> 'b option) -> 'b option
  val is_none : 'a option -> bool
  val equal_optiona : 'a HOL.equal -> 'a option -> 'a option -> bool
  val equal_option : 'a HOL.equal -> ('a option) HOL.equal
end = struct

fun map f (SOME x) = SOME (f x)
  | map f NONE = NONE;

fun the (SOME x) = x;

fun bind NONE f = NONE
  | bind (SOME x) f = f x;

fun is_none (SOME x) = false
  | is_none NONE = true;

fun equal_optiona A_ (SOME a) NONE = false
  | equal_optiona A_ NONE (SOME a) = false
  | equal_optiona A_ (SOME aa) (SOME a) = HOL.eq A_ aa a
  | equal_optiona A_ NONE NONE = true;

fun equal_option A_ = {equal = equal_optiona A_} : ('a option) HOL.equal;

end; (*struct Option*)

structure Map : sig
  val map_of : 'a HOL.equal -> ('a * 'b) list -> 'a -> 'b option
  val map_add : ('a -> 'b option) -> ('a -> 'b option) -> 'a -> 'b option
end = struct

fun map_of A_ ((l, v) :: ps) k =
  (if HOL.eq A_ l k then SOME v else map_of A_ ps k)
  | map_of A_ [] k = NONE;

fun map_add m1 m2 = (fn x => (case m2 x of NONE => m1 x | SOME a => SOME a));

end; (*struct Map*)

structure Opt : sig
  val le : ('a -> 'a -> bool) -> 'a option -> 'a option -> bool
  val sup :
    ('a -> 'a -> 'a Err.err) -> 'a option -> 'a option -> ('a option) Err.err
end = struct

fun le r o_1 o_2 =
  (case o_2 of NONE => Option.is_none o_1
    | SOME y => (case o_1 of NONE => true | SOME x => Semilat.lesub x r y));

fun sup f o_1 o_2 =
  (case o_1 of NONE => Err.Ok o_2
    | SOME x =>
      (case o_2 of NONE => Err.Ok o_1
        | SOME y =>
          (case f x y of Err.Err => Err.Err | Err.Ok z => Err.Ok (SOME z))));

end; (*struct Opt*)

structure RBT_Impl : sig
  datatype color = R | B
  datatype ('a, 'b) rbt = Empty |
    Branch of color * ('a, 'b) rbt * 'a * 'b * ('a, 'b) rbt
  val paint : color -> ('a, 'b) rbt -> ('a, 'b) rbt
  val balance : ('a, 'b) rbt -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val balance_left : ('a, 'b) rbt -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val combine : ('a, 'b) rbt -> ('a, 'b) rbt -> ('a, 'b) rbt
  val balance_right : ('a, 'b) rbt -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val rbt_del : 'a Orderings.ord -> 'a -> ('a, 'b) rbt -> ('a, 'b) rbt
  val rbt_del_from_left :
    'a Orderings.ord ->
      'a -> ('a, 'b) rbt -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val rbt_del_from_right :
    'a Orderings.ord ->
      'a -> ('a, 'b) rbt -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val rbt_ins :
    'a Orderings.ord ->
      ('a -> 'b -> 'b -> 'b) -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val equal_color : color -> color -> bool
  val equal_rbt :
    'a HOL.equal -> 'b HOL.equal -> ('a, 'b) rbt -> ('a, 'b) rbt -> bool
  val rbt_delete : 'a Orderings.ord -> 'a -> ('a, 'b) rbt -> ('a, 'b) rbt
  val rbt_insert_with_key :
    'a Orderings.ord ->
      ('a -> 'b -> 'b -> 'b) -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val rbt_insert : 'a Orderings.ord -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val rbt_lookup : 'a Orderings.ord -> ('a, 'b) rbt -> 'a -> 'b option
end = struct

datatype color = R | B;

datatype ('a, 'b) rbt = Empty |
  Branch of color * ('a, 'b) rbt * 'a * 'b * ('a, 'b) rbt;

fun paint c Empty = Empty
  | paint c (Branch (uu, l, k, v, r)) = Branch (c, l, k, v, r);

fun balance (Branch (R, a, w, x, b)) s t (Branch (R, c, y, z, d)) =
  Branch (R, Branch (B, a, w, x, b), s, t, Branch (B, c, y, z, d))
  | balance (Branch (R, Branch (R, a, w, x, b), s, t, c)) y z Empty =
    Branch (R, Branch (B, a, w, x, b), s, t, Branch (B, c, y, z, Empty))
  | balance (Branch (R, Branch (R, a, w, x, b), s, t, c)) y z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, Branch (B, a, w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance (Branch (R, Empty, w, x, Branch (R, b, s, t, c))) y z Empty =
    Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, Empty))
  | balance
    (Branch (R, Branch (B, va, vb, vc, vd), w, x, Branch (R, b, s, t, c))) y z
    Empty =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, Empty))
  | balance (Branch (R, Empty, w, x, Branch (R, b, s, t, c))) y z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, Branch (B, Empty, w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance
    (Branch (R, Branch (B, ve, vf, vg, vh), w, x, Branch (R, b, s, t, c))) y z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, Branch (B, Branch (B, ve, vf, vg, vh), w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance Empty w x (Branch (R, b, s, t, Branch (R, c, y, z, d))) =
    Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, d))
  | balance (Branch (B, va, vb, vc, vd)) w x
    (Branch (R, b, s, t, Branch (R, c, y, z, d))) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, d))
  | balance Empty w x (Branch (R, Branch (R, b, s, t, c), y, z, Empty)) =
    Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, Empty))
  | balance Empty w x
    (Branch (R, Branch (R, b, s, t, c), y, z, Branch (B, va, vb, vc, vd))) =
    Branch
      (R, Branch (B, Empty, w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance (Branch (B, va, vb, vc, vd)) w x
    (Branch (R, Branch (R, b, s, t, c), y, z, Empty)) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, Empty))
  | balance (Branch (B, va, vb, vc, vd)) w x
    (Branch (R, Branch (R, b, s, t, c), y, z, Branch (B, ve, vf, vg, vh))) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, ve, vf, vg, vh)))
  | balance Empty s t Empty = Branch (B, Empty, s, t, Empty)
  | balance Empty s t (Branch (B, va, vb, vc, vd)) =
    Branch (B, Empty, s, t, Branch (B, va, vb, vc, vd))
  | balance Empty s t (Branch (v, Empty, vb, vc, Empty)) =
    Branch (B, Empty, s, t, Branch (v, Empty, vb, vc, Empty))
  | balance Empty s t (Branch (v, Branch (B, ve, vf, vg, vh), vb, vc, Empty)) =
    Branch
      (B, Empty, s, t, Branch (v, Branch (B, ve, vf, vg, vh), vb, vc, Empty))
  | balance Empty s t (Branch (v, Empty, vb, vc, Branch (B, vf, vg, vh, vi))) =
    Branch
      (B, Empty, s, t, Branch (v, Empty, vb, vc, Branch (B, vf, vg, vh, vi)))
  | balance Empty s t
    (Branch (v, Branch (B, ve, vj, vk, vl), vb, vc, Branch (B, vf, vg, vh, vi)))
    = Branch
        (B, Empty, s, t,
          Branch
            (v, Branch (B, ve, vj, vk, vl), vb, vc, Branch (B, vf, vg, vh, vi)))
  | balance (Branch (B, va, vb, vc, vd)) s t Empty =
    Branch (B, Branch (B, va, vb, vc, vd), s, t, Empty)
  | balance (Branch (B, va, vb, vc, vd)) s t (Branch (B, ve, vf, vg, vh)) =
    Branch (B, Branch (B, va, vb, vc, vd), s, t, Branch (B, ve, vf, vg, vh))
  | balance (Branch (B, va, vb, vc, vd)) s t (Branch (v, Empty, vf, vg, Empty))
    = Branch
        (B, Branch (B, va, vb, vc, vd), s, t, Branch (v, Empty, vf, vg, Empty))
  | balance (Branch (B, va, vb, vc, vd)) s t
    (Branch (v, Branch (B, vi, vj, vk, vl), vf, vg, Empty)) =
    Branch
      (B, Branch (B, va, vb, vc, vd), s, t,
        Branch (v, Branch (B, vi, vj, vk, vl), vf, vg, Empty))
  | balance (Branch (B, va, vb, vc, vd)) s t
    (Branch (v, Empty, vf, vg, Branch (B, vj, vk, vl, vm))) =
    Branch
      (B, Branch (B, va, vb, vc, vd), s, t,
        Branch (v, Empty, vf, vg, Branch (B, vj, vk, vl, vm)))
  | balance (Branch (B, va, vb, vc, vd)) s t
    (Branch (v, Branch (B, vi, vn, vo, vp), vf, vg, Branch (B, vj, vk, vl, vm)))
    = Branch
        (B, Branch (B, va, vb, vc, vd), s, t,
          Branch
            (v, Branch (B, vi, vn, vo, vp), vf, vg, Branch (B, vj, vk, vl, vm)))
  | balance (Branch (v, Empty, vb, vc, Empty)) s t Empty =
    Branch (B, Branch (v, Empty, vb, vc, Empty), s, t, Empty)
  | balance (Branch (v, Empty, vb, vc, Branch (B, ve, vf, vg, vh))) s t Empty =
    Branch
      (B, Branch (v, Empty, vb, vc, Branch (B, ve, vf, vg, vh)), s, t, Empty)
  | balance (Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Empty)) s t Empty =
    Branch
      (B, Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Empty), s, t, Empty)
  | balance
    (Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Branch (B, ve, vj, vk, vl)))
    s t Empty =
    Branch
      (B, Branch
            (v, Branch (B, vf, vg, vh, vi), vb, vc, Branch (B, ve, vj, vk, vl)),
        s, t, Empty)
  | balance (Branch (v, Empty, vf, vg, Empty)) s t (Branch (B, va, vb, vc, vd))
    = Branch
        (B, Branch (v, Empty, vf, vg, Empty), s, t, Branch (B, va, vb, vc, vd))
  | balance (Branch (v, Empty, vf, vg, Branch (B, vi, vj, vk, vl))) s t
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (B, Branch (v, Empty, vf, vg, Branch (B, vi, vj, vk, vl)), s, t,
        Branch (B, va, vb, vc, vd))
  | balance (Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Empty)) s t
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (B, Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Empty), s, t,
        Branch (B, va, vb, vc, vd))
  | balance
    (Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Branch (B, vi, vn, vo, vp)))
    s t (Branch (B, va, vb, vc, vd)) =
    Branch
      (B, Branch
            (v, Branch (B, vj, vk, vl, vm), vf, vg, Branch (B, vi, vn, vo, vp)),
        s, t, Branch (B, va, vb, vc, vd));

fun balance_left (Branch (R, a, k, x, b)) s y c =
  Branch (R, Branch (B, a, k, x, b), s, y, c)
  | balance_left Empty k x (Branch (B, a, s, y, b)) =
    balance Empty k x (Branch (R, a, s, y, b))
  | balance_left (Branch (B, va, vb, vc, vd)) k x (Branch (B, a, s, y, b)) =
    balance (Branch (B, va, vb, vc, vd)) k x (Branch (R, a, s, y, b))
  | balance_left Empty k x (Branch (R, Branch (B, a, s, y, b), t, z, c)) =
    Branch (R, Branch (B, Empty, k, x, a), s, y, balance b t z (paint R c))
  | balance_left (Branch (B, va, vb, vc, vd)) k x
    (Branch (R, Branch (B, a, s, y, b), t, z, c)) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), k, x, a), s, y,
        balance b t z (paint R c))
  | balance_left Empty k x Empty = Empty
  | balance_left Empty k x (Branch (R, Empty, vb, vc, vd)) = Empty
  | balance_left Empty k x (Branch (R, Branch (R, ve, vf, vg, vh), vb, vc, vd))
    = Empty
  | balance_left (Branch (B, va, vb, vc, vd)) k x Empty = Empty
  | balance_left (Branch (B, va, vb, vc, vd)) k x
    (Branch (R, Empty, vf, vg, vh)) = Empty
  | balance_left (Branch (B, va, vb, vc, vd)) k x
    (Branch (R, Branch (R, vi, vj, vk, vl), vf, vg, vh)) = Empty;

fun combine Empty x = x
  | combine (Branch (v, va, vb, vc, vd)) Empty = Branch (v, va, vb, vc, vd)
  | combine (Branch (R, a, k, x, b)) (Branch (R, c, s, y, d)) =
    (case combine b c
      of Empty => Branch (R, a, k, x, Branch (R, Empty, s, y, d))
      | Branch (R, b2, t, z, c2) =>
        Branch (R, Branch (R, a, k, x, b2), t, z, Branch (R, c2, s, y, d))
      | Branch (B, b2, t, z, c2) =>
        Branch (R, a, k, x, Branch (R, Branch (B, b2, t, z, c2), s, y, d)))
  | combine (Branch (B, a, k, x, b)) (Branch (B, c, s, y, d)) =
    (case combine b c
      of Empty => balance_left a k x (Branch (B, Empty, s, y, d))
      | Branch (R, b2, t, z, c2) =>
        Branch (R, Branch (B, a, k, x, b2), t, z, Branch (B, c2, s, y, d))
      | Branch (B, b2, t, z, c2) =>
        balance_left a k x (Branch (B, Branch (B, b2, t, z, c2), s, y, d)))
  | combine (Branch (B, va, vb, vc, vd)) (Branch (R, b, k, x, c)) =
    Branch (R, combine (Branch (B, va, vb, vc, vd)) b, k, x, c)
  | combine (Branch (R, a, k, x, b)) (Branch (B, va, vb, vc, vd)) =
    Branch (R, a, k, x, combine b (Branch (B, va, vb, vc, vd)));

fun balance_right a k x (Branch (R, b, s, y, c)) =
  Branch (R, a, k, x, Branch (B, b, s, y, c))
  | balance_right (Branch (B, a, k, x, b)) s y Empty =
    balance (Branch (R, a, k, x, b)) s y Empty
  | balance_right (Branch (B, a, k, x, b)) s y (Branch (B, va, vb, vc, vd)) =
    balance (Branch (R, a, k, x, b)) s y (Branch (B, va, vb, vc, vd))
  | balance_right (Branch (R, a, k, x, Branch (B, b, s, y, c))) t z Empty =
    Branch (R, balance (paint R a) k x b, s, y, Branch (B, c, t, z, Empty))
  | balance_right (Branch (R, a, k, x, Branch (B, b, s, y, c))) t z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, balance (paint R a) k x b, s, y,
        Branch (B, c, t, z, Branch (B, va, vb, vc, vd)))
  | balance_right Empty k x Empty = Empty
  | balance_right (Branch (R, va, vb, vc, Empty)) k x Empty = Empty
  | balance_right (Branch (R, va, vb, vc, Branch (R, ve, vf, vg, vh))) k x Empty
    = Empty
  | balance_right Empty k x (Branch (B, va, vb, vc, vd)) = Empty
  | balance_right (Branch (R, ve, vf, vg, Empty)) k x
    (Branch (B, va, vb, vc, vd)) = Empty
  | balance_right (Branch (R, ve, vf, vg, Branch (R, vi, vj, vk, vl))) k x
    (Branch (B, va, vb, vc, vd)) = Empty;

fun rbt_del A_ x Empty = Empty
  | rbt_del A_ x (Branch (c, a, y, s, b)) =
    (if Orderings.less A_ x y then rbt_del_from_left A_ x a y s b
      else (if Orderings.less A_ y x then rbt_del_from_right A_ x a y s b
             else combine a b))
and rbt_del_from_left A_ x (Branch (B, lt, z, v, rt)) y s b =
  balance_left (rbt_del A_ x (Branch (B, lt, z, v, rt))) y s b
  | rbt_del_from_left A_ x Empty y s b = Branch (R, rbt_del A_ x Empty, y, s, b)
  | rbt_del_from_left A_ x (Branch (R, va, vb, vc, vd)) y s b =
    Branch (R, rbt_del A_ x (Branch (R, va, vb, vc, vd)), y, s, b)
and rbt_del_from_right A_ x a y s (Branch (B, lt, z, v, rt)) =
  balance_right a y s (rbt_del A_ x (Branch (B, lt, z, v, rt)))
  | rbt_del_from_right A_ x a y s Empty =
    Branch (R, a, y, s, rbt_del A_ x Empty)
  | rbt_del_from_right A_ x a y s (Branch (R, va, vb, vc, vd)) =
    Branch (R, a, y, s, rbt_del A_ x (Branch (R, va, vb, vc, vd)));

fun rbt_ins A_ f k v Empty = Branch (R, Empty, k, v, Empty)
  | rbt_ins A_ f k v (Branch (B, l, x, y, r)) =
    (if Orderings.less A_ k x then balance (rbt_ins A_ f k v l) x y r
      else (if Orderings.less A_ x k then balance l x y (rbt_ins A_ f k v r)
             else Branch (B, l, x, f k y v, r)))
  | rbt_ins A_ f k v (Branch (R, l, x, y, r)) =
    (if Orderings.less A_ k x then Branch (R, rbt_ins A_ f k v l, x, y, r)
      else (if Orderings.less A_ x k
             then Branch (R, l, x, y, rbt_ins A_ f k v r)
             else Branch (R, l, x, f k y v, r)));

fun equal_color B R = false
  | equal_color R B = false
  | equal_color B B = true
  | equal_color R R = true;

fun equal_rbt A_ B_ (Branch (color, rbt1, a, b, rbt2)) Empty = false
  | equal_rbt A_ B_ Empty (Branch (color, rbt1, a, b, rbt2)) = false
  | equal_rbt A_ B_ (Branch (colora, rbt1a, aa, ba, rbt2a))
    (Branch (color, rbt1, a, b, rbt2)) =
    equal_color colora color andalso
      (equal_rbt A_ B_ rbt1a rbt1 andalso
        (HOL.eq A_ aa a andalso
          (HOL.eq B_ ba b andalso equal_rbt A_ B_ rbt2a rbt2)))
  | equal_rbt A_ B_ Empty Empty = true;

fun rbt_delete A_ k t = paint B (rbt_del A_ k t);

fun rbt_insert_with_key A_ f k v t = paint B (rbt_ins A_ f k v t);

fun rbt_insert A_ = rbt_insert_with_key A_ (fn _ => fn _ => fn nv => nv);

fun rbt_lookup A_ Empty k = NONE
  | rbt_lookup A_ (Branch (uu, l, x, y, r)) k =
    (if Orderings.less A_ k x then rbt_lookup A_ l k
      else (if Orderings.less A_ x k then rbt_lookup A_ r k else SOME y));

end; (*struct RBT_Impl*)

structure RBT : sig
  datatype ('a, 'b) rbt = Rbt of ('a, 'b) RBT_Impl.rbt
  val empty : 'a Orderings.linorder -> ('a, 'b) rbt
  val impl_of : 'a Orderings.linorder -> ('a, 'b) rbt -> ('a, 'b) RBT_Impl.rbt
  val delete : 'a Orderings.linorder -> 'a -> ('a, 'b) rbt -> ('a, 'b) rbt
  val insert : 'a Orderings.linorder -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val lookup : 'a Orderings.linorder -> ('a, 'b) rbt -> 'a -> 'b option
end = struct

datatype ('a, 'b) rbt = Rbt of ('a, 'b) RBT_Impl.rbt;

fun empty A_ = Rbt RBT_Impl.Empty;

fun impl_of A_ (Rbt x) = x;

fun delete A_ x xa =
  Rbt (RBT_Impl.rbt_delete
        ((Orderings.ord_preorder o Orderings.preorder_order o
           Orderings.order_linorder)
          A_)
        x (impl_of A_ xa));

fun insert A_ x xa xb =
  Rbt (RBT_Impl.rbt_insert
        ((Orderings.ord_preorder o Orderings.preorder_order o
           Orderings.order_linorder)
          A_)
        x xa (impl_of A_ xb));

fun lookup A_ x =
  RBT_Impl.rbt_lookup
    ((Orderings.ord_preorder o Orderings.preorder_order o
       Orderings.order_linorder)
      A_)
    (impl_of A_ x);

end; (*struct RBT*)

structure Transitive_Closure_Table : sig
  val rtrancl_tab_FioB_i_i_i :
    'a HOL.equal ->
      ('a -> 'a Predicate.pred) -> 'a list -> 'a -> 'a -> unit Predicate.pred
  val rtranclp_FioB_i_i :
    'a HOL.equal -> ('a -> 'a Predicate.pred) -> 'a -> 'a -> unit Predicate.pred
end = struct

fun rtrancl_tab_FioB_i_i_i A_ r xa xb xc =
  Predicate.sup_pred
    (if HOL.eq A_ xb xc then Predicate.single () else Predicate.bot_pred)
    (Predicate.bind (Predicate.if_pred (not (List.member A_ xa xb)))
      (fn () =>
        Predicate.bind (r xb)
          (fn xaa =>
            Predicate.bind (rtrancl_tab_FioB_i_i_i A_ r (xb :: xa) xaa xc)
              (fn () => Predicate.single ()))));

fun rtranclp_FioB_i_i A_ r xa xb =
  Predicate.bind (rtrancl_tab_FioB_i_i_i A_ r [] xa xb)
    (fn () => Predicate.single ());

end; (*struct Transitive_Closure_Table*)

structure TypeRel : sig
  datatype 'a fmod_ext = Fmod_ext of bool * 'a
  datatype 'a prog =
    Program of
      (string *
        (string *
          ((string * (Type.ty * unit fmod_ext)) list *
            (string * (Type.ty list * (Type.ty * 'a option))) list))) list
  val class :
    'a prog ->
      string ->
        (string *
          ((string * (Type.ty * unit fmod_ext)) list *
            (string * (Type.ty list * (Type.ty * 'a option))) list)) option
  val eq_i_i : 'a HOL.equal -> 'a -> 'a -> unit Predicate.pred
  val classes :
    'a prog ->
      (string *
        (string *
          ((string * (Type.ty * unit fmod_ext)) list *
            (string * (Type.ty list * (Type.ty * 'a option))) list))) list
  val is_class : 'a prog -> string -> bool
  val is_type : 'a prog -> Type.ty -> bool
  val sees_field_hoaux_PPiiii_Poooo :
    (string * string) * (Type.ty * unit fmod_ext) ->
      (string * (string * (Type.ty * unit fmod_ext))) Predicate.pred
  val mapP_FioB_i_o :
    ('a -> 'b Predicate.pred) -> 'a list -> ('b list) Predicate.pred
  val fields_i_i_o :
    'a prog ->
      string ->
        (((string * string) * (Type.ty * unit fmod_ext)) list) Predicate.pred
  val sees_field_i_i_i_o_o_o :
    'a prog ->
      string -> string -> (Type.ty * (unit fmod_ext * string)) Predicate.pred
  val equal_fmod_exta : 'a HOL.equal -> 'a fmod_ext -> 'a fmod_ext -> bool
  val equal_fmod_ext : 'a HOL.equal -> 'a fmod_ext HOL.equal
  val field : 'a prog -> string -> string -> string * (Type.ty * unit fmod_ext)
  val subcls1_i_i_o : 'a prog -> string -> string Predicate.pred
  val widen_i_i_ia : 'a prog -> Type.ty -> Type.ty -> unit Predicate.pred
  val widen : 'a prog -> Type.ty -> Type.ty -> bool
  val fields :
    'a prog -> string -> ((string * string) * (Type.ty * unit fmod_ext)) list
  val methods_i_i_o :
    'a prog ->
      string ->
        (string -> ((Type.ty list * (Type.ty * 'a option)) * string) option)
          Predicate.pred
  val method_i_i_i_o_o_o_o :
    'a prog ->
      string ->
        string ->
          (Type.ty list * (Type.ty * ('a option * string))) Predicate.pred
  val method :
    'a HOL.equal ->
      'a prog ->
        string -> string -> string * (Type.ty list * (Type.ty * 'a option))
  val subcls_i_i_i : 'a prog -> string -> string -> unit Predicate.pred
  val subcls : 'a prog -> string -> string -> bool
  val widen_i_i_i : 'a prog -> Type.ty -> Type.ty -> unit Predicate.pred
  val sees_field_i_i_i_o_o_i :
    'a prog ->
      string -> string -> string -> (Type.ty * unit fmod_ext) Predicate.pred
end = struct

datatype 'a fmod_ext = Fmod_ext of bool * 'a;

datatype 'a prog =
  Program of
    (string *
      (string *
        ((string * (Type.ty * unit fmod_ext)) list *
          (string * (Type.ty list * (Type.ty * 'a option))) list))) list;

fun class (Program p) = Map.map_of Stringa.equal_literal p;

fun eq_i_i A_ xa xb =
  (if HOL.eq A_ xa xb then Predicate.single () else Predicate.bot_pred);

fun classes (Program p) = p;

fun is_class p c = not (Option.is_none (class p c));

fun is_type p Type.Void = true
  | is_type p Type.Boolean = true
  | is_type p Type.Integer = true
  | is_type p Type.Nt = true
  | is_type p (Type.Class c) = is_class p c
  | is_type p (Type.Array a) =
    (case Type.ground_type a of Type.Void => true | Type.Boolean => true
      | Type.Integer => true | Type.Nt => false | Type.Class aa => is_class p aa
      | Type.Array _ => true);

fun sees_field_hoaux_PPiiii_Poooo ((x, xf), (xg, xh)) =
  Predicate.single (x, (xf, (xg, xh)));

fun mapP_FioB_i_o h x =
  Predicate.sup_pred
    (case x of [] => Predicate.single [] | _ :: _ => Predicate.bot_pred)
    (case x of [] => Predicate.bot_pred
      | z :: xu =>
        Predicate.bind (mapP_FioB_i_o h xu)
          (fn xa =>
            Predicate.bind (h z) (fn xaa => Predicate.single (xaa :: xa))));

fun fields_i_i_o x xa =
  Predicate.sup_pred
    (Predicate.bind (Predicate.if_pred (not ((xa : string) = Type.object)))
      (fn () =>
        (case class x xa of NONE => Predicate.bot_pred
          | SOME (d, (fs, _)) =>
            Predicate.bind (fields_i_i_o x d)
              (fn xb =>
                Predicate.single
                  (List.map (fn (f, a) => ((f, xa), a)) fs @ xb)))))
    (case class x Type.object of NONE => Predicate.bot_pred
      | SOME (_, (fs, _)) =>
        (if ((xa : string) = Type.object)
          then Predicate.single
                 (List.map (fn (f, a) => ((f, Type.object), a)) fs)
          else Predicate.bot_pred));

fun sees_field_i_i_i_o_o_o x xa xb =
  Predicate.bind (fields_i_i_o x xa)
    (fn xc =>
      Predicate.bind (mapP_FioB_i_o sees_field_hoaux_PPiiii_Poooo xc)
        (fn xd =>
          (case Map.map_of Stringa.equal_literal xd xb
            of NONE => Predicate.bot_pred
            | SOME (dc, (tc, fp)) => Predicate.single (tc, (fp, dc)))));

fun equal_fmod_exta A_ (Fmod_ext (volatilea, morea)) (Fmod_ext (volatile, more))
  = Product_Type.equal_bool volatilea volatile andalso HOL.eq A_ morea more;

fun equal_fmod_ext A_ = {equal = equal_fmod_exta A_} : 'a fmod_ext HOL.equal;

fun field p c f =
  Predicate.the
    (Product_Type.equal_prod Stringa.equal_literal
      (Product_Type.equal_prod Type.equal_ty
        (equal_fmod_ext Product_Type.equal_unit)))
    (Predicate.bind (sees_field_i_i_i_o_o_o p c f)
      (fn (t, (fm, d)) => Predicate.single (d, (t, fm))));

fun subcls1_i_i_o x xa =
  Predicate.bind (Predicate.if_pred (not ((xa : string) = Type.object)))
    (fn () =>
      (case class x xa of NONE => Predicate.bot_pred
        | SOME (d, _) => Predicate.single d));

fun widen_i_i_ia x xa xb =
  Predicate.sup_pred
    (if Type.equal_tya xa xb then Predicate.single () else Predicate.bot_pred)
    (case xa of Type.Void => Predicate.bot_pred
      | Type.Boolean => Predicate.bot_pred | Type.Integer => Predicate.bot_pred
      | Type.Nt =>
        (case xb of Type.Void => Predicate.bot_pred
          | Type.Boolean => Predicate.bot_pred
          | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
          | Type.Class _ => Predicate.single ()
          | Type.Array _ => Predicate.single ())
      | Type.Class c =>
        (case xb of Type.Void => Predicate.bot_pred
          | Type.Boolean => Predicate.bot_pred
          | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
          | Type.Class d =>
            Predicate.bind
              (Transitive_Closure_Table.rtranclp_FioB_i_i Stringa.equal_literal
                (subcls1_i_i_o x) c d)
              (fn () => Predicate.single ())
          | Type.Array _ => Predicate.bot_pred)
      | Type.Array t =>
        (case xb of Type.Void => Predicate.bot_pred
          | Type.Boolean => Predicate.bot_pred
          | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
          | Type.Class c =>
            Predicate.bind (eq_i_i Stringa.equal_literal c Type.object)
              (fn () => Predicate.single ())
          | Type.Array ta =>
            Predicate.bind (widen_i_i_ia x t ta)
              (fn () => Predicate.single ())));

fun widen x1 x2 x3 = Predicate.holds (widen_i_i_ia x1 x2 x3);

fun fields p c =
  Predicate.the
    (List.equal_list
      (Product_Type.equal_prod
        (Product_Type.equal_prod Stringa.equal_literal Stringa.equal_literal)
        (Product_Type.equal_prod Type.equal_ty
          (equal_fmod_ext Product_Type.equal_unit))))
    (fields_i_i_o p c);

fun methods_i_i_o x xa =
  Predicate.sup_pred
    (case class x Type.object of NONE => Predicate.bot_pred
      | SOME (_, (_, ms)) =>
        (if ((xa : string) = Type.object)
          then Predicate.single
                 (Option.map (fn m => (m, Type.object)) o
                   Map.map_of Stringa.equal_literal ms)
          else Predicate.bot_pred))
    (Predicate.bind (Predicate.if_pred (not ((xa : string) = Type.object)))
      (fn () =>
        (case class x xa of NONE => Predicate.bot_pred
          | SOME (d, (_, ms)) =>
            Predicate.bind (methods_i_i_o x d)
              (fn xb =>
                Predicate.single
                  (Map.map_add xb
                    (Option.map (fn m => (m, xa)) o
                      Map.map_of Stringa.equal_literal ms))))));

fun method_i_i_i_o_o_o_o x xa xb =
  Predicate.bind (methods_i_i_o x xa)
    (fn xc =>
      (case xc xb of NONE => Predicate.bot_pred
        | SOME ((tv, (tc, p)), dc) => Predicate.single (tv, (tc, (p, dc)))));

fun method A_ p c m =
  Predicate.the
    (Product_Type.equal_prod Stringa.equal_literal
      (Product_Type.equal_prod (List.equal_list Type.equal_ty)
        (Product_Type.equal_prod Type.equal_ty (Option.equal_option A_))))
    (Predicate.bind (method_i_i_i_o_o_o_o p c m)
      (fn (ts, (t, (ma, d))) => Predicate.single (d, (ts, (t, ma)))));

fun subcls_i_i_i x xa xb =
  Predicate.bind
    (Transitive_Closure_Table.rtranclp_FioB_i_i Stringa.equal_literal
      (subcls1_i_i_o x) xa xb)
    (fn () => Predicate.single ());

fun subcls x1 x2 x3 = Predicate.holds (subcls_i_i_i x1 x2 x3);

fun widen_i_i_i p ta t =
  (if widen p ta t then Predicate.single () else Predicate.bot_pred);

fun sees_field_i_i_i_o_o_i x xa xb xc =
  Predicate.bind (fields_i_i_o x xa)
    (fn xd =>
      Predicate.bind (mapP_FioB_i_o sees_field_hoaux_PPiiii_Poooo xd)
        (fn xe =>
          (case Map.map_of Stringa.equal_literal xe xb
            of NONE => Predicate.bot_pred
            | SOME (dca, (tc, fp)) =>
              (if ((xc : string) = dca) then Predicate.single (tc, fp)
                else Predicate.bot_pred))));

end; (*struct TypeRel*)

structure Exceptions : sig
  val sys_xcpts_list : string list
  val wf_syscls : 'a TypeRel.prog -> bool
end = struct

val sys_xcpts_list : string list =
  ["java/lang/NullPointerException", "java/lang/ClassCastException",
    "java/lang/OutOfMemoryError", "java/lang/ArrayIndexOutOfBoundsException",
    "java/lang/ArrayStoreException", "java/lang/NegativeArraySizeException",
    "java/lang/ArithmeticException", "java/lang/IllegalMonitorStateException",
    "java/lang/IllegalThreadStateException", "java/lang/InterruptedException"];

fun wf_syscls p =
  List.list_all (TypeRel.is_class p)
    [Type.object, Type.throwable, Type.thread] andalso
    List.list_all (fn c => TypeRel.subcls p c Type.throwable) sys_xcpts_list;

end; (*struct Exceptions*)

structure StartConfig : sig
  val pick_addr : ('a * 'b) Set.set -> 'a * 'b
  val initialization_list : string list
  val start_heap_data :
    'a HOL.equal -> 'b HOL.equal * 'b Value.addr ->
      'a -> ('a -> Type.htype -> ('a * 'b) Set.set) -> 'a * ('b list * bool)
  val start_addrs :
    'a HOL.equal -> 'b HOL.equal * 'b Value.addr ->
      'a -> ('a -> Type.htype -> ('a * 'b) Set.set) -> 'b list
  val start_tid :
    'a HOL.equal * 'a Value.addr -> 'c HOL.equal ->
      ('a -> 'b) -> 'c -> ('c -> Type.htype -> ('c * 'a) Set.set) -> 'b
  val start_heap :
    'a HOL.equal -> 'b HOL.equal * 'b Value.addr ->
      'a -> ('a -> Type.htype -> ('a * 'b) Set.set) -> 'a
  val addr_of_sys_xcpt :
    'a HOL.equal -> 'b HOL.equal * 'b Value.addr ->
      'a -> ('a -> Type.htype -> ('a * 'b) Set.set) -> string -> 'b
end = struct

fun pick_addr (Set.Set [ha]) = ha;

val initialization_list : string list =
  Type.thread :: Exceptions.sys_xcpts_list;

fun start_heap_data A_ (B1_, B2_) empty_heap allocate =
  let
    val (h, (ads, b)) =
      List.foldl
        (fn (h, (ads, b)) => fn c =>
          (if b then let
                       val ha = allocate h (Type.Class_type c);
                     in
                       (if Set.equal_set (Product_Type.equal_prod A_ B1_) ha
                             Set.bot_seta
                         then (h, (ads, false))
                         else let
                                val (hb, a) = pick_addr ha;
                              in
                                (hb, (a :: ads, true))
                              end)
                     end
            else (h, (ads, false))))
        (empty_heap, ([], true)) initialization_list;
  in
    (h, (List.rev ads, b))
  end;

fun start_addrs A_ (B1_, B2_) empty_heap allocate =
  Product_Type.fst
    (Product_Type.snd (start_heap_data A_ (B1_, B2_) empty_heap allocate));

fun start_tid (A1_, A2_) C_ addr2thread_id empty_heap allocate =
  addr2thread_id (List.hd (start_addrs C_ (A1_, A2_) empty_heap allocate));

fun start_heap A_ (B1_, B2_) empty_heap allocate =
  Product_Type.fst (start_heap_data A_ (B1_, B2_) empty_heap allocate);

fun addr_of_sys_xcpt A_ (B1_, B2_) empty_heap allocate c =
  Option.the
    (Map.map_of Stringa.equal_literal
      (List.zip initialization_list
        (start_addrs A_ (B1_, B2_) empty_heap allocate))
      c);

end; (*struct StartConfig*)

structure Sum_Type : sig
  datatype ('a, 'b) sum = Inl of 'a | Inr of 'b
end = struct

datatype ('a, 'b) sum = Inl of 'a | Inr of 'b;

end; (*struct Sum_Type*)

structure BinOp : sig
  datatype bop = Eq | NotEq | LessThan | LessOrEqual | GreaterThan |
    GreaterOrEqual | Add | Subtract | Mult | Div | Mod | BinAnd | BinOr | BinXor
    | ShiftLeft | ShiftRightZeros | ShiftRightSigned
  val binop_ShiftRightSigned :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_ShiftRightZeros :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_GreaterOrEqual :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_LessOrEqual :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_GreaterThan :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_ShiftLeft :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_Subtract :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_LessThan :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_BinXor :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_BinAnd :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_BinOr :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_Mult :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val word_smod :
    'a Type_Length.len -> 'a Word.word -> 'a Word.word -> 'a Word.word
  val binop_Mod :
    'a HOL.equal -> 'b HOL.equal * 'b Value.addr ->
      'a -> ('a -> Type.htype -> ('a * 'b) Set.set) ->
              'b Value.vala ->
                'b Value.vala -> ('b Value.vala, 'b) Sum_Type.sum option
  val word_sdiv :
    'a Type_Length.len -> 'a Word.word -> 'a Word.word -> 'a Word.word
  val binop_Div :
    'a HOL.equal -> 'b HOL.equal * 'b Value.addr ->
      'a -> ('a -> Type.htype -> ('a * 'b) Set.set) ->
              'b Value.vala ->
                'b Value.vala -> ('b Value.vala, 'b) Sum_Type.sum option
  val binop_Add :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop :
    'a HOL.equal -> 'b HOL.equal * 'b Value.addr ->
      'a -> ('a -> Type.htype -> ('a * 'b) Set.set) ->
              bop ->
                'b Value.vala ->
                  'b Value.vala -> ('b Value.vala, 'b) Sum_Type.sum option
  val equal_bop : bop -> bop -> bool
  val binop_relevant_class : bop -> 'a TypeRel.prog -> string -> bool
  val wTrt_binop_i_i_i_i_o :
    'a TypeRel.prog -> Type.ty -> bop -> Type.ty -> Type.ty Predicate.pred
end = struct

datatype bop = Eq | NotEq | LessThan | LessOrEqual | GreaterThan |
  GreaterOrEqual | Add | Subtract | Mult | Div | Mod | BinAnd | BinOr | BinXor |
  ShiftLeft | ShiftRightZeros | ShiftRightSigned;

fun binop_ShiftRightSigned (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.sshiftr
             (Type_Length.len_bit0
               (Type_Length.len_bit0
                 (Type_Length.len_bit0
                   (Type_Length.len_bit0
                     (Type_Length.len_bit0 Type_Length.len_num1)))))
             i1 (Word.unat
                  (Type_Length.len0_bit0
                    (Type_Length.len0_bit0
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0 Type_Length.len0_num1)))))
                  (Word.bitAND_word
                    (Type_Length.len0_bit0
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0
                            (Type_Length.len0_bit0 Type_Length.len0_num1)))))
                    i2 (Word.word_of_int
                         (Type_Length.len0_bit0
                           (Type_Length.len0_bit0
                             (Type_Length.len0_bit0
                               (Type_Length.len0_bit0
                                 (Type_Length.len0_bit0
                                   Type_Length.len0_num1)))))
                         (Arith.Int_of_integer (31 : IntInf.int))))))))
  | binop_ShiftRightSigned Value.Unit v2 = NONE
  | binop_ShiftRightSigned Value.Null v2 = NONE
  | binop_ShiftRightSigned (Value.Bool v) v2 = NONE
  | binop_ShiftRightSigned (Value.Addr v) v2 = NONE
  | binop_ShiftRightSigned v1 Value.Unit = NONE
  | binop_ShiftRightSigned v1 Value.Null = NONE
  | binop_ShiftRightSigned v1 (Value.Bool v) = NONE
  | binop_ShiftRightSigned v1 (Value.Addr v) = NONE;

fun binop_ShiftRightZeros (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.shiftr_word
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0 Type_Length.len0_num1)))))
             i1 (Word.unat
                  (Type_Length.len0_bit0
                    (Type_Length.len0_bit0
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0 Type_Length.len0_num1)))))
                  (Word.bitAND_word
                    (Type_Length.len0_bit0
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0
                            (Type_Length.len0_bit0 Type_Length.len0_num1)))))
                    i2 (Word.word_of_int
                         (Type_Length.len0_bit0
                           (Type_Length.len0_bit0
                             (Type_Length.len0_bit0
                               (Type_Length.len0_bit0
                                 (Type_Length.len0_bit0
                                   Type_Length.len0_num1)))))
                         (Arith.Int_of_integer (31 : IntInf.int))))))))
  | binop_ShiftRightZeros Value.Unit v2 = NONE
  | binop_ShiftRightZeros Value.Null v2 = NONE
  | binop_ShiftRightZeros (Value.Bool v) v2 = NONE
  | binop_ShiftRightZeros (Value.Addr v) v2 = NONE
  | binop_ShiftRightZeros v1 Value.Unit = NONE
  | binop_ShiftRightZeros v1 Value.Null = NONE
  | binop_ShiftRightZeros v1 (Value.Bool v) = NONE
  | binop_ShiftRightZeros v1 (Value.Addr v) = NONE;

fun binop_GreaterOrEqual (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Bool
           (Word.word_sle
             (Type_Length.len_bit0
               (Type_Length.len_bit0
                 (Type_Length.len_bit0
                   (Type_Length.len_bit0
                     (Type_Length.len_bit0 Type_Length.len_num1)))))
             i2 i1)))
  | binop_GreaterOrEqual Value.Unit v2 = NONE
  | binop_GreaterOrEqual Value.Null v2 = NONE
  | binop_GreaterOrEqual (Value.Bool v) v2 = NONE
  | binop_GreaterOrEqual (Value.Addr v) v2 = NONE
  | binop_GreaterOrEqual v1 Value.Unit = NONE
  | binop_GreaterOrEqual v1 Value.Null = NONE
  | binop_GreaterOrEqual v1 (Value.Bool v) = NONE
  | binop_GreaterOrEqual v1 (Value.Addr v) = NONE;

fun binop_LessOrEqual (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Bool
           (Word.word_sle
             (Type_Length.len_bit0
               (Type_Length.len_bit0
                 (Type_Length.len_bit0
                   (Type_Length.len_bit0
                     (Type_Length.len_bit0 Type_Length.len_num1)))))
             i1 i2)))
  | binop_LessOrEqual Value.Unit v2 = NONE
  | binop_LessOrEqual Value.Null v2 = NONE
  | binop_LessOrEqual (Value.Bool v) v2 = NONE
  | binop_LessOrEqual (Value.Addr v) v2 = NONE
  | binop_LessOrEqual v1 Value.Unit = NONE
  | binop_LessOrEqual v1 Value.Null = NONE
  | binop_LessOrEqual v1 (Value.Bool v) = NONE
  | binop_LessOrEqual v1 (Value.Addr v) = NONE;

fun binop_GreaterThan (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Bool
           (Word.word_sless
             (Type_Length.len_bit0
               (Type_Length.len_bit0
                 (Type_Length.len_bit0
                   (Type_Length.len_bit0
                     (Type_Length.len_bit0 Type_Length.len_num1)))))
             i2 i1)))
  | binop_GreaterThan Value.Unit v2 = NONE
  | binop_GreaterThan Value.Null v2 = NONE
  | binop_GreaterThan (Value.Bool v) v2 = NONE
  | binop_GreaterThan (Value.Addr v) v2 = NONE
  | binop_GreaterThan v1 Value.Unit = NONE
  | binop_GreaterThan v1 Value.Null = NONE
  | binop_GreaterThan v1 (Value.Bool v) = NONE
  | binop_GreaterThan v1 (Value.Addr v) = NONE;

fun binop_ShiftLeft (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.shiftl_word
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0 Type_Length.len0_num1)))))
             i1 (Word.unat
                  (Type_Length.len0_bit0
                    (Type_Length.len0_bit0
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0 Type_Length.len0_num1)))))
                  (Word.bitAND_word
                    (Type_Length.len0_bit0
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0
                            (Type_Length.len0_bit0 Type_Length.len0_num1)))))
                    i2 (Word.word_of_int
                         (Type_Length.len0_bit0
                           (Type_Length.len0_bit0
                             (Type_Length.len0_bit0
                               (Type_Length.len0_bit0
                                 (Type_Length.len0_bit0
                                   Type_Length.len0_num1)))))
                         (Arith.Int_of_integer (31 : IntInf.int))))))))
  | binop_ShiftLeft Value.Unit v2 = NONE
  | binop_ShiftLeft Value.Null v2 = NONE
  | binop_ShiftLeft (Value.Bool v) v2 = NONE
  | binop_ShiftLeft (Value.Addr v) v2 = NONE
  | binop_ShiftLeft v1 Value.Unit = NONE
  | binop_ShiftLeft v1 Value.Null = NONE
  | binop_ShiftLeft v1 (Value.Bool v) = NONE
  | binop_ShiftLeft v1 (Value.Addr v) = NONE;

fun binop_Subtract (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.minus_word
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0 Type_Length.len0_num1)))))
             i1 i2)))
  | binop_Subtract Value.Unit v2 = NONE
  | binop_Subtract Value.Null v2 = NONE
  | binop_Subtract (Value.Bool v) v2 = NONE
  | binop_Subtract (Value.Addr v) v2 = NONE
  | binop_Subtract v1 Value.Unit = NONE
  | binop_Subtract v1 Value.Null = NONE
  | binop_Subtract v1 (Value.Bool v) = NONE
  | binop_Subtract v1 (Value.Addr v) = NONE;

fun binop_LessThan (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Bool
           (Word.word_sless
             (Type_Length.len_bit0
               (Type_Length.len_bit0
                 (Type_Length.len_bit0
                   (Type_Length.len_bit0
                     (Type_Length.len_bit0 Type_Length.len_num1)))))
             i1 i2)))
  | binop_LessThan Value.Unit v2 = NONE
  | binop_LessThan Value.Null v2 = NONE
  | binop_LessThan (Value.Bool v) v2 = NONE
  | binop_LessThan (Value.Addr v) v2 = NONE
  | binop_LessThan v1 Value.Unit = NONE
  | binop_LessThan v1 Value.Null = NONE
  | binop_LessThan v1 (Value.Bool v) = NONE
  | binop_LessThan v1 (Value.Addr v) = NONE;

fun binop_BinXor (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.bitXOR_word
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0 Type_Length.len0_num1)))))
             i1 i2)))
  | binop_BinXor (Value.Bool b1) (Value.Bool b2) =
    SOME (Sum_Type.Inl (Value.Bool (not (Product_Type.equal_bool b1 b2))))
  | binop_BinXor Value.Unit v2 = NONE
  | binop_BinXor Value.Null v2 = NONE
  | binop_BinXor (Value.Bool v) Value.Unit = NONE
  | binop_BinXor (Value.Bool v) Value.Null = NONE
  | binop_BinXor (Value.Bool v) (Value.Intg va) = NONE
  | binop_BinXor (Value.Bool v) (Value.Addr va) = NONE
  | binop_BinXor (Value.Addr v) v2 = NONE
  | binop_BinXor v1 Value.Unit = NONE
  | binop_BinXor v1 Value.Null = NONE
  | binop_BinXor (Value.Intg va) (Value.Bool v) = NONE
  | binop_BinXor v1 (Value.Addr v) = NONE;

fun binop_BinAnd (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.bitAND_word
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0 Type_Length.len0_num1)))))
             i1 i2)))
  | binop_BinAnd (Value.Bool b1) (Value.Bool b2) =
    SOME (Sum_Type.Inl (Value.Bool (b1 andalso b2)))
  | binop_BinAnd Value.Unit v2 = NONE
  | binop_BinAnd Value.Null v2 = NONE
  | binop_BinAnd (Value.Bool v) Value.Unit = NONE
  | binop_BinAnd (Value.Bool v) Value.Null = NONE
  | binop_BinAnd (Value.Bool v) (Value.Intg va) = NONE
  | binop_BinAnd (Value.Bool v) (Value.Addr va) = NONE
  | binop_BinAnd (Value.Addr v) v2 = NONE
  | binop_BinAnd v1 Value.Unit = NONE
  | binop_BinAnd v1 Value.Null = NONE
  | binop_BinAnd (Value.Intg va) (Value.Bool v) = NONE
  | binop_BinAnd v1 (Value.Addr v) = NONE;

fun binop_BinOr (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.bitOR_word
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0 Type_Length.len0_num1)))))
             i1 i2)))
  | binop_BinOr (Value.Bool b1) (Value.Bool b2) =
    SOME (Sum_Type.Inl (Value.Bool (b1 orelse b2)))
  | binop_BinOr Value.Unit v2 = NONE
  | binop_BinOr Value.Null v2 = NONE
  | binop_BinOr (Value.Bool v) Value.Unit = NONE
  | binop_BinOr (Value.Bool v) Value.Null = NONE
  | binop_BinOr (Value.Bool v) (Value.Intg va) = NONE
  | binop_BinOr (Value.Bool v) (Value.Addr va) = NONE
  | binop_BinOr (Value.Addr v) v2 = NONE
  | binop_BinOr v1 Value.Unit = NONE
  | binop_BinOr v1 Value.Null = NONE
  | binop_BinOr (Value.Intg va) (Value.Bool v) = NONE
  | binop_BinOr v1 (Value.Addr v) = NONE;

fun binop_Mult (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.times_word
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0 Type_Length.len0_num1)))))
             i1 i2)))
  | binop_Mult Value.Unit v2 = NONE
  | binop_Mult Value.Null v2 = NONE
  | binop_Mult (Value.Bool v) v2 = NONE
  | binop_Mult (Value.Addr v) v2 = NONE
  | binop_Mult v1 Value.Unit = NONE
  | binop_Mult v1 Value.Null = NONE
  | binop_Mult v1 (Value.Bool v) = NONE
  | binop_Mult v1 (Value.Addr v) = NONE;

fun word_smod A_ x y =
  let
    val xa = Word.sint A_ x;
    val ya = Word.sint A_ y;
    val negative = Arith.less_int xa Arith.zero_inta;
    val result = Arith.mod_int (Arith.abs_int xa) (Arith.abs_int ya);
  in
    Word.word_of_int (Type_Length.len0_len A_)
      (if negative then Arith.uminus_int result else result)
  end;

fun binop_Mod A_ (B1_, B2_) empty_heap allocate v1 (Value.Addr v) = NONE
  | binop_Mod A_ (B1_, B2_) empty_heap allocate v1 (Value.Bool v) = NONE
  | binop_Mod A_ (B1_, B2_) empty_heap allocate v1 Value.Null = NONE
  | binop_Mod A_ (B1_, B2_) empty_heap allocate v1 Value.Unit = NONE
  | binop_Mod A_ (B1_, B2_) empty_heap allocate (Value.Addr v) v2 = NONE
  | binop_Mod A_ (B1_, B2_) empty_heap allocate (Value.Bool v) v2 = NONE
  | binop_Mod A_ (B1_, B2_) empty_heap allocate Value.Null v2 = NONE
  | binop_Mod A_ (B1_, B2_) empty_heap allocate Value.Unit v2 = NONE
  | binop_Mod A_ (B1_, B2_) empty_heap allocate (Value.Intg i1) (Value.Intg i2)
    = SOME (if Word.equal_word
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0
                       (Type_Length.len0_bit0
                         (Type_Length.len0_bit0 Type_Length.len0_num1)))))
                 i2 (Word.zero_word
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0
                            (Type_Length.len0_bit0
                              (Type_Length.len0_bit0 Type_Length.len0_num1))))))
             then Sum_Type.Inr
                    (StartConfig.addr_of_sys_xcpt A_ (B1_, B2_) empty_heap
                      allocate "java/lang/ArithmeticException")
             else Sum_Type.Inl
                    (Value.Intg
                      (word_smod
                        (Type_Length.len_bit0
                          (Type_Length.len_bit0
                            (Type_Length.len_bit0
                              (Type_Length.len_bit0
                                (Type_Length.len_bit0 Type_Length.len_num1)))))
                        i1 i2)));

fun word_sdiv A_ x y =
  let
    val xa = Word.sint A_ x;
    val ya = Word.sint A_ y;
    val negative =
      not (Product_Type.equal_bool (Arith.less_int xa Arith.zero_inta)
            (Arith.less_int ya Arith.zero_inta));
    val result = Arith.div_int (Arith.abs_int xa) (Arith.abs_int ya);
  in
    Word.word_of_int (Type_Length.len0_len A_)
      (if negative then Arith.uminus_int result else result)
  end;

fun binop_Div A_ (B1_, B2_) empty_heap allocate v1 (Value.Addr v) = NONE
  | binop_Div A_ (B1_, B2_) empty_heap allocate v1 (Value.Bool v) = NONE
  | binop_Div A_ (B1_, B2_) empty_heap allocate v1 Value.Null = NONE
  | binop_Div A_ (B1_, B2_) empty_heap allocate v1 Value.Unit = NONE
  | binop_Div A_ (B1_, B2_) empty_heap allocate (Value.Addr v) v2 = NONE
  | binop_Div A_ (B1_, B2_) empty_heap allocate (Value.Bool v) v2 = NONE
  | binop_Div A_ (B1_, B2_) empty_heap allocate Value.Null v2 = NONE
  | binop_Div A_ (B1_, B2_) empty_heap allocate Value.Unit v2 = NONE
  | binop_Div A_ (B1_, B2_) empty_heap allocate (Value.Intg i1) (Value.Intg i2)
    = SOME (if Word.equal_word
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0
                       (Type_Length.len0_bit0
                         (Type_Length.len0_bit0 Type_Length.len0_num1)))))
                 i2 (Word.zero_word
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0
                            (Type_Length.len0_bit0
                              (Type_Length.len0_bit0 Type_Length.len0_num1))))))
             then Sum_Type.Inr
                    (StartConfig.addr_of_sys_xcpt A_ (B1_, B2_) empty_heap
                      allocate "java/lang/ArithmeticException")
             else Sum_Type.Inl
                    (Value.Intg
                      (word_sdiv
                        (Type_Length.len_bit0
                          (Type_Length.len_bit0
                            (Type_Length.len_bit0
                              (Type_Length.len_bit0
                                (Type_Length.len_bit0 Type_Length.len_num1)))))
                        i1 i2)));

fun binop_Add (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.plus_word
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0 Type_Length.len0_num1)))))
             i1 i2)))
  | binop_Add Value.Unit v2 = NONE
  | binop_Add Value.Null v2 = NONE
  | binop_Add (Value.Bool v) v2 = NONE
  | binop_Add (Value.Addr v) v2 = NONE
  | binop_Add v1 Value.Unit = NONE
  | binop_Add v1 Value.Null = NONE
  | binop_Add v1 (Value.Bool v) = NONE
  | binop_Add v1 (Value.Addr v) = NONE;

fun binop A_ (B1_, B2_) empty_heap allocate ShiftRightSigned a b =
  binop_ShiftRightSigned a b
  | binop A_ (B1_, B2_) empty_heap allocate ShiftRightZeros a b =
    binop_ShiftRightZeros a b
  | binop A_ (B1_, B2_) empty_heap allocate ShiftLeft a b = binop_ShiftLeft a b
  | binop A_ (B1_, B2_) empty_heap allocate BinXor a b = binop_BinXor a b
  | binop A_ (B1_, B2_) empty_heap allocate BinOr a b = binop_BinOr a b
  | binop A_ (B1_, B2_) empty_heap allocate BinAnd a b = binop_BinAnd a b
  | binop A_ (B1_, B2_) empty_heap allocate Div a b =
    binop_Div A_ (B1_, B2_) empty_heap allocate a b
  | binop A_ (B1_, B2_) empty_heap allocate Mod a b =
    binop_Mod A_ (B1_, B2_) empty_heap allocate a b
  | binop A_ (B1_, B2_) empty_heap allocate Mult a b = binop_Mult a b
  | binop A_ (B1_, B2_) empty_heap allocate Subtract a b = binop_Subtract a b
  | binop A_ (B1_, B2_) empty_heap allocate Add a b = binop_Add a b
  | binop A_ (B1_, B2_) empty_heap allocate GreaterOrEqual a b =
    binop_GreaterOrEqual a b
  | binop A_ (B1_, B2_) empty_heap allocate GreaterThan a b =
    binop_GreaterThan a b
  | binop A_ (B1_, B2_) empty_heap allocate LessOrEqual a b =
    binop_LessOrEqual a b
  | binop A_ (B1_, B2_) empty_heap allocate LessThan a b = binop_LessThan a b
  | binop A_ (B1_, B2_) empty_heap allocate NotEq v1 v2 =
    SOME (Sum_Type.Inl (Value.Bool (not (Value.equal_vala B1_ v1 v2))))
  | binop A_ (B1_, B2_) empty_heap allocate Eq v1 v2 =
    SOME (Sum_Type.Inl (Value.Bool (Value.equal_vala B1_ v1 v2)));

fun equal_bop ShiftRightSigned ShiftRightZeros = false
  | equal_bop ShiftRightZeros ShiftRightSigned = false
  | equal_bop ShiftRightSigned ShiftLeft = false
  | equal_bop ShiftLeft ShiftRightSigned = false
  | equal_bop ShiftRightZeros ShiftLeft = false
  | equal_bop ShiftLeft ShiftRightZeros = false
  | equal_bop ShiftRightSigned BinXor = false
  | equal_bop BinXor ShiftRightSigned = false
  | equal_bop ShiftRightZeros BinXor = false
  | equal_bop BinXor ShiftRightZeros = false
  | equal_bop ShiftLeft BinXor = false
  | equal_bop BinXor ShiftLeft = false
  | equal_bop ShiftRightSigned BinOr = false
  | equal_bop BinOr ShiftRightSigned = false
  | equal_bop ShiftRightZeros BinOr = false
  | equal_bop BinOr ShiftRightZeros = false
  | equal_bop ShiftLeft BinOr = false
  | equal_bop BinOr ShiftLeft = false
  | equal_bop BinXor BinOr = false
  | equal_bop BinOr BinXor = false
  | equal_bop ShiftRightSigned BinAnd = false
  | equal_bop BinAnd ShiftRightSigned = false
  | equal_bop ShiftRightZeros BinAnd = false
  | equal_bop BinAnd ShiftRightZeros = false
  | equal_bop ShiftLeft BinAnd = false
  | equal_bop BinAnd ShiftLeft = false
  | equal_bop BinXor BinAnd = false
  | equal_bop BinAnd BinXor = false
  | equal_bop BinOr BinAnd = false
  | equal_bop BinAnd BinOr = false
  | equal_bop ShiftRightSigned Mod = false
  | equal_bop Mod ShiftRightSigned = false
  | equal_bop ShiftRightZeros Mod = false
  | equal_bop Mod ShiftRightZeros = false
  | equal_bop ShiftLeft Mod = false
  | equal_bop Mod ShiftLeft = false
  | equal_bop BinXor Mod = false
  | equal_bop Mod BinXor = false
  | equal_bop BinOr Mod = false
  | equal_bop Mod BinOr = false
  | equal_bop BinAnd Mod = false
  | equal_bop Mod BinAnd = false
  | equal_bop ShiftRightSigned Div = false
  | equal_bop Div ShiftRightSigned = false
  | equal_bop ShiftRightZeros Div = false
  | equal_bop Div ShiftRightZeros = false
  | equal_bop ShiftLeft Div = false
  | equal_bop Div ShiftLeft = false
  | equal_bop BinXor Div = false
  | equal_bop Div BinXor = false
  | equal_bop BinOr Div = false
  | equal_bop Div BinOr = false
  | equal_bop BinAnd Div = false
  | equal_bop Div BinAnd = false
  | equal_bop Mod Div = false
  | equal_bop Div Mod = false
  | equal_bop ShiftRightSigned Mult = false
  | equal_bop Mult ShiftRightSigned = false
  | equal_bop ShiftRightZeros Mult = false
  | equal_bop Mult ShiftRightZeros = false
  | equal_bop ShiftLeft Mult = false
  | equal_bop Mult ShiftLeft = false
  | equal_bop BinXor Mult = false
  | equal_bop Mult BinXor = false
  | equal_bop BinOr Mult = false
  | equal_bop Mult BinOr = false
  | equal_bop BinAnd Mult = false
  | equal_bop Mult BinAnd = false
  | equal_bop Mod Mult = false
  | equal_bop Mult Mod = false
  | equal_bop Div Mult = false
  | equal_bop Mult Div = false
  | equal_bop ShiftRightSigned Subtract = false
  | equal_bop Subtract ShiftRightSigned = false
  | equal_bop ShiftRightZeros Subtract = false
  | equal_bop Subtract ShiftRightZeros = false
  | equal_bop ShiftLeft Subtract = false
  | equal_bop Subtract ShiftLeft = false
  | equal_bop BinXor Subtract = false
  | equal_bop Subtract BinXor = false
  | equal_bop BinOr Subtract = false
  | equal_bop Subtract BinOr = false
  | equal_bop BinAnd Subtract = false
  | equal_bop Subtract BinAnd = false
  | equal_bop Mod Subtract = false
  | equal_bop Subtract Mod = false
  | equal_bop Div Subtract = false
  | equal_bop Subtract Div = false
  | equal_bop Mult Subtract = false
  | equal_bop Subtract Mult = false
  | equal_bop ShiftRightSigned Add = false
  | equal_bop Add ShiftRightSigned = false
  | equal_bop ShiftRightZeros Add = false
  | equal_bop Add ShiftRightZeros = false
  | equal_bop ShiftLeft Add = false
  | equal_bop Add ShiftLeft = false
  | equal_bop BinXor Add = false
  | equal_bop Add BinXor = false
  | equal_bop BinOr Add = false
  | equal_bop Add BinOr = false
  | equal_bop BinAnd Add = false
  | equal_bop Add BinAnd = false
  | equal_bop Mod Add = false
  | equal_bop Add Mod = false
  | equal_bop Div Add = false
  | equal_bop Add Div = false
  | equal_bop Mult Add = false
  | equal_bop Add Mult = false
  | equal_bop Subtract Add = false
  | equal_bop Add Subtract = false
  | equal_bop ShiftRightSigned GreaterOrEqual = false
  | equal_bop GreaterOrEqual ShiftRightSigned = false
  | equal_bop ShiftRightZeros GreaterOrEqual = false
  | equal_bop GreaterOrEqual ShiftRightZeros = false
  | equal_bop ShiftLeft GreaterOrEqual = false
  | equal_bop GreaterOrEqual ShiftLeft = false
  | equal_bop BinXor GreaterOrEqual = false
  | equal_bop GreaterOrEqual BinXor = false
  | equal_bop BinOr GreaterOrEqual = false
  | equal_bop GreaterOrEqual BinOr = false
  | equal_bop BinAnd GreaterOrEqual = false
  | equal_bop GreaterOrEqual BinAnd = false
  | equal_bop Mod GreaterOrEqual = false
  | equal_bop GreaterOrEqual Mod = false
  | equal_bop Div GreaterOrEqual = false
  | equal_bop GreaterOrEqual Div = false
  | equal_bop Mult GreaterOrEqual = false
  | equal_bop GreaterOrEqual Mult = false
  | equal_bop Subtract GreaterOrEqual = false
  | equal_bop GreaterOrEqual Subtract = false
  | equal_bop Add GreaterOrEqual = false
  | equal_bop GreaterOrEqual Add = false
  | equal_bop ShiftRightSigned GreaterThan = false
  | equal_bop GreaterThan ShiftRightSigned = false
  | equal_bop ShiftRightZeros GreaterThan = false
  | equal_bop GreaterThan ShiftRightZeros = false
  | equal_bop ShiftLeft GreaterThan = false
  | equal_bop GreaterThan ShiftLeft = false
  | equal_bop BinXor GreaterThan = false
  | equal_bop GreaterThan BinXor = false
  | equal_bop BinOr GreaterThan = false
  | equal_bop GreaterThan BinOr = false
  | equal_bop BinAnd GreaterThan = false
  | equal_bop GreaterThan BinAnd = false
  | equal_bop Mod GreaterThan = false
  | equal_bop GreaterThan Mod = false
  | equal_bop Div GreaterThan = false
  | equal_bop GreaterThan Div = false
  | equal_bop Mult GreaterThan = false
  | equal_bop GreaterThan Mult = false
  | equal_bop Subtract GreaterThan = false
  | equal_bop GreaterThan Subtract = false
  | equal_bop Add GreaterThan = false
  | equal_bop GreaterThan Add = false
  | equal_bop GreaterOrEqual GreaterThan = false
  | equal_bop GreaterThan GreaterOrEqual = false
  | equal_bop ShiftRightSigned LessOrEqual = false
  | equal_bop LessOrEqual ShiftRightSigned = false
  | equal_bop ShiftRightZeros LessOrEqual = false
  | equal_bop LessOrEqual ShiftRightZeros = false
  | equal_bop ShiftLeft LessOrEqual = false
  | equal_bop LessOrEqual ShiftLeft = false
  | equal_bop BinXor LessOrEqual = false
  | equal_bop LessOrEqual BinXor = false
  | equal_bop BinOr LessOrEqual = false
  | equal_bop LessOrEqual BinOr = false
  | equal_bop BinAnd LessOrEqual = false
  | equal_bop LessOrEqual BinAnd = false
  | equal_bop Mod LessOrEqual = false
  | equal_bop LessOrEqual Mod = false
  | equal_bop Div LessOrEqual = false
  | equal_bop LessOrEqual Div = false
  | equal_bop Mult LessOrEqual = false
  | equal_bop LessOrEqual Mult = false
  | equal_bop Subtract LessOrEqual = false
  | equal_bop LessOrEqual Subtract = false
  | equal_bop Add LessOrEqual = false
  | equal_bop LessOrEqual Add = false
  | equal_bop GreaterOrEqual LessOrEqual = false
  | equal_bop LessOrEqual GreaterOrEqual = false
  | equal_bop GreaterThan LessOrEqual = false
  | equal_bop LessOrEqual GreaterThan = false
  | equal_bop ShiftRightSigned LessThan = false
  | equal_bop LessThan ShiftRightSigned = false
  | equal_bop ShiftRightZeros LessThan = false
  | equal_bop LessThan ShiftRightZeros = false
  | equal_bop ShiftLeft LessThan = false
  | equal_bop LessThan ShiftLeft = false
  | equal_bop BinXor LessThan = false
  | equal_bop LessThan BinXor = false
  | equal_bop BinOr LessThan = false
  | equal_bop LessThan BinOr = false
  | equal_bop BinAnd LessThan = false
  | equal_bop LessThan BinAnd = false
  | equal_bop Mod LessThan = false
  | equal_bop LessThan Mod = false
  | equal_bop Div LessThan = false
  | equal_bop LessThan Div = false
  | equal_bop Mult LessThan = false
  | equal_bop LessThan Mult = false
  | equal_bop Subtract LessThan = false
  | equal_bop LessThan Subtract = false
  | equal_bop Add LessThan = false
  | equal_bop LessThan Add = false
  | equal_bop GreaterOrEqual LessThan = false
  | equal_bop LessThan GreaterOrEqual = false
  | equal_bop GreaterThan LessThan = false
  | equal_bop LessThan GreaterThan = false
  | equal_bop LessOrEqual LessThan = false
  | equal_bop LessThan LessOrEqual = false
  | equal_bop ShiftRightSigned NotEq = false
  | equal_bop NotEq ShiftRightSigned = false
  | equal_bop ShiftRightZeros NotEq = false
  | equal_bop NotEq ShiftRightZeros = false
  | equal_bop ShiftLeft NotEq = false
  | equal_bop NotEq ShiftLeft = false
  | equal_bop BinXor NotEq = false
  | equal_bop NotEq BinXor = false
  | equal_bop BinOr NotEq = false
  | equal_bop NotEq BinOr = false
  | equal_bop BinAnd NotEq = false
  | equal_bop NotEq BinAnd = false
  | equal_bop Mod NotEq = false
  | equal_bop NotEq Mod = false
  | equal_bop Div NotEq = false
  | equal_bop NotEq Div = false
  | equal_bop Mult NotEq = false
  | equal_bop NotEq Mult = false
  | equal_bop Subtract NotEq = false
  | equal_bop NotEq Subtract = false
  | equal_bop Add NotEq = false
  | equal_bop NotEq Add = false
  | equal_bop GreaterOrEqual NotEq = false
  | equal_bop NotEq GreaterOrEqual = false
  | equal_bop GreaterThan NotEq = false
  | equal_bop NotEq GreaterThan = false
  | equal_bop LessOrEqual NotEq = false
  | equal_bop NotEq LessOrEqual = false
  | equal_bop LessThan NotEq = false
  | equal_bop NotEq LessThan = false
  | equal_bop ShiftRightSigned Eq = false
  | equal_bop Eq ShiftRightSigned = false
  | equal_bop ShiftRightZeros Eq = false
  | equal_bop Eq ShiftRightZeros = false
  | equal_bop ShiftLeft Eq = false
  | equal_bop Eq ShiftLeft = false
  | equal_bop BinXor Eq = false
  | equal_bop Eq BinXor = false
  | equal_bop BinOr Eq = false
  | equal_bop Eq BinOr = false
  | equal_bop BinAnd Eq = false
  | equal_bop Eq BinAnd = false
  | equal_bop Mod Eq = false
  | equal_bop Eq Mod = false
  | equal_bop Div Eq = false
  | equal_bop Eq Div = false
  | equal_bop Mult Eq = false
  | equal_bop Eq Mult = false
  | equal_bop Subtract Eq = false
  | equal_bop Eq Subtract = false
  | equal_bop Add Eq = false
  | equal_bop Eq Add = false
  | equal_bop GreaterOrEqual Eq = false
  | equal_bop Eq GreaterOrEqual = false
  | equal_bop GreaterThan Eq = false
  | equal_bop Eq GreaterThan = false
  | equal_bop LessOrEqual Eq = false
  | equal_bop Eq LessOrEqual = false
  | equal_bop LessThan Eq = false
  | equal_bop Eq LessThan = false
  | equal_bop NotEq Eq = false
  | equal_bop Eq NotEq = false
  | equal_bop ShiftRightSigned ShiftRightSigned = true
  | equal_bop ShiftRightZeros ShiftRightZeros = true
  | equal_bop ShiftLeft ShiftLeft = true
  | equal_bop BinXor BinXor = true
  | equal_bop BinOr BinOr = true
  | equal_bop BinAnd BinAnd = true
  | equal_bop Mod Mod = true
  | equal_bop Div Div = true
  | equal_bop Mult Mult = true
  | equal_bop Subtract Subtract = true
  | equal_bop Add Add = true
  | equal_bop GreaterOrEqual GreaterOrEqual = true
  | equal_bop GreaterThan GreaterThan = true
  | equal_bop LessOrEqual LessOrEqual = true
  | equal_bop LessThan LessThan = true
  | equal_bop NotEq NotEq = true
  | equal_bop Eq Eq = true;

fun binop_relevant_class Div =
  (fn p => TypeRel.subcls p "java/lang/ArithmeticException")
  | binop_relevant_class Mod =
    (fn p => TypeRel.subcls p "java/lang/ArithmeticException")
  | binop_relevant_class Eq = (fn _ => fn _ => false)
  | binop_relevant_class NotEq = (fn _ => fn _ => false)
  | binop_relevant_class LessThan = (fn _ => fn _ => false)
  | binop_relevant_class LessOrEqual = (fn _ => fn _ => false)
  | binop_relevant_class GreaterThan = (fn _ => fn _ => false)
  | binop_relevant_class GreaterOrEqual = (fn _ => fn _ => false)
  | binop_relevant_class Add = (fn _ => fn _ => false)
  | binop_relevant_class Subtract = (fn _ => fn _ => false)
  | binop_relevant_class Mult = (fn _ => fn _ => false)
  | binop_relevant_class BinAnd = (fn _ => fn _ => false)
  | binop_relevant_class BinOr = (fn _ => fn _ => false)
  | binop_relevant_class BinXor = (fn _ => fn _ => false)
  | binop_relevant_class ShiftLeft = (fn _ => fn _ => false)
  | binop_relevant_class ShiftRightZeros = (fn _ => fn _ => false)
  | binop_relevant_class ShiftRightSigned = (fn _ => fn _ => false);

fun wTrt_binop_i_i_i_i_o x xa xb xc =
  Predicate.sup_pred
    (case xb of Eq => Predicate.single Type.Boolean
      | NotEq => Predicate.single Type.Boolean | LessThan => Predicate.bot_pred
      | LessOrEqual => Predicate.bot_pred | GreaterThan => Predicate.bot_pred
      | GreaterOrEqual => Predicate.bot_pred | Add => Predicate.bot_pred
      | Subtract => Predicate.bot_pred | Mult => Predicate.bot_pred
      | Div => Predicate.bot_pred | Mod => Predicate.bot_pred
      | BinAnd => Predicate.bot_pred | BinOr => Predicate.bot_pred
      | BinXor => Predicate.bot_pred | ShiftLeft => Predicate.bot_pred
      | ShiftRightZeros => Predicate.bot_pred
      | ShiftRightSigned => Predicate.bot_pred)
    (case xa of Type.Void => Predicate.bot_pred
      | Type.Boolean =>
        (case xb of Eq => Predicate.bot_pred | NotEq => Predicate.bot_pred
          | LessThan => Predicate.bot_pred | LessOrEqual => Predicate.bot_pred
          | GreaterThan => Predicate.bot_pred
          | GreaterOrEqual => Predicate.bot_pred | Add => Predicate.bot_pred
          | Subtract => Predicate.bot_pred | Mult => Predicate.bot_pred
          | Div => Predicate.bot_pred | Mod => Predicate.bot_pred
          | BinAnd =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.single Type.Boolean
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | BinOr =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.single Type.Boolean
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | BinXor =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.single Type.Boolean
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | ShiftLeft => Predicate.bot_pred
          | ShiftRightZeros => Predicate.bot_pred
          | ShiftRightSigned => Predicate.bot_pred)
      | Type.Integer =>
        (case xb of Eq => Predicate.bot_pred | NotEq => Predicate.bot_pred
          | LessThan =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Boolean
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | LessOrEqual =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Boolean
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | GreaterThan =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Boolean
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | GreaterOrEqual =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Boolean
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Add =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Subtract =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Mult =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Div =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Mod =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | BinAnd =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | BinOr =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | BinXor =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | ShiftLeft =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | ShiftRightZeros =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | ShiftRightSigned =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred))
      | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
      | Type.Array _ => Predicate.bot_pred);

end; (*struct BinOp*)

structure Expr : sig
  datatype ('a, 'b, 'c) exp = New of string |
    NewArray of Type.ty * ('a, 'b, 'c) exp | Cast of Type.ty * ('a, 'b, 'c) exp
    | InstanceOf of ('a, 'b, 'c) exp * Type.ty | Val of 'c Value.vala |
    BinOp of ('a, 'b, 'c) exp * BinOp.bop * ('a, 'b, 'c) exp | Var of 'a |
    LAss of 'a * ('a, 'b, 'c) exp | AAcc of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp
    | AAss of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
    ALen of ('a, 'b, 'c) exp | FAcc of ('a, 'b, 'c) exp * string * string |
    FAss of ('a, 'b, 'c) exp * string * string * ('a, 'b, 'c) exp |
    Call of ('a, 'b, 'c) exp * string * ('a, 'b, 'c) exp list |
    Block of 'a * Type.ty * 'c Value.vala option * ('a, 'b, 'c) exp |
    Synchronized of 'b * ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
    InSynchronized of 'b * 'c * ('a, 'b, 'c) exp |
    Seq of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
    Cond of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
    While of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp | Throw of ('a, 'b, 'c) exp |
    TryCatch of ('a, 'b, 'c) exp * string * 'a * ('a, 'b, 'c) exp
end = struct

datatype ('a, 'b, 'c) exp = New of string |
  NewArray of Type.ty * ('a, 'b, 'c) exp | Cast of Type.ty * ('a, 'b, 'c) exp |
  InstanceOf of ('a, 'b, 'c) exp * Type.ty | Val of 'c Value.vala |
  BinOp of ('a, 'b, 'c) exp * BinOp.bop * ('a, 'b, 'c) exp | Var of 'a |
  LAss of 'a * ('a, 'b, 'c) exp | AAcc of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
  AAss of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
  ALen of ('a, 'b, 'c) exp | FAcc of ('a, 'b, 'c) exp * string * string |
  FAss of ('a, 'b, 'c) exp * string * string * ('a, 'b, 'c) exp |
  Call of ('a, 'b, 'c) exp * string * ('a, 'b, 'c) exp list |
  Block of 'a * Type.ty * 'c Value.vala option * ('a, 'b, 'c) exp |
  Synchronized of 'b * ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
  InSynchronized of 'b * 'c * ('a, 'b, 'c) exp |
  Seq of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
  Cond of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
  While of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp | Throw of ('a, 'b, 'c) exp |
  TryCatch of ('a, 'b, 'c) exp * string * 'a * ('a, 'b, 'c) exp;

end; (*struct Expr*)

structure Fifo : sig
  val fifo_empty : unit -> 'a list * 'a list
  val fifo_appendl : 'a -> 'a list * 'a list -> 'a list * 'a list
  val fifo_appendr : 'a -> 'a list * 'a list -> 'a list * 'a list
  val fifo_isEmpty : 'a HOL.equal -> 'a list * 'a list -> bool
  val fifo_removel : 'a list * 'a list -> 'a * ('a list * 'a list)
end = struct

fun fifo_empty x = (fn _ => ([], [])) x;

fun fifo_appendl x f = let
                         val (e, d) = f;
                       in
                         (e, x :: d)
                       end;

fun fifo_appendr a f = (a :: Product_Type.fst f, Product_Type.snd f);

fun fifo_isEmpty A_ f =
  Product_Type.equal_proda (List.equal_list A_) (List.equal_list A_) f ([], []);

fun fifo_removel f =
  (case Product_Type.snd f
    of [] =>
      let
        val rp = List.rev (Product_Type.fst f);
      in
        (List.hd rp, ([], List.tl rp))
      end
    | a :: l => (a, (Product_Type.fst f, l)));

end; (*struct Fifo*)

structure Heap : sig
  datatype addr_loc = CField of string * string | ACell of Arith.nat
  val typeof : 'a Value.vala -> Type.ty option
  val cname_of :
    'b Value.addr -> ('a -> 'b -> Type.htype option) -> 'a -> 'b -> string
  val typeof_h :
    'b Value.addr ->
      ('a -> 'b -> Type.htype option) -> 'a -> 'b Value.vala -> Type.ty option
  val equal_addr_loc : addr_loc -> addr_loc -> bool
end = struct

datatype addr_loc = CField of string * string | ACell of Arith.nat;

fun typeof Value.Unit = SOME Type.Void
  | typeof Value.Null = SOME Type.Nt
  | typeof (Value.Bool b) = SOME Type.Boolean
  | typeof (Value.Intg i) = SOME Type.Integer
  | typeof (Value.Addr a) = NONE;

fun cname_of B_ typeof_addr h a =
  Type.the_Class (Type.ty_of_htype (Option.the (typeof_addr h a)));

fun typeof_h B_ typeof_addr h (Value.Intg va) = typeof (Value.Intg va)
  | typeof_h B_ typeof_addr h (Value.Bool va) = typeof (Value.Bool va)
  | typeof_h B_ typeof_addr h Value.Null = typeof Value.Null
  | typeof_h B_ typeof_addr h Value.Unit = typeof Value.Unit
  | typeof_h B_ typeof_addr h (Value.Addr a) =
    Option.map Type.ty_of_htype (typeof_addr h a);

fun equal_addr_loc (ACell nat) (CField (literal1, literal2)) = false
  | equal_addr_loc (CField (literal1, literal2)) (ACell nat) = false
  | equal_addr_loc (ACell nata) (ACell nat) = Arith.equal_nata nata nat
  | equal_addr_loc (CField (literal1a, literal2a)) (CField (literal1, literal2))
    = ((literal1a : string) = literal1) andalso
        ((literal2a : string) = literal2);

end; (*struct Heap*)

structure Assoc_List : sig
  datatype ('a, 'b) assoc_list = Assoc_List of ('a * 'b) list
  val empty : ('a, 'b) assoc_list
  val impl_of : ('a, 'b) assoc_list -> ('a * 'b) list
  val lookup : 'a HOL.equal -> ('a, 'b) assoc_list -> 'a -> 'b option
  val update_with_aux :
    'b HOL.equal -> 'a -> 'b -> ('a -> 'a) -> ('b * 'a) list -> ('b * 'a) list
  val update_with :
    'a HOL.equal ->
      'b -> 'a -> ('b -> 'b) -> ('a, 'b) assoc_list -> ('a, 'b) assoc_list
  val update :
    'a HOL.equal -> 'a -> 'b -> ('a, 'b) assoc_list -> ('a, 'b) assoc_list
  val equal_assoc_lista :
    'a HOL.equal -> 'b HOL.equal ->
      ('a, 'b) assoc_list -> ('a, 'b) assoc_list -> bool
  val equal_assoc_list :
    'a HOL.equal -> 'b HOL.equal -> ('a, 'b) assoc_list HOL.equal
end = struct

datatype ('a, 'b) assoc_list = Assoc_List of ('a * 'b) list;

val empty : ('a, 'b) assoc_list = Assoc_List [];

fun impl_of (Assoc_List x) = x;

fun lookup A_ al = Map.map_of A_ (impl_of al);

fun update_with_aux B_ v k f [] = [(k, f v)]
  | update_with_aux B_ v k f (p :: ps) =
    (if HOL.eq B_ (Product_Type.fst p) k then (k, f (Product_Type.snd p)) :: ps
      else p :: update_with_aux B_ v k f ps);

fun update_with A_ v k f al =
  Assoc_List (update_with_aux A_ v k f (impl_of al));

fun update A_ k v = update_with A_ v k (fn _ => v);

fun equal_assoc_lista A_ B_ ala al =
  List.equal_lista (Product_Type.equal_prod A_ B_) (impl_of ala) (impl_of al);

fun equal_assoc_list A_ B_ = {equal = equal_assoc_lista A_ B_} :
  ('a, 'b) assoc_list HOL.equal;

end; (*struct Assoc_List*)

structure Foldi : sig
  val foldli : 'a list -> ('b -> bool) -> ('a -> 'b -> 'b) -> 'b -> 'b
end = struct

fun foldli [] c f sigma = sigma
  | foldli (x :: xs) c f sigma =
    (if c sigma then foldli xs c f (f x sigma) else sigma);

end; (*struct Foldi*)

structure Trie_Impl : sig
  datatype ('a, 'b) trie = Trie of 'b option * ('a * ('a, 'b) trie) list
  val empty : ('a, 'b) trie
  val lookup : 'a HOL.equal -> ('a, 'b) trie -> 'a list -> 'b option
  val update : 'a HOL.equal -> ('a, 'b) trie -> 'a list -> 'b -> ('a, 'b) trie
  val iteratei_postfixed :
    'a list ->
      ('a, 'b) trie -> ('c -> bool) -> ('a list * 'b -> 'c -> 'c) -> 'c -> 'c
  val iteratei :
    ('a, 'b) trie -> ('c -> bool) -> ('a list * 'b -> 'c -> 'c) -> 'c -> 'c
  val equal_trie : 'a HOL.equal -> 'b HOL.equal -> ('a, 'b) trie HOL.equal
  val equal_triea :
    'a HOL.equal -> 'b HOL.equal -> ('a, 'b) trie -> ('a, 'b) trie -> bool
end = struct

datatype ('a, 'b) trie = Trie of 'b option * ('a * ('a, 'b) trie) list;

val empty : ('a, 'b) trie = Trie (NONE, []);

fun lookup A_ (Trie (vo, uu)) [] = vo
  | lookup A_ (Trie (uv, ts)) (k :: ks) =
    (case Map.map_of A_ ts k of NONE => NONE | SOME t => lookup A_ t ks);

fun update A_ (Trie (vo, ts)) [] v = Trie (SOME v, ts)
  | update A_ (Trie (vo, ts)) (k :: ks) v =
    Trie (vo, Assoc_List.update_with_aux A_ (Trie (NONE, [])) k
                (fn t => update A_ t ks v) ts);

fun iteratei_postfixed ks (Trie (vo, ts)) c f sigma =
  (if c sigma
    then Foldi.foldli ts c (fn (k, t) => iteratei_postfixed (k :: ks) t c f)
           (case vo of NONE => sigma | SOME v => f (ks, v) sigma)
    else sigma);

fun iteratei t c f sigma = iteratei_postfixed [] t c f sigma;

fun equal_trie A_ B_ = {equal = equal_triea A_ B_} : ('a, 'b) trie HOL.equal
and equal_triea A_ B_ (Trie (optionaa, listaa)) (Trie (optiona, lista)) =
  Option.equal_optiona B_ optionaa optiona andalso
    List.equal_lista (Product_Type.equal_prod A_ (equal_trie A_ B_)) listaa
      lista;

end; (*struct Trie_Impl*)

structure Trie : sig
  datatype ('a, 'b) trie = Trie of ('a, 'b) Trie_Impl.trie
  val empty : ('a, 'b) trie
  val impl_of : ('a, 'b) trie -> ('a, 'b) Trie_Impl.trie
  val lookup : 'a HOL.equal -> ('a, 'b) trie -> 'a list -> 'b option
  val update : 'a HOL.equal -> 'a list -> 'b -> ('a, 'b) trie -> ('a, 'b) trie
  val iteratei :
    ('a, 'b) trie -> ('c -> bool) -> ('a list * 'b -> 'c -> 'c) -> 'c -> 'c
  val equal_trie :
    'a HOL.equal -> 'b HOL.equal -> ('a, 'b) trie -> ('a, 'b) trie -> bool
end = struct

datatype ('a, 'b) trie = Trie of ('a, 'b) Trie_Impl.trie;

val empty : ('a, 'b) trie = Trie Trie_Impl.empty;

fun impl_of (Trie x) = x;

fun lookup A_ t = Trie_Impl.lookup A_ (impl_of t);

fun update A_ ks v t = Trie (Trie_Impl.update A_ (impl_of t) ks v);

fun iteratei t c f =
  Trie_Impl.iteratei (impl_of t) c (fn (ks, v) => f (List.rev ks, v));

fun equal_trie A_ B_ ta t =
  Trie_Impl.equal_triea A_ B_ (impl_of ta) (impl_of t);

end; (*struct Trie*)

structure Listn : sig
  val le : ('a -> 'a -> bool) -> 'a list -> 'a list -> bool
  val coalesce : 'a Err.err list -> ('a list) Err.err
  val map2 : ('a -> 'b -> 'c) -> 'a list -> 'b list -> 'c list
  val sup : ('a -> 'b -> 'c Err.err) -> 'a list -> 'b list -> ('c list) Err.err
end = struct

fun le r = List.list_all2 (fn x => Semilat.lesub x r);

fun coalesce [] = Err.Ok []
  | coalesce (ex :: exs) = Err.sup (fn a => fn b => a :: b) ex (coalesce exs);

fun map2 f = (fn xs => fn ys => List.map (fn (a, b) => f a b) (List.zip xs ys));

fun sup f =
  (fn xs => fn ys =>
    (if Arith.equal_nata (List.size_list xs) (List.size_list ys)
      then coalesce (Semilat.plussub xs (map2 f) ys) else Err.Err));

end; (*struct Listn*)

structure JVMInstructions : sig
  datatype 'a instr = Load of Arith.nat | Store of Arith.nat |
    Push of 'a Value.vala | New of string | NewArray of Type.ty | ALoad | AStore
    | ALength | Getfield of string * string | Putfield of string * string |
    Checkcast of Type.ty | Instanceof of Type.ty | Invoke of string * Arith.nat
    | Return | Pop | Dup | Swap | BinOpInstr of BinOp.bop | Goto of Arith.int |
    IfFalse of Arith.int | ThrowExc | MEnter | MExit
  val equal_instra : 'a HOL.equal -> 'a instr -> 'a instr -> bool
  val equal_instr : 'a HOL.equal -> 'a instr HOL.equal
end = struct

datatype 'a instr = Load of Arith.nat | Store of Arith.nat |
  Push of 'a Value.vala | New of string | NewArray of Type.ty | ALoad | AStore |
  ALength | Getfield of string * string | Putfield of string * string |
  Checkcast of Type.ty | Instanceof of Type.ty | Invoke of string * Arith.nat |
  Return | Pop | Dup | Swap | BinOpInstr of BinOp.bop | Goto of Arith.int |
  IfFalse of Arith.int | ThrowExc | MEnter | MExit;

fun equal_instra A_ MExit MEnter = false
  | equal_instra A_ MEnter MExit = false
  | equal_instra A_ MExit ThrowExc = false
  | equal_instra A_ ThrowExc MExit = false
  | equal_instra A_ MEnter ThrowExc = false
  | equal_instra A_ ThrowExc MEnter = false
  | equal_instra A_ MExit (IfFalse int) = false
  | equal_instra A_ (IfFalse int) MExit = false
  | equal_instra A_ MEnter (IfFalse int) = false
  | equal_instra A_ (IfFalse int) MEnter = false
  | equal_instra A_ ThrowExc (IfFalse int) = false
  | equal_instra A_ (IfFalse int) ThrowExc = false
  | equal_instra A_ MExit (Goto int) = false
  | equal_instra A_ (Goto int) MExit = false
  | equal_instra A_ MEnter (Goto int) = false
  | equal_instra A_ (Goto int) MEnter = false
  | equal_instra A_ ThrowExc (Goto int) = false
  | equal_instra A_ (Goto int) ThrowExc = false
  | equal_instra A_ (IfFalse inta) (Goto int) = false
  | equal_instra A_ (Goto inta) (IfFalse int) = false
  | equal_instra A_ MExit (BinOpInstr bop) = false
  | equal_instra A_ (BinOpInstr bop) MExit = false
  | equal_instra A_ MEnter (BinOpInstr bop) = false
  | equal_instra A_ (BinOpInstr bop) MEnter = false
  | equal_instra A_ ThrowExc (BinOpInstr bop) = false
  | equal_instra A_ (BinOpInstr bop) ThrowExc = false
  | equal_instra A_ (IfFalse int) (BinOpInstr bop) = false
  | equal_instra A_ (BinOpInstr bop) (IfFalse int) = false
  | equal_instra A_ (Goto int) (BinOpInstr bop) = false
  | equal_instra A_ (BinOpInstr bop) (Goto int) = false
  | equal_instra A_ MExit Swap = false
  | equal_instra A_ Swap MExit = false
  | equal_instra A_ MEnter Swap = false
  | equal_instra A_ Swap MEnter = false
  | equal_instra A_ ThrowExc Swap = false
  | equal_instra A_ Swap ThrowExc = false
  | equal_instra A_ (IfFalse int) Swap = false
  | equal_instra A_ Swap (IfFalse int) = false
  | equal_instra A_ (Goto int) Swap = false
  | equal_instra A_ Swap (Goto int) = false
  | equal_instra A_ (BinOpInstr bop) Swap = false
  | equal_instra A_ Swap (BinOpInstr bop) = false
  | equal_instra A_ MExit Dup = false
  | equal_instra A_ Dup MExit = false
  | equal_instra A_ MEnter Dup = false
  | equal_instra A_ Dup MEnter = false
  | equal_instra A_ ThrowExc Dup = false
  | equal_instra A_ Dup ThrowExc = false
  | equal_instra A_ (IfFalse int) Dup = false
  | equal_instra A_ Dup (IfFalse int) = false
  | equal_instra A_ (Goto int) Dup = false
  | equal_instra A_ Dup (Goto int) = false
  | equal_instra A_ (BinOpInstr bop) Dup = false
  | equal_instra A_ Dup (BinOpInstr bop) = false
  | equal_instra A_ Swap Dup = false
  | equal_instra A_ Dup Swap = false
  | equal_instra A_ MExit Pop = false
  | equal_instra A_ Pop MExit = false
  | equal_instra A_ MEnter Pop = false
  | equal_instra A_ Pop MEnter = false
  | equal_instra A_ ThrowExc Pop = false
  | equal_instra A_ Pop ThrowExc = false
  | equal_instra A_ (IfFalse int) Pop = false
  | equal_instra A_ Pop (IfFalse int) = false
  | equal_instra A_ (Goto int) Pop = false
  | equal_instra A_ Pop (Goto int) = false
  | equal_instra A_ (BinOpInstr bop) Pop = false
  | equal_instra A_ Pop (BinOpInstr bop) = false
  | equal_instra A_ Swap Pop = false
  | equal_instra A_ Pop Swap = false
  | equal_instra A_ Dup Pop = false
  | equal_instra A_ Pop Dup = false
  | equal_instra A_ MExit Return = false
  | equal_instra A_ Return MExit = false
  | equal_instra A_ MEnter Return = false
  | equal_instra A_ Return MEnter = false
  | equal_instra A_ ThrowExc Return = false
  | equal_instra A_ Return ThrowExc = false
  | equal_instra A_ (IfFalse int) Return = false
  | equal_instra A_ Return (IfFalse int) = false
  | equal_instra A_ (Goto int) Return = false
  | equal_instra A_ Return (Goto int) = false
  | equal_instra A_ (BinOpInstr bop) Return = false
  | equal_instra A_ Return (BinOpInstr bop) = false
  | equal_instra A_ Swap Return = false
  | equal_instra A_ Return Swap = false
  | equal_instra A_ Dup Return = false
  | equal_instra A_ Return Dup = false
  | equal_instra A_ Pop Return = false
  | equal_instra A_ Return Pop = false
  | equal_instra A_ MExit (Invoke (literal, nat)) = false
  | equal_instra A_ (Invoke (literal, nat)) MExit = false
  | equal_instra A_ MEnter (Invoke (literal, nat)) = false
  | equal_instra A_ (Invoke (literal, nat)) MEnter = false
  | equal_instra A_ ThrowExc (Invoke (literal, nat)) = false
  | equal_instra A_ (Invoke (literal, nat)) ThrowExc = false
  | equal_instra A_ (IfFalse int) (Invoke (literal, nat)) = false
  | equal_instra A_ (Invoke (literal, nat)) (IfFalse int) = false
  | equal_instra A_ (Goto int) (Invoke (literal, nat)) = false
  | equal_instra A_ (Invoke (literal, nat)) (Goto int) = false
  | equal_instra A_ (BinOpInstr bop) (Invoke (literal, nat)) = false
  | equal_instra A_ (Invoke (literal, nat)) (BinOpInstr bop) = false
  | equal_instra A_ Swap (Invoke (literal, nat)) = false
  | equal_instra A_ (Invoke (literal, nat)) Swap = false
  | equal_instra A_ Dup (Invoke (literal, nat)) = false
  | equal_instra A_ (Invoke (literal, nat)) Dup = false
  | equal_instra A_ Pop (Invoke (literal, nat)) = false
  | equal_instra A_ (Invoke (literal, nat)) Pop = false
  | equal_instra A_ Return (Invoke (literal, nat)) = false
  | equal_instra A_ (Invoke (literal, nat)) Return = false
  | equal_instra A_ MExit (Instanceof ty) = false
  | equal_instra A_ (Instanceof ty) MExit = false
  | equal_instra A_ MEnter (Instanceof ty) = false
  | equal_instra A_ (Instanceof ty) MEnter = false
  | equal_instra A_ ThrowExc (Instanceof ty) = false
  | equal_instra A_ (Instanceof ty) ThrowExc = false
  | equal_instra A_ (IfFalse int) (Instanceof ty) = false
  | equal_instra A_ (Instanceof ty) (IfFalse int) = false
  | equal_instra A_ (Goto int) (Instanceof ty) = false
  | equal_instra A_ (Instanceof ty) (Goto int) = false
  | equal_instra A_ (BinOpInstr bop) (Instanceof ty) = false
  | equal_instra A_ (Instanceof ty) (BinOpInstr bop) = false
  | equal_instra A_ Swap (Instanceof ty) = false
  | equal_instra A_ (Instanceof ty) Swap = false
  | equal_instra A_ Dup (Instanceof ty) = false
  | equal_instra A_ (Instanceof ty) Dup = false
  | equal_instra A_ Pop (Instanceof ty) = false
  | equal_instra A_ (Instanceof ty) Pop = false
  | equal_instra A_ Return (Instanceof ty) = false
  | equal_instra A_ (Instanceof ty) Return = false
  | equal_instra A_ (Invoke (literal, nat)) (Instanceof ty) = false
  | equal_instra A_ (Instanceof ty) (Invoke (literal, nat)) = false
  | equal_instra A_ MExit (Checkcast ty) = false
  | equal_instra A_ (Checkcast ty) MExit = false
  | equal_instra A_ MEnter (Checkcast ty) = false
  | equal_instra A_ (Checkcast ty) MEnter = false
  | equal_instra A_ ThrowExc (Checkcast ty) = false
  | equal_instra A_ (Checkcast ty) ThrowExc = false
  | equal_instra A_ (IfFalse int) (Checkcast ty) = false
  | equal_instra A_ (Checkcast ty) (IfFalse int) = false
  | equal_instra A_ (Goto int) (Checkcast ty) = false
  | equal_instra A_ (Checkcast ty) (Goto int) = false
  | equal_instra A_ (BinOpInstr bop) (Checkcast ty) = false
  | equal_instra A_ (Checkcast ty) (BinOpInstr bop) = false
  | equal_instra A_ Swap (Checkcast ty) = false
  | equal_instra A_ (Checkcast ty) Swap = false
  | equal_instra A_ Dup (Checkcast ty) = false
  | equal_instra A_ (Checkcast ty) Dup = false
  | equal_instra A_ Pop (Checkcast ty) = false
  | equal_instra A_ (Checkcast ty) Pop = false
  | equal_instra A_ Return (Checkcast ty) = false
  | equal_instra A_ (Checkcast ty) Return = false
  | equal_instra A_ (Invoke (literal, nat)) (Checkcast ty) = false
  | equal_instra A_ (Checkcast ty) (Invoke (literal, nat)) = false
  | equal_instra A_ (Instanceof tya) (Checkcast ty) = false
  | equal_instra A_ (Checkcast tya) (Instanceof ty) = false
  | equal_instra A_ MExit (Putfield (literal1, literal2)) = false
  | equal_instra A_ (Putfield (literal1, literal2)) MExit = false
  | equal_instra A_ MEnter (Putfield (literal1, literal2)) = false
  | equal_instra A_ (Putfield (literal1, literal2)) MEnter = false
  | equal_instra A_ ThrowExc (Putfield (literal1, literal2)) = false
  | equal_instra A_ (Putfield (literal1, literal2)) ThrowExc = false
  | equal_instra A_ (IfFalse int) (Putfield (literal1, literal2)) = false
  | equal_instra A_ (Putfield (literal1, literal2)) (IfFalse int) = false
  | equal_instra A_ (Goto int) (Putfield (literal1, literal2)) = false
  | equal_instra A_ (Putfield (literal1, literal2)) (Goto int) = false
  | equal_instra A_ (BinOpInstr bop) (Putfield (literal1, literal2)) = false
  | equal_instra A_ (Putfield (literal1, literal2)) (BinOpInstr bop) = false
  | equal_instra A_ Swap (Putfield (literal1, literal2)) = false
  | equal_instra A_ (Putfield (literal1, literal2)) Swap = false
  | equal_instra A_ Dup (Putfield (literal1, literal2)) = false
  | equal_instra A_ (Putfield (literal1, literal2)) Dup = false
  | equal_instra A_ Pop (Putfield (literal1, literal2)) = false
  | equal_instra A_ (Putfield (literal1, literal2)) Pop = false
  | equal_instra A_ Return (Putfield (literal1, literal2)) = false
  | equal_instra A_ (Putfield (literal1, literal2)) Return = false
  | equal_instra A_ (Invoke (literal, nat)) (Putfield (literal1, literal2)) =
    false
  | equal_instra A_ (Putfield (literal1, literal2)) (Invoke (literal, nat)) =
    false
  | equal_instra A_ (Instanceof ty) (Putfield (literal1, literal2)) = false
  | equal_instra A_ (Putfield (literal1, literal2)) (Instanceof ty) = false
  | equal_instra A_ (Checkcast ty) (Putfield (literal1, literal2)) = false
  | equal_instra A_ (Putfield (literal1, literal2)) (Checkcast ty) = false
  | equal_instra A_ MExit (Getfield (literal1, literal2)) = false
  | equal_instra A_ (Getfield (literal1, literal2)) MExit = false
  | equal_instra A_ MEnter (Getfield (literal1, literal2)) = false
  | equal_instra A_ (Getfield (literal1, literal2)) MEnter = false
  | equal_instra A_ ThrowExc (Getfield (literal1, literal2)) = false
  | equal_instra A_ (Getfield (literal1, literal2)) ThrowExc = false
  | equal_instra A_ (IfFalse int) (Getfield (literal1, literal2)) = false
  | equal_instra A_ (Getfield (literal1, literal2)) (IfFalse int) = false
  | equal_instra A_ (Goto int) (Getfield (literal1, literal2)) = false
  | equal_instra A_ (Getfield (literal1, literal2)) (Goto int) = false
  | equal_instra A_ (BinOpInstr bop) (Getfield (literal1, literal2)) = false
  | equal_instra A_ (Getfield (literal1, literal2)) (BinOpInstr bop) = false
  | equal_instra A_ Swap (Getfield (literal1, literal2)) = false
  | equal_instra A_ (Getfield (literal1, literal2)) Swap = false
  | equal_instra A_ Dup (Getfield (literal1, literal2)) = false
  | equal_instra A_ (Getfield (literal1, literal2)) Dup = false
  | equal_instra A_ Pop (Getfield (literal1, literal2)) = false
  | equal_instra A_ (Getfield (literal1, literal2)) Pop = false
  | equal_instra A_ Return (Getfield (literal1, literal2)) = false
  | equal_instra A_ (Getfield (literal1, literal2)) Return = false
  | equal_instra A_ (Invoke (literal, nat)) (Getfield (literal1, literal2)) =
    false
  | equal_instra A_ (Getfield (literal1, literal2)) (Invoke (literal, nat)) =
    false
  | equal_instra A_ (Instanceof ty) (Getfield (literal1, literal2)) = false
  | equal_instra A_ (Getfield (literal1, literal2)) (Instanceof ty) = false
  | equal_instra A_ (Checkcast ty) (Getfield (literal1, literal2)) = false
  | equal_instra A_ (Getfield (literal1, literal2)) (Checkcast ty) = false
  | equal_instra A_ (Putfield (literal1a, literal2a))
    (Getfield (literal1, literal2)) = false
  | equal_instra A_ (Getfield (literal1a, literal2a))
    (Putfield (literal1, literal2)) = false
  | equal_instra A_ MExit ALength = false
  | equal_instra A_ ALength MExit = false
  | equal_instra A_ MEnter ALength = false
  | equal_instra A_ ALength MEnter = false
  | equal_instra A_ ThrowExc ALength = false
  | equal_instra A_ ALength ThrowExc = false
  | equal_instra A_ (IfFalse int) ALength = false
  | equal_instra A_ ALength (IfFalse int) = false
  | equal_instra A_ (Goto int) ALength = false
  | equal_instra A_ ALength (Goto int) = false
  | equal_instra A_ (BinOpInstr bop) ALength = false
  | equal_instra A_ ALength (BinOpInstr bop) = false
  | equal_instra A_ Swap ALength = false
  | equal_instra A_ ALength Swap = false
  | equal_instra A_ Dup ALength = false
  | equal_instra A_ ALength Dup = false
  | equal_instra A_ Pop ALength = false
  | equal_instra A_ ALength Pop = false
  | equal_instra A_ Return ALength = false
  | equal_instra A_ ALength Return = false
  | equal_instra A_ (Invoke (literal, nat)) ALength = false
  | equal_instra A_ ALength (Invoke (literal, nat)) = false
  | equal_instra A_ (Instanceof ty) ALength = false
  | equal_instra A_ ALength (Instanceof ty) = false
  | equal_instra A_ (Checkcast ty) ALength = false
  | equal_instra A_ ALength (Checkcast ty) = false
  | equal_instra A_ (Putfield (literal1, literal2)) ALength = false
  | equal_instra A_ ALength (Putfield (literal1, literal2)) = false
  | equal_instra A_ (Getfield (literal1, literal2)) ALength = false
  | equal_instra A_ ALength (Getfield (literal1, literal2)) = false
  | equal_instra A_ MExit AStore = false
  | equal_instra A_ AStore MExit = false
  | equal_instra A_ MEnter AStore = false
  | equal_instra A_ AStore MEnter = false
  | equal_instra A_ ThrowExc AStore = false
  | equal_instra A_ AStore ThrowExc = false
  | equal_instra A_ (IfFalse int) AStore = false
  | equal_instra A_ AStore (IfFalse int) = false
  | equal_instra A_ (Goto int) AStore = false
  | equal_instra A_ AStore (Goto int) = false
  | equal_instra A_ (BinOpInstr bop) AStore = false
  | equal_instra A_ AStore (BinOpInstr bop) = false
  | equal_instra A_ Swap AStore = false
  | equal_instra A_ AStore Swap = false
  | equal_instra A_ Dup AStore = false
  | equal_instra A_ AStore Dup = false
  | equal_instra A_ Pop AStore = false
  | equal_instra A_ AStore Pop = false
  | equal_instra A_ Return AStore = false
  | equal_instra A_ AStore Return = false
  | equal_instra A_ (Invoke (literal, nat)) AStore = false
  | equal_instra A_ AStore (Invoke (literal, nat)) = false
  | equal_instra A_ (Instanceof ty) AStore = false
  | equal_instra A_ AStore (Instanceof ty) = false
  | equal_instra A_ (Checkcast ty) AStore = false
  | equal_instra A_ AStore (Checkcast ty) = false
  | equal_instra A_ (Putfield (literal1, literal2)) AStore = false
  | equal_instra A_ AStore (Putfield (literal1, literal2)) = false
  | equal_instra A_ (Getfield (literal1, literal2)) AStore = false
  | equal_instra A_ AStore (Getfield (literal1, literal2)) = false
  | equal_instra A_ ALength AStore = false
  | equal_instra A_ AStore ALength = false
  | equal_instra A_ MExit ALoad = false
  | equal_instra A_ ALoad MExit = false
  | equal_instra A_ MEnter ALoad = false
  | equal_instra A_ ALoad MEnter = false
  | equal_instra A_ ThrowExc ALoad = false
  | equal_instra A_ ALoad ThrowExc = false
  | equal_instra A_ (IfFalse int) ALoad = false
  | equal_instra A_ ALoad (IfFalse int) = false
  | equal_instra A_ (Goto int) ALoad = false
  | equal_instra A_ ALoad (Goto int) = false
  | equal_instra A_ (BinOpInstr bop) ALoad = false
  | equal_instra A_ ALoad (BinOpInstr bop) = false
  | equal_instra A_ Swap ALoad = false
  | equal_instra A_ ALoad Swap = false
  | equal_instra A_ Dup ALoad = false
  | equal_instra A_ ALoad Dup = false
  | equal_instra A_ Pop ALoad = false
  | equal_instra A_ ALoad Pop = false
  | equal_instra A_ Return ALoad = false
  | equal_instra A_ ALoad Return = false
  | equal_instra A_ (Invoke (literal, nat)) ALoad = false
  | equal_instra A_ ALoad (Invoke (literal, nat)) = false
  | equal_instra A_ (Instanceof ty) ALoad = false
  | equal_instra A_ ALoad (Instanceof ty) = false
  | equal_instra A_ (Checkcast ty) ALoad = false
  | equal_instra A_ ALoad (Checkcast ty) = false
  | equal_instra A_ (Putfield (literal1, literal2)) ALoad = false
  | equal_instra A_ ALoad (Putfield (literal1, literal2)) = false
  | equal_instra A_ (Getfield (literal1, literal2)) ALoad = false
  | equal_instra A_ ALoad (Getfield (literal1, literal2)) = false
  | equal_instra A_ ALength ALoad = false
  | equal_instra A_ ALoad ALength = false
  | equal_instra A_ AStore ALoad = false
  | equal_instra A_ ALoad AStore = false
  | equal_instra A_ MExit (NewArray ty) = false
  | equal_instra A_ (NewArray ty) MExit = false
  | equal_instra A_ MEnter (NewArray ty) = false
  | equal_instra A_ (NewArray ty) MEnter = false
  | equal_instra A_ ThrowExc (NewArray ty) = false
  | equal_instra A_ (NewArray ty) ThrowExc = false
  | equal_instra A_ (IfFalse int) (NewArray ty) = false
  | equal_instra A_ (NewArray ty) (IfFalse int) = false
  | equal_instra A_ (Goto int) (NewArray ty) = false
  | equal_instra A_ (NewArray ty) (Goto int) = false
  | equal_instra A_ (BinOpInstr bop) (NewArray ty) = false
  | equal_instra A_ (NewArray ty) (BinOpInstr bop) = false
  | equal_instra A_ Swap (NewArray ty) = false
  | equal_instra A_ (NewArray ty) Swap = false
  | equal_instra A_ Dup (NewArray ty) = false
  | equal_instra A_ (NewArray ty) Dup = false
  | equal_instra A_ Pop (NewArray ty) = false
  | equal_instra A_ (NewArray ty) Pop = false
  | equal_instra A_ Return (NewArray ty) = false
  | equal_instra A_ (NewArray ty) Return = false
  | equal_instra A_ (Invoke (literal, nat)) (NewArray ty) = false
  | equal_instra A_ (NewArray ty) (Invoke (literal, nat)) = false
  | equal_instra A_ (Instanceof tya) (NewArray ty) = false
  | equal_instra A_ (NewArray tya) (Instanceof ty) = false
  | equal_instra A_ (Checkcast tya) (NewArray ty) = false
  | equal_instra A_ (NewArray tya) (Checkcast ty) = false
  | equal_instra A_ (Putfield (literal1, literal2)) (NewArray ty) = false
  | equal_instra A_ (NewArray ty) (Putfield (literal1, literal2)) = false
  | equal_instra A_ (Getfield (literal1, literal2)) (NewArray ty) = false
  | equal_instra A_ (NewArray ty) (Getfield (literal1, literal2)) = false
  | equal_instra A_ ALength (NewArray ty) = false
  | equal_instra A_ (NewArray ty) ALength = false
  | equal_instra A_ AStore (NewArray ty) = false
  | equal_instra A_ (NewArray ty) AStore = false
  | equal_instra A_ ALoad (NewArray ty) = false
  | equal_instra A_ (NewArray ty) ALoad = false
  | equal_instra A_ MExit (New literal) = false
  | equal_instra A_ (New literal) MExit = false
  | equal_instra A_ MEnter (New literal) = false
  | equal_instra A_ (New literal) MEnter = false
  | equal_instra A_ ThrowExc (New literal) = false
  | equal_instra A_ (New literal) ThrowExc = false
  | equal_instra A_ (IfFalse int) (New literal) = false
  | equal_instra A_ (New literal) (IfFalse int) = false
  | equal_instra A_ (Goto int) (New literal) = false
  | equal_instra A_ (New literal) (Goto int) = false
  | equal_instra A_ (BinOpInstr bop) (New literal) = false
  | equal_instra A_ (New literal) (BinOpInstr bop) = false
  | equal_instra A_ Swap (New literal) = false
  | equal_instra A_ (New literal) Swap = false
  | equal_instra A_ Dup (New literal) = false
  | equal_instra A_ (New literal) Dup = false
  | equal_instra A_ Pop (New literal) = false
  | equal_instra A_ (New literal) Pop = false
  | equal_instra A_ Return (New literal) = false
  | equal_instra A_ (New literal) Return = false
  | equal_instra A_ (Invoke (literala, nat)) (New literal) = false
  | equal_instra A_ (New literala) (Invoke (literal, nat)) = false
  | equal_instra A_ (Instanceof ty) (New literal) = false
  | equal_instra A_ (New literal) (Instanceof ty) = false
  | equal_instra A_ (Checkcast ty) (New literal) = false
  | equal_instra A_ (New literal) (Checkcast ty) = false
  | equal_instra A_ (Putfield (literal1, literal2)) (New literal) = false
  | equal_instra A_ (New literal) (Putfield (literal1, literal2)) = false
  | equal_instra A_ (Getfield (literal1, literal2)) (New literal) = false
  | equal_instra A_ (New literal) (Getfield (literal1, literal2)) = false
  | equal_instra A_ ALength (New literal) = false
  | equal_instra A_ (New literal) ALength = false
  | equal_instra A_ AStore (New literal) = false
  | equal_instra A_ (New literal) AStore = false
  | equal_instra A_ ALoad (New literal) = false
  | equal_instra A_ (New literal) ALoad = false
  | equal_instra A_ (NewArray ty) (New literal) = false
  | equal_instra A_ (New literal) (NewArray ty) = false
  | equal_instra A_ MExit (Push vala) = false
  | equal_instra A_ (Push vala) MExit = false
  | equal_instra A_ MEnter (Push vala) = false
  | equal_instra A_ (Push vala) MEnter = false
  | equal_instra A_ ThrowExc (Push vala) = false
  | equal_instra A_ (Push vala) ThrowExc = false
  | equal_instra A_ (IfFalse int) (Push vala) = false
  | equal_instra A_ (Push vala) (IfFalse int) = false
  | equal_instra A_ (Goto int) (Push vala) = false
  | equal_instra A_ (Push vala) (Goto int) = false
  | equal_instra A_ (BinOpInstr bop) (Push vala) = false
  | equal_instra A_ (Push vala) (BinOpInstr bop) = false
  | equal_instra A_ Swap (Push vala) = false
  | equal_instra A_ (Push vala) Swap = false
  | equal_instra A_ Dup (Push vala) = false
  | equal_instra A_ (Push vala) Dup = false
  | equal_instra A_ Pop (Push vala) = false
  | equal_instra A_ (Push vala) Pop = false
  | equal_instra A_ Return (Push vala) = false
  | equal_instra A_ (Push vala) Return = false
  | equal_instra A_ (Invoke (literal, nat)) (Push vala) = false
  | equal_instra A_ (Push vala) (Invoke (literal, nat)) = false
  | equal_instra A_ (Instanceof ty) (Push vala) = false
  | equal_instra A_ (Push vala) (Instanceof ty) = false
  | equal_instra A_ (Checkcast ty) (Push vala) = false
  | equal_instra A_ (Push vala) (Checkcast ty) = false
  | equal_instra A_ (Putfield (literal1, literal2)) (Push vala) = false
  | equal_instra A_ (Push vala) (Putfield (literal1, literal2)) = false
  | equal_instra A_ (Getfield (literal1, literal2)) (Push vala) = false
  | equal_instra A_ (Push vala) (Getfield (literal1, literal2)) = false
  | equal_instra A_ ALength (Push vala) = false
  | equal_instra A_ (Push vala) ALength = false
  | equal_instra A_ AStore (Push vala) = false
  | equal_instra A_ (Push vala) AStore = false
  | equal_instra A_ ALoad (Push vala) = false
  | equal_instra A_ (Push vala) ALoad = false
  | equal_instra A_ (NewArray ty) (Push vala) = false
  | equal_instra A_ (Push vala) (NewArray ty) = false
  | equal_instra A_ (New literal) (Push vala) = false
  | equal_instra A_ (Push vala) (New literal) = false
  | equal_instra A_ MExit (Store nat) = false
  | equal_instra A_ (Store nat) MExit = false
  | equal_instra A_ MEnter (Store nat) = false
  | equal_instra A_ (Store nat) MEnter = false
  | equal_instra A_ ThrowExc (Store nat) = false
  | equal_instra A_ (Store nat) ThrowExc = false
  | equal_instra A_ (IfFalse int) (Store nat) = false
  | equal_instra A_ (Store nat) (IfFalse int) = false
  | equal_instra A_ (Goto int) (Store nat) = false
  | equal_instra A_ (Store nat) (Goto int) = false
  | equal_instra A_ (BinOpInstr bop) (Store nat) = false
  | equal_instra A_ (Store nat) (BinOpInstr bop) = false
  | equal_instra A_ Swap (Store nat) = false
  | equal_instra A_ (Store nat) Swap = false
  | equal_instra A_ Dup (Store nat) = false
  | equal_instra A_ (Store nat) Dup = false
  | equal_instra A_ Pop (Store nat) = false
  | equal_instra A_ (Store nat) Pop = false
  | equal_instra A_ Return (Store nat) = false
  | equal_instra A_ (Store nat) Return = false
  | equal_instra A_ (Invoke (literal, nata)) (Store nat) = false
  | equal_instra A_ (Store nata) (Invoke (literal, nat)) = false
  | equal_instra A_ (Instanceof ty) (Store nat) = false
  | equal_instra A_ (Store nat) (Instanceof ty) = false
  | equal_instra A_ (Checkcast ty) (Store nat) = false
  | equal_instra A_ (Store nat) (Checkcast ty) = false
  | equal_instra A_ (Putfield (literal1, literal2)) (Store nat) = false
  | equal_instra A_ (Store nat) (Putfield (literal1, literal2)) = false
  | equal_instra A_ (Getfield (literal1, literal2)) (Store nat) = false
  | equal_instra A_ (Store nat) (Getfield (literal1, literal2)) = false
  | equal_instra A_ ALength (Store nat) = false
  | equal_instra A_ (Store nat) ALength = false
  | equal_instra A_ AStore (Store nat) = false
  | equal_instra A_ (Store nat) AStore = false
  | equal_instra A_ ALoad (Store nat) = false
  | equal_instra A_ (Store nat) ALoad = false
  | equal_instra A_ (NewArray ty) (Store nat) = false
  | equal_instra A_ (Store nat) (NewArray ty) = false
  | equal_instra A_ (New literal) (Store nat) = false
  | equal_instra A_ (Store nat) (New literal) = false
  | equal_instra A_ (Push vala) (Store nat) = false
  | equal_instra A_ (Store nat) (Push vala) = false
  | equal_instra A_ MExit (Load nat) = false
  | equal_instra A_ (Load nat) MExit = false
  | equal_instra A_ MEnter (Load nat) = false
  | equal_instra A_ (Load nat) MEnter = false
  | equal_instra A_ ThrowExc (Load nat) = false
  | equal_instra A_ (Load nat) ThrowExc = false
  | equal_instra A_ (IfFalse int) (Load nat) = false
  | equal_instra A_ (Load nat) (IfFalse int) = false
  | equal_instra A_ (Goto int) (Load nat) = false
  | equal_instra A_ (Load nat) (Goto int) = false
  | equal_instra A_ (BinOpInstr bop) (Load nat) = false
  | equal_instra A_ (Load nat) (BinOpInstr bop) = false
  | equal_instra A_ Swap (Load nat) = false
  | equal_instra A_ (Load nat) Swap = false
  | equal_instra A_ Dup (Load nat) = false
  | equal_instra A_ (Load nat) Dup = false
  | equal_instra A_ Pop (Load nat) = false
  | equal_instra A_ (Load nat) Pop = false
  | equal_instra A_ Return (Load nat) = false
  | equal_instra A_ (Load nat) Return = false
  | equal_instra A_ (Invoke (literal, nata)) (Load nat) = false
  | equal_instra A_ (Load nata) (Invoke (literal, nat)) = false
  | equal_instra A_ (Instanceof ty) (Load nat) = false
  | equal_instra A_ (Load nat) (Instanceof ty) = false
  | equal_instra A_ (Checkcast ty) (Load nat) = false
  | equal_instra A_ (Load nat) (Checkcast ty) = false
  | equal_instra A_ (Putfield (literal1, literal2)) (Load nat) = false
  | equal_instra A_ (Load nat) (Putfield (literal1, literal2)) = false
  | equal_instra A_ (Getfield (literal1, literal2)) (Load nat) = false
  | equal_instra A_ (Load nat) (Getfield (literal1, literal2)) = false
  | equal_instra A_ ALength (Load nat) = false
  | equal_instra A_ (Load nat) ALength = false
  | equal_instra A_ AStore (Load nat) = false
  | equal_instra A_ (Load nat) AStore = false
  | equal_instra A_ ALoad (Load nat) = false
  | equal_instra A_ (Load nat) ALoad = false
  | equal_instra A_ (NewArray ty) (Load nat) = false
  | equal_instra A_ (Load nat) (NewArray ty) = false
  | equal_instra A_ (New literal) (Load nat) = false
  | equal_instra A_ (Load nat) (New literal) = false
  | equal_instra A_ (Push vala) (Load nat) = false
  | equal_instra A_ (Load nat) (Push vala) = false
  | equal_instra A_ (Store nata) (Load nat) = false
  | equal_instra A_ (Load nata) (Store nat) = false
  | equal_instra A_ (IfFalse inta) (IfFalse int) = Arith.equal_int inta int
  | equal_instra A_ (Goto inta) (Goto int) = Arith.equal_int inta int
  | equal_instra A_ (BinOpInstr bopa) (BinOpInstr bop) =
    BinOp.equal_bop bopa bop
  | equal_instra A_ (Invoke (literala, nata)) (Invoke (literal, nat)) =
    ((literala : string) = literal) andalso Arith.equal_nata nata nat
  | equal_instra A_ (Instanceof tya) (Instanceof ty) = Type.equal_tya tya ty
  | equal_instra A_ (Checkcast tya) (Checkcast ty) = Type.equal_tya tya ty
  | equal_instra A_ (Putfield (literal1a, literal2a))
    (Putfield (literal1, literal2)) =
    ((literal1a : string) = literal1) andalso ((literal2a : string) = literal2)
  | equal_instra A_ (Getfield (literal1a, literal2a))
    (Getfield (literal1, literal2)) =
    ((literal1a : string) = literal1) andalso ((literal2a : string) = literal2)
  | equal_instra A_ (NewArray tya) (NewArray ty) = Type.equal_tya tya ty
  | equal_instra A_ (New literala) (New literal) =
    ((literala : string) = literal)
  | equal_instra A_ (Push valaa) (Push vala) = Value.equal_vala A_ valaa vala
  | equal_instra A_ (Store nata) (Store nat) = Arith.equal_nata nata nat
  | equal_instra A_ (Load nata) (Load nat) = Arith.equal_nata nata nat
  | equal_instra A_ MExit MExit = true
  | equal_instra A_ MEnter MEnter = true
  | equal_instra A_ ThrowExc ThrowExc = true
  | equal_instra A_ Swap Swap = true
  | equal_instra A_ Dup Dup = true
  | equal_instra A_ Pop Pop = true
  | equal_instra A_ Return Return = true
  | equal_instra A_ ALength ALength = true
  | equal_instra A_ AStore AStore = true
  | equal_instra A_ ALoad ALoad = true;

fun equal_instr A_ = {equal = equal_instra A_} : 'a instr HOL.equal;

end; (*struct JVMInstructions*)

structure While_Combinator : sig
  val whilea : ('a -> bool) -> ('a -> 'a) -> 'a -> 'a
end = struct

fun whilea b c s = (if b s then whilea b c (c s) else s);

end; (*struct While_Combinator*)

structure SemiType : sig
  val widen1p_i_i_o : 'a TypeRel.prog -> Type.ty -> Type.ty Predicate.pred
  val super : 'a TypeRel.prog -> Type.ty -> Type.ty
  val sup : 'a TypeRel.prog -> Type.ty -> Type.ty -> Type.ty Err.err
end = struct

fun widen1p_i_i_o x xa =
  (case xa of Type.Void => Predicate.bot_pred
    | Type.Boolean => Predicate.bot_pred | Type.Integer => Predicate.bot_pred
    | Type.Nt => Predicate.bot_pred
    | Type.Class c =>
      Predicate.bind (TypeRel.subcls1_i_i_o x c)
        (fn xb => Predicate.single (Type.Class xb))
    | Type.Array t =>
      Predicate.sup_pred
        (case t of Type.Void => Predicate.bot_pred
          | Type.Boolean => Predicate.bot_pred
          | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
          | Type.Class xb =>
            (if ((xb : string) = Type.object)
              then Predicate.single (Type.Class Type.object)
              else Predicate.bot_pred)
          | Type.Array _ => Predicate.bot_pred)
        (Predicate.sup_pred
          (case t of Type.Void => Predicate.bot_pred
            | Type.Boolean => Predicate.bot_pred
            | Type.Integer => Predicate.single (Type.Class Type.object)
            | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
            | Type.Array _ => Predicate.bot_pred)
          (Predicate.sup_pred
            (case t of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.single (Type.Class Type.object)
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
            (Predicate.sup_pred
              (case t of Type.Void => Predicate.single (Type.Class Type.object)
                | Type.Boolean => Predicate.bot_pred
                | Type.Integer => Predicate.bot_pred
                | Type.Nt => Predicate.bot_pred
                | Type.Class _ => Predicate.bot_pred
                | Type.Array _ => Predicate.bot_pred)
              (Predicate.bind
                (Predicate.if_pred
                  (not (Type.equal_tya (Type.ground_type t) Type.Nt)))
                (fn () =>
                  Predicate.bind (widen1p_i_i_o x t)
                    (fn xb => Predicate.single (Type.Array xb))))))));

fun super p (Type.Array Type.Integer) = Type.Class Type.object
  | super p (Type.Array Type.Boolean) = Type.Class Type.object
  | super p (Type.Array Type.Void) = Type.Class Type.object
  | super p (Type.Array (Type.Class c)) =
    (if ((c : string) = Type.object) then Type.Class Type.object
      else Type.Array (super p (Type.Class c)))
  | super p (Type.Array (Type.Array t)) = Type.Array (super p (Type.Array t))
  | super p (Type.Class c) =
    Type.Class (Product_Type.fst (Option.the (TypeRel.class p c)));

fun sup p t u =
  (if Type.is_refT t andalso Type.is_refT u
    then Err.Ok
           (if Type.equal_tya u Type.Nt then t
             else (if Type.equal_tya t Type.Nt then u
                    else While_Combinator.whilea
                           (fn z =>
                             not (Predicate.holds
                                   (Transitive_Closure_Table.rtrancl_tab_FioB_i_i_i
                                     Type.equal_ty (widen1p_i_i_o p) [] t z)))
                           (super p) u))
    else (if Type.equal_tya t u then Err.Ok t else Err.Err));

end; (*struct SemiType*)

structure Product : sig
  val le :
    ('a -> 'a -> bool) -> ('b -> 'b -> bool) -> 'a * 'b -> 'a * 'b -> bool
  val sup :
    ('a -> 'a -> 'a Err.err) ->
      ('b -> 'b -> 'b Err.err) -> 'a * 'b -> 'a * 'b -> ('a * 'b) Err.err
end = struct

fun le r_A r_B =
  (fn (a_1, b_1) => fn (a_2, b_2) =>
    Semilat.lesub a_1 r_A a_2 andalso Semilat.lesub b_1 r_B b_2);

fun sup f g =
  (fn (a_1, b_1) => fn (a_2, b_2) =>
    Err.sup (fn a => fn b => (a, b)) (Semilat.plussub a_1 f a_2)
      (Semilat.plussub b_1 g b_2));

end; (*struct Product*)

structure JVM_SemiType : sig
  val le :
    'a TypeRel.prog ->
      Arith.nat ->
        Arith.nat ->
          ((Type.ty list * Type.ty Err.err list) option) Err.err ->
            ((Type.ty list * Type.ty Err.err list) option) Err.err -> bool
  val sup :
    'a TypeRel.prog ->
      Arith.nat ->
        Arith.nat ->
          ((Type.ty list * Type.ty Err.err list) option) Err.err ->
            ((Type.ty list * Type.ty Err.err list) option) Err.err ->
              ((Type.ty list * Type.ty Err.err list) option) Err.err
end = struct

fun le p m n =
  Err.le
    (Opt.le
      (Product.le (Listn.le (TypeRel.widen p))
        (Listn.le (Err.le (TypeRel.widen p)))));

fun sup s m n =
  Err.lift2
    (Opt.sup
      (Product.sup (Listn.sup (SemiType.sup s))
        (fn x => fn y =>
          Err.Ok (Listn.map2 (Err.lift2 (SemiType.sup s)) x y))));

end; (*struct JVM_SemiType*)

structure Predicate_Compile_Alternative_Defs : sig
  val list_all2_FiiB_i_i :
    ('a -> 'b -> unit Predicate.pred) ->
      'a list -> 'b list -> unit Predicate.pred
end = struct

fun list_all2_FiiB_i_i p xa xb =
  Predicate.sup_pred
    (case xa
      of [] =>
        (case xb of [] => Predicate.single () | _ :: _ => Predicate.bot_pred)
      | _ :: _ => Predicate.bot_pred)
    (case xa of [] => Predicate.bot_pred
      | x :: xs =>
        (case xb of [] => Predicate.bot_pred
          | y :: ys =>
            Predicate.bind (list_all2_FiiB_i_i p xs ys)
              (fn () =>
                Predicate.bind (p x y) (fn () => Predicate.single ()))));

end; (*struct Predicate_Compile_Alternative_Defs*)

structure ExternalCall : sig
  datatype 'a extCallRet = RetVal of 'a Value.vala | RetExc of 'a | RetStaySame
  val equal_extCallReta : 'a HOL.equal -> 'a extCallRet -> 'a extCallRet -> bool
  val equal_extCallRet : 'a HOL.equal -> 'a extCallRet HOL.equal
  val external_WT_defs_i_i_i_i :
    string -> string -> Type.ty list -> Type.ty -> unit Predicate.pred
  val external_WT_defs : string -> string -> Type.ty list -> Type.ty -> bool
end = struct

datatype 'a extCallRet = RetVal of 'a Value.vala | RetExc of 'a | RetStaySame;

fun equal_extCallReta A_ RetStaySame (RetExc addr) = false
  | equal_extCallReta A_ (RetExc addr) RetStaySame = false
  | equal_extCallReta A_ RetStaySame (RetVal vala) = false
  | equal_extCallReta A_ (RetVal vala) RetStaySame = false
  | equal_extCallReta A_ (RetExc addr) (RetVal vala) = false
  | equal_extCallReta A_ (RetVal vala) (RetExc addr) = false
  | equal_extCallReta A_ (RetExc addra) (RetExc addr) = HOL.eq A_ addra addr
  | equal_extCallReta A_ (RetVal valaa) (RetVal vala) =
    Value.equal_vala A_ valaa vala
  | equal_extCallReta A_ RetStaySame RetStaySame = true;

fun equal_extCallRet A_ = {equal = equal_extCallReta A_} :
  'a extCallRet HOL.equal;

fun external_WT_defs_i_i_i_i x xa xb xc =
  Predicate.sup_pred
    (case xb
      of [] =>
        (case xc
          of Type.Void =>
            (if ((xa : string) = "start()V") andalso
                  ((x : string) = Type.thread)
              then Predicate.single () else Predicate.bot_pred)
          | Type.Boolean => Predicate.bot_pred
          | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
          | Type.Class _ => Predicate.bot_pred
          | Type.Array _ => Predicate.bot_pred)
      | _ :: _ => Predicate.bot_pred)
    (Predicate.sup_pred
      (case xb
        of [] =>
          (case xc
            of Type.Void =>
              (if ((xa : string) = "join()V") andalso
                    ((x : string) = Type.thread)
                then Predicate.single () else Predicate.bot_pred)
            | Type.Boolean => Predicate.bot_pred
            | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
            | Type.Class _ => Predicate.bot_pred
            | Type.Array _ => Predicate.bot_pred)
        | _ :: _ => Predicate.bot_pred)
      (Predicate.sup_pred
        (case xb
          of [] =>
            (case xc
              of Type.Void =>
                (if ((xa : string) = "interrupt()V") andalso
                      ((x : string) = Type.thread)
                  then Predicate.single () else Predicate.bot_pred)
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | _ :: _ => Predicate.bot_pred)
        (Predicate.sup_pred
          (case xb
            of [] =>
              (case xc of Type.Void => Predicate.bot_pred
                | Type.Boolean =>
                  (if ((xa : string) = "isInterrupted()Z") andalso
                        ((x : string) = Type.thread)
                    then Predicate.single () else Predicate.bot_pred)
                | Type.Integer => Predicate.bot_pred
                | Type.Nt => Predicate.bot_pred
                | Type.Class _ => Predicate.bot_pred
                | Type.Array _ => Predicate.bot_pred)
            | _ :: _ => Predicate.bot_pred)
          (Predicate.sup_pred
            (case xb
              of [] =>
                (case xc
                  of Type.Void =>
                    (if ((xa : string) = "wait()V") andalso
                          ((x : string) = Type.object)
                      then Predicate.single () else Predicate.bot_pred)
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.bot_pred
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | _ :: _ => Predicate.bot_pred)
            (Predicate.sup_pred
              (case xb
                of [] =>
                  (case xc
                    of Type.Void =>
                      (if ((xa : string) = "notify()V") andalso
                            ((x : string) = Type.object)
                        then Predicate.single () else Predicate.bot_pred)
                    | Type.Boolean => Predicate.bot_pred
                    | Type.Integer => Predicate.bot_pred
                    | Type.Nt => Predicate.bot_pred
                    | Type.Class _ => Predicate.bot_pred
                    | Type.Array _ => Predicate.bot_pred)
                | _ :: _ => Predicate.bot_pred)
              (Predicate.sup_pred
                (case xb
                  of [] =>
                    (case xc
                      of Type.Void =>
                        (if ((xa : string) = "notifyAll()V") andalso
                              ((x : string) = Type.object)
                          then Predicate.single () else Predicate.bot_pred)
                      | Type.Boolean => Predicate.bot_pred
                      | Type.Integer => Predicate.bot_pred
                      | Type.Nt => Predicate.bot_pred
                      | Type.Class _ => Predicate.bot_pred
                      | Type.Array _ => Predicate.bot_pred)
                  | _ :: _ => Predicate.bot_pred)
                (Predicate.sup_pred
                  (case xb
                    of [] =>
                      (case xc of Type.Void => Predicate.bot_pred
                        | Type.Boolean => Predicate.bot_pred
                        | Type.Integer => Predicate.bot_pred
                        | Type.Nt => Predicate.bot_pred
                        | Type.Class xba =>
                          (if ((xba : string) = Type.object) andalso
                                (((xa : string) = "clone()Ljava/lang/Object;") andalso
                                  ((x : string) = Type.object))
                            then Predicate.single () else Predicate.bot_pred)
                        | Type.Array _ => Predicate.bot_pred)
                    | _ :: _ => Predicate.bot_pred)
                  (Predicate.sup_pred
                    (case xb
                      of [] =>
                        (case xc of Type.Void => Predicate.bot_pred
                          | Type.Boolean => Predicate.bot_pred
                          | Type.Integer =>
                            (if ((xa : string) = "hashCode()I") andalso
                                  ((x : string) = Type.object)
                              then Predicate.single () else Predicate.bot_pred)
                          | Type.Nt => Predicate.bot_pred
                          | Type.Class _ => Predicate.bot_pred
                          | Type.Array _ => Predicate.bot_pred)
                      | _ :: _ => Predicate.bot_pred)
                    (Predicate.sup_pred
                      (case xb of [] => Predicate.bot_pred
                        | Type.Void :: lista => Predicate.bot_pred
                        | Type.Boolean :: lista => Predicate.bot_pred
                        | [Type.Integer] =>
                          (case xc
                            of Type.Void =>
                              (if ((xa : string) = "~print(I)V") andalso
                                    ((x : string) = Type.object)
                                then Predicate.single ()
                                else Predicate.bot_pred)
                            | Type.Boolean => Predicate.bot_pred
                            | Type.Integer => Predicate.bot_pred
                            | Type.Nt => Predicate.bot_pred
                            | Type.Class _ => Predicate.bot_pred
                            | Type.Array _ => Predicate.bot_pred)
                        | Type.Integer :: _ :: _ => Predicate.bot_pred
                        | Type.Nt :: lista => Predicate.bot_pred
                        | Type.Class _ :: lista => Predicate.bot_pred
                        | Type.Array _ :: lista => Predicate.bot_pred)
                      (Predicate.sup_pred
                        (case xb
                          of [] =>
                            (case xc of Type.Void => Predicate.bot_pred
                              | Type.Boolean => Predicate.bot_pred
                              | Type.Integer => Predicate.bot_pred
                              | Type.Nt => Predicate.bot_pred
                              | Type.Class xba =>
                                (if ((xba : string) = Type.thread) andalso
                                      (((xa : string) = "~Thread.currentThread()Ljava/lang/Thread;") andalso
((x : string) = Type.object))
                                  then Predicate.single ()
                                  else Predicate.bot_pred)
                              | Type.Array _ => Predicate.bot_pred)
                          | _ :: _ => Predicate.bot_pred)
                        (Predicate.sup_pred
                          (case xb
                            of [] =>
                              (case xc of Type.Void => Predicate.bot_pred
                                | Type.Boolean =>
                                  (if ((xa : string) = "~Thread.interrupted()Z") andalso
((x : string) = Type.object)
                                    then Predicate.single ()
                                    else Predicate.bot_pred)
                                | Type.Integer => Predicate.bot_pred
                                | Type.Nt => Predicate.bot_pred
                                | Type.Class _ => Predicate.bot_pred
                                | Type.Array _ => Predicate.bot_pred)
                            | _ :: _ => Predicate.bot_pred)
                          (case xb
                            of [] =>
                              (case xc
                                of Type.Void =>
                                  (if ((xa : string) = "~Thread.yield()V") andalso
((x : string) = Type.object)
                                    then Predicate.single ()
                                    else Predicate.bot_pred)
                                | Type.Boolean => Predicate.bot_pred
                                | Type.Integer => Predicate.bot_pred
                                | Type.Nt => Predicate.bot_pred
                                | Type.Class _ => Predicate.bot_pred
                                | Type.Array _ => Predicate.bot_pred)
                            | _ :: _ => Predicate.bot_pred))))))))))));

fun external_WT_defs x1 x2 x3 x4 =
  Predicate.holds (external_WT_defs_i_i_i_i x1 x2 x3 x4);

end; (*struct ExternalCall*)

structure Auxiliary : sig
  val distinct_fst : 'a HOL.equal -> ('a * 'b) list -> bool
end = struct

fun distinct_fst A_ = List.distinct A_ o List.map Product_Type.fst;

end; (*struct Auxiliary*)

structure WellForm : sig
  val acyclic_class_hierarchy : 'a TypeRel.prog -> bool
  val wf_overriding_aux_aux_i_i_i_i_i :
    Type.ty ->
      Type.ty ->
        Type.ty list -> Type.ty list -> 'a TypeRel.prog -> unit Predicate.pred
  val wf_overriding_aux_i_i_i_i_i :
    Type.ty ->
      Type.ty list -> string -> string -> 'a TypeRel.prog -> unit Predicate.pred
  val wf_overriding_i_i_i :
    'a TypeRel.prog ->
      string ->
        string * (Type.ty list * (Type.ty * 'a option)) -> unit Predicate.pred
  val wf_overriding :
    'a TypeRel.prog ->
      string -> string * (Type.ty list * (Type.ty * 'a option)) -> bool
  val wf_mdecl :
    ('a TypeRel.prog ->
      string -> string * (Type.ty list * (Type.ty * 'a)) -> bool) ->
      'a TypeRel.prog ->
        string -> string * (Type.ty list * (Type.ty * 'a option)) -> bool
  val wf_fdecl :
    'a TypeRel.prog -> string * (Type.ty * unit TypeRel.fmod_ext) -> bool
  val wf_cdecl :
    ('a TypeRel.prog ->
      string -> string * (Type.ty list * (Type.ty * 'a)) -> bool) ->
      'a TypeRel.prog ->
        string *
          (string *
            ((string * (Type.ty * unit TypeRel.fmod_ext)) list *
              (string * (Type.ty list * (Type.ty * 'a option))) list)) ->
          bool
  val wf_prog :
    ('a TypeRel.prog ->
      string -> string * (Type.ty list * (Type.ty * 'a)) -> bool) ->
      'a TypeRel.prog -> bool
end = struct

fun acyclic_class_hierarchy p =
  List.list_all
    (fn (c, (d, (_, _))) =>
      (if not ((c : string) = Type.object) then not (TypeRel.subcls p d c)
        else true))
    (TypeRel.classes p);

fun wf_overriding_aux_aux_i_i_i_i_i x xa xb xc xd =
  Predicate.sup_pred
    (Predicate.bind
      (Predicate.not_pred
        (Predicate_Compile_Alternative_Defs.list_all2_FiiB_i_i
          (TypeRel.widen_i_i_i xd) xc xb))
      (fn () => Predicate.single ()))
    (Predicate.bind (Predicate.not_pred (TypeRel.widen_i_i_i xd xa x))
      (fn () => Predicate.single ()));

fun wf_overriding_aux_i_i_i_i_i x xa xb xc xd =
  Predicate.bind (TypeRel.method_i_i_i_o_o_o_o xd xc xb)
    (fn (ts, (t, (_, _))) =>
      Predicate.bind (wf_overriding_aux_aux_i_i_i_i_i t x xa ts xd)
        (fn () => Predicate.single ()));

fun wf_overriding_i_i_i x xa (xb, (xc, (xd, xe))) =
  Predicate.bind
    (Predicate.not_pred (wf_overriding_aux_i_i_i_i_i xd xc xb xa x))
    (fn () => Predicate.single ());

fun wf_overriding x1 x2 x3 = Predicate.holds (wf_overriding_i_i_i x1 x2 x3);

fun wf_mdecl wf_md p c =
  (fn (m, (ts, (t, ma))) =>
    List.list_all (TypeRel.is_type p) ts andalso
      (TypeRel.is_type p t andalso
        (case ma of NONE => ExternalCall.external_WT_defs c m ts t
          | SOME mb => wf_md p c (m, (ts, (t, mb))))));

fun wf_fdecl p = (fn (_, (t, _)) => TypeRel.is_type p t);

fun wf_cdecl wf_md p =
  (fn (c, (d, (fs, ms))) =>
    List.list_all (wf_fdecl p) fs andalso
      (Auxiliary.distinct_fst Stringa.equal_literal fs andalso
        (List.list_all (wf_mdecl wf_md p c) ms andalso
          (Auxiliary.distinct_fst Stringa.equal_literal ms andalso
            ((if not ((c : string) = Type.object)
               then TypeRel.is_class p d andalso
                      List.list_all (wf_overriding p d) ms
               else true) andalso
              (if ((c : string) = Type.thread)
                then List.member
                       (Product_Type.equal_prod Stringa.equal_literal
                         (Product_Type.equal_prod
                           (List.equal_list Type.equal_ty) Type.equal_ty))
                       (List.map (fn (m, (ts, (t, _))) => (m, (ts, t))) ms)
                       ("run()V", ([], Type.Void))
                else true))))));

fun wf_prog wf_md p =
  acyclic_class_hierarchy p andalso
    (Exceptions.wf_syscls p andalso
      (Auxiliary.distinct_fst Stringa.equal_literal (TypeRel.classes p) andalso
        List.list_all (wf_cdecl wf_md p) (TypeRel.classes p)));

end; (*struct WellForm*)

structure Typing_Framework : sig
  val stable :
    ('a -> 'a -> bool) ->
      (Arith.nat -> 'a -> (Arith.nat * 'a) list) -> 'a list -> Arith.nat -> bool
end = struct

fun stable r step tau_s p =
  List.list_all (fn (q, tau) => Semilat.lesub tau r (List.nth tau_s q))
    (step p (List.nth tau_s p));

end; (*struct Typing_Framework*)

structure Kildall : sig
  val propa :
    'b HOL.equal ->
      (Arith.nat -> 'a -> 'a) ->
        ('b -> 'b -> 'b) ->
          (Arith.nat * 'b) list -> 'b list -> 'a -> 'b list * 'a
  val iter :
    'b HOL.equal ->
      ('a -> bool) ->
        ('a -> Arith.nat) ->
          (Arith.nat -> 'a -> 'a) ->
            (Arith.nat -> 'a -> 'a) ->
              ('b -> 'b -> 'b) ->
                (Arith.nat -> 'b -> (Arith.nat * 'b) list) ->
                  'b list -> 'a -> 'b list * 'a
  val unstables :
    'a -> (Arith.nat -> 'a -> 'a) ->
            ('b -> 'b -> bool) ->
              (Arith.nat -> 'b -> (Arith.nat * 'b) list) -> 'b list -> 'a
  val kildall :
    'b HOL.equal ->
      'a -> ('a -> bool) ->
              ('a -> Arith.nat) ->
                (Arith.nat -> 'a -> 'a) ->
                  (Arith.nat -> 'a -> 'a) ->
                    ('b -> 'b -> bool) ->
                      ('b -> 'b -> 'b) ->
                        (Arith.nat -> 'b -> (Arith.nat * 'b) list) ->
                          'b list -> 'b list
end = struct

fun propa B_ s_insert f (q :: qs) tau_s w =
  let
    val (qa, tau) = q;
    val u = Semilat.plussub tau f (List.nth tau_s qa);
    val a = (if HOL.eq B_ u (List.nth tau_s qa) then w else s_insert qa w);
  in
    propa B_ s_insert f qs (List.list_update tau_s qa u) a
  end
  | propa B_ s_insert f [] tau_s w = (tau_s, w);

fun iter B_ s_is_empty s_choose s_remove s_insert f step tau_s w =
  While_Combinator.whilea (fn (_, wa) => not (s_is_empty wa))
    (fn (tau_sa, wa) =>
      let
        val p = s_choose wa;
      in
        propa B_ s_insert f (step p (List.nth tau_sa p)) tau_sa (s_remove p wa)
      end)
    (tau_s, w);

fun unstables s_empty s_insert r step tau_s =
  List.foldr s_insert
    (List.filter (fn p => not (Typing_Framework.stable r step tau_s p))
      (List.upt Arith.zero_nat (List.size_list tau_s)))
    s_empty;

fun kildall B_ s_empty s_is_empty s_choose s_remove s_insert r f step tau_s =
  Product_Type.fst
    (iter B_ s_is_empty s_choose s_remove s_insert f step tau_s
      (unstables s_empty s_insert r step tau_s));

end; (*struct Kildall*)

structure Typing_Framework_err : sig
  val error : Arith.nat -> (Arith.nat * 'a Err.err) list
  val map_snd : ('a -> 'b) -> ('c * 'a) list -> ('c * 'b) list
  val err_step :
    Arith.nat ->
      (Arith.nat -> 'a -> bool) ->
        (Arith.nat -> 'a -> (Arith.nat * 'a) list) ->
          Arith.nat -> 'a Err.err -> (Arith.nat * 'a Err.err) list
end = struct

fun error n = List.map (fn x => (x, Err.Err)) (List.upt Arith.zero_nat n);

fun map_snd f = List.map (fn (x, y) => (x, f y));

fun err_step n app step p t =
  (case t of Err.Err => error n
    | Err.Ok tau =>
      (if app p tau then map_snd Err.Ok (step p tau) else error n));

end; (*struct Typing_Framework_err*)

structure Effect : sig
  val is_relevant_class :
    'a JVMInstructions.instr -> 'b TypeRel.prog -> string -> bool
  val is_relevant_entry :
    'a TypeRel.prog ->
      'b JVMInstructions.instr ->
        Arith.nat ->
          Arith.nat * (Arith.nat * (string option * (Arith.nat * Arith.nat))) ->
            bool
  val relevant_entries :
    'a TypeRel.prog ->
      'b JVMInstructions.instr ->
        Arith.nat ->
          (Arith.nat *
            (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list ->
            (Arith.nat *
              (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list
  val xcpt_app :
    'a JVMInstructions.instr ->
      'b TypeRel.prog ->
        Arith.nat ->
          Arith.nat ->
            (Arith.nat *
              (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list ->
              Type.ty list * Type.ty Err.err list -> bool
  val app_i :
    'a JVMInstructions.instr *
      ('b TypeRel.prog *
        (Arith.nat *
          (Arith.nat * (Type.ty * (Type.ty list * Type.ty Err.err list))))) ->
      bool
  val xcpt_eff :
    'a JVMInstructions.instr ->
      'b TypeRel.prog ->
        Arith.nat ->
          Type.ty list * Type.ty Err.err list ->
            (Arith.nat *
              (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list ->
              (Arith.nat * (Type.ty list * Type.ty Err.err list) option) list
  val succs :
    'a JVMInstructions.instr ->
      Type.ty list * Type.ty Err.err list -> Arith.nat -> Arith.nat list
  val eff_i :
    'b HOL.equal ->
      'a JVMInstructions.instr *
        ('b TypeRel.prog * (Type.ty list * Type.ty Err.err list)) ->
        Type.ty list * Type.ty Err.err list
  val norm_eff :
    'b HOL.equal ->
      'a JVMInstructions.instr ->
        'b TypeRel.prog ->
          Arith.nat ->
            Type.ty list * Type.ty Err.err list ->
              (Arith.nat * (Type.ty list * Type.ty Err.err list) option) list
  val eff :
    'b HOL.equal ->
      'a JVMInstructions.instr ->
        'b TypeRel.prog ->
          Arith.nat ->
            (Arith.nat *
              (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list ->
              (Type.ty list * Type.ty Err.err list) option ->
                (Arith.nat * (Type.ty list * Type.ty Err.err list) option) list
  val app :
    'b HOL.equal ->
      'a JVMInstructions.instr ->
        'b TypeRel.prog ->
          Arith.nat ->
            Type.ty ->
              Arith.nat ->
                Arith.nat ->
                  (Arith.nat *
                    (Arith.nat *
                      (string option * (Arith.nat * Arith.nat)))) list ->
                    (Type.ty list * Type.ty Err.err list) option -> bool
end = struct

fun is_relevant_class (JVMInstructions.Getfield (f, d)) =
  (fn p => TypeRel.subcls p "java/lang/NullPointerException")
  | is_relevant_class (JVMInstructions.Putfield (f, d)) =
    (fn p => TypeRel.subcls p "java/lang/NullPointerException")
  | is_relevant_class (JVMInstructions.Checkcast t) =
    (fn p => TypeRel.subcls p "java/lang/ClassCastException")
  | is_relevant_class (JVMInstructions.New d) =
    (fn p => TypeRel.subcls p "java/lang/OutOfMemoryError")
  | is_relevant_class JVMInstructions.ThrowExc = (fn _ => fn _ => true)
  | is_relevant_class (JVMInstructions.Invoke (m, n)) = (fn _ => fn _ => true)
  | is_relevant_class (JVMInstructions.NewArray t) =
    (fn p => fn c =>
      TypeRel.subcls p "java/lang/OutOfMemoryError" c orelse
        TypeRel.subcls p "java/lang/NegativeArraySizeException" c)
  | is_relevant_class JVMInstructions.ALoad =
    (fn p => fn c =>
      TypeRel.subcls p "java/lang/ArrayIndexOutOfBoundsException" c orelse
        TypeRel.subcls p "java/lang/NullPointerException" c)
  | is_relevant_class JVMInstructions.AStore =
    (fn p => fn c =>
      TypeRel.subcls p "java/lang/ArrayIndexOutOfBoundsException" c orelse
        (TypeRel.subcls p "java/lang/ArrayStoreException" c orelse
          TypeRel.subcls p "java/lang/NullPointerException" c))
  | is_relevant_class JVMInstructions.ALength =
    (fn p => TypeRel.subcls p "java/lang/NullPointerException")
  | is_relevant_class JVMInstructions.MEnter =
    (fn p => fn c =>
      TypeRel.subcls p "java/lang/IllegalMonitorStateException" c orelse
        TypeRel.subcls p "java/lang/NullPointerException" c)
  | is_relevant_class JVMInstructions.MExit =
    (fn p => fn c =>
      TypeRel.subcls p "java/lang/IllegalMonitorStateException" c orelse
        TypeRel.subcls p "java/lang/NullPointerException" c)
  | is_relevant_class (JVMInstructions.BinOpInstr bop) =
    BinOp.binop_relevant_class bop
  | is_relevant_class (JVMInstructions.Load v) = (fn _ => fn _ => false)
  | is_relevant_class (JVMInstructions.Store v) = (fn _ => fn _ => false)
  | is_relevant_class (JVMInstructions.Push v) = (fn _ => fn _ => false)
  | is_relevant_class (JVMInstructions.Instanceof v) = (fn _ => fn _ => false)
  | is_relevant_class JVMInstructions.Return = (fn _ => fn _ => false)
  | is_relevant_class JVMInstructions.Pop = (fn _ => fn _ => false)
  | is_relevant_class JVMInstructions.Dup = (fn _ => fn _ => false)
  | is_relevant_class JVMInstructions.Swap = (fn _ => fn _ => false)
  | is_relevant_class (JVMInstructions.Goto v) = (fn _ => fn _ => false)
  | is_relevant_class (JVMInstructions.IfFalse v) = (fn _ => fn _ => false);

fun is_relevant_entry p i pc e =
  let
    val (f, (t, (c, (_, _)))) = e;
  in
    (case c of NONE => true | SOME a => is_relevant_class i p a) andalso
      List.member Arith.equal_nat (List.upt f t) pc
  end;

fun relevant_entries p i pc = List.filter (is_relevant_entry p i pc);

fun xcpt_app i p pc mxs xt tau =
  List.list_all
    (fn (_, (_, (c, (_, d)))) =>
      (case c of NONE => true | SOME a => TypeRel.is_class p a) andalso
        (Arith.less_eq_nat d (List.size_list (Product_Type.fst tau)) andalso
          Arith.less_nat d mxs))
    (relevant_entries p i pc xt);

fun app_i (JVMInstructions.MExit, (p, (pc, (mxs, (t_r, ([], va)))))) = false
  | app_i (JVMInstructions.MEnter, (p, (pc, (mxs, (t_r, ([], va)))))) = false
  | app_i (JVMInstructions.ThrowExc, (p, (pc, (mxs, (t_r, ([], va)))))) = false
  | app_i
    (JVMInstructions.IfFalse v,
      (p, (pc, (mxs, (t_r, (Type.Array ve :: vd, vb))))))
    = false
  | app_i
    (JVMInstructions.IfFalse v,
      (p, (pc, (mxs, (t_r, (Type.Class ve :: vd, vb))))))
    = false
  | app_i
    (JVMInstructions.IfFalse v, (p, (pc, (mxs, (t_r, (Type.Nt :: vd, vb)))))) =
    false
  | app_i
    (JVMInstructions.IfFalse v,
      (p, (pc, (mxs, (t_r, (Type.Integer :: vd, vb))))))
    = false
  | app_i
    (JVMInstructions.IfFalse v, (p, (pc, (mxs, (t_r, (Type.Void :: vd, vb))))))
    = false
  | app_i (JVMInstructions.IfFalse v, (p, (pc, (mxs, (t_r, ([], vb)))))) = false
  | app_i (JVMInstructions.BinOpInstr v, (p, (pc, (mxs, (t_r, ([vc], vb)))))) =
    false
  | app_i (JVMInstructions.BinOpInstr v, (p, (pc, (mxs, (t_r, ([], vb)))))) =
    false
  | app_i (JVMInstructions.Swap, (p, (pc, (mxs, (t_r, ([vb], va)))))) = false
  | app_i (JVMInstructions.Swap, (p, (pc, (mxs, (t_r, ([], va)))))) = false
  | app_i (JVMInstructions.Dup, (p, (pc, (mxs, (t_r, ([], va)))))) = false
  | app_i (JVMInstructions.Pop, (p, (pc, (mxs, (t_r, ([], va)))))) = false
  | app_i (JVMInstructions.Return, (p, (pc, (mxs, (t_r, ([], va)))))) = false
  | app_i (JVMInstructions.Instanceof v, (p, (pc, (mxs, (t_r, ([], vb)))))) =
    false
  | app_i (JVMInstructions.Checkcast v, (p, (pc, (mxs, (t_r, ([], vb)))))) =
    false
  | app_i
    (JVMInstructions.Putfield (v, va), (p, (pc, (mxs, (t_r, ([vd], vc)))))) =
    false
  | app_i (JVMInstructions.Putfield (v, va), (p, (pc, (mxs, (t_r, ([], vc))))))
    = false
  | app_i (JVMInstructions.Getfield (v, va), (p, (pc, (mxs, (t_r, ([], vc))))))
    = false
  | app_i (JVMInstructions.ALength, (p, (pc, (mxs, (t_r, ([], va)))))) = false
  | app_i (JVMInstructions.AStore, (p, (pc, (mxs, (t_r, ([vb, vd], va)))))) =
    false
  | app_i (JVMInstructions.AStore, (p, (pc, (mxs, (t_r, ([vb], va)))))) = false
  | app_i (JVMInstructions.AStore, (p, (pc, (mxs, (t_r, ([], va)))))) = false
  | app_i (JVMInstructions.ALoad, (p, (pc, (mxs, (t_r, ([vb], va)))))) = false
  | app_i (JVMInstructions.ALoad, (p, (pc, (mxs, (t_r, ([], va)))))) = false
  | app_i
    (JVMInstructions.NewArray v,
      (p, (pc, (mxs, (t_r, (Type.Array ve :: vd, vb))))))
    = false
  | app_i
    (JVMInstructions.NewArray v,
      (p, (pc, (mxs, (t_r, (Type.Class ve :: vd, vb))))))
    = false
  | app_i
    (JVMInstructions.NewArray v, (p, (pc, (mxs, (t_r, (Type.Nt :: vd, vb)))))) =
    false
  | app_i
    (JVMInstructions.NewArray v,
      (p, (pc, (mxs, (t_r, (Type.Boolean :: vd, vb))))))
    = false
  | app_i
    (JVMInstructions.NewArray v, (p, (pc, (mxs, (t_r, (Type.Void :: vd, vb))))))
    = false
  | app_i (JVMInstructions.NewArray v, (p, (pc, (mxs, (t_r, ([], vb)))))) =
    false
  | app_i (JVMInstructions.Store v, (p, (pc, (mxs, (t_r, ([], vb)))))) = false
  | app_i (JVMInstructions.MExit, (p, (pc, (mxs, (t_r, (t :: st, lt)))))) =
    Type.is_refT t
  | app_i (JVMInstructions.MEnter, (p, (pc, (mxs, (t_r, (t :: st, lt)))))) =
    Type.is_refT t
  | app_i (JVMInstructions.Invoke (m, n), (p, (pc, (mxs, (t_r, (st, lt)))))) =
    Arith.less_nat n (List.size_list st) andalso
      (if not (Type.equal_tya (List.nth st n) Type.Nt)
        then (case Type.class_type_of (List.nth st n) of NONE => false
               | SOME c =>
                 Predicate.holds
                   (Predicate.bind (TypeRel.method_i_i_i_o_o_o_o p c m)
                     (fn (ts, _) =>
                       (if List.list_all2 (TypeRel.widen p)
                             (List.rev (List.take n st)) ts
                         then Predicate.single () else Predicate.bot_pred))))
        else true)
  | app_i (JVMInstructions.ThrowExc, (p, (pc, (mxs, (t_r, (t :: st, lt)))))) =
    (case t of Type.Void => false | Type.Boolean => false
      | Type.Integer => false | Type.Nt => true
      | Type.Class c => TypeRel.subcls p c Type.throwable
      | Type.Array _ => false)
  | app_i (JVMInstructions.Return, (p, (pc, (mxs, (t_r, (t :: st, lt)))))) =
    TypeRel.widen p t t_r
  | app_i (JVMInstructions.Goto b, (p, (pc, (mxs, (t_r, s))))) =
    Arith.less_eq_int Arith.zero_inta (Arith.plus_int (Arith.int_of_nat pc) b)
  | app_i
    (JVMInstructions.IfFalse b,
      (p, (pc, (mxs, (t_r, (Type.Boolean :: st, lt))))))
    = Arith.less_eq_int Arith.zero_inta (Arith.plus_int (Arith.int_of_nat pc) b)
  | app_i
    (JVMInstructions.BinOpInstr bop,
      (p, (pc, (mxs, (t_r, (t2 :: t1 :: st, lt))))))
    = Predicate.holds
        (Predicate.bind (BinOp.wTrt_binop_i_i_i_i_o p t1 bop t2)
          (fn _ => Predicate.single ()))
  | app_i (JVMInstructions.Swap, (p, (pc, (mxs, (t_r, (t1 :: t2 :: st, lt))))))
    = true
  | app_i (JVMInstructions.Dup, (p, (pc, (mxs, (t_r, (t :: st, lt)))))) =
    Arith.less_nat (Arith.suc (List.size_list st)) mxs
  | app_i (JVMInstructions.Pop, (p, (pc, (mxs, (t_r, (t :: st, lt)))))) = true
  | app_i
    (JVMInstructions.Instanceof ty, (p, (pc, (mxs, (t_r, (t :: st, lt)))))) =
    TypeRel.is_type p ty andalso Type.is_refT t
  | app_i (JVMInstructions.Checkcast ty, (p, (pc, (mxs, (t_r, (t :: st, lt))))))
    = TypeRel.is_type p ty
  | app_i (JVMInstructions.ALength, (p, (pc, (mxs, (t_r, (t1 :: st, lt)))))) =
    (case t1 of Type.Void => false | Type.Boolean => false
      | Type.Integer => false | Type.Nt => true | Type.Class _ => false
      | Type.Array _ => true)
  | app_i
    (JVMInstructions.AStore,
      (p, (pc, (mxs, (t_r, (t1 :: t2 :: t3 :: st, lt))))))
    = Type.equal_tya t2 Type.Integer andalso
        (case t3 of Type.Void => false | Type.Boolean => false
          | Type.Integer => false | Type.Nt => true | Type.Class _ => false
          | Type.Array _ => true)
  | app_i (JVMInstructions.ALoad, (p, (pc, (mxs, (t_r, (t1 :: t2 :: st, lt))))))
    = Type.equal_tya t1 Type.Integer andalso
        (case t2 of Type.Void => false | Type.Boolean => false
          | Type.Integer => false | Type.Nt => true | Type.Class _ => false
          | Type.Array _ => true)
  | app_i
    (JVMInstructions.NewArray ty,
      (p, (pc, (mxs, (t_r, (Type.Integer :: st, lt))))))
    = TypeRel.is_type p (Type.Array ty)
  | app_i (JVMInstructions.New c, (p, (pc, (mxs, (t_r, (st, lt)))))) =
    TypeRel.is_class p c andalso Arith.less_nat (List.size_list st) mxs
  | app_i
    (JVMInstructions.Putfield (f, c),
      (p, (pc, (mxs, (t_r, (t_1 :: t_2 :: st, lt))))))
    = TypeRel.widen p t_2 (Type.Class c) andalso
        Predicate.holds
          (Predicate.bind (TypeRel.sees_field_i_i_i_o_o_i p c f c)
            (fn (t, _) =>
              (if TypeRel.widen p t_1 t then Predicate.single ()
                else Predicate.bot_pred)))
  | app_i
    (JVMInstructions.Getfield (f, c), (p, (pc, (mxs, (t_r, (t :: st, lt)))))) =
    Predicate.holds
      (Predicate.bind (TypeRel.sees_field_i_i_i_o_o_i p c f c)
        (fn _ => Predicate.single ())) andalso
      TypeRel.widen p t (Type.Class c)
  | app_i (JVMInstructions.Push v, (p, (pc, (mxs, (t_r, (st, lt)))))) =
    Arith.less_nat (List.size_list st) mxs andalso
      not (Option.is_none (Heap.typeof v))
  | app_i (JVMInstructions.Store n, (p, (pc, (mxs, (t_r, (t :: st, lt)))))) =
    Arith.less_nat n (List.size_list lt)
  | app_i (JVMInstructions.Load n, (p, (pc, (mxs, (t_r, (st, lt)))))) =
    Arith.less_nat n (List.size_list lt) andalso
      (not (Err.equal_erra Type.equal_ty (List.nth lt n) Err.Err) andalso
        Arith.less_nat (List.size_list st) mxs);

fun xcpt_eff i p pc tau et =
  let
    val (st, lt) = tau;
  in
    List.map
      (fn (_, (_, (c, (h, d)))) =>
        (h, SOME ((case c of NONE => Type.Class Type.throwable
                    | SOME a => Type.Class a) ::
                    List.drop (Arith.minus_nat (List.size_list st) d) st,
                   lt)))
      (relevant_entries p i pc et)
  end;

fun succs (JVMInstructions.Load idx) tau pc = [Arith.plus_nat pc Arith.one_nat]
  | succs (JVMInstructions.Store idx) tau pc = [Arith.plus_nat pc Arith.one_nat]
  | succs (JVMInstructions.Push v) tau pc = [Arith.plus_nat pc Arith.one_nat]
  | succs (JVMInstructions.Getfield (f, c)) tau pc =
    [Arith.plus_nat pc Arith.one_nat]
  | succs (JVMInstructions.Putfield (f, c)) tau pc =
    [Arith.plus_nat pc Arith.one_nat]
  | succs (JVMInstructions.New c) tau pc = [Arith.plus_nat pc Arith.one_nat]
  | succs (JVMInstructions.NewArray t) tau pc =
    [Arith.plus_nat pc Arith.one_nat]
  | succs JVMInstructions.ALoad tau pc =
    (if Type.equal_tya (List.nth (Product_Type.fst tau) Arith.one_nat) Type.Nt
      then [] else [Arith.plus_nat pc Arith.one_nat])
  | succs JVMInstructions.AStore tau pc =
    (if Type.equal_tya
          (List.nth (Product_Type.fst tau)
            (Arith.nat_of_integer (2 : IntInf.int)))
          Type.Nt
      then [] else [Arith.plus_nat pc Arith.one_nat])
  | succs JVMInstructions.ALength tau pc =
    (if Type.equal_tya (List.nth (Product_Type.fst tau) Arith.zero_nat) Type.Nt
      then [] else [Arith.plus_nat pc Arith.one_nat])
  | succs (JVMInstructions.Checkcast c) tau pc =
    [Arith.plus_nat pc Arith.one_nat]
  | succs (JVMInstructions.Instanceof t) tau pc =
    [Arith.plus_nat pc Arith.one_nat]
  | succs JVMInstructions.Pop tau pc = [Arith.plus_nat pc Arith.one_nat]
  | succs JVMInstructions.Dup tau pc = [Arith.plus_nat pc Arith.one_nat]
  | succs JVMInstructions.Swap tau pc = [Arith.plus_nat pc Arith.one_nat]
  | succs (JVMInstructions.BinOpInstr b) tau pc =
    [Arith.plus_nat pc Arith.one_nat]
  | succs (JVMInstructions.IfFalse b) tau pc =
    [Arith.plus_nat pc Arith.one_nat,
      Arith.nat (Arith.plus_int (Arith.int_of_nat pc) b)]
  | succs (JVMInstructions.Goto b) tau pc =
    [Arith.nat (Arith.plus_int (Arith.int_of_nat pc) b)]
  | succs JVMInstructions.Return tau pc = []
  | succs (JVMInstructions.Invoke (m, n)) tau pc =
    (if Type.equal_tya (List.nth (Product_Type.fst tau) n) Type.Nt then []
      else [Arith.plus_nat pc Arith.one_nat])
  | succs JVMInstructions.ThrowExc tau pc = []
  | succs JVMInstructions.MEnter tau pc =
    (if Type.equal_tya (List.nth (Product_Type.fst tau) Arith.zero_nat) Type.Nt
      then [] else [Arith.plus_nat pc Arith.one_nat])
  | succs JVMInstructions.MExit tau pc =
    (if Type.equal_tya (List.nth (Product_Type.fst tau) Arith.zero_nat) Type.Nt
      then [] else [Arith.plus_nat pc Arith.one_nat]);

fun eff_i B_ (JVMInstructions.MExit, (p, (t1 :: st, lt))) = (st, lt)
  | eff_i B_ (JVMInstructions.MEnter, (p, (t1 :: st, lt))) = (st, lt)
  | eff_i B_ (JVMInstructions.Goto n, (p, s)) = s
  | eff_i B_ (JVMInstructions.Invoke (m, n), (p, (st, lt))) =
    let
      val u =
        Product_Type.fst
          (Product_Type.snd
            (Product_Type.snd
              (TypeRel.method B_ p
                (Option.the (Type.class_type_of (List.nth st n))) m)));
    in
      (u :: List.drop (Arith.plus_nat n Arith.one_nat) st, lt)
    end
  | eff_i B_ (JVMInstructions.IfFalse b, (p, (t_1 :: st, lt))) = (st, lt)
  | eff_i B_ (JVMInstructions.BinOpInstr bop, (p, (t2 :: t1 :: st, lt))) =
    (Predicate.the Type.equal_ty (BinOp.wTrt_binop_i_i_i_i_o p t1 bop t2) :: st,
      lt)
  | eff_i B_ (JVMInstructions.Swap, (p, (t1 :: t2 :: st, lt))) =
    (t2 :: t1 :: st, lt)
  | eff_i B_ (JVMInstructions.Dup, (p, (t :: st, lt))) = (t :: t :: st, lt)
  | eff_i B_ (JVMInstructions.Pop, (p, (t :: st, lt))) = (st, lt)
  | eff_i B_ (JVMInstructions.Instanceof ty, (p, (t :: st, lt))) =
    (Type.Boolean :: st, lt)
  | eff_i B_ (JVMInstructions.Checkcast ty, (p, (t :: st, lt))) = (ty :: st, lt)
  | eff_i B_ (JVMInstructions.ALength, (p, (t1 :: st, lt))) =
    (Type.Integer :: st, lt)
  | eff_i B_ (JVMInstructions.AStore, (p, (t1 :: t2 :: t3 :: st, lt))) =
    (st, lt)
  | eff_i B_ (JVMInstructions.ALoad, (p, (t1 :: t2 :: st, lt))) =
    (Type.the_Array t2 :: st, lt)
  | eff_i B_ (JVMInstructions.NewArray ty, (p, (t :: st, lt))) =
    (Type.Array ty :: st, lt)
  | eff_i B_ (JVMInstructions.New c, (p, (st, lt))) = (Type.Class c :: st, lt)
  | eff_i B_ (JVMInstructions.Putfield (f, c), (p, (t_1 :: t_2 :: st, lt))) =
    (st, lt)
  | eff_i B_ (JVMInstructions.Getfield (f, c), (p, (t :: st, lt))) =
    (Product_Type.fst (Product_Type.snd (TypeRel.field p c f)) :: st, lt)
  | eff_i B_ (JVMInstructions.Push v, (p, (st, lt))) =
    (Option.the (Heap.typeof v) :: st, lt)
  | eff_i B_ (JVMInstructions.Store n, (p, (t :: st, lt))) =
    (st, List.list_update lt n (Err.Ok t))
  | eff_i B_ (JVMInstructions.Load n, (p, (st, lt))) =
    (Err.ok_val (List.nth lt n) :: st, lt);

fun norm_eff B_ i p pc tau =
  List.map (fn pca => (pca, SOME (eff_i B_ (i, (p, tau))))) (succs i tau pc);

fun eff B_ i p pc et t =
  (case t of NONE => []
    | SOME tau => norm_eff B_ i p pc tau @ xcpt_eff i p pc tau et);

fun app B_ i p mxs t_r pc mpc xt t =
  (case t of NONE => true
    | SOME tau =>
      app_i (i, (p, (pc, (mxs, (t_r, tau))))) andalso
        (xcpt_app i p pc mxs xt tau andalso
          List.list_all (fn (pca, _) => Arith.less_nat pca mpc)
            (eff B_ i p pc xt t)));

end; (*struct Effect*)

structure TF_JVM : sig
  val exec :
    'a HOL.equal ->
      (Arith.nat *
        (Arith.nat *
          ('a JVMInstructions.instr list *
            (Arith.nat *
              (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list)))
        TypeRel.prog ->
        Arith.nat ->
          Type.ty ->
            (Arith.nat *
              (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list ->
              'a JVMInstructions.instr list ->
                Arith.nat ->
                  ((Type.ty list * Type.ty Err.err list) option) Err.err ->
                    (Arith.nat *
                      ((Type.ty list * Type.ty Err.err list) option)
                        Err.err) list
end = struct

fun exec A_ g maxs rT et bs =
  Typing_Framework_err.err_step (List.size_list bs)
    (fn pc =>
      Effect.app
        (Product_Type.equal_prod Arith.equal_nat
          (Product_Type.equal_prod Arith.equal_nat
            (Product_Type.equal_prod
              (List.equal_list (JVMInstructions.equal_instr A_))
              (List.equal_list
                (Product_Type.equal_prod Arith.equal_nat
                  (Product_Type.equal_prod Arith.equal_nat
                    (Product_Type.equal_prod
                      (Option.equal_option Stringa.equal_literal)
                      (Product_Type.equal_prod Arith.equal_nat
                        Arith.equal_nat))))))))
        (List.nth bs pc) g maxs rT pc (List.size_list bs) et)
    (fn pc =>
      Effect.eff
        (Product_Type.equal_prod Arith.equal_nat
          (Product_Type.equal_prod Arith.equal_nat
            (Product_Type.equal_prod
              (List.equal_list (JVMInstructions.equal_instr A_))
              (List.equal_list
                (Product_Type.equal_prod Arith.equal_nat
                  (Product_Type.equal_prod Arith.equal_nat
                    (Product_Type.equal_prod
                      (Option.equal_option Stringa.equal_literal)
                      (Product_Type.equal_prod Arith.equal_nat
                        Arith.equal_nat))))))))
        (List.nth bs pc) g pc et);

end; (*struct TF_JVM*)

structure BVExec : sig
  val kiljvm :
    'a HOL.equal ->
      (Arith.nat *
        (Arith.nat *
          ('a JVMInstructions.instr list *
            (Arith.nat *
              (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list)))
        TypeRel.prog ->
        Arith.nat ->
          Arith.nat ->
            Type.ty ->
              'a JVMInstructions.instr list ->
                (Arith.nat *
                  (Arith.nat *
                    (string option * (Arith.nat * Arith.nat)))) list ->
                  ((Type.ty list * Type.ty Err.err list) option) Err.err list ->
                    ((Type.ty list * Type.ty Err.err list) option) Err.err list
  val wt_kildall :
    'a HOL.equal ->
      (Arith.nat *
        (Arith.nat *
          ('a JVMInstructions.instr list *
            (Arith.nat *
              (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list)))
        TypeRel.prog ->
        string ->
          Type.ty list ->
            Type.ty ->
              Arith.nat ->
                Arith.nat ->
                  'a JVMInstructions.instr list ->
                    (Arith.nat *
                      (Arith.nat *
                        (string option * (Arith.nat * Arith.nat)))) list ->
                      bool
  val wf_jvm_prog_k :
    'a HOL.equal ->
      (Arith.nat *
        (Arith.nat *
          ('a JVMInstructions.instr list *
            (Arith.nat *
              (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list)))
        TypeRel.prog ->
        bool
end = struct

fun kiljvm A_ p mxs mxl t_r is xt =
  Kildall.kildall
    (Err.equal_err
      (Option.equal_option
        (Product_Type.equal_prod (List.equal_list Type.equal_ty)
          (List.equal_list (Err.equal_err Type.equal_ty)))))
    [] List.null List.hd (List.removeAll Arith.equal_nat)
    (fn a => fn b => a :: b) (JVM_SemiType.le p mxs mxl)
    (JVM_SemiType.sup p mxs mxl) (TF_JVM.exec A_ p mxs t_r xt is);

fun wt_kildall A_ p c ts t_r mxs mxl_0 is xt =
  Arith.less_nat Arith.zero_nat (List.size_list is) andalso
    let
      val first =
        SOME ([], [Err.Ok (Type.Class c)] @
                    List.map Err.Ok ts @ List.replicate mxl_0 Err.Err);
      val start =
        Err.Ok first ::
          List.replicate (Arith.minus_nat (List.size_list is) Arith.one_nat)
            (Err.Ok NONE);
      val result =
        kiljvm A_ p mxs
          (Arith.plus_nat (Arith.plus_nat Arith.one_nat (List.size_list ts))
            mxl_0)
          t_r is xt start;
    in
      List.all_interval_nat
        (fn n =>
          not (Err.equal_erra
                (Option.equal_option
                  (Product_Type.equal_prod (List.equal_list Type.equal_ty)
                    (List.equal_list (Err.equal_err Type.equal_ty))))
                (List.nth result n) Err.Err))
        Arith.zero_nat (List.size_list is)
    end;

fun wf_jvm_prog_k A_ p =
  WellForm.wf_prog
    (fn pa => fn c => fn (_, (ts, (t_r, (mxs, (mxl_0, (a, b)))))) =>
      wt_kildall A_ pa c ts t_r mxs mxl_0 a b)
    p;

end; (*struct BVExec*)

structure FWState : sig
  val shr : 'a * (('b * 'c) * ('d * 'e)) -> 'c
  val thr : 'a * (('b * 'c) * ('d * 'e)) -> 'b
  datatype wake_up_status = WSNotified | WSWokenUp
  datatype 'a wait_set_status = InWS of 'a | PostWS of wake_up_status
  datatype 'a conditional_action = Join of 'a | Yield
  datatype lock_action = Lock | Unlock | UnlockFail | ReleaseAcquire
  val wset : 'a * (('b * 'c) * ('d * 'e)) -> 'd
  val locks : 'a * (('b * 'c) * ('d * 'e)) -> 'a
  datatype ('a, 'b, 'c) new_thread_action = NewThread of 'a * 'b * 'c |
    ThreadExists of 'a * bool
  datatype 'a interrupt_action = IsInterrupted of 'a * bool | Interrupt of 'a |
    ClearInterrupt of 'a
  datatype ('a, 'b) wait_set_action = Suspend of 'b | Notify of 'b |
    NotifyAll of 'b | WakeUp of 'a | Notified | WokenUp
  val obs_a :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      'f list
  val thr_a :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      ('b, 'c, 'd) new_thread_action list
  val cond_a :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      'b conditional_action list
  val wset_a :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      ('b, 'e) wait_set_action list
  val locks_a :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      ('a, (lock_action list)) FinFun.finfun
  datatype ('a, 'b, 'c, 'd, 'e, 'f) thread_action =
    LockAction of (lock_action * 'a) |
    NewThreadAction of ('b, 'c, 'd) new_thread_action |
    ConditionalAction of 'b conditional_action |
    WaitSetAction of ('b, 'e) wait_set_action |
    InterruptAction of 'b interrupt_action | ObsAction of 'f
  val interrupts : 'a * (('b * 'c) * ('d * 'e)) -> 'e
  type 'a obs_action
  val interrupt_a :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      'b interrupt_action list
  val convert_new_thread_action :
    ('a -> 'b) ->
      ('c, 'a, 'd) new_thread_action -> ('c, 'b, 'd) new_thread_action
  val convert_extTA :
    ('a -> 'b) ->
      ('c, (lock_action list)) FinFun.finfun *
        (('d, 'a, 'e) new_thread_action list *
          ('d conditional_action list *
            (('d, 'f) wait_set_action list *
              ('d interrupt_action list * 'g list)))) ->
        ('c, (lock_action list)) FinFun.finfun *
          (('d, 'b, 'e) new_thread_action list *
            ('d conditional_action list *
              (('d, 'f) wait_set_action list *
                ('d interrupt_action list * 'g list))))
  val ta_update_obs :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      'f -> ('a, (lock_action list)) FinFun.finfun *
              (('b, 'c, 'd) new_thread_action list *
                ('b conditional_action list *
                  (('b, 'e) wait_set_action list *
                    ('b interrupt_action list * 'f list))))
  val equal_lock_actiona : lock_action -> lock_action -> bool
  val equal_lock_action : lock_action HOL.equal
  val ta_update_locks :
    'a HOL.equal ->
      ('a, (lock_action list)) FinFun.finfun *
        (('b, 'c, 'd) new_thread_action list *
          ('b conditional_action list *
            (('b, 'e) wait_set_action list *
              ('b interrupt_action list * 'f list)))) ->
        lock_action ->
          'a -> ('a, (lock_action list)) FinFun.finfun *
                  (('b, 'c, 'd) new_thread_action list *
                    ('b conditional_action list *
                      (('b, 'e) wait_set_action list *
                        ('b interrupt_action list * 'f list))))
  val ta_update_wait_set :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      ('b, 'e) wait_set_action ->
        ('a, (lock_action list)) FinFun.finfun *
          (('b, 'c, 'd) new_thread_action list *
            ('b conditional_action list *
              (('b, 'e) wait_set_action list *
                ('b interrupt_action list * 'f list))))
  val ta_update_NewThread :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      ('b, 'c, 'd) new_thread_action ->
        ('a, (lock_action list)) FinFun.finfun *
          (('b, 'c, 'd) new_thread_action list *
            ('b conditional_action list *
              (('b, 'e) wait_set_action list *
                ('b interrupt_action list * 'f list))))
  val ta_update_interrupt :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      'b interrupt_action ->
        ('a, (lock_action list)) FinFun.finfun *
          (('b, 'c, 'd) new_thread_action list *
            ('b conditional_action list *
              (('b, 'e) wait_set_action list *
                ('b interrupt_action list * 'f list))))
  val equal_wait_set_actiona :
    'a HOL.equal -> 'b HOL.equal ->
      ('a, 'b) wait_set_action -> ('a, 'b) wait_set_action -> bool
  val equal_wait_set_action :
    'a HOL.equal -> 'b HOL.equal -> ('a, 'b) wait_set_action HOL.equal
  val equal_wake_up_status : wake_up_status -> wake_up_status -> bool
  val equal_wait_set_statusa :
    'a HOL.equal -> 'a wait_set_status -> 'a wait_set_status -> bool
  val equal_wait_set_status : 'a HOL.equal -> 'a wait_set_status HOL.equal
  val equal_interrupt_actiona :
    'a HOL.equal -> 'a interrupt_action -> 'a interrupt_action -> bool
  val equal_interrupt_action : 'a HOL.equal -> 'a interrupt_action HOL.equal
  val ta_update_Conditional :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      'b conditional_action ->
        ('a, (lock_action list)) FinFun.finfun *
          (('b, 'c, 'd) new_thread_action list *
            ('b conditional_action list *
              (('b, 'e) wait_set_action list *
                ('b interrupt_action list * 'f list))))
  val equal_new_thread_actiona :
    'a HOL.equal -> 'b HOL.equal -> 'c HOL.equal ->
      ('a, 'b, 'c) new_thread_action -> ('a, 'b, 'c) new_thread_action -> bool
  val equal_new_thread_action :
    'a HOL.equal -> 'b HOL.equal -> 'c HOL.equal ->
      ('a, 'b, 'c) new_thread_action HOL.equal
  val equal_conditional_actiona :
    'a HOL.equal -> 'a conditional_action -> 'a conditional_action -> bool
  val equal_conditional_action : 'a HOL.equal -> 'a conditional_action HOL.equal
  val thread_action_to_thread_action :
    'a HOL.equal -> 'f obs_action ->
      ('a, 'b, 'c, 'd, 'e, 'f) thread_action ->
        ('a, (lock_action list)) FinFun.finfun *
          (('b, 'c, 'd) new_thread_action list *
            ('b conditional_action list *
              (('b, 'e) wait_set_action list *
                ('b interrupt_action list * 'f list)))) ->
          ('a, (lock_action list)) FinFun.finfun *
            (('b, 'c, 'd) new_thread_action list *
              ('b conditional_action list *
                (('b, 'e) wait_set_action list *
                  ('b interrupt_action list * 'f list))))
end = struct

fun shr lstsmws =
  Product_Type.snd (Product_Type.fst (Product_Type.snd lstsmws));

fun thr lstsmws =
  Product_Type.fst (Product_Type.fst (Product_Type.snd lstsmws));

datatype wake_up_status = WSNotified | WSWokenUp;

datatype 'a wait_set_status = InWS of 'a | PostWS of wake_up_status;

datatype 'a conditional_action = Join of 'a | Yield;

datatype lock_action = Lock | Unlock | UnlockFail | ReleaseAcquire;

fun wset lstsmws =
  Product_Type.fst (Product_Type.snd (Product_Type.snd lstsmws));

fun locks lstsmws = Product_Type.fst lstsmws;

datatype ('a, 'b, 'c) new_thread_action = NewThread of 'a * 'b * 'c |
  ThreadExists of 'a * bool;

datatype 'a interrupt_action = IsInterrupted of 'a * bool | Interrupt of 'a |
  ClearInterrupt of 'a;

datatype ('a, 'b) wait_set_action = Suspend of 'b | Notify of 'b |
  NotifyAll of 'b | WakeUp of 'a | Notified | WokenUp;

fun obs_a x =
  (Product_Type.snd o Product_Type.snd o Product_Type.snd o Product_Type.snd o
    Product_Type.snd)
    x;

fun thr_a x = (Product_Type.fst o Product_Type.snd) x;

fun cond_a x = (Product_Type.fst o Product_Type.snd o Product_Type.snd) x;

fun wset_a x =
  (Product_Type.fst o Product_Type.snd o Product_Type.snd o Product_Type.snd) x;

fun locks_a x = Product_Type.fst x;

datatype ('a, 'b, 'c, 'd, 'e, 'f) thread_action =
  LockAction of (lock_action * 'a) |
  NewThreadAction of ('b, 'c, 'd) new_thread_action |
  ConditionalAction of 'b conditional_action |
  WaitSetAction of ('b, 'e) wait_set_action |
  InterruptAction of 'b interrupt_action | ObsAction of 'f;

fun interrupts lstsmws =
  Product_Type.snd (Product_Type.snd (Product_Type.snd lstsmws));

type 'a obs_action = {};

fun interrupt_a x =
  (Product_Type.fst o Product_Type.snd o Product_Type.snd o Product_Type.snd o
    Product_Type.snd)
    x;

fun convert_new_thread_action f (NewThread (t, x, m)) = NewThread (t, f x, m)
  | convert_new_thread_action f (ThreadExists (t, b)) = ThreadExists (t, b);

fun convert_extTA f ta =
  (locks_a ta,
    (List.map (convert_new_thread_action f) (thr_a ta),
      Product_Type.snd (Product_Type.snd ta)));

fun ta_update_obs (ls, (nts, (js, (wss, (is, obs))))) ob =
  (ls, (nts, (js, (wss, (is, obs @ [ob])))));

fun equal_lock_actiona ReleaseAcquire UnlockFail = false
  | equal_lock_actiona UnlockFail ReleaseAcquire = false
  | equal_lock_actiona ReleaseAcquire Unlock = false
  | equal_lock_actiona Unlock ReleaseAcquire = false
  | equal_lock_actiona UnlockFail Unlock = false
  | equal_lock_actiona Unlock UnlockFail = false
  | equal_lock_actiona ReleaseAcquire Lock = false
  | equal_lock_actiona Lock ReleaseAcquire = false
  | equal_lock_actiona UnlockFail Lock = false
  | equal_lock_actiona Lock UnlockFail = false
  | equal_lock_actiona Unlock Lock = false
  | equal_lock_actiona Lock Unlock = false
  | equal_lock_actiona ReleaseAcquire ReleaseAcquire = true
  | equal_lock_actiona UnlockFail UnlockFail = true
  | equal_lock_actiona Unlock Unlock = true
  | equal_lock_actiona Lock Lock = true;

val equal_lock_action = {equal = equal_lock_actiona} : lock_action HOL.equal;

fun ta_update_locks A_ (ls, (nts, (js, (wss, obs)))) lta l =
  (FinFun.finfun_update A_ (List.equal_list equal_lock_action) ls l
     (FinFun.finfun_apply A_ ls l @ [lta]),
    (nts, (js, (wss, obs))));

fun ta_update_wait_set (ls, (nts, (js, (wss, (is, obs))))) ws =
  (ls, (nts, (js, (wss @ [ws], (is, obs)))));

fun ta_update_NewThread (ls, (nts, (js, (wss, (is, obs))))) nt =
  (ls, (nts @ [nt], (js, (wss, (is, obs)))));

fun ta_update_interrupt (ls, (nts, (js, (wss, (is, obs))))) i =
  (ls, (nts, (js, (wss, (is @ [i], obs)))));

fun equal_wait_set_actiona A_ B_ WokenUp Notified = false
  | equal_wait_set_actiona A_ B_ Notified WokenUp = false
  | equal_wait_set_actiona A_ B_ WokenUp (WakeUp t) = false
  | equal_wait_set_actiona A_ B_ (WakeUp t) WokenUp = false
  | equal_wait_set_actiona A_ B_ Notified (WakeUp t) = false
  | equal_wait_set_actiona A_ B_ (WakeUp t) Notified = false
  | equal_wait_set_actiona A_ B_ WokenUp (NotifyAll w) = false
  | equal_wait_set_actiona A_ B_ (NotifyAll w) WokenUp = false
  | equal_wait_set_actiona A_ B_ Notified (NotifyAll w) = false
  | equal_wait_set_actiona A_ B_ (NotifyAll w) Notified = false
  | equal_wait_set_actiona A_ B_ (WakeUp t) (NotifyAll w) = false
  | equal_wait_set_actiona A_ B_ (NotifyAll w) (WakeUp t) = false
  | equal_wait_set_actiona A_ B_ WokenUp (Notify w) = false
  | equal_wait_set_actiona A_ B_ (Notify w) WokenUp = false
  | equal_wait_set_actiona A_ B_ Notified (Notify w) = false
  | equal_wait_set_actiona A_ B_ (Notify w) Notified = false
  | equal_wait_set_actiona A_ B_ (WakeUp t) (Notify w) = false
  | equal_wait_set_actiona A_ B_ (Notify w) (WakeUp t) = false
  | equal_wait_set_actiona A_ B_ (NotifyAll wa) (Notify w) = false
  | equal_wait_set_actiona A_ B_ (Notify wa) (NotifyAll w) = false
  | equal_wait_set_actiona A_ B_ WokenUp (Suspend w) = false
  | equal_wait_set_actiona A_ B_ (Suspend w) WokenUp = false
  | equal_wait_set_actiona A_ B_ Notified (Suspend w) = false
  | equal_wait_set_actiona A_ B_ (Suspend w) Notified = false
  | equal_wait_set_actiona A_ B_ (WakeUp t) (Suspend w) = false
  | equal_wait_set_actiona A_ B_ (Suspend w) (WakeUp t) = false
  | equal_wait_set_actiona A_ B_ (NotifyAll wa) (Suspend w) = false
  | equal_wait_set_actiona A_ B_ (Suspend wa) (NotifyAll w) = false
  | equal_wait_set_actiona A_ B_ (Notify wa) (Suspend w) = false
  | equal_wait_set_actiona A_ B_ (Suspend wa) (Notify w) = false
  | equal_wait_set_actiona A_ B_ (WakeUp ta) (WakeUp t) = HOL.eq A_ ta t
  | equal_wait_set_actiona A_ B_ (NotifyAll wa) (NotifyAll w) = HOL.eq B_ wa w
  | equal_wait_set_actiona A_ B_ (Notify wa) (Notify w) = HOL.eq B_ wa w
  | equal_wait_set_actiona A_ B_ (Suspend wa) (Suspend w) = HOL.eq B_ wa w
  | equal_wait_set_actiona A_ B_ WokenUp WokenUp = true
  | equal_wait_set_actiona A_ B_ Notified Notified = true;

fun equal_wait_set_action A_ B_ = {equal = equal_wait_set_actiona A_ B_} :
  ('a, 'b) wait_set_action HOL.equal;

fun equal_wake_up_status WSWokenUp WSNotified = false
  | equal_wake_up_status WSNotified WSWokenUp = false
  | equal_wake_up_status WSWokenUp WSWokenUp = true
  | equal_wake_up_status WSNotified WSNotified = true;

fun equal_wait_set_statusa A_ (PostWS wake_up_status) (InWS w) = false
  | equal_wait_set_statusa A_ (InWS w) (PostWS wake_up_status) = false
  | equal_wait_set_statusa A_ (PostWS wake_up_statusa) (PostWS wake_up_status) =
    equal_wake_up_status wake_up_statusa wake_up_status
  | equal_wait_set_statusa A_ (InWS wa) (InWS w) = HOL.eq A_ wa w;

fun equal_wait_set_status A_ = {equal = equal_wait_set_statusa A_} :
  'a wait_set_status HOL.equal;

fun equal_interrupt_actiona A_ (ClearInterrupt ta) (Interrupt t) = false
  | equal_interrupt_actiona A_ (Interrupt ta) (ClearInterrupt t) = false
  | equal_interrupt_actiona A_ (ClearInterrupt ta) (IsInterrupted (t, boola)) =
    false
  | equal_interrupt_actiona A_ (IsInterrupted (ta, boola)) (ClearInterrupt t) =
    false
  | equal_interrupt_actiona A_ (Interrupt ta) (IsInterrupted (t, boola)) = false
  | equal_interrupt_actiona A_ (IsInterrupted (ta, boola)) (Interrupt t) = false
  | equal_interrupt_actiona A_ (ClearInterrupt ta) (ClearInterrupt t) =
    HOL.eq A_ ta t
  | equal_interrupt_actiona A_ (Interrupt ta) (Interrupt t) = HOL.eq A_ ta t
  | equal_interrupt_actiona A_ (IsInterrupted (ta, boolaa))
    (IsInterrupted (t, boola)) =
    HOL.eq A_ ta t andalso Product_Type.equal_bool boolaa boola;

fun equal_interrupt_action A_ = {equal = equal_interrupt_actiona A_} :
  'a interrupt_action HOL.equal;

fun ta_update_Conditional (ls, (nts, (js, (wss, (is, obs))))) j =
  (ls, (nts, (js @ [j], (wss, (is, obs)))));

fun equal_new_thread_actiona A_ B_ C_ (ThreadExists (ta, boola))
  (NewThread (t, x, m)) = false
  | equal_new_thread_actiona A_ B_ C_ (NewThread (ta, x, m))
    (ThreadExists (t, boola)) = false
  | equal_new_thread_actiona A_ B_ C_ (ThreadExists (ta, boolaa))
    (ThreadExists (t, boola)) =
    HOL.eq A_ ta t andalso Product_Type.equal_bool boolaa boola
  | equal_new_thread_actiona A_ B_ C_ (NewThread (ta, xa, ma))
    (NewThread (t, x, m)) =
    HOL.eq A_ ta t andalso (HOL.eq B_ xa x andalso HOL.eq C_ ma m);

fun equal_new_thread_action A_ B_ C_ =
  {equal = equal_new_thread_actiona A_ B_ C_} :
  ('a, 'b, 'c) new_thread_action HOL.equal;

fun equal_conditional_actiona A_ Yield (Join t) = false
  | equal_conditional_actiona A_ (Join t) Yield = false
  | equal_conditional_actiona A_ (Join ta) (Join t) = HOL.eq A_ ta t
  | equal_conditional_actiona A_ Yield Yield = true;

fun equal_conditional_action A_ = {equal = equal_conditional_actiona A_} :
  'a conditional_action HOL.equal;

fun thread_action_to_thread_action A_ F_ (LockAction (la, l)) ta =
  ta_update_locks A_ ta la l
  | thread_action_to_thread_action A_ F_ (NewThreadAction nt) ta =
    ta_update_NewThread ta nt
  | thread_action_to_thread_action A_ F_ (ConditionalAction ca) ta =
    ta_update_Conditional ta ca
  | thread_action_to_thread_action A_ F_ (WaitSetAction wa) ta =
    ta_update_wait_set ta wa
  | thread_action_to_thread_action A_ F_ (InterruptAction ia) ta =
    ta_update_interrupt ta ia
  | thread_action_to_thread_action A_ F_ (ObsAction ob) ta =
    ta_update_obs ta ob;

end; (*struct FWState*)

structure FWLock : sig
  val may_lock : 'a HOL.equal -> ('a * Arith.nat) option -> 'a -> bool
  val unlock_lock : ('a * Arith.nat) option -> ('a * Arith.nat) option
  val release_all :
    'a HOL.equal -> ('a * Arith.nat) option -> 'a -> ('a * Arith.nat) option
  val lock_lock : ('a * Arith.nat) option -> 'a -> ('a * Arith.nat) option
  val upd_lock :
    'a HOL.equal ->
      ('a * Arith.nat) option ->
        'a -> FWState.lock_action -> ('a * Arith.nat) option
  val has_locks : 'a HOL.equal -> ('a * Arith.nat) option -> 'a -> Arith.nat
  val upd_locks :
    'a HOL.equal ->
      ('a * Arith.nat) option ->
        'a -> FWState.lock_action list -> ('a * Arith.nat) option
  val acquire_locks :
    ('a * Arith.nat) option -> 'a -> Arith.nat -> ('a * Arith.nat) option
  val lock_action_ok :
    'a HOL.equal -> ('a * Arith.nat) option -> 'a -> FWState.lock_action -> bool
  val lock_actions_ok :
    'a HOL.equal ->
      ('a * Arith.nat) option -> 'a -> FWState.lock_action list -> bool
end = struct

fun may_lock A_ l t = (case l of NONE => true | SOME (ta, _) => HOL.eq A_ t ta);

fun unlock_lock NONE = NONE
  | unlock_lock (SOME (t, n)) =
    (if Arith.equal_nata n Arith.zero_nat then NONE
      else SOME (t, Arith.minus_nat n Arith.one_nat));

fun release_all A_ NONE t = NONE
  | release_all A_ (SOME (ta, n)) t =
    (if HOL.eq A_ t ta then NONE else SOME (ta, n));

fun lock_lock NONE t = SOME (t, Arith.zero_nat)
  | lock_lock (SOME (ta, n)) t = SOME (ta, Arith.suc n);

fun upd_lock A_ l t FWState.Lock = lock_lock l t
  | upd_lock A_ l t FWState.Unlock = unlock_lock l
  | upd_lock A_ l t FWState.UnlockFail = l
  | upd_lock A_ l t FWState.ReleaseAcquire = release_all A_ l t;

fun has_locks A_ NONE t = Arith.zero_nat
  | has_locks A_ (SOME (ta, n)) t =
    (if HOL.eq A_ t ta then Arith.suc n else Arith.zero_nat);

fun upd_locks A_ l t [] = l
  | upd_locks A_ la t (l :: ls) = upd_locks A_ (upd_lock A_ la t l) t ls;

fun acquire_locks l t m =
  (if Arith.equal_nata m Arith.zero_nat then l
    else acquire_locks (lock_lock l t) t (Arith.minus_nat m Arith.one_nat));

fun lock_action_ok A_ l t FWState.Lock = may_lock A_ l t
  | lock_action_ok A_ l t FWState.Unlock =
    Arith.less_nat Arith.zero_nat (has_locks A_ l t)
  | lock_action_ok A_ l t FWState.UnlockFail =
    not (Arith.less_nat Arith.zero_nat (has_locks A_ l t))
  | lock_action_ok A_ l t FWState.ReleaseAcquire = true;

fun lock_actions_ok A_ l t [] = true
  | lock_actions_ok A_ la t (l :: ls) =
    lock_action_ok A_ la t l andalso
      lock_actions_ok A_ (upd_lock A_ la t l) t ls;

end; (*struct FWLock*)

structure FWWait : sig
  val waiting : 'a FWState.wait_set_status option -> bool
end = struct

fun waiting (SOME (FWState.InWS w)) = true
  | waiting (SOME (FWState.PostWS w)) = false
  | waiting NONE = false;

end; (*struct FWWait*)

structure Random : sig
  val log : Arith.natural -> Arith.natural -> Arith.natural
  val minus_shift :
    Arith.natural -> Arith.natural -> Arith.natural -> Arith.natural
  val next :
    Arith.natural * Arith.natural ->
      Arith.natural * (Arith.natural * Arith.natural)
  val iterate : Arith.natural -> ('a -> 'b -> 'a * 'b) -> 'a -> 'b -> 'a * 'b
  val range :
    Arith.natural ->
      Arith.natural * Arith.natural ->
        Arith.natural * (Arith.natural * Arith.natural)
  val select :
    'a list ->
      Arith.natural * Arith.natural -> 'a * (Arith.natural * Arith.natural)
end = struct

fun log b i =
  (if Arith.less_eq_natural b Arith.one_natural orelse Arith.less_natural i b
    then Arith.one_natural
    else Arith.plus_natural Arith.one_natural (log b (Arith.div_natural i b)));

fun minus_shift r k l =
  (if Arith.less_natural k l then Arith.minus_natural (Arith.plus_natural r k) l
    else Arith.minus_natural k l);

fun next (v, w) =
  let
    val k = Arith.div_natural v (Arith.natural_of_integer (53668 : IntInf.int));
    val va =
      minus_shift (Arith.natural_of_integer (2147483563 : IntInf.int))
        (Arith.times_natural
          (Arith.mod_natural v (Arith.natural_of_integer (53668 : IntInf.int)))
          (Arith.natural_of_integer (40014 : IntInf.int)))
        (Arith.times_natural k (Arith.natural_of_integer (12211 : IntInf.int)));
    val l = Arith.div_natural w (Arith.natural_of_integer (52774 : IntInf.int));
    val wa =
      minus_shift (Arith.natural_of_integer (2147483399 : IntInf.int))
        (Arith.times_natural
          (Arith.mod_natural w (Arith.natural_of_integer (52774 : IntInf.int)))
          (Arith.natural_of_integer (40692 : IntInf.int)))
        (Arith.times_natural l (Arith.natural_of_integer (3791 : IntInf.int)));
    val z =
      Arith.plus_natural
        (minus_shift (Arith.natural_of_integer (2147483562 : IntInf.int)) va
          (Arith.plus_natural wa Arith.one_natural))
        Arith.one_natural;
  in
    (z, (va, wa))
  end;

fun iterate k f x =
  (if Arith.equal_natural k Arith.zero_natural then (fn a => (x, a))
    else Product_Type.scomp (f x)
           (iterate (Arith.minus_natural k Arith.one_natural) f));

fun range k =
  Product_Type.scomp
    (iterate (log (Arith.natural_of_integer (2147483561 : IntInf.int)) k)
      (fn l =>
        Product_Type.scomp next
          (fn v =>
            (fn a =>
              (Arith.plus_natural v
                 (Arith.times_natural l
                   (Arith.natural_of_integer (2147483561 : IntInf.int))),
                a))))
      Arith.one_natural)
    (fn v => (fn a => (Arith.mod_natural v k, a)));

fun select xs =
  Product_Type.scomp (range (Arith.natural_of_nat (List.size_list xs)))
    (fn k => (fn a => (List.nth xs (Arith.nat_of_natural k), a)));

end; (*struct Random*)

structure Coinductive_List : sig
  datatype 'a llist = Lazy_llist of (unit -> ('a * 'a llist) option)
  val force : 'a llist -> ('a * 'a llist) option
  val lNil : 'a llist
  val lmap : ('a -> 'b) -> 'a llist -> 'b llist
  val lappend : 'a llist -> 'a llist -> 'a llist
  val lconcat : 'a llist llist -> 'a llist
  val lfilter : ('a -> bool) -> 'a llist -> 'a llist
  val llist_of : 'a list -> 'a llist
  val llist_case : 'a -> ('b -> 'b llist -> 'a) -> 'b llist -> 'a
end = struct

datatype 'a llist = Lazy_llist of (unit -> ('a * 'a llist) option);

fun force (Lazy_llist xs) = xs ();

val lNil : 'a llist = Lazy_llist (fn _ => NONE);

fun lmap f (Lazy_llist xs) =
  Lazy_llist (fn _ => Option.map (Product_Type.map_pair f (lmap f)) (xs ()));

fun lappend (Lazy_llist xs) ys =
  Lazy_llist
    (fn _ =>
      (case xs () of NONE => force ys
        | SOME (x, xsa) => SOME (x, lappend xsa ys)));

fun lconcat (Lazy_llist xss) =
  Lazy_llist
    (fn _ =>
      (case xss () of NONE => NONE
        | SOME (xs, xssa) => force (lappend xs (lconcat xssa))));

fun lfilter p (Lazy_llist xs) =
  Lazy_llist
    (fn _ =>
      (case xs () of NONE => NONE
        | SOME (x, ys) =>
          (if p x then SOME (x, lfilter p ys) else force (lfilter p ys))));

fun llist_of [] = lNil
  | llist_of (x :: xs) = Lazy_llist (fn _ => SOME (x, llist_of xs));

fun llist_case n c (Lazy_llist xs) =
  (case xs () of NONE => n | SOME (a, b) => c a b);

end; (*struct Coinductive_List*)

structure TLList : sig
  datatype ('a, 'b) tllist =
    Lazy_tllist of (unit -> (('a * ('a, 'b) tllist), 'b) Sum_Type.sum)
  val tNil : 'a -> ('b, 'a) tllist
  val tmap : ('a -> 'b) -> ('c -> 'd) -> ('a, 'c) tllist -> ('b, 'd) tllist
  val force : ('a, 'b) tllist -> (('a * ('a, 'b) tllist), 'b) Sum_Type.sum
  val tfilter :
    (unit -> 'a) -> ('b -> bool) -> ('b, 'a) tllist -> ('b, 'a) tllist
  val terminal : ('a, 'b) tllist -> 'b
  val tllist_case :
    ('a -> 'b) -> ('c -> ('c, 'a) tllist -> 'b) -> ('c, 'a) tllist -> 'b
  val llist_of_tllist : ('a, 'b) tllist -> 'a Coinductive_List.llist
end = struct

datatype ('a, 'b) tllist =
  Lazy_tllist of (unit -> (('a * ('a, 'b) tllist), 'b) Sum_Type.sum);

fun tNil b = Lazy_tllist (fn _ => Sum_Type.Inr b);

fun tmap f g (Lazy_tllist xs) =
  Lazy_tllist
    (fn _ =>
      (case xs () of Sum_Type.Inl (x, ys) => Sum_Type.Inl (f x, tmap f g ys)
        | Sum_Type.Inr b => Sum_Type.Inr (g b)));

fun force (Lazy_tllist xs) = xs ();

fun tfilter b p (Lazy_tllist xs) =
  Lazy_tllist
    (fn _ =>
      (case xs ()
        of Sum_Type.Inl (x, xsa) =>
          (if p x then Sum_Type.Inl (x, tfilter b p xsa)
            else force (tfilter b p xsa))
        | Sum_Type.Inr a => Sum_Type.Inr a));

fun terminal (Lazy_tllist xs) =
  (case xs () of Sum_Type.Inl (_, a) => terminal a | Sum_Type.Inr b => b);

fun tllist_case n c (Lazy_tllist xs) =
  (case xs () of Sum_Type.Inl (a, b) => c a b | Sum_Type.Inr a => n a);

fun llist_of_tllist (Lazy_tllist xs) =
  Coinductive_List.Lazy_llist
    (fn _ =>
      (case xs () of Sum_Type.Inl (x, ys) => SOME (x, llist_of_tllist ys)
        | Sum_Type.Inr _ => NONE));

end; (*struct TLList*)

structure BCVExec : sig
  val wf_jvm_prog :
    'a HOL.equal ->
      (Arith.nat *
        (Arith.nat *
          ('a JVMInstructions.instr list *
            (Arith.nat *
              (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list)))
        TypeRel.prog ->
        bool
end = struct

fun wf_jvm_prog A_ = BVExec.wf_jvm_prog_k A_;

end; (*struct BCVExec*)

structure J1State : sig
  val max_vars : ('a, 'b, 'c) Expr.exp -> Arith.nat
  val max_varss : ('a, 'b, 'c) Expr.exp list -> Arith.nat
end = struct

fun max_vars (Expr.New c) = Arith.zero_nat
  | max_vars (Expr.NewArray (t, e)) = max_vars e
  | max_vars (Expr.Cast (c, e)) = max_vars e
  | max_vars (Expr.InstanceOf (e, t)) = max_vars e
  | max_vars (Expr.Val v) = Arith.zero_nat
  | max_vars (Expr.BinOp (ea, bop, e)) =
    Orderings.max Arith.ord_nat (max_vars ea) (max_vars e)
  | max_vars (Expr.Var v) = Arith.zero_nat
  | max_vars (Expr.LAss (v, e)) = max_vars e
  | max_vars (Expr.AAcc (a, i)) =
    Orderings.max Arith.ord_nat (max_vars a) (max_vars i)
  | max_vars (Expr.AAss (a, i, e)) =
    Orderings.max Arith.ord_nat
      (Orderings.max Arith.ord_nat (max_vars a) (max_vars i)) (max_vars e)
  | max_vars (Expr.ALen a) = max_vars a
  | max_vars (Expr.FAcc (e, f, d)) = max_vars e
  | max_vars (Expr.FAss (e_1, f, d, e_2)) =
    Orderings.max Arith.ord_nat (max_vars e_1) (max_vars e_2)
  | max_vars (Expr.Call (e, m, es)) =
    Orderings.max Arith.ord_nat (max_vars e) (max_varss es)
  | max_vars (Expr.Block (v, t, vo, e)) =
    Arith.plus_nat (max_vars e) Arith.one_nat
  | max_vars (Expr.Synchronized (v, ea, e)) =
    Orderings.max Arith.ord_nat (max_vars ea)
      (Arith.plus_nat (max_vars e) Arith.one_nat)
  | max_vars (Expr.InSynchronized (v, a, e)) =
    Arith.plus_nat (max_vars e) Arith.one_nat
  | max_vars (Expr.Seq (e_1, e_2)) =
    Orderings.max Arith.ord_nat (max_vars e_1) (max_vars e_2)
  | max_vars (Expr.Cond (e, e_1, e_2)) =
    Orderings.max Arith.ord_nat (max_vars e)
      (Orderings.max Arith.ord_nat (max_vars e_1) (max_vars e_2))
  | max_vars (Expr.While (b, e)) =
    Orderings.max Arith.ord_nat (max_vars b) (max_vars e)
  | max_vars (Expr.Throw e) = max_vars e
  | max_vars (Expr.TryCatch (e_1, c, v, e_2)) =
    Orderings.max Arith.ord_nat (max_vars e_1)
      (Arith.plus_nat (max_vars e_2) Arith.one_nat)
and max_varss [] = Arith.zero_nat
  | max_varss (e :: es) =
    Orderings.max Arith.ord_nat (max_vars e) (max_varss es);

end; (*struct J1State*)

structure RBT_add : sig
  val rm_iterateoi :
    ('a, 'b) RBT_Impl.rbt -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
  val rm_reverse_iterateoi :
    ('a, 'b) RBT_Impl.rbt -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
end = struct

fun rm_iterateoi RBT_Impl.Empty c f sigma = sigma
  | rm_iterateoi (RBT_Impl.Branch (col, l, k, v, r)) c f sigma =
    (if c sigma
      then let
             val sigmaa = rm_iterateoi l c f sigma;
           in
             (if c sigmaa then rm_iterateoi r c f (f (k, v) sigmaa) else sigmaa)
           end
      else sigma);

fun rm_reverse_iterateoi RBT_Impl.Empty c f sigma = sigma
  | rm_reverse_iterateoi (RBT_Impl.Branch (col, l, k, v, r)) c f sigma =
    (if c sigma
      then let
             val sigmaa = rm_reverse_iterateoi r c f sigma;
           in
             (if c sigmaa then rm_reverse_iterateoi l c f (f (k, v) sigmaa)
               else sigmaa)
           end
      else sigma);

end; (*struct RBT_add*)

structure PCompiler : sig
  val compM :
    (string -> Type.ty list -> Type.ty -> 'a -> 'b) ->
      string * (Type.ty list * (Type.ty * 'a option)) ->
        string * (Type.ty list * (Type.ty * 'b option))
  val compC :
    (string -> string -> Type.ty list -> Type.ty -> 'a -> 'b) ->
      string *
        (string *
          ((string * (Type.ty * unit TypeRel.fmod_ext)) list *
            (string * (Type.ty list * (Type.ty * 'a option))) list)) ->
        string *
          (string *
            ((string * (Type.ty * unit TypeRel.fmod_ext)) list *
              (string * (Type.ty list * (Type.ty * 'b option))) list))
  val compP :
    (string -> string -> Type.ty list -> Type.ty -> 'a -> 'b) ->
      'a TypeRel.prog -> 'b TypeRel.prog
end = struct

fun compM f =
  (fn (m, (ts, (t, ma))) => (m, (ts, (t, Option.map (f m ts t) ma))));

fun compC f =
  (fn (c, (d, (fdecls, mdecls))) =>
    (c, (d, (fdecls, List.map (compM (f c)) mdecls))));

fun compP f (TypeRel.Program p) = TypeRel.Program (List.map (compC f) p);

end; (*struct PCompiler*)

structure Compiler2 : sig
  val compE2 :
    (Arith.nat, Arith.nat, 'a) Expr.exp -> 'a JVMInstructions.instr list
  val compEs2 :
    (Arith.nat, Arith.nat, 'a) Expr.exp list -> 'a JVMInstructions.instr list
  val max_stack : (Arith.nat, Arith.nat, 'a) Expr.exp -> Arith.nat
  val max_stacks : (Arith.nat, Arith.nat, 'a) Expr.exp list -> Arith.nat
  val compxE2 :
    (Arith.nat, Arith.nat, 'a) Expr.exp ->
      Arith.nat ->
        Arith.nat ->
          (Arith.nat *
            (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list
  val compxEs2 :
    (Arith.nat, Arith.nat, 'a) Expr.exp list ->
      Arith.nat ->
        Arith.nat ->
          (Arith.nat *
            (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list
  val compMb2 :
    (Arith.nat, Arith.nat, 'a) Expr.exp ->
      Arith.nat *
        (Arith.nat *
          ('a JVMInstructions.instr list *
            (Arith.nat *
              (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list))
end = struct

fun compE2 (Expr.New c) = [JVMInstructions.New c]
  | compE2 (Expr.NewArray (t, e)) = compE2 e @ [JVMInstructions.NewArray t]
  | compE2 (Expr.Cast (t, e)) = compE2 e @ [JVMInstructions.Checkcast t]
  | compE2 (Expr.InstanceOf (e, t)) = compE2 e @ [JVMInstructions.Instanceof t]
  | compE2 (Expr.Val v) = [JVMInstructions.Push v]
  | compE2 (Expr.BinOp (e1, bop, e2)) =
    compE2 e1 @ compE2 e2 @ [JVMInstructions.BinOpInstr bop]
  | compE2 (Expr.Var i) = [JVMInstructions.Load i]
  | compE2 (Expr.LAss (i, e)) =
    compE2 e @ [JVMInstructions.Store i, JVMInstructions.Push Value.Unit]
  | compE2 (Expr.AAcc (a, i)) = compE2 a @ compE2 i @ [JVMInstructions.ALoad]
  | compE2 (Expr.AAss (a, i, e)) =
    compE2 a @
      compE2 i @
        compE2 e @ [JVMInstructions.AStore, JVMInstructions.Push Value.Unit]
  | compE2 (Expr.ALen a) = compE2 a @ [JVMInstructions.ALength]
  | compE2 (Expr.FAcc (e, f, d)) = compE2 e @ [JVMInstructions.Getfield (f, d)]
  | compE2 (Expr.FAss (e1, f, d, e2)) =
    compE2 e1 @
      compE2 e2 @
        [JVMInstructions.Putfield (f, d), JVMInstructions.Push Value.Unit]
  | compE2 (Expr.Call (e, m, es)) =
    compE2 e @ compEs2 es @ [JVMInstructions.Invoke (m, List.size_list es)]
  | compE2 (Expr.Block (i, t, vo, e)) =
    (case vo of NONE => []
      | SOME v => [JVMInstructions.Push v, JVMInstructions.Store i]) @
      compE2 e
  | compE2 (Expr.Synchronized (v, oa, e)) =
    compE2 oa @
      [JVMInstructions.Dup, JVMInstructions.Store v, JVMInstructions.MEnter] @
        compE2 e @
          [JVMInstructions.Load v, JVMInstructions.MExit,
            JVMInstructions.Goto (Arith.Int_of_integer (4 : IntInf.int)),
            JVMInstructions.Load v, JVMInstructions.MExit,
            JVMInstructions.ThrowExc]
  | compE2 (Expr.InSynchronized (v, a, e)) =
    [JVMInstructions.Goto (Arith.Int_of_integer (1 : IntInf.int))]
  | compE2 (Expr.Seq (e1, e2)) = compE2 e1 @ [JVMInstructions.Pop] @ compE2 e2
  | compE2 (Expr.Cond (e, e_1, e_2)) =
    let
      val cnd = compE2 e;
      val thn = compE2 e_1;
      val els = compE2 e_2;
      val test =
        JVMInstructions.IfFalse
          (Arith.int_of_nat
            (Arith.plus_nat (List.size_list thn)
              (Arith.nat_of_integer (2 : IntInf.int))));
      val thnex =
        JVMInstructions.Goto
          (Arith.int_of_nat
            (Arith.plus_nat (List.size_list els) Arith.one_nat));
    in
      cnd @ [test] @ thn @ [thnex] @ els
    end
  | compE2 (Expr.While (e, c)) =
    let
      val cnd = compE2 e;
      val bdy = compE2 c;
      val test =
        JVMInstructions.IfFalse
          (Arith.int_of_nat
            (Arith.plus_nat (List.size_list bdy)
              (Arith.nat_of_integer (3 : IntInf.int))));
      val loop =
        JVMInstructions.Goto
          (Arith.uminus_int
            (Arith.int_of_nat
              (Arith.plus_nat
                (Arith.plus_nat (List.size_list bdy) (List.size_list cnd))
                (Arith.nat_of_integer (2 : IntInf.int)))));
    in
      cnd @ [test] @
              bdy @ [JVMInstructions.Pop] @
                      [loop] @ [JVMInstructions.Push Value.Unit]
    end
  | compE2 (Expr.Throw e) = compE2 e @ [JVMInstructions.ThrowExc]
  | compE2 (Expr.TryCatch (e1, c, i, e2)) =
    let
      val catch = compE2 e2;
    in
      compE2 e1 @
        [JVMInstructions.Goto
           (Arith.plus_int (Arith.int_of_nat (List.size_list catch))
             (Arith.Int_of_integer (2 : IntInf.int))),
          JVMInstructions.Store i] @
          catch
    end
and compEs2 [] = []
  | compEs2 (e :: es) = compE2 e @ compEs2 es;

fun max_stack (Expr.New c) = Arith.one_nat
  | max_stack (Expr.NewArray (t, e)) = max_stack e
  | max_stack (Expr.Cast (c, e)) = max_stack e
  | max_stack (Expr.InstanceOf (e, t)) = max_stack e
  | max_stack (Expr.Val v) = Arith.one_nat
  | max_stack (Expr.BinOp (e1, bop, e2)) =
    Arith.plus_nat (Orderings.max Arith.ord_nat (max_stack e1) (max_stack e2))
      Arith.one_nat
  | max_stack (Expr.Var i) = Arith.one_nat
  | max_stack (Expr.LAss (i, e)) = max_stack e
  | max_stack (Expr.AAcc (a, i)) =
    Orderings.max Arith.ord_nat (max_stack a)
      (Arith.plus_nat (max_stack i) Arith.one_nat)
  | max_stack (Expr.AAss (a, i, e)) =
    Orderings.max Arith.ord_nat
      (Orderings.max Arith.ord_nat (max_stack a)
        (Arith.plus_nat (max_stack i) Arith.one_nat))
      (Arith.plus_nat (max_stack e) (Arith.nat_of_integer (2 : IntInf.int)))
  | max_stack (Expr.ALen a) = max_stack a
  | max_stack (Expr.FAcc (e, f, d)) = max_stack e
  | max_stack (Expr.FAss (e1, f, d, e2)) =
    Arith.plus_nat (Orderings.max Arith.ord_nat (max_stack e1) (max_stack e2))
      Arith.one_nat
  | max_stack (Expr.Call (e, m, es)) =
    Arith.plus_nat (Orderings.max Arith.ord_nat (max_stack e) (max_stacks es))
      Arith.one_nat
  | max_stack (Expr.Block (i, t, vo, e)) = max_stack e
  | max_stack (Expr.Synchronized (v, oa, e)) =
    Orderings.max Arith.ord_nat (max_stack oa)
      (Orderings.max Arith.ord_nat (max_stack e)
        (Arith.nat_of_integer (2 : IntInf.int)))
  | max_stack (Expr.InSynchronized (v, a, e)) = Arith.one_nat
  | max_stack (Expr.Seq (e1, e2)) =
    Orderings.max Arith.ord_nat (max_stack e1) (max_stack e2)
  | max_stack (Expr.Cond (e, e_1, e_2)) =
    Orderings.max Arith.ord_nat (max_stack e)
      (Orderings.max Arith.ord_nat (max_stack e_1) (max_stack e_2))
  | max_stack (Expr.While (e, c)) =
    Orderings.max Arith.ord_nat (max_stack e) (max_stack c)
  | max_stack (Expr.Throw e) = max_stack e
  | max_stack (Expr.TryCatch (e1, c, i, e2)) =
    Orderings.max Arith.ord_nat (max_stack e1) (max_stack e2)
and max_stacks [] = Arith.zero_nat
  | max_stacks (e :: es) =
    Orderings.max Arith.ord_nat (max_stack e)
      (Arith.plus_nat Arith.one_nat (max_stacks es));

fun compxE2 (Expr.New c) pc d = []
  | compxE2 (Expr.NewArray (t, e)) pc d = compxE2 e pc d
  | compxE2 (Expr.Cast (t, e)) pc d = compxE2 e pc d
  | compxE2 (Expr.InstanceOf (e, t)) pc d = compxE2 e pc d
  | compxE2 (Expr.Val v) pc d = []
  | compxE2 (Expr.BinOp (e1, bop, e2)) pc d =
    compxE2 e1 pc d @
      compxE2 e2 (Arith.plus_nat pc (List.size_list (compE2 e1)))
        (Arith.plus_nat d Arith.one_nat)
  | compxE2 (Expr.Var i) pc d = []
  | compxE2 (Expr.LAss (i, e)) pc d = compxE2 e pc d
  | compxE2 (Expr.AAcc (a, i)) pc d =
    compxE2 a pc d @
      compxE2 i (Arith.plus_nat pc (List.size_list (compE2 a)))
        (Arith.plus_nat d Arith.one_nat)
  | compxE2 (Expr.AAss (a, i, e)) pc d =
    let
      val pc1 = Arith.plus_nat pc (List.size_list (compE2 a));
      val pc2 = Arith.plus_nat pc1 (List.size_list (compE2 i));
    in
      compxE2 a pc d @
        compxE2 i pc1 (Arith.plus_nat d Arith.one_nat) @
          compxE2 e pc2
            (Arith.plus_nat d (Arith.nat_of_integer (2 : IntInf.int)))
    end
  | compxE2 (Expr.ALen a) pc d = compxE2 a pc d
  | compxE2 (Expr.FAcc (e, f, da)) pc d = compxE2 e pc d
  | compxE2 (Expr.FAss (e1, f, da, e2)) pc d =
    compxE2 e1 pc d @
      compxE2 e2 (Arith.plus_nat pc (List.size_list (compE2 e1)))
        (Arith.plus_nat d Arith.one_nat)
  | compxE2 (Expr.Call (e, m, es)) pc d =
    compxE2 e pc d @
      compxEs2 es (Arith.plus_nat pc (List.size_list (compE2 e)))
        (Arith.plus_nat d Arith.one_nat)
  | compxE2 (Expr.Block (i, t, vo, e)) pc d =
    compxE2 e (case vo of NONE => pc | SOME _ => Arith.suc (Arith.suc pc)) d
  | compxE2 (Expr.Synchronized (v, oa, e)) pc d =
    let
      val pc1 =
        Arith.plus_nat (Arith.plus_nat pc (List.size_list (compE2 oa)))
          (Arith.nat_of_integer (3 : IntInf.int));
      val pc2 = Arith.plus_nat pc1 (List.size_list (compE2 e));
    in
      compxE2 oa pc d @
        compxE2 e pc1 d @
          [(pc1, (pc2, (NONE, (Arith.suc (Arith.suc (Arith.suc pc2)), d))))]
    end
  | compxE2 (Expr.InSynchronized (v, a, e)) pc d = []
  | compxE2 (Expr.Seq (e1, e2)) pc d =
    compxE2 e1 pc d @
      compxE2 e2
        (Arith.plus_nat (Arith.plus_nat pc (List.size_list (compE2 e1)))
          Arith.one_nat)
        d
  | compxE2 (Expr.Cond (e, e_1, e_2)) pc d =
    let
      val pc_1 =
        Arith.plus_nat (Arith.plus_nat pc (List.size_list (compE2 e)))
          Arith.one_nat;
      val pc_2 =
        Arith.plus_nat (Arith.plus_nat pc_1 (List.size_list (compE2 e_1)))
          Arith.one_nat;
    in
      compxE2 e pc d @ compxE2 e_1 pc_1 d @ compxE2 e_2 pc_2 d
    end
  | compxE2 (Expr.While (b, e)) pc d =
    compxE2 b pc d @
      compxE2 e
        (Arith.plus_nat (Arith.plus_nat pc (List.size_list (compE2 b)))
          Arith.one_nat)
        d
  | compxE2 (Expr.Throw e) pc d = compxE2 e pc d
  | compxE2 (Expr.TryCatch (e1, c, i, e2)) pc d =
    let
      val pc1 = Arith.plus_nat pc (List.size_list (compE2 e1));
    in
      compxE2 e1 pc d @
        compxE2 e2 (Arith.plus_nat pc1 (Arith.nat_of_integer (2 : IntInf.int)))
          d @
          [(pc, (pc1, (SOME c, (Arith.plus_nat pc1 Arith.one_nat, d))))]
    end
and compxEs2 [] pc d = []
  | compxEs2 (e :: es) pc d =
    compxE2 e pc d @
      compxEs2 es (Arith.plus_nat pc (List.size_list (compE2 e)))
        (Arith.plus_nat d Arith.one_nat);

fun compMb2 x =
  (fn body =>
    let
      val ins = compE2 body @ [JVMInstructions.Return];
      val xt = compxE2 body Arith.zero_nat Arith.zero_nat;
    in
      (max_stack body, (J1State.max_vars body, (ins, xt)))
    end)
    x;

end; (*struct Compiler2*)

structure ListIndex : sig
  val index : 'a HOL.equal -> 'a list -> 'a -> Arith.nat
end = struct

fun index A_ [] y = Arith.zero_nat
  | index A_ (x :: xs) y =
    (if HOL.eq A_ x y
      then (if List.member A_ xs x
             then Arith.plus_nat (index A_ xs y) Arith.one_nat
             else Arith.zero_nat)
      else Arith.plus_nat (index A_ xs y) Arith.one_nat);

end; (*struct ListIndex*)

structure Compiler1 : sig
  val fresh_var : string list -> string
  val compE1 :
    string list ->
      (string, unit, 'a) Expr.exp -> (Arith.nat, Arith.nat, 'a) Expr.exp
  val compEs1 :
    string list ->
      (string, unit, 'a) Expr.exp list ->
        (Arith.nat, Arith.nat, 'a) Expr.exp list
end = struct

fun fresh_var vs = String.concat ("V" :: vs);

fun compE1 vs (Expr.New c) = Expr.New c
  | compE1 vs (Expr.NewArray (t, e)) = Expr.NewArray (t, compE1 vs e)
  | compE1 vs (Expr.Cast (t, e)) = Expr.Cast (t, compE1 vs e)
  | compE1 vs (Expr.InstanceOf (e, t)) = Expr.InstanceOf (compE1 vs e, t)
  | compE1 vs (Expr.Val v) = Expr.Val v
  | compE1 vs (Expr.Var v) =
    Expr.Var (ListIndex.index Stringa.equal_literal vs v)
  | compE1 vs (Expr.BinOp (ea, bop, e)) =
    Expr.BinOp (compE1 vs ea, bop, compE1 vs e)
  | compE1 vs (Expr.LAss (v, e)) =
    Expr.LAss (ListIndex.index Stringa.equal_literal vs v, compE1 vs e)
  | compE1 vs (Expr.AAcc (a, i)) = Expr.AAcc (compE1 vs a, compE1 vs i)
  | compE1 vs (Expr.AAss (a, i, e)) =
    Expr.AAss (compE1 vs a, compE1 vs i, compE1 vs e)
  | compE1 vs (Expr.ALen a) = Expr.ALen (compE1 vs a)
  | compE1 vs (Expr.FAcc (e, f, d)) = Expr.FAcc (compE1 vs e, f, d)
  | compE1 vs (Expr.FAss (ea, f, d, e)) =
    Expr.FAss (compE1 vs ea, f, d, compE1 vs e)
  | compE1 vs (Expr.Call (e, m, es)) = Expr.Call (compE1 vs e, m, compEs1 vs es)
  | compE1 vs (Expr.Block (v, t, vo, e)) =
    Expr.Block (List.size_list vs, t, vo, compE1 (vs @ [v]) e)
  | compE1 vs (Expr.Synchronized (u, oa, e)) =
    Expr.Synchronized
      (List.size_list vs, compE1 vs oa, compE1 (vs @ [fresh_var vs]) e)
  | compE1 vs (Expr.InSynchronized (u, a, e)) =
    Expr.InSynchronized (List.size_list vs, a, compE1 (vs @ [fresh_var vs]) e)
  | compE1 vs (Expr.Seq (e1, e2)) = Expr.Seq (compE1 vs e1, compE1 vs e2)
  | compE1 vs (Expr.Cond (b, e1, e2)) =
    Expr.Cond (compE1 vs b, compE1 vs e1, compE1 vs e2)
  | compE1 vs (Expr.While (b, e)) = Expr.While (compE1 vs b, compE1 vs e)
  | compE1 vs (Expr.Throw e) = Expr.Throw (compE1 vs e)
  | compE1 vs (Expr.TryCatch (e1, c, v, e2)) =
    Expr.TryCatch (compE1 vs e1, c, List.size_list vs, compE1 (vs @ [v]) e2)
and compEs1 vs [] = []
  | compEs1 vs (e :: es) = compE1 vs e :: compEs1 vs es;

end; (*struct Compiler1*)

structure Compiler : sig
  val j2JVM :
    (string list * (string, unit, 'a) Expr.exp) TypeRel.prog ->
      (Arith.nat *
        (Arith.nat *
          ('a JVMInstructions.instr list *
            (Arith.nat *
              (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list)))
        TypeRel.prog
end = struct

fun j2JVM x =
  PCompiler.compP
    (fn _ => fn _ => fn _ => fn _ => fn (pns, body) =>
      Compiler2.compMb2 (Compiler1.compE1 ("this" :: pns) body))
    x;

end; (*struct Compiler*)

structure Observable_Events : sig
  datatype ('a, 'b) obs_event =
    ExternalCall of 'a * string * 'a Value.vala list * 'a Value.vala |
    ReadMem of 'a * Heap.addr_loc * 'a Value.vala |
    WriteMem of 'a * Heap.addr_loc * 'a Value.vala |
    NewHeapElem of 'a * Type.htype | ThreadStart of 'b | ThreadJoin of 'b |
    SyncLock of 'a | SyncUnlock of 'a | ObsInterrupt of 'b |
    ObsInterrupted of 'b
  val convert_RA :
    'a HOL.equal * 'a Value.addr ->
      ('a, Arith.nat) FinFun.finfun -> ('a, 'b) obs_event list
  val equal_obs_eventa :
    'a HOL.equal -> 'b HOL.equal ->
      ('a, 'b) obs_event -> ('a, 'b) obs_event -> bool
  val equal_obs_event :
    'a HOL.equal -> 'b HOL.equal -> ('a, 'b) obs_event HOL.equal
  val obs_action_obs_event : ('a, 'b) obs_event FWState.obs_action
end = struct

datatype ('a, 'b) obs_event =
  ExternalCall of 'a * string * 'a Value.vala list * 'a Value.vala |
  ReadMem of 'a * Heap.addr_loc * 'a Value.vala |
  WriteMem of 'a * Heap.addr_loc * 'a Value.vala |
  NewHeapElem of 'a * Type.htype | ThreadStart of 'b | ThreadJoin of 'b |
  SyncLock of 'a | SyncUnlock of 'a | ObsInterrupt of 'b | ObsInterrupted of 'b;

fun convert_RA (A1_, A2_) ln =
  List.maps
    (fn ad => List.replicate (FinFun.finfun_apply A1_ ln ad) (SyncLock ad))
    (Value.monitor_finfun_to_list A2_ ln);

fun equal_obs_eventa A_ B_ (ObsInterrupted thread_ida) (ObsInterrupt thread_id)
  = false
  | equal_obs_eventa A_ B_ (ObsInterrupt thread_ida) (ObsInterrupted thread_id)
    = false
  | equal_obs_eventa A_ B_ (ObsInterrupted thread_id) (SyncUnlock addr) = false
  | equal_obs_eventa A_ B_ (SyncUnlock addr) (ObsInterrupted thread_id) = false
  | equal_obs_eventa A_ B_ (ObsInterrupt thread_id) (SyncUnlock addr) = false
  | equal_obs_eventa A_ B_ (SyncUnlock addr) (ObsInterrupt thread_id) = false
  | equal_obs_eventa A_ B_ (ObsInterrupted thread_id) (SyncLock addr) = false
  | equal_obs_eventa A_ B_ (SyncLock addr) (ObsInterrupted thread_id) = false
  | equal_obs_eventa A_ B_ (ObsInterrupt thread_id) (SyncLock addr) = false
  | equal_obs_eventa A_ B_ (SyncLock addr) (ObsInterrupt thread_id) = false
  | equal_obs_eventa A_ B_ (SyncUnlock addra) (SyncLock addr) = false
  | equal_obs_eventa A_ B_ (SyncLock addra) (SyncUnlock addr) = false
  | equal_obs_eventa A_ B_ (ObsInterrupted thread_ida) (ThreadJoin thread_id) =
    false
  | equal_obs_eventa A_ B_ (ThreadJoin thread_ida) (ObsInterrupted thread_id) =
    false
  | equal_obs_eventa A_ B_ (ObsInterrupt thread_ida) (ThreadJoin thread_id) =
    false
  | equal_obs_eventa A_ B_ (ThreadJoin thread_ida) (ObsInterrupt thread_id) =
    false
  | equal_obs_eventa A_ B_ (SyncUnlock addr) (ThreadJoin thread_id) = false
  | equal_obs_eventa A_ B_ (ThreadJoin thread_id) (SyncUnlock addr) = false
  | equal_obs_eventa A_ B_ (SyncLock addr) (ThreadJoin thread_id) = false
  | equal_obs_eventa A_ B_ (ThreadJoin thread_id) (SyncLock addr) = false
  | equal_obs_eventa A_ B_ (ObsInterrupted thread_ida) (ThreadStart thread_id) =
    false
  | equal_obs_eventa A_ B_ (ThreadStart thread_ida) (ObsInterrupted thread_id) =
    false
  | equal_obs_eventa A_ B_ (ObsInterrupt thread_ida) (ThreadStart thread_id) =
    false
  | equal_obs_eventa A_ B_ (ThreadStart thread_ida) (ObsInterrupt thread_id) =
    false
  | equal_obs_eventa A_ B_ (SyncUnlock addr) (ThreadStart thread_id) = false
  | equal_obs_eventa A_ B_ (ThreadStart thread_id) (SyncUnlock addr) = false
  | equal_obs_eventa A_ B_ (SyncLock addr) (ThreadStart thread_id) = false
  | equal_obs_eventa A_ B_ (ThreadStart thread_id) (SyncLock addr) = false
  | equal_obs_eventa A_ B_ (ThreadJoin thread_ida) (ThreadStart thread_id) =
    false
  | equal_obs_eventa A_ B_ (ThreadStart thread_ida) (ThreadJoin thread_id) =
    false
  | equal_obs_eventa A_ B_ (ObsInterrupted thread_id)
    (NewHeapElem (addr, htype)) = false
  | equal_obs_eventa A_ B_ (NewHeapElem (addr, htype))
    (ObsInterrupted thread_id) = false
  | equal_obs_eventa A_ B_ (ObsInterrupt thread_id) (NewHeapElem (addr, htype))
    = false
  | equal_obs_eventa A_ B_ (NewHeapElem (addr, htype)) (ObsInterrupt thread_id)
    = false
  | equal_obs_eventa A_ B_ (SyncUnlock addra) (NewHeapElem (addr, htype)) =
    false
  | equal_obs_eventa A_ B_ (NewHeapElem (addra, htype)) (SyncUnlock addr) =
    false
  | equal_obs_eventa A_ B_ (SyncLock addra) (NewHeapElem (addr, htype)) = false
  | equal_obs_eventa A_ B_ (NewHeapElem (addra, htype)) (SyncLock addr) = false
  | equal_obs_eventa A_ B_ (ThreadJoin thread_id) (NewHeapElem (addr, htype)) =
    false
  | equal_obs_eventa A_ B_ (NewHeapElem (addr, htype)) (ThreadJoin thread_id) =
    false
  | equal_obs_eventa A_ B_ (ThreadStart thread_id) (NewHeapElem (addr, htype)) =
    false
  | equal_obs_eventa A_ B_ (NewHeapElem (addr, htype)) (ThreadStart thread_id) =
    false
  | equal_obs_eventa A_ B_ (ObsInterrupted thread_id)
    (WriteMem (addr, addr_loc, vala)) = false
  | equal_obs_eventa A_ B_ (WriteMem (addr, addr_loc, vala))
    (ObsInterrupted thread_id) = false
  | equal_obs_eventa A_ B_ (ObsInterrupt thread_id)
    (WriteMem (addr, addr_loc, vala)) = false
  | equal_obs_eventa A_ B_ (WriteMem (addr, addr_loc, vala))
    (ObsInterrupt thread_id) = false
  | equal_obs_eventa A_ B_ (SyncUnlock addra) (WriteMem (addr, addr_loc, vala))
    = false
  | equal_obs_eventa A_ B_ (WriteMem (addra, addr_loc, vala)) (SyncUnlock addr)
    = false
  | equal_obs_eventa A_ B_ (SyncLock addra) (WriteMem (addr, addr_loc, vala)) =
    false
  | equal_obs_eventa A_ B_ (WriteMem (addra, addr_loc, vala)) (SyncLock addr) =
    false
  | equal_obs_eventa A_ B_ (ThreadJoin thread_id)
    (WriteMem (addr, addr_loc, vala)) = false
  | equal_obs_eventa A_ B_ (WriteMem (addr, addr_loc, vala))
    (ThreadJoin thread_id) = false
  | equal_obs_eventa A_ B_ (ThreadStart thread_id)
    (WriteMem (addr, addr_loc, vala)) = false
  | equal_obs_eventa A_ B_ (WriteMem (addr, addr_loc, vala))
    (ThreadStart thread_id) = false
  | equal_obs_eventa A_ B_ (NewHeapElem (addra, htype))
    (WriteMem (addr, addr_loc, vala)) = false
  | equal_obs_eventa A_ B_ (WriteMem (addra, addr_loc, vala))
    (NewHeapElem (addr, htype)) = false
  | equal_obs_eventa A_ B_ (ObsInterrupted thread_id)
    (ReadMem (addr, addr_loc, vala)) = false
  | equal_obs_eventa A_ B_ (ReadMem (addr, addr_loc, vala))
    (ObsInterrupted thread_id) = false
  | equal_obs_eventa A_ B_ (ObsInterrupt thread_id)
    (ReadMem (addr, addr_loc, vala)) = false
  | equal_obs_eventa A_ B_ (ReadMem (addr, addr_loc, vala))
    (ObsInterrupt thread_id) = false
  | equal_obs_eventa A_ B_ (SyncUnlock addra) (ReadMem (addr, addr_loc, vala)) =
    false
  | equal_obs_eventa A_ B_ (ReadMem (addra, addr_loc, vala)) (SyncUnlock addr) =
    false
  | equal_obs_eventa A_ B_ (SyncLock addra) (ReadMem (addr, addr_loc, vala)) =
    false
  | equal_obs_eventa A_ B_ (ReadMem (addra, addr_loc, vala)) (SyncLock addr) =
    false
  | equal_obs_eventa A_ B_ (ThreadJoin thread_id)
    (ReadMem (addr, addr_loc, vala)) = false
  | equal_obs_eventa A_ B_ (ReadMem (addr, addr_loc, vala))
    (ThreadJoin thread_id) = false
  | equal_obs_eventa A_ B_ (ThreadStart thread_id)
    (ReadMem (addr, addr_loc, vala)) = false
  | equal_obs_eventa A_ B_ (ReadMem (addr, addr_loc, vala))
    (ThreadStart thread_id) = false
  | equal_obs_eventa A_ B_ (NewHeapElem (addra, htype))
    (ReadMem (addr, addr_loc, vala)) = false
  | equal_obs_eventa A_ B_ (ReadMem (addra, addr_loc, vala))
    (NewHeapElem (addr, htype)) = false
  | equal_obs_eventa A_ B_ (WriteMem (addra, addr_loca, valaa))
    (ReadMem (addr, addr_loc, vala)) = false
  | equal_obs_eventa A_ B_ (ReadMem (addra, addr_loca, valaa))
    (WriteMem (addr, addr_loc, vala)) = false
  | equal_obs_eventa A_ B_ (ObsInterrupted thread_id)
    (ExternalCall (addr, literal, lista, vala)) = false
  | equal_obs_eventa A_ B_ (ExternalCall (addr, literal, lista, vala))
    (ObsInterrupted thread_id) = false
  | equal_obs_eventa A_ B_ (ObsInterrupt thread_id)
    (ExternalCall (addr, literal, lista, vala)) = false
  | equal_obs_eventa A_ B_ (ExternalCall (addr, literal, lista, vala))
    (ObsInterrupt thread_id) = false
  | equal_obs_eventa A_ B_ (SyncUnlock addra)
    (ExternalCall (addr, literal, lista, vala)) = false
  | equal_obs_eventa A_ B_ (ExternalCall (addra, literal, lista, vala))
    (SyncUnlock addr) = false
  | equal_obs_eventa A_ B_ (SyncLock addra)
    (ExternalCall (addr, literal, lista, vala)) = false
  | equal_obs_eventa A_ B_ (ExternalCall (addra, literal, lista, vala))
    (SyncLock addr) = false
  | equal_obs_eventa A_ B_ (ThreadJoin thread_id)
    (ExternalCall (addr, literal, lista, vala)) = false
  | equal_obs_eventa A_ B_ (ExternalCall (addr, literal, lista, vala))
    (ThreadJoin thread_id) = false
  | equal_obs_eventa A_ B_ (ThreadStart thread_id)
    (ExternalCall (addr, literal, lista, vala)) = false
  | equal_obs_eventa A_ B_ (ExternalCall (addr, literal, lista, vala))
    (ThreadStart thread_id) = false
  | equal_obs_eventa A_ B_ (NewHeapElem (addra, htype))
    (ExternalCall (addr, literal, lista, vala)) = false
  | equal_obs_eventa A_ B_ (ExternalCall (addra, literal, lista, vala))
    (NewHeapElem (addr, htype)) = false
  | equal_obs_eventa A_ B_ (WriteMem (addra, addr_loc, valaa))
    (ExternalCall (addr, literal, lista, vala)) = false
  | equal_obs_eventa A_ B_ (ExternalCall (addra, literal, lista, valaa))
    (WriteMem (addr, addr_loc, vala)) = false
  | equal_obs_eventa A_ B_ (ReadMem (addra, addr_loc, valaa))
    (ExternalCall (addr, literal, lista, vala)) = false
  | equal_obs_eventa A_ B_ (ExternalCall (addra, literal, lista, valaa))
    (ReadMem (addr, addr_loc, vala)) = false
  | equal_obs_eventa A_ B_ (ObsInterrupted thread_ida)
    (ObsInterrupted thread_id) = HOL.eq B_ thread_ida thread_id
  | equal_obs_eventa A_ B_ (ObsInterrupt thread_ida) (ObsInterrupt thread_id) =
    HOL.eq B_ thread_ida thread_id
  | equal_obs_eventa A_ B_ (SyncUnlock addra) (SyncUnlock addr) =
    HOL.eq A_ addra addr
  | equal_obs_eventa A_ B_ (SyncLock addra) (SyncLock addr) =
    HOL.eq A_ addra addr
  | equal_obs_eventa A_ B_ (ThreadJoin thread_ida) (ThreadJoin thread_id) =
    HOL.eq B_ thread_ida thread_id
  | equal_obs_eventa A_ B_ (ThreadStart thread_ida) (ThreadStart thread_id) =
    HOL.eq B_ thread_ida thread_id
  | equal_obs_eventa A_ B_ (NewHeapElem (addra, htypea))
    (NewHeapElem (addr, htype)) =
    HOL.eq A_ addra addr andalso Type.equal_htype htypea htype
  | equal_obs_eventa A_ B_ (WriteMem (addra, addr_loca, valaa))
    (WriteMem (addr, addr_loc, vala)) =
    HOL.eq A_ addra addr andalso
      (Heap.equal_addr_loc addr_loca addr_loc andalso
        Value.equal_vala A_ valaa vala)
  | equal_obs_eventa A_ B_ (ReadMem (addra, addr_loca, valaa))
    (ReadMem (addr, addr_loc, vala)) =
    HOL.eq A_ addra addr andalso
      (Heap.equal_addr_loc addr_loca addr_loc andalso
        Value.equal_vala A_ valaa vala)
  | equal_obs_eventa A_ B_ (ExternalCall (addra, literala, listaa, valaa))
    (ExternalCall (addr, literal, lista, vala)) =
    HOL.eq A_ addra addr andalso
      (((literala : string) = literal) andalso
        (List.equal_lista (Value.equal_val A_) listaa lista andalso
          Value.equal_vala A_ valaa vala));

fun equal_obs_event A_ B_ = {equal = equal_obs_eventa A_ B_} :
  ('a, 'b) obs_event HOL.equal;

val obs_action_obs_event = {} : ('a, 'b) obs_event FWState.obs_action;

end; (*struct Observable_Events*)

structure TrieMapImpl : sig
  val iteratei_bmap_op_list_it_tm_basic_ops :
    ('a, 'b) Trie.trie -> ('c -> bool) -> ('a list * 'b -> 'c -> 'c) -> 'c -> 'c
  val g_to_list_tm_basic_ops : ('a, 'b) Trie.trie -> ('a list * 'b) list
end = struct

fun iteratei_bmap_op_list_it_tm_basic_ops s = Trie.iteratei s;

fun g_to_list_tm_basic_ops m =
  iteratei_bmap_op_list_it_tm_basic_ops m (fn _ => true)
    (fn a => fn b => a :: b) [];

end; (*struct TrieMapImpl*)

structure RBTMapImpl : sig
  val rev_iterateoi_bmap_op_rev_list_it_rm_basic_ops :
    'a Orderings.linorder ->
      ('a, 'b) RBT.rbt -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
  val g_max_rm_basic_ops :
    'a Orderings.linorder ->
      ('a, 'b) RBT.rbt -> ('a * 'b -> bool) -> ('a * 'b) option
  val iteratei_bmap_op_list_it_rm_basic_ops :
    'a Orderings.linorder ->
      ('a, 'b) RBT.rbt -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
  val g_sel_rm_basic_ops :
    'a Orderings.linorder ->
      ('a, 'b) RBT.rbt -> ('a * 'b -> bool) -> ('a * 'b) option
  val g_to_list_rm_basic_ops :
    'a Orderings.linorder -> ('a, 'b) RBT.rbt -> ('a * 'b) list
  val iteratei_map_op_list_it_rm_ops :
    'a Orderings.linorder ->
      ('a, 'b) RBT.rbt -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
end = struct

fun rev_iterateoi_bmap_op_rev_list_it_rm_basic_ops A_ s =
  RBT_add.rm_reverse_iterateoi (RBT.impl_of A_ s);

fun g_max_rm_basic_ops A_ m p =
  rev_iterateoi_bmap_op_rev_list_it_rm_basic_ops A_ m Option.is_none
    (fn x => fn _ => (if p x then SOME x else NONE)) NONE;

fun iteratei_bmap_op_list_it_rm_basic_ops A_ s =
  RBT_add.rm_iterateoi (RBT.impl_of A_ s);

fun g_sel_rm_basic_ops A_ m p =
  iteratei_bmap_op_list_it_rm_basic_ops A_ m Option.is_none
    (fn x => fn _ => (if p x then SOME x else NONE)) NONE;

fun g_to_list_rm_basic_ops A_ m =
  iteratei_bmap_op_list_it_rm_basic_ops A_ m (fn _ => true)
    (fn a => fn b => a :: b) [];

fun iteratei_map_op_list_it_rm_ops A_ s =
  RBT_add.rm_iterateoi (RBT.impl_of A_ s);

end; (*struct RBTMapImpl*)

structure ToString : sig
  type 'a toString
  val toString : 'a toString -> 'a -> string
  val digit_toString : Arith.int -> string
  val int_toString : Arith.int -> string list
  val toString_int : Arith.int -> string
  val toString_nata : Arith.nat -> string
  val toString_nat : Arith.nat toString
  val toString_literala : string -> string
  val toString_ty : Type.ty -> string
  val toString_word : 'a Type_Length.len -> 'a Word.word -> string
  val toString_bool : bool -> string
  val toString_vala : 'a toString -> 'a Value.vala -> string
  val toString_val : 'a toString -> 'a Value.vala toString
  val toString_bop : BinOp.bop -> string
  val toString_chara : char -> string
  val toString_char : char toString
  val list_toString : 'a toString -> string -> 'a list -> string list
  val toString_lista : 'a toString -> 'a list -> string
  val toString_list : 'a toString -> ('a list) toString
  val toString_proda : 'a toString -> 'b toString -> 'a * 'b -> string
  val toString_prod : 'a toString -> 'b toString -> ('a * 'b) toString
  val toString_rbt :
    'a Orderings.linorder * 'a toString -> 'b toString ->
      ('a, 'b) RBT.rbt -> string
  val toString_instra : 'a toString -> 'a JVMInstructions.instr -> string
  val toString_instr : 'a toString -> 'a JVMInstructions.instr toString
  val toString_trie : 'a toString -> 'b toString -> ('a, 'b) Trie.trie -> string
  val toString_finfuna :
    'a Set.card_UNIV * 'a HOL.equal * 'a Orderings.linorder * 'a toString ->
      'b HOL.equal * 'b toString -> ('a, 'b) FinFun.finfun -> string
  val toString_finfun :
    'a Set.card_UNIV * 'a HOL.equal * 'a Orderings.linorder * 'a toString ->
      'b HOL.equal * 'b toString -> ('a, 'b) FinFun.finfun toString
  val toString_htype : Type.htype -> string
  val toString_optiona : 'a toString -> 'a option -> string
  val toString_option : 'a toString -> ('a option) toString
  val toString_literal : string toString
  val toString_addr_loc : Heap.addr_loc -> string
  val toString_obs_eventa :
    'a toString -> 'b toString -> ('a, 'b) Observable_Events.obs_event -> string
  val toString_obs_event :
    'a toString -> 'b toString -> ('a, 'b) Observable_Events.obs_event toString
  val toString_assoc_lista :
    'a toString -> 'b toString -> ('a, 'b) Assoc_List.assoc_list -> string
  val toString_assoc_list :
    'a toString -> 'b toString -> ('a, 'b) Assoc_List.assoc_list toString
end = struct

type 'a toString = {toString : 'a -> string};
val toString = #toString : 'a toString -> 'a -> string;

fun digit_toString k =
  (if Arith.equal_int k Arith.zero_inta then "0"
    else (if Arith.equal_int k (Arith.Int_of_integer (1 : IntInf.int)) then "1"
           else (if Arith.equal_int k (Arith.Int_of_integer (2 : IntInf.int))
                  then "2"
                  else (if Arith.equal_int k
                             (Arith.Int_of_integer (3 : IntInf.int))
                         then "3"
                         else (if Arith.equal_int k
                                    (Arith.Int_of_integer (4 : IntInf.int))
                                then "4"
                                else (if Arith.equal_int k
   (Arith.Int_of_integer (5 : IntInf.int))
                                       then "5"
                                       else (if Arith.equal_int k
          (Arith.Int_of_integer (6 : IntInf.int))
      then "6"
      else (if Arith.equal_int k (Arith.Int_of_integer (7 : IntInf.int))
             then "7"
             else (if Arith.equal_int k (Arith.Int_of_integer (8 : IntInf.int))
                    then "8"
                    else (if Arith.equal_int k
                               (Arith.Int_of_integer (9 : IntInf.int))
                           then "9" else (raise Fail "undefined")))))))))));

fun int_toString n =
  (if Arith.less_int n Arith.zero_inta
    then "-" :: int_toString (Arith.uminus_int n)
    else (if Arith.less_int n (Arith.Int_of_integer (10 : IntInf.int))
           then [digit_toString n]
           else int_toString
                  (Arith.div_int n (Arith.Int_of_integer (10 : IntInf.int))) @
                  [digit_toString
                     (Arith.mod_int n
                       (Arith.Int_of_integer (10 : IntInf.int)))]));

fun toString_int i = String.concat (int_toString i);

fun toString_nata n = toString_int (Arith.int_of_nat n);

val toString_nat = {toString = toString_nata} : Arith.nat toString;

fun toString_literala s = s;

fun toString_ty Type.Void = "Void"
  | toString_ty Type.Boolean = "Boolean"
  | toString_ty Type.Integer = "Integer"
  | toString_ty Type.Nt = "NT"
  | toString_ty (Type.Class c) = String.concat ["Class ", toString_literala c]
  | toString_ty (Type.Array t) = String.concat [toString_ty t, "[]"];

fun toString_word A_ w = toString_int (Word.sint A_ w);

fun toString_bool b = (case b of true => "True" | false => "False");

fun toString_vala A_ Value.Unit = "Unit"
  | toString_vala A_ Value.Null = "Null"
  | toString_vala A_ (Value.Bool b) = String.concat ["Bool ", toString_bool b]
  | toString_vala A_ (Value.Intg i) =
    String.concat
      ["Intg ",
        toString_word
          (Type_Length.len_bit0
            (Type_Length.len_bit0
              (Type_Length.len_bit0
                (Type_Length.len_bit0
                  (Type_Length.len_bit0 Type_Length.len_num1)))))
          i]
  | toString_vala A_ (Value.Addr a) = String.concat ["Addr ", toString A_ a];

fun toString_val A_ = {toString = toString_vala A_} : 'a Value.vala toString;

fun toString_bop BinOp.Eq = "=="
  | toString_bop BinOp.NotEq = "!="
  | toString_bop BinOp.LessThan = "<"
  | toString_bop BinOp.LessOrEqual = "<="
  | toString_bop BinOp.GreaterThan = ">"
  | toString_bop BinOp.GreaterOrEqual = ">="
  | toString_bop BinOp.Add = "+"
  | toString_bop BinOp.Subtract = "-"
  | toString_bop BinOp.Mult = "*"
  | toString_bop BinOp.Div = "/"
  | toString_bop BinOp.Mod = "%"
  | toString_bop BinOp.BinAnd = "&"
  | toString_bop BinOp.BinOr = "|"
  | toString_bop BinOp.BinXor = "^"
  | toString_bop BinOp.ShiftLeft = "<<"
  | toString_bop BinOp.ShiftRightZeros = ">>"
  | toString_bop BinOp.ShiftRightSigned = ">>>";

fun toString_chara c = String.implode [c];

val toString_char = {toString = toString_chara} : char toString;

fun list_toString A_ sep [] = []
  | list_toString A_ sep [x] = [toString A_ x]
  | list_toString A_ sep (x :: v :: va) =
    toString A_ x :: sep :: list_toString A_ sep (v :: va);

fun toString_lista A_ xs =
  String.concat ("[" :: list_toString A_ "," xs @ ["]"]);

fun toString_list A_ = {toString = toString_lista A_} : ('a list) toString;

fun toString_proda A_ B_ =
  (fn (a, b) => String.concat ["(", toString A_ a, ", ", toString B_ b, ")"]);

fun toString_prod A_ B_ = {toString = toString_proda A_ B_} :
  ('a * 'b) toString;

fun toString_rbt (A1_, A2_) B_ t =
  String.concat
    (list_toString (toString_prod A2_ B_) ",\n"
      (RBTMapImpl.g_to_list_rm_basic_ops A1_ t));

fun toString_instra A_ (JVMInstructions.Load i) =
  String.concat ["Load (", toString_nata i, ")"]
  | toString_instra A_ (JVMInstructions.Store i) =
    String.concat ["Store (", toString_nata i, ")"]
  | toString_instra A_ (JVMInstructions.Push v) =
    String.concat ["Push (", toString_vala A_ v, ")"]
  | toString_instra A_ (JVMInstructions.New c) =
    String.concat ["New ", toString_literala c]
  | toString_instra A_ (JVMInstructions.NewArray t) =
    String.concat ["NewArray ", toString_ty t]
  | toString_instra A_ JVMInstructions.ALoad = "ALoad"
  | toString_instra A_ JVMInstructions.AStore = "AStore"
  | toString_instra A_ JVMInstructions.ALength = "ALength"
  | toString_instra A_ (JVMInstructions.Getfield (f, d)) =
    String.concat ["Getfield  ", toString_literala f, " ", toString_literala d]
  | toString_instra A_ (JVMInstructions.Putfield (f, d)) =
    String.concat ["Putfield  ", toString_literala f, " ", toString_literala d]
  | toString_instra A_ (JVMInstructions.Checkcast t) =
    String.concat ["Checkcast ", toString_ty t]
  | toString_instra A_ (JVMInstructions.Instanceof t) =
    String.concat ["Instanceof ", toString_ty t]
  | toString_instra A_ (JVMInstructions.Invoke (m, n)) =
    String.concat ["Invoke ", toString_literala m, " ", toString_nata n]
  | toString_instra A_ JVMInstructions.Return = "Return"
  | toString_instra A_ JVMInstructions.Pop = "Pop"
  | toString_instra A_ JVMInstructions.Dup = "Dup"
  | toString_instra A_ JVMInstructions.Swap = "Swap"
  | toString_instra A_ (JVMInstructions.BinOpInstr bop) =
    String.concat ["BinOpInstr  ", toString_bop bop]
  | toString_instra A_ (JVMInstructions.Goto i) =
    String.concat ["Goto ", toString_int i]
  | toString_instra A_ (JVMInstructions.IfFalse i) =
    String.concat ["IfFalse ", toString_int i]
  | toString_instra A_ JVMInstructions.ThrowExc = "ThrowExc"
  | toString_instra A_ JVMInstructions.MEnter = "monitorenter"
  | toString_instra A_ JVMInstructions.MExit = "monitorexit";

fun toString_instr A_ = {toString = toString_instra A_} :
  'a JVMInstructions.instr toString;

fun toString_trie A_ B_ t =
  toString_lista (toString_prod (toString_list A_) B_)
    (TrieMapImpl.g_to_list_tm_basic_ops t);

fun toString_finfuna (A1_, A2_, A3_, A4_) (B1_, B2_) f =
  String.concat
    ("(" ::
      toString B2_ (FinFun.finfun_default (A1_, A2_) f) ::
        List.maps
          (fn x =>
            [",", toString A4_ x, "|->",
              toString B2_ (FinFun.finfun_apply A2_ f x)])
          (FinFun.finfun_to_list (A1_, A2_, A3_) B1_ f) @
          [")"]);

fun toString_finfun (A1_, A2_, A3_, A4_) (B1_, B2_) =
  {toString = toString_finfuna (A1_, A2_, A3_, A4_) (B1_, B2_)} :
  ('a, 'b) FinFun.finfun toString;

fun toString_htype (Type.Class_type c) = c
  | toString_htype (Type.Array_type (t, n)) =
    String.concat [toString_ty t, "[", toString_nata n, "]"];

fun toString_optiona A_ NONE = "None"
  | toString_optiona A_ (SOME a) = String.concat ["Some (", toString A_ a, ")"];

fun toString_option A_ = {toString = toString_optiona A_} :
  ('a option) toString;

val toString_literal = {toString = toString_literala} : string toString;

fun toString_addr_loc (Heap.CField (c, f)) =
  String.concat ["CField ", f, "{", c, "}"]
  | toString_addr_loc (Heap.ACell n) =
    String.concat ["ACell ", toString_nata n];

fun toString_obs_eventa A_ B_ (Observable_Events.ExternalCall (ad, m, vs, v)) =
  String.concat
    ["ExternalCall ", m, "(", toString_lista (toString_val A_) vs, ") = ",
      toString_vala A_ v]
  | toString_obs_eventa A_ B_ (Observable_Events.ReadMem (ad, al, v)) =
    String.concat
      ["ReadMem ", toString A_ ad, "@", toString_addr_loc al, "=",
        toString_vala A_ v]
  | toString_obs_eventa A_ B_ (Observable_Events.WriteMem (ad, al, v)) =
    String.concat
      ["WriteMem ", toString A_ ad, "@", toString_addr_loc al, "=",
        toString_vala A_ v]
  | toString_obs_eventa A_ B_ (Observable_Events.NewHeapElem (ad, hT)) =
    String.concat ["Allocate ", toString A_ ad, ":", toString_htype hT]
  | toString_obs_eventa A_ B_ (Observable_Events.ThreadStart t) =
    String.concat ["ThreadStart ", toString B_ t]
  | toString_obs_eventa A_ B_ (Observable_Events.ThreadJoin t) =
    String.concat ["ThreadJoin ", toString B_ t]
  | toString_obs_eventa A_ B_ (Observable_Events.SyncLock ad) =
    String.concat ["SyncLock ", toString A_ ad]
  | toString_obs_eventa A_ B_ (Observable_Events.SyncUnlock ad) =
    String.concat ["SyncUnlock ", toString A_ ad]
  | toString_obs_eventa A_ B_ (Observable_Events.ObsInterrupt t) =
    String.concat ["Interrupt ", toString B_ t]
  | toString_obs_eventa A_ B_ (Observable_Events.ObsInterrupted t) =
    String.concat ["Interrupted ", toString B_ t];

fun toString_obs_event A_ B_ = {toString = toString_obs_eventa A_ B_} :
  ('a, 'b) Observable_Events.obs_event toString;

fun toString_assoc_lista A_ B_ =
  toString_lista (toString_prod A_ B_) o Assoc_List.impl_of;

fun toString_assoc_list A_ B_ = {toString = toString_assoc_lista A_ B_} :
  ('a, 'b) Assoc_List.assoc_list toString;

end; (*struct ToString*)

structure FWLocking : sig
  val redT_updLs :
    'a HOL.equal -> 'b HOL.equal ->
      ('a, (('b * Arith.nat) option)) FinFun.finfun ->
        'b -> ('a, (FWState.lock_action list)) FinFun.finfun ->
                ('a, (('b * Arith.nat) option)) FinFun.finfun
  val acquire_all :
    'a HOL.equal -> 'b HOL.equal ->
      ('a, (('b * Arith.nat) option)) FinFun.finfun ->
        'b -> ('a, Arith.nat) FinFun.finfun ->
                ('a, (('b * Arith.nat) option)) FinFun.finfun
  val lock_ok_las :
    'a Set.card_UNIV * 'a HOL.equal -> 'b HOL.equal ->
      ('a, (('b * Arith.nat) option)) FinFun.finfun ->
        'b -> ('a, (FWState.lock_action list)) FinFun.finfun -> bool
  val may_acquire_all :
    'a Set.card_UNIV * 'a HOL.equal -> 'b HOL.equal ->
      ('a, (('b * Arith.nat) option)) FinFun.finfun ->
        'b -> ('a, Arith.nat) FinFun.finfun -> bool
end = struct

fun redT_updLs A_ B_ ls t las =
  FinFun.finfun_comp (fn (l, a) => FWLock.upd_locks B_ l t a)
    (FinFun.finfun_Diag A_
      (Option.equal_option (Product_Type.equal_prod B_ Arith.equal_nat))
      (List.equal_list FWState.equal_lock_action) ls las);

fun acquire_all A_ B_ ls t ln =
  FinFun.finfun_comp (fn (l, a) => FWLock.acquire_locks l t a)
    (FinFun.finfun_Diag A_
      (Option.equal_option (Product_Type.equal_prod B_ Arith.equal_nat))
      Arith.equal_nat ls ln);

fun lock_ok_las (A1_, A2_) B_ ls t las =
  FinFun.finfun_All (A1_, A2_)
    (FinFun.finfun_comp (fn (l, a) => FWLock.lock_actions_ok B_ l t a)
      (FinFun.finfun_Diag A2_
        (Option.equal_option (Product_Type.equal_prod B_ Arith.equal_nat))
        (List.equal_list FWState.equal_lock_action) ls las));

fun may_acquire_all (A1_, A2_) B_ ls t ln =
  FinFun.finfun_All (A1_, A2_)
    (FinFun.finfun_comp
      (fn (lock, n) =>
        (if Arith.less_nat Arith.zero_nat n then FWLock.may_lock B_ lock t
          else true))
      (FinFun.finfun_Diag A2_
        (Option.equal_option (Product_Type.equal_prod B_ Arith.equal_nat))
        Arith.equal_nat ls ln));

end; (*struct FWLocking*)

structure FWLockingThread : sig
  val upd_threadR :
    'a HOL.equal ->
      Arith.nat ->
        ('a * Arith.nat) option -> 'a -> FWState.lock_action -> Arith.nat
  val upd_threadRs :
    'a HOL.equal ->
      Arith.nat ->
        ('a * Arith.nat) option -> 'a -> FWState.lock_action list -> Arith.nat
  val redT_updLns :
    'a HOL.equal -> 'b HOL.equal ->
      ('a, (('b * Arith.nat) option)) FinFun.finfun ->
        'b -> ('a, Arith.nat) FinFun.finfun ->
                ('a, (FWState.lock_action list)) FinFun.finfun ->
                  ('a, Arith.nat) FinFun.finfun
end = struct

fun upd_threadR A_ n l t FWState.ReleaseAcquire =
  Arith.plus_nat n (FWLock.has_locks A_ l t)
  | upd_threadR A_ n l t FWState.Lock = n
  | upd_threadR A_ n l t FWState.Unlock = n
  | upd_threadR A_ n l t FWState.UnlockFail = n;

fun upd_threadRs A_ n l t [] = n
  | upd_threadRs A_ n l t (la :: las) =
    upd_threadRs A_ (upd_threadR A_ n l t la) (FWLock.upd_lock A_ l t la) t las;

fun redT_updLns A_ B_ ls t ln las =
  FinFun.finfun_comp (fn (l, (n, a)) => upd_threadRs B_ n l t a)
    (FinFun.finfun_Diag A_
      (Option.equal_option (Product_Type.equal_prod B_ Arith.equal_nat))
      (Product_Type.equal_prod Arith.equal_nat
        (List.equal_list FWState.equal_lock_action))
      ls (FinFun.finfun_Diag A_ Arith.equal_nat
           (List.equal_list FWState.equal_lock_action) ln las));

end; (*struct FWLockingThread*)

structure Scheduler : sig
  val tllist_unfold :
    ('a -> (('b * 'a), 'c) Sum_Type.sum) -> 'a -> ('b, 'c) TLList.tllist
  val redT_updT :
    ('a -> 'b * ('c, Arith.nat) FinFun.finfun -> 'd -> 'd) ->
      'd -> ('a, 'b, 'e) FWState.new_thread_action -> 'd
  val redT_updTs :
    ('a -> 'b * ('c, Arith.nat) FinFun.finfun -> 'd -> 'd) ->
      'd -> ('a, 'b, 'e) FWState.new_thread_action list -> 'd
  val redT_updI :
    ('a -> 'b -> 'b) ->
      ('a -> 'b -> 'b) -> 'b -> 'a FWState.interrupt_action -> 'b
  val redT_updIs :
    ('a -> 'b -> 'b) ->
      ('a -> 'b -> 'b) -> 'b -> 'a FWState.interrupt_action list -> 'b
  val exec_updW :
    'c HOL.equal ->
      ('a -> 'b -> 'c -> 'd -> 'b option) ->
        ('b -> 'd -> 'c FWState.wait_set_status option) ->
          ('b -> 'c FWState.wait_set_status -> 'd -> 'd) ->
            ('b -> 'd -> 'd) ->
              ('d ->
                ('d -> bool) ->
                  ('b * 'c FWState.wait_set_status -> 'd -> 'd) -> 'd -> 'd) ->
                'a -> 'b -> 'd -> ('b, 'c) FWState.wait_set_action -> 'd
  val exec_updWs :
    'c HOL.equal ->
      ('a -> 'b -> 'c -> 'd -> 'b option) ->
        ('b -> 'd -> 'c FWState.wait_set_status option) ->
          ('b -> 'c FWState.wait_set_status -> 'd -> 'd) ->
            ('b -> 'd -> 'd) ->
              ('d ->
                ('d -> bool) ->
                  ('b * 'c FWState.wait_set_status -> 'd -> 'd) -> 'd -> 'd) ->
                'a -> 'b -> 'd -> ('b, 'c) FWState.wait_set_action list -> 'd
  val exec_upd :
    'b HOL.equal -> 'c HOL.equal -> 'g HOL.equal ->
      ('a -> 'b -> 'c -> 'd -> 'b option) ->
        ('b -> 'e -> ('f * ('g, Arith.nat) FinFun.finfun) option) ->
          ('b -> 'f * ('g, Arith.nat) FinFun.finfun -> 'e -> 'e) ->
            ('b -> 'd -> 'c FWState.wait_set_status option) ->
              ('b -> 'c FWState.wait_set_status -> 'd -> 'd) ->
                ('b -> 'd -> 'd) ->
                  ('d ->
                    ('d -> bool) ->
                      ('b * 'c FWState.wait_set_status -> 'd -> 'd) ->
                        'd -> 'd) ->
                    ('b -> 'h -> 'h) ->
                      ('b -> 'h -> 'h) ->
                        'a -> ('g, (('b * Arith.nat) option)) FinFun.finfun *
                                (('e * 'i) * ('d * 'h)) ->
                                'b -> ('g, (FWState.lock_action list))
FinFun.finfun *
(('b, 'f, 'i) FWState.new_thread_action list *
  ('b FWState.conditional_action list *
    (('b, 'c) FWState.wait_set_action list *
      ('b FWState.interrupt_action list * 'j list)))) ->
'f -> 'i -> ('g, (('b * Arith.nat) option)) FinFun.finfun *
              (('e * 'i) * ('d * 'h))
  val execT :
    'a HOL.equal -> 'd HOL.equal -> 'j HOL.equal ->
      (('a, Arith.nat) FinFun.finfun -> 'b list) ->
        ('c ->
          ('a, (('d * Arith.nat) option)) FinFun.finfun *
            (('e * 'f) * ('g * 'h)) ->
            ('d * (((('a, (FWState.lock_action list)) FinFun.finfun *
                      (('d, 'i, 'f) FWState.new_thread_action list *
                        ('d FWState.conditional_action list *
                          (('d, 'j) FWState.wait_set_action list *
                            ('d FWState.interrupt_action list * 'b list))))) *
                     ('i * 'f)) option *
                    'c)) option) ->
          ('c -> 'd -> 'j -> 'g -> 'd option) ->
            ('d -> 'e -> ('i * ('a, Arith.nat) FinFun.finfun) option) ->
              ('d -> 'i * ('a, Arith.nat) FinFun.finfun -> 'e -> 'e) ->
                ('d -> 'g -> 'j FWState.wait_set_status option) ->
                  ('d -> 'j FWState.wait_set_status -> 'g -> 'g) ->
                    ('d -> 'g -> 'g) ->
                      ('g ->
                        ('g -> bool) ->
                          ('d * 'j FWState.wait_set_status -> 'g -> 'g) ->
                            'g -> 'g) ->
                        ('d -> 'h -> 'h) ->
                          ('d -> 'h -> 'h) ->
                            'c -> ('a, (('d * Arith.nat) option))
                                    FinFun.finfun *
                                    (('e * 'f) * ('g * 'h)) ->
                                    ('c * ('d *
    ((('a, (FWState.lock_action list)) FinFun.finfun *
       (('d, 'i, 'f) FWState.new_thread_action list *
         ('d FWState.conditional_action list *
           (('d, 'j) FWState.wait_set_action list *
             ('d FWState.interrupt_action list * 'b list))))) *
      (('a, (('d * Arith.nat) option)) FinFun.finfun *
        (('e * 'f) * ('g * 'h)))))) option
  val exec_step :
    'a HOL.equal -> 'd HOL.equal -> 'j HOL.equal ->
      (('a, Arith.nat) FinFun.finfun -> 'b list) ->
        ('c ->
          ('a, (('d * Arith.nat) option)) FinFun.finfun *
            (('e * 'f) * ('g * 'h)) ->
            ('d * (((('a, (FWState.lock_action list)) FinFun.finfun *
                      (('d, 'i, 'f) FWState.new_thread_action list *
                        ('d FWState.conditional_action list *
                          (('d, 'j) FWState.wait_set_action list *
                            ('d FWState.interrupt_action list * 'b list))))) *
                     ('i * 'f)) option *
                    'c)) option) ->
          ('c -> 'd -> 'j -> 'g -> 'd option) ->
            ('d -> 'e -> ('i * ('a, Arith.nat) FinFun.finfun) option) ->
              ('d -> 'i * ('a, Arith.nat) FinFun.finfun -> 'e -> 'e) ->
                ('d -> 'g -> 'j FWState.wait_set_status option) ->
                  ('d -> 'j FWState.wait_set_status -> 'g -> 'g) ->
                    ('d -> 'g -> 'g) ->
                      ('g ->
                        ('g -> bool) ->
                          ('d * 'j FWState.wait_set_status -> 'g -> 'g) ->
                            'g -> 'g) ->
                        ('d -> 'h -> 'h) ->
                          ('d -> 'h -> 'h) ->
                            'c * (('a, (('d * Arith.nat) option))
                                    FinFun.finfun *
                                   (('e * 'f) * ('g * 'h))) ->
                              ((('c * ('d *
(('a, (FWState.lock_action list)) FinFun.finfun *
  (('d, 'i, 'f) FWState.new_thread_action list *
    ('d FWState.conditional_action list *
      (('d, 'j) FWState.wait_set_action list *
        ('d FWState.interrupt_action list * 'b list))))))) *
                                 ('c * (('a, (('d * Arith.nat) option))
  FinFun.finfun *
 (('e * 'f) * ('g * 'h))))),
                                (('a, (('d * Arith.nat) option)) FinFun.finfun *
                                  (('e * 'f) * ('g * 'h))))
                                Sum_Type.sum
  val exec_aux :
    'a HOL.equal -> 'd HOL.equal -> 'j HOL.equal ->
      (('a, Arith.nat) FinFun.finfun -> 'b list) ->
        ('c ->
          ('a, (('d * Arith.nat) option)) FinFun.finfun *
            (('e * 'f) * ('g * 'h)) ->
            ('d * (((('a, (FWState.lock_action list)) FinFun.finfun *
                      (('d, 'i, 'f) FWState.new_thread_action list *
                        ('d FWState.conditional_action list *
                          (('d, 'j) FWState.wait_set_action list *
                            ('d FWState.interrupt_action list * 'b list))))) *
                     ('i * 'f)) option *
                    'c)) option) ->
          ('c -> 'd -> 'j -> 'g -> 'd option) ->
            ('d -> 'e -> ('i * ('a, Arith.nat) FinFun.finfun) option) ->
              ('d -> 'i * ('a, Arith.nat) FinFun.finfun -> 'e -> 'e) ->
                ('d -> 'g -> 'j FWState.wait_set_status option) ->
                  ('d -> 'j FWState.wait_set_status -> 'g -> 'g) ->
                    ('d -> 'g -> 'g) ->
                      ('g ->
                        ('g -> bool) ->
                          ('d * 'j FWState.wait_set_status -> 'g -> 'g) ->
                            'g -> 'g) ->
                        ('d -> 'h -> 'h) ->
                          ('d -> 'h -> 'h) ->
                            'c * (('a, (('d * Arith.nat) option))
                                    FinFun.finfun *
                                   (('e * 'f) * ('g * 'h))) ->
                              (('c * ('d * (('a, (FWState.lock_action list))
      FinFun.finfun *
     (('d, 'i, 'f) FWState.new_thread_action list *
       ('d FWState.conditional_action list *
         (('d, 'j) FWState.wait_set_action list *
           ('d FWState.interrupt_action list * 'b list))))))),
                                (('a, (('d * Arith.nat) option)) FinFun.finfun *
                                  (('e * 'f) * ('g * 'h))))
                                TLList.tllist
  val exec :
    'a HOL.equal -> 'd HOL.equal -> 'j HOL.equal ->
      (('a, Arith.nat) FinFun.finfun -> 'b list) ->
        ('c ->
          ('a, (('d * Arith.nat) option)) FinFun.finfun *
            (('e * 'f) * ('g * 'h)) ->
            ('d * (((('a, (FWState.lock_action list)) FinFun.finfun *
                      (('d, 'i, 'f) FWState.new_thread_action list *
                        ('d FWState.conditional_action list *
                          (('d, 'j) FWState.wait_set_action list *
                            ('d FWState.interrupt_action list * 'b list))))) *
                     ('i * 'f)) option *
                    'c)) option) ->
          ('c ->
            'd -> ('a, (FWState.lock_action list)) FinFun.finfun *
                    (('d, 'i, 'f) FWState.new_thread_action list *
                      ('d FWState.conditional_action list *
                        (('d, 'j) FWState.wait_set_action list *
                          ('d FWState.interrupt_action list * 'b list)))) ->
                    'k option) ->
            ('c -> 'd -> 'j -> 'g -> 'd option) ->
              ('d -> 'e -> ('i * ('a, Arith.nat) FinFun.finfun) option) ->
                ('d -> 'i * ('a, Arith.nat) FinFun.finfun -> 'e -> 'e) ->
                  ('d -> 'g -> 'j FWState.wait_set_status option) ->
                    ('d -> 'j FWState.wait_set_status -> 'g -> 'g) ->
                      ('d -> 'g -> 'g) ->
                        ('g ->
                          ('g -> bool) ->
                            ('d * 'j FWState.wait_set_status -> 'g -> 'g) ->
                              'g -> 'g) ->
                          ('d -> 'h -> 'h) ->
                            ('d -> 'h -> 'h) ->
                              'c -> ('a, (('d * Arith.nat) option))
                                      FinFun.finfun *
                                      (('e * 'f) * ('g * 'h)) ->
                                      ('k,
(('a, (('d * Arith.nat) option)) FinFun.finfun * (('e * 'f) * ('g * 'h))))
TLList.tllist
  val singleton2 : (unit -> 'a) -> 'a Predicate.pred -> 'a
  val the_only2 : (unit -> 'a) -> 'a Predicate.seq -> 'a
  val the2 : 'a Predicate.pred -> 'a
  val free_thread_id :
    ('a -> 'b -> ('c * ('d, Arith.nat) FinFun.finfun) option) ->
      'b -> 'a -> bool
  val thread_ok :
    ('a -> 'b -> ('c * ('d, Arith.nat) FinFun.finfun) option) ->
      'b -> ('a, 'c, 'e) FWState.new_thread_action -> bool
  val interrupt_action_ok :
    ('a -> 'b -> bool) -> 'b -> 'a FWState.interrupt_action -> bool
  val interrupt_actions_ok :
    ('a -> 'b -> bool) ->
      ('a -> 'b -> 'b) ->
        ('a -> 'b -> 'b) -> 'b -> 'a FWState.interrupt_action list -> bool
  val wset_actions_ok :
    'a HOL.equal -> 'c HOL.equal ->
      ('a -> 'b -> 'c FWState.wait_set_status option) ->
        'b -> 'a -> ('a, 'c) FWState.wait_set_action list -> bool
  val cond_action_ok :
    'b HOL.equal -> 'd Set.card_UNIV * 'd HOL.equal ->
      ('a -> bool) ->
        ('b -> 'c -> ('a * ('d, Arith.nat) FinFun.finfun) option) ->
          ('b -> 'e -> 'f FWState.wait_set_status option) ->
            ('d, (('b * Arith.nat) option)) FinFun.finfun *
              (('c * 'g) * ('e * 'h)) ->
              'b -> 'b FWState.conditional_action -> bool
  val cond_action_oks :
    'b HOL.equal -> 'd Set.card_UNIV * 'd HOL.equal ->
      ('a -> bool) ->
        ('b -> 'c -> ('a * ('d, Arith.nat) FinFun.finfun) option) ->
          ('b -> 'e -> 'f FWState.wait_set_status option) ->
            ('d, (('b * Arith.nat) option)) FinFun.finfun *
              (('c * 'g) * ('e * 'h)) ->
              'b -> 'b FWState.conditional_action list -> bool
  val thread_oks :
    ('a -> 'b -> ('c * ('d, Arith.nat) FinFun.finfun) option) ->
      ('a -> 'c * ('d, Arith.nat) FinFun.finfun -> 'b -> 'b) ->
        'b -> ('a, 'c, 'e) FWState.new_thread_action list -> bool
  val actions_ok :
    'b HOL.equal -> 'd Set.card_UNIV * 'd HOL.equal -> 'f HOL.equal ->
      ('a -> bool) ->
        ('b -> 'c -> ('a * ('d, Arith.nat) FinFun.finfun) option) ->
          ('b -> 'a * ('d, Arith.nat) FinFun.finfun -> 'c -> 'c) ->
            ('b -> 'e -> 'f FWState.wait_set_status option) ->
              ('b -> 'g -> bool) ->
                ('b -> 'g -> 'g) ->
                  ('b -> 'g -> 'g) ->
                    ('d, (('b * Arith.nat) option)) FinFun.finfun *
                      (('c * 'h) * ('e * 'g)) ->
                      'b -> ('d, (FWState.lock_action list)) FinFun.finfun *
                              (('b, 'a, 'h) FWState.new_thread_action list *
                                ('b FWState.conditional_action list *
                                  (('b, 'f) FWState.wait_set_action list *
                                    ('b FWState.interrupt_action list *
                                      'i list)))) ->
                              bool
  val step_thread :
    'b HOL.equal -> 'd Set.card_UNIV * 'd HOL.equal -> 'e HOL.equal ->
      ('a -> bool) ->
        ('b ->
          'a * 'c ->
            ((('d, (FWState.lock_action list)) FinFun.finfun *
               (('b, 'a, 'c) FWState.new_thread_action list *
                 ('b FWState.conditional_action list *
                   (('b, 'e) FWState.wait_set_action list *
                     ('b FWState.interrupt_action list * 'f list))))) *
              ('a * 'c))
              Predicate.pred) ->
          (('d, Arith.nat) FinFun.finfun -> 'f list) ->
            ('b -> 'g -> ('a * ('d, Arith.nat) FinFun.finfun) option) ->
              ('b -> 'a * ('d, Arith.nat) FinFun.finfun -> 'g -> 'g) ->
                ('b -> 'h -> 'e FWState.wait_set_status option) ->
                  ('b -> 'i -> bool) ->
                    ('b -> 'i -> 'i) ->
                      ('b -> 'i -> 'i) ->
                        (('d, (FWState.lock_action list)) FinFun.finfun *
                           (('b, 'a, 'c) FWState.new_thread_action list *
                             ('b FWState.conditional_action list *
                               (('b, 'e) FWState.wait_set_action list *
                                 ('b FWState.interrupt_action list *
                                   'f list)))) ->
                          'j) ->
                          ('d, (('b * Arith.nat) option)) FinFun.finfun *
                            (('g * 'c) * ('h * 'i)) ->
                            'b -> ('b * (((('d, (FWState.lock_action list))
     FinFun.finfun *
    (('b, 'a, 'c) FWState.new_thread_action list *
      ('b FWState.conditional_action list *
        (('b, 'e) FWState.wait_set_action list *
          ('b FWState.interrupt_action list * 'f list))))) *
   ('a * 'c)) option *
  'j)) option
  val active_threads :
    'b HOL.equal -> 'd Set.card_UNIV * 'd HOL.equal -> 'e HOL.equal ->
      ('a -> bool) ->
        ('b ->
          'a * 'c ->
            ((('d, (FWState.lock_action list)) FinFun.finfun *
               (('b, 'a, 'c) FWState.new_thread_action list *
                 ('b FWState.conditional_action list *
                   (('b, 'e) FWState.wait_set_action list *
                     ('b FWState.interrupt_action list * 'f list))))) *
              ('a * 'c))
              Predicate.pred) ->
          ('b -> 'g -> ('a * ('d, Arith.nat) FinFun.finfun) option) ->
            ('b -> 'a * ('d, Arith.nat) FinFun.finfun -> 'g -> 'g) ->
              ('g ->
                ('h -> bool) ->
                  ('b * ('a * ('d, Arith.nat) FinFun.finfun) -> 'h -> 'h) ->
                    'h -> 'h) ->
                ('b -> 'i -> 'e FWState.wait_set_status option) ->
                  ('b -> 'j -> bool) ->
                    ('b -> 'j -> 'j) ->
                      ('b -> 'j -> 'j) ->
                        (unit -> 'h) ->
                          ('b -> 'h -> 'h) ->
                            ('d, (('b * Arith.nat) option)) FinFun.finfun *
                              (('g * 'c) * ('i * 'j)) ->
                              'h
  val pick_wakeup_via_sel :
    'c HOL.equal ->
      ('a ->
        ('b -> 'c FWState.wait_set_status -> bool) ->
          ('b * 'c FWState.wait_set_status) option) ->
        'd -> 'b -> 'c -> 'a -> 'b option
end = struct

fun tllist_unfold f a =
  (case f a
    of Sum_Type.Inl (b, aa) =>
      TLList.Lazy_tllist (fn _ => Sum_Type.Inl (b, tllist_unfold f aa))
    | Sum_Type.Inr aa => TLList.tNil aa);

fun redT_updT thr_update ts (FWState.ThreadExists (v, va)) = ts
  | redT_updT thr_update ts (FWState.NewThread (t, x, m)) =
    thr_update t (x, FinFun.Finfun_const Arith.zero_nat) ts;

fun redT_updTs thr_update = List.foldl (redT_updT thr_update);

fun redT_updI is_ins is_delete is (FWState.IsInterrupted (t, b)) = is
  | redT_updI is_ins is_delete is (FWState.ClearInterrupt t) = is_delete t is
  | redT_updI is_ins is_delete is (FWState.Interrupt t) = is_ins t is;

fun redT_updIs is_ins is_delete is (ia :: ias) =
  redT_updIs is_ins is_delete (redT_updI is_ins is_delete is ia) ias
  | redT_updIs is_ins is_delete is [] = is;

fun exec_updW C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma t ws
  FWState.WokenUp = ws_delete t ws
  | exec_updW C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma t ws
    FWState.Notified = ws_delete t ws
  | exec_updW C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma ta
    ws (FWState.WakeUp t) =
    (case ws_lookup t ws of NONE => ws
      | SOME (FWState.InWS _) =>
        ws_update t (FWState.PostWS FWState.WSWokenUp) ws
      | SOME (FWState.PostWS _) => ws)
  | exec_updW C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma t ws
    (FWState.Suspend w) = ws_update t (FWState.InWS w) ws
  | exec_updW C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma t ws
    (FWState.NotifyAll w) =
    ws_iterate ws (fn _ => true)
      (fn (ta, wa) => fn wsa =>
        (if FWState.equal_wait_set_statusa C_ wa (FWState.InWS w)
          then ws_update ta (FWState.PostWS FWState.WSNotified) wsa else wsa))
      ws
  | exec_updW C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma t ws
    (FWState.Notify w) =
    (case pick_wakeup sigma t w ws of NONE => ws
      | SOME ta => ws_update ta (FWState.PostWS FWState.WSNotified) ws);

fun exec_updWs C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma t =
  List.foldl
    (exec_updW C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma t);

fun exec_upd B_ C_ G_ pick_wakeup thr_lookup thr_update ws_lookup ws_update
  ws_delete ws_iterate is_ins is_delete sigma s t ta x m =
  (FWLocking.redT_updLs G_ B_ (FWState.locks s) t (FWState.locks_a ta),
    ((thr_update t
        (x, FWLockingThread.redT_updLns G_ B_ (FWState.locks s) t
              (Product_Type.snd (Option.the (thr_lookup t (FWState.thr s))))
              (FWState.locks_a ta))
        (redT_updTs thr_update (FWState.thr s) (FWState.thr_a ta)),
       m),
      (exec_updWs C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma
         t (FWState.wset s) (FWState.wset_a ta),
        redT_updIs is_ins is_delete (FWState.interrupts s)
          (FWState.interrupt_a ta))));

fun execT A_ D_ J_ convert_RA schedule pick_wakeup thr_lookup thr_update
  ws_lookup ws_update ws_delete ws_iterate is_ins is_delete sigma s =
  Option.bind (schedule sigma s)
    (fn a =>
      (case a
        of (t, (NONE, sigmaa)) =>
          let
            val (x, ln) = Option.the (thr_lookup t (FWState.thr s));
            val ta =
              (FinFun.Finfun_const [], ([], ([], ([], ([], convert_RA ln)))));
            val sa =
              (FWLocking.acquire_all A_ D_ (FWState.locks s) t ln,
                ((thr_update t (x, FinFun.Finfun_const Arith.zero_nat)
                    (FWState.thr s),
                   FWState.shr s),
                  (FWState.wset s, FWState.interrupts s)));
          in
            SOME (sigmaa, (t, (ta, sa)))
          end
        | (t, (SOME (ta, (x, m)), sigmaa)) =>
          SOME (sigmaa,
                 (t, (ta, exec_upd D_ J_ A_ pick_wakeup thr_lookup thr_update
                            ws_lookup ws_update ws_delete ws_iterate is_ins
                            is_delete sigma s t ta x m)))));

fun exec_step A_ D_ J_ convert_RA schedule pick_wakeup thr_lookup thr_update
  ws_lookup ws_update ws_delete ws_iterate is_ins is_delete (sigma, s) =
  (case execT A_ D_ J_ convert_RA schedule pick_wakeup thr_lookup thr_update
          ws_lookup ws_update ws_delete ws_iterate is_ins is_delete sigma s
    of NONE => Sum_Type.Inr s
    | SOME (sigmaa, (t, (ta, sa))) =>
      Sum_Type.Inl ((sigma, (t, ta)), (sigmaa, sa)));

fun exec_aux A_ D_ J_ convert_RA schedule pick_wakeup thr_lookup thr_update
  ws_lookup ws_update ws_delete ws_iterate is_ins is_delete sigma_s =
  tllist_unfold
    (exec_step A_ D_ J_ convert_RA schedule pick_wakeup thr_lookup thr_update
      ws_lookup ws_update ws_delete ws_iterate is_ins is_delete)
    sigma_s;

fun exec A_ D_ J_ convert_RA schedule output pick_wakeup thr_lookup thr_update
  ws_lookup ws_update ws_delete ws_iterate is_ins is_delete sigma s =
  TLList.tmap Option.the Fun.id
    (TLList.tfilter (fn _ => (raise Fail "undefined"))
      (fn q => not (Option.is_none q))
      (TLList.tmap (fn (sigmaa, (a, b)) => output sigmaa a b) Fun.id
        (exec_aux A_ D_ J_ convert_RA schedule pick_wakeup thr_lookup thr_update
          ws_lookup ws_update ws_delete ws_iterate is_ins is_delete
          (sigma, s))));

fun singleton2 dfault (Predicate.Seq f) =
  (case f () of Predicate.Empty => dfault ()
    | Predicate.Insert (x, p) =>
      (if Predicate.is_empty p then x
        else (raise Fail "singleton2 not unique")
               (fn _ => singleton2 dfault (Predicate.Seq f)))
    | Predicate.Join (p, xq) =>
      (if Predicate.is_empty p then the_only2 dfault xq
        else (if Predicate.null xq then singleton2 dfault p
               else (raise Fail "singleton2 not unique")
                      (fn _ => singleton2 dfault (Predicate.Seq f)))))
and the_only2 dfault (Predicate.Join (p, xq)) =
  (if Predicate.is_empty p then the_only2 dfault xq
    else (if Predicate.null xq then singleton2 dfault p
           else (raise Fail "the_only2 not unique")
                  (fn _ => the_only2 dfault (Predicate.Join (p, xq)))))
  | the_only2 dfault (Predicate.Insert (x, p)) =
    (if Predicate.is_empty p then x
      else (raise Fail "the_only2 not unique")
             (fn _ => the_only2 dfault (Predicate.Insert (x, p))))
  | the_only2 dfault Predicate.Empty = (raise Fail "the_only2 empty") dfault;

fun the2 a = singleton2 (fn _ => (raise Fail "not_unique") (fn _ => the2 a)) a;

fun free_thread_id thr_lookup ts t = Option.is_none (thr_lookup t ts);

fun thread_ok thr_lookup ts (FWState.ThreadExists (t, b)) =
  not (Product_Type.equal_bool b (free_thread_id thr_lookup ts t))
  | thread_ok thr_lookup ts (FWState.NewThread (t, x, m)) =
    free_thread_id thr_lookup ts t;

fun interrupt_action_ok is_memb is (FWState.IsInterrupted (t, b)) =
  Product_Type.equal_bool b (is_memb t is)
  | interrupt_action_ok is_memb is (FWState.ClearInterrupt t) = true
  | interrupt_action_ok is_memb is (FWState.Interrupt t) = true;

fun interrupt_actions_ok is_memb is_ins is_delete is (ia :: ias) =
  interrupt_action_ok is_memb is ia andalso
    interrupt_actions_ok is_memb is_ins is_delete
      (redT_updI is_ins is_delete is ia) ias
  | interrupt_actions_ok is_memb is_ins is_delete is [] = true;

fun wset_actions_ok A_ C_ ws_lookup ws t was =
  Option.equal_optiona (FWState.equal_wait_set_status C_) (ws_lookup t ws)
    (if List.member (FWState.equal_wait_set_action A_ C_) was FWState.Notified
      then SOME (FWState.PostWS FWState.WSNotified)
      else (if List.member (FWState.equal_wait_set_action A_ C_) was
                 FWState.WokenUp
             then SOME (FWState.PostWS FWState.WSWokenUp) else NONE));

fun cond_action_ok B_ (D1_, D2_) final thr_lookup ws_lookup s t FWState.Yield =
  true
  | cond_action_ok B_ (D1_, D2_) final thr_lookup ws_lookup s ta
    (FWState.Join t) =
    (case thr_lookup t (FWState.thr s) of NONE => true
      | SOME (x, ln) =>
        not (HOL.eq B_ ta t) andalso
          (final x andalso
            (HOL.eq (FinFun.equal_finfun (D1_, D2_) Arith.equal_nat) ln
               (FinFun.Finfun_const Arith.zero_nat) andalso
              Option.is_none (ws_lookup t (FWState.wset s)))));

fun cond_action_oks B_ (D1_, D2_) final thr_lookup ws_lookup s t cts =
  List.list_all (cond_action_ok B_ (D1_, D2_) final thr_lookup ws_lookup s t)
    cts;

fun thread_oks thr_lookup thr_update ts (ta :: tas) =
  thread_ok thr_lookup ts ta andalso
    thread_oks thr_lookup thr_update (redT_updT thr_update ts ta) tas
  | thread_oks thr_lookup thr_update ts [] = true;

fun actions_ok B_ (D1_, D2_) F_ final thr_lookup thr_update ws_lookup is_memb
  is_ins is_delete s t ta =
  FWLocking.lock_ok_las (D1_, D2_) B_ (FWState.locks s) t
    (FWState.locks_a ta) andalso
    (thread_oks thr_lookup thr_update (FWState.thr s) (FWState.thr_a ta) andalso
      (cond_action_oks B_ (D1_, D2_) final thr_lookup ws_lookup s t
         (FWState.cond_a ta) andalso
        (wset_actions_ok B_ F_ ws_lookup (FWState.wset s) t
           (FWState.wset_a ta) andalso
          interrupt_actions_ok is_memb is_ins is_delete (FWState.interrupts s)
            (FWState.interrupt_a ta))));

fun step_thread B_ (D1_, D2_) E_ final r convert_RA thr_lookup thr_update
  ws_lookup is_memb is_ins is_delete update_state s t =
  (case thr_lookup t (FWState.thr s) of NONE => NONE
    | SOME (x, ln) =>
      (if HOL.eq (FinFun.equal_finfun (D1_, D2_) Arith.equal_nat) ln
            (FinFun.Finfun_const Arith.zero_nat)
        then let
               val reds =
                 Predicate.bind (r t (x, FWState.shr s))
                   (fn (ta, (xa, m)) =>
                     (if actions_ok B_ (D1_, D2_) E_ final thr_lookup thr_update
                           ws_lookup is_memb is_ins is_delete s t ta
                       then Predicate.single (ta, (xa, m))
                       else Predicate.bot_pred));
             in
               (if Predicate.holds
                     (Predicate.bind reds (fn _ => Predicate.single ()))
                 then let
                        val (ta, (xa, m)) = the2 reds;
                      in
                        SOME (t, (SOME (ta, (xa, m)), update_state ta))
                      end
                 else NONE)
             end
        else (if FWLocking.may_acquire_all (D1_, D2_) B_ (FWState.locks s) t
                   ln andalso
                   not (FWWait.waiting (ws_lookup t (FWState.wset s)))
               then SOME (t, (NONE,
                               update_state
                                 (FinFun.Finfun_const [],
                                   ([], ([], ([], ([], convert_RA ln)))))))
               else NONE)));

fun active_threads B_ (D1_, D2_) E_ final r thr_lookup thr_update thr_iterate
  ws_lookup is_memb is_ins is_delete thr_empty thr_ins_dj
  (ls, ((ts, m), (ws, is))) =
  thr_iterate ts (fn _ => true)
    (fn (t, (x, ln)) => fn tsa =>
      (if HOL.eq (FinFun.equal_finfun (D1_, D2_) Arith.equal_nat) ln
            (FinFun.Finfun_const Arith.zero_nat)
        then (if Predicate.holds
                   (Predicate.bind (r t (x, m))
                     (fn (ta, _) =>
                       Predicate.if_pred
                         (actions_ok B_ (D1_, D2_) E_ final thr_lookup
                           thr_update ws_lookup is_memb is_ins is_delete
                           (ls, ((ts, m), (ws, is))) t ta)))
               then thr_ins_dj t tsa else tsa)
        else (if not (FWWait.waiting (ws_lookup t ws)) andalso
                   FWLocking.may_acquire_all (D1_, D2_) B_ ls t ln
               then thr_ins_dj t tsa else tsa)))
    (thr_empty ());

fun pick_wakeup_via_sel C_ ws_sel sigma t w ws =
  Option.map Product_Type.fst
    (ws_sel ws
      (fn _ => fn wa => FWState.equal_wait_set_statusa C_ wa (FWState.InWS w)));

end; (*struct Scheduler*)

structure RBTSetImpl : sig
  val ins_rm_basic_ops :
    'a Orderings.linorder -> 'a -> ('a, unit) RBT.rbt -> ('a, unit) RBT.rbt
  val memb_rm_basic_ops :
    'a Orderings.linorder -> 'a -> ('a, unit) RBT.rbt -> bool
  val empty_rm_basic_ops : 'a Orderings.linorder -> unit -> ('a, unit) RBT.rbt
  val delete_rm_basic_ops :
    'a Orderings.linorder -> 'a -> ('a, unit) RBT.rbt -> ('a, unit) RBT.rbt
end = struct

fun ins_rm_basic_ops A_ x s = RBT.insert A_ x () s;

fun memb_rm_basic_ops A_ x s = not (Option.is_none (RBT.lookup A_ s x));

fun empty_rm_basic_ops A_ = (fn _ => RBT.empty A_);

fun delete_rm_basic_ops A_ x s = RBT.delete A_ x s;

end; (*struct RBTSetImpl*)

structure ListMapImpl : sig
  val g_list_to_map_lm_basic_ops :
    'a HOL.equal -> ('a * 'b) list -> ('a, 'b) Assoc_List.assoc_list
end = struct

fun g_list_to_map_lm_basic_ops A_ l =
  List.foldl (fn m => fn (k, v) => Assoc_List.update A_ k v m) Assoc_List.empty
    (List.rev l);

end; (*struct ListMapImpl*)

structure Round_Robin : sig
  val enqueue_new_thread :
    ('a -> 'b -> 'b) -> 'b -> ('a, 'c, 'd) FWState.new_thread_action -> 'b
  val enqueue_new_threads :
    ('a -> 'b -> 'b) -> 'b -> ('a, 'c, 'd) FWState.new_thread_action list -> 'b
  val queue_rotate1 : ('a -> 'b -> 'b) -> ('b -> 'a * 'b) -> 'b -> 'b
  val round_robin_update_state :
    'a HOL.equal ->
      ('a -> 'b -> 'b) ->
        ('b -> 'a * 'b) ->
          Arith.nat ->
            'b * Arith.nat ->
              'a -> ('c, (FWState.lock_action list)) FinFun.finfun *
                      (('a, 'd, 'e) FWState.new_thread_action list *
                        ('a FWState.conditional_action list *
                          (('a, 'f) FWState.wait_set_action list *
                            ('a FWState.interrupt_action list * 'g list)))) ->
                      'b * Arith.nat
  val round_robin_reschedule :
    'b HOL.equal -> 'd Set.card_UNIV * 'd HOL.equal -> 'e HOL.equal ->
      ('a -> bool) ->
        ('b ->
          'a * 'c ->
            ((('d, (FWState.lock_action list)) FinFun.finfun *
               (('b, 'a, 'c) FWState.new_thread_action list *
                 ('b FWState.conditional_action list *
                   (('b, 'e) FWState.wait_set_action list *
                     ('b FWState.interrupt_action list * 'f list))))) *
              ('a * 'c))
              Predicate.pred) ->
          (('d, Arith.nat) FinFun.finfun -> 'f list) ->
            ('b -> 'g -> ('a * ('d, Arith.nat) FinFun.finfun) option) ->
              ('b -> 'a * ('d, Arith.nat) FinFun.finfun -> 'g -> 'g) ->
                ('b -> 'h -> 'e FWState.wait_set_status option) ->
                  ('b -> 'i -> bool) ->
                    ('b -> 'i -> 'i) ->
                      ('b -> 'i -> 'i) ->
                        ('b -> 'j -> 'j) ->
                          ('j -> 'b * 'j) ->
                            ('b -> 'j -> 'j) ->
                              'b -> 'j -> Arith.nat ->
    ('d, (('b * Arith.nat) option)) FinFun.finfun * (('g * 'c) * ('h * 'i)) ->
      ('b * (((('d, (FWState.lock_action list)) FinFun.finfun *
                (('b, 'a, 'c) FWState.new_thread_action list *
                  ('b FWState.conditional_action list *
                    (('b, 'e) FWState.wait_set_action list *
                      ('b FWState.interrupt_action list * 'f list))))) *
               ('a * 'c)) option *
              ('j * Arith.nat))) option
  val round_robin :
    'b HOL.equal -> 'd Set.card_UNIV * 'd HOL.equal -> 'e HOL.equal ->
      ('a -> bool) ->
        ('b ->
          'a * 'c ->
            ((('d, (FWState.lock_action list)) FinFun.finfun *
               (('b, 'a, 'c) FWState.new_thread_action list *
                 ('b FWState.conditional_action list *
                   (('b, 'e) FWState.wait_set_action list *
                     ('b FWState.interrupt_action list * 'f list))))) *
              ('a * 'c))
              Predicate.pred) ->
          (('d, Arith.nat) FinFun.finfun -> 'f list) ->
            ('b -> 'g -> ('a * ('d, Arith.nat) FinFun.finfun) option) ->
              ('b -> 'a * ('d, Arith.nat) FinFun.finfun -> 'g -> 'g) ->
                ('b -> 'h -> 'e FWState.wait_set_status option) ->
                  ('b -> 'i -> bool) ->
                    ('b -> 'i -> 'i) ->
                      ('b -> 'i -> 'i) ->
                        ('j -> bool) ->
                          ('b -> 'j -> 'j) ->
                            ('j -> 'b * 'j) ->
                              ('b -> 'j -> 'j) ->
                                Arith.nat ->
                                  'j * Arith.nat ->
                                    ('d, (('b * Arith.nat) option))
                                      FinFun.finfun *
                                      (('g * 'c) * ('h * 'i)) ->
                                      ('b *
(((('d, (FWState.lock_action list)) FinFun.finfun *
    (('b, 'a, 'c) FWState.new_thread_action list *
      ('b FWState.conditional_action list *
        (('b, 'e) FWState.wait_set_action list *
          ('b FWState.interrupt_action list * 'f list))))) *
   ('a * 'c)) option *
  ('j * Arith.nat))) option
  val round_robin_start :
    (unit -> 'a) -> ('b -> 'a -> 'a) -> Arith.nat -> 'b -> 'a * Arith.nat
end = struct

fun enqueue_new_thread queue_enqueue ts (FWState.ThreadExists (t, b)) = ts
  | enqueue_new_thread queue_enqueue ts (FWState.NewThread (t, x, m)) =
    queue_enqueue t ts;

fun enqueue_new_threads queue_enqueue =
  List.foldl (enqueue_new_thread queue_enqueue);

fun queue_rotate1 queue_enqueue queue_dequeue =
  (fn (a, b) => queue_enqueue a b) o queue_dequeue;

fun round_robin_update_state A_ queue_enqueue queue_dequeue n0 (queue, n) t ta =
  let
    val queuea = enqueue_new_threads queue_enqueue queue (FWState.thr_a ta);
  in
    (if Arith.equal_nata n Arith.zero_nat orelse
          List.member (FWState.equal_conditional_action A_) (FWState.cond_a ta)
            FWState.Yield
      then (queue_rotate1 queue_enqueue queue_dequeue queuea, n0)
      else (queuea, Arith.minus_nat n Arith.one_nat))
  end;

fun round_robin_reschedule B_ (D1_, D2_) E_ final r convert_RA thr_lookup
  thr_update ws_lookup is_memb is_ins is_delete queue_enqueue queue_dequeue
  queue_push t0 queue n0 s =
  let
    val (t, queuea) = queue_dequeue queue;
  in
    (if HOL.eq B_ t t0 then NONE
      else (case Scheduler.step_thread B_ (D1_, D2_) E_ final r convert_RA
                   thr_lookup thr_update ws_lookup is_memb is_ins is_delete
                   (round_robin_update_state B_ queue_enqueue queue_dequeue n0
                     (queue_push t queuea, n0) t)
                   s t
             of NONE =>
               round_robin_reschedule B_ (D1_, D2_) E_ final r convert_RA
                 thr_lookup thr_update ws_lookup is_memb is_ins is_delete
                 queue_enqueue queue_dequeue queue_push t0
                 (queue_enqueue t queuea) n0 s
             | SOME a => SOME a))
  end;

fun round_robin B_ (D1_, D2_) E_ final r convert_RA thr_lookup thr_update
  ws_lookup is_memb is_ins is_delete queue_isEmpty queue_enqueue queue_dequeue
  queue_push n0 (queue, n) s =
  (if queue_isEmpty queue then NONE
    else let
           val (t, queuea) = queue_dequeue queue;
         in
           (case Scheduler.step_thread B_ (D1_, D2_) E_ final r convert_RA
                   thr_lookup thr_update ws_lookup is_memb is_ins is_delete
                   (round_robin_update_state B_ queue_enqueue queue_dequeue n0
                     (queue_push t queuea, n) t)
                   s t
             of NONE =>
               round_robin_reschedule B_ (D1_, D2_) E_ final r convert_RA
                 thr_lookup thr_update ws_lookup is_memb is_ins is_delete
                 queue_enqueue queue_dequeue queue_push t
                 (queue_enqueue t queuea) n0 s
             | SOME a => SOME a)
         end);

fun round_robin_start queue_empty queue_enqueue n0 t =
  (queue_enqueue t (queue_empty ()), n0);

end; (*struct Round_Robin*)

structure ListSetImpl_Invar : sig
  val iteratei_bset_op_list_it_lsi_basic_ops :
    'a list -> ('b -> bool) -> ('a -> 'b -> 'b) -> 'b -> 'b
  val g_to_list_lsi_basic_ops : 'a list -> 'a list
end = struct

fun iteratei_bset_op_list_it_lsi_basic_ops s = Foldi.foldli s;

fun g_to_list_lsi_basic_ops s =
  iteratei_bset_op_list_it_lsi_basic_ops s (fn _ => true)
    (fn a => fn b => a :: b) [];

end; (*struct ListSetImpl_Invar*)

structure State_Refinement : sig
  val jinja_output :
    'a -> 'b -> ('c, (FWState.lock_action list)) FinFun.finfun *
                  (('b, 'd, 'e) FWState.new_thread_action list *
                    ('b FWState.conditional_action list *
                      (('b, 'c) FWState.wait_set_action list *
                        ('b FWState.interrupt_action list *
                          ('c, 'b) Observable_Events.obs_event list)))) ->
                  ('b * ('c, 'b) Observable_Events.obs_event list) option
  val start_state_refine :
    'a HOL.equal * 'a Value.addr -> 'c HOL.equal -> 'h HOL.equal ->
      ('a -> 'b) ->
        'c -> ('c -> Type.htype -> ('c * 'a) Set.set) ->
                'd -> ('b -> 'e * ('a, Arith.nat) FinFun.finfun -> 'd -> 'd) ->
                        'f -> 'g -> (string ->
                                      string ->
Type.ty list -> Type.ty -> 'h -> 'a Value.vala list -> 'e) ->
                                      'h TypeRel.prog ->
string ->
  string ->
    'a Value.vala list ->
      ('a, (('b * Arith.nat) option)) FinFun.finfun * (('d * 'c) * ('f * 'g))
end = struct

fun jinja_output sigma t ta =
  (if List.null (FWState.obs_a ta) then NONE else SOME (t, FWState.obs_a ta));

fun start_state_refine (A1_, A2_) C_ H_ addr2thread_id empty_heap allocate
  thr_empty thr_update ws_empty is_empty f p c m vs =
  let
    val (d, (ts, (t, ma))) = TypeRel.method H_ p c m;
  in
    (FinFun.Finfun_const NONE,
      ((thr_update
          (StartConfig.start_tid (A1_, A2_) C_ addr2thread_id empty_heap
            allocate)
          (f d m ts t (Option.the ma) vs, FinFun.Finfun_const Arith.zero_nat)
          thr_empty,
         StartConfig.start_heap C_ (A1_, A2_) empty_heap allocate),
        (ws_empty, is_empty)))
  end;

end; (*struct State_Refinement*)

structure Random_Scheduler : sig
  val next_thread :
    ('a -> 'b list) ->
      Arith.natural * Arith.natural ->
        'a -> ('b * (Arith.natural * Arith.natural)) option
  val random_scheduler :
    'b HOL.equal -> 'd Set.card_UNIV * 'd HOL.equal -> 'e HOL.equal ->
      ('a -> bool) ->
        ('b ->
          'a * 'c ->
            ((('d, (FWState.lock_action list)) FinFun.finfun *
               (('b, 'a, 'c) FWState.new_thread_action list *
                 ('b FWState.conditional_action list *
                   (('b, 'e) FWState.wait_set_action list *
                     ('b FWState.interrupt_action list * 'f list))))) *
              ('a * 'c))
              Predicate.pred) ->
          (('d, Arith.nat) FinFun.finfun -> 'f list) ->
            ('b -> 'g -> ('a * ('d, Arith.nat) FinFun.finfun) option) ->
              ('b -> 'a * ('d, Arith.nat) FinFun.finfun -> 'g -> 'g) ->
                ('g ->
                  ('h -> bool) ->
                    ('b * ('a * ('d, Arith.nat) FinFun.finfun) -> 'h -> 'h) ->
                      'h -> 'h) ->
                  ('b -> 'i -> 'e FWState.wait_set_status option) ->
                    ('b -> 'j -> bool) ->
                      ('b -> 'j -> 'j) ->
                        ('b -> 'j -> 'j) ->
                          (unit -> 'h) ->
                            ('b -> 'h -> 'h) ->
                              ('h -> 'b list) ->
                                Arith.natural * Arith.natural ->
                                  ('d, (('b * Arith.nat) option))
                                    FinFun.finfun *
                                    (('g * 'c) * ('i * 'j)) ->
                                    ('b * (((('d, (FWState.lock_action list))
       FinFun.finfun *
      (('b, 'a, 'c) FWState.new_thread_action list *
        ('b FWState.conditional_action list *
          (('b, 'e) FWState.wait_set_action list *
            ('b FWState.interrupt_action list * 'f list))))) *
     ('a * 'c)) option *
    (Arith.natural * Arith.natural))) option
end = struct

fun next_thread thr_to_list seed active =
  let
    val ts = thr_to_list active;
  in
    (if List.null ts then NONE
      else SOME (Random.select (thr_to_list active) seed))
  end;

fun random_scheduler B_ (D1_, D2_) E_ final r convert_RA thr_lookup thr_update
  thr_iterate ws_lookup is_memb is_ins is_delete thr_empty thr_ins_dj
  thr_to_list seed s =
  Option.bind
    (next_thread thr_to_list seed
      (Scheduler.active_threads B_ (D1_, D2_) E_ final r thr_lookup thr_update
        thr_iterate ws_lookup is_memb is_ins is_delete thr_empty thr_ins_dj s))
    (fn (t, seeda) =>
      Scheduler.step_thread B_ (D1_, D2_) E_ final r convert_RA thr_lookup
        thr_update ws_lookup is_memb is_ins is_delete (fn _ => seeda) s t);

end; (*struct Random_Scheduler*)

structure ExternalCall_Execute : sig
  val heap_copy_loc :
    'a HOL.equal -> 'b HOL.equal * 'b Value.addr -> 'c HOL.equal ->
      ('a -> 'b -> Heap.addr_loc -> 'b Value.vala Set.set) ->
        ('a -> 'b -> Heap.addr_loc -> 'b Value.vala -> 'a Set.set) ->
          'b -> 'b -> Heap.addr_loc ->
                        'a -> (('b, 'c) Observable_Events.obs_event list * 'a)
                                Set.set
  val heap_copies :
    'a HOL.equal -> 'b HOL.equal * 'b Value.addr -> 'c HOL.equal ->
      ('a -> 'b -> Heap.addr_loc -> 'b Value.vala Set.set) ->
        ('a -> 'b -> Heap.addr_loc -> 'b Value.vala -> 'a Set.set) ->
          'b -> 'b -> Heap.addr_loc list ->
                        'a -> (('b, 'c) Observable_Events.obs_event list * 'a)
                                Set.set
  val heap_clone :
    'a HOL.equal -> 'b HOL.equal * 'b Value.addr -> 'd HOL.equal ->
      ('a -> Type.htype -> ('a * 'b) Set.set) ->
        ('a -> 'b -> Type.htype option) ->
          ('a -> 'b -> Heap.addr_loc -> 'b Value.vala Set.set) ->
            ('a -> 'b -> Heap.addr_loc -> 'b Value.vala -> 'a Set.set) ->
              'c TypeRel.prog ->
                'a -> 'b -> ('a * (('b, 'd) Observable_Events.obs_event list *
                                    'b) option)
                              Set.set
  val red_external_aggr :
    'a Set.card_UNIV * 'a HOL.equal * 'a Value.addr -> 'b HOL.equal ->
      'c HOL.equal ->
      ('a -> 'b) ->
        ('b -> 'a) ->
          bool ->
            'c -> ('c -> Type.htype -> ('c * 'a) Set.set) ->
                    ('c -> 'a -> Type.htype option) ->
                      ('c -> 'a -> Heap.addr_loc -> 'a Value.vala Set.set) ->
                        ('c ->
                          'a -> Heap.addr_loc -> 'a Value.vala -> 'c Set.set) ->
                          'd TypeRel.prog ->
                            'b -> 'a -> string ->
  'a Value.vala list ->
    'c -> ((('a, (FWState.lock_action list)) FinFun.finfun *
             (('b, (string * (string * 'a)), 'c)
                FWState.new_thread_action list *
               ('b FWState.conditional_action list *
                 (('b, 'a) FWState.wait_set_action list *
                   ('b FWState.interrupt_action list *
                     ('a, 'b) Observable_Events.obs_event list))))) *
            ('a ExternalCall.extCallRet * 'c))
            Set.set
end = struct

fun heap_copy_loc A_ (B1_, B2_) C_ heap_read heap_write aa a al h =
  Set.bind
    (Product_Type.equal_prod
      (List.equal_list (Observable_Events.equal_obs_event B1_ C_)) A_)
    (heap_read h aa al)
    (fn v =>
      Set.bind
        (Product_Type.equal_prod
          (List.equal_list (Observable_Events.equal_obs_event B1_ C_)) A_)
        (heap_write h a al v)
        (fn ha =>
          Set.insert
            ([Observable_Events.ReadMem (aa, al, v),
               Observable_Events.WriteMem (a, al, v)],
              ha)
            Set.bot_seta));

fun heap_copies A_ (B1_, B2_) C_ heap_read heap_write aa a (al :: als) h =
  Set.bind
    (Product_Type.equal_prod
      (List.equal_list (Observable_Events.equal_obs_event B1_ C_)) A_)
    (heap_copy_loc A_ (B1_, B2_) C_ heap_read heap_write aa a al h)
    (fn (ob, ha) =>
      Set.bind
        (Product_Type.equal_prod
          (List.equal_list (Observable_Events.equal_obs_event B1_ C_)) A_)
        (heap_copies A_ (B1_, B2_) C_ heap_read heap_write aa a als ha)
        (fn (obs, hb) => Set.insert (ob @ obs, hb) Set.bot_seta))
  | heap_copies A_ (B1_, B2_) C_ heap_read heap_write aa a [] h =
    Set.insert ([], h) Set.bot_seta;

fun heap_clone A_ (B1_, B2_) D_ allocate typeof_addr heap_read heap_write p h a
  = (case typeof_addr h a of NONE => Set.bot_seta
      | SOME (Type.Class_type c) =>
        let
          val ha = allocate h (Type.Class_type c);
        in
          (if Set.equal_set (Product_Type.equal_prod A_ B1_) ha Set.bot_seta
            then Set.insert (h, NONE) Set.bot_seta
            else Set.bind
                   (Product_Type.equal_prod A_
                     (Option.equal_option
                       (Product_Type.equal_prod
                         (List.equal_list
                           (Observable_Events.equal_obs_event B1_ D_))
                         B1_)))
                   ha (fn (hb, aa) =>
                        Set.bind
                          (Product_Type.equal_prod A_
                            (Option.equal_option
                              (Product_Type.equal_prod
                                (List.equal_list
                                  (Observable_Events.equal_obs_event B1_ D_))
                                B1_)))
                          (Predicate.set_of_pred (TypeRel.fields_i_i_o p c))
                          (fn fDTs =>
                            Set.bind
                              (Product_Type.equal_prod A_
                                (Option.equal_option
                                  (Product_Type.equal_prod
                                    (List.equal_list
                                      (Observable_Events.equal_obs_event B1_
D_))
                                    B1_)))
                              (heap_copies A_ (B1_, B2_) D_ heap_read heap_write
                                a aa
                                (List.map
                                  (fn (ab, b) =>
                                    let
                                      val (f, d) = ab;
                                    in
                                      (fn _ => Heap.CField (d, f))
                                    end
                                      b)
                                  fDTs)
                                hb)
                              (fn (obs, hc) =>
                                Set.insert
                                  (hc, SOME
 (Observable_Events.NewHeapElem (aa, Type.Class_type c) :: obs, aa))
                                  Set.bot_seta))))
        end
      | SOME (Type.Array_type (t, n)) =>
        let
          val ha = allocate h (Type.Array_type (t, n));
        in
          (if Set.equal_set (Product_Type.equal_prod A_ B1_) ha Set.bot_seta
            then Set.insert (h, NONE) Set.bot_seta
            else Set.bind
                   (Product_Type.equal_prod A_
                     (Option.equal_option
                       (Product_Type.equal_prod
                         (List.equal_list
                           (Observable_Events.equal_obs_event B1_ D_))
                         B1_)))
                   ha (fn (hb, aa) =>
                        Set.bind
                          (Product_Type.equal_prod A_
                            (Option.equal_option
                              (Product_Type.equal_prod
                                (List.equal_list
                                  (Observable_Events.equal_obs_event B1_ D_))
                                B1_)))
                          (Predicate.set_of_pred
                            (TypeRel.fields_i_i_o p Type.object))
                          (fn fDTs =>
                            Set.bind
                              (Product_Type.equal_prod A_
                                (Option.equal_option
                                  (Product_Type.equal_prod
                                    (List.equal_list
                                      (Observable_Events.equal_obs_event B1_
D_))
                                    B1_)))
                              (heap_copies A_ (B1_, B2_) D_ heap_read heap_write
                                a aa
                                (List.map
                                   (fn (ab, b) =>
                                     let
                                       val (f, d) = ab;
                                     in
                                       (fn _ => Heap.CField (d, f))
                                     end
                                       b)
                                   fDTs @
                                  List.map Heap.ACell
                                    (List.upt Arith.zero_nat n))
                                hb)
                              (fn (obs, hc) =>
                                Set.insert
                                  (hc, SOME
 (Observable_Events.NewHeapElem (aa, Type.Array_type (t, n)) :: obs, aa))
                                  Set.bot_seta))))
        end);

fun red_external_aggr (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr
  spurious_wakeups empty_heap allocate typeof_addr heap_read heap_write p t a m
  vs h =
  (if ((m : string) = "wait()V")
    then let
           val _ = thread_id2addr t;
         in
           Set.union
             (Set.insert
               (FWState.thread_action_to_thread_action A2_
                  Observable_Events.obs_action_obs_event
                  (FWState.ObsAction (Observable_Events.ObsInterrupted t))
                  (FWState.thread_action_to_thread_action A2_
                    Observable_Events.obs_action_obs_event
                    (FWState.InterruptAction (FWState.ClearInterrupt t))
                    (FWState.thread_action_to_thread_action A2_
                      Observable_Events.obs_action_obs_event
                      (FWState.InterruptAction
                        (FWState.IsInterrupted (t, true)))
                      (FWState.thread_action_to_thread_action A2_
                        Observable_Events.obs_action_obs_event
                        (FWState.LockAction (FWState.Lock, a))
                        (FWState.thread_action_to_thread_action A2_
                          Observable_Events.obs_action_obs_event
                          (FWState.LockAction (FWState.Unlock, a))
                          (FinFun.Finfun_const [],
                            ([], ([], ([], ([], []))))))))),
                 (ExternalCall.RetExc
                    (StartConfig.addr_of_sys_xcpt C_ (A2_, A3_) empty_heap
                      allocate "java/lang/InterruptedException"),
                   h))
               (Set.insert
                 (FWState.thread_action_to_thread_action A2_
                    Observable_Events.obs_action_obs_event
                    (FWState.ObsAction (Observable_Events.SyncUnlock a))
                    (FWState.thread_action_to_thread_action A2_
                      Observable_Events.obs_action_obs_event
                      (FWState.InterruptAction
                        (FWState.IsInterrupted (t, false)))
                      (FWState.thread_action_to_thread_action A2_
                        Observable_Events.obs_action_obs_event
                        (FWState.LockAction (FWState.ReleaseAcquire, a))
                        (FWState.thread_action_to_thread_action A2_
                          Observable_Events.obs_action_obs_event
                          (FWState.LockAction (FWState.Lock, a))
                          (FWState.thread_action_to_thread_action A2_
                            Observable_Events.obs_action_obs_event
                            (FWState.LockAction (FWState.Unlock, a))
                            (FWState.thread_action_to_thread_action A2_
                              Observable_Events.obs_action_obs_event
                              (FWState.WaitSetAction (FWState.Suspend a))
                              (FinFun.Finfun_const [],
                                ([], ([], ([], ([], [])))))))))),
                   (ExternalCall.RetStaySame, h))
                 (Set.insert
                   (FWState.thread_action_to_thread_action A2_
                      Observable_Events.obs_action_obs_event
                      (FWState.LockAction (FWState.UnlockFail, a))
                      (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                     (ExternalCall.RetExc
                        (StartConfig.addr_of_sys_xcpt C_ (A2_, A3_) empty_heap
                          allocate "java/lang/IllegalMonitorStateException"),
                       h))
                   (Set.insert
                     (FWState.thread_action_to_thread_action A2_
                        Observable_Events.obs_action_obs_event
                        (FWState.WaitSetAction FWState.Notified)
                        (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                       (ExternalCall.RetVal Value.Unit, h))
                     (Set.insert
                       (FWState.thread_action_to_thread_action A2_
                          Observable_Events.obs_action_obs_event
                          (FWState.ObsAction
                            (Observable_Events.ObsInterrupted t))
                          (FWState.thread_action_to_thread_action A2_
                            Observable_Events.obs_action_obs_event
                            (FWState.InterruptAction (FWState.ClearInterrupt t))
                            (FWState.thread_action_to_thread_action A2_
                              Observable_Events.obs_action_obs_event
                              (FWState.WaitSetAction FWState.WokenUp)
                              (FinFun.Finfun_const [],
                                ([], ([], ([], ([], []))))))),
                         (ExternalCall.RetExc
                            (StartConfig.addr_of_sys_xcpt C_ (A2_, A3_)
                              empty_heap allocate
                              "java/lang/InterruptedException"),
                           h))
                       Set.bot_seta)))))
             (if spurious_wakeups
               then Set.insert
                      (FWState.thread_action_to_thread_action A2_
                         Observable_Events.obs_action_obs_event
                         (FWState.ObsAction (Observable_Events.SyncUnlock a))
                         (FWState.thread_action_to_thread_action A2_
                           Observable_Events.obs_action_obs_event
                           (FWState.InterruptAction
                             (FWState.IsInterrupted (t, false)))
                           (FWState.thread_action_to_thread_action A2_
                             Observable_Events.obs_action_obs_event
                             (FWState.LockAction (FWState.ReleaseAcquire, a))
                             (FWState.thread_action_to_thread_action A2_
                               Observable_Events.obs_action_obs_event
                               (FWState.LockAction (FWState.Lock, a))
                               (FWState.thread_action_to_thread_action A2_
                                 Observable_Events.obs_action_obs_event
                                 (FWState.LockAction (FWState.Unlock, a))
                                 (FinFun.Finfun_const [],
                                   ([], ([], ([], ([], []))))))))),
                        (ExternalCall.RetVal Value.Unit, h))
                      Set.bot_seta
               else Set.bot_seta)
         end
    else (if ((m : string) = "notify()V")
           then Set.insert
                  (FWState.thread_action_to_thread_action A2_
                     Observable_Events.obs_action_obs_event
                     (FWState.LockAction (FWState.Lock, a))
                     (FWState.thread_action_to_thread_action A2_
                       Observable_Events.obs_action_obs_event
                       (FWState.LockAction (FWState.Unlock, a))
                       (FWState.thread_action_to_thread_action A2_
                         Observable_Events.obs_action_obs_event
                         (FWState.WaitSetAction (FWState.Notify a))
                         (FinFun.Finfun_const [], ([], ([], ([], ([], []))))))),
                    (ExternalCall.RetVal Value.Unit, h))
                  (Set.insert
                    (FWState.thread_action_to_thread_action A2_
                       Observable_Events.obs_action_obs_event
                       (FWState.LockAction (FWState.UnlockFail, a))
                       (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                      (ExternalCall.RetExc
                         (StartConfig.addr_of_sys_xcpt C_ (A2_, A3_) empty_heap
                           allocate "java/lang/IllegalMonitorStateException"),
                        h))
                    Set.bot_seta)
           else (if ((m : string) = "notifyAll()V")
                  then Set.insert
                         (FWState.thread_action_to_thread_action A2_
                            Observable_Events.obs_action_obs_event
                            (FWState.LockAction (FWState.Lock, a))
                            (FWState.thread_action_to_thread_action A2_
                              Observable_Events.obs_action_obs_event
                              (FWState.LockAction (FWState.Unlock, a))
                              (FWState.thread_action_to_thread_action A2_
                                Observable_Events.obs_action_obs_event
                                (FWState.WaitSetAction (FWState.NotifyAll a))
                                (FinFun.Finfun_const [],
                                  ([], ([], ([], ([], []))))))),
                           (ExternalCall.RetVal Value.Unit, h))
                         (Set.insert
                           (FWState.thread_action_to_thread_action A2_
                              Observable_Events.obs_action_obs_event
                              (FWState.LockAction (FWState.UnlockFail, a))
                              (FinFun.Finfun_const [],
                                ([], ([], ([], ([], []))))),
                             (ExternalCall.RetExc
                                (StartConfig.addr_of_sys_xcpt C_ (A2_, A3_)
                                  empty_heap allocate
                                  "java/lang/IllegalMonitorStateException"),
                               h))
                           Set.bot_seta)
                  else (if ((m : string) = "clone()Ljava/lang/Object;")
                         then Set.bind
                                (Product_Type.equal_prod
                                  (Product_Type.equal_prod
                                    (FinFun.equal_finfun (A1_, A2_)
                                      (List.equal_list
FWState.equal_lock_action))
                                    (Product_Type.equal_prod
                                      (List.equal_list
(FWState.equal_new_thread_action B_
  (Product_Type.equal_prod Stringa.equal_literal
    (Product_Type.equal_prod Stringa.equal_literal A2_))
  C_))
                                      (Product_Type.equal_prod
(List.equal_list (FWState.equal_conditional_action B_))
(Product_Type.equal_prod
  (List.equal_list (FWState.equal_wait_set_action B_ A2_))
  (Product_Type.equal_prod (List.equal_list (FWState.equal_interrupt_action B_))
    (List.equal_list (Observable_Events.equal_obs_event A2_ B_)))))))
                                  (Product_Type.equal_prod
                                    (ExternalCall.equal_extCallRet A2_) C_))
                                (heap_clone C_ (A2_, A3_) B_ allocate
                                  typeof_addr heap_read heap_write p h a)
                                (fn (ha, obsa) =>
                                  Set.insert
                                    (case obsa
                                      of NONE =>
((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
  (ExternalCall.RetExc
     (StartConfig.addr_of_sys_xcpt C_ (A2_, A3_) empty_heap allocate
       "java/lang/OutOfMemoryError"),
    ha))
                                      | SOME (obs, aa) =>
((FinFun.Finfun_const [], ([], ([], ([], ([], obs))))),
  (ExternalCall.RetVal (Value.Addr aa), ha)))
                                    Set.bot_seta)
                         else (if ((m : string) = "hashCode()I")
                                then Set.insert
                                       ((FinFun.Finfun_const [],
  ([], ([], ([], ([], []))))),
 (ExternalCall.RetVal
    (Value.Intg
      (Word.word_of_int
        (Type_Length.len0_bit0
          (Type_Length.len0_bit0
            (Type_Length.len0_bit0
              (Type_Length.len0_bit0
                (Type_Length.len0_bit0 Type_Length.len0_num1)))))
        (Value.hash_addr A3_ a))),
   h))
                                       Set.bot_seta
                                else (if ((m : string) = "~print(I)V")
                                       then Set.insert
      (FWState.thread_action_to_thread_action A2_
         Observable_Events.obs_action_obs_event
         (FWState.ObsAction
           (Observable_Events.ExternalCall (a, m, vs, Value.Unit)))
         (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
        (ExternalCall.RetVal Value.Unit, h))
      Set.bot_seta
                                       else (if ((m : string) = "~Thread.currentThread()Ljava/lang/Thread;")
      then Set.insert
             ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
               (ExternalCall.RetVal (Value.Addr (thread_id2addr t)), h))
             Set.bot_seta
      else (if ((m : string) = "~Thread.interrupted()Z")
             then Set.insert
                    (FWState.thread_action_to_thread_action A2_
                       Observable_Events.obs_action_obs_event
                       (FWState.ObsAction (Observable_Events.ObsInterrupted t))
                       (FWState.thread_action_to_thread_action A2_
                         Observable_Events.obs_action_obs_event
                         (FWState.InterruptAction (FWState.ClearInterrupt t))
                         (FWState.thread_action_to_thread_action A2_
                           Observable_Events.obs_action_obs_event
                           (FWState.InterruptAction
                             (FWState.IsInterrupted (t, true)))
                           (FinFun.Finfun_const [],
                             ([], ([], ([], ([], []))))))),
                      (ExternalCall.RetVal (Value.Bool true), h))
                    (Set.insert
                      (FWState.thread_action_to_thread_action A2_
                         Observable_Events.obs_action_obs_event
                         (FWState.InterruptAction
                           (FWState.IsInterrupted (t, false)))
                         (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                        (ExternalCall.RetVal (Value.Bool false), h))
                      Set.bot_seta)
             else (if ((m : string) = "~Thread.yield()V")
                    then Set.insert
                           (FWState.thread_action_to_thread_action A2_
                              Observable_Events.obs_action_obs_event
                              (FWState.ConditionalAction FWState.Yield)
                              (FinFun.Finfun_const [],
                                ([], ([], ([], ([], []))))),
                             (ExternalCall.RetVal Value.Unit, h))
                           Set.bot_seta
                    else let
                           val ta =
                             Type.ty_of_htype (Option.the (typeof_addr h a));
                         in
                           (if TypeRel.widen p ta (Type.Class Type.thread)
                             then let
                                    val t_a = addr2thread_id a;
                                  in
                                    (if ((m : string) = "start()V")
                                      then Set.insert
     (FWState.thread_action_to_thread_action A2_
        Observable_Events.obs_action_obs_event
        (FWState.ObsAction (Observable_Events.ThreadStart t_a))
        (FWState.thread_action_to_thread_action A2_
          Observable_Events.obs_action_obs_event
          (FWState.NewThreadAction
            (FWState.NewThread (t_a, (Type.the_Class ta, ("run()V", a)), h)))
          (FinFun.Finfun_const [], ([], ([], ([], ([], [])))))),
       (ExternalCall.RetVal Value.Unit, h))
     (Set.insert
       (FWState.thread_action_to_thread_action A2_
          Observable_Events.obs_action_obs_event
          (FWState.NewThreadAction (FWState.ThreadExists (t_a, true)))
          (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
         (ExternalCall.RetExc
            (StartConfig.addr_of_sys_xcpt C_ (A2_, A3_) empty_heap allocate
              "java/lang/IllegalThreadStateException"),
           h))
       Set.bot_seta)
                                      else (if ((m : string) = "join()V")
     then Set.insert
            (FWState.thread_action_to_thread_action A2_
               Observable_Events.obs_action_obs_event
               (FWState.ObsAction (Observable_Events.ThreadJoin t_a))
               (FWState.thread_action_to_thread_action A2_
                 Observable_Events.obs_action_obs_event
                 (FWState.InterruptAction (FWState.IsInterrupted (t, false)))
                 (FWState.thread_action_to_thread_action A2_
                   Observable_Events.obs_action_obs_event
                   (FWState.ConditionalAction (FWState.Join t_a))
                   (FinFun.Finfun_const [], ([], ([], ([], ([], []))))))),
              (ExternalCall.RetVal Value.Unit, h))
            (Set.insert
              (FWState.thread_action_to_thread_action A2_
                 Observable_Events.obs_action_obs_event
                 (FWState.ObsAction (Observable_Events.ObsInterrupted t))
                 (FWState.thread_action_to_thread_action A2_
                   Observable_Events.obs_action_obs_event
                   (FWState.InterruptAction (FWState.ClearInterrupt t))
                   (FWState.thread_action_to_thread_action A2_
                     Observable_Events.obs_action_obs_event
                     (FWState.InterruptAction (FWState.IsInterrupted (t, true)))
                     (FinFun.Finfun_const [], ([], ([], ([], ([], []))))))),
                (ExternalCall.RetExc
                   (StartConfig.addr_of_sys_xcpt C_ (A2_, A3_) empty_heap
                     allocate "java/lang/InterruptedException"),
                  h))
              Set.bot_seta)
     else (if ((m : string) = "interrupt()V")
            then Set.insert
                   (FWState.thread_action_to_thread_action A2_
                      Observable_Events.obs_action_obs_event
                      (FWState.ObsAction (Observable_Events.ObsInterrupt t_a))
                      (FWState.thread_action_to_thread_action A2_
                        Observable_Events.obs_action_obs_event
                        (FWState.InterruptAction (FWState.Interrupt t_a))
                        (FWState.thread_action_to_thread_action A2_
                          Observable_Events.obs_action_obs_event
                          (FWState.WaitSetAction (FWState.WakeUp t_a))
                          (FWState.thread_action_to_thread_action A2_
                            Observable_Events.obs_action_obs_event
                            (FWState.NewThreadAction
                              (FWState.ThreadExists (t_a, true)))
                            (FinFun.Finfun_const [],
                              ([], ([], ([], ([], [])))))))),
                     (ExternalCall.RetVal Value.Unit, h))
                   (Set.insert
                     (FWState.thread_action_to_thread_action A2_
                        Observable_Events.obs_action_obs_event
                        (FWState.NewThreadAction
                          (FWState.ThreadExists (t_a, false)))
                        (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                       (ExternalCall.RetVal Value.Unit, h))
                     Set.bot_seta)
            else (if ((m : string) = "isInterrupted()Z")
                   then Set.insert
                          (FWState.thread_action_to_thread_action A2_
                             Observable_Events.obs_action_obs_event
                             (FWState.InterruptAction
                               (FWState.IsInterrupted (t_a, false)))
                             (FinFun.Finfun_const [],
                               ([], ([], ([], ([], []))))),
                            (ExternalCall.RetVal (Value.Bool false), h))
                          (Set.insert
                            (FWState.thread_action_to_thread_action A2_
                               Observable_Events.obs_action_obs_event
                               (FWState.ObsAction
                                 (Observable_Events.ObsInterrupted t_a))
                               (FWState.thread_action_to_thread_action A2_
                                 Observable_Events.obs_action_obs_event
                                 (FWState.InterruptAction
                                   (FWState.IsInterrupted (t_a, true)))
                                 (FinFun.Finfun_const [],
                                   ([], ([], ([], ([], [])))))),
                              (ExternalCall.RetVal (Value.Bool true), h))
                            Set.bot_seta)
                   else Set.insert
                          ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                            (raise Fail "undefined"))
                          Set.bot_seta))))
                                  end
                             else Set.insert
                                    ((FinFun.Finfun_const [],
                                       ([], ([], ([], ([], []))))),
                                      (raise Fail "undefined"))
                                    Set.bot_seta)
                         end)))))))));

end; (*struct ExternalCall_Execute*)

structure JVMExceptions : sig
  val matches_ex_entry :
    'a TypeRel.prog ->
      string ->
        Arith.nat ->
          Arith.nat * (Arith.nat * (string option * (Arith.nat * Arith.nat))) ->
            bool
  val match_ex_table :
    'a TypeRel.prog ->
      string ->
        Arith.nat ->
          (Arith.nat *
            (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list ->
            (Arith.nat * Arith.nat) option
end = struct

fun matches_ex_entry p c pc xcp =
  let
    val (s, (e, (ca, (_, _)))) = xcp;
  in
    Arith.less_eq_nat s pc andalso
      (Arith.less_nat pc e andalso
        (case ca of NONE => true | SOME a => TypeRel.subcls p c a))
  end;

fun match_ex_table p c pc [] = NONE
  | match_ex_table p c pc (e :: es) =
    (if matches_ex_entry p c pc e
      then SOME (Product_Type.snd (Product_Type.snd (Product_Type.snd e)))
      else match_ex_table p c pc es);

end; (*struct JVMExceptions*)

structure JVMExec_Execute2 : sig
  val exception_step :
    'b Value.addr ->
      ('a -> 'b -> Type.htype option) ->
        (Arith.nat *
          (Arith.nat *
            ('b JVMInstructions.instr list *
              (Arith.nat *
                (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list)))
          TypeRel.prog ->
          'b -> 'a -> ('b JVMInstructions.instr list *
                        ('b JVMInstructions.instr list *
                          (Arith.nat *
                            (Arith.nat *
                              (string option *
                                (Arith.nat * Arith.nat)))) list)) *
                        ('b Value.vala list *
                          ('b Value.vala list *
                            (string * (string * Arith.nat)))) ->
                        (('b JVMInstructions.instr list *
                           ('b JVMInstructions.instr list *
                             (Arith.nat *
                               (Arith.nat *
                                 (string option *
                                   (Arith.nat * Arith.nat)))) list)) *
                          ('b Value.vala list *
                            ('b Value.vala list *
                              (string * (string * Arith.nat))))) list ->
                          'b option *
                            ('a * (('b JVMInstructions.instr list *
                                     ('b JVMInstructions.instr list *
                                       (Arith.nat *
 (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list)) *
                                    ('b Value.vala list *
                                      ('b Value.vala list *
(string * (string * Arith.nat))))) list)
  val extRet2JVM :
    'a JVMInstructions.instr list ->
      'a JVMInstructions.instr list ->
        (Arith.nat *
          (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list ->
          Arith.nat ->
            'b -> 'a Value.vala list ->
                    'a Value.vala list ->
                      string ->
                        string ->
                          Arith.nat ->
                            (('a JVMInstructions.instr list *
                               ('a JVMInstructions.instr list *
                                 (Arith.nat *
                                   (Arith.nat *
                                     (string option *
                                       (Arith.nat * Arith.nat)))) list)) *
                              ('a Value.vala list *
                                ('a Value.vala list *
                                  (string * (string * Arith.nat))))) list ->
                              'a ExternalCall.extCallRet ->
                                'a option *
                                  ('b * (('a JVMInstructions.instr list *
   ('a JVMInstructions.instr list *
     (Arith.nat *
       (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list)) *
  ('a Value.vala list *
    ('a Value.vala list * (string * (string * Arith.nat))))) list)
  val extNTA2JVM :
    'a HOL.equal ->
      (Arith.nat *
        (Arith.nat *
          ('a JVMInstructions.instr list *
            (Arith.nat *
              (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list)))
        TypeRel.prog ->
        string * (string * 'a) ->
          'a option *
            (('a JVMInstructions.instr list *
               ('a JVMInstructions.instr list *
                 (Arith.nat *
                   (Arith.nat *
                     (string option * (Arith.nat * Arith.nat)))) list)) *
              ('a Value.vala list *
                ('a Value.vala list * (string * (string * Arith.nat))))) list
  val exec_instr :
    'a Set.card_UNIV * 'a HOL.equal * 'a Value.addr -> 'b HOL.equal ->
      'c HOL.equal ->
      ('a -> 'b) ->
        ('b -> 'a) ->
          bool ->
            'c -> ('c -> Type.htype -> ('c * 'a) Set.set) ->
                    ('c -> 'a -> Type.htype option) ->
                      ('c -> 'a -> Heap.addr_loc -> 'a Value.vala Set.set) ->
                        ('c ->
                          'a -> Heap.addr_loc -> 'a Value.vala -> 'c Set.set) ->
                          'a JVMInstructions.instr list ->
                            'a JVMInstructions.instr list ->
                              (Arith.nat *
                                (Arith.nat *
                                  (string option *
                                    (Arith.nat * Arith.nat)))) list ->
                                'a JVMInstructions.instr ->
                                  (Arith.nat *
                                    (Arith.nat *
                                      ('a JVMInstructions.instr list *
(Arith.nat * (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list)))
                                    TypeRel.prog ->
                                    'b -> 'c ->
    'a Value.vala list ->
      'a Value.vala list ->
        string ->
          string ->
            Arith.nat ->
              (('a JVMInstructions.instr list *
                 ('a JVMInstructions.instr list *
                   (Arith.nat *
                     (Arith.nat *
                       (string option * (Arith.nat * Arith.nat)))) list)) *
                ('a Value.vala list *
                  ('a Value.vala list *
                    (string * (string * Arith.nat))))) list ->
                ((('a, (FWState.lock_action list)) FinFun.finfun *
                   (('b, ('a option *
                           (('a JVMInstructions.instr list *
                              ('a JVMInstructions.instr list *
                                (Arith.nat *
                                  (Arith.nat *
                                    (string option *
                                      (Arith.nat * Arith.nat)))) list)) *
                             ('a Value.vala list *
                               ('a Value.vala list *
                                 (string * (string * Arith.nat))))) list),
                      'c)
                      FWState.new_thread_action list *
                     ('b FWState.conditional_action list *
                       (('b, 'a) FWState.wait_set_action list *
                         ('b FWState.interrupt_action list *
                           ('a, 'b) Observable_Events.obs_event list))))) *
                  ('a option *
                    ('c * (('a JVMInstructions.instr list *
                             ('a JVMInstructions.instr list *
                               (Arith.nat *
                                 (Arith.nat *
                                   (string option *
                                     (Arith.nat * Arith.nat)))) list)) *
                            ('a Value.vala list *
                              ('a Value.vala list *
                                (string * (string * Arith.nat))))) list)))
                  Set.set
  val exec :
    'a Set.card_UNIV * 'a HOL.equal * 'a Value.addr -> 'b HOL.equal ->
      'c HOL.equal ->
      ('a -> 'b) ->
        ('b -> 'a) ->
          bool ->
            'c -> ('c -> Type.htype -> ('c * 'a) Set.set) ->
                    ('c -> 'a -> Type.htype option) ->
                      ('c -> 'a -> Heap.addr_loc -> 'a Value.vala Set.set) ->
                        ('c ->
                          'a -> Heap.addr_loc -> 'a Value.vala -> 'c Set.set) ->
                          (Arith.nat *
                            (Arith.nat *
                              ('a JVMInstructions.instr list *
                                (Arith.nat *
                                  (Arith.nat *
                                    (string option *
                                      (Arith.nat * Arith.nat)))) list)))
                            TypeRel.prog ->
                            'b -> 'a option *
                                    ('c * (('a JVMInstructions.instr list *
     ('a JVMInstructions.instr list *
       (Arith.nat *
         (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list)) *
    ('a Value.vala list *
      ('a Value.vala list * (string * (string * Arith.nat))))) list) ->
                                    ((('a, (FWState.lock_action list))
FinFun.finfun *
                                       (('b,
  ('a option *
    (('a JVMInstructions.instr list *
       ('a JVMInstructions.instr list *
         (Arith.nat *
           (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list)) *
      ('a Value.vala list *
        ('a Value.vala list * (string * (string * Arith.nat))))) list),
  'c)
  FWState.new_thread_action list *
 ('b FWState.conditional_action list *
   (('b, 'a) FWState.wait_set_action list *
     ('b FWState.interrupt_action list *
       ('a, 'b) Observable_Events.obs_event list))))) *
                                      ('a option *
('c * (('a JVMInstructions.instr list *
         ('a JVMInstructions.instr list *
           (Arith.nat *
             (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list)) *
        ('a Value.vala list *
          ('a Value.vala list * (string * (string * Arith.nat))))) list)))
                                      Set.set
  val exec_1 :
    'a Set.card_UNIV * 'a HOL.equal * 'a Value.addr -> 'b HOL.equal ->
      'c HOL.equal ->
      ('a -> 'b) ->
        ('b -> 'a) ->
          bool ->
            'c -> ('c -> Type.htype -> ('c * 'a) Set.set) ->
                    ('c -> 'a -> Type.htype option) ->
                      ('c -> 'a -> Heap.addr_loc -> 'a Value.vala Set.set) ->
                        ('c ->
                          'a -> Heap.addr_loc -> 'a Value.vala -> 'c Set.set) ->
                          (Arith.nat *
                            (Arith.nat *
                              ('a JVMInstructions.instr list *
                                (Arith.nat *
                                  (Arith.nat *
                                    (string option *
                                      (Arith.nat * Arith.nat)))) list)))
                            TypeRel.prog ->
                            'b -> 'a option *
                                    ('c * (('a JVMInstructions.instr list *
     ('a JVMInstructions.instr list *
       (Arith.nat *
         (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list)) *
    ('a Value.vala list *
      ('a Value.vala list * (string * (string * Arith.nat))))) list) ->
                                    ((('a, (FWState.lock_action list))
FinFun.finfun *
                                       (('b,
  ('a option *
    (('a JVMInstructions.instr list *
       ('a JVMInstructions.instr list *
         (Arith.nat *
           (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list)) *
      ('a Value.vala list *
        ('a Value.vala list * (string * (string * Arith.nat))))) list),
  'c)
  FWState.new_thread_action list *
 ('b FWState.conditional_action list *
   (('b, 'a) FWState.wait_set_action list *
     ('b FWState.interrupt_action list *
       ('a, 'b) Observable_Events.obs_event list))))) *
                                      ('a option *
('c * (('a JVMInstructions.instr list *
         ('a JVMInstructions.instr list *
           (Arith.nat *
             (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list)) *
        ('a Value.vala list *
          ('a Value.vala list * (string * (string * Arith.nat))))) list)))
                                      Predicate.pred
end = struct

fun exception_step B_ typeof_addr p a h
  ((insa, (ins, xt)), (stk, (loc, (c, (m, pc))))) frs =
  (case JVMExceptions.match_ex_table p (Heap.cname_of B_ typeof_addr h a) pc xt
    of NONE => (SOME a, (h, frs))
    | SOME (pca, d) =>
      (NONE,
        (h, ((List.drop pca ins, (ins, xt)),
              (Value.Addr a ::
                 List.drop (Arith.minus_nat (List.size_list stk) d) stk,
                (loc, (c, (m, pca))))) ::
              frs)));

fun extRet2JVM insa ins xt n h stk loc c m pc frs (ExternalCall.RetVal v) =
  (NONE,
    (h, ((List.tl insa, (ins, xt)),
          (v :: List.drop (Arith.suc n) stk,
            (loc, (c, (m, Arith.plus_nat pc Arith.one_nat))))) ::
          frs))
  | extRet2JVM insa ins xt n h stk loc c m pc frs (ExternalCall.RetExc a) =
    (SOME a, (h, ((insa, (ins, xt)), (stk, (loc, (c, (m, pc))))) :: frs))
  | extRet2JVM insa ins xt n h stk loc c m pc frs ExternalCall.RetStaySame =
    (NONE, (h, ((insa, (ins, xt)), (stk, (loc, (c, (m, pc))))) :: frs));

fun extNTA2JVM A_ p =
  (fn (c, (m, a)) =>
    let
      val (d, (_, (_, meth))) =
        TypeRel.method
          (Product_Type.equal_prod Arith.equal_nat
            (Product_Type.equal_prod Arith.equal_nat
              (Product_Type.equal_prod
                (List.equal_list (JVMInstructions.equal_instr A_))
                (List.equal_list
                  (Product_Type.equal_prod Arith.equal_nat
                    (Product_Type.equal_prod Arith.equal_nat
                      (Product_Type.equal_prod
                        (Option.equal_option Stringa.equal_literal)
                        (Product_Type.equal_prod Arith.equal_nat
                          Arith.equal_nat))))))))
          p c m;
      val (_, (mxl0, (ins, xt))) = Option.the meth;
    in
      (NONE,
        [((ins, (ins, xt)),
           ([], (Value.Addr a :: List.replicate mxl0 Value.undefined_value,
                  (d, (m, Arith.zero_nat)))))])
    end);

fun exec_instr (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr
  spurious_wakeups empty_heap allocate typeof_addr heap_read heap_write insa ins
  xt JVMInstructions.MExit p t h stk loc c_0 m_0 pc frs =
  let
    val v = List.hd stk;
  in
    (if Value.equal_vala A2_ v Value.Null
      then Set.insert
             ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
               (SOME (StartConfig.addr_of_sys_xcpt C_ (A2_, A3_) empty_heap
                       allocate "java/lang/NullPointerException"),
                 (h, ((insa, (ins, xt)), (stk, (loc, (c_0, (m_0, pc))))) ::
                       frs)))
             Set.bot_seta
      else Set.insert
             (FWState.thread_action_to_thread_action A2_
                Observable_Events.obs_action_obs_event
                (FWState.ObsAction
                  (Observable_Events.SyncUnlock (Value.the_Addr v)))
                (FWState.thread_action_to_thread_action A2_
                  Observable_Events.obs_action_obs_event
                  (FWState.LockAction (FWState.Unlock, Value.the_Addr v))
                  (FinFun.Finfun_const [], ([], ([], ([], ([], [])))))),
               (NONE,
                 (h, ((List.tl insa, (ins, xt)),
                       (List.tl stk,
                         (loc, (c_0, (m_0, Arith.plus_nat pc
     Arith.one_nat))))) ::
                       frs)))
             (Set.insert
               (FWState.thread_action_to_thread_action A2_
                  Observable_Events.obs_action_obs_event
                  (FWState.LockAction (FWState.UnlockFail, Value.the_Addr v))
                  (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                 (SOME (StartConfig.addr_of_sys_xcpt C_ (A2_, A3_) empty_heap
                         allocate "java/lang/IllegalMonitorStateException"),
                   (h, ((insa, (ins, xt)), (stk, (loc, (c_0, (m_0, pc))))) ::
                         frs)))
               Set.bot_seta))
  end
  | exec_instr (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr
    spurious_wakeups empty_heap allocate typeof_addr heap_read heap_write insa
    ins xt JVMInstructions.MEnter p t h stk loc c_0 m_0 pc frs =
    Set.insert
      let
        val v = List.hd stk;
      in
        (if Value.equal_vala A2_ v Value.Null
          then ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                 (SOME (StartConfig.addr_of_sys_xcpt C_ (A2_, A3_) empty_heap
                         allocate "java/lang/NullPointerException"),
                   (h, ((insa, (ins, xt)), (stk, (loc, (c_0, (m_0, pc))))) ::
                         frs)))
          else (FWState.thread_action_to_thread_action A2_
                  Observable_Events.obs_action_obs_event
                  (FWState.ObsAction
                    (Observable_Events.SyncLock (Value.the_Addr v)))
                  (FWState.thread_action_to_thread_action A2_
                    Observable_Events.obs_action_obs_event
                    (FWState.LockAction (FWState.Lock, Value.the_Addr v))
                    (FinFun.Finfun_const [], ([], ([], ([], ([], [])))))),
                 (NONE,
                   (h, ((List.tl insa, (ins, xt)),
                         (List.tl stk,
                           (loc, (c_0, (m_0,
 Arith.plus_nat pc Arith.one_nat))))) ::
                         frs))))
      end
      Set.bot_seta
  | exec_instr (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr
    spurious_wakeups empty_heap allocate typeof_addr heap_read heap_write insa
    ins xt JVMInstructions.ThrowExc p t h stk loc c_0 m_0 pc frs =
    Set.insert
      ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
        let
          val xp =
            (if Value.equal_vala A2_ (List.hd stk) Value.Null
              then SOME (StartConfig.addr_of_sys_xcpt C_ (A2_, A3_) empty_heap
                          allocate "java/lang/NullPointerException")
              else SOME (Value.the_Addr (List.hd stk)));
        in
          (xp, (h, ((insa, (ins, xt)), (stk, (loc, (c_0, (m_0, pc))))) :: frs))
        end)
      Set.bot_seta
  | exec_instr (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr
    spurious_wakeups empty_heap allocate typeof_addr heap_read heap_write insa
    ins xt (JVMInstructions.Goto i) p t h stk loc c_0 m_0 pc frs =
    Set.insert
      let
        val pca = Arith.nat (Arith.plus_int (Arith.int_of_nat pc) i);
      in
        ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
          (NONE,
            (h, ((List.drop pca ins, (ins, xt)),
                  (stk, (loc, (c_0, (m_0, pca))))) ::
                  frs)))
      end
      Set.bot_seta
  | exec_instr (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr
    spurious_wakeups empty_heap allocate typeof_addr heap_read heap_write insa
    ins xt (JVMInstructions.IfFalse i) p t h stk loc c_0 m_0 pc frs =
    Set.insert
      ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
        let
          val pca =
            (if Value.equal_vala A2_ (List.hd stk) (Value.Bool false)
              then Arith.nat (Arith.plus_int (Arith.int_of_nat pc) i)
              else Arith.plus_nat pc Arith.one_nat);
        in
          (NONE,
            (h, ((List.drop pca ins, (ins, xt)),
                  (List.tl stk, (loc, (c_0, (m_0, pca))))) ::
                  frs))
        end)
      Set.bot_seta
  | exec_instr (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr
    spurious_wakeups empty_heap allocate typeof_addr heap_read heap_write insa
    ins xt (JVMInstructions.BinOpInstr bop) p t h stk loc c0 m0 pc frs =
    Set.insert
      ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
        (case Option.the
                (BinOp.binop C_ (A2_, A3_) empty_heap allocate bop
                  (List.hd (List.tl stk)) (List.hd stk))
          of Sum_Type.Inl v =>
            (NONE,
              (h, ((List.tl insa, (ins, xt)),
                    (v :: List.tl (List.tl stk),
                      (loc, (c0, (m0, Arith.plus_nat pc Arith.one_nat))))) ::
                    frs))
          | Sum_Type.Inr a =>
            (SOME a,
              (h, ((insa, (ins, xt)), (stk, (loc, (c0, (m0, pc))))) :: frs))))
      Set.bot_seta
  | exec_instr (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr
    spurious_wakeups empty_heap allocate typeof_addr heap_read heap_write insa
    ins xt JVMInstructions.Swap p t h stk loc c_0 m_0 pc frs =
    Set.insert
      ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
        (NONE,
          (h, ((List.tl insa, (ins, xt)),
                (List.hd (List.tl stk) :: List.hd stk :: List.tl (List.tl stk),
                  (loc, (c_0, (m_0, Arith.plus_nat pc Arith.one_nat))))) ::
                frs)))
      Set.bot_seta
  | exec_instr (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr
    spurious_wakeups empty_heap allocate typeof_addr heap_read heap_write insa
    ins xt JVMInstructions.Dup p t h stk loc c_0 m_0 pc frs =
    Set.insert
      ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
        (NONE,
          (h, ((List.tl insa, (ins, xt)),
                (List.hd stk :: stk,
                  (loc, (c_0, (m_0, Arith.plus_nat pc Arith.one_nat))))) ::
                frs)))
      Set.bot_seta
  | exec_instr (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr
    spurious_wakeups empty_heap allocate typeof_addr heap_read heap_write insa
    ins xt JVMInstructions.Pop p t h stk loc c_0 m_0 pc frs =
    Set.insert
      ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
        (NONE,
          (h, ((List.tl insa, (ins, xt)),
                (List.tl stk,
                  (loc, (c_0, (m_0, Arith.plus_nat pc Arith.one_nat))))) ::
                frs)))
      Set.bot_seta
  | exec_instr (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr
    spurious_wakeups empty_heap allocate typeof_addr heap_read heap_write insa
    ins xt JVMInstructions.Return p t h stk_0 loc_0 c_0 m_0 pc frs =
    Set.insert
      ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
        (if List.null frs then (NONE, (h, []))
          else let
                 val v = List.hd stk_0;
                 val (a, b) = List.hd frs;
               in
                 let
                   val (insb, (insaa, xta)) = a;
                 in
                   (fn (stk, (loc, (c, (m, pca)))) =>
                     let
                       val n =
                         List.size_list
                           (Product_Type.fst
                             (Product_Type.snd
                               (TypeRel.method
                                 (Product_Type.equal_prod Arith.equal_nat
                                   (Product_Type.equal_prod Arith.equal_nat
                                     (Product_Type.equal_prod
                                       (List.equal_list
 (JVMInstructions.equal_instr A2_))
                                       (List.equal_list
 (Product_Type.equal_prod Arith.equal_nat
   (Product_Type.equal_prod Arith.equal_nat
     (Product_Type.equal_prod (Option.equal_option Stringa.equal_literal)
       (Product_Type.equal_prod Arith.equal_nat Arith.equal_nat))))))))
                                 p c_0 m_0)));
                     in
                       (NONE,
                         (h, ((List.tl insb, (insaa, xta)),
                               (v :: List.drop (Arith.plus_nat n Arith.one_nat)
                                       stk,
                                 (loc, (c,
 (m, Arith.plus_nat pca Arith.one_nat))))) ::
                               List.tl frs))
                     end)
                 end
                   b
               end))
      Set.bot_seta
  | exec_instr (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr
    spurious_wakeups empty_heap allocate typeof_addr heap_read heap_write insa
    ins xt (JVMInstructions.Invoke (m, n)) p t h stk loc c0 m0 pc frs =
    let
      val r = List.nth stk n;
    in
      (if Value.equal_vala A2_ r Value.Null
        then Set.insert
               ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                 (SOME (StartConfig.addr_of_sys_xcpt C_ (A2_, A3_) empty_heap
                         allocate "java/lang/NullPointerException"),
                   (h, ((insa, (ins, xt)), (stk, (loc, (c0, (m0, pc))))) ::
                         frs)))
               Set.bot_seta
        else let
               val ps = List.rev (List.take n stk);
               val a = Value.the_Addr r;
               val ta = Option.the (typeof_addr h a);
             in
               (case TypeRel.method
                       (Product_Type.equal_prod Arith.equal_nat
                         (Product_Type.equal_prod Arith.equal_nat
                           (Product_Type.equal_prod
                             (List.equal_list (JVMInstructions.equal_instr A2_))
                             (List.equal_list
                               (Product_Type.equal_prod Arith.equal_nat
                                 (Product_Type.equal_prod Arith.equal_nat
                                   (Product_Type.equal_prod
                                     (Option.equal_option Stringa.equal_literal)
                                     (Product_Type.equal_prod Arith.equal_nat
                                       Arith.equal_nat))))))))
                       p (Type.class_type_ofa ta) m
                 of (d, (_, (_, NONE))) =>
                   Set.bind
                     (Product_Type.equal_prod
                       (Product_Type.equal_prod
                         (FinFun.equal_finfun (A1_, A2_)
                           (List.equal_list FWState.equal_lock_action))
                         (Product_Type.equal_prod
                           (List.equal_list
                             (FWState.equal_new_thread_action B_
                               (Product_Type.equal_prod
                                 (Option.equal_option A2_)
                                 (List.equal_list
                                   (Product_Type.equal_prod
                                     (Product_Type.equal_prod
                                       (List.equal_list
 (JVMInstructions.equal_instr A2_))
                                       (Product_Type.equal_prod
 (List.equal_list (JVMInstructions.equal_instr A2_))
 (List.equal_list
   (Product_Type.equal_prod Arith.equal_nat
     (Product_Type.equal_prod Arith.equal_nat
       (Product_Type.equal_prod (Option.equal_option Stringa.equal_literal)
         (Product_Type.equal_prod Arith.equal_nat Arith.equal_nat)))))))
                                     (Product_Type.equal_prod
                                       (List.equal_list (Value.equal_val A2_))
                                       (Product_Type.equal_prod
 (List.equal_list (Value.equal_val A2_))
 (Product_Type.equal_prod Stringa.equal_literal
   (Product_Type.equal_prod Stringa.equal_literal Arith.equal_nat)))))))
                               C_))
                           (Product_Type.equal_prod
                             (List.equal_list
                               (FWState.equal_conditional_action B_))
                             (Product_Type.equal_prod
                               (List.equal_list
                                 (FWState.equal_wait_set_action B_ A2_))
                               (Product_Type.equal_prod
                                 (List.equal_list
                                   (FWState.equal_interrupt_action B_))
                                 (List.equal_list
                                   (Observable_Events.equal_obs_event A2_
                                     B_)))))))
                       (Product_Type.equal_prod (Option.equal_option A2_)
                         (Product_Type.equal_prod C_
                           (List.equal_list
                             (Product_Type.equal_prod
                               (Product_Type.equal_prod
                                 (List.equal_list
                                   (JVMInstructions.equal_instr A2_))
                                 (Product_Type.equal_prod
                                   (List.equal_list
                                     (JVMInstructions.equal_instr A2_))
                                   (List.equal_list
                                     (Product_Type.equal_prod Arith.equal_nat
                                       (Product_Type.equal_prod Arith.equal_nat
 (Product_Type.equal_prod (Option.equal_option Stringa.equal_literal)
   (Product_Type.equal_prod Arith.equal_nat Arith.equal_nat)))))))
                               (Product_Type.equal_prod
                                 (List.equal_list (Value.equal_val A2_))
                                 (Product_Type.equal_prod
                                   (List.equal_list (Value.equal_val A2_))
                                   (Product_Type.equal_prod
                                     Stringa.equal_literal
                                     (Product_Type.equal_prod
                                       Stringa.equal_literal
                                       Arith.equal_nat)))))))))
                     (ExternalCall_Execute.red_external_aggr (A1_, A2_, A3_) B_
                       C_ addr2thread_id thread_id2addr spurious_wakeups
                       empty_heap allocate typeof_addr heap_read heap_write p t
                       a m ps h)
                     (fn (taa, (va, ha)) =>
                       Set.insert
                         (FWState.convert_extTA (extNTA2JVM A2_ p) taa,
                           extRet2JVM insa ins xt n ha stk loc c0 m0 pc frs va)
                         Set.bot_seta)
                 | (d, (_, (_, SOME (_, (mxl_0, (insb, xta)))))) =>
                   let
                     val f =
                       ((insb, (insb, xta)),
                         ([], ([r] @ ps @ List.replicate mxl_0
    Value.undefined_value,
                                (d, (m, Arith.zero_nat)))));
                   in
                     Set.insert
                       ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                         (NONE,
                           (h, f :: ((insa, (ins, xt)),
                                      (stk, (loc, (c0, (m0, pc))))) ::
                                      frs)))
                       Set.bot_seta
                   end)
             end)
    end
  | exec_instr (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr
    spurious_wakeups empty_heap allocate typeof_addr heap_read heap_write insa
    ins xt (JVMInstructions.Instanceof ta) p t h stk loc c_0 m_0 pc frs =
    Set.insert
      ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
        (NONE,
          (h, ((List.tl insa, (ins, xt)),
                (Value.Bool
                   (not (Value.equal_vala A2_ (List.hd stk) Value.Null) andalso
                     TypeRel.widen p
                       (Option.the
                         (Heap.typeof_h A3_ typeof_addr h (List.hd stk)))
                       ta) ::
                   List.tl stk,
                  (loc, (c_0, (m_0, Arith.plus_nat pc Arith.one_nat))))) ::
                frs)))
      Set.bot_seta
  | exec_instr (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr
    spurious_wakeups empty_heap allocate typeof_addr heap_read heap_write insa
    ins xt (JVMInstructions.Checkcast ta) p t h stk loc c_0 m_0 pc frs =
    Set.insert
      ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
        let
          val u = Option.the (Heap.typeof_h A3_ typeof_addr h (List.hd stk));
        in
          (if TypeRel.widen p u ta
            then (NONE,
                   (h, ((List.tl insa, (ins, xt)),
                         (stk, (loc, (c_0, (m_0,
     Arith.plus_nat pc Arith.one_nat))))) ::
                         frs))
            else (SOME (StartConfig.addr_of_sys_xcpt C_ (A2_, A3_) empty_heap
                         allocate "java/lang/ClassCastException"),
                   (h, ((insa, (ins, xt)), (stk, (loc, (c_0, (m_0, pc))))) ::
                         frs)))
        end)
      Set.bot_seta
  | exec_instr (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr
    spurious_wakeups empty_heap allocate typeof_addr heap_read heap_write insa
    ins xt (JVMInstructions.Putfield (f, c)) p t h stk loc c_0 m_0 pc frs =
    let
      val v = List.hd stk;
      val r = List.hd (List.tl stk);
    in
      (if Value.equal_vala A2_ r Value.Null
        then Set.insert
               ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                 (SOME (StartConfig.addr_of_sys_xcpt C_ (A2_, A3_) empty_heap
                         allocate "java/lang/NullPointerException"),
                   (h, ((insa, (ins, xt)), (stk, (loc, (c_0, (m_0, pc))))) ::
                         frs)))
               Set.bot_seta
        else let
               val a = Value.the_Addr r;
             in
               Set.bind
                 (Product_Type.equal_prod
                   (Product_Type.equal_prod
                     (FinFun.equal_finfun (A1_, A2_)
                       (List.equal_list FWState.equal_lock_action))
                     (Product_Type.equal_prod
                       (List.equal_list
                         (FWState.equal_new_thread_action B_
                           (Product_Type.equal_prod (Option.equal_option A2_)
                             (List.equal_list
                               (Product_Type.equal_prod
                                 (Product_Type.equal_prod
                                   (List.equal_list
                                     (JVMInstructions.equal_instr A2_))
                                   (Product_Type.equal_prod
                                     (List.equal_list
                                       (JVMInstructions.equal_instr A2_))
                                     (List.equal_list
                                       (Product_Type.equal_prod Arith.equal_nat
 (Product_Type.equal_prod Arith.equal_nat
   (Product_Type.equal_prod (Option.equal_option Stringa.equal_literal)
     (Product_Type.equal_prod Arith.equal_nat Arith.equal_nat)))))))
                                 (Product_Type.equal_prod
                                   (List.equal_list (Value.equal_val A2_))
                                   (Product_Type.equal_prod
                                     (List.equal_list (Value.equal_val A2_))
                                     (Product_Type.equal_prod
                                       Stringa.equal_literal
                                       (Product_Type.equal_prod
 Stringa.equal_literal Arith.equal_nat)))))))
                           C_))
                       (Product_Type.equal_prod
                         (List.equal_list (FWState.equal_conditional_action B_))
                         (Product_Type.equal_prod
                           (List.equal_list
                             (FWState.equal_wait_set_action B_ A2_))
                           (Product_Type.equal_prod
                             (List.equal_list
                               (FWState.equal_interrupt_action B_))
                             (List.equal_list
                               (Observable_Events.equal_obs_event A2_ B_)))))))
                   (Product_Type.equal_prod (Option.equal_option A2_)
                     (Product_Type.equal_prod C_
                       (List.equal_list
                         (Product_Type.equal_prod
                           (Product_Type.equal_prod
                             (List.equal_list (JVMInstructions.equal_instr A2_))
                             (Product_Type.equal_prod
                               (List.equal_list
                                 (JVMInstructions.equal_instr A2_))
                               (List.equal_list
                                 (Product_Type.equal_prod Arith.equal_nat
                                   (Product_Type.equal_prod Arith.equal_nat
                                     (Product_Type.equal_prod
                                       (Option.equal_option
 Stringa.equal_literal)
                                       (Product_Type.equal_prod Arith.equal_nat
 Arith.equal_nat)))))))
                           (Product_Type.equal_prod
                             (List.equal_list (Value.equal_val A2_))
                             (Product_Type.equal_prod
                               (List.equal_list (Value.equal_val A2_))
                               (Product_Type.equal_prod Stringa.equal_literal
                                 (Product_Type.equal_prod Stringa.equal_literal
                                   Arith.equal_nat)))))))))
                 (heap_write h a (Heap.CField (c, f)) v)
                 (fn ha =>
                   Set.insert
                     (FWState.thread_action_to_thread_action A2_
                        Observable_Events.obs_action_obs_event
                        (FWState.ObsAction
                          (Observable_Events.WriteMem
                            (a, Heap.CField (c, f), v)))
                        (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                       (NONE,
                         (ha, ((List.tl insa, (ins, xt)),
                                (List.tl (List.tl stk),
                                  (loc, (c_0,
  (m_0, Arith.plus_nat pc Arith.one_nat))))) ::
                                frs)))
                     Set.bot_seta)
             end)
    end
  | exec_instr (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr
    spurious_wakeups empty_heap allocate typeof_addr heap_read heap_write insa
    ins xt (JVMInstructions.Getfield (f, c)) p t h stk loc c_0 m_0 pc frs =
    let
      val v = List.hd stk;
    in
      (if Value.equal_vala A2_ v Value.Null
        then Set.insert
               ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                 (SOME (StartConfig.addr_of_sys_xcpt C_ (A2_, A3_) empty_heap
                         allocate "java/lang/NullPointerException"),
                   (h, ((insa, (ins, xt)), (stk, (loc, (c_0, (m_0, pc))))) ::
                         frs)))
               Set.bot_seta
        else let
               val a = Value.the_Addr v;
             in
               Set.bind
                 (Product_Type.equal_prod
                   (Product_Type.equal_prod
                     (FinFun.equal_finfun (A1_, A2_)
                       (List.equal_list FWState.equal_lock_action))
                     (Product_Type.equal_prod
                       (List.equal_list
                         (FWState.equal_new_thread_action B_
                           (Product_Type.equal_prod (Option.equal_option A2_)
                             (List.equal_list
                               (Product_Type.equal_prod
                                 (Product_Type.equal_prod
                                   (List.equal_list
                                     (JVMInstructions.equal_instr A2_))
                                   (Product_Type.equal_prod
                                     (List.equal_list
                                       (JVMInstructions.equal_instr A2_))
                                     (List.equal_list
                                       (Product_Type.equal_prod Arith.equal_nat
 (Product_Type.equal_prod Arith.equal_nat
   (Product_Type.equal_prod (Option.equal_option Stringa.equal_literal)
     (Product_Type.equal_prod Arith.equal_nat Arith.equal_nat)))))))
                                 (Product_Type.equal_prod
                                   (List.equal_list (Value.equal_val A2_))
                                   (Product_Type.equal_prod
                                     (List.equal_list (Value.equal_val A2_))
                                     (Product_Type.equal_prod
                                       Stringa.equal_literal
                                       (Product_Type.equal_prod
 Stringa.equal_literal Arith.equal_nat)))))))
                           C_))
                       (Product_Type.equal_prod
                         (List.equal_list (FWState.equal_conditional_action B_))
                         (Product_Type.equal_prod
                           (List.equal_list
                             (FWState.equal_wait_set_action B_ A2_))
                           (Product_Type.equal_prod
                             (List.equal_list
                               (FWState.equal_interrupt_action B_))
                             (List.equal_list
                               (Observable_Events.equal_obs_event A2_ B_)))))))
                   (Product_Type.equal_prod (Option.equal_option A2_)
                     (Product_Type.equal_prod C_
                       (List.equal_list
                         (Product_Type.equal_prod
                           (Product_Type.equal_prod
                             (List.equal_list (JVMInstructions.equal_instr A2_))
                             (Product_Type.equal_prod
                               (List.equal_list
                                 (JVMInstructions.equal_instr A2_))
                               (List.equal_list
                                 (Product_Type.equal_prod Arith.equal_nat
                                   (Product_Type.equal_prod Arith.equal_nat
                                     (Product_Type.equal_prod
                                       (Option.equal_option
 Stringa.equal_literal)
                                       (Product_Type.equal_prod Arith.equal_nat
 Arith.equal_nat)))))))
                           (Product_Type.equal_prod
                             (List.equal_list (Value.equal_val A2_))
                             (Product_Type.equal_prod
                               (List.equal_list (Value.equal_val A2_))
                               (Product_Type.equal_prod Stringa.equal_literal
                                 (Product_Type.equal_prod Stringa.equal_literal
                                   Arith.equal_nat)))))))))
                 (heap_read h a (Heap.CField (c, f)))
                 (fn va =>
                   Set.insert
                     (FWState.thread_action_to_thread_action A2_
                        Observable_Events.obs_action_obs_event
                        (FWState.ObsAction
                          (Observable_Events.ReadMem
                            (a, Heap.CField (c, f), va)))
                        (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                       (NONE,
                         (h, ((List.tl insa, (ins, xt)),
                               (va :: List.tl stk,
                                 (loc, (c_0,
 (m_0, Arith.plus_nat pc Arith.one_nat))))) ::
                               frs)))
                     Set.bot_seta)
             end)
    end
  | exec_instr (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr
    spurious_wakeups empty_heap allocate typeof_addr heap_read heap_write insa
    ins xt JVMInstructions.ALength p t h stk loc c0 m0 pc frs =
    Set.insert
      ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
        let
          val va = List.hd stk;
        in
          (if Value.equal_vala A2_ va Value.Null
            then (SOME (StartConfig.addr_of_sys_xcpt C_ (A2_, A3_) empty_heap
                         allocate "java/lang/NullPointerException"),
                   (h, ((insa, (ins, xt)), (stk, (loc, (c0, (m0, pc))))) ::
                         frs))
            else (NONE,
                   (h, ((List.tl insa, (ins, xt)),
                         (Value.Intg
                            (Word.word_of_int
                              (Type_Length.len0_bit0
                                (Type_Length.len0_bit0
                                  (Type_Length.len0_bit0
                                    (Type_Length.len0_bit0
                                      (Type_Length.len0_bit0
Type_Length.len0_num1)))))
                              (Arith.int_of_nat
                                (Type.alen_of_htype
                                  (Option.the
                                    (typeof_addr h (Value.the_Addr va)))))) ::
                            List.tl stk,
                           (loc, (c0, (m0,
Arith.plus_nat pc Arith.one_nat))))) ::
                         frs)))
        end)
      Set.bot_seta
  | exec_instr (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr
    spurious_wakeups empty_heap allocate typeof_addr heap_read heap_write insa
    ins xt JVMInstructions.AStore p t h stk loc c0 m0 pc frs =
    let
      val ve = List.hd stk;
      val vi = List.hd (List.tl stk);
      val va = List.hd (List.tl (List.tl stk));
    in
      (if Value.equal_vala A2_ va Value.Null
        then Set.insert
               ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                 (SOME (StartConfig.addr_of_sys_xcpt C_ (A2_, A3_) empty_heap
                         allocate "java/lang/NullPointerException"),
                   (h, ((insa, (ins, xt)), (stk, (loc, (c0, (m0, pc))))) ::
                         frs)))
               Set.bot_seta
        else let
               val i = Value.the_Intg vi;
               val idx =
                 Arith.nat
                   (Word.sint
                     (Type_Length.len_bit0
                       (Type_Length.len_bit0
                         (Type_Length.len_bit0
                           (Type_Length.len_bit0
                             (Type_Length.len_bit0 Type_Length.len_num1)))))
                     i);
               val a = Value.the_Addr va;
               val hT = Option.the (typeof_addr h a);
               val ta = Type.ty_of_htype hT;
               val len = Type.alen_of_htype hT;
               val u = Option.the (Heap.typeof_h A3_ typeof_addr h ve);
             in
               (if Word.word_sless
                     (Type_Length.len_bit0
                       (Type_Length.len_bit0
                         (Type_Length.len_bit0
                           (Type_Length.len_bit0
                             (Type_Length.len_bit0 Type_Length.len_num1)))))
                     i (Word.zero_word
                         (Type_Length.len0_bit0
                           (Type_Length.len0_bit0
                             (Type_Length.len0_bit0
                               (Type_Length.len0_bit0
                                 (Type_Length.len0_bit0
                                   Type_Length.len0_num1)))))) orelse
                     Arith.less_eq_int (Arith.int_of_nat len)
                       (Word.sint
                         (Type_Length.len_bit0
                           (Type_Length.len_bit0
                             (Type_Length.len_bit0
                               (Type_Length.len_bit0
                                 (Type_Length.len_bit0 Type_Length.len_num1)))))
                         i)
                 then Set.insert
                        ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                          (SOME (StartConfig.addr_of_sys_xcpt C_ (A2_, A3_)
                                  empty_heap allocate
                                  "java/lang/ArrayIndexOutOfBoundsException"),
                            (h, ((insa, (ins, xt)),
                                  (stk, (loc, (c0, (m0, pc))))) ::
                                  frs)))
                        Set.bot_seta
                 else (if TypeRel.widen p u (Type.the_Array ta)
                        then Set.bind
                               (Product_Type.equal_prod
                                 (Product_Type.equal_prod
                                   (FinFun.equal_finfun (A1_, A2_)
                                     (List.equal_list
                                       FWState.equal_lock_action))
                                   (Product_Type.equal_prod
                                     (List.equal_list
                                       (FWState.equal_new_thread_action B_
 (Product_Type.equal_prod (Option.equal_option A2_)
   (List.equal_list
     (Product_Type.equal_prod
       (Product_Type.equal_prod
         (List.equal_list (JVMInstructions.equal_instr A2_))
         (Product_Type.equal_prod
           (List.equal_list (JVMInstructions.equal_instr A2_))
           (List.equal_list
             (Product_Type.equal_prod Arith.equal_nat
               (Product_Type.equal_prod Arith.equal_nat
                 (Product_Type.equal_prod
                   (Option.equal_option Stringa.equal_literal)
                   (Product_Type.equal_prod Arith.equal_nat
                     Arith.equal_nat)))))))
       (Product_Type.equal_prod (List.equal_list (Value.equal_val A2_))
         (Product_Type.equal_prod (List.equal_list (Value.equal_val A2_))
           (Product_Type.equal_prod Stringa.equal_literal
             (Product_Type.equal_prod Stringa.equal_literal
               Arith.equal_nat)))))))
 C_))
                                     (Product_Type.equal_prod
                                       (List.equal_list
 (FWState.equal_conditional_action B_))
                                       (Product_Type.equal_prod
 (List.equal_list (FWState.equal_wait_set_action B_ A2_))
 (Product_Type.equal_prod (List.equal_list (FWState.equal_interrupt_action B_))
   (List.equal_list (Observable_Events.equal_obs_event A2_ B_)))))))
                                 (Product_Type.equal_prod
                                   (Option.equal_option A2_)
                                   (Product_Type.equal_prod C_
                                     (List.equal_list
                                       (Product_Type.equal_prod
 (Product_Type.equal_prod (List.equal_list (JVMInstructions.equal_instr A2_))
   (Product_Type.equal_prod (List.equal_list (JVMInstructions.equal_instr A2_))
     (List.equal_list
       (Product_Type.equal_prod Arith.equal_nat
         (Product_Type.equal_prod Arith.equal_nat
           (Product_Type.equal_prod (Option.equal_option Stringa.equal_literal)
             (Product_Type.equal_prod Arith.equal_nat Arith.equal_nat)))))))
 (Product_Type.equal_prod (List.equal_list (Value.equal_val A2_))
   (Product_Type.equal_prod (List.equal_list (Value.equal_val A2_))
     (Product_Type.equal_prod Stringa.equal_literal
       (Product_Type.equal_prod Stringa.equal_literal Arith.equal_nat)))))))))
                               (heap_write h a (Heap.ACell idx) ve)
                               (fn ha =>
                                 Set.insert
                                   (FWState.thread_action_to_thread_action A2_
                                      Observable_Events.obs_action_obs_event
                                      (FWState.ObsAction
(Observable_Events.WriteMem (a, Heap.ACell idx, ve)))
                                      (FinFun.Finfun_const [],
([], ([], ([], ([], []))))),
                                     (NONE,
                                       (ha,
 ((List.tl insa, (ins, xt)),
   (List.tl (List.tl (List.tl stk)),
     (loc, (c0, (m0, Arith.plus_nat pc Arith.one_nat))))) ::
   frs)))
                                   Set.bot_seta)
                        else Set.insert
                               ((FinFun.Finfun_const [],
                                  ([], ([], ([], ([], []))))),
                                 (SOME (StartConfig.addr_of_sys_xcpt C_
 (A2_, A3_) empty_heap allocate "java/lang/ArrayStoreException"),
                                   (h, ((insa, (ins, xt)),
 (stk, (loc, (c0, (m0, pc))))) ::
 frs)))
                               Set.bot_seta))
             end)
    end
  | exec_instr (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr
    spurious_wakeups empty_heap allocate typeof_addr heap_read heap_write insa
    ins xt JVMInstructions.ALoad p t h stk loc c0 m0 pc frs =
    let
      val va = List.hd (List.tl stk);
    in
      (if Value.equal_vala A2_ va Value.Null
        then Set.insert
               ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                 (SOME (StartConfig.addr_of_sys_xcpt C_ (A2_, A3_) empty_heap
                         allocate "java/lang/NullPointerException"),
                   (h, ((insa, (ins, xt)), (stk, (loc, (c0, (m0, pc))))) ::
                         frs)))
               Set.bot_seta
        else let
               val i = Value.the_Intg (List.hd stk);
               val a = Value.the_Addr va;
               val len = Type.alen_of_htype (Option.the (typeof_addr h a));
             in
               (if Word.word_sless
                     (Type_Length.len_bit0
                       (Type_Length.len_bit0
                         (Type_Length.len_bit0
                           (Type_Length.len_bit0
                             (Type_Length.len_bit0 Type_Length.len_num1)))))
                     i (Word.zero_word
                         (Type_Length.len0_bit0
                           (Type_Length.len0_bit0
                             (Type_Length.len0_bit0
                               (Type_Length.len0_bit0
                                 (Type_Length.len0_bit0
                                   Type_Length.len0_num1)))))) orelse
                     Arith.less_eq_int (Arith.int_of_nat len)
                       (Word.sint
                         (Type_Length.len_bit0
                           (Type_Length.len_bit0
                             (Type_Length.len_bit0
                               (Type_Length.len_bit0
                                 (Type_Length.len_bit0 Type_Length.len_num1)))))
                         i)
                 then Set.insert
                        ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                          (SOME (StartConfig.addr_of_sys_xcpt C_ (A2_, A3_)
                                  empty_heap allocate
                                  "java/lang/ArrayIndexOutOfBoundsException"),
                            (h, ((insa, (ins, xt)),
                                  (stk, (loc, (c0, (m0, pc))))) ::
                                  frs)))
                        Set.bot_seta
                 else Set.bind
                        (Product_Type.equal_prod
                          (Product_Type.equal_prod
                            (FinFun.equal_finfun (A1_, A2_)
                              (List.equal_list FWState.equal_lock_action))
                            (Product_Type.equal_prod
                              (List.equal_list
                                (FWState.equal_new_thread_action B_
                                  (Product_Type.equal_prod
                                    (Option.equal_option A2_)
                                    (List.equal_list
                                      (Product_Type.equal_prod
(Product_Type.equal_prod (List.equal_list (JVMInstructions.equal_instr A2_))
  (Product_Type.equal_prod (List.equal_list (JVMInstructions.equal_instr A2_))
    (List.equal_list
      (Product_Type.equal_prod Arith.equal_nat
        (Product_Type.equal_prod Arith.equal_nat
          (Product_Type.equal_prod (Option.equal_option Stringa.equal_literal)
            (Product_Type.equal_prod Arith.equal_nat Arith.equal_nat)))))))
(Product_Type.equal_prod (List.equal_list (Value.equal_val A2_))
  (Product_Type.equal_prod (List.equal_list (Value.equal_val A2_))
    (Product_Type.equal_prod Stringa.equal_literal
      (Product_Type.equal_prod Stringa.equal_literal Arith.equal_nat)))))))
                                  C_))
                              (Product_Type.equal_prod
                                (List.equal_list
                                  (FWState.equal_conditional_action B_))
                                (Product_Type.equal_prod
                                  (List.equal_list
                                    (FWState.equal_wait_set_action B_ A2_))
                                  (Product_Type.equal_prod
                                    (List.equal_list
                                      (FWState.equal_interrupt_action B_))
                                    (List.equal_list
                                      (Observable_Events.equal_obs_event A2_
B_)))))))
                          (Product_Type.equal_prod (Option.equal_option A2_)
                            (Product_Type.equal_prod C_
                              (List.equal_list
                                (Product_Type.equal_prod
                                  (Product_Type.equal_prod
                                    (List.equal_list
                                      (JVMInstructions.equal_instr A2_))
                                    (Product_Type.equal_prod
                                      (List.equal_list
(JVMInstructions.equal_instr A2_))
                                      (List.equal_list
(Product_Type.equal_prod Arith.equal_nat
  (Product_Type.equal_prod Arith.equal_nat
    (Product_Type.equal_prod (Option.equal_option Stringa.equal_literal)
      (Product_Type.equal_prod Arith.equal_nat Arith.equal_nat)))))))
                                  (Product_Type.equal_prod
                                    (List.equal_list (Value.equal_val A2_))
                                    (Product_Type.equal_prod
                                      (List.equal_list (Value.equal_val A2_))
                                      (Product_Type.equal_prod
Stringa.equal_literal
(Product_Type.equal_prod Stringa.equal_literal Arith.equal_nat)))))))))
                        (heap_read h a
                          (Heap.ACell
                            (Arith.nat
                              (Word.sint
                                (Type_Length.len_bit0
                                  (Type_Length.len_bit0
                                    (Type_Length.len_bit0
                                      (Type_Length.len_bit0
(Type_Length.len_bit0 Type_Length.len_num1)))))
                                i))))
                        (fn v =>
                          Set.insert
                            (FWState.thread_action_to_thread_action A2_
                               Observable_Events.obs_action_obs_event
                               (FWState.ObsAction
                                 (Observable_Events.ReadMem
                                   (a, Heap.ACell
 (Arith.nat
   (Word.sint
     (Type_Length.len_bit0
       (Type_Length.len_bit0
         (Type_Length.len_bit0
           (Type_Length.len_bit0 (Type_Length.len_bit0 Type_Length.len_num1)))))
     i)),
                                     v)))
                               (FinFun.Finfun_const [],
                                 ([], ([], ([], ([], []))))),
                              (NONE,
                                (h, ((List.tl insa, (ins, xt)),
                                      (v :: List.tl (List.tl stk),
(loc, (c0, (m0, Arith.plus_nat pc Arith.one_nat))))) ::
                                      frs)))
                            Set.bot_seta))
             end)
    end
  | exec_instr (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr
    spurious_wakeups empty_heap allocate typeof_addr heap_read heap_write insa
    ins xt (JVMInstructions.NewArray ta) p t h stk loc c0 m0 pc frs =
    let
      val si = Value.the_Intg (List.hd stk);
      val i =
        Arith.nat
          (Word.sint
            (Type_Length.len_bit0
              (Type_Length.len_bit0
                (Type_Length.len_bit0
                  (Type_Length.len_bit0
                    (Type_Length.len_bit0 Type_Length.len_num1)))))
            si);
    in
      (if Word.word_sless
            (Type_Length.len_bit0
              (Type_Length.len_bit0
                (Type_Length.len_bit0
                  (Type_Length.len_bit0
                    (Type_Length.len_bit0 Type_Length.len_num1)))))
            si (Word.zero_word
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0
                       (Type_Length.len0_bit0
                         (Type_Length.len0_bit0 Type_Length.len0_num1))))))
        then Set.insert
               ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                 (SOME (StartConfig.addr_of_sys_xcpt C_ (A2_, A3_) empty_heap
                         allocate "java/lang/NegativeArraySizeException"),
                   (h, ((insa, (ins, xt)), (stk, (loc, (c0, (m0, pc))))) ::
                         frs)))
               Set.bot_seta
        else let
               val ha = allocate h (Type.Array_type (ta, i));
             in
               (if Set.equal_set (Product_Type.equal_prod C_ A2_) ha
                     Set.bot_seta
                 then Set.insert
                        ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                          (SOME (StartConfig.addr_of_sys_xcpt C_ (A2_, A3_)
                                  empty_heap allocate
                                  "java/lang/OutOfMemoryError"),
                            (h, ((insa, (ins, xt)),
                                  (stk, (loc, (c0, (m0, pc))))) ::
                                  frs)))
                        Set.bot_seta
                 else Set.bind
                        (Product_Type.equal_prod
                          (Product_Type.equal_prod
                            (FinFun.equal_finfun (A1_, A2_)
                              (List.equal_list FWState.equal_lock_action))
                            (Product_Type.equal_prod
                              (List.equal_list
                                (FWState.equal_new_thread_action B_
                                  (Product_Type.equal_prod
                                    (Option.equal_option A2_)
                                    (List.equal_list
                                      (Product_Type.equal_prod
(Product_Type.equal_prod (List.equal_list (JVMInstructions.equal_instr A2_))
  (Product_Type.equal_prod (List.equal_list (JVMInstructions.equal_instr A2_))
    (List.equal_list
      (Product_Type.equal_prod Arith.equal_nat
        (Product_Type.equal_prod Arith.equal_nat
          (Product_Type.equal_prod (Option.equal_option Stringa.equal_literal)
            (Product_Type.equal_prod Arith.equal_nat Arith.equal_nat)))))))
(Product_Type.equal_prod (List.equal_list (Value.equal_val A2_))
  (Product_Type.equal_prod (List.equal_list (Value.equal_val A2_))
    (Product_Type.equal_prod Stringa.equal_literal
      (Product_Type.equal_prod Stringa.equal_literal Arith.equal_nat)))))))
                                  C_))
                              (Product_Type.equal_prod
                                (List.equal_list
                                  (FWState.equal_conditional_action B_))
                                (Product_Type.equal_prod
                                  (List.equal_list
                                    (FWState.equal_wait_set_action B_ A2_))
                                  (Product_Type.equal_prod
                                    (List.equal_list
                                      (FWState.equal_interrupt_action B_))
                                    (List.equal_list
                                      (Observable_Events.equal_obs_event A2_
B_)))))))
                          (Product_Type.equal_prod (Option.equal_option A2_)
                            (Product_Type.equal_prod C_
                              (List.equal_list
                                (Product_Type.equal_prod
                                  (Product_Type.equal_prod
                                    (List.equal_list
                                      (JVMInstructions.equal_instr A2_))
                                    (Product_Type.equal_prod
                                      (List.equal_list
(JVMInstructions.equal_instr A2_))
                                      (List.equal_list
(Product_Type.equal_prod Arith.equal_nat
  (Product_Type.equal_prod Arith.equal_nat
    (Product_Type.equal_prod (Option.equal_option Stringa.equal_literal)
      (Product_Type.equal_prod Arith.equal_nat Arith.equal_nat)))))))
                                  (Product_Type.equal_prod
                                    (List.equal_list (Value.equal_val A2_))
                                    (Product_Type.equal_prod
                                      (List.equal_list (Value.equal_val A2_))
                                      (Product_Type.equal_prod
Stringa.equal_literal
(Product_Type.equal_prod Stringa.equal_literal Arith.equal_nat)))))))))
                        ha (fn (hb, a) =>
                             Set.insert
                               (FWState.thread_action_to_thread_action A2_
                                  Observable_Events.obs_action_obs_event
                                  (FWState.ObsAction
                                    (Observable_Events.NewHeapElem
                                      (a, Type.Array_type (ta, i))))
                                  (FinFun.Finfun_const [],
                                    ([], ([], ([], ([], []))))),
                                 (NONE,
                                   (hb, ((List.tl insa, (ins, xt)),
  (Value.Addr a :: List.tl stk,
    (loc, (c0, (m0, Arith.plus_nat pc Arith.one_nat))))) ::
  frs)))
                               Set.bot_seta))
             end)
    end
  | exec_instr (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr
    spurious_wakeups empty_heap allocate typeof_addr heap_read heap_write insa
    ins xt (JVMInstructions.New c) p t h stk loc c_0 m_0 pc frs =
    let
      val ha = allocate h (Type.Class_type c);
    in
      (if Set.equal_set (Product_Type.equal_prod C_ A2_) ha Set.bot_seta
        then Set.insert
               ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                 (SOME (StartConfig.addr_of_sys_xcpt C_ (A2_, A3_) empty_heap
                         allocate "java/lang/OutOfMemoryError"),
                   (h, ((insa, (ins, xt)), (stk, (loc, (c_0, (m_0, pc))))) ::
                         frs)))
               Set.bot_seta
        else Set.bind
               (Product_Type.equal_prod
                 (Product_Type.equal_prod
                   (FinFun.equal_finfun (A1_, A2_)
                     (List.equal_list FWState.equal_lock_action))
                   (Product_Type.equal_prod
                     (List.equal_list
                       (FWState.equal_new_thread_action B_
                         (Product_Type.equal_prod (Option.equal_option A2_)
                           (List.equal_list
                             (Product_Type.equal_prod
                               (Product_Type.equal_prod
                                 (List.equal_list
                                   (JVMInstructions.equal_instr A2_))
                                 (Product_Type.equal_prod
                                   (List.equal_list
                                     (JVMInstructions.equal_instr A2_))
                                   (List.equal_list
                                     (Product_Type.equal_prod Arith.equal_nat
                                       (Product_Type.equal_prod Arith.equal_nat
 (Product_Type.equal_prod (Option.equal_option Stringa.equal_literal)
   (Product_Type.equal_prod Arith.equal_nat Arith.equal_nat)))))))
                               (Product_Type.equal_prod
                                 (List.equal_list (Value.equal_val A2_))
                                 (Product_Type.equal_prod
                                   (List.equal_list (Value.equal_val A2_))
                                   (Product_Type.equal_prod
                                     Stringa.equal_literal
                                     (Product_Type.equal_prod
                                       Stringa.equal_literal
                                       Arith.equal_nat)))))))
                         C_))
                     (Product_Type.equal_prod
                       (List.equal_list (FWState.equal_conditional_action B_))
                       (Product_Type.equal_prod
                         (List.equal_list
                           (FWState.equal_wait_set_action B_ A2_))
                         (Product_Type.equal_prod
                           (List.equal_list (FWState.equal_interrupt_action B_))
                           (List.equal_list
                             (Observable_Events.equal_obs_event A2_ B_)))))))
                 (Product_Type.equal_prod (Option.equal_option A2_)
                   (Product_Type.equal_prod C_
                     (List.equal_list
                       (Product_Type.equal_prod
                         (Product_Type.equal_prod
                           (List.equal_list (JVMInstructions.equal_instr A2_))
                           (Product_Type.equal_prod
                             (List.equal_list (JVMInstructions.equal_instr A2_))
                             (List.equal_list
                               (Product_Type.equal_prod Arith.equal_nat
                                 (Product_Type.equal_prod Arith.equal_nat
                                   (Product_Type.equal_prod
                                     (Option.equal_option Stringa.equal_literal)
                                     (Product_Type.equal_prod Arith.equal_nat
                                       Arith.equal_nat)))))))
                         (Product_Type.equal_prod
                           (List.equal_list (Value.equal_val A2_))
                           (Product_Type.equal_prod
                             (List.equal_list (Value.equal_val A2_))
                             (Product_Type.equal_prod Stringa.equal_literal
                               (Product_Type.equal_prod Stringa.equal_literal
                                 Arith.equal_nat)))))))))
               ha (fn (hb, a) =>
                    Set.insert
                      (FWState.thread_action_to_thread_action A2_
                         Observable_Events.obs_action_obs_event
                         (FWState.ObsAction
                           (Observable_Events.NewHeapElem
                             (a, Type.Class_type c)))
                         (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                        (NONE,
                          (hb, ((List.tl insa, (ins, xt)),
                                 (Value.Addr a :: stk,
                                   (loc, (c_0,
   (m_0, Arith.plus_nat pc Arith.one_nat))))) ::
                                 frs)))
                      Set.bot_seta))
    end
  | exec_instr (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr
    spurious_wakeups empty_heap allocate typeof_addr heap_read heap_write insa
    ins xt (JVMInstructions.Push v) p t h stk loc c_0 m_0 pc frs =
    Set.insert
      ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
        (NONE,
          (h, ((List.tl insa, (ins, xt)),
                (v :: stk,
                  (loc, (c_0, (m_0, Arith.plus_nat pc Arith.one_nat))))) ::
                frs)))
      Set.bot_seta
  | exec_instr (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr
    spurious_wakeups empty_heap allocate typeof_addr heap_read heap_write insa
    ins xt (JVMInstructions.Store n) p t h stk loc c_0 m_0 pc frs =
    Set.insert
      ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
        (NONE,
          (h, ((List.tl insa, (ins, xt)),
                (List.tl stk,
                  (List.list_update loc n (List.hd stk),
                    (c_0, (m_0, Arith.plus_nat pc Arith.one_nat))))) ::
                frs)))
      Set.bot_seta
  | exec_instr (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr
    spurious_wakeups empty_heap allocate typeof_addr heap_read heap_write insa
    ins xt (JVMInstructions.Load n) p t h stk loc c_0 m_0 pc frs =
    Set.insert
      ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
        (NONE,
          (h, ((List.tl insa, (ins, xt)),
                (List.nth loc n :: stk,
                  (loc, (c_0, (m_0, Arith.plus_nat pc Arith.one_nat))))) ::
                frs)))
      Set.bot_seta;

fun exec (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr spurious_wakeups
  empty_heap allocate typeof_addr heap_read heap_write p t
  (SOME a, (h, fr :: frs)) =
  Set.insert
    ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
      exception_step A3_ typeof_addr p a h fr frs)
    Set.bot_seta
  | exec (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr spurious_wakeups
    empty_heap allocate typeof_addr heap_read heap_write p t
    (NONE, (h, ((insa, (ins, xt)), (stk, (loc, (c, (m, pc))))) :: frs)) =
    exec_instr (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr
      spurious_wakeups empty_heap allocate typeof_addr heap_read heap_write insa
      ins xt (List.hd insa) p t h stk loc c m pc frs
  | exec (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr spurious_wakeups
    empty_heap allocate typeof_addr heap_read heap_write p t (xcp, (h, [])) =
    Set.bot_seta;

fun exec_1 (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr spurious_wakeups
  empty_heap allocate typeof_addr heap_read heap_write p t sigma =
  Predicate.pred_of_set
    (exec (A1_, A2_, A3_) B_ C_ addr2thread_id thread_id2addr spurious_wakeups
      empty_heap allocate typeof_addr heap_read heap_write p t sigma);

end; (*struct JVMExec_Execute2*)

structure CollectionsV1 : sig
  val equal_rbta :
    'a HOL.equal * 'a Orderings.linorder -> 'b HOL.equal ->
      ('a, 'b) RBT.rbt -> ('a, 'b) RBT.rbt -> bool
  val equal_rbt :
    'a HOL.equal * 'a Orderings.linorder -> 'b HOL.equal ->
      ('a, 'b) RBT.rbt HOL.equal
end = struct

fun equal_rbta (A1_, A2_) B_ ra r =
  RBT_Impl.equal_rbt A1_ B_ (RBT.impl_of A2_ ra) (RBT.impl_of A2_ r);

fun equal_rbt (A1_, A2_) B_ = {equal = equal_rbta (A1_, A2_) B_} :
  ('a, 'b) RBT.rbt HOL.equal;

end; (*struct CollectionsV1*)

structure SC_Collections : sig
  datatype heapobj =
    Obj of
      string *
        (char, (string, Arith.nat Value.vala) Assoc_List.assoc_list) Trie.trie
    | Arr of
        Type.ty * Arith.nat *
          (string, Arith.nat Value.vala) Assoc_List.assoc_list *
          (Arith.nat, Arith.nat Value.vala) RBT.rbt
  val init_fields_array :
    (string * Type.ty) list ->
      (string, Arith.nat Value.vala) Assoc_List.assoc_list
  val init_fields :
    ((string * string) * Type.ty) list ->
      (char, (string, Arith.nat Value.vala) Assoc_List.assoc_list) Trie.trie
  val init_cells :
    Type.ty -> Arith.nat -> (Arith.nat, Arith.nat Value.vala) RBT.rbt
  val blank : 'a TypeRel.prog -> Type.htype -> heapobj
  val obj_ty : heapobj -> Type.htype
  val new_Addr : (Arith.nat, heapobj) RBT.rbt -> Arith.nat option
  val sc_allocate :
    'a TypeRel.prog ->
      (Arith.nat, heapobj) RBT.rbt ->
        Type.htype -> ((Arith.nat, heapobj) RBT.rbt * Arith.nat) Set.set
  val equal_heapobja : heapobj -> heapobj -> bool
  val equal_heapobj : heapobj HOL.equal
  val sc_typeof_addr :
    (Arith.nat, heapobj) RBT.rbt -> Arith.nat -> Type.htype option
  val sc_spurious_wakeups : bool
  val sc_heap_read_i_i_i_o :
    (Arith.nat, heapobj) RBT.rbt ->
      Arith.nat -> Heap.addr_loc -> Arith.nat Value.vala Predicate.pred
  val sc_heap_write_i_i_i_i_o :
    (Arith.nat, heapobj) RBT.rbt ->
      Arith.nat ->
        Heap.addr_loc ->
          Arith.nat Value.vala -> (Arith.nat, heapobj) RBT.rbt Predicate.pred
end = struct

datatype heapobj =
  Obj of
    string *
      (char, (string, Arith.nat Value.vala) Assoc_List.assoc_list) Trie.trie
  | Arr of
      Type.ty * Arith.nat *
        (string, Arith.nat Value.vala) Assoc_List.assoc_list *
        (Arith.nat, Arith.nat Value.vala) RBT.rbt;

fun init_fields_array x =
  (ListMapImpl.g_list_to_map_lm_basic_ops Stringa.equal_literal o
    List.map (fn (f, t) => (f, Value.default_val t)))
    x;

fun init_fields fDTs =
  List.foldr
    (fn (a, b) =>
      let
        val (f, d) = a;
      in
        (fn t => fn fields =>
          let
            val fa = String.explode f;
          in
            Trie.update Stringa.equal_char fa
              (Assoc_List.update Stringa.equal_literal d (Value.default_val t)
                (case Trie.lookup Stringa.equal_char fields fa
                  of NONE => Assoc_List.empty | SOME lm => lm))
              fields
          end)
      end
        b)
    fDTs Trie.empty;

fun init_cells t n =
  List.foldl
    (fn cells => fn i =>
      RBT.insert Arith.linorder_nat i (Value.default_val t) cells)
    (RBT.empty Arith.linorder_nat) (List.upt Arith.zero_nat n);

fun blank p (Type.Class_type c) =
  Obj (c, init_fields
            (List.map (fn (fd, (t, _)) => (fd, t)) (TypeRel.fields p c)))
  | blank p (Type.Array_type (t, n)) =
    Arr (t, n,
          init_fields_array
            (List.map
              (fn (a, b) =>
                let
                  val (f, _) = a;
                in
                  (fn (ta, _) => (f, ta))
                end
                  b)
              (TypeRel.fields p Type.object)),
          init_cells t n);

fun obj_ty (Obj (c, f)) = Type.Class_type c
  | obj_ty (Arr (t, si, f, e)) = Type.Array_type (t, si);

fun new_Addr h =
  SOME (case RBTMapImpl.g_max_rm_basic_ops Arith.linorder_nat h (fn _ => true)
         of NONE => Arith.zero_nat
         | SOME (a, _) => Arith.plus_nat a Arith.one_nat);

fun sc_allocate p h hT =
  (case new_Addr h of NONE => Set.bot_seta
    | SOME a =>
      Set.insert (RBT.insert Arith.linorder_nat a (blank p hT) h, a)
        Set.bot_seta);

fun equal_heapobja (Arr (ty, nat, assoc_list, rbt)) (Obj (literal, trie)) =
  false
  | equal_heapobja (Obj (literal, trie)) (Arr (ty, nat, assoc_list, rbt)) =
    false
  | equal_heapobja (Arr (tya, nata, assoc_lista, rbta))
    (Arr (ty, nat, assoc_list, rbt)) =
    Type.equal_tya tya ty andalso
      (Arith.equal_nata nata nat andalso
        (Assoc_List.equal_assoc_lista Stringa.equal_literal
           (Value.equal_val Arith.equal_nat) assoc_lista assoc_list andalso
          CollectionsV1.equal_rbta (Arith.equal_nat, Arith.linorder_nat)
            (Value.equal_val Arith.equal_nat) rbta rbt))
  | equal_heapobja (Obj (literala, triea)) (Obj (literal, trie)) =
    ((literala : string) = literal) andalso
      Trie.equal_trie Stringa.equal_char
        (Assoc_List.equal_assoc_list Stringa.equal_literal
          (Value.equal_val Arith.equal_nat))
        triea trie;

val equal_heapobj = {equal = equal_heapobja} : heapobj HOL.equal;

fun sc_typeof_addr h a = Option.map obj_ty (RBT.lookup Arith.linorder_nat h a);

val sc_spurious_wakeups : bool = false;

fun sc_heap_read_i_i_i_o x xa xb =
  Predicate.sup_pred
    (case xb
      of Heap.CField (d, f) =>
        (case RBT.lookup Arith.linorder_nat x xa of NONE => Predicate.bot_pred
          | SOME (Obj (_, fs)) =>
            (case Trie.lookup Stringa.equal_char fs (String.explode f)
              of NONE => Predicate.bot_pred
              | SOME fsa =>
                (case Assoc_List.lookup Stringa.equal_literal fsa d
                  of NONE => Predicate.bot_pred | SOME a => Predicate.single a))
          | SOME (Arr (_, _, _, _)) => Predicate.bot_pred)
      | Heap.ACell _ => Predicate.bot_pred)
    (Predicate.sup_pred
      (case xb of Heap.CField (_, _) => Predicate.bot_pred
        | Heap.ACell n =>
          (case RBT.lookup Arith.linorder_nat x xa of NONE => Predicate.bot_pred
            | SOME (Obj (_, _)) => Predicate.bot_pred
            | SOME (Arr (_, si, _, el)) =>
              Predicate.bind (Predicate.if_pred (Arith.less_nat n si))
                (fn () =>
                  Predicate.single
                    (Option.the (RBT.lookup Arith.linorder_nat el n)))))
      (case xb
        of Heap.CField (xba, f) =>
          (case RBT.lookup Arith.linorder_nat x xa of NONE => Predicate.bot_pred
            | SOME (Obj (_, _)) => Predicate.bot_pred
            | SOME (Arr (_, _, fa, _)) =>
              (case Assoc_List.lookup Stringa.equal_literal fa f
                of NONE => Predicate.bot_pred
                | SOME v =>
                  (if ((xba : string) = Type.object) then Predicate.single v
                    else Predicate.bot_pred)))
        | Heap.ACell _ => Predicate.bot_pred));

fun sc_heap_write_i_i_i_i_o x xa xb xc =
  Predicate.sup_pred
    (case xb
      of Heap.CField (d, f) =>
        (case RBT.lookup Arith.linorder_nat x xa of NONE => Predicate.bot_pred
          | SOME (Obj (c, fs)) =>
            Predicate.single
              (RBT.insert Arith.linorder_nat xa
                (Obj (c, Trie.update Stringa.equal_char (String.explode f)
                           (Assoc_List.update Stringa.equal_literal d xc
                             (case Trie.lookup Stringa.equal_char fs
                                     (String.explode f)
                               of NONE => Assoc_List.empty | SOME fsa => fsa))
                           fs))
                x)
          | SOME (Arr (_, _, _, _)) => Predicate.bot_pred)
      | Heap.ACell _ => Predicate.bot_pred)
    (Predicate.sup_pred
      (case xb of Heap.CField (_, _) => Predicate.bot_pred
        | Heap.ACell n =>
          (case RBT.lookup Arith.linorder_nat x xa of NONE => Predicate.bot_pred
            | SOME (Obj (_, _)) => Predicate.bot_pred
            | SOME (Arr (t, si, f, el)) =>
              Predicate.single
                (RBT.insert Arith.linorder_nat xa
                  (Arr (t, si, f, RBT.insert Arith.linorder_nat n xc el)) x)))
      (case xb
        of Heap.CField (xba, f) =>
          (case RBT.lookup Arith.linorder_nat x xa of NONE => Predicate.bot_pred
            | SOME (Obj (_, _)) => Predicate.bot_pred
            | SOME (Arr (t, si, fa, el)) =>
              (if ((xba : string) = Type.object)
                then Predicate.single
                       (RBT.insert Arith.linorder_nat xa
                         (Arr (t, si,
                                Assoc_List.update Stringa.equal_literal f xc fa,
                                el))
                         x)
                else Predicate.bot_pred))
        | Heap.ACell _ => Predicate.bot_pred));

end; (*struct SC_Collections*)

structure JVM_Execute2 : sig
  val sc_mexec :
    (Arith.nat *
      (Arith.nat *
        (Arith.nat JVMInstructions.instr list *
          (Arith.nat *
            (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list)))
      TypeRel.prog ->
      Arith.nat ->
        (Arith.nat option *
          ((Arith.nat JVMInstructions.instr list *
             (Arith.nat JVMInstructions.instr list *
               (Arith.nat *
                 (Arith.nat *
                   (string option * (Arith.nat * Arith.nat)))) list)) *
            (Arith.nat Value.vala list *
              (Arith.nat Value.vala list *
                (string * (string * Arith.nat))))) list) *
          (Arith.nat, SC_Collections.heapobj) RBT.rbt ->
          (((Arith.nat, (FWState.lock_action list)) FinFun.finfun *
             ((Arith.nat,
                (Arith.nat option *
                  ((Arith.nat JVMInstructions.instr list *
                     (Arith.nat JVMInstructions.instr list *
                       (Arith.nat *
                         (Arith.nat *
                           (string option * (Arith.nat * Arith.nat)))) list)) *
                    (Arith.nat Value.vala list *
                      (Arith.nat Value.vala list *
                        (string * (string * Arith.nat))))) list),
                (Arith.nat, SC_Collections.heapobj) RBT.rbt)
                FWState.new_thread_action list *
               (Arith.nat FWState.conditional_action list *
                 ((Arith.nat, Arith.nat) FWState.wait_set_action list *
                   (Arith.nat FWState.interrupt_action list *
                     (Arith.nat, Arith.nat)
                       Observable_Events.obs_event list))))) *
            ((Arith.nat option *
               ((Arith.nat JVMInstructions.instr list *
                  (Arith.nat JVMInstructions.instr list *
                    (Arith.nat *
                      (Arith.nat *
                        (string option * (Arith.nat * Arith.nat)))) list)) *
                 (Arith.nat Value.vala list *
                   (Arith.nat Value.vala list *
                     (string * (string * Arith.nat))))) list) *
              (Arith.nat, SC_Collections.heapobj) RBT.rbt))
            Predicate.pred
  val jVM_final :
    'a option *
      (('a JVMInstructions.instr list *
         ('a JVMInstructions.instr list *
           (Arith.nat *
             (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list)) *
        ('a Value.vala list *
          ('a Value.vala list * (string * (string * Arith.nat))))) list ->
      bool
  val sc_rr_JVM_start_state :
    Arith.nat ->
      'a TypeRel.prog -> (Arith.nat list * Arith.nat list) * Arith.nat
  val exec_JVM_rr :
    Arith.nat ->
      (Arith.nat *
        (Arith.nat *
          (Arith.nat JVMInstructions.instr list *
            (Arith.nat *
              (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list)))
        TypeRel.prog ->
        string ->
          string ->
            Arith.nat Value.vala list ->
              ((Arith.nat *
                 (Arith.nat, Arith.nat) Observable_Events.obs_event list),
                ((Arith.nat, ((Arith.nat * Arith.nat) option)) FinFun.finfun *
                  (((Arith.nat,
                      ((Arith.nat option *
                         ((Arith.nat JVMInstructions.instr list *
                            (Arith.nat JVMInstructions.instr list *
                              (Arith.nat *
                                (Arith.nat *
                                  (string option *
                                    (Arith.nat * Arith.nat)))) list)) *
                           (Arith.nat Value.vala list *
                             (Arith.nat Value.vala list *
                               (string * (string * Arith.nat))))) list) *
                        (Arith.nat, Arith.nat) FinFun.finfun))
                      RBT.rbt *
                     (Arith.nat, SC_Collections.heapobj) RBT.rbt) *
                    ((Arith.nat, Arith.nat FWState.wait_set_status) RBT.rbt *
                      (Arith.nat, unit) RBT.rbt))))
                TLList.tllist
  val sc_rnd_JVM_start_state :
    Arith.natural * Arith.natural -> Arith.natural * Arith.natural
  val exec_JVM_rnd :
    Arith.natural * Arith.natural ->
      (Arith.nat *
        (Arith.nat *
          (Arith.nat JVMInstructions.instr list *
            (Arith.nat *
              (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list)))
        TypeRel.prog ->
        string ->
          string ->
            Arith.nat Value.vala list ->
              ((Arith.nat *
                 (Arith.nat, Arith.nat) Observable_Events.obs_event list),
                ((Arith.nat, ((Arith.nat * Arith.nat) option)) FinFun.finfun *
                  (((Arith.nat,
                      ((Arith.nat option *
                         ((Arith.nat JVMInstructions.instr list *
                            (Arith.nat JVMInstructions.instr list *
                              (Arith.nat *
                                (Arith.nat *
                                  (string option *
                                    (Arith.nat * Arith.nat)))) list)) *
                           (Arith.nat Value.vala list *
                             (Arith.nat Value.vala list *
                               (string * (string * Arith.nat))))) list) *
                        (Arith.nat, Arith.nat) FinFun.finfun))
                      RBT.rbt *
                     (Arith.nat, SC_Collections.heapobj) RBT.rbt) *
                    ((Arith.nat, Arith.nat FWState.wait_set_status) RBT.rbt *
                      (Arith.nat, unit) RBT.rbt))))
                TLList.tllist
end = struct

fun sc_mexec p t ((xcp, frs), h) =
  Predicate.bind
    (JVMExec_Execute2.exec_1 (Set.card_UNIV_nat, Arith.equal_nat, MM.addr_nat)
      Arith.equal_nat
      (CollectionsV1.equal_rbt (Arith.equal_nat, Arith.linorder_nat)
        SC_Collections.equal_heapobj)
      (fn x => x) (fn x => x) SC_Collections.sc_spurious_wakeups
      (RBT.empty Arith.linorder_nat) (SC_Collections.sc_allocate p)
      SC_Collections.sc_typeof_addr
      (fn ha => fn ad => fn al =>
        Predicate.set_of_pred (SC_Collections.sc_heap_read_i_i_i_o ha ad al))
      (fn ha => fn ad => fn al => fn v =>
        Predicate.set_of_pred
          (SC_Collections.sc_heap_write_i_i_i_i_o ha ad al v))
      p t (xcp, (h, frs)))
    (fn (ta, (xcpa, (ha, frsa))) => Predicate.single (ta, ((xcpa, frsa), ha)));

fun jVM_final (xcp, frs) = List.null frs;

fun sc_rr_JVM_start_state n0 p =
  Round_Robin.round_robin_start Fifo.fifo_empty Fifo.fifo_appendr n0
    (StartConfig.start_tid (Arith.equal_nat, MM.addr_nat)
      (CollectionsV1.equal_rbt (Arith.equal_nat, Arith.linorder_nat)
        SC_Collections.equal_heapobj)
      (fn x => x) (RBT.empty Arith.linorder_nat)
      (SC_Collections.sc_allocate p));

fun exec_JVM_rr n0 p c m vs =
  Scheduler.exec Arith.equal_nat Arith.equal_nat Arith.equal_nat
    (Observable_Events.convert_RA (Arith.equal_nat, MM.addr_nat))
    (Round_Robin.round_robin Arith.equal_nat
      (Set.card_UNIV_nat, Arith.equal_nat) Arith.equal_nat jVM_final
      (sc_mexec p) (Observable_Events.convert_RA (Arith.equal_nat, MM.addr_nat))
      (fn k => fn ma => RBT.lookup Arith.linorder_nat ma k)
      (RBT.insert Arith.linorder_nat)
      (fn k => fn ma => RBT.lookup Arith.linorder_nat ma k)
      (RBTSetImpl.memb_rm_basic_ops Arith.linorder_nat)
      (RBTSetImpl.ins_rm_basic_ops Arith.linorder_nat)
      (RBTSetImpl.delete_rm_basic_ops Arith.linorder_nat)
      (Fifo.fifo_isEmpty Arith.equal_nat) Fifo.fifo_appendr Fifo.fifo_removel
      Fifo.fifo_appendl n0)
    State_Refinement.jinja_output
    (Scheduler.pick_wakeup_via_sel Arith.equal_nat
      (fn s => fn pa =>
        RBTMapImpl.g_sel_rm_basic_ops Arith.linorder_nat s
          (fn (a, b) => pa a b)))
    (fn k => fn ma => RBT.lookup Arith.linorder_nat ma k)
    (RBT.insert Arith.linorder_nat)
    (fn k => fn ma => RBT.lookup Arith.linorder_nat ma k)
    (RBT.insert Arith.linorder_nat) (RBT.delete Arith.linorder_nat)
    (RBTMapImpl.iteratei_map_op_list_it_rm_ops Arith.linorder_nat)
    (RBTSetImpl.ins_rm_basic_ops Arith.linorder_nat)
    (RBTSetImpl.delete_rm_basic_ops Arith.linorder_nat)
    (sc_rr_JVM_start_state n0 p)
    (State_Refinement.start_state_refine (Arith.equal_nat, MM.addr_nat)
      (CollectionsV1.equal_rbt (Arith.equal_nat, Arith.linorder_nat)
        SC_Collections.equal_heapobj)
      (Product_Type.equal_prod Arith.equal_nat
        (Product_Type.equal_prod Arith.equal_nat
          (Product_Type.equal_prod
            (List.equal_list (JVMInstructions.equal_instr Arith.equal_nat))
            (List.equal_list
              (Product_Type.equal_prod Arith.equal_nat
                (Product_Type.equal_prod Arith.equal_nat
                  (Product_Type.equal_prod
                    (Option.equal_option Stringa.equal_literal)
                    (Product_Type.equal_prod Arith.equal_nat
                      Arith.equal_nat))))))))
      (fn x => x) (RBT.empty Arith.linorder_nat) (SC_Collections.sc_allocate p)
      (RBT.empty Arith.linorder_nat) (RBT.insert Arith.linorder_nat)
      (RBT.empty Arith.linorder_nat)
      (RBTSetImpl.empty_rm_basic_ops Arith.linorder_nat ())
      (fn ca => fn ma => fn _ => fn _ => fn (_, (mxl0, (ins, xt))) => fn vsa =>
        (NONE,
          [((ins, (ins, xt)),
             ([], (Value.Null ::
                     vsa @ List.replicate mxl0 Value.undefined_value,
                    (ca, (ma, Arith.zero_nat)))))]))
      p c m vs);

fun sc_rnd_JVM_start_state seed = seed;

fun exec_JVM_rnd seed p c m vs =
  Scheduler.exec Arith.equal_nat Arith.equal_nat Arith.equal_nat
    (Observable_Events.convert_RA (Arith.equal_nat, MM.addr_nat))
    (Random_Scheduler.random_scheduler Arith.equal_nat
      (Set.card_UNIV_nat, Arith.equal_nat) Arith.equal_nat jVM_final
      (sc_mexec p) (Observable_Events.convert_RA (Arith.equal_nat, MM.addr_nat))
      (fn k => fn ma => RBT.lookup Arith.linorder_nat ma k)
      (RBT.insert Arith.linorder_nat)
      (RBTMapImpl.iteratei_map_op_list_it_rm_ops Arith.linorder_nat)
      (fn k => fn ma => RBT.lookup Arith.linorder_nat ma k)
      (RBTSetImpl.memb_rm_basic_ops Arith.linorder_nat)
      (RBTSetImpl.ins_rm_basic_ops Arith.linorder_nat)
      (RBTSetImpl.delete_rm_basic_ops Arith.linorder_nat) (fn _ => [])
      (fn a => fn b => a :: b) ListSetImpl_Invar.g_to_list_lsi_basic_ops)
    State_Refinement.jinja_output
    (Scheduler.pick_wakeup_via_sel Arith.equal_nat
      (fn s => fn pa =>
        RBTMapImpl.g_sel_rm_basic_ops Arith.linorder_nat s
          (fn (a, b) => pa a b)))
    (fn k => fn ma => RBT.lookup Arith.linorder_nat ma k)
    (RBT.insert Arith.linorder_nat)
    (fn k => fn ma => RBT.lookup Arith.linorder_nat ma k)
    (RBT.insert Arith.linorder_nat) (RBT.delete Arith.linorder_nat)
    (RBTMapImpl.iteratei_map_op_list_it_rm_ops Arith.linorder_nat)
    (RBTSetImpl.ins_rm_basic_ops Arith.linorder_nat)
    (RBTSetImpl.delete_rm_basic_ops Arith.linorder_nat)
    (sc_rnd_JVM_start_state seed)
    (State_Refinement.start_state_refine (Arith.equal_nat, MM.addr_nat)
      (CollectionsV1.equal_rbt (Arith.equal_nat, Arith.linorder_nat)
        SC_Collections.equal_heapobj)
      (Product_Type.equal_prod Arith.equal_nat
        (Product_Type.equal_prod Arith.equal_nat
          (Product_Type.equal_prod
            (List.equal_list (JVMInstructions.equal_instr Arith.equal_nat))
            (List.equal_list
              (Product_Type.equal_prod Arith.equal_nat
                (Product_Type.equal_prod Arith.equal_nat
                  (Product_Type.equal_prod
                    (Option.equal_option Stringa.equal_literal)
                    (Product_Type.equal_prod Arith.equal_nat
                      Arith.equal_nat))))))))
      (fn x => x) (RBT.empty Arith.linorder_nat) (SC_Collections.sc_allocate p)
      (RBT.empty Arith.linorder_nat) (RBT.insert Arith.linorder_nat)
      (RBT.empty Arith.linorder_nat)
      (RBTSetImpl.empty_rm_basic_ops Arith.linorder_nat ())
      (fn ca => fn ma => fn _ => fn _ => fn (_, (mxl0, (ins, xt))) => fn vsa =>
        (NONE,
          [((ins, (ins, xt)),
             ([], (Value.Null ::
                     vsa @ List.replicate mxl0 Value.undefined_value,
                    (ca, (ma, Arith.zero_nat)))))]))
      p c m vs);

end; (*struct JVM_Execute2*)

structure Code_Generation : sig
  val shr : 'a * (('b * 'c) * ('d * 'e)) -> 'c
  val thr : 'a * (('b * 'c) * ('d * 'e)) -> 'b
  val j2jvm :
    (string list * (string, unit, Arith.nat) Expr.exp) TypeRel.prog ->
      (Arith.nat *
        (Arith.nat *
          (Arith.nat JVMInstructions.instr list *
            (Arith.nat *
              (Arith.nat * (string option * (Arith.nat * Arith.nat)))) list)))
        TypeRel.prog
  val purge :
    (('a * ('b, 'c) Observable_Events.obs_event list), 'd) TLList.tllist ->
      Numeral_Type.num1 Numeral_Type.bit0 Numeral_Type.bit0 Numeral_Type.bit0
        Numeral_Type.bit0
        Numeral_Type.bit0
        Word.word
        Coinductive_List.llist
  val terminal : ('a, 'b) TLList.tllist -> 'b
  val j_Program :
    (string *
      (string *
        ((string * (Type.ty * unit TypeRel.fmod_ext)) list *
          (string *
            (Type.ty list *
              (Type.ty *
                (string list *
                  (string, unit, Arith.nat) Expr.exp) option))) list))) list ->
      (string list * (string, unit, Arith.nat) Expr.exp) TypeRel.prog
  val llist_case :
    'a -> ('b -> 'b Coinductive_List.llist -> 'a) ->
            'b Coinductive_List.llist -> 'a
  val tllist_case :
    ('a -> 'b) ->
      ('c -> ('c, 'a) TLList.tllist -> 'b) -> ('c, 'a) TLList.tllist -> 'b
  val toString_heapobja : SC_Collections.heapobj -> string
  val toString_heapobj : SC_Collections.heapobj ToString.toString
  val heap_toString : (Arith.nat, SC_Collections.heapobj) RBT.rbt -> string
  val trace_toString :
    Arith.nat * (Arith.nat, Arith.nat) Observable_Events.obs_event list ->
      string
  val llist_of_tllist : ('a, 'b) TLList.tllist -> 'a Coinductive_List.llist
  val thread_toString :
    (Arith.nat,
      ((Arith.nat option *
         ((Arith.nat JVMInstructions.instr list *
            (Arith.nat JVMInstructions.instr list *
              (Arith.nat *
                (Arith.nat *
                  (string option * (Arith.nat * Arith.nat)))) list)) *
           (Arith.nat Value.vala list *
             (Arith.nat Value.vala list *
               (string * (string * Arith.nat))))) list) *
        (Arith.nat, Arith.nat) FinFun.finfun))
      RBT.rbt ->
      string
end = struct

fun shr x = FWState.shr x;

fun thr x = FWState.thr x;

fun j2jvm x = Compiler.j2JVM x;

fun purge run =
  Coinductive_List.lmap
    (fn Observable_Events.ExternalCall (_, _, Value.Intg i :: _, _) => i)
    (Coinductive_List.lfilter
      (fn a =>
        (case a of Observable_Events.ExternalCall (_, m, [], _) => false
          | Observable_Events.ExternalCall (_, m, Value.Unit :: lista, _) =>
            false
          | Observable_Events.ExternalCall (_, m, Value.Null :: lista, _) =>
            false
          | Observable_Events.ExternalCall (_, m, Value.Bool _ :: lista, _) =>
            false
          | Observable_Events.ExternalCall (_, m, [Value.Intg _], _) =>
            ((m : string) = "~print(I)V")
          | Observable_Events.ExternalCall (_, m, Value.Intg _ :: _ :: _, _) =>
            false
          | Observable_Events.ExternalCall (_, m, Value.Addr _ :: lista, _) =>
            false
          | Observable_Events.ReadMem (_, _, _) => false
          | Observable_Events.WriteMem (_, _, _) => false
          | Observable_Events.NewHeapElem (_, _) => false
          | Observable_Events.ThreadStart _ => false
          | Observable_Events.ThreadJoin _ => false
          | Observable_Events.SyncLock _ => false
          | Observable_Events.SyncUnlock _ => false
          | Observable_Events.ObsInterrupt _ => false
          | Observable_Events.ObsInterrupted _ => false))
      (Coinductive_List.lconcat
        (Coinductive_List.lmap (Coinductive_List.llist_of o Product_Type.snd)
          (TLList.llist_of_tllist run))));

fun terminal x = TLList.terminal x;

fun j_Program x = TypeRel.Program x;

fun llist_case x = Coinductive_List.llist_case x;

fun tllist_case x = TLList.tllist_case x;

fun toString_heapobja (SC_Collections.Obj (c, fs)) =
  String.concat
    ["(Obj ", ToString.toString_literala c, ", ",
      ToString.toString_trie ToString.toString_char
        (ToString.toString_assoc_list ToString.toString_literal
          (ToString.toString_val ToString.toString_nat))
        fs,
      ")"]
  | toString_heapobja (SC_Collections.Arr (t, si, fs, el)) =
    String.concat
      ["([", ToString.toString_nata si, "]", ToString.toString_ty t, ", ",
        ToString.toString_assoc_lista ToString.toString_literal
          (ToString.toString_val ToString.toString_nat) fs,
        ", ",
        ToString.toString_lista (ToString.toString_val ToString.toString_nat)
          (List.map Product_Type.snd
            (RBTMapImpl.g_to_list_rm_basic_ops Arith.linorder_nat el)),
        ")"];

val toString_heapobj = {toString = toString_heapobja} :
  SC_Collections.heapobj ToString.toString;

fun heap_toString x =
  ToString.toString_rbt (Arith.linorder_nat, ToString.toString_nat)
    toString_heapobj x;

fun trace_toString x =
  ToString.toString_proda ToString.toString_nat
    (ToString.toString_list
      (ToString.toString_obs_event ToString.toString_nat ToString.toString_nat))
    x;

fun llist_of_tllist x = TLList.llist_of_tllist x;

fun thread_toString x =
  ToString.toString_rbt (Arith.linorder_nat, ToString.toString_nat)
    (ToString.toString_prod
      (ToString.toString_prod (ToString.toString_option ToString.toString_nat)
        (ToString.toString_list
          (ToString.toString_prod
            (ToString.toString_prod
              (ToString.toString_list
                (ToString.toString_instr ToString.toString_nat))
              (ToString.toString_prod
                (ToString.toString_list
                  (ToString.toString_instr ToString.toString_nat))
                (ToString.toString_list
                  (ToString.toString_prod ToString.toString_nat
                    (ToString.toString_prod ToString.toString_nat
                      (ToString.toString_prod
                        (ToString.toString_option ToString.toString_literal)
                        (ToString.toString_prod ToString.toString_nat
                          ToString.toString_nat)))))))
            (ToString.toString_prod
              (ToString.toString_list
                (ToString.toString_val ToString.toString_nat))
              (ToString.toString_prod
                (ToString.toString_list
                  (ToString.toString_val ToString.toString_nat))
                (ToString.toString_prod ToString.toString_literal
                  (ToString.toString_prod ToString.toString_literal
                    ToString.toString_nat)))))))
      (ToString.toString_finfun
        (Set.card_UNIV_nat, Arith.equal_nat, Arith.linorder_nat,
          ToString.toString_nat)
        (Arith.equal_nat, ToString.toString_nat)))
    x;

end; (*struct Code_Generation*)
