
structure STArray = struct

datatype 'a Cell = Invalid | Value of 'a array;

exception AccessedOldVersion;

type 'a array = 'a Cell Unsynchronized.ref;

fun fromList l = Unsynchronized.ref (Value (Array.fromList l));
fun array (size, v) = Unsynchronized.ref (Value (Array.array (size,v)));
fun tabulate (size, f) = Unsynchronized.ref (Value (Array.tabulate(size, f)));
fun sub (Unsynchronized.ref Invalid, idx) = raise AccessedOldVersion |
    sub (Unsynchronized.ref (Value a), idx) = Array.sub (a,idx);
fun update (aref,idx,v) =
  case aref of
    (Unsynchronized.ref Invalid) => raise AccessedOldVersion |
    (Unsynchronized.ref (Value a)) => (
      aref := Invalid;
      Array.update (a,idx,v);
      Unsynchronized.ref (Value a)
    );

fun length (Unsynchronized.ref Invalid) = raise AccessedOldVersion |
    length (Unsynchronized.ref (Value a)) = Array.length a

fun grow (aref, i, x) = case aref of 
  (Unsynchronized.ref Invalid) => raise AccessedOldVersion |
  (Unsynchronized.ref (Value a)) => (
    let val len=Array.length a;
        val na = Array.array (len+i,x) 
    in
      aref := Invalid;
      Array.copy {src=a, dst=na, di=0};
      Unsynchronized.ref (Value na)
    end
    );

fun shrink (aref, sz) = case aref of
  (Unsynchronized.ref Invalid) => raise AccessedOldVersion |
  (Unsynchronized.ref (Value a)) => (
    if sz > Array.length a then 
      raise Size
    else (
      aref:=Invalid;
      Unsynchronized.ref (Value (Array.tabulate (sz,fn i => Array.sub (a,i))))
    )
  );

structure IsabelleMapping = struct
type 'a ArrayType = 'a array;

fun new_array (a:'a) (n:int) = array (n, a);

fun array_length (a:'a ArrayType) = length a;

fun array_get (a:'a ArrayType) (i:int) = sub (a, i);

fun array_set (a:'a ArrayType) (i:int) (e:'a) = update (a, i, e);

fun array_of_list (xs:'a list) = fromList xs;

fun array_grow (a:'a ArrayType) (i:int) (x:'a) = grow (a, i, x);

fun array_shrink (a:'a ArrayType) (sz:int) = shrink (a,sz);

end;

end;

structure FArray = struct
  datatype 'a Cell = Value of 'a Array.array | Upd of (int*'a*'a Cell Unsynchronized.ref);

  type 'a array = 'a Cell Unsynchronized.ref;

  fun array (size,v) = Unsynchronized.ref (Value (Array.array (size,v)));
  fun tabulate (size, f) = Unsynchronized.ref (Value (Array.tabulate(size, f)));
  fun fromList l = Unsynchronized.ref (Value (Array.fromList l));

  fun sub (Unsynchronized.ref (Value a), idx) = Array.sub (a,idx) |
      sub (Unsynchronized.ref (Upd (i,v,cr)),idx) =
        if i=idx then v
        else sub (cr,idx);

  fun length (Unsynchronized.ref (Value a)) = Array.length a |
      length (Unsynchronized.ref (Upd (i,v,cr))) = length cr;

  fun realize_aux (aref, v) =
    case aref of
      (Unsynchronized.ref (Value a)) => (
        let
          val len = Array.length a;
          val a' = Array.array (len,v);
        in
          Array.copy {src=a, dst=a', di=0};
          Unsynchronized.ref (Value a')
        end
      ) |
      (Unsynchronized.ref (Upd (i,v,cr))) => (
        let val res=realize_aux (cr,v) in
          case res of
            (Unsynchronized.ref (Value a)) => (Array.update (a,i,v); res)
        end
      );

  fun realize aref =
    case aref of
      (Unsynchronized.ref (Value _)) => aref |
      (Unsynchronized.ref (Upd (i,v,cr))) => realize_aux(aref,v);

  fun update (aref,idx,v) =
    case aref of
      (Unsynchronized.ref (Value a)) => (
        let val nref=Unsynchronized.ref (Value a) in
          aref := Upd (idx,Array.sub(a,idx),nref);
          Array.update (a,idx,v);
          nref
        end
      ) |
      (Unsynchronized.ref (Upd _)) =>
        let val ra = realize_aux(aref,v) in
          case ra of
            (Unsynchronized.ref (Value a)) => Array.update (a,idx,v);
          ra
        end
      ;

  fun grow (aref, inc, x) = case aref of 
    (Unsynchronized.ref (Value a)) => (
      let val len=Array.length a;
          val na = Array.array (len+inc,x) 
      in
        Array.copy {src=a, dst=na, di=0};
        Unsynchronized.ref (Value na)
      end
      )
  | (Unsynchronized.ref (Upd _)) => (
    grow (realize aref, inc, x)
  );

  fun shrink (aref, sz) = case aref of
    (Unsynchronized.ref (Value a)) => (
      if sz > Array.length a then 
        raise Size
      else (
        Unsynchronized.ref (Value (Array.tabulate (sz,fn i => Array.sub (a,i))))
      )
    ) |
    (Unsynchronized.ref (Upd _)) => (
      shrink (realize aref,sz)
    );

structure IsabelleMapping = struct
type 'a ArrayType = 'a array;

fun new_array (a:'a) (n:int) = array (n, a);

fun array_length (a:'a ArrayType) = length a;

fun array_get (a:'a ArrayType) (i:int) = sub (a, i);

fun array_set (a:'a ArrayType) (i:int) (e:'a) = update (a, i, e);

fun array_of_list (xs:'a list) = fromList xs;

fun array_grow (a:'a ArrayType) (i:int) (x:'a) = grow (a, i, x);

fun array_shrink (a:'a ArrayType) (sz:int) = shrink (a,sz);

fun array_get_oo (d:'a) (a:'a ArrayType) (i:int) = 
  sub (a,i) handle Subscript => d

fun array_set_oo (d:(unit->'a ArrayType)) (a:'a ArrayType) (i:int) (e:'a) = 
  update (a, i, e) handle Subscript => d ()

end;
end;




structure HOL : sig
  type 'a equal
  val equal : 'a equal -> 'a -> 'a -> bool
  val eq : 'a equal -> 'a -> 'a -> bool
  datatype 'a itself = Type
end = struct

type 'a equal = {equal : 'a -> 'a -> bool};
val equal = #equal : 'a equal -> 'a -> 'a -> bool;

fun eq A_ a b = equal A_ a b;

datatype 'a itself = Type;

end; (*struct HOL*)

structure Product_Type : sig
  val fst : 'a * 'b -> 'a
  val snd : 'a * 'b -> 'b
  val apsnd : ('a -> 'b) -> 'c * 'a -> 'c * 'b
  val scomp : ('a -> 'b * 'c) -> ('b -> 'c -> 'd) -> 'a -> 'd
  val map_pair : ('a -> 'b) -> ('c -> 'd) -> 'a * 'c -> 'b * 'd
  val equal_boola : bool -> bool -> bool
  val equal_bool : bool HOL.equal
  val equal_proda : 'a HOL.equal -> 'b HOL.equal -> 'a * 'b -> 'a * 'b -> bool
  val equal_prod : 'a HOL.equal -> 'b HOL.equal -> ('a * 'b) HOL.equal
  val equal_unita : unit -> unit -> bool
  val equal_unit : unit HOL.equal
end = struct

fun fst (a, b) = a;

fun snd (a, b) = b;

fun apsnd f (x, y) = (x, f y);

fun scomp f g = (fn x => let
                           val (a, b) = f x;
                         in
                           g a b
                         end);

fun map_pair f g (a, b) = (f a, g b);

fun equal_boola p true = p
  | equal_boola p false = not p
  | equal_boola true p = p
  | equal_boola false p = not p;

val equal_bool = {equal = equal_boola} : bool HOL.equal;

fun equal_proda A_ B_ (aa, ba) (a, b) = HOL.eq A_ aa a andalso HOL.eq B_ ba b;

fun equal_prod A_ B_ = {equal = equal_proda A_ B_} : ('a * 'b) HOL.equal;

fun equal_unita u v = true;

val equal_unit = {equal = equal_unita} : unit HOL.equal;

end; (*struct Product_Type*)

structure Orderings : sig
  type 'a ord
  val less_eq : 'a ord -> 'a -> 'a -> bool
  val less : 'a ord -> 'a -> 'a -> bool
  val max : 'a ord -> 'a -> 'a -> 'a
  type 'a preorder
  val ord_preorder : 'a preorder -> 'a ord
  type 'a order
  val preorder_order : 'a order -> 'a preorder
  type 'a linorder
  val order_linorder : 'a linorder -> 'a order
end = struct

type 'a ord = {less_eq : 'a -> 'a -> bool, less : 'a -> 'a -> bool};
val less_eq = #less_eq : 'a ord -> 'a -> 'a -> bool;
val less = #less : 'a ord -> 'a -> 'a -> bool;

fun max A_ a b = (if less_eq A_ a b then b else a);

type 'a preorder = {ord_preorder : 'a ord};
val ord_preorder = #ord_preorder : 'a preorder -> 'a ord;

type 'a order = {preorder_order : 'a preorder};
val preorder_order = #preorder_order : 'a order -> 'a preorder;

type 'a linorder = {order_linorder : 'a order};
val order_linorder = #order_linorder : 'a linorder -> 'a order;

end; (*struct Orderings*)

structure Stringa : sig
  datatype nibble = Nibble0 | Nibble1 | Nibble2 | Nibble3 | Nibble4 | Nibble5 |
    Nibble6 | Nibble7 | Nibble8 | Nibble9 | NibbleA | NibbleB | NibbleC |
    NibbleD | NibbleE | NibbleF
  val equal_char : char HOL.equal
  val equal_literal : string HOL.equal
end = struct

datatype nibble = Nibble0 | Nibble1 | Nibble2 | Nibble3 | Nibble4 | Nibble5 |
  Nibble6 | Nibble7 | Nibble8 | Nibble9 | NibbleA | NibbleB | NibbleC | NibbleD
  | NibbleE | NibbleF;

val equal_char = {equal = (fn a => fn b => ((a : char) = b))} : char HOL.equal;

val equal_literal = {equal = (fn a => fn b => ((a : string) = b))} :
  string HOL.equal;

end; (*struct Stringa*)

structure Fun : sig
  val id : 'a -> 'a
  val fun_upd : 'a HOL.equal -> ('a -> 'b) -> 'a -> 'b -> 'a -> 'b
end = struct

fun id x = (fn xa => xa) x;

fun fun_upd A_ f a b = (fn x => (if HOL.eq A_ x a then b else f x));

end; (*struct Fun*)

structure Arith : sig
  datatype int = Int_of_integer of IntInf.int
  datatype nat = Nat of IntInf.int
  datatype num = One | Bit0 of num | Bit1 of num
  val integer_of_int : int -> IntInf.int
  val ord_integer : IntInf.int Orderings.ord
  val nat : int -> nat
  val integer_of_nat : nat -> IntInf.int
  val plus_nat : nat -> nat -> nat
  val one_nat : nat
  val suc : nat -> nat
  type 'a one
  val one : 'a one -> 'a
  val one_inta : int
  val one_int : int one
  val minus_nat : nat -> nat -> nat
  val equal_nata : nat -> nat -> bool
  val zero_nat : nat
  val funpow : nat -> ('a -> 'a) -> 'a -> 'a
  val less_eq_nat : nat -> nat -> bool
  val less_nat : nat -> nat -> bool
  val ord_nat : nat Orderings.ord
  type 'a zero
  val zero : 'a zero -> 'a
  val uminus_int : int -> int
  val zero_inta : int
  val less_int : int -> int -> bool
  val abs_int : int -> int
  val zero_int : int zero
  type 'a times
  val times : 'a times -> 'a -> 'a -> 'a
  type 'a power
  val one_power : 'a power -> 'a one
  val times_power : 'a power -> 'a times
  val powera : 'a -> ('a -> 'a -> 'a) -> 'a -> nat -> 'a
  val power : 'a power -> 'a -> nat -> 'a
  val plus_int : int -> int -> int
  val times_inta : int -> int -> int
  val times_int : int times
  val power_int : int power
  val equal_nat : nat HOL.equal
  val preorder_nat : nat Orderings.preorder
  val order_nat : nat Orderings.order
  val equal_int : int -> int -> bool
  val minus_int : int -> int -> int
  val times_nat : nat -> nat -> nat
  val sgn_integer : IntInf.int -> IntInf.int
  val abs_integer : IntInf.int -> IntInf.int
  val divmod_integer : IntInf.int -> IntInf.int -> IntInf.int * IntInf.int
  val div_integer : IntInf.int -> IntInf.int -> IntInf.int
  val div_int : int -> int -> int
  val mod_integer : IntInf.int -> IntInf.int -> IntInf.int
  val mod_int : int -> int -> int
  val less_eq_int : int -> int -> bool
  val linorder_nat : nat Orderings.linorder
  datatype natural = Nata of IntInf.int
  type 'a zero_neq_one
  val one_zero_neq_one : 'a zero_neq_one -> 'a one
  val zero_zero_neq_one : 'a zero_neq_one -> 'a zero
  val zero_neq_one_int : int zero_neq_one
  val integer_of_natural : natural -> IntInf.int
  val div_natural : natural -> natural -> natural
  val mod_natural : natural -> natural -> natural
  val one_natural : natural
  val less_natural : natural -> natural -> bool
  val plus_natural : natural -> natural -> natural
  val zero_natural : natural
  val equal_natural : natural -> natural -> bool
  val minus_natural : natural -> natural -> natural
  val times_natural : natural -> natural -> natural
  val int_of_nat : nat -> int
  val nat_of_integer : IntInf.int -> nat
  val nat_of_natural : natural -> nat
  val natural_of_nat : nat -> natural
  val less_eq_natural : natural -> natural -> bool
  val natural_of_integer : IntInf.int -> natural
end = struct

datatype int = Int_of_integer of IntInf.int;

datatype nat = Nat of IntInf.int;

datatype num = One | Bit0 of num | Bit1 of num;

fun integer_of_int (Int_of_integer k) = k;

val ord_integer =
  {less_eq = (fn a => fn b => IntInf.<= (a, b)),
    less = (fn a => fn b => IntInf.< (a, b))}
  : IntInf.int Orderings.ord;

fun nat k = Nat (Orderings.max ord_integer 0 (integer_of_int k));

fun integer_of_nat (Nat x) = x;

fun plus_nat m n = Nat (IntInf.+ (integer_of_nat m, integer_of_nat n));

val one_nat : nat = Nat (1 : IntInf.int);

fun suc n = plus_nat n one_nat;

type 'a one = {one : 'a};
val one = #one : 'a one -> 'a;

val one_inta : int = Int_of_integer (1 : IntInf.int);

val one_int = {one = one_inta} : int one;

fun minus_nat m n =
  Nat (Orderings.max ord_integer 0
        (IntInf.- (integer_of_nat m, integer_of_nat n)));

fun equal_nata m n = (((integer_of_nat m) : IntInf.int) = (integer_of_nat n));

val zero_nat : nat = Nat 0;

fun funpow n f =
  (if equal_nata n zero_nat then Fun.id
    else f o funpow (minus_nat n one_nat) f);

fun less_eq_nat m n = IntInf.<= (integer_of_nat m, integer_of_nat n);

fun less_nat m n = IntInf.< (integer_of_nat m, integer_of_nat n);

val ord_nat = {less_eq = less_eq_nat, less = less_nat} : nat Orderings.ord;

type 'a zero = {zero : 'a};
val zero = #zero : 'a zero -> 'a;

fun uminus_int k = Int_of_integer (IntInf.~ (integer_of_int k));

val zero_inta : int = Int_of_integer 0;

fun less_int k l = IntInf.< (integer_of_int k, integer_of_int l);

fun abs_int i = (if less_int i zero_inta then uminus_int i else i);

val zero_int = {zero = zero_inta} : int zero;

type 'a times = {times : 'a -> 'a -> 'a};
val times = #times : 'a times -> 'a -> 'a -> 'a;

type 'a power = {one_power : 'a one, times_power : 'a times};
val one_power = #one_power : 'a power -> 'a one;
val times_power = #times_power : 'a power -> 'a times;

fun powera one times a n =
  (if equal_nata n zero_nat then one
    else times a (powera one times a (minus_nat n one_nat)));

fun power A_ = powera (one (one_power A_)) (times (times_power A_));

fun plus_int k l =
  Int_of_integer (IntInf.+ (integer_of_int k, integer_of_int l));

fun times_inta k l =
  Int_of_integer (IntInf.* (integer_of_int k, integer_of_int l));

val times_int = {times = times_inta} : int times;

val power_int = {one_power = one_int, times_power = times_int} : int power;

val equal_nat = {equal = equal_nata} : nat HOL.equal;

val preorder_nat = {ord_preorder = ord_nat} : nat Orderings.preorder;

val order_nat = {preorder_order = preorder_nat} : nat Orderings.order;

fun equal_int k l = (((integer_of_int k) : IntInf.int) = (integer_of_int l));

fun minus_int k l =
  Int_of_integer (IntInf.- (integer_of_int k, integer_of_int l));

fun times_nat m n = Nat (IntInf.* (integer_of_nat m, integer_of_nat n));

fun sgn_integer k =
  (if ((k : IntInf.int) = 0) then 0
    else (if IntInf.< (k, 0) then IntInf.~ (1 : IntInf.int)
           else (1 : IntInf.int)));

fun abs_integer k = (if IntInf.< (k, 0) then IntInf.~ k else k);

fun divmod_integer k l =
  (if ((k : IntInf.int) = 0) then (0, 0)
    else (if ((l : IntInf.int) = 0) then (0, k)
           else (Product_Type.apsnd o (fn a => fn b => IntInf.* (a, b)) o
                  sgn_integer)
                  l (if (((sgn_integer k) : IntInf.int) = (sgn_integer l))
                      then IntInf.divMod (IntInf.abs k, IntInf.abs l)
                      else let
                             val (r, s) =
                               IntInf.divMod (IntInf.abs k, IntInf.abs l);
                           in
                             (if ((s : IntInf.int) = 0) then (IntInf.~ r, 0)
                               else (IntInf.- (IntInf.~ r, (1 : IntInf.int)),
                                      IntInf.- (abs_integer l, s)))
                           end)));

fun div_integer k l = Product_Type.fst (divmod_integer k l);

fun div_int k l =
  Int_of_integer (div_integer (integer_of_int k) (integer_of_int l));

fun mod_integer k l = Product_Type.snd (divmod_integer k l);

fun mod_int k l =
  Int_of_integer (mod_integer (integer_of_int k) (integer_of_int l));

fun less_eq_int k l = IntInf.<= (integer_of_int k, integer_of_int l);

val linorder_nat = {order_linorder = order_nat} : nat Orderings.linorder;

datatype natural = Nata of IntInf.int;

type 'a zero_neq_one = {one_zero_neq_one : 'a one, zero_zero_neq_one : 'a zero};
val one_zero_neq_one = #one_zero_neq_one : 'a zero_neq_one -> 'a one;
val zero_zero_neq_one = #zero_zero_neq_one : 'a zero_neq_one -> 'a zero;

val zero_neq_one_int =
  {one_zero_neq_one = one_int, zero_zero_neq_one = zero_int} : int zero_neq_one;

fun integer_of_natural (Nata x) = x;

fun div_natural m n =
  Nata (div_integer (integer_of_natural m) (integer_of_natural n));

fun mod_natural m n =
  Nata (mod_integer (integer_of_natural m) (integer_of_natural n));

val one_natural : natural = Nata (1 : IntInf.int);

fun less_natural m n = IntInf.< (integer_of_natural m, integer_of_natural n);

fun plus_natural m n =
  Nata (IntInf.+ (integer_of_natural m, integer_of_natural n));

val zero_natural : natural = Nata 0;

fun equal_natural m n =
  (((integer_of_natural m) : IntInf.int) = (integer_of_natural n));

fun minus_natural m n =
  Nata (Orderings.max ord_integer 0
         (IntInf.- (integer_of_natural m, integer_of_natural n)));

fun times_natural m n =
  Nata (IntInf.* (integer_of_natural m, integer_of_natural n));

fun int_of_nat n = Int_of_integer (integer_of_nat n);

fun nat_of_integer k = Nat (Orderings.max ord_integer 0 k);

fun nat_of_natural x = (nat_of_integer o integer_of_natural) x;

fun natural_of_nat n = Nata (integer_of_nat n);

fun less_eq_natural m n =
  IntInf.<= (integer_of_natural m, integer_of_natural n);

fun natural_of_integer k = Nata (Orderings.max ord_integer 0 k);

end; (*struct Arith*)

structure List : sig
  val hd : 'a list -> 'a
  val tl : 'a list -> 'a list
  val map : ('a -> 'b) -> 'a list -> 'b list
  val nth : 'a list -> Arith.nat -> 'a
  val fold : ('a -> 'b -> 'b) -> 'a list -> 'b -> 'b
  val rev : 'a list -> 'a list
  val upt : Arith.nat -> Arith.nat -> Arith.nat list
  val zip : 'a list -> 'b list -> ('a * 'b) list
  val maps : ('a -> 'b list) -> 'a list -> 'b list
  val null : 'a list -> bool
  val foldl : ('a -> 'b -> 'a) -> 'a -> 'b list -> 'a
  val foldr : ('a -> 'b -> 'b) -> 'a list -> 'b -> 'b
  val filter : ('a -> bool) -> 'a list -> 'a list
  val member : 'a HOL.equal -> 'a list -> 'a -> bool
  val insert : 'a HOL.equal -> 'a -> 'a list -> 'a list
  val remdups : 'a HOL.equal -> 'a list -> 'a list
  val remove1 : 'a HOL.equal -> 'a -> 'a list -> 'a list
  val distinct : 'a HOL.equal -> 'a list -> bool
  val list_all : ('a -> bool) -> 'a list -> bool
  val dropWhile : ('a -> bool) -> 'a list -> 'a list
  val equal_lista : 'a HOL.equal -> 'a list -> 'a list -> bool
  val equal_list : 'a HOL.equal -> ('a list) HOL.equal
  val removeAll : 'a HOL.equal -> 'a -> 'a list -> 'a list
  val replicate : Arith.nat -> 'a -> 'a list
  val gen_length : Arith.nat -> 'a list -> Arith.nat
  val size_list : 'a list -> Arith.nat
  val insort_key :
    'b Orderings.linorder -> ('a -> 'b) -> 'a -> 'a list -> 'a list
  val insort_insert_key :
    'b HOL.equal * 'b Orderings.linorder ->
      ('a -> 'b) -> 'a -> 'a list -> 'a list
end = struct

fun hd (x :: xs) = x;

fun tl [] = []
  | tl (x :: xs) = xs;

fun map f [] = []
  | map f (x :: xs) = f x :: map f xs;

fun nth (x :: xs) n =
  (if Arith.equal_nata n Arith.zero_nat then x
    else nth xs (Arith.minus_nat n Arith.one_nat));

fun fold f (x :: xs) s = fold f xs (f x s)
  | fold f [] s = s;

fun rev xs = fold (fn a => fn b => a :: b) xs [];

fun upt i j = (if Arith.less_nat i j then i :: upt (Arith.suc i) j else []);

fun zip (x :: xs) (y :: ys) = (x, y) :: zip xs ys
  | zip xs [] = []
  | zip [] ys = [];

fun maps f [] = []
  | maps f (x :: xs) = f x @ maps f xs;

fun null [] = true
  | null (x :: xs) = false;

fun foldl f a [] = a
  | foldl f a (x :: xs) = foldl f (f a x) xs;

fun foldr f [] = Fun.id
  | foldr f (x :: xs) = f x o foldr f xs;

fun filter p [] = []
  | filter p (x :: xs) = (if p x then x :: filter p xs else filter p xs);

fun member A_ [] y = false
  | member A_ (x :: xs) y = HOL.eq A_ x y orelse member A_ xs y;

fun insert A_ x xs = (if member A_ xs x then xs else x :: xs);

fun remdups A_ [] = []
  | remdups A_ (x :: xs) =
    (if member A_ xs x then remdups A_ xs else x :: remdups A_ xs);

fun remove1 A_ x [] = []
  | remove1 A_ x (y :: xs) =
    (if HOL.eq A_ x y then xs else y :: remove1 A_ x xs);

fun distinct A_ [] = true
  | distinct A_ (x :: xs) = not (member A_ xs x) andalso distinct A_ xs;

fun list_all p [] = true
  | list_all p (x :: xs) = p x andalso list_all p xs;

fun dropWhile p [] = []
  | dropWhile p (x :: xs) = (if p x then dropWhile p xs else x :: xs);

fun equal_lista A_ (a :: lista) [] = false
  | equal_lista A_ [] (a :: lista) = false
  | equal_lista A_ (aa :: listaa) (a :: lista) =
    HOL.eq A_ aa a andalso equal_lista A_ listaa lista
  | equal_lista A_ [] [] = true;

fun equal_list A_ = {equal = equal_lista A_} : ('a list) HOL.equal;

fun removeAll A_ x [] = []
  | removeAll A_ x (y :: xs) =
    (if HOL.eq A_ x y then removeAll A_ x xs else y :: removeAll A_ x xs);

fun replicate n x =
  (if Arith.equal_nata n Arith.zero_nat then []
    else x :: replicate (Arith.minus_nat n Arith.one_nat) x);

fun gen_length n (x :: xs) = gen_length (Arith.suc n) xs
  | gen_length n [] = n;

fun size_list x = gen_length Arith.zero_nat x;

fun insort_key B_ f x [] = [x]
  | insort_key B_ f x (y :: ys) =
    (if Orderings.less_eq
          ((Orderings.ord_preorder o Orderings.preorder_order o
             Orderings.order_linorder)
            B_)
          (f x) (f y)
      then x :: y :: ys else y :: insort_key B_ f x ys);

fun insort_insert_key (B1_, B2_) f x xs =
  (if member B1_ (map f xs) (f x) then xs else insort_key B2_ f x xs);

end; (*struct List*)

structure Phantom_Type : sig
  datatype ('a, 'b) phantom = Phantom of 'b
  val of_phantom : ('a, 'b) phantom -> 'b
end = struct

datatype ('a, 'b) phantom = Phantom of 'b;

fun of_phantom (Phantom x) = x;

end; (*struct Phantom_Type*)

structure Set : sig
  datatype 'a set = Set of 'a list | Coset of 'a list
  val member : 'a HOL.equal -> 'a -> 'a set -> bool
  val remove : 'a HOL.equal -> 'a -> 'a set -> 'a set
  val bot_set : 'a set
  val inf_set : 'a HOL.equal -> 'a set -> 'a set -> 'a set
  val top_set : 'a set
  val less_eq_set : 'a HOL.equal -> 'a set -> 'a set -> bool
  val equal_seta : 'a HOL.equal -> 'a set -> 'a set -> bool
  val equal_set : 'a HOL.equal -> 'a set HOL.equal
  val minus_set : 'a HOL.equal -> 'a set -> 'a set -> 'a set
  type 'a finite_UNIV
  val finite_UNIV : 'a finite_UNIV -> ('a, bool) Phantom_Type.phantom
  type 'a card_UNIV
  val finite_UNIV_card_UNIV : 'a card_UNIV -> 'a finite_UNIV
  val card_UNIV : 'a card_UNIV -> ('a, Arith.nat) Phantom_Type.phantom
  val card : 'a card_UNIV * 'a HOL.equal -> 'a set -> Arith.nat
  val card_UNIVa : 'a card_UNIV -> ('a, Arith.nat) Phantom_Type.phantom
  val union : 'a set -> 'a set -> 'a set
  val finite_UNIV_nata : (Arith.nat, bool) Phantom_Type.phantom
  val finite_UNIV_nat : Arith.nat finite_UNIV
  val card_UNIV_nata : (Arith.nat, Arith.nat) Phantom_Type.phantom
  val card_UNIV_nat : Arith.nat card_UNIV
  val is_list_UNIV : 'a card_UNIV * 'a HOL.equal -> 'a list -> bool
  val insert : 'a -> 'a set -> 'a set
end = struct

datatype 'a set = Set of 'a list | Coset of 'a list;

fun member A_ x (Coset xs) = not (List.member A_ xs x)
  | member A_ x (Set xs) = List.member A_ xs x;

fun remove A_ x (Coset xs) = Coset (List.insert A_ x xs)
  | remove A_ x (Set xs) = Set (List.removeAll A_ x xs);

val bot_set : 'a set = Set [];

fun inf_set A_ a (Coset xs) = List.fold (remove A_) xs a
  | inf_set A_ a (Set xs) = Set (List.filter (fn x => member A_ x a) xs);

val top_set : 'a set = Coset [];

fun less_eq_set A_ (Coset xs) (Set ys) =
  (if List.null xs andalso List.null ys then false
    else (raise Fail
           "subset_eq (List.coset _) (List.set _) requires type class instance card_UNIV")
           (fn _ => less_eq_set A_ (Coset xs) (Set ys)))
  | less_eq_set A_ a (Coset ys) = List.list_all (fn y => not (member A_ y a)) ys
  | less_eq_set A_ (Set xs) b = List.list_all (fn x => member A_ x b) xs;

fun equal_seta A_ a b = less_eq_set A_ a b andalso less_eq_set A_ b a;

fun equal_set A_ = {equal = equal_seta A_} : 'a set HOL.equal;

fun minus_set A_ a (Coset xs) = Set (List.filter (fn x => member A_ x a) xs)
  | minus_set A_ a (Set xs) = List.fold (remove A_) xs a;

type 'a finite_UNIV = {finite_UNIV : ('a, bool) Phantom_Type.phantom};
val finite_UNIV = #finite_UNIV :
  'a finite_UNIV -> ('a, bool) Phantom_Type.phantom;

type 'a card_UNIV =
  {finite_UNIV_card_UNIV : 'a finite_UNIV,
    card_UNIV : ('a, Arith.nat) Phantom_Type.phantom};
val finite_UNIV_card_UNIV = #finite_UNIV_card_UNIV :
  'a card_UNIV -> 'a finite_UNIV;
val card_UNIV = #card_UNIV :
  'a card_UNIV -> ('a, Arith.nat) Phantom_Type.phantom;

fun card (A1_, A2_) (Coset xs) =
  Arith.minus_nat (Phantom_Type.of_phantom (card_UNIV A1_))
    (List.size_list (List.remdups A2_ xs))
  | card (A1_, A2_) (Set xs) = List.size_list (List.remdups A2_ xs);

fun card_UNIVa A_ = card_UNIV A_;

fun union (Set xs) (Set ys) = Set (xs @ ys);

val finite_UNIV_nata : (Arith.nat, bool) Phantom_Type.phantom =
  Phantom_Type.Phantom false;

val finite_UNIV_nat = {finite_UNIV = finite_UNIV_nata} : Arith.nat finite_UNIV;

val card_UNIV_nata : (Arith.nat, Arith.nat) Phantom_Type.phantom =
  Phantom_Type.Phantom Arith.zero_nat;

val card_UNIV_nat =
  {finite_UNIV_card_UNIV = finite_UNIV_nat, card_UNIV = card_UNIV_nata} :
  Arith.nat card_UNIV;

fun is_list_UNIV (A1_, A2_) xs =
  let
    val c = Phantom_Type.of_phantom (card_UNIVa A1_);
  in
    (if Arith.equal_nata c Arith.zero_nat then false
      else Arith.equal_nata (List.size_list (List.remdups A2_ xs)) c)
  end;

fun insert x (Set xs) = Set (x :: xs);

end; (*struct Set*)

structure FinFun : sig
  datatype ('a, 'b) finfun = Finfun_const of 'b |
    Finfun_update_code of ('a, 'b) finfun * 'a * 'b
  val finfun_All_except :
    'a Set.card_UNIV * 'a HOL.equal -> 'a list -> ('a, bool) finfun -> bool
  val finfun_All : 'a Set.card_UNIV * 'a HOL.equal -> ('a, bool) finfun -> bool
  val finfun_comp : ('a -> 'b) -> ('c, 'a) finfun -> ('c, 'b) finfun
  val finfun_update :
    'a HOL.equal -> 'b HOL.equal ->
      ('a, 'b) finfun -> 'a -> 'b -> ('a, 'b) finfun
  val finfun_apply : 'a HOL.equal -> ('a, 'b) finfun -> 'a -> 'b
  val finfun_Diag :
    'a HOL.equal -> 'b HOL.equal -> 'c HOL.equal ->
      ('a, 'b) finfun -> ('a, 'c) finfun -> ('a, ('b * 'c)) finfun
  val equal_finfuna :
    'a Set.card_UNIV * 'a HOL.equal -> 'b HOL.equal ->
      ('a, 'b) finfun -> ('a, 'b) finfun -> bool
  val equal_finfun :
    'a Set.card_UNIV * 'a HOL.equal -> 'b HOL.equal -> ('a, 'b) finfun HOL.equal
  val finfun_default : 'a Set.card_UNIV * 'a HOL.equal -> ('a, 'b) finfun -> 'b
  val finfun_to_list :
    'a Set.card_UNIV * 'a HOL.equal * 'a Orderings.linorder -> 'b HOL.equal ->
      ('a, 'b) finfun -> 'a list
end = struct

datatype ('a, 'b) finfun = Finfun_const of 'b |
  Finfun_update_code of ('a, 'b) finfun * 'a * 'b;

fun finfun_All_except (A1_, A2_) aa (Finfun_update_code (f, a, b)) =
  (List.member A2_ aa a orelse b) andalso
    finfun_All_except (A1_, A2_) (a :: aa) f
  | finfun_All_except (A1_, A2_) a (Finfun_const b) =
    b orelse Set.is_list_UNIV (A1_, A2_) a;

fun finfun_All (A1_, A2_) = finfun_All_except (A1_, A2_) [];

fun finfun_comp g (Finfun_update_code (f, a, b)) =
  Finfun_update_code (finfun_comp g f, a, g b)
  | finfun_comp g (Finfun_const c) = Finfun_const (g c);

fun finfun_update A_ B_ (Finfun_update_code (f, aa, ba)) a b =
  (if HOL.eq A_ aa a then finfun_update A_ B_ f aa b
    else Finfun_update_code (finfun_update A_ B_ f a b, aa, ba))
  | finfun_update A_ B_ (Finfun_const ba) a b =
    (if HOL.eq B_ ba b then Finfun_const ba
      else Finfun_update_code (Finfun_const ba, a, b));

fun finfun_apply A_ (Finfun_const b) a = b
  | finfun_apply A_ (Finfun_update_code (f, aa, b)) a =
    (if HOL.eq A_ aa a then b else finfun_apply A_ f a);

fun finfun_Diag A_ B_ C_ (Finfun_update_code (f, a, b)) g =
  finfun_update A_ (Product_Type.equal_prod B_ C_) (finfun_Diag A_ B_ C_ f g) a
    (b, finfun_apply A_ g a)
  | finfun_Diag A_ B_ C_ (Finfun_const b) (Finfun_update_code (g, a, c)) =
    Finfun_update_code (finfun_Diag A_ B_ C_ (Finfun_const b) g, a, (b, c))
  | finfun_Diag A_ B_ C_ (Finfun_const b) (Finfun_const c) =
    Finfun_const (b, c);

fun equal_finfuna (A1_, A2_) B_ f g =
  finfun_All (A1_, A2_)
    (finfun_comp (fn (a, b) => HOL.eq B_ a b) (finfun_Diag A2_ B_ B_ f g));

fun equal_finfun (A1_, A2_) B_ = {equal = equal_finfuna (A1_, A2_) B_} :
  ('a, 'b) finfun HOL.equal;

fun finfun_default (A1_, A2_) (Finfun_update_code (f, a, b)) =
  finfun_default (A1_, A2_) f
  | finfun_default (A1_, A2_) (Finfun_const c) =
    (if Arith.equal_nata (Set.card (A1_, A2_) Set.top_set) Arith.zero_nat then c
      else (raise Fail "undefined"));

fun finfun_to_list (A1_, A2_, A3_) B_ (Finfun_update_code (f, a, b)) =
  (if HOL.eq B_ b (finfun_default (A1_, A2_) f)
    then List.remove1 A2_ a (finfun_to_list (A1_, A2_, A3_) B_ f)
    else List.insort_insert_key (A2_, A3_) (fn x => x) a
           (finfun_to_list (A1_, A2_, A3_) B_ f))
  | finfun_to_list (A1_, A2_, A3_) B_ (Finfun_const c) =
    (if Arith.equal_nata (Set.card (A1_, A2_) Set.top_set) Arith.zero_nat
      then []
      else (raise Fail "finfun_to_list called on finite type")
             (fn _ => finfun_to_list (A1_, A2_, A3_) B_ (Finfun_const c)));

end; (*struct FinFun*)

structure Countable : sig
  type 'a countable
end = struct

type 'a countable = {};

end; (*struct Countable*)

structure Finite_Set : sig
  type 'a finite
  val countable_finite : 'a finite -> 'a Countable.countable
end = struct

type 'a finite = {countable_finite : 'a Countable.countable};
val countable_finite = #countable_finite : 'a finite -> 'a Countable.countable;

end; (*struct Finite_Set*)

structure Numeral_Type : sig
  datatype 'a bit0 = Abs_bit0 of Arith.int
  datatype num1 = One_num1
end = struct

datatype 'a bit0 = Abs_bit0 of Arith.int;

datatype num1 = One_num1;

end; (*struct Numeral_Type*)

structure Type_Length : sig
  type 'a len0
  val len_of : 'a len0 -> 'a HOL.itself -> Arith.nat
  type 'a len
  val len0_len : 'a len -> 'a len0
  val len_of_bit0 : 'a len0 -> 'a Numeral_Type.bit0 HOL.itself -> Arith.nat
  val len0_bit0 : 'a len0 -> 'a Numeral_Type.bit0 len0
  val len_bit0 : 'a len -> 'a Numeral_Type.bit0 len
  val len_of_num1 : Numeral_Type.num1 HOL.itself -> Arith.nat
  val len0_num1 : Numeral_Type.num1 len0
  val len_num1 : Numeral_Type.num1 len
end = struct

type 'a len0 = {len_of : 'a HOL.itself -> Arith.nat};
val len_of = #len_of : 'a len0 -> 'a HOL.itself -> Arith.nat;

type 'a len = {len0_len : 'a len0};
val len0_len = #len0_len : 'a len -> 'a len0;

fun len_of_bit0 A_ x =
  Arith.times_nat (Arith.nat_of_integer (2 : IntInf.int)) (len_of A_ HOL.Type);

fun len0_bit0 A_ = {len_of = len_of_bit0 A_} : 'a Numeral_Type.bit0 len0;

fun len_bit0 A_ = {len0_len = len0_bit0 (len0_len A_)} :
  'a Numeral_Type.bit0 len;

fun len_of_num1 x = Arith.one_nat;

val len0_num1 = {len_of = len_of_num1} : Numeral_Type.num1 len0;

val len_num1 = {len0_len = len0_num1} : Numeral_Type.num1 len;

end; (*struct Type_Length*)

structure Bit : sig
  datatype bit = Zero_bit | One_bit
end = struct

datatype bit = Zero_bit | One_bit;

end; (*struct Bit*)

structure Bit_Representation : sig
  val bitval : 'a Arith.zero_neq_one -> Bit.bit -> 'a
  val bit : Arith.int -> Bit.bit -> Arith.int
  val bin_last : Arith.int -> Bit.bit
  val bin_rest : Arith.int -> Arith.int
  val sbintrunc : Arith.nat -> Arith.int -> Arith.int
end = struct

fun bitval A_ =
  (fn a =>
    (case a of Bit.Zero_bit => Arith.zero (Arith.zero_zero_neq_one A_)
      | Bit.One_bit => Arith.one (Arith.one_zero_neq_one A_)));

fun bit k b =
  Arith.plus_int (Arith.plus_int (bitval Arith.zero_neq_one_int b) k) k;

fun bin_last w =
  (if Arith.equal_int (Arith.mod_int w (Arith.Int_of_integer (2 : IntInf.int)))
        Arith.zero_inta
    then Bit.Zero_bit else Bit.One_bit);

fun bin_rest w = Arith.div_int w (Arith.Int_of_integer (2 : IntInf.int));

fun sbintrunc n bin =
  (if Arith.equal_nata n Arith.zero_nat
    then (case bin_last bin of Bit.Zero_bit => Arith.zero_inta
           | Bit.One_bit => Arith.Int_of_integer (~1 : IntInf.int))
    else bit (sbintrunc (Arith.minus_nat n Arith.one_nat) (bin_rest bin))
           (bin_last bin));

end; (*struct Bit_Representation*)

structure Bit_Operations : sig
  val bitAND_bit : Bit.bit -> Bit.bit -> Bit.bit
end = struct

fun bitAND_bit Bit.Zero_bit y = Bit.Zero_bit
  | bitAND_bit Bit.One_bit y = y;

end; (*struct Bit_Operations*)

structure Bit_Int : sig
  val bitNOT_int : Arith.int -> Arith.int
  val bitAND_int : Arith.int -> Arith.int -> Arith.int
  val bitOR_int : Arith.int -> Arith.int -> Arith.int
  val bitXOR_int : Arith.int -> Arith.int -> Arith.int
end = struct

fun bitNOT_int x =
  (fn xa =>
    Arith.minus_int (Arith.uminus_int xa)
      (Arith.Int_of_integer (1 : IntInf.int)))
    x;

fun bitAND_int x y =
  (if Arith.equal_int x Arith.zero_inta then Arith.zero_inta
    else (if Arith.equal_int x (Arith.Int_of_integer (~1 : IntInf.int)) then y
           else Bit_Representation.bit
                  (bitAND_int (Bit_Representation.bin_rest x)
                    (Bit_Representation.bin_rest y))
                  (Bit_Operations.bitAND_bit (Bit_Representation.bin_last x)
                    (Bit_Representation.bin_last y))));

fun bitOR_int x =
  (fn xa => fn y => bitNOT_int (bitAND_int (bitNOT_int xa) (bitNOT_int y))) x;

fun bitXOR_int x =
  (fn xa => fn y =>
    bitOR_int (bitAND_int xa (bitNOT_int y)) (bitAND_int (bitNOT_int xa) y))
    x;

end; (*struct Bit_Int*)

structure Word : sig
  datatype 'a word = Word of Arith.int
  val uint : 'a Type_Length.len0 -> 'a word -> Arith.int
  val sint : 'a Type_Length.len -> 'a word -> Arith.int
  val unat : 'a Type_Length.len0 -> 'a word -> Arith.nat
  val word_of_int : 'a Type_Length.len0 -> Arith.int -> 'a word
  val shiftl1 : 'a Type_Length.len0 -> 'a word -> 'a word
  val shiftr1 : 'a Type_Length.len0 -> 'a word -> 'a word
  val sshiftr1 : 'a Type_Length.len -> 'a word -> 'a word
  val sshiftr : 'a Type_Length.len -> 'a word -> Arith.nat -> 'a word
  val word_sle : 'a Type_Length.len -> 'a word -> 'a word -> bool
  val plus_word : 'a Type_Length.len0 -> 'a word -> 'a word -> 'a word
  val zero_word : 'a Type_Length.len0 -> 'a word
  val bitOR_word : 'a Type_Length.len0 -> 'a word -> 'a word -> 'a word
  val equal_word : 'a Type_Length.len0 -> 'a word -> 'a word -> bool
  val minus_word : 'a Type_Length.len0 -> 'a word -> 'a word -> 'a word
  val times_word : 'a Type_Length.len0 -> 'a word -> 'a word -> 'a word
  val word_sless : 'a Type_Length.len -> 'a word -> 'a word -> bool
  val bitAND_word : 'a Type_Length.len0 -> 'a word -> 'a word -> 'a word
  val bitXOR_word : 'a Type_Length.len0 -> 'a word -> 'a word -> 'a word
  val shiftl_word : 'a Type_Length.len0 -> 'a word -> Arith.nat -> 'a word
  val shiftr_word : 'a Type_Length.len0 -> 'a word -> Arith.nat -> 'a word
end = struct

datatype 'a word = Word of Arith.int;

fun uint A_ (Word x) = x;

fun sint A_ w =
  Bit_Representation.sbintrunc
    (Arith.minus_nat (Type_Length.len_of (Type_Length.len0_len A_) HOL.Type)
      Arith.one_nat)
    (uint (Type_Length.len0_len A_) w);

fun unat A_ w = Arith.nat (uint A_ w);

fun word_of_int A_ k =
  Word (Arith.mod_int k
         (Arith.power Arith.power_int (Arith.Int_of_integer (2 : IntInf.int))
           (Type_Length.len_of A_ HOL.Type)));

fun shiftl1 A_ w =
  word_of_int A_ (Bit_Representation.bit (uint A_ w) Bit.Zero_bit);

fun shiftr1 A_ w = word_of_int A_ (Bit_Representation.bin_rest (uint A_ w));

fun sshiftr1 A_ w =
  word_of_int (Type_Length.len0_len A_)
    (Bit_Representation.bin_rest (sint A_ w));

fun sshiftr A_ w n = Arith.funpow n (sshiftr1 A_) w;

fun word_sle A_ a b = Arith.less_eq_int (sint A_ a) (sint A_ b);

fun plus_word A_ a b = word_of_int A_ (Arith.plus_int (uint A_ a) (uint A_ b));

fun zero_word A_ = word_of_int A_ Arith.zero_inta;

fun bitOR_word A_ a b =
  word_of_int A_ (Bit_Int.bitOR_int (uint A_ a) (uint A_ b));

fun equal_word A_ k l = Arith.equal_int (uint A_ k) (uint A_ l);

fun minus_word A_ a b =
  word_of_int A_ (Arith.minus_int (uint A_ a) (uint A_ b));

fun times_word A_ a b =
  word_of_int A_ (Arith.times_inta (uint A_ a) (uint A_ b));

fun word_sless A_ x y =
  word_sle A_ x y andalso not (equal_word (Type_Length.len0_len A_) x y);

fun bitAND_word A_ a b =
  word_of_int A_ (Bit_Int.bitAND_int (uint A_ a) (uint A_ b));

fun bitXOR_word A_ a b =
  word_of_int A_ (Bit_Int.bitXOR_int (uint A_ a) (uint A_ b));

fun shiftl_word A_ w n = Arith.funpow n (shiftl1 A_) w;

fun shiftr_word A_ w n = Arith.funpow n (shiftr1 A_) w;

end; (*struct Word*)

structure Predicate : sig
  datatype 'a seq = Empty | Insert of 'a * 'a pred | Join of 'a pred * 'a seq
  and 'a pred = Seq of (unit -> 'a seq)
  val is_empty : 'a pred -> bool
  val null : 'a seq -> bool
  val singleton : 'a HOL.equal -> (unit -> 'a) -> 'a pred -> 'a
  val the_only : 'a HOL.equal -> (unit -> 'a) -> 'a seq -> 'a
  val the : 'a HOL.equal -> 'a pred -> 'a
  val bind : 'a pred -> ('a -> 'b pred) -> 'b pred
  val apply : ('a -> 'b pred) -> 'a seq -> 'b seq
  val eval : 'a HOL.equal -> 'a pred -> 'a -> bool
  val member : 'a HOL.equal -> 'a seq -> 'a -> bool
  val holds : unit pred -> bool
  val bot_pred : 'a pred
  val single : 'a -> 'a pred
  val adjunct : 'a pred -> 'a seq -> 'a seq
  val sup_pred : 'a pred -> 'a pred -> 'a pred
  val if_pred : bool -> unit pred
  val not_pred : unit pred -> unit pred
  val pred_of_set : 'a Set.set -> 'a pred
end = struct

datatype 'a seq = Empty | Insert of 'a * 'a pred | Join of 'a pred * 'a seq
and 'a pred = Seq of (unit -> 'a seq);

fun is_empty (Seq f) = null (f ())
and null Empty = true
  | null (Insert (x, p)) = false
  | null (Join (p, xq)) = is_empty p andalso null xq;

fun singleton A_ dfault (Seq f) =
  (case f () of Empty => dfault ()
    | Insert (x, p) =>
      (if is_empty p then x
        else let
               val y = singleton A_ dfault p;
             in
               (if HOL.eq A_ x y then x else dfault ())
             end)
    | Join (p, xq) =>
      (if is_empty p then the_only A_ dfault xq
        else (if null xq then singleton A_ dfault p
               else let
                      val x = singleton A_ dfault p;
                      val y = the_only A_ dfault xq;
                    in
                      (if HOL.eq A_ x y then x else dfault ())
                    end)))
and the_only A_ dfault (Insert (x, p)) =
  (if is_empty p then x
    else let
           val y = singleton A_ dfault p;
         in
           (if HOL.eq A_ x y then x else dfault ())
         end)
  | the_only A_ dfault (Join (p, xq)) =
    (if is_empty p then the_only A_ dfault xq
      else (if null xq then singleton A_ dfault p
             else let
                    val x = singleton A_ dfault p;
                    val y = the_only A_ dfault xq;
                  in
                    (if HOL.eq A_ x y then x else dfault ())
                  end));

fun the A_ a =
  singleton A_ (fn _ => (raise Fail "not_unique") (fn _ => the A_ a)) a;

fun bind (Seq g) f = Seq (fn _ => apply f (g ()))
and apply f Empty = Empty
  | apply f (Insert (x, p)) = Join (f x, Join (bind p f, Empty))
  | apply f (Join (p, xq)) = Join (bind p f, apply f xq);

fun eval A_ (Seq f) = member A_ (f ())
and member A_ Empty x = false
  | member A_ (Insert (y, p)) x = HOL.eq A_ x y orelse eval A_ p x
  | member A_ (Join (p, xq)) x = eval A_ p x orelse member A_ xq x;

fun holds p = eval Product_Type.equal_unit p ();

val bot_pred : 'a pred = Seq (fn _ => Empty);

fun single x = Seq (fn _ => Insert (x, bot_pred));

fun adjunct p Empty = Join (p, Empty)
  | adjunct p (Insert (x, q)) = Insert (x, sup_pred q p)
  | adjunct p (Join (q, xq)) = Join (q, adjunct p xq)
and sup_pred (Seq f) (Seq g) =
  Seq (fn _ =>
        (case f () of Empty => g ()
          | Insert (x, p) => Insert (x, sup_pred p (Seq g))
          | Join (p, xq) => adjunct (Seq g) (Join (p, xq))));

fun if_pred b = (if b then single () else bot_pred);

fun not_pred p =
  (if eval Product_Type.equal_unit p () then bot_pred else single ());

fun pred_of_set (Set.Set xs) = List.foldr (sup_pred o single) xs bot_pred;

end; (*struct Predicate*)

structure Type : sig
  datatype ty = Void | Boolean | Integer | Nt | Class of string | Array of ty
  datatype htype = Class_type of string | Array_type of ty * Arith.nat
  val object : string
  val thread : string
  val equal_tya : ty -> ty -> bool
  val equal_ty : ty HOL.equal
  val is_refT_i : ty -> unit Predicate.pred
  val is_refT : ty -> bool
  val throwable : string
  val ground_type : ty -> ty
  val ty_of_htype : htype -> ty
  val class_type_of : ty -> string option
  val class_type_ofa : htype -> string
end = struct

datatype ty = Void | Boolean | Integer | Nt | Class of string | Array of ty;

datatype htype = Class_type of string | Array_type of ty * Arith.nat;

val object : string = "java/lang/Object";

val thread : string = "java/lang/Thread";

fun equal_tya (Array ty) (Class literal) = false
  | equal_tya (Class literal) (Array ty) = false
  | equal_tya (Array ty) Nt = false
  | equal_tya Nt (Array ty) = false
  | equal_tya (Class literal) Nt = false
  | equal_tya Nt (Class literal) = false
  | equal_tya (Array ty) Integer = false
  | equal_tya Integer (Array ty) = false
  | equal_tya (Class literal) Integer = false
  | equal_tya Integer (Class literal) = false
  | equal_tya Nt Integer = false
  | equal_tya Integer Nt = false
  | equal_tya (Array ty) Boolean = false
  | equal_tya Boolean (Array ty) = false
  | equal_tya (Class literal) Boolean = false
  | equal_tya Boolean (Class literal) = false
  | equal_tya Nt Boolean = false
  | equal_tya Boolean Nt = false
  | equal_tya Integer Boolean = false
  | equal_tya Boolean Integer = false
  | equal_tya (Array ty) Void = false
  | equal_tya Void (Array ty) = false
  | equal_tya (Class literal) Void = false
  | equal_tya Void (Class literal) = false
  | equal_tya Nt Void = false
  | equal_tya Void Nt = false
  | equal_tya Integer Void = false
  | equal_tya Void Integer = false
  | equal_tya Boolean Void = false
  | equal_tya Void Boolean = false
  | equal_tya (Array tya) (Array ty) = equal_tya tya ty
  | equal_tya (Class literala) (Class literal) = ((literala : string) = literal)
  | equal_tya Nt Nt = true
  | equal_tya Integer Integer = true
  | equal_tya Boolean Boolean = true
  | equal_tya Void Void = true;

val equal_ty = {equal = equal_tya} : ty HOL.equal;

fun is_refT_i x =
  (case x of Void => Predicate.bot_pred | Boolean => Predicate.bot_pred
    | Integer => Predicate.bot_pred | Nt => Predicate.single ()
    | Class _ => Predicate.single () | Array _ => Predicate.single ());

fun is_refT x1 = Predicate.holds (is_refT_i x1);

val throwable : string = "java/lang/Throwable";

fun ground_type (Array t) = ground_type t
  | ground_type Void = Void
  | ground_type Boolean = Boolean
  | ground_type Integer = Integer
  | ground_type Nt = Nt
  | ground_type (Class v) = Class v;

fun ty_of_htype (Class_type c) = Class c
  | ty_of_htype (Array_type (t, n)) = Array t;

fun class_type_of (Class c) = SOME c
  | class_type_of (Array t) = SOME object
  | class_type_of Void = NONE
  | class_type_of Boolean = NONE
  | class_type_of Integer = NONE
  | class_type_of Nt = NONE;

fun class_type_ofa (Class_type c) = c
  | class_type_ofa (Array_type (t, n)) = object;

end; (*struct Type*)

structure Value : sig
  datatype 'a vala = Unit | Null | Bool of bool |
    Intg of
      Numeral_Type.num1 Numeral_Type.bit0 Numeral_Type.bit0 Numeral_Type.bit0
        Numeral_Type.bit0
        Numeral_Type.bit0
        Word.word
    | Addr of 'a
  type 'a addr
  val hash_addr : 'a addr -> 'a -> Arith.int
  val monitor_finfun_to_list :
    'a addr -> ('a, Arith.nat) FinFun.finfun -> 'a list
  val equal_vala : 'a HOL.equal -> 'a vala -> 'a vala -> bool
  val equal_val : 'a HOL.equal -> 'a vala HOL.equal
  val default_val : Type.ty -> 'a vala
end = struct

datatype 'a vala = Unit | Null | Bool of bool |
  Intg of
    Numeral_Type.num1 Numeral_Type.bit0 Numeral_Type.bit0 Numeral_Type.bit0
      Numeral_Type.bit0
      Numeral_Type.bit0
      Word.word
  | Addr of 'a;

type 'a addr =
  {hash_addr : 'a -> Arith.int,
    monitor_finfun_to_list : ('a, Arith.nat) FinFun.finfun -> 'a list};
val hash_addr = #hash_addr : 'a addr -> 'a -> Arith.int;
val monitor_finfun_to_list = #monitor_finfun_to_list :
  'a addr -> ('a, Arith.nat) FinFun.finfun -> 'a list;

fun equal_vala A_ (Addr addr) (Intg word) = false
  | equal_vala A_ (Intg word) (Addr addr) = false
  | equal_vala A_ (Addr addr) (Bool boola) = false
  | equal_vala A_ (Bool boola) (Addr addr) = false
  | equal_vala A_ (Intg word) (Bool boola) = false
  | equal_vala A_ (Bool boola) (Intg word) = false
  | equal_vala A_ (Addr addr) Null = false
  | equal_vala A_ Null (Addr addr) = false
  | equal_vala A_ (Intg word) Null = false
  | equal_vala A_ Null (Intg word) = false
  | equal_vala A_ (Bool boola) Null = false
  | equal_vala A_ Null (Bool boola) = false
  | equal_vala A_ (Addr addr) Unit = false
  | equal_vala A_ Unit (Addr addr) = false
  | equal_vala A_ (Intg word) Unit = false
  | equal_vala A_ Unit (Intg word) = false
  | equal_vala A_ (Bool boola) Unit = false
  | equal_vala A_ Unit (Bool boola) = false
  | equal_vala A_ Null Unit = false
  | equal_vala A_ Unit Null = false
  | equal_vala A_ (Addr addra) (Addr addr) = HOL.eq A_ addra addr
  | equal_vala A_ (Intg worda) (Intg word) =
    Word.equal_word
      (Type_Length.len0_bit0
        (Type_Length.len0_bit0
          (Type_Length.len0_bit0
            (Type_Length.len0_bit0
              (Type_Length.len0_bit0 Type_Length.len0_num1)))))
      worda word
  | equal_vala A_ (Bool boolaa) (Bool boola) =
    Product_Type.equal_boola boolaa boola
  | equal_vala A_ Null Null = true
  | equal_vala A_ Unit Unit = true;

fun equal_val A_ = {equal = equal_vala A_} : 'a vala HOL.equal;

fun default_val Type.Void = Unit
  | default_val Type.Boolean = Bool false
  | default_val Type.Integer =
    Intg (Word.zero_word
           (Type_Length.len0_bit0
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0 Type_Length.len0_num1))))))
  | default_val Type.Nt = Null
  | default_val (Type.Class c) = Null
  | default_val (Type.Array a) = Null;

end; (*struct Value*)

structure MM : sig
  val monitor_finfun_to_list_nat :
    (Arith.nat, Arith.nat) FinFun.finfun -> Arith.nat list
  val hash_addr_nat : Arith.nat -> Arith.int
  val addr_nat : Arith.nat Value.addr
end = struct

fun monitor_finfun_to_list_nat x =
  FinFun.finfun_to_list (Set.card_UNIV_nat, Arith.equal_nat, Arith.linorder_nat)
    Arith.equal_nat x;

fun hash_addr_nat x = Arith.int_of_nat x;

val addr_nat =
  {hash_addr = hash_addr_nat,
    monitor_finfun_to_list = monitor_finfun_to_list_nat}
  : Arith.nat Value.addr;

end; (*struct MM*)

structure Err : sig
  datatype 'a err = Err | Ok of 'a
end = struct

datatype 'a err = Err | Ok of 'a;

end; (*struct Err*)

structure Option : sig
  val map : ('a -> 'b) -> 'a option -> 'b option
  val the : 'a option -> 'a
  val bind : 'a option -> ('a -> 'b option) -> 'b option
  val is_none : 'a option -> bool
  val equal_optiona : 'a HOL.equal -> 'a option -> 'a option -> bool
  val equal_option : 'a HOL.equal -> ('a option) HOL.equal
end = struct

fun map f (SOME x) = SOME (f x)
  | map f NONE = NONE;

fun the (SOME x) = x;

fun bind NONE f = NONE
  | bind (SOME x) f = f x;

fun is_none (SOME x) = false
  | is_none NONE = true;

fun equal_optiona A_ (SOME a) NONE = false
  | equal_optiona A_ NONE (SOME a) = false
  | equal_optiona A_ (SOME aa) (SOME a) = HOL.eq A_ aa a
  | equal_optiona A_ NONE NONE = true;

fun equal_option A_ = {equal = equal_optiona A_} : ('a option) HOL.equal;

end; (*struct Option*)

structure Map : sig
  val map_of : 'a HOL.equal -> ('a * 'b) list -> 'a -> 'b option
  val map_add : ('a -> 'b option) -> ('a -> 'b option) -> 'a -> 'b option
  val map_upds :
    'a HOL.equal -> ('a -> 'b option) -> 'a list -> 'b list -> 'a -> 'b option
end = struct

fun map_of A_ ((l, v) :: ps) k =
  (if HOL.eq A_ l k then SOME v else map_of A_ ps k)
  | map_of A_ [] k = NONE;

fun map_add m1 m2 = (fn x => (case m2 x of NONE => m1 x | SOME a => SOME a));

fun map_upds A_ m xs ys = map_add m (map_of A_ (List.rev (List.zip xs ys)));

end; (*struct Map*)

structure RBT_Impl : sig
  datatype color = R | B
  datatype ('a, 'b) rbt = Empty |
    Branch of color * ('a, 'b) rbt * 'a * 'b * ('a, 'b) rbt
  val paint : color -> ('a, 'b) rbt -> ('a, 'b) rbt
  val balance : ('a, 'b) rbt -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val balance_left : ('a, 'b) rbt -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val combine : ('a, 'b) rbt -> ('a, 'b) rbt -> ('a, 'b) rbt
  val balance_right : ('a, 'b) rbt -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val rbt_del : 'a Orderings.ord -> 'a -> ('a, 'b) rbt -> ('a, 'b) rbt
  val rbt_del_from_left :
    'a Orderings.ord ->
      'a -> ('a, 'b) rbt -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val rbt_del_from_right :
    'a Orderings.ord ->
      'a -> ('a, 'b) rbt -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val rbt_ins :
    'a Orderings.ord ->
      ('a -> 'b -> 'b -> 'b) -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val equal_color : color -> color -> bool
  val equal_rbt :
    'a HOL.equal -> 'b HOL.equal -> ('a, 'b) rbt -> ('a, 'b) rbt -> bool
  val rbt_delete : 'a Orderings.ord -> 'a -> ('a, 'b) rbt -> ('a, 'b) rbt
  val rbt_insert_with_key :
    'a Orderings.ord ->
      ('a -> 'b -> 'b -> 'b) -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val rbt_insert : 'a Orderings.ord -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val rbt_lookup : 'a Orderings.ord -> ('a, 'b) rbt -> 'a -> 'b option
end = struct

datatype color = R | B;

datatype ('a, 'b) rbt = Empty |
  Branch of color * ('a, 'b) rbt * 'a * 'b * ('a, 'b) rbt;

fun paint c Empty = Empty
  | paint c (Branch (uu, l, k, v, r)) = Branch (c, l, k, v, r);

fun balance (Branch (R, a, w, x, b)) s t (Branch (R, c, y, z, d)) =
  Branch (R, Branch (B, a, w, x, b), s, t, Branch (B, c, y, z, d))
  | balance (Branch (R, Branch (R, a, w, x, b), s, t, c)) y z Empty =
    Branch (R, Branch (B, a, w, x, b), s, t, Branch (B, c, y, z, Empty))
  | balance (Branch (R, Branch (R, a, w, x, b), s, t, c)) y z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, Branch (B, a, w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance (Branch (R, Empty, w, x, Branch (R, b, s, t, c))) y z Empty =
    Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, Empty))
  | balance
    (Branch (R, Branch (B, va, vb, vc, vd), w, x, Branch (R, b, s, t, c))) y z
    Empty =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, Empty))
  | balance (Branch (R, Empty, w, x, Branch (R, b, s, t, c))) y z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, Branch (B, Empty, w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance
    (Branch (R, Branch (B, ve, vf, vg, vh), w, x, Branch (R, b, s, t, c))) y z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, Branch (B, Branch (B, ve, vf, vg, vh), w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance Empty w x (Branch (R, b, s, t, Branch (R, c, y, z, d))) =
    Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, d))
  | balance (Branch (B, va, vb, vc, vd)) w x
    (Branch (R, b, s, t, Branch (R, c, y, z, d))) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, d))
  | balance Empty w x (Branch (R, Branch (R, b, s, t, c), y, z, Empty)) =
    Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, Empty))
  | balance Empty w x
    (Branch (R, Branch (R, b, s, t, c), y, z, Branch (B, va, vb, vc, vd))) =
    Branch
      (R, Branch (B, Empty, w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance (Branch (B, va, vb, vc, vd)) w x
    (Branch (R, Branch (R, b, s, t, c), y, z, Empty)) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, Empty))
  | balance (Branch (B, va, vb, vc, vd)) w x
    (Branch (R, Branch (R, b, s, t, c), y, z, Branch (B, ve, vf, vg, vh))) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, ve, vf, vg, vh)))
  | balance Empty s t Empty = Branch (B, Empty, s, t, Empty)
  | balance Empty s t (Branch (B, va, vb, vc, vd)) =
    Branch (B, Empty, s, t, Branch (B, va, vb, vc, vd))
  | balance Empty s t (Branch (v, Empty, vb, vc, Empty)) =
    Branch (B, Empty, s, t, Branch (v, Empty, vb, vc, Empty))
  | balance Empty s t (Branch (v, Branch (B, ve, vf, vg, vh), vb, vc, Empty)) =
    Branch
      (B, Empty, s, t, Branch (v, Branch (B, ve, vf, vg, vh), vb, vc, Empty))
  | balance Empty s t (Branch (v, Empty, vb, vc, Branch (B, vf, vg, vh, vi))) =
    Branch
      (B, Empty, s, t, Branch (v, Empty, vb, vc, Branch (B, vf, vg, vh, vi)))
  | balance Empty s t
    (Branch (v, Branch (B, ve, vj, vk, vl), vb, vc, Branch (B, vf, vg, vh, vi)))
    = Branch
        (B, Empty, s, t,
          Branch
            (v, Branch (B, ve, vj, vk, vl), vb, vc, Branch (B, vf, vg, vh, vi)))
  | balance (Branch (B, va, vb, vc, vd)) s t Empty =
    Branch (B, Branch (B, va, vb, vc, vd), s, t, Empty)
  | balance (Branch (B, va, vb, vc, vd)) s t (Branch (B, ve, vf, vg, vh)) =
    Branch (B, Branch (B, va, vb, vc, vd), s, t, Branch (B, ve, vf, vg, vh))
  | balance (Branch (B, va, vb, vc, vd)) s t (Branch (v, Empty, vf, vg, Empty))
    = Branch
        (B, Branch (B, va, vb, vc, vd), s, t, Branch (v, Empty, vf, vg, Empty))
  | balance (Branch (B, va, vb, vc, vd)) s t
    (Branch (v, Branch (B, vi, vj, vk, vl), vf, vg, Empty)) =
    Branch
      (B, Branch (B, va, vb, vc, vd), s, t,
        Branch (v, Branch (B, vi, vj, vk, vl), vf, vg, Empty))
  | balance (Branch (B, va, vb, vc, vd)) s t
    (Branch (v, Empty, vf, vg, Branch (B, vj, vk, vl, vm))) =
    Branch
      (B, Branch (B, va, vb, vc, vd), s, t,
        Branch (v, Empty, vf, vg, Branch (B, vj, vk, vl, vm)))
  | balance (Branch (B, va, vb, vc, vd)) s t
    (Branch (v, Branch (B, vi, vn, vo, vp), vf, vg, Branch (B, vj, vk, vl, vm)))
    = Branch
        (B, Branch (B, va, vb, vc, vd), s, t,
          Branch
            (v, Branch (B, vi, vn, vo, vp), vf, vg, Branch (B, vj, vk, vl, vm)))
  | balance (Branch (v, Empty, vb, vc, Empty)) s t Empty =
    Branch (B, Branch (v, Empty, vb, vc, Empty), s, t, Empty)
  | balance (Branch (v, Empty, vb, vc, Branch (B, ve, vf, vg, vh))) s t Empty =
    Branch
      (B, Branch (v, Empty, vb, vc, Branch (B, ve, vf, vg, vh)), s, t, Empty)
  | balance (Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Empty)) s t Empty =
    Branch
      (B, Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Empty), s, t, Empty)
  | balance
    (Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Branch (B, ve, vj, vk, vl)))
    s t Empty =
    Branch
      (B, Branch
            (v, Branch (B, vf, vg, vh, vi), vb, vc, Branch (B, ve, vj, vk, vl)),
        s, t, Empty)
  | balance (Branch (v, Empty, vf, vg, Empty)) s t (Branch (B, va, vb, vc, vd))
    = Branch
        (B, Branch (v, Empty, vf, vg, Empty), s, t, Branch (B, va, vb, vc, vd))
  | balance (Branch (v, Empty, vf, vg, Branch (B, vi, vj, vk, vl))) s t
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (B, Branch (v, Empty, vf, vg, Branch (B, vi, vj, vk, vl)), s, t,
        Branch (B, va, vb, vc, vd))
  | balance (Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Empty)) s t
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (B, Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Empty), s, t,
        Branch (B, va, vb, vc, vd))
  | balance
    (Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Branch (B, vi, vn, vo, vp)))
    s t (Branch (B, va, vb, vc, vd)) =
    Branch
      (B, Branch
            (v, Branch (B, vj, vk, vl, vm), vf, vg, Branch (B, vi, vn, vo, vp)),
        s, t, Branch (B, va, vb, vc, vd));

fun balance_left (Branch (R, a, k, x, b)) s y c =
  Branch (R, Branch (B, a, k, x, b), s, y, c)
  | balance_left Empty k x (Branch (B, a, s, y, b)) =
    balance Empty k x (Branch (R, a, s, y, b))
  | balance_left (Branch (B, va, vb, vc, vd)) k x (Branch (B, a, s, y, b)) =
    balance (Branch (B, va, vb, vc, vd)) k x (Branch (R, a, s, y, b))
  | balance_left Empty k x (Branch (R, Branch (B, a, s, y, b), t, z, c)) =
    Branch (R, Branch (B, Empty, k, x, a), s, y, balance b t z (paint R c))
  | balance_left (Branch (B, va, vb, vc, vd)) k x
    (Branch (R, Branch (B, a, s, y, b), t, z, c)) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), k, x, a), s, y,
        balance b t z (paint R c))
  | balance_left Empty k x Empty = Empty
  | balance_left Empty k x (Branch (R, Empty, vb, vc, vd)) = Empty
  | balance_left Empty k x (Branch (R, Branch (R, ve, vf, vg, vh), vb, vc, vd))
    = Empty
  | balance_left (Branch (B, va, vb, vc, vd)) k x Empty = Empty
  | balance_left (Branch (B, va, vb, vc, vd)) k x
    (Branch (R, Empty, vf, vg, vh)) = Empty
  | balance_left (Branch (B, va, vb, vc, vd)) k x
    (Branch (R, Branch (R, vi, vj, vk, vl), vf, vg, vh)) = Empty;

fun combine Empty x = x
  | combine (Branch (v, va, vb, vc, vd)) Empty = Branch (v, va, vb, vc, vd)
  | combine (Branch (R, a, k, x, b)) (Branch (R, c, s, y, d)) =
    (case combine b c
      of Empty => Branch (R, a, k, x, Branch (R, Empty, s, y, d))
      | Branch (R, b2, t, z, c2) =>
        Branch (R, Branch (R, a, k, x, b2), t, z, Branch (R, c2, s, y, d))
      | Branch (B, b2, t, z, c2) =>
        Branch (R, a, k, x, Branch (R, Branch (B, b2, t, z, c2), s, y, d)))
  | combine (Branch (B, a, k, x, b)) (Branch (B, c, s, y, d)) =
    (case combine b c
      of Empty => balance_left a k x (Branch (B, Empty, s, y, d))
      | Branch (R, b2, t, z, c2) =>
        Branch (R, Branch (B, a, k, x, b2), t, z, Branch (B, c2, s, y, d))
      | Branch (B, b2, t, z, c2) =>
        balance_left a k x (Branch (B, Branch (B, b2, t, z, c2), s, y, d)))
  | combine (Branch (B, va, vb, vc, vd)) (Branch (R, b, k, x, c)) =
    Branch (R, combine (Branch (B, va, vb, vc, vd)) b, k, x, c)
  | combine (Branch (R, a, k, x, b)) (Branch (B, va, vb, vc, vd)) =
    Branch (R, a, k, x, combine b (Branch (B, va, vb, vc, vd)));

fun balance_right a k x (Branch (R, b, s, y, c)) =
  Branch (R, a, k, x, Branch (B, b, s, y, c))
  | balance_right (Branch (B, a, k, x, b)) s y Empty =
    balance (Branch (R, a, k, x, b)) s y Empty
  | balance_right (Branch (B, a, k, x, b)) s y (Branch (B, va, vb, vc, vd)) =
    balance (Branch (R, a, k, x, b)) s y (Branch (B, va, vb, vc, vd))
  | balance_right (Branch (R, a, k, x, Branch (B, b, s, y, c))) t z Empty =
    Branch (R, balance (paint R a) k x b, s, y, Branch (B, c, t, z, Empty))
  | balance_right (Branch (R, a, k, x, Branch (B, b, s, y, c))) t z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, balance (paint R a) k x b, s, y,
        Branch (B, c, t, z, Branch (B, va, vb, vc, vd)))
  | balance_right Empty k x Empty = Empty
  | balance_right (Branch (R, va, vb, vc, Empty)) k x Empty = Empty
  | balance_right (Branch (R, va, vb, vc, Branch (R, ve, vf, vg, vh))) k x Empty
    = Empty
  | balance_right Empty k x (Branch (B, va, vb, vc, vd)) = Empty
  | balance_right (Branch (R, ve, vf, vg, Empty)) k x
    (Branch (B, va, vb, vc, vd)) = Empty
  | balance_right (Branch (R, ve, vf, vg, Branch (R, vi, vj, vk, vl))) k x
    (Branch (B, va, vb, vc, vd)) = Empty;

fun rbt_del A_ x Empty = Empty
  | rbt_del A_ x (Branch (c, a, y, s, b)) =
    (if Orderings.less A_ x y then rbt_del_from_left A_ x a y s b
      else (if Orderings.less A_ y x then rbt_del_from_right A_ x a y s b
             else combine a b))
and rbt_del_from_left A_ x (Branch (B, lt, z, v, rt)) y s b =
  balance_left (rbt_del A_ x (Branch (B, lt, z, v, rt))) y s b
  | rbt_del_from_left A_ x Empty y s b = Branch (R, rbt_del A_ x Empty, y, s, b)
  | rbt_del_from_left A_ x (Branch (R, va, vb, vc, vd)) y s b =
    Branch (R, rbt_del A_ x (Branch (R, va, vb, vc, vd)), y, s, b)
and rbt_del_from_right A_ x a y s (Branch (B, lt, z, v, rt)) =
  balance_right a y s (rbt_del A_ x (Branch (B, lt, z, v, rt)))
  | rbt_del_from_right A_ x a y s Empty =
    Branch (R, a, y, s, rbt_del A_ x Empty)
  | rbt_del_from_right A_ x a y s (Branch (R, va, vb, vc, vd)) =
    Branch (R, a, y, s, rbt_del A_ x (Branch (R, va, vb, vc, vd)));

fun rbt_ins A_ f k v Empty = Branch (R, Empty, k, v, Empty)
  | rbt_ins A_ f k v (Branch (B, l, x, y, r)) =
    (if Orderings.less A_ k x then balance (rbt_ins A_ f k v l) x y r
      else (if Orderings.less A_ x k then balance l x y (rbt_ins A_ f k v r)
             else Branch (B, l, x, f k y v, r)))
  | rbt_ins A_ f k v (Branch (R, l, x, y, r)) =
    (if Orderings.less A_ k x then Branch (R, rbt_ins A_ f k v l, x, y, r)
      else (if Orderings.less A_ x k
             then Branch (R, l, x, y, rbt_ins A_ f k v r)
             else Branch (R, l, x, f k y v, r)));

fun equal_color B R = false
  | equal_color R B = false
  | equal_color B B = true
  | equal_color R R = true;

fun equal_rbt A_ B_ (Branch (color, rbt1, a, b, rbt2)) Empty = false
  | equal_rbt A_ B_ Empty (Branch (color, rbt1, a, b, rbt2)) = false
  | equal_rbt A_ B_ (Branch (colora, rbt1a, aa, ba, rbt2a))
    (Branch (color, rbt1, a, b, rbt2)) =
    equal_color colora color andalso
      (equal_rbt A_ B_ rbt1a rbt1 andalso
        (HOL.eq A_ aa a andalso
          (HOL.eq B_ ba b andalso equal_rbt A_ B_ rbt2a rbt2)))
  | equal_rbt A_ B_ Empty Empty = true;

fun rbt_delete A_ k t = paint B (rbt_del A_ k t);

fun rbt_insert_with_key A_ f k v t = paint B (rbt_ins A_ f k v t);

fun rbt_insert A_ = rbt_insert_with_key A_ (fn _ => fn _ => fn nv => nv);

fun rbt_lookup A_ Empty k = NONE
  | rbt_lookup A_ (Branch (uu, l, x, y, r)) k =
    (if Orderings.less A_ k x then rbt_lookup A_ l k
      else (if Orderings.less A_ x k then rbt_lookup A_ r k else SOME y));

end; (*struct RBT_Impl*)

structure RBT : sig
  datatype ('a, 'b) rbt = Rbt of ('a, 'b) RBT_Impl.rbt
  val empty : 'a Orderings.linorder -> ('a, 'b) rbt
  val impl_of : 'a Orderings.linorder -> ('a, 'b) rbt -> ('a, 'b) RBT_Impl.rbt
  val delete : 'a Orderings.linorder -> 'a -> ('a, 'b) rbt -> ('a, 'b) rbt
  val insert : 'a Orderings.linorder -> 'a -> 'b -> ('a, 'b) rbt -> ('a, 'b) rbt
  val lookup : 'a Orderings.linorder -> ('a, 'b) rbt -> 'a -> 'b option
end = struct

datatype ('a, 'b) rbt = Rbt of ('a, 'b) RBT_Impl.rbt;

fun empty A_ = Rbt RBT_Impl.Empty;

fun impl_of A_ (Rbt x) = x;

fun delete A_ x xa =
  Rbt (RBT_Impl.rbt_delete
        ((Orderings.ord_preorder o Orderings.preorder_order o
           Orderings.order_linorder)
          A_)
        x (impl_of A_ xa));

fun insert A_ x xa xb =
  Rbt (RBT_Impl.rbt_insert
        ((Orderings.ord_preorder o Orderings.preorder_order o
           Orderings.order_linorder)
          A_)
        x xa (impl_of A_ xb));

fun lookup A_ x =
  RBT_Impl.rbt_lookup
    ((Orderings.ord_preorder o Orderings.preorder_order o
       Orderings.order_linorder)
      A_)
    (impl_of A_ x);

end; (*struct RBT*)

structure Transitive_Closure_Table : sig
  val rtrancl_tab_FioB_i_i_i :
    'a HOL.equal ->
      ('a -> 'a Predicate.pred) -> 'a list -> 'a -> 'a -> unit Predicate.pred
  val rtranclp_FioB_i_i :
    'a HOL.equal -> ('a -> 'a Predicate.pred) -> 'a -> 'a -> unit Predicate.pred
end = struct

fun rtrancl_tab_FioB_i_i_i A_ r xa xb xc =
  Predicate.sup_pred
    (if HOL.eq A_ xb xc then Predicate.single () else Predicate.bot_pred)
    (Predicate.bind (Predicate.if_pred (not (List.member A_ xa xb)))
      (fn () =>
        Predicate.bind (r xb)
          (fn xaa =>
            Predicate.bind (rtrancl_tab_FioB_i_i_i A_ r (xb :: xa) xaa xc)
              (fn () => Predicate.single ()))));

fun rtranclp_FioB_i_i A_ r xa xb =
  Predicate.bind (rtrancl_tab_FioB_i_i_i A_ r [] xa xb)
    (fn () => Predicate.single ());

end; (*struct Transitive_Closure_Table*)

structure FWState : sig
  val shr : 'a * (('b * 'c) * ('d * 'e)) -> 'c
  val thr : 'a * (('b * 'c) * ('d * 'e)) -> 'b
  datatype wake_up_status = WSNotified | WSWokenUp
  datatype 'a wait_set_status = InWS of 'a | PostWS of wake_up_status
  datatype 'a conditional_action = Join of 'a | Yield
  datatype lock_action = Lock | Unlock | UnlockFail | ReleaseAcquire
  val wset : 'a * (('b * 'c) * ('d * 'e)) -> 'd
  val locks : 'a * (('b * 'c) * ('d * 'e)) -> 'a
  datatype ('a, 'b, 'c) new_thread_action = NewThread of 'a * 'b * 'c |
    ThreadExists of 'a * bool
  datatype 'a interrupt_action = IsInterrupted of 'a * bool | Interrupt of 'a |
    ClearInterrupt of 'a
  datatype ('a, 'b) wait_set_action = Suspend of 'b | Notify of 'b |
    NotifyAll of 'b | WakeUp of 'a | Notified | WokenUp
  val obs_a :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      'f list
  val thr_a :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      ('b, 'c, 'd) new_thread_action list
  val cond_a :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      'b conditional_action list
  val wset_a :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      ('b, 'e) wait_set_action list
  val locks_a :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      ('a, (lock_action list)) FinFun.finfun
  datatype ('a, 'b, 'c, 'd, 'e, 'f) thread_action =
    LockAction of (lock_action * 'a) |
    NewThreadAction of ('b, 'c, 'd) new_thread_action |
    ConditionalAction of 'b conditional_action |
    WaitSetAction of ('b, 'e) wait_set_action |
    InterruptAction of 'b interrupt_action | ObsAction of 'f
  val interrupts : 'a * (('b * 'c) * ('d * 'e)) -> 'e
  type 'a obs_action
  val interrupt_a :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      'b interrupt_action list
  val convert_new_thread_action :
    ('a -> 'b) ->
      ('c, 'a, 'd) new_thread_action -> ('c, 'b, 'd) new_thread_action
  val convert_extTA :
    ('a -> 'b) ->
      ('c, (lock_action list)) FinFun.finfun *
        (('d, 'a, 'e) new_thread_action list *
          ('d conditional_action list *
            (('d, 'f) wait_set_action list *
              ('d interrupt_action list * 'g list)))) ->
        ('c, (lock_action list)) FinFun.finfun *
          (('d, 'b, 'e) new_thread_action list *
            ('d conditional_action list *
              (('d, 'f) wait_set_action list *
                ('d interrupt_action list * 'g list))))
  val ta_update_obs :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      'f -> ('a, (lock_action list)) FinFun.finfun *
              (('b, 'c, 'd) new_thread_action list *
                ('b conditional_action list *
                  (('b, 'e) wait_set_action list *
                    ('b interrupt_action list * 'f list))))
  val equal_lock_actiona : lock_action -> lock_action -> bool
  val equal_lock_action : lock_action HOL.equal
  val ta_update_locks :
    'a HOL.equal ->
      ('a, (lock_action list)) FinFun.finfun *
        (('b, 'c, 'd) new_thread_action list *
          ('b conditional_action list *
            (('b, 'e) wait_set_action list *
              ('b interrupt_action list * 'f list)))) ->
        lock_action ->
          'a -> ('a, (lock_action list)) FinFun.finfun *
                  (('b, 'c, 'd) new_thread_action list *
                    ('b conditional_action list *
                      (('b, 'e) wait_set_action list *
                        ('b interrupt_action list * 'f list))))
  val ta_update_wait_set :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      ('b, 'e) wait_set_action ->
        ('a, (lock_action list)) FinFun.finfun *
          (('b, 'c, 'd) new_thread_action list *
            ('b conditional_action list *
              (('b, 'e) wait_set_action list *
                ('b interrupt_action list * 'f list))))
  val ta_update_NewThread :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      ('b, 'c, 'd) new_thread_action ->
        ('a, (lock_action list)) FinFun.finfun *
          (('b, 'c, 'd) new_thread_action list *
            ('b conditional_action list *
              (('b, 'e) wait_set_action list *
                ('b interrupt_action list * 'f list))))
  val ta_update_interrupt :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      'b interrupt_action ->
        ('a, (lock_action list)) FinFun.finfun *
          (('b, 'c, 'd) new_thread_action list *
            ('b conditional_action list *
              (('b, 'e) wait_set_action list *
                ('b interrupt_action list * 'f list))))
  val equal_wait_set_actiona :
    'a HOL.equal -> 'b HOL.equal ->
      ('a, 'b) wait_set_action -> ('a, 'b) wait_set_action -> bool
  val equal_wait_set_action :
    'a HOL.equal -> 'b HOL.equal -> ('a, 'b) wait_set_action HOL.equal
  val equal_wake_up_status : wake_up_status -> wake_up_status -> bool
  val equal_wait_set_statusa :
    'a HOL.equal -> 'a wait_set_status -> 'a wait_set_status -> bool
  val equal_wait_set_status : 'a HOL.equal -> 'a wait_set_status HOL.equal
  val ta_update_Conditional :
    ('a, (lock_action list)) FinFun.finfun *
      (('b, 'c, 'd) new_thread_action list *
        ('b conditional_action list *
          (('b, 'e) wait_set_action list *
            ('b interrupt_action list * 'f list)))) ->
      'b conditional_action ->
        ('a, (lock_action list)) FinFun.finfun *
          (('b, 'c, 'd) new_thread_action list *
            ('b conditional_action list *
              (('b, 'e) wait_set_action list *
                ('b interrupt_action list * 'f list))))
  val equal_conditional_actiona :
    'a HOL.equal -> 'a conditional_action -> 'a conditional_action -> bool
  val equal_conditional_action : 'a HOL.equal -> 'a conditional_action HOL.equal
  val thread_action_to_thread_action :
    'a HOL.equal -> 'f obs_action ->
      ('a, 'b, 'c, 'd, 'e, 'f) thread_action ->
        ('a, (lock_action list)) FinFun.finfun *
          (('b, 'c, 'd) new_thread_action list *
            ('b conditional_action list *
              (('b, 'e) wait_set_action list *
                ('b interrupt_action list * 'f list)))) ->
          ('a, (lock_action list)) FinFun.finfun *
            (('b, 'c, 'd) new_thread_action list *
              ('b conditional_action list *
                (('b, 'e) wait_set_action list *
                  ('b interrupt_action list * 'f list))))
end = struct

fun shr lstsmws =
  Product_Type.snd (Product_Type.fst (Product_Type.snd lstsmws));

fun thr lstsmws =
  Product_Type.fst (Product_Type.fst (Product_Type.snd lstsmws));

datatype wake_up_status = WSNotified | WSWokenUp;

datatype 'a wait_set_status = InWS of 'a | PostWS of wake_up_status;

datatype 'a conditional_action = Join of 'a | Yield;

datatype lock_action = Lock | Unlock | UnlockFail | ReleaseAcquire;

fun wset lstsmws =
  Product_Type.fst (Product_Type.snd (Product_Type.snd lstsmws));

fun locks lstsmws = Product_Type.fst lstsmws;

datatype ('a, 'b, 'c) new_thread_action = NewThread of 'a * 'b * 'c |
  ThreadExists of 'a * bool;

datatype 'a interrupt_action = IsInterrupted of 'a * bool | Interrupt of 'a |
  ClearInterrupt of 'a;

datatype ('a, 'b) wait_set_action = Suspend of 'b | Notify of 'b |
  NotifyAll of 'b | WakeUp of 'a | Notified | WokenUp;

fun obs_a x =
  (Product_Type.snd o Product_Type.snd o Product_Type.snd o Product_Type.snd o
    Product_Type.snd)
    x;

fun thr_a x = (Product_Type.fst o Product_Type.snd) x;

fun cond_a x = (Product_Type.fst o Product_Type.snd o Product_Type.snd) x;

fun wset_a x =
  (Product_Type.fst o Product_Type.snd o Product_Type.snd o Product_Type.snd) x;

fun locks_a x = Product_Type.fst x;

datatype ('a, 'b, 'c, 'd, 'e, 'f) thread_action =
  LockAction of (lock_action * 'a) |
  NewThreadAction of ('b, 'c, 'd) new_thread_action |
  ConditionalAction of 'b conditional_action |
  WaitSetAction of ('b, 'e) wait_set_action |
  InterruptAction of 'b interrupt_action | ObsAction of 'f;

fun interrupts lstsmws =
  Product_Type.snd (Product_Type.snd (Product_Type.snd lstsmws));

type 'a obs_action = {};

fun interrupt_a x =
  (Product_Type.fst o Product_Type.snd o Product_Type.snd o Product_Type.snd o
    Product_Type.snd)
    x;

fun convert_new_thread_action f (NewThread (t, x, m)) = NewThread (t, f x, m)
  | convert_new_thread_action f (ThreadExists (t, b)) = ThreadExists (t, b);

fun convert_extTA f ta =
  (locks_a ta,
    (List.map (convert_new_thread_action f) (thr_a ta),
      Product_Type.snd (Product_Type.snd ta)));

fun ta_update_obs (ls, (nts, (js, (wss, (is, obs))))) ob =
  (ls, (nts, (js, (wss, (is, obs @ [ob])))));

fun equal_lock_actiona ReleaseAcquire UnlockFail = false
  | equal_lock_actiona UnlockFail ReleaseAcquire = false
  | equal_lock_actiona ReleaseAcquire Unlock = false
  | equal_lock_actiona Unlock ReleaseAcquire = false
  | equal_lock_actiona UnlockFail Unlock = false
  | equal_lock_actiona Unlock UnlockFail = false
  | equal_lock_actiona ReleaseAcquire Lock = false
  | equal_lock_actiona Lock ReleaseAcquire = false
  | equal_lock_actiona UnlockFail Lock = false
  | equal_lock_actiona Lock UnlockFail = false
  | equal_lock_actiona Unlock Lock = false
  | equal_lock_actiona Lock Unlock = false
  | equal_lock_actiona ReleaseAcquire ReleaseAcquire = true
  | equal_lock_actiona UnlockFail UnlockFail = true
  | equal_lock_actiona Unlock Unlock = true
  | equal_lock_actiona Lock Lock = true;

val equal_lock_action = {equal = equal_lock_actiona} : lock_action HOL.equal;

fun ta_update_locks A_ (ls, (nts, (js, (wss, obs)))) lta l =
  (FinFun.finfun_update A_ (List.equal_list equal_lock_action) ls l
     (FinFun.finfun_apply A_ ls l @ [lta]),
    (nts, (js, (wss, obs))));

fun ta_update_wait_set (ls, (nts, (js, (wss, (is, obs))))) ws =
  (ls, (nts, (js, (wss @ [ws], (is, obs)))));

fun ta_update_NewThread (ls, (nts, (js, (wss, (is, obs))))) nt =
  (ls, (nts @ [nt], (js, (wss, (is, obs)))));

fun ta_update_interrupt (ls, (nts, (js, (wss, (is, obs))))) i =
  (ls, (nts, (js, (wss, (is @ [i], obs)))));

fun equal_wait_set_actiona A_ B_ WokenUp Notified = false
  | equal_wait_set_actiona A_ B_ Notified WokenUp = false
  | equal_wait_set_actiona A_ B_ WokenUp (WakeUp t) = false
  | equal_wait_set_actiona A_ B_ (WakeUp t) WokenUp = false
  | equal_wait_set_actiona A_ B_ Notified (WakeUp t) = false
  | equal_wait_set_actiona A_ B_ (WakeUp t) Notified = false
  | equal_wait_set_actiona A_ B_ WokenUp (NotifyAll w) = false
  | equal_wait_set_actiona A_ B_ (NotifyAll w) WokenUp = false
  | equal_wait_set_actiona A_ B_ Notified (NotifyAll w) = false
  | equal_wait_set_actiona A_ B_ (NotifyAll w) Notified = false
  | equal_wait_set_actiona A_ B_ (WakeUp t) (NotifyAll w) = false
  | equal_wait_set_actiona A_ B_ (NotifyAll w) (WakeUp t) = false
  | equal_wait_set_actiona A_ B_ WokenUp (Notify w) = false
  | equal_wait_set_actiona A_ B_ (Notify w) WokenUp = false
  | equal_wait_set_actiona A_ B_ Notified (Notify w) = false
  | equal_wait_set_actiona A_ B_ (Notify w) Notified = false
  | equal_wait_set_actiona A_ B_ (WakeUp t) (Notify w) = false
  | equal_wait_set_actiona A_ B_ (Notify w) (WakeUp t) = false
  | equal_wait_set_actiona A_ B_ (NotifyAll wa) (Notify w) = false
  | equal_wait_set_actiona A_ B_ (Notify wa) (NotifyAll w) = false
  | equal_wait_set_actiona A_ B_ WokenUp (Suspend w) = false
  | equal_wait_set_actiona A_ B_ (Suspend w) WokenUp = false
  | equal_wait_set_actiona A_ B_ Notified (Suspend w) = false
  | equal_wait_set_actiona A_ B_ (Suspend w) Notified = false
  | equal_wait_set_actiona A_ B_ (WakeUp t) (Suspend w) = false
  | equal_wait_set_actiona A_ B_ (Suspend w) (WakeUp t) = false
  | equal_wait_set_actiona A_ B_ (NotifyAll wa) (Suspend w) = false
  | equal_wait_set_actiona A_ B_ (Suspend wa) (NotifyAll w) = false
  | equal_wait_set_actiona A_ B_ (Notify wa) (Suspend w) = false
  | equal_wait_set_actiona A_ B_ (Suspend wa) (Notify w) = false
  | equal_wait_set_actiona A_ B_ (WakeUp ta) (WakeUp t) = HOL.eq A_ ta t
  | equal_wait_set_actiona A_ B_ (NotifyAll wa) (NotifyAll w) = HOL.eq B_ wa w
  | equal_wait_set_actiona A_ B_ (Notify wa) (Notify w) = HOL.eq B_ wa w
  | equal_wait_set_actiona A_ B_ (Suspend wa) (Suspend w) = HOL.eq B_ wa w
  | equal_wait_set_actiona A_ B_ WokenUp WokenUp = true
  | equal_wait_set_actiona A_ B_ Notified Notified = true;

fun equal_wait_set_action A_ B_ = {equal = equal_wait_set_actiona A_ B_} :
  ('a, 'b) wait_set_action HOL.equal;

fun equal_wake_up_status WSWokenUp WSNotified = false
  | equal_wake_up_status WSNotified WSWokenUp = false
  | equal_wake_up_status WSWokenUp WSWokenUp = true
  | equal_wake_up_status WSNotified WSNotified = true;

fun equal_wait_set_statusa A_ (PostWS wake_up_status) (InWS w) = false
  | equal_wait_set_statusa A_ (InWS w) (PostWS wake_up_status) = false
  | equal_wait_set_statusa A_ (PostWS wake_up_statusa) (PostWS wake_up_status) =
    equal_wake_up_status wake_up_statusa wake_up_status
  | equal_wait_set_statusa A_ (InWS wa) (InWS w) = HOL.eq A_ wa w;

fun equal_wait_set_status A_ = {equal = equal_wait_set_statusa A_} :
  'a wait_set_status HOL.equal;

fun ta_update_Conditional (ls, (nts, (js, (wss, (is, obs))))) j =
  (ls, (nts, (js @ [j], (wss, (is, obs)))));

fun equal_conditional_actiona A_ Yield (Join t) = false
  | equal_conditional_actiona A_ (Join t) Yield = false
  | equal_conditional_actiona A_ (Join ta) (Join t) = HOL.eq A_ ta t
  | equal_conditional_actiona A_ Yield Yield = true;

fun equal_conditional_action A_ = {equal = equal_conditional_actiona A_} :
  'a conditional_action HOL.equal;

fun thread_action_to_thread_action A_ F_ (LockAction (la, l)) ta =
  ta_update_locks A_ ta la l
  | thread_action_to_thread_action A_ F_ (NewThreadAction nt) ta =
    ta_update_NewThread ta nt
  | thread_action_to_thread_action A_ F_ (ConditionalAction ca) ta =
    ta_update_Conditional ta ca
  | thread_action_to_thread_action A_ F_ (WaitSetAction wa) ta =
    ta_update_wait_set ta wa
  | thread_action_to_thread_action A_ F_ (InterruptAction ia) ta =
    ta_update_interrupt ta ia
  | thread_action_to_thread_action A_ F_ (ObsAction ob) ta =
    ta_update_obs ta ob;

end; (*struct FWState*)

structure Heap : sig
  datatype addr_loc = CField of string * string | ACell of Arith.nat
  val typeof : 'a Value.vala -> Type.ty option
  val typeof_h :
    'b Value.addr ->
      ('a -> 'b -> Type.htype option) -> 'a -> 'b Value.vala -> Type.ty option
end = struct

datatype addr_loc = CField of string * string | ACell of Arith.nat;

fun typeof Value.Unit = SOME Type.Void
  | typeof Value.Null = SOME Type.Nt
  | typeof (Value.Bool b) = SOME Type.Boolean
  | typeof (Value.Intg i) = SOME Type.Integer
  | typeof (Value.Addr a) = NONE;

fun typeof_h B_ typeof_addr h (Value.Intg va) = typeof (Value.Intg va)
  | typeof_h B_ typeof_addr h (Value.Bool va) = typeof (Value.Bool va)
  | typeof_h B_ typeof_addr h Value.Null = typeof Value.Null
  | typeof_h B_ typeof_addr h Value.Unit = typeof Value.Unit
  | typeof_h B_ typeof_addr h (Value.Addr a) =
    Option.map Type.ty_of_htype (typeof_addr h a);

end; (*struct Heap*)

structure Observable_Events : sig
  datatype ('a, 'b) obs_event =
    ExternalCall of 'a * string * 'a Value.vala list * 'a Value.vala |
    ReadMem of 'a * Heap.addr_loc * 'a Value.vala |
    WriteMem of 'a * Heap.addr_loc * 'a Value.vala |
    NewHeapElem of 'a * Type.htype | ThreadStart of 'b | ThreadJoin of 'b |
    SyncLock of 'a | SyncUnlock of 'a | ObsInterrupt of 'b |
    ObsInterrupted of 'b
  val convert_RA :
    'a HOL.equal * 'a Value.addr ->
      ('a, Arith.nat) FinFun.finfun -> ('a, 'b) obs_event list
  val obs_action_obs_event : ('a, 'b) obs_event FWState.obs_action
end = struct

datatype ('a, 'b) obs_event =
  ExternalCall of 'a * string * 'a Value.vala list * 'a Value.vala |
  ReadMem of 'a * Heap.addr_loc * 'a Value.vala |
  WriteMem of 'a * Heap.addr_loc * 'a Value.vala |
  NewHeapElem of 'a * Type.htype | ThreadStart of 'b | ThreadJoin of 'b |
  SyncLock of 'a | SyncUnlock of 'a | ObsInterrupt of 'b | ObsInterrupted of 'b;

fun convert_RA (A1_, A2_) ln =
  List.maps
    (fn ad => List.replicate (FinFun.finfun_apply A1_ ln ad) (SyncLock ad))
    (Value.monitor_finfun_to_list A2_ ln);

val obs_action_obs_event = {} : ('a, 'b) obs_event FWState.obs_action;

end; (*struct Observable_Events*)

structure TypeRel : sig
  datatype 'a fmod_ext = Fmod_ext of bool * 'a
  datatype 'a prog =
    Program of
      (string *
        (string *
          ((string * (Type.ty * unit fmod_ext)) list *
            (string * (Type.ty list * (Type.ty * 'a option))) list))) list
  val class :
    'a prog ->
      string ->
        (string *
          ((string * (Type.ty * unit fmod_ext)) list *
            (string * (Type.ty list * (Type.ty * 'a option))) list)) option
  val eq_i_i : 'a HOL.equal -> 'a -> 'a -> unit Predicate.pred
  val classes :
    'a prog ->
      (string *
        (string *
          ((string * (Type.ty * unit fmod_ext)) list *
            (string * (Type.ty list * (Type.ty * 'a option))) list))) list
  val is_class : 'a prog -> string -> bool
  val is_type : 'a prog -> Type.ty -> bool
  val subcls1_i_i_o : 'a prog -> string -> string Predicate.pred
  val widen_i_i_ia : 'a prog -> Type.ty -> Type.ty -> unit Predicate.pred
  val widen : 'a prog -> Type.ty -> Type.ty -> bool
  val fields_i_i_o :
    'a prog ->
      string ->
        (((string * string) * (Type.ty * unit fmod_ext)) list) Predicate.pred
  val equal_fmod_exta : 'a HOL.equal -> 'a fmod_ext -> 'a fmod_ext -> bool
  val equal_fmod_ext : 'a HOL.equal -> 'a fmod_ext HOL.equal
  val fields :
    'a prog -> string -> ((string * string) * (Type.ty * unit fmod_ext)) list
  val methods_i_i_o :
    'a prog ->
      string ->
        (string -> ((Type.ty list * (Type.ty * 'a option)) * string) option)
          Predicate.pred
  val method_i_i_i_o_o_o_o :
    'a prog ->
      string ->
        string ->
          (Type.ty list * (Type.ty * ('a option * string))) Predicate.pred
  val method :
    'a HOL.equal ->
      'a prog ->
        string -> string -> string * (Type.ty list * (Type.ty * 'a option))
  val subcls_i_i_i : 'a prog -> string -> string -> unit Predicate.pred
  val subcls : 'a prog -> string -> string -> bool
  val is_class_i_i : 'a prog -> string -> unit Predicate.pred
  val widen_i_i_i : 'a prog -> Type.ty -> Type.ty -> unit Predicate.pred
  val mapP_FioB_i_o :
    ('a -> 'b Predicate.pred) -> 'a list -> ('b list) Predicate.pred
  val sees_field_hoaux_PPiiii_Poooo :
    (string * string) * (Type.ty * unit fmod_ext) ->
      (string * (string * (Type.ty * unit fmod_ext))) Predicate.pred
  val sees_field_i_i_i_o_o_i :
    'a prog ->
      string -> string -> string -> (Type.ty * unit fmod_ext) Predicate.pred
end = struct

datatype 'a fmod_ext = Fmod_ext of bool * 'a;

datatype 'a prog =
  Program of
    (string *
      (string *
        ((string * (Type.ty * unit fmod_ext)) list *
          (string * (Type.ty list * (Type.ty * 'a option))) list))) list;

fun class (Program p) = Map.map_of Stringa.equal_literal p;

fun eq_i_i A_ xa xb =
  (if HOL.eq A_ xa xb then Predicate.single () else Predicate.bot_pred);

fun classes (Program p) = p;

fun is_class p c = not (Option.is_none (class p c));

fun is_type p Type.Void = true
  | is_type p Type.Boolean = true
  | is_type p Type.Integer = true
  | is_type p Type.Nt = true
  | is_type p (Type.Class c) = is_class p c
  | is_type p (Type.Array a) =
    (case Type.ground_type a of Type.Void => true | Type.Boolean => true
      | Type.Integer => true | Type.Nt => false | Type.Class aa => is_class p aa
      | Type.Array _ => true);

fun subcls1_i_i_o x xa =
  Predicate.bind (Predicate.if_pred (not ((xa : string) = Type.object)))
    (fn () =>
      (case class x xa of NONE => Predicate.bot_pred
        | SOME (d, _) => Predicate.single d));

fun widen_i_i_ia x xa xb =
  Predicate.sup_pred
    (if Type.equal_tya xa xb then Predicate.single () else Predicate.bot_pred)
    (case xa of Type.Void => Predicate.bot_pred
      | Type.Boolean => Predicate.bot_pred | Type.Integer => Predicate.bot_pred
      | Type.Nt =>
        (case xb of Type.Void => Predicate.bot_pred
          | Type.Boolean => Predicate.bot_pred
          | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
          | Type.Class _ => Predicate.single ()
          | Type.Array _ => Predicate.single ())
      | Type.Class c =>
        (case xb of Type.Void => Predicate.bot_pred
          | Type.Boolean => Predicate.bot_pred
          | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
          | Type.Class d =>
            Predicate.bind
              (Transitive_Closure_Table.rtranclp_FioB_i_i Stringa.equal_literal
                (subcls1_i_i_o x) c d)
              (fn () => Predicate.single ())
          | Type.Array _ => Predicate.bot_pred)
      | Type.Array t =>
        (case xb of Type.Void => Predicate.bot_pred
          | Type.Boolean => Predicate.bot_pred
          | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
          | Type.Class c =>
            Predicate.bind (eq_i_i Stringa.equal_literal c Type.object)
              (fn () => Predicate.single ())
          | Type.Array ta =>
            Predicate.bind (widen_i_i_ia x t ta)
              (fn () => Predicate.single ())));

fun widen x1 x2 x3 = Predicate.holds (widen_i_i_ia x1 x2 x3);

fun fields_i_i_o x xa =
  Predicate.sup_pred
    (Predicate.bind (Predicate.if_pred (not ((xa : string) = Type.object)))
      (fn () =>
        (case class x xa of NONE => Predicate.bot_pred
          | SOME (d, (fs, _)) =>
            Predicate.bind (fields_i_i_o x d)
              (fn xb =>
                Predicate.single
                  (List.map (fn (f, a) => ((f, xa), a)) fs @ xb)))))
    (case class x Type.object of NONE => Predicate.bot_pred
      | SOME (_, (fs, _)) =>
        (if ((xa : string) = Type.object)
          then Predicate.single
                 (List.map (fn (f, a) => ((f, Type.object), a)) fs)
          else Predicate.bot_pred));

fun equal_fmod_exta A_ (Fmod_ext (volatilea, morea)) (Fmod_ext (volatile, more))
  = Product_Type.equal_boola volatilea volatile andalso HOL.eq A_ morea more;

fun equal_fmod_ext A_ = {equal = equal_fmod_exta A_} : 'a fmod_ext HOL.equal;

fun fields p c =
  Predicate.the
    (List.equal_list
      (Product_Type.equal_prod
        (Product_Type.equal_prod Stringa.equal_literal Stringa.equal_literal)
        (Product_Type.equal_prod Type.equal_ty
          (equal_fmod_ext Product_Type.equal_unit))))
    (fields_i_i_o p c);

fun methods_i_i_o x xa =
  Predicate.sup_pred
    (case class x Type.object of NONE => Predicate.bot_pred
      | SOME (_, (_, ms)) =>
        (if ((xa : string) = Type.object)
          then Predicate.single
                 (Option.map (fn m => (m, Type.object)) o
                   Map.map_of Stringa.equal_literal ms)
          else Predicate.bot_pred))
    (Predicate.bind (Predicate.if_pred (not ((xa : string) = Type.object)))
      (fn () =>
        (case class x xa of NONE => Predicate.bot_pred
          | SOME (d, (_, ms)) =>
            Predicate.bind (methods_i_i_o x d)
              (fn xb =>
                Predicate.single
                  (Map.map_add xb
                    (Option.map (fn m => (m, xa)) o
                      Map.map_of Stringa.equal_literal ms))))));

fun method_i_i_i_o_o_o_o x xa xb =
  Predicate.bind (methods_i_i_o x xa)
    (fn xc =>
      (case xc xb of NONE => Predicate.bot_pred
        | SOME ((tv, (tc, p)), dc) => Predicate.single (tv, (tc, (p, dc)))));

fun method A_ p c m =
  Predicate.the
    (Product_Type.equal_prod Stringa.equal_literal
      (Product_Type.equal_prod (List.equal_list Type.equal_ty)
        (Product_Type.equal_prod Type.equal_ty (Option.equal_option A_))))
    (Predicate.bind (method_i_i_i_o_o_o_o p c m)
      (fn (ts, (t, (ma, d))) => Predicate.single (d, (ts, (t, ma)))));

fun subcls_i_i_i x xa xb =
  Predicate.bind
    (Transitive_Closure_Table.rtranclp_FioB_i_i Stringa.equal_literal
      (subcls1_i_i_o x) xa xb)
    (fn () => Predicate.single ());

fun subcls x1 x2 x3 = Predicate.holds (subcls_i_i_i x1 x2 x3);

fun is_class_i_i x xa =
  Predicate.bind (Predicate.if_pred (not (Option.is_none (class x xa))))
    (fn () => Predicate.single ());

fun widen_i_i_i p ta t =
  (if widen p ta t then Predicate.single () else Predicate.bot_pred);

fun mapP_FioB_i_o h x =
  Predicate.sup_pred
    (case x of [] => Predicate.single [] | _ :: _ => Predicate.bot_pred)
    (case x of [] => Predicate.bot_pred
      | z :: xu =>
        Predicate.bind (mapP_FioB_i_o h xu)
          (fn xa =>
            Predicate.bind (h z) (fn xaa => Predicate.single (xaa :: xa))));

fun sees_field_hoaux_PPiiii_Poooo ((x, xf), (xg, xh)) =
  Predicate.single (x, (xf, (xg, xh)));

fun sees_field_i_i_i_o_o_i x xa xb xc =
  Predicate.bind (fields_i_i_o x xa)
    (fn xd =>
      Predicate.bind (mapP_FioB_i_o sees_field_hoaux_PPiiii_Poooo xd)
        (fn xe =>
          (case Map.map_of Stringa.equal_literal xe xb
            of NONE => Predicate.bot_pred
            | SOME (dca, (tc, fp)) =>
              (if ((xc : string) = dca) then Predicate.single (tc, fp)
                else Predicate.bot_pred))));

end; (*struct TypeRel*)

structure Exceptions : sig
  val sys_xcpts_list : string list
  val wf_syscls : 'a TypeRel.prog -> bool
end = struct

val sys_xcpts_list : string list =
  ["java/lang/NullPointerException", "java/lang/ClassCastException",
    "java/lang/OutOfMemoryError", "java/lang/ArrayIndexOutOfBoundsException",
    "java/lang/ArrayStoreException", "java/lang/NegativeArraySizeException",
    "java/lang/ArithmeticException", "java/lang/IllegalMonitorStateException",
    "java/lang/IllegalThreadStateException", "java/lang/InterruptedException"];

fun wf_syscls p =
  List.list_all (TypeRel.is_class p)
    [Type.object, Type.throwable, Type.thread] andalso
    List.list_all (fn c => TypeRel.subcls p c Type.throwable) sys_xcpts_list;

end; (*struct Exceptions*)

structure StartConfig : sig
  val pick_addr : ('a * 'b) Set.set -> 'a * 'b
  val initialization_list : string list
  val start_heap_data :
    'a HOL.equal -> 'b HOL.equal * 'b Value.addr ->
      'a -> ('a -> Type.htype -> ('a * 'b) Set.set) -> 'a * ('b list * bool)
  val start_addrs :
    'a HOL.equal -> 'b HOL.equal * 'b Value.addr ->
      'a -> ('a -> Type.htype -> ('a * 'b) Set.set) -> 'b list
  val start_tid :
    'a HOL.equal * 'a Value.addr -> 'c HOL.equal ->
      ('a -> 'b) -> 'c -> ('c -> Type.htype -> ('c * 'a) Set.set) -> 'b
  val start_heap :
    'a HOL.equal -> 'b HOL.equal * 'b Value.addr ->
      'a -> ('a -> Type.htype -> ('a * 'b) Set.set) -> 'a
  val addr_of_sys_xcpt :
    'a HOL.equal -> 'b HOL.equal * 'b Value.addr ->
      'a -> ('a -> Type.htype -> ('a * 'b) Set.set) -> string -> 'b
end = struct

fun pick_addr (Set.Set [ha]) = ha;

val initialization_list : string list =
  Type.thread :: Exceptions.sys_xcpts_list;

fun start_heap_data A_ (B1_, B2_) empty_heap allocate =
  let
    val (h, (ads, b)) =
      List.foldl
        (fn (h, (ads, b)) => fn c =>
          (if b then let
                       val ha = allocate h (Type.Class_type c);
                     in
                       (if Set.equal_seta (Product_Type.equal_prod A_ B1_) ha
                             Set.bot_set
                         then (h, (ads, false))
                         else let
                                val (hb, a) = pick_addr ha;
                              in
                                (hb, (a :: ads, true))
                              end)
                     end
            else (h, (ads, false))))
        (empty_heap, ([], true)) initialization_list;
  in
    (h, (List.rev ads, b))
  end;

fun start_addrs A_ (B1_, B2_) empty_heap allocate =
  Product_Type.fst
    (Product_Type.snd (start_heap_data A_ (B1_, B2_) empty_heap allocate));

fun start_tid (A1_, A2_) C_ addr2thread_id empty_heap allocate =
  addr2thread_id (List.hd (start_addrs C_ (A1_, A2_) empty_heap allocate));

fun start_heap A_ (B1_, B2_) empty_heap allocate =
  Product_Type.fst (start_heap_data A_ (B1_, B2_) empty_heap allocate);

fun addr_of_sys_xcpt A_ (B1_, B2_) empty_heap allocate c =
  Option.the
    (Map.map_of Stringa.equal_literal
      (List.zip initialization_list
        (start_addrs A_ (B1_, B2_) empty_heap allocate))
      c);

end; (*struct StartConfig*)

structure ExternalCall : sig
  datatype 'a extCallRet = RetVal of 'a Value.vala | RetExc of 'a | RetStaySame
  val external_WT_defs_i_i_i_i :
    string -> string -> Type.ty list -> Type.ty -> unit Predicate.pred
  val external_WT_defs : string -> string -> Type.ty list -> Type.ty -> bool
  val heap_copy_loc_Fii_i_oB_Fii_i_i_oB_i_i_i_i_o_o :
    'b Value.addr ->
      ('a -> 'b -> Heap.addr_loc -> 'b Value.vala Predicate.pred) ->
        ('a -> 'b -> Heap.addr_loc -> 'b Value.vala -> 'a Predicate.pred) ->
          'b -> 'b -> Heap.addr_loc ->
                        'a -> (('b, 'c) Observable_Events.obs_event list * 'a)
                                Predicate.pred
  val heap_copies_Fii_i_oB_Fii_i_i_oB_i_i_i_i_o_o :
    'b Value.addr ->
      ('a -> 'b -> Heap.addr_loc -> 'b Value.vala Predicate.pred) ->
        ('a -> 'b -> Heap.addr_loc -> 'b Value.vala -> 'a Predicate.pred) ->
          'b -> 'b -> Heap.addr_loc list ->
                        'a -> (('b, 'c) Observable_Events.obs_event list * 'a)
                                Predicate.pred
  val heap_clone_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_o_o :
    'a HOL.equal -> 'b HOL.equal * 'b Value.addr ->
      ('a -> Type.htype -> ('a * 'b) Set.set) ->
        ('a -> 'b -> Type.htype option) ->
          ('a -> 'b -> Heap.addr_loc -> 'b Value.vala Predicate.pred) ->
            ('a -> 'b -> Heap.addr_loc -> 'b Value.vala -> 'a Predicate.pred) ->
              'c TypeRel.prog ->
                'a -> 'b -> ('a * (('b, 'd) Observable_Events.obs_event list *
                                    'b) option)
                              Predicate.pred
  val red_external_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_i_o_o_o :
    'a HOL.equal * 'a Value.addr -> 'c HOL.equal ->
      ('a -> 'b) ->
        ('b -> 'a) ->
          bool ->
            'c -> ('c -> Type.htype -> ('c * 'a) Set.set) ->
                    ('c -> 'a -> Type.htype option) ->
                      ('c ->
                        'a -> Heap.addr_loc -> 'a Value.vala Predicate.pred) ->
                        ('c ->
                          'a -> Heap.addr_loc ->
                                  'a Value.vala -> 'c Predicate.pred) ->
                          'd TypeRel.prog ->
                            'b -> 'c -> 'a ->
  string ->
    'a Value.vala list ->
      ((('a, (FWState.lock_action list)) FinFun.finfun *
         (('b, (string * (string * 'a)), 'c) FWState.new_thread_action list *
           ('b FWState.conditional_action list *
             (('b, 'a) FWState.wait_set_action list *
               ('b FWState.interrupt_action list *
                 ('a, 'b) Observable_Events.obs_event list))))) *
        ('a extCallRet * 'c))
        Predicate.pred
end = struct

datatype 'a extCallRet = RetVal of 'a Value.vala | RetExc of 'a | RetStaySame;

fun external_WT_defs_i_i_i_i x xa xb xc =
  Predicate.sup_pred
    (case xb
      of [] =>
        (case xc
          of Type.Void =>
            (if ((xa : string) = "start()V") andalso
                  ((x : string) = Type.thread)
              then Predicate.single () else Predicate.bot_pred)
          | Type.Boolean => Predicate.bot_pred
          | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
          | Type.Class _ => Predicate.bot_pred
          | Type.Array _ => Predicate.bot_pred)
      | _ :: _ => Predicate.bot_pred)
    (Predicate.sup_pred
      (case xb
        of [] =>
          (case xc
            of Type.Void =>
              (if ((xa : string) = "join()V") andalso
                    ((x : string) = Type.thread)
                then Predicate.single () else Predicate.bot_pred)
            | Type.Boolean => Predicate.bot_pred
            | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
            | Type.Class _ => Predicate.bot_pred
            | Type.Array _ => Predicate.bot_pred)
        | _ :: _ => Predicate.bot_pred)
      (Predicate.sup_pred
        (case xb
          of [] =>
            (case xc
              of Type.Void =>
                (if ((xa : string) = "interrupt()V") andalso
                      ((x : string) = Type.thread)
                  then Predicate.single () else Predicate.bot_pred)
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | _ :: _ => Predicate.bot_pred)
        (Predicate.sup_pred
          (case xb
            of [] =>
              (case xc of Type.Void => Predicate.bot_pred
                | Type.Boolean =>
                  (if ((xa : string) = "isInterrupted()Z") andalso
                        ((x : string) = Type.thread)
                    then Predicate.single () else Predicate.bot_pred)
                | Type.Integer => Predicate.bot_pred
                | Type.Nt => Predicate.bot_pred
                | Type.Class _ => Predicate.bot_pred
                | Type.Array _ => Predicate.bot_pred)
            | _ :: _ => Predicate.bot_pred)
          (Predicate.sup_pred
            (case xb
              of [] =>
                (case xc
                  of Type.Void =>
                    (if ((xa : string) = "wait()V") andalso
                          ((x : string) = Type.object)
                      then Predicate.single () else Predicate.bot_pred)
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.bot_pred
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | _ :: _ => Predicate.bot_pred)
            (Predicate.sup_pred
              (case xb
                of [] =>
                  (case xc
                    of Type.Void =>
                      (if ((xa : string) = "notify()V") andalso
                            ((x : string) = Type.object)
                        then Predicate.single () else Predicate.bot_pred)
                    | Type.Boolean => Predicate.bot_pred
                    | Type.Integer => Predicate.bot_pred
                    | Type.Nt => Predicate.bot_pred
                    | Type.Class _ => Predicate.bot_pred
                    | Type.Array _ => Predicate.bot_pred)
                | _ :: _ => Predicate.bot_pred)
              (Predicate.sup_pred
                (case xb
                  of [] =>
                    (case xc
                      of Type.Void =>
                        (if ((xa : string) = "notifyAll()V") andalso
                              ((x : string) = Type.object)
                          then Predicate.single () else Predicate.bot_pred)
                      | Type.Boolean => Predicate.bot_pred
                      | Type.Integer => Predicate.bot_pred
                      | Type.Nt => Predicate.bot_pred
                      | Type.Class _ => Predicate.bot_pred
                      | Type.Array _ => Predicate.bot_pred)
                  | _ :: _ => Predicate.bot_pred)
                (Predicate.sup_pred
                  (case xb
                    of [] =>
                      (case xc of Type.Void => Predicate.bot_pred
                        | Type.Boolean => Predicate.bot_pred
                        | Type.Integer => Predicate.bot_pred
                        | Type.Nt => Predicate.bot_pred
                        | Type.Class xba =>
                          (if ((xba : string) = Type.object) andalso
                                (((xa : string) = "clone()Ljava/lang/Object;") andalso
                                  ((x : string) = Type.object))
                            then Predicate.single () else Predicate.bot_pred)
                        | Type.Array _ => Predicate.bot_pred)
                    | _ :: _ => Predicate.bot_pred)
                  (Predicate.sup_pred
                    (case xb
                      of [] =>
                        (case xc of Type.Void => Predicate.bot_pred
                          | Type.Boolean => Predicate.bot_pred
                          | Type.Integer =>
                            (if ((xa : string) = "hashCode()I") andalso
                                  ((x : string) = Type.object)
                              then Predicate.single () else Predicate.bot_pred)
                          | Type.Nt => Predicate.bot_pred
                          | Type.Class _ => Predicate.bot_pred
                          | Type.Array _ => Predicate.bot_pred)
                      | _ :: _ => Predicate.bot_pred)
                    (Predicate.sup_pred
                      (case xb of [] => Predicate.bot_pred
                        | Type.Void :: lista => Predicate.bot_pred
                        | Type.Boolean :: lista => Predicate.bot_pred
                        | [Type.Integer] =>
                          (case xc
                            of Type.Void =>
                              (if ((xa : string) = "~print(I)V") andalso
                                    ((x : string) = Type.object)
                                then Predicate.single ()
                                else Predicate.bot_pred)
                            | Type.Boolean => Predicate.bot_pred
                            | Type.Integer => Predicate.bot_pred
                            | Type.Nt => Predicate.bot_pred
                            | Type.Class _ => Predicate.bot_pred
                            | Type.Array _ => Predicate.bot_pred)
                        | Type.Integer :: _ :: _ => Predicate.bot_pred
                        | Type.Nt :: lista => Predicate.bot_pred
                        | Type.Class _ :: lista => Predicate.bot_pred
                        | Type.Array _ :: lista => Predicate.bot_pred)
                      (Predicate.sup_pred
                        (case xb
                          of [] =>
                            (case xc of Type.Void => Predicate.bot_pred
                              | Type.Boolean => Predicate.bot_pred
                              | Type.Integer => Predicate.bot_pred
                              | Type.Nt => Predicate.bot_pred
                              | Type.Class xba =>
                                (if ((xba : string) = Type.thread) andalso
                                      (((xa : string) = "~Thread.currentThread()Ljava/lang/Thread;") andalso
((x : string) = Type.object))
                                  then Predicate.single ()
                                  else Predicate.bot_pred)
                              | Type.Array _ => Predicate.bot_pred)
                          | _ :: _ => Predicate.bot_pred)
                        (Predicate.sup_pred
                          (case xb
                            of [] =>
                              (case xc of Type.Void => Predicate.bot_pred
                                | Type.Boolean =>
                                  (if ((xa : string) = "~Thread.interrupted()Z") andalso
((x : string) = Type.object)
                                    then Predicate.single ()
                                    else Predicate.bot_pred)
                                | Type.Integer => Predicate.bot_pred
                                | Type.Nt => Predicate.bot_pred
                                | Type.Class _ => Predicate.bot_pred
                                | Type.Array _ => Predicate.bot_pred)
                            | _ :: _ => Predicate.bot_pred)
                          (case xb
                            of [] =>
                              (case xc
                                of Type.Void =>
                                  (if ((xa : string) = "~Thread.yield()V") andalso
((x : string) = Type.object)
                                    then Predicate.single ()
                                    else Predicate.bot_pred)
                                | Type.Boolean => Predicate.bot_pred
                                | Type.Integer => Predicate.bot_pred
                                | Type.Nt => Predicate.bot_pred
                                | Type.Class _ => Predicate.bot_pred
                                | Type.Array _ => Predicate.bot_pred)
                            | _ :: _ => Predicate.bot_pred))))))))))));

fun external_WT_defs x1 x2 x3 x4 =
  Predicate.holds (external_WT_defs_i_i_i_i x1 x2 x3 x4);

fun heap_copy_loc_Fii_i_oB_Fii_i_i_oB_i_i_i_i_o_o B_ heap_read heap_write x xa
  xb xc =
  Predicate.bind (heap_read xc x xb)
    (fn xd =>
      Predicate.bind (heap_write xc xa xb xd)
        (fn xca =>
          Predicate.single
            ([Observable_Events.ReadMem (x, xb, xd),
               Observable_Events.WriteMem (xa, xb, xd)],
              xca)));

fun heap_copies_Fii_i_oB_Fii_i_i_oB_i_i_i_i_o_o B_ heap_read heap_write x xa xb
  xc =
  Predicate.sup_pred
    (case xb of [] => Predicate.single ([], xc) | _ :: _ => Predicate.bot_pred)
    (case xb of [] => Predicate.bot_pred
      | al :: als =>
        Predicate.bind
          (heap_copy_loc_Fii_i_oB_Fii_i_i_oB_i_i_i_i_o_o B_ heap_read heap_write
            x xa al xc)
          (fn (ob, h) =>
            Predicate.bind
              (heap_copies_Fii_i_oB_Fii_i_i_oB_i_i_i_i_o_o B_ heap_read
                heap_write x xa als h)
              (fn (obs, ha) => Predicate.single (ob @ obs, ha))));

fun heap_clone_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_o_o A_ (B1_, B2_) x xa heap_read
  heap_write xb xc xd =
  Predicate.sup_pred
    (case xa xc xd of NONE => Predicate.bot_pred
      | SOME hT =>
        Predicate.bind
          (TypeRel.eq_i_i (Set.equal_set (Product_Type.equal_prod A_ B1_))
            (x xc hT) Set.bot_set)
          (fn () => Predicate.single (xc, NONE)))
    (Predicate.sup_pred
      (case xa xc xd of NONE => Predicate.bot_pred
        | SOME (Type.Class_type c) =>
          Predicate.bind (Predicate.pred_of_set (x xc (Type.Class_type c)))
            (fn (h, a) =>
              Predicate.bind (TypeRel.fields_i_i_o xb c)
                (fn xe =>
                  Predicate.bind
                    (heap_copies_Fii_i_oB_Fii_i_i_oB_i_i_i_i_o_o B2_ heap_read
                      heap_write xd a
                      (List.map
                        (fn (aa, b) =>
                          let
                            val (f, d) = aa;
                          in
                            (fn _ => Heap.CField (d, f))
                          end
                            b)
                        xe)
                      h)
                    (fn (obs, ha) =>
                      Predicate.single
                        (ha, SOME (Observable_Events.NewHeapElem
                                     (a, Type.Class_type c) ::
                                     obs,
                                    a)))))
        | SOME (Type.Array_type (_, _)) => Predicate.bot_pred)
      (case xa xc xd of NONE => Predicate.bot_pred
        | SOME (Type.Class_type _) => Predicate.bot_pred
        | SOME (Type.Array_type (t, n)) =>
          Predicate.bind (Predicate.pred_of_set (x xc (Type.Array_type (t, n))))
            (fn (h, a) =>
              Predicate.bind (TypeRel.fields_i_i_o xb Type.object)
                (fn xe =>
                  Predicate.bind
                    (heap_copies_Fii_i_oB_Fii_i_i_oB_i_i_i_i_o_o B2_ heap_read
                      heap_write xd a
                      (List.map
                         (fn (aa, b) =>
                           let
                             val (f, d) = aa;
                           in
                             (fn _ => Heap.CField (d, f))
                           end
                             b)
                         xe @
                        List.map Heap.ACell (List.upt Arith.zero_nat n))
                      h)
                    (fn (obs, ha) =>
                      Predicate.single
                        (ha, SOME (Observable_Events.NewHeapElem
                                     (a, Type.Array_type (t, n)) ::
                                     obs,
                                    a)))))));

fun red_external_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_i_o_o_o (A1_, A2_) C_
  x xa xb xc xd xe heap_read heap_write xf xg xh xi xj xk =
  Predicate.sup_pred
    (case xk
      of [] =>
        (case xe xh xi of NONE => Predicate.bot_pred
          | SOME (Type.Class_type c) =>
            Predicate.bind
              (Transitive_Closure_Table.rtranclp_FioB_i_i Stringa.equal_literal
                (TypeRel.subcls1_i_i_o xf) c Type.thread)
              (fn () =>
                (if ((xj : string) = "start()V")
                  then Predicate.single
                         (FWState.thread_action_to_thread_action A1_
                            Observable_Events.obs_action_obs_event
                            (FWState.ObsAction
                              (Observable_Events.ThreadStart (x xi)))
                            (FWState.thread_action_to_thread_action A1_
                              Observable_Events.obs_action_obs_event
                              (FWState.NewThreadAction
                                (FWState.NewThread
                                  (x xi, (c, ("run()V", xi)), xh)))
                              (FinFun.Finfun_const [],
                                ([], ([], ([], ([], [])))))),
                           (RetVal Value.Unit, xh))
                  else Predicate.bot_pred))
          | SOME (Type.Array_type (_, _)) => Predicate.bot_pred)
      | _ :: _ => Predicate.bot_pred)
    (Predicate.sup_pred
      (case xk
        of [] =>
          (case xe xh xi of NONE => Predicate.bot_pred
            | SOME (Type.Class_type c) =>
              Predicate.bind
                (Transitive_Closure_Table.rtranclp_FioB_i_i
                  Stringa.equal_literal (TypeRel.subcls1_i_i_o xf) c
                  Type.thread)
                (fn () =>
                  (if ((xj : string) = "start()V")
                    then Predicate.single
                           (FWState.thread_action_to_thread_action A1_
                              Observable_Events.obs_action_obs_event
                              (FWState.NewThreadAction
                                (FWState.ThreadExists (x xi, true)))
                              (FinFun.Finfun_const [],
                                ([], ([], ([], ([], []))))),
                             (RetExc
                                (StartConfig.addr_of_sys_xcpt C_ (A1_, A2_) xc
                                  xd "java/lang/IllegalThreadStateException"),
                               xh))
                    else Predicate.bot_pred))
            | SOME (Type.Array_type (_, _)) => Predicate.bot_pred)
        | _ :: _ => Predicate.bot_pred)
      (Predicate.sup_pred
        (case xk
          of [] =>
            (case xe xh xi of NONE => Predicate.bot_pred
              | SOME (Type.Class_type c) =>
                Predicate.bind
                  (Transitive_Closure_Table.rtranclp_FioB_i_i
                    Stringa.equal_literal (TypeRel.subcls1_i_i_o xf) c
                    Type.thread)
                  (fn () =>
                    (if ((xj : string) = "join()V")
                      then Predicate.single
                             (FWState.thread_action_to_thread_action A1_
                                Observable_Events.obs_action_obs_event
                                (FWState.ObsAction
                                  (Observable_Events.ThreadJoin (x xi)))
                                (FWState.thread_action_to_thread_action A1_
                                  Observable_Events.obs_action_obs_event
                                  (FWState.InterruptAction
                                    (FWState.IsInterrupted (xg, false)))
                                  (FWState.thread_action_to_thread_action A1_
                                    Observable_Events.obs_action_obs_event
                                    (FWState.ConditionalAction
                                      (FWState.Join (x xi)))
                                    (FinFun.Finfun_const [],
                                      ([], ([], ([], ([], []))))))),
                               (RetVal Value.Unit, xh))
                      else Predicate.bot_pred))
              | SOME (Type.Array_type (_, _)) => Predicate.bot_pred)
          | _ :: _ => Predicate.bot_pred)
        (Predicate.sup_pred
          (case xk
            of [] =>
              (case xe xh xi of NONE => Predicate.bot_pred
                | SOME (Type.Class_type c) =>
                  Predicate.bind
                    (Transitive_Closure_Table.rtranclp_FioB_i_i
                      Stringa.equal_literal (TypeRel.subcls1_i_i_o xf) c
                      Type.thread)
                    (fn () =>
                      (if ((xj : string) = "join()V")
                        then Predicate.single
                               (FWState.thread_action_to_thread_action A1_
                                  Observable_Events.obs_action_obs_event
                                  (FWState.ObsAction
                                    (Observable_Events.ObsInterrupted xg))
                                  (FWState.thread_action_to_thread_action A1_
                                    Observable_Events.obs_action_obs_event
                                    (FWState.InterruptAction
                                      (FWState.ClearInterrupt xg))
                                    (FWState.thread_action_to_thread_action A1_
                                      Observable_Events.obs_action_obs_event
                                      (FWState.InterruptAction
(FWState.IsInterrupted (xg, true)))
                                      (FinFun.Finfun_const [],
([], ([], ([], ([], []))))))),
                                 (RetExc
                                    (StartConfig.addr_of_sys_xcpt C_ (A1_, A2_)
                                      xc xd "java/lang/InterruptedException"),
                                   xh))
                        else Predicate.bot_pred))
                | SOME (Type.Array_type (_, _)) => Predicate.bot_pred)
            | _ :: _ => Predicate.bot_pred)
          (Predicate.sup_pred
            (case xk
              of [] =>
                (case xe xh xi of NONE => Predicate.bot_pred
                  | SOME (Type.Class_type c) =>
                    Predicate.bind
                      (Transitive_Closure_Table.rtranclp_FioB_i_i
                        Stringa.equal_literal (TypeRel.subcls1_i_i_o xf) c
                        Type.thread)
                      (fn () =>
                        (if ((xj : string) = "interrupt()V")
                          then Predicate.single
                                 (FWState.thread_action_to_thread_action A1_
                                    Observable_Events.obs_action_obs_event
                                    (FWState.ObsAction
                                      (Observable_Events.ObsInterrupt (x xi)))
                                    (FWState.thread_action_to_thread_action A1_
                                      Observable_Events.obs_action_obs_event
                                      (FWState.InterruptAction
(FWState.Interrupt (x xi)))
                                      (FWState.thread_action_to_thread_action
A1_ Observable_Events.obs_action_obs_event
(FWState.WaitSetAction (FWState.WakeUp (x xi)))
(FWState.thread_action_to_thread_action A1_
  Observable_Events.obs_action_obs_event
  (FWState.NewThreadAction (FWState.ThreadExists (x xi, true)))
  (FinFun.Finfun_const [], ([], ([], ([], ([], [])))))))),
                                   (RetVal Value.Unit, xh))
                          else Predicate.bot_pred))
                  | SOME (Type.Array_type (_, _)) => Predicate.bot_pred)
              | _ :: _ => Predicate.bot_pred)
            (Predicate.sup_pred
              (case xk
                of [] =>
                  (case xe xh xi of NONE => Predicate.bot_pred
                    | SOME (Type.Class_type c) =>
                      Predicate.bind
                        (Transitive_Closure_Table.rtranclp_FioB_i_i
                          Stringa.equal_literal (TypeRel.subcls1_i_i_o xf) c
                          Type.thread)
                        (fn () =>
                          (if ((xj : string) = "interrupt()V")
                            then Predicate.single
                                   (FWState.thread_action_to_thread_action A1_
                                      Observable_Events.obs_action_obs_event
                                      (FWState.NewThreadAction
(FWState.ThreadExists (x xi, false)))
                                      (FinFun.Finfun_const [],
([], ([], ([], ([], []))))),
                                     (RetVal Value.Unit, xh))
                            else Predicate.bot_pred))
                    | SOME (Type.Array_type (_, _)) => Predicate.bot_pred)
                | _ :: _ => Predicate.bot_pred)
              (Predicate.sup_pred
                (case xk
                  of [] =>
                    (case xe xh xi of NONE => Predicate.bot_pred
                      | SOME (Type.Class_type c) =>
                        Predicate.bind
                          (Transitive_Closure_Table.rtranclp_FioB_i_i
                            Stringa.equal_literal (TypeRel.subcls1_i_i_o xf) c
                            Type.thread)
                          (fn () =>
                            (if ((xj : string) = "isInterrupted()Z")
                              then Predicate.single
                                     (FWState.thread_action_to_thread_action A1_
Observable_Events.obs_action_obs_event
(FWState.ObsAction (Observable_Events.ObsInterrupted (x xi)))
(FWState.thread_action_to_thread_action A1_
  Observable_Events.obs_action_obs_event
  (FWState.InterruptAction (FWState.IsInterrupted (x xi, true)))
  (FinFun.Finfun_const [], ([], ([], ([], ([], [])))))),
                                       (RetVal (Value.Bool true), xh))
                              else Predicate.bot_pred))
                      | SOME (Type.Array_type (_, _)) => Predicate.bot_pred)
                  | _ :: _ => Predicate.bot_pred)
                (Predicate.sup_pred
                  (case xk
                    of [] =>
                      (case xe xh xi of NONE => Predicate.bot_pred
                        | SOME (Type.Class_type c) =>
                          Predicate.bind
                            (Transitive_Closure_Table.rtranclp_FioB_i_i
                              Stringa.equal_literal (TypeRel.subcls1_i_i_o xf) c
                              Type.thread)
                            (fn () =>
                              (if ((xj : string) = "isInterrupted()Z")
                                then Predicate.single
                                       (FWState.thread_action_to_thread_action
  A1_ Observable_Events.obs_action_obs_event
  (FWState.InterruptAction (FWState.IsInterrupted (x xi, false)))
  (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
 (RetVal (Value.Bool false), xh))
                                else Predicate.bot_pred))
                        | SOME (Type.Array_type (_, _)) => Predicate.bot_pred)
                    | _ :: _ => Predicate.bot_pred)
                  (Predicate.sup_pred
                    (case xk
                      of [] =>
                        (if ((xj : string) = "wait()V")
                          then Predicate.single
                                 (FWState.thread_action_to_thread_action A1_
                                    Observable_Events.obs_action_obs_event
                                    (FWState.ObsAction
                                      (Observable_Events.ObsInterrupted xg))
                                    (FWState.thread_action_to_thread_action A1_
                                      Observable_Events.obs_action_obs_event
                                      (FWState.InterruptAction
(FWState.ClearInterrupt xg))
                                      (FWState.thread_action_to_thread_action
A1_ Observable_Events.obs_action_obs_event
(FWState.InterruptAction (FWState.IsInterrupted (xg, true)))
(FWState.thread_action_to_thread_action A1_
  Observable_Events.obs_action_obs_event (FWState.LockAction (FWState.Lock, xi))
  (FWState.thread_action_to_thread_action A1_
    Observable_Events.obs_action_obs_event
    (FWState.LockAction (FWState.Unlock, xi))
    (FinFun.Finfun_const [], ([], ([], ([], ([], []))))))))),
                                   (RetExc
                                      (StartConfig.addr_of_sys_xcpt C_
(A1_, A2_) xc xd "java/lang/InterruptedException"),
                                     xh))
                          else Predicate.bot_pred)
                      | _ :: _ => Predicate.bot_pred)
                    (Predicate.sup_pred
                      (case xk
                        of [] =>
                          (if ((xj : string) = "wait()V")
                            then Predicate.single
                                   (FWState.thread_action_to_thread_action A1_
                                      Observable_Events.obs_action_obs_event
                                      (FWState.ObsAction
(Observable_Events.SyncUnlock xi))
                                      (FWState.thread_action_to_thread_action
A1_ Observable_Events.obs_action_obs_event
(FWState.InterruptAction (FWState.IsInterrupted (xg, false)))
(FWState.thread_action_to_thread_action A1_
  Observable_Events.obs_action_obs_event
  (FWState.LockAction (FWState.ReleaseAcquire, xi))
  (FWState.thread_action_to_thread_action A1_
    Observable_Events.obs_action_obs_event
    (FWState.LockAction (FWState.Lock, xi))
    (FWState.thread_action_to_thread_action A1_
      Observable_Events.obs_action_obs_event
      (FWState.LockAction (FWState.Unlock, xi))
      (FWState.thread_action_to_thread_action A1_
        Observable_Events.obs_action_obs_event
        (FWState.WaitSetAction (FWState.Suspend xi))
        (FinFun.Finfun_const [], ([], ([], ([], ([], [])))))))))),
                                     (RetStaySame, xh))
                            else Predicate.bot_pred)
                        | _ :: _ => Predicate.bot_pred)
                      (Predicate.sup_pred
                        (case xk
                          of [] =>
                            (if ((xj : string) = "wait()V")
                              then Predicate.single
                                     (FWState.thread_action_to_thread_action A1_
Observable_Events.obs_action_obs_event
(FWState.LockAction (FWState.UnlockFail, xi))
(FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                                       (RetExc
  (StartConfig.addr_of_sys_xcpt C_ (A1_, A2_) xc xd
    "java/lang/IllegalMonitorStateException"),
 xh))
                              else Predicate.bot_pred)
                          | _ :: _ => Predicate.bot_pred)
                        (Predicate.sup_pred
                          (case xk
                            of [] =>
                              (if ((xj : string) = "wait()V")
                                then Predicate.single
                                       (FWState.thread_action_to_thread_action
  A1_ Observable_Events.obs_action_obs_event
  (FWState.WaitSetAction FWState.Notified)
  (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
 (RetVal Value.Unit, xh))
                                else Predicate.bot_pred)
                            | _ :: _ => Predicate.bot_pred)
                          (Predicate.sup_pred
                            (case xk
                              of [] =>
                                (if ((xj : string) = "wait()V")
                                  then Predicate.single
 (FWState.thread_action_to_thread_action A1_
    Observable_Events.obs_action_obs_event
    (FWState.ObsAction (Observable_Events.ObsInterrupted xg))
    (FWState.thread_action_to_thread_action A1_
      Observable_Events.obs_action_obs_event
      (FWState.InterruptAction (FWState.ClearInterrupt xg))
      (FWState.thread_action_to_thread_action A1_
        Observable_Events.obs_action_obs_event
        (FWState.WaitSetAction FWState.WokenUp)
        (FinFun.Finfun_const [], ([], ([], ([], ([], []))))))),
   (RetExc
      (StartConfig.addr_of_sys_xcpt C_ (A1_, A2_) xc xd
        "java/lang/InterruptedException"),
     xh))
                                  else Predicate.bot_pred)
                              | _ :: _ => Predicate.bot_pred)
                            (Predicate.sup_pred
                              (case xk
                                of [] =>
                                  Predicate.bind
                                    (TypeRel.eq_i_i Product_Type.equal_bool xb
                                      true)
                                    (fn () =>
                                      (if ((xj : string) = "wait()V")
then Predicate.single
       (FWState.thread_action_to_thread_action A1_
          Observable_Events.obs_action_obs_event
          (FWState.ObsAction (Observable_Events.SyncUnlock xi))
          (FWState.thread_action_to_thread_action A1_
            Observable_Events.obs_action_obs_event
            (FWState.InterruptAction (FWState.IsInterrupted (xg, false)))
            (FWState.thread_action_to_thread_action A1_
              Observable_Events.obs_action_obs_event
              (FWState.LockAction (FWState.ReleaseAcquire, xi))
              (FWState.thread_action_to_thread_action A1_
                Observable_Events.obs_action_obs_event
                (FWState.LockAction (FWState.Lock, xi))
                (FWState.thread_action_to_thread_action A1_
                  Observable_Events.obs_action_obs_event
                  (FWState.LockAction (FWState.Unlock, xi))
                  (FinFun.Finfun_const [], ([], ([], ([], ([], []))))))))),
         (RetVal Value.Unit, xh))
else Predicate.bot_pred))
                                | _ :: _ => Predicate.bot_pred)
                              (Predicate.sup_pred
                                (case xk
                                  of [] =>
                                    (if ((xj : string) = "notify()V")
                                      then Predicate.single
     (FWState.thread_action_to_thread_action A1_
        Observable_Events.obs_action_obs_event
        (FWState.LockAction (FWState.Lock, xi))
        (FWState.thread_action_to_thread_action A1_
          Observable_Events.obs_action_obs_event
          (FWState.LockAction (FWState.Unlock, xi))
          (FWState.thread_action_to_thread_action A1_
            Observable_Events.obs_action_obs_event
            (FWState.WaitSetAction (FWState.Notify xi))
            (FinFun.Finfun_const [], ([], ([], ([], ([], []))))))),
       (RetVal Value.Unit, xh))
                                      else Predicate.bot_pred)
                                  | _ :: _ => Predicate.bot_pred)
                                (Predicate.sup_pred
                                  (case xk
                                    of [] =>
                                      (if ((xj : string) = "notify()V")
then Predicate.single
       (FWState.thread_action_to_thread_action A1_
          Observable_Events.obs_action_obs_event
          (FWState.LockAction (FWState.UnlockFail, xi))
          (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
         (RetExc
            (StartConfig.addr_of_sys_xcpt C_ (A1_, A2_) xc xd
              "java/lang/IllegalMonitorStateException"),
           xh))
else Predicate.bot_pred)
                                    | _ :: _ => Predicate.bot_pred)
                                  (Predicate.sup_pred
                                    (case xk
                                      of [] =>
(if ((xj : string) = "notifyAll()V")
  then Predicate.single
         (FWState.thread_action_to_thread_action A1_
            Observable_Events.obs_action_obs_event
            (FWState.LockAction (FWState.Lock, xi))
            (FWState.thread_action_to_thread_action A1_
              Observable_Events.obs_action_obs_event
              (FWState.LockAction (FWState.Unlock, xi))
              (FWState.thread_action_to_thread_action A1_
                Observable_Events.obs_action_obs_event
                (FWState.WaitSetAction (FWState.NotifyAll xi))
                (FinFun.Finfun_const [], ([], ([], ([], ([], []))))))),
           (RetVal Value.Unit, xh))
  else Predicate.bot_pred)
                                      | _ :: _ => Predicate.bot_pred)
                                    (Predicate.sup_pred
                                      (case xk
of [] =>
  (if ((xj : string) = "notifyAll()V")
    then Predicate.single
           (FWState.thread_action_to_thread_action A1_
              Observable_Events.obs_action_obs_event
              (FWState.LockAction (FWState.UnlockFail, xi))
              (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
             (RetExc
                (StartConfig.addr_of_sys_xcpt C_ (A1_, A2_) xc xd
                  "java/lang/IllegalMonitorStateException"),
               xh))
    else Predicate.bot_pred)
| _ :: _ => Predicate.bot_pred)
                                      (Predicate.sup_pred
(case xk
  of [] =>
    Predicate.bind
      (heap_clone_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_o_o C_ (A1_, A2_) xd xe
        heap_read heap_write xf xh xi)
      (fn a =>
        (case a of (h, NONE) => Predicate.bot_pred
          | (h, SOME (obs, aa)) =>
            (if ((xj : string) = "clone()Ljava/lang/Object;")
              then Predicate.single
                     ((FinFun.Finfun_const [], ([], ([], ([], ([], obs))))),
                       (RetVal (Value.Addr aa), h))
              else Predicate.bot_pred)))
  | _ :: _ => Predicate.bot_pred)
(Predicate.sup_pred
  (case xk
    of [] =>
      Predicate.bind
        (heap_clone_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_o_o C_ (A1_, A2_) xd xe
          heap_read heap_write xf xh xi)
        (fn a =>
          (case a
            of (h, NONE) =>
              (if ((xj : string) = "clone()Ljava/lang/Object;")
                then Predicate.single
                       ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                         (RetExc
                            (StartConfig.addr_of_sys_xcpt C_ (A1_, A2_) xc xd
                              "java/lang/OutOfMemoryError"),
                           h))
                else Predicate.bot_pred)
            | (h, SOME _) => Predicate.bot_pred))
    | _ :: _ => Predicate.bot_pred)
  (Predicate.sup_pred
    (case xk
      of [] =>
        (if ((xj : string) = "hashCode()I")
          then Predicate.single
                 ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                   (RetVal
                      (Value.Intg
                        (Word.word_of_int
                          (Type_Length.len0_bit0
                            (Type_Length.len0_bit0
                              (Type_Length.len0_bit0
                                (Type_Length.len0_bit0
                                  (Type_Length.len0_bit0
                                    Type_Length.len0_num1)))))
                          (Value.hash_addr A2_ xi))),
                     xh))
          else Predicate.bot_pred)
      | _ :: _ => Predicate.bot_pred)
    (Predicate.sup_pred
      (if ((xj : string) = "~print(I)V")
        then Predicate.single
               (FWState.thread_action_to_thread_action A1_
                  Observable_Events.obs_action_obs_event
                  (FWState.ObsAction
                    (Observable_Events.ExternalCall
                      (xi, "~print(I)V", xk, Value.Unit)))
                  (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                 (RetVal Value.Unit, xh))
        else Predicate.bot_pred)
      (Predicate.sup_pred
        (case xk
          of [] =>
            (if ((xj : string) = "~Thread.currentThread()Ljava/lang/Thread;")
              then Predicate.single
                     ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                       (RetVal (Value.Addr (xa xg)), xh))
              else Predicate.bot_pred)
          | _ :: _ => Predicate.bot_pred)
        (Predicate.sup_pred
          (case xk
            of [] =>
              (if ((xj : string) = "~Thread.interrupted()Z")
                then Predicate.single
                       (FWState.thread_action_to_thread_action A1_
                          Observable_Events.obs_action_obs_event
                          (FWState.ObsAction
                            (Observable_Events.ObsInterrupted xg))
                          (FWState.thread_action_to_thread_action A1_
                            Observable_Events.obs_action_obs_event
                            (FWState.InterruptAction
                              (FWState.ClearInterrupt xg))
                            (FWState.thread_action_to_thread_action A1_
                              Observable_Events.obs_action_obs_event
                              (FWState.InterruptAction
                                (FWState.IsInterrupted (xg, true)))
                              (FinFun.Finfun_const [],
                                ([], ([], ([], ([], []))))))),
                         (RetVal (Value.Bool true), xh))
                else Predicate.bot_pred)
            | _ :: _ => Predicate.bot_pred)
          (Predicate.sup_pred
            (case xk
              of [] =>
                (if ((xj : string) = "~Thread.interrupted()Z")
                  then Predicate.single
                         (FWState.thread_action_to_thread_action A1_
                            Observable_Events.obs_action_obs_event
                            (FWState.InterruptAction
                              (FWState.IsInterrupted (xg, false)))
                            (FinFun.Finfun_const [],
                              ([], ([], ([], ([], []))))),
                           (RetVal (Value.Bool false), xh))
                  else Predicate.bot_pred)
              | _ :: _ => Predicate.bot_pred)
            (case xk
              of [] =>
                (if ((xj : string) = "~Thread.yield()V")
                  then Predicate.single
                         (FWState.thread_action_to_thread_action A1_
                            Observable_Events.obs_action_obs_event
                            (FWState.ConditionalAction FWState.Yield)
                            (FinFun.Finfun_const [],
                              ([], ([], ([], ([], []))))),
                           (RetVal Value.Unit, xh))
                  else Predicate.bot_pred)
              | _ :: _ => Predicate.bot_pred)))))))))))))))))))))))));

end; (*struct ExternalCall*)

structure Sum_Type : sig
  datatype ('a, 'b) sum = Inl of 'a | Inr of 'b
end = struct

datatype ('a, 'b) sum = Inl of 'a | Inr of 'b;

end; (*struct Sum_Type*)

structure BinOp : sig
  datatype bop = Eq | NotEq | LessThan | LessOrEqual | GreaterThan |
    GreaterOrEqual | Add | Subtract | Mult | Div | Mod | BinAnd | BinOr | BinXor
    | ShiftLeft | ShiftRightZeros | ShiftRightSigned
  val binop_ShiftRightSigned :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_ShiftRightZeros :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_GreaterOrEqual :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_LessOrEqual :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_GreaterThan :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_ShiftLeft :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_Subtract :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_LessThan :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_BinXor :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_BinAnd :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_BinOr :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop_Mult :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val word_smod :
    'a Type_Length.len -> 'a Word.word -> 'a Word.word -> 'a Word.word
  val binop_Mod :
    'a HOL.equal -> 'b HOL.equal * 'b Value.addr ->
      'a -> ('a -> Type.htype -> ('a * 'b) Set.set) ->
              'b Value.vala ->
                'b Value.vala -> ('b Value.vala, 'b) Sum_Type.sum option
  val word_sdiv :
    'a Type_Length.len -> 'a Word.word -> 'a Word.word -> 'a Word.word
  val binop_Div :
    'a HOL.equal -> 'b HOL.equal * 'b Value.addr ->
      'a -> ('a -> Type.htype -> ('a * 'b) Set.set) ->
              'b Value.vala ->
                'b Value.vala -> ('b Value.vala, 'b) Sum_Type.sum option
  val binop_Add :
    'a Value.vala -> 'a Value.vala -> ('a Value.vala, 'a) Sum_Type.sum option
  val binop :
    'a HOL.equal -> 'b HOL.equal * 'b Value.addr ->
      'a -> ('a -> Type.htype -> ('a * 'b) Set.set) ->
              bop ->
                'b Value.vala ->
                  'b Value.vala -> ('b Value.vala, 'b) Sum_Type.sum option
  val equal_bop : bop -> bop -> bool
  val wT_binop_i_i_i_i_i :
    'a TypeRel.prog ->
      Type.ty -> bop -> Type.ty -> Type.ty -> unit Predicate.pred
  val wT_binop_i_i_i_i_o :
    'a TypeRel.prog -> Type.ty -> bop -> Type.ty -> Type.ty Predicate.pred
end = struct

datatype bop = Eq | NotEq | LessThan | LessOrEqual | GreaterThan |
  GreaterOrEqual | Add | Subtract | Mult | Div | Mod | BinAnd | BinOr | BinXor |
  ShiftLeft | ShiftRightZeros | ShiftRightSigned;

fun binop_ShiftRightSigned (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.sshiftr
             (Type_Length.len_bit0
               (Type_Length.len_bit0
                 (Type_Length.len_bit0
                   (Type_Length.len_bit0
                     (Type_Length.len_bit0 Type_Length.len_num1)))))
             i1 (Word.unat
                  (Type_Length.len0_bit0
                    (Type_Length.len0_bit0
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0 Type_Length.len0_num1)))))
                  (Word.bitAND_word
                    (Type_Length.len0_bit0
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0
                            (Type_Length.len0_bit0 Type_Length.len0_num1)))))
                    i2 (Word.word_of_int
                         (Type_Length.len0_bit0
                           (Type_Length.len0_bit0
                             (Type_Length.len0_bit0
                               (Type_Length.len0_bit0
                                 (Type_Length.len0_bit0
                                   Type_Length.len0_num1)))))
                         (Arith.Int_of_integer (31 : IntInf.int))))))))
  | binop_ShiftRightSigned Value.Unit v2 = NONE
  | binop_ShiftRightSigned Value.Null v2 = NONE
  | binop_ShiftRightSigned (Value.Bool v) v2 = NONE
  | binop_ShiftRightSigned (Value.Addr v) v2 = NONE
  | binop_ShiftRightSigned v1 Value.Unit = NONE
  | binop_ShiftRightSigned v1 Value.Null = NONE
  | binop_ShiftRightSigned v1 (Value.Bool v) = NONE
  | binop_ShiftRightSigned v1 (Value.Addr v) = NONE;

fun binop_ShiftRightZeros (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.shiftr_word
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0 Type_Length.len0_num1)))))
             i1 (Word.unat
                  (Type_Length.len0_bit0
                    (Type_Length.len0_bit0
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0 Type_Length.len0_num1)))))
                  (Word.bitAND_word
                    (Type_Length.len0_bit0
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0
                            (Type_Length.len0_bit0 Type_Length.len0_num1)))))
                    i2 (Word.word_of_int
                         (Type_Length.len0_bit0
                           (Type_Length.len0_bit0
                             (Type_Length.len0_bit0
                               (Type_Length.len0_bit0
                                 (Type_Length.len0_bit0
                                   Type_Length.len0_num1)))))
                         (Arith.Int_of_integer (31 : IntInf.int))))))))
  | binop_ShiftRightZeros Value.Unit v2 = NONE
  | binop_ShiftRightZeros Value.Null v2 = NONE
  | binop_ShiftRightZeros (Value.Bool v) v2 = NONE
  | binop_ShiftRightZeros (Value.Addr v) v2 = NONE
  | binop_ShiftRightZeros v1 Value.Unit = NONE
  | binop_ShiftRightZeros v1 Value.Null = NONE
  | binop_ShiftRightZeros v1 (Value.Bool v) = NONE
  | binop_ShiftRightZeros v1 (Value.Addr v) = NONE;

fun binop_GreaterOrEqual (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Bool
           (Word.word_sle
             (Type_Length.len_bit0
               (Type_Length.len_bit0
                 (Type_Length.len_bit0
                   (Type_Length.len_bit0
                     (Type_Length.len_bit0 Type_Length.len_num1)))))
             i2 i1)))
  | binop_GreaterOrEqual Value.Unit v2 = NONE
  | binop_GreaterOrEqual Value.Null v2 = NONE
  | binop_GreaterOrEqual (Value.Bool v) v2 = NONE
  | binop_GreaterOrEqual (Value.Addr v) v2 = NONE
  | binop_GreaterOrEqual v1 Value.Unit = NONE
  | binop_GreaterOrEqual v1 Value.Null = NONE
  | binop_GreaterOrEqual v1 (Value.Bool v) = NONE
  | binop_GreaterOrEqual v1 (Value.Addr v) = NONE;

fun binop_LessOrEqual (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Bool
           (Word.word_sle
             (Type_Length.len_bit0
               (Type_Length.len_bit0
                 (Type_Length.len_bit0
                   (Type_Length.len_bit0
                     (Type_Length.len_bit0 Type_Length.len_num1)))))
             i1 i2)))
  | binop_LessOrEqual Value.Unit v2 = NONE
  | binop_LessOrEqual Value.Null v2 = NONE
  | binop_LessOrEqual (Value.Bool v) v2 = NONE
  | binop_LessOrEqual (Value.Addr v) v2 = NONE
  | binop_LessOrEqual v1 Value.Unit = NONE
  | binop_LessOrEqual v1 Value.Null = NONE
  | binop_LessOrEqual v1 (Value.Bool v) = NONE
  | binop_LessOrEqual v1 (Value.Addr v) = NONE;

fun binop_GreaterThan (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Bool
           (Word.word_sless
             (Type_Length.len_bit0
               (Type_Length.len_bit0
                 (Type_Length.len_bit0
                   (Type_Length.len_bit0
                     (Type_Length.len_bit0 Type_Length.len_num1)))))
             i2 i1)))
  | binop_GreaterThan Value.Unit v2 = NONE
  | binop_GreaterThan Value.Null v2 = NONE
  | binop_GreaterThan (Value.Bool v) v2 = NONE
  | binop_GreaterThan (Value.Addr v) v2 = NONE
  | binop_GreaterThan v1 Value.Unit = NONE
  | binop_GreaterThan v1 Value.Null = NONE
  | binop_GreaterThan v1 (Value.Bool v) = NONE
  | binop_GreaterThan v1 (Value.Addr v) = NONE;

fun binop_ShiftLeft (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.shiftl_word
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0 Type_Length.len0_num1)))))
             i1 (Word.unat
                  (Type_Length.len0_bit0
                    (Type_Length.len0_bit0
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0 Type_Length.len0_num1)))))
                  (Word.bitAND_word
                    (Type_Length.len0_bit0
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0
                            (Type_Length.len0_bit0 Type_Length.len0_num1)))))
                    i2 (Word.word_of_int
                         (Type_Length.len0_bit0
                           (Type_Length.len0_bit0
                             (Type_Length.len0_bit0
                               (Type_Length.len0_bit0
                                 (Type_Length.len0_bit0
                                   Type_Length.len0_num1)))))
                         (Arith.Int_of_integer (31 : IntInf.int))))))))
  | binop_ShiftLeft Value.Unit v2 = NONE
  | binop_ShiftLeft Value.Null v2 = NONE
  | binop_ShiftLeft (Value.Bool v) v2 = NONE
  | binop_ShiftLeft (Value.Addr v) v2 = NONE
  | binop_ShiftLeft v1 Value.Unit = NONE
  | binop_ShiftLeft v1 Value.Null = NONE
  | binop_ShiftLeft v1 (Value.Bool v) = NONE
  | binop_ShiftLeft v1 (Value.Addr v) = NONE;

fun binop_Subtract (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.minus_word
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0 Type_Length.len0_num1)))))
             i1 i2)))
  | binop_Subtract Value.Unit v2 = NONE
  | binop_Subtract Value.Null v2 = NONE
  | binop_Subtract (Value.Bool v) v2 = NONE
  | binop_Subtract (Value.Addr v) v2 = NONE
  | binop_Subtract v1 Value.Unit = NONE
  | binop_Subtract v1 Value.Null = NONE
  | binop_Subtract v1 (Value.Bool v) = NONE
  | binop_Subtract v1 (Value.Addr v) = NONE;

fun binop_LessThan (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Bool
           (Word.word_sless
             (Type_Length.len_bit0
               (Type_Length.len_bit0
                 (Type_Length.len_bit0
                   (Type_Length.len_bit0
                     (Type_Length.len_bit0 Type_Length.len_num1)))))
             i1 i2)))
  | binop_LessThan Value.Unit v2 = NONE
  | binop_LessThan Value.Null v2 = NONE
  | binop_LessThan (Value.Bool v) v2 = NONE
  | binop_LessThan (Value.Addr v) v2 = NONE
  | binop_LessThan v1 Value.Unit = NONE
  | binop_LessThan v1 Value.Null = NONE
  | binop_LessThan v1 (Value.Bool v) = NONE
  | binop_LessThan v1 (Value.Addr v) = NONE;

fun binop_BinXor (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.bitXOR_word
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0 Type_Length.len0_num1)))))
             i1 i2)))
  | binop_BinXor (Value.Bool b1) (Value.Bool b2) =
    SOME (Sum_Type.Inl (Value.Bool (not (Product_Type.equal_boola b1 b2))))
  | binop_BinXor Value.Unit v2 = NONE
  | binop_BinXor Value.Null v2 = NONE
  | binop_BinXor (Value.Bool v) Value.Unit = NONE
  | binop_BinXor (Value.Bool v) Value.Null = NONE
  | binop_BinXor (Value.Bool v) (Value.Intg va) = NONE
  | binop_BinXor (Value.Bool v) (Value.Addr va) = NONE
  | binop_BinXor (Value.Addr v) v2 = NONE
  | binop_BinXor v1 Value.Unit = NONE
  | binop_BinXor v1 Value.Null = NONE
  | binop_BinXor (Value.Intg va) (Value.Bool v) = NONE
  | binop_BinXor v1 (Value.Addr v) = NONE;

fun binop_BinAnd (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.bitAND_word
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0 Type_Length.len0_num1)))))
             i1 i2)))
  | binop_BinAnd (Value.Bool b1) (Value.Bool b2) =
    SOME (Sum_Type.Inl (Value.Bool (b1 andalso b2)))
  | binop_BinAnd Value.Unit v2 = NONE
  | binop_BinAnd Value.Null v2 = NONE
  | binop_BinAnd (Value.Bool v) Value.Unit = NONE
  | binop_BinAnd (Value.Bool v) Value.Null = NONE
  | binop_BinAnd (Value.Bool v) (Value.Intg va) = NONE
  | binop_BinAnd (Value.Bool v) (Value.Addr va) = NONE
  | binop_BinAnd (Value.Addr v) v2 = NONE
  | binop_BinAnd v1 Value.Unit = NONE
  | binop_BinAnd v1 Value.Null = NONE
  | binop_BinAnd (Value.Intg va) (Value.Bool v) = NONE
  | binop_BinAnd v1 (Value.Addr v) = NONE;

fun binop_BinOr (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.bitOR_word
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0 Type_Length.len0_num1)))))
             i1 i2)))
  | binop_BinOr (Value.Bool b1) (Value.Bool b2) =
    SOME (Sum_Type.Inl (Value.Bool (b1 orelse b2)))
  | binop_BinOr Value.Unit v2 = NONE
  | binop_BinOr Value.Null v2 = NONE
  | binop_BinOr (Value.Bool v) Value.Unit = NONE
  | binop_BinOr (Value.Bool v) Value.Null = NONE
  | binop_BinOr (Value.Bool v) (Value.Intg va) = NONE
  | binop_BinOr (Value.Bool v) (Value.Addr va) = NONE
  | binop_BinOr (Value.Addr v) v2 = NONE
  | binop_BinOr v1 Value.Unit = NONE
  | binop_BinOr v1 Value.Null = NONE
  | binop_BinOr (Value.Intg va) (Value.Bool v) = NONE
  | binop_BinOr v1 (Value.Addr v) = NONE;

fun binop_Mult (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.times_word
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0 Type_Length.len0_num1)))))
             i1 i2)))
  | binop_Mult Value.Unit v2 = NONE
  | binop_Mult Value.Null v2 = NONE
  | binop_Mult (Value.Bool v) v2 = NONE
  | binop_Mult (Value.Addr v) v2 = NONE
  | binop_Mult v1 Value.Unit = NONE
  | binop_Mult v1 Value.Null = NONE
  | binop_Mult v1 (Value.Bool v) = NONE
  | binop_Mult v1 (Value.Addr v) = NONE;

fun word_smod A_ x y =
  let
    val xa = Word.sint A_ x;
    val ya = Word.sint A_ y;
    val negative = Arith.less_int xa Arith.zero_inta;
    val result = Arith.mod_int (Arith.abs_int xa) (Arith.abs_int ya);
  in
    Word.word_of_int (Type_Length.len0_len A_)
      (if negative then Arith.uminus_int result else result)
  end;

fun binop_Mod A_ (B1_, B2_) empty_heap allocate v1 (Value.Addr v) = NONE
  | binop_Mod A_ (B1_, B2_) empty_heap allocate v1 (Value.Bool v) = NONE
  | binop_Mod A_ (B1_, B2_) empty_heap allocate v1 Value.Null = NONE
  | binop_Mod A_ (B1_, B2_) empty_heap allocate v1 Value.Unit = NONE
  | binop_Mod A_ (B1_, B2_) empty_heap allocate (Value.Addr v) v2 = NONE
  | binop_Mod A_ (B1_, B2_) empty_heap allocate (Value.Bool v) v2 = NONE
  | binop_Mod A_ (B1_, B2_) empty_heap allocate Value.Null v2 = NONE
  | binop_Mod A_ (B1_, B2_) empty_heap allocate Value.Unit v2 = NONE
  | binop_Mod A_ (B1_, B2_) empty_heap allocate (Value.Intg i1) (Value.Intg i2)
    = SOME (if Word.equal_word
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0
                       (Type_Length.len0_bit0
                         (Type_Length.len0_bit0 Type_Length.len0_num1)))))
                 i2 (Word.zero_word
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0
                            (Type_Length.len0_bit0
                              (Type_Length.len0_bit0 Type_Length.len0_num1))))))
             then Sum_Type.Inr
                    (StartConfig.addr_of_sys_xcpt A_ (B1_, B2_) empty_heap
                      allocate "java/lang/ArithmeticException")
             else Sum_Type.Inl
                    (Value.Intg
                      (word_smod
                        (Type_Length.len_bit0
                          (Type_Length.len_bit0
                            (Type_Length.len_bit0
                              (Type_Length.len_bit0
                                (Type_Length.len_bit0 Type_Length.len_num1)))))
                        i1 i2)));

fun word_sdiv A_ x y =
  let
    val xa = Word.sint A_ x;
    val ya = Word.sint A_ y;
    val negative =
      not (Product_Type.equal_boola (Arith.less_int xa Arith.zero_inta)
            (Arith.less_int ya Arith.zero_inta));
    val result = Arith.div_int (Arith.abs_int xa) (Arith.abs_int ya);
  in
    Word.word_of_int (Type_Length.len0_len A_)
      (if negative then Arith.uminus_int result else result)
  end;

fun binop_Div A_ (B1_, B2_) empty_heap allocate v1 (Value.Addr v) = NONE
  | binop_Div A_ (B1_, B2_) empty_heap allocate v1 (Value.Bool v) = NONE
  | binop_Div A_ (B1_, B2_) empty_heap allocate v1 Value.Null = NONE
  | binop_Div A_ (B1_, B2_) empty_heap allocate v1 Value.Unit = NONE
  | binop_Div A_ (B1_, B2_) empty_heap allocate (Value.Addr v) v2 = NONE
  | binop_Div A_ (B1_, B2_) empty_heap allocate (Value.Bool v) v2 = NONE
  | binop_Div A_ (B1_, B2_) empty_heap allocate Value.Null v2 = NONE
  | binop_Div A_ (B1_, B2_) empty_heap allocate Value.Unit v2 = NONE
  | binop_Div A_ (B1_, B2_) empty_heap allocate (Value.Intg i1) (Value.Intg i2)
    = SOME (if Word.equal_word
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0
                       (Type_Length.len0_bit0
                         (Type_Length.len0_bit0 Type_Length.len0_num1)))))
                 i2 (Word.zero_word
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0
                            (Type_Length.len0_bit0
                              (Type_Length.len0_bit0 Type_Length.len0_num1))))))
             then Sum_Type.Inr
                    (StartConfig.addr_of_sys_xcpt A_ (B1_, B2_) empty_heap
                      allocate "java/lang/ArithmeticException")
             else Sum_Type.Inl
                    (Value.Intg
                      (word_sdiv
                        (Type_Length.len_bit0
                          (Type_Length.len_bit0
                            (Type_Length.len_bit0
                              (Type_Length.len_bit0
                                (Type_Length.len_bit0 Type_Length.len_num1)))))
                        i1 i2)));

fun binop_Add (Value.Intg i1) (Value.Intg i2) =
  SOME (Sum_Type.Inl
         (Value.Intg
           (Word.plus_word
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0
                     (Type_Length.len0_bit0 Type_Length.len0_num1)))))
             i1 i2)))
  | binop_Add Value.Unit v2 = NONE
  | binop_Add Value.Null v2 = NONE
  | binop_Add (Value.Bool v) v2 = NONE
  | binop_Add (Value.Addr v) v2 = NONE
  | binop_Add v1 Value.Unit = NONE
  | binop_Add v1 Value.Null = NONE
  | binop_Add v1 (Value.Bool v) = NONE
  | binop_Add v1 (Value.Addr v) = NONE;

fun binop A_ (B1_, B2_) empty_heap allocate ShiftRightSigned a b =
  binop_ShiftRightSigned a b
  | binop A_ (B1_, B2_) empty_heap allocate ShiftRightZeros a b =
    binop_ShiftRightZeros a b
  | binop A_ (B1_, B2_) empty_heap allocate ShiftLeft a b = binop_ShiftLeft a b
  | binop A_ (B1_, B2_) empty_heap allocate BinXor a b = binop_BinXor a b
  | binop A_ (B1_, B2_) empty_heap allocate BinOr a b = binop_BinOr a b
  | binop A_ (B1_, B2_) empty_heap allocate BinAnd a b = binop_BinAnd a b
  | binop A_ (B1_, B2_) empty_heap allocate Div a b =
    binop_Div A_ (B1_, B2_) empty_heap allocate a b
  | binop A_ (B1_, B2_) empty_heap allocate Mod a b =
    binop_Mod A_ (B1_, B2_) empty_heap allocate a b
  | binop A_ (B1_, B2_) empty_heap allocate Mult a b = binop_Mult a b
  | binop A_ (B1_, B2_) empty_heap allocate Subtract a b = binop_Subtract a b
  | binop A_ (B1_, B2_) empty_heap allocate Add a b = binop_Add a b
  | binop A_ (B1_, B2_) empty_heap allocate GreaterOrEqual a b =
    binop_GreaterOrEqual a b
  | binop A_ (B1_, B2_) empty_heap allocate GreaterThan a b =
    binop_GreaterThan a b
  | binop A_ (B1_, B2_) empty_heap allocate LessOrEqual a b =
    binop_LessOrEqual a b
  | binop A_ (B1_, B2_) empty_heap allocate LessThan a b = binop_LessThan a b
  | binop A_ (B1_, B2_) empty_heap allocate NotEq v1 v2 =
    SOME (Sum_Type.Inl (Value.Bool (not (Value.equal_vala B1_ v1 v2))))
  | binop A_ (B1_, B2_) empty_heap allocate Eq v1 v2 =
    SOME (Sum_Type.Inl (Value.Bool (Value.equal_vala B1_ v1 v2)));

fun equal_bop ShiftRightSigned ShiftRightZeros = false
  | equal_bop ShiftRightZeros ShiftRightSigned = false
  | equal_bop ShiftRightSigned ShiftLeft = false
  | equal_bop ShiftLeft ShiftRightSigned = false
  | equal_bop ShiftRightZeros ShiftLeft = false
  | equal_bop ShiftLeft ShiftRightZeros = false
  | equal_bop ShiftRightSigned BinXor = false
  | equal_bop BinXor ShiftRightSigned = false
  | equal_bop ShiftRightZeros BinXor = false
  | equal_bop BinXor ShiftRightZeros = false
  | equal_bop ShiftLeft BinXor = false
  | equal_bop BinXor ShiftLeft = false
  | equal_bop ShiftRightSigned BinOr = false
  | equal_bop BinOr ShiftRightSigned = false
  | equal_bop ShiftRightZeros BinOr = false
  | equal_bop BinOr ShiftRightZeros = false
  | equal_bop ShiftLeft BinOr = false
  | equal_bop BinOr ShiftLeft = false
  | equal_bop BinXor BinOr = false
  | equal_bop BinOr BinXor = false
  | equal_bop ShiftRightSigned BinAnd = false
  | equal_bop BinAnd ShiftRightSigned = false
  | equal_bop ShiftRightZeros BinAnd = false
  | equal_bop BinAnd ShiftRightZeros = false
  | equal_bop ShiftLeft BinAnd = false
  | equal_bop BinAnd ShiftLeft = false
  | equal_bop BinXor BinAnd = false
  | equal_bop BinAnd BinXor = false
  | equal_bop BinOr BinAnd = false
  | equal_bop BinAnd BinOr = false
  | equal_bop ShiftRightSigned Mod = false
  | equal_bop Mod ShiftRightSigned = false
  | equal_bop ShiftRightZeros Mod = false
  | equal_bop Mod ShiftRightZeros = false
  | equal_bop ShiftLeft Mod = false
  | equal_bop Mod ShiftLeft = false
  | equal_bop BinXor Mod = false
  | equal_bop Mod BinXor = false
  | equal_bop BinOr Mod = false
  | equal_bop Mod BinOr = false
  | equal_bop BinAnd Mod = false
  | equal_bop Mod BinAnd = false
  | equal_bop ShiftRightSigned Div = false
  | equal_bop Div ShiftRightSigned = false
  | equal_bop ShiftRightZeros Div = false
  | equal_bop Div ShiftRightZeros = false
  | equal_bop ShiftLeft Div = false
  | equal_bop Div ShiftLeft = false
  | equal_bop BinXor Div = false
  | equal_bop Div BinXor = false
  | equal_bop BinOr Div = false
  | equal_bop Div BinOr = false
  | equal_bop BinAnd Div = false
  | equal_bop Div BinAnd = false
  | equal_bop Mod Div = false
  | equal_bop Div Mod = false
  | equal_bop ShiftRightSigned Mult = false
  | equal_bop Mult ShiftRightSigned = false
  | equal_bop ShiftRightZeros Mult = false
  | equal_bop Mult ShiftRightZeros = false
  | equal_bop ShiftLeft Mult = false
  | equal_bop Mult ShiftLeft = false
  | equal_bop BinXor Mult = false
  | equal_bop Mult BinXor = false
  | equal_bop BinOr Mult = false
  | equal_bop Mult BinOr = false
  | equal_bop BinAnd Mult = false
  | equal_bop Mult BinAnd = false
  | equal_bop Mod Mult = false
  | equal_bop Mult Mod = false
  | equal_bop Div Mult = false
  | equal_bop Mult Div = false
  | equal_bop ShiftRightSigned Subtract = false
  | equal_bop Subtract ShiftRightSigned = false
  | equal_bop ShiftRightZeros Subtract = false
  | equal_bop Subtract ShiftRightZeros = false
  | equal_bop ShiftLeft Subtract = false
  | equal_bop Subtract ShiftLeft = false
  | equal_bop BinXor Subtract = false
  | equal_bop Subtract BinXor = false
  | equal_bop BinOr Subtract = false
  | equal_bop Subtract BinOr = false
  | equal_bop BinAnd Subtract = false
  | equal_bop Subtract BinAnd = false
  | equal_bop Mod Subtract = false
  | equal_bop Subtract Mod = false
  | equal_bop Div Subtract = false
  | equal_bop Subtract Div = false
  | equal_bop Mult Subtract = false
  | equal_bop Subtract Mult = false
  | equal_bop ShiftRightSigned Add = false
  | equal_bop Add ShiftRightSigned = false
  | equal_bop ShiftRightZeros Add = false
  | equal_bop Add ShiftRightZeros = false
  | equal_bop ShiftLeft Add = false
  | equal_bop Add ShiftLeft = false
  | equal_bop BinXor Add = false
  | equal_bop Add BinXor = false
  | equal_bop BinOr Add = false
  | equal_bop Add BinOr = false
  | equal_bop BinAnd Add = false
  | equal_bop Add BinAnd = false
  | equal_bop Mod Add = false
  | equal_bop Add Mod = false
  | equal_bop Div Add = false
  | equal_bop Add Div = false
  | equal_bop Mult Add = false
  | equal_bop Add Mult = false
  | equal_bop Subtract Add = false
  | equal_bop Add Subtract = false
  | equal_bop ShiftRightSigned GreaterOrEqual = false
  | equal_bop GreaterOrEqual ShiftRightSigned = false
  | equal_bop ShiftRightZeros GreaterOrEqual = false
  | equal_bop GreaterOrEqual ShiftRightZeros = false
  | equal_bop ShiftLeft GreaterOrEqual = false
  | equal_bop GreaterOrEqual ShiftLeft = false
  | equal_bop BinXor GreaterOrEqual = false
  | equal_bop GreaterOrEqual BinXor = false
  | equal_bop BinOr GreaterOrEqual = false
  | equal_bop GreaterOrEqual BinOr = false
  | equal_bop BinAnd GreaterOrEqual = false
  | equal_bop GreaterOrEqual BinAnd = false
  | equal_bop Mod GreaterOrEqual = false
  | equal_bop GreaterOrEqual Mod = false
  | equal_bop Div GreaterOrEqual = false
  | equal_bop GreaterOrEqual Div = false
  | equal_bop Mult GreaterOrEqual = false
  | equal_bop GreaterOrEqual Mult = false
  | equal_bop Subtract GreaterOrEqual = false
  | equal_bop GreaterOrEqual Subtract = false
  | equal_bop Add GreaterOrEqual = false
  | equal_bop GreaterOrEqual Add = false
  | equal_bop ShiftRightSigned GreaterThan = false
  | equal_bop GreaterThan ShiftRightSigned = false
  | equal_bop ShiftRightZeros GreaterThan = false
  | equal_bop GreaterThan ShiftRightZeros = false
  | equal_bop ShiftLeft GreaterThan = false
  | equal_bop GreaterThan ShiftLeft = false
  | equal_bop BinXor GreaterThan = false
  | equal_bop GreaterThan BinXor = false
  | equal_bop BinOr GreaterThan = false
  | equal_bop GreaterThan BinOr = false
  | equal_bop BinAnd GreaterThan = false
  | equal_bop GreaterThan BinAnd = false
  | equal_bop Mod GreaterThan = false
  | equal_bop GreaterThan Mod = false
  | equal_bop Div GreaterThan = false
  | equal_bop GreaterThan Div = false
  | equal_bop Mult GreaterThan = false
  | equal_bop GreaterThan Mult = false
  | equal_bop Subtract GreaterThan = false
  | equal_bop GreaterThan Subtract = false
  | equal_bop Add GreaterThan = false
  | equal_bop GreaterThan Add = false
  | equal_bop GreaterOrEqual GreaterThan = false
  | equal_bop GreaterThan GreaterOrEqual = false
  | equal_bop ShiftRightSigned LessOrEqual = false
  | equal_bop LessOrEqual ShiftRightSigned = false
  | equal_bop ShiftRightZeros LessOrEqual = false
  | equal_bop LessOrEqual ShiftRightZeros = false
  | equal_bop ShiftLeft LessOrEqual = false
  | equal_bop LessOrEqual ShiftLeft = false
  | equal_bop BinXor LessOrEqual = false
  | equal_bop LessOrEqual BinXor = false
  | equal_bop BinOr LessOrEqual = false
  | equal_bop LessOrEqual BinOr = false
  | equal_bop BinAnd LessOrEqual = false
  | equal_bop LessOrEqual BinAnd = false
  | equal_bop Mod LessOrEqual = false
  | equal_bop LessOrEqual Mod = false
  | equal_bop Div LessOrEqual = false
  | equal_bop LessOrEqual Div = false
  | equal_bop Mult LessOrEqual = false
  | equal_bop LessOrEqual Mult = false
  | equal_bop Subtract LessOrEqual = false
  | equal_bop LessOrEqual Subtract = false
  | equal_bop Add LessOrEqual = false
  | equal_bop LessOrEqual Add = false
  | equal_bop GreaterOrEqual LessOrEqual = false
  | equal_bop LessOrEqual GreaterOrEqual = false
  | equal_bop GreaterThan LessOrEqual = false
  | equal_bop LessOrEqual GreaterThan = false
  | equal_bop ShiftRightSigned LessThan = false
  | equal_bop LessThan ShiftRightSigned = false
  | equal_bop ShiftRightZeros LessThan = false
  | equal_bop LessThan ShiftRightZeros = false
  | equal_bop ShiftLeft LessThan = false
  | equal_bop LessThan ShiftLeft = false
  | equal_bop BinXor LessThan = false
  | equal_bop LessThan BinXor = false
  | equal_bop BinOr LessThan = false
  | equal_bop LessThan BinOr = false
  | equal_bop BinAnd LessThan = false
  | equal_bop LessThan BinAnd = false
  | equal_bop Mod LessThan = false
  | equal_bop LessThan Mod = false
  | equal_bop Div LessThan = false
  | equal_bop LessThan Div = false
  | equal_bop Mult LessThan = false
  | equal_bop LessThan Mult = false
  | equal_bop Subtract LessThan = false
  | equal_bop LessThan Subtract = false
  | equal_bop Add LessThan = false
  | equal_bop LessThan Add = false
  | equal_bop GreaterOrEqual LessThan = false
  | equal_bop LessThan GreaterOrEqual = false
  | equal_bop GreaterThan LessThan = false
  | equal_bop LessThan GreaterThan = false
  | equal_bop LessOrEqual LessThan = false
  | equal_bop LessThan LessOrEqual = false
  | equal_bop ShiftRightSigned NotEq = false
  | equal_bop NotEq ShiftRightSigned = false
  | equal_bop ShiftRightZeros NotEq = false
  | equal_bop NotEq ShiftRightZeros = false
  | equal_bop ShiftLeft NotEq = false
  | equal_bop NotEq ShiftLeft = false
  | equal_bop BinXor NotEq = false
  | equal_bop NotEq BinXor = false
  | equal_bop BinOr NotEq = false
  | equal_bop NotEq BinOr = false
  | equal_bop BinAnd NotEq = false
  | equal_bop NotEq BinAnd = false
  | equal_bop Mod NotEq = false
  | equal_bop NotEq Mod = false
  | equal_bop Div NotEq = false
  | equal_bop NotEq Div = false
  | equal_bop Mult NotEq = false
  | equal_bop NotEq Mult = false
  | equal_bop Subtract NotEq = false
  | equal_bop NotEq Subtract = false
  | equal_bop Add NotEq = false
  | equal_bop NotEq Add = false
  | equal_bop GreaterOrEqual NotEq = false
  | equal_bop NotEq GreaterOrEqual = false
  | equal_bop GreaterThan NotEq = false
  | equal_bop NotEq GreaterThan = false
  | equal_bop LessOrEqual NotEq = false
  | equal_bop NotEq LessOrEqual = false
  | equal_bop LessThan NotEq = false
  | equal_bop NotEq LessThan = false
  | equal_bop ShiftRightSigned Eq = false
  | equal_bop Eq ShiftRightSigned = false
  | equal_bop ShiftRightZeros Eq = false
  | equal_bop Eq ShiftRightZeros = false
  | equal_bop ShiftLeft Eq = false
  | equal_bop Eq ShiftLeft = false
  | equal_bop BinXor Eq = false
  | equal_bop Eq BinXor = false
  | equal_bop BinOr Eq = false
  | equal_bop Eq BinOr = false
  | equal_bop BinAnd Eq = false
  | equal_bop Eq BinAnd = false
  | equal_bop Mod Eq = false
  | equal_bop Eq Mod = false
  | equal_bop Div Eq = false
  | equal_bop Eq Div = false
  | equal_bop Mult Eq = false
  | equal_bop Eq Mult = false
  | equal_bop Subtract Eq = false
  | equal_bop Eq Subtract = false
  | equal_bop Add Eq = false
  | equal_bop Eq Add = false
  | equal_bop GreaterOrEqual Eq = false
  | equal_bop Eq GreaterOrEqual = false
  | equal_bop GreaterThan Eq = false
  | equal_bop Eq GreaterThan = false
  | equal_bop LessOrEqual Eq = false
  | equal_bop Eq LessOrEqual = false
  | equal_bop LessThan Eq = false
  | equal_bop Eq LessThan = false
  | equal_bop NotEq Eq = false
  | equal_bop Eq NotEq = false
  | equal_bop ShiftRightSigned ShiftRightSigned = true
  | equal_bop ShiftRightZeros ShiftRightZeros = true
  | equal_bop ShiftLeft ShiftLeft = true
  | equal_bop BinXor BinXor = true
  | equal_bop BinOr BinOr = true
  | equal_bop BinAnd BinAnd = true
  | equal_bop Mod Mod = true
  | equal_bop Div Div = true
  | equal_bop Mult Mult = true
  | equal_bop Subtract Subtract = true
  | equal_bop Add Add = true
  | equal_bop GreaterOrEqual GreaterOrEqual = true
  | equal_bop GreaterThan GreaterThan = true
  | equal_bop LessOrEqual LessOrEqual = true
  | equal_bop LessThan LessThan = true
  | equal_bop NotEq NotEq = true
  | equal_bop Eq Eq = true;

fun wT_binop_i_i_i_i_i x xa xb xc xd =
  Predicate.sup_pred
    (case xb
      of Eq =>
        (case xd of Type.Void => Predicate.bot_pred
          | Type.Boolean =>
            Predicate.bind
              (Predicate.if_pred
                (TypeRel.widen x xa xc orelse TypeRel.widen x xc xa))
              (fn () => Predicate.single ())
          | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
          | Type.Class _ => Predicate.bot_pred
          | Type.Array _ => Predicate.bot_pred)
      | NotEq =>
        (case xd of Type.Void => Predicate.bot_pred
          | Type.Boolean =>
            Predicate.bind
              (Predicate.if_pred
                (TypeRel.widen x xa xc orelse TypeRel.widen x xc xa))
              (fn () => Predicate.single ())
          | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
          | Type.Class _ => Predicate.bot_pred
          | Type.Array _ => Predicate.bot_pred)
      | LessThan => Predicate.bot_pred | LessOrEqual => Predicate.bot_pred
      | GreaterThan => Predicate.bot_pred | GreaterOrEqual => Predicate.bot_pred
      | Add => Predicate.bot_pred | Subtract => Predicate.bot_pred
      | Mult => Predicate.bot_pred | Div => Predicate.bot_pred
      | Mod => Predicate.bot_pred | BinAnd => Predicate.bot_pred
      | BinOr => Predicate.bot_pred | BinXor => Predicate.bot_pred
      | ShiftLeft => Predicate.bot_pred | ShiftRightZeros => Predicate.bot_pred
      | ShiftRightSigned => Predicate.bot_pred)
    (case xa of Type.Void => Predicate.bot_pred
      | Type.Boolean =>
        (case xb of Eq => Predicate.bot_pred | NotEq => Predicate.bot_pred
          | LessThan => Predicate.bot_pred | LessOrEqual => Predicate.bot_pred
          | GreaterThan => Predicate.bot_pred
          | GreaterOrEqual => Predicate.bot_pred | Add => Predicate.bot_pred
          | Subtract => Predicate.bot_pred | Mult => Predicate.bot_pred
          | Div => Predicate.bot_pred | Mod => Predicate.bot_pred
          | BinAnd =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.single ()
                  | Type.Integer => Predicate.bot_pred
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | BinOr =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.single ()
                  | Type.Integer => Predicate.bot_pred
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | BinXor =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.single ()
                  | Type.Integer => Predicate.bot_pred
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | ShiftLeft => Predicate.bot_pred
          | ShiftRightZeros => Predicate.bot_pred
          | ShiftRightSigned => Predicate.bot_pred)
      | Type.Integer =>
        (case xb of Eq => Predicate.bot_pred | NotEq => Predicate.bot_pred
          | LessThan =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.single ()
                  | Type.Integer => Predicate.bot_pred
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | LessOrEqual =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.single ()
                  | Type.Integer => Predicate.bot_pred
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | GreaterThan =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.single ()
                  | Type.Integer => Predicate.bot_pred
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | GreaterOrEqual =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.single ()
                  | Type.Integer => Predicate.bot_pred
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Add =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.single ()
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Subtract =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.single ()
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Mult =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.single ()
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Div =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.single ()
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Mod =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.single ()
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | BinAnd =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.single ()
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | BinOr =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.single ()
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | BinXor =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.single ()
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | ShiftLeft =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.single ()
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | ShiftRightZeros =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.single ()
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | ShiftRightSigned =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer =>
                (case xd of Type.Void => Predicate.bot_pred
                  | Type.Boolean => Predicate.bot_pred
                  | Type.Integer => Predicate.single ()
                  | Type.Nt => Predicate.bot_pred
                  | Type.Class _ => Predicate.bot_pred
                  | Type.Array _ => Predicate.bot_pred)
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred))
      | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
      | Type.Array _ => Predicate.bot_pred);

fun wT_binop_i_i_i_i_o x xa xb xc =
  Predicate.sup_pred
    (case xb
      of Eq =>
        Predicate.bind
          (Predicate.if_pred
            (TypeRel.widen x xa xc orelse TypeRel.widen x xc xa))
          (fn () => Predicate.single Type.Boolean)
      | NotEq =>
        Predicate.bind
          (Predicate.if_pred
            (TypeRel.widen x xa xc orelse TypeRel.widen x xc xa))
          (fn () => Predicate.single Type.Boolean)
      | LessThan => Predicate.bot_pred | LessOrEqual => Predicate.bot_pred
      | GreaterThan => Predicate.bot_pred | GreaterOrEqual => Predicate.bot_pred
      | Add => Predicate.bot_pred | Subtract => Predicate.bot_pred
      | Mult => Predicate.bot_pred | Div => Predicate.bot_pred
      | Mod => Predicate.bot_pred | BinAnd => Predicate.bot_pred
      | BinOr => Predicate.bot_pred | BinXor => Predicate.bot_pred
      | ShiftLeft => Predicate.bot_pred | ShiftRightZeros => Predicate.bot_pred
      | ShiftRightSigned => Predicate.bot_pred)
    (case xa of Type.Void => Predicate.bot_pred
      | Type.Boolean =>
        (case xb of Eq => Predicate.bot_pred | NotEq => Predicate.bot_pred
          | LessThan => Predicate.bot_pred | LessOrEqual => Predicate.bot_pred
          | GreaterThan => Predicate.bot_pred
          | GreaterOrEqual => Predicate.bot_pred | Add => Predicate.bot_pred
          | Subtract => Predicate.bot_pred | Mult => Predicate.bot_pred
          | Div => Predicate.bot_pred | Mod => Predicate.bot_pred
          | BinAnd =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.single Type.Boolean
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | BinOr =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.single Type.Boolean
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | BinXor =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.single Type.Boolean
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | ShiftLeft => Predicate.bot_pred
          | ShiftRightZeros => Predicate.bot_pred
          | ShiftRightSigned => Predicate.bot_pred)
      | Type.Integer =>
        (case xb of Eq => Predicate.bot_pred | NotEq => Predicate.bot_pred
          | LessThan =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Boolean
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | LessOrEqual =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Boolean
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | GreaterThan =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Boolean
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | GreaterOrEqual =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Boolean
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Add =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Subtract =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Mult =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Div =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | Mod =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | BinAnd =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | BinOr =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | BinXor =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | ShiftLeft =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | ShiftRightZeros =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
          | ShiftRightSigned =>
            (case xc of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.bot_pred
              | Type.Integer => Predicate.single Type.Integer
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred))
      | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
      | Type.Array _ => Predicate.bot_pred);

end; (*struct BinOp*)

structure Expr : sig
  datatype ('a, 'b, 'c) exp = New of string |
    NewArray of Type.ty * ('a, 'b, 'c) exp | Cast of Type.ty * ('a, 'b, 'c) exp
    | InstanceOf of ('a, 'b, 'c) exp * Type.ty | Val of 'c Value.vala |
    BinOp of ('a, 'b, 'c) exp * BinOp.bop * ('a, 'b, 'c) exp | Var of 'a |
    LAss of 'a * ('a, 'b, 'c) exp | AAcc of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp
    | AAss of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
    ALen of ('a, 'b, 'c) exp | FAcc of ('a, 'b, 'c) exp * string * string |
    FAss of ('a, 'b, 'c) exp * string * string * ('a, 'b, 'c) exp |
    Call of ('a, 'b, 'c) exp * string * ('a, 'b, 'c) exp list |
    Block of 'a * Type.ty * 'c Value.vala option * ('a, 'b, 'c) exp |
    Synchronized of 'b * ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
    InSynchronized of 'b * 'c * ('a, 'b, 'c) exp |
    Seq of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
    Cond of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
    While of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp | Throw of ('a, 'b, 'c) exp |
    TryCatch of ('a, 'b, 'c) exp * string * 'a * ('a, 'b, 'c) exp
  val final_i : ('a, 'b, 'c) exp -> unit Predicate.pred
  val final : ('a, 'b, 'c) exp -> bool
  val blocks :
    'a list ->
      Type.ty list -> 'b Value.vala list -> ('a, 'c, 'b) exp -> ('a, 'c, 'b) exp
  val is_val_i : ('a, 'b, 'c) exp -> unit Predicate.pred
  val is_val : ('a, 'b, 'c) exp -> bool
  val is_vals : ('a, 'b, 'c) exp list -> bool
  val the_Val : ('a, 'b, 'c) exp -> 'c Value.vala
  val equal_exp :
    'a HOL.equal -> 'b HOL.equal -> 'c HOL.equal -> ('a, 'b, 'c) exp HOL.equal
  val equal_expa :
    'a HOL.equal -> 'b HOL.equal -> 'c HOL.equal ->
      ('a, 'b, 'c) exp -> ('a, 'b, 'c) exp -> bool
  val extRet2J :
    ('a, 'b, 'c) exp -> 'c ExternalCall.extCallRet -> ('a, 'b, 'c) exp
  val is_Throws_i : ('a, 'b, 'c) exp list -> unit Predicate.pred
end = struct

datatype ('a, 'b, 'c) exp = New of string |
  NewArray of Type.ty * ('a, 'b, 'c) exp | Cast of Type.ty * ('a, 'b, 'c) exp |
  InstanceOf of ('a, 'b, 'c) exp * Type.ty | Val of 'c Value.vala |
  BinOp of ('a, 'b, 'c) exp * BinOp.bop * ('a, 'b, 'c) exp | Var of 'a |
  LAss of 'a * ('a, 'b, 'c) exp | AAcc of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
  AAss of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
  ALen of ('a, 'b, 'c) exp | FAcc of ('a, 'b, 'c) exp * string * string |
  FAss of ('a, 'b, 'c) exp * string * string * ('a, 'b, 'c) exp |
  Call of ('a, 'b, 'c) exp * string * ('a, 'b, 'c) exp list |
  Block of 'a * Type.ty * 'c Value.vala option * ('a, 'b, 'c) exp |
  Synchronized of 'b * ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
  InSynchronized of 'b * 'c * ('a, 'b, 'c) exp |
  Seq of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
  Cond of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp * ('a, 'b, 'c) exp |
  While of ('a, 'b, 'c) exp * ('a, 'b, 'c) exp | Throw of ('a, 'b, 'c) exp |
  TryCatch of ('a, 'b, 'c) exp * string * 'a * ('a, 'b, 'c) exp;

fun final_i x =
  (case x of New _ => Predicate.bot_pred | NewArray (_, _) => Predicate.bot_pred
    | Cast (_, _) => Predicate.bot_pred
    | InstanceOf (_, _) => Predicate.bot_pred | Val _ => Predicate.single ()
    | BinOp (_, _, _) => Predicate.bot_pred | Var _ => Predicate.bot_pred
    | LAss (_, _) => Predicate.bot_pred | AAcc (_, _) => Predicate.bot_pred
    | AAss (_, _, _) => Predicate.bot_pred | ALen _ => Predicate.bot_pred
    | FAcc (_, _, _) => Predicate.bot_pred
    | FAss (_, _, _, _) => Predicate.bot_pred
    | Call (_, _, _) => Predicate.bot_pred
    | Block (_, _, _, _) => Predicate.bot_pred
    | Synchronized (_, _, _) => Predicate.bot_pred
    | InSynchronized (_, _, _) => Predicate.bot_pred
    | Seq (_, _) => Predicate.bot_pred | Cond (_, _, _) => Predicate.bot_pred
    | While (_, _) => Predicate.bot_pred | Throw (New _) => Predicate.bot_pred
    | Throw (NewArray (_, _)) => Predicate.bot_pred
    | Throw (Cast (_, _)) => Predicate.bot_pred
    | Throw (InstanceOf (_, _)) => Predicate.bot_pred
    | Throw (Val Value.Unit) => Predicate.bot_pred
    | Throw (Val Value.Null) => Predicate.bot_pred
    | Throw (Val (Value.Bool _)) => Predicate.bot_pred
    | Throw (Val (Value.Intg _)) => Predicate.bot_pred
    | Throw (Val (Value.Addr _)) => Predicate.single ()
    | Throw (BinOp (_, _, _)) => Predicate.bot_pred
    | Throw (Var _) => Predicate.bot_pred
    | Throw (LAss (_, _)) => Predicate.bot_pred
    | Throw (AAcc (_, _)) => Predicate.bot_pred
    | Throw (AAss (_, _, _)) => Predicate.bot_pred
    | Throw (ALen _) => Predicate.bot_pred
    | Throw (FAcc (_, _, _)) => Predicate.bot_pred
    | Throw (FAss (_, _, _, _)) => Predicate.bot_pred
    | Throw (Call (_, _, _)) => Predicate.bot_pred
    | Throw (Block (_, _, _, _)) => Predicate.bot_pred
    | Throw (Synchronized (_, _, _)) => Predicate.bot_pred
    | Throw (InSynchronized (_, _, _)) => Predicate.bot_pred
    | Throw (Seq (_, _)) => Predicate.bot_pred
    | Throw (Cond (_, _, _)) => Predicate.bot_pred
    | Throw (While (_, _)) => Predicate.bot_pred
    | Throw (Throw _) => Predicate.bot_pred
    | Throw (TryCatch (_, _, _, _)) => Predicate.bot_pred
    | TryCatch (_, _, _, _) => Predicate.bot_pred);

fun final x1 = Predicate.holds (final_i x1);

fun blocks (va :: vsa) (t :: ts) (v :: vs) e =
  Block (va, t, SOME v, blocks vsa ts vs e)
  | blocks [] [] [] e = e;

fun is_val_i x =
  (case x of New _ => Predicate.bot_pred | NewArray (_, _) => Predicate.bot_pred
    | Cast (_, _) => Predicate.bot_pred
    | InstanceOf (_, _) => Predicate.bot_pred | Val _ => Predicate.single ()
    | BinOp (_, _, _) => Predicate.bot_pred | Var _ => Predicate.bot_pred
    | LAss (_, _) => Predicate.bot_pred | AAcc (_, _) => Predicate.bot_pred
    | AAss (_, _, _) => Predicate.bot_pred | ALen _ => Predicate.bot_pred
    | FAcc (_, _, _) => Predicate.bot_pred
    | FAss (_, _, _, _) => Predicate.bot_pred
    | Call (_, _, _) => Predicate.bot_pred
    | Block (_, _, _, _) => Predicate.bot_pred
    | Synchronized (_, _, _) => Predicate.bot_pred
    | InSynchronized (_, _, _) => Predicate.bot_pred
    | Seq (_, _) => Predicate.bot_pred | Cond (_, _, _) => Predicate.bot_pred
    | While (_, _) => Predicate.bot_pred | Throw _ => Predicate.bot_pred
    | TryCatch (_, _, _, _) => Predicate.bot_pred);

fun is_val x1 = Predicate.holds (is_val_i x1);

fun is_vals [] = true
  | is_vals (e :: es) = is_val e andalso is_vals es;

fun the_Val (Val v) = v;

fun equal_exp A_ B_ C_ = {equal = equal_expa A_ B_ C_} :
  ('a, 'b, 'c) exp HOL.equal
and equal_expa A_ B_ C_ (TryCatch (exp1, literal, a, exp2)) (Throw exp) = false
  | equal_expa A_ B_ C_ (Throw exp) (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1a, literal, a, exp2a))
    (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (While (exp1a, exp2a))
    (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw exp) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (Throw exp) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1a, literal, a, exp2a))
    (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Cond (exp1a, exp2a, exp3))
    (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw exp) (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1a, exp2a)) (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Cond (exp1a, exp2a, exp3)) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1a, literal, a, exp2a)) (Seq (exp1, exp2))
    = false
  | equal_expa A_ B_ C_ (Seq (exp1a, exp2a)) (TryCatch (exp1, literal, a, exp2))
    = false
  | equal_expa A_ B_ C_ (Throw exp) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1a, exp2a)) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Seq (exp1a, exp2a)) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1a, exp2a, exp3)) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Seq (exp1a, exp2a)) (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1, literal, a, exp2))
    (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, exp))
    (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw expa) (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, expa)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (InSynchronized (b, addr, exp)) =
    false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, exp)) (While (exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3))
    (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, exp))
    (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (InSynchronized (b, addr, exp)) =
    false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, exp)) (Seq (exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (TryCatch (exp1a, literal, a, exp2a))
    (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1a, exp2a))
    (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw exp) (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1, exp2)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1a, exp2a)) (Synchronized (b, exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1a, exp2a)) (While (exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (Cond (exp1a, exp2a, exp3))
    (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1a, exp2a))
    (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1a, exp2a)) (Synchronized (b, exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1a, exp2a)) (Seq (exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (InSynchronized (ba, addr, exp))
    (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Synchronized (ba, exp1, exp2))
    (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1, literal, aa, exp2))
    (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Block (aa, ty, optiona, exp))
    (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw expa) (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, expa)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (Block (a, ty, optiona, exp)) =
    false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, exp)) (While (exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3)) (Block (a, ty, optiona, exp))
    = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, exp)) (Cond (exp1, exp2, exp3))
    = false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, exp)) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, expa))
    (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, expa))
    (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1, exp2))
    (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, exp))
    (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1, literala, a, exp2))
    (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (Call (exp, literala, lista))
    (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw expa) (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (Call (expa, literal, lista)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (Call (exp, literal, lista)) =
    false
  | equal_expa A_ B_ C_ (Call (exp, literal, lista)) (While (exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3)) (Call (exp, literal, lista)) =
    false
  | equal_expa A_ B_ C_ (Call (exp, literal, lista)) (Cond (exp1, exp2, exp3)) =
    false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (Call (exp, literal, lista)) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, expa))
    (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (Call (expa, literal, lista))
    (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1, exp2))
    (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (Call (exp, literal, lista))
    (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, expa))
    (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (Call (expa, literal, lista))
    (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1a, literal, a, exp2a))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAss (exp1a, literal1, literal2, exp2a))
    (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw exp) (FAss (exp1, literal1, literal2, exp2)) =
    false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1, literal2, exp2)) (Throw exp) =
    false
  | equal_expa A_ B_ C_ (While (exp1a, exp2a))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAss (exp1a, literal1, literal2, exp2a))
    (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1a, exp2a, exp3))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAss (exp1a, literal1, literal2, exp2a))
    (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1a, exp2a))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAss (exp1a, literal1, literal2, exp2a))
    (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, exp))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1, literal2, exp2))
    (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1a, exp2a))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAss (exp1a, literal1, literal2, exp2a))
    (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, exp))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1, literal2, exp2))
    (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Call (exp, literal, lista))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1, literal2, exp2))
    (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1, literal, a, exp2))
    (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (FAcc (exp, literal1, literal2))
    (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw expa) (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (FAcc (expa, literal1, literal2)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (FAcc (exp, literal1, literal2)) =
    false
  | equal_expa A_ B_ C_ (FAcc (exp, literal1, literal2)) (While (exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3))
    (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (FAcc (exp, literal1, literal2))
    (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (FAcc (exp, literal1, literal2)) =
    false
  | equal_expa A_ B_ C_ (FAcc (exp, literal1, literal2)) (Seq (exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, expa))
    (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (FAcc (expa, literal1, literal2))
    (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1, exp2))
    (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (FAcc (exp, literal1, literal2))
    (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, expa))
    (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (FAcc (expa, literal1, literal2))
    (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Call (expa, literal, lista))
    (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (FAcc (expa, literal1, literal2))
    (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1a, literal2a, exp2))
    (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (FAcc (exp, literal1a, literal2a))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1, literal, a, exp2)) (ALen exp) = false
  | equal_expa A_ B_ C_ (ALen exp) (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw expa) (ALen exp) = false
  | equal_expa A_ B_ C_ (ALen expa) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (ALen exp) = false
  | equal_expa A_ B_ C_ (ALen exp) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3)) (ALen exp) = false
  | equal_expa A_ B_ C_ (ALen exp) (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (ALen exp) = false
  | equal_expa A_ B_ C_ (ALen exp) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, expa)) (ALen exp) = false
  | equal_expa A_ B_ C_ (ALen expa) (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1, exp2)) (ALen exp) = false
  | equal_expa A_ B_ C_ (ALen exp) (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, expa)) (ALen exp) = false
  | equal_expa A_ B_ C_ (ALen expa) (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Call (expa, literal, lista)) (ALen exp) = false
  | equal_expa A_ B_ C_ (ALen expa) (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1, literal2, exp2)) (ALen exp) =
    false
  | equal_expa A_ B_ C_ (ALen exp) (FAss (exp1, literal1, literal2, exp2)) =
    false
  | equal_expa A_ B_ C_ (FAcc (expa, literal1, literal2)) (ALen exp) = false
  | equal_expa A_ B_ C_ (ALen expa) (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1a, literal, a, exp2a))
    (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAss (exp1a, exp2a, exp3))
    (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw exp) (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1a, exp2a)) (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAss (exp1a, exp2a, exp3)) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1a, exp2a, exp3a)) (AAss (exp1, exp2, exp3)) =
    false
  | equal_expa A_ B_ C_ (AAss (exp1a, exp2a, exp3a)) (Cond (exp1, exp2, exp3)) =
    false
  | equal_expa A_ B_ C_ (Seq (exp1a, exp2a)) (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAss (exp1a, exp2a, exp3)) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, exp))
    (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3))
    (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1a, exp2a))
    (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAss (exp1a, exp2a, exp3))
    (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, exp)) (AAss (exp1, exp2, exp3))
    = false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3)) (Block (a, ty, optiona, exp))
    = false
  | equal_expa A_ B_ C_ (Call (exp, literal, lista)) (AAss (exp1, exp2, exp3)) =
    false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3)) (Call (exp, literal, lista)) =
    false
  | equal_expa A_ B_ C_ (FAss (exp1a, literal1, literal2, exp2a))
    (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAss (exp1a, exp2a, exp3))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAcc (exp, literal1, literal2))
    (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3))
    (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (ALen exp) (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3)) (ALen exp) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1a, literal, a, exp2a))
    (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1a, exp2a))
    (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw exp) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1a, exp2a)) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1a, exp2a)) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1a, exp2a, exp3)) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1a, exp2a)) (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1a, exp2a)) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1a, exp2a)) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, exp)) (AAcc (exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (InSynchronized (b, addr, exp)) =
    false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1a, exp2a)) (AAcc (exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (AAcc (exp1a, exp2a)) (Synchronized (b, exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, exp)) (AAcc (exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (Block (a, ty, optiona, exp)) =
    false
  | equal_expa A_ B_ C_ (Call (exp, literal, lista)) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (FAss (exp1a, literal1, literal2, exp2a))
    (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1a, exp2a))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAcc (exp, literal1, literal2)) (AAcc (exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (FAcc (exp, literal1, literal2)) =
    false
  | equal_expa A_ B_ C_ (ALen exp) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (ALen exp) = false
  | equal_expa A_ B_ C_ (AAss (exp1a, exp2a, exp3)) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1a, exp2a)) (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1, literal, aa, exp2)) (LAss (a, exp)) =
    false
  | equal_expa A_ B_ C_ (LAss (aa, exp)) (TryCatch (exp1, literal, a, exp2)) =
    false
  | equal_expa A_ B_ C_ (Throw expa) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (LAss (a, expa)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (LAss (a, exp)) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (LAss (a, exp)) (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (LAss (a, exp)) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, expa)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (LAss (a, expa)) (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1, exp2)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (LAss (a, exp)) (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Block (aa, ty, optiona, expa)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (LAss (aa, expa)) (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Call (expa, literal, lista)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (LAss (a, expa)) (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1, literal2, exp2)) (LAss (a, exp))
    = false
  | equal_expa A_ B_ C_ (LAss (a, exp)) (FAss (exp1, literal1, literal2, exp2))
    = false
  | equal_expa A_ B_ C_ (FAcc (expa, literal1, literal2)) (LAss (a, exp)) =
    false
  | equal_expa A_ B_ C_ (LAss (a, expa)) (FAcc (exp, literal1, literal2)) =
    false
  | equal_expa A_ B_ C_ (ALen expa) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (LAss (a, expa)) (ALen exp) = false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (LAss (a, exp)) (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (LAss (a, exp)) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1, literal, aa, exp2)) (Var a) = false
  | equal_expa A_ B_ C_ (Var aa) (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw exp) (Var a) = false
  | equal_expa A_ B_ C_ (Var a) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (Var a) = false
  | equal_expa A_ B_ C_ (Var a) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3)) (Var a) = false
  | equal_expa A_ B_ C_ (Var a) (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (Var a) = false
  | equal_expa A_ B_ C_ (Var a) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, exp)) (Var a) = false
  | equal_expa A_ B_ C_ (Var a) (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1, exp2)) (Var a) = false
  | equal_expa A_ B_ C_ (Var a) (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Block (aa, ty, optiona, exp)) (Var a) = false
  | equal_expa A_ B_ C_ (Var aa) (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Call (exp, literal, lista)) (Var a) = false
  | equal_expa A_ B_ C_ (Var a) (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1, literal2, exp2)) (Var a) = false
  | equal_expa A_ B_ C_ (Var a) (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAcc (exp, literal1, literal2)) (Var a) = false
  | equal_expa A_ B_ C_ (Var a) (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (ALen exp) (Var a) = false
  | equal_expa A_ B_ C_ (Var a) (ALen exp) = false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3)) (Var a) = false
  | equal_expa A_ B_ C_ (Var a) (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (Var a) = false
  | equal_expa A_ B_ C_ (Var a) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (LAss (aa, exp)) (Var a) = false
  | equal_expa A_ B_ C_ (Var aa) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1a, literal, a, exp2a))
    (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (BinOp (exp1a, bop, exp2a))
    (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw exp) (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1a, exp2a)) (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (BinOp (exp1a, bop, exp2a)) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1a, exp2a, exp3)) (BinOp (exp1, bop, exp2)) =
    false
  | equal_expa A_ B_ C_ (BinOp (exp1a, bop, exp2a)) (Cond (exp1, exp2, exp3)) =
    false
  | equal_expa A_ B_ C_ (Seq (exp1a, exp2a)) (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (BinOp (exp1a, bop, exp2a)) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, exp))
    (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2))
    (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1a, exp2a))
    (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (BinOp (exp1a, bop, exp2a))
    (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, exp)) (BinOp (exp1, bop, exp2))
    = false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2)) (Block (a, ty, optiona, exp))
    = false
  | equal_expa A_ B_ C_ (Call (exp, literal, lista)) (BinOp (exp1, bop, exp2)) =
    false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2)) (Call (exp, literal, lista)) =
    false
  | equal_expa A_ B_ C_ (FAss (exp1a, literal1, literal2, exp2a))
    (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (BinOp (exp1a, bop, exp2a))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAcc (exp, literal1, literal2))
    (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2))
    (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (ALen exp) (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2)) (ALen exp) = false
  | equal_expa A_ B_ C_ (AAss (exp1a, exp2a, exp3)) (BinOp (exp1, bop, exp2)) =
    false
  | equal_expa A_ B_ C_ (BinOp (exp1a, bop, exp2a)) (AAss (exp1, exp2, exp3)) =
    false
  | equal_expa A_ B_ C_ (AAcc (exp1a, exp2a)) (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (BinOp (exp1a, bop, exp2a)) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (LAss (a, exp)) (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (Var a) (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2)) (Var a) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1, literal, a, exp2)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw exp) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, exp)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1, exp2)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, exp)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Call (exp, literal, lista)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1, literal2, exp2)) (Val vala) =
    false
  | equal_expa A_ B_ C_ (Val vala) (FAss (exp1, literal1, literal2, exp2)) =
    false
  | equal_expa A_ B_ C_ (FAcc (exp, literal1, literal2)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (ALen exp) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (ALen exp) = false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (LAss (a, exp)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (Var a) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (Var a) = false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2)) (Val vala) = false
  | equal_expa A_ B_ C_ (Val vala) (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1, literal, a, exp2))
    (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (exp, ty))
    (TryCatch (exp1, literal, a, exp2)) = false
  | equal_expa A_ B_ C_ (Throw expa) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (expa, ty)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (exp, ty)) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3)) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (exp, ty)) (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (exp, ty)) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, expa)) (InstanceOf (exp, ty))
    = false
  | equal_expa A_ B_ C_ (InstanceOf (expa, ty)) (InSynchronized (b, addr, exp))
    = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1, exp2)) (InstanceOf (exp, ty)) =
    false
  | equal_expa A_ B_ C_ (InstanceOf (exp, ty)) (Synchronized (b, exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (Block (a, tya, optiona, expa)) (InstanceOf (exp, ty)) =
    false
  | equal_expa A_ B_ C_ (InstanceOf (expa, tya)) (Block (a, ty, optiona, exp)) =
    false
  | equal_expa A_ B_ C_ (Call (expa, literal, lista)) (InstanceOf (exp, ty)) =
    false
  | equal_expa A_ B_ C_ (InstanceOf (expa, ty)) (Call (exp, literal, lista)) =
    false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1, literal2, exp2))
    (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (exp, ty))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAcc (expa, literal1, literal2)) (InstanceOf (exp, ty))
    = false
  | equal_expa A_ B_ C_ (InstanceOf (expa, ty)) (FAcc (exp, literal1, literal2))
    = false
  | equal_expa A_ B_ C_ (ALen expa) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (expa, ty)) (ALen exp) = false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3)) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (exp, ty)) (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (exp, ty)) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (LAss (a, expa)) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (expa, ty)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (Var a) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (exp, ty)) (Var a) = false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2)) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (exp, ty)) (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (Val vala) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (InstanceOf (exp, ty)) (Val vala) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1, literal, a, exp2)) (Cast (ty, exp)) =
    false
  | equal_expa A_ B_ C_ (Cast (ty, exp)) (TryCatch (exp1, literal, a, exp2)) =
    false
  | equal_expa A_ B_ C_ (Throw expa) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, expa)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, exp)) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3)) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, exp)) (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, exp)) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, expa)) (Cast (ty, exp)) =
    false
  | equal_expa A_ B_ C_ (Cast (ty, expa)) (InSynchronized (b, addr, exp)) =
    false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1, exp2)) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, exp)) (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Block (a, tya, optiona, expa)) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (tya, expa)) (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Call (expa, literal, lista)) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, expa)) (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1, literal2, exp2)) (Cast (ty, exp))
    = false
  | equal_expa A_ B_ C_ (Cast (ty, exp)) (FAss (exp1, literal1, literal2, exp2))
    = false
  | equal_expa A_ B_ C_ (FAcc (expa, literal1, literal2)) (Cast (ty, exp)) =
    false
  | equal_expa A_ B_ C_ (Cast (ty, expa)) (FAcc (exp, literal1, literal2)) =
    false
  | equal_expa A_ B_ C_ (ALen expa) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, expa)) (ALen exp) = false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3)) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, exp)) (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, exp)) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (LAss (a, expa)) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, expa)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (Var a) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, exp)) (Var a) = false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2)) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, exp)) (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (Val vala) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (ty, exp)) (Val vala) = false
  | equal_expa A_ B_ C_ (InstanceOf (expa, tya)) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (Cast (tya, expa)) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1, literal, a, exp2)) (NewArray (ty, exp))
    = false
  | equal_expa A_ B_ C_ (NewArray (ty, exp)) (TryCatch (exp1, literal, a, exp2))
    = false
  | equal_expa A_ B_ C_ (Throw expa) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, expa)) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, exp)) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3)) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, exp)) (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, exp)) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, expa)) (NewArray (ty, exp)) =
    false
  | equal_expa A_ B_ C_ (NewArray (ty, expa)) (InSynchronized (b, addr, exp)) =
    false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1, exp2)) (NewArray (ty, exp)) =
    false
  | equal_expa A_ B_ C_ (NewArray (ty, exp)) (Synchronized (b, exp1, exp2)) =
    false
  | equal_expa A_ B_ C_ (Block (a, tya, optiona, expa)) (NewArray (ty, exp)) =
    false
  | equal_expa A_ B_ C_ (NewArray (tya, expa)) (Block (a, ty, optiona, exp)) =
    false
  | equal_expa A_ B_ C_ (Call (expa, literal, lista)) (NewArray (ty, exp)) =
    false
  | equal_expa A_ B_ C_ (NewArray (ty, expa)) (Call (exp, literal, lista)) =
    false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1, literal2, exp2))
    (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, exp))
    (FAss (exp1, literal1, literal2, exp2)) = false
  | equal_expa A_ B_ C_ (FAcc (expa, literal1, literal2)) (NewArray (ty, exp)) =
    false
  | equal_expa A_ B_ C_ (NewArray (ty, expa)) (FAcc (exp, literal1, literal2)) =
    false
  | equal_expa A_ B_ C_ (ALen expa) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, expa)) (ALen exp) = false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3)) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, exp)) (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, exp)) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (LAss (a, expa)) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, expa)) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (Var a) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, exp)) (Var a) = false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2)) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, exp)) (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (Val vala) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, exp)) (Val vala) = false
  | equal_expa A_ B_ C_ (InstanceOf (expa, tya)) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (tya, expa)) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (Cast (tya, expa)) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (tya, expa)) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1, literala, a, exp2)) (New literal) =
    false
  | equal_expa A_ B_ C_ (New literala) (TryCatch (exp1, literal, a, exp2)) =
    false
  | equal_expa A_ B_ C_ (Throw exp) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (Throw exp) = false
  | equal_expa A_ B_ C_ (While (exp1, exp2)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (While (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Cond (exp1, exp2, exp3)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (Cond (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (Seq (exp1, exp2)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (Seq (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (InSynchronized (b, addr, exp)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (InSynchronized (b, addr, exp)) = false
  | equal_expa A_ B_ C_ (Synchronized (b, exp1, exp2)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (Synchronized (b, exp1, exp2)) = false
  | equal_expa A_ B_ C_ (Block (a, ty, optiona, exp)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (Block (a, ty, optiona, exp)) = false
  | equal_expa A_ B_ C_ (Call (exp, literala, lista)) (New literal) = false
  | equal_expa A_ B_ C_ (New literala) (Call (exp, literal, lista)) = false
  | equal_expa A_ B_ C_ (FAss (exp1, literal1, literal2, exp2)) (New literal) =
    false
  | equal_expa A_ B_ C_ (New literal) (FAss (exp1, literal1, literal2, exp2)) =
    false
  | equal_expa A_ B_ C_ (FAcc (exp, literal1, literal2)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (FAcc (exp, literal1, literal2)) = false
  | equal_expa A_ B_ C_ (ALen exp) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (ALen exp) = false
  | equal_expa A_ B_ C_ (AAss (exp1, exp2, exp3)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (AAss (exp1, exp2, exp3)) = false
  | equal_expa A_ B_ C_ (AAcc (exp1, exp2)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (AAcc (exp1, exp2)) = false
  | equal_expa A_ B_ C_ (LAss (a, exp)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (LAss (a, exp)) = false
  | equal_expa A_ B_ C_ (Var a) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (Var a) = false
  | equal_expa A_ B_ C_ (BinOp (exp1, bop, exp2)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (BinOp (exp1, bop, exp2)) = false
  | equal_expa A_ B_ C_ (Val vala) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (Val vala) = false
  | equal_expa A_ B_ C_ (InstanceOf (exp, ty)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (InstanceOf (exp, ty)) = false
  | equal_expa A_ B_ C_ (Cast (ty, exp)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (Cast (ty, exp)) = false
  | equal_expa A_ B_ C_ (NewArray (ty, exp)) (New literal) = false
  | equal_expa A_ B_ C_ (New literal) (NewArray (ty, exp)) = false
  | equal_expa A_ B_ C_ (TryCatch (exp1a, literala, aa, exp2a))
    (TryCatch (exp1, literal, a, exp2)) =
    equal_expa A_ B_ C_ exp1a exp1 andalso
      (((literala : string) = literal) andalso
        (HOL.eq A_ aa a andalso equal_expa A_ B_ C_ exp2a exp2))
  | equal_expa A_ B_ C_ (Throw expa) (Throw exp) = equal_expa A_ B_ C_ expa exp
  | equal_expa A_ B_ C_ (While (exp1a, exp2a)) (While (exp1, exp2)) =
    equal_expa A_ B_ C_ exp1a exp1 andalso equal_expa A_ B_ C_ exp2a exp2
  | equal_expa A_ B_ C_ (Cond (exp1a, exp2a, exp3a)) (Cond (exp1, exp2, exp3)) =
    equal_expa A_ B_ C_ exp1a exp1 andalso
      (equal_expa A_ B_ C_ exp2a exp2 andalso equal_expa A_ B_ C_ exp3a exp3)
  | equal_expa A_ B_ C_ (Seq (exp1a, exp2a)) (Seq (exp1, exp2)) =
    equal_expa A_ B_ C_ exp1a exp1 andalso equal_expa A_ B_ C_ exp2a exp2
  | equal_expa A_ B_ C_ (InSynchronized (ba, addra, expa))
    (InSynchronized (b, addr, exp)) =
    HOL.eq B_ ba b andalso
      (HOL.eq C_ addra addr andalso equal_expa A_ B_ C_ expa exp)
  | equal_expa A_ B_ C_ (Synchronized (ba, exp1a, exp2a))
    (Synchronized (b, exp1, exp2)) =
    HOL.eq B_ ba b andalso
      (equal_expa A_ B_ C_ exp1a exp1 andalso equal_expa A_ B_ C_ exp2a exp2)
  | equal_expa A_ B_ C_ (Block (aa, tya, optionaa, expa))
    (Block (a, ty, optiona, exp)) =
    HOL.eq A_ aa a andalso
      (Type.equal_tya tya ty andalso
        (Option.equal_optiona (Value.equal_val C_) optionaa optiona andalso
          equal_expa A_ B_ C_ expa exp))
  | equal_expa A_ B_ C_ (Call (expa, literala, listaa))
    (Call (exp, literal, lista)) =
    equal_expa A_ B_ C_ expa exp andalso
      (((literala : string) = literal) andalso
        List.equal_lista (equal_exp A_ B_ C_) listaa lista)
  | equal_expa A_ B_ C_ (FAss (exp1a, literal1a, literal2a, exp2a))
    (FAss (exp1, literal1, literal2, exp2)) =
    equal_expa A_ B_ C_ exp1a exp1 andalso
      (((literal1a : string) = literal1) andalso
        (((literal2a : string) = literal2) andalso
          equal_expa A_ B_ C_ exp2a exp2))
  | equal_expa A_ B_ C_ (FAcc (expa, literal1a, literal2a))
    (FAcc (exp, literal1, literal2)) =
    equal_expa A_ B_ C_ expa exp andalso
      (((literal1a : string) = literal1) andalso
        ((literal2a : string) = literal2))
  | equal_expa A_ B_ C_ (ALen expa) (ALen exp) = equal_expa A_ B_ C_ expa exp
  | equal_expa A_ B_ C_ (AAss (exp1a, exp2a, exp3a)) (AAss (exp1, exp2, exp3)) =
    equal_expa A_ B_ C_ exp1a exp1 andalso
      (equal_expa A_ B_ C_ exp2a exp2 andalso equal_expa A_ B_ C_ exp3a exp3)
  | equal_expa A_ B_ C_ (AAcc (exp1a, exp2a)) (AAcc (exp1, exp2)) =
    equal_expa A_ B_ C_ exp1a exp1 andalso equal_expa A_ B_ C_ exp2a exp2
  | equal_expa A_ B_ C_ (LAss (aa, expa)) (LAss (a, exp)) =
    HOL.eq A_ aa a andalso equal_expa A_ B_ C_ expa exp
  | equal_expa A_ B_ C_ (Var aa) (Var a) = HOL.eq A_ aa a
  | equal_expa A_ B_ C_ (BinOp (exp1a, bopa, exp2a)) (BinOp (exp1, bop, exp2)) =
    equal_expa A_ B_ C_ exp1a exp1 andalso
      (BinOp.equal_bop bopa bop andalso equal_expa A_ B_ C_ exp2a exp2)
  | equal_expa A_ B_ C_ (Val valaa) (Val vala) = Value.equal_vala C_ valaa vala
  | equal_expa A_ B_ C_ (InstanceOf (expa, tya)) (InstanceOf (exp, ty)) =
    equal_expa A_ B_ C_ expa exp andalso Type.equal_tya tya ty
  | equal_expa A_ B_ C_ (Cast (tya, expa)) (Cast (ty, exp)) =
    Type.equal_tya tya ty andalso equal_expa A_ B_ C_ expa exp
  | equal_expa A_ B_ C_ (NewArray (tya, expa)) (NewArray (ty, exp)) =
    Type.equal_tya tya ty andalso equal_expa A_ B_ C_ expa exp
  | equal_expa A_ B_ C_ (New literala) (New literal) =
    ((literala : string) = literal);

fun extRet2J e (ExternalCall.RetVal v) = Val v
  | extRet2J e (ExternalCall.RetExc a) = Throw (Val (Value.Addr a))
  | extRet2J e ExternalCall.RetStaySame = e;

fun is_Throws_i x =
  Predicate.sup_pred
    (case x of [] => Predicate.bot_pred | New _ :: _ => Predicate.bot_pred
      | NewArray (_, _) :: _ => Predicate.bot_pred
      | Cast (_, _) :: _ => Predicate.bot_pred
      | InstanceOf (_, _) :: _ => Predicate.bot_pred
      | Val _ :: _ => Predicate.bot_pred
      | BinOp (_, _, _) :: _ => Predicate.bot_pred
      | Var _ :: _ => Predicate.bot_pred
      | LAss (_, _) :: _ => Predicate.bot_pred
      | AAcc (_, _) :: _ => Predicate.bot_pred
      | AAss (_, _, _) :: _ => Predicate.bot_pred
      | ALen _ :: _ => Predicate.bot_pred
      | FAcc (_, _, _) :: _ => Predicate.bot_pred
      | FAss (_, _, _, _) :: _ => Predicate.bot_pred
      | Call (_, _, _) :: _ => Predicate.bot_pred
      | Block (_, _, _, _) :: _ => Predicate.bot_pred
      | Synchronized (_, _, _) :: _ => Predicate.bot_pred
      | InSynchronized (_, _, _) :: _ => Predicate.bot_pred
      | Seq (_, _) :: _ => Predicate.bot_pred
      | Cond (_, _, _) :: _ => Predicate.bot_pred
      | While (_, _) :: _ => Predicate.bot_pred
      | Throw (New _) :: _ => Predicate.bot_pred
      | Throw (NewArray (_, _)) :: _ => Predicate.bot_pred
      | Throw (Cast (_, _)) :: _ => Predicate.bot_pred
      | Throw (InstanceOf (_, _)) :: _ => Predicate.bot_pred
      | Throw (Val Value.Unit) :: _ => Predicate.bot_pred
      | Throw (Val Value.Null) :: _ => Predicate.bot_pred
      | Throw (Val (Value.Bool _)) :: _ => Predicate.bot_pred
      | Throw (Val (Value.Intg _)) :: _ => Predicate.bot_pred
      | Throw (Val (Value.Addr _)) :: _ => Predicate.single ()
      | Throw (BinOp (_, _, _)) :: _ => Predicate.bot_pred
      | Throw (Var _) :: _ => Predicate.bot_pred
      | Throw (LAss (_, _)) :: _ => Predicate.bot_pred
      | Throw (AAcc (_, _)) :: _ => Predicate.bot_pred
      | Throw (AAss (_, _, _)) :: _ => Predicate.bot_pred
      | Throw (ALen _) :: _ => Predicate.bot_pred
      | Throw (FAcc (_, _, _)) :: _ => Predicate.bot_pred
      | Throw (FAss (_, _, _, _)) :: _ => Predicate.bot_pred
      | Throw (Call (_, _, _)) :: _ => Predicate.bot_pred
      | Throw (Block (_, _, _, _)) :: _ => Predicate.bot_pred
      | Throw (Synchronized (_, _, _)) :: _ => Predicate.bot_pred
      | Throw (InSynchronized (_, _, _)) :: _ => Predicate.bot_pred
      | Throw (Seq (_, _)) :: _ => Predicate.bot_pred
      | Throw (Cond (_, _, _)) :: _ => Predicate.bot_pred
      | Throw (While (_, _)) :: _ => Predicate.bot_pred
      | Throw (Throw _) :: _ => Predicate.bot_pred
      | Throw (TryCatch (_, _, _, _)) :: _ => Predicate.bot_pred
      | TryCatch (_, _, _, _) :: _ => Predicate.bot_pred)
    (case x of [] => Predicate.bot_pred | New _ :: es => Predicate.bot_pred
      | NewArray (_, _) :: es => Predicate.bot_pred
      | Cast (_, _) :: es => Predicate.bot_pred
      | InstanceOf (_, _) :: es => Predicate.bot_pred
      | Val _ :: es =>
        Predicate.bind (is_Throws_i es) (fn () => Predicate.single ())
      | BinOp (_, _, _) :: es => Predicate.bot_pred
      | Var _ :: es => Predicate.bot_pred
      | LAss (_, _) :: es => Predicate.bot_pred
      | AAcc (_, _) :: es => Predicate.bot_pred
      | AAss (_, _, _) :: es => Predicate.bot_pred
      | ALen _ :: es => Predicate.bot_pred
      | FAcc (_, _, _) :: es => Predicate.bot_pred
      | FAss (_, _, _, _) :: es => Predicate.bot_pred
      | Call (_, _, _) :: es => Predicate.bot_pred
      | Block (_, _, _, _) :: es => Predicate.bot_pred
      | Synchronized (_, _, _) :: es => Predicate.bot_pred
      | InSynchronized (_, _, _) :: es => Predicate.bot_pred
      | Seq (_, _) :: es => Predicate.bot_pred
      | Cond (_, _, _) :: es => Predicate.bot_pred
      | While (_, _) :: es => Predicate.bot_pred
      | Throw _ :: es => Predicate.bot_pred
      | TryCatch (_, _, _, _) :: es => Predicate.bot_pred);

end; (*struct Expr*)

structure Fifo : sig
  val fifo_empty : unit -> 'a list * 'a list
  val fifo_appendl : 'a -> 'a list * 'a list -> 'a list * 'a list
  val fifo_appendr : 'a -> 'a list * 'a list -> 'a list * 'a list
  val fifo_isEmpty : 'a HOL.equal -> 'a list * 'a list -> bool
  val fifo_removel : 'a list * 'a list -> 'a * ('a list * 'a list)
end = struct

fun fifo_empty x = (fn _ => ([], [])) x;

fun fifo_appendl x f = let
                         val (e, d) = f;
                       in
                         (e, x :: d)
                       end;

fun fifo_appendr a f = (a :: Product_Type.fst f, Product_Type.snd f);

fun fifo_isEmpty A_ f =
  Product_Type.equal_proda (List.equal_list A_) (List.equal_list A_) f ([], []);

fun fifo_removel f =
  (case Product_Type.snd f
    of [] =>
      let
        val rp = List.rev (Product_Type.fst f);
      in
        (List.hd rp, ([], List.tl rp))
      end
    | a :: l => (a, (Product_Type.fst f, l)));

end; (*struct Fifo*)

structure Assoc_List : sig
  datatype ('a, 'b) assoc_list = Assoc_List of ('a * 'b) list
  val empty : ('a, 'b) assoc_list
  val impl_of : ('a, 'b) assoc_list -> ('a * 'b) list
  val lookup : 'a HOL.equal -> ('a, 'b) assoc_list -> 'a -> 'b option
  val update_with_aux :
    'b HOL.equal -> 'a -> 'b -> ('a -> 'a) -> ('b * 'a) list -> ('b * 'a) list
  val update_with :
    'a HOL.equal ->
      'b -> 'a -> ('b -> 'b) -> ('a, 'b) assoc_list -> ('a, 'b) assoc_list
  val update :
    'a HOL.equal -> 'a -> 'b -> ('a, 'b) assoc_list -> ('a, 'b) assoc_list
  val equal_assoc_lista :
    'a HOL.equal -> 'b HOL.equal ->
      ('a, 'b) assoc_list -> ('a, 'b) assoc_list -> bool
  val equal_assoc_list :
    'a HOL.equal -> 'b HOL.equal -> ('a, 'b) assoc_list HOL.equal
end = struct

datatype ('a, 'b) assoc_list = Assoc_List of ('a * 'b) list;

val empty : ('a, 'b) assoc_list = Assoc_List [];

fun impl_of (Assoc_List x) = x;

fun lookup A_ al = Map.map_of A_ (impl_of al);

fun update_with_aux B_ v k f [] = [(k, f v)]
  | update_with_aux B_ v k f (p :: ps) =
    (if HOL.eq B_ (Product_Type.fst p) k then (k, f (Product_Type.snd p)) :: ps
      else p :: update_with_aux B_ v k f ps);

fun update_with A_ v k f al =
  Assoc_List (update_with_aux A_ v k f (impl_of al));

fun update A_ k v = update_with A_ v k (fn _ => v);

fun equal_assoc_lista A_ B_ ala al =
  List.equal_lista (Product_Type.equal_prod A_ B_) (impl_of ala) (impl_of al);

fun equal_assoc_list A_ B_ = {equal = equal_assoc_lista A_ B_} :
  ('a, 'b) assoc_list HOL.equal;

end; (*struct Assoc_List*)

structure Foldi : sig
  val foldli : 'a list -> ('b -> bool) -> ('a -> 'b -> 'b) -> 'b -> 'b
end = struct

fun foldli [] c f sigma = sigma
  | foldli (x :: xs) c f sigma =
    (if c sigma then foldli xs c f (f x sigma) else sigma);

end; (*struct Foldi*)

structure Trie_Impl : sig
  datatype ('a, 'b) trie = Trie of 'b option * ('a * ('a, 'b) trie) list
  val empty : ('a, 'b) trie
  val lookup : 'a HOL.equal -> ('a, 'b) trie -> 'a list -> 'b option
  val update : 'a HOL.equal -> ('a, 'b) trie -> 'a list -> 'b -> ('a, 'b) trie
  val iteratei_postfixed :
    'a list ->
      ('a, 'b) trie -> ('c -> bool) -> ('a list * 'b -> 'c -> 'c) -> 'c -> 'c
  val iteratei :
    ('a, 'b) trie -> ('c -> bool) -> ('a list * 'b -> 'c -> 'c) -> 'c -> 'c
  val equal_trie : 'a HOL.equal -> 'b HOL.equal -> ('a, 'b) trie HOL.equal
  val equal_triea :
    'a HOL.equal -> 'b HOL.equal -> ('a, 'b) trie -> ('a, 'b) trie -> bool
end = struct

datatype ('a, 'b) trie = Trie of 'b option * ('a * ('a, 'b) trie) list;

val empty : ('a, 'b) trie = Trie (NONE, []);

fun lookup A_ (Trie (vo, uu)) [] = vo
  | lookup A_ (Trie (uv, ts)) (k :: ks) =
    (case Map.map_of A_ ts k of NONE => NONE | SOME t => lookup A_ t ks);

fun update A_ (Trie (vo, ts)) [] v = Trie (SOME v, ts)
  | update A_ (Trie (vo, ts)) (k :: ks) v =
    Trie (vo, Assoc_List.update_with_aux A_ (Trie (NONE, [])) k
                (fn t => update A_ t ks v) ts);

fun iteratei_postfixed ks (Trie (vo, ts)) c f sigma =
  (if c sigma
    then Foldi.foldli ts c (fn (k, t) => iteratei_postfixed (k :: ks) t c f)
           (case vo of NONE => sigma | SOME v => f (ks, v) sigma)
    else sigma);

fun iteratei t c f sigma = iteratei_postfixed [] t c f sigma;

fun equal_trie A_ B_ = {equal = equal_triea A_ B_} : ('a, 'b) trie HOL.equal
and equal_triea A_ B_ (Trie (optionaa, listaa)) (Trie (optiona, lista)) =
  Option.equal_optiona B_ optionaa optiona andalso
    List.equal_lista (Product_Type.equal_prod A_ (equal_trie A_ B_)) listaa
      lista;

end; (*struct Trie_Impl*)

structure Trie : sig
  datatype ('a, 'b) trie = Trie of ('a, 'b) Trie_Impl.trie
  val empty : ('a, 'b) trie
  val impl_of : ('a, 'b) trie -> ('a, 'b) Trie_Impl.trie
  val lookup : 'a HOL.equal -> ('a, 'b) trie -> 'a list -> 'b option
  val update : 'a HOL.equal -> 'a list -> 'b -> ('a, 'b) trie -> ('a, 'b) trie
  val iteratei :
    ('a, 'b) trie -> ('c -> bool) -> ('a list * 'b -> 'c -> 'c) -> 'c -> 'c
  val equal_trie :
    'a HOL.equal -> 'b HOL.equal -> ('a, 'b) trie -> ('a, 'b) trie -> bool
end = struct

datatype ('a, 'b) trie = Trie of ('a, 'b) Trie_Impl.trie;

val empty : ('a, 'b) trie = Trie Trie_Impl.empty;

fun impl_of (Trie x) = x;

fun lookup A_ t = Trie_Impl.lookup A_ (impl_of t);

fun update A_ ks v t = Trie (Trie_Impl.update A_ (impl_of t) ks v);

fun iteratei t c f =
  Trie_Impl.iteratei (impl_of t) c (fn (ks, v) => f (List.rev ks, v));

fun equal_trie A_ B_ ta t =
  Trie_Impl.equal_triea A_ B_ (impl_of ta) (impl_of t);

end; (*struct Trie*)

structure State : sig
  val hp : 'a * 'b -> 'a
  val lcl : 'a * 'b -> 'b
end = struct

fun hp x = Product_Type.fst x;

fun lcl x = Product_Type.snd x;

end; (*struct State*)

structure DefAss : sig
  val hyperDiff1 : 'a HOL.equal -> 'a Set.set option -> 'a -> 'a Set.set option
  val hyperInt :
    'a HOL.equal -> 'a Set.set option -> 'a Set.set option -> 'a Set.set option
  val hyperUn : 'a Set.set option -> 'a Set.set option -> 'a Set.set option
  val a : 'a HOL.equal -> ('a, 'b, 'c) Expr.exp -> 'a Set.set option
  val a_s : 'a HOL.equal -> ('a, 'b, 'c) Expr.exp list -> 'a Set.set option
  val hyper_isin : 'a HOL.equal -> 'a -> 'a Set.set option -> bool
  val d : 'a HOL.equal -> ('a, 'b, 'c) Expr.exp -> 'a Set.set option -> bool
  val d_s :
    'a HOL.equal -> ('a, 'b, 'c) Expr.exp list -> 'a Set.set option -> bool
end = struct

fun hyperDiff1 A_ aa a =
  (case aa of NONE => NONE
    | SOME aaa => SOME (Set.minus_set A_ aaa (Set.insert a Set.bot_set)));

fun hyperInt A_ a b =
  (case a of NONE => b
    | SOME aa =>
      (case b of NONE => SOME aa | SOME ba => SOME (Set.inf_set A_ aa ba)));

fun hyperUn a b =
  (case a of NONE => NONE
    | SOME aa => (case b of NONE => NONE | SOME ba => SOME (Set.union aa ba)));

fun a A_ (Expr.New c) = SOME Set.bot_set
  | a A_ (Expr.NewArray (t, e)) = a A_ e
  | a A_ (Expr.Cast (c, e)) = a A_ e
  | a A_ (Expr.InstanceOf (e, t)) = a A_ e
  | a A_ (Expr.Val v) = SOME Set.bot_set
  | a A_ (Expr.BinOp (e_1, bop, e_2)) = hyperUn (a A_ e_1) (a A_ e_2)
  | a A_ (Expr.Var v) = SOME Set.bot_set
  | a A_ (Expr.LAss (v, e)) = hyperUn (SOME (Set.insert v Set.bot_set)) (a A_ e)
  | a A_ (Expr.AAcc (aa, i)) = hyperUn (a A_ aa) (a A_ i)
  | a A_ (Expr.AAss (aa, i, e)) = hyperUn (hyperUn (a A_ aa) (a A_ i)) (a A_ e)
  | a A_ (Expr.ALen aa) = a A_ aa
  | a A_ (Expr.FAcc (e, f, d)) = a A_ e
  | a A_ (Expr.FAss (e_1, f, d, e_2)) = hyperUn (a A_ e_1) (a A_ e_2)
  | a A_ (Expr.Call (e, m, es)) = hyperUn (a A_ e) (a_s A_ es)
  | a A_ (Expr.Block (v, t, vo, e)) = hyperDiff1 A_ (a A_ e) v
  | a A_ (Expr.Synchronized (v, oa, e)) = hyperUn (a A_ oa) (a A_ e)
  | a A_ (Expr.InSynchronized (v, aa, e)) = a A_ e
  | a A_ (Expr.Seq (e_1, e_2)) = hyperUn (a A_ e_1) (a A_ e_2)
  | a A_ (Expr.Cond (e, e_1, e_2)) =
    hyperUn (a A_ e) (hyperInt A_ (a A_ e_1) (a A_ e_2))
  | a A_ (Expr.While (b, e)) = a A_ b
  | a A_ (Expr.Throw e) = NONE
  | a A_ (Expr.TryCatch (e_1, c, v, e_2)) =
    hyperInt A_ (a A_ e_1) (hyperDiff1 A_ (a A_ e_2) v)
and a_s A_ [] = SOME Set.bot_set
  | a_s A_ (e :: es) = hyperUn (a A_ e) (a_s A_ es);

fun hyper_isin A_ aa a =
  (case a of NONE => true | SOME ab => Set.member A_ aa ab);

fun d A_ (Expr.New c) a = true
  | d A_ (Expr.NewArray (t, e)) a = d A_ e a
  | d A_ (Expr.Cast (c, e)) a = d A_ e a
  | d A_ (Expr.InstanceOf (e, t)) a = d A_ e a
  | d A_ (Expr.Val v) a = true
  | d A_ (Expr.BinOp (e_1, bop, e_2)) aa =
    d A_ e_1 aa andalso d A_ e_2 (hyperUn aa (a A_ e_1))
  | d A_ (Expr.Var v) a = hyper_isin A_ v a
  | d A_ (Expr.LAss (v, e)) a = d A_ e a
  | d A_ (Expr.AAcc (aaa, i)) aa =
    d A_ aaa aa andalso d A_ i (hyperUn aa (a A_ aaa))
  | d A_ (Expr.AAss (aaa, i, e)) aa =
    d A_ aaa aa andalso
      (d A_ i (hyperUn aa (a A_ aaa)) andalso
        d A_ e (hyperUn (hyperUn aa (a A_ aaa)) (a A_ i)))
  | d A_ (Expr.ALen aa) a = d A_ aa a
  | d A_ (Expr.FAcc (e, f, da)) a = d A_ e a
  | d A_ (Expr.FAss (e_1, f, da, e_2)) aa =
    d A_ e_1 aa andalso d A_ e_2 (hyperUn aa (a A_ e_1))
  | d A_ (Expr.Call (e, m, es)) aa =
    d A_ e aa andalso d_s A_ es (hyperUn aa (a A_ e))
  | d A_ (Expr.Block (v, t, vo, e)) a =
    (if Option.is_none vo then d A_ e (hyperDiff1 A_ a v)
      else d A_ e (hyperUn a (SOME (Set.insert v Set.bot_set))))
  | d A_ (Expr.Synchronized (v, oa, e)) aa =
    d A_ oa aa andalso d A_ e (hyperUn aa (a A_ oa))
  | d A_ (Expr.InSynchronized (v, aa, e)) a = d A_ e a
  | d A_ (Expr.Seq (e_1, e_2)) aa =
    d A_ e_1 aa andalso d A_ e_2 (hyperUn aa (a A_ e_1))
  | d A_ (Expr.Cond (e, e_1, e_2)) aa =
    d A_ e aa andalso
      (d A_ e_1 (hyperUn aa (a A_ e)) andalso d A_ e_2 (hyperUn aa (a A_ e)))
  | d A_ (Expr.While (e, c)) aa = d A_ e aa andalso d A_ c (hyperUn aa (a A_ e))
  | d A_ (Expr.Throw e) a = d A_ e a
  | d A_ (Expr.TryCatch (e_1, c, v, e_2)) a =
    d A_ e_1 a andalso d A_ e_2 (hyperUn a (SOME (Set.insert v Set.bot_set)))
and d_s A_ [] a = true
  | d_s A_ (e :: es) aa = d A_ e aa andalso d_s A_ es (hyperUn aa (a A_ e));

end; (*struct DefAss*)

structure FWLock : sig
  val may_lock : 'a HOL.equal -> ('a * Arith.nat) option -> 'a -> bool
  val unlock_lock : ('a * Arith.nat) option -> ('a * Arith.nat) option
  val release_all :
    'a HOL.equal -> ('a * Arith.nat) option -> 'a -> ('a * Arith.nat) option
  val lock_lock : ('a * Arith.nat) option -> 'a -> ('a * Arith.nat) option
  val upd_lock :
    'a HOL.equal ->
      ('a * Arith.nat) option ->
        'a -> FWState.lock_action -> ('a * Arith.nat) option
  val has_locks : 'a HOL.equal -> ('a * Arith.nat) option -> 'a -> Arith.nat
  val upd_locks :
    'a HOL.equal ->
      ('a * Arith.nat) option ->
        'a -> FWState.lock_action list -> ('a * Arith.nat) option
  val acquire_locks :
    ('a * Arith.nat) option -> 'a -> Arith.nat -> ('a * Arith.nat) option
  val lock_action_ok :
    'a HOL.equal -> ('a * Arith.nat) option -> 'a -> FWState.lock_action -> bool
  val lock_actions_ok :
    'a HOL.equal ->
      ('a * Arith.nat) option -> 'a -> FWState.lock_action list -> bool
end = struct

fun may_lock A_ l t = (case l of NONE => true | SOME (ta, _) => HOL.eq A_ t ta);

fun unlock_lock NONE = NONE
  | unlock_lock (SOME (t, n)) =
    (if Arith.equal_nata n Arith.zero_nat then NONE
      else SOME (t, Arith.minus_nat n Arith.one_nat));

fun release_all A_ NONE t = NONE
  | release_all A_ (SOME (ta, n)) t =
    (if HOL.eq A_ t ta then NONE else SOME (ta, n));

fun lock_lock NONE t = SOME (t, Arith.zero_nat)
  | lock_lock (SOME (ta, n)) t = SOME (ta, Arith.suc n);

fun upd_lock A_ l t FWState.Lock = lock_lock l t
  | upd_lock A_ l t FWState.Unlock = unlock_lock l
  | upd_lock A_ l t FWState.UnlockFail = l
  | upd_lock A_ l t FWState.ReleaseAcquire = release_all A_ l t;

fun has_locks A_ NONE t = Arith.zero_nat
  | has_locks A_ (SOME (ta, n)) t =
    (if HOL.eq A_ t ta then Arith.suc n else Arith.zero_nat);

fun upd_locks A_ l t [] = l
  | upd_locks A_ la t (l :: ls) = upd_locks A_ (upd_lock A_ la t l) t ls;

fun acquire_locks l t m =
  (if Arith.equal_nata m Arith.zero_nat then l
    else acquire_locks (lock_lock l t) t (Arith.minus_nat m Arith.one_nat));

fun lock_action_ok A_ l t FWState.Lock = may_lock A_ l t
  | lock_action_ok A_ l t FWState.Unlock =
    Arith.less_nat Arith.zero_nat (has_locks A_ l t)
  | lock_action_ok A_ l t FWState.UnlockFail =
    not (Arith.less_nat Arith.zero_nat (has_locks A_ l t))
  | lock_action_ok A_ l t FWState.ReleaseAcquire = true;

fun lock_actions_ok A_ l t [] = true
  | lock_actions_ok A_ la t (l :: ls) =
    lock_action_ok A_ la t l andalso
      lock_actions_ok A_ (upd_lock A_ la t l) t ls;

end; (*struct FWLock*)

structure FWWait : sig
  val waiting : 'a FWState.wait_set_status option -> bool
end = struct

fun waiting (SOME (FWState.InWS w)) = true
  | waiting (SOME (FWState.PostWS w)) = false
  | waiting NONE = false;

end; (*struct FWWait*)

structure Random : sig
  val log : Arith.natural -> Arith.natural -> Arith.natural
  val minus_shift :
    Arith.natural -> Arith.natural -> Arith.natural -> Arith.natural
  val next :
    Arith.natural * Arith.natural ->
      Arith.natural * (Arith.natural * Arith.natural)
  val iterate : Arith.natural -> ('a -> 'b -> 'a * 'b) -> 'a -> 'b -> 'a * 'b
  val range :
    Arith.natural ->
      Arith.natural * Arith.natural ->
        Arith.natural * (Arith.natural * Arith.natural)
  val select :
    'a list ->
      Arith.natural * Arith.natural -> 'a * (Arith.natural * Arith.natural)
end = struct

fun log b i =
  (if Arith.less_eq_natural b Arith.one_natural orelse Arith.less_natural i b
    then Arith.one_natural
    else Arith.plus_natural Arith.one_natural (log b (Arith.div_natural i b)));

fun minus_shift r k l =
  (if Arith.less_natural k l then Arith.minus_natural (Arith.plus_natural r k) l
    else Arith.minus_natural k l);

fun next (v, w) =
  let
    val k = Arith.div_natural v (Arith.natural_of_integer (53668 : IntInf.int));
    val va =
      minus_shift (Arith.natural_of_integer (2147483563 : IntInf.int))
        (Arith.times_natural
          (Arith.mod_natural v (Arith.natural_of_integer (53668 : IntInf.int)))
          (Arith.natural_of_integer (40014 : IntInf.int)))
        (Arith.times_natural k (Arith.natural_of_integer (12211 : IntInf.int)));
    val l = Arith.div_natural w (Arith.natural_of_integer (52774 : IntInf.int));
    val wa =
      minus_shift (Arith.natural_of_integer (2147483399 : IntInf.int))
        (Arith.times_natural
          (Arith.mod_natural w (Arith.natural_of_integer (52774 : IntInf.int)))
          (Arith.natural_of_integer (40692 : IntInf.int)))
        (Arith.times_natural l (Arith.natural_of_integer (3791 : IntInf.int)));
    val z =
      Arith.plus_natural
        (minus_shift (Arith.natural_of_integer (2147483562 : IntInf.int)) va
          (Arith.plus_natural wa Arith.one_natural))
        Arith.one_natural;
  in
    (z, (va, wa))
  end;

fun iterate k f x =
  (if Arith.equal_natural k Arith.zero_natural then (fn a => (x, a))
    else Product_Type.scomp (f x)
           (iterate (Arith.minus_natural k Arith.one_natural) f));

fun range k =
  Product_Type.scomp
    (iterate (log (Arith.natural_of_integer (2147483561 : IntInf.int)) k)
      (fn l =>
        Product_Type.scomp next
          (fn v =>
            (fn a =>
              (Arith.plus_natural v
                 (Arith.times_natural l
                   (Arith.natural_of_integer (2147483561 : IntInf.int))),
                a))))
      Arith.one_natural)
    (fn v => (fn a => (Arith.mod_natural v k, a)));

fun select xs =
  Product_Type.scomp (range (Arith.natural_of_nat (List.size_list xs)))
    (fn k => (fn a => (List.nth xs (Arith.nat_of_natural k), a)));

end; (*struct Random*)

structure Coinductive_List : sig
  datatype 'a llist = Lazy_llist of (unit -> ('a * 'a llist) option)
  val force : 'a llist -> ('a * 'a llist) option
  val lNil : 'a llist
  val lmap : ('a -> 'b) -> 'a llist -> 'b llist
  val lappend : 'a llist -> 'a llist -> 'a llist
  val lconcat : 'a llist llist -> 'a llist
  val lfilter : ('a -> bool) -> 'a llist -> 'a llist
  val llist_of : 'a list -> 'a llist
  val llist_case : 'a -> ('b -> 'b llist -> 'a) -> 'b llist -> 'a
end = struct

datatype 'a llist = Lazy_llist of (unit -> ('a * 'a llist) option);

fun force (Lazy_llist xs) = xs ();

val lNil : 'a llist = Lazy_llist (fn _ => NONE);

fun lmap f (Lazy_llist xs) =
  Lazy_llist (fn _ => Option.map (Product_Type.map_pair f (lmap f)) (xs ()));

fun lappend (Lazy_llist xs) ys =
  Lazy_llist
    (fn _ =>
      (case xs () of NONE => force ys
        | SOME (x, xsa) => SOME (x, lappend xsa ys)));

fun lconcat (Lazy_llist xss) =
  Lazy_llist
    (fn _ =>
      (case xss () of NONE => NONE
        | SOME (xs, xssa) => force (lappend xs (lconcat xssa))));

fun lfilter p (Lazy_llist xs) =
  Lazy_llist
    (fn _ =>
      (case xs () of NONE => NONE
        | SOME (x, ys) =>
          (if p x then SOME (x, lfilter p ys) else force (lfilter p ys))));

fun llist_of [] = lNil
  | llist_of (x :: xs) = Lazy_llist (fn _ => SOME (x, llist_of xs));

fun llist_case n c (Lazy_llist xs) =
  (case xs () of NONE => n | SOME (a, b) => c a b);

end; (*struct Coinductive_List*)

structure TLList : sig
  datatype ('a, 'b) tllist =
    Lazy_tllist of (unit -> (('a * ('a, 'b) tllist), 'b) Sum_Type.sum)
  val tNil : 'a -> ('b, 'a) tllist
  val tmap : ('a -> 'b) -> ('c -> 'd) -> ('a, 'c) tllist -> ('b, 'd) tllist
  val force : ('a, 'b) tllist -> (('a * ('a, 'b) tllist), 'b) Sum_Type.sum
  val tfilter :
    (unit -> 'a) -> ('b -> bool) -> ('b, 'a) tllist -> ('b, 'a) tllist
  val terminal : ('a, 'b) tllist -> 'b
  val tllist_case :
    ('a -> 'b) -> ('c -> ('c, 'a) tllist -> 'b) -> ('c, 'a) tllist -> 'b
  val llist_of_tllist : ('a, 'b) tllist -> 'a Coinductive_List.llist
end = struct

datatype ('a, 'b) tllist =
  Lazy_tllist of (unit -> (('a * ('a, 'b) tllist), 'b) Sum_Type.sum);

fun tNil b = Lazy_tllist (fn _ => Sum_Type.Inr b);

fun tmap f g (Lazy_tllist xs) =
  Lazy_tllist
    (fn _ =>
      (case xs () of Sum_Type.Inl (x, ys) => Sum_Type.Inl (f x, tmap f g ys)
        | Sum_Type.Inr b => Sum_Type.Inr (g b)));

fun force (Lazy_tllist xs) = xs ();

fun tfilter b p (Lazy_tllist xs) =
  Lazy_tllist
    (fn _ =>
      (case xs ()
        of Sum_Type.Inl (x, xsa) =>
          (if p x then Sum_Type.Inl (x, tfilter b p xsa)
            else force (tfilter b p xsa))
        | Sum_Type.Inr a => Sum_Type.Inr a));

fun terminal (Lazy_tllist xs) =
  (case xs () of Sum_Type.Inl (_, a) => terminal a | Sum_Type.Inr b => b);

fun tllist_case n c (Lazy_tllist xs) =
  (case xs () of Sum_Type.Inl (a, b) => c a b | Sum_Type.Inr a => n a);

fun llist_of_tllist (Lazy_tllist xs) =
  Coinductive_List.Lazy_llist
    (fn _ =>
      (case xs () of Sum_Type.Inl (x, ys) => SOME (x, llist_of_tllist ys)
        | Sum_Type.Inr _ => NONE));

end; (*struct TLList*)

structure RBT_add : sig
  val rm_iterateoi :
    ('a, 'b) RBT_Impl.rbt -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
  val rm_reverse_iterateoi :
    ('a, 'b) RBT_Impl.rbt -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
end = struct

fun rm_iterateoi RBT_Impl.Empty c f sigma = sigma
  | rm_iterateoi (RBT_Impl.Branch (col, l, k, v, r)) c f sigma =
    (if c sigma
      then let
             val sigmaa = rm_iterateoi l c f sigma;
           in
             (if c sigmaa then rm_iterateoi r c f (f (k, v) sigmaa) else sigmaa)
           end
      else sigma);

fun rm_reverse_iterateoi RBT_Impl.Empty c f sigma = sigma
  | rm_reverse_iterateoi (RBT_Impl.Branch (col, l, k, v, r)) c f sigma =
    (if c sigma
      then let
             val sigmaa = rm_reverse_iterateoi r c f sigma;
           in
             (if c sigmaa then rm_reverse_iterateoi l c f (f (k, v) sigmaa)
               else sigmaa)
           end
      else sigma);

end; (*struct RBT_add*)

structure While_Combinator : sig
  val whilea : ('a -> bool) -> ('a -> 'a) -> 'a -> 'a
end = struct

fun whilea b c s = (if b s then whilea b c (c s) else s);

end; (*struct While_Combinator*)

structure SemiType : sig
  val widen1p_i_i_o : 'a TypeRel.prog -> Type.ty -> Type.ty Predicate.pred
  val super : 'a TypeRel.prog -> Type.ty -> Type.ty
  val sup : 'a TypeRel.prog -> Type.ty -> Type.ty -> Type.ty Err.err
end = struct

fun widen1p_i_i_o x xa =
  (case xa of Type.Void => Predicate.bot_pred
    | Type.Boolean => Predicate.bot_pred | Type.Integer => Predicate.bot_pred
    | Type.Nt => Predicate.bot_pred
    | Type.Class c =>
      Predicate.bind (TypeRel.subcls1_i_i_o x c)
        (fn xb => Predicate.single (Type.Class xb))
    | Type.Array t =>
      Predicate.sup_pred
        (case t of Type.Void => Predicate.bot_pred
          | Type.Boolean => Predicate.bot_pred
          | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
          | Type.Class xb =>
            (if ((xb : string) = Type.object)
              then Predicate.single (Type.Class Type.object)
              else Predicate.bot_pred)
          | Type.Array _ => Predicate.bot_pred)
        (Predicate.sup_pred
          (case t of Type.Void => Predicate.bot_pred
            | Type.Boolean => Predicate.bot_pred
            | Type.Integer => Predicate.single (Type.Class Type.object)
            | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
            | Type.Array _ => Predicate.bot_pred)
          (Predicate.sup_pred
            (case t of Type.Void => Predicate.bot_pred
              | Type.Boolean => Predicate.single (Type.Class Type.object)
              | Type.Integer => Predicate.bot_pred
              | Type.Nt => Predicate.bot_pred
              | Type.Class _ => Predicate.bot_pred
              | Type.Array _ => Predicate.bot_pred)
            (Predicate.sup_pred
              (case t of Type.Void => Predicate.single (Type.Class Type.object)
                | Type.Boolean => Predicate.bot_pred
                | Type.Integer => Predicate.bot_pred
                | Type.Nt => Predicate.bot_pred
                | Type.Class _ => Predicate.bot_pred
                | Type.Array _ => Predicate.bot_pred)
              (Predicate.bind
                (Predicate.if_pred
                  (not (Type.equal_tya (Type.ground_type t) Type.Nt)))
                (fn () =>
                  Predicate.bind (widen1p_i_i_o x t)
                    (fn xb => Predicate.single (Type.Array xb))))))));

fun super p (Type.Array Type.Integer) = Type.Class Type.object
  | super p (Type.Array Type.Boolean) = Type.Class Type.object
  | super p (Type.Array Type.Void) = Type.Class Type.object
  | super p (Type.Array (Type.Class c)) =
    (if ((c : string) = Type.object) then Type.Class Type.object
      else Type.Array (super p (Type.Class c)))
  | super p (Type.Array (Type.Array t)) = Type.Array (super p (Type.Array t))
  | super p (Type.Class c) =
    Type.Class (Product_Type.fst (Option.the (TypeRel.class p c)));

fun sup p t u =
  (if Type.is_refT t andalso Type.is_refT u
    then Err.Ok
           (if Type.equal_tya u Type.Nt then t
             else (if Type.equal_tya t Type.Nt then u
                    else While_Combinator.whilea
                           (fn z =>
                             not (Predicate.holds
                                   (Transitive_Closure_Table.rtrancl_tab_FioB_i_i_i
                                     Type.equal_ty (widen1p_i_i_o p) [] t z)))
                           (super p) u))
    else (if Type.equal_tya t u then Err.Ok t else Err.Err));

end; (*struct SemiType*)

structure TrieMapImpl : sig
  val iteratei_bmap_op_list_it_tm_basic_ops :
    ('a, 'b) Trie.trie -> ('c -> bool) -> ('a list * 'b -> 'c -> 'c) -> 'c -> 'c
  val g_to_list_tm_basic_ops : ('a, 'b) Trie.trie -> ('a list * 'b) list
end = struct

fun iteratei_bmap_op_list_it_tm_basic_ops s = Trie.iteratei s;

fun g_to_list_tm_basic_ops m =
  iteratei_bmap_op_list_it_tm_basic_ops m (fn _ => true)
    (fn a => fn b => a :: b) [];

end; (*struct TrieMapImpl*)

structure RBTMapImpl : sig
  val rev_iterateoi_bmap_op_rev_list_it_rm_basic_ops :
    'a Orderings.linorder ->
      ('a, 'b) RBT.rbt -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
  val g_max_rm_basic_ops :
    'a Orderings.linorder ->
      ('a, 'b) RBT.rbt -> ('a * 'b -> bool) -> ('a * 'b) option
  val iteratei_bmap_op_list_it_rm_basic_ops :
    'a Orderings.linorder ->
      ('a, 'b) RBT.rbt -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
  val g_sel_rm_basic_ops :
    'a Orderings.linorder ->
      ('a, 'b) RBT.rbt -> ('a * 'b -> bool) -> ('a * 'b) option
  val g_to_list_rm_basic_ops :
    'a Orderings.linorder -> ('a, 'b) RBT.rbt -> ('a * 'b) list
  val iteratei_map_op_list_it_rm_ops :
    'a Orderings.linorder ->
      ('a, 'b) RBT.rbt -> ('c -> bool) -> ('a * 'b -> 'c -> 'c) -> 'c -> 'c
end = struct

fun rev_iterateoi_bmap_op_rev_list_it_rm_basic_ops A_ s =
  RBT_add.rm_reverse_iterateoi (RBT.impl_of A_ s);

fun g_max_rm_basic_ops A_ m p =
  rev_iterateoi_bmap_op_rev_list_it_rm_basic_ops A_ m Option.is_none
    (fn x => fn _ => (if p x then SOME x else NONE)) NONE;

fun iteratei_bmap_op_list_it_rm_basic_ops A_ s =
  RBT_add.rm_iterateoi (RBT.impl_of A_ s);

fun g_sel_rm_basic_ops A_ m p =
  iteratei_bmap_op_list_it_rm_basic_ops A_ m Option.is_none
    (fn x => fn _ => (if p x then SOME x else NONE)) NONE;

fun g_to_list_rm_basic_ops A_ m =
  iteratei_bmap_op_list_it_rm_basic_ops A_ m (fn _ => true)
    (fn a => fn b => a :: b) [];

fun iteratei_map_op_list_it_rm_ops A_ s =
  RBT_add.rm_iterateoi (RBT.impl_of A_ s);

end; (*struct RBTMapImpl*)

structure ToString : sig
  type 'a toString
  val toString : 'a toString -> 'a -> string
  val digit_toString : Arith.int -> string
  val int_toString : Arith.int -> string list
  val toString_int : Arith.int -> string
  val toString_word : 'a Type_Length.len -> 'a Word.word -> string
  val toString_bool : bool -> string
  val toString_vala : 'a toString -> 'a Value.vala -> string
  val toString_bop : BinOp.bop -> string
  val toString_literala : string -> string
  val toString_ty : Type.ty -> string
  val toString_expa :
    'a toString -> 'b toString -> 'c toString -> ('a, 'b, 'c) Expr.exp -> string
  val toString_exp :
    'a toString -> 'b toString -> 'c toString -> ('a, 'b, 'c) Expr.exp toString
  val toString_funa : ('a -> 'b) -> string
  val toString_fun : ('a -> 'b) toString
  val toString_nata : Arith.nat -> string
  val toString_nat : Arith.nat toString
  val toString_val : 'a toString -> 'a Value.vala toString
  val toString_chara : char -> string
  val toString_char : char toString
  val list_toString : 'a toString -> string -> 'a list -> string list
  val toString_lista : 'a toString -> 'a list -> string
  val toString_list : 'a toString -> ('a list) toString
  val toString_proda : 'a toString -> 'b toString -> 'a * 'b -> string
  val toString_prod : 'a toString -> 'b toString -> ('a * 'b) toString
  val toString_rbt :
    'a Orderings.linorder * 'a toString -> 'b toString ->
      ('a, 'b) RBT.rbt -> string
  val toString_unita : unit -> string
  val toString_unit : unit toString
  val toString_trie : 'a toString -> 'b toString -> ('a, 'b) Trie.trie -> string
  val toString_finfuna :
    'a Set.card_UNIV * 'a HOL.equal * 'a Orderings.linorder * 'a toString ->
      'b HOL.equal * 'b toString -> ('a, 'b) FinFun.finfun -> string
  val toString_finfun :
    'a Set.card_UNIV * 'a HOL.equal * 'a Orderings.linorder * 'a toString ->
      'b HOL.equal * 'b toString -> ('a, 'b) FinFun.finfun toString
  val toString_htype : Type.htype -> string
  val toString_literal : string toString
  val toString_addr_loc : Heap.addr_loc -> string
  val toString_obs_eventa :
    'a toString -> 'b toString -> ('a, 'b) Observable_Events.obs_event -> string
  val toString_obs_event :
    'a toString -> 'b toString -> ('a, 'b) Observable_Events.obs_event toString
  val toString_assoc_lista :
    'a toString -> 'b toString -> ('a, 'b) Assoc_List.assoc_list -> string
  val toString_assoc_list :
    'a toString -> 'b toString -> ('a, 'b) Assoc_List.assoc_list toString
end = struct

type 'a toString = {toString : 'a -> string};
val toString = #toString : 'a toString -> 'a -> string;

fun digit_toString k =
  (if Arith.equal_int k Arith.zero_inta then "0"
    else (if Arith.equal_int k (Arith.Int_of_integer (1 : IntInf.int)) then "1"
           else (if Arith.equal_int k (Arith.Int_of_integer (2 : IntInf.int))
                  then "2"
                  else (if Arith.equal_int k
                             (Arith.Int_of_integer (3 : IntInf.int))
                         then "3"
                         else (if Arith.equal_int k
                                    (Arith.Int_of_integer (4 : IntInf.int))
                                then "4"
                                else (if Arith.equal_int k
   (Arith.Int_of_integer (5 : IntInf.int))
                                       then "5"
                                       else (if Arith.equal_int k
          (Arith.Int_of_integer (6 : IntInf.int))
      then "6"
      else (if Arith.equal_int k (Arith.Int_of_integer (7 : IntInf.int))
             then "7"
             else (if Arith.equal_int k (Arith.Int_of_integer (8 : IntInf.int))
                    then "8"
                    else (if Arith.equal_int k
                               (Arith.Int_of_integer (9 : IntInf.int))
                           then "9" else (raise Fail "undefined")))))))))));

fun int_toString n =
  (if Arith.less_int n Arith.zero_inta
    then "-" :: int_toString (Arith.uminus_int n)
    else (if Arith.less_int n (Arith.Int_of_integer (10 : IntInf.int))
           then [digit_toString n]
           else int_toString
                  (Arith.div_int n (Arith.Int_of_integer (10 : IntInf.int))) @
                  [digit_toString
                     (Arith.mod_int n
                       (Arith.Int_of_integer (10 : IntInf.int)))]));

fun toString_int i = String.concat (int_toString i);

fun toString_word A_ w = toString_int (Word.sint A_ w);

fun toString_bool b = (case b of true => "True" | false => "False");

fun toString_vala A_ Value.Unit = "Unit"
  | toString_vala A_ Value.Null = "Null"
  | toString_vala A_ (Value.Bool b) = String.concat ["Bool ", toString_bool b]
  | toString_vala A_ (Value.Intg i) =
    String.concat
      ["Intg ",
        toString_word
          (Type_Length.len_bit0
            (Type_Length.len_bit0
              (Type_Length.len_bit0
                (Type_Length.len_bit0
                  (Type_Length.len_bit0 Type_Length.len_num1)))))
          i]
  | toString_vala A_ (Value.Addr a) = String.concat ["Addr ", toString A_ a];

fun toString_bop BinOp.Eq = "=="
  | toString_bop BinOp.NotEq = "!="
  | toString_bop BinOp.LessThan = "<"
  | toString_bop BinOp.LessOrEqual = "<="
  | toString_bop BinOp.GreaterThan = ">"
  | toString_bop BinOp.GreaterOrEqual = ">="
  | toString_bop BinOp.Add = "+"
  | toString_bop BinOp.Subtract = "-"
  | toString_bop BinOp.Mult = "*"
  | toString_bop BinOp.Div = "/"
  | toString_bop BinOp.Mod = "%"
  | toString_bop BinOp.BinAnd = "&"
  | toString_bop BinOp.BinOr = "|"
  | toString_bop BinOp.BinXor = "^"
  | toString_bop BinOp.ShiftLeft = "<<"
  | toString_bop BinOp.ShiftRightZeros = ">>"
  | toString_bop BinOp.ShiftRightSigned = ">>>";

fun toString_literala s = s;

fun toString_ty Type.Void = "Void"
  | toString_ty Type.Boolean = "Boolean"
  | toString_ty Type.Integer = "Integer"
  | toString_ty Type.Nt = "NT"
  | toString_ty (Type.Class c) = String.concat ["Class ", toString_literala c]
  | toString_ty (Type.Array t) = String.concat [toString_ty t, "[]"];

fun toString_expa A_ B_ C_ (Expr.New c) = String.concat ["new ", c]
  | toString_expa A_ B_ C_ (Expr.NewArray (t, e)) =
    String.concat ["new ", toString_ty t, "[", toString_expa A_ B_ C_ e, "]"]
  | toString_expa A_ B_ C_ (Expr.Cast (t, e)) =
    String.concat ["(", toString_ty t, ") (", toString_expa A_ B_ C_ e, ")"]
  | toString_expa A_ B_ C_ (Expr.InstanceOf (e, t)) =
    String.concat
      ["(", toString_expa A_ B_ C_ e, ") instanceof ", toString_ty t]
  | toString_expa A_ B_ C_ (Expr.Val v) =
    String.concat ["Val (", toString_vala C_ v, ")"]
  | toString_expa A_ B_ C_ (Expr.BinOp (e1, bop, e2)) =
    String.concat
      ["(", toString_expa A_ B_ C_ e1, ") ", toString_bop bop, " (",
        toString_expa A_ B_ C_ e2, ")"]
  | toString_expa A_ B_ C_ (Expr.Var v) = String.concat ["Var ", toString A_ v]
  | toString_expa A_ B_ C_ (Expr.LAss (v, e)) =
    String.concat [toString A_ v, " := (", toString_expa A_ B_ C_ e, ")"]
  | toString_expa A_ B_ C_ (Expr.AAcc (a, i)) =
    String.concat
      ["(", toString_expa A_ B_ C_ a, ")[", toString_expa A_ B_ C_ i, "]"]
  | toString_expa A_ B_ C_ (Expr.AAss (a, i, e)) =
    String.concat
      ["(", toString_expa A_ B_ C_ a, ")[", toString_expa A_ B_ C_ i, "] := (",
        toString_expa A_ B_ C_ e, ")"]
  | toString_expa A_ B_ C_ (Expr.ALen a) =
    String.concat ["(", toString_expa A_ B_ C_ a, ").length"]
  | toString_expa A_ B_ C_ (Expr.FAcc (e, f, d)) =
    String.concat ["(", toString_expa A_ B_ C_ e, ").", f, "{", d, "}"]
  | toString_expa A_ B_ C_ (Expr.FAss (ea, f, d, e)) =
    String.concat
      ["(", toString_expa A_ B_ C_ ea, ").", f, "{", d, "} := (",
        toString_expa A_ B_ C_ e, ")"]
  | toString_expa A_ B_ C_ (Expr.Call (e, m, es)) =
    String.concat
      (["(", toString_expa A_ B_ C_ e, ").", m, "("] @
        List.map (toString_expa A_ B_ C_) es @ [")"])
  | toString_expa A_ B_ C_ (Expr.Block (v, t, vo, e)) =
    String.concat
      (["{", toString A_ v, ":", toString_ty t] @
        (case vo of NONE => [] | SOME va => ["=", toString_vala C_ va]) @
          ["; ", toString_expa A_ B_ C_ e, "}"])
  | toString_expa A_ B_ C_ (Expr.Synchronized (v, ea, e)) =
    String.concat
      ["synchronized_", toString B_ v, "_(", toString_expa A_ B_ C_ ea, ") {",
        toString_expa A_ B_ C_ e, "}"]
  | toString_expa A_ B_ C_ (Expr.InSynchronized (v, ad, e)) =
    String.concat
      ["insynchronized_", toString B_ v, "_(", toString C_ ad, ") {",
        toString_expa A_ B_ C_ e, "}"]
  | toString_expa A_ B_ C_ (Expr.Seq (ea, e)) =
    String.concat [toString_expa A_ B_ C_ ea, "; ", toString_expa A_ B_ C_ e]
  | toString_expa A_ B_ C_ (Expr.Cond (eb, ea, e)) =
    String.concat
      ["if (", toString_expa A_ B_ C_ eb, ") { ", toString_expa A_ B_ C_ ea,
        " } else { ", toString_expa A_ B_ C_ e, "}"]
  | toString_expa A_ B_ C_ (Expr.While (ea, e)) =
    String.concat
      ["while (", toString_expa A_ B_ C_ ea, ") { ", toString_expa A_ B_ C_ e,
        " }"]
  | toString_expa A_ B_ C_ (Expr.Throw e) =
    String.concat ["throw (", toString_expa A_ B_ C_ e, ")"]
  | toString_expa A_ B_ C_ (Expr.TryCatch (ea, c, v, e)) =
    String.concat
      ["try { ", toString_expa A_ B_ C_ ea, " } catch (", c, " ", toString A_ v,
        ") { ", toString_expa A_ B_ C_ e, " }"];

fun toString_exp A_ B_ C_ = {toString = toString_expa A_ B_ C_} :
  ('a, 'b, 'c) Expr.exp toString;

fun toString_funa f = "fn";

val toString_fun = {toString = toString_funa} : ('a -> 'b) toString;

fun toString_nata n = toString_int (Arith.int_of_nat n);

val toString_nat = {toString = toString_nata} : Arith.nat toString;

fun toString_val A_ = {toString = toString_vala A_} : 'a Value.vala toString;

fun toString_chara c = String.implode [c];

val toString_char = {toString = toString_chara} : char toString;

fun list_toString A_ sep [] = []
  | list_toString A_ sep [x] = [toString A_ x]
  | list_toString A_ sep (x :: v :: va) =
    toString A_ x :: sep :: list_toString A_ sep (v :: va);

fun toString_lista A_ xs =
  String.concat ("[" :: list_toString A_ "," xs @ ["]"]);

fun toString_list A_ = {toString = toString_lista A_} : ('a list) toString;

fun toString_proda A_ B_ =
  (fn (a, b) => String.concat ["(", toString A_ a, ", ", toString B_ b, ")"]);

fun toString_prod A_ B_ = {toString = toString_proda A_ B_} :
  ('a * 'b) toString;

fun toString_rbt (A1_, A2_) B_ t =
  String.concat
    (list_toString (toString_prod A2_ B_) ",\n"
      (RBTMapImpl.g_to_list_rm_basic_ops A1_ t));

fun toString_unita u = "()";

val toString_unit = {toString = toString_unita} : unit toString;

fun toString_trie A_ B_ t =
  toString_lista (toString_prod (toString_list A_) B_)
    (TrieMapImpl.g_to_list_tm_basic_ops t);

fun toString_finfuna (A1_, A2_, A3_, A4_) (B1_, B2_) f =
  String.concat
    ("(" ::
      toString B2_ (FinFun.finfun_default (A1_, A2_) f) ::
        List.maps
          (fn x =>
            [",", toString A4_ x, "|->",
              toString B2_ (FinFun.finfun_apply A2_ f x)])
          (FinFun.finfun_to_list (A1_, A2_, A3_) B1_ f) @
          [")"]);

fun toString_finfun (A1_, A2_, A3_, A4_) (B1_, B2_) =
  {toString = toString_finfuna (A1_, A2_, A3_, A4_) (B1_, B2_)} :
  ('a, 'b) FinFun.finfun toString;

fun toString_htype (Type.Class_type c) = c
  | toString_htype (Type.Array_type (t, n)) =
    String.concat [toString_ty t, "[", toString_nata n, "]"];

val toString_literal = {toString = toString_literala} : string toString;

fun toString_addr_loc (Heap.CField (c, f)) =
  String.concat ["CField ", f, "{", c, "}"]
  | toString_addr_loc (Heap.ACell n) =
    String.concat ["ACell ", toString_nata n];

fun toString_obs_eventa A_ B_ (Observable_Events.ExternalCall (ad, m, vs, v)) =
  String.concat
    ["ExternalCall ", m, "(", toString_lista (toString_val A_) vs, ") = ",
      toString_vala A_ v]
  | toString_obs_eventa A_ B_ (Observable_Events.ReadMem (ad, al, v)) =
    String.concat
      ["ReadMem ", toString A_ ad, "@", toString_addr_loc al, "=",
        toString_vala A_ v]
  | toString_obs_eventa A_ B_ (Observable_Events.WriteMem (ad, al, v)) =
    String.concat
      ["WriteMem ", toString A_ ad, "@", toString_addr_loc al, "=",
        toString_vala A_ v]
  | toString_obs_eventa A_ B_ (Observable_Events.NewHeapElem (ad, hT)) =
    String.concat ["Allocate ", toString A_ ad, ":", toString_htype hT]
  | toString_obs_eventa A_ B_ (Observable_Events.ThreadStart t) =
    String.concat ["ThreadStart ", toString B_ t]
  | toString_obs_eventa A_ B_ (Observable_Events.ThreadJoin t) =
    String.concat ["ThreadJoin ", toString B_ t]
  | toString_obs_eventa A_ B_ (Observable_Events.SyncLock ad) =
    String.concat ["SyncLock ", toString A_ ad]
  | toString_obs_eventa A_ B_ (Observable_Events.SyncUnlock ad) =
    String.concat ["SyncUnlock ", toString A_ ad]
  | toString_obs_eventa A_ B_ (Observable_Events.ObsInterrupt t) =
    String.concat ["Interrupt ", toString B_ t]
  | toString_obs_eventa A_ B_ (Observable_Events.ObsInterrupted t) =
    String.concat ["Interrupted ", toString B_ t];

fun toString_obs_event A_ B_ = {toString = toString_obs_eventa A_ B_} :
  ('a, 'b) Observable_Events.obs_event toString;

fun toString_assoc_lista A_ B_ =
  toString_lista (toString_prod A_ B_) o Assoc_List.impl_of;

fun toString_assoc_list A_ B_ = {toString = toString_assoc_lista A_ B_} :
  ('a, 'b) Assoc_List.assoc_list toString;

end; (*struct ToString*)

structure Predicate_Compile_Alternative_Defs : sig
  val list_all2_FiiB_i_i :
    ('a -> 'b -> unit Predicate.pred) ->
      'a list -> 'b list -> unit Predicate.pred
end = struct

fun list_all2_FiiB_i_i p xa xb =
  Predicate.sup_pred
    (case xa
      of [] =>
        (case xb of [] => Predicate.single () | _ :: _ => Predicate.bot_pred)
      | _ :: _ => Predicate.bot_pred)
    (case xa of [] => Predicate.bot_pred
      | x :: xs =>
        (case xb of [] => Predicate.bot_pred
          | y :: ys =>
            Predicate.bind (list_all2_FiiB_i_i p xs ys)
              (fn () =>
                Predicate.bind (p x y) (fn () => Predicate.single ()))));

end; (*struct Predicate_Compile_Alternative_Defs*)

structure Auxiliary : sig
  val distinct_fst : 'a HOL.equal -> ('a * 'b) list -> bool
end = struct

fun distinct_fst A_ = List.distinct A_ o List.map Product_Type.fst;

end; (*struct Auxiliary*)

structure WellForm : sig
  val acyclic_class_hierarchy : 'a TypeRel.prog -> bool
  val wf_overriding_aux_aux_i_i_i_i_i :
    Type.ty ->
      Type.ty ->
        Type.ty list -> Type.ty list -> 'a TypeRel.prog -> unit Predicate.pred
  val wf_overriding_aux_i_i_i_i_i :
    Type.ty ->
      Type.ty list -> string -> string -> 'a TypeRel.prog -> unit Predicate.pred
  val wf_overriding_i_i_i :
    'a TypeRel.prog ->
      string ->
        string * (Type.ty list * (Type.ty * 'a option)) -> unit Predicate.pred
  val wf_overriding :
    'a TypeRel.prog ->
      string -> string * (Type.ty list * (Type.ty * 'a option)) -> bool
  val wf_mdecl :
    ('a TypeRel.prog ->
      string -> string * (Type.ty list * (Type.ty * 'a)) -> bool) ->
      'a TypeRel.prog ->
        string -> string * (Type.ty list * (Type.ty * 'a option)) -> bool
  val wf_fdecl :
    'a TypeRel.prog -> string * (Type.ty * unit TypeRel.fmod_ext) -> bool
  val wf_cdecl :
    ('a TypeRel.prog ->
      string -> string * (Type.ty list * (Type.ty * 'a)) -> bool) ->
      'a TypeRel.prog ->
        string *
          (string *
            ((string * (Type.ty * unit TypeRel.fmod_ext)) list *
              (string * (Type.ty list * (Type.ty * 'a option))) list)) ->
          bool
  val wf_prog :
    ('a TypeRel.prog ->
      string -> string * (Type.ty list * (Type.ty * 'a)) -> bool) ->
      'a TypeRel.prog -> bool
end = struct

fun acyclic_class_hierarchy p =
  List.list_all
    (fn (c, (d, (_, _))) =>
      (if not ((c : string) = Type.object) then not (TypeRel.subcls p d c)
        else true))
    (TypeRel.classes p);

fun wf_overriding_aux_aux_i_i_i_i_i x xa xb xc xd =
  Predicate.sup_pred
    (Predicate.bind
      (Predicate.not_pred
        (Predicate_Compile_Alternative_Defs.list_all2_FiiB_i_i
          (TypeRel.widen_i_i_i xd) xc xb))
      (fn () => Predicate.single ()))
    (Predicate.bind (Predicate.not_pred (TypeRel.widen_i_i_i xd xa x))
      (fn () => Predicate.single ()));

fun wf_overriding_aux_i_i_i_i_i x xa xb xc xd =
  Predicate.bind (TypeRel.method_i_i_i_o_o_o_o xd xc xb)
    (fn (ts, (t, (_, _))) =>
      Predicate.bind (wf_overriding_aux_aux_i_i_i_i_i t x xa ts xd)
        (fn () => Predicate.single ()));

fun wf_overriding_i_i_i x xa (xb, (xc, (xd, xe))) =
  Predicate.bind
    (Predicate.not_pred (wf_overriding_aux_i_i_i_i_i xd xc xb xa x))
    (fn () => Predicate.single ());

fun wf_overriding x1 x2 x3 = Predicate.holds (wf_overriding_i_i_i x1 x2 x3);

fun wf_mdecl wf_md p c =
  (fn (m, (ts, (t, ma))) =>
    List.list_all (TypeRel.is_type p) ts andalso
      (TypeRel.is_type p t andalso
        (case ma of NONE => ExternalCall.external_WT_defs c m ts t
          | SOME mb => wf_md p c (m, (ts, (t, mb))))));

fun wf_fdecl p = (fn (_, (t, _)) => TypeRel.is_type p t);

fun wf_cdecl wf_md p =
  (fn (c, (d, (fs, ms))) =>
    List.list_all (wf_fdecl p) fs andalso
      (Auxiliary.distinct_fst Stringa.equal_literal fs andalso
        (List.list_all (wf_mdecl wf_md p c) ms andalso
          (Auxiliary.distinct_fst Stringa.equal_literal ms andalso
            ((if not ((c : string) = Type.object)
               then TypeRel.is_class p d andalso
                      List.list_all (wf_overriding p d) ms
               else true) andalso
              (if ((c : string) = Type.thread)
                then List.member
                       (Product_Type.equal_prod Stringa.equal_literal
                         (Product_Type.equal_prod
                           (List.equal_list Type.equal_ty) Type.equal_ty))
                       (List.map (fn (m, (ts, (t, _))) => (m, (ts, t))) ms)
                       ("run()V", ([], Type.Void))
                else true))))));

fun wf_prog wf_md p =
  acyclic_class_hierarchy p andalso
    (Exceptions.wf_syscls p andalso
      (Auxiliary.distinct_fst Stringa.equal_literal (TypeRel.classes p) andalso
        List.list_all (wf_cdecl wf_md p) (TypeRel.classes p)));

end; (*struct WellForm*)

structure WellType : sig
  val is_lub_sup_i_i_i_o :
    'a TypeRel.prog -> Type.ty -> Type.ty -> Type.ty Predicate.pred
  val wT_Fii_oB_i_i_i_o :
    (Type.ty -> Type.ty -> Type.ty Predicate.pred) ->
      (string list * (string, unit, 'a) Expr.exp) TypeRel.prog ->
        (string -> Type.ty option) ->
          (string, unit, 'a) Expr.exp -> Type.ty Predicate.pred
  val wTs_Fii_oB_i_i_i_o :
    (Type.ty -> Type.ty -> Type.ty Predicate.pred) ->
      (string list * (string, unit, 'a) Expr.exp) TypeRel.prog ->
        (string -> Type.ty option) ->
          (string, unit, 'a) Expr.exp list -> (Type.ty list) Predicate.pred
  val wT_Fii_oB_i_i_i_i :
    (Type.ty -> Type.ty -> Type.ty Predicate.pred) ->
      (string list * (string, unit, 'a) Expr.exp) TypeRel.prog ->
        (string -> Type.ty option) ->
          (string, unit, 'a) Expr.exp -> Type.ty -> unit Predicate.pred
  val wT_code_i_i_i_o :
    (string list * (string, unit, 'a) Expr.exp) TypeRel.prog ->
      (string -> Type.ty option) ->
        (string, unit, 'a) Expr.exp -> Type.ty Predicate.pred
end = struct

fun is_lub_sup_i_i_i_o x xa xb =
  (case SemiType.sup x xa xb of Err.Err => Predicate.bot_pred
    | Err.Ok a => Predicate.single a);

fun wT_Fii_oB_i_i_i_o is_lub x xa xb =
  (case xb
    of Expr.New c1 =>
      Predicate.bind (TypeRel.is_class_i_i x c1)
        (fn () => Predicate.single (Type.Class c1))
    | Expr.NewArray (c1, c2) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c2 Type.Integer)
        (fn () =>
          Predicate.bind (Predicate.if_pred (TypeRel.is_type x (Type.Array c1)))
            (fn () => Predicate.single (Type.Array c1)))
    | Expr.Cast (c1, c2) =>
      Predicate.bind (Predicate.if_pred (TypeRel.is_type x c1))
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
            (fn xaa =>
              Predicate.bind
                (Predicate.if_pred
                  (TypeRel.widen x c1 xaa orelse TypeRel.widen x xaa c1))
                (fn () => Predicate.single c1)))
    | Expr.InstanceOf (c1, c2) =>
      Predicate.bind (Predicate.if_pred (TypeRel.is_type x c2))
        (fn () =>
          Predicate.bind (Type.is_refT_i c2)
            (fn () =>
              Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
                (fn xaa =>
                  Predicate.bind
                    (Predicate.if_pred
                      (TypeRel.widen x c2 xaa orelse TypeRel.widen x xaa c2))
                    (fn () => Predicate.single Type.Boolean))))
    | Expr.Val c1 =>
      (case Heap.typeof c1 of NONE => Predicate.bot_pred
        | SOME a => Predicate.single a)
    | Expr.BinOp (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn xba =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c3)
            (BinOp.wT_binop_i_i_i_i_o x xba c2))
    | Expr.Var c1 =>
      (case xa c1 of NONE => Predicate.bot_pred | SOME a => Predicate.single a)
    | Expr.LAss (c1, c2) =>
      Predicate.bind (Predicate.if_pred (not ((c1 : string) = "this")))
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
            (fn xba =>
              (case xa c1 of NONE => Predicate.bot_pred
                | SOME t =>
                  Predicate.bind (TypeRel.widen_i_i_i x xba t)
                    (fn () => Predicate.single Type.Void))))
    | Expr.AAcc (c1, c2) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c2 Type.Integer)
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
            (fn a =>
              (case a of Type.Void => Predicate.bot_pred
                | Type.Boolean => Predicate.bot_pred
                | Type.Integer => Predicate.bot_pred
                | Type.Nt => Predicate.bot_pred
                | Type.Class _ => Predicate.bot_pred
                | Type.Array aa => Predicate.single aa)))
    | Expr.AAss (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c2 Type.Integer)
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
            (fn a =>
              (case a of Type.Void => Predicate.bot_pred
                | Type.Boolean => Predicate.bot_pred
                | Type.Integer => Predicate.bot_pred
                | Type.Nt => Predicate.bot_pred
                | Type.Class _ => Predicate.bot_pred
                | Type.Array t =>
                  Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c3)
                    (fn xaa =>
                      Predicate.bind (TypeRel.widen_i_i_i x xaa t)
                        (fn () => Predicate.single Type.Void)))))
    | Expr.ALen c1 =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn a =>
          (case a of Type.Void => Predicate.bot_pred
            | Type.Boolean => Predicate.bot_pred
            | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
            | Type.Class _ => Predicate.bot_pred
            | Type.Array _ => Predicate.single Type.Integer))
    | Expr.FAcc (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn xaa =>
          (case Type.class_type_of xaa of NONE => Predicate.bot_pred
            | SOME c =>
              Predicate.bind (TypeRel.sees_field_i_i_i_o_o_i x c c2 c3)
                (fn (t, _) => Predicate.single t)))
    | Expr.FAss (c1, c2, c3, c4) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn xba =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c4)
            (fn xaa =>
              (case Type.class_type_of xba of NONE => Predicate.bot_pred
                | SOME c =>
                  Predicate.bind (TypeRel.sees_field_i_i_i_o_o_i x c c2 c3)
                    (fn (t, _) =>
                      Predicate.bind (TypeRel.widen_i_i_i x xaa t)
                        (fn () => Predicate.single Type.Void)))))
    | Expr.Call (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn xba =>
          (case Type.class_type_of xba of NONE => Predicate.bot_pred
            | SOME c =>
              Predicate.bind (wTs_Fii_oB_i_i_i_o is_lub x xa c3)
                (fn xaa =>
                  Predicate.bind (TypeRel.method_i_i_i_o_o_o_o x c c2)
                    (fn (ts, (t, (_, _))) =>
                      Predicate.bind
                        (Predicate_Compile_Alternative_Defs.list_all2_FiiB_i_i
                          (TypeRel.widen_i_i_i x) xaa ts)
                        (fn () => Predicate.single t)))))
    | Expr.Block (c1, c2, c3, c4) =>
      Predicate.bind (Predicate.if_pred (TypeRel.is_type x c2))
        (fn () =>
          Predicate.bind
            (Predicate.if_pred
              (case c3 of NONE => true
                | SOME v =>
                  (case Heap.typeof v of NONE => false
                    | SOME t => TypeRel.widen x t c2)))
            (fn () =>
              wT_Fii_oB_i_i_i_o is_lub x
                (Fun.fun_upd Stringa.equal_literal xa c1 (SOME c2)) c4))
    | Expr.Synchronized ((), c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
        (fn xba =>
          Predicate.bind (Type.is_refT_i xba)
            (fn () =>
              Predicate.bind
                (Predicate.if_pred (not (Type.equal_tya xba Type.Nt)))
                (fn () => wT_Fii_oB_i_i_i_o is_lub x xa c3)))
    | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
    | Expr.Seq (c1, c2) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn _ => wT_Fii_oB_i_i_i_o is_lub x xa c2)
    | Expr.Cond (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c1 Type.Boolean)
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
            (fn xba =>
              Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c3) (is_lub xba)))
    | Expr.While (c1, c2) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c1 Type.Boolean)
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
            (fn _ => Predicate.single Type.Void))
    | Expr.Throw c1 =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn a =>
          (case a of Type.Void => Predicate.bot_pred
            | Type.Boolean => Predicate.bot_pred
            | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
            | Type.Class c =>
              Predicate.bind
                (Transitive_Closure_Table.rtranclp_FioB_i_i
                  Stringa.equal_literal (TypeRel.subcls1_i_i_o x) c
                  Type.throwable)
                (fn () => Predicate.single Type.Void)
            | Type.Array _ => Predicate.bot_pred))
    | Expr.TryCatch (c1, c2, c3, c4) =>
      Predicate.bind
        (Transitive_Closure_Table.rtranclp_FioB_i_i Stringa.equal_literal
          (TypeRel.subcls1_i_i_o x) c2 Type.throwable)
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
            (fn xba =>
              Predicate.bind
                (wT_Fii_oB_i_i_i_i is_lub x
                  (Fun.fun_upd Stringa.equal_literal xa c3
                    (SOME (Type.Class c2)))
                  c4 xba)
                (fn () => Predicate.single xba))))
and wTs_Fii_oB_i_i_i_o is_lub x xa xb =
  (case xb of [] => Predicate.single []
    | c1 :: c2 =>
      Predicate.bind (wTs_Fii_oB_i_i_i_o is_lub x xa c2)
        (fn xba =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
            (fn xc => Predicate.single (xc :: xba))))
and wT_Fii_oB_i_i_i_i is_lub x xa xb xc =
  (case xb
    of Expr.New _ =>
      (case xc of Type.Void => Predicate.bot_pred
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class c1 =>
          Predicate.bind (TypeRel.is_class_i_i x c1)
            (fn () => Predicate.single ())
        | Type.Array _ => Predicate.bot_pred)
    | Expr.NewArray (_, c2) =>
      (case xc of Type.Void => Predicate.bot_pred
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class _ => Predicate.bot_pred
        | Type.Array c1 =>
          Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c2 Type.Integer)
            (fn () =>
              Predicate.bind
                (Predicate.if_pred (TypeRel.is_type x (Type.Array c1)))
                (fn () => Predicate.single ())))
    | Expr.Cast (_, c2) =>
      Predicate.bind (Predicate.if_pred (TypeRel.is_type x xc))
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
            (fn xaa =>
              Predicate.bind
                (Predicate.if_pred
                  (TypeRel.widen x xc xaa orelse TypeRel.widen x xaa xc))
                (fn () => Predicate.single ())))
    | Expr.InstanceOf (c1, c2) =>
      (case xc of Type.Void => Predicate.bot_pred
        | Type.Boolean =>
          Predicate.bind (Predicate.if_pred (TypeRel.is_type x c2))
            (fn () =>
              Predicate.bind (Type.is_refT_i c2)
                (fn () =>
                  Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
                    (fn xaa =>
                      Predicate.bind
                        (Predicate.if_pred
                          (TypeRel.widen x c2 xaa orelse
                            TypeRel.widen x xaa c2))
                        (fn () => Predicate.single ()))))
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class _ => Predicate.bot_pred
        | Type.Array _ => Predicate.bot_pred)
    | Expr.Val c1 =>
      Predicate.bind
        (TypeRel.eq_i_i (Option.equal_option Type.equal_ty) (Heap.typeof c1)
          (SOME xc))
        (fn () => Predicate.single ())
    | Expr.BinOp (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn xba =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c3)
            (fn xaa =>
              Predicate.bind (BinOp.wT_binop_i_i_i_i_i x xba c2 xaa xc)
                (fn () => Predicate.single ())))
    | Expr.Var c1 =>
      Predicate.bind
        (TypeRel.eq_i_i (Option.equal_option Type.equal_ty) (xa c1) (SOME xc))
        (fn () => Predicate.single ())
    | Expr.LAss (c1, c2) =>
      (case xc
        of Type.Void =>
          Predicate.bind (Predicate.if_pred (not ((c1 : string) = "this")))
            (fn () =>
              Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
                (fn xba =>
                  (case xa c1 of NONE => Predicate.bot_pred
                    | SOME t =>
                      Predicate.bind (TypeRel.widen_i_i_i x xba t)
                        (fn () => Predicate.single ()))))
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class _ => Predicate.bot_pred
        | Type.Array _ => Predicate.bot_pred)
    | Expr.AAcc (c1, c2) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c1 (Type.Array xc))
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c2 Type.Integer)
            (fn () => Predicate.single ()))
    | Expr.AAss (c1, c2, c3) =>
      (case xc
        of Type.Void =>
          Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c2 Type.Integer)
            (fn () =>
              Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
                (fn a =>
                  (case a of Type.Void => Predicate.bot_pred
                    | Type.Boolean => Predicate.bot_pred
                    | Type.Integer => Predicate.bot_pred
                    | Type.Nt => Predicate.bot_pred
                    | Type.Class _ => Predicate.bot_pred
                    | Type.Array t =>
                      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c3)
                        (fn xaa =>
                          Predicate.bind (TypeRel.widen_i_i_i x xaa t)
                            (fn () => Predicate.single ())))))
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class _ => Predicate.bot_pred
        | Type.Array _ => Predicate.bot_pred)
    | Expr.ALen c1 =>
      (case xc of Type.Void => Predicate.bot_pred
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
            (fn a =>
              (case a of Type.Void => Predicate.bot_pred
                | Type.Boolean => Predicate.bot_pred
                | Type.Integer => Predicate.bot_pred
                | Type.Nt => Predicate.bot_pred
                | Type.Class _ => Predicate.bot_pred
                | Type.Array _ => Predicate.single ()))
        | Type.Nt => Predicate.bot_pred | Type.Class _ => Predicate.bot_pred
        | Type.Array _ => Predicate.bot_pred)
    | Expr.FAcc (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn xaa =>
          (case Type.class_type_of xaa of NONE => Predicate.bot_pred
            | SOME c =>
              Predicate.bind (TypeRel.sees_field_i_i_i_o_o_i x c c2 c3)
                (fn (_, _) => Predicate.single ())))
    | Expr.FAss (c1, c2, c3, c4) =>
      (case xc
        of Type.Void =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
            (fn xba =>
              Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c4)
                (fn xaa =>
                  (case Type.class_type_of xba of NONE => Predicate.bot_pred
                    | SOME c =>
                      Predicate.bind (TypeRel.sees_field_i_i_i_o_o_i x c c2 c3)
                        (fn (t, _) =>
                          Predicate.bind (TypeRel.widen_i_i_i x xaa t)
                            (fn () => Predicate.single ())))))
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class _ => Predicate.bot_pred
        | Type.Array _ => Predicate.bot_pred)
    | Expr.Call (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
        (fn xba =>
          (case Type.class_type_of xba of NONE => Predicate.bot_pred
            | SOME c =>
              Predicate.bind (wTs_Fii_oB_i_i_i_o is_lub x xa c3)
                (fn xaa =>
                  Predicate.bind (TypeRel.method_i_i_i_o_o_o_o x c c2)
                    (fn (ts, (_, (_, _))) =>
                      Predicate.bind
                        (Predicate_Compile_Alternative_Defs.list_all2_FiiB_i_i
                          (TypeRel.widen_i_i_i x) xaa ts)
                        (fn () => Predicate.single ())))))
    | Expr.Block (c1, c2, c3, c4) =>
      Predicate.bind
        (wT_Fii_oB_i_i_i_i is_lub x
          (Fun.fun_upd Stringa.equal_literal xa c1 (SOME c2)) c4 xc)
        (fn () =>
          Predicate.bind (Predicate.if_pred (TypeRel.is_type x c2))
            (fn () =>
              Predicate.bind
                (Predicate.if_pred
                  (case c3 of NONE => true
                    | SOME v =>
                      (case Heap.typeof v of NONE => false
                        | SOME t => TypeRel.widen x t c2)))
                (fn () => Predicate.single ())))
    | Expr.Synchronized ((), c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c3 xc)
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
            (fn xd =>
              Predicate.bind (Type.is_refT_i xd)
                (fn () =>
                  Predicate.bind
                    (Predicate.if_pred (not (Type.equal_tya xd Type.Nt)))
                    (fn () => Predicate.single ()))))
    | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
    | Expr.Seq (c1, c2) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c2 xc)
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
            (fn _ => Predicate.single ()))
    | Expr.Cond (c1, c2, c3) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c1 Type.Boolean)
        (fn () =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
            (fn xba =>
              Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c3)
                (fn xd =>
                  Predicate.bind (is_lub xba xd)
                    (fn _ => Predicate.single ()))))
    | Expr.While (c1, c2) =>
      (case xc
        of Type.Void =>
          Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c1 Type.Boolean)
            (fn () =>
              Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c2)
                (fn _ => Predicate.single ()))
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class _ => Predicate.bot_pred
        | Type.Array _ => Predicate.bot_pred)
    | Expr.Throw c1 =>
      (case xc
        of Type.Void =>
          Predicate.bind (wT_Fii_oB_i_i_i_o is_lub x xa c1)
            (fn a =>
              (case a of Type.Void => Predicate.bot_pred
                | Type.Boolean => Predicate.bot_pred
                | Type.Integer => Predicate.bot_pred
                | Type.Nt => Predicate.bot_pred
                | Type.Class c =>
                  Predicate.bind
                    (Transitive_Closure_Table.rtranclp_FioB_i_i
                      Stringa.equal_literal (TypeRel.subcls1_i_i_o x) c
                      Type.throwable)
                    (fn () => Predicate.single ())
                | Type.Array _ => Predicate.bot_pred))
        | Type.Boolean => Predicate.bot_pred
        | Type.Integer => Predicate.bot_pred | Type.Nt => Predicate.bot_pred
        | Type.Class _ => Predicate.bot_pred
        | Type.Array _ => Predicate.bot_pred)
    | Expr.TryCatch (c1, c2, c3, c4) =>
      Predicate.bind (wT_Fii_oB_i_i_i_i is_lub x xa c1 xc)
        (fn () =>
          Predicate.bind
            (wT_Fii_oB_i_i_i_i is_lub x
              (Fun.fun_upd Stringa.equal_literal xa c3 (SOME (Type.Class c2)))
              c4 xc)
            (fn () =>
              Predicate.bind
                (Transitive_Closure_Table.rtranclp_FioB_i_i
                  Stringa.equal_literal (TypeRel.subcls1_i_i_o x) c2
                  Type.throwable)
                (fn () => Predicate.single ()))));

fun wT_code_i_i_i_o x xa xb = wT_Fii_oB_i_i_i_o (is_lub_sup_i_i_i_o x) x xa xb;

end; (*struct WellType*)

structure FWLocking : sig
  val redT_updLs :
    'a HOL.equal -> 'b HOL.equal ->
      ('a, (('b * Arith.nat) option)) FinFun.finfun ->
        'b -> ('a, (FWState.lock_action list)) FinFun.finfun ->
                ('a, (('b * Arith.nat) option)) FinFun.finfun
  val acquire_all :
    'a HOL.equal -> 'b HOL.equal ->
      ('a, (('b * Arith.nat) option)) FinFun.finfun ->
        'b -> ('a, Arith.nat) FinFun.finfun ->
                ('a, (('b * Arith.nat) option)) FinFun.finfun
  val lock_ok_las :
    'a Set.card_UNIV * 'a HOL.equal -> 'b HOL.equal ->
      ('a, (('b * Arith.nat) option)) FinFun.finfun ->
        'b -> ('a, (FWState.lock_action list)) FinFun.finfun -> bool
  val may_acquire_all :
    'a Set.card_UNIV * 'a HOL.equal -> 'b HOL.equal ->
      ('a, (('b * Arith.nat) option)) FinFun.finfun ->
        'b -> ('a, Arith.nat) FinFun.finfun -> bool
end = struct

fun redT_updLs A_ B_ ls t las =
  FinFun.finfun_comp (fn (l, a) => FWLock.upd_locks B_ l t a)
    (FinFun.finfun_Diag A_
      (Option.equal_option (Product_Type.equal_prod B_ Arith.equal_nat))
      (List.equal_list FWState.equal_lock_action) ls las);

fun acquire_all A_ B_ ls t ln =
  FinFun.finfun_comp (fn (l, a) => FWLock.acquire_locks l t a)
    (FinFun.finfun_Diag A_
      (Option.equal_option (Product_Type.equal_prod B_ Arith.equal_nat))
      Arith.equal_nat ls ln);

fun lock_ok_las (A1_, A2_) B_ ls t las =
  FinFun.finfun_All (A1_, A2_)
    (FinFun.finfun_comp (fn (l, a) => FWLock.lock_actions_ok B_ l t a)
      (FinFun.finfun_Diag A2_
        (Option.equal_option (Product_Type.equal_prod B_ Arith.equal_nat))
        (List.equal_list FWState.equal_lock_action) ls las));

fun may_acquire_all (A1_, A2_) B_ ls t ln =
  FinFun.finfun_All (A1_, A2_)
    (FinFun.finfun_comp
      (fn (lock, n) =>
        (if Arith.less_nat Arith.zero_nat n then FWLock.may_lock B_ lock t
          else true))
      (FinFun.finfun_Diag A2_
        (Option.equal_option (Product_Type.equal_prod B_ Arith.equal_nat))
        Arith.equal_nat ls ln));

end; (*struct FWLocking*)

structure JWellForm : sig
  val typeable_with_i_i_i_i :
    (string list * (string, unit, 'a) Expr.exp) TypeRel.prog ->
      (string -> Type.ty option) ->
        (string, unit, 'a) Expr.exp -> Type.ty -> unit Predicate.pred
  val typeable_with :
    (string list * (string, unit, 'a) Expr.exp) TypeRel.prog ->
      (string -> Type.ty option) ->
        (string, unit, 'a) Expr.exp -> Type.ty -> bool
  val wf_J_mdecl :
    (string list * (string, unit, 'a) Expr.exp) TypeRel.prog ->
      string ->
        string *
          (Type.ty list *
            (Type.ty * (string list * (string, unit, 'a) Expr.exp))) ->
          bool
  val wf_J_prog :
    (string list * (string, unit, 'a) Expr.exp) TypeRel.prog -> bool
end = struct

fun typeable_with_i_i_i_i x xa xb xc =
  Predicate.bind (WellType.wT_code_i_i_i_o x xa xb)
    (fn xaa =>
      Predicate.bind (TypeRel.widen_i_i_i x xaa xc)
        (fn () => Predicate.single ()));

fun typeable_with x1 x2 x3 x4 =
  Predicate.holds (typeable_with_i_i_i_i x1 x2 x3 x4);

fun wf_J_mdecl p c =
  (fn (_, (ts, (t, (pns, body)))) =>
    Arith.equal_nata (List.size_list ts) (List.size_list pns) andalso
      (List.distinct Stringa.equal_literal pns andalso
        (not (List.member Stringa.equal_literal pns "this") andalso
          (typeable_with p
             (Map.map_upds Stringa.equal_literal
               (Fun.fun_upd Stringa.equal_literal (fn _ => NONE) "this"
                 (SOME (Type.Class c)))
               pns ts)
             body t andalso
            DefAss.d Stringa.equal_literal body
              (SOME (Set.union (Set.insert "this" Set.bot_set)
                      (Set.Set pns)))))));

fun wf_J_prog x = WellForm.wf_prog wf_J_mdecl x;

end; (*struct JWellForm*)

structure ListSetImpl_Invar : sig
  val iteratei_bset_op_list_it_lsi_basic_ops :
    'a list -> ('b -> bool) -> ('a -> 'b -> 'b) -> 'b -> 'b
  val g_to_list_lsi_basic_ops : 'a list -> 'a list
end = struct

fun iteratei_bset_op_list_it_lsi_basic_ops s = Foldi.foldli s;

fun g_to_list_lsi_basic_ops s =
  iteratei_bset_op_list_it_lsi_basic_ops s (fn _ => true)
    (fn a => fn b => a :: b) [];

end; (*struct ListSetImpl_Invar*)

structure State_Refinement : sig
  val jinja_output :
    'a -> 'b -> ('c, (FWState.lock_action list)) FinFun.finfun *
                  (('b, 'd, 'e) FWState.new_thread_action list *
                    ('b FWState.conditional_action list *
                      (('b, 'c) FWState.wait_set_action list *
                        ('b FWState.interrupt_action list *
                          ('c, 'b) Observable_Events.obs_event list)))) ->
                  ('b * ('c, 'b) Observable_Events.obs_event list) option
  val start_state_refine :
    'a HOL.equal * 'a Value.addr -> 'c HOL.equal -> 'h HOL.equal ->
      ('a -> 'b) ->
        'c -> ('c -> Type.htype -> ('c * 'a) Set.set) ->
                'd -> ('b -> 'e * ('a, Arith.nat) FinFun.finfun -> 'd -> 'd) ->
                        'f -> 'g -> (string ->
                                      string ->
Type.ty list -> Type.ty -> 'h -> 'a Value.vala list -> 'e) ->
                                      'h TypeRel.prog ->
string ->
  string ->
    'a Value.vala list ->
      ('a, (('b * Arith.nat) option)) FinFun.finfun * (('d * 'c) * ('f * 'g))
end = struct

fun jinja_output sigma t ta =
  (if List.null (FWState.obs_a ta) then NONE else SOME (t, FWState.obs_a ta));

fun start_state_refine (A1_, A2_) C_ H_ addr2thread_id empty_heap allocate
  thr_empty thr_update ws_empty is_empty f p c m vs =
  let
    val (d, (ts, (t, ma))) = TypeRel.method H_ p c m;
  in
    (FinFun.Finfun_const NONE,
      ((thr_update
          (StartConfig.start_tid (A1_, A2_) C_ addr2thread_id empty_heap
            allocate)
          (f d m ts t (Option.the ma) vs, FinFun.Finfun_const Arith.zero_nat)
          thr_empty,
         StartConfig.start_heap C_ (A1_, A2_) empty_heap allocate),
        (ws_empty, is_empty)))
  end;

end; (*struct State_Refinement*)

structure FWLockingThread : sig
  val upd_threadR :
    'a HOL.equal ->
      Arith.nat ->
        ('a * Arith.nat) option -> 'a -> FWState.lock_action -> Arith.nat
  val upd_threadRs :
    'a HOL.equal ->
      Arith.nat ->
        ('a * Arith.nat) option -> 'a -> FWState.lock_action list -> Arith.nat
  val redT_updLns :
    'a HOL.equal -> 'b HOL.equal ->
      ('a, (('b * Arith.nat) option)) FinFun.finfun ->
        'b -> ('a, Arith.nat) FinFun.finfun ->
                ('a, (FWState.lock_action list)) FinFun.finfun ->
                  ('a, Arith.nat) FinFun.finfun
end = struct

fun upd_threadR A_ n l t FWState.ReleaseAcquire =
  Arith.plus_nat n (FWLock.has_locks A_ l t)
  | upd_threadR A_ n l t FWState.Lock = n
  | upd_threadR A_ n l t FWState.Unlock = n
  | upd_threadR A_ n l t FWState.UnlockFail = n;

fun upd_threadRs A_ n l t [] = n
  | upd_threadRs A_ n l t (la :: las) =
    upd_threadRs A_ (upd_threadR A_ n l t la) (FWLock.upd_lock A_ l t la) t las;

fun redT_updLns A_ B_ ls t ln las =
  FinFun.finfun_comp (fn (l, (n, a)) => upd_threadRs B_ n l t a)
    (FinFun.finfun_Diag A_
      (Option.equal_option (Product_Type.equal_prod B_ Arith.equal_nat))
      (Product_Type.equal_prod Arith.equal_nat
        (List.equal_list FWState.equal_lock_action))
      ls (FinFun.finfun_Diag A_ Arith.equal_nat
           (List.equal_list FWState.equal_lock_action) ln las));

end; (*struct FWLockingThread*)

structure Scheduler : sig
  val tllist_unfold :
    ('a -> (('b * 'a), 'c) Sum_Type.sum) -> 'a -> ('b, 'c) TLList.tllist
  val redT_updT :
    ('a -> 'b * ('c, Arith.nat) FinFun.finfun -> 'd -> 'd) ->
      'd -> ('a, 'b, 'e) FWState.new_thread_action -> 'd
  val redT_updTs :
    ('a -> 'b * ('c, Arith.nat) FinFun.finfun -> 'd -> 'd) ->
      'd -> ('a, 'b, 'e) FWState.new_thread_action list -> 'd
  val redT_updI :
    ('a -> 'b -> 'b) ->
      ('a -> 'b -> 'b) -> 'b -> 'a FWState.interrupt_action -> 'b
  val redT_updIs :
    ('a -> 'b -> 'b) ->
      ('a -> 'b -> 'b) -> 'b -> 'a FWState.interrupt_action list -> 'b
  val exec_updW :
    'c HOL.equal ->
      ('a -> 'b -> 'c -> 'd -> 'b option) ->
        ('b -> 'd -> 'c FWState.wait_set_status option) ->
          ('b -> 'c FWState.wait_set_status -> 'd -> 'd) ->
            ('b -> 'd -> 'd) ->
              ('d ->
                ('d -> bool) ->
                  ('b * 'c FWState.wait_set_status -> 'd -> 'd) -> 'd -> 'd) ->
                'a -> 'b -> 'd -> ('b, 'c) FWState.wait_set_action -> 'd
  val exec_updWs :
    'c HOL.equal ->
      ('a -> 'b -> 'c -> 'd -> 'b option) ->
        ('b -> 'd -> 'c FWState.wait_set_status option) ->
          ('b -> 'c FWState.wait_set_status -> 'd -> 'd) ->
            ('b -> 'd -> 'd) ->
              ('d ->
                ('d -> bool) ->
                  ('b * 'c FWState.wait_set_status -> 'd -> 'd) -> 'd -> 'd) ->
                'a -> 'b -> 'd -> ('b, 'c) FWState.wait_set_action list -> 'd
  val exec_upd :
    'b HOL.equal -> 'c HOL.equal -> 'g HOL.equal ->
      ('a -> 'b -> 'c -> 'd -> 'b option) ->
        ('b -> 'e -> ('f * ('g, Arith.nat) FinFun.finfun) option) ->
          ('b -> 'f * ('g, Arith.nat) FinFun.finfun -> 'e -> 'e) ->
            ('b -> 'd -> 'c FWState.wait_set_status option) ->
              ('b -> 'c FWState.wait_set_status -> 'd -> 'd) ->
                ('b -> 'd -> 'd) ->
                  ('d ->
                    ('d -> bool) ->
                      ('b * 'c FWState.wait_set_status -> 'd -> 'd) ->
                        'd -> 'd) ->
                    ('b -> 'h -> 'h) ->
                      ('b -> 'h -> 'h) ->
                        'a -> ('g, (('b * Arith.nat) option)) FinFun.finfun *
                                (('e * 'i) * ('d * 'h)) ->
                                'b -> ('g, (FWState.lock_action list))
FinFun.finfun *
(('b, 'f, 'i) FWState.new_thread_action list *
  ('b FWState.conditional_action list *
    (('b, 'c) FWState.wait_set_action list *
      ('b FWState.interrupt_action list * 'j list)))) ->
'f -> 'i -> ('g, (('b * Arith.nat) option)) FinFun.finfun *
              (('e * 'i) * ('d * 'h))
  val execT :
    'a HOL.equal -> 'd HOL.equal -> 'j HOL.equal ->
      (('a, Arith.nat) FinFun.finfun -> 'b list) ->
        ('c ->
          ('a, (('d * Arith.nat) option)) FinFun.finfun *
            (('e * 'f) * ('g * 'h)) ->
            ('d * (((('a, (FWState.lock_action list)) FinFun.finfun *
                      (('d, 'i, 'f) FWState.new_thread_action list *
                        ('d FWState.conditional_action list *
                          (('d, 'j) FWState.wait_set_action list *
                            ('d FWState.interrupt_action list * 'b list))))) *
                     ('i * 'f)) option *
                    'c)) option) ->
          ('c -> 'd -> 'j -> 'g -> 'd option) ->
            ('d -> 'e -> ('i * ('a, Arith.nat) FinFun.finfun) option) ->
              ('d -> 'i * ('a, Arith.nat) FinFun.finfun -> 'e -> 'e) ->
                ('d -> 'g -> 'j FWState.wait_set_status option) ->
                  ('d -> 'j FWState.wait_set_status -> 'g -> 'g) ->
                    ('d -> 'g -> 'g) ->
                      ('g ->
                        ('g -> bool) ->
                          ('d * 'j FWState.wait_set_status -> 'g -> 'g) ->
                            'g -> 'g) ->
                        ('d -> 'h -> 'h) ->
                          ('d -> 'h -> 'h) ->
                            'c -> ('a, (('d * Arith.nat) option))
                                    FinFun.finfun *
                                    (('e * 'f) * ('g * 'h)) ->
                                    ('c * ('d *
    ((('a, (FWState.lock_action list)) FinFun.finfun *
       (('d, 'i, 'f) FWState.new_thread_action list *
         ('d FWState.conditional_action list *
           (('d, 'j) FWState.wait_set_action list *
             ('d FWState.interrupt_action list * 'b list))))) *
      (('a, (('d * Arith.nat) option)) FinFun.finfun *
        (('e * 'f) * ('g * 'h)))))) option
  val exec_step :
    'a HOL.equal -> 'd HOL.equal -> 'j HOL.equal ->
      (('a, Arith.nat) FinFun.finfun -> 'b list) ->
        ('c ->
          ('a, (('d * Arith.nat) option)) FinFun.finfun *
            (('e * 'f) * ('g * 'h)) ->
            ('d * (((('a, (FWState.lock_action list)) FinFun.finfun *
                      (('d, 'i, 'f) FWState.new_thread_action list *
                        ('d FWState.conditional_action list *
                          (('d, 'j) FWState.wait_set_action list *
                            ('d FWState.interrupt_action list * 'b list))))) *
                     ('i * 'f)) option *
                    'c)) option) ->
          ('c -> 'd -> 'j -> 'g -> 'd option) ->
            ('d -> 'e -> ('i * ('a, Arith.nat) FinFun.finfun) option) ->
              ('d -> 'i * ('a, Arith.nat) FinFun.finfun -> 'e -> 'e) ->
                ('d -> 'g -> 'j FWState.wait_set_status option) ->
                  ('d -> 'j FWState.wait_set_status -> 'g -> 'g) ->
                    ('d -> 'g -> 'g) ->
                      ('g ->
                        ('g -> bool) ->
                          ('d * 'j FWState.wait_set_status -> 'g -> 'g) ->
                            'g -> 'g) ->
                        ('d -> 'h -> 'h) ->
                          ('d -> 'h -> 'h) ->
                            'c * (('a, (('d * Arith.nat) option))
                                    FinFun.finfun *
                                   (('e * 'f) * ('g * 'h))) ->
                              ((('c * ('d *
(('a, (FWState.lock_action list)) FinFun.finfun *
  (('d, 'i, 'f) FWState.new_thread_action list *
    ('d FWState.conditional_action list *
      (('d, 'j) FWState.wait_set_action list *
        ('d FWState.interrupt_action list * 'b list))))))) *
                                 ('c * (('a, (('d * Arith.nat) option))
  FinFun.finfun *
 (('e * 'f) * ('g * 'h))))),
                                (('a, (('d * Arith.nat) option)) FinFun.finfun *
                                  (('e * 'f) * ('g * 'h))))
                                Sum_Type.sum
  val exec_aux :
    'a HOL.equal -> 'd HOL.equal -> 'j HOL.equal ->
      (('a, Arith.nat) FinFun.finfun -> 'b list) ->
        ('c ->
          ('a, (('d * Arith.nat) option)) FinFun.finfun *
            (('e * 'f) * ('g * 'h)) ->
            ('d * (((('a, (FWState.lock_action list)) FinFun.finfun *
                      (('d, 'i, 'f) FWState.new_thread_action list *
                        ('d FWState.conditional_action list *
                          (('d, 'j) FWState.wait_set_action list *
                            ('d FWState.interrupt_action list * 'b list))))) *
                     ('i * 'f)) option *
                    'c)) option) ->
          ('c -> 'd -> 'j -> 'g -> 'd option) ->
            ('d -> 'e -> ('i * ('a, Arith.nat) FinFun.finfun) option) ->
              ('d -> 'i * ('a, Arith.nat) FinFun.finfun -> 'e -> 'e) ->
                ('d -> 'g -> 'j FWState.wait_set_status option) ->
                  ('d -> 'j FWState.wait_set_status -> 'g -> 'g) ->
                    ('d -> 'g -> 'g) ->
                      ('g ->
                        ('g -> bool) ->
                          ('d * 'j FWState.wait_set_status -> 'g -> 'g) ->
                            'g -> 'g) ->
                        ('d -> 'h -> 'h) ->
                          ('d -> 'h -> 'h) ->
                            'c * (('a, (('d * Arith.nat) option))
                                    FinFun.finfun *
                                   (('e * 'f) * ('g * 'h))) ->
                              (('c * ('d * (('a, (FWState.lock_action list))
      FinFun.finfun *
     (('d, 'i, 'f) FWState.new_thread_action list *
       ('d FWState.conditional_action list *
         (('d, 'j) FWState.wait_set_action list *
           ('d FWState.interrupt_action list * 'b list))))))),
                                (('a, (('d * Arith.nat) option)) FinFun.finfun *
                                  (('e * 'f) * ('g * 'h))))
                                TLList.tllist
  val exec :
    'a HOL.equal -> 'd HOL.equal -> 'j HOL.equal ->
      (('a, Arith.nat) FinFun.finfun -> 'b list) ->
        ('c ->
          ('a, (('d * Arith.nat) option)) FinFun.finfun *
            (('e * 'f) * ('g * 'h)) ->
            ('d * (((('a, (FWState.lock_action list)) FinFun.finfun *
                      (('d, 'i, 'f) FWState.new_thread_action list *
                        ('d FWState.conditional_action list *
                          (('d, 'j) FWState.wait_set_action list *
                            ('d FWState.interrupt_action list * 'b list))))) *
                     ('i * 'f)) option *
                    'c)) option) ->
          ('c ->
            'd -> ('a, (FWState.lock_action list)) FinFun.finfun *
                    (('d, 'i, 'f) FWState.new_thread_action list *
                      ('d FWState.conditional_action list *
                        (('d, 'j) FWState.wait_set_action list *
                          ('d FWState.interrupt_action list * 'b list)))) ->
                    'k option) ->
            ('c -> 'd -> 'j -> 'g -> 'd option) ->
              ('d -> 'e -> ('i * ('a, Arith.nat) FinFun.finfun) option) ->
                ('d -> 'i * ('a, Arith.nat) FinFun.finfun -> 'e -> 'e) ->
                  ('d -> 'g -> 'j FWState.wait_set_status option) ->
                    ('d -> 'j FWState.wait_set_status -> 'g -> 'g) ->
                      ('d -> 'g -> 'g) ->
                        ('g ->
                          ('g -> bool) ->
                            ('d * 'j FWState.wait_set_status -> 'g -> 'g) ->
                              'g -> 'g) ->
                          ('d -> 'h -> 'h) ->
                            ('d -> 'h -> 'h) ->
                              'c -> ('a, (('d * Arith.nat) option))
                                      FinFun.finfun *
                                      (('e * 'f) * ('g * 'h)) ->
                                      ('k,
(('a, (('d * Arith.nat) option)) FinFun.finfun * (('e * 'f) * ('g * 'h))))
TLList.tllist
  val singleton2 : (unit -> 'a) -> 'a Predicate.pred -> 'a
  val the_only2 : (unit -> 'a) -> 'a Predicate.seq -> 'a
  val the2 : 'a Predicate.pred -> 'a
  val free_thread_id :
    ('a -> 'b -> ('c * ('d, Arith.nat) FinFun.finfun) option) ->
      'b -> 'a -> bool
  val thread_ok :
    ('a -> 'b -> ('c * ('d, Arith.nat) FinFun.finfun) option) ->
      'b -> ('a, 'c, 'e) FWState.new_thread_action -> bool
  val interrupt_action_ok :
    ('a -> 'b -> bool) -> 'b -> 'a FWState.interrupt_action -> bool
  val interrupt_actions_ok :
    ('a -> 'b -> bool) ->
      ('a -> 'b -> 'b) ->
        ('a -> 'b -> 'b) -> 'b -> 'a FWState.interrupt_action list -> bool
  val wset_actions_ok :
    'a HOL.equal -> 'c HOL.equal ->
      ('a -> 'b -> 'c FWState.wait_set_status option) ->
        'b -> 'a -> ('a, 'c) FWState.wait_set_action list -> bool
  val cond_action_ok :
    'b HOL.equal -> 'd Set.card_UNIV * 'd HOL.equal ->
      ('a -> bool) ->
        ('b -> 'c -> ('a * ('d, Arith.nat) FinFun.finfun) option) ->
          ('b -> 'e -> 'f FWState.wait_set_status option) ->
            ('d, (('b * Arith.nat) option)) FinFun.finfun *
              (('c * 'g) * ('e * 'h)) ->
              'b -> 'b FWState.conditional_action -> bool
  val cond_action_oks :
    'b HOL.equal -> 'd Set.card_UNIV * 'd HOL.equal ->
      ('a -> bool) ->
        ('b -> 'c -> ('a * ('d, Arith.nat) FinFun.finfun) option) ->
          ('b -> 'e -> 'f FWState.wait_set_status option) ->
            ('d, (('b * Arith.nat) option)) FinFun.finfun *
              (('c * 'g) * ('e * 'h)) ->
              'b -> 'b FWState.conditional_action list -> bool
  val thread_oks :
    ('a -> 'b -> ('c * ('d, Arith.nat) FinFun.finfun) option) ->
      ('a -> 'c * ('d, Arith.nat) FinFun.finfun -> 'b -> 'b) ->
        'b -> ('a, 'c, 'e) FWState.new_thread_action list -> bool
  val actions_ok :
    'b HOL.equal -> 'd Set.card_UNIV * 'd HOL.equal -> 'f HOL.equal ->
      ('a -> bool) ->
        ('b -> 'c -> ('a * ('d, Arith.nat) FinFun.finfun) option) ->
          ('b -> 'a * ('d, Arith.nat) FinFun.finfun -> 'c -> 'c) ->
            ('b -> 'e -> 'f FWState.wait_set_status option) ->
              ('b -> 'g -> bool) ->
                ('b -> 'g -> 'g) ->
                  ('b -> 'g -> 'g) ->
                    ('d, (('b * Arith.nat) option)) FinFun.finfun *
                      (('c * 'h) * ('e * 'g)) ->
                      'b -> ('d, (FWState.lock_action list)) FinFun.finfun *
                              (('b, 'a, 'h) FWState.new_thread_action list *
                                ('b FWState.conditional_action list *
                                  (('b, 'f) FWState.wait_set_action list *
                                    ('b FWState.interrupt_action list *
                                      'i list)))) ->
                              bool
  val step_thread :
    'b HOL.equal -> 'd Set.card_UNIV * 'd HOL.equal -> 'e HOL.equal ->
      ('a -> bool) ->
        ('b ->
          'a * 'c ->
            ((('d, (FWState.lock_action list)) FinFun.finfun *
               (('b, 'a, 'c) FWState.new_thread_action list *
                 ('b FWState.conditional_action list *
                   (('b, 'e) FWState.wait_set_action list *
                     ('b FWState.interrupt_action list * 'f list))))) *
              ('a * 'c))
              Predicate.pred) ->
          (('d, Arith.nat) FinFun.finfun -> 'f list) ->
            ('b -> 'g -> ('a * ('d, Arith.nat) FinFun.finfun) option) ->
              ('b -> 'a * ('d, Arith.nat) FinFun.finfun -> 'g -> 'g) ->
                ('b -> 'h -> 'e FWState.wait_set_status option) ->
                  ('b -> 'i -> bool) ->
                    ('b -> 'i -> 'i) ->
                      ('b -> 'i -> 'i) ->
                        (('d, (FWState.lock_action list)) FinFun.finfun *
                           (('b, 'a, 'c) FWState.new_thread_action list *
                             ('b FWState.conditional_action list *
                               (('b, 'e) FWState.wait_set_action list *
                                 ('b FWState.interrupt_action list *
                                   'f list)))) ->
                          'j) ->
                          ('d, (('b * Arith.nat) option)) FinFun.finfun *
                            (('g * 'c) * ('h * 'i)) ->
                            'b -> ('b * (((('d, (FWState.lock_action list))
     FinFun.finfun *
    (('b, 'a, 'c) FWState.new_thread_action list *
      ('b FWState.conditional_action list *
        (('b, 'e) FWState.wait_set_action list *
          ('b FWState.interrupt_action list * 'f list))))) *
   ('a * 'c)) option *
  'j)) option
  val active_threads :
    'b HOL.equal -> 'd Set.card_UNIV * 'd HOL.equal -> 'e HOL.equal ->
      ('a -> bool) ->
        ('b ->
          'a * 'c ->
            ((('d, (FWState.lock_action list)) FinFun.finfun *
               (('b, 'a, 'c) FWState.new_thread_action list *
                 ('b FWState.conditional_action list *
                   (('b, 'e) FWState.wait_set_action list *
                     ('b FWState.interrupt_action list * 'f list))))) *
              ('a * 'c))
              Predicate.pred) ->
          ('b -> 'g -> ('a * ('d, Arith.nat) FinFun.finfun) option) ->
            ('b -> 'a * ('d, Arith.nat) FinFun.finfun -> 'g -> 'g) ->
              ('g ->
                ('h -> bool) ->
                  ('b * ('a * ('d, Arith.nat) FinFun.finfun) -> 'h -> 'h) ->
                    'h -> 'h) ->
                ('b -> 'i -> 'e FWState.wait_set_status option) ->
                  ('b -> 'j -> bool) ->
                    ('b -> 'j -> 'j) ->
                      ('b -> 'j -> 'j) ->
                        (unit -> 'h) ->
                          ('b -> 'h -> 'h) ->
                            ('d, (('b * Arith.nat) option)) FinFun.finfun *
                              (('g * 'c) * ('i * 'j)) ->
                              'h
  val pick_wakeup_via_sel :
    'c HOL.equal ->
      ('a ->
        ('b -> 'c FWState.wait_set_status -> bool) ->
          ('b * 'c FWState.wait_set_status) option) ->
        'd -> 'b -> 'c -> 'a -> 'b option
end = struct

fun tllist_unfold f a =
  (case f a
    of Sum_Type.Inl (b, aa) =>
      TLList.Lazy_tllist (fn _ => Sum_Type.Inl (b, tllist_unfold f aa))
    | Sum_Type.Inr aa => TLList.tNil aa);

fun redT_updT thr_update ts (FWState.ThreadExists (v, va)) = ts
  | redT_updT thr_update ts (FWState.NewThread (t, x, m)) =
    thr_update t (x, FinFun.Finfun_const Arith.zero_nat) ts;

fun redT_updTs thr_update = List.foldl (redT_updT thr_update);

fun redT_updI is_ins is_delete is (FWState.IsInterrupted (t, b)) = is
  | redT_updI is_ins is_delete is (FWState.ClearInterrupt t) = is_delete t is
  | redT_updI is_ins is_delete is (FWState.Interrupt t) = is_ins t is;

fun redT_updIs is_ins is_delete is (ia :: ias) =
  redT_updIs is_ins is_delete (redT_updI is_ins is_delete is ia) ias
  | redT_updIs is_ins is_delete is [] = is;

fun exec_updW C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma t ws
  FWState.WokenUp = ws_delete t ws
  | exec_updW C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma t ws
    FWState.Notified = ws_delete t ws
  | exec_updW C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma ta
    ws (FWState.WakeUp t) =
    (case ws_lookup t ws of NONE => ws
      | SOME (FWState.InWS _) =>
        ws_update t (FWState.PostWS FWState.WSWokenUp) ws
      | SOME (FWState.PostWS _) => ws)
  | exec_updW C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma t ws
    (FWState.Suspend w) = ws_update t (FWState.InWS w) ws
  | exec_updW C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma t ws
    (FWState.NotifyAll w) =
    ws_iterate ws (fn _ => true)
      (fn (ta, wa) => fn wsa =>
        (if FWState.equal_wait_set_statusa C_ wa (FWState.InWS w)
          then ws_update ta (FWState.PostWS FWState.WSNotified) wsa else wsa))
      ws
  | exec_updW C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma t ws
    (FWState.Notify w) =
    (case pick_wakeup sigma t w ws of NONE => ws
      | SOME ta => ws_update ta (FWState.PostWS FWState.WSNotified) ws);

fun exec_updWs C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma t =
  List.foldl
    (exec_updW C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma t);

fun exec_upd B_ C_ G_ pick_wakeup thr_lookup thr_update ws_lookup ws_update
  ws_delete ws_iterate is_ins is_delete sigma s t ta x m =
  (FWLocking.redT_updLs G_ B_ (FWState.locks s) t (FWState.locks_a ta),
    ((thr_update t
        (x, FWLockingThread.redT_updLns G_ B_ (FWState.locks s) t
              (Product_Type.snd (Option.the (thr_lookup t (FWState.thr s))))
              (FWState.locks_a ta))
        (redT_updTs thr_update (FWState.thr s) (FWState.thr_a ta)),
       m),
      (exec_updWs C_ pick_wakeup ws_lookup ws_update ws_delete ws_iterate sigma
         t (FWState.wset s) (FWState.wset_a ta),
        redT_updIs is_ins is_delete (FWState.interrupts s)
          (FWState.interrupt_a ta))));

fun execT A_ D_ J_ convert_RA schedule pick_wakeup thr_lookup thr_update
  ws_lookup ws_update ws_delete ws_iterate is_ins is_delete sigma s =
  Option.bind (schedule sigma s)
    (fn a =>
      (case a
        of (t, (NONE, sigmaa)) =>
          let
            val (x, ln) = Option.the (thr_lookup t (FWState.thr s));
            val ta =
              (FinFun.Finfun_const [], ([], ([], ([], ([], convert_RA ln)))));
            val sa =
              (FWLocking.acquire_all A_ D_ (FWState.locks s) t ln,
                ((thr_update t (x, FinFun.Finfun_const Arith.zero_nat)
                    (FWState.thr s),
                   FWState.shr s),
                  (FWState.wset s, FWState.interrupts s)));
          in
            SOME (sigmaa, (t, (ta, sa)))
          end
        | (t, (SOME (ta, (x, m)), sigmaa)) =>
          SOME (sigmaa,
                 (t, (ta, exec_upd D_ J_ A_ pick_wakeup thr_lookup thr_update
                            ws_lookup ws_update ws_delete ws_iterate is_ins
                            is_delete sigma s t ta x m)))));

fun exec_step A_ D_ J_ convert_RA schedule pick_wakeup thr_lookup thr_update
  ws_lookup ws_update ws_delete ws_iterate is_ins is_delete (sigma, s) =
  (case execT A_ D_ J_ convert_RA schedule pick_wakeup thr_lookup thr_update
          ws_lookup ws_update ws_delete ws_iterate is_ins is_delete sigma s
    of NONE => Sum_Type.Inr s
    | SOME (sigmaa, (t, (ta, sa))) =>
      Sum_Type.Inl ((sigma, (t, ta)), (sigmaa, sa)));

fun exec_aux A_ D_ J_ convert_RA schedule pick_wakeup thr_lookup thr_update
  ws_lookup ws_update ws_delete ws_iterate is_ins is_delete sigma_s =
  tllist_unfold
    (exec_step A_ D_ J_ convert_RA schedule pick_wakeup thr_lookup thr_update
      ws_lookup ws_update ws_delete ws_iterate is_ins is_delete)
    sigma_s;

fun exec A_ D_ J_ convert_RA schedule output pick_wakeup thr_lookup thr_update
  ws_lookup ws_update ws_delete ws_iterate is_ins is_delete sigma s =
  TLList.tmap Option.the Fun.id
    (TLList.tfilter (fn _ => (raise Fail "undefined"))
      (fn q => not (Option.is_none q))
      (TLList.tmap (fn (sigmaa, (a, b)) => output sigmaa a b) Fun.id
        (exec_aux A_ D_ J_ convert_RA schedule pick_wakeup thr_lookup thr_update
          ws_lookup ws_update ws_delete ws_iterate is_ins is_delete
          (sigma, s))));

fun singleton2 dfault (Predicate.Seq f) =
  (case f () of Predicate.Empty => dfault ()
    | Predicate.Insert (x, p) =>
      (if Predicate.is_empty p then x
        else (raise Fail "singleton2 not unique")
               (fn _ => singleton2 dfault (Predicate.Seq f)))
    | Predicate.Join (p, xq) =>
      (if Predicate.is_empty p then the_only2 dfault xq
        else (if Predicate.null xq then singleton2 dfault p
               else (raise Fail "singleton2 not unique")
                      (fn _ => singleton2 dfault (Predicate.Seq f)))))
and the_only2 dfault (Predicate.Join (p, xq)) =
  (if Predicate.is_empty p then the_only2 dfault xq
    else (if Predicate.null xq then singleton2 dfault p
           else (raise Fail "the_only2 not unique")
                  (fn _ => the_only2 dfault (Predicate.Join (p, xq)))))
  | the_only2 dfault (Predicate.Insert (x, p)) =
    (if Predicate.is_empty p then x
      else (raise Fail "the_only2 not unique")
             (fn _ => the_only2 dfault (Predicate.Insert (x, p))))
  | the_only2 dfault Predicate.Empty = (raise Fail "the_only2 empty") dfault;

fun the2 a = singleton2 (fn _ => (raise Fail "not_unique") (fn _ => the2 a)) a;

fun free_thread_id thr_lookup ts t = Option.is_none (thr_lookup t ts);

fun thread_ok thr_lookup ts (FWState.ThreadExists (t, b)) =
  not (Product_Type.equal_boola b (free_thread_id thr_lookup ts t))
  | thread_ok thr_lookup ts (FWState.NewThread (t, x, m)) =
    free_thread_id thr_lookup ts t;

fun interrupt_action_ok is_memb is (FWState.IsInterrupted (t, b)) =
  Product_Type.equal_boola b (is_memb t is)
  | interrupt_action_ok is_memb is (FWState.ClearInterrupt t) = true
  | interrupt_action_ok is_memb is (FWState.Interrupt t) = true;

fun interrupt_actions_ok is_memb is_ins is_delete is (ia :: ias) =
  interrupt_action_ok is_memb is ia andalso
    interrupt_actions_ok is_memb is_ins is_delete
      (redT_updI is_ins is_delete is ia) ias
  | interrupt_actions_ok is_memb is_ins is_delete is [] = true;

fun wset_actions_ok A_ C_ ws_lookup ws t was =
  Option.equal_optiona (FWState.equal_wait_set_status C_) (ws_lookup t ws)
    (if List.member (FWState.equal_wait_set_action A_ C_) was FWState.Notified
      then SOME (FWState.PostWS FWState.WSNotified)
      else (if List.member (FWState.equal_wait_set_action A_ C_) was
                 FWState.WokenUp
             then SOME (FWState.PostWS FWState.WSWokenUp) else NONE));

fun cond_action_ok B_ (D1_, D2_) final thr_lookup ws_lookup s t FWState.Yield =
  true
  | cond_action_ok B_ (D1_, D2_) final thr_lookup ws_lookup s ta
    (FWState.Join t) =
    (case thr_lookup t (FWState.thr s) of NONE => true
      | SOME (x, ln) =>
        not (HOL.eq B_ ta t) andalso
          (final x andalso
            (HOL.eq (FinFun.equal_finfun (D1_, D2_) Arith.equal_nat) ln
               (FinFun.Finfun_const Arith.zero_nat) andalso
              Option.is_none (ws_lookup t (FWState.wset s)))));

fun cond_action_oks B_ (D1_, D2_) final thr_lookup ws_lookup s t cts =
  List.list_all (cond_action_ok B_ (D1_, D2_) final thr_lookup ws_lookup s t)
    cts;

fun thread_oks thr_lookup thr_update ts (ta :: tas) =
  thread_ok thr_lookup ts ta andalso
    thread_oks thr_lookup thr_update (redT_updT thr_update ts ta) tas
  | thread_oks thr_lookup thr_update ts [] = true;

fun actions_ok B_ (D1_, D2_) F_ final thr_lookup thr_update ws_lookup is_memb
  is_ins is_delete s t ta =
  FWLocking.lock_ok_las (D1_, D2_) B_ (FWState.locks s) t
    (FWState.locks_a ta) andalso
    (thread_oks thr_lookup thr_update (FWState.thr s) (FWState.thr_a ta) andalso
      (cond_action_oks B_ (D1_, D2_) final thr_lookup ws_lookup s t
         (FWState.cond_a ta) andalso
        (wset_actions_ok B_ F_ ws_lookup (FWState.wset s) t
           (FWState.wset_a ta) andalso
          interrupt_actions_ok is_memb is_ins is_delete (FWState.interrupts s)
            (FWState.interrupt_a ta))));

fun step_thread B_ (D1_, D2_) E_ final r convert_RA thr_lookup thr_update
  ws_lookup is_memb is_ins is_delete update_state s t =
  (case thr_lookup t (FWState.thr s) of NONE => NONE
    | SOME (x, ln) =>
      (if HOL.eq (FinFun.equal_finfun (D1_, D2_) Arith.equal_nat) ln
            (FinFun.Finfun_const Arith.zero_nat)
        then let
               val reds =
                 Predicate.bind (r t (x, FWState.shr s))
                   (fn (ta, (xa, m)) =>
                     (if actions_ok B_ (D1_, D2_) E_ final thr_lookup thr_update
                           ws_lookup is_memb is_ins is_delete s t ta
                       then Predicate.single (ta, (xa, m))
                       else Predicate.bot_pred));
             in
               (if Predicate.holds
                     (Predicate.bind reds (fn _ => Predicate.single ()))
                 then let
                        val (ta, (xa, m)) = the2 reds;
                      in
                        SOME (t, (SOME (ta, (xa, m)), update_state ta))
                      end
                 else NONE)
             end
        else (if FWLocking.may_acquire_all (D1_, D2_) B_ (FWState.locks s) t
                   ln andalso
                   not (FWWait.waiting (ws_lookup t (FWState.wset s)))
               then SOME (t, (NONE,
                               update_state
                                 (FinFun.Finfun_const [],
                                   ([], ([], ([], ([], convert_RA ln)))))))
               else NONE)));

fun active_threads B_ (D1_, D2_) E_ final r thr_lookup thr_update thr_iterate
  ws_lookup is_memb is_ins is_delete thr_empty thr_ins_dj
  (ls, ((ts, m), (ws, is))) =
  thr_iterate ts (fn _ => true)
    (fn (t, (x, ln)) => fn tsa =>
      (if HOL.eq (FinFun.equal_finfun (D1_, D2_) Arith.equal_nat) ln
            (FinFun.Finfun_const Arith.zero_nat)
        then (if Predicate.holds
                   (Predicate.bind (r t (x, m))
                     (fn (ta, _) =>
                       Predicate.if_pred
                         (actions_ok B_ (D1_, D2_) E_ final thr_lookup
                           thr_update ws_lookup is_memb is_ins is_delete
                           (ls, ((ts, m), (ws, is))) t ta)))
               then thr_ins_dj t tsa else tsa)
        else (if not (FWWait.waiting (ws_lookup t ws)) andalso
                   FWLocking.may_acquire_all (D1_, D2_) B_ ls t ln
               then thr_ins_dj t tsa else tsa)))
    (thr_empty ());

fun pick_wakeup_via_sel C_ ws_sel sigma t w ws =
  Option.map Product_Type.fst
    (ws_sel ws
      (fn _ => fn wa => FWState.equal_wait_set_statusa C_ wa (FWState.InWS w)));

end; (*struct Scheduler*)

structure Random_Scheduler : sig
  val next_thread :
    ('a -> 'b list) ->
      Arith.natural * Arith.natural ->
        'a -> ('b * (Arith.natural * Arith.natural)) option
  val random_scheduler :
    'b HOL.equal -> 'd Set.card_UNIV * 'd HOL.equal -> 'e HOL.equal ->
      ('a -> bool) ->
        ('b ->
          'a * 'c ->
            ((('d, (FWState.lock_action list)) FinFun.finfun *
               (('b, 'a, 'c) FWState.new_thread_action list *
                 ('b FWState.conditional_action list *
                   (('b, 'e) FWState.wait_set_action list *
                     ('b FWState.interrupt_action list * 'f list))))) *
              ('a * 'c))
              Predicate.pred) ->
          (('d, Arith.nat) FinFun.finfun -> 'f list) ->
            ('b -> 'g -> ('a * ('d, Arith.nat) FinFun.finfun) option) ->
              ('b -> 'a * ('d, Arith.nat) FinFun.finfun -> 'g -> 'g) ->
                ('g ->
                  ('h -> bool) ->
                    ('b * ('a * ('d, Arith.nat) FinFun.finfun) -> 'h -> 'h) ->
                      'h -> 'h) ->
                  ('b -> 'i -> 'e FWState.wait_set_status option) ->
                    ('b -> 'j -> bool) ->
                      ('b -> 'j -> 'j) ->
                        ('b -> 'j -> 'j) ->
                          (unit -> 'h) ->
                            ('b -> 'h -> 'h) ->
                              ('h -> 'b list) ->
                                Arith.natural * Arith.natural ->
                                  ('d, (('b * Arith.nat) option))
                                    FinFun.finfun *
                                    (('g * 'c) * ('i * 'j)) ->
                                    ('b * (((('d, (FWState.lock_action list))
       FinFun.finfun *
      (('b, 'a, 'c) FWState.new_thread_action list *
        ('b FWState.conditional_action list *
          (('b, 'e) FWState.wait_set_action list *
            ('b FWState.interrupt_action list * 'f list))))) *
     ('a * 'c)) option *
    (Arith.natural * Arith.natural))) option
end = struct

fun next_thread thr_to_list seed active =
  let
    val ts = thr_to_list active;
  in
    (if List.null ts then NONE
      else SOME (Random.select (thr_to_list active) seed))
  end;

fun random_scheduler B_ (D1_, D2_) E_ final r convert_RA thr_lookup thr_update
  thr_iterate ws_lookup is_memb is_ins is_delete thr_empty thr_ins_dj
  thr_to_list seed s =
  Option.bind
    (next_thread thr_to_list seed
      (Scheduler.active_threads B_ (D1_, D2_) E_ final r thr_lookup thr_update
        thr_iterate ws_lookup is_memb is_ins is_delete thr_empty thr_ins_dj s))
    (fn (t, seeda) =>
      Scheduler.step_thread B_ (D1_, D2_) E_ final r convert_RA thr_lookup
        thr_update ws_lookup is_memb is_ins is_delete (fn _ => seeda) s t);

end; (*struct Random_Scheduler*)

structure CollectionsV1 : sig
  val equal_rbta :
    'a HOL.equal * 'a Orderings.linorder -> 'b HOL.equal ->
      ('a, 'b) RBT.rbt -> ('a, 'b) RBT.rbt -> bool
  val equal_rbt :
    'a HOL.equal * 'a Orderings.linorder -> 'b HOL.equal ->
      ('a, 'b) RBT.rbt HOL.equal
end = struct

fun equal_rbta (A1_, A2_) B_ ra r =
  RBT_Impl.equal_rbt A1_ B_ (RBT.impl_of A2_ ra) (RBT.impl_of A2_ r);

fun equal_rbt (A1_, A2_) B_ = {equal = equal_rbta (A1_, A2_) B_} :
  ('a, 'b) RBT.rbt HOL.equal;

end; (*struct CollectionsV1*)

structure ListMapImpl : sig
  val g_list_to_map_lm_basic_ops :
    'a HOL.equal -> ('a * 'b) list -> ('a, 'b) Assoc_List.assoc_list
end = struct

fun g_list_to_map_lm_basic_ops A_ l =
  List.foldl (fn m => fn (k, v) => Assoc_List.update A_ k v m) Assoc_List.empty
    (List.rev l);

end; (*struct ListMapImpl*)

structure SC_Collections : sig
  datatype heapobj =
    Obj of
      string *
        (char, (string, Arith.nat Value.vala) Assoc_List.assoc_list) Trie.trie
    | Arr of
        Type.ty * Arith.nat *
          (string, Arith.nat Value.vala) Assoc_List.assoc_list *
          (Arith.nat, Arith.nat Value.vala) RBT.rbt
  val init_fields_array :
    (string * Type.ty) list ->
      (string, Arith.nat Value.vala) Assoc_List.assoc_list
  val init_fields :
    ((string * string) * Type.ty) list ->
      (char, (string, Arith.nat Value.vala) Assoc_List.assoc_list) Trie.trie
  val init_cells :
    Type.ty -> Arith.nat -> (Arith.nat, Arith.nat Value.vala) RBT.rbt
  val blank : 'a TypeRel.prog -> Type.htype -> heapobj
  val obj_ty : heapobj -> Type.htype
  val new_Addr : (Arith.nat, heapobj) RBT.rbt -> Arith.nat option
  val sc_allocate :
    'a TypeRel.prog ->
      (Arith.nat, heapobj) RBT.rbt ->
        Type.htype -> ((Arith.nat, heapobj) RBT.rbt * Arith.nat) Set.set
  val equal_heapobja : heapobj -> heapobj -> bool
  val equal_heapobj : heapobj HOL.equal
  val sc_typeof_addr :
    (Arith.nat, heapobj) RBT.rbt -> Arith.nat -> Type.htype option
  val sc_spurious_wakeups : bool
  val sc_heap_read_i_i_i_o :
    (Arith.nat, heapobj) RBT.rbt ->
      Arith.nat -> Heap.addr_loc -> Arith.nat Value.vala Predicate.pred
  val sc_heap_write_i_i_i_i_o :
    (Arith.nat, heapobj) RBT.rbt ->
      Arith.nat ->
        Heap.addr_loc ->
          Arith.nat Value.vala -> (Arith.nat, heapobj) RBT.rbt Predicate.pred
end = struct

datatype heapobj =
  Obj of
    string *
      (char, (string, Arith.nat Value.vala) Assoc_List.assoc_list) Trie.trie
  | Arr of
      Type.ty * Arith.nat *
        (string, Arith.nat Value.vala) Assoc_List.assoc_list *
        (Arith.nat, Arith.nat Value.vala) RBT.rbt;

fun init_fields_array x =
  (ListMapImpl.g_list_to_map_lm_basic_ops Stringa.equal_literal o
    List.map (fn (f, t) => (f, Value.default_val t)))
    x;

fun init_fields fDTs =
  List.foldr
    (fn (a, b) =>
      let
        val (f, d) = a;
      in
        (fn t => fn fields =>
          let
            val fa = String.explode f;
          in
            Trie.update Stringa.equal_char fa
              (Assoc_List.update Stringa.equal_literal d (Value.default_val t)
                (case Trie.lookup Stringa.equal_char fields fa
                  of NONE => Assoc_List.empty | SOME lm => lm))
              fields
          end)
      end
        b)
    fDTs Trie.empty;

fun init_cells t n =
  List.foldl
    (fn cells => fn i =>
      RBT.insert Arith.linorder_nat i (Value.default_val t) cells)
    (RBT.empty Arith.linorder_nat) (List.upt Arith.zero_nat n);

fun blank p (Type.Class_type c) =
  Obj (c, init_fields
            (List.map (fn (fd, (t, _)) => (fd, t)) (TypeRel.fields p c)))
  | blank p (Type.Array_type (t, n)) =
    Arr (t, n,
          init_fields_array
            (List.map
              (fn (a, b) =>
                let
                  val (f, _) = a;
                in
                  (fn (ta, _) => (f, ta))
                end
                  b)
              (TypeRel.fields p Type.object)),
          init_cells t n);

fun obj_ty (Obj (c, f)) = Type.Class_type c
  | obj_ty (Arr (t, si, f, e)) = Type.Array_type (t, si);

fun new_Addr h =
  SOME (case RBTMapImpl.g_max_rm_basic_ops Arith.linorder_nat h (fn _ => true)
         of NONE => Arith.zero_nat
         | SOME (a, _) => Arith.plus_nat a Arith.one_nat);

fun sc_allocate p h hT =
  (case new_Addr h of NONE => Set.bot_set
    | SOME a =>
      Set.insert (RBT.insert Arith.linorder_nat a (blank p hT) h, a)
        Set.bot_set);

fun equal_heapobja (Arr (ty, nat, assoc_list, rbt)) (Obj (literal, trie)) =
  false
  | equal_heapobja (Obj (literal, trie)) (Arr (ty, nat, assoc_list, rbt)) =
    false
  | equal_heapobja (Arr (tya, nata, assoc_lista, rbta))
    (Arr (ty, nat, assoc_list, rbt)) =
    Type.equal_tya tya ty andalso
      (Arith.equal_nata nata nat andalso
        (Assoc_List.equal_assoc_lista Stringa.equal_literal
           (Value.equal_val Arith.equal_nat) assoc_lista assoc_list andalso
          CollectionsV1.equal_rbta (Arith.equal_nat, Arith.linorder_nat)
            (Value.equal_val Arith.equal_nat) rbta rbt))
  | equal_heapobja (Obj (literala, triea)) (Obj (literal, trie)) =
    ((literala : string) = literal) andalso
      Trie.equal_trie Stringa.equal_char
        (Assoc_List.equal_assoc_list Stringa.equal_literal
          (Value.equal_val Arith.equal_nat))
        triea trie;

val equal_heapobj = {equal = equal_heapobja} : heapobj HOL.equal;

fun sc_typeof_addr h a = Option.map obj_ty (RBT.lookup Arith.linorder_nat h a);

val sc_spurious_wakeups : bool = false;

fun sc_heap_read_i_i_i_o x xa xb =
  Predicate.sup_pred
    (case xb
      of Heap.CField (d, f) =>
        (case RBT.lookup Arith.linorder_nat x xa of NONE => Predicate.bot_pred
          | SOME (Obj (_, fs)) =>
            (case Trie.lookup Stringa.equal_char fs (String.explode f)
              of NONE => Predicate.bot_pred
              | SOME fsa =>
                (case Assoc_List.lookup Stringa.equal_literal fsa d
                  of NONE => Predicate.bot_pred | SOME a => Predicate.single a))
          | SOME (Arr (_, _, _, _)) => Predicate.bot_pred)
      | Heap.ACell _ => Predicate.bot_pred)
    (Predicate.sup_pred
      (case xb of Heap.CField (_, _) => Predicate.bot_pred
        | Heap.ACell n =>
          (case RBT.lookup Arith.linorder_nat x xa of NONE => Predicate.bot_pred
            | SOME (Obj (_, _)) => Predicate.bot_pred
            | SOME (Arr (_, si, _, el)) =>
              Predicate.bind (Predicate.if_pred (Arith.less_nat n si))
                (fn () =>
                  Predicate.single
                    (Option.the (RBT.lookup Arith.linorder_nat el n)))))
      (case xb
        of Heap.CField (xba, f) =>
          (case RBT.lookup Arith.linorder_nat x xa of NONE => Predicate.bot_pred
            | SOME (Obj (_, _)) => Predicate.bot_pred
            | SOME (Arr (_, _, fa, _)) =>
              (case Assoc_List.lookup Stringa.equal_literal fa f
                of NONE => Predicate.bot_pred
                | SOME v =>
                  (if ((xba : string) = Type.object) then Predicate.single v
                    else Predicate.bot_pred)))
        | Heap.ACell _ => Predicate.bot_pred));

fun sc_heap_write_i_i_i_i_o x xa xb xc =
  Predicate.sup_pred
    (case xb
      of Heap.CField (d, f) =>
        (case RBT.lookup Arith.linorder_nat x xa of NONE => Predicate.bot_pred
          | SOME (Obj (c, fs)) =>
            Predicate.single
              (RBT.insert Arith.linorder_nat xa
                (Obj (c, Trie.update Stringa.equal_char (String.explode f)
                           (Assoc_List.update Stringa.equal_literal d xc
                             (case Trie.lookup Stringa.equal_char fs
                                     (String.explode f)
                               of NONE => Assoc_List.empty | SOME fsa => fsa))
                           fs))
                x)
          | SOME (Arr (_, _, _, _)) => Predicate.bot_pred)
      | Heap.ACell _ => Predicate.bot_pred)
    (Predicate.sup_pred
      (case xb of Heap.CField (_, _) => Predicate.bot_pred
        | Heap.ACell n =>
          (case RBT.lookup Arith.linorder_nat x xa of NONE => Predicate.bot_pred
            | SOME (Obj (_, _)) => Predicate.bot_pred
            | SOME (Arr (t, si, f, el)) =>
              Predicate.single
                (RBT.insert Arith.linorder_nat xa
                  (Arr (t, si, f, RBT.insert Arith.linorder_nat n xc el)) x)))
      (case xb
        of Heap.CField (xba, f) =>
          (case RBT.lookup Arith.linorder_nat x xa of NONE => Predicate.bot_pred
            | SOME (Obj (_, _)) => Predicate.bot_pred
            | SOME (Arr (t, si, fa, el)) =>
              (if ((xba : string) = Type.object)
                then Predicate.single
                       (RBT.insert Arith.linorder_nat xa
                         (Arr (t, si,
                                Assoc_List.update Stringa.equal_literal f xc fa,
                                el))
                         x)
                else Predicate.bot_pred))
        | Heap.ACell _ => Predicate.bot_pred));

end; (*struct SC_Collections*)

structure Round_Robin : sig
  val enqueue_new_thread :
    ('a -> 'b -> 'b) -> 'b -> ('a, 'c, 'd) FWState.new_thread_action -> 'b
  val enqueue_new_threads :
    ('a -> 'b -> 'b) -> 'b -> ('a, 'c, 'd) FWState.new_thread_action list -> 'b
  val queue_rotate1 : ('a -> 'b -> 'b) -> ('b -> 'a * 'b) -> 'b -> 'b
  val round_robin_update_state :
    'a HOL.equal ->
      ('a -> 'b -> 'b) ->
        ('b -> 'a * 'b) ->
          Arith.nat ->
            'b * Arith.nat ->
              'a -> ('c, (FWState.lock_action list)) FinFun.finfun *
                      (('a, 'd, 'e) FWState.new_thread_action list *
                        ('a FWState.conditional_action list *
                          (('a, 'f) FWState.wait_set_action list *
                            ('a FWState.interrupt_action list * 'g list)))) ->
                      'b * Arith.nat
  val round_robin_reschedule :
    'b HOL.equal -> 'd Set.card_UNIV * 'd HOL.equal -> 'e HOL.equal ->
      ('a -> bool) ->
        ('b ->
          'a * 'c ->
            ((('d, (FWState.lock_action list)) FinFun.finfun *
               (('b, 'a, 'c) FWState.new_thread_action list *
                 ('b FWState.conditional_action list *
                   (('b, 'e) FWState.wait_set_action list *
                     ('b FWState.interrupt_action list * 'f list))))) *
              ('a * 'c))
              Predicate.pred) ->
          (('d, Arith.nat) FinFun.finfun -> 'f list) ->
            ('b -> 'g -> ('a * ('d, Arith.nat) FinFun.finfun) option) ->
              ('b -> 'a * ('d, Arith.nat) FinFun.finfun -> 'g -> 'g) ->
                ('b -> 'h -> 'e FWState.wait_set_status option) ->
                  ('b -> 'i -> bool) ->
                    ('b -> 'i -> 'i) ->
                      ('b -> 'i -> 'i) ->
                        ('b -> 'j -> 'j) ->
                          ('j -> 'b * 'j) ->
                            ('b -> 'j -> 'j) ->
                              'b -> 'j -> Arith.nat ->
    ('d, (('b * Arith.nat) option)) FinFun.finfun * (('g * 'c) * ('h * 'i)) ->
      ('b * (((('d, (FWState.lock_action list)) FinFun.finfun *
                (('b, 'a, 'c) FWState.new_thread_action list *
                  ('b FWState.conditional_action list *
                    (('b, 'e) FWState.wait_set_action list *
                      ('b FWState.interrupt_action list * 'f list))))) *
               ('a * 'c)) option *
              ('j * Arith.nat))) option
  val round_robin :
    'b HOL.equal -> 'd Set.card_UNIV * 'd HOL.equal -> 'e HOL.equal ->
      ('a -> bool) ->
        ('b ->
          'a * 'c ->
            ((('d, (FWState.lock_action list)) FinFun.finfun *
               (('b, 'a, 'c) FWState.new_thread_action list *
                 ('b FWState.conditional_action list *
                   (('b, 'e) FWState.wait_set_action list *
                     ('b FWState.interrupt_action list * 'f list))))) *
              ('a * 'c))
              Predicate.pred) ->
          (('d, Arith.nat) FinFun.finfun -> 'f list) ->
            ('b -> 'g -> ('a * ('d, Arith.nat) FinFun.finfun) option) ->
              ('b -> 'a * ('d, Arith.nat) FinFun.finfun -> 'g -> 'g) ->
                ('b -> 'h -> 'e FWState.wait_set_status option) ->
                  ('b -> 'i -> bool) ->
                    ('b -> 'i -> 'i) ->
                      ('b -> 'i -> 'i) ->
                        ('j -> bool) ->
                          ('b -> 'j -> 'j) ->
                            ('j -> 'b * 'j) ->
                              ('b -> 'j -> 'j) ->
                                Arith.nat ->
                                  'j * Arith.nat ->
                                    ('d, (('b * Arith.nat) option))
                                      FinFun.finfun *
                                      (('g * 'c) * ('h * 'i)) ->
                                      ('b *
(((('d, (FWState.lock_action list)) FinFun.finfun *
    (('b, 'a, 'c) FWState.new_thread_action list *
      ('b FWState.conditional_action list *
        (('b, 'e) FWState.wait_set_action list *
          ('b FWState.interrupt_action list * 'f list))))) *
   ('a * 'c)) option *
  ('j * Arith.nat))) option
  val round_robin_start :
    (unit -> 'a) -> ('b -> 'a -> 'a) -> Arith.nat -> 'b -> 'a * Arith.nat
end = struct

fun enqueue_new_thread queue_enqueue ts (FWState.ThreadExists (t, b)) = ts
  | enqueue_new_thread queue_enqueue ts (FWState.NewThread (t, x, m)) =
    queue_enqueue t ts;

fun enqueue_new_threads queue_enqueue =
  List.foldl (enqueue_new_thread queue_enqueue);

fun queue_rotate1 queue_enqueue queue_dequeue =
  (fn (a, b) => queue_enqueue a b) o queue_dequeue;

fun round_robin_update_state A_ queue_enqueue queue_dequeue n0 (queue, n) t ta =
  let
    val queuea = enqueue_new_threads queue_enqueue queue (FWState.thr_a ta);
  in
    (if Arith.equal_nata n Arith.zero_nat orelse
          List.member (FWState.equal_conditional_action A_) (FWState.cond_a ta)
            FWState.Yield
      then (queue_rotate1 queue_enqueue queue_dequeue queuea, n0)
      else (queuea, Arith.minus_nat n Arith.one_nat))
  end;

fun round_robin_reschedule B_ (D1_, D2_) E_ final r convert_RA thr_lookup
  thr_update ws_lookup is_memb is_ins is_delete queue_enqueue queue_dequeue
  queue_push t0 queue n0 s =
  let
    val (t, queuea) = queue_dequeue queue;
  in
    (if HOL.eq B_ t t0 then NONE
      else (case Scheduler.step_thread B_ (D1_, D2_) E_ final r convert_RA
                   thr_lookup thr_update ws_lookup is_memb is_ins is_delete
                   (round_robin_update_state B_ queue_enqueue queue_dequeue n0
                     (queue_push t queuea, n0) t)
                   s t
             of NONE =>
               round_robin_reschedule B_ (D1_, D2_) E_ final r convert_RA
                 thr_lookup thr_update ws_lookup is_memb is_ins is_delete
                 queue_enqueue queue_dequeue queue_push t0
                 (queue_enqueue t queuea) n0 s
             | SOME a => SOME a))
  end;

fun round_robin B_ (D1_, D2_) E_ final r convert_RA thr_lookup thr_update
  ws_lookup is_memb is_ins is_delete queue_isEmpty queue_enqueue queue_dequeue
  queue_push n0 (queue, n) s =
  (if queue_isEmpty queue then NONE
    else let
           val (t, queuea) = queue_dequeue queue;
         in
           (case Scheduler.step_thread B_ (D1_, D2_) E_ final r convert_RA
                   thr_lookup thr_update ws_lookup is_memb is_ins is_delete
                   (round_robin_update_state B_ queue_enqueue queue_dequeue n0
                     (queue_push t queuea, n) t)
                   s t
             of NONE =>
               round_robin_reschedule B_ (D1_, D2_) E_ final r convert_RA
                 thr_lookup thr_update ws_lookup is_memb is_ins is_delete
                 queue_enqueue queue_dequeue queue_push t
                 (queue_enqueue t queuea) n0 s
             | SOME a => SOME a)
         end);

fun round_robin_start queue_empty queue_enqueue n0 t =
  (queue_enqueue t (queue_empty ()), n0);

end; (*struct Round_Robin*)

structure RBTSetImpl : sig
  val ins_rm_basic_ops :
    'a Orderings.linorder -> 'a -> ('a, unit) RBT.rbt -> ('a, unit) RBT.rbt
  val memb_rm_basic_ops :
    'a Orderings.linorder -> 'a -> ('a, unit) RBT.rbt -> bool
  val empty_rm_basic_ops : 'a Orderings.linorder -> unit -> ('a, unit) RBT.rbt
  val delete_rm_basic_ops :
    'a Orderings.linorder -> 'a -> ('a, unit) RBT.rbt -> ('a, unit) RBT.rbt
end = struct

fun ins_rm_basic_ops A_ x s = RBT.insert A_ x () s;

fun memb_rm_basic_ops A_ x s = not (Option.is_none (RBT.lookup A_ s x));

fun empty_rm_basic_ops A_ = (fn _ => RBT.empty A_);

fun delete_rm_basic_ops A_ x s = RBT.delete A_ x s;

end; (*struct RBTSetImpl*)

structure SmallStep : sig
  val extNTA2J :
    'a HOL.equal ->
      (string list * (string, unit, 'a) Expr.exp) TypeRel.prog ->
        string * (string * 'a) ->
          (string, unit, 'a) Expr.exp * (string -> 'a Value.vala option)
  val red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o :
    'a HOL.equal * 'a Value.addr -> 'c HOL.equal ->
      ('a -> 'b) ->
        ('b -> 'a) ->
          bool ->
            'c -> ('c -> Type.htype -> ('c * 'a) Set.set) ->
                    ('c -> 'a -> Type.htype option) ->
                      ('c ->
                        'a -> Heap.addr_loc -> 'a Value.vala Predicate.pred) ->
                        ('c ->
                          'a -> Heap.addr_loc ->
                                  'a Value.vala -> 'c Predicate.pred) ->
                          (('a, (FWState.lock_action list)) FinFun.finfun *
                             (('b, (string * (string * 'a)), 'c)
                                FWState.new_thread_action list *
                               ('b FWState.conditional_action list *
                                 (('b, 'a) FWState.wait_set_action list *
                                   ('b FWState.interrupt_action list *
                                     ('a, 'b)
                                       Observable_Events.obs_event list)))) ->
                            ('a, (FWState.lock_action list)) FinFun.finfun *
                              (('b, 'd, 'c) FWState.new_thread_action list *
                                ('b FWState.conditional_action list *
                                  (('b, 'a) FWState.wait_set_action list *
                                    ('b FWState.interrupt_action list *
                                      ('a, 'b)
Observable_Events.obs_event list))))) ->
                            (string list * (string, unit, 'a) Expr.exp)
                              TypeRel.prog ->
                              'b -> (string, unit, 'a) Expr.exp ->
                                      'c * (string -> 'a Value.vala option) ->
((('a, (FWState.lock_action list)) FinFun.finfun *
   (('b, 'd, 'c) FWState.new_thread_action list *
     ('b FWState.conditional_action list *
       (('b, 'a) FWState.wait_set_action list *
         ('b FWState.interrupt_action list *
           ('a, 'b) Observable_Events.obs_event list))))) *
  ((string, unit, 'a) Expr.exp * ('c * (string -> 'a Value.vala option))))
  Predicate.pred
  val reds_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o :
    'a HOL.equal * 'a Value.addr -> 'c HOL.equal ->
      ('a -> 'b) ->
        ('b -> 'a) ->
          bool ->
            'c -> ('c -> Type.htype -> ('c * 'a) Set.set) ->
                    ('c -> 'a -> Type.htype option) ->
                      ('c ->
                        'a -> Heap.addr_loc -> 'a Value.vala Predicate.pred) ->
                        ('c ->
                          'a -> Heap.addr_loc ->
                                  'a Value.vala -> 'c Predicate.pred) ->
                          (('a, (FWState.lock_action list)) FinFun.finfun *
                             (('b, (string * (string * 'a)), 'c)
                                FWState.new_thread_action list *
                               ('b FWState.conditional_action list *
                                 (('b, 'a) FWState.wait_set_action list *
                                   ('b FWState.interrupt_action list *
                                     ('a, 'b)
                                       Observable_Events.obs_event list)))) ->
                            ('a, (FWState.lock_action list)) FinFun.finfun *
                              (('b, 'd, 'c) FWState.new_thread_action list *
                                ('b FWState.conditional_action list *
                                  (('b, 'a) FWState.wait_set_action list *
                                    ('b FWState.interrupt_action list *
                                      ('a, 'b)
Observable_Events.obs_event list))))) ->
                            (string list * (string, unit, 'a) Expr.exp)
                              TypeRel.prog ->
                              'b -> (string, unit, 'a) Expr.exp list ->
                                      'c * (string -> 'a Value.vala option) ->
((('a, (FWState.lock_action list)) FinFun.finfun *
   (('b, 'd, 'c) FWState.new_thread_action list *
     ('b FWState.conditional_action list *
       (('b, 'a) FWState.wait_set_action list *
         ('b FWState.interrupt_action list *
           ('a, 'b) Observable_Events.obs_event list))))) *
  ((string, unit, 'a) Expr.exp list * ('c * (string -> 'a Value.vala option))))
  Predicate.pred
end = struct

fun extNTA2J A_ p =
  (fn (c, (m, a)) =>
    let
      val (d, (_, (_, meth))) =
        TypeRel.method
          (Product_Type.equal_prod (List.equal_list Stringa.equal_literal)
            (Expr.equal_exp Stringa.equal_literal Product_Type.equal_unit A_))
          p c m;
      val (_, body) = Option.the meth;
    in
      (Expr.Block ("this", Type.Class d, SOME (Value.Addr a), body),
        (fn _ => NONE))
    end);

fun red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) C_ xl xm xn
  xo xp xq heap_read heap_write xr xs xt xu (xv, xw) =
  (case xu
    of Expr.New c1 =>
      Predicate.sup_pred
        (Predicate.bind
          (TypeRel.eq_i_i (Set.equal_set (Product_Type.equal_prod C_ A1_))
            (xp xv (Type.Class_type c1)) Set.bot_set)
          (fn () =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw
                   (Expr.Val
                     (Value.Addr
                       (StartConfig.addr_of_sys_xcpt C_ (A1_, A2_) xo xp
                         "java/lang/OutOfMemoryError"))),
                  (xv, xw)))))
        (Predicate.bind (Predicate.pred_of_set (xp xv (Type.Class_type c1)))
          (fn (h, a) =>
            Predicate.single
              (FWState.thread_action_to_thread_action A1_
                 Observable_Events.obs_action_obs_event
                 (FWState.ObsAction
                   (Observable_Events.NewHeapElem (a, Type.Class_type c1)))
                 (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Val (Value.Addr a), (h, xw)))))
    | Expr.NewArray (c1, c2) =>
      Predicate.sup_pred
        (case c2 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val Value.Unit => Predicate.bot_pred
          | Expr.Val Value.Null => Predicate.bot_pred
          | Expr.Val (Value.Bool _) => Predicate.bot_pred
          | Expr.Val (Value.Intg i) =>
            Predicate.sup_pred
              (Predicate.bind
                (Predicate.if_pred
                  (Word.word_sle
                    (Type_Length.len_bit0
                      (Type_Length.len_bit0
                        (Type_Length.len_bit0
                          (Type_Length.len_bit0
                            (Type_Length.len_bit0 Type_Length.len_num1)))))
                    (Word.zero_word
                      (Type_Length.len0_bit0
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0
                            (Type_Length.len0_bit0
                              (Type_Length.len0_bit0 Type_Length.len0_num1))))))
                    i))
                (fn () =>
                  Predicate.bind
                    (TypeRel.eq_i_i
                      (Set.equal_set (Product_Type.equal_prod C_ A1_))
                      (xp xv
                        (Type.Array_type
                          (c1, Arith.nat
                                 (Word.sint
                                   (Type_Length.len_bit0
                                     (Type_Length.len_bit0
                                       (Type_Length.len_bit0
 (Type_Length.len_bit0 (Type_Length.len_bit0 Type_Length.len_num1)))))
                                   i))))
                      Set.bot_set)
                    (fn () =>
                      Predicate.single
                        ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                          (Expr.Throw
                             (Expr.Val
                               (Value.Addr
                                 (StartConfig.addr_of_sys_xcpt C_ (A1_, A2_) xo
                                   xp "java/lang/OutOfMemoryError"))),
                            (xv, xw))))))
              (Predicate.sup_pred
                (Predicate.bind
                  (Predicate.if_pred
                    (Word.word_sless
                      (Type_Length.len_bit0
                        (Type_Length.len_bit0
                          (Type_Length.len_bit0
                            (Type_Length.len_bit0
                              (Type_Length.len_bit0 Type_Length.len_num1)))))
                      i (Word.zero_word
                          (Type_Length.len0_bit0
                            (Type_Length.len0_bit0
                              (Type_Length.len0_bit0
                                (Type_Length.len0_bit0
                                  (Type_Length.len0_bit0
                                    Type_Length.len0_num1))))))))
                  (fn () =>
                    Predicate.single
                      ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                        (Expr.Throw
                           (Expr.Val
                             (Value.Addr
                               (StartConfig.addr_of_sys_xcpt C_ (A1_, A2_) xo xp
                                 "java/lang/NegativeArraySizeException"))),
                          (xv, xw)))))
                (Predicate.bind
                  (Predicate.if_pred
                    (Word.word_sle
                      (Type_Length.len_bit0
                        (Type_Length.len_bit0
                          (Type_Length.len_bit0
                            (Type_Length.len_bit0
                              (Type_Length.len_bit0 Type_Length.len_num1)))))
                      (Word.zero_word
                        (Type_Length.len0_bit0
                          (Type_Length.len0_bit0
                            (Type_Length.len0_bit0
                              (Type_Length.len0_bit0
                                (Type_Length.len0_bit0
                                  Type_Length.len0_num1))))))
                      i))
                  (fn () =>
                    Predicate.bind
                      (Predicate.pred_of_set
                        (xp xv
                          (Type.Array_type
                            (c1, Arith.nat
                                   (Word.sint
                                     (Type_Length.len_bit0
                                       (Type_Length.len_bit0
 (Type_Length.len_bit0
   (Type_Length.len_bit0 (Type_Length.len_bit0 Type_Length.len_num1)))))
                                     i)))))
                      (fn (h, a) =>
                        Predicate.single
                          (FWState.thread_action_to_thread_action A1_
                             Observable_Events.obs_action_obs_event
                             (FWState.ObsAction
                               (Observable_Events.NewHeapElem
                                 (a, Type.Array_type
                                       (c1,
 Arith.nat
   (Word.sint
     (Type_Length.len_bit0
       (Type_Length.len_bit0
         (Type_Length.len_bit0
           (Type_Length.len_bit0 (Type_Length.len_bit0 Type_Length.len_num1)))))
     i)))))
                             (FinFun.Finfun_const [],
                               ([], ([], ([], ([], []))))),
                            (Expr.Val (Value.Addr a), (h, xw)))))))
          | Expr.Val (Value.Addr _) => Predicate.bot_pred
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw (Expr.Val (Value.Addr a)), (xv, xw)))
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.bind
          (red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) C_ xl
            xm xn xo xp xq heap_read heap_write xr xs xt c2 (xv, xw))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))),
                (Expr.NewArray (c1, e), (xf, yf)))))
    | Expr.Cast (c1, c2) =>
      Predicate.sup_pred
        (case c2 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val v =>
            Predicate.sup_pred
              (case Heap.typeof_h A2_ xq (State.hp (xv, xw)) v
                of NONE => Predicate.bot_pred
                | SOME u =>
                  Predicate.bind
                    (Predicate.not_pred (TypeRel.widen_i_i_i xs u c1))
                    (fn () =>
                      Predicate.single
                        ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                          (Expr.Throw
                             (Expr.Val
                               (Value.Addr
                                 (StartConfig.addr_of_sys_xcpt C_ (A1_, A2_) xo
                                   xp "java/lang/ClassCastException"))),
                            (xv, xw)))))
              (case Heap.typeof_h A2_ xq (State.hp (xv, xw)) v
                of NONE => Predicate.bot_pred
                | SOME u =>
                  Predicate.bind (TypeRel.widen_i_i_i xs u c1)
                    (fn () =>
                      Predicate.single
                        ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                          (Expr.Val v, (xv, xw)))))
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw (Expr.Val (Value.Addr a)), (xv, xw)))
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.bind
          (red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) C_ xl
            xm xn xo xp xq heap_read heap_write xr xs xt c2 (xv, xw))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))),
                (Expr.Cast (c1, e), (xf, yf)))))
    | Expr.InstanceOf (c1, c2) =>
      Predicate.sup_pred
        (case c1 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val v =>
            (case Heap.typeof_h A2_ xq (State.hp (xv, xw)) v
              of NONE => Predicate.bot_pred
              | SOME u =>
                Predicate.single
                  ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                    (Expr.Val
                       (Value.Bool
                         (not (Value.equal_vala A1_ v Value.Null) andalso
                           TypeRel.widen xs u c2)),
                      (xv, xw))))
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw (Expr.Val (Value.Addr a)), (xv, xw)))
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.bind
          (red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) C_ xl
            xm xn xo xp xq heap_read heap_write xr xs xt c1 (xv, xw))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))),
                (Expr.InstanceOf (e, c2), (xf, yf)))))
    | Expr.Val _ => Predicate.bot_pred
    | Expr.BinOp (c1, c2, c3) =>
      Predicate.sup_pred
        (case c3 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val _ => Predicate.bot_pred
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            (case c1 of Expr.New _ => Predicate.bot_pred
              | Expr.NewArray (_, _) => Predicate.bot_pred
              | Expr.Cast (_, _) => Predicate.bot_pred
              | Expr.InstanceOf (_, _) => Predicate.bot_pred
              | Expr.Val _ =>
                Predicate.single
                  ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                    (Expr.Throw (Expr.Val (Value.Addr a)), (xv, xw)))
              | Expr.BinOp (_, _, _) => Predicate.bot_pred
              | Expr.Var _ => Predicate.bot_pred
              | Expr.LAss (_, _) => Predicate.bot_pred
              | Expr.AAcc (_, _) => Predicate.bot_pred
              | Expr.AAss (_, _, _) => Predicate.bot_pred
              | Expr.ALen _ => Predicate.bot_pred
              | Expr.FAcc (_, _, _) => Predicate.bot_pred
              | Expr.FAss (_, _, _, _) => Predicate.bot_pred
              | Expr.Call (_, _, _) => Predicate.bot_pred
              | Expr.Block (_, _, _, _) => Predicate.bot_pred
              | Expr.Synchronized (_, _, _) => Predicate.bot_pred
              | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
              | Expr.Seq (_, _) => Predicate.bot_pred
              | Expr.Cond (_, _, _) => Predicate.bot_pred
              | Expr.While (_, _) => Predicate.bot_pred
              | Expr.Throw _ => Predicate.bot_pred
              | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.sup_pred
          (case c1 of Expr.New _ => Predicate.bot_pred
            | Expr.NewArray (_, _) => Predicate.bot_pred
            | Expr.Cast (_, _) => Predicate.bot_pred
            | Expr.InstanceOf (_, _) => Predicate.bot_pred
            | Expr.Val _ => Predicate.bot_pred
            | Expr.BinOp (_, _, _) => Predicate.bot_pred
            | Expr.Var _ => Predicate.bot_pred
            | Expr.LAss (_, _) => Predicate.bot_pred
            | Expr.AAcc (_, _) => Predicate.bot_pred
            | Expr.AAss (_, _, _) => Predicate.bot_pred
            | Expr.ALen _ => Predicate.bot_pred
            | Expr.FAcc (_, _, _) => Predicate.bot_pred
            | Expr.FAss (_, _, _, _) => Predicate.bot_pred
            | Expr.Call (_, _, _) => Predicate.bot_pred
            | Expr.Block (_, _, _, _) => Predicate.bot_pred
            | Expr.Synchronized (_, _, _) => Predicate.bot_pred
            | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
            | Expr.Seq (_, _) => Predicate.bot_pred
            | Expr.Cond (_, _, _) => Predicate.bot_pred
            | Expr.While (_, _) => Predicate.bot_pred
            | Expr.Throw (Expr.New _) => Predicate.bot_pred
            | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
            | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
            | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Val (Value.Addr a)) =>
              Predicate.single
                ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                  (Expr.Throw (Expr.Val (Value.Addr a)), (xv, xw)))
            | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Var _) => Predicate.bot_pred
            | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
            | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
            | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
            | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
          (Predicate.sup_pred
            (case c3 of Expr.New _ => Predicate.bot_pred
              | Expr.NewArray (_, _) => Predicate.bot_pred
              | Expr.Cast (_, _) => Predicate.bot_pred
              | Expr.InstanceOf (_, _) => Predicate.bot_pred
              | Expr.Val v =>
                (case c1 of Expr.New _ => Predicate.bot_pred
                  | Expr.NewArray (_, _) => Predicate.bot_pred
                  | Expr.Cast (_, _) => Predicate.bot_pred
                  | Expr.InstanceOf (_, _) => Predicate.bot_pred
                  | Expr.Val va =>
                    Predicate.sup_pred
                      (case BinOp.binop C_ (A1_, A2_) xo xp c2 va v
                        of NONE => Predicate.bot_pred
                        | SOME (Sum_Type.Inl _) => Predicate.bot_pred
                        | SOME (Sum_Type.Inr a) =>
                          Predicate.single
                            ((FinFun.Finfun_const [],
                               ([], ([], ([], ([], []))))),
                              (Expr.Throw (Expr.Val (Value.Addr a)), (xv, xw))))
                      (case BinOp.binop C_ (A1_, A2_) xo xp c2 va v
                        of NONE => Predicate.bot_pred
                        | SOME (Sum_Type.Inl vb) =>
                          Predicate.single
                            ((FinFun.Finfun_const [],
                               ([], ([], ([], ([], []))))),
                              (Expr.Val vb, (xv, xw)))
                        | SOME (Sum_Type.Inr _) => Predicate.bot_pred)
                  | Expr.BinOp (_, _, _) => Predicate.bot_pred
                  | Expr.Var _ => Predicate.bot_pred
                  | Expr.LAss (_, _) => Predicate.bot_pred
                  | Expr.AAcc (_, _) => Predicate.bot_pred
                  | Expr.AAss (_, _, _) => Predicate.bot_pred
                  | Expr.ALen _ => Predicate.bot_pred
                  | Expr.FAcc (_, _, _) => Predicate.bot_pred
                  | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                  | Expr.Call (_, _, _) => Predicate.bot_pred
                  | Expr.Block (_, _, _, _) => Predicate.bot_pred
                  | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                  | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                  | Expr.Seq (_, _) => Predicate.bot_pred
                  | Expr.Cond (_, _, _) => Predicate.bot_pred
                  | Expr.While (_, _) => Predicate.bot_pred
                  | Expr.Throw _ => Predicate.bot_pred
                  | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
              | Expr.BinOp (_, _, _) => Predicate.bot_pred
              | Expr.Var _ => Predicate.bot_pred
              | Expr.LAss (_, _) => Predicate.bot_pred
              | Expr.AAcc (_, _) => Predicate.bot_pred
              | Expr.AAss (_, _, _) => Predicate.bot_pred
              | Expr.ALen _ => Predicate.bot_pred
              | Expr.FAcc (_, _, _) => Predicate.bot_pred
              | Expr.FAss (_, _, _, _) => Predicate.bot_pred
              | Expr.Call (_, _, _) => Predicate.bot_pred
              | Expr.Block (_, _, _, _) => Predicate.bot_pred
              | Expr.Synchronized (_, _, _) => Predicate.bot_pred
              | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
              | Expr.Seq (_, _) => Predicate.bot_pred
              | Expr.Cond (_, _, _) => Predicate.bot_pred
              | Expr.While (_, _) => Predicate.bot_pred
              | Expr.Throw _ => Predicate.bot_pred
              | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
            (Predicate.sup_pred
              (case c1 of Expr.New _ => Predicate.bot_pred
                | Expr.NewArray (_, _) => Predicate.bot_pred
                | Expr.Cast (_, _) => Predicate.bot_pred
                | Expr.InstanceOf (_, _) => Predicate.bot_pred
                | Expr.Val v =>
                  Predicate.bind
                    (red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o
                      (A1_, A2_) C_ xl xm xn xo xp xq heap_read heap_write xr xs
                      xt c3 (xv, xw))
                    (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
                      Predicate.single
                        ((xa, (xb, (xc, (xd, (xe, ye))))),
                          (Expr.BinOp (Expr.Val v, c2, e), (xf, yf))))
                | Expr.BinOp (_, _, _) => Predicate.bot_pred
                | Expr.Var _ => Predicate.bot_pred
                | Expr.LAss (_, _) => Predicate.bot_pred
                | Expr.AAcc (_, _) => Predicate.bot_pred
                | Expr.AAss (_, _, _) => Predicate.bot_pred
                | Expr.ALen _ => Predicate.bot_pred
                | Expr.FAcc (_, _, _) => Predicate.bot_pred
                | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                | Expr.Call (_, _, _) => Predicate.bot_pred
                | Expr.Block (_, _, _, _) => Predicate.bot_pred
                | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                | Expr.Seq (_, _) => Predicate.bot_pred
                | Expr.Cond (_, _, _) => Predicate.bot_pred
                | Expr.While (_, _) => Predicate.bot_pred
                | Expr.Throw _ => Predicate.bot_pred
                | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
              (Predicate.bind
                (red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_)
                  C_ xl xm xn xo xp xq heap_read heap_write xr xs xt c1
                  (xv, xw))
                (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
                  Predicate.single
                    ((xa, (xb, (xc, (xd, (xe, ye))))),
                      (Expr.BinOp (e, c2, c3), (xf, yf))))))))
    | Expr.Var c1 =>
      (case State.lcl (xv, xw) c1 of NONE => Predicate.bot_pred
        | SOME v =>
          Predicate.single
            ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
              (Expr.Val v, (xv, xw))))
    | Expr.LAss (c1, c2) =>
      Predicate.sup_pred
        (case c2 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val v =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Val Value.Unit,
                  (xv, Fun.fun_upd Stringa.equal_literal xw c1 (SOME v))))
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw (Expr.Val (Value.Addr a)), (xv, xw)))
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.bind
          (red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) C_ xl
            xm xn xo xp xq heap_read heap_write xr xs xt c2 (xv, xw))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))),
                (Expr.LAss (c1, e), (xf, yf)))))
    | Expr.AAcc (c1, c2) =>
      Predicate.sup_pred
        (case c2 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val _ => Predicate.bot_pred
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            (case c1 of Expr.New _ => Predicate.bot_pred
              | Expr.NewArray (_, _) => Predicate.bot_pred
              | Expr.Cast (_, _) => Predicate.bot_pred
              | Expr.InstanceOf (_, _) => Predicate.bot_pred
              | Expr.Val _ =>
                Predicate.single
                  ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                    (Expr.Throw (Expr.Val (Value.Addr a)), (xv, xw)))
              | Expr.BinOp (_, _, _) => Predicate.bot_pred
              | Expr.Var _ => Predicate.bot_pred
              | Expr.LAss (_, _) => Predicate.bot_pred
              | Expr.AAcc (_, _) => Predicate.bot_pred
              | Expr.AAss (_, _, _) => Predicate.bot_pred
              | Expr.ALen _ => Predicate.bot_pred
              | Expr.FAcc (_, _, _) => Predicate.bot_pred
              | Expr.FAss (_, _, _, _) => Predicate.bot_pred
              | Expr.Call (_, _, _) => Predicate.bot_pred
              | Expr.Block (_, _, _, _) => Predicate.bot_pred
              | Expr.Synchronized (_, _, _) => Predicate.bot_pred
              | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
              | Expr.Seq (_, _) => Predicate.bot_pred
              | Expr.Cond (_, _, _) => Predicate.bot_pred
              | Expr.While (_, _) => Predicate.bot_pred
              | Expr.Throw _ => Predicate.bot_pred
              | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.sup_pred
          (case c1 of Expr.New _ => Predicate.bot_pred
            | Expr.NewArray (_, _) => Predicate.bot_pred
            | Expr.Cast (_, _) => Predicate.bot_pred
            | Expr.InstanceOf (_, _) => Predicate.bot_pred
            | Expr.Val _ => Predicate.bot_pred
            | Expr.BinOp (_, _, _) => Predicate.bot_pred
            | Expr.Var _ => Predicate.bot_pred
            | Expr.LAss (_, _) => Predicate.bot_pred
            | Expr.AAcc (_, _) => Predicate.bot_pred
            | Expr.AAss (_, _, _) => Predicate.bot_pred
            | Expr.ALen _ => Predicate.bot_pred
            | Expr.FAcc (_, _, _) => Predicate.bot_pred
            | Expr.FAss (_, _, _, _) => Predicate.bot_pred
            | Expr.Call (_, _, _) => Predicate.bot_pred
            | Expr.Block (_, _, _, _) => Predicate.bot_pred
            | Expr.Synchronized (_, _, _) => Predicate.bot_pred
            | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
            | Expr.Seq (_, _) => Predicate.bot_pred
            | Expr.Cond (_, _, _) => Predicate.bot_pred
            | Expr.While (_, _) => Predicate.bot_pred
            | Expr.Throw (Expr.New _) => Predicate.bot_pred
            | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
            | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
            | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Val (Value.Addr a)) =>
              Predicate.single
                ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                  (Expr.Throw (Expr.Val (Value.Addr a)), (xv, xw)))
            | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Var _) => Predicate.bot_pred
            | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
            | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
            | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
            | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
          (Predicate.sup_pred
            (case c2 of Expr.New _ => Predicate.bot_pred
              | Expr.NewArray (_, _) => Predicate.bot_pred
              | Expr.Cast (_, _) => Predicate.bot_pred
              | Expr.InstanceOf (_, _) => Predicate.bot_pred
              | Expr.Val v =>
                Predicate.sup_pred
                  (case v of Value.Unit => Predicate.bot_pred
                    | Value.Null => Predicate.bot_pred
                    | Value.Bool _ => Predicate.bot_pred
                    | Value.Intg i =>
                      (case c1 of Expr.New _ => Predicate.bot_pred
                        | Expr.NewArray (_, _) => Predicate.bot_pred
                        | Expr.Cast (_, _) => Predicate.bot_pred
                        | Expr.InstanceOf (_, _) => Predicate.bot_pred
                        | Expr.Val Value.Unit => Predicate.bot_pred
                        | Expr.Val Value.Null => Predicate.bot_pred
                        | Expr.Val (Value.Bool _) => Predicate.bot_pred
                        | Expr.Val (Value.Intg _) => Predicate.bot_pred
                        | Expr.Val (Value.Addr a) =>
                          Predicate.bind
                            (Predicate.if_pred
                              (Word.word_sle
                                (Type_Length.len_bit0
                                  (Type_Length.len_bit0
                                    (Type_Length.len_bit0
                                      (Type_Length.len_bit0
(Type_Length.len_bit0 Type_Length.len_num1)))))
                                (Word.zero_word
                                  (Type_Length.len0_bit0
                                    (Type_Length.len0_bit0
                                      (Type_Length.len0_bit0
(Type_Length.len0_bit0 (Type_Length.len0_bit0 Type_Length.len0_num1))))))
                                i))
                            (fn () =>
                              (case xq xv a of NONE => Predicate.bot_pred
                                | SOME (Type.Class_type _) => Predicate.bot_pred
                                | SOME (Type.Array_type (_, n)) =>
                                  Predicate.bind
                                    (Predicate.if_pred
                                      (Arith.less_int
(Word.sint
  (Type_Length.len_bit0
    (Type_Length.len_bit0
      (Type_Length.len_bit0
        (Type_Length.len_bit0 (Type_Length.len_bit0 Type_Length.len_num1)))))
  i)
(Arith.int_of_nat n)))
                                    (fn () =>
                                      Predicate.bind
(heap_read xv a
  (Heap.ACell
    (Arith.nat
      (Word.sint
        (Type_Length.len_bit0
          (Type_Length.len_bit0
            (Type_Length.len_bit0
              (Type_Length.len_bit0
                (Type_Length.len_bit0 Type_Length.len_num1)))))
        i))))
(fn x =>
  Predicate.single
    (FWState.thread_action_to_thread_action A1_
       Observable_Events.obs_action_obs_event
       (FWState.ObsAction
         (Observable_Events.ReadMem
           (a, Heap.ACell
                 (Arith.nat
                   (Word.sint
                     (Type_Length.len_bit0
                       (Type_Length.len_bit0
                         (Type_Length.len_bit0
                           (Type_Length.len_bit0
                             (Type_Length.len_bit0 Type_Length.len_num1)))))
                     i)),
             x)))
       (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
      (Expr.Val x, (xv, xw)))))))
                        | Expr.BinOp (_, _, _) => Predicate.bot_pred
                        | Expr.Var _ => Predicate.bot_pred
                        | Expr.LAss (_, _) => Predicate.bot_pred
                        | Expr.AAcc (_, _) => Predicate.bot_pred
                        | Expr.AAss (_, _, _) => Predicate.bot_pred
                        | Expr.ALen _ => Predicate.bot_pred
                        | Expr.FAcc (_, _, _) => Predicate.bot_pred
                        | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                        | Expr.Call (_, _, _) => Predicate.bot_pred
                        | Expr.Block (_, _, _, _) => Predicate.bot_pred
                        | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                        | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                        | Expr.Seq (_, _) => Predicate.bot_pred
                        | Expr.Cond (_, _, _) => Predicate.bot_pred
                        | Expr.While (_, _) => Predicate.bot_pred
                        | Expr.Throw _ => Predicate.bot_pred
                        | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
                    | Value.Addr _ => Predicate.bot_pred)
                  (Predicate.sup_pred
                    (case v of Value.Unit => Predicate.bot_pred
                      | Value.Null => Predicate.bot_pred
                      | Value.Bool _ => Predicate.bot_pred
                      | Value.Intg i =>
                        (case c1 of Expr.New _ => Predicate.bot_pred
                          | Expr.NewArray (_, _) => Predicate.bot_pred
                          | Expr.Cast (_, _) => Predicate.bot_pred
                          | Expr.InstanceOf (_, _) => Predicate.bot_pred
                          | Expr.Val Value.Unit => Predicate.bot_pred
                          | Expr.Val Value.Null => Predicate.bot_pred
                          | Expr.Val (Value.Bool _) => Predicate.bot_pred
                          | Expr.Val (Value.Intg _) => Predicate.bot_pred
                          | Expr.Val (Value.Addr a) =>
                            (case xq (State.hp (xv, xw)) a
                              of NONE => Predicate.bot_pred
                              | SOME (Type.Class_type _) => Predicate.bot_pred
                              | SOME (Type.Array_type (_, n)) =>
                                Predicate.bind
                                  (Predicate.if_pred
                                    (Word.word_sless
                                       (Type_Length.len_bit0
 (Type_Length.len_bit0
   (Type_Length.len_bit0
     (Type_Length.len_bit0 (Type_Length.len_bit0 Type_Length.len_num1)))))
                                       i (Word.zero_word
   (Type_Length.len0_bit0
     (Type_Length.len0_bit0
       (Type_Length.len0_bit0
         (Type_Length.len0_bit0
           (Type_Length.len0_bit0 Type_Length.len0_num1)))))) orelse
                                      Arith.less_eq_int (Arith.int_of_nat n)
(Word.sint
  (Type_Length.len_bit0
    (Type_Length.len_bit0
      (Type_Length.len_bit0
        (Type_Length.len_bit0 (Type_Length.len_bit0 Type_Length.len_num1)))))
  i)))
                                  (fn () =>
                                    Predicate.single
                                      ((FinFun.Finfun_const [],
 ([], ([], ([], ([], []))))),
(Expr.Throw
   (Expr.Val
     (Value.Addr
       (StartConfig.addr_of_sys_xcpt C_ (A1_, A2_) xo xp
         "java/lang/ArrayIndexOutOfBoundsException"))),
  (xv, xw)))))
                          | Expr.BinOp (_, _, _) => Predicate.bot_pred
                          | Expr.Var _ => Predicate.bot_pred
                          | Expr.LAss (_, _) => Predicate.bot_pred
                          | Expr.AAcc (_, _) => Predicate.bot_pred
                          | Expr.AAss (_, _, _) => Predicate.bot_pred
                          | Expr.ALen _ => Predicate.bot_pred
                          | Expr.FAcc (_, _, _) => Predicate.bot_pred
                          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                          | Expr.Call (_, _, _) => Predicate.bot_pred
                          | Expr.Block (_, _, _, _) => Predicate.bot_pred
                          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                          | Expr.Seq (_, _) => Predicate.bot_pred
                          | Expr.Cond (_, _, _) => Predicate.bot_pred
                          | Expr.While (_, _) => Predicate.bot_pred
                          | Expr.Throw _ => Predicate.bot_pred
                          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
                      | Value.Addr _ => Predicate.bot_pred)
                    (case c1 of Expr.New _ => Predicate.bot_pred
                      | Expr.NewArray (_, _) => Predicate.bot_pred
                      | Expr.Cast (_, _) => Predicate.bot_pred
                      | Expr.InstanceOf (_, _) => Predicate.bot_pred
                      | Expr.Val Value.Unit => Predicate.bot_pred
                      | Expr.Val Value.Null =>
                        Predicate.single
                          ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                            (Expr.Throw
                               (Expr.Val
                                 (Value.Addr
                                   (StartConfig.addr_of_sys_xcpt C_ (A1_, A2_)
                                     xo xp "java/lang/NullPointerException"))),
                              (xv, xw)))
                      | Expr.Val (Value.Bool _) => Predicate.bot_pred
                      | Expr.Val (Value.Intg _) => Predicate.bot_pred
                      | Expr.Val (Value.Addr _) => Predicate.bot_pred
                      | Expr.BinOp (_, _, _) => Predicate.bot_pred
                      | Expr.Var _ => Predicate.bot_pred
                      | Expr.LAss (_, _) => Predicate.bot_pred
                      | Expr.AAcc (_, _) => Predicate.bot_pred
                      | Expr.AAss (_, _, _) => Predicate.bot_pred
                      | Expr.ALen _ => Predicate.bot_pred
                      | Expr.FAcc (_, _, _) => Predicate.bot_pred
                      | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                      | Expr.Call (_, _, _) => Predicate.bot_pred
                      | Expr.Block (_, _, _, _) => Predicate.bot_pred
                      | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                      | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                      | Expr.Seq (_, _) => Predicate.bot_pred
                      | Expr.Cond (_, _, _) => Predicate.bot_pred
                      | Expr.While (_, _) => Predicate.bot_pred
                      | Expr.Throw _ => Predicate.bot_pred
                      | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred))
              | Expr.BinOp (_, _, _) => Predicate.bot_pred
              | Expr.Var _ => Predicate.bot_pred
              | Expr.LAss (_, _) => Predicate.bot_pred
              | Expr.AAcc (_, _) => Predicate.bot_pred
              | Expr.AAss (_, _, _) => Predicate.bot_pred
              | Expr.ALen _ => Predicate.bot_pred
              | Expr.FAcc (_, _, _) => Predicate.bot_pred
              | Expr.FAss (_, _, _, _) => Predicate.bot_pred
              | Expr.Call (_, _, _) => Predicate.bot_pred
              | Expr.Block (_, _, _, _) => Predicate.bot_pred
              | Expr.Synchronized (_, _, _) => Predicate.bot_pred
              | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
              | Expr.Seq (_, _) => Predicate.bot_pred
              | Expr.Cond (_, _, _) => Predicate.bot_pred
              | Expr.While (_, _) => Predicate.bot_pred
              | Expr.Throw _ => Predicate.bot_pred
              | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
            (Predicate.sup_pred
              (case c1 of Expr.New _ => Predicate.bot_pred
                | Expr.NewArray (_, _) => Predicate.bot_pred
                | Expr.Cast (_, _) => Predicate.bot_pred
                | Expr.InstanceOf (_, _) => Predicate.bot_pred
                | Expr.Val ab =>
                  Predicate.bind
                    (red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o
                      (A1_, A2_) C_ xl xm xn xo xp xq heap_read heap_write xr xs
                      xt c2 (xv, xw))
                    (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (i, (xf, yf))) =>
                      Predicate.single
                        ((xa, (xb, (xc, (xd, (xe, ye))))),
                          (Expr.AAcc (Expr.Val ab, i), (xf, yf))))
                | Expr.BinOp (_, _, _) => Predicate.bot_pred
                | Expr.Var _ => Predicate.bot_pred
                | Expr.LAss (_, _) => Predicate.bot_pred
                | Expr.AAcc (_, _) => Predicate.bot_pred
                | Expr.AAss (_, _, _) => Predicate.bot_pred
                | Expr.ALen _ => Predicate.bot_pred
                | Expr.FAcc (_, _, _) => Predicate.bot_pred
                | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                | Expr.Call (_, _, _) => Predicate.bot_pred
                | Expr.Block (_, _, _, _) => Predicate.bot_pred
                | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                | Expr.Seq (_, _) => Predicate.bot_pred
                | Expr.Cond (_, _, _) => Predicate.bot_pred
                | Expr.While (_, _) => Predicate.bot_pred
                | Expr.Throw _ => Predicate.bot_pred
                | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
              (Predicate.bind
                (red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_)
                  C_ xl xm xn xo xp xq heap_read heap_write xr xs xt c1
                  (xv, xw))
                (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (a, (xf, yf))) =>
                  Predicate.single
                    ((xa, (xb, (xc, (xd, (xe, ye))))),
                      (Expr.AAcc (a, c2), (xf, yf))))))))
    | Expr.AAss (c1, c2, c3) =>
      Predicate.sup_pred
        (case c3 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val _ => Predicate.bot_pred
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            (case c2 of Expr.New _ => Predicate.bot_pred
              | Expr.NewArray (_, _) => Predicate.bot_pred
              | Expr.Cast (_, _) => Predicate.bot_pred
              | Expr.InstanceOf (_, _) => Predicate.bot_pred
              | Expr.Val _ =>
                (case c1 of Expr.New _ => Predicate.bot_pred
                  | Expr.NewArray (_, _) => Predicate.bot_pred
                  | Expr.Cast (_, _) => Predicate.bot_pred
                  | Expr.InstanceOf (_, _) => Predicate.bot_pred
                  | Expr.Val _ =>
                    Predicate.single
                      ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                        (Expr.Throw (Expr.Val (Value.Addr a)), (xv, xw)))
                  | Expr.BinOp (_, _, _) => Predicate.bot_pred
                  | Expr.Var _ => Predicate.bot_pred
                  | Expr.LAss (_, _) => Predicate.bot_pred
                  | Expr.AAcc (_, _) => Predicate.bot_pred
                  | Expr.AAss (_, _, _) => Predicate.bot_pred
                  | Expr.ALen _ => Predicate.bot_pred
                  | Expr.FAcc (_, _, _) => Predicate.bot_pred
                  | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                  | Expr.Call (_, _, _) => Predicate.bot_pred
                  | Expr.Block (_, _, _, _) => Predicate.bot_pred
                  | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                  | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                  | Expr.Seq (_, _) => Predicate.bot_pred
                  | Expr.Cond (_, _, _) => Predicate.bot_pred
                  | Expr.While (_, _) => Predicate.bot_pred
                  | Expr.Throw _ => Predicate.bot_pred
                  | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
              | Expr.BinOp (_, _, _) => Predicate.bot_pred
              | Expr.Var _ => Predicate.bot_pred
              | Expr.LAss (_, _) => Predicate.bot_pred
              | Expr.AAcc (_, _) => Predicate.bot_pred
              | Expr.AAss (_, _, _) => Predicate.bot_pred
              | Expr.ALen _ => Predicate.bot_pred
              | Expr.FAcc (_, _, _) => Predicate.bot_pred
              | Expr.FAss (_, _, _, _) => Predicate.bot_pred
              | Expr.Call (_, _, _) => Predicate.bot_pred
              | Expr.Block (_, _, _, _) => Predicate.bot_pred
              | Expr.Synchronized (_, _, _) => Predicate.bot_pred
              | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
              | Expr.Seq (_, _) => Predicate.bot_pred
              | Expr.Cond (_, _, _) => Predicate.bot_pred
              | Expr.While (_, _) => Predicate.bot_pred
              | Expr.Throw _ => Predicate.bot_pred
              | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.sup_pred
          (case c2 of Expr.New _ => Predicate.bot_pred
            | Expr.NewArray (_, _) => Predicate.bot_pred
            | Expr.Cast (_, _) => Predicate.bot_pred
            | Expr.InstanceOf (_, _) => Predicate.bot_pred
            | Expr.Val _ => Predicate.bot_pred
            | Expr.BinOp (_, _, _) => Predicate.bot_pred
            | Expr.Var _ => Predicate.bot_pred
            | Expr.LAss (_, _) => Predicate.bot_pred
            | Expr.AAcc (_, _) => Predicate.bot_pred
            | Expr.AAss (_, _, _) => Predicate.bot_pred
            | Expr.ALen _ => Predicate.bot_pred
            | Expr.FAcc (_, _, _) => Predicate.bot_pred
            | Expr.FAss (_, _, _, _) => Predicate.bot_pred
            | Expr.Call (_, _, _) => Predicate.bot_pred
            | Expr.Block (_, _, _, _) => Predicate.bot_pred
            | Expr.Synchronized (_, _, _) => Predicate.bot_pred
            | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
            | Expr.Seq (_, _) => Predicate.bot_pred
            | Expr.Cond (_, _, _) => Predicate.bot_pred
            | Expr.While (_, _) => Predicate.bot_pred
            | Expr.Throw (Expr.New _) => Predicate.bot_pred
            | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
            | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
            | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Val (Value.Addr a)) =>
              (case c1 of Expr.New _ => Predicate.bot_pred
                | Expr.NewArray (_, _) => Predicate.bot_pred
                | Expr.Cast (_, _) => Predicate.bot_pred
                | Expr.InstanceOf (_, _) => Predicate.bot_pred
                | Expr.Val _ =>
                  Predicate.single
                    ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                      (Expr.Throw (Expr.Val (Value.Addr a)), (xv, xw)))
                | Expr.BinOp (_, _, _) => Predicate.bot_pred
                | Expr.Var _ => Predicate.bot_pred
                | Expr.LAss (_, _) => Predicate.bot_pred
                | Expr.AAcc (_, _) => Predicate.bot_pred
                | Expr.AAss (_, _, _) => Predicate.bot_pred
                | Expr.ALen _ => Predicate.bot_pred
                | Expr.FAcc (_, _, _) => Predicate.bot_pred
                | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                | Expr.Call (_, _, _) => Predicate.bot_pred
                | Expr.Block (_, _, _, _) => Predicate.bot_pred
                | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                | Expr.Seq (_, _) => Predicate.bot_pred
                | Expr.Cond (_, _, _) => Predicate.bot_pred
                | Expr.While (_, _) => Predicate.bot_pred
                | Expr.Throw _ => Predicate.bot_pred
                | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
            | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Var _) => Predicate.bot_pred
            | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
            | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
            | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
            | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
          (Predicate.sup_pred
            (case c1 of Expr.New _ => Predicate.bot_pred
              | Expr.NewArray (_, _) => Predicate.bot_pred
              | Expr.Cast (_, _) => Predicate.bot_pred
              | Expr.InstanceOf (_, _) => Predicate.bot_pred
              | Expr.Val _ => Predicate.bot_pred
              | Expr.BinOp (_, _, _) => Predicate.bot_pred
              | Expr.Var _ => Predicate.bot_pred
              | Expr.LAss (_, _) => Predicate.bot_pred
              | Expr.AAcc (_, _) => Predicate.bot_pred
              | Expr.AAss (_, _, _) => Predicate.bot_pred
              | Expr.ALen _ => Predicate.bot_pred
              | Expr.FAcc (_, _, _) => Predicate.bot_pred
              | Expr.FAss (_, _, _, _) => Predicate.bot_pred
              | Expr.Call (_, _, _) => Predicate.bot_pred
              | Expr.Block (_, _, _, _) => Predicate.bot_pred
              | Expr.Synchronized (_, _, _) => Predicate.bot_pred
              | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
              | Expr.Seq (_, _) => Predicate.bot_pred
              | Expr.Cond (_, _, _) => Predicate.bot_pred
              | Expr.While (_, _) => Predicate.bot_pred
              | Expr.Throw (Expr.New _) => Predicate.bot_pred
              | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
              | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
              | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
              | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
              | Expr.Throw (Expr.Val (Value.Addr a)) =>
                Predicate.single
                  ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                    (Expr.Throw (Expr.Val (Value.Addr a)), (xv, xw)))
              | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.Var _) => Predicate.bot_pred
              | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
              | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
              | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
              | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
              | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
            (Predicate.sup_pred
              (case c3 of Expr.New _ => Predicate.bot_pred
                | Expr.NewArray (_, _) => Predicate.bot_pred
                | Expr.Cast (_, _) => Predicate.bot_pred
                | Expr.InstanceOf (_, _) => Predicate.bot_pred
                | Expr.Val v =>
                  (case c2 of Expr.New _ => Predicate.bot_pred
                    | Expr.NewArray (_, _) => Predicate.bot_pred
                    | Expr.Cast (_, _) => Predicate.bot_pred
                    | Expr.InstanceOf (_, _) => Predicate.bot_pred
                    | Expr.Val va =>
                      Predicate.sup_pred
                        (case va of Value.Unit => Predicate.bot_pred
                          | Value.Null => Predicate.bot_pred
                          | Value.Bool _ => Predicate.bot_pred
                          | Value.Intg i =>
                            (case c1 of Expr.New _ => Predicate.bot_pred
                              | Expr.NewArray (_, _) => Predicate.bot_pred
                              | Expr.Cast (_, _) => Predicate.bot_pred
                              | Expr.InstanceOf (_, _) => Predicate.bot_pred
                              | Expr.Val Value.Unit => Predicate.bot_pred
                              | Expr.Val Value.Null => Predicate.bot_pred
                              | Expr.Val (Value.Bool _) => Predicate.bot_pred
                              | Expr.Val (Value.Intg _) => Predicate.bot_pred
                              | Expr.Val (Value.Addr a) =>
                                Predicate.bind
                                  (Predicate.if_pred
                                    (Word.word_sle
                                      (Type_Length.len_bit0
(Type_Length.len_bit0
  (Type_Length.len_bit0
    (Type_Length.len_bit0 (Type_Length.len_bit0 Type_Length.len_num1)))))
                                      (Word.zero_word
(Type_Length.len0_bit0
  (Type_Length.len0_bit0
    (Type_Length.len0_bit0
      (Type_Length.len0_bit0 (Type_Length.len0_bit0 Type_Length.len0_num1))))))
                                      i))
                                  (fn () =>
                                    (case xq xv a of NONE => Predicate.bot_pred
                                      | SOME (Type.Class_type _) =>
Predicate.bot_pred
                                      | SOME (Type.Array_type (t, n)) =>
Predicate.bind
  (Predicate.if_pred
    (Arith.less_int
      (Word.sint
        (Type_Length.len_bit0
          (Type_Length.len_bit0
            (Type_Length.len_bit0
              (Type_Length.len_bit0
                (Type_Length.len_bit0 Type_Length.len_num1)))))
        i)
      (Arith.int_of_nat n)))
  (fn () =>
    (case Heap.typeof_h A2_ xq xv v of NONE => Predicate.bot_pred
      | SOME u =>
        Predicate.bind (TypeRel.widen_i_i_i xs u t)
          (fn () =>
            Predicate.bind
              (heap_write xv a
                 (Heap.ACell
                   (Arith.nat
                     (Word.sint
                       (Type_Length.len_bit0
                         (Type_Length.len_bit0
                           (Type_Length.len_bit0
                             (Type_Length.len_bit0
                               (Type_Length.len_bit0 Type_Length.len_num1)))))
                       i)))
                v)
              (fn x =>
                Predicate.single
                  (FWState.thread_action_to_thread_action A1_
                     Observable_Events.obs_action_obs_event
                     (FWState.ObsAction
                       (Observable_Events.WriteMem
                         (a, Heap.ACell
                               (Arith.nat
                                 (Word.sint
                                   (Type_Length.len_bit0
                                     (Type_Length.len_bit0
                                       (Type_Length.len_bit0
 (Type_Length.len_bit0 (Type_Length.len_bit0 Type_Length.len_num1)))))
                                   i)),
                           v)))
                     (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                    (Expr.Val Value.Unit, (x, xw)))))))))
                              | Expr.BinOp (_, _, _) => Predicate.bot_pred
                              | Expr.Var _ => Predicate.bot_pred
                              | Expr.LAss (_, _) => Predicate.bot_pred
                              | Expr.AAcc (_, _) => Predicate.bot_pred
                              | Expr.AAss (_, _, _) => Predicate.bot_pred
                              | Expr.ALen _ => Predicate.bot_pred
                              | Expr.FAcc (_, _, _) => Predicate.bot_pred
                              | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                              | Expr.Call (_, _, _) => Predicate.bot_pred
                              | Expr.Block (_, _, _, _) => Predicate.bot_pred
                              | Expr.Synchronized (_, _, _) =>
                                Predicate.bot_pred
                              | Expr.InSynchronized (_, _, _) =>
                                Predicate.bot_pred
                              | Expr.Seq (_, _) => Predicate.bot_pred
                              | Expr.Cond (_, _, _) => Predicate.bot_pred
                              | Expr.While (_, _) => Predicate.bot_pred
                              | Expr.Throw _ => Predicate.bot_pred
                              | Expr.TryCatch (_, _, _, _) =>
                                Predicate.bot_pred)
                          | Value.Addr _ => Predicate.bot_pred)
                        (Predicate.sup_pred
                          (case va of Value.Unit => Predicate.bot_pred
                            | Value.Null => Predicate.bot_pred
                            | Value.Bool _ => Predicate.bot_pred
                            | Value.Intg i =>
                              (case c1 of Expr.New _ => Predicate.bot_pred
                                | Expr.NewArray (_, _) => Predicate.bot_pred
                                | Expr.Cast (_, _) => Predicate.bot_pred
                                | Expr.InstanceOf (_, _) => Predicate.bot_pred
                                | Expr.Val Value.Unit => Predicate.bot_pred
                                | Expr.Val Value.Null => Predicate.bot_pred
                                | Expr.Val (Value.Bool _) => Predicate.bot_pred
                                | Expr.Val (Value.Intg _) => Predicate.bot_pred
                                | Expr.Val (Value.Addr a) =>
                                  Predicate.bind
                                    (Predicate.if_pred
                                      (Word.word_sle
(Type_Length.len_bit0
  (Type_Length.len_bit0
    (Type_Length.len_bit0
      (Type_Length.len_bit0 (Type_Length.len_bit0 Type_Length.len_num1)))))
(Word.zero_word
  (Type_Length.len0_bit0
    (Type_Length.len0_bit0
      (Type_Length.len0_bit0
        (Type_Length.len0_bit0
          (Type_Length.len0_bit0 Type_Length.len0_num1))))))
i))
                                    (fn () =>
                                      (case xq (State.hp (xv, xw)) a
of NONE => Predicate.bot_pred | SOME (Type.Class_type _) => Predicate.bot_pred
| SOME (Type.Array_type (t, n)) =>
  Predicate.bind
    (Predicate.if_pred
      (Arith.less_int
        (Word.sint
          (Type_Length.len_bit0
            (Type_Length.len_bit0
              (Type_Length.len_bit0
                (Type_Length.len_bit0
                  (Type_Length.len_bit0 Type_Length.len_num1)))))
          i)
        (Arith.int_of_nat n)))
    (fn () =>
      (case Heap.typeof_h A2_ xq (State.hp (xv, xw)) v
        of NONE => Predicate.bot_pred
        | SOME u =>
          Predicate.bind (Predicate.not_pred (TypeRel.widen_i_i_i xs u t))
            (fn () =>
              Predicate.single
                ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                  (Expr.Throw
                     (Expr.Val
                       (Value.Addr
                         (StartConfig.addr_of_sys_xcpt C_ (A1_, A2_) xo xp
                           "java/lang/ArrayStoreException"))),
                    (xv, xw))))))))
                                | Expr.BinOp (_, _, _) => Predicate.bot_pred
                                | Expr.Var _ => Predicate.bot_pred
                                | Expr.LAss (_, _) => Predicate.bot_pred
                                | Expr.AAcc (_, _) => Predicate.bot_pred
                                | Expr.AAss (_, _, _) => Predicate.bot_pred
                                | Expr.ALen _ => Predicate.bot_pred
                                | Expr.FAcc (_, _, _) => Predicate.bot_pred
                                | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                                | Expr.Call (_, _, _) => Predicate.bot_pred
                                | Expr.Block (_, _, _, _) => Predicate.bot_pred
                                | Expr.Synchronized (_, _, _) =>
                                  Predicate.bot_pred
                                | Expr.InSynchronized (_, _, _) =>
                                  Predicate.bot_pred
                                | Expr.Seq (_, _) => Predicate.bot_pred
                                | Expr.Cond (_, _, _) => Predicate.bot_pred
                                | Expr.While (_, _) => Predicate.bot_pred
                                | Expr.Throw _ => Predicate.bot_pred
                                | Expr.TryCatch (_, _, _, _) =>
                                  Predicate.bot_pred)
                            | Value.Addr _ => Predicate.bot_pred)
                          (Predicate.sup_pred
                            (case va of Value.Unit => Predicate.bot_pred
                              | Value.Null => Predicate.bot_pred
                              | Value.Bool _ => Predicate.bot_pred
                              | Value.Intg i =>
                                (case c1 of Expr.New _ => Predicate.bot_pred
                                  | Expr.NewArray (_, _) => Predicate.bot_pred
                                  | Expr.Cast (_, _) => Predicate.bot_pred
                                  | Expr.InstanceOf (_, _) => Predicate.bot_pred
                                  | Expr.Val Value.Unit => Predicate.bot_pred
                                  | Expr.Val Value.Null => Predicate.bot_pred
                                  | Expr.Val (Value.Bool _) =>
                                    Predicate.bot_pred
                                  | Expr.Val (Value.Intg _) =>
                                    Predicate.bot_pred
                                  | Expr.Val (Value.Addr a) =>
                                    (case xq (State.hp (xv, xw)) a
                                      of NONE => Predicate.bot_pred
                                      | SOME (Type.Class_type _) =>
Predicate.bot_pred
                                      | SOME (Type.Array_type (_, n)) =>
Predicate.bind
  (Predicate.if_pred
    (Word.word_sless
       (Type_Length.len_bit0
         (Type_Length.len_bit0
           (Type_Length.len_bit0
             (Type_Length.len_bit0
               (Type_Length.len_bit0 Type_Length.len_num1)))))
       i (Word.zero_word
           (Type_Length.len0_bit0
             (Type_Length.len0_bit0
               (Type_Length.len0_bit0
                 (Type_Length.len0_bit0
                   (Type_Length.len0_bit0 Type_Length.len0_num1)))))) orelse
      Arith.less_eq_int (Arith.int_of_nat n)
        (Word.sint
          (Type_Length.len_bit0
            (Type_Length.len_bit0
              (Type_Length.len_bit0
                (Type_Length.len_bit0
                  (Type_Length.len_bit0 Type_Length.len_num1)))))
          i)))
  (fn () =>
    Predicate.single
      ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
        (Expr.Throw
           (Expr.Val
             (Value.Addr
               (StartConfig.addr_of_sys_xcpt C_ (A1_, A2_) xo xp
                 "java/lang/ArrayIndexOutOfBoundsException"))),
          (xv, xw)))))
                                  | Expr.BinOp (_, _, _) => Predicate.bot_pred
                                  | Expr.Var _ => Predicate.bot_pred
                                  | Expr.LAss (_, _) => Predicate.bot_pred
                                  | Expr.AAcc (_, _) => Predicate.bot_pred
                                  | Expr.AAss (_, _, _) => Predicate.bot_pred
                                  | Expr.ALen _ => Predicate.bot_pred
                                  | Expr.FAcc (_, _, _) => Predicate.bot_pred
                                  | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                                  | Expr.Call (_, _, _) => Predicate.bot_pred
                                  | Expr.Block (_, _, _, _) =>
                                    Predicate.bot_pred
                                  | Expr.Synchronized (_, _, _) =>
                                    Predicate.bot_pred
                                  | Expr.InSynchronized (_, _, _) =>
                                    Predicate.bot_pred
                                  | Expr.Seq (_, _) => Predicate.bot_pred
                                  | Expr.Cond (_, _, _) => Predicate.bot_pred
                                  | Expr.While (_, _) => Predicate.bot_pred
                                  | Expr.Throw _ => Predicate.bot_pred
                                  | Expr.TryCatch (_, _, _, _) =>
                                    Predicate.bot_pred)
                              | Value.Addr _ => Predicate.bot_pred)
                            (case c1 of Expr.New _ => Predicate.bot_pred
                              | Expr.NewArray (_, _) => Predicate.bot_pred
                              | Expr.Cast (_, _) => Predicate.bot_pred
                              | Expr.InstanceOf (_, _) => Predicate.bot_pred
                              | Expr.Val Value.Unit => Predicate.bot_pred
                              | Expr.Val Value.Null =>
                                Predicate.single
                                  ((FinFun.Finfun_const [],
                                     ([], ([], ([], ([], []))))),
                                    (Expr.Throw
                                       (Expr.Val
 (Value.Addr
   (StartConfig.addr_of_sys_xcpt C_ (A1_, A2_) xo xp
     "java/lang/NullPointerException"))),
                                      (xv, xw)))
                              | Expr.Val (Value.Bool _) => Predicate.bot_pred
                              | Expr.Val (Value.Intg _) => Predicate.bot_pred
                              | Expr.Val (Value.Addr _) => Predicate.bot_pred
                              | Expr.BinOp (_, _, _) => Predicate.bot_pred
                              | Expr.Var _ => Predicate.bot_pred
                              | Expr.LAss (_, _) => Predicate.bot_pred
                              | Expr.AAcc (_, _) => Predicate.bot_pred
                              | Expr.AAss (_, _, _) => Predicate.bot_pred
                              | Expr.ALen _ => Predicate.bot_pred
                              | Expr.FAcc (_, _, _) => Predicate.bot_pred
                              | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                              | Expr.Call (_, _, _) => Predicate.bot_pred
                              | Expr.Block (_, _, _, _) => Predicate.bot_pred
                              | Expr.Synchronized (_, _, _) =>
                                Predicate.bot_pred
                              | Expr.InSynchronized (_, _, _) =>
                                Predicate.bot_pred
                              | Expr.Seq (_, _) => Predicate.bot_pred
                              | Expr.Cond (_, _, _) => Predicate.bot_pred
                              | Expr.While (_, _) => Predicate.bot_pred
                              | Expr.Throw _ => Predicate.bot_pred
                              | Expr.TryCatch (_, _, _, _) =>
                                Predicate.bot_pred)))
                    | Expr.BinOp (_, _, _) => Predicate.bot_pred
                    | Expr.Var _ => Predicate.bot_pred
                    | Expr.LAss (_, _) => Predicate.bot_pred
                    | Expr.AAcc (_, _) => Predicate.bot_pred
                    | Expr.AAss (_, _, _) => Predicate.bot_pred
                    | Expr.ALen _ => Predicate.bot_pred
                    | Expr.FAcc (_, _, _) => Predicate.bot_pred
                    | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                    | Expr.Call (_, _, _) => Predicate.bot_pred
                    | Expr.Block (_, _, _, _) => Predicate.bot_pred
                    | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                    | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                    | Expr.Seq (_, _) => Predicate.bot_pred
                    | Expr.Cond (_, _, _) => Predicate.bot_pred
                    | Expr.While (_, _) => Predicate.bot_pred
                    | Expr.Throw _ => Predicate.bot_pred
                    | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
                | Expr.BinOp (_, _, _) => Predicate.bot_pred
                | Expr.Var _ => Predicate.bot_pred
                | Expr.LAss (_, _) => Predicate.bot_pred
                | Expr.AAcc (_, _) => Predicate.bot_pred
                | Expr.AAss (_, _, _) => Predicate.bot_pred
                | Expr.ALen _ => Predicate.bot_pred
                | Expr.FAcc (_, _, _) => Predicate.bot_pred
                | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                | Expr.Call (_, _, _) => Predicate.bot_pred
                | Expr.Block (_, _, _, _) => Predicate.bot_pred
                | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                | Expr.Seq (_, _) => Predicate.bot_pred
                | Expr.Cond (_, _, _) => Predicate.bot_pred
                | Expr.While (_, _) => Predicate.bot_pred
                | Expr.Throw _ => Predicate.bot_pred
                | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
              (Predicate.sup_pred
                (case c2 of Expr.New _ => Predicate.bot_pred
                  | Expr.NewArray (_, _) => Predicate.bot_pred
                  | Expr.Cast (_, _) => Predicate.bot_pred
                  | Expr.InstanceOf (_, _) => Predicate.bot_pred
                  | Expr.Val ib =>
                    (case c1 of Expr.New _ => Predicate.bot_pred
                      | Expr.NewArray (_, _) => Predicate.bot_pred
                      | Expr.Cast (_, _) => Predicate.bot_pred
                      | Expr.InstanceOf (_, _) => Predicate.bot_pred
                      | Expr.Val ab =>
                        Predicate.bind
                          (red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o
                            (A1_, A2_) C_ xl xm xn xo xp xq heap_read heap_write
                            xr xs xt c3 (xv, xw))
                          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf)))
                             =>
                            Predicate.single
                              ((xa, (xb, (xc, (xd, (xe, ye))))),
                                (Expr.AAss (Expr.Val ab, Expr.Val ib, e),
                                  (xf, yf))))
                      | Expr.BinOp (_, _, _) => Predicate.bot_pred
                      | Expr.Var _ => Predicate.bot_pred
                      | Expr.LAss (_, _) => Predicate.bot_pred
                      | Expr.AAcc (_, _) => Predicate.bot_pred
                      | Expr.AAss (_, _, _) => Predicate.bot_pred
                      | Expr.ALen _ => Predicate.bot_pred
                      | Expr.FAcc (_, _, _) => Predicate.bot_pred
                      | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                      | Expr.Call (_, _, _) => Predicate.bot_pred
                      | Expr.Block (_, _, _, _) => Predicate.bot_pred
                      | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                      | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                      | Expr.Seq (_, _) => Predicate.bot_pred
                      | Expr.Cond (_, _, _) => Predicate.bot_pred
                      | Expr.While (_, _) => Predicate.bot_pred
                      | Expr.Throw _ => Predicate.bot_pred
                      | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
                  | Expr.BinOp (_, _, _) => Predicate.bot_pred
                  | Expr.Var _ => Predicate.bot_pred
                  | Expr.LAss (_, _) => Predicate.bot_pred
                  | Expr.AAcc (_, _) => Predicate.bot_pred
                  | Expr.AAss (_, _, _) => Predicate.bot_pred
                  | Expr.ALen _ => Predicate.bot_pred
                  | Expr.FAcc (_, _, _) => Predicate.bot_pred
                  | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                  | Expr.Call (_, _, _) => Predicate.bot_pred
                  | Expr.Block (_, _, _, _) => Predicate.bot_pred
                  | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                  | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                  | Expr.Seq (_, _) => Predicate.bot_pred
                  | Expr.Cond (_, _, _) => Predicate.bot_pred
                  | Expr.While (_, _) => Predicate.bot_pred
                  | Expr.Throw _ => Predicate.bot_pred
                  | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
                (Predicate.sup_pred
                  (case c1 of Expr.New _ => Predicate.bot_pred
                    | Expr.NewArray (_, _) => Predicate.bot_pred
                    | Expr.Cast (_, _) => Predicate.bot_pred
                    | Expr.InstanceOf (_, _) => Predicate.bot_pred
                    | Expr.Val ab =>
                      Predicate.bind
                        (red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o
                          (A1_, A2_) C_ xl xm xn xo xp xq heap_read heap_write
                          xr xs xt c2 (xv, xw))
                        (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (i, (xf, yf))) =>
                          Predicate.single
                            ((xa, (xb, (xc, (xd, (xe, ye))))),
                              (Expr.AAss (Expr.Val ab, i, c3), (xf, yf))))
                    | Expr.BinOp (_, _, _) => Predicate.bot_pred
                    | Expr.Var _ => Predicate.bot_pred
                    | Expr.LAss (_, _) => Predicate.bot_pred
                    | Expr.AAcc (_, _) => Predicate.bot_pred
                    | Expr.AAss (_, _, _) => Predicate.bot_pred
                    | Expr.ALen _ => Predicate.bot_pred
                    | Expr.FAcc (_, _, _) => Predicate.bot_pred
                    | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                    | Expr.Call (_, _, _) => Predicate.bot_pred
                    | Expr.Block (_, _, _, _) => Predicate.bot_pred
                    | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                    | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                    | Expr.Seq (_, _) => Predicate.bot_pred
                    | Expr.Cond (_, _, _) => Predicate.bot_pred
                    | Expr.While (_, _) => Predicate.bot_pred
                    | Expr.Throw _ => Predicate.bot_pred
                    | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
                  (Predicate.bind
                    (red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o
                      (A1_, A2_) C_ xl xm xn xo xp xq heap_read heap_write xr xs
                      xt c1 (xv, xw))
                    (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (a, (xf, yf))) =>
                      Predicate.single
                        ((xa, (xb, (xc, (xd, (xe, ye))))),
                          (Expr.AAss (a, c2, c3), (xf, yf))))))))))
    | Expr.ALen c1 =>
      Predicate.sup_pred
        (case c1 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val Value.Unit => Predicate.bot_pred
          | Expr.Val Value.Null =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw
                   (Expr.Val
                     (Value.Addr
                       (StartConfig.addr_of_sys_xcpt C_ (A1_, A2_) xo xp
                         "java/lang/NullPointerException"))),
                  (xv, xw)))
          | Expr.Val (Value.Bool _) => Predicate.bot_pred
          | Expr.Val (Value.Intg _) => Predicate.bot_pred
          | Expr.Val (Value.Addr a) =>
            (case xq xv a of NONE => Predicate.bot_pred
              | SOME (Type.Class_type _) => Predicate.bot_pred
              | SOME (Type.Array_type (_, n)) =>
                Predicate.single
                  ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                    (Expr.Val
                       (Value.Intg
                         (Word.word_of_int
                           (Type_Length.len0_bit0
                             (Type_Length.len0_bit0
                               (Type_Length.len0_bit0
                                 (Type_Length.len0_bit0
                                   (Type_Length.len0_bit0
                                     Type_Length.len0_num1)))))
                           (Arith.int_of_nat n))),
                      (xv, xw))))
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw (Expr.Val (Value.Addr a)), (xv, xw)))
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.bind
          (red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) C_ xl
            xm xn xo xp xq heap_read heap_write xr xs xt c1 (xv, xw))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (a, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))), (Expr.ALen a, (xf, yf)))))
    | Expr.FAcc (c1, c2, c3) =>
      Predicate.sup_pred
        (case c1 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val Value.Unit => Predicate.bot_pred
          | Expr.Val Value.Null =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw
                   (Expr.Val
                     (Value.Addr
                       (StartConfig.addr_of_sys_xcpt C_ (A1_, A2_) xo xp
                         "java/lang/NullPointerException"))),
                  (xv, xw)))
          | Expr.Val (Value.Bool _) => Predicate.bot_pred
          | Expr.Val (Value.Intg _) => Predicate.bot_pred
          | Expr.Val (Value.Addr a) =>
            Predicate.bind (heap_read xv a (Heap.CField (c3, c2)))
              (fn x =>
                Predicate.single
                  (FWState.thread_action_to_thread_action A1_
                     Observable_Events.obs_action_obs_event
                     (FWState.ObsAction
                       (Observable_Events.ReadMem (a, Heap.CField (c3, c2), x)))
                     (FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                    (Expr.Val x, (xv, xw))))
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw (Expr.Val (Value.Addr a)), (xv, xw)))
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.bind
          (red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) C_ xl
            xm xn xo xp xq heap_read heap_write xr xs xt c1 (xv, xw))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))),
                (Expr.FAcc (e, c2, c3), (xf, yf)))))
    | Expr.FAss (c1, c2, c3, c4) =>
      Predicate.sup_pred
        (case c4 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val _ => Predicate.bot_pred
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            (case c1 of Expr.New _ => Predicate.bot_pred
              | Expr.NewArray (_, _) => Predicate.bot_pred
              | Expr.Cast (_, _) => Predicate.bot_pred
              | Expr.InstanceOf (_, _) => Predicate.bot_pred
              | Expr.Val _ =>
                Predicate.single
                  ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                    (Expr.Throw (Expr.Val (Value.Addr a)), (xv, xw)))
              | Expr.BinOp (_, _, _) => Predicate.bot_pred
              | Expr.Var _ => Predicate.bot_pred
              | Expr.LAss (_, _) => Predicate.bot_pred
              | Expr.AAcc (_, _) => Predicate.bot_pred
              | Expr.AAss (_, _, _) => Predicate.bot_pred
              | Expr.ALen _ => Predicate.bot_pred
              | Expr.FAcc (_, _, _) => Predicate.bot_pred
              | Expr.FAss (_, _, _, _) => Predicate.bot_pred
              | Expr.Call (_, _, _) => Predicate.bot_pred
              | Expr.Block (_, _, _, _) => Predicate.bot_pred
              | Expr.Synchronized (_, _, _) => Predicate.bot_pred
              | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
              | Expr.Seq (_, _) => Predicate.bot_pred
              | Expr.Cond (_, _, _) => Predicate.bot_pred
              | Expr.While (_, _) => Predicate.bot_pred
              | Expr.Throw _ => Predicate.bot_pred
              | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.sup_pred
          (case c1 of Expr.New _ => Predicate.bot_pred
            | Expr.NewArray (_, _) => Predicate.bot_pred
            | Expr.Cast (_, _) => Predicate.bot_pred
            | Expr.InstanceOf (_, _) => Predicate.bot_pred
            | Expr.Val _ => Predicate.bot_pred
            | Expr.BinOp (_, _, _) => Predicate.bot_pred
            | Expr.Var _ => Predicate.bot_pred
            | Expr.LAss (_, _) => Predicate.bot_pred
            | Expr.AAcc (_, _) => Predicate.bot_pred
            | Expr.AAss (_, _, _) => Predicate.bot_pred
            | Expr.ALen _ => Predicate.bot_pred
            | Expr.FAcc (_, _, _) => Predicate.bot_pred
            | Expr.FAss (_, _, _, _) => Predicate.bot_pred
            | Expr.Call (_, _, _) => Predicate.bot_pred
            | Expr.Block (_, _, _, _) => Predicate.bot_pred
            | Expr.Synchronized (_, _, _) => Predicate.bot_pred
            | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
            | Expr.Seq (_, _) => Predicate.bot_pred
            | Expr.Cond (_, _, _) => Predicate.bot_pred
            | Expr.While (_, _) => Predicate.bot_pred
            | Expr.Throw (Expr.New _) => Predicate.bot_pred
            | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
            | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
            | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Val (Value.Addr a)) =>
              Predicate.single
                ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                  (Expr.Throw (Expr.Val (Value.Addr a)), (xv, xw)))
            | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Var _) => Predicate.bot_pred
            | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
            | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
            | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
            | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
            | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
          (Predicate.sup_pred
            (case c4 of Expr.New _ => Predicate.bot_pred
              | Expr.NewArray (_, _) => Predicate.bot_pred
              | Expr.Cast (_, _) => Predicate.bot_pred
              | Expr.InstanceOf (_, _) => Predicate.bot_pred
              | Expr.Val v =>
                (case c1 of Expr.New _ => Predicate.bot_pred
                  | Expr.NewArray (_, _) => Predicate.bot_pred
                  | Expr.Cast (_, _) => Predicate.bot_pred
                  | Expr.InstanceOf (_, _) => Predicate.bot_pred
                  | Expr.Val Value.Unit => Predicate.bot_pred
                  | Expr.Val Value.Null =>
                    Predicate.single
                      ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                        (Expr.Throw
                           (Expr.Val
                             (Value.Addr
                               (StartConfig.addr_of_sys_xcpt C_ (A1_, A2_) xo xp
                                 "java/lang/NullPointerException"))),
                          (xv, xw)))
                  | Expr.Val (Value.Bool _) => Predicate.bot_pred
                  | Expr.Val (Value.Intg _) => Predicate.bot_pred
                  | Expr.Val (Value.Addr a) =>
                    Predicate.bind (heap_write xv a (Heap.CField (c3, c2)) v)
                      (fn x =>
                        Predicate.single
                          (FWState.thread_action_to_thread_action A1_
                             Observable_Events.obs_action_obs_event
                             (FWState.ObsAction
                               (Observable_Events.WriteMem
                                 (a, Heap.CField (c3, c2), v)))
                             (FinFun.Finfun_const [],
                               ([], ([], ([], ([], []))))),
                            (Expr.Val Value.Unit, (x, xw))))
                  | Expr.BinOp (_, _, _) => Predicate.bot_pred
                  | Expr.Var _ => Predicate.bot_pred
                  | Expr.LAss (_, _) => Predicate.bot_pred
                  | Expr.AAcc (_, _) => Predicate.bot_pred
                  | Expr.AAss (_, _, _) => Predicate.bot_pred
                  | Expr.ALen _ => Predicate.bot_pred
                  | Expr.FAcc (_, _, _) => Predicate.bot_pred
                  | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                  | Expr.Call (_, _, _) => Predicate.bot_pred
                  | Expr.Block (_, _, _, _) => Predicate.bot_pred
                  | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                  | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                  | Expr.Seq (_, _) => Predicate.bot_pred
                  | Expr.Cond (_, _, _) => Predicate.bot_pred
                  | Expr.While (_, _) => Predicate.bot_pred
                  | Expr.Throw _ => Predicate.bot_pred
                  | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
              | Expr.BinOp (_, _, _) => Predicate.bot_pred
              | Expr.Var _ => Predicate.bot_pred
              | Expr.LAss (_, _) => Predicate.bot_pred
              | Expr.AAcc (_, _) => Predicate.bot_pred
              | Expr.AAss (_, _, _) => Predicate.bot_pred
              | Expr.ALen _ => Predicate.bot_pred
              | Expr.FAcc (_, _, _) => Predicate.bot_pred
              | Expr.FAss (_, _, _, _) => Predicate.bot_pred
              | Expr.Call (_, _, _) => Predicate.bot_pred
              | Expr.Block (_, _, _, _) => Predicate.bot_pred
              | Expr.Synchronized (_, _, _) => Predicate.bot_pred
              | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
              | Expr.Seq (_, _) => Predicate.bot_pred
              | Expr.Cond (_, _, _) => Predicate.bot_pred
              | Expr.While (_, _) => Predicate.bot_pred
              | Expr.Throw _ => Predicate.bot_pred
              | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
            (Predicate.sup_pred
              (case c1 of Expr.New _ => Predicate.bot_pred
                | Expr.NewArray (_, _) => Predicate.bot_pred
                | Expr.Cast (_, _) => Predicate.bot_pred
                | Expr.InstanceOf (_, _) => Predicate.bot_pred
                | Expr.Val v =>
                  Predicate.bind
                    (red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o
                      (A1_, A2_) C_ xl xm xn xo xp xq heap_read heap_write xr xs
                      xt c4 (xv, xw))
                    (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
                      Predicate.single
                        ((xa, (xb, (xc, (xd, (xe, ye))))),
                          (Expr.FAss (Expr.Val v, c2, c3, e), (xf, yf))))
                | Expr.BinOp (_, _, _) => Predicate.bot_pred
                | Expr.Var _ => Predicate.bot_pred
                | Expr.LAss (_, _) => Predicate.bot_pred
                | Expr.AAcc (_, _) => Predicate.bot_pred
                | Expr.AAss (_, _, _) => Predicate.bot_pred
                | Expr.ALen _ => Predicate.bot_pred
                | Expr.FAcc (_, _, _) => Predicate.bot_pred
                | Expr.FAss (_, _, _, _) => Predicate.bot_pred
                | Expr.Call (_, _, _) => Predicate.bot_pred
                | Expr.Block (_, _, _, _) => Predicate.bot_pred
                | Expr.Synchronized (_, _, _) => Predicate.bot_pred
                | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
                | Expr.Seq (_, _) => Predicate.bot_pred
                | Expr.Cond (_, _, _) => Predicate.bot_pred
                | Expr.While (_, _) => Predicate.bot_pred
                | Expr.Throw _ => Predicate.bot_pred
                | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
              (Predicate.bind
                (red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_)
                  C_ xl xm xn xo xp xq heap_read heap_write xr xs xt c1
                  (xv, xw))
                (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
                  Predicate.single
                    ((xa, (xb, (xc, (xd, (xe, ye))))),
                      (Expr.FAss (e, c2, c3, c4), (xf, yf))))))))
    | Expr.Call (c1, c2, c3) =>
      Predicate.sup_pred
        (case c1 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val v =>
            Predicate.sup_pred
              (Predicate.bind (Expr.is_Throws_i c3)
                (fn () =>
                  Predicate.single
                    ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                      (List.hd (List.dropWhile Expr.is_val c3), (xv, xw)))))
              (Predicate.sup_pred
                (case v of Value.Unit => Predicate.bot_pred
                  | Value.Null =>
                    Predicate.bind (Predicate.if_pred (Expr.is_vals c3))
                      (fn () =>
                        Predicate.single
                          ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                            (Expr.Throw
                               (Expr.Val
                                 (Value.Addr
                                   (StartConfig.addr_of_sys_xcpt C_ (A1_, A2_)
                                     xo xp "java/lang/NullPointerException"))),
                              (xv, xw))))
                  | Value.Bool _ => Predicate.bot_pred
                  | Value.Intg _ => Predicate.bot_pred
                  | Value.Addr _ => Predicate.bot_pred)
                (Predicate.sup_pred
                  (case v of Value.Unit => Predicate.bot_pred
                    | Value.Null => Predicate.bot_pred
                    | Value.Bool _ => Predicate.bot_pred
                    | Value.Intg _ => Predicate.bot_pred
                    | Value.Addr a =>
                      Predicate.bind (Predicate.if_pred (Expr.is_vals c3))
                        (fn () =>
                          (case xq (State.hp (xv, xw)) a
                            of NONE => Predicate.bot_pred
                            | SOME hU =>
                              Predicate.bind
                                (ExternalCall.red_external_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_i_o_o_o
                                  (A1_, A2_) C_ xl xm xn xo xp xq heap_read
                                  heap_write xs xt (State.hp (xv, xw)) a c2
                                  (List.map Expr.the_Val c3))
                                (fn ((xfa, (xg, (xh, (xi, (xj, yj))))), (va, h))
                                   =>
                                  Predicate.bind
                                    (TypeRel.method_i_i_i_o_o_o_o xs
                                      (Type.class_type_ofa hU) c2)
                                    (fn b =>
                                      (case b
of (_, (_, (NONE, _))) =>
  Predicate.single
    (xr (xfa, (xg, (xh, (xi, (xj, yj))))),
      (Expr.extRet2J (Expr.Call (Expr.Val (Value.Addr a), c2, c3)) va,
        (h, State.lcl (xv, xw))))
| (_, (_, (SOME _, _))) => Predicate.bot_pred))))))
                  (Predicate.sup_pred
                    (case v of Value.Unit => Predicate.bot_pred
                      | Value.Null => Predicate.bot_pred
                      | Value.Bool _ => Predicate.bot_pred
                      | Value.Intg _ => Predicate.bot_pred
                      | Value.Addr a =>
                        Predicate.bind (Predicate.if_pred (Expr.is_vals c3))
                          (fn () =>
                            (case xq (State.hp (xv, xw)) a
                              of NONE => Predicate.bot_pred
                              | SOME hU =>
                                Predicate.bind
                                  (TypeRel.method_i_i_i_o_o_o_o xs
                                    (Type.class_type_ofa hU) c2)
                                  (fn b =>
                                    (case b
                                      of (ts, (_, (NONE, d))) =>
Predicate.bot_pred
                                      | (ts, (_, (SOME (pns, body), d))) =>
Predicate.bind
  (TypeRel.eq_i_i Arith.equal_nat (List.size_list c3) (List.size_list pns))
  (fn () =>
    Predicate.bind
      (TypeRel.eq_i_i Arith.equal_nat (List.size_list ts) (List.size_list pns))
      (fn () =>
        Predicate.single
          ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
            (Expr.blocks ("this" :: pns) (Type.Class d :: ts)
               (Value.Addr a :: List.map Expr.the_Val c3) body,
              (xv, xw))))))))))
                    (Predicate.bind
                      (reds_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o
                        (A1_, A2_) C_ xl xm xn xo xp xq heap_read heap_write xr
                        xs xt c3 (xv, xw))
                      (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (es, (xf, yf))) =>
                        Predicate.single
                          ((xa, (xb, (xc, (xd, (xe, ye))))),
                            (Expr.Call (Expr.Val v, c2, es), (xf, yf))))))))
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw (Expr.Val (Value.Addr a)), (xv, xw)))
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.bind
          (red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) C_ xl
            xm xn xo xp xq heap_read heap_write xr xs xt c1 (xv, xw))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))),
                (Expr.Call (e, c2, c3), (xf, yf)))))
    | Expr.Block (c1, c2, c3, c4) =>
      Predicate.sup_pred
        (case c4 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val v =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Val v, (xv, xw)))
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw (Expr.Val (Value.Addr a)), (xv, xw)))
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.bind
          (red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) C_ xl
            xm xn xo xp xq heap_read heap_write xr xs xt c4
            (xv, Fun.fun_upd Stringa.equal_literal xw c1 c3))
          (fn ((xk, (xaa, (xba, (xca, (xda, yd))))), (e, (h, l))) =>
            Predicate.single
              ((xk, (xaa, (xba, (xca, (xda, yd))))),
                (Expr.Block (c1, c2, l c1, e),
                  (h, Fun.fun_upd Stringa.equal_literal l c1 (xw c1))))))
    | Expr.Synchronized (c1, c2, c3) =>
      Predicate.sup_pred
        (case c2 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val Value.Unit => Predicate.bot_pred
          | Expr.Val Value.Null =>
            let
              val () = c1;
            in
              Predicate.single
                ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                  (Expr.Throw
                     (Expr.Val
                       (Value.Addr
                         (StartConfig.addr_of_sys_xcpt C_ (A1_, A2_) xo xp
                           "java/lang/NullPointerException"))),
                    (xv, xw)))
            end
          | Expr.Val (Value.Bool _) => Predicate.bot_pred
          | Expr.Val (Value.Intg _) => Predicate.bot_pred
          | Expr.Val (Value.Addr a) =>
            let
              val () = c1;
            in
              Predicate.single
                (FWState.thread_action_to_thread_action A1_
                   Observable_Events.obs_action_obs_event
                   (FWState.ObsAction (Observable_Events.SyncLock a))
                   (FWState.thread_action_to_thread_action A1_
                     Observable_Events.obs_action_obs_event
                     (FWState.LockAction (FWState.Lock, a))
                     (FinFun.Finfun_const [], ([], ([], ([], ([], [])))))),
                  (Expr.InSynchronized ((), a, c3), (xv, xw)))
            end
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            let
              val () = c1;
            in
              Predicate.single
                ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                  (Expr.Throw (Expr.Val (Value.Addr a)), (xv, xw)))
            end
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        let
          val () = c1;
        in
          Predicate.bind
            (red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) C_
              xl xm xn xo xp xq heap_read heap_write xr xs xt c2 (xv, xw))
            (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (oa, (xf, yf))) =>
              Predicate.single
                ((xa, (xb, (xc, (xd, (xe, ye))))),
                  (Expr.Synchronized ((), oa, c3), (xf, yf))))
        end
    | Expr.InSynchronized (c1, c2, c3) =>
      Predicate.sup_pred
        (case c3 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val v =>
            let
              val () = c1;
            in
              Predicate.single
                (FWState.thread_action_to_thread_action A1_
                   Observable_Events.obs_action_obs_event
                   (FWState.ObsAction (Observable_Events.SyncUnlock c2))
                   (FWState.thread_action_to_thread_action A1_
                     Observable_Events.obs_action_obs_event
                     (FWState.LockAction (FWState.Unlock, c2))
                     (FinFun.Finfun_const [], ([], ([], ([], ([], [])))))),
                  (Expr.Val v, (xv, xw)))
            end
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr ad)) =>
            let
              val () = c1;
            in
              Predicate.single
                (FWState.thread_action_to_thread_action A1_
                   Observable_Events.obs_action_obs_event
                   (FWState.ObsAction (Observable_Events.SyncUnlock c2))
                   (FWState.thread_action_to_thread_action A1_
                     Observable_Events.obs_action_obs_event
                     (FWState.LockAction (FWState.Unlock, c2))
                     (FinFun.Finfun_const [], ([], ([], ([], ([], [])))))),
                  (Expr.Throw (Expr.Val (Value.Addr ad)), (xv, xw)))
            end
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        let
          val () = c1;
        in
          Predicate.bind
            (red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) C_
              xl xm xn xo xp xq heap_read heap_write xr xs xt c3 (xv, xw))
            (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
              Predicate.single
                ((xa, (xb, (xc, (xd, (xe, ye))))),
                  (Expr.InSynchronized ((), c2, e), (xf, yf))))
        end
    | Expr.Seq (c1, c2) =>
      Predicate.sup_pred
        (case c1 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val _ =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (c2, (xv, xw)))
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw (Expr.Val (Value.Addr a)), (xv, xw)))
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.bind
          (red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) C_ xl
            xm xn xo xp xq heap_read heap_write xr xs xt c1 (xv, xw))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))), (Expr.Seq (e, c2), (xf, yf)))))
    | Expr.Cond (c1, c2, c3) =>
      Predicate.sup_pred
        (case c1 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val Value.Unit => Predicate.bot_pred
          | Expr.Val Value.Null => Predicate.bot_pred
          | Expr.Val (Value.Bool true) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (c2, (xv, xw)))
          | Expr.Val (Value.Bool false) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (c3, (xv, xw)))
          | Expr.Val (Value.Intg _) => Predicate.bot_pred
          | Expr.Val (Value.Addr _) => Predicate.bot_pred
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw (Expr.Val (Value.Addr a)), (xv, xw)))
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.bind
          (red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) C_ xl
            xm xn xo xp xq heap_read heap_write xr xs xt c1 (xv, xw))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (b, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))),
                (Expr.Cond (b, c2, c3), (xf, yf)))))
    | Expr.While (c1, c2) =>
      Predicate.single
        ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
          (Expr.Cond
             (c1, Expr.Seq (c2, Expr.While (c1, c2)), Expr.Val Value.Unit),
            (xv, xw)))
    | Expr.Throw c1 =>
      Predicate.sup_pred
        (case c1 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val Value.Unit => Predicate.bot_pred
          | Expr.Val Value.Null =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw
                   (Expr.Val
                     (Value.Addr
                       (StartConfig.addr_of_sys_xcpt C_ (A1_, A2_) xo xp
                         "java/lang/NullPointerException"))),
                  (xv, xw)))
          | Expr.Val (Value.Bool _) => Predicate.bot_pred
          | Expr.Val (Value.Intg _) => Predicate.bot_pred
          | Expr.Val (Value.Addr _) => Predicate.bot_pred
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Throw (Expr.Val (Value.Addr a)), (xv, xw)))
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.bind
          (red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) C_ xl
            xm xn xo xp xq heap_read heap_write xr xs xt c1 (xv, xw))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))), (Expr.Throw e, (xf, yf)))))
    | Expr.TryCatch (c1, c2, c3, c4) =>
      Predicate.sup_pred
        (case c1 of Expr.New _ => Predicate.bot_pred
          | Expr.NewArray (_, _) => Predicate.bot_pred
          | Expr.Cast (_, _) => Predicate.bot_pred
          | Expr.InstanceOf (_, _) => Predicate.bot_pred
          | Expr.Val v =>
            Predicate.single
              ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                (Expr.Val v, (xv, xw)))
          | Expr.BinOp (_, _, _) => Predicate.bot_pred
          | Expr.Var _ => Predicate.bot_pred
          | Expr.LAss (_, _) => Predicate.bot_pred
          | Expr.AAcc (_, _) => Predicate.bot_pred
          | Expr.AAss (_, _, _) => Predicate.bot_pred
          | Expr.ALen _ => Predicate.bot_pred
          | Expr.FAcc (_, _, _) => Predicate.bot_pred
          | Expr.FAss (_, _, _, _) => Predicate.bot_pred
          | Expr.Call (_, _, _) => Predicate.bot_pred
          | Expr.Block (_, _, _, _) => Predicate.bot_pred
          | Expr.Synchronized (_, _, _) => Predicate.bot_pred
          | Expr.InSynchronized (_, _, _) => Predicate.bot_pred
          | Expr.Seq (_, _) => Predicate.bot_pred
          | Expr.Cond (_, _, _) => Predicate.bot_pred
          | Expr.While (_, _) => Predicate.bot_pred
          | Expr.Throw (Expr.New _) => Predicate.bot_pred
          | Expr.Throw (Expr.NewArray (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cast (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InstanceOf (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Unit) => Predicate.bot_pred
          | Expr.Throw (Expr.Val Value.Null) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Bool _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Intg _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Val (Value.Addr a)) =>
            Predicate.sup_pred
              (case xq (State.hp (xv, xw)) a of NONE => Predicate.bot_pred
                | SOME (Type.Class_type d) =>
                  Predicate.bind
                    (Predicate.not_pred
                      (Transitive_Closure_Table.rtranclp_FioB_i_i
                        Stringa.equal_literal (TypeRel.subcls1_i_i_o xs) d c2))
                    (fn () =>
                      Predicate.single
                        ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                          (Expr.Throw (Expr.Val (Value.Addr a)), (xv, xw))))
                | SOME (Type.Array_type (_, _)) => Predicate.bot_pred)
              (case xq (State.hp (xv, xw)) a of NONE => Predicate.bot_pred
                | SOME (Type.Class_type d) =>
                  Predicate.bind
                    (Transitive_Closure_Table.rtranclp_FioB_i_i
                      Stringa.equal_literal (TypeRel.subcls1_i_i_o xs) d c2)
                    (fn () =>
                      Predicate.single
                        ((FinFun.Finfun_const [], ([], ([], ([], ([], []))))),
                          (Expr.Block
                             (c3, Type.Class c2, SOME (Value.Addr a), c4),
                            (xv, xw))))
                | SOME (Type.Array_type (_, _)) => Predicate.bot_pred)
          | Expr.Throw (Expr.BinOp (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Var _) => Predicate.bot_pred
          | Expr.Throw (Expr.LAss (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAcc (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.AAss (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.ALen _) => Predicate.bot_pred
          | Expr.Throw (Expr.FAcc (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.FAss (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Call (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Block (_, _, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Synchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.InSynchronized (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Seq (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Cond (_, _, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.While (_, _)) => Predicate.bot_pred
          | Expr.Throw (Expr.Throw _) => Predicate.bot_pred
          | Expr.Throw (Expr.TryCatch (_, _, _, _)) => Predicate.bot_pred
          | Expr.TryCatch (_, _, _, _) => Predicate.bot_pred)
        (Predicate.bind
          (red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) C_ xl
            xm xn xo xp xq heap_read heap_write xr xs xt c1 (xv, xw))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (e, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))),
                (Expr.TryCatch (e, c2, c3, c4), (xf, yf))))))
and reds_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) C_ xl xm xn
  xo xp xq heap_read heap_write xr xs xt xu (xv, xw) =
  Predicate.sup_pred
    (case xu of [] => Predicate.bot_pred
      | e :: es =>
        Predicate.bind
          (red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) C_ xl
            xm xn xo xp xq heap_read heap_write xr xs xt e (xv, xw))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (ea, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))), (ea :: es, (xf, yf)))))
    (case xu of [] => Predicate.bot_pred
      | Expr.New _ :: es => Predicate.bot_pred
      | Expr.NewArray (_, _) :: es => Predicate.bot_pred
      | Expr.Cast (_, _) :: es => Predicate.bot_pred
      | Expr.InstanceOf (_, _) :: es => Predicate.bot_pred
      | Expr.Val v :: es =>
        Predicate.bind
          (reds_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o (A1_, A2_) C_ xl
            xm xn xo xp xq heap_read heap_write xr xs xt es (xv, xw))
          (fn ((xa, (xb, (xc, (xd, (xe, ye))))), (esa, (xf, yf))) =>
            Predicate.single
              ((xa, (xb, (xc, (xd, (xe, ye))))), (Expr.Val v :: esa, (xf, yf))))
      | Expr.BinOp (_, _, _) :: es => Predicate.bot_pred
      | Expr.Var _ :: es => Predicate.bot_pred
      | Expr.LAss (_, _) :: es => Predicate.bot_pred
      | Expr.AAcc (_, _) :: es => Predicate.bot_pred
      | Expr.AAss (_, _, _) :: es => Predicate.bot_pred
      | Expr.ALen _ :: es => Predicate.bot_pred
      | Expr.FAcc (_, _, _) :: es => Predicate.bot_pred
      | Expr.FAss (_, _, _, _) :: es => Predicate.bot_pred
      | Expr.Call (_, _, _) :: es => Predicate.bot_pred
      | Expr.Block (_, _, _, _) :: es => Predicate.bot_pred
      | Expr.Synchronized (_, _, _) :: es => Predicate.bot_pred
      | Expr.InSynchronized (_, _, _) :: es => Predicate.bot_pred
      | Expr.Seq (_, _) :: es => Predicate.bot_pred
      | Expr.Cond (_, _, _) :: es => Predicate.bot_pred
      | Expr.While (_, _) :: es => Predicate.bot_pred
      | Expr.Throw _ :: es => Predicate.bot_pred
      | Expr.TryCatch (_, _, _, _) :: es => Predicate.bot_pred);

end; (*struct SmallStep*)

structure J_Execute : sig
  val sc_red_i_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o :
    (string list * (string, unit, Arith.nat) Expr.exp) TypeRel.prog ->
      Arith.nat ->
        ((string, unit, Arith.nat) Expr.exp *
          (string -> Arith.nat Value.vala option)) *
          (Arith.nat, SC_Collections.heapobj) RBT.rbt ->
          (((Arith.nat, (FWState.lock_action list)) FinFun.finfun *
             ((Arith.nat,
                ((string, unit, Arith.nat) Expr.exp *
                  (string -> Arith.nat Value.vala option)),
                (Arith.nat, SC_Collections.heapobj) RBT.rbt)
                FWState.new_thread_action list *
               (Arith.nat FWState.conditional_action list *
                 ((Arith.nat, Arith.nat) FWState.wait_set_action list *
                   (Arith.nat FWState.interrupt_action list *
                     (Arith.nat, Arith.nat)
                       Observable_Events.obs_event list))))) *
            (((string, unit, Arith.nat) Expr.exp *
               (string -> Arith.nat Value.vala option)) *
              (Arith.nat, SC_Collections.heapobj) RBT.rbt))
            Predicate.pred
  val sc_rr_J_start_state :
    Arith.nat ->
      'a TypeRel.prog -> (Arith.nat list * Arith.nat list) * Arith.nat
  val exec_J_rr :
    Arith.nat ->
      (string list * (string, unit, Arith.nat) Expr.exp) TypeRel.prog ->
        string ->
          string ->
            Arith.nat Value.vala list ->
              ((Arith.nat *
                 (Arith.nat, Arith.nat) Observable_Events.obs_event list),
                ((Arith.nat, ((Arith.nat * Arith.nat) option)) FinFun.finfun *
                  (((Arith.nat,
                      (((string, unit, Arith.nat) Expr.exp *
                         (string -> Arith.nat Value.vala option)) *
                        (Arith.nat, Arith.nat) FinFun.finfun))
                      RBT.rbt *
                     (Arith.nat, SC_Collections.heapobj) RBT.rbt) *
                    ((Arith.nat, Arith.nat FWState.wait_set_status) RBT.rbt *
                      (Arith.nat, unit) RBT.rbt))))
                TLList.tllist
  val sc_rnd_J_start_state :
    Arith.natural * Arith.natural -> Arith.natural * Arith.natural
  val exec_J_rnd :
    Arith.natural * Arith.natural ->
      (string list * (string, unit, Arith.nat) Expr.exp) TypeRel.prog ->
        string ->
          string ->
            Arith.nat Value.vala list ->
              ((Arith.nat *
                 (Arith.nat, Arith.nat) Observable_Events.obs_event list),
                ((Arith.nat, ((Arith.nat * Arith.nat) option)) FinFun.finfun *
                  (((Arith.nat,
                      (((string, unit, Arith.nat) Expr.exp *
                         (string -> Arith.nat Value.vala option)) *
                        (Arith.nat, Arith.nat) FinFun.finfun))
                      RBT.rbt *
                     (Arith.nat, SC_Collections.heapobj) RBT.rbt) *
                    ((Arith.nat, Arith.nat FWState.wait_set_status) RBT.rbt *
                      (Arith.nat, unit) RBT.rbt))))
                TLList.tllist
end = struct

fun sc_red_i_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o p t ((e, xs), h)
  = Predicate.bind
      (SmallStep.red_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o
        (Arith.equal_nat, MM.addr_nat)
        (CollectionsV1.equal_rbt (Arith.equal_nat, Arith.linorder_nat)
          SC_Collections.equal_heapobj)
        (fn x => x) (fn x => x) SC_Collections.sc_spurious_wakeups
        (RBT.empty Arith.linorder_nat) (SC_Collections.sc_allocate p)
        SC_Collections.sc_typeof_addr SC_Collections.sc_heap_read_i_i_i_o
        SC_Collections.sc_heap_write_i_i_i_i_o
        (FWState.convert_extTA (SmallStep.extNTA2J Arith.equal_nat p)) p t e
        (h, xs))
      (fn (ta, (ea, (ha, xsa))) => Predicate.single (ta, ((ea, xsa), ha)));

fun sc_rr_J_start_state n0 p =
  Round_Robin.round_robin_start Fifo.fifo_empty Fifo.fifo_appendr n0
    (StartConfig.start_tid (Arith.equal_nat, MM.addr_nat)
      (CollectionsV1.equal_rbt (Arith.equal_nat, Arith.linorder_nat)
        SC_Collections.equal_heapobj)
      (fn x => x) (RBT.empty Arith.linorder_nat)
      (SC_Collections.sc_allocate p));

fun exec_J_rr n0 p c m vs =
  Scheduler.exec Arith.equal_nat Arith.equal_nat Arith.equal_nat
    (Observable_Events.convert_RA (Arith.equal_nat, MM.addr_nat))
    (Round_Robin.round_robin Arith.equal_nat
      (Set.card_UNIV_nat, Arith.equal_nat) Arith.equal_nat
      (fn (e, _) => Expr.final e)
      (sc_red_i_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o p)
      (Observable_Events.convert_RA (Arith.equal_nat, MM.addr_nat))
      (fn k => fn ma => RBT.lookup Arith.linorder_nat ma k)
      (RBT.insert Arith.linorder_nat)
      (fn k => fn ma => RBT.lookup Arith.linorder_nat ma k)
      (RBTSetImpl.memb_rm_basic_ops Arith.linorder_nat)
      (RBTSetImpl.ins_rm_basic_ops Arith.linorder_nat)
      (RBTSetImpl.delete_rm_basic_ops Arith.linorder_nat)
      (Fifo.fifo_isEmpty Arith.equal_nat) Fifo.fifo_appendr Fifo.fifo_removel
      Fifo.fifo_appendl n0)
    State_Refinement.jinja_output
    (Scheduler.pick_wakeup_via_sel Arith.equal_nat
      (fn s => fn pa =>
        RBTMapImpl.g_sel_rm_basic_ops Arith.linorder_nat s
          (fn (a, b) => pa a b)))
    (fn k => fn ma => RBT.lookup Arith.linorder_nat ma k)
    (RBT.insert Arith.linorder_nat)
    (fn k => fn ma => RBT.lookup Arith.linorder_nat ma k)
    (RBT.insert Arith.linorder_nat) (RBT.delete Arith.linorder_nat)
    (RBTMapImpl.iteratei_map_op_list_it_rm_ops Arith.linorder_nat)
    (RBTSetImpl.ins_rm_basic_ops Arith.linorder_nat)
    (RBTSetImpl.delete_rm_basic_ops Arith.linorder_nat)
    (sc_rr_J_start_state n0 p)
    (State_Refinement.start_state_refine (Arith.equal_nat, MM.addr_nat)
      (CollectionsV1.equal_rbt (Arith.equal_nat, Arith.linorder_nat)
        SC_Collections.equal_heapobj)
      (Product_Type.equal_prod (List.equal_list Stringa.equal_literal)
        (Expr.equal_exp Stringa.equal_literal Product_Type.equal_unit
          Arith.equal_nat))
      (fn x => x) (RBT.empty Arith.linorder_nat) (SC_Collections.sc_allocate p)
      (RBT.empty Arith.linorder_nat) (RBT.insert Arith.linorder_nat)
      (RBT.empty Arith.linorder_nat)
      (RBTSetImpl.empty_rm_basic_ops Arith.linorder_nat ())
      (fn ca => fn _ => fn ts => fn _ => fn (pns, body) => fn vsa =>
        (Expr.blocks ("this" :: pns) (Type.Class ca :: ts) (Value.Null :: vsa)
           body,
          (fn _ => NONE)))
      p c m vs);

fun sc_rnd_J_start_state seed = seed;

fun exec_J_rnd seed p c m vs =
  Scheduler.exec Arith.equal_nat Arith.equal_nat Arith.equal_nat
    (Observable_Events.convert_RA (Arith.equal_nat, MM.addr_nat))
    (Random_Scheduler.random_scheduler Arith.equal_nat
      (Set.card_UNIV_nat, Arith.equal_nat) Arith.equal_nat
      (fn (e, _) => Expr.final e)
      (sc_red_i_i_i_i_i_i_i_i_Fii_i_oB_Fii_i_i_oB_i_i_i_i_i_o_o_o p)
      (Observable_Events.convert_RA (Arith.equal_nat, MM.addr_nat))
      (fn k => fn ma => RBT.lookup Arith.linorder_nat ma k)
      (RBT.insert Arith.linorder_nat)
      (RBTMapImpl.iteratei_map_op_list_it_rm_ops Arith.linorder_nat)
      (fn k => fn ma => RBT.lookup Arith.linorder_nat ma k)
      (RBTSetImpl.memb_rm_basic_ops Arith.linorder_nat)
      (RBTSetImpl.ins_rm_basic_ops Arith.linorder_nat)
      (RBTSetImpl.delete_rm_basic_ops Arith.linorder_nat) (fn _ => [])
      (fn a => fn b => a :: b) ListSetImpl_Invar.g_to_list_lsi_basic_ops)
    State_Refinement.jinja_output
    (Scheduler.pick_wakeup_via_sel Arith.equal_nat
      (fn s => fn pa =>
        RBTMapImpl.g_sel_rm_basic_ops Arith.linorder_nat s
          (fn (a, b) => pa a b)))
    (fn k => fn ma => RBT.lookup Arith.linorder_nat ma k)
    (RBT.insert Arith.linorder_nat)
    (fn k => fn ma => RBT.lookup Arith.linorder_nat ma k)
    (RBT.insert Arith.linorder_nat) (RBT.delete Arith.linorder_nat)
    (RBTMapImpl.iteratei_map_op_list_it_rm_ops Arith.linorder_nat)
    (RBTSetImpl.ins_rm_basic_ops Arith.linorder_nat)
    (RBTSetImpl.delete_rm_basic_ops Arith.linorder_nat)
    (sc_rnd_J_start_state seed)
    (State_Refinement.start_state_refine (Arith.equal_nat, MM.addr_nat)
      (CollectionsV1.equal_rbt (Arith.equal_nat, Arith.linorder_nat)
        SC_Collections.equal_heapobj)
      (Product_Type.equal_prod (List.equal_list Stringa.equal_literal)
        (Expr.equal_exp Stringa.equal_literal Product_Type.equal_unit
          Arith.equal_nat))
      (fn x => x) (RBT.empty Arith.linorder_nat) (SC_Collections.sc_allocate p)
      (RBT.empty Arith.linorder_nat) (RBT.insert Arith.linorder_nat)
      (RBT.empty Arith.linorder_nat)
      (RBTSetImpl.empty_rm_basic_ops Arith.linorder_nat ())
      (fn ca => fn _ => fn ts => fn _ => fn (pns, body) => fn vsa =>
        (Expr.blocks ("this" :: pns) (Type.Class ca :: ts) (Value.Null :: vsa)
           body,
          (fn _ => NONE)))
      p c m vs);

end; (*struct J_Execute*)

structure Code_Generation : sig
  val shr : 'a * (('b * 'c) * ('d * 'e)) -> 'c
  val thr : 'a * (('b * 'c) * ('d * 'e)) -> 'b
  val purge :
    (('a * ('b, 'c) Observable_Events.obs_event list), 'd) TLList.tllist ->
      Numeral_Type.num1 Numeral_Type.bit0 Numeral_Type.bit0 Numeral_Type.bit0
        Numeral_Type.bit0
        Numeral_Type.bit0
        Word.word
        Coinductive_List.llist
  val terminal : ('a, 'b) TLList.tllist -> 'b
  val j_Program :
    (string *
      (string *
        ((string * (Type.ty * unit TypeRel.fmod_ext)) list *
          (string *
            (Type.ty list *
              (Type.ty *
                (string list *
                  (string, unit, Arith.nat) Expr.exp) option))) list))) list ->
      (string list * (string, unit, Arith.nat) Expr.exp) TypeRel.prog
  val llist_case :
    'a -> ('b -> 'b Coinductive_List.llist -> 'a) ->
            'b Coinductive_List.llist -> 'a
  val tllist_case :
    ('a -> 'b) ->
      ('c -> ('c, 'a) TLList.tllist -> 'b) -> ('c, 'a) TLList.tllist -> 'b
  val toString_heapobja : SC_Collections.heapobj -> string
  val toString_heapobj : SC_Collections.heapobj ToString.toString
  val heap_toString : (Arith.nat, SC_Collections.heapobj) RBT.rbt -> string
  val trace_toString :
    Arith.nat * (Arith.nat, Arith.nat) Observable_Events.obs_event list ->
      string
  val llist_of_tllist : ('a, 'b) TLList.tllist -> 'a Coinductive_List.llist
  val thread_toString :
    (Arith.nat,
      (((string, unit, Arith.nat) Expr.exp *
         (string -> Arith.nat Value.vala option)) *
        (Arith.nat, Arith.nat) FinFun.finfun))
      RBT.rbt ->
      string
end = struct

fun shr x = FWState.shr x;

fun thr x = FWState.thr x;

fun purge run =
  Coinductive_List.lmap
    (fn Observable_Events.ExternalCall (_, _, Value.Intg i :: _, _) => i)
    (Coinductive_List.lfilter
      (fn a =>
        (case a of Observable_Events.ExternalCall (_, m, [], _) => false
          | Observable_Events.ExternalCall (_, m, Value.Unit :: lista, _) =>
            false
          | Observable_Events.ExternalCall (_, m, Value.Null :: lista, _) =>
            false
          | Observable_Events.ExternalCall (_, m, Value.Bool _ :: lista, _) =>
            false
          | Observable_Events.ExternalCall (_, m, [Value.Intg _], _) =>
            ((m : string) = "~print(I)V")
          | Observable_Events.ExternalCall (_, m, Value.Intg _ :: _ :: _, _) =>
            false
          | Observable_Events.ExternalCall (_, m, Value.Addr _ :: lista, _) =>
            false
          | Observable_Events.ReadMem (_, _, _) => false
          | Observable_Events.WriteMem (_, _, _) => false
          | Observable_Events.NewHeapElem (_, _) => false
          | Observable_Events.ThreadStart _ => false
          | Observable_Events.ThreadJoin _ => false
          | Observable_Events.SyncLock _ => false
          | Observable_Events.SyncUnlock _ => false
          | Observable_Events.ObsInterrupt _ => false
          | Observable_Events.ObsInterrupted _ => false))
      (Coinductive_List.lconcat
        (Coinductive_List.lmap (Coinductive_List.llist_of o Product_Type.snd)
          (TLList.llist_of_tllist run))));

fun terminal x = TLList.terminal x;

fun j_Program x = TypeRel.Program x;

fun llist_case x = Coinductive_List.llist_case x;

fun tllist_case x = TLList.tllist_case x;

fun toString_heapobja (SC_Collections.Obj (c, fs)) =
  String.concat
    ["(Obj ", ToString.toString_literala c, ", ",
      ToString.toString_trie ToString.toString_char
        (ToString.toString_assoc_list ToString.toString_literal
          (ToString.toString_val ToString.toString_nat))
        fs,
      ")"]
  | toString_heapobja (SC_Collections.Arr (t, si, fs, el)) =
    String.concat
      ["([", ToString.toString_nata si, "]", ToString.toString_ty t, ", ",
        ToString.toString_assoc_lista ToString.toString_literal
          (ToString.toString_val ToString.toString_nat) fs,
        ", ",
        ToString.toString_lista (ToString.toString_val ToString.toString_nat)
          (List.map Product_Type.snd
            (RBTMapImpl.g_to_list_rm_basic_ops Arith.linorder_nat el)),
        ")"];

val toString_heapobj = {toString = toString_heapobja} :
  SC_Collections.heapobj ToString.toString;

fun heap_toString x =
  ToString.toString_rbt (Arith.linorder_nat, ToString.toString_nat)
    toString_heapobj x;

fun trace_toString x =
  ToString.toString_proda ToString.toString_nat
    (ToString.toString_list
      (ToString.toString_obs_event ToString.toString_nat ToString.toString_nat))
    x;

fun llist_of_tllist x = TLList.llist_of_tllist x;

fun thread_toString x =
  ToString.toString_rbt (Arith.linorder_nat, ToString.toString_nat)
    (ToString.toString_prod
      (ToString.toString_prod
        (ToString.toString_exp ToString.toString_literal ToString.toString_unit
          ToString.toString_nat)
        ToString.toString_fun)
      (ToString.toString_finfun
        (Set.card_UNIV_nat, Arith.equal_nat, Arith.linorder_nat,
          ToString.toString_nat)
        (Arith.equal_nat, ToString.toString_nat)))
    x;

end; (*struct Code_Generation*)
