theory Sort_Min imports Tutorial_Basic begin

text \<open>In this theory, we illustrate very basic concepts of refinement\<close>

text \<open>Let's assume we have a type t, with natural number priorities\<close>
typedecl t
consts prio :: "t \<Rightarrow> nat"


text \<open>Sorting a list by priority can be specified as follows\<close>
definition sort_spec :: "t list \<Rightarrow> t list nres" where
  "sort_spec l \<equiv> SPEC (\<lambda>l'. multiset_of l = multiset_of l' \<and> sorted (map prio l'))"

text \<open>It can be implemented using standard list functions\<close>
lemma sort_impl: "RETURN (sort_key prio l) \<le> sort_spec l" 
  unfolding sort_spec_def
  by simp

text \<open>Next, we specify retrieval of a minimal-priority element\<close>
definition "min_spec l \<equiv> do {ASSERT (l\<noteq>[]); SPEC (\<lambda>x. x\<in>set l \<and> prio x = Min (prio`set l))}"
text \<open>Note that we use an assertion to express the precondition that the list is not empty\<close>

text \<open>An abstract algorithmic idea to retrieve the minimal-priority element is to
  sort the list, and return its head\<close>

definition "min_abs l \<equiv> do {
  ASSERT (l\<noteq>[]);
  l' \<leftarrow> sort_spec l;
  RETURN (hd l')
}"

text \<open>We prove that the abstract algorithm satisfies its specification.

  First, we require an auxiliary lemma:\<close>

lemma aux:
  fixes l 
  assumes S: "sorted (map f l)"
  assumes "l\<noteq>[]"  
  shows "f (hd l) = Min (f`set l)"
proof (cases l)
  case (Cons x xs)[simp]
  from S have "\<forall>v\<in>f`set xs. f x \<le> v" by (auto simp: sorted_Cons)
  hence "f x \<le> Min (f`set l)" by auto
  also have "f x \<ge> Min (f`set l)" by auto
  finally show ?thesis by simp
qed (simp add: `l\<noteq>[]`)

text \<open>With the auxiliary lemma, the proof is straightforward: \<close>
lemma min_abs_correct: "min_abs l \<le> min_spec l"
  unfolding min_abs_def min_spec_def sort_spec_def -- \<open>Unfolding definitions.\<close>
  apply (refine_vcg le_ASSERTI) -- \<open>VCG generates verification conditions ...\<close>

  txt \<open>... which are discharged using standard reasoning and the auxiliary lemma\<close>
  apply (clarsimp)
  apply (metis hd_in_set multiset_of_eq_setD multiset_of_zero_iff)

  apply (auto dest!: aux dest: multiset_of_eq_setD)
  done

text \<open>Next, we implement the algorithm, using @{const sort_key}\<close>
definition "min_impl l \<equiv> do {
  ASSERT (l\<noteq>[]);
  l' \<leftarrow> RETURN (sort_key prio l);
  RETURN (hd l')
}"

text \<open>We show that the implementation refines the abstract algorithm.
  First, we present a very straightforward proof, 
  exploiting monotonicity of bind.
\<close>
lemma "min_impl l \<le> min_abs l"
  unfolding min_impl_def min_abs_def
  apply (rule bind_mono)
  apply (rule order_refl)
  apply (rule bind_mono)
  apply (simp add: sort_impl)
  apply (rule order_refl)
  done

text \<open>There is also a verification condition generator for refinements,
  which automates the above task. For first, ignore the @{term "\<Down>Id"},
  which indicates a vacuous data refinement, and is required to trigger the vcg\<close>

lemmas sort_impl' = order_trans[OF sort_impl Id_refine]

lemma "min_impl l \<le> \<Down>Id (min_abs l)"
  unfolding min_impl_def min_abs_def
  apply (refine_rcg sort_impl')
  by auto

end

