
structure Uint : sig
  val set_bit : Word.word -> IntInf.int -> bool -> Word.word
  val shiftl : Word.word -> IntInf.int -> Word.word
  val shiftr : Word.word -> IntInf.int -> Word.word
  val shiftr_signed : Word.word -> IntInf.int -> Word.word
  val test_bit : Word.word -> IntInf.int -> bool
end = struct

fun set_bit x n b =
  let val mask = Word.<< (0wx1, Word.fromLargeInt (IntInf.toLarge n))
  in if b then Word.orb (x, mask)
     else Word.andb (x, Word.notb mask)
  end

fun shiftl x n =
  Word.<< (x, Word.fromLargeInt (IntInf.toLarge n))

fun shiftr x n =
  Word.>> (x, Word.fromLargeInt (IntInf.toLarge n))

fun shiftr_signed x n =
  Word.~>> (x, Word.fromLargeInt (IntInf.toLarge n))

fun test_bit x n =
  Word.andb (x, Word.<< (0wx1, Word.fromLargeInt (IntInf.toLarge n))) <> Word.fromInt 0

end; (* struct Uint *)

(* Test that words can handle numbers between 0 and 31 *)
val _ = if 5 <= Word.wordSize then () else raise (Fail ("wordSize less than 5"));

structure Uint32 : sig
  val set_bit : Word32.word -> IntInf.int -> bool -> Word32.word
  val shiftl : Word32.word -> IntInf.int -> Word32.word
  val shiftr : Word32.word -> IntInf.int -> Word32.word
  val shiftr_signed : Word32.word -> IntInf.int -> Word32.word
  val test_bit : Word32.word -> IntInf.int -> bool
end = struct

fun set_bit x n b =
  let val mask = Word32.<< (0wx1, Word.fromLargeInt (IntInf.toLarge n))
  in if b then Word32.orb (x, mask)
     else Word32.andb (x, Word32.notb mask)
  end

fun shiftl x n =
  Word32.<< (x, Word.fromLargeInt (IntInf.toLarge n))

fun shiftr x n =
  Word32.>> (x, Word.fromLargeInt (IntInf.toLarge n))

fun shiftr_signed x n =
  Word32.~>> (x, Word.fromLargeInt (IntInf.toLarge n))

fun test_bit x n =
  Word32.andb (x, Word32.<< (0wx1, Word.fromLargeInt (IntInf.toLarge n))) <> Word32.fromInt 0

end; (* struct Uint32 *)


structure STArray = struct

datatype 'a Cell = Invalid | Value of 'a array;

exception AccessedOldVersion;

type 'a array = 'a Cell Unsynchronized.ref;

fun fromList l = Unsynchronized.ref (Value (Array.fromList l));
fun array (size, v) = Unsynchronized.ref (Value (Array.array (size,v)));
fun tabulate (size, f) = Unsynchronized.ref (Value (Array.tabulate(size, f)));
fun sub (Unsynchronized.ref Invalid, idx) = raise AccessedOldVersion |
    sub (Unsynchronized.ref (Value a), idx) = Array.sub (a,idx);
fun update (aref,idx,v) =
  case aref of
    (Unsynchronized.ref Invalid) => raise AccessedOldVersion |
    (Unsynchronized.ref (Value a)) => (
      aref := Invalid;
      Array.update (a,idx,v);
      Unsynchronized.ref (Value a)
    );

fun length (Unsynchronized.ref Invalid) = raise AccessedOldVersion |
    length (Unsynchronized.ref (Value a)) = Array.length a

fun grow (aref, i, x) = case aref of 
  (Unsynchronized.ref Invalid) => raise AccessedOldVersion |
  (Unsynchronized.ref (Value a)) => (
    let val len=Array.length a;
        val na = Array.array (len+i,x) 
    in
      aref := Invalid;
      Array.copy {src=a, dst=na, di=0};
      Unsynchronized.ref (Value na)
    end
    );

fun shrink (aref, sz) = case aref of
  (Unsynchronized.ref Invalid) => raise AccessedOldVersion |
  (Unsynchronized.ref (Value a)) => (
    if sz > Array.length a then 
      raise Size
    else (
      aref:=Invalid;
      Unsynchronized.ref (Value (Array.tabulate (sz,fn i => Array.sub (a,i))))
    )
  );

structure IsabelleMapping = struct
type 'a ArrayType = 'a array;

fun new_array (a:'a) (n:int) = array (n, a);

fun array_length (a:'a ArrayType) = length a;

fun array_get (a:'a ArrayType) (i:int) = sub (a, i);

fun array_set (a:'a ArrayType) (i:int) (e:'a) = update (a, i, e);

fun array_of_list (xs:'a list) = fromList xs;

fun array_grow (a:'a ArrayType) (i:int) (x:'a) = grow (a, i, x);

fun array_shrink (a:'a ArrayType) (sz:int) = shrink (a,sz);

end;

end;

structure FArray = struct
  datatype 'a Cell = Value of 'a Array.array | Upd of (int*'a*'a Cell Unsynchronized.ref);

  type 'a array = 'a Cell Unsynchronized.ref;

  fun array (size,v) = Unsynchronized.ref (Value (Array.array (size,v)));
  fun tabulate (size, f) = Unsynchronized.ref (Value (Array.tabulate(size, f)));
  fun fromList l = Unsynchronized.ref (Value (Array.fromList l));

  fun sub (Unsynchronized.ref (Value a), idx) = Array.sub (a,idx) |
      sub (Unsynchronized.ref (Upd (i,v,cr)),idx) =
        if i=idx then v
        else sub (cr,idx);

  fun length (Unsynchronized.ref (Value a)) = Array.length a |
      length (Unsynchronized.ref (Upd (i,v,cr))) = length cr;

  fun realize_aux (aref, v) =
    case aref of
      (Unsynchronized.ref (Value a)) => (
        let
          val len = Array.length a;
          val a' = Array.array (len,v);
        in
          Array.copy {src=a, dst=a', di=0};
          Unsynchronized.ref (Value a')
        end
      ) |
      (Unsynchronized.ref (Upd (i,v,cr))) => (
        let val res=realize_aux (cr,v) in
          case res of
            (Unsynchronized.ref (Value a)) => (Array.update (a,i,v); res)
        end
      );

  fun realize aref =
    case aref of
      (Unsynchronized.ref (Value _)) => aref |
      (Unsynchronized.ref (Upd (i,v,cr))) => realize_aux(aref,v);

  fun update (aref,idx,v) =
    case aref of
      (Unsynchronized.ref (Value a)) => (
        let val nref=Unsynchronized.ref (Value a) in
          aref := Upd (idx,Array.sub(a,idx),nref);
          Array.update (a,idx,v);
          nref
        end
      ) |
      (Unsynchronized.ref (Upd _)) =>
        let val ra = realize_aux(aref,v) in
          case ra of
            (Unsynchronized.ref (Value a)) => Array.update (a,idx,v);
          ra
        end
      ;

  fun grow (aref, inc, x) = case aref of 
    (Unsynchronized.ref (Value a)) => (
      let val len=Array.length a;
          val na = Array.array (len+inc,x) 
      in
        Array.copy {src=a, dst=na, di=0};
        Unsynchronized.ref (Value na)
      end
      )
  | (Unsynchronized.ref (Upd _)) => (
    grow (realize aref, inc, x)
  );

  fun shrink (aref, sz) = case aref of
    (Unsynchronized.ref (Value a)) => (
      if sz > Array.length a then 
        raise Size
      else (
        Unsynchronized.ref (Value (Array.tabulate (sz,fn i => Array.sub (a,i))))
      )
    ) |
    (Unsynchronized.ref (Upd _)) => (
      shrink (realize aref,sz)
    );

structure IsabelleMapping = struct
type 'a ArrayType = 'a array;

fun new_array (a:'a) (n:int) = array (n, a);

fun array_length (a:'a ArrayType) = length a;

fun array_get (a:'a ArrayType) (i:int) = sub (a, i);

fun array_set (a:'a ArrayType) (i:int) (e:'a) = update (a, i, e);

fun array_of_list (xs:'a list) = fromList xs;

fun array_grow (a:'a ArrayType) (i:int) (x:'a) = grow (a, i, x);

fun array_shrink (a:'a ArrayType) (sz:int) = shrink (a,sz);

fun array_get_oo (d:'a) (a:'a ArrayType) (i:int) = 
  sub (a,i) handle Subscript => d

fun array_set_oo (d:(unit->'a ArrayType)) (a:'a ArrayType) (i:int) (e:'a) = 
  update (a, i, e) handle Subscript => d ()

end;
end;





  structure Statistics : sig
    type stat_entry = string * (unit -> bool) * (unit -> string)
  
    val register_stat : stat_entry -> unit
    val get_active_stats : unit -> (string * string) list
    val pretty_stats : (string * string) list -> string

  end = struct
    type stat_entry = string * (unit -> bool) * (unit -> string)
    val stats : stat_entry list Unsynchronized.ref = Unsynchronized.ref []
  
    fun register_stat e = stats := e :: !stats

    fun get_active_stats () = let
      fun flt [] = []
        | flt ((n,a,s)::l) = if a () then (n,s ()) :: flt l else flt l

    in flt (!stats)
    end

    fun pretty_stats [] = ""
      | pretty_stats ((n,s)::l) = "=== " ^ n ^ " ===\n" ^ s ^ "\n" ^ pretty_stats l
  end

(* Argh! Functors not compatible with ML_val command!
  functor Timer () : sig 
    val reset : unit -> unit
    val start : unit -> unit
    val stop : unit -> unit
    val set : Time.time -> unit
    val get : unit -> Time.time
    val pretty : unit -> string
  end = struct

    open Time;

    val time : Time.time Unsynchronized.ref = Unsynchronized.ref Time.zeroTime
    val running : bool Unsynchronized.ref = Unsynchronized.ref false
    val start_time : Time.time Unsynchronized.ref = Unsynchronized.ref Time.zeroTime
        
    fun reset () = (
      time := Time.zeroTime;
      running := false;
      start_time := Time.zeroTime
    )

    fun start () = 
      if !running then 
        () 
      else (
        running := true;
        start_time := Time.now ()
      )

    fun this_runs_time () = 
      if !running then 
        Time.now () - !start_time 
      else 
        Time.zeroTime

    fun stop () = (
      time := !time + this_runs_time ();
      running := false
    )

    fun get () = !time + this_runs_time ()
    fun set t = time := t - this_runs_time ()
  
    fun pretty () = Time.toString (!time) ^ "s"
  end
  *)



structure Bits_Integer : sig
  val set_bit : IntInf.int -> IntInf.int -> bool -> IntInf.int
  val shiftl : IntInf.int -> IntInf.int -> IntInf.int
  val shiftr : IntInf.int -> IntInf.int -> IntInf.int
  val test_bit : IntInf.int -> IntInf.int -> bool
end = struct

val maxWord = IntInf.pow (2, Word.wordSize);

fun set_bit x n b =
  if n < maxWord then
    if b then IntInf.orb (x, IntInf.<< (1, Word.fromLargeInt (IntInf.toLarge n)))
    else IntInf.andb (x, IntInf.notb (IntInf.<< (1, Word.fromLargeInt (IntInf.toLarge n))))
  else raise (Fail ("Bit index too large: " ^ IntInf.toString n));

fun shiftl x n =
  if n < maxWord then IntInf.<< (x, Word.fromLargeInt (IntInf.toLarge n))
  else raise (Fail ("Shift operand too large: " ^ IntInf.toString n));

fun shiftr x n =
  if n < maxWord then IntInf.~>> (x, Word.fromLargeInt (IntInf.toLarge n))
  else raise (Fail ("Shift operand too large: " ^ IntInf.toString n));

fun test_bit x n =
  if n < maxWord then IntInf.andb (x, IntInf.<< (1, Word.fromLargeInt (IntInf.toLarge n))) <> 0
  else raise (Fail ("Bit index too large: " ^ IntInf.toString n));

end; (*struct Bits_Integer*)


    structure PromelaUtils = struct
      exception UnsupportedConstruct of string
      exception StaticError of string
      exception RuntimeError of string
      fun warn msg = TextIO.print ("Warning: " ^ msg ^ "\n")
      fun usc  c   = raise (UnsupportedConstruct c)
      fun err  e   = raise (StaticError e)
      fun abort msg _ = raise (RuntimeError msg)
    end 


    structure Gerth_Statistics = struct
      val active = Unsynchronized.ref false
      val data = Unsynchronized.ref (0,0,0)
 
      fun is_active () = !active
      fun set_data num_states num_init num_acc = (
        active := true; 
        data := (num_states, num_init, num_acc)
      )

      fun to_string () = let
        val (num_states, num_init, num_acc) = !data
      in
          "Num states: " ^ IntInf.toString (num_states) ^ "\n"
        ^ "  Num initial: " ^ IntInf.toString num_init ^ "\n"
        ^ "  Num acc-classes: " ^ IntInf.toString num_acc ^ "\n"
      end
        
      val _ = Statistics.register_stat ("Gerth LTL_to_GBA",is_active,to_string)
    end



    structure PromelaStatistics = struct
      val active = Unsynchronized.ref false
      val parseTime = Unsynchronized.ref Time.zeroTime
      val time = Unsynchronized.ref Time.zeroTime

      fun is_active () = !active

      fun start () = (
            active := true; 
            if !parseTime = Time.zeroTime then () else parseTime := Time.- (Time.now (), !parseTime);
            time := Time.now ())
      fun start_parse () = (active := true; parseTime := Time.now())
      fun stop_timer () = (time := Time.- (Time.now (), !time))

      fun to_string () = let
        val t = Time.toMilliseconds (!time)
        val pt = Time.toMilliseconds (!parseTime)
      in
        (if pt = 0 then "" else 
        "Parsing time : " ^ IntInf.toString pt ^ "ms\n")
      ^ "Building time: " ^ IntInf.toString t ^ "ms\n"
      end
        
      val _ = Statistics.register_stat ("Promela",is_active,to_string)
    end



    structure NDFS_SI_Statistics = struct
      val active = Unsynchronized.ref false
      val cur_limit = Unsynchronized.ref 1000000
      val blue_vis = Unsynchronized.ref 0
      val blue_match = Unsynchronized.ref 0
      val red_vis = Unsynchronized.ref 0
      val time = Unsynchronized.ref Time.zeroTime

      fun is_active () = !active
      fun vis_blue () = (
            if !cur_limit < !blue_vis then (
              TextIO.print("*** " ^ IntInf.toString((!cur_limit) div 1000000) ^ "e+06 states\n");
              cur_limit := !cur_limit + 1000000)
            else ();
            blue_vis := !blue_vis + 1)
      fun match_blue () = blue_match := !blue_match + 1
      fun vis_red () = red_vis := !red_vis + 1

      fun start () = (active := true; time := Time.now ())
      fun stop () = (time := Time.- (Time.now (), !time))

      fun to_string () = let
        val t = Time.toReal (!time)
        val states_per_s = real (!blue_vis) / t
        val realStr = Real.fmt (StringCvt.FIX (SOME 2))

      in
        "Required time  : " ^ realStr t ^ "s\n"
      ^ "States per sec : " ^ realStr states_per_s ^ "\n"
      ^ "Blue states (v): " ^ IntInf.toString (!blue_vis) ^ "\n"
      ^ "Blue states (m): " ^ IntInf.toString (!blue_match) ^ "\n"
      ^ "Blue edges     : " ^ IntInf.toString (!blue_match + !blue_vis) ^ "\n"
      ^ "Red states     : " ^ IntInf.toString (!red_vis) ^ "\n"
      end
        
      val _ = Statistics.register_stat ("NDFS",is_active,to_string)
    end



    structure Gabow_Skeleton_Statistics = struct
      val active = Unsynchronized.ref false
      val num_vis = Unsynchronized.ref 0

      val time = Unsynchronized.ref Time.zeroTime

      fun is_active () = !active
      fun newnode () = num_vis := !num_vis + 1

      fun start () = (active := true; time := Time.now ())
      fun stop () = (time := Time.- (Time.now (), !time))

      fun to_string () = let
        val t = Time.toMilliseconds (!time)
        val states_per_ms = real (!num_vis) / real t
        val realStr = Real.fmt (StringCvt.FIX (SOME 2))
      in
        "Required time: " ^ IntInf.toString (t) ^ "ms\n"
      ^ "States per ms: " ^ realStr states_per_ms ^ "\n"
      ^ "# states: " ^ IntInf.toString (!num_vis) ^ "\n"
      end
        
      val _ = Statistics.register_stat ("Gabow-Skeleton",is_active,to_string)

    end


structure Test : sig
  type 'a ltlc
  type aprop
  type cmd
  type 'a local_state_impl_ext
  type ('a, 'b, 'c, 'd) pid_global_config_ext
  type 'a global_state_impl_ext
  type config_ce
  type ('a, 'b) lasso_ext
  datatype sm_result = TY_ERR | SAT | UNSAT |
    UNSAT_CE of
      ((cmd, unit local_state_impl_ext, unit global_state_impl_ext, unit)
         pid_global_config_ext,
        unit)
        lasso_ext
  type config_l2b
  type 'a program_ext
  val test : sm_result
end = struct

datatype int = Int_of_integer of IntInf.int;

datatype num = One | Bit0 of num | Bit1 of num;

val one_inta : int = Int_of_integer (1 : IntInf.int);

type 'a one = {one : 'a};
val one = #one : 'a one -> 'a;

val one_int = {one = one_inta} : int one;

fun integer_of_int (Int_of_integer k) = k;

fun times_inta k l =
  Int_of_integer (IntInf.* (integer_of_int k, integer_of_int l));

type 'a times = {times : 'a -> 'a -> 'a};
val times = #times : 'a times -> 'a -> 'a -> 'a;

type 'a power = {one_power : 'a one, times_power : 'a times};
val one_power = #one_power : 'a power -> 'a one;
val times_power = #times_power : 'a power -> 'a times;

val times_int = {times = times_inta} : int times;

val power_int = {one_power = one_int, times_power = times_int} : int power;

datatype nat = Nat of IntInf.int;

fun integer_of_nat (Nat x) = x;

fun equal_nata m n = (((integer_of_nat m) : IntInf.int) = (integer_of_nat n));

type 'a equal = {equal : 'a -> 'a -> bool};
val equal = #equal : 'a equal -> 'a -> 'a -> bool;

val equal_nat = {equal = equal_nata} : nat equal;

fun times_nata m n = Nat (IntInf.* (integer_of_nat m, integer_of_nat n));

type 'a dvd = {times_dvd : 'a times};
val times_dvd = #times_dvd : 'a dvd -> 'a times;

val times_nat = {times = times_nata} : nat times;

val dvd_nat = {times_dvd = times_nat} : nat dvd;

val one_nata : nat = Nat (1 : IntInf.int);

val one_nat = {one = one_nata} : nat one;

fun sgn_integer k =
  (if ((k : IntInf.int) = 0) then 0
    else (if IntInf.< (k, 0) then (~1 : IntInf.int) else (1 : IntInf.int)));

fun abs_integer k = (if IntInf.< (k, 0) then IntInf.~ k else k);

fun apsnd f (x, y) = (x, f y);

fun divmod_integer k l =
  (if ((k : IntInf.int) = 0) then (0, 0)
    else (if ((l : IntInf.int) = 0) then (0, k)
           else (apsnd o (fn a => fn b => IntInf.* (a, b)) o sgn_integer) l
                  (if (((sgn_integer k) : IntInf.int) = (sgn_integer l))
                    then IntInf.divMod (IntInf.abs k, IntInf.abs l)
                    else let
                           val (r, s) =
                             IntInf.divMod (IntInf.abs k, IntInf.abs l);
                         in
                           (if ((s : IntInf.int) = 0) then (IntInf.~ r, 0)
                             else (IntInf.- (IntInf.~ r, (1 : IntInf.int)),
                                    IntInf.- (abs_integer l, s)))
                         end)));

fun snd (x1, x2) = x2;

fun mod_integer k l = snd (divmod_integer k l);

fun mod_nat m n = Nat (mod_integer (integer_of_nat m) (integer_of_nat n));

fun fst (x1, x2) = x1;

fun div_integer k l = fst (divmod_integer k l);

fun div_nata m n = Nat (div_integer (integer_of_nat m) (integer_of_nat n));

type 'a diva = {dvd_div : 'a dvd, diva : 'a -> 'a -> 'a, moda : 'a -> 'a -> 'a};
val dvd_div = #dvd_div : 'a diva -> 'a dvd;
val diva = #diva : 'a diva -> 'a -> 'a -> 'a;
val moda = #moda : 'a diva -> 'a -> 'a -> 'a;

val div_nat = {dvd_div = dvd_nat, diva = div_nata, moda = mod_nat} : nat diva;

fun plus_nata m n = Nat (IntInf.+ (integer_of_nat m, integer_of_nat n));

type 'a plus = {plus : 'a -> 'a -> 'a};
val plus = #plus : 'a plus -> 'a -> 'a -> 'a;

val plus_nat = {plus = plus_nata} : nat plus;

val zero_nata : nat = Nat 0;

type 'a zero = {zero : 'a};
val zero = #zero : 'a zero -> 'a;

val zero_nat = {zero = zero_nata} : nat zero;

type 'a semigroup_add = {plus_semigroup_add : 'a plus};
val plus_semigroup_add = #plus_semigroup_add : 'a semigroup_add -> 'a plus;

type 'a numeral =
  {one_numeral : 'a one, semigroup_add_numeral : 'a semigroup_add};
val one_numeral = #one_numeral : 'a numeral -> 'a one;
val semigroup_add_numeral = #semigroup_add_numeral :
  'a numeral -> 'a semigroup_add;

val semigroup_add_nat = {plus_semigroup_add = plus_nat} : nat semigroup_add;

val numeral_nat =
  {one_numeral = one_nat, semigroup_add_numeral = semigroup_add_nat} :
  nat numeral;

fun less_eq_nat m n = IntInf.<= (integer_of_nat m, integer_of_nat n);

type 'a ord = {less_eq : 'a -> 'a -> bool, less : 'a -> 'a -> bool};
val less_eq = #less_eq : 'a ord -> 'a -> 'a -> bool;
val less = #less : 'a ord -> 'a -> 'a -> bool;

fun less_nat m n = IntInf.< (integer_of_nat m, integer_of_nat n);

val ord_nat = {less_eq = less_eq_nat, less = less_nat} : nat ord;

fun max A_ a b = (if less_eq A_ a b then b else a);

val ord_integer =
  {less_eq = (fn a => fn b => IntInf.<= (a, b)),
    less = (fn a => fn b => IntInf.< (a, b))}
  : IntInf.int ord;

fun nat_of_integer k = Nat (max ord_integer 0 k);

datatype 'a itself = Type;

fun def_hashmap_size_nat x = (fn _ => nat_of_integer (16 : IntInf.int)) x;

type 'a hashable =
  {hashcode : 'a -> Word32.word, def_hashmap_size : 'a itself -> nat};
val hashcode = #hashcode : 'a hashable -> 'a -> Word32.word;
val def_hashmap_size = #def_hashmap_size : 'a hashable -> 'a itself -> nat;

fun int_of_nat n = Int_of_integer (integer_of_nat n);

fun uint32_of_int i = Word32.fromInt (integer_of_int i);

fun hashcode_nat n = uint32_of_int (int_of_nat n);

val hashable_nat =
  {hashcode = hashcode_nat, def_hashmap_size = def_hashmap_size_nat} :
  nat hashable;

datatype 'a set = Set of 'a list | Coset of 'a list;

fun filter p [] = []
  | filter p (x :: xs) = (if p x then x :: filter p xs else filter p xs);

fun eq A_ a b = equal A_ a b;

fun removeAll A_ x [] = []
  | removeAll A_ x (y :: xs) =
    (if eq A_ x y then removeAll A_ x xs else y :: removeAll A_ x xs);

fun membera A_ [] y = false
  | membera A_ (x :: xs) y = eq A_ x y orelse membera A_ xs y;

fun insertb A_ x xs = (if membera A_ xs x then xs else x :: xs);

fun remove A_ x (Coset xs) = Coset (insertb A_ x xs)
  | remove A_ x (Set xs) = Set (removeAll A_ x xs);

fun member A_ x (Coset xs) = not (membera A_ xs x)
  | member A_ x (Set xs) = membera A_ xs x;

fun fold f (x :: xs) s = fold f xs (f x s)
  | fold f [] s = s;

fun inf_setb A_ a (Coset xs) = fold (remove A_) xs a
  | inf_setb A_ a (Set xs) = Set (filter (fn x => member A_ x a) xs);

type 'a inf = {inf : 'a -> 'a -> 'a};
val inf = #inf : 'a inf -> 'a -> 'a -> 'a;

fun inf_set A_ = {inf = inf_setb A_} : 'a set inf;

fun inserta A_ x (Coset xs) = Coset (removeAll A_ x xs)
  | inserta A_ x (Set xs) = Set (insertb A_ x xs);

fun sup_setb A_ (Coset xs) a = Coset (filter (fn x => not (member A_ x a)) xs)
  | sup_setb A_ (Set xs) a = fold (inserta A_) xs a;

type 'a sup = {sup : 'a -> 'a -> 'a};
val sup = #sup : 'a sup -> 'a -> 'a -> 'a;

fun sup_set A_ = {sup = sup_setb A_} : 'a set sup;

val bot_seta : 'a set = Set [];

type 'a bot = {bot : 'a};
val bot = #bot : 'a bot -> 'a;

val bot_set = {bot = bot_seta} : 'a set bot;

datatype nibble = Nibble0 | Nibble1 | Nibble2 | Nibble3 | Nibble4 | Nibble5 |
  Nibble6 | Nibble7 | Nibble8 | Nibble9 | NibbleA | NibbleB | NibbleC | NibbleD
  | NibbleE | NibbleF;

fun pred_list p [] = true
  | pred_list p (x :: xs) = p x andalso pred_list p xs;

fun null [] = true
  | null (x :: xs) = false;

fun less_eq_set A_ (Coset xs) (Set ys) =
  (if null xs andalso null ys then false
    else (raise Fail
           "subset_eq (List.coset _) (List.set _) requires type class instance card_UNIV")
           (fn _ => less_eq_set A_ (Coset xs) (Set ys)))
  | less_eq_set A_ a (Coset ys) = pred_list (fn y => not (member A_ y a)) ys
  | less_eq_set A_ (Set xs) b = pred_list (fn x => member A_ x b) xs;

fun less_set A_ a b = less_eq_set A_ a b andalso not (less_eq_set A_ b a);

fun ord_set A_ = {less_eq = less_eq_set A_, less = less_set A_} : 'a set ord;

val top_seta : 'a set = Coset [];

type 'a top = {top : 'a};
val top = #top : 'a top -> 'a;

val top_set = {top = top_seta} : 'a set top;

type 'a preorder = {ord_preorder : 'a ord};
val ord_preorder = #ord_preorder : 'a preorder -> 'a ord;

type 'a order = {preorder_order : 'a preorder};
val preorder_order = #preorder_order : 'a order -> 'a preorder;

fun preorder_set A_ = {ord_preorder = ord_set A_} : 'a set preorder;

fun order_set A_ = {preorder_order = preorder_set A_} : 'a set order;

type 'a semilattice_sup =
  {sup_semilattice_sup : 'a sup, order_semilattice_sup : 'a order};
val sup_semilattice_sup = #sup_semilattice_sup : 'a semilattice_sup -> 'a sup;
val order_semilattice_sup = #order_semilattice_sup :
  'a semilattice_sup -> 'a order;

type 'a semilattice_inf =
  {inf_semilattice_inf : 'a inf, order_semilattice_inf : 'a order};
val inf_semilattice_inf = #inf_semilattice_inf : 'a semilattice_inf -> 'a inf;
val order_semilattice_inf = #order_semilattice_inf :
  'a semilattice_inf -> 'a order;

type 'a lattice =
  {semilattice_inf_lattice : 'a semilattice_inf,
    semilattice_sup_lattice : 'a semilattice_sup};
val semilattice_inf_lattice = #semilattice_inf_lattice :
  'a lattice -> 'a semilattice_inf;
val semilattice_sup_lattice = #semilattice_sup_lattice :
  'a lattice -> 'a semilattice_sup;

fun semilattice_sup_set A_ =
  {sup_semilattice_sup = sup_set A_, order_semilattice_sup = order_set A_} :
  'a set semilattice_sup;

fun semilattice_inf_set A_ =
  {inf_semilattice_inf = inf_set A_, order_semilattice_inf = order_set A_} :
  'a set semilattice_inf;

fun lattice_set A_ =
  {semilattice_inf_lattice = semilattice_inf_set A_,
    semilattice_sup_lattice = semilattice_sup_set A_}
  : 'a set lattice;

type 'a order_bot = {bot_order_bot : 'a bot, order_order_bot : 'a order};
val bot_order_bot = #bot_order_bot : 'a order_bot -> 'a bot;
val order_order_bot = #order_order_bot : 'a order_bot -> 'a order;

fun order_bot_set A_ = {bot_order_bot = bot_set, order_order_bot = order_set A_}
  : 'a set order_bot;

type 'a order_top = {order_order_top : 'a order, top_order_top : 'a top};
val order_order_top = #order_order_top : 'a order_top -> 'a order;
val top_order_top = #top_order_top : 'a order_top -> 'a top;

fun order_top_set A_ = {order_order_top = order_set A_, top_order_top = top_set}
  : 'a set order_top;

fun inf_setc A_ (Set xs) = fold (inf_setb A_) xs top_seta;

type 'a infa = {infa : 'a set -> 'a};
val infa = #infa : 'a infa -> 'a set -> 'a;

fun inf_seta A_ = {infa = inf_setc A_} : 'a set infa;

fun sup_setc A_ (Set xs) = fold (sup_setb A_) xs bot_seta;

type 'a supa = {supa : 'a set -> 'a};
val supa = #supa : 'a supa -> 'a set -> 'a;

fun sup_seta A_ = {supa = sup_setc A_} : 'a set supa;

type 'a bounded_semilattice_inf_top =
  {semilattice_inf_bounded_semilattice_inf_top : 'a semilattice_inf,
    order_top_bounded_semilattice_inf_top : 'a order_top};
val semilattice_inf_bounded_semilattice_inf_top =
  #semilattice_inf_bounded_semilattice_inf_top :
  'a bounded_semilattice_inf_top -> 'a semilattice_inf;
val order_top_bounded_semilattice_inf_top =
  #order_top_bounded_semilattice_inf_top :
  'a bounded_semilattice_inf_top -> 'a order_top;

type 'a bounded_lattice_top =
  {bounded_semilattice_inf_top_bounded_lattice_top :
     'a bounded_semilattice_inf_top,
    lattice_bounded_lattice_top : 'a lattice};
val bounded_semilattice_inf_top_bounded_lattice_top =
  #bounded_semilattice_inf_top_bounded_lattice_top :
  'a bounded_lattice_top -> 'a bounded_semilattice_inf_top;
val lattice_bounded_lattice_top = #lattice_bounded_lattice_top :
  'a bounded_lattice_top -> 'a lattice;

type 'a bounded_semilattice_sup_bot =
  {semilattice_sup_bounded_semilattice_sup_bot : 'a semilattice_sup,
    order_bot_bounded_semilattice_sup_bot : 'a order_bot};
val semilattice_sup_bounded_semilattice_sup_bot =
  #semilattice_sup_bounded_semilattice_sup_bot :
  'a bounded_semilattice_sup_bot -> 'a semilattice_sup;
val order_bot_bounded_semilattice_sup_bot =
  #order_bot_bounded_semilattice_sup_bot :
  'a bounded_semilattice_sup_bot -> 'a order_bot;

type 'a bounded_lattice_bot =
  {bounded_semilattice_sup_bot_bounded_lattice_bot :
     'a bounded_semilattice_sup_bot,
    lattice_bounded_lattice_bot : 'a lattice};
val bounded_semilattice_sup_bot_bounded_lattice_bot =
  #bounded_semilattice_sup_bot_bounded_lattice_bot :
  'a bounded_lattice_bot -> 'a bounded_semilattice_sup_bot;
val lattice_bounded_lattice_bot = #lattice_bounded_lattice_bot :
  'a bounded_lattice_bot -> 'a lattice;

type 'a bounded_lattice =
  {bounded_lattice_bot_bounded_lattice : 'a bounded_lattice_bot,
    bounded_lattice_top_bounded_lattice : 'a bounded_lattice_top};
val bounded_lattice_bot_bounded_lattice = #bounded_lattice_bot_bounded_lattice :
  'a bounded_lattice -> 'a bounded_lattice_bot;
val bounded_lattice_top_bounded_lattice = #bounded_lattice_top_bounded_lattice :
  'a bounded_lattice -> 'a bounded_lattice_top;

fun bounded_semilattice_inf_top_set A_ =
  {semilattice_inf_bounded_semilattice_inf_top = semilattice_inf_set A_,
    order_top_bounded_semilattice_inf_top = order_top_set A_}
  : 'a set bounded_semilattice_inf_top;

fun bounded_lattice_top_set A_ =
  {bounded_semilattice_inf_top_bounded_lattice_top =
     bounded_semilattice_inf_top_set A_,
    lattice_bounded_lattice_top = lattice_set A_}
  : 'a set bounded_lattice_top;

fun bounded_semilattice_sup_bot_set A_ =
  {semilattice_sup_bounded_semilattice_sup_bot = semilattice_sup_set A_,
    order_bot_bounded_semilattice_sup_bot = order_bot_set A_}
  : 'a set bounded_semilattice_sup_bot;

fun bounded_lattice_bot_set A_ =
  {bounded_semilattice_sup_bot_bounded_lattice_bot =
     bounded_semilattice_sup_bot_set A_,
    lattice_bounded_lattice_bot = lattice_set A_}
  : 'a set bounded_lattice_bot;

fun bounded_lattice_set A_ =
  {bounded_lattice_bot_bounded_lattice = bounded_lattice_bot_set A_,
    bounded_lattice_top_bounded_lattice = bounded_lattice_top_set A_}
  : 'a set bounded_lattice;

type 'a ccpo = {sup_ccpo : 'a supa, order_ccpo : 'a order};
val sup_ccpo = #sup_ccpo : 'a ccpo -> 'a supa;
val order_ccpo = #order_ccpo : 'a ccpo -> 'a order;

fun ccpo_set A_ = {sup_ccpo = sup_seta A_, order_ccpo = order_set A_} :
  'a set ccpo;

type 'a complete_lattice =
  {inf_complete_lattice : 'a infa, ccpo_complete_lattice : 'a ccpo,
    bounded_lattice_complete_lattice : 'a bounded_lattice};
val inf_complete_lattice = #inf_complete_lattice :
  'a complete_lattice -> 'a infa;
val ccpo_complete_lattice = #ccpo_complete_lattice :
  'a complete_lattice -> 'a ccpo;
val bounded_lattice_complete_lattice = #bounded_lattice_complete_lattice :
  'a complete_lattice -> 'a bounded_lattice;

fun complete_lattice_set A_ =
  {inf_complete_lattice = inf_seta A_, ccpo_complete_lattice = ccpo_set A_,
    bounded_lattice_complete_lattice = bounded_lattice_set A_}
  : 'a set complete_lattice;

datatype 'a ltlc = LTLcTrue | LTLcFalse | LTLcProp of 'a | LTLcNeg of 'a ltlc |
  LTLcAnd of 'a ltlc * 'a ltlc | LTLcOr of 'a ltlc * 'a ltlc |
  LTLcImplies of 'a ltlc * 'a ltlc | LTLcIff of 'a ltlc * 'a ltlc |
  LTLcNext of 'a ltlc | LTLcFinal of 'a ltlc | LTLcGlobal of 'a ltlc |
  LTLcUntil of 'a ltlc * 'a ltlc | LTLcRelease of 'a ltlc * 'a ltlc;

fun equal_ltlca A_ (LTLcUntil (x121, x122)) (LTLcRelease (x131, x132)) = false
  | equal_ltlca A_ (LTLcRelease (x131, x132)) (LTLcUntil (x121, x122)) = false
  | equal_ltlca A_ (LTLcGlobal x11) (LTLcRelease (x131, x132)) = false
  | equal_ltlca A_ (LTLcRelease (x131, x132)) (LTLcGlobal x11) = false
  | equal_ltlca A_ (LTLcGlobal x11) (LTLcUntil (x121, x122)) = false
  | equal_ltlca A_ (LTLcUntil (x121, x122)) (LTLcGlobal x11) = false
  | equal_ltlca A_ (LTLcFinal x10) (LTLcRelease (x131, x132)) = false
  | equal_ltlca A_ (LTLcRelease (x131, x132)) (LTLcFinal x10) = false
  | equal_ltlca A_ (LTLcFinal x10) (LTLcUntil (x121, x122)) = false
  | equal_ltlca A_ (LTLcUntil (x121, x122)) (LTLcFinal x10) = false
  | equal_ltlca A_ (LTLcFinal x10) (LTLcGlobal x11) = false
  | equal_ltlca A_ (LTLcGlobal x11) (LTLcFinal x10) = false
  | equal_ltlca A_ (LTLcNext x9) (LTLcRelease (x131, x132)) = false
  | equal_ltlca A_ (LTLcRelease (x131, x132)) (LTLcNext x9) = false
  | equal_ltlca A_ (LTLcNext x9) (LTLcUntil (x121, x122)) = false
  | equal_ltlca A_ (LTLcUntil (x121, x122)) (LTLcNext x9) = false
  | equal_ltlca A_ (LTLcNext x9) (LTLcGlobal x11) = false
  | equal_ltlca A_ (LTLcGlobal x11) (LTLcNext x9) = false
  | equal_ltlca A_ (LTLcNext x9) (LTLcFinal x10) = false
  | equal_ltlca A_ (LTLcFinal x10) (LTLcNext x9) = false
  | equal_ltlca A_ (LTLcIff (x81, x82)) (LTLcRelease (x131, x132)) = false
  | equal_ltlca A_ (LTLcRelease (x131, x132)) (LTLcIff (x81, x82)) = false
  | equal_ltlca A_ (LTLcIff (x81, x82)) (LTLcUntil (x121, x122)) = false
  | equal_ltlca A_ (LTLcUntil (x121, x122)) (LTLcIff (x81, x82)) = false
  | equal_ltlca A_ (LTLcIff (x81, x82)) (LTLcGlobal x11) = false
  | equal_ltlca A_ (LTLcGlobal x11) (LTLcIff (x81, x82)) = false
  | equal_ltlca A_ (LTLcIff (x81, x82)) (LTLcFinal x10) = false
  | equal_ltlca A_ (LTLcFinal x10) (LTLcIff (x81, x82)) = false
  | equal_ltlca A_ (LTLcIff (x81, x82)) (LTLcNext x9) = false
  | equal_ltlca A_ (LTLcNext x9) (LTLcIff (x81, x82)) = false
  | equal_ltlca A_ (LTLcImplies (x71, x72)) (LTLcRelease (x131, x132)) = false
  | equal_ltlca A_ (LTLcRelease (x131, x132)) (LTLcImplies (x71, x72)) = false
  | equal_ltlca A_ (LTLcImplies (x71, x72)) (LTLcUntil (x121, x122)) = false
  | equal_ltlca A_ (LTLcUntil (x121, x122)) (LTLcImplies (x71, x72)) = false
  | equal_ltlca A_ (LTLcImplies (x71, x72)) (LTLcGlobal x11) = false
  | equal_ltlca A_ (LTLcGlobal x11) (LTLcImplies (x71, x72)) = false
  | equal_ltlca A_ (LTLcImplies (x71, x72)) (LTLcFinal x10) = false
  | equal_ltlca A_ (LTLcFinal x10) (LTLcImplies (x71, x72)) = false
  | equal_ltlca A_ (LTLcImplies (x71, x72)) (LTLcNext x9) = false
  | equal_ltlca A_ (LTLcNext x9) (LTLcImplies (x71, x72)) = false
  | equal_ltlca A_ (LTLcImplies (x71, x72)) (LTLcIff (x81, x82)) = false
  | equal_ltlca A_ (LTLcIff (x81, x82)) (LTLcImplies (x71, x72)) = false
  | equal_ltlca A_ (LTLcOr (x61, x62)) (LTLcRelease (x131, x132)) = false
  | equal_ltlca A_ (LTLcRelease (x131, x132)) (LTLcOr (x61, x62)) = false
  | equal_ltlca A_ (LTLcOr (x61, x62)) (LTLcUntil (x121, x122)) = false
  | equal_ltlca A_ (LTLcUntil (x121, x122)) (LTLcOr (x61, x62)) = false
  | equal_ltlca A_ (LTLcOr (x61, x62)) (LTLcGlobal x11) = false
  | equal_ltlca A_ (LTLcGlobal x11) (LTLcOr (x61, x62)) = false
  | equal_ltlca A_ (LTLcOr (x61, x62)) (LTLcFinal x10) = false
  | equal_ltlca A_ (LTLcFinal x10) (LTLcOr (x61, x62)) = false
  | equal_ltlca A_ (LTLcOr (x61, x62)) (LTLcNext x9) = false
  | equal_ltlca A_ (LTLcNext x9) (LTLcOr (x61, x62)) = false
  | equal_ltlca A_ (LTLcOr (x61, x62)) (LTLcIff (x81, x82)) = false
  | equal_ltlca A_ (LTLcIff (x81, x82)) (LTLcOr (x61, x62)) = false
  | equal_ltlca A_ (LTLcOr (x61, x62)) (LTLcImplies (x71, x72)) = false
  | equal_ltlca A_ (LTLcImplies (x71, x72)) (LTLcOr (x61, x62)) = false
  | equal_ltlca A_ (LTLcAnd (x51, x52)) (LTLcRelease (x131, x132)) = false
  | equal_ltlca A_ (LTLcRelease (x131, x132)) (LTLcAnd (x51, x52)) = false
  | equal_ltlca A_ (LTLcAnd (x51, x52)) (LTLcUntil (x121, x122)) = false
  | equal_ltlca A_ (LTLcUntil (x121, x122)) (LTLcAnd (x51, x52)) = false
  | equal_ltlca A_ (LTLcAnd (x51, x52)) (LTLcGlobal x11) = false
  | equal_ltlca A_ (LTLcGlobal x11) (LTLcAnd (x51, x52)) = false
  | equal_ltlca A_ (LTLcAnd (x51, x52)) (LTLcFinal x10) = false
  | equal_ltlca A_ (LTLcFinal x10) (LTLcAnd (x51, x52)) = false
  | equal_ltlca A_ (LTLcAnd (x51, x52)) (LTLcNext x9) = false
  | equal_ltlca A_ (LTLcNext x9) (LTLcAnd (x51, x52)) = false
  | equal_ltlca A_ (LTLcAnd (x51, x52)) (LTLcIff (x81, x82)) = false
  | equal_ltlca A_ (LTLcIff (x81, x82)) (LTLcAnd (x51, x52)) = false
  | equal_ltlca A_ (LTLcAnd (x51, x52)) (LTLcImplies (x71, x72)) = false
  | equal_ltlca A_ (LTLcImplies (x71, x72)) (LTLcAnd (x51, x52)) = false
  | equal_ltlca A_ (LTLcAnd (x51, x52)) (LTLcOr (x61, x62)) = false
  | equal_ltlca A_ (LTLcOr (x61, x62)) (LTLcAnd (x51, x52)) = false
  | equal_ltlca A_ (LTLcNeg x4) (LTLcRelease (x131, x132)) = false
  | equal_ltlca A_ (LTLcRelease (x131, x132)) (LTLcNeg x4) = false
  | equal_ltlca A_ (LTLcNeg x4) (LTLcUntil (x121, x122)) = false
  | equal_ltlca A_ (LTLcUntil (x121, x122)) (LTLcNeg x4) = false
  | equal_ltlca A_ (LTLcNeg x4) (LTLcGlobal x11) = false
  | equal_ltlca A_ (LTLcGlobal x11) (LTLcNeg x4) = false
  | equal_ltlca A_ (LTLcNeg x4) (LTLcFinal x10) = false
  | equal_ltlca A_ (LTLcFinal x10) (LTLcNeg x4) = false
  | equal_ltlca A_ (LTLcNeg x4) (LTLcNext x9) = false
  | equal_ltlca A_ (LTLcNext x9) (LTLcNeg x4) = false
  | equal_ltlca A_ (LTLcNeg x4) (LTLcIff (x81, x82)) = false
  | equal_ltlca A_ (LTLcIff (x81, x82)) (LTLcNeg x4) = false
  | equal_ltlca A_ (LTLcNeg x4) (LTLcImplies (x71, x72)) = false
  | equal_ltlca A_ (LTLcImplies (x71, x72)) (LTLcNeg x4) = false
  | equal_ltlca A_ (LTLcNeg x4) (LTLcOr (x61, x62)) = false
  | equal_ltlca A_ (LTLcOr (x61, x62)) (LTLcNeg x4) = false
  | equal_ltlca A_ (LTLcNeg x4) (LTLcAnd (x51, x52)) = false
  | equal_ltlca A_ (LTLcAnd (x51, x52)) (LTLcNeg x4) = false
  | equal_ltlca A_ (LTLcProp x3) (LTLcRelease (x131, x132)) = false
  | equal_ltlca A_ (LTLcRelease (x131, x132)) (LTLcProp x3) = false
  | equal_ltlca A_ (LTLcProp x3) (LTLcUntil (x121, x122)) = false
  | equal_ltlca A_ (LTLcUntil (x121, x122)) (LTLcProp x3) = false
  | equal_ltlca A_ (LTLcProp x3) (LTLcGlobal x11) = false
  | equal_ltlca A_ (LTLcGlobal x11) (LTLcProp x3) = false
  | equal_ltlca A_ (LTLcProp x3) (LTLcFinal x10) = false
  | equal_ltlca A_ (LTLcFinal x10) (LTLcProp x3) = false
  | equal_ltlca A_ (LTLcProp x3) (LTLcNext x9) = false
  | equal_ltlca A_ (LTLcNext x9) (LTLcProp x3) = false
  | equal_ltlca A_ (LTLcProp x3) (LTLcIff (x81, x82)) = false
  | equal_ltlca A_ (LTLcIff (x81, x82)) (LTLcProp x3) = false
  | equal_ltlca A_ (LTLcProp x3) (LTLcImplies (x71, x72)) = false
  | equal_ltlca A_ (LTLcImplies (x71, x72)) (LTLcProp x3) = false
  | equal_ltlca A_ (LTLcProp x3) (LTLcOr (x61, x62)) = false
  | equal_ltlca A_ (LTLcOr (x61, x62)) (LTLcProp x3) = false
  | equal_ltlca A_ (LTLcProp x3) (LTLcAnd (x51, x52)) = false
  | equal_ltlca A_ (LTLcAnd (x51, x52)) (LTLcProp x3) = false
  | equal_ltlca A_ (LTLcProp x3) (LTLcNeg x4) = false
  | equal_ltlca A_ (LTLcNeg x4) (LTLcProp x3) = false
  | equal_ltlca A_ LTLcFalse (LTLcRelease (x131, x132)) = false
  | equal_ltlca A_ (LTLcRelease (x131, x132)) LTLcFalse = false
  | equal_ltlca A_ LTLcFalse (LTLcUntil (x121, x122)) = false
  | equal_ltlca A_ (LTLcUntil (x121, x122)) LTLcFalse = false
  | equal_ltlca A_ LTLcFalse (LTLcGlobal x11) = false
  | equal_ltlca A_ (LTLcGlobal x11) LTLcFalse = false
  | equal_ltlca A_ LTLcFalse (LTLcFinal x10) = false
  | equal_ltlca A_ (LTLcFinal x10) LTLcFalse = false
  | equal_ltlca A_ LTLcFalse (LTLcNext x9) = false
  | equal_ltlca A_ (LTLcNext x9) LTLcFalse = false
  | equal_ltlca A_ LTLcFalse (LTLcIff (x81, x82)) = false
  | equal_ltlca A_ (LTLcIff (x81, x82)) LTLcFalse = false
  | equal_ltlca A_ LTLcFalse (LTLcImplies (x71, x72)) = false
  | equal_ltlca A_ (LTLcImplies (x71, x72)) LTLcFalse = false
  | equal_ltlca A_ LTLcFalse (LTLcOr (x61, x62)) = false
  | equal_ltlca A_ (LTLcOr (x61, x62)) LTLcFalse = false
  | equal_ltlca A_ LTLcFalse (LTLcAnd (x51, x52)) = false
  | equal_ltlca A_ (LTLcAnd (x51, x52)) LTLcFalse = false
  | equal_ltlca A_ LTLcFalse (LTLcNeg x4) = false
  | equal_ltlca A_ (LTLcNeg x4) LTLcFalse = false
  | equal_ltlca A_ LTLcFalse (LTLcProp x3) = false
  | equal_ltlca A_ (LTLcProp x3) LTLcFalse = false
  | equal_ltlca A_ LTLcTrue (LTLcRelease (x131, x132)) = false
  | equal_ltlca A_ (LTLcRelease (x131, x132)) LTLcTrue = false
  | equal_ltlca A_ LTLcTrue (LTLcUntil (x121, x122)) = false
  | equal_ltlca A_ (LTLcUntil (x121, x122)) LTLcTrue = false
  | equal_ltlca A_ LTLcTrue (LTLcGlobal x11) = false
  | equal_ltlca A_ (LTLcGlobal x11) LTLcTrue = false
  | equal_ltlca A_ LTLcTrue (LTLcFinal x10) = false
  | equal_ltlca A_ (LTLcFinal x10) LTLcTrue = false
  | equal_ltlca A_ LTLcTrue (LTLcNext x9) = false
  | equal_ltlca A_ (LTLcNext x9) LTLcTrue = false
  | equal_ltlca A_ LTLcTrue (LTLcIff (x81, x82)) = false
  | equal_ltlca A_ (LTLcIff (x81, x82)) LTLcTrue = false
  | equal_ltlca A_ LTLcTrue (LTLcImplies (x71, x72)) = false
  | equal_ltlca A_ (LTLcImplies (x71, x72)) LTLcTrue = false
  | equal_ltlca A_ LTLcTrue (LTLcOr (x61, x62)) = false
  | equal_ltlca A_ (LTLcOr (x61, x62)) LTLcTrue = false
  | equal_ltlca A_ LTLcTrue (LTLcAnd (x51, x52)) = false
  | equal_ltlca A_ (LTLcAnd (x51, x52)) LTLcTrue = false
  | equal_ltlca A_ LTLcTrue (LTLcNeg x4) = false
  | equal_ltlca A_ (LTLcNeg x4) LTLcTrue = false
  | equal_ltlca A_ LTLcTrue (LTLcProp x3) = false
  | equal_ltlca A_ (LTLcProp x3) LTLcTrue = false
  | equal_ltlca A_ LTLcTrue LTLcFalse = false
  | equal_ltlca A_ LTLcFalse LTLcTrue = false
  | equal_ltlca A_ (LTLcRelease (x131, x132)) (LTLcRelease (y131, y132)) =
    equal_ltlca A_ x131 y131 andalso equal_ltlca A_ x132 y132
  | equal_ltlca A_ (LTLcUntil (x121, x122)) (LTLcUntil (y121, y122)) =
    equal_ltlca A_ x121 y121 andalso equal_ltlca A_ x122 y122
  | equal_ltlca A_ (LTLcGlobal x11) (LTLcGlobal y11) = equal_ltlca A_ x11 y11
  | equal_ltlca A_ (LTLcFinal x10) (LTLcFinal y10) = equal_ltlca A_ x10 y10
  | equal_ltlca A_ (LTLcNext x9) (LTLcNext y9) = equal_ltlca A_ x9 y9
  | equal_ltlca A_ (LTLcIff (x81, x82)) (LTLcIff (y81, y82)) =
    equal_ltlca A_ x81 y81 andalso equal_ltlca A_ x82 y82
  | equal_ltlca A_ (LTLcImplies (x71, x72)) (LTLcImplies (y71, y72)) =
    equal_ltlca A_ x71 y71 andalso equal_ltlca A_ x72 y72
  | equal_ltlca A_ (LTLcOr (x61, x62)) (LTLcOr (y61, y62)) =
    equal_ltlca A_ x61 y61 andalso equal_ltlca A_ x62 y62
  | equal_ltlca A_ (LTLcAnd (x51, x52)) (LTLcAnd (y51, y52)) =
    equal_ltlca A_ x51 y51 andalso equal_ltlca A_ x52 y52
  | equal_ltlca A_ (LTLcNeg x4) (LTLcNeg y4) = equal_ltlca A_ x4 y4
  | equal_ltlca A_ (LTLcProp x3) (LTLcProp y3) = eq A_ x3 y3
  | equal_ltlca A_ LTLcFalse LTLcFalse = true
  | equal_ltlca A_ LTLcTrue LTLcTrue = true;

fun equal_ltlc A_ = {equal = equal_ltlca A_} : 'a ltlc equal;

datatype 'a ltln = LTLnTrue | LTLnFalse | LTLnProp of 'a | LTLnNProp of 'a |
  LTLnAnd of 'a ltln * 'a ltln | LTLnOr of 'a ltln * 'a ltln |
  LTLnNext of 'a ltln | LTLnUntil of 'a ltln * 'a ltln |
  LTLnRelease of 'a ltln * 'a ltln;

fun equal_ltlna A_ (LTLnUntil (ltln1a, ltln2a)) (LTLnRelease (ltln1, ltln2)) =
  false
  | equal_ltlna A_ (LTLnRelease (ltln1a, ltln2a)) (LTLnUntil (ltln1, ltln2)) =
    false
  | equal_ltlna A_ (LTLnNext ltln) (LTLnRelease (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnRelease (ltln1, ltln2)) (LTLnNext ltln) = false
  | equal_ltlna A_ (LTLnNext ltln) (LTLnUntil (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnUntil (ltln1, ltln2)) (LTLnNext ltln) = false
  | equal_ltlna A_ (LTLnOr (ltln1a, ltln2a)) (LTLnRelease (ltln1, ltln2)) =
    false
  | equal_ltlna A_ (LTLnRelease (ltln1a, ltln2a)) (LTLnOr (ltln1, ltln2)) =
    false
  | equal_ltlna A_ (LTLnOr (ltln1a, ltln2a)) (LTLnUntil (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnUntil (ltln1a, ltln2a)) (LTLnOr (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnOr (ltln1, ltln2)) (LTLnNext ltln) = false
  | equal_ltlna A_ (LTLnNext ltln) (LTLnOr (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnAnd (ltln1a, ltln2a)) (LTLnRelease (ltln1, ltln2)) =
    false
  | equal_ltlna A_ (LTLnRelease (ltln1a, ltln2a)) (LTLnAnd (ltln1, ltln2)) =
    false
  | equal_ltlna A_ (LTLnAnd (ltln1a, ltln2a)) (LTLnUntil (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnUntil (ltln1a, ltln2a)) (LTLnAnd (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnAnd (ltln1, ltln2)) (LTLnNext ltln) = false
  | equal_ltlna A_ (LTLnNext ltln) (LTLnAnd (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnAnd (ltln1a, ltln2a)) (LTLnOr (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnOr (ltln1a, ltln2a)) (LTLnAnd (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnNProp a) (LTLnRelease (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnRelease (ltln1, ltln2)) (LTLnNProp a) = false
  | equal_ltlna A_ (LTLnNProp a) (LTLnUntil (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnUntil (ltln1, ltln2)) (LTLnNProp a) = false
  | equal_ltlna A_ (LTLnNProp a) (LTLnNext ltln) = false
  | equal_ltlna A_ (LTLnNext ltln) (LTLnNProp a) = false
  | equal_ltlna A_ (LTLnNProp a) (LTLnOr (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnOr (ltln1, ltln2)) (LTLnNProp a) = false
  | equal_ltlna A_ (LTLnNProp a) (LTLnAnd (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnAnd (ltln1, ltln2)) (LTLnNProp a) = false
  | equal_ltlna A_ (LTLnProp a) (LTLnRelease (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnRelease (ltln1, ltln2)) (LTLnProp a) = false
  | equal_ltlna A_ (LTLnProp a) (LTLnUntil (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnUntil (ltln1, ltln2)) (LTLnProp a) = false
  | equal_ltlna A_ (LTLnProp a) (LTLnNext ltln) = false
  | equal_ltlna A_ (LTLnNext ltln) (LTLnProp a) = false
  | equal_ltlna A_ (LTLnProp a) (LTLnOr (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnOr (ltln1, ltln2)) (LTLnProp a) = false
  | equal_ltlna A_ (LTLnProp a) (LTLnAnd (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnAnd (ltln1, ltln2)) (LTLnProp a) = false
  | equal_ltlna A_ (LTLnProp aa) (LTLnNProp a) = false
  | equal_ltlna A_ (LTLnNProp aa) (LTLnProp a) = false
  | equal_ltlna A_ LTLnFalse (LTLnRelease (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnRelease (ltln1, ltln2)) LTLnFalse = false
  | equal_ltlna A_ LTLnFalse (LTLnUntil (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnUntil (ltln1, ltln2)) LTLnFalse = false
  | equal_ltlna A_ LTLnFalse (LTLnNext ltln) = false
  | equal_ltlna A_ (LTLnNext ltln) LTLnFalse = false
  | equal_ltlna A_ LTLnFalse (LTLnOr (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnOr (ltln1, ltln2)) LTLnFalse = false
  | equal_ltlna A_ LTLnFalse (LTLnAnd (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnAnd (ltln1, ltln2)) LTLnFalse = false
  | equal_ltlna A_ LTLnFalse (LTLnNProp a) = false
  | equal_ltlna A_ (LTLnNProp a) LTLnFalse = false
  | equal_ltlna A_ LTLnFalse (LTLnProp a) = false
  | equal_ltlna A_ (LTLnProp a) LTLnFalse = false
  | equal_ltlna A_ LTLnTrue (LTLnRelease (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnRelease (ltln1, ltln2)) LTLnTrue = false
  | equal_ltlna A_ LTLnTrue (LTLnUntil (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnUntil (ltln1, ltln2)) LTLnTrue = false
  | equal_ltlna A_ LTLnTrue (LTLnNext ltln) = false
  | equal_ltlna A_ (LTLnNext ltln) LTLnTrue = false
  | equal_ltlna A_ LTLnTrue (LTLnOr (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnOr (ltln1, ltln2)) LTLnTrue = false
  | equal_ltlna A_ LTLnTrue (LTLnAnd (ltln1, ltln2)) = false
  | equal_ltlna A_ (LTLnAnd (ltln1, ltln2)) LTLnTrue = false
  | equal_ltlna A_ LTLnTrue (LTLnNProp a) = false
  | equal_ltlna A_ (LTLnNProp a) LTLnTrue = false
  | equal_ltlna A_ LTLnTrue (LTLnProp a) = false
  | equal_ltlna A_ (LTLnProp a) LTLnTrue = false
  | equal_ltlna A_ LTLnTrue LTLnFalse = false
  | equal_ltlna A_ LTLnFalse LTLnTrue = false
  | equal_ltlna A_ (LTLnRelease (ltln1a, ltln2a)) (LTLnRelease (ltln1, ltln2)) =
    equal_ltlna A_ ltln1a ltln1 andalso equal_ltlna A_ ltln2a ltln2
  | equal_ltlna A_ (LTLnUntil (ltln1a, ltln2a)) (LTLnUntil (ltln1, ltln2)) =
    equal_ltlna A_ ltln1a ltln1 andalso equal_ltlna A_ ltln2a ltln2
  | equal_ltlna A_ (LTLnNext ltlna) (LTLnNext ltln) = equal_ltlna A_ ltlna ltln
  | equal_ltlna A_ (LTLnOr (ltln1a, ltln2a)) (LTLnOr (ltln1, ltln2)) =
    equal_ltlna A_ ltln1a ltln1 andalso equal_ltlna A_ ltln2a ltln2
  | equal_ltlna A_ (LTLnAnd (ltln1a, ltln2a)) (LTLnAnd (ltln1, ltln2)) =
    equal_ltlna A_ ltln1a ltln1 andalso equal_ltlna A_ ltln2a ltln2
  | equal_ltlna A_ (LTLnNProp aa) (LTLnNProp a) = eq A_ aa a
  | equal_ltlna A_ (LTLnProp aa) (LTLnProp a) = eq A_ aa a
  | equal_ltlna A_ LTLnFalse LTLnFalse = true
  | equal_ltlna A_ LTLnTrue LTLnTrue = true;

fun equal_ltln A_ = {equal = equal_ltlna A_} : 'a ltln equal;

fun rec_ltln f1 f2 f3 f4 f5 f6 f7 f8 f9 LTLnTrue = f1
  | rec_ltln f1 f2 f3 f4 f5 f6 f7 f8 f9 LTLnFalse = f2
  | rec_ltln f1 f2 f3 f4 f5 f6 f7 f8 f9 (LTLnProp a) = f3 a
  | rec_ltln f1 f2 f3 f4 f5 f6 f7 f8 f9 (LTLnNProp a) = f4 a
  | rec_ltln f1 f2 f3 f4 f5 f6 f7 f8 f9 (LTLnAnd (ltln1, ltln2)) =
    f5 ltln1 ltln2 (rec_ltln f1 f2 f3 f4 f5 f6 f7 f8 f9 ltln1)
      (rec_ltln f1 f2 f3 f4 f5 f6 f7 f8 f9 ltln2)
  | rec_ltln f1 f2 f3 f4 f5 f6 f7 f8 f9 (LTLnOr (ltln1, ltln2)) =
    f6 ltln1 ltln2 (rec_ltln f1 f2 f3 f4 f5 f6 f7 f8 f9 ltln1)
      (rec_ltln f1 f2 f3 f4 f5 f6 f7 f8 f9 ltln2)
  | rec_ltln f1 f2 f3 f4 f5 f6 f7 f8 f9 (LTLnNext ltln) =
    f7 ltln (rec_ltln f1 f2 f3 f4 f5 f6 f7 f8 f9 ltln)
  | rec_ltln f1 f2 f3 f4 f5 f6 f7 f8 f9 (LTLnUntil (ltln1, ltln2)) =
    f8 ltln1 ltln2 (rec_ltln f1 f2 f3 f4 f5 f6 f7 f8 f9 ltln1)
      (rec_ltln f1 f2 f3 f4 f5 f6 f7 f8 f9 ltln2)
  | rec_ltln f1 f2 f3 f4 f5 f6 f7 f8 f9 (LTLnRelease (ltln1, ltln2)) =
    f9 ltln1 ltln2 (rec_ltln f1 f2 f3 f4 f5 f6 f7 f8 f9 ltln1)
      (rec_ltln f1 f2 f3 f4 f5 f6 f7 f8 f9 ltln2);

fun less_eq_ltln (A1_, A2_) =
  (fn x => fn y =>
    rec_ltln
      (fn a =>
        (case a of LTLnTrue => false | LTLnFalse => true | LTLnProp _ => true
          | LTLnNProp _ => true | LTLnAnd (_, _) => true | LTLnOr (_, _) => true
          | LTLnNext _ => true | LTLnUntil (_, _) => true
          | LTLnRelease (_, _) => true))
      (fn a =>
        (case a of LTLnTrue => false | LTLnFalse => false | LTLnProp _ => true
          | LTLnNProp _ => true | LTLnAnd (_, _) => true | LTLnOr (_, _) => true
          | LTLnNext _ => true | LTLnUntil (_, _) => true
          | LTLnRelease (_, _) => true))
      (fn x_0 => fn a =>
        (case a of LTLnTrue => false | LTLnFalse => false
          | LTLnProp aa => less A2_ x_0 aa | LTLnNProp _ => true
          | LTLnAnd (_, _) => true | LTLnOr (_, _) => true | LTLnNext _ => true
          | LTLnUntil (_, _) => true | LTLnRelease (_, _) => true))
      (fn x_0 => fn a =>
        (case a of LTLnTrue => false | LTLnFalse => false | LTLnProp _ => false
          | LTLnNProp aa => less A2_ x_0 aa | LTLnAnd (_, _) => true
          | LTLnOr (_, _) => true | LTLnNext _ => true
          | LTLnUntil (_, _) => true | LTLnRelease (_, _) => true))
      (fn x_0 => fn _ => fn res_0 => fn res_1 => fn a =>
        (case a of LTLnTrue => false | LTLnFalse => false | LTLnProp _ => false
          | LTLnNProp _ => false
          | LTLnAnd (y_0, y_1) =>
            res_0 y_0 orelse equal_ltlna A1_ x_0 y_0 andalso res_1 y_1
          | LTLnOr (_, _) => true | LTLnNext _ => true
          | LTLnUntil (_, _) => true | LTLnRelease (_, _) => true))
      (fn x_0 => fn _ => fn res_0 => fn res_1 => fn a =>
        (case a of LTLnTrue => false | LTLnFalse => false | LTLnProp _ => false
          | LTLnNProp _ => false | LTLnAnd (_, _) => false
          | LTLnOr (y_0, y_1) =>
            res_0 y_0 orelse equal_ltlna A1_ x_0 y_0 andalso res_1 y_1
          | LTLnNext _ => true | LTLnUntil (_, _) => true
          | LTLnRelease (_, _) => true))
      (fn _ => fn res_0 => fn a =>
        (case a of LTLnTrue => false | LTLnFalse => false | LTLnProp _ => false
          | LTLnNProp _ => false | LTLnAnd (_, _) => false
          | LTLnOr (_, _) => false | LTLnNext aa => res_0 aa
          | LTLnUntil (_, _) => true | LTLnRelease (_, _) => true))
      (fn x_0 => fn _ => fn res_0 => fn res_1 => fn a =>
        (case a of LTLnTrue => false | LTLnFalse => false | LTLnProp _ => false
          | LTLnNProp _ => false | LTLnAnd (_, _) => false
          | LTLnOr (_, _) => false | LTLnNext _ => false
          | LTLnUntil (y_0, y_1) =>
            res_0 y_0 orelse equal_ltlna A1_ x_0 y_0 andalso res_1 y_1
          | LTLnRelease (_, _) => true))
      (fn x_0 => fn _ => fn res_0 => fn res_1 => fn a =>
        (case a of LTLnTrue => false | LTLnFalse => false | LTLnProp _ => false
          | LTLnNProp _ => false | LTLnAnd (_, _) => false
          | LTLnOr (_, _) => false | LTLnNext _ => false
          | LTLnUntil (_, _) => false
          | LTLnRelease (y_0, y_1) =>
            res_0 y_0 orelse equal_ltlna A1_ x_0 y_0 andalso res_1 y_1))
      x y orelse
      equal_ltlna A1_ x y);

fun less_ltln (A1_, A2_) =
  rec_ltln
    (fn a =>
      (case a of LTLnTrue => false | LTLnFalse => true | LTLnProp _ => true
        | LTLnNProp _ => true | LTLnAnd (_, _) => true | LTLnOr (_, _) => true
        | LTLnNext _ => true | LTLnUntil (_, _) => true
        | LTLnRelease (_, _) => true))
    (fn a =>
      (case a of LTLnTrue => false | LTLnFalse => false | LTLnProp _ => true
        | LTLnNProp _ => true | LTLnAnd (_, _) => true | LTLnOr (_, _) => true
        | LTLnNext _ => true | LTLnUntil (_, _) => true
        | LTLnRelease (_, _) => true))
    (fn x_0 => fn a =>
      (case a of LTLnTrue => false | LTLnFalse => false
        | LTLnProp aa => less A2_ x_0 aa | LTLnNProp _ => true
        | LTLnAnd (_, _) => true | LTLnOr (_, _) => true | LTLnNext _ => true
        | LTLnUntil (_, _) => true | LTLnRelease (_, _) => true))
    (fn x_0 => fn a =>
      (case a of LTLnTrue => false | LTLnFalse => false | LTLnProp _ => false
        | LTLnNProp aa => less A2_ x_0 aa | LTLnAnd (_, _) => true
        | LTLnOr (_, _) => true | LTLnNext _ => true | LTLnUntil (_, _) => true
        | LTLnRelease (_, _) => true))
    (fn x_0 => fn _ => fn res_0 => fn res_1 => fn a =>
      (case a of LTLnTrue => false | LTLnFalse => false | LTLnProp _ => false
        | LTLnNProp _ => false
        | LTLnAnd (y_0, y_1) =>
          res_0 y_0 orelse equal_ltlna A1_ x_0 y_0 andalso res_1 y_1
        | LTLnOr (_, _) => true | LTLnNext _ => true | LTLnUntil (_, _) => true
        | LTLnRelease (_, _) => true))
    (fn x_0 => fn _ => fn res_0 => fn res_1 => fn a =>
      (case a of LTLnTrue => false | LTLnFalse => false | LTLnProp _ => false
        | LTLnNProp _ => false | LTLnAnd (_, _) => false
        | LTLnOr (y_0, y_1) =>
          res_0 y_0 orelse equal_ltlna A1_ x_0 y_0 andalso res_1 y_1
        | LTLnNext _ => true | LTLnUntil (_, _) => true
        | LTLnRelease (_, _) => true))
    (fn _ => fn res_0 => fn a =>
      (case a of LTLnTrue => false | LTLnFalse => false | LTLnProp _ => false
        | LTLnNProp _ => false | LTLnAnd (_, _) => false
        | LTLnOr (_, _) => false | LTLnNext aa => res_0 aa
        | LTLnUntil (_, _) => true | LTLnRelease (_, _) => true))
    (fn x_0 => fn _ => fn res_0 => fn res_1 => fn a =>
      (case a of LTLnTrue => false | LTLnFalse => false | LTLnProp _ => false
        | LTLnNProp _ => false | LTLnAnd (_, _) => false
        | LTLnOr (_, _) => false | LTLnNext _ => false
        | LTLnUntil (y_0, y_1) =>
          res_0 y_0 orelse equal_ltlna A1_ x_0 y_0 andalso res_1 y_1
        | LTLnRelease (_, _) => true))
    (fn x_0 => fn _ => fn res_0 => fn res_1 => fn a =>
      (case a of LTLnTrue => false | LTLnFalse => false | LTLnProp _ => false
        | LTLnNProp _ => false | LTLnAnd (_, _) => false
        | LTLnOr (_, _) => false | LTLnNext _ => false
        | LTLnUntil (_, _) => false
        | LTLnRelease (y_0, y_1) =>
          res_0 y_0 orelse equal_ltlna A1_ x_0 y_0 andalso res_1 y_1));

fun ord_ltln (A1_, A2_) =
  {less_eq = less_eq_ltln (A1_, A2_), less = less_ltln (A1_, A2_)} :
  'a ltln ord;

fun preorder_ltln (A1_, A2_) =
  {ord_preorder = ord_ltln (A1_, (ord_preorder o preorder_order) A2_)} :
  'a ltln preorder;

fun order_ltln (A1_, A2_) = {preorder_order = preorder_ltln (A1_, A2_)} :
  'a ltln order;

type 'a linorder = {order_linorder : 'a order};
val order_linorder = #order_linorder : 'a linorder -> 'a order;

fun linorder_ltln (A1_, A2_) =
  {order_linorder = order_ltln (A1_, order_linorder A2_)} : 'a ltln linorder;

fun def_hashmap_size_char x = (fn _ => nat_of_integer (32 : IntInf.int)) x;

fun nat_of_char x = (nat_of_integer o (IntInf.fromInt o Char.ord)) x;

fun hashcode_char c = uint32_of_int (int_of_nat (nat_of_char c));

val hashable_char =
  {hashcode = hashcode_char, def_hashmap_size = def_hashmap_size_char} :
  char hashable;

datatype aprop = AP_isNull of string;

fun equal_apropa (AP_isNull literala) (AP_isNull literal) =
  ((literala : string) = literal);

val equal_aprop = {equal = equal_apropa} : aprop equal;

fun rec_aprop f1 (AP_isNull literal) = f1 literal;

fun less_eq_aprop x =
  (fn xa => fn y =>
    rec_aprop (fn x_0 => fn AP_isNull a => ((x_0 : string) < a)) xa y orelse
      equal_apropa xa y)
    x;

fun less_aprop x =
  rec_aprop (fn x_0 => fn AP_isNull a => ((x_0 : string) < a)) x;

val ord_aprop = {less_eq = less_eq_aprop, less = less_aprop} : aprop ord;

val preorder_aprop = {ord_preorder = ord_aprop} : aprop preorder;

val order_aprop = {preorder_order = preorder_aprop} : aprop order;

val linorder_aprop = {order_linorder = order_aprop} : aprop linorder;

datatype ('a, 'b) sum = Inl of 'a | Inr of 'b;

fun equal_suma A_ B_ (Inl x1) (Inr x2) = false
  | equal_suma A_ B_ (Inr x2) (Inl x1) = false
  | equal_suma A_ B_ (Inr x2) (Inr y2) = eq B_ x2 y2
  | equal_suma A_ B_ (Inl x1) (Inl y1) = eq A_ x1 y1;

fun equal_sum A_ B_ = {equal = equal_suma A_ B_} : ('a, 'b) sum equal;

datatype bin_op = Bo_plus | Bo_minus | Bo_mul | Bo_div | Bo_mod | Bo_less |
  Bo_less_eq | Bo_eq | Bo_and | Bo_or | Bo_xor;

fun equal_bin_op Bo_or Bo_xor = false
  | equal_bin_op Bo_xor Bo_or = false
  | equal_bin_op Bo_and Bo_xor = false
  | equal_bin_op Bo_xor Bo_and = false
  | equal_bin_op Bo_and Bo_or = false
  | equal_bin_op Bo_or Bo_and = false
  | equal_bin_op Bo_eq Bo_xor = false
  | equal_bin_op Bo_xor Bo_eq = false
  | equal_bin_op Bo_eq Bo_or = false
  | equal_bin_op Bo_or Bo_eq = false
  | equal_bin_op Bo_eq Bo_and = false
  | equal_bin_op Bo_and Bo_eq = false
  | equal_bin_op Bo_less_eq Bo_xor = false
  | equal_bin_op Bo_xor Bo_less_eq = false
  | equal_bin_op Bo_less_eq Bo_or = false
  | equal_bin_op Bo_or Bo_less_eq = false
  | equal_bin_op Bo_less_eq Bo_and = false
  | equal_bin_op Bo_and Bo_less_eq = false
  | equal_bin_op Bo_less_eq Bo_eq = false
  | equal_bin_op Bo_eq Bo_less_eq = false
  | equal_bin_op Bo_less Bo_xor = false
  | equal_bin_op Bo_xor Bo_less = false
  | equal_bin_op Bo_less Bo_or = false
  | equal_bin_op Bo_or Bo_less = false
  | equal_bin_op Bo_less Bo_and = false
  | equal_bin_op Bo_and Bo_less = false
  | equal_bin_op Bo_less Bo_eq = false
  | equal_bin_op Bo_eq Bo_less = false
  | equal_bin_op Bo_less Bo_less_eq = false
  | equal_bin_op Bo_less_eq Bo_less = false
  | equal_bin_op Bo_mod Bo_xor = false
  | equal_bin_op Bo_xor Bo_mod = false
  | equal_bin_op Bo_mod Bo_or = false
  | equal_bin_op Bo_or Bo_mod = false
  | equal_bin_op Bo_mod Bo_and = false
  | equal_bin_op Bo_and Bo_mod = false
  | equal_bin_op Bo_mod Bo_eq = false
  | equal_bin_op Bo_eq Bo_mod = false
  | equal_bin_op Bo_mod Bo_less_eq = false
  | equal_bin_op Bo_less_eq Bo_mod = false
  | equal_bin_op Bo_mod Bo_less = false
  | equal_bin_op Bo_less Bo_mod = false
  | equal_bin_op Bo_div Bo_xor = false
  | equal_bin_op Bo_xor Bo_div = false
  | equal_bin_op Bo_div Bo_or = false
  | equal_bin_op Bo_or Bo_div = false
  | equal_bin_op Bo_div Bo_and = false
  | equal_bin_op Bo_and Bo_div = false
  | equal_bin_op Bo_div Bo_eq = false
  | equal_bin_op Bo_eq Bo_div = false
  | equal_bin_op Bo_div Bo_less_eq = false
  | equal_bin_op Bo_less_eq Bo_div = false
  | equal_bin_op Bo_div Bo_less = false
  | equal_bin_op Bo_less Bo_div = false
  | equal_bin_op Bo_div Bo_mod = false
  | equal_bin_op Bo_mod Bo_div = false
  | equal_bin_op Bo_mul Bo_xor = false
  | equal_bin_op Bo_xor Bo_mul = false
  | equal_bin_op Bo_mul Bo_or = false
  | equal_bin_op Bo_or Bo_mul = false
  | equal_bin_op Bo_mul Bo_and = false
  | equal_bin_op Bo_and Bo_mul = false
  | equal_bin_op Bo_mul Bo_eq = false
  | equal_bin_op Bo_eq Bo_mul = false
  | equal_bin_op Bo_mul Bo_less_eq = false
  | equal_bin_op Bo_less_eq Bo_mul = false
  | equal_bin_op Bo_mul Bo_less = false
  | equal_bin_op Bo_less Bo_mul = false
  | equal_bin_op Bo_mul Bo_mod = false
  | equal_bin_op Bo_mod Bo_mul = false
  | equal_bin_op Bo_mul Bo_div = false
  | equal_bin_op Bo_div Bo_mul = false
  | equal_bin_op Bo_minus Bo_xor = false
  | equal_bin_op Bo_xor Bo_minus = false
  | equal_bin_op Bo_minus Bo_or = false
  | equal_bin_op Bo_or Bo_minus = false
  | equal_bin_op Bo_minus Bo_and = false
  | equal_bin_op Bo_and Bo_minus = false
  | equal_bin_op Bo_minus Bo_eq = false
  | equal_bin_op Bo_eq Bo_minus = false
  | equal_bin_op Bo_minus Bo_less_eq = false
  | equal_bin_op Bo_less_eq Bo_minus = false
  | equal_bin_op Bo_minus Bo_less = false
  | equal_bin_op Bo_less Bo_minus = false
  | equal_bin_op Bo_minus Bo_mod = false
  | equal_bin_op Bo_mod Bo_minus = false
  | equal_bin_op Bo_minus Bo_div = false
  | equal_bin_op Bo_div Bo_minus = false
  | equal_bin_op Bo_minus Bo_mul = false
  | equal_bin_op Bo_mul Bo_minus = false
  | equal_bin_op Bo_plus Bo_xor = false
  | equal_bin_op Bo_xor Bo_plus = false
  | equal_bin_op Bo_plus Bo_or = false
  | equal_bin_op Bo_or Bo_plus = false
  | equal_bin_op Bo_plus Bo_and = false
  | equal_bin_op Bo_and Bo_plus = false
  | equal_bin_op Bo_plus Bo_eq = false
  | equal_bin_op Bo_eq Bo_plus = false
  | equal_bin_op Bo_plus Bo_less_eq = false
  | equal_bin_op Bo_less_eq Bo_plus = false
  | equal_bin_op Bo_plus Bo_less = false
  | equal_bin_op Bo_less Bo_plus = false
  | equal_bin_op Bo_plus Bo_mod = false
  | equal_bin_op Bo_mod Bo_plus = false
  | equal_bin_op Bo_plus Bo_div = false
  | equal_bin_op Bo_div Bo_plus = false
  | equal_bin_op Bo_plus Bo_mul = false
  | equal_bin_op Bo_mul Bo_plus = false
  | equal_bin_op Bo_plus Bo_minus = false
  | equal_bin_op Bo_minus Bo_plus = false
  | equal_bin_op Bo_xor Bo_xor = true
  | equal_bin_op Bo_or Bo_or = true
  | equal_bin_op Bo_and Bo_and = true
  | equal_bin_op Bo_eq Bo_eq = true
  | equal_bin_op Bo_less_eq Bo_less_eq = true
  | equal_bin_op Bo_less Bo_less = true
  | equal_bin_op Bo_mod Bo_mod = true
  | equal_bin_op Bo_div Bo_div = true
  | equal_bin_op Bo_mul Bo_mul = true
  | equal_bin_op Bo_minus Bo_minus = true
  | equal_bin_op Bo_plus Bo_plus = true;

datatype un_op = Uo_minus | Uo_not;

fun equal_un_op Uo_minus Uo_not = false
  | equal_un_op Uo_not Uo_minus = false
  | equal_un_op Uo_not Uo_not = true
  | equal_un_op Uo_minus Uo_minus = true;

fun equal_int k l = (((integer_of_int k) : IntInf.int) = (integer_of_int l));

datatype exp = E_var of string | E_localvar of string | E_globalvar of string |
  E_const of int | E_bin of bin_op * exp * exp | E_un of un_op * exp;

fun equal_exp (E_bin (bin_op, exp1, exp2)) (E_un (un_op, exp)) = false
  | equal_exp (E_un (un_op, exp)) (E_bin (bin_op, exp1, exp2)) = false
  | equal_exp (E_const int) (E_un (un_op, exp)) = false
  | equal_exp (E_un (un_op, exp)) (E_const int) = false
  | equal_exp (E_const int) (E_bin (bin_op, exp1, exp2)) = false
  | equal_exp (E_bin (bin_op, exp1, exp2)) (E_const int) = false
  | equal_exp (E_globalvar literal) (E_un (un_op, exp)) = false
  | equal_exp (E_un (un_op, exp)) (E_globalvar literal) = false
  | equal_exp (E_globalvar literal) (E_bin (bin_op, exp1, exp2)) = false
  | equal_exp (E_bin (bin_op, exp1, exp2)) (E_globalvar literal) = false
  | equal_exp (E_globalvar literal) (E_const int) = false
  | equal_exp (E_const int) (E_globalvar literal) = false
  | equal_exp (E_localvar literal) (E_un (un_op, exp)) = false
  | equal_exp (E_un (un_op, exp)) (E_localvar literal) = false
  | equal_exp (E_localvar literal) (E_bin (bin_op, exp1, exp2)) = false
  | equal_exp (E_bin (bin_op, exp1, exp2)) (E_localvar literal) = false
  | equal_exp (E_localvar literal) (E_const int) = false
  | equal_exp (E_const int) (E_localvar literal) = false
  | equal_exp (E_localvar literala) (E_globalvar literal) = false
  | equal_exp (E_globalvar literala) (E_localvar literal) = false
  | equal_exp (E_var literal) (E_un (un_op, exp)) = false
  | equal_exp (E_un (un_op, exp)) (E_var literal) = false
  | equal_exp (E_var literal) (E_bin (bin_op, exp1, exp2)) = false
  | equal_exp (E_bin (bin_op, exp1, exp2)) (E_var literal) = false
  | equal_exp (E_var literal) (E_const int) = false
  | equal_exp (E_const int) (E_var literal) = false
  | equal_exp (E_var literala) (E_globalvar literal) = false
  | equal_exp (E_globalvar literala) (E_var literal) = false
  | equal_exp (E_var literala) (E_localvar literal) = false
  | equal_exp (E_localvar literala) (E_var literal) = false
  | equal_exp (E_un (un_opa, expa)) (E_un (un_op, exp)) =
    equal_un_op un_opa un_op andalso equal_exp expa exp
  | equal_exp (E_bin (bin_opa, exp1a, exp2a)) (E_bin (bin_op, exp1, exp2)) =
    equal_bin_op bin_opa bin_op andalso
      (equal_exp exp1a exp1 andalso equal_exp exp2a exp2)
  | equal_exp (E_const inta) (E_const int) = equal_int inta int
  | equal_exp (E_globalvar literala) (E_globalvar literal) =
    ((literala : string) = literal)
  | equal_exp (E_localvar literala) (E_localvar literal) =
    ((literala : string) = literal)
  | equal_exp (E_var literala) (E_var literal) =
    ((literala : string) = literal);

datatype action = AAssign of exp * string * exp |
  AAssign_local of exp * string * exp | AAssign_global of exp * string * exp |
  ATest of exp | ASkip;

fun equal_actiona (ATest x4) ASkip = false
  | equal_actiona ASkip (ATest x4) = false
  | equal_actiona (AAssign_global (x31, x32, x33)) ASkip = false
  | equal_actiona ASkip (AAssign_global (x31, x32, x33)) = false
  | equal_actiona (AAssign_global (x31, x32, x33)) (ATest x4) = false
  | equal_actiona (ATest x4) (AAssign_global (x31, x32, x33)) = false
  | equal_actiona (AAssign_local (x21, x22, x23)) ASkip = false
  | equal_actiona ASkip (AAssign_local (x21, x22, x23)) = false
  | equal_actiona (AAssign_local (x21, x22, x23)) (ATest x4) = false
  | equal_actiona (ATest x4) (AAssign_local (x21, x22, x23)) = false
  | equal_actiona (AAssign_local (x21, x22, x23))
    (AAssign_global (x31, x32, x33)) = false
  | equal_actiona (AAssign_global (x31, x32, x33))
    (AAssign_local (x21, x22, x23)) = false
  | equal_actiona (AAssign (x11, x12, x13)) ASkip = false
  | equal_actiona ASkip (AAssign (x11, x12, x13)) = false
  | equal_actiona (AAssign (x11, x12, x13)) (ATest x4) = false
  | equal_actiona (ATest x4) (AAssign (x11, x12, x13)) = false
  | equal_actiona (AAssign (x11, x12, x13)) (AAssign_global (x31, x32, x33)) =
    false
  | equal_actiona (AAssign_global (x31, x32, x33)) (AAssign (x11, x12, x13)) =
    false
  | equal_actiona (AAssign (x11, x12, x13)) (AAssign_local (x21, x22, x23)) =
    false
  | equal_actiona (AAssign_local (x21, x22, x23)) (AAssign (x11, x12, x13)) =
    false
  | equal_actiona (ATest x4) (ATest y4) = equal_exp x4 y4
  | equal_actiona (AAssign_global (x31, x32, x33))
    (AAssign_global (y31, y32, y33)) =
    equal_exp x31 y31 andalso (((x32 : string) = y32) andalso equal_exp x33 y33)
  | equal_actiona (AAssign_local (x21, x22, x23))
    (AAssign_local (y21, y22, y23)) =
    equal_exp x21 y21 andalso (((x22 : string) = y22) andalso equal_exp x23 y23)
  | equal_actiona (AAssign (x11, x12, x13)) (AAssign (y11, y12, y13)) =
    equal_exp x11 y11 andalso (((x12 : string) = y12) andalso equal_exp x13 y13)
  | equal_actiona ASkip ASkip = true;

val equal_action = {equal = equal_actiona} : action equal;

datatype cmd = Assign of exp * string * exp | Assign_local of exp * string * exp
  | Assign_global of exp * string * exp | Test of exp | Skip | Seq of cmd * cmd
  | Or of cmd * cmd | Loop of cmd | Break | Continue | Iterate of cmd * cmd |
  Invalid;

fun equal_cmda (Iterate (cmd1, cmd2)) Invalid = false
  | equal_cmda Invalid (Iterate (cmd1, cmd2)) = false
  | equal_cmda Continue Invalid = false
  | equal_cmda Invalid Continue = false
  | equal_cmda Continue (Iterate (cmd1, cmd2)) = false
  | equal_cmda (Iterate (cmd1, cmd2)) Continue = false
  | equal_cmda Break Invalid = false
  | equal_cmda Invalid Break = false
  | equal_cmda Break (Iterate (cmd1, cmd2)) = false
  | equal_cmda (Iterate (cmd1, cmd2)) Break = false
  | equal_cmda Break Continue = false
  | equal_cmda Continue Break = false
  | equal_cmda (Loop cmd) Invalid = false
  | equal_cmda Invalid (Loop cmd) = false
  | equal_cmda (Loop cmd) (Iterate (cmd1, cmd2)) = false
  | equal_cmda (Iterate (cmd1, cmd2)) (Loop cmd) = false
  | equal_cmda (Loop cmd) Continue = false
  | equal_cmda Continue (Loop cmd) = false
  | equal_cmda (Loop cmd) Break = false
  | equal_cmda Break (Loop cmd) = false
  | equal_cmda (Or (cmd1, cmd2)) Invalid = false
  | equal_cmda Invalid (Or (cmd1, cmd2)) = false
  | equal_cmda (Or (cmd1a, cmd2a)) (Iterate (cmd1, cmd2)) = false
  | equal_cmda (Iterate (cmd1a, cmd2a)) (Or (cmd1, cmd2)) = false
  | equal_cmda (Or (cmd1, cmd2)) Continue = false
  | equal_cmda Continue (Or (cmd1, cmd2)) = false
  | equal_cmda (Or (cmd1, cmd2)) Break = false
  | equal_cmda Break (Or (cmd1, cmd2)) = false
  | equal_cmda (Or (cmd1, cmd2)) (Loop cmd) = false
  | equal_cmda (Loop cmd) (Or (cmd1, cmd2)) = false
  | equal_cmda (Seq (cmd1, cmd2)) Invalid = false
  | equal_cmda Invalid (Seq (cmd1, cmd2)) = false
  | equal_cmda (Seq (cmd1a, cmd2a)) (Iterate (cmd1, cmd2)) = false
  | equal_cmda (Iterate (cmd1a, cmd2a)) (Seq (cmd1, cmd2)) = false
  | equal_cmda (Seq (cmd1, cmd2)) Continue = false
  | equal_cmda Continue (Seq (cmd1, cmd2)) = false
  | equal_cmda (Seq (cmd1, cmd2)) Break = false
  | equal_cmda Break (Seq (cmd1, cmd2)) = false
  | equal_cmda (Seq (cmd1, cmd2)) (Loop cmd) = false
  | equal_cmda (Loop cmd) (Seq (cmd1, cmd2)) = false
  | equal_cmda (Seq (cmd1a, cmd2a)) (Or (cmd1, cmd2)) = false
  | equal_cmda (Or (cmd1a, cmd2a)) (Seq (cmd1, cmd2)) = false
  | equal_cmda Skip Invalid = false
  | equal_cmda Invalid Skip = false
  | equal_cmda Skip (Iterate (cmd1, cmd2)) = false
  | equal_cmda (Iterate (cmd1, cmd2)) Skip = false
  | equal_cmda Skip Continue = false
  | equal_cmda Continue Skip = false
  | equal_cmda Skip Break = false
  | equal_cmda Break Skip = false
  | equal_cmda Skip (Loop cmd) = false
  | equal_cmda (Loop cmd) Skip = false
  | equal_cmda Skip (Or (cmd1, cmd2)) = false
  | equal_cmda (Or (cmd1, cmd2)) Skip = false
  | equal_cmda Skip (Seq (cmd1, cmd2)) = false
  | equal_cmda (Seq (cmd1, cmd2)) Skip = false
  | equal_cmda (Test exp) Invalid = false
  | equal_cmda Invalid (Test exp) = false
  | equal_cmda (Test exp) (Iterate (cmd1, cmd2)) = false
  | equal_cmda (Iterate (cmd1, cmd2)) (Test exp) = false
  | equal_cmda (Test exp) Continue = false
  | equal_cmda Continue (Test exp) = false
  | equal_cmda (Test exp) Break = false
  | equal_cmda Break (Test exp) = false
  | equal_cmda (Test exp) (Loop cmd) = false
  | equal_cmda (Loop cmd) (Test exp) = false
  | equal_cmda (Test exp) (Or (cmd1, cmd2)) = false
  | equal_cmda (Or (cmd1, cmd2)) (Test exp) = false
  | equal_cmda (Test exp) (Seq (cmd1, cmd2)) = false
  | equal_cmda (Seq (cmd1, cmd2)) (Test exp) = false
  | equal_cmda (Test exp) Skip = false
  | equal_cmda Skip (Test exp) = false
  | equal_cmda (Assign_global (exp1, literal, exp2)) Invalid = false
  | equal_cmda Invalid (Assign_global (exp1, literal, exp2)) = false
  | equal_cmda (Assign_global (exp1, literal, exp2)) (Iterate (cmd1, cmd2)) =
    false
  | equal_cmda (Iterate (cmd1, cmd2)) (Assign_global (exp1, literal, exp2)) =
    false
  | equal_cmda (Assign_global (exp1, literal, exp2)) Continue = false
  | equal_cmda Continue (Assign_global (exp1, literal, exp2)) = false
  | equal_cmda (Assign_global (exp1, literal, exp2)) Break = false
  | equal_cmda Break (Assign_global (exp1, literal, exp2)) = false
  | equal_cmda (Assign_global (exp1, literal, exp2)) (Loop cmd) = false
  | equal_cmda (Loop cmd) (Assign_global (exp1, literal, exp2)) = false
  | equal_cmda (Assign_global (exp1, literal, exp2)) (Or (cmd1, cmd2)) = false
  | equal_cmda (Or (cmd1, cmd2)) (Assign_global (exp1, literal, exp2)) = false
  | equal_cmda (Assign_global (exp1, literal, exp2)) (Seq (cmd1, cmd2)) = false
  | equal_cmda (Seq (cmd1, cmd2)) (Assign_global (exp1, literal, exp2)) = false
  | equal_cmda (Assign_global (exp1, literal, exp2)) Skip = false
  | equal_cmda Skip (Assign_global (exp1, literal, exp2)) = false
  | equal_cmda (Assign_global (exp1, literal, exp2)) (Test exp) = false
  | equal_cmda (Test exp) (Assign_global (exp1, literal, exp2)) = false
  | equal_cmda (Assign_local (exp1, literal, exp2)) Invalid = false
  | equal_cmda Invalid (Assign_local (exp1, literal, exp2)) = false
  | equal_cmda (Assign_local (exp1, literal, exp2)) (Iterate (cmd1, cmd2)) =
    false
  | equal_cmda (Iterate (cmd1, cmd2)) (Assign_local (exp1, literal, exp2)) =
    false
  | equal_cmda (Assign_local (exp1, literal, exp2)) Continue = false
  | equal_cmda Continue (Assign_local (exp1, literal, exp2)) = false
  | equal_cmda (Assign_local (exp1, literal, exp2)) Break = false
  | equal_cmda Break (Assign_local (exp1, literal, exp2)) = false
  | equal_cmda (Assign_local (exp1, literal, exp2)) (Loop cmd) = false
  | equal_cmda (Loop cmd) (Assign_local (exp1, literal, exp2)) = false
  | equal_cmda (Assign_local (exp1, literal, exp2)) (Or (cmd1, cmd2)) = false
  | equal_cmda (Or (cmd1, cmd2)) (Assign_local (exp1, literal, exp2)) = false
  | equal_cmda (Assign_local (exp1, literal, exp2)) (Seq (cmd1, cmd2)) = false
  | equal_cmda (Seq (cmd1, cmd2)) (Assign_local (exp1, literal, exp2)) = false
  | equal_cmda (Assign_local (exp1, literal, exp2)) Skip = false
  | equal_cmda Skip (Assign_local (exp1, literal, exp2)) = false
  | equal_cmda (Assign_local (exp1, literal, exp2)) (Test exp) = false
  | equal_cmda (Test exp) (Assign_local (exp1, literal, exp2)) = false
  | equal_cmda (Assign_local (exp1a, literala, exp2a))
    (Assign_global (exp1, literal, exp2)) = false
  | equal_cmda (Assign_global (exp1a, literala, exp2a))
    (Assign_local (exp1, literal, exp2)) = false
  | equal_cmda (Assign (exp1, literal, exp2)) Invalid = false
  | equal_cmda Invalid (Assign (exp1, literal, exp2)) = false
  | equal_cmda (Assign (exp1, literal, exp2)) (Iterate (cmd1, cmd2)) = false
  | equal_cmda (Iterate (cmd1, cmd2)) (Assign (exp1, literal, exp2)) = false
  | equal_cmda (Assign (exp1, literal, exp2)) Continue = false
  | equal_cmda Continue (Assign (exp1, literal, exp2)) = false
  | equal_cmda (Assign (exp1, literal, exp2)) Break = false
  | equal_cmda Break (Assign (exp1, literal, exp2)) = false
  | equal_cmda (Assign (exp1, literal, exp2)) (Loop cmd) = false
  | equal_cmda (Loop cmd) (Assign (exp1, literal, exp2)) = false
  | equal_cmda (Assign (exp1, literal, exp2)) (Or (cmd1, cmd2)) = false
  | equal_cmda (Or (cmd1, cmd2)) (Assign (exp1, literal, exp2)) = false
  | equal_cmda (Assign (exp1, literal, exp2)) (Seq (cmd1, cmd2)) = false
  | equal_cmda (Seq (cmd1, cmd2)) (Assign (exp1, literal, exp2)) = false
  | equal_cmda (Assign (exp1, literal, exp2)) Skip = false
  | equal_cmda Skip (Assign (exp1, literal, exp2)) = false
  | equal_cmda (Assign (exp1, literal, exp2)) (Test exp) = false
  | equal_cmda (Test exp) (Assign (exp1, literal, exp2)) = false
  | equal_cmda (Assign (exp1a, literala, exp2a))
    (Assign_global (exp1, literal, exp2)) = false
  | equal_cmda (Assign_global (exp1a, literala, exp2a))
    (Assign (exp1, literal, exp2)) = false
  | equal_cmda (Assign (exp1a, literala, exp2a))
    (Assign_local (exp1, literal, exp2)) = false
  | equal_cmda (Assign_local (exp1a, literala, exp2a))
    (Assign (exp1, literal, exp2)) = false
  | equal_cmda (Iterate (cmd1a, cmd2a)) (Iterate (cmd1, cmd2)) =
    equal_cmda cmd1a cmd1 andalso equal_cmda cmd2a cmd2
  | equal_cmda (Loop cmda) (Loop cmd) = equal_cmda cmda cmd
  | equal_cmda (Or (cmd1a, cmd2a)) (Or (cmd1, cmd2)) =
    equal_cmda cmd1a cmd1 andalso equal_cmda cmd2a cmd2
  | equal_cmda (Seq (cmd1a, cmd2a)) (Seq (cmd1, cmd2)) =
    equal_cmda cmd1a cmd1 andalso equal_cmda cmd2a cmd2
  | equal_cmda (Test expa) (Test exp) = equal_exp expa exp
  | equal_cmda (Assign_global (exp1a, literala, exp2a))
    (Assign_global (exp1, literal, exp2)) =
    equal_exp exp1a exp1 andalso
      (((literala : string) = literal) andalso equal_exp exp2a exp2)
  | equal_cmda (Assign_local (exp1a, literala, exp2a))
    (Assign_local (exp1, literal, exp2)) =
    equal_exp exp1a exp1 andalso
      (((literala : string) = literal) andalso equal_exp exp2a exp2)
  | equal_cmda (Assign (exp1a, literala, exp2a)) (Assign (exp1, literal, exp2))
    = equal_exp exp1a exp1 andalso
        (((literala : string) = literal) andalso equal_exp exp2a exp2)
  | equal_cmda Invalid Invalid = true
  | equal_cmda Continue Continue = true
  | equal_cmda Break Break = true
  | equal_cmda Skip Skip = true;

val equal_cmd = {equal = equal_cmda} : cmd equal;

fun def_hashmap_size_cmd x = (fn _ => nat_of_integer (10 : IntInf.int)) x;

fun foldl f a [] = a
  | foldl f a (x :: xs) = foldl f (f a x) xs;

fun hashcode_list A_ =
  foldl (fn h => fn x =>
          Word32.+ (Word32.* (h, Word32.fromInt
                                   (33 : IntInf.int)), hashcode A_ x))
    (Word32.fromInt (5381 : IntInf.int));

fun hashcode_literal s = hashcode_list hashable_char (String.explode s);

fun rec_bin_op f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 Bo_plus = f1
  | rec_bin_op f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 Bo_minus = f2
  | rec_bin_op f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 Bo_mul = f3
  | rec_bin_op f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 Bo_div = f4
  | rec_bin_op f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 Bo_mod = f5
  | rec_bin_op f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 Bo_less = f6
  | rec_bin_op f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 Bo_less_eq = f7
  | rec_bin_op f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 Bo_eq = f8
  | rec_bin_op f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 Bo_and = f9
  | rec_bin_op f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 Bo_or = f10
  | rec_bin_op f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 Bo_xor = f11;

fun hashcode_bin_op x =
  rec_bin_op (Word32.fromInt (980429261 : IntInf.int))
    (Word32.fromInt (2027524386 : IntInf.int))
    (Word32.fromInt (295356509 : IntInf.int))
    (Word32.fromInt (1946686921 : IntInf.int))
    (Word32.fromInt (1981935517 : IntInf.int))
    (Word32.fromInt (1888951127 : IntInf.int))
    (Word32.fromInt (503674811 : IntInf.int))
    (Word32.fromInt (1139324734 : IntInf.int))
    (Word32.fromInt (1494669842 : IntInf.int))
    (Word32.fromInt (1234494377 : IntInf.int))
    (Word32.fromInt (108976710 : IntInf.int)) x;

fun rec_un_op f1 f2 Uo_minus = f1
  | rec_un_op f1 f2 Uo_not = f2;

fun hashcode_un_op x =
  rec_un_op (Word32.fromInt (965085014 : IntInf.int))
    (Word32.fromInt (743012010 : IntInf.int)) x;

fun hashcode_int i = uint32_of_int i;

fun rec_exp f1 f2 f3 f4 f5 f6 (E_var literal) = f1 literal
  | rec_exp f1 f2 f3 f4 f5 f6 (E_localvar literal) = f2 literal
  | rec_exp f1 f2 f3 f4 f5 f6 (E_globalvar literal) = f3 literal
  | rec_exp f1 f2 f3 f4 f5 f6 (E_const int) = f4 int
  | rec_exp f1 f2 f3 f4 f5 f6 (E_bin (bin_op, exp1, exp2)) =
    f5 bin_op exp1 exp2 (rec_exp f1 f2 f3 f4 f5 f6 exp1)
      (rec_exp f1 f2 f3 f4 f5 f6 exp2)
  | rec_exp f1 f2 f3 f4 f5 f6 (E_un (un_op, exp)) =
    f6 un_op exp (rec_exp f1 f2 f3 f4 f5 f6 exp);

fun hashcode_exp x =
  rec_exp
    (fn x_0 =>
      Word32.+ (Word32.* (Word32.fromInt
                            (1722761561 : IntInf.int), hashcode_literal
                 x_0), Word32.fromInt (1729421928 : IntInf.int)))
    (fn x_0 =>
      Word32.+ (Word32.* (Word32.fromInt
                            (95816375 : IntInf.int), hashcode_literal
               x_0), Word32.fromInt (102476742 : IntInf.int)))
    (fn x_0 =>
      Word32.+ (Word32.* (Word32.fromInt
                            (1123140918 : IntInf.int), hashcode_literal
                 x_0), Word32.fromInt (1129801285 : IntInf.int)))
    (fn x_0 =>
      Word32.+ (Word32.* (Word32.fromInt
                            (1080875490 : IntInf.int), hashcode_int
                 x_0), Word32.fromInt (1087535857 : IntInf.int)))
    (fn x_0 => fn _ => fn _ => fn res_0 => fn res_1 =>
      Word32.+ (Word32.* (Word32.fromInt
                            (140914973 : IntInf.int), hashcode_bin_op
                x_0), Word32.+ (Word32.* (Word32.fromInt
    (134254606 : IntInf.int), res_0), Word32.+ (Word32.* (Word32.fromInt
                    (127594239 : IntInf.int), res_1), Word32.fromInt
                (147575340 : IntInf.int)))))
    (fn x_0 => fn _ => fn res_0 =>
      Word32.+ (Word32.* (Word32.fromInt
                            (324287792 : IntInf.int), hashcode_un_op
                x_0), Word32.+ (Word32.* (Word32.fromInt
    (317627425 : IntInf.int), res_0), Word32.fromInt (330948159 : IntInf.int))))
    x;

fun rec_cmd f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12
  (Assign (exp1, literal, exp2)) = f1 exp1 literal exp2
  | rec_cmd f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12
    (Assign_local (exp1, literal, exp2)) = f2 exp1 literal exp2
  | rec_cmd f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12
    (Assign_global (exp1, literal, exp2)) = f3 exp1 literal exp2
  | rec_cmd f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 (Test exp) = f4 exp
  | rec_cmd f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 Skip = f5
  | rec_cmd f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 (Seq (cmd1, cmd2)) =
    f6 cmd1 cmd2 (rec_cmd f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 cmd1)
      (rec_cmd f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 cmd2)
  | rec_cmd f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 (Or (cmd1, cmd2)) =
    f7 cmd1 cmd2 (rec_cmd f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 cmd1)
      (rec_cmd f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 cmd2)
  | rec_cmd f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 (Loop cmd) =
    f8 cmd (rec_cmd f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 cmd)
  | rec_cmd f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 Break = f9
  | rec_cmd f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 Continue = f10
  | rec_cmd f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 (Iterate (cmd1, cmd2)) =
    f11 cmd1 cmd2 (rec_cmd f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 cmd1)
      (rec_cmd f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 cmd2)
  | rec_cmd f1 f2 f3 f4 f5 f6 f7 f8 f9 f10 f11 f12 Invalid = f12;

fun hashcode_cmd x =
  rec_cmd
    (fn x_0 => fn x_1 => fn x_2 =>
      Word32.+ (Word32.* (Word32.fromInt
                            (1570866865 : IntInf.int), hashcode_exp
                 x_0), Word32.+ (Word32.* (Word32.fromInt
     (1564206498 : IntInf.int), hashcode_literal
                                  x_1), Word32.+ (Word32.* (Word32.fromInt
                      (1557546131 : IntInf.int), hashcode_exp
           x_2), Word32.fromInt (1577527232 : IntInf.int)))))
    (fn x_0 => fn x_1 => fn x_2 =>
      Word32.+ (Word32.* (Word32.fromInt
                            (1826932132 : IntInf.int), hashcode_exp
                 x_0), Word32.+ (Word32.* (Word32.fromInt
     (1820271765 : IntInf.int), hashcode_literal
                                  x_1), Word32.+ (Word32.* (Word32.fromInt
                      (1813611398 : IntInf.int), hashcode_exp
           x_2), Word32.fromInt (1833592499 : IntInf.int)))))
    (fn x_0 => fn x_1 => fn x_2 =>
      Word32.+ (Word32.* (Word32.fromInt
                            (1099820219 : IntInf.int), hashcode_exp
                 x_0), Word32.+ (Word32.* (Word32.fromInt
     (1093159852 : IntInf.int), hashcode_literal
                                  x_1), Word32.+ (Word32.* (Word32.fromInt
                      (1086499485 : IntInf.int), hashcode_exp
           x_2), Word32.fromInt (1106480586 : IntInf.int)))))
    (fn x_0 =>
      Word32.+ (Word32.* (Word32.fromInt
                            (1940597353 : IntInf.int), hashcode_exp
                 x_0), Word32.fromInt (1947257720 : IntInf.int)))
    (Word32.fromInt (403702018 : IntInf.int))
    (fn _ => fn _ => fn res_0 => fn res_1 =>
      Word32.+ (Word32.* (Word32.fromInt
                            (844900734 : IntInf.int), res_0), Word32.+ (Word32.* (Word32.fromInt
    (838240367 : IntInf.int), res_1), Word32.fromInt (851561101 : IntInf.int))))
    (fn _ => fn _ => fn res_0 => fn res_1 =>
      Word32.+ (Word32.* (Word32.fromInt
                            (1248959763 : IntInf.int), res_0), Word32.+ (Word32.* (Word32.fromInt
     (1242299396 : IntInf.int), res_1), Word32.fromInt
  (1255620130 : IntInf.int))))
    (fn _ => fn res_0 =>
      Word32.+ (Word32.* (Word32.fromInt
                            (1380907979 : IntInf.int), res_0), Word32.fromInt
                         (1387568346 : IntInf.int)))
    (Word32.fromInt (641569256 : IntInf.int))
    (Word32.fromInt (992741301 : IntInf.int))
    (fn _ => fn _ => fn res_0 => fn res_1 =>
      Word32.+ (Word32.* (Word32.fromInt
                            (55136646 : IntInf.int), res_0), Word32.+ (Word32.* (Word32.fromInt
   (48476279 : IntInf.int), res_1), Word32.fromInt (61797013 : IntInf.int))))
    (Word32.fromInt (281889773 : IntInf.int)) x;

val hashable_cmd =
  {hashcode = hashcode_cmd, def_hashmap_size = def_hashmap_size_cmd} :
  cmd hashable;

val equal_uint32 = {equal = (fn a => fn b => ((a : Word32.word) = b))} :
  Word32.word equal;

val plus_uint32 = {plus = (fn a => fn b => Word32.+ (a, b))} : Word32.word plus;

val zero_uint32 = {zero = (Word32.fromInt 0)} : Word32.word zero;

type 'a monoid_add =
  {semigroup_add_monoid_add : 'a semigroup_add, zero_monoid_add : 'a zero};
val semigroup_add_monoid_add = #semigroup_add_monoid_add :
  'a monoid_add -> 'a semigroup_add;
val zero_monoid_add = #zero_monoid_add : 'a monoid_add -> 'a zero;

val semigroup_add_uint32 = {plus_semigroup_add = plus_uint32} :
  Word32.word semigroup_add;

val monoid_add_uint32 =
  {semigroup_add_monoid_add = semigroup_add_uint32,
    zero_monoid_add = zero_uint32}
  : Word32.word monoid_add;

fun def_hashmap_size_uint32 t = nat_of_integer (8 : IntInf.int);

fun hashcode_uint32 x = x;

val hashable_uint32 =
  {hashcode = hashcode_uint32, def_hashmap_size = def_hashmap_size_uint32} :
  Word32.word hashable;

type 'a ab_semigroup_add = {semigroup_add_ab_semigroup_add : 'a semigroup_add};
val semigroup_add_ab_semigroup_add = #semigroup_add_ab_semigroup_add :
  'a ab_semigroup_add -> 'a semigroup_add;

type 'a comm_monoid_add =
  {ab_semigroup_add_comm_monoid_add : 'a ab_semigroup_add,
    monoid_add_comm_monoid_add : 'a monoid_add};
val ab_semigroup_add_comm_monoid_add = #ab_semigroup_add_comm_monoid_add :
  'a comm_monoid_add -> 'a ab_semigroup_add;
val monoid_add_comm_monoid_add = #monoid_add_comm_monoid_add :
  'a comm_monoid_add -> 'a monoid_add;

val ab_semigroup_add_uint32 =
  {semigroup_add_ab_semigroup_add = semigroup_add_uint32} :
  Word32.word ab_semigroup_add;

val comm_monoid_add_uint32 =
  {ab_semigroup_add_comm_monoid_add = ab_semigroup_add_uint32,
    monoid_add_comm_monoid_add = monoid_add_uint32}
  : Word32.word comm_monoid_add;

datatype brk_ctd = AIBreak | AIContinue;

fun equal_brk_ctda AIBreak AIContinue = false
  | equal_brk_ctda AIContinue AIBreak = false
  | equal_brk_ctda AIContinue AIContinue = true
  | equal_brk_ctda AIBreak AIBreak = true;

val equal_brk_ctd = {equal = equal_brk_ctda} : brk_ctd equal;

val equal_literal = {equal = (fn a => fn b => ((a : string) = b))} :
  string equal;

val ord_literal =
  {less_eq = (fn a => fn b => ((a : string) <= b)),
    less = (fn a => fn b => ((a : string) < b))}
  : string ord;

val preorder_literal = {ord_preorder = ord_literal} : string preorder;

val order_literal = {preorder_order = preorder_literal} : string order;

val linorder_literal = {order_linorder = order_literal} : string linorder;

datatype ('a, 'b) phantom = Phantom of 'b;

val finite_UNIV_literala : (string, bool) phantom = Phantom false;

val card_UNIV_literala : (string, nat) phantom = Phantom zero_nata;

type 'a finite_UNIV = {finite_UNIV : ('a, bool) phantom};
val finite_UNIV = #finite_UNIV : 'a finite_UNIV -> ('a, bool) phantom;

type 'a card_UNIV =
  {finite_UNIV_card_UNIV : 'a finite_UNIV, card_UNIV : ('a, nat) phantom};
val finite_UNIV_card_UNIV = #finite_UNIV_card_UNIV :
  'a card_UNIV -> 'a finite_UNIV;
val card_UNIV = #card_UNIV : 'a card_UNIV -> ('a, nat) phantom;

val finite_UNIV_literal = {finite_UNIV = finite_UNIV_literala} :
  string finite_UNIV;

val card_UNIV_literal =
  {finite_UNIV_card_UNIV = finite_UNIV_literal, card_UNIV = card_UNIV_literala}
  : string card_UNIV;

type 'a len0 = {len_of : 'a itself -> nat};
val len_of = #len_of : 'a len0 -> 'a itself -> nat;

type 'a countable = {};

type 'a finite = {countable_finite : 'a countable};
val countable_finite = #countable_finite : 'a finite -> 'a countable;

datatype 'a bit0 = Abs_bit0 of int;

fun len_of_bit0 A_ x =
  times_nata (nat_of_integer (2 : IntInf.int)) (len_of A_ Type);

type 'a len = {len0_len : 'a len0};
val len0_len = #len0_len : 'a len -> 'a len0;

fun len0_bit0 A_ = {len_of = len_of_bit0 A_} : 'a bit0 len0;

fun len_bit0 A_ = {len0_len = len0_bit0 (len0_len A_)} : 'a bit0 len;

datatype num1 = One_num1;

fun len_of_num1 x = one_nata;

val len0_num1 = {len_of = len_of_num1} : num1 len0;

val len_num1 = {len0_len = len0_num1} : num1 len;

fun equal_proda A_ B_ (x1, x2) (y1, y2) = eq A_ x1 y1 andalso eq B_ x2 y2;

fun equal_prod A_ B_ = {equal = equal_proda A_ B_} : ('a * 'b) equal;

fun def_hashmap_size_prod A_ B_ =
  (fn _ => plus_nata (def_hashmap_size A_ Type) (def_hashmap_size B_ Type));

fun hashcode_prod A_ B_ x =
  Word32.+ (Word32.* (hashcode A_
                        (fst x), Word32.fromInt
                                   (33 : IntInf.int)), hashcode B_ (snd x));

fun hashable_prod A_ B_ =
  {hashcode = hashcode_prod A_ B_,
    def_hashmap_size = def_hashmap_size_prod A_ B_}
  : ('a * 'b) hashable;

fun equal_unita u v = true;

val equal_unit = {equal = equal_unita} : unit equal;

fun def_hashmap_size_unit x = (fn _ => nat_of_integer (2 : IntInf.int)) x;

fun hashcode_unit u = (Word32.fromInt 0);

val hashable_unit =
  {hashcode = hashcode_unit, def_hashmap_size = def_hashmap_size_unit} :
  unit hashable;

datatype ('a, 'b, 'c) local_config_ext = Local_config_ext of 'a * 'b * 'c;

fun equal_local_config_exta A_ B_ C_
  (Local_config_ext (commanda, statea, morea))
  (Local_config_ext (command, state, more)) =
  eq A_ commanda command andalso (eq B_ statea state andalso eq C_ morea more);

fun equal_local_config_ext A_ B_ C_ = {equal = equal_local_config_exta A_ B_ C_}
  : ('a, 'b, 'c) local_config_ext equal;

fun def_hashmap_size_local_config_ext A_ B_ C_ t =
  nat_of_integer (8 : IntInf.int);

fun command (Local_config_ext (command, state, more)) = command;

fun state (Local_config_ext (command, state, more)) = state;

fun more (Local_config_ext (command, state, more)) = more;

fun hashcode_local_config_ext A_ B_ C_ lc =
  Word32.+ (Word32.+ (hashcode A_
                        (command
                          lc), hashcode B_ (state lc)), hashcode C_ (more lc));

fun hashable_local_config_ext A_ B_ C_ =
  {hashcode = hashcode_local_config_ext A_ B_ C_,
    def_hashmap_size = def_hashmap_size_local_config_ext A_ B_ C_}
  : ('a, 'b, 'c) local_config_ext hashable;

datatype color = R | B;

datatype ('a, 'b) rbta = Empty |
  Branch of color * ('a, 'b) rbta * 'a * 'b * ('a, 'b) rbta;

datatype ('b, 'a) rbt = RBT of ('b, 'a) rbta;

datatype ('a, 'b) mapping = Mapping of ('a, 'b) rbt;

datatype 'a local_state_impl_ext =
  Local_state_impl_ext of (string, Word32.word) mapping * 'a;

fun equal_list A_ [] (x21 :: x22) = false
  | equal_list A_ (x21 :: x22) [] = false
  | equal_list A_ (x21 :: x22) (y21 :: y22) =
    eq A_ x21 y21 andalso equal_list A_ x22 y22
  | equal_list A_ [] [] = true;

fun gen_entries kvts (Branch (c, l, k, v, r)) =
  gen_entries (((k, v), r) :: kvts) l
  | gen_entries ((kv, t) :: kvts) Empty = kv :: gen_entries kvts t
  | gen_entries [] Empty = [];

fun entriesa x = gen_entries [] x;

fun impl_of B_ (RBT x) = x;

fun entries A_ x = entriesa (impl_of A_ x);

fun equal_mapping (A1_, A2_) B_ (Mapping t1) (Mapping t2) =
  equal_list (equal_prod A1_ B_) (entries A2_ t1) (entries A2_ t2);

fun equal_local_state_impl_exta A_ (Local_state_impl_ext (variablesa, morea))
  (Local_state_impl_ext (variables, more)) =
  equal_mapping (equal_literal, linorder_literal) equal_uint32 variablesa
    variables andalso
    eq A_ morea more;

fun equal_local_state_impl_ext A_ = {equal = equal_local_state_impl_exta A_} :
  'a local_state_impl_ext equal;

fun def_hashmap_size_local_state_impl_ext A_ t =
  nat_of_integer (8 : IntInf.int);

fun variables (Local_state_impl_ext (variables, more)) = variables;

fun moreb (Local_state_impl_ext (variables, more)) = more;

fun hashcode_option A_ opt =
  (case opt of NONE => (Word32.fromInt 0)
    | SOME a => Word32.+ (hashcode A_ a, (Word32.fromInt 1)));

fun id x = (fn xa => xa) x;

fun foldr f [] = id
  | foldr f (x :: xs) = f x o foldr f xs;

fun listsum A_ xs =
  foldr (plus ((plus_semigroup_add o semigroup_add_monoid_add) A_)) xs
    (zero (zero_monoid_add A_));

fun map fi [] = []
  | map fi (x21a :: x22) = fi x21a :: map fi x22;

fun remdups A_ [] = []
  | remdups A_ (x :: xs) =
    (if membera A_ xs x then remdups A_ xs else x :: remdups A_ xs);

fun setsum A_ B_ f (Set xs) =
  listsum (monoid_add_comm_monoid_add B_) (map f (remdups A_ xs));

fun rbt_lookupa A_ Empty k = NONE
  | rbt_lookupa A_ (Branch (uu, l, x, y, r)) k =
    (if less A_ k x then rbt_lookupa A_ l k
      else (if less A_ x k then rbt_lookupa A_ r k else SOME y));

fun lookup A_ x =
  rbt_lookupa ((ord_preorder o preorder_order o order_linorder) A_)
    (impl_of A_ x);

fun lookupa A_ (Mapping t) = lookup A_ t;

fun gen_keys kts (Branch (c, l, k, v, r)) = gen_keys ((k, r) :: kts) l
  | gen_keys ((k, t) :: kts) Empty = k :: gen_keys kts t
  | gen_keys [] Empty = [];

fun keysb x = gen_keys [] x;

fun keys A_ x = keysb (impl_of A_ x);

fun keysa A_ (Mapping t) = Set (keys A_ t);

fun image f (Set xs) = Set (map f xs);

fun mapping_val_hashcode A_ B_ m =
  setsum equal_uint32 comm_monoid_add_uint32 (fn x => x)
    (image (hashcode_option B_ o lookupa A_ m) (keysa A_ m));

fun hashcode_local_state_impl_ext A_ ls =
  Word32.+ (mapping_val_hashcode linorder_literal hashable_uint32
              (variables ls), hashcode A_ (moreb ls));

fun hashable_local_state_impl_ext A_ =
  {hashcode = hashcode_local_state_impl_ext A_,
    def_hashmap_size = def_hashmap_size_local_state_impl_ext A_}
  : 'a local_state_impl_ext hashable;

datatype ('a, 'b, 'c, 'd) pid_global_config_ext =
  Pid_global_config_ext of ('a, 'b, unit) local_config_ext list * 'c * 'd;

fun equal_pid_global_config_exta A_ B_ C_ D_
  (Pid_global_config_ext (processesa, statea, morea))
  (Pid_global_config_ext (processes, state, more)) =
  equal_list (equal_local_config_ext A_ B_ equal_unit) processesa
    processes andalso
    (eq C_ statea state andalso eq D_ morea more);

fun equal_pid_global_config_ext A_ B_ C_ D_ =
  {equal = equal_pid_global_config_exta A_ B_ C_ D_} :
  ('a, 'b, 'c, 'd) pid_global_config_ext equal;

fun def_hashmap_size_pid_global_config_ext A_ B_ C_ D_ t =
  nat_of_integer (8 : IntInf.int);

fun processesa (Pid_global_config_ext (processes, state, more)) = processes;

fun statea (Pid_global_config_ext (processes, state, more)) = state;

fun morec (Pid_global_config_ext (processes, state, more)) = more;

fun hashcode_pid_global_config_ext A_ B_ C_ D_ gc =
  Word32.+ (Word32.+ (hashcode_list
                        (hashable_local_config_ext A_ B_ hashable_unit)
                        (processesa
                          gc), hashcode C_
                                 (statea gc)), hashcode D_ (morec gc));

fun hashable_pid_global_config_ext A_ B_ C_ D_ =
  {hashcode = hashcode_pid_global_config_ext A_ B_ C_ D_,
    def_hashmap_size = def_hashmap_size_pid_global_config_ext A_ B_ C_ D_}
  : ('a, 'b, 'c, 'd) pid_global_config_ext hashable;

datatype 'a global_state_impl_ext =
  Global_state_impl_ext of (string, Word32.word) mapping * 'a;

fun equal_global_state_impl_exta A_ (Global_state_impl_ext (variablesa, morea))
  (Global_state_impl_ext (variables, more)) =
  equal_mapping (equal_literal, linorder_literal) equal_uint32 variablesa
    variables andalso
    eq A_ morea more;

fun equal_global_state_impl_ext A_ = {equal = equal_global_state_impl_exta A_} :
  'a global_state_impl_ext equal;

fun def_hashmap_size_global_state_impl_ext A_ t =
  nat_of_integer (8 : IntInf.int);

fun variablesa (Global_state_impl_ext (variables, more)) = variables;

fun mored (Global_state_impl_ext (variables, more)) = more;

fun hashcode_global_state_impl_ext A_ gs =
  Word32.+ (mapping_val_hashcode linorder_literal hashable_uint32
              (variablesa gs), hashcode A_ (mored gs));

fun hashable_global_state_impl_ext A_ =
  {hashcode = hashcode_global_state_impl_ext A_,
    def_hashmap_size = def_hashmap_size_global_state_impl_ext A_}
  : 'a global_state_impl_ext hashable;

datatype 'a ltl = LTLTrue | LTLFalse | LTLProp of 'a | LTLNeg of 'a ltl |
  LTLAnd of 'a ltl * 'a ltl | LTLOr of 'a ltl * 'a ltl | LTLNext of 'a ltl |
  LTLUntil of 'a ltl * 'a ltl | LTLRelease of 'a ltl * 'a ltl;

datatype 'a word = Word of int;

datatype 'a dres = DSUCCEEDi | DFAILi | DRETURN of 'a;

datatype compare = LT | GT | EQ;

datatype comp_res = LESS | EQUAL | GREATER;

datatype config_ce = CFG_CE_SCC_GABOW | CFG_CE_NDFS;

datatype ('a, 'b) lasso_ext = Lasso_ext of 'a list * 'a * 'a list * 'b;

datatype sm_result = TY_ERR | SAT | UNSAT |
  UNSAT_CE of
    ((cmd, unit local_state_impl_ext, unit global_state_impl_ext, unit)
       pid_global_config_ext,
      unit)
      lasso_ext;

datatype ('a, 'b) hashmapa =
  HashMapa of (('a * 'b) list) FArray.IsabelleMapping.ArrayType * nat;

datatype ('b, 'a) hashmap = HashMap of ('b, 'a) hashmapa;

datatype config_l2b = CFG_L2B_GERTHS;

datatype 'a blue_witness = NO_CYC | REACH of 'a * 'a list |
  CIRC of 'a list * 'a list;

datatype ('a, 'b) hashmapb =
  HashMapb of (('a * 'b) list) FArray.IsabelleMapping.ArrayType * nat;

datatype 'a proc_decl_ext = Proc_decl_ext of cmd * string list * 'a;

datatype 'a program_ext =
  Program_ext of unit proc_decl_ext list * string list * 'a;

datatype ('a, 'b) node_impl_ext =
  Node_impl_ext of
    nat * (nat, unit) rbta * 'a ltln list * 'a ltln list * 'a ltln list * 'b;

datatype ('a, 'b) gen_bg_impl_ext = Gen_bg_impl_ext of 'a * 'b;

datatype ('a, 'b) gen_sa_impl_ext = Gen_sa_impl_ext of 'a * 'b;

datatype ('a, 'b, 'c, 'd) gen_frg_impl_ext =
  Gen_frg_impl_ext of 'a * 'b * 'c * 'd;

datatype ('a, 'b) gen_gba_impl_ext = Gen_gba_impl_ext of 'a * 'b;

datatype ('a, 'b) gen_gbg_impl_ext = Gen_gbg_impl_ext of 'a * 'b;

datatype ('a, 'b) gen_igba_impl_ext = Gen_igba_impl_ext of 'a * 'b;

datatype ('a, 'b) gen_igbg_impl_ext = Gen_igbg_impl_ext of nat * 'a * 'b;

datatype ('a, 'b) fas_state_impl_ext = Fas_state_impl_ext of 'a * 'b;

datatype ('a, 'b, 'c) simple_state_impl_ext =
  Simple_state_impl_ext of 'a * 'b * 'b * 'c;

datatype ('a, 'b, 'c) simple_state_nos_impl_ext =
  Simple_state_nos_impl_ext of 'a * 'b * 'c;

fun nat k = Nat (max ord_integer 0 (integer_of_int k));

fun suc n = plus_nata n one_nata;

fun list_ex p [] = false
  | list_ex p (x :: xs) = p x orelse list_ex p xs;

fun bex (Set xs) p = list_ex p xs;

fun minus_nat m n =
  Nat (max ord_integer 0 (IntInf.- (integer_of_nat m, integer_of_nat n)));

fun nth (x :: xs) n =
  (if equal_nata n zero_nata then x else nth xs (minus_nat n one_nata));

fun rev xs = fold (fn a => fn b => a :: b) xs [];

fun maps f [] = []
  | maps f (x :: xs) = f x @ maps f xs;

fun empty A_ = RBT Empty;

val mulog_property : aprop ltlc = LTLcGlobal (LTLcProp (AP_isNull "x"));

val zero_int : int = Int_of_integer 0;

val mulog_program : unit program_ext =
  Program_ext
    ([Proc_decl_ext
        (Seq (Seq (Seq (Assign
                          (E_const (Int_of_integer (1 : IntInf.int)),
                            "foo_handshake_sender_lock", E_const zero_int),
                         Seq (Assign
                                (E_const (Int_of_integer (1 : IntInf.int)),
                                  "foo_handshake_receiver_lock",
                                  E_const zero_int),
                               Seq (Assign
                                      (E_const
 (Int_of_integer (1 : IntInf.int)),
"foo_handshake_message", E_const zero_int),
                                     Seq (Assign
    (E_const (Int_of_integer (1 : IntInf.int)), "foo_handshake_sent",
      E_const zero_int),
   Assign
     (E_const (Int_of_integer (1 : IntInf.int)), "foo_handshake_received",
       E_const zero_int))))),
                    Seq (Assign
                           (E_const (Int_of_integer (1 : IntInf.int)),
                             "foo_msg", E_const zero_int),
                          Seq (Assign
                                 (E_const (Int_of_integer (1 : IntInf.int)),
                                   "foo_idle",
                                   E_const (Int_of_integer (1 : IntInf.int))),
                                Seq (Assign
                                       (E_const
  (Int_of_integer (1 : IntInf.int)),
 "foo_parent", E_const zero_int),
                                      Seq
(Assign
   (E_const (Int_of_integer (1 : IntInf.int)), "foo_next", E_const zero_int),
  Seq (Assign
         (E_const (Int_of_integer (1 : IntInf.int)), "foo_requesting",
           E_const zero_int),
        Assign
          (E_const (Int_of_integer (1 : IntInf.int)), "foo_token",
            E_const (Int_of_integer (1 : IntInf.int))))))))),
               Loop (Or (Seq (Test (E_var "foo_handshake_sent"),
                               Seq (Seq (Assign
   (E_un (Uo_not, E_var "foo_handshake_receiver_lock"),
     "foo_handshake_receiver_lock", E_const (Int_of_integer (1 : IntInf.int))),
  Seq (Seq (Test (E_var "foo_handshake_sent"),
             Seq (Assign
                    (E_const (Int_of_integer (1 : IntInf.int)),
                      "foo_handshake_sent", E_const zero_int),
                   Seq (Assign
                          (E_const (Int_of_integer (1 : IntInf.int)), "foo_msg",
                            E_var "foo_handshake_message"),
                         Assign
                           (E_const (Int_of_integer (1 : IntInf.int)),
                             "foo_handshake_received",
                             E_const (Int_of_integer (1 : IntInf.int)))))),
        Assign
          (E_const (Int_of_integer (1 : IntInf.int)),
            "foo_handshake_receiver_lock", E_const zero_int))),
                                     Or (Seq
   (Test (E_bin (Bo_eq, E_var "foo_msg", E_const zero_int)),
     Assign
       (E_const (Int_of_integer (1 : IntInf.int)), "foo_token",
         E_const (Int_of_integer (1 : IntInf.int)))),
  Seq (Test (E_un (Uo_not, E_bin (Bo_eq, E_var "foo_msg", E_const zero_int))),
        Seq (Or (Seq (Test (E_bin
                             (Bo_eq, E_var "foo_parent", E_const zero_int)),
                       Or (Seq (Test (E_var "foo_requesting"),
                                 Assign
                                   (E_const (Int_of_integer (1 : IntInf.int)),
                                     "foo_next", E_var "foo_msg")),
                            Seq (Test (E_un (Uo_not, E_var "foo_requesting")),
                                  Seq (Assign
 (E_const (Int_of_integer (1 : IntInf.int)), "foo_token", E_const zero_int),
Or (Seq (Test (E_bin
                (Bo_eq, E_var "foo_msg",
                  E_const (Int_of_integer (3 : IntInf.int)))),
          Seq (Assign
                 (E_un (Uo_not, E_var "baz_handshake_sender_lock"),
                   "baz_handshake_sender_lock",
                   E_const (Int_of_integer (1 : IntInf.int))),
                Seq (Seq (Assign
                            (E_const (Int_of_integer (1 : IntInf.int)),
                              "baz_handshake_message", E_const zero_int),
                           Seq (Assign
                                  (E_const (Int_of_integer (1 : IntInf.int)),
                                    "baz_handshake_sent",
                                    E_const (Int_of_integer (1 : IntInf.int))),
                                 Seq (Test (E_var "baz_handshake_received"),
                                       Assign
 (E_const (Int_of_integer (1 : IntInf.int)), "baz_handshake_received",
   E_const zero_int)))),
                      Assign
                        (E_const (Int_of_integer (1 : IntInf.int)),
                          "baz_handshake_sender_lock", E_const zero_int)))),
     Seq (Test (E_un (Uo_not,
                       E_bin (Bo_eq, E_var "foo_msg",
                               E_const (Int_of_integer (3 : IntInf.int))))),
           Or (Seq (Test (E_bin
                           (Bo_eq, E_var "foo_msg",
                             E_const (Int_of_integer (2 : IntInf.int)))),
                     Seq (Assign
                            (E_un (Uo_not, E_var "bar_handshake_sender_lock"),
                              "bar_handshake_sender_lock",
                              E_const (Int_of_integer (1 : IntInf.int))),
                           Seq (Seq (Assign
                                       (E_const
  (Int_of_integer (1 : IntInf.int)),
 "bar_handshake_message", E_const zero_int),
                                      Seq
(Assign
   (E_const (Int_of_integer (1 : IntInf.int)), "bar_handshake_sent",
     E_const (Int_of_integer (1 : IntInf.int))),
  Seq (Test (E_var "bar_handshake_received"),
        Assign
          (E_const (Int_of_integer (1 : IntInf.int)), "bar_handshake_received",
            E_const zero_int)))),
                                 Assign
                                   (E_const (Int_of_integer (1 : IntInf.int)),
                                     "bar_handshake_sender_lock",
                                     E_const zero_int)))),
                Seq (Test (E_un (Uo_not,
                                  E_bin (Bo_eq, E_var "foo_msg",
  E_const (Int_of_integer (2 : IntInf.int))))),
                      Or (Seq (Test (E_bin
                                      (Bo_eq, E_var "foo_msg",
E_const (Int_of_integer (1 : IntInf.int)))),
                                Seq (Assign
                                       (E_un
  (Uo_not, E_var "foo_handshake_sender_lock"),
 "foo_handshake_sender_lock", E_const (Int_of_integer (1 : IntInf.int))),
                                      Seq
(Seq (Assign
        (E_const (Int_of_integer (1 : IntInf.int)), "foo_handshake_message",
          E_const zero_int),
       Seq (Assign
              (E_const (Int_of_integer (1 : IntInf.int)), "foo_handshake_sent",
                E_const (Int_of_integer (1 : IntInf.int))),
             Seq (Test (E_var "foo_handshake_received"),
                   Assign
                     (E_const (Int_of_integer (1 : IntInf.int)),
                       "foo_handshake_received", E_const zero_int)))),
  Assign
    (E_const (Int_of_integer (1 : IntInf.int)), "foo_handshake_sender_lock",
      E_const zero_int)))),
                           Seq (Test (E_un (Uo_not,
     E_bin (Bo_eq, E_var "foo_msg",
             E_const (Int_of_integer (1 : IntInf.int))))),
                                 Skip)))))))))),
                  Seq (Test (E_un (Uo_not,
                                    E_bin (Bo_eq, E_var "foo_parent",
    E_const zero_int))),
                        Or (Seq (Test (E_bin
(Bo_eq, E_var "foo_parent", E_const (Int_of_integer (3 : IntInf.int)))),
                                  Seq (Assign
 (E_un (Uo_not, E_var "baz_handshake_sender_lock"), "baz_handshake_sender_lock",
   E_const (Int_of_integer (1 : IntInf.int))),
Seq (Seq (Assign
            (E_const (Int_of_integer (1 : IntInf.int)), "baz_handshake_message",
              E_var "foo_msg"),
           Seq (Assign
                  (E_const (Int_of_integer (1 : IntInf.int)),
                    "baz_handshake_sent",
                    E_const (Int_of_integer (1 : IntInf.int))),
                 Seq (Test (E_var "baz_handshake_received"),
                       Assign
                         (E_const (Int_of_integer (1 : IntInf.int)),
                           "baz_handshake_received", E_const zero_int)))),
      Assign
        (E_const (Int_of_integer (1 : IntInf.int)), "baz_handshake_sender_lock",
          E_const zero_int)))),
                             Seq (Test (E_un
 (Uo_not,
   E_bin (Bo_eq, E_var "foo_parent",
           E_const (Int_of_integer (3 : IntInf.int))))),
                                   Or (Seq
 (Test (E_bin
         (Bo_eq, E_var "foo_parent",
           E_const (Int_of_integer (2 : IntInf.int)))),
   Seq (Assign
          (E_un (Uo_not, E_var "bar_handshake_sender_lock"),
            "bar_handshake_sender_lock",
            E_const (Int_of_integer (1 : IntInf.int))),
         Seq (Seq (Assign
                     (E_const (Int_of_integer (1 : IntInf.int)),
                       "bar_handshake_message", E_var "foo_msg"),
                    Seq (Assign
                           (E_const (Int_of_integer (1 : IntInf.int)),
                             "bar_handshake_sent",
                             E_const (Int_of_integer (1 : IntInf.int))),
                          Seq (Test (E_var "bar_handshake_received"),
                                Assign
                                  (E_const (Int_of_integer (1 : IntInf.int)),
                                    "bar_handshake_received",
                                    E_const zero_int)))),
               Assign
                 (E_const (Int_of_integer (1 : IntInf.int)),
                   "bar_handshake_sender_lock", E_const zero_int)))),
Seq (Test (E_un (Uo_not,
                  E_bin (Bo_eq, E_var "foo_parent",
                          E_const (Int_of_integer (2 : IntInf.int))))),
      Or (Seq (Test (E_bin
                      (Bo_eq, E_var "foo_parent",
                        E_const (Int_of_integer (1 : IntInf.int)))),
                Seq (Assign
                       (E_un (Uo_not, E_var "foo_handshake_sender_lock"),
                         "foo_handshake_sender_lock",
                         E_const (Int_of_integer (1 : IntInf.int))),
                      Seq (Seq (Assign
                                  (E_const (Int_of_integer (1 : IntInf.int)),
                                    "foo_handshake_message", E_var "foo_msg"),
                                 Seq (Assign
(E_const (Int_of_integer (1 : IntInf.int)), "foo_handshake_sent",
  E_const (Int_of_integer (1 : IntInf.int))),
                                       Seq
 (Test (E_var "foo_handshake_received"),
   Assign
     (E_const (Int_of_integer (1 : IntInf.int)), "foo_handshake_received",
       E_const zero_int)))),
                            Assign
                              (E_const (Int_of_integer (1 : IntInf.int)),
                                "foo_handshake_sender_lock",
                                E_const zero_int)))),
           Seq (Test (E_un (Uo_not,
                             E_bin (Bo_eq, E_var "foo_parent",
                                     E_const
                                       (Int_of_integer (1 : IntInf.int))))),
                 Skip)))))))),
              Assign
                (E_const (Int_of_integer (1 : IntInf.int)), "foo_parent",
                  E_var "foo_msg")))))),
                          Seq (Test (E_un (Uo_not, E_var "foo_handshake_sent")),
                                Or (Seq (Test (E_var "foo_idle"),
  Seq (Assign
         (E_const (Int_of_integer (1 : IntInf.int)), "foo_idle",
           E_const zero_int),
        Seq (Assign
               (E_const (Int_of_integer (1 : IntInf.int)), "foo_requesting",
                 E_const (Int_of_integer (1 : IntInf.int))),
              Or (Seq (Test (E_bin
                              (Bo_eq, E_var "foo_parent", E_const zero_int)),
                        Skip),
                   Seq (Test (E_un (Uo_not,
                                     E_bin (Bo_eq, E_var "foo_parent",
     E_const zero_int))),
                         Seq (Or (Seq (Test
 (E_bin (Bo_eq, E_var "foo_parent", E_const (Int_of_integer (3 : IntInf.int)))),
Seq (Assign
       (E_un (Uo_not, E_var "baz_handshake_sender_lock"),
         "baz_handshake_sender_lock",
         E_const (Int_of_integer (1 : IntInf.int))),
      Seq (Seq (Assign
                  (E_const (Int_of_integer (1 : IntInf.int)),
                    "baz_handshake_message",
                    E_const (Int_of_integer (1 : IntInf.int))),
                 Seq (Assign
                        (E_const (Int_of_integer (1 : IntInf.int)),
                          "baz_handshake_sent",
                          E_const (Int_of_integer (1 : IntInf.int))),
                       Seq (Test (E_var "baz_handshake_received"),
                             Assign
                               (E_const (Int_of_integer (1 : IntInf.int)),
                                 "baz_handshake_received", E_const zero_int)))),
            Assign
              (E_const (Int_of_integer (1 : IntInf.int)),
                "baz_handshake_sender_lock", E_const zero_int)))),
                                   Seq (Test
  (E_un (Uo_not,
          E_bin (Bo_eq, E_var "foo_parent",
                  E_const (Int_of_integer (3 : IntInf.int))))),
 Or (Seq (Test (E_bin
                 (Bo_eq, E_var "foo_parent",
                   E_const (Int_of_integer (2 : IntInf.int)))),
           Seq (Assign
                  (E_un (Uo_not, E_var "bar_handshake_sender_lock"),
                    "bar_handshake_sender_lock",
                    E_const (Int_of_integer (1 : IntInf.int))),
                 Seq (Seq (Assign
                             (E_const (Int_of_integer (1 : IntInf.int)),
                               "bar_handshake_message",
                               E_const (Int_of_integer (1 : IntInf.int))),
                            Seq (Assign
                                   (E_const (Int_of_integer (1 : IntInf.int)),
                                     "bar_handshake_sent",
                                     E_const (Int_of_integer (1 : IntInf.int))),
                                  Seq (Test (E_var "bar_handshake_received"),
Assign
  (E_const (Int_of_integer (1 : IntInf.int)), "bar_handshake_received",
    E_const zero_int)))),
                       Assign
                         (E_const (Int_of_integer (1 : IntInf.int)),
                           "bar_handshake_sender_lock", E_const zero_int)))),
      Seq (Test (E_un (Uo_not,
                        E_bin (Bo_eq, E_var "foo_parent",
                                E_const (Int_of_integer (2 : IntInf.int))))),
            Or (Seq (Test (E_bin
                            (Bo_eq, E_var "foo_parent",
                              E_const (Int_of_integer (1 : IntInf.int)))),
                      Seq (Assign
                             (E_un (Uo_not, E_var "foo_handshake_sender_lock"),
                               "foo_handshake_sender_lock",
                               E_const (Int_of_integer (1 : IntInf.int))),
                            Seq (Seq (Assign
(E_const (Int_of_integer (1 : IntInf.int)), "foo_handshake_message",
  E_const (Int_of_integer (1 : IntInf.int))),
                                       Seq
 (Assign
    (E_const (Int_of_integer (1 : IntInf.int)), "foo_handshake_sent",
      E_const (Int_of_integer (1 : IntInf.int))),
   Seq (Test (E_var "foo_handshake_received"),
         Assign
           (E_const (Int_of_integer (1 : IntInf.int)), "foo_handshake_received",
             E_const zero_int)))),
                                  Assign
                                    (E_const (Int_of_integer (1 : IntInf.int)),
                                      "foo_handshake_sender_lock",
                                      E_const zero_int)))),
                 Seq (Test (E_un (Uo_not,
                                   E_bin (Bo_eq, E_var "foo_parent",
   E_const (Int_of_integer (1 : IntInf.int))))),
                       Skip)))))),
                               Assign
                                 (E_const (Int_of_integer (1 : IntInf.int)),
                                   "foo_parent", E_const zero_int))))))),
                                     Seq (Test
    (E_un (Uo_not, E_var "foo_idle")),
   Or (Seq (Test (E_var "foo_token"),
             Seq (Seq (Assign
                         (E_const (Int_of_integer (1 : IntInf.int)),
                           "foo_requesting", E_const zero_int),
                        Or (Seq (Test (E_bin
(Bo_eq, E_var "foo_next", E_const zero_int)),
                                  Skip),
                             Seq (Test (E_un
 (Uo_not, E_bin (Bo_eq, E_var "foo_next", E_const zero_int))),
                                   Seq (Assign
  (E_const (Int_of_integer (1 : IntInf.int)), "foo_token", E_const zero_int),
 Seq (Or (Seq (Test (E_bin
                      (Bo_eq, E_var "foo_next",
                        E_const (Int_of_integer (3 : IntInf.int)))),
                Seq (Assign
                       (E_un (Uo_not, E_var "baz_handshake_sender_lock"),
                         "baz_handshake_sender_lock",
                         E_const (Int_of_integer (1 : IntInf.int))),
                      Seq (Seq (Assign
                                  (E_const (Int_of_integer (1 : IntInf.int)),
                                    "baz_handshake_message", E_const zero_int),
                                 Seq (Assign
(E_const (Int_of_integer (1 : IntInf.int)), "baz_handshake_sent",
  E_const (Int_of_integer (1 : IntInf.int))),
                                       Seq
 (Test (E_var "baz_handshake_received"),
   Assign
     (E_const (Int_of_integer (1 : IntInf.int)), "baz_handshake_received",
       E_const zero_int)))),
                            Assign
                              (E_const (Int_of_integer (1 : IntInf.int)),
                                "baz_handshake_sender_lock",
                                E_const zero_int)))),
           Seq (Test (E_un (Uo_not,
                             E_bin (Bo_eq, E_var "foo_next",
                                     E_const
                                       (Int_of_integer (3 : IntInf.int))))),
                 Or (Seq (Test (E_bin
                                 (Bo_eq, E_var "foo_next",
                                   E_const (Int_of_integer (2 : IntInf.int)))),
                           Seq (Assign
                                  (E_un (Uo_not,
  E_var "bar_handshake_sender_lock"),
                                    "bar_handshake_sender_lock",
                                    E_const (Int_of_integer (1 : IntInf.int))),
                                 Seq (Seq
(Assign
   (E_const (Int_of_integer (1 : IntInf.int)), "bar_handshake_message",
     E_const zero_int),
  Seq (Assign
         (E_const (Int_of_integer (1 : IntInf.int)), "bar_handshake_sent",
           E_const (Int_of_integer (1 : IntInf.int))),
        Seq (Test (E_var "bar_handshake_received"),
              Assign
                (E_const (Int_of_integer (1 : IntInf.int)),
                  "bar_handshake_received", E_const zero_int)))),
                                       Assign
 (E_const (Int_of_integer (1 : IntInf.int)), "bar_handshake_sender_lock",
   E_const zero_int)))),
                      Seq (Test (E_un (Uo_not,
E_bin (Bo_eq, E_var "foo_next", E_const (Int_of_integer (2 : IntInf.int))))),
                            Or (Seq (Test (E_bin
    (Bo_eq, E_var "foo_next", E_const (Int_of_integer (1 : IntInf.int)))),
                                      Seq
(Assign
   (E_un (Uo_not, E_var "foo_handshake_sender_lock"),
     "foo_handshake_sender_lock", E_const (Int_of_integer (1 : IntInf.int))),
  Seq (Seq (Assign
              (E_const (Int_of_integer (1 : IntInf.int)),
                "foo_handshake_message", E_const zero_int),
             Seq (Assign
                    (E_const (Int_of_integer (1 : IntInf.int)),
                      "foo_handshake_sent",
                      E_const (Int_of_integer (1 : IntInf.int))),
                   Seq (Test (E_var "foo_handshake_received"),
                         Assign
                           (E_const (Int_of_integer (1 : IntInf.int)),
                             "foo_handshake_received", E_const zero_int)))),
        Assign
          (E_const (Int_of_integer (1 : IntInf.int)),
            "foo_handshake_sender_lock", E_const zero_int)))),
                                 Seq (Test
(E_un (Uo_not,
        E_bin (Bo_eq, E_var "foo_next",
                E_const (Int_of_integer (1 : IntInf.int))))),
                                       Skip)))))),
       Assign
         (E_const (Int_of_integer (1 : IntInf.int)), "foo_next",
           E_const zero_int)))))),
                   Assign
                     (E_const (Int_of_integer (1 : IntInf.int)), "foo_idle",
                       E_const (Int_of_integer (1 : IntInf.int))))),
        Seq (Test (E_un (Uo_not, E_var "foo_token")), Skip)))))))),
          ["foo_msg", "foo_idle", "foo_token", "foo_requesting", "foo_parent",
            "foo_next"],
          ()),
       Proc_decl_ext
         (Seq (Seq (Seq (Assign
                           (E_const (Int_of_integer (1 : IntInf.int)),
                             "bar_handshake_sender_lock", E_const zero_int),
                          Seq (Assign
                                 (E_const (Int_of_integer (1 : IntInf.int)),
                                   "bar_handshake_receiver_lock",
                                   E_const zero_int),
                                Seq (Assign
                                       (E_const
  (Int_of_integer (1 : IntInf.int)),
 "bar_handshake_message", E_const zero_int),
                                      Seq
(Assign
   (E_const (Int_of_integer (1 : IntInf.int)), "bar_handshake_sent",
     E_const zero_int),
  Assign
    (E_const (Int_of_integer (1 : IntInf.int)), "bar_handshake_received",
      E_const zero_int))))),
                     Seq (Assign
                            (E_const (Int_of_integer (1 : IntInf.int)),
                              "bar_msg", E_const zero_int),
                           Seq (Assign
                                  (E_const (Int_of_integer (1 : IntInf.int)),
                                    "bar_idle",
                                    E_const (Int_of_integer (1 : IntInf.int))),
                                 Seq (Assign
(E_const (Int_of_integer (1 : IntInf.int)), "bar_parent",
  E_const (Int_of_integer (1 : IntInf.int))),
                                       Seq
 (Assign
    (E_const (Int_of_integer (1 : IntInf.int)), "bar_next", E_const zero_int),
   Seq (Assign
          (E_const (Int_of_integer (1 : IntInf.int)), "bar_requesting",
            E_const zero_int),
         Assign
           (E_const (Int_of_integer (1 : IntInf.int)), "bar_token",
             E_const zero_int))))))),
                Loop (Or (Seq (Test (E_var "bar_handshake_sent"),
                                Seq (Seq (Assign
    (E_un (Uo_not, E_var "bar_handshake_receiver_lock"),
      "bar_handshake_receiver_lock", E_const (Int_of_integer (1 : IntInf.int))),
   Seq (Seq (Test (E_var "bar_handshake_sent"),
              Seq (Assign
                     (E_const (Int_of_integer (1 : IntInf.int)),
                       "bar_handshake_sent", E_const zero_int),
                    Seq (Assign
                           (E_const (Int_of_integer (1 : IntInf.int)),
                             "bar_msg", E_var "bar_handshake_message"),
                          Assign
                            (E_const (Int_of_integer (1 : IntInf.int)),
                              "bar_handshake_received",
                              E_const (Int_of_integer (1 : IntInf.int)))))),
         Assign
           (E_const (Int_of_integer (1 : IntInf.int)),
             "bar_handshake_receiver_lock", E_const zero_int))),
                                      Or
(Seq (Test (E_bin (Bo_eq, E_var "bar_msg", E_const zero_int)),
       Assign
         (E_const (Int_of_integer (1 : IntInf.int)), "bar_token",
           E_const (Int_of_integer (1 : IntInf.int)))),
  Seq (Test (E_un (Uo_not, E_bin (Bo_eq, E_var "bar_msg", E_const zero_int))),
        Seq (Or (Seq (Test (E_bin
                             (Bo_eq, E_var "bar_parent", E_const zero_int)),
                       Or (Seq (Test (E_var "bar_requesting"),
                                 Assign
                                   (E_const (Int_of_integer (1 : IntInf.int)),
                                     "bar_next", E_var "bar_msg")),
                            Seq (Test (E_un (Uo_not, E_var "bar_requesting")),
                                  Seq (Assign
 (E_const (Int_of_integer (1 : IntInf.int)), "bar_token", E_const zero_int),
Or (Seq (Test (E_bin
                (Bo_eq, E_var "bar_msg",
                  E_const (Int_of_integer (3 : IntInf.int)))),
          Seq (Assign
                 (E_un (Uo_not, E_var "baz_handshake_sender_lock"),
                   "baz_handshake_sender_lock",
                   E_const (Int_of_integer (1 : IntInf.int))),
                Seq (Seq (Assign
                            (E_const (Int_of_integer (1 : IntInf.int)),
                              "baz_handshake_message", E_const zero_int),
                           Seq (Assign
                                  (E_const (Int_of_integer (1 : IntInf.int)),
                                    "baz_handshake_sent",
                                    E_const (Int_of_integer (1 : IntInf.int))),
                                 Seq (Test (E_var "baz_handshake_received"),
                                       Assign
 (E_const (Int_of_integer (1 : IntInf.int)), "baz_handshake_received",
   E_const zero_int)))),
                      Assign
                        (E_const (Int_of_integer (1 : IntInf.int)),
                          "baz_handshake_sender_lock", E_const zero_int)))),
     Seq (Test (E_un (Uo_not,
                       E_bin (Bo_eq, E_var "bar_msg",
                               E_const (Int_of_integer (3 : IntInf.int))))),
           Or (Seq (Test (E_bin
                           (Bo_eq, E_var "bar_msg",
                             E_const (Int_of_integer (2 : IntInf.int)))),
                     Seq (Assign
                            (E_un (Uo_not, E_var "bar_handshake_sender_lock"),
                              "bar_handshake_sender_lock",
                              E_const (Int_of_integer (1 : IntInf.int))),
                           Seq (Seq (Assign
                                       (E_const
  (Int_of_integer (1 : IntInf.int)),
 "bar_handshake_message", E_const zero_int),
                                      Seq
(Assign
   (E_const (Int_of_integer (1 : IntInf.int)), "bar_handshake_sent",
     E_const (Int_of_integer (1 : IntInf.int))),
  Seq (Test (E_var "bar_handshake_received"),
        Assign
          (E_const (Int_of_integer (1 : IntInf.int)), "bar_handshake_received",
            E_const zero_int)))),
                                 Assign
                                   (E_const (Int_of_integer (1 : IntInf.int)),
                                     "bar_handshake_sender_lock",
                                     E_const zero_int)))),
                Seq (Test (E_un (Uo_not,
                                  E_bin (Bo_eq, E_var "bar_msg",
  E_const (Int_of_integer (2 : IntInf.int))))),
                      Or (Seq (Test (E_bin
                                      (Bo_eq, E_var "bar_msg",
E_const (Int_of_integer (1 : IntInf.int)))),
                                Seq (Assign
                                       (E_un
  (Uo_not, E_var "foo_handshake_sender_lock"),
 "foo_handshake_sender_lock", E_const (Int_of_integer (1 : IntInf.int))),
                                      Seq
(Seq (Assign
        (E_const (Int_of_integer (1 : IntInf.int)), "foo_handshake_message",
          E_const zero_int),
       Seq (Assign
              (E_const (Int_of_integer (1 : IntInf.int)), "foo_handshake_sent",
                E_const (Int_of_integer (1 : IntInf.int))),
             Seq (Test (E_var "foo_handshake_received"),
                   Assign
                     (E_const (Int_of_integer (1 : IntInf.int)),
                       "foo_handshake_received", E_const zero_int)))),
  Assign
    (E_const (Int_of_integer (1 : IntInf.int)), "foo_handshake_sender_lock",
      E_const zero_int)))),
                           Seq (Test (E_un (Uo_not,
     E_bin (Bo_eq, E_var "bar_msg",
             E_const (Int_of_integer (1 : IntInf.int))))),
                                 Skip)))))))))),
                  Seq (Test (E_un (Uo_not,
                                    E_bin (Bo_eq, E_var "bar_parent",
    E_const zero_int))),
                        Or (Seq (Test (E_bin
(Bo_eq, E_var "bar_parent", E_const (Int_of_integer (3 : IntInf.int)))),
                                  Seq (Assign
 (E_un (Uo_not, E_var "baz_handshake_sender_lock"), "baz_handshake_sender_lock",
   E_const (Int_of_integer (1 : IntInf.int))),
Seq (Seq (Assign
            (E_const (Int_of_integer (1 : IntInf.int)), "baz_handshake_message",
              E_var "bar_msg"),
           Seq (Assign
                  (E_const (Int_of_integer (1 : IntInf.int)),
                    "baz_handshake_sent",
                    E_const (Int_of_integer (1 : IntInf.int))),
                 Seq (Test (E_var "baz_handshake_received"),
                       Assign
                         (E_const (Int_of_integer (1 : IntInf.int)),
                           "baz_handshake_received", E_const zero_int)))),
      Assign
        (E_const (Int_of_integer (1 : IntInf.int)), "baz_handshake_sender_lock",
          E_const zero_int)))),
                             Seq (Test (E_un
 (Uo_not,
   E_bin (Bo_eq, E_var "bar_parent",
           E_const (Int_of_integer (3 : IntInf.int))))),
                                   Or (Seq
 (Test (E_bin
         (Bo_eq, E_var "bar_parent",
           E_const (Int_of_integer (2 : IntInf.int)))),
   Seq (Assign
          (E_un (Uo_not, E_var "bar_handshake_sender_lock"),
            "bar_handshake_sender_lock",
            E_const (Int_of_integer (1 : IntInf.int))),
         Seq (Seq (Assign
                     (E_const (Int_of_integer (1 : IntInf.int)),
                       "bar_handshake_message", E_var "bar_msg"),
                    Seq (Assign
                           (E_const (Int_of_integer (1 : IntInf.int)),
                             "bar_handshake_sent",
                             E_const (Int_of_integer (1 : IntInf.int))),
                          Seq (Test (E_var "bar_handshake_received"),
                                Assign
                                  (E_const (Int_of_integer (1 : IntInf.int)),
                                    "bar_handshake_received",
                                    E_const zero_int)))),
               Assign
                 (E_const (Int_of_integer (1 : IntInf.int)),
                   "bar_handshake_sender_lock", E_const zero_int)))),
Seq (Test (E_un (Uo_not,
                  E_bin (Bo_eq, E_var "bar_parent",
                          E_const (Int_of_integer (2 : IntInf.int))))),
      Or (Seq (Test (E_bin
                      (Bo_eq, E_var "bar_parent",
                        E_const (Int_of_integer (1 : IntInf.int)))),
                Seq (Assign
                       (E_un (Uo_not, E_var "foo_handshake_sender_lock"),
                         "foo_handshake_sender_lock",
                         E_const (Int_of_integer (1 : IntInf.int))),
                      Seq (Seq (Assign
                                  (E_const (Int_of_integer (1 : IntInf.int)),
                                    "foo_handshake_message", E_var "bar_msg"),
                                 Seq (Assign
(E_const (Int_of_integer (1 : IntInf.int)), "foo_handshake_sent",
  E_const (Int_of_integer (1 : IntInf.int))),
                                       Seq
 (Test (E_var "foo_handshake_received"),
   Assign
     (E_const (Int_of_integer (1 : IntInf.int)), "foo_handshake_received",
       E_const zero_int)))),
                            Assign
                              (E_const (Int_of_integer (1 : IntInf.int)),
                                "foo_handshake_sender_lock",
                                E_const zero_int)))),
           Seq (Test (E_un (Uo_not,
                             E_bin (Bo_eq, E_var "bar_parent",
                                     E_const
                                       (Int_of_integer (1 : IntInf.int))))),
                 Skip)))))))),
              Assign
                (E_const (Int_of_integer (1 : IntInf.int)), "bar_parent",
                  E_var "bar_msg")))))),
                           Seq (Test (E_un (Uo_not,
     E_var "bar_handshake_sent")),
                                 Or (Seq (Test (E_var "bar_idle"),
   Seq (Assign
          (E_const (Int_of_integer (1 : IntInf.int)), "bar_idle",
            E_const zero_int),
         Seq (Assign
                (E_const (Int_of_integer (1 : IntInf.int)), "bar_requesting",
                  E_const (Int_of_integer (1 : IntInf.int))),
               Or (Seq (Test (E_bin
                               (Bo_eq, E_var "bar_parent", E_const zero_int)),
                         Skip),
                    Seq (Test (E_un (Uo_not,
                                      E_bin
(Bo_eq, E_var "bar_parent", E_const zero_int))),
                          Seq (Or (Seq (Test
  (E_bin
    (Bo_eq, E_var "bar_parent", E_const (Int_of_integer (3 : IntInf.int)))),
 Seq (Assign
        (E_un (Uo_not, E_var "baz_handshake_sender_lock"),
          "baz_handshake_sender_lock",
          E_const (Int_of_integer (1 : IntInf.int))),
       Seq (Seq (Assign
                   (E_const (Int_of_integer (1 : IntInf.int)),
                     "baz_handshake_message",
                     E_const (Int_of_integer (2 : IntInf.int))),
                  Seq (Assign
                         (E_const (Int_of_integer (1 : IntInf.int)),
                           "baz_handshake_sent",
                           E_const (Int_of_integer (1 : IntInf.int))),
                        Seq (Test (E_var "baz_handshake_received"),
                              Assign
                                (E_const (Int_of_integer (1 : IntInf.int)),
                                  "baz_handshake_received",
                                  E_const zero_int)))),
             Assign
               (E_const (Int_of_integer (1 : IntInf.int)),
                 "baz_handshake_sender_lock", E_const zero_int)))),
                                    Seq (Test
   (E_un (Uo_not,
           E_bin (Bo_eq, E_var "bar_parent",
                   E_const (Int_of_integer (3 : IntInf.int))))),
  Or (Seq (Test (E_bin
                  (Bo_eq, E_var "bar_parent",
                    E_const (Int_of_integer (2 : IntInf.int)))),
            Seq (Assign
                   (E_un (Uo_not, E_var "bar_handshake_sender_lock"),
                     "bar_handshake_sender_lock",
                     E_const (Int_of_integer (1 : IntInf.int))),
                  Seq (Seq (Assign
                              (E_const (Int_of_integer (1 : IntInf.int)),
                                "bar_handshake_message",
                                E_const (Int_of_integer (2 : IntInf.int))),
                             Seq (Assign
                                    (E_const (Int_of_integer (1 : IntInf.int)),
                                      "bar_handshake_sent",
                                      E_const
(Int_of_integer (1 : IntInf.int))),
                                   Seq (Test (E_var "bar_handshake_received"),
 Assign
   (E_const (Int_of_integer (1 : IntInf.int)), "bar_handshake_received",
     E_const zero_int)))),
                        Assign
                          (E_const (Int_of_integer (1 : IntInf.int)),
                            "bar_handshake_sender_lock", E_const zero_int)))),
       Seq (Test (E_un (Uo_not,
                         E_bin (Bo_eq, E_var "bar_parent",
                                 E_const (Int_of_integer (2 : IntInf.int))))),
             Or (Seq (Test (E_bin
                             (Bo_eq, E_var "bar_parent",
                               E_const (Int_of_integer (1 : IntInf.int)))),
                       Seq (Assign
                              (E_un (Uo_not, E_var "foo_handshake_sender_lock"),
                                "foo_handshake_sender_lock",
                                E_const (Int_of_integer (1 : IntInf.int))),
                             Seq (Seq (Assign
 (E_const (Int_of_integer (1 : IntInf.int)), "foo_handshake_message",
   E_const (Int_of_integer (2 : IntInf.int))),
Seq (Assign
       (E_const (Int_of_integer (1 : IntInf.int)), "foo_handshake_sent",
         E_const (Int_of_integer (1 : IntInf.int))),
      Seq (Test (E_var "foo_handshake_received"),
            Assign
              (E_const (Int_of_integer (1 : IntInf.int)),
                "foo_handshake_received", E_const zero_int)))),
                                   Assign
                                     (E_const (Int_of_integer (1 : IntInf.int)),
                                       "foo_handshake_sender_lock",
                                       E_const zero_int)))),
                  Seq (Test (E_un (Uo_not,
                                    E_bin (Bo_eq, E_var "bar_parent",
    E_const (Int_of_integer (1 : IntInf.int))))),
                        Skip)))))),
                                Assign
                                  (E_const (Int_of_integer (1 : IntInf.int)),
                                    "bar_parent", E_const zero_int))))))),
                                      Seq
(Test (E_un (Uo_not, E_var "bar_idle")),
  Or (Seq (Test (E_var "bar_token"),
            Seq (Seq (Assign
                        (E_const (Int_of_integer (1 : IntInf.int)),
                          "bar_requesting", E_const zero_int),
                       Or (Seq (Test (E_bin
                                       (Bo_eq, E_var "bar_next",
 E_const zero_int)),
                                 Skip),
                            Seq (Test (E_un
(Uo_not, E_bin (Bo_eq, E_var "bar_next", E_const zero_int))),
                                  Seq (Assign
 (E_const (Int_of_integer (1 : IntInf.int)), "bar_token", E_const zero_int),
Seq (Or (Seq (Test (E_bin
                     (Bo_eq, E_var "bar_next",
                       E_const (Int_of_integer (3 : IntInf.int)))),
               Seq (Assign
                      (E_un (Uo_not, E_var "baz_handshake_sender_lock"),
                        "baz_handshake_sender_lock",
                        E_const (Int_of_integer (1 : IntInf.int))),
                     Seq (Seq (Assign
                                 (E_const (Int_of_integer (1 : IntInf.int)),
                                   "baz_handshake_message", E_const zero_int),
                                Seq (Assign
                                       (E_const
  (Int_of_integer (1 : IntInf.int)),
 "baz_handshake_sent", E_const (Int_of_integer (1 : IntInf.int))),
                                      Seq
(Test (E_var "baz_handshake_received"),
  Assign
    (E_const (Int_of_integer (1 : IntInf.int)), "baz_handshake_received",
      E_const zero_int)))),
                           Assign
                             (E_const (Int_of_integer (1 : IntInf.int)),
                               "baz_handshake_sender_lock",
                               E_const zero_int)))),
          Seq (Test (E_un (Uo_not,
                            E_bin (Bo_eq, E_var "bar_next",
                                    E_const
                                      (Int_of_integer (3 : IntInf.int))))),
                Or (Seq (Test (E_bin
                                (Bo_eq, E_var "bar_next",
                                  E_const (Int_of_integer (2 : IntInf.int)))),
                          Seq (Assign
                                 (E_un (Uo_not,
 E_var "bar_handshake_sender_lock"),
                                   "bar_handshake_sender_lock",
                                   E_const (Int_of_integer (1 : IntInf.int))),
                                Seq (Seq (Assign
    (E_const (Int_of_integer (1 : IntInf.int)), "bar_handshake_message",
      E_const zero_int),
   Seq (Assign
          (E_const (Int_of_integer (1 : IntInf.int)), "bar_handshake_sent",
            E_const (Int_of_integer (1 : IntInf.int))),
         Seq (Test (E_var "bar_handshake_received"),
               Assign
                 (E_const (Int_of_integer (1 : IntInf.int)),
                   "bar_handshake_received", E_const zero_int)))),
                                      Assign
(E_const (Int_of_integer (1 : IntInf.int)), "bar_handshake_sender_lock",
  E_const zero_int)))),
                     Seq (Test (E_un (Uo_not,
                                       E_bin
 (Bo_eq, E_var "bar_next", E_const (Int_of_integer (2 : IntInf.int))))),
                           Or (Seq (Test (E_bin
   (Bo_eq, E_var "bar_next", E_const (Int_of_integer (1 : IntInf.int)))),
                                     Seq (Assign
    (E_un (Uo_not, E_var "foo_handshake_sender_lock"),
      "foo_handshake_sender_lock", E_const (Int_of_integer (1 : IntInf.int))),
   Seq (Seq (Assign
               (E_const (Int_of_integer (1 : IntInf.int)),
                 "foo_handshake_message", E_const zero_int),
              Seq (Assign
                     (E_const (Int_of_integer (1 : IntInf.int)),
                       "foo_handshake_sent",
                       E_const (Int_of_integer (1 : IntInf.int))),
                    Seq (Test (E_var "foo_handshake_received"),
                          Assign
                            (E_const (Int_of_integer (1 : IntInf.int)),
                              "foo_handshake_received", E_const zero_int)))),
         Assign
           (E_const (Int_of_integer (1 : IntInf.int)),
             "foo_handshake_sender_lock", E_const zero_int)))),
                                Seq (Test (E_un
    (Uo_not,
      E_bin (Bo_eq, E_var "bar_next",
              E_const (Int_of_integer (1 : IntInf.int))))),
                                      Skip)))))),
      Assign
        (E_const (Int_of_integer (1 : IntInf.int)), "bar_next",
          E_const zero_int)))))),
                  Assign
                    (E_const (Int_of_integer (1 : IntInf.int)), "bar_idle",
                      E_const (Int_of_integer (1 : IntInf.int))))),
       Seq (Test (E_un (Uo_not, E_var "bar_token")), Skip)))))))),
           ["bar_msg", "bar_idle", "bar_token", "bar_requesting", "bar_parent",
             "bar_next"],
           ()),
       Proc_decl_ext
         (Seq (Seq (Seq (Assign
                           (E_const (Int_of_integer (1 : IntInf.int)),
                             "baz_handshake_sender_lock", E_const zero_int),
                          Seq (Assign
                                 (E_const (Int_of_integer (1 : IntInf.int)),
                                   "baz_handshake_receiver_lock",
                                   E_const zero_int),
                                Seq (Assign
                                       (E_const
  (Int_of_integer (1 : IntInf.int)),
 "baz_handshake_message", E_const zero_int),
                                      Seq
(Assign
   (E_const (Int_of_integer (1 : IntInf.int)), "baz_handshake_sent",
     E_const zero_int),
  Assign
    (E_const (Int_of_integer (1 : IntInf.int)), "baz_handshake_received",
      E_const zero_int))))),
                     Seq (Assign
                            (E_const (Int_of_integer (1 : IntInf.int)),
                              "baz_msg", E_const zero_int),
                           Seq (Assign
                                  (E_const (Int_of_integer (1 : IntInf.int)),
                                    "baz_idle",
                                    E_const (Int_of_integer (1 : IntInf.int))),
                                 Seq (Assign
(E_const (Int_of_integer (1 : IntInf.int)), "baz_parent",
  E_const (Int_of_integer (1 : IntInf.int))),
                                       Seq
 (Assign
    (E_const (Int_of_integer (1 : IntInf.int)), "baz_next", E_const zero_int),
   Seq (Assign
          (E_const (Int_of_integer (1 : IntInf.int)), "baz_requesting",
            E_const zero_int),
         Assign
           (E_const (Int_of_integer (1 : IntInf.int)), "baz_token",
             E_const zero_int))))))),
                Loop (Or (Seq (Test (E_var "baz_handshake_sent"),
                                Seq (Seq (Assign
    (E_un (Uo_not, E_var "baz_handshake_receiver_lock"),
      "baz_handshake_receiver_lock", E_const (Int_of_integer (1 : IntInf.int))),
   Seq (Seq (Test (E_var "baz_handshake_sent"),
              Seq (Assign
                     (E_const (Int_of_integer (1 : IntInf.int)),
                       "baz_handshake_sent", E_const zero_int),
                    Seq (Assign
                           (E_const (Int_of_integer (1 : IntInf.int)),
                             "baz_msg", E_var "baz_handshake_message"),
                          Assign
                            (E_const (Int_of_integer (1 : IntInf.int)),
                              "baz_handshake_received",
                              E_const (Int_of_integer (1 : IntInf.int)))))),
         Assign
           (E_const (Int_of_integer (1 : IntInf.int)),
             "baz_handshake_receiver_lock", E_const zero_int))),
                                      Or
(Seq (Test (E_bin (Bo_eq, E_var "baz_msg", E_const zero_int)),
       Assign
         (E_const (Int_of_integer (1 : IntInf.int)), "baz_token",
           E_const (Int_of_integer (1 : IntInf.int)))),
  Seq (Test (E_un (Uo_not, E_bin (Bo_eq, E_var "baz_msg", E_const zero_int))),
        Seq (Or (Seq (Test (E_bin
                             (Bo_eq, E_var "baz_parent", E_const zero_int)),
                       Or (Seq (Test (E_var "baz_requesting"),
                                 Assign
                                   (E_const (Int_of_integer (1 : IntInf.int)),
                                     "baz_next", E_var "baz_msg")),
                            Seq (Test (E_un (Uo_not, E_var "baz_requesting")),
                                  Seq (Assign
 (E_const (Int_of_integer (1 : IntInf.int)), "baz_token", E_const zero_int),
Or (Seq (Test (E_bin
                (Bo_eq, E_var "baz_msg",
                  E_const (Int_of_integer (3 : IntInf.int)))),
          Seq (Assign
                 (E_un (Uo_not, E_var "baz_handshake_sender_lock"),
                   "baz_handshake_sender_lock",
                   E_const (Int_of_integer (1 : IntInf.int))),
                Seq (Seq (Assign
                            (E_const (Int_of_integer (1 : IntInf.int)),
                              "baz_handshake_message", E_const zero_int),
                           Seq (Assign
                                  (E_const (Int_of_integer (1 : IntInf.int)),
                                    "baz_handshake_sent",
                                    E_const (Int_of_integer (1 : IntInf.int))),
                                 Seq (Test (E_var "baz_handshake_received"),
                                       Assign
 (E_const (Int_of_integer (1 : IntInf.int)), "baz_handshake_received",
   E_const zero_int)))),
                      Assign
                        (E_const (Int_of_integer (1 : IntInf.int)),
                          "baz_handshake_sender_lock", E_const zero_int)))),
     Seq (Test (E_un (Uo_not,
                       E_bin (Bo_eq, E_var "baz_msg",
                               E_const (Int_of_integer (3 : IntInf.int))))),
           Or (Seq (Test (E_bin
                           (Bo_eq, E_var "baz_msg",
                             E_const (Int_of_integer (2 : IntInf.int)))),
                     Seq (Assign
                            (E_un (Uo_not, E_var "bar_handshake_sender_lock"),
                              "bar_handshake_sender_lock",
                              E_const (Int_of_integer (1 : IntInf.int))),
                           Seq (Seq (Assign
                                       (E_const
  (Int_of_integer (1 : IntInf.int)),
 "bar_handshake_message", E_const zero_int),
                                      Seq
(Assign
   (E_const (Int_of_integer (1 : IntInf.int)), "bar_handshake_sent",
     E_const (Int_of_integer (1 : IntInf.int))),
  Seq (Test (E_var "bar_handshake_received"),
        Assign
          (E_const (Int_of_integer (1 : IntInf.int)), "bar_handshake_received",
            E_const zero_int)))),
                                 Assign
                                   (E_const (Int_of_integer (1 : IntInf.int)),
                                     "bar_handshake_sender_lock",
                                     E_const zero_int)))),
                Seq (Test (E_un (Uo_not,
                                  E_bin (Bo_eq, E_var "baz_msg",
  E_const (Int_of_integer (2 : IntInf.int))))),
                      Or (Seq (Test (E_bin
                                      (Bo_eq, E_var "baz_msg",
E_const (Int_of_integer (1 : IntInf.int)))),
                                Seq (Assign
                                       (E_un
  (Uo_not, E_var "foo_handshake_sender_lock"),
 "foo_handshake_sender_lock", E_const (Int_of_integer (1 : IntInf.int))),
                                      Seq
(Seq (Assign
        (E_const (Int_of_integer (1 : IntInf.int)), "foo_handshake_message",
          E_const zero_int),
       Seq (Assign
              (E_const (Int_of_integer (1 : IntInf.int)), "foo_handshake_sent",
                E_const (Int_of_integer (1 : IntInf.int))),
             Seq (Test (E_var "foo_handshake_received"),
                   Assign
                     (E_const (Int_of_integer (1 : IntInf.int)),
                       "foo_handshake_received", E_const zero_int)))),
  Assign
    (E_const (Int_of_integer (1 : IntInf.int)), "foo_handshake_sender_lock",
      E_const zero_int)))),
                           Seq (Test (E_un (Uo_not,
     E_bin (Bo_eq, E_var "baz_msg",
             E_const (Int_of_integer (1 : IntInf.int))))),
                                 Skip)))))))))),
                  Seq (Test (E_un (Uo_not,
                                    E_bin (Bo_eq, E_var "baz_parent",
    E_const zero_int))),
                        Or (Seq (Test (E_bin
(Bo_eq, E_var "baz_parent", E_const (Int_of_integer (3 : IntInf.int)))),
                                  Seq (Assign
 (E_un (Uo_not, E_var "baz_handshake_sender_lock"), "baz_handshake_sender_lock",
   E_const (Int_of_integer (1 : IntInf.int))),
Seq (Seq (Assign
            (E_const (Int_of_integer (1 : IntInf.int)), "baz_handshake_message",
              E_var "baz_msg"),
           Seq (Assign
                  (E_const (Int_of_integer (1 : IntInf.int)),
                    "baz_handshake_sent",
                    E_const (Int_of_integer (1 : IntInf.int))),
                 Seq (Test (E_var "baz_handshake_received"),
                       Assign
                         (E_const (Int_of_integer (1 : IntInf.int)),
                           "baz_handshake_received", E_const zero_int)))),
      Assign
        (E_const (Int_of_integer (1 : IntInf.int)), "baz_handshake_sender_lock",
          E_const zero_int)))),
                             Seq (Test (E_un
 (Uo_not,
   E_bin (Bo_eq, E_var "baz_parent",
           E_const (Int_of_integer (3 : IntInf.int))))),
                                   Or (Seq
 (Test (E_bin
         (Bo_eq, E_var "baz_parent",
           E_const (Int_of_integer (2 : IntInf.int)))),
   Seq (Assign
          (E_un (Uo_not, E_var "bar_handshake_sender_lock"),
            "bar_handshake_sender_lock",
            E_const (Int_of_integer (1 : IntInf.int))),
         Seq (Seq (Assign
                     (E_const (Int_of_integer (1 : IntInf.int)),
                       "bar_handshake_message", E_var "baz_msg"),
                    Seq (Assign
                           (E_const (Int_of_integer (1 : IntInf.int)),
                             "bar_handshake_sent",
                             E_const (Int_of_integer (1 : IntInf.int))),
                          Seq (Test (E_var "bar_handshake_received"),
                                Assign
                                  (E_const (Int_of_integer (1 : IntInf.int)),
                                    "bar_handshake_received",
                                    E_const zero_int)))),
               Assign
                 (E_const (Int_of_integer (1 : IntInf.int)),
                   "bar_handshake_sender_lock", E_const zero_int)))),
Seq (Test (E_un (Uo_not,
                  E_bin (Bo_eq, E_var "baz_parent",
                          E_const (Int_of_integer (2 : IntInf.int))))),
      Or (Seq (Test (E_bin
                      (Bo_eq, E_var "baz_parent",
                        E_const (Int_of_integer (1 : IntInf.int)))),
                Seq (Assign
                       (E_un (Uo_not, E_var "foo_handshake_sender_lock"),
                         "foo_handshake_sender_lock",
                         E_const (Int_of_integer (1 : IntInf.int))),
                      Seq (Seq (Assign
                                  (E_const (Int_of_integer (1 : IntInf.int)),
                                    "foo_handshake_message", E_var "baz_msg"),
                                 Seq (Assign
(E_const (Int_of_integer (1 : IntInf.int)), "foo_handshake_sent",
  E_const (Int_of_integer (1 : IntInf.int))),
                                       Seq
 (Test (E_var "foo_handshake_received"),
   Assign
     (E_const (Int_of_integer (1 : IntInf.int)), "foo_handshake_received",
       E_const zero_int)))),
                            Assign
                              (E_const (Int_of_integer (1 : IntInf.int)),
                                "foo_handshake_sender_lock",
                                E_const zero_int)))),
           Seq (Test (E_un (Uo_not,
                             E_bin (Bo_eq, E_var "baz_parent",
                                     E_const
                                       (Int_of_integer (1 : IntInf.int))))),
                 Skip)))))))),
              Assign
                (E_const (Int_of_integer (1 : IntInf.int)), "baz_parent",
                  E_var "baz_msg")))))),
                           Seq (Test (E_un (Uo_not,
     E_var "baz_handshake_sent")),
                                 Or (Seq (Test (E_var "baz_idle"),
   Seq (Assign
          (E_const (Int_of_integer (1 : IntInf.int)), "baz_idle",
            E_const zero_int),
         Seq (Assign
                (E_const (Int_of_integer (1 : IntInf.int)), "baz_requesting",
                  E_const (Int_of_integer (1 : IntInf.int))),
               Or (Seq (Test (E_bin
                               (Bo_eq, E_var "baz_parent", E_const zero_int)),
                         Skip),
                    Seq (Test (E_un (Uo_not,
                                      E_bin
(Bo_eq, E_var "baz_parent", E_const zero_int))),
                          Seq (Or (Seq (Test
  (E_bin
    (Bo_eq, E_var "baz_parent", E_const (Int_of_integer (3 : IntInf.int)))),
 Seq (Assign
        (E_un (Uo_not, E_var "baz_handshake_sender_lock"),
          "baz_handshake_sender_lock",
          E_const (Int_of_integer (1 : IntInf.int))),
       Seq (Seq (Assign
                   (E_const (Int_of_integer (1 : IntInf.int)),
                     "baz_handshake_message",
                     E_const (Int_of_integer (3 : IntInf.int))),
                  Seq (Assign
                         (E_const (Int_of_integer (1 : IntInf.int)),
                           "baz_handshake_sent",
                           E_const (Int_of_integer (1 : IntInf.int))),
                        Seq (Test (E_var "baz_handshake_received"),
                              Assign
                                (E_const (Int_of_integer (1 : IntInf.int)),
                                  "baz_handshake_received",
                                  E_const zero_int)))),
             Assign
               (E_const (Int_of_integer (1 : IntInf.int)),
                 "baz_handshake_sender_lock", E_const zero_int)))),
                                    Seq (Test
   (E_un (Uo_not,
           E_bin (Bo_eq, E_var "baz_parent",
                   E_const (Int_of_integer (3 : IntInf.int))))),
  Or (Seq (Test (E_bin
                  (Bo_eq, E_var "baz_parent",
                    E_const (Int_of_integer (2 : IntInf.int)))),
            Seq (Assign
                   (E_un (Uo_not, E_var "bar_handshake_sender_lock"),
                     "bar_handshake_sender_lock",
                     E_const (Int_of_integer (1 : IntInf.int))),
                  Seq (Seq (Assign
                              (E_const (Int_of_integer (1 : IntInf.int)),
                                "bar_handshake_message",
                                E_const (Int_of_integer (3 : IntInf.int))),
                             Seq (Assign
                                    (E_const (Int_of_integer (1 : IntInf.int)),
                                      "bar_handshake_sent",
                                      E_const
(Int_of_integer (1 : IntInf.int))),
                                   Seq (Test (E_var "bar_handshake_received"),
 Assign
   (E_const (Int_of_integer (1 : IntInf.int)), "bar_handshake_received",
     E_const zero_int)))),
                        Assign
                          (E_const (Int_of_integer (1 : IntInf.int)),
                            "bar_handshake_sender_lock", E_const zero_int)))),
       Seq (Test (E_un (Uo_not,
                         E_bin (Bo_eq, E_var "baz_parent",
                                 E_const (Int_of_integer (2 : IntInf.int))))),
             Or (Seq (Test (E_bin
                             (Bo_eq, E_var "baz_parent",
                               E_const (Int_of_integer (1 : IntInf.int)))),
                       Seq (Assign
                              (E_un (Uo_not, E_var "foo_handshake_sender_lock"),
                                "foo_handshake_sender_lock",
                                E_const (Int_of_integer (1 : IntInf.int))),
                             Seq (Seq (Assign
 (E_const (Int_of_integer (1 : IntInf.int)), "foo_handshake_message",
   E_const (Int_of_integer (3 : IntInf.int))),
Seq (Assign
       (E_const (Int_of_integer (1 : IntInf.int)), "foo_handshake_sent",
         E_const (Int_of_integer (1 : IntInf.int))),
      Seq (Test (E_var "foo_handshake_received"),
            Assign
              (E_const (Int_of_integer (1 : IntInf.int)),
                "foo_handshake_received", E_const zero_int)))),
                                   Assign
                                     (E_const (Int_of_integer (1 : IntInf.int)),
                                       "foo_handshake_sender_lock",
                                       E_const zero_int)))),
                  Seq (Test (E_un (Uo_not,
                                    E_bin (Bo_eq, E_var "baz_parent",
    E_const (Int_of_integer (1 : IntInf.int))))),
                        Skip)))))),
                                Assign
                                  (E_const (Int_of_integer (1 : IntInf.int)),
                                    "baz_parent", E_const zero_int))))))),
                                      Seq
(Test (E_un (Uo_not, E_var "baz_idle")),
  Or (Seq (Test (E_var "baz_token"),
            Seq (Seq (Assign
                        (E_const (Int_of_integer (1 : IntInf.int)),
                          "baz_requesting", E_const zero_int),
                       Or (Seq (Test (E_bin
                                       (Bo_eq, E_var "baz_next",
 E_const zero_int)),
                                 Skip),
                            Seq (Test (E_un
(Uo_not, E_bin (Bo_eq, E_var "baz_next", E_const zero_int))),
                                  Seq (Assign
 (E_const (Int_of_integer (1 : IntInf.int)), "baz_token", E_const zero_int),
Seq (Or (Seq (Test (E_bin
                     (Bo_eq, E_var "baz_next",
                       E_const (Int_of_integer (3 : IntInf.int)))),
               Seq (Assign
                      (E_un (Uo_not, E_var "baz_handshake_sender_lock"),
                        "baz_handshake_sender_lock",
                        E_const (Int_of_integer (1 : IntInf.int))),
                     Seq (Seq (Assign
                                 (E_const (Int_of_integer (1 : IntInf.int)),
                                   "baz_handshake_message", E_const zero_int),
                                Seq (Assign
                                       (E_const
  (Int_of_integer (1 : IntInf.int)),
 "baz_handshake_sent", E_const (Int_of_integer (1 : IntInf.int))),
                                      Seq
(Test (E_var "baz_handshake_received"),
  Assign
    (E_const (Int_of_integer (1 : IntInf.int)), "baz_handshake_received",
      E_const zero_int)))),
                           Assign
                             (E_const (Int_of_integer (1 : IntInf.int)),
                               "baz_handshake_sender_lock",
                               E_const zero_int)))),
          Seq (Test (E_un (Uo_not,
                            E_bin (Bo_eq, E_var "baz_next",
                                    E_const
                                      (Int_of_integer (3 : IntInf.int))))),
                Or (Seq (Test (E_bin
                                (Bo_eq, E_var "baz_next",
                                  E_const (Int_of_integer (2 : IntInf.int)))),
                          Seq (Assign
                                 (E_un (Uo_not,
 E_var "bar_handshake_sender_lock"),
                                   "bar_handshake_sender_lock",
                                   E_const (Int_of_integer (1 : IntInf.int))),
                                Seq (Seq (Assign
    (E_const (Int_of_integer (1 : IntInf.int)), "bar_handshake_message",
      E_const zero_int),
   Seq (Assign
          (E_const (Int_of_integer (1 : IntInf.int)), "bar_handshake_sent",
            E_const (Int_of_integer (1 : IntInf.int))),
         Seq (Test (E_var "bar_handshake_received"),
               Assign
                 (E_const (Int_of_integer (1 : IntInf.int)),
                   "bar_handshake_received", E_const zero_int)))),
                                      Assign
(E_const (Int_of_integer (1 : IntInf.int)), "bar_handshake_sender_lock",
  E_const zero_int)))),
                     Seq (Test (E_un (Uo_not,
                                       E_bin
 (Bo_eq, E_var "baz_next", E_const (Int_of_integer (2 : IntInf.int))))),
                           Or (Seq (Test (E_bin
   (Bo_eq, E_var "baz_next", E_const (Int_of_integer (1 : IntInf.int)))),
                                     Seq (Assign
    (E_un (Uo_not, E_var "foo_handshake_sender_lock"),
      "foo_handshake_sender_lock", E_const (Int_of_integer (1 : IntInf.int))),
   Seq (Seq (Assign
               (E_const (Int_of_integer (1 : IntInf.int)),
                 "foo_handshake_message", E_const zero_int),
              Seq (Assign
                     (E_const (Int_of_integer (1 : IntInf.int)),
                       "foo_handshake_sent",
                       E_const (Int_of_integer (1 : IntInf.int))),
                    Seq (Test (E_var "foo_handshake_received"),
                          Assign
                            (E_const (Int_of_integer (1 : IntInf.int)),
                              "foo_handshake_received", E_const zero_int)))),
         Assign
           (E_const (Int_of_integer (1 : IntInf.int)),
             "foo_handshake_sender_lock", E_const zero_int)))),
                                Seq (Test (E_un
    (Uo_not,
      E_bin (Bo_eq, E_var "baz_next",
              E_const (Int_of_integer (1 : IntInf.int))))),
                                      Skip)))))),
      Assign
        (E_const (Int_of_integer (1 : IntInf.int)), "baz_next",
          E_const zero_int)))))),
                  Assign
                    (E_const (Int_of_integer (1 : IntInf.int)), "baz_idle",
                      E_const (Int_of_integer (1 : IntInf.int))))),
       Seq (Test (E_un (Uo_not, E_var "baz_token")), Skip)))))))),
           ["baz_msg", "baz_idle", "baz_token", "baz_requesting", "baz_parent",
             "baz_next"],
           ())],
      ["foo_handshake_sender_lock", "foo_handshake_receiver_lock",
        "foo_handshake_message", "foo_handshake_sent", "foo_handshake_received",
        "bar_handshake_sender_lock", "bar_handshake_receiver_lock",
        "bar_handshake_message", "bar_handshake_sent", "bar_handshake_received",
        "baz_handshake_sender_lock", "baz_handshake_receiver_lock",
        "baz_handshake_message", "baz_handshake_sent",
        "baz_handshake_received"],
      ());

val dflt_cfg : config_l2b * (unit * config_ce) =
  (CFG_L2B_GERTHS, ((), CFG_CE_SCC_GABOW));

fun ltlc_next_free LTLcTrue = true
  | ltlc_next_free LTLcFalse = true
  | ltlc_next_free (LTLcProp q) = true
  | ltlc_next_free (LTLcNeg phi) = ltlc_next_free phi
  | ltlc_next_free (LTLcAnd (phi, psi)) =
    ltlc_next_free phi andalso ltlc_next_free psi
  | ltlc_next_free (LTLcOr (phi, psi)) =
    ltlc_next_free phi andalso ltlc_next_free psi
  | ltlc_next_free (LTLcImplies (phi, psi)) =
    ltlc_next_free phi andalso ltlc_next_free psi
  | ltlc_next_free (LTLcIff (phi, psi)) =
    ltlc_next_free phi andalso ltlc_next_free psi
  | ltlc_next_free (LTLcNext phi) = false
  | ltlc_next_free (LTLcFinal phi) = ltlc_next_free phi
  | ltlc_next_free (LTLcGlobal phi) = ltlc_next_free phi
  | ltlc_next_free (LTLcUntil (phi, psi)) =
    ltlc_next_free phi andalso ltlc_next_free psi
  | ltlc_next_free (LTLcRelease (phi, psi)) =
    ltlc_next_free phi andalso ltlc_next_free psi;

fun balance (Branch (R, a, w, x, b)) s t (Branch (R, c, y, z, d)) =
  Branch (R, Branch (B, a, w, x, b), s, t, Branch (B, c, y, z, d))
  | balance (Branch (R, Branch (R, a, w, x, b), s, t, c)) y z Empty =
    Branch (R, Branch (B, a, w, x, b), s, t, Branch (B, c, y, z, Empty))
  | balance (Branch (R, Branch (R, a, w, x, b), s, t, c)) y z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, Branch (B, a, w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance (Branch (R, Empty, w, x, Branch (R, b, s, t, c))) y z Empty =
    Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, Empty))
  | balance
    (Branch (R, Branch (B, va, vb, vc, vd), w, x, Branch (R, b, s, t, c))) y z
    Empty =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, Empty))
  | balance (Branch (R, Empty, w, x, Branch (R, b, s, t, c))) y z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, Branch (B, Empty, w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance
    (Branch (R, Branch (B, ve, vf, vg, vh), w, x, Branch (R, b, s, t, c))) y z
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (R, Branch (B, Branch (B, ve, vf, vg, vh), w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance Empty w x (Branch (R, b, s, t, Branch (R, c, y, z, d))) =
    Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, d))
  | balance (Branch (B, va, vb, vc, vd)) w x
    (Branch (R, b, s, t, Branch (R, c, y, z, d))) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, d))
  | balance Empty w x (Branch (R, Branch (R, b, s, t, c), y, z, Empty)) =
    Branch (R, Branch (B, Empty, w, x, b), s, t, Branch (B, c, y, z, Empty))
  | balance Empty w x
    (Branch (R, Branch (R, b, s, t, c), y, z, Branch (B, va, vb, vc, vd))) =
    Branch
      (R, Branch (B, Empty, w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, va, vb, vc, vd)))
  | balance (Branch (B, va, vb, vc, vd)) w x
    (Branch (R, Branch (R, b, s, t, c), y, z, Empty)) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, Empty))
  | balance (Branch (B, va, vb, vc, vd)) w x
    (Branch (R, Branch (R, b, s, t, c), y, z, Branch (B, ve, vf, vg, vh))) =
    Branch
      (R, Branch (B, Branch (B, va, vb, vc, vd), w, x, b), s, t,
        Branch (B, c, y, z, Branch (B, ve, vf, vg, vh)))
  | balance Empty s t Empty = Branch (B, Empty, s, t, Empty)
  | balance Empty s t (Branch (B, va, vb, vc, vd)) =
    Branch (B, Empty, s, t, Branch (B, va, vb, vc, vd))
  | balance Empty s t (Branch (v, Empty, vb, vc, Empty)) =
    Branch (B, Empty, s, t, Branch (v, Empty, vb, vc, Empty))
  | balance Empty s t (Branch (v, Branch (B, ve, vf, vg, vh), vb, vc, Empty)) =
    Branch
      (B, Empty, s, t, Branch (v, Branch (B, ve, vf, vg, vh), vb, vc, Empty))
  | balance Empty s t (Branch (v, Empty, vb, vc, Branch (B, vf, vg, vh, vi))) =
    Branch
      (B, Empty, s, t, Branch (v, Empty, vb, vc, Branch (B, vf, vg, vh, vi)))
  | balance Empty s t
    (Branch (v, Branch (B, ve, vj, vk, vl), vb, vc, Branch (B, vf, vg, vh, vi)))
    = Branch
        (B, Empty, s, t,
          Branch
            (v, Branch (B, ve, vj, vk, vl), vb, vc, Branch (B, vf, vg, vh, vi)))
  | balance (Branch (B, va, vb, vc, vd)) s t Empty =
    Branch (B, Branch (B, va, vb, vc, vd), s, t, Empty)
  | balance (Branch (B, va, vb, vc, vd)) s t (Branch (B, ve, vf, vg, vh)) =
    Branch (B, Branch (B, va, vb, vc, vd), s, t, Branch (B, ve, vf, vg, vh))
  | balance (Branch (B, va, vb, vc, vd)) s t (Branch (v, Empty, vf, vg, Empty))
    = Branch
        (B, Branch (B, va, vb, vc, vd), s, t, Branch (v, Empty, vf, vg, Empty))
  | balance (Branch (B, va, vb, vc, vd)) s t
    (Branch (v, Branch (B, vi, vj, vk, vl), vf, vg, Empty)) =
    Branch
      (B, Branch (B, va, vb, vc, vd), s, t,
        Branch (v, Branch (B, vi, vj, vk, vl), vf, vg, Empty))
  | balance (Branch (B, va, vb, vc, vd)) s t
    (Branch (v, Empty, vf, vg, Branch (B, vj, vk, vl, vm))) =
    Branch
      (B, Branch (B, va, vb, vc, vd), s, t,
        Branch (v, Empty, vf, vg, Branch (B, vj, vk, vl, vm)))
  | balance (Branch (B, va, vb, vc, vd)) s t
    (Branch (v, Branch (B, vi, vn, vo, vp), vf, vg, Branch (B, vj, vk, vl, vm)))
    = Branch
        (B, Branch (B, va, vb, vc, vd), s, t,
          Branch
            (v, Branch (B, vi, vn, vo, vp), vf, vg, Branch (B, vj, vk, vl, vm)))
  | balance (Branch (v, Empty, vb, vc, Empty)) s t Empty =
    Branch (B, Branch (v, Empty, vb, vc, Empty), s, t, Empty)
  | balance (Branch (v, Empty, vb, vc, Branch (B, ve, vf, vg, vh))) s t Empty =
    Branch
      (B, Branch (v, Empty, vb, vc, Branch (B, ve, vf, vg, vh)), s, t, Empty)
  | balance (Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Empty)) s t Empty =
    Branch
      (B, Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Empty), s, t, Empty)
  | balance
    (Branch (v, Branch (B, vf, vg, vh, vi), vb, vc, Branch (B, ve, vj, vk, vl)))
    s t Empty =
    Branch
      (B, Branch
            (v, Branch (B, vf, vg, vh, vi), vb, vc, Branch (B, ve, vj, vk, vl)),
        s, t, Empty)
  | balance (Branch (v, Empty, vf, vg, Empty)) s t (Branch (B, va, vb, vc, vd))
    = Branch
        (B, Branch (v, Empty, vf, vg, Empty), s, t, Branch (B, va, vb, vc, vd))
  | balance (Branch (v, Empty, vf, vg, Branch (B, vi, vj, vk, vl))) s t
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (B, Branch (v, Empty, vf, vg, Branch (B, vi, vj, vk, vl)), s, t,
        Branch (B, va, vb, vc, vd))
  | balance (Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Empty)) s t
    (Branch (B, va, vb, vc, vd)) =
    Branch
      (B, Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Empty), s, t,
        Branch (B, va, vb, vc, vd))
  | balance
    (Branch (v, Branch (B, vj, vk, vl, vm), vf, vg, Branch (B, vi, vn, vo, vp)))
    s t (Branch (B, va, vb, vc, vd)) =
    Branch
      (B, Branch
            (v, Branch (B, vj, vk, vl, vm), vf, vg, Branch (B, vi, vn, vo, vp)),
        s, t, Branch (B, va, vb, vc, vd));

fun rbt_insa A_ f k v Empty = Branch (R, Empty, k, v, Empty)
  | rbt_insa A_ f k v (Branch (B, l, x, y, r)) =
    (if less A_ k x then balance (rbt_insa A_ f k v l) x y r
      else (if less A_ x k then balance l x y (rbt_insa A_ f k v r)
             else Branch (B, l, x, f k y v, r)))
  | rbt_insa A_ f k v (Branch (R, l, x, y, r)) =
    (if less A_ k x then Branch (R, rbt_insa A_ f k v l, x, y, r)
      else (if less A_ x k then Branch (R, l, x, y, rbt_insa A_ f k v r)
             else Branch (R, l, x, f k y v, r)));

fun paint c Empty = Empty
  | paint c (Branch (uu, l, k, v, r)) = Branch (c, l, k, v, r);

fun rbt_insert_with_keya A_ f k v t = paint B (rbt_insa A_ f k v t);

fun rbt_inserta A_ = rbt_insert_with_keya A_ (fn _ => fn _ => fn nv => nv);

fun insert A_ xc xd xe =
  RBT (rbt_inserta ((ord_preorder o preorder_order o order_linorder) A_) xc xd
        (impl_of A_ xe));

fun updatea A_ k v (Mapping t) = Mapping (insert A_ k v t);

fun vi_update x = updatea linorder_literal x;

fun emptya A_ = Mapping (empty A_);

val vi_empty : (string, Word32.word) mapping = emptya linorder_literal;

fun init_valuation_impl vd =
  fold (fn x => vi_update x (Word32.fromInt 0)) vd vi_empty;

fun global_vars (Program_ext (processes, global_vars, more)) = global_vars;

fun processes (Program_ext (processes, global_vars, more)) = processes;

fun local_vars (Proc_decl_ext (body, local_vars, more)) = local_vars;

fun body (Proc_decl_ext (body, local_vars, more)) = body;

fun init_pc_impl pd =
  Local_config_ext
    (body pd, Local_state_impl_ext (init_valuation_impl (local_vars pd), ()),
      ());

fun pid_init_gc_impl prog =
  Pid_global_config_ext
    (map init_pc_impl (processes prog),
      Global_state_impl_ext (init_valuation_impl (global_vars prog), ()), ());

fun test_aprop_impl (AP_isNull x) s =
  (case s x of NONE => false
    | SOME v => ((v : Word32.word) = (Word32.fromInt 0)));

fun vi_lookup x = lookupa linorder_literal x;

fun pid_test_gc_impl is_vis_var gci ap =
  test_aprop_impl ap
    (fn x =>
      (if is_vis_var x then vi_lookup (variablesa (statea gci)) x else NONE));

fun glist_member eq x [] = false
  | glist_member eq x (y :: ys) = eq x y orelse glist_member eq x ys;

fun glist_insert eq x xs = (if glist_member eq x xs then xs else x :: xs);

fun gen_image it1 emp2 ins2 f s1 =
  it1 s1 (fn _ => true) (fn x => ins2 (f x)) emp2;

fun foldli [] c f sigma = sigma
  | foldli (x :: xs) c f sigma =
    (if c sigma then foldli xs c f (f x sigma) else sigma);

fun succ_of_enex_list C_ =
  (fn x => fn xa =>
    let
      val (xb, xc) = x;
    in
      gen_image (fn xd => foldli (id xd)) [] (glist_insert (eq C_))
        (fn xd => xc xd xa) (xb xa)
    end);

fun nat_of_uint32 x = nat_of_integer (Word32.toInt x : IntInf.int);

fun nat_of_hashcode x = nat_of_uint32 x;

fun bounded_hashcode_nat A_ n x = mod_nat (nat_of_hashcode (hashcode A_ x)) n;

fun is_Nil a = (case a of [] => true | _ :: _ => false);

fun stutter_extend_en_list x =
  (fn xa => fn xaa =>
    let
      val xb = xa xaa;
    in
      (if is_Nil xb then [NONE] else map SOME xb)
    end)
    x;

fun set_iterator_image g it = (fn c => fn f => it c (fn x => f (g x)));

fun map_iterator_dom it = set_iterator_image fst it;

fun ssnos_visited_impl_update ssnos_visited_impla
  (Simple_state_nos_impl_ext (ssnos_stack_impl, ssnos_visited_impl, more)) =
  Simple_state_nos_impl_ext
    (ssnos_stack_impl, ssnos_visited_impla ssnos_visited_impl, more);

fun ssnos_stack_impl_update ssnos_stack_impla
  (Simple_state_nos_impl_ext (ssnos_stack_impl, ssnos_visited_impl, more)) =
  Simple_state_nos_impl_ext
    (ssnos_stack_impla ssnos_stack_impl, ssnos_visited_impl, more);

fun ssnos_visited_impl
  (Simple_state_nos_impl_ext (ssnos_stack_impl, ssnos_visited_impl, more)) =
  ssnos_visited_impl;

fun ssnos_stack_impl
  (Simple_state_nos_impl_ext (ssnos_stack_impl, ssnos_visited_impl, more)) =
  ssnos_stack_impl;

fun ras_singleton B_ x = (FArray.IsabelleMapping.array_of_list [x], one B_);

fun ras_is_empty s = equal_nata (snd s) zero_nata;

fun ras_empty B_ uu = (FArray.IsabelleMapping.array_of_list [], zero B_);

fun list_map_update_aux eq k v [] accu = (k, v) :: accu
  | list_map_update_aux eq k v (x :: xs) accu =
    (if eq (fst x) k then (k, v) :: xs @ accu
      else list_map_update_aux eq k v xs (x :: accu));

fun list_map_update eq k v m = list_map_update_aux eq k v m [];

fun list_map_lookup eq uu [] = NONE
  | list_map_lookup eq k (y :: ys) =
    (if eq (fst y) k then SOME (snd y) else list_map_lookup eq k ys);

fun array_length x = (nat_of_integer o FArray.IsabelleMapping.array_length) x;

fun array_set a = FArray.IsabelleMapping.array_set a o integer_of_nat;

fun array_get a = FArray.IsabelleMapping.array_get a o integer_of_nat;

fun is_none (SOME x) = false
  | is_none NONE = true;

fun ahm_update_auxa eq bhc (HashMapb (a, n)) k v =
  let
    val h = bhc (array_length a) k;
    val m = array_get a h;
    val insert = is_none (list_map_lookup eq k m);
  in
    HashMapb
      (array_set a h (list_map_update eq k v m),
        (if insert then plus_nata n one_nata else n))
  end;

fun idx_iteratei_aux get sz i l c f sigma =
  (if equal_nata i zero_nata orelse not (c sigma) then sigma
    else idx_iteratei_aux get sz (minus_nat i one_nata) l c f
           (f (get l (minus_nat sz i)) sigma));

fun idx_iteratei get sz l c f sigma =
  idx_iteratei_aux get (sz l) (sz l) l c f sigma;

fun ahm_iteratei_auxa a c f sigma =
  idx_iteratei array_get array_length a c (fn x => foldli x c f) sigma;

fun ahm_rehash_auxc bhc n kv a =
  let
    val h = bhc n (fst kv);
  in
    array_set a h (kv :: array_get a h)
  end;

fun new_array v = FArray.IsabelleMapping.new_array v o integer_of_nat;

fun ahm_rehash_auxb bhc a sz =
  ahm_iteratei_auxa a (fn _ => true) (ahm_rehash_auxc bhc sz) (new_array [] sz);

fun ahm_rehasha bhc (HashMapb (a, n)) sz =
  HashMapb (ahm_rehash_auxb bhc a sz, n);

val load_factora : nat = nat_of_integer (75 : IntInf.int);

fun ahm_filleda (HashMapb (a, n)) =
  less_eq_nat (times_nata (array_length a) load_factora)
    (times_nata n (nat_of_integer (100 : IntInf.int)));

fun hm_growa (HashMapb (a, n)) =
  plus_nata (times_nata (nat_of_integer (2 : IntInf.int)) (array_length a))
    (nat_of_integer (3 : IntInf.int));

fun ahm_updateb eq bhc k v hm =
  let
    val hma = ahm_update_auxa eq bhc hm k v;
  in
    (if ahm_filleda hma then ahm_rehasha bhc hma (hm_growa hma) else hma)
  end;

fun ahm_lookup_aux eq bhc k a =
  list_map_lookup eq k (array_get a (bhc (array_length a) k));

fun ahm_lookupb eq bhc k (HashMapb (a, uu)) = ahm_lookup_aux eq bhc k a;

fun array_grow a = FArray.IsabelleMapping.array_grow a o integer_of_nat;

fun ras_push x s =
  let
    val (a, n) = s;
    val aa =
      (if equal_nata n (array_length a)
        then array_grow a
               (max ord_nat (nat_of_integer (4 : IntInf.int))
                 (times_nata (nat_of_integer (2 : IntInf.int)) n))
               x
        else a);
    val ab = array_set aa n x;
  in
    (ab, plus_nata n one_nata)
  end;

fun new_hashmap_witha size = HashMapb (new_array [] size, zero_nata);

fun ahm_emptyb def_size = new_hashmap_witha def_size;

fun ras_top s = let
                  val (a, n) = s;
                in
                  array_get a (minus_nat n one_nata)
                end;

fun array_shrink a = FArray.IsabelleMapping.array_shrink a o integer_of_nat;

fun ras_shrink s =
  let
    val (a, n) = s;
    val aa =
      (if less_eq_nat (times_nata (nat_of_integer (128 : IntInf.int)) n)
            (array_length a) andalso
            less_nat (nat_of_integer (4 : IntInf.int)) n
        then array_shrink a n else a);
  in
    (aa, n)
  end;

fun ras_pop s = let
                  val (a, n) = s;
                in
                  ras_shrink (a, minus_nat n one_nata)
                end;

fun rev_append [] ac = ac
  | rev_append (x :: xs) ac = rev_append xs (x :: ac);

fun glist_delete_aux eq x [] asa = asa
  | glist_delete_aux eq x (y :: ys) asa =
    (if eq x y then rev_append asa ys else glist_delete_aux eq x ys (y :: asa));

fun glist_delete eq x l = glist_delete_aux eq x l [];

fun map2set_insert i k s = i k () s;

fun map2set_memb l k s = (case l k s of NONE => false | SOME _ => true);

fun whilea b c s = (if b s then whilea b c (c s) else s);

fun the (SOME x2) = x2;

fun gen_pick it s =
  the (it s (fn a => (case a of NONE => true | SOME _ => false))
         (fn x => fn _ => SOME x)
        NONE);

fun find_reachable_codeT eq bhc sz succi v0 =
  let
    val a =
      let
        val a =
          let
            val a = ();
          in
            Simple_state_nos_impl_ext (ras_empty zero_nat (), ahm_emptyb sz, a)
          end;
      in
        foldli v0 (fn _ => not false)
          (fn xa => fn sigma =>
            let
              val _ = sigma;
            in
              (if map2set_memb (ahm_lookupb eq bhc) xa
                    (ssnos_visited_impl sigma)
                then sigma
                else let
                       val aa =
                         let
                           val xc =
                             let
                               val xd =
                                 ssnos_stack_impl_update
                                   (fn _ =>
                                     ras_singleton one_nat (xa, succi xa))
                                   sigma;
                               val xe =
                                 ssnos_visited_impl_update
                                   (fn _ =>
                                     map2set_insert (ahm_updateb eq bhc) xa
                                       (ssnos_visited_impl xd))
                                   xd;
                             in
                               xe
                             end;
                           val _ = ();
                         in
                           xc
                         end;
                     in
                       whilea
                         (fn xd =>
                           not false andalso
                             not (ras_is_empty (ssnos_stack_impl xd)))
                         (fn xd =>
                           (case let
                                   val (ab, b) = ras_top (ssnos_stack_impl xd);
                                 in
                                   (if is_Nil b then (ab, (NONE, xd))
                                     else let
    val xf = gen_pick foldli b;
    val xg = glist_delete eq xf b;
    val xh =
      ssnos_stack_impl_update
        (fn _ => ras_push (ab, xg) (ras_pop (ssnos_stack_impl xd))) xd;
  in
    (ab, (SOME xf, xh))
  end)
                                 end
                             of (_, (NONE, ba)) =>
                               let
                                 val xf =
                                   let
                                     val xg =
                                       ssnos_stack_impl_update
 (fn _ => ras_pop (ssnos_stack_impl ba)) ba;
                                   in
                                     xg
                                   end;
                                 val _ = ();
                               in
                                 xf
                               end
                             | (_, (SOME xf, ba)) =>
                               (if map2set_memb (ahm_lookupb eq bhc) xf
                                     (ssnos_visited_impl ba)
                                 then let
val xg = ba;
val _ = ();
                                      in
xg
                                      end
                                 else let
val xg =
  let
    val xh =
      ssnos_stack_impl_update
        (fn _ => ras_push (xf, succi xf) (ssnos_stack_impl ba)) ba;
    val xi =
      ssnos_visited_impl_update
        (fn _ => map2set_insert (ahm_updateb eq bhc) xf (ssnos_visited_impl xh))
        xh;
  in
    xi
  end;
val _ = ();
                                      in
xg
                                      end)))
                         aa
                     end)
            end)
          a
      end;
  in
    ssnos_visited_impl a
  end;

fun numeral A_ (Bit1 n) =
  let
    val m = numeral A_ n;
  in
    plus ((plus_semigroup_add o semigroup_add_numeral) A_)
      (plus ((plus_semigroup_add o semigroup_add_numeral) A_) m m)
      (one (one_numeral A_))
  end
  | numeral A_ (Bit0 n) =
    let
      val m = numeral A_ n;
    in
      plus ((plus_semigroup_add o semigroup_add_numeral) A_) m m
    end
  | numeral A_ One = one (one_numeral A_);

fun prod_bhc (A1_, A2_) bhc1 bhc2 =
  (fn n => fn (a, b) =>
    moda A1_
      (plus ((plus_semigroup_add o semigroup_add_numeral) A2_)
        (times ((times_dvd o dvd_div) A1_) (bhc1 n a)
          (numeral A2_ (Bit1 (Bit0 (Bit0 (Bit0 (Bit0 One)))))))
        (bhc2 n b))
      n);

fun ahm_iteratei (HashMapb (a, n)) = ahm_iteratei_auxa a;

fun prod_eq eqa eqb x1 x2 =
  let
    val (a1, b1) = x1;
    val (a2, b2) = x2;
  in
    eqa a1 a2 andalso eqb b1 b2
  end;

fun ss_on_stack_impl_update ss_on_stack_impla
  (Simple_state_impl_ext
    (ss_stack_impl, ss_on_stack_impl, ss_visited_impl, more))
  = Simple_state_impl_ext
      (ss_stack_impl, ss_on_stack_impla ss_on_stack_impl, ss_visited_impl,
        more);

fun ss_visited_impl_update ss_visited_impla
  (Simple_state_impl_ext
    (ss_stack_impl, ss_on_stack_impl, ss_visited_impl, more))
  = Simple_state_impl_ext
      (ss_stack_impl, ss_on_stack_impl, ss_visited_impla ss_visited_impl, more);

fun ss_stack_impl_update ss_stack_impla
  (Simple_state_impl_ext
    (ss_stack_impl, ss_on_stack_impl, ss_visited_impl, more))
  = Simple_state_impl_ext
      (ss_stack_impla ss_stack_impl, ss_on_stack_impl, ss_visited_impl, more);

fun ss_on_stack_impl
  (Simple_state_impl_ext
    (ss_stack_impl, ss_on_stack_impl, ss_visited_impl, more))
  = ss_on_stack_impl;

fun ss_visited_impl
  (Simple_state_impl_ext
    (ss_stack_impl, ss_on_stack_impl, ss_visited_impl, more))
  = ss_visited_impl;

fun ss_stack_impl
  (Simple_state_impl_ext
    (ss_stack_impl, ss_on_stack_impl, ss_visited_impl, more))
  = ss_stack_impl;

fun more_update morea
  (Simple_state_impl_ext
    (ss_stack_impl, ss_on_stack_impl, ss_visited_impl, more))
  = Simple_state_impl_ext
      (ss_stack_impl, ss_on_stack_impl, ss_visited_impl, morea more);

fun fas_impl
  (Simple_state_impl_ext
    (ss_stack_impl, ss_on_stack_impl, ss_visited_impl,
      Fas_state_impl_ext (fas_impl, more)))
  = fas_impl;

fun morea
  (Simple_state_impl_ext
    (ss_stack_impl, ss_on_stack_impl, ss_visited_impl, more))
  = more;

fun list_map_delete_aux eq k [] accu = accu
  | list_map_delete_aux eq k (x :: xs) accu =
    (if eq (fst x) k then xs @ accu
      else list_map_delete_aux eq k xs (x :: accu));

fun list_map_delete eq k m = list_map_delete_aux eq k m [];

fun ahm_deleteb eq bhc k (HashMapb (a, n)) =
  let
    val h = bhc (array_length a) k;
    val m = array_get a h;
    val deleted = not (is_none (list_map_lookup eq k m));
  in
    HashMapb
      (array_set a h (list_map_delete eq k m),
        (if deleted then minus_nat n one_nata else n))
  end;

fun the_res (DRETURN x) = x;

fun dbind DFAILi f = DFAILi
  | dbind DSUCCEEDi f = DSUCCEEDi
  | dbind (DRETURN x) f = f x;

fun dWHILET b f s = (if b s then dbind (f s) (dWHILET b f) else DRETURN s);

fun gen_diff del1 it2 s1 s2 = it2 s2 (fn _ => true) del1 s1;

fun find_fas_code eq bhc sz succi v0 =
  the_res
    (dbind
      (foldli v0
        (fn a =>
          (case a of DSUCCEEDi => false | DFAILi => false
            | DRETURN _ => not false))
        (fn x => fn s =>
          dbind s
            (fn sigma =>
              let
                val _ = sigma;
              in
                (if map2set_memb (ahm_lookupb eq bhc) x (ss_visited_impl sigma)
                  then DRETURN sigma
                  else dbind let
                               val xc =
                                 ss_stack_impl_update
                                   (fn _ => ras_singleton one_nat (x, succi x))
                                   sigma;
                               val xd =
                                 ss_on_stack_impl_update
                                   (fn _ =>
                                     map2set_insert (ahm_updateb eq bhc) x
                                       (ahm_emptyb sz))
                                   xc;
                             in
                               DRETURN
                                 (ss_visited_impl_update
                                   (fn _ =>
                                     map2set_insert (ahm_updateb eq bhc) x
                                       (ss_visited_impl xd))
                                   xd)
                             end
                         (fn xa =>
                           dWHILET
                             (fn xd =>
                               not false andalso
                                 not (ras_is_empty (ss_stack_impl xd)))
                             (fn xd =>
                               dbind let
                                       val (a, b) = ras_top (ss_stack_impl xd);
                                     in
                                       (if is_Nil b then DRETURN (a, (NONE, xd))
 else let
        val xg = glist_delete eq (gen_pick foldli b) b;
        val xh =
          ss_stack_impl_update
            (fn _ => ras_push (a, xg) (ras_pop (ss_stack_impl xd))) xd;
      in
        DRETURN (a, (SOME (gen_pick foldli b), xh))
      end)
                                     end
                                 (fn a =>
                                   (case a
                                     of (aa, (NONE, ba)) =>
                                       dbind
 let
   val xf = ss_stack_impl_update (fn _ => ras_pop (ss_stack_impl ba)) ba;
 in
   DRETURN
     (ss_on_stack_impl_update
       (fn _ => ahm_deleteb eq bhc aa (ss_on_stack_impl xf)) xf)
 end
 (fn xf => DRETURN (more_update (fn _ => morea xf) xf))
                                     | (aa, (SOME xf, ba)) =>
                                       (if map2set_memb (ahm_lookupb eq bhc) xf
     (ss_visited_impl ba)
 then (if map2set_memb (ahm_lookupb eq bhc) xf
            (gen_diff (ahm_deleteb eq bhc) (map_iterator_dom o ahm_iteratei)
              (ss_visited_impl ba) (ss_on_stack_impl ba))
        then DRETURN (more_update (fn _ => morea ba) ba)
        else DRETURN
               (more_update
                 (fn _ =>
                   Fas_state_impl_ext
                     (map2set_insert
                        (ahm_updateb (prod_eq eq eq)
                          (prod_bhc (div_nat, numeral_nat) bhc bhc))
                        (aa, xf) (fas_impl ba),
                       ()))
                 ba))
 else dbind let
              val xg =
                ss_stack_impl_update
                  (fn _ => ras_push (xf, succi xf) (ss_stack_impl ba)) ba;
              val xh =
                ss_on_stack_impl_update
                  (fn _ =>
                    map2set_insert (ahm_updateb eq bhc) xf
                      (ss_on_stack_impl xg))
                  xg;
            in
              DRETURN
                (ss_visited_impl_update
                  (fn _ =>
                    map2set_insert (ahm_updateb eq bhc) xf (ss_visited_impl xh))
                  xh)
            end
        (fn xg => DRETURN (more_update (fn _ => morea xg) xg))))))
                             (more_update (fn _ => morea xa) xa)))
              end))
        (DRETURN
          (Simple_state_impl_ext
            (ras_empty zero_nat (), ahm_emptyb sz, ahm_emptyb sz,
              Fas_state_impl_ext (ahm_emptyb (plus_nata sz sz), ())))))
      (fn x => DRETURN (fas_impl x)));

fun gen_union it ins a b = it a (fn _ => true) ins b;

fun find_fas_init_code eq bhc sz succi v0 fIi =
  let
    val x =
      gen_union (map_iterator_dom o ahm_iteratei) (glist_insert eq)
        (find_reachable_codeT eq bhc sz succi v0) [];
    val xa =
      find_fas_code eq bhc sz
        (fn xa => filter (fn xc => not (fIi (xa, xc))) (succi xa)) x;
  in
    (fn xb =>
      fIi xb orelse
        map2set_memb
          (ahm_lookupb (prod_eq eq eq)
            (prod_bhc (div_nat, numeral_nat) bhc bhc))
          xb xa)
  end;

fun collecti_index_list i a c [] = a
  | collecti_index_list i a c (x :: xs) =
    (case c i x of (true, s) => map (fn aa => (i, aa)) s
      | (false, s) =>
        collecti_index_list (suc i) (a @ map (fn aa => (i, aa)) s) c xs);

fun cfg_succ_list (Assign (ge, x, e)) = [(Inl (AAssign (ge, x, e)), Skip)]
  | cfg_succ_list (Assign_local (ge, x, e)) =
    [(Inl (AAssign_local (ge, x, e)), Skip)]
  | cfg_succ_list (Assign_global (ge, x, e)) =
    [(Inl (AAssign_global (ge, x, e)), Skip)]
  | cfg_succ_list (Test e) = [(Inl (ATest e), Skip)]
  | cfg_succ_list (Seq (c1, c2)) =
    (if equal_cmda c1 Skip then [(Inl ASkip, c2)] else []) @
      map (fn (e, c1a) => (e, Seq (c1a, c2))) (cfg_succ_list c1)
  | cfg_succ_list (Or (c1, c2)) =
    remdups (equal_prod (equal_sum equal_action equal_brk_ctd) equal_cmd)
      [(Inl ASkip, c1), (Inl ASkip, c2)]
  | cfg_succ_list (Loop c) = [(Inl ASkip, Iterate (c, c))]
  | cfg_succ_list Break = [(Inr AIBreak, Invalid)]
  | cfg_succ_list Continue = [(Inr AIContinue, Invalid)]
  | cfg_succ_list (Iterate (c, cb)) =
    (if equal_cmda c Skip then [(Inl ASkip, Iterate (cb, cb))] else []) @
      remdups (equal_prod (equal_sum equal_action equal_brk_ctd) equal_cmd)
        (map (fn a =>
               (case a of (Inl e, ca) => (Inl e, Iterate (ca, cb))
                 | (Inr AIBreak, ca) => (Inl ASkip, Skip)
                 | (Inr AIContinue, ca) => (Inl ASkip, Iterate (cb, cb))))
          (cfg_succ_list c))
  | cfg_succ_list Skip = []
  | cfg_succ_list Invalid = [];

fun map_filter f [] = []
  | map_filter f (x :: xs) =
    (case f x of NONE => map_filter f xs | SOME y => y :: map_filter f xs);

fun cfg_succ_lista c =
  map_filter
    (fn a => (case a of (Inl aa, ca) => SOME (aa, ca) | (Inr _, ca) => NONE))
    (cfg_succ_list c);

fun write_globals (AAssign (uu, x, uv)) = inserta equal_literal x bot_seta
  | write_globals (AAssign_local (uw, ux, uy)) = bot_seta
  | write_globals (AAssign_global (uz, x, va)) =
    inserta equal_literal x bot_seta
  | write_globals (ATest e) = bot_seta
  | write_globals ASkip = bot_seta;

fun udvars_of_exp (E_var x) = inserta equal_literal x bot_seta
  | udvars_of_exp (E_localvar x) = bot_seta
  | udvars_of_exp (E_globalvar x) = bot_seta
  | udvars_of_exp (E_const c) = bot_seta
  | udvars_of_exp (E_bin (bop, e1, e2)) =
    sup_setb equal_literal (udvars_of_exp e1) (udvars_of_exp e2)
  | udvars_of_exp (E_un (uop, e)) = udvars_of_exp e;

fun gvars_of_exp (E_var x) = bot_seta
  | gvars_of_exp (E_localvar x) = bot_seta
  | gvars_of_exp (E_globalvar x) = inserta equal_literal x bot_seta
  | gvars_of_exp (E_const c) = bot_seta
  | gvars_of_exp (E_bin (bop, e1, e2)) =
    sup_setb equal_literal (gvars_of_exp e1) (gvars_of_exp e2)
  | gvars_of_exp (E_un (uop, e)) = gvars_of_exp e;

fun read_globals (AAssign (c, uu, e)) =
  sup_setb equal_literal
    (sup_setb equal_literal
      (sup_setb equal_literal (gvars_of_exp c) (udvars_of_exp c))
      (gvars_of_exp e))
    (udvars_of_exp e)
  | read_globals (AAssign_local (c, uv, e)) =
    sup_setb equal_literal
      (sup_setb equal_literal
        (sup_setb equal_literal (gvars_of_exp c) (udvars_of_exp c))
        (gvars_of_exp e))
      (udvars_of_exp e)
  | read_globals (AAssign_global (c, uw, e)) =
    sup_setb equal_literal
      (sup_setb equal_literal
        (sup_setb equal_literal (gvars_of_exp c) (udvars_of_exp c))
        (gvars_of_exp e))
      (udvars_of_exp e)
  | read_globals (ATest e) =
    sup_setb equal_literal (gvars_of_exp e) (udvars_of_exp e)
  | read_globals ASkip = bot_seta;

fun bool_of_val_impl v = not ((v : Word32.word) = (Word32.fromInt 0));

fun val_of_bool_impl b = (if b then (Word32.fromInt 1) else (Word32.fromInt 0));

fun equal_bool p true = p
  | equal_bool p false = not p
  | equal_bool true p = p
  | equal_bool false p = not p;

fun uminus_int k = Int_of_integer (IntInf.~ (integer_of_int k));

fun div_int k l =
  Int_of_integer (div_integer (integer_of_int k) (integer_of_int l));

fun less_int k l = IntInf.< (integer_of_int k, integer_of_int l);

fun abs_int i = (if less_int i zero_int then uminus_int i else i);

fun bitAND_int (Int_of_integer i) (Int_of_integer j) =
  Int_of_integer (IntInf.andb (i, j));

fun shiftl_integer x n = Bits_Integer.shiftl x (integer_of_nat n);

fun bit_integer i true = IntInf.+ (shiftl_integer i one_nata, (1 : IntInf.int))
  | bit_integer i false = shiftl_integer i one_nata;

fun bit (Int_of_integer i) b = Int_of_integer (bit_integer i b);

fun bin_mask n =
  (if equal_nata n zero_nata then zero_int
    else bit (bin_mask (minus_nat n one_nata)) true);

fun word_of_int A_ x = Word (bitAND_int x (bin_mask (len_of A_ Type)));

fun test_bit_integer x n = Bits_Integer.test_bit x (integer_of_nat n);

fun test_bit_int (Int_of_integer x) n = test_bit_integer x n;

fun shiftl_int (Int_of_integer x) n = Int_of_integer (shiftl_integer x n);

fun minus_int k l =
  Int_of_integer (IntInf.- (integer_of_int k, integer_of_int l));

fun sbintrunc n bin =
  let
    val bina = bitAND_int bin (bin_mask (plus_nata n one_nata));
  in
    (if test_bit_int bina n
      then minus_int bina (shiftl_int (Int_of_integer (2 : IntInf.int)) n)
      else bina)
  end;

fun uint A_ (Word x) = x;

fun sint A_ w =
  sbintrunc (minus_nat (len_of (len0_len A_) Type) one_nata)
    (uint (len0_len A_) w);

fun word_sdiv A_ x y =
  let
    val xa = sint A_ x;
    val ya = sint A_ y;
    val negative =
      not (equal_bool (less_int xa zero_int) (less_int ya zero_int));
    val result = div_int (abs_int xa) (abs_int ya);
  in
    word_of_int (len0_len A_) (if negative then uminus_int result else result)
  end;

fun test_bit_uint32 x n =
  less_nat n (nat_of_integer (32 : IntInf.int)) andalso
    Uint32.test_bit x (integer_of_nat n);

fun bl_of_nth n f =
  (if equal_nata n zero_nata then []
    else f (minus_nat n one_nata) :: bl_of_nth (minus_nat n one_nata) f);

fun bl_to_bin_aux [] w = w
  | bl_to_bin_aux (b :: bs) w = bl_to_bin_aux bs (bit w b);

fun bl_to_bin bs = bl_to_bin_aux bs zero_int;

fun of_bl A_ bl = word_of_int A_ (bl_to_bin bl);

fun set_bits_word A_ f = of_bl A_ (bl_of_nth (len_of A_ Type) f);

fun rep_uint32 x =
  set_bits_word
    (len0_bit0 (len0_bit0 (len0_bit0 (len0_bit0 (len0_bit0 len0_num1)))))
    (test_bit_uint32 x);

fun abs_uint32 x =
  Word32.fromInt
    (integer_of_int
      (uint (len0_bit0
              (len0_bit0 (len0_bit0 (len0_bit0 (len0_bit0 len0_num1)))))
        x));

fun sdiv_impl x y =
  abs_uint32
    (word_sdiv (len_bit0 (len_bit0 (len_bit0 (len_bit0 (len_bit0 len_num1)))))
      (rep_uint32 x) (rep_uint32 y));

fun smod_impl a b = Word32.- (a, Word32.* (sdiv_impl a b, b));

fun eval_bin_op_impl_aux Bo_plus v1 v2 = Word32.+ (v1, v2)
  | eval_bin_op_impl_aux Bo_minus v1 v2 = Word32.- (v1, v2)
  | eval_bin_op_impl_aux Bo_mul v1 v2 = Word32.* (v1, v2)
  | eval_bin_op_impl_aux Bo_div v1 v2 = sdiv_impl v1 v2
  | eval_bin_op_impl_aux Bo_mod v1 v2 = smod_impl v1 v2
  | eval_bin_op_impl_aux Bo_less v1 v2 = val_of_bool_impl (Word32.< (v1, v2))
  | eval_bin_op_impl_aux Bo_less_eq v1 v2 =
    val_of_bool_impl (Word32.<= (v1, v2))
  | eval_bin_op_impl_aux Bo_eq v1 v2 =
    val_of_bool_impl ((v1 : Word32.word) = v2)
  | eval_bin_op_impl_aux Bo_and v1 v2 = Word32.andb (v1, v2)
  | eval_bin_op_impl_aux Bo_or v1 v2 = Word32.orb (v1, v2)
  | eval_bin_op_impl_aux Bo_xor v1 v2 = Word32.xorb (v1, v2);

fun eval_bin_op_impl bop v1 v2 = eval_bin_op_impl_aux bop v1 v2;

fun eval_un_op_impl_aux Uo_minus v = Word32.~ v
  | eval_un_op_impl_aux Uo_not v = Word32.notb v;

fun eval_un_op_impl uop v = eval_un_op_impl_aux uop v;

fun less_eq_int k l = IntInf.<= (integer_of_int k, integer_of_int l);

fun assert_option true = SOME ()
  | assert_option false = NONE;

fun power one times a n =
  (if equal_nata n zero_nata then one
    else times a (power one times a (minus_nat n one_nata)));

fun powera A_ = power (one (one_power A_)) (times (times_power A_));

val min_signed : int =
  uminus_int
    (powera power_int (Int_of_integer (2 : IntInf.int))
      (minus_nat
        (len_of_bit0 (len0_bit0 (len0_bit0 (len0_bit0 (len0_bit0 len0_num1))))
          Type)
        one_nata));

val max_signed : int =
  minus_int
    (powera power_int (Int_of_integer (2 : IntInf.int))
      (minus_nat
        (len_of_bit0 (len0_bit0 (len0_bit0 (len0_bit0 (len0_bit0 len0_num1))))
          Type)
        one_nata))
    (Int_of_integer (1 : IntInf.int));

fun bind NONE f = NONE
  | bind (SOME x) f = f x;

fun eval_exp_impl (E_var x) (ls, gs) =
  let
    val lv = variables ls;
    val gv = variablesa gs;
  in
    (case vi_lookup lv x
      of NONE => (case vi_lookup gv x of NONE => NONE | SOME a => SOME a)
      | SOME a => SOME a)
  end
  | eval_exp_impl (E_localvar x) (ls, gs) = vi_lookup (variables ls) x
  | eval_exp_impl (E_globalvar x) (ls, gs) = vi_lookup (variablesa gs) x
  | eval_exp_impl (E_const n) fs =
    bind (assert_option
           (less_eq_int min_signed n andalso less_eq_int n max_signed))
      (fn _ => SOME (uint32_of_int n))
  | eval_exp_impl (E_bin (bop, e1, e2)) fs =
    bind (eval_exp_impl e1 fs)
      (fn v1 =>
        bind (eval_exp_impl e2 fs) (fn v2 => SOME (eval_bin_op_impl bop v1 v2)))
  | eval_exp_impl (E_un (uop, e)) fs =
    bind (eval_exp_impl e fs) (fn v => SOME (eval_un_op_impl uop v));

fun la_en_impl fs (AAssign (e, uu, uv)) =
  bind (eval_exp_impl e fs) (fn v => SOME (bool_of_val_impl v))
  | la_en_impl fs (AAssign_local (e, uw, ux)) =
    bind (eval_exp_impl e fs) (fn v => SOME (bool_of_val_impl v))
  | la_en_impl fs (AAssign_global (e, uy, uz)) =
    bind (eval_exp_impl e fs) (fn v => SOME (bool_of_val_impl v))
  | la_en_impl fs (ATest e) =
    bind (eval_exp_impl e fs) (fn v => SOME (bool_of_val_impl v))
  | la_en_impl va ASkip = SOME true;

fun la_en_impla fs a = (case la_en_impl fs a of NONE => false | SOME b => b);

fun gen_length n (x :: xs) = gen_length (suc n) xs
  | gen_length n [] = n;

fun size_list x = gen_length zero_nata x;

fun of_phantom (Phantom x) = x;

fun card (A1_, A2_) (Coset xs) =
  minus_nat (of_phantom (card_UNIV A1_)) (size_list (remdups A2_ xs))
  | card (A1_, A2_) (Set xs) = size_list (remdups A2_ xs);

fun eq_set (A1_, A2_) (Coset xs) (Coset ys) =
  pred_list (membera A2_ ys) xs andalso pred_list (membera A2_ xs) ys
  | eq_set (A1_, A2_) (Set xs) (Set ys) =
    pred_list (membera A2_ ys) xs andalso pred_list (membera A2_ xs) ys
  | eq_set (A1_, A2_) (Set ys) (Coset xs) =
    let
      val n = card (A1_, A2_) top_seta;
    in
      (if equal_nata n zero_nata then false
        else let
               val xsa = remdups A2_ xs;
               val ysa = remdups A2_ ys;
             in
               equal_nata (plus_nata (size_list xsa) (size_list ysa)) n andalso
                 (pred_list (fn x => not (membera A2_ ysa x)) xsa andalso
                   pred_list (fn y => not (membera A2_ xsa y)) ysa)
             end)
    end
  | eq_set (A1_, A2_) (Coset xs) (Set ys) =
    let
      val n = card (A1_, A2_) top_seta;
    in
      (if equal_nata n zero_nata then false
        else let
               val xsa = remdups A2_ xs;
               val ysa = remdups A2_ ys;
             in
               equal_nata (plus_nata (size_list xsa) (size_list ysa)) n andalso
                 (pred_list (fn x => not (membera A2_ ysa x)) xsa andalso
                   pred_list (fn y => not (membera A2_ xsa y)) ysa)
             end)
    end;

fun ga_gample_mil4 mem sticky_Ei gc =
  let
    val lcs = processesa gc;
    val gs = statea gc;
  in
    collecti_index_list zero_nata []
      (fn _ => fn lc =>
        let
          val c = command lc;
          val ls = state lc;
          val asa = cfg_succ_lista c;
          val s = filter (la_en_impla (ls, gs) o fst) asa;
          val ample =
            not (null s) andalso
              (pred_list
                 (fn (a, _) =>
                   eq_set (card_UNIV_literal, equal_literal) (read_globals a)
                     bot_seta andalso
                     eq_set (card_UNIV_literal, equal_literal) (write_globals a)
                       bot_seta)
                 asa andalso
                pred_list (fn (_, ca) => not (mem (c, ca) sticky_Ei)) s);
        in
          (ample, map (fn a => (c, a)) s)
        end)
      lcs
  end;

fun cfgE_succ_list c =
  remdups equal_cmd
    (map_filter
      (fn a => (case a of (Inl _, ca) => SOME ca | (Inr _, ca) => NONE))
      (cfg_succ_list c));

fun is_vis_edge is_vis_var =
  (fn (c, ca) =>
    list_ex
      (fn (a, cc) => equal_cmda ca cc andalso bex (write_globals a) is_vis_var)
      (cfg_succ_lista c));

fun cr_ample_impl4 prog is_vis_var =
  let
    val sticky =
      find_fas_init_code equal_cmda (bounded_hashcode_nat hashable_cmd)
        (def_hashmap_size_cmd Type) cfgE_succ_list
        (fold (glist_insert equal_cmda o body) (processes prog) [])
        (is_vis_edge is_vis_var);
  in
    stutter_extend_en_list (ga_gample_mil4 (fn x => fn f => f x) sticky)
  end;

fun stutter_extend_ex ex = (fn a => (case a of NONE => id | SOME aa => ex aa));

fun variables_updatea variablesa (Global_state_impl_ext (variables, more)) =
  Global_state_impl_ext (variablesa variables, more);

fun variables_update variablesa (Local_state_impl_ext (variables, more)) =
  Local_state_impl_ext (variablesa variables, more);

fun la_ex_impl (ls, gs) (AAssign (ge, x, e)) =
  bind (eval_exp_impl ge (ls, gs))
    (fn v =>
      bind (assert_option (bool_of_val_impl v))
        (fn _ =>
          bind (eval_exp_impl e (ls, gs))
            (fn va =>
              (if not (is_none (vi_lookup (variables ls) x))
                then let
                       val lsa = variables_update (vi_update x va) ls;
                     in
                       SOME (lsa, gs)
                     end
                else bind (assert_option
                            (not (is_none (vi_lookup (variablesa gs) x))))
                       (fn _ =>
                         let
                           val gsa = variables_updatea (vi_update x va) gs;
                         in
                           SOME (ls, gsa)
                         end)))))
  | la_ex_impl (ls, gs) (AAssign_local (ge, x, e)) =
    bind (eval_exp_impl ge (ls, gs))
      (fn v =>
        bind (assert_option (bool_of_val_impl v))
          (fn _ =>
            bind (eval_exp_impl e (ls, gs))
              (fn va =>
                bind (assert_option
                       (not (is_none (vi_lookup (variables ls) x))))
                  (fn _ =>
                    let
                      val lsa = variables_update (vi_update x va) ls;
                    in
                      SOME (lsa, gs)
                    end))))
  | la_ex_impl (ls, gs) (AAssign_global (ge, x, e)) =
    bind (eval_exp_impl ge (ls, gs))
      (fn v =>
        bind (assert_option (bool_of_val_impl v))
          (fn _ =>
            bind (eval_exp_impl e (ls, gs))
              (fn va =>
                bind (assert_option
                       (not (is_none (vi_lookup (variablesa gs) x))))
                  (fn _ =>
                    let
                      val gsa = variables_updatea (vi_update x va) gs;
                    in
                      SOME (ls, gsa)
                    end))))
  | la_ex_impl fs (ATest e) =
    bind (eval_exp_impl e fs)
      (fn v => bind (assert_option (bool_of_val_impl v)) (fn _ => SOME fs))
  | la_ex_impl fs ASkip = SOME fs;

fun la_ex_impla fs a = (case la_ex_impl fs a of NONE => fs | SOME fsa => fsa);

fun list_update [] i y = []
  | list_update (x :: xs) i y =
    (if equal_nata i zero_nata then y :: xs
      else x :: list_update xs (minus_nat i one_nata) y);

fun ga_gex_impl ga gc =
  let
    val (pid, (_, (a, c))) = ga;
    val fs = (state (nth (processesa gc) pid), statea gc);
    val (ls, gs) = la_ex_impla fs a;
    val gca =
      Pid_global_config_ext
        (list_update (processesa gc) pid (Local_config_ext (c, ls, ())), gs,
          ());
  in
    gca
  end;

fun ga_ex_impl x = stutter_extend_ex ga_gex_impl x;

fun impl4_succ prog is_vis_var =
  succ_of_enex_list
    (equal_pid_global_config_ext equal_cmd
      (equal_local_state_impl_ext equal_unit)
      (equal_global_state_impl_ext equal_unit) equal_unit)
    (cr_ample_impl4 prog is_vis_var, ga_ex_impl);

fun sm_to_sa_impl prog is_vis_var =
  Gen_frg_impl_ext
    ((fn _ => true),
      remdups
        (equal_pid_global_config_ext equal_cmd
          (equal_local_state_impl_ext equal_unit)
          (equal_global_state_impl_ext equal_unit) equal_unit) o
        impl4_succ prog is_vis_var,
      [pid_init_gc_impl prog],
      Gen_sa_impl_ext (pid_test_gc_impl is_vis_var, ()));

fun ty_expr (gamma_l, gamma_g) (E_var x) =
  (if (case gamma_l x of NONE => false | SOME _ => true) orelse
        (case gamma_g x of NONE => false | SOME _ => true)
    then SOME () else NONE)
  | ty_expr (gamma_l, gamma_g) (E_localvar x) =
    (if (case gamma_l x of NONE => false | SOME _ => true) then SOME ()
      else NONE)
  | ty_expr (gamma_l, gamma_g) (E_globalvar x) =
    (if (case gamma_g x of NONE => false | SOME _ => true) then SOME ()
      else NONE)
  | ty_expr uu (E_const c) =
    (if less_eq_int min_signed c andalso less_eq_int c max_signed then SOME ()
      else NONE)
  | ty_expr (gamma_l, gamma_g) (E_bin (bop, e1, e2)) =
    bind (ty_expr (gamma_l, gamma_g) e1)
      (fn _ => bind (ty_expr (gamma_l, gamma_g) e2) (fn _ => SOME ()))
  | ty_expr (gamma_l, gamma_g) (E_un (uop, e)) =
    bind (ty_expr (gamma_l, gamma_g) e) (fn _ => SOME ());

fun ty_cmd ((gamma_l, gamma_g), loop) (Assign (c, x, e)) =
  not (is_none (ty_expr (gamma_l, gamma_g) c)) andalso
    (((case gamma_l x of NONE => false | SOME _ => true) orelse
       (case gamma_g x of NONE => false | SOME _ => true)) andalso
      not (is_none (ty_expr (gamma_l, gamma_g) e)))
  | ty_cmd ((gamma_l, gamma_g), loop) (Assign_local (c, x, e)) =
    not (is_none (ty_expr (gamma_l, gamma_g) c)) andalso
      ((case gamma_l x of NONE => false | SOME _ => true) andalso
        not (is_none (ty_expr (gamma_l, gamma_g) e)))
  | ty_cmd ((gamma_l, gamma_g), loop) (Assign_global (c, x, e)) =
    not (is_none (ty_expr (gamma_l, gamma_g) c)) andalso
      ((case gamma_g x of NONE => false | SOME _ => true) andalso
        not (is_none (ty_expr (gamma_l, gamma_g) e)))
  | ty_cmd ((gamma_l, gamma_g), loop) (Test e) =
    not (is_none (ty_expr (gamma_l, gamma_g) e))
  | ty_cmd ((gamma_l, gamma_g), loop) Skip = true
  | ty_cmd ((gamma_l, gamma_g), loop) (Seq (c1, c2)) =
    ty_cmd ((gamma_l, gamma_g), loop) c1 andalso
      ty_cmd ((gamma_l, gamma_g), loop) c2
  | ty_cmd ((gamma_l, gamma_g), loop) (Or (c1, c2)) =
    ty_cmd ((gamma_l, gamma_g), loop) c1 andalso
      ty_cmd ((gamma_l, gamma_g), loop) c2
  | ty_cmd ((gamma_l, gamma_g), loop) (Loop c) =
    ty_cmd ((gamma_l, gamma_g), loop) (Iterate (c, c))
  | ty_cmd ((gamma_l, gamma_g), loop) Break = loop
  | ty_cmd ((gamma_l, gamma_g), loop) Continue = loop
  | ty_cmd ((gamma_l, gamma_g), loop) (Iterate (c1, c2)) =
    ty_cmd ((gamma_l, gamma_g), true) c1 andalso
      ty_cmd ((gamma_l, gamma_g), true) c2
  | ty_cmd ((gamma_l, gamma_g), loop) Invalid = false;

fun ty_proc_decl gamma_g pd =
  let
    val gamma_l =
      (fn x =>
        (if membera equal_literal (local_vars pd) x then SOME () else NONE));
  in
    ty_cmd ((gamma_l, gamma_g), false) (body pd)
  end;

fun ty_program prog =
  let
    val gamma_g =
      (fn x =>
        (if membera equal_literal (global_vars prog) x then SOME () else NONE));
  in
    pred_list (ty_proc_decl gamma_g) (processes prog)
  end;

fun vars_of_aprop (AP_isNull v) = inserta equal_literal v bot_seta;

fun supremum B_ (Set xs) f =
  fold (sup ((sup_semilattice_sup o semilattice_sup_lattice o
               lattice_bounded_lattice_bot o
               bounded_lattice_bot_bounded_lattice o
               bounded_lattice_complete_lattice)
              B_) o
         f)
    xs (bot ((bot_order_bot o order_bot_bounded_semilattice_sup_bot o
               bounded_semilattice_sup_bot_bounded_lattice_bot o
               bounded_lattice_bot_bounded_lattice o
               bounded_lattice_complete_lattice)
              B_));

fun ltlc_aprops A_ LTLcTrue = bot_seta
  | ltlc_aprops A_ LTLcFalse = bot_seta
  | ltlc_aprops A_ (LTLcProp x3a) = inserta A_ x3a bot_seta
  | ltlc_aprops A_ (LTLcNeg x4) = ltlc_aprops A_ x4
  | ltlc_aprops A_ (LTLcAnd (x51, x52)) =
    supremum (complete_lattice_set A_)
      (sup_setb (equal_ltlc A_) (inserta (equal_ltlc A_) x51 bot_seta)
        (inserta (equal_ltlc A_) x52 bot_seta))
      (ltlc_aprops A_)
  | ltlc_aprops A_ (LTLcOr (x61, x62)) =
    supremum (complete_lattice_set A_)
      (sup_setb (equal_ltlc A_) (inserta (equal_ltlc A_) x61 bot_seta)
        (inserta (equal_ltlc A_) x62 bot_seta))
      (ltlc_aprops A_)
  | ltlc_aprops A_ (LTLcImplies (x71, x72)) =
    supremum (complete_lattice_set A_)
      (sup_setb (equal_ltlc A_) (inserta (equal_ltlc A_) x71 bot_seta)
        (inserta (equal_ltlc A_) x72 bot_seta))
      (ltlc_aprops A_)
  | ltlc_aprops A_ (LTLcIff (x81, x82)) =
    supremum (complete_lattice_set A_)
      (sup_setb (equal_ltlc A_) (inserta (equal_ltlc A_) x81 bot_seta)
        (inserta (equal_ltlc A_) x82 bot_seta))
      (ltlc_aprops A_)
  | ltlc_aprops A_ (LTLcNext x9) = ltlc_aprops A_ x9
  | ltlc_aprops A_ (LTLcFinal x10) = ltlc_aprops A_ x10
  | ltlc_aprops A_ (LTLcGlobal x11) = ltlc_aprops A_ x11
  | ltlc_aprops A_ (LTLcUntil (x121, x122)) =
    supremum (complete_lattice_set A_)
      (sup_setb (equal_ltlc A_) (inserta (equal_ltlc A_) x121 bot_seta)
        (inserta (equal_ltlc A_) x122 bot_seta))
      (ltlc_aprops A_)
  | ltlc_aprops A_ (LTLcRelease (x131, x132)) =
    supremum (complete_lattice_set A_)
      (sup_setb (equal_ltlc A_) (inserta (equal_ltlc A_) x131 bot_seta)
        (inserta (equal_ltlc A_) x132 bot_seta))
      (ltlc_aprops A_);

fun vars_of_ltlc ltlc =
  sup_setc equal_literal (image vars_of_aprop (ltlc_aprops equal_aprop ltlc));

fun cfg_l2b (c1, (c2, c3)) = c1;

fun cfg_ce (c1, (c2, c3)) = c3;

fun igbgi_num_acc
  (Gen_frg_impl_ext
    (frgi_V, frgi_E, frgi_V0,
      Gen_igbg_impl_ext (igbgi_num_acc, igbgi_acc, more)))
  = igbgi_num_acc;

fun incoming_impl
  (Node_impl_ext
    (name_impl, incoming_impl, new_impl, old_impl, next_impl, more))
  = incoming_impl;

fun name_impl
  (Node_impl_ext
    (name_impl, incoming_impl, new_impl, old_impl, next_impl, more))
  = name_impl;

fun array_get_oo x a = FArray.IsabelleMapping.array_get_oo x a o integer_of_nat;

fun iam_alpha a i = array_get_oo NONE a i;

fun iam_increment l idx =
  max ord_nat (minus_nat (plus_nata idx one_nata) l)
    (plus_nata (times_nata (nat_of_integer (2 : IntInf.int)) l)
      (nat_of_integer (3 : IntInf.int)));

fun array_set_oo f a = FArray.IsabelleMapping.array_set_oo f a o integer_of_nat;

fun iam_update k v a =
  array_set_oo
    (fn _ =>
      let
        val l = array_length a;
      in
        array_set (array_grow a (iam_increment l k) NONE) k (SOME v)
      end)
    a k (SOME v);

fun iam_empty x = (fn _ => FArray.IsabelleMapping.array_of_list []) x;

fun rm_iterateoi Empty c f sigma = sigma
  | rm_iterateoi (Branch (col, l, k, v, r)) c f sigma =
    (if c sigma
      then let
             val sigmaa = rm_iterateoi l c f sigma;
           in
             (if c sigmaa then rm_iterateoi r c f (f (k, v) sigmaa) else sigmaa)
           end
      else sigma);

fun the_default uu (SOME x) = x
  | the_default x NONE = x;

fun build_succ_code A_ x =
  foldli x (fn _ => true)
    (fn xa =>
      (map_iterator_dom o rm_iterateoi) (incoming_impl xa) (fn _ => true)
        (fn xaa => fn sigma =>
          (if equal_nata xaa zero_nata then sigma
            else iam_update xaa
                   (glist_insert equal_nata (name_impl xa)
                     (the_default [] (iam_alpha sigma xaa)))
                   sigma)))
    (iam_empty ());

fun old_impl
  (Node_impl_ext
    (name_impl, incoming_impl, new_impl, old_impl, next_impl, more))
  = old_impl;

fun rbt_ins less f k v (Branch (R, l, x, y, r)) =
  (if less k x then Branch (R, rbt_ins less f k v l, x, y, r)
    else (if less x k then Branch (R, l, x, y, rbt_ins less f k v r)
           else Branch (R, l, x, f k y v, r)))
  | rbt_ins less f k v (Branch (B, l, x, y, r)) =
    (if less k x then balance (rbt_ins less f k v l) x y r
      else (if less x k then balance l x y (rbt_ins less f k v r)
             else Branch (B, l, x, f k y v, r)))
  | rbt_ins less f k v Empty = Branch (R, Empty, k, v, Empty);

fun rbt_insert_with_key less f k v t = paint B (rbt_ins less f k v t);

fun rbt_insert less = rbt_insert_with_key less (fn _ => fn _ => fn nv => nv);

fun sunion_with less f asa [] = asa
  | sunion_with less f [] bs = bs
  | sunion_with less f ((ka, va) :: asa) ((k, v) :: bs) =
    (if less k ka then (k, v) :: sunion_with less f ((ka, va) :: asa) bs
      else (if less ka k then (ka, va) :: sunion_with less f asa ((k, v) :: bs)
             else (ka, f ka va v) :: sunion_with less f asa bs));

fun skip_red (Branch (R, l, k, v, r)) = l
  | skip_red Empty = Empty
  | skip_red (Branch (B, va, vb, vc, vd)) = Branch (B, va, vb, vc, vd);

fun skip_black t =
  let
    val ta = skip_red t;
  in
    (case ta of Empty => ta | Branch (R, l, _, _, _) => ta
      | Branch (B, l, _, _, _) => l)
  end;

fun compare_height sx s t tx =
  (case (skip_red sx, (skip_red s, (skip_red t, skip_red tx)))
    of (Empty, (Empty, (_, Empty))) => EQ
    | (Empty, (Empty, (_, Branch (_, _, _, _, _)))) => LT
    | (Empty, (Branch (_, sa, _, _, _), (Empty, b))) => EQ
    | (Empty, (Branch (_, sa, _, _, _), (Branch (_, ta, _, _, _), Empty))) => EQ
    | (Empty,
        (Branch (_, sa, _, _, _),
          (Branch (_, ta, _, _, _), Branch (_, txa, _, _, _))))
      => compare_height Empty sa ta (skip_black txa)
    | (Branch (_, sxa, _, _, _), (Empty, (Empty, Empty))) => GT
    | (Branch (_, sxa, _, _, _), (Empty, (Empty, Branch (_, _, _, _, _)))) => LT
    | (Branch (_, sxa, _, _, _), (Empty, (Branch (_, _, _, _, _), Empty))) => EQ
    | (Branch (_, sxa, _, _, _),
        (Empty, (Branch (_, _, _, _, _), Branch (_, _, _, _, _))))
      => LT
    | (Branch (_, sxa, _, _, _), (Branch (_, sa, _, _, _), (Empty, xf))) => GT
    | (Branch (_, sxa, _, _, _),
        (Branch (_, sa, _, _, _), (Branch (_, ta, _, _, _), Empty)))
      => compare_height (skip_black sxa) sa ta Empty
    | (Branch (_, sxa, _, _, _),
        (Branch (_, sa, _, _, _),
          (Branch (_, ta, _, _, _), Branch (_, txa, _, _, _))))
      => compare_height (skip_black sxa) sa ta (skip_black txa));

fun apfst f (x, y) = (f x, y);

fun divmod_nat m n = (div_nata m n, mod_nat m n);

fun rbtreeify_g n kvs =
  (if equal_nata n zero_nata orelse equal_nata n one_nata then (Empty, kvs)
    else let
           val (na, r) = divmod_nat n (nat_of_integer (2 : IntInf.int));
         in
           (if equal_nata r zero_nata
             then let
                    val (t1, (k, v) :: kvsa) = rbtreeify_g na kvs;
                  in
                    apfst (fn a => Branch (B, t1, k, v, a))
                      (rbtreeify_g na kvsa)
                  end
             else let
                    val (t1, (k, v) :: kvsa) = rbtreeify_f na kvs;
                  in
                    apfst (fn a => Branch (B, t1, k, v, a))
                      (rbtreeify_g na kvsa)
                  end)
         end)
and rbtreeify_f n kvs =
  (if equal_nata n zero_nata then (Empty, kvs)
    else (if equal_nata n one_nata
           then let
                  val (k, v) :: kvsa = kvs;
                in
                  (Branch (R, Empty, k, v, Empty), kvsa)
                end
           else let
                  val (na, r) = divmod_nat n (nat_of_integer (2 : IntInf.int));
                in
                  (if equal_nata r zero_nata
                    then let
                           val (t1, (k, v) :: kvsa) = rbtreeify_f na kvs;
                         in
                           apfst (fn a => Branch (B, t1, k, v, a))
                             (rbtreeify_g na kvsa)
                         end
                    else let
                           val (t1, (k, v) :: kvsa) = rbtreeify_f na kvs;
                         in
                           apfst (fn a => Branch (B, t1, k, v, a))
                             (rbtreeify_f na kvsa)
                         end)
                end));

fun rbtreeify kvs = fst (rbtreeify_g (suc (size_list kvs)) kvs);

fun folda f (Branch (c, lt, k, v, rt)) x = folda f rt (f k v (folda f lt x))
  | folda f Empty x = x;

fun rbt_union_with_key less f t1 t2 =
  (case compare_height t1 t1 t2 t2
    of LT =>
      folda (rbt_insert_with_key less (fn k => fn v => fn w => f k w v)) t1 t2
    | GT => folda (rbt_insert_with_key less f) t2 t1
    | EQ => rbtreeify (sunion_with less f (entriesa t1) (entriesa t2)));

fun rbt_union less = rbt_union_with_key less (fn _ => fn _ => fn rv => rv);

fun dflt_cmp le lt a b =
  (if lt a b then LESS else (if le a b then EQUAL else GREATER));

fun cmp_prod cmp1 cmp2 (a1, a2) (b1, b2) =
  (case cmp1 a1 b1 of LESS => LESS | EQUAL => cmp2 a2 b2 | GREATER => GREATER);

fun comp2lt cmp a b =
  (case cmp a b of LESS => true | EQUAL => false | GREATER => false);

fun until_frmlsn_impl (A1_, A2_) =
  rec_ltln Empty Empty (fn _ => Empty) (fn _ => Empty)
    (fn _ => fn _ =>
      rbt_union
        (comp2lt
          (cmp_prod
            (dflt_cmp
              (less_eq_ltln
                (A1_, (ord_preorder o preorder_order o order_linorder) A2_))
              (less_ltln
                (A1_, (ord_preorder o preorder_order o order_linorder) A2_)))
            (dflt_cmp
              (less_eq_ltln
                (A1_, (ord_preorder o preorder_order o order_linorder) A2_))
              (less_ltln
                (A1_, (ord_preorder o preorder_order o order_linorder) A2_))))))
    (fn _ => fn _ =>
      rbt_union
        (comp2lt
          (cmp_prod
            (dflt_cmp
              (less_eq_ltln
                (A1_, (ord_preorder o preorder_order o order_linorder) A2_))
              (less_ltln
                (A1_, (ord_preorder o preorder_order o order_linorder) A2_)))
            (dflt_cmp
              (less_eq_ltln
                (A1_, (ord_preorder o preorder_order o order_linorder) A2_))
              (less_ltln
                (A1_, (ord_preorder o preorder_order o order_linorder) A2_))))))
    (fn _ => fn xa => xa)
    (fn x => fn xa => fn xb => fn xc =>
      map2set_insert
        (rbt_insert
          (comp2lt
            (cmp_prod
              (dflt_cmp
                (less_eq_ltln
                  (A1_, (ord_preorder o preorder_order o order_linorder) A2_))
                (less_ltln
                  (A1_, (ord_preorder o preorder_order o order_linorder) A2_)))
              (dflt_cmp
                (less_eq_ltln
                  (A1_, (ord_preorder o preorder_order o order_linorder) A2_))
                (less_ltln
                  (A1_, (ord_preorder o preorder_order o order_linorder)
                          A2_))))))
        (x, xa)
        (rbt_union
          (comp2lt
            (cmp_prod
              (dflt_cmp
                (less_eq_ltln
                  (A1_, (ord_preorder o preorder_order o order_linorder) A2_))
                (less_ltln
                  (A1_, (ord_preorder o preorder_order o order_linorder) A2_)))
              (dflt_cmp
                (less_eq_ltln
                  (A1_, (ord_preorder o preorder_order o order_linorder) A2_))
                (less_ltln
                  (A1_, (ord_preorder o preorder_order o order_linorder)
                          A2_)))))
          xb xc))
    (fn _ => fn _ =>
      rbt_union
        (comp2lt
          (cmp_prod
            (dflt_cmp
              (less_eq_ltln
                (A1_, (ord_preorder o preorder_order o order_linorder) A2_))
              (less_ltln
                (A1_, (ord_preorder o preorder_order o order_linorder) A2_)))
            (dflt_cmp
              (less_eq_ltln
                (A1_, (ord_preorder o preorder_order o order_linorder) A2_))
              (less_ltln
                (A1_, (ord_preorder o preorder_order o order_linorder)
                        A2_))))));

fun memb_sorted (A1_, A2_) [] x = false
  | memb_sorted (A1_, A2_) (y :: xs) x =
    (if less A2_ y x then memb_sorted (A1_, A2_) xs x else eq A1_ x y);

fun lss_memb (A1_, A2_) x l =
  memb_sorted (A1_, (ord_preorder o preorder_order o order_linorder) A2_) l x;

fun it_to_list it s = it s (fn _ => true) (fn x => fn l => l @ [x]) [];

fun gen_subseteq ball1 mem2 s1 s2 = ball1 s1 (fn x => mem2 x s2);

fun gen_equal ss1 ss2 s1 s2 = ss1 s1 s2 andalso ss2 s2 s1;

fun gen_ball it s p = it s (fn x => x) (fn x => fn _ => p x) true;

fun build_F_impl (A1_, A2_) =
  (fn x => fn xa =>
    gen_image
      (fn xb =>
        foldli
          (it_to_list (map_iterator_dom o (foldli o it_to_list rm_iterateoi))
            xb))
      [] (glist_insert
           (gen_equal
             (gen_subseteq (gen_ball (fn xb => foldli (id xb)))
               (glist_member equal_nata))
             (gen_subseteq (gen_ball (fn xb => foldli (id xb)))
               (glist_member equal_nata))))
      (fn (xb, xc) =>
        gen_image (fn xd => foldli (id xd)) [] (glist_insert equal_nata)
          name_impl
          (filter
            (fn xd =>
              (if lss_memb (equal_ltln A1_, linorder_ltln (A1_, A2_))
                    (LTLnUntil (xb, xc)) (old_impl xd)
                then lss_memb (equal_ltln A1_, linorder_ltln (A1_, A2_)) xc
                       (old_impl xd)
                else true))
            x))
      (until_frmlsn_impl (A1_, A2_) xa));

fun lss_iteratei l = foldli l;

fun iteratei_set_op_list_it_lss_ops A_ s = lss_iteratei s;

fun comp2eq cmp a b =
  (case cmp a b of LESS => false | EQUAL => true | GREATER => false);

fun pn_map_code (A1_, A2_) x =
  foldli x (fn _ => true)
    (fn xa => fn sigma =>
      let
        val xaa =
          iteratei_set_op_list_it_lss_ops (linorder_ltln (A1_, A2_))
            (old_impl xa) (fn _ => true)
            (fn xaa => fn (a, b) =>
              (case xaa of LTLnTrue => (a, b) | LTLnFalse => (a, b)
                | LTLnProp aa =>
                  (glist_insert
                     (comp2eq
                       (dflt_cmp
                         (less_eq
                           ((ord_preorder o preorder_order o order_linorder)
                             A2_))
                         (less ((ord_preorder o preorder_order o order_linorder)
                                 A2_))))
                     aa a,
                    b)
                | LTLnNProp aa =>
                  (a, glist_insert
                        (comp2eq
                          (dflt_cmp
                            (less_eq
                              ((ord_preorder o preorder_order o order_linorder)
                                A2_))
                            (less ((ord_preorder o preorder_order o
                                     order_linorder)
                                    A2_))))
                        aa b)
                | LTLnAnd (_, _) => (a, b) | LTLnOr (_, _) => (a, b)
                | LTLnNext _ => (a, b) | LTLnUntil (_, _) => (a, b)
                | LTLnRelease (_, _) => (a, b)))
            ([], []);
      in
        iam_update (name_impl xa) xaa sigma
      end)
    (iam_empty ());

fun rbt_lookup less (Branch (uu, l, x, y, r)) k =
  (if less k x then rbt_lookup less l k
    else (if less x k then rbt_lookup less r k else SOME y))
  | rbt_lookup less Empty k = NONE;

fun cr_rename_gba_code (A1_, A2_) x y =
  let
    val xa = gen_image foldli [] (glist_insert equal_nata) name_impl x;
    val xaa =
      gen_image foldli [] (glist_insert equal_nata) name_impl
        (filter
          (fn xc =>
            map2set_memb (fn k => fn t => rbt_lookup less_nat t k) zero_nata
              (incoming_impl xc))
          x);
    val xb = build_succ_code A2_ x;
    val xc = (fn xe => the_default [] (iam_alpha xb xe));
    val xd = build_F_impl (A1_, A2_) x y;
    val xe = pn_map_code (A1_, A2_) x;
    val xf =
      (fn xh => fn xi =>
        (case iam_alpha xe xh of NONE => false
          | SOME (xj, xk) =>
            gen_ball foldli xj xi andalso
              gen_ball foldli xk (fn xl => not (xi xl))));
  in
    Gen_frg_impl_ext
      (xa, xc, xaa, Gen_gbg_impl_ext (xd, Gen_gba_impl_ext (xf, ())))
  end;

fun insertion_sort (A1_, A2_) x [] = [x]
  | insertion_sort (A1_, A2_) x (y :: xs) =
    (if less A2_ y x then y :: insertion_sort (A1_, A2_) x xs
      else (if eq A1_ x y then y :: xs else x :: y :: xs));

fun lss_ins (A1_, A2_) x l =
  insertion_sort (A1_, (ord_preorder o preorder_order o order_linorder) A2_) x
    l;

fun lss_ins_dj (A1_, A2_) = lss_ins (A1_, A2_);

fun lss_empty x = (fn _ => []) x;

fun next_impl_update next_impla
  (Node_impl_ext
    (name_impl, incoming_impl, new_impl, old_impl, next_impl, more))
  = Node_impl_ext
      (name_impl, incoming_impl, new_impl, old_impl, next_impla next_impl,
        more);

fun name_impl_update name_impla
  (Node_impl_ext
    (name_impl, incoming_impl, new_impl, old_impl, next_impl, more))
  = Node_impl_ext
      (name_impla name_impl, incoming_impl, new_impl, old_impl, next_impl,
        more);

fun old_impl_update old_impla
  (Node_impl_ext
    (name_impl, incoming_impl, new_impl, old_impl, next_impl, more))
  = Node_impl_ext
      (name_impl, incoming_impl, new_impl, old_impla old_impl, next_impl, more);

fun new_impl_update new_impla
  (Node_impl_ext
    (name_impl, incoming_impl, new_impl, old_impl, next_impl, more))
  = Node_impl_ext
      (name_impl, incoming_impl, new_impla new_impl, old_impl, next_impl, more);

fun iteratei_bset_op_list_it_lss_basic_ops A_ s = lss_iteratei s;

fun g_ball_lss_basic_ops A_ s p =
  iteratei_bset_op_list_it_lss_basic_ops A_ s (fn c => c) (fn x => fn _ => p x)
    true;

fun g_subset_lss_basic_ops (A1_, A2_) s1 s2 =
  g_ball_lss_basic_ops A2_ s1 (fn x => lss_memb (A1_, A2_) x s2);

fun g_equal_lss_basic_ops (A1_, A2_) s1 s2 =
  g_subset_lss_basic_ops (A1_, A2_) s1 s2 andalso
    g_subset_lss_basic_ops (A1_, A2_) s2 s1;

fun g_sel_lss_basic_ops A_ s p =
  iteratei_bset_op_list_it_lss_basic_ops A_ s is_none
    (fn x => fn _ => (if p x then SOME x else NONE)) NONE;

fun next_impl
  (Node_impl_ext
    (name_impl, incoming_impl, new_impl, old_impl, next_impl, more))
  = next_impl;

fun new_impl
  (Node_impl_ext
    (name_impl, incoming_impl, new_impl, old_impl, next_impl, more))
  = new_impl;

fun incoming_impl_update incoming_impla
  (Node_impl_ext
    (name_impl, incoming_impl, new_impl, old_impl, next_impl, more))
  = Node_impl_ext
      (name_impl, incoming_impla incoming_impl, new_impl, old_impl, next_impl,
        more);

fun upd_incoming_impl (A1_, A2_) =
  (fn x =>
    map (fn xa =>
          (if g_equal_lss_basic_ops (equal_ltln A1_, linorder_ltln (A1_, A2_))
                (old_impl xa) (old_impl x) andalso
                g_equal_lss_basic_ops (equal_ltln A1_, linorder_ltln (A1_, A2_))
                  (next_impl xa) (next_impl x)
            then incoming_impl_update
                   (fn _ =>
                     rbt_union less_nat (incoming_impl x) (incoming_impl xa))
                   xa
            else xa)));

fun merge (A1_, A2_) [] l2 = l2
  | merge (A1_, A2_) (v :: va) [] = v :: va
  | merge (A1_, A2_) (x1 :: l1) (x2 :: l2) =
    (if less ((ord_preorder o preorder_order o order_linorder) A2_) x1 x2
      then x1 :: merge (A1_, A2_) l1 (x2 :: l2)
      else (if eq A1_ x1 x2 then x1 :: merge (A1_, A2_) l1 l2
             else x2 :: merge (A1_, A2_) (x1 :: l1) l2));

fun lss_union (A1_, A2_) s1 s2 = merge (A1_, A2_) s1 s2;

fun lss_union_dj (A1_, A2_) = lss_union (A1_, A2_);

fun lss_isEmpty s = null s;

fun delete_sorted (A1_, A2_) x [] = []
  | delete_sorted (A1_, A2_) x (y :: xs) =
    (if less A2_ y x then y :: delete_sorted (A1_, A2_) x xs
      else (if eq A1_ x y then xs else y :: xs));

fun lss_delete (A1_, A2_) x l =
  delete_sorted (A1_, (ord_preorder o preorder_order o order_linorder) A2_) x l;

fun gen_bex it s p = it s not (fn x => fn _ => p x) false;

fun expand_code_0 (A1_, A2_) x =
  let
    val (a, b) = x;
  in
    (if lss_isEmpty (new_impl a)
      then (if gen_bex (fn xa => foldli (id xa)) b
                 (fn xc =>
                   g_equal_lss_basic_ops
                     (equal_ltln A1_, linorder_ltln (A1_, A2_)) (old_impl xc)
                     (old_impl a) andalso
                     g_equal_lss_basic_ops
                       (equal_ltln A1_, linorder_ltln (A1_, A2_)) (next_impl xc)
                       (next_impl a))
             then DRETURN (name_impl a, upd_incoming_impl (A1_, A2_) a b)
             else expand_code_0 (A1_, A2_)
                    (Node_impl_ext
                       (suc (name_impl a),
                         map2set_insert (rbt_inserta ord_nat) (name_impl a)
                           Empty,
                         next_impl a, lss_empty (), lss_empty (), ()),
                      a :: b))
      else dbind (DRETURN
                   (the (g_sel_lss_basic_ops (linorder_ltln (A1_, A2_))
                          (new_impl a) (fn _ => true))))
             (fn xa =>
               let
                 val xb =
                   new_impl_update
                     (fn _ =>
                       lss_delete (equal_ltln A1_, linorder_ltln (A1_, A2_)) xa
                         (new_impl a))
                     a;
               in
                 (case xa
                   of LTLnTrue =>
                     expand_code_0 (A1_, A2_)
                       (old_impl_update
                          (fn _ =>
                            lss_union (equal_ltln A1_, linorder_ltln (A1_, A2_))
                              (lss_ins_dj
                                (equal_ltln A1_, linorder_ltln (A1_, A2_)) xa
                                (lss_empty ()))
                              (old_impl xb))
                          xb,
                         b)
                   | LTLnFalse => DRETURN (name_impl xb, b)
                   | LTLnProp aa =>
                     (if lss_memb (equal_ltln A1_, linorder_ltln (A1_, A2_))
                           (LTLnNProp aa) (old_impl xb)
                       then DRETURN (name_impl xb, b)
                       else expand_code_0 (A1_, A2_)
                              (old_impl_update
                                 (fn _ =>
                                   lss_union
                                     (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                     (lss_ins_dj
                                       (equal_ltln A1_,
 linorder_ltln (A1_, A2_))
                                       xa (lss_empty ()))
                                     (old_impl xb))
                                 xb,
                                b))
                   | LTLnNProp aa =>
                     (if lss_memb (equal_ltln A1_, linorder_ltln (A1_, A2_))
                           (LTLnProp aa) (old_impl xb)
                       then DRETURN (name_impl xb, b)
                       else expand_code_0 (A1_, A2_)
                              (old_impl_update
                                 (fn _ =>
                                   lss_union
                                     (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                     (lss_ins_dj
                                       (equal_ltln A1_,
 linorder_ltln (A1_, A2_))
                                       xa (lss_empty ()))
                                     (old_impl xb))
                                 xb,
                                b))
                   | LTLnAnd (ltln1, ltln2) =>
                     expand_code_0 (A1_, A2_)
                       (next_impl_update (fn _ => next_impl xb)
                          (old_impl_update
                            (fn _ =>
                              lss_union
                                (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                (lss_ins_dj
                                  (equal_ltln A1_, linorder_ltln (A1_, A2_)) xa
                                  (lss_empty ()))
                                (old_impl xb))
                            (new_impl_update
                              (fn _ =>
                                lss_ins
                                  (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                  ltln1
                                  (lss_ins
                                    (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                    ltln2 (new_impl xb)))
                              xb)),
                         b)
                   | LTLnOr (ltln1, ltln2) =>
                     dbind (expand_code_0 (A1_, A2_)
                             (next_impl_update
                                (fn _ =>
                                  lss_union_dj
                                    (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                    (lss_empty ()) (next_impl xb))
                                (old_impl_update
                                  (fn _ =>
                                    lss_ins
                                      (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                      xa (old_impl xb))
                                  (new_impl_update
                                    (fn _ =>
                                      lss_ins
(equal_ltln A1_, linorder_ltln (A1_, A2_)) ltln1 (new_impl xb))
                                    xb)),
                               b))
                       (fn (aa, ba) =>
                         expand_code_0 (A1_, A2_)
                           (old_impl_update
                              (fn _ =>
                                lss_union
                                  (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                  (lss_ins_dj
                                    (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                    xa (lss_empty ()))
                                  (old_impl xb))
                              (new_impl_update
                                (fn _ =>
                                  lss_union
                                    (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                    (lss_ins_dj
                                      (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                      ltln2 (lss_empty ()))
                                    (new_impl xb))
                                (name_impl_update (fn _ => aa) xb)),
                             ba))
                   | LTLnNext ltln =>
                     expand_code_0 (A1_, A2_)
                       (next_impl_update
                          (fn _ =>
                            lss_ins (equal_ltln A1_, linorder_ltln (A1_, A2_))
                              ltln (next_impl xb))
                          (old_impl_update
                            (fn _ =>
                              lss_union
                                (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                (lss_ins_dj
                                  (equal_ltln A1_, linorder_ltln (A1_, A2_)) xa
                                  (lss_empty ()))
                                (old_impl xb))
                            (new_impl_update (fn _ => new_impl xb) xb)),
                         b)
                   | LTLnUntil (ltln1, ltln2) =>
                     dbind (expand_code_0 (A1_, A2_)
                             (next_impl_update
                                (fn _ =>
                                  lss_union
                                    (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                    (lss_ins_dj
                                      (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                      xa (lss_empty ()))
                                    (next_impl xb))
                                (old_impl_update
                                  (fn _ =>
                                    lss_ins
                                      (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                      xa (old_impl xb))
                                  (new_impl_update
                                    (fn _ =>
                                      lss_ins
(equal_ltln A1_, linorder_ltln (A1_, A2_)) ltln1 (new_impl xb))
                                    xb)),
                               b))
                       (fn (aa, ba) =>
                         expand_code_0 (A1_, A2_)
                           (old_impl_update
                              (fn _ =>
                                lss_union
                                  (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                  (lss_ins_dj
                                    (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                    xa (lss_empty ()))
                                  (old_impl xb))
                              (new_impl_update
                                (fn _ =>
                                  lss_union
                                    (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                    (lss_ins_dj
                                      (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                      ltln2 (lss_empty ()))
                                    (new_impl xb))
                                (name_impl_update (fn _ => aa) xb)),
                             ba))
                   | LTLnRelease (ltln1, ltln2) =>
                     dbind (expand_code_0 (A1_, A2_)
                             (next_impl_update
                                (fn _ =>
                                  lss_union
                                    (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                    (lss_ins_dj
                                      (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                      xa (lss_empty ()))
                                    (next_impl xb))
                                (old_impl_update
                                  (fn _ =>
                                    lss_ins
                                      (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                      xa (old_impl xb))
                                  (new_impl_update
                                    (fn _ =>
                                      lss_ins
(equal_ltln A1_, linorder_ltln (A1_, A2_)) ltln2 (new_impl xb))
                                    xb)),
                               b))
                       (fn (aa, ba) =>
                         expand_code_0 (A1_, A2_)
                           (old_impl_update
                              (fn _ =>
                                lss_union
                                  (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                  (lss_ins_dj
                                    (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                    xa (lss_empty ()))
                                  (old_impl xb))
                              (new_impl_update
                                (fn _ =>
                                  lss_union
                                    (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                    (lss_ins
                                      (equal_ltln A1_, linorder_ltln (A1_, A2_))
                                      ltln1
                                      (lss_ins_dj
(equal_ltln A1_, linorder_ltln (A1_, A2_)) ltln2 (lss_empty ())))
                                    (new_impl xb))
                                (name_impl_update (fn _ => aa) xb)),
                             ba)))
               end))
  end;

fun expand_code (A1_, A2_) n_ns = the_res (expand_code_0 (A1_, A2_) n_ns);

fun create_graph_code (A1_, A2_) phi =
  let
    val (_, b) =
      expand_code (A1_, A2_)
        (Node_impl_ext
           (suc zero_nata, map2set_insert (rbt_inserta ord_nat) zero_nata Empty,
             lss_ins_dj (equal_ltln A1_, linorder_ltln (A1_, A2_)) phi
               (lss_empty ()),
             lss_empty (), lss_empty (), ()),
          []);
  in
    b
  end;

fun create_name_gba_code (A1_, A2_) phi =
  let
    val x = create_graph_code (A1_, A2_) phi;
  in
    cr_rename_gba_code (A1_, A2_) x phi
  end;

fun frgi_V0 (Gen_frg_impl_ext (frgi_V, frgi_E, frgi_V0, more)) = frgi_V0;

fun gbai_L
  (Gen_frg_impl_ext
    (frgi_V, frgi_E, frgi_V0,
      Gen_gbg_impl_ext (gbgi_F, Gen_gba_impl_ext (gbai_L, more))))
  = gbai_L;

fun gbgi_F
  (Gen_frg_impl_ext (frgi_V, frgi_E, frgi_V0, Gen_gbg_impl_ext (gbgi_F, more)))
  = gbgi_F;

fun frgi_V (Gen_frg_impl_ext (frgi_V, frgi_E, frgi_V0, more)) = frgi_V;

fun frgi_E (Gen_frg_impl_ext (frgi_V, frgi_E, frgi_V0, more)) = frgi_E;

fun set_bit_integer x i b = Bits_Integer.set_bit x (integer_of_nat i) b;

fun bs_insert i s = set_bit_integer s i true;

fun bs_empty x = (fn _ => 0) x;

fun gbg_to_idx_ext_code def_size eq bhc ecnv g =
  let
    val (a, b) =
      let
        val x = id (gbgi_F g);
        val xa = size_list x;
        val a =
          let
            val xb = ahm_emptyb def_size;
            val (_, b) =
              foldli x (fn _ => true)
                (fn xc => fn (a, b) =>
                  let
                    val ba =
                      foldli (id xc) (fn _ => true)
                        (fn xd => fn sigma =>
                          ahm_updateb eq bhc xd
                            (bs_insert a
                              (the_default (bs_empty ())
                                (ahm_lookupb eq bhc xd sigma)))
                            sigma)
                        b;
                  in
                    (suc a, ba)
                  end)
                (zero_nata, xb);
          in
            (fn xg => the_default (bs_empty ()) (ahm_lookupb eq bhc xg b))
          end;
      in
        (xa, a)
      end;
  in
    Gen_frg_impl_ext
      (frgi_V g, frgi_E g, frgi_V0 g, Gen_igbg_impl_ext (a, b, ecnv g))
  end;

fun gba_to_idx_ext_code eq bhc def_size ecnv g =
  gbg_to_idx_ext_code eq bhc def_size
    (fn xa => Gen_igba_impl_ext (gbai_L xa, ecnv xa)) g;

fun stat_set_data ns ni na =
  Gerth_Statistics.set_data (integer_of_nat ns) (integer_of_nat ni)
    (integer_of_nat na);

fun gen_card D_ it s = it s (fn _ => true) (fn _ => suc) (zero D_);

fun create_name_igba_code (A1_, A2_) phi =
  let
    val x = create_name_gba_code (A1_, A2_) phi;
    val xa =
      gba_to_idx_ext_code (def_hashmap_size_nat Type) equal_nata
        (bounded_hashcode_nat hashable_nat) (fn _ => ()) x;
    val _ =
      stat_set_data (gen_card zero_nat foldli (frgi_V x))
        (gen_card zero_nat foldli (frgi_V0 xa)) (igbgi_num_acc xa);
  in
    xa
  end;

fun ltln_pure_universal_frmls_impl (LTLnRelease (LTLnFalse, phi)) = true
  | ltln_pure_universal_frmls_impl (LTLnAnd (nu, mu)) =
    ltln_pure_universal_frmls_impl nu andalso ltln_pure_universal_frmls_impl mu
  | ltln_pure_universal_frmls_impl (LTLnOr (nu, mu)) =
    ltln_pure_universal_frmls_impl nu andalso ltln_pure_universal_frmls_impl mu
  | ltln_pure_universal_frmls_impl (LTLnUntil (nu, mu)) =
    ltln_pure_universal_frmls_impl nu andalso ltln_pure_universal_frmls_impl mu
  | ltln_pure_universal_frmls_impl (LTLnRelease (LTLnTrue, mu)) =
    ltln_pure_universal_frmls_impl LTLnTrue andalso
      ltln_pure_universal_frmls_impl mu
  | ltln_pure_universal_frmls_impl (LTLnRelease (LTLnProp v, mu)) =
    ltln_pure_universal_frmls_impl (LTLnProp v) andalso
      ltln_pure_universal_frmls_impl mu
  | ltln_pure_universal_frmls_impl (LTLnRelease (LTLnNProp v, mu)) =
    ltln_pure_universal_frmls_impl (LTLnNProp v) andalso
      ltln_pure_universal_frmls_impl mu
  | ltln_pure_universal_frmls_impl (LTLnRelease (LTLnAnd (v, va), mu)) =
    ltln_pure_universal_frmls_impl (LTLnAnd (v, va)) andalso
      ltln_pure_universal_frmls_impl mu
  | ltln_pure_universal_frmls_impl (LTLnRelease (LTLnOr (v, va), mu)) =
    ltln_pure_universal_frmls_impl (LTLnOr (v, va)) andalso
      ltln_pure_universal_frmls_impl mu
  | ltln_pure_universal_frmls_impl (LTLnRelease (LTLnNext v, mu)) =
    ltln_pure_universal_frmls_impl (LTLnNext v) andalso
      ltln_pure_universal_frmls_impl mu
  | ltln_pure_universal_frmls_impl (LTLnRelease (LTLnUntil (v, va), mu)) =
    ltln_pure_universal_frmls_impl (LTLnUntil (v, va)) andalso
      ltln_pure_universal_frmls_impl mu
  | ltln_pure_universal_frmls_impl (LTLnRelease (LTLnRelease (v, va), mu)) =
    ltln_pure_universal_frmls_impl (LTLnRelease (v, va)) andalso
      ltln_pure_universal_frmls_impl mu
  | ltln_pure_universal_frmls_impl LTLnTrue = false
  | ltln_pure_universal_frmls_impl LTLnFalse = false
  | ltln_pure_universal_frmls_impl (LTLnProp v) = false
  | ltln_pure_universal_frmls_impl (LTLnNProp v) = false
  | ltln_pure_universal_frmls_impl (LTLnNext v) = false;

fun ltln_pure_eventual_frmls_impl (LTLnUntil (LTLnTrue, phi)) = true
  | ltln_pure_eventual_frmls_impl (LTLnAnd (nu, mu)) =
    ltln_pure_eventual_frmls_impl nu andalso ltln_pure_eventual_frmls_impl mu
  | ltln_pure_eventual_frmls_impl (LTLnOr (nu, mu)) =
    ltln_pure_eventual_frmls_impl nu andalso ltln_pure_eventual_frmls_impl mu
  | ltln_pure_eventual_frmls_impl (LTLnUntil (LTLnFalse, mu)) =
    ltln_pure_eventual_frmls_impl LTLnFalse andalso
      ltln_pure_eventual_frmls_impl mu
  | ltln_pure_eventual_frmls_impl (LTLnUntil (LTLnProp v, mu)) =
    ltln_pure_eventual_frmls_impl (LTLnProp v) andalso
      ltln_pure_eventual_frmls_impl mu
  | ltln_pure_eventual_frmls_impl (LTLnUntil (LTLnNProp v, mu)) =
    ltln_pure_eventual_frmls_impl (LTLnNProp v) andalso
      ltln_pure_eventual_frmls_impl mu
  | ltln_pure_eventual_frmls_impl (LTLnUntil (LTLnAnd (v, va), mu)) =
    ltln_pure_eventual_frmls_impl (LTLnAnd (v, va)) andalso
      ltln_pure_eventual_frmls_impl mu
  | ltln_pure_eventual_frmls_impl (LTLnUntil (LTLnOr (v, va), mu)) =
    ltln_pure_eventual_frmls_impl (LTLnOr (v, va)) andalso
      ltln_pure_eventual_frmls_impl mu
  | ltln_pure_eventual_frmls_impl (LTLnUntil (LTLnNext v, mu)) =
    ltln_pure_eventual_frmls_impl (LTLnNext v) andalso
      ltln_pure_eventual_frmls_impl mu
  | ltln_pure_eventual_frmls_impl (LTLnUntil (LTLnUntil (v, va), mu)) =
    ltln_pure_eventual_frmls_impl (LTLnUntil (v, va)) andalso
      ltln_pure_eventual_frmls_impl mu
  | ltln_pure_eventual_frmls_impl (LTLnUntil (LTLnRelease (v, va), mu)) =
    ltln_pure_eventual_frmls_impl (LTLnRelease (v, va)) andalso
      ltln_pure_eventual_frmls_impl mu
  | ltln_pure_eventual_frmls_impl (LTLnRelease (nu, mu)) =
    ltln_pure_eventual_frmls_impl nu andalso ltln_pure_eventual_frmls_impl mu
  | ltln_pure_eventual_frmls_impl LTLnTrue = false
  | ltln_pure_eventual_frmls_impl LTLnFalse = false
  | ltln_pure_eventual_frmls_impl (LTLnProp v) = false
  | ltln_pure_eventual_frmls_impl (LTLnNProp v) = false
  | ltln_pure_eventual_frmls_impl (LTLnNext v) = false;

fun ltln_rewrite_step_impl A_ psi =
  (case psi of LTLnTrue => psi | LTLnFalse => psi | LTLnProp _ => psi
    | LTLnNProp _ => psi
    | LTLnAnd (psi_1, psi_2) =>
      (case (psi_1, psi_2) of (LTLnTrue, b) => psi | (LTLnFalse, b) => psi
        | (LTLnProp _, b) => psi | (LTLnNProp _, b) => psi
        | (LTLnAnd (_, _), b) => psi | (LTLnOr (_, _), b) => psi
        | (LTLnNext _, b) => psi | (LTLnUntil (phi, mu), LTLnTrue) => psi
        | (LTLnUntil (phi, mu), LTLnFalse) => psi
        | (LTLnUntil (phi, mu), LTLnProp _) => psi
        | (LTLnUntil (phi, mu), LTLnNProp _) => psi
        | (LTLnUntil (phi, mu), LTLnAnd (_, _)) => psi
        | (LTLnUntil (phi, mu), LTLnOr (_, _)) => psi
        | (LTLnUntil (phi, mu), LTLnNext _) => psi
        | (LTLnUntil (phi, mu), LTLnUntil (nu, mua)) =>
          (if equal_ltlna A_ mu mua then LTLnUntil (LTLnAnd (phi, nu), mu)
            else psi)
        | (LTLnUntil (phi, mu), LTLnRelease (_, _)) => psi
        | (LTLnRelease (phi, nu), LTLnTrue) => psi
        | (LTLnRelease (phi, nu), LTLnFalse) => psi
        | (LTLnRelease (phi, nu), LTLnProp _) => psi
        | (LTLnRelease (phi, nu), LTLnNProp _) => psi
        | (LTLnRelease (phi, nu), LTLnAnd (_, _)) => psi
        | (LTLnRelease (phi, nu), LTLnOr (_, _)) => psi
        | (LTLnRelease (phi, nu), LTLnNext _) => psi
        | (LTLnRelease (phi, nu), LTLnUntil (_, _)) => psi
        | (LTLnRelease (phi, nu), LTLnRelease (phia, mu)) =>
          (if equal_ltlna A_ phi phia then LTLnRelease (phi, LTLnAnd (nu, mu))
            else psi))
    | LTLnOr (psi_1, psi_2) =>
      (case (psi_1, psi_2) of (LTLnTrue, b) => psi | (LTLnFalse, b) => psi
        | (LTLnProp _, b) => psi | (LTLnNProp _, b) => psi
        | (LTLnAnd (_, _), b) => psi | (LTLnOr (_, _), b) => psi
        | (LTLnNext _, b) => psi | (LTLnUntil (phi, nu), LTLnTrue) => psi
        | (LTLnUntil (phi, nu), LTLnFalse) => psi
        | (LTLnUntil (phi, nu), LTLnProp _) => psi
        | (LTLnUntil (phi, nu), LTLnNProp _) => psi
        | (LTLnUntil (phi, nu), LTLnAnd (_, _)) => psi
        | (LTLnUntil (phi, nu), LTLnOr (_, _)) => psi
        | (LTLnUntil (phi, nu), LTLnNext _) => psi
        | (LTLnUntil (phi, nu), LTLnUntil (phia, mu)) =>
          (if equal_ltlna A_ phi phia then LTLnUntil (phi, LTLnOr (nu, mu))
            else psi)
        | (LTLnUntil (phi, nu), LTLnRelease (_, _)) => psi
        | (LTLnRelease (phi, mu), LTLnTrue) => psi
        | (LTLnRelease (phi, mu), LTLnFalse) => psi
        | (LTLnRelease (phi, mu), LTLnProp _) => psi
        | (LTLnRelease (phi, mu), LTLnNProp _) => psi
        | (LTLnRelease (phi, mu), LTLnAnd (_, _)) => psi
        | (LTLnRelease (phi, mu), LTLnOr (_, _)) => psi
        | (LTLnRelease (phi, mu), LTLnNext _) => psi
        | (LTLnRelease (phi, mu), LTLnUntil (_, _)) => psi
        | (LTLnRelease (phi, mu), LTLnRelease (nu, mua)) =>
          (if equal_ltlna A_ mu mua then LTLnRelease (LTLnOr (phi, nu), mu)
            else psi))
    | LTLnNext _ => psi
    | LTLnUntil (nu, mu) =>
      (if equal_ltlna A_ mu LTLnTrue then LTLnTrue
        else (case (nu, mu)
               of (LTLnTrue, LTLnTrue) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_eventual_frmls_impl mu then mu else psi))
               | (LTLnTrue, LTLnFalse) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_eventual_frmls_impl mu then mu else psi))
               | (LTLnTrue, LTLnProp _) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_eventual_frmls_impl mu then mu else psi))
               | (LTLnTrue, LTLnNProp _) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_eventual_frmls_impl mu then mu else psi))
               | (LTLnTrue, LTLnAnd (_, _)) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_eventual_frmls_impl mu then mu else psi))
               | (LTLnTrue, LTLnOr (_, _)) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_eventual_frmls_impl mu then mu else psi))
               | (LTLnTrue, LTLnNext _) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_eventual_frmls_impl mu then mu else psi))
               | (LTLnTrue, LTLnUntil (_, a)) => LTLnUntil (LTLnTrue, a)
               | (LTLnTrue, LTLnRelease (_, _)) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_eventual_frmls_impl mu then mu else psi))
               | (LTLnFalse, b) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_eventual_frmls_impl mu then mu else psi))
               | (LTLnProp _, b) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_eventual_frmls_impl mu then mu else psi))
               | (LTLnNProp _, b) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_eventual_frmls_impl mu then mu else psi))
               | (LTLnAnd (_, _), b) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_eventual_frmls_impl mu then mu else psi))
               | (LTLnOr (_, _), b) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_eventual_frmls_impl mu then mu else psi))
               | (LTLnNext _, b) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_eventual_frmls_impl mu then mu else psi))
               | (LTLnUntil (_, _), b) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_eventual_frmls_impl mu then mu else psi))
               | (LTLnRelease (_, _), b) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_eventual_frmls_impl mu then mu
                          else psi))))
    | LTLnRelease (nu, mu) =>
      (if equal_ltlna A_ mu LTLnFalse then LTLnFalse
        else (case (nu, mu)
               of (LTLnTrue, b) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_universal_frmls_impl mu then mu else psi))
               | (LTLnFalse, LTLnTrue) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_universal_frmls_impl mu then mu else psi))
               | (LTLnFalse, LTLnFalse) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_universal_frmls_impl mu then mu else psi))
               | (LTLnFalse, LTLnProp _) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_universal_frmls_impl mu then mu else psi))
               | (LTLnFalse, LTLnNProp _) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_universal_frmls_impl mu then mu else psi))
               | (LTLnFalse, LTLnAnd (_, _)) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_universal_frmls_impl mu then mu else psi))
               | (LTLnFalse, LTLnOr (_, _)) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_universal_frmls_impl mu then mu else psi))
               | (LTLnFalse, LTLnNext _) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_universal_frmls_impl mu then mu else psi))
               | (LTLnFalse, LTLnUntil (_, _)) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_universal_frmls_impl mu then mu else psi))
               | (LTLnFalse, LTLnRelease (_, a)) => LTLnRelease (LTLnFalse, a)
               | (LTLnProp _, b) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_universal_frmls_impl mu then mu else psi))
               | (LTLnNProp _, b) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_universal_frmls_impl mu then mu else psi))
               | (LTLnAnd (_, _), b) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_universal_frmls_impl mu then mu else psi))
               | (LTLnOr (_, _), b) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_universal_frmls_impl mu then mu else psi))
               | (LTLnNext _, b) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_universal_frmls_impl mu then mu else psi))
               | (LTLnUntil (_, _), b) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_universal_frmls_impl mu then mu else psi))
               | (LTLnRelease (_, _), b) =>
                 (if equal_ltlna A_ nu mu then nu
                   else (if ltln_pure_universal_frmls_impl mu then mu
                          else psi)))));

fun ltln_rewrite_rec A_ psi =
  (case ltln_rewrite_step_impl A_ psi of LTLnTrue => LTLnTrue
    | LTLnFalse => LTLnFalse | LTLnProp a => LTLnProp a
    | LTLnNProp a => LTLnNProp a
    | LTLnAnd (nu, mu) =>
      LTLnAnd (ltln_rewrite_rec A_ nu, ltln_rewrite_rec A_ mu)
    | LTLnOr (nu, mu) => LTLnOr (ltln_rewrite_rec A_ nu, ltln_rewrite_rec A_ mu)
    | LTLnNext nu => LTLnNext (ltln_rewrite_rec A_ nu)
    | LTLnUntil (nu, mu) =>
      LTLnUntil (ltln_rewrite_rec A_ nu, ltln_rewrite_rec A_ mu)
    | LTLnRelease (nu, mu) =>
      LTLnRelease (ltln_rewrite_rec A_ nu, ltln_rewrite_rec A_ mu));

fun ltln_rewrite A_ psi =
  let
    val phi = ltln_rewrite_rec A_ psi;
  in
    (if not (equal_ltlna A_ phi psi) then ltln_rewrite A_ phi else psi)
  end;

fun ltlc_to_ltl LTLcTrue = LTLTrue
  | ltlc_to_ltl LTLcFalse = LTLFalse
  | ltlc_to_ltl (LTLcProp q) = LTLProp q
  | ltlc_to_ltl (LTLcNeg phi) = LTLNeg (ltlc_to_ltl phi)
  | ltlc_to_ltl (LTLcAnd (phi, psi)) = LTLAnd (ltlc_to_ltl phi, ltlc_to_ltl psi)
  | ltlc_to_ltl (LTLcOr (phi, psi)) = LTLOr (ltlc_to_ltl phi, ltlc_to_ltl psi)
  | ltlc_to_ltl (LTLcImplies (phi, psi)) =
    LTLOr (LTLNeg (ltlc_to_ltl phi), ltlc_to_ltl psi)
  | ltlc_to_ltl (LTLcIff (phi, psi)) =
    let
      val phia = ltlc_to_ltl phi;
      val psia = ltlc_to_ltl psi;
    in
      LTLAnd (LTLOr (LTLNeg phia, psia), LTLOr (LTLNeg psia, phia))
    end
  | ltlc_to_ltl (LTLcNext phi) = LTLNext (ltlc_to_ltl phi)
  | ltlc_to_ltl (LTLcFinal phi) = LTLUntil (LTLTrue, ltlc_to_ltl phi)
  | ltlc_to_ltl (LTLcGlobal phi) = LTLRelease (LTLFalse, ltlc_to_ltl phi)
  | ltlc_to_ltl (LTLcUntil (phi, psi)) =
    LTLUntil (ltlc_to_ltl phi, ltlc_to_ltl psi)
  | ltlc_to_ltl (LTLcRelease (phi, psi)) =
    LTLRelease (ltlc_to_ltl phi, ltlc_to_ltl psi);

fun ltl_to_ltln LTLTrue = LTLnTrue
  | ltl_to_ltln LTLFalse = LTLnFalse
  | ltl_to_ltln (LTLProp q) = LTLnProp q
  | ltl_to_ltln (LTLNeg (LTLProp q)) = LTLnNProp q
  | ltl_to_ltln (LTLAnd (phi, psi)) = LTLnAnd (ltl_to_ltln phi, ltl_to_ltln psi)
  | ltl_to_ltln (LTLOr (phi, psi)) = LTLnOr (ltl_to_ltln phi, ltl_to_ltln psi)
  | ltl_to_ltln (LTLNext phi) = LTLnNext (ltl_to_ltln phi)
  | ltl_to_ltln (LTLUntil (phi, psi)) =
    LTLnUntil (ltl_to_ltln phi, ltl_to_ltln psi)
  | ltl_to_ltln (LTLRelease (phi, psi)) =
    LTLnRelease (ltl_to_ltln phi, ltl_to_ltln psi);

fun ltl_pushneg LTLTrue = LTLTrue
  | ltl_pushneg LTLFalse = LTLFalse
  | ltl_pushneg (LTLProp q) = LTLProp q
  | ltl_pushneg (LTLNeg LTLTrue) = LTLFalse
  | ltl_pushneg (LTLNeg LTLFalse) = LTLTrue
  | ltl_pushneg (LTLNeg (LTLProp q)) = LTLNeg (LTLProp q)
  | ltl_pushneg (LTLNeg (LTLNeg psi)) = ltl_pushneg psi
  | ltl_pushneg (LTLNeg (LTLAnd (nu, mu))) =
    LTLOr (ltl_pushneg (LTLNeg nu), ltl_pushneg (LTLNeg mu))
  | ltl_pushneg (LTLNeg (LTLOr (nu, mu))) =
    LTLAnd (ltl_pushneg (LTLNeg nu), ltl_pushneg (LTLNeg mu))
  | ltl_pushneg (LTLNeg (LTLNext psi)) = LTLNext (ltl_pushneg (LTLNeg psi))
  | ltl_pushneg (LTLNeg (LTLUntil (nu, mu))) =
    LTLRelease (ltl_pushneg (LTLNeg nu), ltl_pushneg (LTLNeg mu))
  | ltl_pushneg (LTLNeg (LTLRelease (nu, mu))) =
    LTLUntil (ltl_pushneg (LTLNeg nu), ltl_pushneg (LTLNeg mu))
  | ltl_pushneg (LTLAnd (phi, psi)) = LTLAnd (ltl_pushneg phi, ltl_pushneg psi)
  | ltl_pushneg (LTLOr (phi, psi)) = LTLOr (ltl_pushneg phi, ltl_pushneg psi)
  | ltl_pushneg (LTLNext phi) = LTLNext (ltl_pushneg phi)
  | ltl_pushneg (LTLUntil (phi, psi)) =
    LTLUntil (ltl_pushneg phi, ltl_pushneg psi)
  | ltl_pushneg (LTLRelease (phi, psi)) =
    LTLRelease (ltl_pushneg phi, ltl_pushneg psi);

fun gerth_ltl_to_gba_code (A1_, A2_) phi =
  create_name_igba_code (A1_, A2_)
    (ltln_rewrite A1_ (ltl_to_ltln (ltl_pushneg (ltlc_to_ltl phi))));

fun ltl_to_gba_code (A1_, A2_) cfg =
  let
    val CFG_L2B_GERTHS = cfg;
  in
    gerth_ltl_to_gba_code (A1_, A2_)
  end;

fun igbgi_acc
  (Gen_frg_impl_ext
    (frgi_V, frgi_E, frgi_V0,
      Gen_igbg_impl_ext (igbgi_num_acc, igbgi_acc, more)))
  = igbgi_acc;

fun igbai_L
  (Gen_frg_impl_ext
    (frgi_V, frgi_E, frgi_V0,
      Gen_igbg_impl_ext
        (igbgi_num_acc, igbgi_acc, Gen_igba_impl_ext (igbai_L, more))))
  = igbai_L;

fun sai_L
  (Gen_frg_impl_ext (frgi_V, frgi_E, frgi_V0, Gen_sa_impl_ext (sai_L, more))) =
  sai_L;

fun gbav_sys_prod_impl_cava_reorder eqq gi si =
  Gen_frg_impl_ext
    ((fn (x, xa) => glist_member eqq x (frgi_V gi) andalso frgi_V si xa),
      (fn (x, xa) =>
        (if igbai_L gi x (sai_L si xa)
          then maps (fn xaa => map (fn xb => (xb, xaa)) (frgi_E gi x))
                 (rev (frgi_E si xa))
          else [])),
      maps (fn x => map (fn a => (x, a)) (frgi_V0 si)) (frgi_V0 gi),
      Gen_igbg_impl_ext
        (igbgi_num_acc gi,
          (fn (x, xa) =>
            (if frgi_V si xa then igbgi_acc gi x else bs_empty ())),
          ()));

fun dflt_inter_impl eqq si gi =
  (gbav_sys_prod_impl_cava_reorder eqq gi si, snd);

fun graph_restrict_right_impl memr =
  (fn x => fn xa => fn xb => filter (fn xc => memr xc x) (xa xb));

fun graph_restrict_left_impl meml =
  (fn x => fn xa => fn xb => (if meml xb x then xa xb else []));

fun list_map_pick_remove [] = (raise Fail "undefined")
  | list_map_pick_remove (kv :: l) = (kv, l);

fun list_map_isEmpty x = null x;

fun is_None a = (case a of NONE => true | SOME _ => false);

fun wset_find_path_code (A1_, A2_) e u0 p =
  let
    val x =
      let
        val (a, b) =
          foldli u0 (fn _ => true)
            (fn x => fn (a, b) =>
              (map2set_insert (ahm_updateb (eq A1_) (bounded_hashcode_nat A2_))
                 x a,
                list_map_update (eq A1_) x [] b))
            (ahm_emptyb (def_hashmap_size A2_ Type), []);
      in
        (NONE, (a, b))
      end;
    val (a, (_, _)) =
      whilea (fn (xb, (_, xd)) => is_None xb andalso not (list_map_isEmpty xd))
        (fn (_, (aa, ba)) =>
          let
            val ((ac, bc), bb) = list_map_pick_remove ba;
          in
            (if p ac then (SOME (rev bc, ac), (aa, bb))
              else let
                     val (ad, bd) =
                       foldli (rev (e ac)) (fn _ => true)
                         (fn xc => fn (ad, bd) =>
                           (if map2set_memb
                                 (ahm_lookupb (eq A1_)
                                   (bounded_hashcode_nat A2_))
                                 xc ad
                             then (ad, bd)
                             else (map2set_insert
                                     (ahm_updateb (eq A1_)
                                       (bounded_hashcode_nat A2_))
                                     xc ad,
                                    (xc, ac :: bc) :: bd)))
                         (aa, bb);
                   in
                     (NONE, (ad, bd))
                   end)
          end)
        x;
  in
    a
  end;

fun find_path1_code (A1_, A2_) e u p =
  (case wset_find_path_code (A1_, A2_) e (e u) p of NONE => NONE
    | SOME (a, b) => SOME (u :: a, b));

fun atLeastLessThan_tr emp ins a b =
  let
    val (_, ba) =
      whilea (fn (xc, _) => less_nat xc b)
        (fn (aa, ba) => (plus_nata aa one_nata, ins aa ba)) (a, emp);
  in
    ba
  end;

fun bs_subset_eq s1 s2 =
  (((IntInf.andb (s1, IntInf.notb s2)) : IntInf.int) = 0);

fun bs_union s1 s2 = IntInf.orb (s1, s2);

fun bs_eq s1 s2 = ((s1 : IntInf.int) = s2);

fun reconstruct_lasso_tr (A1_, A2_) vr vl g_impl =
  let
    val (a, b) =
      let
        val a =
          wset_find_path_code (A1_, A2_)
            (graph_restrict_left_impl (fn x => fn f => f x) vr (frgi_E g_impl))
            (frgi_V0 g_impl) vl;
      in
        the a
      end;
    val xa =
      atLeastLessThan_tr (bs_empty ()) bs_insert zero_nata
        (igbgi_num_acc g_impl);
    val xb = graph_restrict_right_impl (fn x => fn f => f x) vl (frgi_E g_impl);
    val (aa, (ab, _)) =
      whilea (fn (_, (_, xi)) => not (bs_eq xi xa))
        (fn (aa, (ab, bb)) =>
          let
            val xd =
              wset_find_path_code (A1_, A2_) xb [aa]
                (fn ac => not (bs_subset_eq (igbgi_acc g_impl ac) bb));
            val (ac, bc) = the xd;
          in
            (bc, (ab @ ac, bs_union bb (igbgi_acc g_impl bc)))
          end)
        (b, ([], igbgi_acc g_impl b));
    val xd =
      (if is_Nil ab then find_path1_code (A1_, A2_) xb aa (eq A1_ b)
        else wset_find_path_code (A1_, A2_) xb [aa] (eq A1_ b));
    val (ac, _) = the xd;
  in
    (a, ab @ ac)
  end;

fun as_is_empty s = equal_nata (snd s) zero_nata;

fun as_length x = snd x;

fun as_top s = let
                 val (a, n) = s;
               in
                 array_get a (minus_nat n one_nata)
               end;

fun as_set s i x = let
                     val (a, b) = s;
                   in
                     (array_set a i x, b)
                   end;

fun as_shrink s =
  let
    val (a, n) = s;
    val aa =
      (if less_eq_nat (times_nata (nat_of_integer (128 : IntInf.int)) n)
            (array_length a) andalso
            less_nat (nat_of_integer (4 : IntInf.int)) n
        then array_shrink a n else a);
  in
    (aa, n)
  end;

fun as_pop s = let
                 val (a, n) = s;
               in
                 as_shrink (a, minus_nat n one_nata)
               end;

fun as_get s i = let
                   val (a, _) = s;
                 in
                   array_get a i
                 end;

fun select_edge_tr (A1_, A2_) s =
  let
    val (a, (aa, (ab, bb))) = s;
  in
    (if as_is_empty bb then (NONE, (a, (aa, (ab, bb))))
      else let
             val (ac, bc) = as_top bb;
           in
             (if less_eq_nat (as_get aa (minus_nat (as_length aa) one_nata)) ac
               then let
                      val xa = gen_pick (fn x => foldli (id x)) bc;
                      val xb = glist_delete (eq A1_) xa bc;
                      val xc =
                        (if is_Nil xb then as_pop bb
                          else as_set bb (minus_nat (as_length bb) one_nata)
                                 (ac, xb));
                    in
                      (SOME xa, (a, (aa, (ab, xc))))
                    end
               else (NONE, (a, (aa, (ab, bb)))))
           end)
  end;

fun go_is_no_brk_code A_ = (fn x => is_None (fst x));

fun go_is_done_code (A1_, A2_) =
  (fn x => fn xa =>
    (case ahm_lookupb (eq A1_) (bounded_hashcode_nat A2_) x (snd xa)
      of NONE => false
      | SOME i => (if less_eq_int zero_int i then false else true)));

fun as_singleton B_ x = (FArray.IsabelleMapping.array_of_list [x], one B_);

fun as_push s x =
  let
    val (a, n) = s;
    val aa =
      (if equal_nata n (array_length a)
        then array_grow a
               (max ord_nat (nat_of_integer (4 : IntInf.int))
                 (times_nata (nat_of_integer (2 : IntInf.int)) n))
               x
        else a);
    val ab = array_set aa n x;
  in
    (ab, plus_nata n one_nata)
  end;

fun push_code (A1_, A2_) g_impl =
  (fn x => fn (xa, (xb, (xc, xd))) =>
    let
      val _ = Gabow_Skeleton_Statistics.newnode ();
      val xf = as_length xa;
      val xg = as_push xa x;
      val xh = as_push xb xf;
      val xi =
        ahm_updateb (eq A1_) (bounded_hashcode_nat A2_) x (int_of_nat xf) xc;
      val xj =
        (if is_Nil (frgi_E g_impl x) then xd
          else as_push xd (xf, frgi_E g_impl x));
    in
      (xg, (xh, (xi, xj)))
    end);

fun find_max_nat n uu =
  (if equal_nata n zero_nata then zero_nata
    else (if uu (minus_nat n one_nata) then minus_nat n one_nata
           else find_max_nat (minus_nat n one_nata) uu));

fun idx_of_tr (A1_, A2_) s v =
  let
    val (_, (aa, (ab, _))) = v;
    val x =
      let
        val SOME i = ahm_lookupb (eq A1_) (bounded_hashcode_nat A2_) s ab;
        val true = less_eq_int zero_int i;
      in
        nat i
      end;
    val xa = find_max_nat (as_length aa) (fn j => less_eq_nat (as_get aa j) x);
  in
    xa
  end;

fun gto_outer_code A_ = (fn x => fn (_, (_, (_, (xe, _)))) => (x, xe));

fun goinitial_code A_ = (NONE, ahm_emptyb (def_hashmap_size A_ Type));

fun as_take m s =
  let
    val (a, n) = s;
  in
    (if less_nat m n then as_shrink (a, m) else (a, n))
  end;

fun pop_tr (A1_, A2_) s =
  let
    val (a, (aa, (ab, bb))) = s;
    val x = minus_nat (as_length aa) one_nata;
    val xa =
      let
        val (_, bc) =
          whilea
            (fn (xf, _) =>
              less_nat xf
                (if equal_nata (plus_nata x one_nata) (as_length aa)
                  then as_length a else as_get aa (plus_nata x one_nata)))
            (fn (ac, bc) =>
              (suc ac,
                ahm_updateb (eq A1_) (bounded_hashcode_nat A2_) (as_get a ac)
                  (uminus_int (Int_of_integer (1 : IntInf.int))) bc))
            (as_get aa x, ab);
      in
        bc
      end;
    val xb = as_take (as_top aa) a;
    val xc = as_pop aa;
  in
    (xb, (xc, (xa, bb)))
  end;

fun as_empty B_ uu = (FArray.IsabelleMapping.array_of_list [], zero B_);

fun goBrk_code A_ = fst;

fun un_set_drop_tr es_impl un_impl i a =
  let
    val (_, b) =
      whilea (fn (xc, _) => less_nat xc (as_length a))
        (fn (aa, b) =>
          let
            val xa = un_impl (as_get a aa) b;
            val xb = plus_nata aa one_nata;
          in
            (xb, xa)
          end)
        (i, es_impl);
  in
    b
  end;

fun all_interval_nat p i j =
  less_eq_nat j i orelse p i andalso all_interval_nat p (suc i) j;

fun bs_mem i s = test_bit_integer s i;

fun find_ce_tr (A1_, A2_) g_impl =
  let
    val _ = Gabow_Skeleton_Statistics.start ();
    val xa = goinitial_code A2_;
    val xb =
      foldli (id (frgi_V0 g_impl)) (go_is_no_brk_code A2_)
        (fn xb => fn sigma =>
          (if not (go_is_done_code (A1_, A2_) xb sigma)
            then let
                   val xc =
                     (NONE,
                       (as_singleton one_nat (igbgi_acc g_impl xb),
                         (as_singleton one_nat xb,
                           (as_singleton one_nat zero_nata,
                             (ahm_updateb (eq A1_) (bounded_hashcode_nat A2_) xb
                                (int_of_nat zero_nata) (snd sigma),
                               (if is_Nil (frgi_E g_impl xb)
                                 then as_empty zero_nat ()
                                 else as_singleton one_nat
(zero_nata, frgi_E g_impl xb)))))));
                   val (a, b) =
                     whilea
                       (fn (xf, xg) =>
                         is_None xf andalso
                           not (as_is_empty
                                 let
                                   val (xaa, (_, (_, _))) = snd xg;
                                 in
                                   xaa
                                 end))
                       (fn (_, b) =>
                         (case let
                                 val (aa, ba) = b;
                                 val (ab, bb) = select_edge_tr (A1_, A2_) ba;
                               in
                                 (ab, (aa, bb))
                               end
                           of (NONE, ba) =>
                             let
                               val a =
                                 let
                                   val (ab, bb) = ba;
                                   val xg = pop_tr (A1_, A2_) bb;
                                   val xh = as_pop ab;
                                 in
                                   (xh, xg)
                                 end;
                             in
                               (NONE, a)
                             end
                           | (SOME xf, ba) =>
                             (if (case ahm_lookupb (eq A1_)
 (bounded_hashcode_nat A2_) xf let
                                 val (_, (_, (xd, _))) = snd ba;
                               in
                                 xd
                               end
                                   of NONE => false
                                   | SOME i =>
                                     (if less_eq_int zero_int i then true
                                       else false))
                               then let
                                      val xg =
let
  val (ab, (ac, (ad, (ae, be)))) = ba;
  val xh = idx_of_tr (A1_, A2_) xf (ac, (ad, (ae, be)));
  val xi = as_take (plus_nata xh one_nata) ad;
  val xj = un_set_drop_tr (bs_empty ()) bs_union xh ab;
  val xk = as_push (as_take xh ab) xj;
in
  (xk, (ac, (xi, (ae, be))))
end;
                                    in
                                      (case
let
  val (ab, _) = xg;
in
  all_interval_nat (fn xca => bs_mem xca (as_top ab)) zero_nata
    (igbgi_num_acc g_impl)
end
of true =>
  let
    val (ab, (ac, (ad, (ae, be)))) = xg;
    val xj =
      (fn xfa =>
        (case ahm_lookupb (eq A1_) (bounded_hashcode_nat A2_) xfa ae
          of NONE => false
          | SOME i =>
            (if less_eq_int zero_int i then less_nat (nat i) (as_top ad)
              else false)));
    val xk =
      (fn xga =>
        (case ahm_lookupb (eq A1_) (bounded_hashcode_nat A2_) xga ae
          of NONE => false
          | SOME i =>
            (if less_eq_int zero_int i then less_eq_nat (as_top ad) (nat i)
              else false)));
  in
    (SOME (xj, xk), (ab, (ac, (ad, (ae, be)))))
  end
| false => (NONE, xg))
                                    end
                               else (if not
  (case ahm_lookupb (eq A1_) (bounded_hashcode_nat A2_) xf
          let
            val (_, (_, (xd, _))) = snd ba;
          in
            xd
          end
    of NONE => false
    | SOME i => (if less_eq_int zero_int i then false else true))
                                      then (NONE,
     let
       val (xba, xca) = ba;
     in
       (as_push xba (igbgi_acc g_impl xf), push_code (A1_, A2_) g_impl xf xca)
     end)
                                      else (NONE, ba)))))
                       xc;
                 in
                   gto_outer_code A2_ a b
                 end
            else sigma))
        xa;
    val _ = Gabow_Skeleton_Statistics.stop ();
  in
    goBrk_code A2_ xb
  end;

fun find_lasso_tr (A1_, A2_) g_impl =
  (case find_ce_tr (A1_, A2_) g_impl of NONE => NONE
    | SOME (a, b) =>
      let
        val aa = reconstruct_lasso_tr (A1_, A2_) a b g_impl;
      in
        SOME aa
      end);

fun map_option fi NONE = NONE
  | map_option fi (SOME x2a) = SOME (fi x2a);

fun tl [] = []
  | tl (x21 :: x22) = x22;

fun hd (x21 :: x22) = x21;

fun lasso_of_prpl prpl =
  let
    val (pr, pl) = prpl;
  in
    Lasso_ext (pr, hd pl, tl pl, ())
  end;

fun gabow_find_ce_code (A1_, A2_) =
  map_option (SOME o lasso_of_prpl) o find_lasso_tr (A1_, A2_);

fun degeneralize_ext_impl gi x =
  (if equal_nata (igbgi_num_acc gi) zero_nata
    then Gen_frg_impl_ext
           ((fn (xa, xb) => equal_nata xb zero_nata andalso frgi_V gi xa),
             (fn (xa, xb) =>
               (if equal_nata xb zero_nata
                 then map (fn xc => (xc, zero_nata)) (frgi_E gi xa) else [])),
             map (fn xa => (xa, zero_nata)) (frgi_V0 gi),
             Gen_bg_impl_ext
               ((fn (xa, xb) => equal_nata xb zero_nata andalso frgi_V gi xa),
                 x gi))
    else Gen_frg_impl_ext
           ((fn (xa, xb) =>
              less_nat xb (igbgi_num_acc gi) andalso frgi_V gi xa),
             (fn (xa, xb) =>
               (if less_nat xb (igbgi_num_acc gi)
                 then let
                        val xc =
                          (if bs_mem xb (igbgi_acc gi xa)
                            then mod_nat (plus_nata xb one_nata)
                                   (igbgi_num_acc gi)
                            else xb);
                      in
                        map (fn xd => (xd, xc)) (frgi_E gi xa)
                      end
                 else [])),
             map (fn xa => (xa, zero_nata)) (frgi_V0 gi),
             Gen_bg_impl_ext
               ((fn (xa, xb) =>
                  equal_nata xb zero_nata andalso
                    bs_mem zero_nata (igbgi_acc gi xa)),
                 x gi)));

fun equal_blue_witness A_ (REACH (v, lista)) (CIRC (list1, list2)) = false
  | equal_blue_witness A_ (CIRC (list1, list2)) (REACH (v, lista)) = false
  | equal_blue_witness A_ NO_CYC (CIRC (list1, list2)) = false
  | equal_blue_witness A_ (CIRC (list1, list2)) NO_CYC = false
  | equal_blue_witness A_ NO_CYC (REACH (v, lista)) = false
  | equal_blue_witness A_ (REACH (v, lista)) NO_CYC = false
  | equal_blue_witness A_ (CIRC (list1a, list2a)) (CIRC (list1, list2)) =
    equal_list A_ list1a list1 andalso equal_list A_ list2a list2
  | equal_blue_witness A_ (REACH (va, listaa)) (REACH (v, lista)) =
    eq A_ va v andalso equal_list A_ listaa lista
  | equal_blue_witness A_ NO_CYC NO_CYC = true;

fun new_hashmap_with A_ size = HashMapa (new_array [] size, zero_nata);

fun ahm_emptya A_ = (fn _ => new_hashmap_with A_ (def_hashmap_size A_ Type));

fun ahm_empty_const A_ = HashMap (ahm_emptya A_ ());

fun ahm_empty A_ = (fn _ => ahm_empty_const A_);

fun empty_ahm_basic_ops A_ = ahm_empty A_;

fun map_of A_ ((l, v) :: ps) k = (if eq A_ l k then SOME v else map_of A_ ps k)
  | map_of A_ [] k = NONE;

fun ahm_alpha_aux (A1_, A2_) a k =
  map_of A1_ (array_get a (bounded_hashcode_nat A2_ (array_length a) k)) k;

fun ahm_alpha (A1_, A2_) (HashMapa (a, uu)) = ahm_alpha_aux (A1_, A2_) a;

fun ahm_lookupa (A1_, A2_) k hm = ahm_alpha (A1_, A2_) hm k;

fun impl_ofa B_ (HashMap x) = x;

fun ahm_lookup (A1_, A2_) k hm = ahm_lookupa (A1_, A2_) k (impl_ofa A2_ hm);

fun memb_ahm_basic_ops (A1_, A2_) x s =
  not (is_none (ahm_lookup (A1_, A2_) x s));

fun delete A_ k = filter (fn (ka, _) => not (eq A_ k ka));

fun ahm_deletea (A1_, A2_) k (HashMapa (a, n)) =
  let
    val h = bounded_hashcode_nat A2_ (array_length a) k;
    val m = array_get a h;
    val deleted = not (is_none (map_of A1_ m k));
  in
    HashMapa
      (array_set a h (delete A1_ k m),
        (if deleted then minus_nat n one_nata else n))
  end;

fun ahm_delete (A1_, A2_) k hm =
  HashMap (ahm_deletea (A1_, A2_) k (impl_ofa A2_ hm));

fun delete_ahm_basic_ops (A1_, A2_) x s = ahm_delete (A1_, A2_) x s;

fun bgi_F
  (Gen_frg_impl_ext (frgi_V, frgi_E, frgi_V0, Gen_bg_impl_ext (bgi_F, more))) =
  bgi_F;

fun update A_ k v [] = [(k, v)]
  | update A_ k v (p :: ps) =
    (if eq A_ (fst p) k then (k, v) :: ps else p :: update A_ k v ps);

fun ahm_update_aux (A1_, A2_) (HashMapa (a, n)) k v =
  let
    val h = bounded_hashcode_nat A2_ (array_length a) k;
    val m = array_get a h;
    val insert = is_none (map_of A1_ m k);
  in
    HashMapa
      (array_set a h (update A1_ k v m),
        (if insert then plus_nata n one_nata else n))
  end;

fun idx_iteratei_aux_array_get sz i l c f sigma =
  (if equal_nata i zero_nata orelse not (c sigma) then sigma
    else idx_iteratei_aux_array_get sz (minus_nat i one_nata) l c f
           (f (array_get l (minus_nat sz i)) sigma));

fun idx_iteratei_array_length_array_get l c f sigma =
  idx_iteratei_aux_array_get (array_length l) (array_length l) l c f sigma;

fun ahm_iteratei_aux A_ a c f sigma =
  idx_iteratei_array_length_array_get a c (fn x => foldli x c f) sigma;

fun ahm_rehash_auxa A_ n kv a =
  let
    val h = bounded_hashcode_nat A_ n (fst kv);
  in
    array_set a h (kv :: array_get a h)
  end;

fun ahm_rehash_aux A_ a sz =
  ahm_iteratei_aux A_ a (fn _ => true) (ahm_rehash_auxa A_ sz)
    (new_array [] sz);

fun ahm_rehash A_ (HashMapa (a, n)) sz = HashMapa (ahm_rehash_aux A_ a sz, n);

val load_factor : nat = nat_of_integer (75 : IntInf.int);

fun ahm_filled A_ (HashMapa (a, n)) =
  less_eq_nat (times_nata (array_length a) load_factor)
    (times_nata n (nat_of_integer (100 : IntInf.int)));

fun hm_grow A_ (HashMapa (a, n)) =
  plus_nata (times_nata (nat_of_integer (2 : IntInf.int)) (array_length a))
    (nat_of_integer (3 : IntInf.int));

fun ahm_updatea (A1_, A2_) k v hm =
  let
    val hma = ahm_update_aux (A1_, A2_) hm k v;
  in
    (if ahm_filled A2_ hma then ahm_rehash A2_ hma (hm_grow A2_ hma) else hma)
  end;

fun ahm_update (A1_, A2_) k v hm =
  HashMap (ahm_updatea (A1_, A2_) k v (impl_ofa A2_ hm));

fun ins_ahm_basic_ops (A1_, A2_) x s = ahm_update (A1_, A2_) x () s;

fun init_wit_blue_early A_ s t =
  (if eq A_ s t then CIRC ([], [s]) else REACH (t, [s]));

fun red_init_witness u v = SOME ([u], v);

fun prep_wit_red v NONE = NONE
  | prep_wit_red v (SOME (p, u)) = SOME (v :: p, u);

fun code_red_dfs_ahs_0 (A1_, A2_) e onstack x =
  let
    val (a, b) = x;
    val xa = ins_ahm_basic_ops (A1_, A2_) b a;
  in
    dbind (DRETURN (NDFS_SI_Statistics.vis_red ()))
      (fn _ =>
        dbind (dbind (DRETURN (id (e b)))
                (fn xs =>
                  foldli xs
                    (fn aa =>
                      (case aa of DSUCCEEDi => false | DFAILi => false
                        | DRETURN ab => is_None ab))
                    (fn xb => fn s =>
                      dbind s
                        (fn _ =>
                          (if memb_ahm_basic_ops (A1_, A2_) xb onstack
                            then DRETURN (red_init_witness b xb)
                            else DRETURN NONE)))
                    (DRETURN NONE)))
          (fn xc =>
            (case xc
              of NONE =>
                dbind (DRETURN (id (e b)))
                  (fn xs =>
                    foldli xs
                      (fn aa =>
                        (case aa of DSUCCEEDi => false | DFAILi => false
                          | DRETURN (_, ab) => is_None ab))
                      (fn xb => fn s =>
                        dbind s
                          (fn (aa, _) =>
                            (if not (memb_ahm_basic_ops (A1_, A2_) xb aa)
                              then dbind (code_red_dfs_ahs_0 (A1_, A2_) e
   onstack (aa, xb))
                                     (fn (ab, bb) =>
                                       DRETURN (ab, prep_wit_red b bb))
                              else DRETURN (aa, NONE))))
                      (DRETURN (xa, NONE)))
              | SOME _ => DRETURN (xa, xc))))
  end;

fun code_red_dfs_ahs (A1_, A2_) e onstack v u =
  the_res (code_red_dfs_ahs_0 (A1_, A2_) e onstack (v, u));

fun prep_wit_blue A_ u0 NO_CYC = NO_CYC
  | prep_wit_blue A_ u0 (REACH (u, p)) =
    (if eq A_ u0 u then CIRC ([], u0 :: p) else REACH (u, u0 :: p))
  | prep_wit_blue A_ u0 (CIRC (pr, pl)) = CIRC (u0 :: pr, pl);

fun init_wit_blue A_ u0 NONE = NO_CYC
  | init_wit_blue A_ u0 (SOME (p, u)) =
    (if eq A_ u u0 then CIRC ([], p) else REACH (u, p));

fun code_blue_dfs_ahs_0 (A1_, A2_) g x =
  let
    val (ab, (ac, (ad, bd))) = x;
    val xc = ins_ahm_basic_ops (A1_, A2_) bd ab;
    val xd = ins_ahm_basic_ops (A1_, A2_) bd ad;
    val xe = bgi_F g bd;
  in
    dbind (DRETURN (NDFS_SI_Statistics.vis_blue ()))
      (fn _ =>
        dbind (dbind (DRETURN (id (frgi_E g bd)))
                (fn xs =>
                  foldli xs
                    (fn a =>
                      (case a of DSUCCEEDi => false | DFAILi => false
                        | DRETURN (_, (_, (_, xr))) =>
                          equal_blue_witness A1_ xr NO_CYC))
                    (fn xa => fn s =>
                      dbind s
                        (fn (ae, (af, (ag, bg))) =>
                          (if memb_ahm_basic_ops (A1_, A2_) xa ag andalso
                                (xe orelse bgi_F g xa)
                            then DRETURN
                                   (ae, (af,
  (ag, init_wit_blue_early A1_ bd xa)))
                            else (if not (memb_ahm_basic_ops (A1_, A2_) xa ae)
                                   then dbind
  (code_blue_dfs_ahs_0 (A1_, A2_) g (ae, (af, (ag, xa))))
  (fn (ah, (ai, (aj, bj))) => DRETURN (ah, (ai, (aj, prep_wit_blue A1_ bd bj))))
                                   else dbind
  (DRETURN (NDFS_SI_Statistics.match_blue ()))
  (fn _ => DRETURN (ae, (af, (ag, bg))))))))
                    (DRETURN (xc, (ac, (xd, NO_CYC))))))
          (fn (ae, (af, (ag, bg))) =>
            dbind (if equal_blue_witness A1_ bg NO_CYC andalso xe
                    then dbind (DRETURN
                                 (code_red_dfs_ahs (A1_, A2_) (frgi_E g) ag af
                                   bd))
                           (fn (ah, bh) =>
                             DRETURN (ah, init_wit_blue A1_ bd bh))
                    else DRETURN (af, bg))
              (fn (ah, bh) =>
                let
                  val xi = delete_ahm_basic_ops (A1_, A2_) bd ag;
                in
                  DRETURN (ae, (ah, (xi, bh)))
                end)))
  end;

fun extract_res cyc =
  (case cyc of NO_CYC => NONE | REACH (_, _) => NONE
    | CIRC (pr, pl) => SOME (pr, pl));

fun code_blue_dfs_ahs (A1_, A2_) g =
  the_res
    (dbind (DRETURN (NDFS_SI_Statistics.start ()))
      (fn _ =>
        dbind (dbind (DRETURN (id (frgi_V0 g)))
                (fn xs =>
                  foldli xs
                    (fn a =>
                      (case a of DSUCCEEDi => false | DFAILi => false
                        | DRETURN (_, (_, xd)) =>
                          equal_blue_witness A1_ xd NO_CYC))
                    (fn x => fn s =>
                      dbind s
                        (fn (a, (aa, _)) =>
                          (if not (memb_ahm_basic_ops (A1_, A2_) x a)
                            then dbind (code_blue_dfs_ahs_0 (A1_, A2_) g
 (a, (aa, (empty_ahm_basic_ops A2_ (), x))))
                                   (fn (ab, (ac, (_, bd))) =>
                                     DRETURN (ab, (ac, bd)))
                            else DRETURN (a, (aa, NO_CYC)))))
                    (DRETURN
                      (empty_ahm_basic_ops A2_ (),
                        (empty_ahm_basic_ops A2_ (), NO_CYC)))))
          (fn (_, (_, ba)) =>
            dbind (DRETURN (NDFS_SI_Statistics.stop ()))
              (fn _ => DRETURN (extract_res ba)))));

fun lasso_reach (Lasso_ext (lasso_reach, lasso_va, lasso_cysfx, more)) =
  lasso_reach;

fun lasso_cysfx (Lasso_ext (lasso_reach, lasso_va, lasso_cysfx, more)) =
  lasso_cysfx;

fun lasso_va (Lasso_ext (lasso_reach, lasso_va, lasso_cysfx, more)) = lasso_va;

fun map_lasso f l =
  Lasso_ext (map f (lasso_reach l), f (lasso_va l), map f (lasso_cysfx l), ());

fun ndfs_find_ce_code (A1_, A2_) g =
  let
    val x = degeneralize_ext_impl g (fn _ => ());
  in
    (case code_blue_dfs_ahs
            (equal_prod A1_ equal_nat, hashable_prod A2_ hashable_nat) x
      of NONE => NONE
      | SOME xb => SOME (SOME (map_lasso fst (lasso_of_prpl xb))))
  end;

fun find_ce_code (A1_, A2_) cfg =
  (case cfg of CFG_CE_SCC_GABOW => gabow_find_ce_code (A1_, A2_)
    | CFG_CE_NDFS => ndfs_find_ce_code (A1_, A2_));

fun impl_model_check_ltl_to_gba_code_uu_find_ce_code_map_lasso (A1_, A2_)
  (B1_, B2_) cfg sys phi =
  let
    val ba = ltl_to_gba_code (B1_, B2_) (cfg_l2b cfg) (LTLcNeg phi);
    val (g, map_q) = dflt_inter_impl equal_nata sys ba;
  in
    (case find_ce_code
            (equal_prod equal_nat A1_, hashable_prod hashable_nat A2_)
            (cfg_ce cfg) g
      of NONE => NONE | SOME NONE => SOME NONE
      | SOME (SOME ce) => SOME (SOME (map_lasso map_q ce)))
  end;

fun cava_sys_agn (A1_, A2_) (B1_, B2_) =
  impl_model_check_ltl_to_gba_code_uu_find_ce_code_map_lasso (A1_, A2_)
    (B1_, B2_);

fun cava_sm cfg prog phi =
  (if ty_program prog andalso ltlc_next_free phi
    then (case cava_sys_agn
                 (equal_pid_global_config_ext equal_cmd
                    (equal_local_state_impl_ext equal_unit)
                    (equal_global_state_impl_ext equal_unit) equal_unit,
                   hashable_pid_global_config_ext hashable_cmd
                     (hashable_local_state_impl_ext hashable_unit)
                     (hashable_global_state_impl_ext hashable_unit)
                     hashable_unit)
                 (equal_aprop, linorder_aprop) cfg
                 (sm_to_sa_impl prog
                   (fn x => member equal_literal x (vars_of_ltlc phi)))
                 phi
           of NONE => SAT | SOME NONE => UNSAT | SOME (SOME a) => UNSAT_CE a)
    else TY_ERR);

val test : sm_result = cava_sm dflt_cfg mulog_program mulog_property;

end; (*struct Test*)
